/* 
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: calibrate.c,v 1.29 2001/11/15 02:03:22 damian Exp $ 
 */

#include <fgl.h>
#include "ipl.h"

#include <string.h>
#include "calibrate.h"

#define SHOW_CURSOR 0

int state = 0;
fgl_context_t *gc;
fgl_color_t color1;
fgl_color_t color2;
fgl_sprite_t *sp,*cr;
int bw,bh,bs,btop,bbot;
int test_x[6];
int test_y[6];
int sample_x[6];
int sample_y[6];

int size_x = 0;
int size_y = 0;


void getPoints (int size_x, int size_y, int* test_x, int* test_y)
{
  /* calculate test positions */
  test_x[0] = size_x / 2;
  test_y[0] = size_y / 2;

  test_x[1] = 25;
  test_y[1] = size_y - 25;

  test_x[2] = size_x - 25;
  test_y[2] = size_y - 25;

  test_x[3] = size_x - 25;
  test_y[3] = 25;

  test_x[4] = 25;
  test_y[4] = 25;

  test_x[5] = size_x / 2;
  test_y[5] = size_y / 2;
}

int calibrate (int *tx, int *ty, int *sx, int *sy, ts_config_t *cfg, int rot, int w, int h) {

        int i,j;
        int temp;
	int off_x[6];
	int off_y[6];
        int tmp_x[6];
        int tmp_y[6];
        int smp_x[6];
        int smp_y[6];
	int temp_xscale, temp_yscale;

        switch (rot) {
          case 1:
            getPoints (h, w, tmp_x, tmp_y);
            smp_x[0] = sx[0];
            smp_y[0] = sy[0];
            for (i = 1; i < 5; i++ )
              {
                smp_x[i] = sx[(i%4)+1];
                smp_y[i] = sy[(i%4)+1];
              }
            tx = tmp_x;
            ty = tmp_y;
            sx = smp_x;
            sy = smp_y;
            break;
          case 2:
            smp_x[0] = sx[0];
            smp_y[0] = sy[0];
            for (i = 1; i < 5; i++ )
              {
                smp_x[i] = sx[((i+1)%4)+1];
                smp_y[i] = sy[((i+1)%4)+1];
              }
            sx = smp_x;
            sy = smp_y;
            break;
          case 3:
            getPoints (h, w, tmp_x, tmp_y);
            smp_x[0] = sx[0];
            smp_y[0] = sy[0];
            for (i = 1; i < 5; i++ )
              {
                smp_x[i] = sx[((i+2)%4)+1];
                smp_y[i] = sy[((i+2)%4)+1];
              }
            tx = tmp_x;
            ty = tmp_y;
            sx = smp_x;
            sy = smp_y;
            break;
          default:
        }

	/* determnie if swapped */
	if (( sy[1] - sy[2] ) * ( sy[1] - sy[2] ) > ( sx[1] - sx[2] ) * ( sx[1] - sx[2] )) {
		  cfg->swap = 1;
		  /* swap taken samples */
		  for(i=0;i<5;i++) { 
			  temp = sx[i]; 
			  sx[i] = sy[i]; 
			  sy[i] = temp;
		  }
		  
	} else {
		cfg->swap = 0;
	}

	/* determine the scaling factors */
	/* use fixpoint */

	/* check for divide by 0 error */
	if (( sx[0] - sx[1] ) == 0 || ( sx[0] - sx[4] ) == 0 || ( sx[2] - sx[0] ) == 0 ||
	    ( sx[3] - sx[0] ) == 0 || ( sx[2] - sx[1] ) == 0 || ( sx[3] - sx[4] ) == 0 ||
	    ( sy[1] - sy[0] ) == 0 || ( sy[2] - sy[0] ) == 0 || ( sy[0] - sy[3] ) == 0 ||
	    ( sy[0] - sy[4] ) == 0 || ( sy[3] - sy[2] ) == 0 || ( sy[4] - sy[1] ) == 0 ) {
		return 1;
	}

	temp_xscale  = (( tx[0] << 11 ) - ( tx[1] << 11 ))/( sx[0] - sx[1] );
	temp_xscale += (( tx[0] << 11 ) - ( tx[4] << 11 ))/( sx[0] - sx[4] );
	temp_xscale += (( tx[2] << 11 ) - ( tx[0] << 11 ))/( sx[2] - sx[0] );
	temp_xscale += (( tx[3] << 11 ) - ( tx[0] << 11 ))/( sx[3] - sx[0] );
	temp_xscale += (( tx[2] << 12 ) - ( tx[1] << 12 ))/( sx[2] - sx[1] );
	temp_xscale += (( tx[3] << 12 ) - ( tx[4] << 12 ))/( sx[3] - sx[4] );

	
	cfg->xscale  = temp_xscale >> 4;

	temp_yscale  = (( ty[1] << 11 ) - ( ty[0] << 11 ))/( sy[1] - sy[0] );
	temp_yscale += (( ty[2] << 11 ) - ( ty[0] << 11 ))/( sy[2] - sy[0] );
	temp_yscale += (( ty[0] << 11 ) - ( ty[3] << 11 ))/( sy[0] - sy[3] );
	temp_yscale += (( ty[0] << 11 ) - ( ty[4] << 11 ))/( sy[0] - sy[4] );
	temp_yscale += (( ty[3] << 12 ) - ( ty[2] << 12 ))/( sy[3] - sy[2] );
	temp_yscale += (( ty[4] << 12 ) - ( ty[1] << 12 ))/( sy[4] - sy[1] );

	cfg->yscale  = temp_yscale >> 4;

	if (cfg->xscale == 0 || cfg->yscale == 0) {
		return 1;
	}

	/* determine the offsets */
	
	for(i=0;i<5;i++) {
		off_x[i] = (( tx[i] << 16) / cfg->xscale);
		off_y[i] = (( ty[i] << 16) / cfg->yscale);
	}
	
	cfg->xoff = 0;
	cfg->yoff = 0;
	
	for(i=0;i<5;i++) {
		cfg->xoff += (off_x[i] - ( sx[i] << 6 ));
		cfg->yoff += (off_y[i] - ( sy[i] << 6 ));
	}
	cfg->xoff = cfg->xoff / 320;
	cfg->yoff = cfg->yoff / 320;

	return 0;
}

void print_usage()
{
  fprintf( stderr, "FGL/IPL calibration utility\n");
  fprintf( stderr, "usage: calibrate [options]\n");
  fprintf( stderr, "options are as follows:\n");
  fprintf( stderr, "  --width=<pixels>      set framebuffer width\n");
  fprintf( stderr, "  --height=<pixels>     set framebuffer height\n");
  fprintf( stderr, "  --rotation=<rot>      set screen rotation\n");
  fprintf( stderr, "  --color1=<hex color>  set lefthand-side  color\n");
  fprintf( stderr, "  --color2=<hex color>  set righthand-side color\n");
  fprintf( stderr, "  -h  --help            display this message\n");
  fprintf( stderr, "  -v  --version         show version and build information\n");
}

void draw_result_screen ()
{
	draw_screen(gc,size_x,size_y);
		  
	fgl_setColor( gc, fgl_clrBlue);
	fgl_drawRect( gc, size_x/2-30, size_y/2-30, size_x/2+30, size_y/2+30);
	fgl_setColor( gc, fgl_clrRed);
	fgl_drawRect( gc, sample_x[5]-30,sample_y[5]-30,
		      sample_x[5]+30,sample_y[5]+30); 
}

void draw_state ()
{
	int txh;

	if (state > 0 && SHOW_CURSOR) {
		fgl_hideSprite( gc, cr);
	}
	switch (state) {
	case 0:
		draw_init( gc, color1, color2, size_x, size_y );
		break;
	case 1:
		draw_screen1( gc, size_x, size_y);
		break;
	case 2:
		draw_screen2(gc,size_x,size_y);
		break;
	case 3:
		draw_screen6( gc, size_x, size_y);
		break;
	case 4:
		draw_result_screen();
		break;
	case 5:
	  //		draw_screen(gc, size_x, size_y);
		draw_result_screen();
		draw_message(gc, size_x, size_y,
			     "Calibration failed!",
			     "Tap anywhere to start again.");
		break;
	case 6:
		draw_result_screen();
		draw_message(gc,size_x,size_y,"Press button to finish,",
			     "redo or cancel calibration:");
		  
		  /* draw buttons */
		  
		  txh = font->fontHeight;
		  
		  bw = size_x / 4;
		  bh = txh * 2;
		  bs = bw / 4;
		  
		  btop = size_y - bh - bh/4;
		  bbot = size_y - bh/4;
		  
		  draw_button( gc, bs, btop, bw, bh, "Finish", 0); 
		  draw_button( gc, 2*bs+bw, btop, bw, bh, "Redo", 0); 
		  draw_button( gc, 3*bs+2*bw, btop, bw, bh, "Cancel", 0); 
		  
		  break;
	default:
		// Shouldn't happen
		exit(1);
	}
	if (state == 2 || state == 3) {
		sp->isVisible=0;
		fgl_showSprite( gc, sp);
	}
	if (state > 0 && SHOW_CURSOR) {
		fgl_showSprite( gc, cr);
	}
}

void check_vt_switch_event (ipl_event_t *e)
{
#ifdef VT_SWITCH
  if ((int)e->type == EVT_VT_REL_REQ) {
    // Linux VT switch
    ipl_vtSwitchRelease();
  } else if ((int)(e->type) == EVT_VT_ACQ_REQ) {
    ipl_vtSwitchAcquire();
    draw_state();
  }
#endif
}

int main ( int argc, char* argv[] ) /*{{{*/
{
	ipl_event_t e;

	int dx = 0;
	int dy = 0;

	int i;
	int debug = 0;

	int message = 0;
	ts_config_t   new_cfg;

	char config_old[256];
	char config_new[256];
	char *cfgFile;
	FILE *fp;
	struct stat config_stats;

        int degree, scr_rot;
        int ipl_x, ipl_y;

        if (!fgl_init (0, 0, 0)) {
            printf ("fgl initializtion failed\n");
            return 1;
        }
                
	color1 = fgl_clrRed;
	color2 = fgl_clrWhite;

        /* parse command line args */
        for (i=1; i<argc; i++) 
          {
            if (!strncmp(argv[i],"-h",2) ||
                !strncmp(argv[i],"--help",6)) 
              {
                fgl_terminate();
                print_usage(); 
                exit(0);
              } 
            else if (!strncmp(argv[i], "--width=", 8)) 
              {
                size_x = atoi(argv[i]+8);
              } 
            else if (!strncmp(argv[i], "--height=", 9))
              {
                size_y = atoi(argv[i]+9);
              } 
            else if (!strncmp(argv[i], "--rotation=", 11))
              {
                degree = atoi(argv[i]+11);
                switch (degree) 
                  {
                    case 90: 
                      scr_rot = 1;
                      break;
                    case 180: 
                      scr_rot = 2;
                      break;
                    case 270:
                      scr_rot = 3;
                      break;
                    default:
                      scr_rot =0;
                  }
                fgl_setScreenRotation (scr_rot);
              } 
            else if (!strncmp(argv[i], "--color1=", 9))
              {
                color1 = fgl_getPixelValue (strtol(argv[i]+9,NULL,16));
              } 
            else if (!strncmp(argv[i], "--color2=", 9))
              {
                color2 = fgl_getPixelValue (strtol(argv[i]+9,NULL,16));
              } 
            else if (!strncmp(argv[i], "--debug", 7))
              {
                debug = 1;
              } 
            else if (!strncmp(argv[i], "-v",2) ||
                !strncmp(argv[i], "--version", 9))
              {
                fgl_terminate();
                printf("%s v%s\n", PACKAGE, VERSION);
                printf("fgl calibration program\n");
                printf("Compiled by  : %s\n", BUILD_USER);
                printf("Compiled on  : %s (%s)\n", BUILD_HOST, BUILD_ARCH);
                printf("Compile date : %s\n", BUILD_DATE);
                exit (0);
              }
          }

	gc = fgl_getContext( fgl_screenBmp);

        if (!size_x) size_x = fgl_getScreenWidth();
	if (!size_y) size_y = fgl_getScreenHeight();

	fgl_setClip( gc, 0, 0, size_x, size_y);

        getPoints ( size_x, size_y, test_x, test_y);
        
	state = 0;
	draw_state();

        if (scr_rot & 0x1)
          {
            ipl_x = size_y;
            ipl_y = size_x;
          }
        else
          {
            ipl_x = size_x;
            ipl_y = size_y;
          }
       
	if ( !ipl_init( ipl_x, ipl_y) ){

	  draw_error( gc, size_x, size_y, "Error initializing touchscreen/mouse!");
	  fgl_terminate();

	  return 1;
	}
        
	/* do not convert any values */

	ipl_wakeUp();

	sp = fgl_createSprite( &cross );
	cr = fgl_createSprite( &spd );


	state = 1;
	draw_state();

	fgl_moveSprite(gc, cr, size_x/2, size_y/2);
	if (SHOW_CURSOR) fgl_showSprite( gc, cr);
	while (1) {
	  e.type = 0;
	  if (ipl_getEvent( &e )) {
	    if ((int)e.type == EVT_BUTTON_UP) {
	      break;
	    }
	    if ((int)e.type == EVT_MOUSE_MOVE) fgl_moveSprite(gc, cr, e.x, e.y);
	    check_vt_switch_event( &e );
	  }
	}

	while (!message) {
	  /* do not convert any values */
	  ipl_cfg.xoff = TS_UNDEF;
	  ipl_cfg.xscale = TS_UNDEF;
	  ipl_cfg.yoff = TS_UNDEF;
	  ipl_cfg.yscale = TS_UNDEF;
	  ipl_cfg.swap = 0;
	  ipl_cfg.rate = TS_UNDEF;
	  ipl_cfg.pressure = TS_UNDEF;

          ipl_setScreenRotation (0);

	  state = 2;
	  draw_state();

	  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
	  fgl_moveSprite( gc, sp, test_x[0]-10, test_y[0]-10);
	  if (SHOW_CURSOR) fgl_showSprite( gc, cr);

	  for(i=0;i<5;i++) {
	    
	    while (1) {
	      e.type = 0;
	      if (ipl_getEvent( &e )) {
		if ((int)e.type == EVT_BUTTON_UP) {
		  sample_x[i] = e.x;
		  sample_y[i] = e.y;
		  if (debug) fprintf(stderr,"sample (%3i,%3i)\n",e.x,e.y);
		  if (i == 0) {
			  break;
		  } else {
			  if ( (sample_x[i] - sample_x[i-1])*(sample_x[i] - sample_x[i-1]) > 400 ||
			       (sample_y[i] - sample_y[i-1])*(sample_y[i] - sample_y[i-1]) > 400 ) {
				  break;
			  }
		  }
		}
		if ((int)e.type == EVT_MOUSE_MOVE)
		  fgl_moveSprite(gc, cr, e.x, e.y);
		check_vt_switch_event( &e );
	      }
	    }
	    if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
	    if (i != 4) { 
		    move(gc,sp,cr,test_x[i]-10,test_y[i]-10,test_x[i+1]-10,test_y[i+1]-10,200,30000);
		    if (SHOW_CURSOR) fgl_showSprite( gc, cr);
	    }
	  }

	  /* calculate the calibration values */
	  
	  if (!calibrate( test_x, test_y, sample_x, sample_y, &new_cfg, scr_rot, size_x, size_y)) {
		  /* test the calibration */
		  
		  ipl_cfg.xoff   = new_cfg.xoff;
		  ipl_cfg.xscale = new_cfg.xscale;
		  ipl_cfg.yoff   = new_cfg.yoff;
		  ipl_cfg.yscale = new_cfg.yscale;
		  ipl_cfg.swap   = new_cfg.swap;

                  ipl_setScreenRotation (scr_rot);

		  if (debug) {
		    fprintf(stderr,"xoff     (%i)\n",ipl_cfg.xoff);
		    fprintf(stderr,"xscale   (%i)\n",ipl_cfg.xscale);
		    fprintf(stderr,"yoff     (%i)\n",ipl_cfg.yoff);
		    fprintf(stderr,"yscale   (%i)\n",ipl_cfg.yscale);
		    fprintf(stderr,"swap     (%i)\n",ipl_cfg.swap);
		  }

		  state = 3;
		  draw_state();
		  
		  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
		  move(gc,sp,cr,test_x[4]-10,test_y[4]-10,test_x[5]-10,test_y[5]-10,200,30000);
		  if (SHOW_CURSOR) fgl_showSprite( gc, cr);

		  while (1) {
			  e.type = 0;
			  if (ipl_getEvent( &e )) {
				  if ((int)e.type == EVT_BUTTON_UP) {
					  sample_x[5] = e.x;
					  sample_y[5] = e.y;
					  if (debug) fprintf(stderr,"test (%3i,%3i)\n",e.x,e.y);
					  break;
				  }
				  if ((int)e.type == EVT_MOUSE_MOVE) fgl_moveSprite(gc, cr, e.x, e.y);
				  check_vt_switch_event( &e );
			  }
		  }   
		  
		  fgl_hideSprite(gc, sp);

		  dx = sample_x[5] - test_x[5];
		  dy = sample_y[5] - test_y[5];

		  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);

		  state = 4;
		  draw_state();
	  } else {
		  dx = 100;
	  }

	  if ( dx*dx > 70 || dy*dy > 70 ) {
		  state = 5;
		  draw_state();

		  while (1) {
			  e.type = 0;
			  if (ipl_getEvent( &e )) {
				  if ((int)e.type == EVT_BUTTON_UP) {
					  break;
				  }
				  if ((int)e.type == EVT_MOUSE_MOVE) fgl_moveSprite(gc, cr, e.x, e.y);
				  check_vt_switch_event( &e );
			  }
		  }
	  } else {

		  state = 6;
		  draw_state();
		  		  
		  while (1) {
			  e.type = 0;
			  if (ipl_getEvent( &e )) {
				  if ((int)e.type == EVT_BUTTON_DOWN) {
					  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
					  if (e.y > btop && e.y < bbot) {
						  if (e.x +dx > bs && e.x + dx < bs + bw) {
							  draw_button( gc, bs, btop, bw, bh, "Finish", 1); 
						  }
						  if (e.x +dx > 2*bs +bw && e.x + dx < 2*bs + 2*bw) {
							  draw_button( gc, 2*bs+bw, btop, bw, bh, "Redo", 1); 
						  }
						  if (e.x +dx > 3*bs +2*bw && e.x + dx < 3*bs + 3* bw) {
							  draw_button( gc, 3*bs+2*bw, btop, bw, bh, "Cancel", 1); 
		    }
					  }
					  if (SHOW_CURSOR) fgl_showSprite( gc, cr);
				  }
				  if ((int)e.type == EVT_BUTTON_UP) {
					  if (e.y > btop && e.y < bbot) {
						  if (e.x > bs && e.x < bs + bw) {
							  message = 1;
							  break;
						  }
						  if (e.x > 2*bs +bw && e.x < 2*bs + 2*bw) {
							  break;
						  }
						  if (e.x > 3*bs +2*bw && e.x < 3*bs + 3* bw) {
							  message = 2;
							  break;
						  }
					  }
				  }
				  if ((int)e.type == EVT_MOUSE_MOVE) fgl_moveSprite(gc, cr, e.x, e.y);
				  check_vt_switch_event( &e );
			  }
		  }   
	  }
	} 
	
	if (message == 1) {
	  /* saving results to a file */
	  
          config_old[0] = '\0';
          config_new[0] = '\0';

          if ( (cfgFile = getenv( "TP_CONFIG")) ) {
            strncat( config_new, cfgFile, 256 - 1);
          }
          else {
            strncat( config_new, DEFAULT_TP_CONFIG, 256 - 1);
          }

	  /* create the backup */
	  if (!stat(config_new,&config_stats)) rename (config_new,config_old);	  

	  if ( (fp = fopen( config_new, "w")) ) {
		  fprintf(fp,"xoffset  %i\n",ipl_cfg.xoff);
		  fprintf(fp,"xscale   %i\n",ipl_cfg.xscale);
		  fprintf(fp,"yoffset  %i\n",ipl_cfg.yoff);
		  fprintf(fp,"yscale   %i\n",ipl_cfg.yscale);
		  fprintf(fp,"rate     %i\n",ipl_cfg.rate);
		  fprintf(fp,"swap     %i\n",ipl_cfg.swap);
		  fprintf(fp,"pressure %i\n",ipl_cfg.pressure);
		  fclose(fp);
		  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
		  draw_screen5( gc, size_x, size_y,config_new);
		  if (SHOW_CURSOR) fgl_showSprite( gc, cr);
	  } else {
		  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
		  draw_error( gc, size_x, size_y,"could not write file!");
		  if (SHOW_CURSOR) fgl_showSprite( gc, cr);
	  }
	} 
	if (message == 2) {
	  if (SHOW_CURSOR) fgl_hideSprite( gc, cr);
	  draw_screen4( gc, size_x, size_y);
	  if (SHOW_CURSOR) fgl_showSprite( gc, cr);
	}
	ipl_reset();
	fgl_terminate();
   fgl_init(0,0,0);
   fgl_terminate();

	return 0;
}
/*}}}*/
