/* 
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: screens.c,v 1.10 2001/11/15 02:03:22 damian Exp $ 
 */

#include <string.h>

#include <fgl.h>
#include "config.h"
#include "xoe.h"

extern fgl_font_t       *font;
extern fgl_bitmap_t     *background;

void draw_init( fgl_context_t *gc, fgl_color_t color1, fgl_color_t color2, int size_x, int size_y ) {
  
  fgl_context_t    *tc;
  fgl_bitmap_t     *xoe_bmp;
  fgl_monoBitmap_t *xoe_mask;
  fgl_a8Bitmap_t   *alpha;
  fgl_font_t       *fnt;

  char version[256];
  int length, height;

  int step, value;
  int i, k;

  font = fgl_getFont("arial",10);
  fnt = fgl_getFont("arial",8);

  fgl_setFont( gc, font);
  get_xoe(&xoe_bmp, &xoe_mask);
  
  background = fgl_createBitmap( size_x, size_y, -1);

  tc = fgl_getContext( background );
  fgl_setFont( tc, fnt);

  fgl_setColor( tc, fgl_clrWhite );
  fgl_fillRect( tc, 0, 0, size_x-1, size_y -1 );

  fgl_maskBltBmp( tc, 30, 30, xoe_bmp, xoe_mask );

  sprintf( version, "%s v%s", PACKAGE, VERSION );
  length = fgl_getTextLength( fnt, version, -1 );
  height = fnt->fontHeight;

  fgl_setColor( tc, fgl_clrGray );
  fgl_drawText( tc, size_x - length -1, 1 + height, version , -1 );
  
  fgl_disposeBitmap( xoe_bmp );
  fgl_disposeMonoBitmap ( xoe_mask );
 
  free( tc );
}

void free_background() {
   fgl_disposeBitmap( background );
}

void draw_screen( fgl_context_t *gc, int size_x, int size_y) {

  fgl_bitBlt( gc, 0, 0, background, 0, 0, background->width-1, background->height-1);

}

void draw_message( fgl_context_t *gc, int size_x, int size_y, char *screen11, char *screen12) {

  int th;
  int tl1;
  int tl2;

  int offset_x;
  int offset_y;
  th = font->fontHeight;
  tl1 = fgl_getTextLength(font,screen11,-1);
  tl2 = fgl_getTextLength(font,screen12,-1);

  offset_x = size_x/2;
  offset_y = 160 + (size_y - 160)/2;

  fgl_setFont( gc, font);
  fgl_setColor( gc, fgl_clrBlack);
  fgl_drawText( gc, offset_x - tl1/2, offset_y-th,screen11,-1);
  fgl_drawText( gc, offset_x - tl2/2, offset_y,screen12,-1);

}

void draw_screen1( fgl_context_t *gc, int size_x, int size_y) {

	char screen11[] = "Tap on the screen";
	char screen12[] = "to start calibration";

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11,screen12);

}

void draw_screen2( fgl_context_t *gc, int size_x, int size_y) {

	char screen11[] = "Tap precisely on";
	char screen12[] = "the target mark.";

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11,screen12);

}

void draw_screen3( fgl_context_t *gc, int size_x, int size_y) {

	char screen11[] = "Press button to finish,";
	char screen12[] = "redo or cancel?";

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11,screen12);
	
}

void draw_screen4( fgl_context_t *gc, int size_x, int size_y) {

	char screen11[] = "Calibration canceled:";
	char screen12[] = "file not saved.";

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11,screen12);

}

void draw_screen5( fgl_context_t *gc, int size_x, int size_y, char *screen12) {

	int  length;
	char filename[24];
	char screen11[] = "Saving results to:";

	length = strlen( screen12 );

	if (length >= 20 ) {
		filename[0] = '.';
		filename[1] = '.';
		filename[2] = '.';
		strcpy(filename + 3, screen12 + ( length - 20 ));
	} else {
		strcpy( filename, screen12 );
	}

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11, filename);

}

void draw_screen6( fgl_context_t *gc, int size_x, int size_y) {

	char screen11[] = "Done,  tap on target to";
	char screen12[] = "test current calibration.";

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11,screen12);

}

void draw_error( fgl_context_t *gc, int size_x, int size_y, char *screen12) {

	char screen11[] = "Error encountered:";

	draw_screen( gc, size_x, size_y );
	draw_message( gc, size_x, size_y, screen11,screen12);

}
