/*
 * a2bmp.c - fgl module to handle monochrome bitmaps (depth 1)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 * 
 * $tvt: a2bmp.c,v 1.11 2001/08/16 01:36:35 samc Exp $
 */

/* fgl module epilog */ /*{{{*/

#ifndef FGL_A2BMP_INCLUDED
#define FGL_A2BMP_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include "util.c"
#include FGL_DRAW_MODULE
#include FGL_HEAP_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_A2BMP
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif

/*}}}*/


#if !defined(FGL_PROTO)
u8 _a2Bits[8] = { 0xc0, 0x30, 0x0c, 0x03};
#else
extern u8 _a2Bits[8];
#endif


static inline
void fgl_setA2Bit ( fgl_a2Bitmap_t* bmp, u32 col, u32 row ) /*{{{*/
{
	u8 *p;
#ifdef ROT_90
	rotate32( (fgl_bitmap_t *)bmp, &col, &row);
#endif
	p = bmp->surface + (row * bmp->sLength) + (col / 4);
	*p |= _a2Bits[col % 4];
}
/*}}}*/

static inline
void fgl_unsetA2Bit ( fgl_a2Bitmap_t* bmp, u32 col, u32 row ) /*{{{*/
{
	u8 *p;
#ifdef ROT_90
	rotate32( (fgl_bitmap_t *)bmp, &col, &row);
#endif
	p = bmp->surface + (row * bmp->sLength) + (col / 4);
	*p &= ~_a2Bits[col % 4];
}
/*}}}*/

static inline
int fgl_isA2BitSet ( fgl_a2Bitmap_t* bmp, u32 col, u32 row ) /*{{{*/
{
	u8 *p;
#ifdef ROT_90
	rotate32( (fgl_bitmap_t *)bmp, &col, &row);
#endif
	p = bmp->surface + (row * bmp->sLength) + (col / 4);
	return (*p & _a2Bits[col % 4]);
}
/*}}}*/

FUNC_ATTR
fgl_a2Bitmap_t* fgl_createA2Bitmap ( u32 width, u32 height, int init ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_A2BMP)
{
#ifdef ROT_90
	u32 sLength = (height + 3) / 4;
	u32 size =  (sLength * width);
#else
	u32 sLength = (width + 3) / 4;
	u32 size =  (sLength * height);
#endif
	fgl_a2Bitmap_t *bmp = (fgl_a2Bitmap_t*) FGL_MALLOC( sizeof(fgl_a2Bitmap_t));

	bmp->sLength = sLength;
	bmp->width = width;
	bmp->height = height;

	bmp->surface = (u8*) FGL_MALLOC(  sizeof(u8) * size);

	if ( init != -1 ){
		memset( bmp->surface, init, size);
	}

	return bmp;
}
#endif /* !FGL_PROTO */
; /*}}}*/


static inline
void fgl_disposeA2Bitmap ( fgl_a2Bitmap_t* bmp ) /*{{{*/
{
	FGL_FREE( bmp);
}
/*}}}*/


FUNC_ATTR
void fgl_a2Blt ( fgl_context_t* dst, i32 u0, i32 v0,
				   fgl_a2Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
				   fgl_color_t clr ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_A2BMP)
{
	u8 *ps;
	fgl_slc_t *pd;
	fgl_color_t pix;
	i32 u, u1, v1, sinc, dinc;
	i32 w, h;
	register u8 a, b, i;
	u8 i0, off;
	u8 blend[4] = {0,85,170,255};

	if ( x0 > x1 )  swap32( &x0, &x1);
	if ( y0 > y1 )  swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( ( fgl_bitmap_t *) src, (u32*)&x0, (u32*)&y0);
	rotate32( ( fgl_bitmap_t *) src, (u32*)&x1, (u32*)&y1);
	rotate32( dst->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( dst->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif

	w = x1 - x0;
	h = y1 - y0;
	off = x0 % 4;

	ps = src->surface + (y0 * src->sLength) + (x0 / 4);
	pd = fgl_getSlc( dst->bmp, u0, v0);

	a = *ps;
	u = u0;
	u1 = u0 + w;
	i = i0 = 4-off;

	dinc = (dst->bmp->sLength - u1) + u0;                 /* in pixels */
	sinc = src->sLength + (x0>>2) - (x1>>2);            /* in bytes */

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( h-- >= 0 ){
	  a <<= off;

	  while ( 1 ){
	    
	    b = a >> 6;
	    if (b){
	      pix = fgl_blendPixelValues(blend[b] , fgl_getSlcPixel( pd, u)
					 , clr);
	      fgl_setSlcPixel( pd, u, pix);
	    }

#ifdef DEBUG_A2BMP
	    putchar( (b) ? 'O' : '.');
	    if ( u == u1 ) putchar( '\n');
#endif	
		  
	    if ( u == u1 ){
	      break;
	    }
	    
	    if ( --i == 0 ){
	      /* next source byte */
	      i = 4;
	      a = *(++ps);
	    }
	    else {
	      a <<= 2;
	    }
	    
	    fgl_incSlc( &pd, u);
	    u++;
	  }

	  /* next source and destination scanlines */
	  fgl_addSlc( &pd, u, dinc);
	  ps += sinc;
	  a = *ps;
	  u = u0;
	  i = i0;
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}
#endif /* !FGL_PROTO */
;/*}}}*/


static inline
void fgl_a2BltBmp ( fgl_context_t* dst, i32 u0, i32 v0,
					  fgl_a2Bitmap_t* src, fgl_color_t clr ) /*{{{*/
{
	fgl_a2Blt( dst, u0, v0, src, 0, 0, src->width-1, src->height-1, clr);
}
/*}}}*/


/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_GBMP_INCLUDED */
/*}}}*/

