/*
 * a8bmp.c - fgl module to handle alpha bitmaps (alpha value [0..255]) 
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: a8bmp.c,v 1.10 2001/08/16 01:36:35 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_A8BMP_INCLUDED
#define FGL_A8BMP_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* these are our fgl imports */
#include FGL_CMAP_MODULE
#include "util.c"
#include FGL_DRAW_MODULE
#include FGL_HEAP_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif

#ifdef FGL_INLINE_A8BMP
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/

FUNC_ATTR
fgl_a8Bitmap_t* fgl_createA8Bitmap ( u32 width, u32 height, int init ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_A8BMP)
{
	fgl_a8Bitmap_t *bmp = (fgl_a8Bitmap_t*) FGL_MALLOC( sizeof(fgl_a8Bitmap_t)
														           + (width * height));

#ifdef ROT_90
	bmp->sLength = height;
#else	
	bmp->sLength = width;
#endif
	bmp->width   = width;
	bmp->height  = height;

	bmp->surface = (u8*)bmp + sizeof(fgl_a8Bitmap_t);

	if ( init != -1 ){
		memset( bmp->surface, init, (width * height));
	}

	return bmp;
}
#endif /* !FGL_PROTO */
;/*}}}*/

static inline
void fgl_disposeA8Bitmap ( fgl_a8Bitmap_t* bmp ) /*{{{*/
{
	FGL_FREE( bmp);
}
/*}}}*/

static inline
void fgl_setPixelA8 ( fgl_a8Bitmap_t* bmp,  u32 x, u32 y, u8 alpha ) /*{{{*/
{
	if ( (x < bmp->width) && (y < bmp->height) ){
#ifdef ROT_90
		rotate32( (fgl_bitmap_t *)bmp, &x, &y);
#endif
	        bmp->surface[ (y * bmp->sLength) + x ] = alpha;
	}
}
/*}}}*/

static inline
fgl_color_t fgl_getPixelA8 ( fgl_a8Bitmap_t* bmp,  u32 x, u32 y ) /*{{{*/
{
	if ( (x < bmp->width) && (y < bmp->height) ){
#ifdef ROT_90
       	        return bmp->surface[ ((bmp->width - x) * bmp->height) + y ];
#else
		return bmp->surface[ (y * bmp->sLength) + x ];
#endif	
	}
	else {
		return 0;
	}
}
/*}}}*/

FUNC_ATTR
void fgl_a8Blt ( fgl_context_t* dst, i32 u0, i32 v0,
					fgl_a8Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
					fgl_color_t clr ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_A8BMP)
{
	i32 u1, v1, u, da, dd;
	fgl_slc_t *pde, *pd;
	u8 *pa;
	fgl_color_t pix;

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90
	rotate32( (fgl_bitmap_t *) src, (u32*)&x0, (u32*)&y0);
	rotate32( (fgl_bitmap_t *) src, (u32*)&x1, (u32*)&y1);
	rotate32( dst->bmp, (u32*)&u0, (u32*)&v0);
	rotate32( dst->bmp, (u32*)&u1, (u32*)&v1);
	swap32( &x0, &x1);
	swap32( &u0, &u1);
#endif

	pd  = fgl_getSlc( dst->bmp, u0, v0);
	pde = fgl_getSlc( dst->bmp, u0, v1);
	dd  = dst->bmp->sLength - u1 + u0 - 1;

	pa  = src->surface;
	da  = src->sLength - x1 - 1;

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( pd <= pde ){
		for ( u=u0; u<=u1; u++ ){
		        pix = fgl_blendPixelValues( *pa, fgl_getSlcPixel( pd, u), clr);
			fgl_setSlcPixel( pd, u, pix);

			fgl_incSlc( &pd, u);
			pa++;
		}

		pa += da;
		fgl_addSlc( &pd, u1, dd);
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}
#endif /* !FGL_PROTO */
;/*}}}*/

static inline
void fgl_a8BltBmp ( fgl_context_t* dst, i32 u0, i32 v0,
					   fgl_a8Bitmap_t* src, fgl_color_t clr ) /*{{{*/
{
	fgl_a8Blt( dst, u0, v0, src, 0, 0, src->width-1, src->height-1, clr);
}
/*}}}*/


/* fgl module epilog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_ABMP_INCLUDED */
/*}}}*/


