/* -*- mode: C; mode: fold -*- */ 

#include "fgl.h"
#include <freetype/freetype.h>


#define TTLOAD_FLAGS (TTLOAD_SCALE_GLYPH | TTLOAD_HINT_GLYPH)

inline
fgl_font_t* fontSpace ( u16 nGlyphs ) /*{{{*/
{
	fgl_font_t *fnt = 0;
	u32        arr32Size = sizeof(u32) * nGlyphs;
	u32        arr8Size = sizeof(u8) * nGlyphs;

	fnt = FGL_CALLOC( 1, sizeof(fgl_font_t) + 5*arr8Size + arr32Size);

	fnt->surfOff  = (u32*)((u8*)fnt + sizeof(fgl_font_t));
	fnt->width    = (i8*) ((u8*)(fnt->surfOff)  + arr32Size); 
	fnt->height   = (i8*) ((u8*)(fnt->width)    + arr8Size); 
	fnt->xBearing = (i8*) ((u8*)(fnt->height)   + arr8Size);
	fnt->yBearing = (i8*) ((u8*)(fnt->xBearing) + arr8Size);
	fnt->advance  = (i8*) ((u8*)(fnt->yBearing) + arr8Size);

	fnt->nGlyphs = nGlyphs;

	return fnt;
}
/*}}}*/


int fgl_ripFont ( char* fname, char* sname, u8 ptSize, u16 start, u16 count, int antialias) /*{{{*/
{
	TT_Error             error;
	TT_Face              face;
	TT_Instance          instance;
	TT_Glyph             glyph;
	TT_CharMap           charMap;
	TT_Glyph_Metrics     metrics;
	TT_Face_Properties   properties;
	TT_Raster_Map        rasterMap;
	TT_Engine engine;

	char                 buf[128];
	int                  i, j, e, w, h, m;
    u16                  platform, encoding, nGlyphs;
	u8                   *pSurf;
	fgl_font_t           *fnt = 0;
	int                  ptres, em72;
	u16                  *map;
	u16                  offset;
	u32                  temp;
	u32                  tsf,tse;
	int                  out;
	char *fntDir;
	u8                   isFileMsb = (__BYTE_ORDER == __BIG_ENDIAN);
	TT_Byte              grayScale[5] = {0,0,85,170,255};
	u8  byte;
	u16 k;
	u8   *tempS;
	fntDir = "../fonts";

	sprintf( buf, "%s/%s.ttf", fntDir, fname);
	printf( "loading font %s\n", buf);
	if ( (error = TT_Init_FreeType( &engine)) ){
		err( "TT_Init_FreeType failed: %d\n", error);
		return 0;
	}

	if ( (error = TT_Open_Face( engine, buf, &face)) ){
		err( "TT_Open_Face \"%s\" failed: %d\n", buf, error);
		return 0;
	}
	if ( (error = TT_New_Instance( face, &instance)) ){
		err( "TT_New_Instance failed: %d\n", error);
		return 0;
	}
    if ( (error = TT_Set_Instance_Resolutions( instance, 96, 96)) ){
		err( "TT_New_Instance failed: %d\n", error);
		return 0;
	}
    if ( (error = TT_Set_Instance_CharSize( instance, ptSize * 64)) ){
		err( "TT_Set_Instance_CharSize %d failed: %d\n", ptSize, error);
		return 0;		
	}
	if ( (error = TT_New_Glyph( face, &glyph)) ){
		err( "TT_New_Glyph failed: %d\n", error);
		return 0;
	}
	if ( (error = TT_Get_Face_Properties( face, &properties)) ){
		err( "TT_New_Glyph failed: %d\n", error);
		return 0;
	}

	/* try to get a unicode char map */
	for ( i=0; i < properties.num_CharMaps; i++ ){
		TT_Get_CharMap_ID( face, i, &platform, &encoding );
		if ( (platform == 3 && encoding == 1 )  ||
			 (platform == 0 && encoding == 0 ) ) {
			TT_Get_CharMap( face, i, &charMap );
			break;
		}
    }
    if ( i == properties.num_CharMaps ){
      err( "no unicode mapping found");
      return 0;		
    }
    

    if (antialias) TT_Set_Raster_Gray_Palette(engine,grayScale);

    nGlyphs = properties.num_Glyphs;
    offset  = 0;
	/* BAD! reverse mapping eats up memory, and it probably is used
	 * just for gaps (all referring to the same void glyph). But how
	 * can we detect if we already processed a certain glyph index,
	 * without walking the whole unicode array?
	 */
	map = FGL_MALLOC( nGlyphs * sizeof(u16));
	memset( map, 0, nGlyphs * sizeof(u16));	

	fnt = fontSpace(nGlyphs);
	m = 0;

	/* first pass, fill in font data */
        for ( i = 0; i < nGlyphs; i++ ){

		k = TT_Char_Index( charMap, i+offset );

		//		printf( "%5i : %0.4x : %0.4x : %0.4x\n", i, i+offset, k,offset);

		if ( map[k-offset] && (map[k-offset] < nGlyphs) ){
			/* if we already have a map entry for k, simply copy the data (if it
			 * doesn't refer to a out-of-range glyph
			 */
			j = map[k-offset];

			fnt->width[i]    = fnt->width[j];
			fnt->height[i]   = fnt->height[j];			
			fnt->xBearing[i] = fnt->xBearing[j];
			fnt->yBearing[i] = fnt->yBearing[j];
			fnt->advance[i]  = fnt->advance[j];			
		}
		else {
			/* Ok, we don't have this one in our glyph space yet, query it */
			
			if ( !(error = TT_Load_Glyph( instance, glyph, k, TTLOAD_FLAGS )) ){
				error = TT_Get_Glyph_Metrics( glyph, &metrics );
				
				w = (metrics.bbox.xMax - metrics.bbox.xMin) / 64;
				h = (metrics.bbox.yMax - metrics.bbox.yMin) / 64;
				//printf("%i %i %i\n",w,h,error);
				if (antialias) {
				  m += A2_SLENGTH(w) * h;
				} else {
				  m += MONO_SLENGTH(w) * h;
				}
				fnt->width[i]    = w;
				fnt->height[i]   = h;
				fnt->xBearing[i] = metrics.bearingX / 64;
				fnt->yBearing[i] = metrics.bearingY / 64;
				fnt->advance[i]  = metrics.advance  / 64;

				map[k-offset] = i+offset;
			} else {
			  printf("error\n");
			}
		}
	}

	if ( properties.horizontal ){
		em72 = properties.header->Units_Per_EM * 72;
		ptres = ptSize * 96;

		/* must be ceiled, or we might loose pixels */
		fnt->ascend = (properties.horizontal->Ascender * ptres + em72/2) / em72;
		fnt->descend = (-properties.horizontal->Descender * ptres + em72/2) / em72;

		fnt->fontHeight = fnt->ascend + fnt->descend +
			(properties.horizontal->Line_Gap * ptres / em72);
	}

	pSurf = fnt->surface = FGL_CALLOC(1, m);
	rasterMap.flow = TT_Flow_Down;

	/* second pass, render glyphs */
	for ( i = 0; i < nGlyphs; i++ ){

		k = TT_Char_Index( charMap, (short)(i+offset) );

		if ( map[k-offset] == i+offset ) {
			/* this is a new one, render the glyph */
			rasterMap.width = fnt->width[i];

			if ( rasterMap.width ) {
				fnt->surfOff[i]  = (unsigned long)pSurf - (unsigned long)fnt->surface;


				/* take different approach if antialias or not */

				if (antialias ) {

				 
				  rasterMap.cols   = ((rasterMap.width+3)/4)*4;
				  rasterMap.rows   = fnt->height[i];
				  rasterMap.size   = rasterMap.cols * rasterMap.rows;
				  tempS = (u8 *)malloc(rasterMap.size);
				  rasterMap.bitmap = tempS;
				  memset(tempS,0,rasterMap.size);
				  if ( !(error = TT_Load_Glyph( instance, glyph, k, TTLOAD_FLAGS )) ){
				    TT_Get_Glyph_Metrics( glyph, &metrics );
				    TT_Get_Glyph_Pixmap( glyph, &rasterMap,
							 -metrics.bbox.xMin, -metrics.bbox.yMin);
				  }
				  for ( j = 0 ; j < rasterMap.size; j+=4 ) {
				    for ( e = 0 ; e < 4 ; e++ ) {
				      if (tempS[j+e] == 85)  tempS[j+e] = 0x01; 
				      if (tempS[j+e] == 170) tempS[j+e] = 0x02; 
				      if (tempS[j+e] == 255) tempS[j+e] = 0x03; 
				    }
				    byte = (tempS[j] << 6) + (tempS[j+1] << 4) + 
				      (tempS[j+2] << 2) + tempS[j+3];

				    *pSurf++ = byte;
				  }
				  free(rasterMap.bitmap);
				} else {
				  rasterMap.bitmap = pSurf;
				  rasterMap.cols   = MONO_SLENGTH(rasterMap.width);
				  rasterMap.rows   = fnt->height[i];
				  rasterMap.size   = rasterMap.cols * rasterMap.rows;
				
				  if ( !(error = TT_Load_Glyph( instance, glyph, k, TTLOAD_FLAGS )) ){
				    TT_Get_Glyph_Metrics( glyph, &metrics );
				    TT_Get_Glyph_Bitmap( glyph, &rasterMap,
							 -metrics.bbox.xMin, -metrics.bbox.yMin);
				  }
				pSurf += (rasterMap.cols * rasterMap.rows);
				}
			}
		}
		else {
			/* We already had this one, copy the offset. Note that map[k]
			 * can't be outside our glyph space because of the way we set it in pass 1
			 */
			fnt->surfOff[i] = fnt->surfOff[ map[k-offset] ];
		}
	}

	TT_Flush_Face( face);

	FGL_FREE( map);

	fnt->fname  = strdup( fname);
	fnt->ptSize = ptSize;
	fnt->uniOffset = offset;

	sprintf( buf, "%s/%s.fnt", fntDir, sname);

	printf( "writing font %s\n", buf);

	out = open(buf,O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | 
		   S_IRGRP | S_IWGRP | S_IROTH);

	if (out > 0) printf("Succes opening file.\n");

	/* write the byte order and antialiasing */

	if (antialias) isFileMsb |= 0x2;
	write(out,&isFileMsb,1);
	//printf("isFileMsb: %x\n",isFileMsb);
	write(out,&count,sizeof(count));
	write(out,&start,sizeof(start));

	write(out,&(fnt->fontHeight),sizeof(fnt->fontHeight));
	write(out,&(fnt->ascend),sizeof(fnt->ascend));
	write(out,&(fnt->descend),sizeof(fnt->descend));

	tsf = 0;
	tse = 0;
	for( i = 0 ; i < count ; i++) {
	  /* encoding the positions of the file */
	  if(!tsf) tsf = fnt->surfOff[start+i];
	  temp = fnt->surfOff[start+i] - tsf;
	  if (fnt->surfOff[start+i] == 0) temp = 0;
	  if (antialias) tse += A2_SLENGTH(fnt->width[start+i])*fnt->height[start+i];
	  else tse += MONO_SLENGTH(fnt->width[start+i])*fnt->height[start+i];
	  //printf("%0.4x %i\n",start+i,temp);
	  write(out,&temp,sizeof(temp));
	  
	}

	write(out,&(fnt->width[start]),sizeof(u8)*count);
	write(out,&(fnt->height[start]),sizeof(u8)*count);
	write(out,&(fnt->xBearing[start]),sizeof(u8)*count);
	write(out,&(fnt->yBearing[start]),sizeof(u8)*count);
	write(out,&(fnt->advance[start]),sizeof(u8)*count);

	write(out,fnt->surface+tsf,tse);
	close(out);
	return 0;
}
/*}}}*/


int main ( int argc, char* argv[] ) /*{{{*/
{

  char out[255];
  int antialias = 0;

  sprintf(out,"%s_%s",argv[2],argv[3]);  
  if (argc > 6 && strcmp(argv[6],"-a") == 0 ) antialias = 1;

  fgl_ripFont(argv[1],out,atoi(argv[3]),strtol(argv[4],NULL,16),atoi(argv[5]),antialias);

  return 1;
}
/*}}}*/
