#include "fgl.h"

void storeFont ( fgl_font_t* fnt, char* fname, int ptSize )
{
	int  i, j, m, so;
	
	for ( i=0, m=0, so=-1; i<fnt->nGlyphs; i++ ){
		if ( (fnt->width[i]) && ((int)fnt->surfOff[i] > so) ){
			m += fnt->height[i] * MONO_SLENGTH(fnt->width[i]);
			so = fnt->surfOff[i];
		}
	}

	printf( "u8 %s_%d_glyphs[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<m; j++ ){
		printf( "0x%02x", fnt->surface[i]);
		if ( ++i < m ) {
			printf( ", ");
			if ( j == 8 ){
				j = 0;
				printf( "\t/* %5d */\n  ", i-8);
			}
		}
	}
	printf( "\n};\n\n");

	printf( "u32 %s_%d_glyphOffsets[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<fnt->nGlyphs; i++, j++ ){
		printf( "    %5d%c  /* %3d */\n", fnt->surfOff[i], (i<m)?',':' ',i);
	}
	printf( "\n};\n\n");

	printf( "u8 %s_%d_widths[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<fnt->nGlyphs; i++, j++ ){
		printf( "    %5d%c  /* %3d */\n", fnt->width[i], (i<m)?',':' ',i);
	}
	printf( "\n};\n\n");

	printf( "u8 %s_%d_heights[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<fnt->nGlyphs; i++, j++ ){
		printf( "    %5d%c  /* %3d */\n", fnt->height[i], (i<m)?',':' ',i);
	}
	printf( "\n};\n\n");

	printf( "i8 %s_%d_xBearing[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<fnt->nGlyphs; i++, j++ ){
		printf( "    %5d%c  /* %3d */\n", fnt->xBearing[i], (i<m)?',':' ',i);
	}
	printf( "\n};\n\n");

	printf( "i8 %s_%d_yBearing[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<fnt->nGlyphs; i++, j++ ){
		printf( "    %5d%c  /* %3d */\n", fnt->yBearing[i], (i<m)?',':' ',i);
	}
	printf( "\n};\n\n");

	printf( "i8 %s_%d_advance[] = {\n  ", fname, ptSize);
	for ( i=0, j=1; i<fnt->nGlyphs; i++, j++ ){
		printf( "    %5d%c  /* %3d */\n", fnt->advance[i], (i<m)?',':' ',i);
	}
	printf( "\n};\n\n");


	printf( "fgl_font_t %s_%d = {\n", fname, ptSize); 
	printf( "  %d,\t/* number of glyphs */\n", fnt->nGlyphs);
	printf( "  %d,\t/* unicode offset of first glyph */\n", fnt->uniOffset);
	printf( "  (u8*)  &%s_%d_glyphs,\n", fname, ptSize);
	printf( "  (u32*) &%s_%d_glyphOffsets,\n", fname, ptSize);
	printf( "  (u8*)  &%s_%d_widths,\n", fname, ptSize);
	printf( "  (u8*)  &%s_%d_heights,\n", fname, ptSize);
	printf( "  (i8*)  &%s_%d_xBearing,\n", fname, ptSize);
	printf( "  (i8*)  &%s_%d_yBearing,\n", fname, ptSize);
	printf( "  (i8*)  &%s_%d_advance,\n", fname, ptSize);
	printf( "  %d,\t/* fontHeight */\n", fnt->fontHeight);
	printf( "  %d,\t/* ascend */\n", fnt->ascend);
	printf( "  %d,\t/* descend */\n", fnt->descend);
    printf( "  \"%s\",\t/* fname */\n", fname);
    printf( "  %d,\t/* ptSize */\n", ptSize);
	printf( "  %d \t/* nRefs */\n", 0);
	printf( "};\n");
}

int main ( int argc, char* argv[] )
{
	fgl_font_t* fnt;
	char  *fname, *stem, *c;
	int ptSize, ptSizeSav;

	if ( argc < 3 ){
		puts( "usage savefonthdr <ttf-file stem> <ptSize> [<fake-ptSize>]");
		return 0;
	}

	fname = argv[1];
	ptSize = atoi( argv[2]);

	if ( argc == 4 ){
		ptSizeSav = atoi( argv[3]);
	}
	else {
		ptSizeSav = ptSize;
	}

	fnt = fgl_getFont( fname, ptSize);

	for ( c = fname, stem = fname; *c; c++ ){
		if ( *c == '/' ){
			stem = c+1;
		}
	}

	if ( fnt ){
		storeFont( fnt, stem, ptSizeSav);
	}
	fgl_disposeFont( fnt);

	return 1;
}
