/* -*- mode: C; mode: fold -*- */
/*
 * cmap_g4r.c - fgl module for pixelvalue/RGB mapping (4bpp grayscale)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: cmap_g4r.c,v 1.2 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_CMAP_INCLUDED
#define FGL_CMAP_INCLUDED

#define FGL_MODULE
#include "fgl.h"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif
/*}}}*/


#if !defined(FGL_PROTO)
u8 grayLevel[16] = { 255, 238, 221, 204, 187, 170, 153, 136,
                     119, 102, 85, 68, 51, 34, 17, 0 };
#else
extern u8 grayLevel[16];
#endif


/*
 * these are our public APIs for RGB<->pixVal conversion
 */

static inline
fgl_color_t fgl_getPixelValue ( u32 argb ) /*{{{*/
{
	u8 r = (argb >> 16) & 0xff;
	u8 g = (argb >> 8) & 0xff;
	u8 b = argb & 0xff;

	u8 min, max;

	min = (r < g) ? r : g;
	min = (min < b) ? min : b;

	max = (r > g) ? r : g;
	max = (max > b) ? max : b;

	return ((min + max) / 34);
}
/*}}}*/

static inline
void fgl_getRgbValues ( fgl_color_t pix, int* r, int* g, int* b ) /*{{{*/
{
	*r = *g = *b = grayLevel[ pix & 0xf];
}

/*}}}*/

static inline
fgl_color_t fgl_modPixelValue( fgl_color_t pix, int mod ) /*{{{*/
{
	mod *= 2;
	pix += mod;

	if ( pix > 0xf ) pix = 0xf;

	return pix;
}

/*}}}*/


static inline
fgl_color_t fgl_blendPixelValues ( u32 alpha, fgl_color_t d, fgl_color_t s ) /*{{{*/
{
	u32 ialpha = 255 - alpha;

	return (((alpha * s + ialpha * d) + 255) >> 8);
}


/*
 * This is the color mapping setup, which is called from init_XX, and
 * therefore can't be inlined
 */

int fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld )
#if !defined(FGL_PROTO)
{
	fgl_clrBlack     = fgl_getPixelValue( 0);
	fgl_clrWhite     = fgl_getPixelValue( 0xffffff);
	fgl_clrRed       = fgl_getPixelValue( 0xff0000);
	fgl_clrGreen     = fgl_getPixelValue( 0x00ff00);
	fgl_clrBlue      = fgl_getPixelValue( 0x0000ff);
	fgl_clrYellow    = fgl_getPixelValue( 0xffff00);
	fgl_clrCyan      = fgl_getPixelValue( 0x00ffff);
	fgl_clrGray      = fgl_getPixelValue( 0x808080);
	fgl_clrLightGray = fgl_getPixelValue( 0xc0c0c0);
	fgl_clrOrange    = fgl_getPixelValue( 0xffc800);

	/* we don't have to set up a colormap */
	return 1;
}
#endif /* FGL_PROTO */
;

static inline
fgl_cmap_t* fgl_getCmap () {
	return 0;
}

#endif /* FGL_CMAP_INCLUDED */
