/* -*- mode: C; mode: fold -*- */
/*
 * cmap_t444.c - fgl module for pixelvalue/RGB mapping (16 bit 444 TrueColor)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: cmap_t444.c,v 1.3 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_CMAP_INCLUDED
#define FGL_CMAP_INCLUDED

#define FGL_MODULE
#include "fgl.h"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif
/*}}}*/


/*
 * these are our public APIs for RGB<->pixVal conversion
 */

/*
 * RRRRRRRR.GGGGGGGG.BBBBBBBB
 * 00000000.0000rrrr.ggggbbbb
 */

static inline
fgl_color_t fgl_getPixelValue ( u32 argb ) /*{{{*/
{
	return (((argb & 0xf00000) >> 12) |
			((argb & 0xf000) >> 8) |
			((argb & 0xf0) >> 4));
}
/*}}}*/

static inline
void fgl_getRgbValues ( fgl_color_t pix, int* r, int* g, int* b ) /*{{{*/
{
	*r = (pix & 0xf00) >> 4;
	*g = (pix & 0x0f0);
	*b = (pix & 0x00f) << 4;
}
/*}}}*/

/*
 * This is here because we can't simply increase/dicrease channel values
 * in a seriously restricted colorspace (VGA, 4-bit grayscale etc.)
 */
static inline
fgl_color_t fgl_modPixelValue( fgl_color_t pix, int mod ) /*{{{*/
{
	u32 r, g, b;
	fgl_getRgbValues( pix, &r, &g, &b);

	mod = 3 + mod;
	if ( mod < 0 ) mod = 0;

	r = (r * mod) / 3;
	g = (g * mod) / 3;
	b = (b * mod) / 3;

	if ( r > 0xff ) r = 0xff;
	if ( g > 0xff ) g = 0xff;
	if ( b > 0xff ) b = 0xff;

	return ((r << 4) & 0xf000) | (g & 0x0f0) | (b >> 4);
} 
/*}}}*/


static inline
fgl_color_t fgl_blendPixelValues ( u32 alpha, fgl_color_t d, fgl_color_t s ) /*{{{*/
{
	u32 ialpha = 255 - alpha;
	u32  dr, dg, db, sr, sg, sb;
	
	fgl_getRgbValues( d, &dr, &dg, &db);
	fgl_getRgbValues( s, &sr, &sg, &sb);
	
	alpha++;
	
	dr = ((alpha * sr + ialpha * dr) + 255) >> 8;
	dg = ((alpha * sg + ialpha * dg) + 255) >> 8;
	db = ((alpha * sb + ialpha * db) + 255) >> 8;
	
	return fgl_getPixelValue( RGB_888(dr, dg, db));
}
/*}}}*/

/*
 * This is the color mapping setup, which is called from init_XX, and
 * therefore can't be inlined
 */

int fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld ) /*{{{*/
#if !defined(FGL_PROTO)
{
	fgl_clrBlack     = fgl_getPixelValue( 0);
	fgl_clrWhite     = fgl_getPixelValue( 0xffffff);
	fgl_clrRed       = fgl_getPixelValue( 0xff0000);
	fgl_clrGreen     = fgl_getPixelValue( 0x00ff00);
	fgl_clrBlue      = fgl_getPixelValue( 0x0000ff);
	fgl_clrYellow    = fgl_getPixelValue( 0xffff00);
	fgl_clrCyan      = fgl_getPixelValue( 0x00ffff);
	fgl_clrGray      = fgl_getPixelValue( 0x808080);
	fgl_clrLightGray = fgl_getPixelValue( 0xc0c0c0);
	fgl_clrOrange    = fgl_getPixelValue( 0xffc800);

	/* we don't have to set up a colormap */
	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/


static inline
fgl_cmap_t* fgl_getCmap () /*{{{*/
{
	return 0;
}
/*}}}*/

/* fgl module epilog */ /*{{{*/
#endif /* FGL_CMAP_INCLUDED */
/*}}}*/
