/* -*- mode: C; mode: fold -*- */
/*
 * draw_gen.c - fgl module to keep track of "dirty rectangles" (areas
 *              where drawing ops take place). This is the generic
 *              version which just takes care of the cursor (hide,re-display)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: draw_gen.c,v 1.10 2001/07/04 00:47:55 samc Exp $
 */

#ifndef FGL_DRAW_INCLUDED
#define FGL_DRAW_INCLUDED

#define FGL_MODULE
#include "fgl.h"
// #include FGL_INIT_MODULE

/*
 * Hrmm, we have to forward-define & de-inline them because begin/endDraw
 * are more valuable inlines
 */
void fgl_hideCursor();
void fgl_showCursor();

static inline
void fgl_beginDraw ( fgl_bitmap_t* bmp, int x0, int y0, int x1, int y1 , u32 op) /*{{{*/
{
        bmp->nextOperation = op;

	if ( bmp == fgl_screenBmp ) {
		if ( fgl_cursor && (_fgl_inCursorOp == 0) &&
			 (fgl_cursor->x0 <= x1) && (fgl_cursor->x1 >= x0) &&
			 (fgl_cursor->y0 <= y1) && (fgl_cursor->y1 >= y0) ){
			fgl_hideCursor();
			_fgl_redrawCursor = 1;
		}
#ifdef FGL_ACCELFN_PENDING
		fgl_accel_pending( bmp );
#endif
	}
}
/*}}}*/

static inline
void fgl_endDraw ( fgl_bitmap_t* bmp, int x0, int y0, int x1, int y1 ) /*{{{*/
{
	if ( bmp == fgl_screenBmp ) {
		if ( _fgl_redrawCursor && (_fgl_inCursorOp == 0)) {
			_fgl_redrawCursor = 0;
			fgl_showCursor();
		}
	}
	bmp->lastOperation = bmp->nextOperation;
}
/*}}}*/

static inline
void fgl_waitReady ( fgl_bitmap_t* bmp ) /*{{{*/
{
#ifdef FGL_ACCELFN_WAIT
		fgl_accel_wait( bmp );
#endif
}
/*}}}*/

#endif /* FGL_DRAW_INCLUDED */

