/* -*- mode: C; mode: fold -*- */
/*
 * draw_x.c   - fgl module to keep track of "dirty rectangles" (areas
 *              where drawing ops take place). This is the X simulator
 *              version which has to copy the framebuffer contents of
 *              the draw rectangle to the X drawable/image
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: draw_x.c,v 1.9 2001/07/04 00:47:55 samc Exp $
 */

#ifndef FGL_DRAW_INCLUDED
#define FGL_DRAW_INCLUDED

#include "X11/Xlib.h"
#include <X11/extensions/XShm.h>

#define FGL_MODULE
#include "fgl.h"

extern Window win;
extern Display *display;
extern XImage *image;
extern GC gc;
extern Atom UPDATE;

/*
 * Hrmm, we have to forward-define & de-inline them because begin/endDraw
 * are more valuable inlines
 */
void fgl_hideCursor();
void fgl_showCursor();

static inline
void fgl_beginDraw ( fgl_bitmap_t* bmp, int x0, int y0, int x1, int y1 , u32 op) /*{{{*/
{
	
	if ( bmp == fgl_screenBmp ) {
		if ( fgl_cursor && (_fgl_inCursorOp == 0) &&
			 (fgl_cursor->x0 <= x1) && (fgl_cursor->x1 >= x0) &&
			 (fgl_cursor->y0 <= y1) && (fgl_cursor->y1 >= y0) ){
			fgl_hideCursor();
			_fgl_redrawCursor = 1;
		}
	}

	//printf(" < %d,%d %d,%d\n", x0, y0, x1, y1);
}
/*}}}*/

extern void updateScreen ( int x0, int y0, int x1, int y1 );

static inline
void fgl_endDraw ( fgl_bitmap_t* bmp, int x0, int y0, int x1, int y1 ) /*{{{*/
{
	if ( bmp == fgl_screenBmp ) {
		if ( _fgl_redrawCursor && (_fgl_inCursorOp == 0)) {
			_fgl_redrawCursor = 0;
			fgl_showCursor();
		}
		updateScreen( x0, y0, x1, y1 );
	}
}
/*}}}*/

static inline
void fgl_waitReady ( fgl_bitmap_t* bmp ) /*{{{*/
{

}
#endif /* FGL_DRAW_INCLUDED */
