/* -*- mode: C; mode: fold -*- */
/*
 * fgld_8  - fgl header to handle framebuffer access and scanline cells
 *           for 8-bpp visuals (read/write value, compute/alter scanline 
 *           cell addresses), generic part (might be partly overriden by
 *           platform specific optimizations)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: fgld_8.h,v 1.10 2001/07/04 00:47:55 samc Exp $
 */

#define IS_CMAP_MODE 1  /* define this when we have a colormap based mode */

#if !defined(FGL_PROTO)

typedef u8 fgl_slc_t;     /* scanline cell type */
typedef u8 fgl_color_t;   /* pixel color value */

#else /* defined FGL_PROTO */

/*****************************************************************************
 * internal helper types and functions
 */

#if defined (TARGET_CPU_I386)  /* simple intel optimizations */
#include "fgld_8_i386.h"
#endif /* defined I386 */


/***************************************************************************
 * public functions
 */


static inline
int fgl_getSlcSize () /*{{{*/
{
	return 1;
}
/*}}}*/

static inline
int fgl_getBitsPerPixel () /*{{{*/
{
	return 8;
}
/*}}}*/

// return length of scanline in cell units
static inline
int fgl_getScanlineLength ( int length ) /*{{{*/
{
	return length;
}
/*}}}*/

static inline
fgl_slc_t* fgl_getSlc ( fgl_bitmap_t* bmp, u32 x, u32 y ) /*{{{*/
{
	return (bmp->surface + (y * bmp->sLength) + x);
}
/*}}}*/

static inline
void fgl_incSlc ( fgl_slc_t **p, u32 x ) /*{{{*/
{
	(*p)++;
}
/*}}}*/

static inline
void fgl_decSlc ( fgl_slc_t **p, u32 x ) /*{{{*/
{
	(*p)--;
}
/*}}}*/

static inline
void fgl_addSlc ( fgl_slc_t **p, u32 x, u32 pixels ) /*{{{*/
{
	(*p) += pixels;
}
/*}}}*/

static inline
void fgl_subSlc ( fgl_slc_t **p, u32 x, u32 pixels ) /*{{{*/
{
	(*p) -= pixels;
}
/*}}}*/

static inline
fgl_color_t fgl_getSlcPixel ( fgl_slc_t *p, u32 x ) /*{{{*/
{
	return (fgl_color_t) *p;
}
/*}}}*/

static inline
void fgl_setSlcPixel ( fgl_slc_t *p, u32 x, fgl_color_t clr ) /*{{{*/
{
	*p = clr;
}
/*}}}*/

static inline
void fgl_setSlcPixelXor ( fgl_slc_t *p, u32 x, fgl_color_t clr ) /*{{{*/
{
	*p ^= clr;
}
/*}}}*/


#ifndef HAVE_SETSLCCOLS
static inline
void fgl_setSlcCols ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	fgl_slc_t *pmax = p + len;

	while ( p < pmax  ){
		*p++ = clr;
	}
}
/*}}}*/
#endif /* HAVE_SETSLCCOLS */

static inline
void fgl_setSlcColsXor ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	fgl_slc_t *pmax = p + len;

	while ( p < pmax  ){
		*p++ ^= clr;
	}
}
/*}}}*/

static inline
void fgl_setSlcRows ( fgl_slc_t *p, u32 inc, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	do {
		*p = clr;
		p += inc;
	} while ( --len );
}
/*}}}*/

static inline
void fgl_setSlcRowsXor ( fgl_slc_t *p, u32 inc, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	do {
		*p ^= clr;
		p += inc;
	} while ( --len );
}
/*}}}*/


#ifndef HAVE_INITBMP
static inline
void fgl_initBmp ( fgl_bitmap_t* bmp, i32 clr ) /*{{{*/
{
	fgl_slc_t* p = bmp->surface;
	u32 len = bmp->sLength * bmp->height;
	u32 *pp;
	u32 cc;
	
	clr &= 0xff;
	cc = (clr << 8) | clr;
	cc = (cc << 16) | cc;

	pp = (u32*) p;
	while (len > 31){
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		*pp++ = cc;
		
		len -= 32;
	}

	if ( len ){
		p = ((fgl_slc_t*)pp);
		while ( len-- ){
			*p++ = (fgl_slc_t)clr;
		}
	}
}
/*}}}*/
#endif /* HAVE_INITBMP */


#ifndef HAVE_CPYSLC
static inline
void fgl_cpySlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len ) /*{{{*/
{
	memcpy( dest, src, len); 
}
/*}}}*/
#endif /* HAVE_CPYSLC */


#ifndef HAVE_MOVESLC
static inline
void fgl_moveSlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len ) /*{{{*/
{
	memmove( dest, src, len); 
}
/*}}}*/
#endif /* HAVE_MOVESLC */

#endif  /* defined FGL_PROTO */

