/* -*- mode: C; mode: fold -*- */
/*
 * init_x.c - FGL_INIT_MODULE to initialize the framebuffer device and
 *             global fgl variables (Linux version)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: init_x.c,v 1.18 2001/07/25 21:42:56 jsimmons Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_INIT_INCLUDED
#define FGL_INIT_INCLUDED

#include <asm/page.h>

#define FGL_MODULE
#define FGL_MAIN
#include "fgl.h"

#include FGL_HEAP_MODULE
#include "util.c"
#include "screencfg.c"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif
/*}}}*/


/* globals and platform specifics */ /*{{{*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <pthread.h>
#include <sched.h>

/* the non page boundary fb address fix */
#include <sys/user.h>
#if defined(USE_VT_CTRL)
#include <linux/kd.h>
#endif

FGL_GLOBAL Display *display;
FGL_GLOBAL Window win;
FGL_GLOBAL XImage *image;
FGL_GLOBAL GC gc;
FGL_GLOBAL Atom       UPDATE;
FGL_GLOBAL int inputListener;

XShmSegmentInfo *shmSegmentInfo;
pthread_t  loopThread;

Atom       STOP;
Atom       WM_PROTOCOLS;
Atom       WM_DELETE_WINDOW;
int        iFifoM;

/*}}}*/

static inline
int fgl_getScreenWidth () /*{{{*/
{
	return fgl_screenCtx->bmp->width;
}
/*}}}*/

static inline
int fgl_getScreenHeight () /*{{{*/
{
	return fgl_screenCtx->bmp->height;
}
/*}}}*/

static inline
int fgl_getScreenXoff () /*{{{*/
{
	return fgl_screenCtx->xoff;
}
/*}}}*/

static inline
int fgl_getScreenYoff () /*{{{*/
{
	return fgl_screenCtx->yoff;
}
/*}}}*/


/*
 * fgl_terminate - clean up system resources
 * 
 * This is automatically called during process exit, so don't inline.
 * It also has to be able to detect multiple calls
 */
void fgl_terminate () /*{{{*/
#if !defined(FGL_PROTO)
{
        XEvent event;

        if ( loopThread ){
                event.xclient.type = ClientMessage;
                event.xclient.message_type = STOP;
                event.xclient.format = 32;
                event.xclient.window = win;
                XSendEvent( display, win, False, 0, &event);
        }
}
#endif /* FGL_PROTO */
;/*}}}*/


#if !defined(FGL_PROTO)

void updateScreen(int x0, int y0, int x1, int y1) {

  XShmPutImage( display, win, gc, image, x0, y0, x0, y0
		, x1 - x0 + 1, y1 - y0 + 1, True);
  XFlush( display );

}

void *eventLoop ( void *p)
{
  int x0, y0, x1, y1;
  XEvent event;
  char buffer[5];
  char status;
  KeySym keysym;
  XComposeStatus ioStatus;

  iFifoM = open("/dev/xps2", O_WRONLY);

  while ( 1 ){
#if defined(USE_UNIX_JTHREADS)
    jthreadedPollFd(ConnectionNumber(display));
#endif
    XNextEvent( display, &event );
    switch ( event.type ){
    case Expose:
      x0 = event.xexpose.x;
      y0 = event.xexpose.y;
      x1 = x0 + event.xexpose.width  - 1;      
      y1 = y0 + event.xexpose.height - 1;      
      while (XCheckMaskEvent( display, Expose, &event)) {
	if ( event.xexpose.x < x0 ) x0 = event.xexpose.x;
	if ( event.xexpose.y < y0 ) y0 = event.xexpose.y;
	
	if ( event.xexpose.width + event.xexpose.x - 1 > x1 ) {
	  x1 = event.xexpose.width + event.xexpose.x - 1;
	}
	if ( event.xexpose.height + event.xexpose.y - 1 > y1 ) {
	  y1 = event.xexpose.height + event.xexpose.y - 1;
	}
      }
      updateScreen( x0, y0, x1, y1 ); 
      break;
    case ButtonPress:
    case ButtonRelease:
      if ( iFifoM ) {
	      status = 0;
	      if ( event.xbutton.type == ButtonPress )   status |= 1;
	      if ( event.xbutton.type == ButtonRelease ) status |= 2;
	      if ( event.xbutton.button == Button1 )     status |= 16;
	      if ( event.xbutton.button == Button2 )     status |= 64;
	      if ( event.xbutton.button == Button3 )     status |= 32;
	      buffer[0] = status;
	      buffer[1] = 0xff & event.xbutton.x;
	      buffer[2] = event.xbutton.x >> 8;
	      buffer[3] = 0xff & event.xbutton.y;
	      buffer[4] = event.xbutton.y >> 8;
	      write(iFifoM,buffer,5);
      }
      break;
    case MotionNotify:
      if ( iFifoM ) {
	    status = 4;
	    if ( event.xmotion.state & Button1Mask ) status |= 16;
	    if ( event.xmotion.state & Button2Mask ) status |= 64;
	    if ( event.xmotion.state & Button3Mask ) status |= 32;
	    buffer[0] = status;
	    buffer[1] = 0xff & event.xmotion.x;
	    buffer[2] = event.xmotion.x >> 8;
	    buffer[3] = 0xff & event.xmotion.y;
	    buffer[4] = event.xmotion.y >> 8;
	    write(iFifoM,buffer,5);	    
      }
      break;    
    case KeyPress:
    case KeyRelease:
      if ( iFifoM ) {
	    status = 0;
	    if ( event.xkey.type == KeyPress)    status |= 1;
	    if ( event.xkey.type == KeyRelease)  status |= 2;
	    if ( event.xkey.state & ShiftMask)   status |= 16;
	    if ( event.xkey.state & ControlMask) status |= 64;
	    if ( event.xkey.state & Mod1Mask)    status |= 128;
	    buffer[3] = XLookupString( &event.xkey, &buffer[4], 1, 0, &ioStatus);
	    keysym = XKeycodeToKeysym( display, event.xkey.keycode, 0);
	    if (keysym >= 0xff00 ) status |= 4;
	    buffer[0] = 0;
	    buffer[1] = status;
	    buffer[2] = keysym & 0xff;
	    write(iFifoM,buffer,5);	    
      }
      break;
    case ClientMessage:
      if ( event.xclient.message_type == STOP ){
	XShmDetach( display, shmSegmentInfo );
	image->data = 0;
	XDestroyImage( image);
	free( shmSegmentInfo );
	
	if ( win ){
	  XDestroyWindow( display, win);
	  win = 0;
	}
	XCloseDisplay( display);
	
	loopThread = 0;

	return 0;
      }
      else if ( event.xclient.message_type == WM_PROTOCOLS) {
	if ( event.xclient.data.l[0] == WM_DELETE_WINDOW){
	  fgl_terminate();
	}
      }
      
      break;
    }
    
  }
  return 0;
}

static 
void
startLoopThread ()
{
#if defined(USE_UNIX_JTHREADS)
	loopThread = (pthread_t)createDaemon(eventLoop, "eventloop", 5);
#else
	int min, max;
	pthread_attr_t attr;
	struct sched_param sp;

	min = sched_get_priority_min( SCHED_OTHER);
	max = sched_get_priority_max( SCHED_OTHER);

	sp.sched_priority = min+1;  /* low */

	pthread_attr_init( &attr);
	pthread_attr_setschedparam( &attr, &sp);
	pthread_attr_setstacksize( &attr, 4096);

	pthread_create( &loopThread, &attr, eventLoop, 0);
#endif
}

#endif /* FGL_PROTO */



/*
 * This is our color mapping set up defined in the cmap_XX module. We
 * can't inline because cmap uses our colormap functions (which are
 * device dependent)
 */
extern int fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld );

/*
 * fgl_init - initialize system resources, screenBmp and color map
 * 
 *   nReq    : number of requested RGB cells (pseudocolor only)
 *   rgbReqs : 888 RGB values of requested colors (pseudocolor only)
 *   restore : 1: restore colormap when process exits (pseudocolor only)
 */
int fgl_init ( int nReq, u32* rgbReqs, int restore ) /*{{{*/
#if !defined(FGL_PROTO)
{

	int screen;
	XSizeHints *sizeHints;
	XWMHints   *wmHints;
	XClassHint *classHint;
	Visual *visual;
	XGCValues values;
	unsigned long valueMask;
	int i;
	long nBytes;
	fgl_config_t cfg;
	int xoff = 100;
	int yoff = 100;
	int width = 640;
	int height = 480;

	if (fgl_readConfig( &cfg)) {
	  if ( cfg.xoff != -1 && cfg.yoff != -1 &&
	       cfg.width != -1 && cfg.height != -1 ) {
	    xoff   = cfg.xoff;
	    yoff   = cfg.yoff;
	    width  = cfg.width;
	    height = cfg.height;
	  }
	}

	sizeHints = XAllocSizeHints();
	wmHints   = XAllocWMHints();
	classHint = XAllocClassHint();

	display = XOpenDisplay( NULL );

	if (!display)
		err("Can't connect to X display\n");

	screen  = DefaultScreen( display );
	win = XCreateSimpleWindow( display, RootWindow( display, screen ),
				   xoff, yoff, width, height, 2, 
				   BlackPixel( display, screen ),
				   WhitePixel( display, screen ));

	sizeHints->flags = PPosition | PSize | PMinSize;
	sizeHints->min_width  = width;
	sizeHints->min_height = height;

	wmHints->initial_state = NormalState;
	wmHints->input = True;
	wmHints->flags = StateHint | InputHint;

	classHint->res_name  = "window";
	classHint->res_class = "FGL";

	XSetWMProperties( display, win, NULL, NULL, NULL, NULL, sizeHints,
			  0, classHint );

	XSelectInput( display, win, ExposureMask | KeyPressMask | 
		      KeyReleaseMask | ButtonPressMask | ButtonReleaseMask |
		      PointerMotionMask);

        values.graphics_exposures = False;
        valueMask = GCGraphicsExposures;

	gc = XCreateGC( display, win, valueMask, &values );


	XMapWindow( display, win );
	visual = DefaultVisual( display, screen );
	XFlush( display );

	shmSegmentInfo = (XShmSegmentInfo *) malloc(sizeof(XShmSegmentInfo));
	image = XShmCreateImage( display, visual, 16, ZPixmap, 0, shmSegmentInfo,
				width, height);
	if ( !image ){
		free( shmSegmentInfo );
		return 0;
	}

	nBytes = image->bytes_per_line * height;
	shmSegmentInfo->shmid = shmget(IPC_PRIVATE, nBytes, (IPC_CREAT | 0777));

	if ( shmSegmentInfo->shmid == -1 ) {
		XShmDetach( display, shmSegmentInfo );
		image->data = 0;
		XDestroyImage( image );
		free( shmSegmentInfo );
		return 0;
	}

	image->data = shmSegmentInfo->shmaddr = shmat( shmSegmentInfo->shmid, 0, 0);
	shmSegmentInfo->readOnly = False;
	XShmAttach( display, shmSegmentInfo);

	shmctl( shmSegmentInfo->shmid, IPC_RMID, 0);
	fgl_screenBmp->surface = (fgl_slc_t*)shmSegmentInfo->shmaddr; 

	for( i = 0; i< width * height; i++) {
		fgl_screenBmp->surface[i] = 0;
	}

	XShmPutImage( display, win, gc, image, 0, 0, 0, 0, width, height, False);
	XFlush( display );

	fgl_initRgbCmap ( nReq, rgbReqs, restore );

	fgl_screenBmp->width  = width;
	fgl_screenBmp->height = height;
	fgl_screenBmp->state  |= IMAGE_ON_SCREEN;
	fgl_screenBmp->sLength = width;
	fgl_screenCtx->bmp     = fgl_screenBmp;
	fgl_screenCtx->alpha   = 0xff;
	fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
	fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
	fgl_screenCtx->xoff    = 0;
	fgl_screenCtx->yoff    = 0;

	UPDATE = XInternAtom( display, "UPDATE", False);
	STOP = XInternAtom( display, "STOP", False);
	WM_PROTOCOLS = XInternAtom( display, "WM_PROTOCOLS", False);
	WM_DELETE_WINDOW = XInternAtom( display, "WM_DELETE_WINDOW", False);

	startLoopThread();
	
	return 1;
}
#endif
;/*}}}*/


/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 */
static inline
int fgl_initialize () /*{{{*/
{	
	return fgl_init( 0, 0, 1);
}
/*}}}*/

/*
 * fgl_initialize - convenience wrapper for fgl_init w/o RGB requests
 * specifying size
 */
static inline
int fgl_open (u32 x, u32 y, u32 w, u32 h) /*{{{*/
{	
        int r;

        r = fgl_init( 0, 0, 1);

	fgl_screenBmp->width   = w; 
	fgl_screenBmp->height  = h;

#ifdef ROT_90
	rotate32( fgl_screenBmp, &x, &y);
	fgl_screenCtx->x1Clip  = h-1;
	fgl_screenCtx->y1Clip  = w-1;
#else
	fgl_screenCtx->x1Clip  = w-1;
	fgl_screenCtx->y1Clip  = h-1;
#endif
	fgl_screenBmp->surface = fgl_getSlc( fgl_screenBmp, x, y);
	return r;
}
/*}}}*/

/****************************************************************************
 * These are the low level functions to manage device specific colormaps.
 * Note that we can't inline them, because they are used from the cmap_XX
 * module, which is included
 */

static inline
int fgl_hasCmap () /*{{{*/
{
#if defined (IS_CMAP_MODE)
	return 1;
#else
	return 0;
#endif
}/*}}}*/

fgl_cmap_t* fgl_allocCmap () /*{{{*/
#if !defined(FGL_PROTO)
{
	fgl_cmap_t* cm = 0;

#if defined (IS_CMAP_MODE)
	int i, n;

	if ( fgl_hasCmap() ){
		for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
		cm = FGL_CALLOC( 1, sizeof(fgl_cmap_t) + 3 * n);

		cm->r = (u8*)cm + sizeof(fgl_cmap_t);
		cm->g = cm->r + n;
		cm->b = cm->g + n;
	}
#endif /* IS_CMAP_MODE */

	return cm;
}
#endif /* FGL_PROTO */
;/*}}}*/

void fgl_freeCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
	FGL_FREE( cm);
}
#endif /* FGL_PROTO */
;/*}}}*/

int fgl_getHWCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i, n;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;

	for ( i=0, n=1; i<_fgl_varInfo.bits_per_pixel; i++, n <<= 1 );
	cmap.len = n;

	buf = (u16*) FGL_MALLOC( n * 3 * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + n;
	cmap.blue  = cmap.green + n;
	cmap.transp = 0;

	if ( ioctl( _fgl_fh, FBIOGETCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}

	cm->len = cmap.len;
	for ( i=0; i<cmap.len; i++ ){
		cm->r[i] = cmap.red[i] >> 8;
		cm->g[i] = cmap.green[i] >> 8;
		cm->b[i] = cmap.blue[i] >> 8;
	}

	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/

int fgl_setHWCmap ( fgl_cmap_t* cm ) /*{{{*/
#if !defined(FGL_PROTO)
{
#if defined (IS_CMAP_MODE)
	struct fb_cmap cmap;
	int i;
	u16 *buf;

	if ( !fgl_hasCmap() ){
		return 0;
	}

	cmap.start = 0;
	cmap.len = cm->len;

	buf = (u16*) FGL_MALLOC( 3 * cm->len * sizeof(u16));
	
	cmap.red   = buf;
	cmap.green = cmap.red + cmap.len;
	cmap.blue  = cmap.green + cmap.len;
	cmap.transp = 0;

	for ( i=0; i<cm->len; i++ ){
		cmap.red[i]   = (u16)(cm->r[i]) << 8;
		cmap.green[i] = (u16)(cm->g[i]) << 8;
		cmap.blue[i]  = (u16)(cm->b[i]) << 8;
	}

	if ( ioctl( _fgl_fh, FBIOPUTCMAP, &cmap) ){
		FGL_FREE( buf);
		return 0;
	}
	
	FGL_FREE( buf);
#endif /* IS_CMAP_MODE */

	return 1;
}
#endif /* FGL_PROTO */
;/*}}}*/

/* fgl module prolog */ /*{{{*/
#endif /* FGL_INIT_INCLUDED */
/*}}}*/
