/* -*- mode: C; mode: fold -*- */
/*
 * line.c - fgl module to draw lines
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: line.c,v 1.24 2001/11/30 04:05:18 peter Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_LINE_INCLUDED
#define FGL_LINE_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include "cntx.c"
#include FGL_DRAW_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_LINE
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/

#ifdef ROT_90
static inline
void fgl_drawVLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 y1 ) /*{{{*/
{
	fgl_slc_t *p;
	i32 x1 = 0;

	if ( y1 < y0 ) swap32( &y0, &y1);

	if ( (x0 > ctx->x1Clip) || (x0 < ctx->x0Clip) ||
		 (y1 < ctx->y0Clip) || (y0 > ctx->y1Clip) )
		return;

	if ( y0 < ctx->y0Clip ) y0 = ctx->y0Clip;
	if ( y1 > ctx->y1Clip ) y1 = ctx->y1Clip;

	rotate32( ctx->bmp, (u32*)&x0, (u32*)&y0);
	rotate32( ctx->bmp, (u32*)&x1, (u32*)&y1);

	swap32( &x0, &x1);

#else
static inline
void fgl_drawHLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1 ) /*{{{*/
{
	fgl_slc_t *p;

	if ( x1 < x0 ) swap32( &x0, &x1);

	if ( (y0 > ctx->y1Clip) || (y0 < ctx->y0Clip) ||
		 (x1 < ctx->x0Clip) || (x0 > ctx->x1Clip) )
		return;

	if ( x0 < ctx->x0Clip ) x0 = ctx->x0Clip;
	if ( x1 > ctx->x1Clip ) x1 = ctx->x1Clip;

#endif

#ifdef FGL_ACCELFN_DRAWHLINE
	if (ctx->bmp->state & IMAGE_ON_SCREEN) {
	  fgl_screenCtx->clr = ctx->clr;
	  fgl_screenCtx->mode = ctx->mode;
	  fgl_beginDraw( ctx->bmp, x0, y0, x1, y0, FGL_ACCEL_DRAWLINE_OP );
	  fgl_accel_drawHLine( fgl_screenCtx, x0, y0, x1 );
	  fgl_endDraw( ctx->bmp, x0, y0, x1, y0);
	  return ;
	}
#endif

	p = fgl_getSlc( ctx->bmp, x0, y0);

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y0, FGL_FRAMEBUFFER_OP );
	MODE_BRANCH( ctx, fgl_setSlcCols, (p, x0, (x1-x0)+1, ctx->clr));
	fgl_endDraw( ctx->bmp, x0, y0, x1, y0);
}
/*}}}*/
#ifdef ROT_90
static inline
void fgl_drawHLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1 ) /*{{{*/
{
	fgl_slc_t *p;
	i32 y1 = 0;

	if ( x1 < x0 ) swap32( &x0, &x1);

	if ( (y0 > ctx->y1Clip) || (y0 < ctx->y0Clip) ||
		 (x1 < ctx->x0Clip) || (x0 > ctx->x1Clip) )
		return;

	if ( x0 < ctx->x0Clip ) x0 = ctx->x0Clip;
	if ( x1 > ctx->x1Clip ) x1 = ctx->x1Clip;

	rotate32( ctx->bmp, (u32*)&x0, (u32*)&y0);
	rotate32( ctx->bmp, (u32*)&x1, (u32*)&y1);
#else 
static inline
void fgl_drawVLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 y1 ) /*{{{*/
{
	fgl_slc_t *p;

	if ( y1 < y0 ) swap32( &y0, &y1);

	if ( (x0 > ctx->x1Clip) || (x0 < ctx->x0Clip) ||
		 (y1 < ctx->y0Clip) || (y0 > ctx->y1Clip) )
		return;

	if ( y0 < ctx->y0Clip ) y0 = ctx->y0Clip;
	if ( y1 > ctx->y1Clip ) y1 = ctx->y1Clip;

#endif

#ifdef FGL_ACCELFN_DRAWVLINE
	if (ctx->bmp->state & IMAGE_ON_SCREEN) {
	  fgl_screenCtx->clr = ctx->clr;
	  fgl_screenCtx->mode = ctx->mode;
	  fgl_beginDraw( ctx->bmp, x0, y0, x0, y1, FGL_ACCEL_DRAWLINE_OP ); 
	  fgl_accel_drawVLine( fgl_screenCtx, x0, y0, y1 );
	  fgl_endDraw( ctx->bmp, x0, y0, x0, y1);
	  return ;
	}
#endif

	p = fgl_getSlc( ctx->bmp, x0, y0);

	fgl_beginDraw( ctx->bmp, x0, y0, x0, y1, FGL_FRAMEBUFFER_OP );

	MODE_BRANCH( ctx, fgl_setSlcRows, (p, ctx->bmp->sLength, x0, (y1-y0)+1, ctx->clr));

	fgl_endDraw( ctx->bmp, x0, y0, x0, y1);
}
/*}}}*/


static inline
i32 rnd ( i32 a, i32 b, i32 c )
{
	const int D = 8192;
	i32 d = (a * b) * D;
	d += (d <0) ? -D/2 : D/2;
	return (d / c) / D;
}

static inline
int clipLine ( fgl_context_t* ctx, i32* px0, i32* py0, i32* px1, i32* py1 ) /*{{{*/
{
	u8 c0, c1;
	i32 dx, dy;

	c0 = fgl_getClipCode( ctx, *px0, *py0);
	c1 = fgl_getClipCode( ctx, *px1, *py1);
	
	while ( 1 ){
		if ( (c0 | c1) == 0 ){         /* completely visible */
			return 1;
		}
		else if ( (c0 & c1) != 0 ){    /* completely invisible */
			return 0;
		}
		else {
			if ( !c0 ){                /* p0 needs to be outside */
				swap32( px0, px1);
				swap32( py0, py1);
				swap8( &c0, &c1);
			}

			dx = (*px1 - *px0);
			dy = (*py1 - *py0);

			if ( c0 & CC_LEFT ){
				*py0 += rnd((ctx->x0Clip - *px0), dy, dx);
				*px0 = ctx->x0Clip;
				
				c0 &= ~CC_LEFT;
				if ( *py0 > ctx->y1Clip )      c0 |= CC_BOTTOM;
				else if ( *py0 < ctx->y0Clip ) c0 |= CC_TOP;
				else                           c0 &= ~(CC_BOTTOM | CC_TOP);
			}
			else if ( c0 & CC_RIGHT ){
				*py0 += rnd((ctx->x1Clip - *px0), dy, dx);								
				*px0 = ctx->x1Clip;
				
				c0 &= ~CC_RIGHT;
				if ( *py0 > ctx->y1Clip )      c0 |= CC_BOTTOM;
				else if ( *py0 < ctx->y0Clip ) c0 |= CC_TOP;
				else                           c0 &= ~(CC_BOTTOM | CC_TOP);
			}
			else if ( c0 & CC_BOTTOM ){
				if ( *px0 != *px1 ) {
					*px0 += rnd((ctx->y1Clip - *py0), dx, dy);
				}
				*py0 = ctx->y1Clip;

				c0 &= ~CC_BOTTOM;
				if( *px0 > ctx->x1Clip )       c0 |= CC_RIGHT;
				else if ( *px0 < ctx->x0Clip ) c0 |= CC_LEFT;
				else                           c0 &= ~(CC_RIGHT | CC_LEFT);
			}
			else {  /* if (c0 & CC_TOP) */
				if ( *px0 != *px1 ) {
					*px0 += rnd((ctx->y0Clip - *py0), dx, dy);
				}
				*py0 = ctx->y0Clip;

				c0 &= ~CC_TOP;
				if( *px0 > ctx->x1Clip )       c0 |= CC_RIGHT;
				else if ( *px0 < ctx->x0Clip ) c0 |= CC_LEFT;
				else                           c0 &= ~(CC_RIGHT | CC_LEFT);
			}
		}
	}
}
/*}}}*/


FUNC_ATTR
void fgl_drawLine ( fgl_context_t *ctx,  i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_LINE)
{
	fgl_slc_t *p;

	i32 x, dx, dy, a, b, c, slinc, slinc1;

	if ( !clipLine( ctx, &x0, &y0, &x1, &y1) ){
		return;
	}

#ifdef ROT_90	
	rotate32( ctx->bmp, (u32*)&x0, (u32*)&y0);
	rotate32( ctx->bmp, (u32*)&x1, (u32*)&y1);
#endif

#ifdef FGL_ACCELFN_DRAWLINE
	if (ctx->bmp->state & IMAGE_ON_SCREEN) {
	  fgl_screenCtx->clr = ctx->clr;
	  fgl_screenCtx->mode = ctx->mode;
	  fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_ACCEL_DRAWLINE_OP ); 
	  fgl_accel_drawLine( fgl_screenCtx, x0, y0, x1, y1);
	  fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
	  return ;
	}
#endif

	if ( x1 < x0 ) {
		swap32( &x0, &x1);
		swap32( &y0, &y1);
	}

	dx = (i32)x1 - (i32)x0;
	dy = (i32)y1 - (i32)y0;

	p = fgl_getSlc( ctx->bmp, x0, y0);

	if ( dy < 0 ){
		swap32( &y0, &y1);
	}
	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP ); 

	if ( dy == 0 ){        /* horizontal line */
		MODE_BRANCH( ctx, fgl_setSlcCols,(p, x0, dx+1, ctx->clr));
	}
	else if ( dx == 0 ){   /* vertical line */
		i32 inc, len;
		if ( dy > 0 ){
			inc = ctx->bmp->sLength; len = dy;
		}
		else {
			inc = - (i32)ctx->bmp->sLength; len = -dy;
		}
		MODE_BRANCH( ctx, fgl_setSlcRows,(p, inc, x0, len+1, ctx->clr));
	}
	else if ( dx == -dy ){ /* diagonal up */
		MODE_BRANCH( ctx, fgl_setSlcRows,(p, -(ctx->bmp->sLength-1), x0, -dy+1, ctx->clr));
	}
	else if ( dx == dy ){  /* diagonal down */
		MODE_BRANCH( ctx, fgl_setSlcRows,(p, ctx->bmp->sLength+1, x0, dy+1, ctx->clr));
	}
	else {
		x = x0;
		slinc = ctx->bmp->sLength;
		if ( dy < 0 ){
			dy = - dy;
			slinc = -slinc; 
		}
		slinc1 = slinc+1;

		if ( dx > dy ){ /* m < 1 */
			a = 2 * dy;
			b = 2 * (dy-dx);
			c = 2 * dy - dx;
				
			/* uggly, but definitely more efficient then to branch pixelwise */
			if ( ctx->mode == PAINT_MODE ){
				fgl_setSlcPixel( p, x, ctx->clr);
				while ( x < x1 ){
					if ( c < 0 ){
						fgl_incSlc( &p, x);          c += a;
					}
					else{
						fgl_addSlc( &p, x, slinc1);  c += b;
					}
					x++;
					fgl_setSlcPixel( p, x, ctx->clr);
				}
			}
			else {
				fgl_setSlcPixelXor( p, x, ctx->clr);
				while ( x <= x1 ){
					if ( c < 0 ){
						fgl_incSlc( &p, x);          c += a;
					}
					else{
						fgl_addSlc( &p, x, slinc1);  c += b;
					}	
					x++;
					fgl_setSlcPixelXor( p, x, ctx->clr);
				}
			}
		}
		else {   /* m > 1 */
			a = 2 * dx;
			b = 2 * (dx-dy);
			c = 2 * dx - dy;
				
			if ( ctx->mode == PAINT_MODE ){
				fgl_setSlcPixel( p, x, ctx->clr);
				while ( dy-- ){
					if ( c < 0 ){
						fgl_addSlc( &p, x, slinc);       c += a;
					}
					else{
						fgl_addSlc( &p, x, slinc1);      c += b;
						x++;
					}
					fgl_setSlcPixel( p, x, ctx->clr);
				}
			}
			else {
				fgl_setSlcPixelXor( p, x, ctx->clr);
				while ( dy-- ){
					if ( c < 0 ){
						fgl_addSlc( &p, x, slinc);       c += a;
					}
					else{
						fgl_addSlc( &p, x, slinc1);      c += b;
						x++;
					}
					fgl_setSlcPixelXor( p, x, ctx->clr);
				}
			}
		}
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
#endif /* !FGL_PROTO */
;
/*}}}*/

/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_LINE_INCLUDED */
/*}}}*/
