/* -*- mode: C; mode: fold -*- */
/*
 * oval.c - fgl module to draw and fill circles and ellipses
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: oval.c,v 1.18 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_OVAL_INCLUDED
#define FGL_OVAL_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include "cbmp.c"
#include "util.c"
#include FGL_DRAW_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_OVAL
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/

static inline
void setCirclePoints ( fgl_context_t* ctx,
					   i32 cx, i32 cy, i32 dx, i32 dy, u8 clip ) /*{{{*/
{
	u32 x0 = cx + dx;
	u32 y0 = cy + dy;
	u32 x1 = cx - dx;
	u32 y1 = cy - dy;
	u32 x2 = cx + dy;
	u32 y2 = cy + dx;
	u32 x3 = cx - dy;
	u32 y3 = cy - dx;

	if ( clip == CR_IN ){
		fgl_setPixel( ctx, x0, y0, ctx->clr); 
		fgl_setPixel( ctx, x1, y0, ctx->clr); 
		fgl_setPixel( ctx, x0, y1, ctx->clr); 
		fgl_setPixel( ctx, x1, y1, ctx->clr); 
		fgl_setPixel( ctx, x2, y2, ctx->clr); 
		fgl_setPixel( ctx, x3, y2, ctx->clr); 
		fgl_setPixel( ctx, x2, y3, ctx->clr); 
		fgl_setPixel( ctx, x3, y3, ctx->clr); 
	}
	else {
		if ( inInterval( x0, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x0, y0, ctx->clr);
			}
			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x0, y1, ctx->clr);
			}
		}
		if ( inInterval( x1, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x1, y0, ctx->clr);
			}
			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x1, y1, ctx->clr);
			}
		}		
		if ( inInterval( x2, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y2, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x2, y2, ctx->clr);
			}
			if ( inInterval( y3, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x2, y3, ctx->clr);
			}
		}		
		if ( inInterval( x3, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y2, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x3, y2, ctx->clr);
			}
			if ( inInterval( y3, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x3, y3, ctx->clr);
			}
		}		
	}
}
/*}}}*/

FUNC_ATTR
void fgl_drawCircle ( fgl_context_t* ctx, i32 cx, i32 cy, u32 r ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_OVAL)
{
	i32 x0, y0, x1, y1;
	i32 x = 0, y = r;
	i32 a = 1 - r;
	u8  clip;

	x0 = cx - r;
	x1 = cx + r;
	y0 = cy - r;
	y1 = cy + r;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	fgl_setPixel( ctx, cx, cy+r, ctx->clr);
	fgl_setPixel( ctx, cx, cy-r, ctx->clr);
	fgl_setPixel( ctx, cx+r, cy, ctx->clr);
	fgl_setPixel( ctx, cx-r, cy, ctx->clr);

	while ( x < y ){
		x++;
		if ( a < 0 ){
			a += 2*x + 1;
		}
		else {
			y--;
			a += 2 * (x - y) + 1;
		}

		setCirclePoints( ctx, cx, cy, x, y, clip);
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
#endif /* !FGL_PROTO */
;/*}}}*/




static inline
void fillCirclePoints ( fgl_context_t* ctx,
						i32 cx, i32 cy, i32 dx, i32 dy, u8 clip ) /*{{{*/
{
	fgl_slc_t *p;

	i32 x0 = cx - dx;
	i32 y0 = cy - dy;
	i32 x1 = cx + dx;
	i32 y1 = cy + dy;
	i32 x2 = cx - dy;
	i32 y2 = cy - dx;
	i32 x3 = cx + dy;
	i32 y3 = cy + dx;
	i32 w1, w2;
	
	if ( clip == CR_PART ){
		x0 = (ctx->x0Clip > x0) ? ctx->x0Clip : x0;
		x2 = (ctx->x0Clip > x2) ? ctx->x0Clip : x2;
		x1 = (ctx->x1Clip < x1) ? ctx->x1Clip : x1;
		x3 = (ctx->x1Clip < x3) ? ctx->x1Clip : x3;

		w1 = x1 - x0 +1;
		w2 = x3 - x2 +1;

		if ( w1 > 0 ) {
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				p = fgl_getSlc( ctx->bmp, x0, y0);
				fgl_setSlcCols( p, x0, w1, ctx->clr);
			}

			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				p = fgl_getSlc( ctx->bmp, x0, y1);
				fgl_setSlcCols( p, x0, w1, ctx->clr);
			}
		}

		if ( w2 > 0 ) {
			if ( inInterval( y2, ctx->y0Clip, ctx->y1Clip) ){
				p = fgl_getSlc( ctx->bmp, x2, y2);
				fgl_setSlcCols(p, x2, w2, ctx->clr);
			}
			
			if ( inInterval( y3, ctx->y0Clip, ctx->y1Clip) ){
				p = fgl_getSlc( ctx->bmp, x2, y3);
				fgl_setSlcCols(p, x2, w2, ctx->clr);
			}
		}
	}
	else {
		p = fgl_getSlc( ctx->bmp, x0, y0);
		fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);

		p = fgl_getSlc( ctx->bmp, x0, y1);
		fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);

		p = fgl_getSlc( ctx->bmp, x2, y2);
		fgl_setSlcCols(p, x2, (x3-x2)+1, ctx->clr);

		p = fgl_getSlc( ctx->bmp, x2, y3);
		fgl_setSlcCols(p, x2, (x3-x2)+1, ctx->clr);
	}
}
/*}}}*/


FUNC_ATTR
void fgl_fillCircle ( fgl_context_t* ctx, i32 cx, i32 cy, u32 r ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_OVAL)
{
	i32 x0, y0, x1, y1;
	i32 x = 0, y = r;
	i32 a = 1 - r;
	u8  clip;

	x0 = cx - r;
	x1 = cx + r;
	y0 = cy - r;
	y1 = cy + r;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

#ifdef ROT_90	
	rotate32( ctx->bmp, (u32*)&cx, (u32*)&cy);
#endif

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );
	
	while ( x < y ){
		if ( a < 0 ){
			a += 2*x + 1;
		}
		else {
			y--;
			a += 2 * (x - y) + 1;
		}

		fillCirclePoints( ctx, cx, cy, x, y, clip);
		x++;
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
#endif /* !FGL_PROTO */
;/*}}}*/



static inline
void setEllipsePoints ( fgl_context_t* ctx, i32 cx, i32 cy,
					   i32 dx, i32 dy, i8 clip ) /*{{{*/
{

	u32 x0 = cx + dx;
	u32 y0 = cy + dy;
	u32 x1 = cx - dx;
	u32 y1 = cy - dy;

	if ( clip == CR_IN ){
		fgl_setPixel( ctx, x0, y0, ctx->clr);
		fgl_setPixel( ctx, x1, y0, ctx->clr);
		fgl_setPixel( ctx, x0, y1, ctx->clr);
		fgl_setPixel( ctx, x1, y1, ctx->clr);
	}
	else {
		if ( inInterval( x0, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x0, y0, ctx->clr);
			}
			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x0, y1, ctx->clr);
			}
		}
		if ( inInterval( x1, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x1, y0, ctx->clr);
			}
			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				fgl_setPixel( ctx, x1, y1, ctx->clr);
			}
		}
	}
}
/*}}}*/


FUNC_ATTR
void fgl_drawEllipse ( fgl_context_t* ctx, i32 cx, i32 cy, u32 rx, u32 ry ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_OVAL)
{
	i32 x0, y0, x1, y1;
	i32 x, y, a, ax, ay;
	i32 rxrx, ryry,  rxrx2, ryry2;
	i8  clip;

	if ( rx == ry ) {
		fgl_drawCircle( ctx, cx, cy, rx);
		return;
	}

	x0 = cx - rx;
	x1 = cx + rx;
	y0 = cy - ry;
	y1 = cy + ry;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	rxrx = rx*rx;
	ryry = ry*ry;
	rxrx2 = 2 * rxrx;
	ryry2 = 2 * ryry;

	x = 0;   ax = 0;
	y = ry;  ay = rxrx2 * y;
	a = ( (20*(ryry - (rxrx * (i32)ry)) + 5*rxrx) + 10) / 20;

	fgl_setPixel( ctx, cx, cy + ry, ctx->clr);
	fgl_setPixel( ctx, cx, cy - ry, ctx->clr);

	while ( ax < ay ){
		x++;
		ax += ryry2;
		if ( a < 0 ){
			a += ryry + ax;
		}
		else {
			y--;
			ay -= rxrx2;
			a += ryry + ax - ay;
		}

		setEllipsePoints( ctx, cx, cy, x, y, clip);
	}
	
	a = (ryry*(20*x*x +20*x + 5) + 20*(rxrx*(y-1)*(y-1) - rxrx*ryry) + 10) / 20;
	while ( y > 0 ){
		y--;
		ay -= rxrx2;
		if ( a > 0 ){
			a += rxrx - ay;
		}
		else {
			x++;
			ax += ryry2;
			a += rxrx - ay + ax;
		}

		setEllipsePoints( ctx, cx, cy, x, y, clip);
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
#endif /* !FGL_PROTO */
;/*}}}*/



static inline
void fillEllipsePoints ( fgl_context_t* ctx, i32 cx, i32 cy,
						 i32 dx, i32 dy, i8 clip ) /*{{{*/
{
	fgl_slc_t *p;

	i32 x0 = cx - dx;
	i32 y0 = cy - dy;
	i32 x1 = cx + dx;
	i32 y1 = cy + dy;

	if ( clip == CR_PART ){
		x0 = (ctx->x0Clip > x0) ? ctx->x0Clip : x0;
		x1 = (ctx->x1Clip < x1) ? ctx->x1Clip : x1;
		y0 = (ctx->y0Clip > y0) ? ctx->y0Clip : y0;
		y1 = (ctx->y1Clip < y1) ? ctx->y1Clip : y1;
		
		if ( x1 < x0 ){
			return;
		}

		if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
			p = fgl_getSlc( ctx->bmp, x0, y0);
			fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);
		}

		if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
			p = fgl_getSlc( ctx->bmp, x0, y1);
			fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);
		}
		
	}
	else {	
		p = fgl_getSlc( ctx->bmp, x0, y0);
		fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);

		p = fgl_getSlc( ctx->bmp, x0, y1);
		fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);
	}
}
/*}}}*/


FUNC_ATTR
void fgl_fillEllipse ( fgl_context_t* ctx, i32 cx, i32 cy, u32 rx, u32 ry ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_OVAL)
{
	i32 x0, y0, x1, y1;
	i32 x, y, a, ax, ay;
	i32 rxrx, ryry,  rxrx2, ryry2;
	i8  clip;

	if ( rx == ry ) {
		fgl_fillCircle( ctx, cx, cy, rx);
		return;
	}

#ifdef ROT_90	
	rotate32( ctx->bmp, (u32*)&cx, (u32*)&cy);
	swap32( (i32*)&rx, (i32*)&ry);
#endif

	x0 = cx - rx;
	x1 = cx + rx;
	y0 = cy - ry;
	y1 = cy + ry;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	rxrx = rx*rx;
	ryry = ry*ry;
	rxrx2 = 2 * rxrx;
	ryry2 = 2 * ryry;

	x = 0;   ax = 0;
	y = ry;  ay = rxrx2 * y;

	a = ( (20*(ryry - (rxrx * (i32)ry)) + 5*rxrx) + 10) / 20;
	while ( ax < ay ){
		ax += ryry2;
		if ( a < 0 ){
			a += ryry + ax;
		}
		else {
			y--;
			ay -= rxrx2;
			a += ryry + ax - ay;
		}

		fillEllipsePoints( ctx, cx, cy, x, y, clip);
		x++;
	}

	a = (ryry*(20*x*x +20*x + 5) + 20*(rxrx*(y-1)*(y-1) - rxrx*ryry) + 10) / 20;
	while ( y > 0 ){
		y--;
		ay -= rxrx2;
		if ( a > 0 ){
			a += rxrx - ay;
		}
		else {
			x++;
			ax += ryry2;
			a += rxrx - ay + ax;
		}

		fillEllipsePoints( ctx, cx, cy, x, y, clip);
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
#endif /* !FGL_PROTO */
;/*}}}*/


/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_OVAL_INCLUDED */
/*}}}*/
