/* -*- mode: C; mode: fold -*- */
/*
 * rect.c - fgl module to draw and fill rectangles
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: rect.c,v 1.16 2001/07/04 00:47:55 samc Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_RECT_INCLUDED
#define FGL_RECT_INCLUDED

#define FGL_MODULE
#include "fgl.h"

/* fgl imports */
#include "cntx.c"
#include FGL_DRAW_MODULE
#include "line.c"

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif


#ifdef FGL_INLINE_RECT
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/


FUNC_ATTR
void fgl_drawRect ( fgl_context_t *ctx,	i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_RECT)
{

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	fgl_drawHLine( ctx, x0+1, y0, x1-1);
	fgl_drawHLine( ctx, x0+1, y1, x1-1);
	fgl_drawVLine( ctx, x0, y0, y1);
	fgl_drawVLine( ctx, x1, y0, y1);
}
#endif /* !FGL_PROTO */
;/*}}}*/


FUNC_ATTR
void fgl_fillRectClr ( fgl_context_t *ctx, i32 x0, i32 y0,
					   i32 x1, i32 y1, fgl_color_t clr ) /*{{{*/
#if !defined(FGL_PROTO) || defined(FGL_INLINE_RECT)
{
	int w, h;
	fgl_slc_t *p;

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( (x1 < ctx->x0Clip) || (x0 > ctx->x1Clip) ||
		 (y1 < ctx->y0Clip) || (y0 > ctx->y1Clip) ){
		return;
	}

	if ( x0 < ctx->x0Clip ) x0 = ctx->x0Clip;
	if ( x1 > ctx->x1Clip ) x1 = ctx->x1Clip;
	if ( y0 < ctx->y0Clip ) y0 = ctx->y0Clip;
	if ( y1 > ctx->y1Clip ) y1 = ctx->y1Clip;

#ifdef ROT_90	
	rotate32( ctx->bmp, (u32*)&x0, (u32*)&y0);
	rotate32( ctx->bmp, (u32*)&x1, (u32*)&y1);
	swap32( &x0, &x1);
#endif

#ifdef FGL_ACCELFN_FILLRECT
	if (ctx->bmp->state & IMAGE_ON_SCREEN) {
	  fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_ACCEL_FILLRECT_OP );
	  fgl_accel_fillRect( fgl_screenCtx, x0, y0, x1, y1, clr );
	  fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
	  return ;
	}
#endif
	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	p = fgl_getSlc( ctx->bmp, x0, y0);
	w = x1 - x0+1;
	h = y1 - y0;

	do {
		fgl_setSlcCols( p, x0, w, clr);
		fgl_addSlc( &p, x0, ctx->bmp->sLength);
	} while ( h-- );

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
#endif /* !FGL_PROTO */
;/*}}}*/

static inline
void fgl_fillRect ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
{
	fgl_fillRectClr( ctx, x0, y0, x1, y1, ctx->clr);
}
/*}}}*/


/* fgl module prolog */ /*{{{*/
#undef FUNC_ATTR
#endif /* FGL_RECT_INCLUDED */
/*}}}*/
