/* -*- mode: C; mode: fold -*- */
/*
 * util.c - fgl module for general auxiliary functions (the dreaded misc)
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: util.c,v 1.14 2001/11/19 08:21:27 peter Exp $
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_UTIL_INCLUDED
#define FGL_UTIL_INCLUDED

#define FGL_MODULE
#include "fgl.h"

#include FGL_CMAP_MODULE

#if (__INCLUDE_LEVEL__ == 0)
#  undef FGL_PROTO
#endif

#ifdef FGL_INLINE_ABMP
#  define FUNC_ATTR static inline
#else
#  define FUNC_ATTR
#endif
/*}}}*/

void err (const char *fmt, ...) /*{{{*/
#if !defined(FGL_PROTO)
{
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "fgl: ");
    vfprintf(stderr, fmt, ap);
	fflush( stderr);

    va_end(ap);
}
#endif /* FGL_PROTO */
;/*}}}*/


void printU8( u8 a ) /*{{{*/
#if !defined(FGL_PROTO)
{
    u8 i;
    for ( i=0; i<8; i++ ){
        putchar( ( a & 0x80 ) ? 'O' : '.' );
        a <<= 1;
    }
}
#endif /* FGL_PROTO */
;/*}}}*/

static inline
u8 clipBlt ( fgl_context_t* dst, i32* u0, i32* v0, i32* u1, i32* v1,
				i32 sw, i32 sh, i32* x0, i32* y0, i32* x1, i32* y1 ) /*{{{*/
{
	if ( (*x0 >= sw) || (*y0 >= sh) ) return 0;
	if ( (*x1 < 0) || (*y1 < 0) ) return 0;

	if ( *x0 < 0 ){
		*u0 -= *x0;
		*x0 = 0;
	}
	if ( *y0 < 0 ){
		*v0 -= *y0;
		*y0 = 0;
	}
	if ( *x1 >= sw ) *x1 = sw-1;
	if ( *y1 >= sh ) *y1 = sh-1;

	if ( *u0 < dst->x0Clip ) {
		*x0 += (dst->x0Clip - *u0);
		*u0 = dst->x0Clip;
	}
	if ( *v0 < dst->y0Clip ) {
		*y0 += (dst->y0Clip - *v0);
		*v0 = dst->y0Clip;
	}

	*u1 = *u0 + (*x1 - *x0);
	*v1 = *v0 + (*y1 - *y0);

	if ( *u1 > dst->x1Clip ){
		*x1 -= (*u1 - dst->x1Clip);
		*u1 = dst->x1Clip;
	}
	if ( *v1 > dst->y1Clip ){
		*y1 -= (*v1 - dst->y1Clip);
		*v1 = dst->y1Clip;
	}

	return ((*u1 >= *u0) && (*v1 >= *v0));
}
/*}}}*/

static inline
u8 clipRel ( fgl_context_t* ctx, i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
{
	if ( (x0 > ctx->x1Clip) || (x1 < ctx->x0Clip) ||
		 (y0 > ctx->y1Clip) || (y1 < ctx->y0Clip) ){
		return CR_OUT;
	}

	if ( (x0 >= ctx->x0Clip) && (x1 <= ctx->x1Clip) &&
		 (y0 >= ctx->y0Clip) && (y1 <= ctx->y1Clip) ){
		return CR_IN;
	}

	return CR_PART;
}
/*}}}*/


static inline
u16 swapU16Bytes ( u16 a ) /*{{{*/
{
	u8 t[2];
	u8 *p = (u8*) &a;
	
	t[1] = *p++;
	t[0] = *p;
	
	return *((u16*)t);
}
/*}}}*/

static inline
u32 swapU32Bytes ( u32 a ) /*{{{*/
{
	u8 t[4];
	u8 *p = (u8*) &a;
	
	t[3] = *p++;
	t[2] = *p++;
	t[1] = *p++;
	t[0] = *p;

	return *((u32*)t);
}
/*}}}*/


/* fgl module epilog */ /*{{{*/
#endif /* FGL_UTIL_INCLUDED */
/*}}}*/
