#include <stdio.h>

#include "ipl.h"
#include "vk_names.h"

#ifdef DEBUG_VBI_FGL
  #include "fgl.h"

  extern fgl_context_t *debug_vbi_gc;
#endif

inline char* eventName ( int type )
{
	switch ( type ){
	case EVT_NONE:         return "NONE";
	case EVT_KEY_DOWN:     return "KEY_DOWN";
	case EVT_KEY_UP:       return "KEY_UP";
	case EVT_MOUSE_MOVE:   return "MOUSE_MOVE";
	case EVT_BUTTON_DOWN:  return "BUTTON_DOWN";
	case EVT_BUTTON_UP:    return "BUTTON_UP";
	case EVT_VT_REL_REQ:   return "VT_REL_REQ";
        case EVT_VT_ACQ_REQ:   return "VT_ACQ_REQ";
	}
	
	if ( (type >= EVT_GENERIC_FIRST) && (type <= EVT_GENERIC_LAST) ){
		return "GENERIC";
	}

	return "?";
}

inline char* modNames ( int mods )
{
	static char buf[5];

	buf[0] = (mods & SHIFT) ? 'S' : '-';
	buf[1] = (mods & CTRL)  ? 'C' : '-';
	buf[2] = (mods & ALT)   ? 'A' : '-';
	buf[3] = (mods & META)  ? 'M' : '-';
	buf[4] = 0;

	return buf;
}

inline char* buttonName ( int btn )
{
	static char buf[4];

	buf[0] = (btn & IPL_BTN_LEFT)    ? 'L' : '-';
	buf[1] = (btn & IPL_BTN_MIDDLE)  ? 'M' : '-';
	buf[2] = (btn & IPL_BTN_RIGHT)   ? 'R' : '-';
	buf[3] = 0;

	return buf;
}

int main ( int argc, char* argv[] )
{
	ipl_event_t e;
	struct timeval tv;
	long   sec1, sec2;

	gettimeofday( &tv, 0);
	sec1 = tv.tv_sec;
	if ( tv.tv_usec >= 500000 ) sec1++;
	
	if ( !ipl_init( 1024, 768) ){
		puts( "init failed");
		return 0;
	}  

#ifdef DEBUG_VBI_FGL
	if ( !fgl_init( 0, 0, 0 ) ) {
		puts( "could not init fgl" );
		return 1;
	}
	debug_vbi_gc = fgl_getContext( fgl_screenBmp );
#endif

	ipl_wakeUp();

	printf( "Press ESCAPE key to finish\n");

	while ( 1 ) {
	    e.type = 0;
		if ( ipl_getEvent( &e) ){
			if ( e.type < EVT_GENERIC_FIRST ){
				printf( "%-14s  pos: %d,%d, button: %s, char: %c (%d), code: %s (%2x), mods: %s\n",
					   eventName( (int)e.type),
					   e.x, e.y,
					   buttonName( e.buttons),
					   (e.keyChar > 31) ? e.keyChar : ' ',
					   e.keyChar,
					   (e.keyCode) ? vkNames[e.keyCode] : "",
					   e.keyCode,
					   modNames( (int)e.mods));
				if (e.keyCode == 27)
					break;
				#ifdef VT_SWITCH
					if (e.type == EVT_VT_REL_REQ)
						ipl_vtSwitchRelease();
					if (e.type == EVT_VT_ACQ_REQ)
						ipl_vtSwitchAcquire();
				#endif
			} else {
				gettimeofday( &tv, 0);
			    sec2 = tv.tv_sec;
				if ( tv.tv_usec >= 500000 ) sec2++;
				
				printf( "%ld: %s\n", (sec2 - sec1), e.data);
			}
		}
	}

	ipl_reset();

#ifdef DEBUG_VBI_FGL
	fgl_terminate();
#endif

}
