/*
 * btnkbd.c - generic button-based pseudo keyboard funcs
 *
 * our environment has to profide the following defines
 * 
 *   KBD_FD   : file descriptor for button device (e.g. btnFd)
 *   KBD_DEV	 : name of device node (e.g. "/dev/r39xxbuttons")
 *   KBD_IPL     : ipl device (e.g. r39xxButtons)
 * 
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 08/12/00 Peter Mehlitz
 */


static int btnkbd_init(ipl_device_t **dev)
{
	if ( KBD_FD == -1 ){
		if ( (KBD_FD = open(KBD_DEV, O_RDONLY)) == -1 ){
			fprintf(stderr, "ipl: Error opening %s\n", KBD_DEV);
			*dev = NULL;
			return 0;
		}
	}
	KBD_IPL.fd = KBD_FD;
	*dev = &KBD_IPL;
	return 1;
}

static
void btnkbd_reset (ipl_device_t *dev)
{
	if ( dev->fd != -1 ){
		close( dev->fd);
		dev->fd = -1;
	}
}


static
int btnkbd_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
	static int mod;
	static fd_set rs;
	static unsigned char b, b1, isDown, eatUp, eatUp1;
	static short kCharUp, kCodeUp;

	struct timeval to;
	short kChar = 0, kCode = 0;

	if ( dev->pending ) {
		b = b1;
		b1 = 0;
		dev->pending = 0;
	}
	else {
		if ( read( KBD_FD, &b, sizeof(b)) < sizeof(b) ){
			return 0;
		}
	}

	isDown = isKeyPressed( b);
	b = getKeyId( b);
	
	/* first, check if this might be a 'up' button we have to merge into a combo up */
	if ( !isDown && (eatUp | eatUp1) ){
		if ( b == eatUp ){
			eatUp = 0;
		}
		else if ( b == eatUp1 ){
			eatUp1 = 0;
		}
		if ( !(eatUp | eatUp1) ){
			/* Ok, merge it into (artificial) combo up */
			kChar = kCharUp;
			kCode = kCodeUp;

			kCharUp = kCodeUp = 0;
		}
		else {
			/* eat it up */
			return 0;
		}
	}
	else {
		/* check if this might be part of a multi-button "key" */
		if ( isDown && kCombo[b][0] ){
			FD_ZERO( &rs);
			FD_SET( KBD_FD, &rs);
			
			to.tv_sec = 0;
			to.tv_usec = COMBO_DELAY;
			
			if ( (select( KBD_FD+1, &rs, 0,0, &to) > 0) && FD_ISSET( KBD_FD, &rs) ){				
				read( KBD_FD, &b1, sizeof(b1));

				if ( (b1 & 0x80) != isDown ) {
					/* no, that was just a fast click, carry the b1, report b */
					dev->pending = 1;
				}
				else if ( (b1 & 0x7f) == kCombo[b][0] ){
					/* yes, that was a combo key, merge */
					b1 &= ~0x80;
					eatUp = b;
					eatUp1 = b1;

					kChar = kCharUp = kCombo[b1][1];
					kCode = kCodeUp = kCombo[b1][2];
				}
			}
		}
		else if ( kMod[b] ){
			/* this is a internal modifier, carry it */
			mod = isDown ? kMod[b] : 0;
			return 0;
		}
	}

	if ( !(kChar | kCode) ) {
		kChar = kChars[b][mod];
		kCode = kCodes[b][mod];
	}

	evt->x = 0;
	evt->y = 0;

	evt->mods    = 0;   /* nothing */
	evt->buttons = 0;   /* this refers to the toucpad - we are a keyboard, here */

	evt->type = (isDown) ? EVT_KEY_DOWN : EVT_KEY_UP;
	evt->keyChar = kChar;
	evt->keyCode = kCode;

	gettimeofday( &(evt->tv), 0);

	return 1;
}
