/*
 * gen_v4l_vbi.c - generic v4l vbi device (tested with bttv/v4l)
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/ioctl.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

//#include <linux/videodev.h>

#include "ipl.h"

// #define DEBUG_PARSER

#ifdef DEBUG_VBI_FGL
  #include "fgl.h"
  fgl_context_t *debug_vbi_gc;
  static int sample_counter = 0;  // Only show graphically a subset
                                  // of the samples
#endif

static int vbiFd = -1;
static int atvef_on = 0;
static int dIdx = 0;
static unsigned char data[256];

ipl_device_t vbiv4lDevice;

static int v4l_vbi_init(ipl_device_t **dev)
{
  char  *vbidev = getenv( "VBI");
  //struct v4l2_format fmt;	

  if ( vbiFd != -1 ) {
    *dev = &vbiv4lDevice;
    return 1;
  }

  if ( !vbidev ){
    vbidev = "/dev/vbi";
  }

  if ( (vbiFd = open( vbidev, O_RDONLY)) == -1 ) {
    fprintf(stderr, "ipl: Error opening VBI device: %s", vbidev);
    *dev = NULL;
    return 0;
  }
  vbiv4lDevice.fd = vbiFd;	
  *dev = &vbiv4lDevice;
  return 1;
}

static
void v4l_vbi_reset (ipl_device_t *dev)
{
  if ( vbiFd != -1 ){
    close( vbiFd);
  }
}

static int parityok( int n )
{
	int mask=0;
	int j, k;

	for (k = 1, j = 0; j < 7; j++)
		if (n & (1<<j)) k++;
	if ((k & 1) == ((n>>7)&1))
		mask|=0x00FF;
	for (k = 1, j = 8; j < 15; j++)
		if (n & (1<<j)) k++;
	if ((k & 1) == ((n>>15)&1))
		mask|=0xFF00;
	return mask;
}

static int decodebit( unsigned char *data,
		      int threshold )
{
	int i, sum = 0;

	for (i = 0; i < 23; i++)
		sum += data[i];
	return (sum > threshold*23);
}


static void decode( unsigned char *vbi_data,
		    unsigned char *vbibyte1,
		    unsigned char *vbibyte2 )
{
	// VBI signal starts on line 21
	int offset = 2048 * 11;

	int clkmax[7], clkmin[7], clkval[7], clk;
	int i, tmp, sample, threshold;
	int packedbits = 0;

#ifdef DEBUG_VBI_FGL
	int x;

	sample_counter = (sample_counter++) % 100;
	if (sample_counter == 1) {
		fgl_setColor( debug_vbi_gc, fgl_clrBlack );
		fgl_fillRect( debug_vbi_gc, 0, 0, 1024, 85 );
//		fgl_setColor( debug_vbi_gc, fgl_clrRed );
//		fgl_drawLine( debug_vbi_gc, 0, 85-85/2, 1024, 85-85/2 );
		fgl_setColor( debug_vbi_gc, fgl_clrWhite );
		for (x=0; x<1024; x++) {
			if (vbi_data[offset + x*2] / 2 < 85 &&
			    vbi_data[offset + x*2 + 2] / 2 < 85) {
		        fgl_drawLine( debug_vbi_gc,
		           x,   85 - vbi_data[offset + x*2] / 2,
		           x+1, 85 - vbi_data[offset + x*2 + 2] / 2);
			}
		}
	}
#endif

       	*vbibyte1 = 0;
	*vbibyte2 = 0;

	// Find the 7 clock signals
	for (clk = 0; clk < 7; clk++)
		clkmax[clk] = clkmin[clk] = clkval[clk] = -1;
	clk = 0;
	tmp = 0;
	for (i = 30; i < 600 && clk < 7; i++) {
		sample = vbi_data[offset + i];
		if (clkmax[clk] < 0) {
			// Find max before drop
			if (sample > 85 && sample > clkval[clk]) {
				// Find max
				clkval[clk] = sample;
				tmp = i;
			}
			else if (clkval[clk] - sample > 30) {
				// Falling edge detected
				clkmax[clk] = tmp;
				i = tmp + 10;
			}
		} else {
			// Find min after drop
			if (sample < 85 && sample < clkval[clk]) {
				// Find min
				clkval[clk] = sample;
				tmp = i;
			}
			else if (sample - clkval[clk] > 30) {
				// Rising edge detected
				clkmin[clk] = tmp;
				i = tmp + 10;
#ifdef DEBUG_VBI_FGL
	if (sample_counter == 1) {
		fgl_setColor( debug_vbi_gc, fgl_clrGreen );
		fgl_drawLine( debug_vbi_gc, clkmax[clk] / 2, 0,
			                 clkmax[clk] / 2, 85 );
		if (clk < 6) {
			fgl_setColor( debug_vbi_gc, fgl_clrBlue );
			fgl_drawLine( debug_vbi_gc, clkmin[clk] / 2, 0,
				      clkmin[clk] / 2, 85 );
		}
	}
#endif 
				clk++;  // Do next clock
			}
		}
	}

	if (clk != 7 || vbi_data[ offset + clkmax[3] ] -
	    vbi_data[ offset + clkmin[5] ] < 45) {
                // Failed to find clock signal 
		return;   
	}
 
	// Calculate threshold by doing weighted average of clocks
	for (clk = 0, threshold = 85; clk < 7; clk++) {
		threshold = (threshold +
			  vbi_data[ offset + clkmin[clk] ] +
			  vbi_data[ offset + clkmax[clk] ]) / 3;
	}
#ifdef DEBUG_VBI_FGL
	if (sample_counter == 1) {
		fgl_setColor( debug_vbi_gc, fgl_clrYellow );
		fgl_drawLine( debug_vbi_gc, 0, 85 - threshold / 2,
			                 1024, 85 - threshold / 2 );
	}
#endif 

	// Find first rising edge after clock signal
	for (i=clkmin[6]; vbi_data[ offset + i] < threshold; i++);

#ifdef DEBUG_VBI_FGL
	// Draw bit edges
	if (sample_counter == 1) {
		fgl_setColor( debug_vbi_gc, fgl_clrRed );
		for (x=i; x < i+57*18; x+=57) {
			fgl_drawLine( debug_vbi_gc,
				      x/2, 0,
			              x/2, 85 );
		}
	}
#endif 

	// Sense bits
	tmp = i + 57;
	for (i = 0; i < 16; i++) {
		if (decodebit( &vbi_data[ offset + tmp + i * 57 ],
			       threshold)) {
			packedbits |= 1<<i;
		}
	}

	if (parityok(packedbits)) {
		*vbibyte1 = packedbits & 0x7f;
		*vbibyte2 = (packedbits >> 8) & 0x7f;
	}
	return;
}

      
static
unsigned short checkSum( unsigned short* val, int byte_count) {
  unsigned long sum = 0;
 
  while ( byte_count > 1 ) {
    sum += *(val++);
    byte_count -= 2;
  }
 
  if ( byte_count > 0 ) {
    sum += *(unsigned char*)val;
  }
 
  while ( sum >> 16 ) {
    sum = (sum & 0xFFFF) + (sum >> 16);
  }
 
  return ((unsigned short) ~sum);
}       

static
int isReady() {
  //check for trailing checksum [xxxx]
  int idx;
 
  if ( (dIdx < 5) || (data[dIdx-4] != '[')) {
    return 0;
  }
  for ( idx = 0; idx < 4; idx++ ) {
    if ( ! isxdigit( data[dIdx-idx] ) ) {
      return 0;
    }
  }

  return 1;
}

static
int checkReady( ipl_event_t* evt) {
  //check for trailing checksum [xxxx]
  char csb[5];
  unsigned short icheck, check;

  //swap checksum byte order
  csb[0] = data[dIdx-1];
  csb[1] = data[dIdx-0];
  csb[2] = data[dIdx-3];
  csb[3] = data[dIdx-2];
  csb[4] = 0;
 
  icheck = (unsigned short)strtol( csb, (char**)0, 16);
  check = checkSum( (unsigned short*)data, dIdx-4) ;
 
  printf( "trigger checksum: %s(%04X/%04X)\n", (check == icheck) ? "passed" : "failed", check, icheck);
   if ( icheck == check ) {
 
    data[++dIdx] = ']';                    
    data[++dIdx] = 0;                    

    evt->x = evt->y = 0;
    evt->buttons = 0;
    evt->mods = 0;
    evt->keyChar = evt->keyCode = 0;
 
    gettimeofday( &(evt->tv), 0);
    evt->type = EVT_GENERIC_FIRST;
    evt->data = data;
    evt->dlen = dIdx;                                         
    
    dIdx = 0;
 
    printf( "trigger ready: %s\n", data);
 
    return 1;                                                                           
  }
  else {
    //bad checksum, start with next trigger
    printf( "trigger was: %s\n", data );
    dIdx = 0;
    return 0;
  }
 
}  

static
int v4l_vbi_getEvent ( ipl_device_t *dev, ipl_event_t* evt ){
   int  n, bIdx;
   static unsigned char vbi_data[65536];
   unsigned char vbibyte[2];

   n = read( vbiFd, vbi_data, 65536);
   if (n != 65536)
	   perror("ipl: error reading v4l VBI data");
   else {
	   decode( vbi_data, &vbibyte[0], &vbibyte[1] );
   }

   if ((vbibyte[0] != 0x00) && (vbibyte[1] != 0x00)) {
      
#if defined(DEBUG_VBI_FGL) || defined(DEBUG_PARSER)
      printf("%02x %02x %c %c\n", vbibyte[0], vbibyte[1],
	     (vbibyte[0] >= ' ' ? vbibyte[0] : ' '),
	     (vbibyte[1] >= ' ' ? vbibyte[1] : ' '));
#endif
      if( (vbibyte[0] == 0x1c) && (vbibyte[1] == 0x2b) ) {
	 atvef_on = 1;
      }
      else if( (vbibyte[0] == 0x1c) && (vbibyte[1] == 0x2a) ) {
	 atvef_on = 1;
      }
      else if( (vbibyte[0] == 0x1c) && (vbibyte[1] == 0x2d) ) {
      }
      else if( (vbibyte[0] >= 0x10) && (vbibyte[0] <= 0x1f) ) {
	 atvef_on = 0;
      }
      else if( atvef_on ) {

         unsigned char b0 = vbibyte[0];
         unsigned char b1 = vbibyte[1];
    
         if ( dIdx == 0 ) {
            if ( (b0 != '<') && (b1 != '<') ) {
	       return 0;
	    }
	    if ( b1 == '<' ) {
	       data[dIdx++] = '<';
	       return 0;
	    }
	 }
	 for ( bIdx = 0; bIdx < 2; bIdx++ ) {
	    if ( vbibyte[bIdx] != 0x00 ) {
	       data[dIdx] = vbibyte[bIdx];
	       data[dIdx+1] = 0;
	       if (dIdx >= 10) {
		  if ( data[dIdx-4] == '[' ) {
		     if ( isReady() ) {
			return checkReady( evt);
		     }
		  }   
	       }
	       dIdx++;

	       if (dIdx > 200) {
		  dIdx = 0;
		  break;
	       }
	    }
	 }
      } 
   }
      
   return 0;
}

ipl_driver_t vbiv4lGen = { v4l_vbi_init, v4l_vbi_getEvent, v4l_vbi_reset };  
ipl_device_t vbiv4lDevice = { 10, -1, 0, NULL, &vbiv4lGen, 0 };

