/*
 * hotplug.c - This file detects when a device is added or removed. 
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <sys/ioctl.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "ipl.h"

// #define DEBUG_PARSER

static int plugFd = -1;
static int atvef_on = 0;
static int dIdx = 0;
static unsigned char data[256];

ipl_device_t plugDevice;

static int hotplug_init(ipl_device_t **dev)
{
  if (plugFd != -1 ) {
    *dev = &plugDevice;
    return 1;
  }

  if ((plugFd = open("/proc/bus/usb/devices", O_RDONLY)) == -1 ) {
    fprintf("ipl: Error opening /proc/bus/usb/devices\n");
    *dev = NULL;
    return 0;
  }
  plugDevice.fd = plugFd;	
  *dev = &plugDevice;
  return 1;
}

static
void hotplug_reset (ipl_device_t *dev)
{
  if (plugFd != -1 ){
    close(plugFd);
  }
}

static
int hotplug_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
   int  n, bIdx;
   static unsigned char usb_data[65536];

   n = read(dev->fd, usb_data, 65536);
   if (n != 65536)
	   perror("ipl: error reading data");
   }
   return 0;
}

ipl_driver_t plugDetector = { hotplug_init, hotplug_getEvent, hotplug_reset };  
ipl_device_t plugDevice = { 10, -1, 0, NULL, &plugDetector, 0 };
