/*
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * X-windows emulation mode.
 *
 * $Id: input_fbx.c,v 1.6 2001/09/22 18:35:32 tim Exp $
 */


#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <arpa/inet.h>

#include "ipl.h"
#include "keysyms.h"

#define IPL_UNIX_SOCKET "/tmp/ipl_socket"

static int mouFd = -1;
static int x, y;
static int xmax, ymax;
static int threshold = 4, accel = 2;

ipl_driver_t fbxInput;
ipl_device_t fbxDevice;

#define ACC_TIME 0
#define TRASH 100

#define ABS(_a) (((_a) > 0) ? (_a) : -(_a))

static int fbx_init(ipl_device_t **dev)
{
	char* env;
	struct sockaddr_in serveraddr;
	
	if ( mouFd != -1 ) {
	    *dev = &fbxDevice;
	    return 1;
	}

	/* try to connect to emulation server */
	mouFd = socket( AF_INET, SOCK_STREAM, 0);

	bzero( &serveraddr, sizeof(serveraddr) );
	serveraddr.sin_family = AF_INET;
	serveraddr.sin_port = htons(6034);
	inet_pton( AF_INET, "127.0.0.1", &serveraddr.sin_addr);

	if (connect( mouFd, &serveraddr, sizeof(serveraddr))){
		perror("ipl error");
		*dev = NULL;
		return 0;
	}

	ipl_xPointer = x = ipl_width / 2;
	ipl_yPointer = y = ipl_height / 2;

	xmax = ipl_width-1;
	ymax = ipl_height-1;

	if ( (env = getenv("IPL_MOUSE")) ){
		sscanf( env, "%d %d", &threshold, &accel);
	}
	fbxDevice.fd = mouFd;
	*dev = &fbxDevice;
	return 1;
}

static
void fbx_reset (ipl_device_t *dev)
{
	if ( mouFd != -1 ){
		close( mouFd);
	}
}

static
int fbx_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{

  static unsigned char b[5];  
  int i, n;
  
  while ( 1 ) {
    /* logically, PS/2 mouse packets come in 3 bytes, but lots of drivers
     * like to deliver them in pieces
     */
    for ( i=0; i< 5; ){
      /* make sure we never block if we get to this point */
      if ( (n = read( mouFd, &(b[i]), 5 - i)) <= 0 ){
	if ( i == 0 )
	  return 0;
      }
      else {
	i += n;
      }
    }
    
    //    printf( "%d %d %d %d %d\n", b[0], b[1], b[2], b[3], b[4]); 
    gettimeofday( &(evt->tv), 0);
    
    if ( b[0] != 0 ) {
	    x = (int)b[1] + ((int)b[2] << 8);
	    y = (int)b[3] + ((int)b[4] << 8);

	    ROTATE(evt->x, evt->y, x, y);

	    evt->buttons = 0;
	    if (b[0] & 1) evt->type = EVT_BUTTON_DOWN;
	    if (b[0] & 2) evt->type = EVT_BUTTON_UP;
	    if (b[0] & 4) evt->type = EVT_MOUSE_MOVE;
	    if (b[0] & 16) evt->buttons |= IPL_BTN_LEFT;
	    if (b[0] & 32) evt->buttons |= IPL_BTN_RIGHT;
	    if (b[0] & 64) evt->buttons |= IPL_BTN_MIDDLE;
	    evt->mods = 0;
	    evt->keyChar = evt->keyCode = 0;
	    
	    ipl_xPointer = evt->x;
	    ipl_yPointer = evt->y;

	    return 1;
    } else {
	    evt->x = evt->y = evt->buttons = 0;
	    evt->mods = (b[1] >> 4);
	    if (( b[1] & 4 ) || ( b[3] == 0 )) {
		    evt->keyCode = FKeyCode[b[2]];
		    if ( b[3] > 0 ) {
			    evt->keyChar = (unsigned char) b[4];
		    } else {
			    switch ( evt->keyCode ){
			    case 0x08: 
			    case 0x09: 
			    case 0x0a: 
			    case 0x1b: 
			    case 0x20: 
			    case 0x7f:
				    evt->keyChar = evt->keyCode;
				    break;
			    default:
				    evt->keyChar = 0;
			    }
		    }
	    } else {
		    evt->keyChar = (unsigned char) b[4];
		    evt->keyCode = LKeyCode[b[2]];
	    }
	    if ( b[1] & 1 ) evt->type = EVT_KEY_DOWN;
	    if ( b[1] & 2 ) evt->type = EVT_KEY_UP;	    
	    return 1;
    }
  }
}
ipl_driver_t fbxInput = { fbx_init, fbx_getEvent, fbx_reset };
ipl_device_t fbxDevice = { 10, -1, 0, NULL, &fbxInput, 0 };
