/*
 * ipl.h - common (public) IPL header
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#ifndef IPL_H_INCLUDED
#define IPL_H_INCLUDED

#include "ipl_config.h"

#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>

#ifndef IPL_FUNC_ATTR
#  define IPL_FUNC_ATTR
#endif

/* make TS_UNDEF accesible globally */

#ifndef TS_UNDEF
#  define TS_UNDEF -1
#endif

/* default place to store touch panal configuration  */
/* can be overwridden with TP_CONFIG environment var */
#define DEFAULT_TP_CONFIG "/etc/tpconfig"

/* these values correspond to the Java AWT */

typedef enum {
	IPL_BTN_LEFT         = 16,
	IPL_BTN_MIDDLE       = 8,
	IPL_BTN_RIGHT        = 4,
} ipl_button_t;

typedef enum {
	EVT_NONE         = 0,
	EVT_MOUSE_MOVE   = 503,
	EVT_BUTTON_DOWN  = 501,
	EVT_BUTTON_UP    = 502,
	EVT_KEY_DOWN     = 401,
	EVT_KEY_UP       = 402,
	EVT_VT_REL_REQ   = 601,
        EVT_VT_ACQ_REQ   = 602,
	EVT_GENERIC_FIRST  = 10001,
	EVT_GENERIC_LAST   = 10010
} ipl_eventType_t;

typedef enum {
    SHIFT = 1,
    CTRL  = 2,
    META  = 4,
    ALT   = 8
} ipl_mod_t;


typedef struct _ipl_event_t {
	ipl_eventType_t    type;
	short              x, y;
	ipl_button_t       buttons;
	
	unsigned short     keyChar;
	unsigned short     keyCode;

	int                mods;

	unsigned char      *data;   /* for generic events (should be turned into a */
	int                dlen;    /* XEvent like 'hierarchy' */

	struct timeval     tv;
	struct _ipl_event_t  *next;
} ipl_event_t;

struct _ipl_device_t;

typedef struct _ipl_driver_t {
       	int    (*initDevice) (struct _ipl_device_t **dev);
        int    (*getEvent) (struct _ipl_device_t *dev, ipl_event_t *evt);
        void   (*resetDevice)(struct _ipl_device_t *dev);
} ipl_driver_t;

typedef struct _ipl_device_t {
        int                prio;
        int                fd;
        int                pending;
        unsigned char      *data;   /* Device specific data */
        struct _ipl_driver_t *driver;
        struct _ipl_device_t *next;
} ipl_device_t;

typedef struct {
       int xoff;
       int xscale;
       int yoff;
       int yscale;
       int rate;
       int swap;
       int pressure;
} ts_config_t;

/*
 * these are our helpers for the driver modules
 */
void _ipl_flush ( int fd );

/*
 * this is our public API
 */
int   ipl_init ( int width, int height );
int   ipl_getEvent ( ipl_event_t* evt );
int   ipl_queueEvent ( ipl_event_t* evt );
int   ipl_wakeUp ();
void  ipl_reset ();
void  ipl_vtSwitchRelease ();
void  ipl_vtSwitchAcquire ();

void  ipl_setScreenRotation (char rot);

#ifdef IPL_MAIN
#  define IPL_GLOBAL
#  define IPL_INIT(_a) = _a
#else
#  define IPL_GLOBAL extern
#  define IPL_INIT(_a)
#endif

IPL_GLOBAL ipl_event_t   ipl_curEvent;
IPL_GLOBAL int           ipl_xoff,  ipl_yoff;
IPL_GLOBAL int           ipl_width, ipl_height;
IPL_GLOBAL int           ipl_xPointer, ipl_yPointer;

IPL_GLOBAL ipl_device_t *ipl_devices;
IPL_GLOBAL ts_config_t   ipl_cfg;

IPL_GLOBAL int		 rotation;

/* Prototypes */
int vtSwitchInit(void);

/* Rotation */
#define	ROTATE(XOUT,YOUT,XIN,YIN) \
	switch (rotation) { \
	default: \
	case 0: \
		(XOUT) = (XIN); \
		(YOUT) = (YIN); \
		break; \
	case 1: \
		(XOUT) = (YIN); \
		(YOUT) = ipl_width - (XIN); \
		break; \
	case 2: \
		(XOUT) = ipl_width - (XIN); \
		(YOUT) = ipl_height - (YIN); \
		break; \
	case 3: \
		(XOUT) = ipl_height - (YIN); \
		(YOUT) = (XIN); \
		break; \
	}

/* this defines how we access devices */
#include IPL_IO_INTERFACE

#endif /* IPL_H_INCLUDED */
