/*
 * btn_r39xx.c - r39xx button pseudo-keyboard device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 00/00/00 Peter Mehlitz
 */


#include "ipl.h"
#include "vk_constants.h"

#include <fcntl.h>

static int btnFd = -1;


/*************************************************************************************
 * this defines our button->key mapping (should eventually be in a separate header)
 */

#define MAX_MOD 2
#define MAX_BTN 10   /* number of buttons plus 1 (there's no button '0') */

static short kChars [MAX_BTN] [MAX_MOD] = {
	/* kChar[mod1] kChar[mod2] */
	{ 0, 0 },         /* 0: no button */
	{ 0, 0 },         /* 1: bottom left outer */
	{ 0, 0 },         /* 2: bottom left inner */
	{ 0, 0 },         /* 3: bottom right inner */
	{ 0, 0 },         /* 4: bottom right outer */
	{ 0, 0 },         /* 5: upper left top */
	{ 0, 0 },         /* 6: upper left bottom */
	{ 0, 0 },         /* 7: top back left */
	{ 0, 0 },         /* 8: top back middle */
	{ 0, 0 }          /* 9: top back right */
}; 

static short kCodes [MAX_BTN] [MAX_MOD] = {
	/* kCodde[mod1] kCode[mod2] */
	{ 0, 0 },         /* 0: no button */
	{ 0, 0 },         /* 1: bottom left outer */
	{ 0, 0 },         /* 2: bottom left inner */
	{ VK_ESCAPE, VK_ESCAPE }, /* 3: bottom right inner */
	{ 0, 0 },         /* 4: bottom right outer */
	{ VK_UP, VK_RIGHT },    /* 5: upper left top */
	{ VK_DOWN, VK_LEFT }, /* 6: upper left bottom */
	{ 0, 0 },         /* 7: top back left */
	{ 0, 0 },         /* 8: top back middle */
	{ 0, 0 }          /* 9: top back right */
};

static char kMod [MAX_BTN] = {
	/* MOD index */
	0,         /* 0: no button */
	0,         /* 1: bottom left outer */
	0,         /* 2: bottom left inner */
	0,         /* 3: bottom right inner */
	1,         /* 4: bottom right outer */
	0,         /* 5: upper left top */
	0,         /* 6: upper left bottom */
	0,         /* 7: top back left */
	0,         /* 8: top back middle */
	0          /* 9: top back right */
};

static char kCombo [MAX_BTN] [3] = {
	/* combo button, kChar, kCode */
	{ 0, 0, 0 },        /* 0: no button */
	{ 0, 0, 0 },        /* 1: bottom left outer */
	{ 0, 0, 0 },        /* 2: bottom left inner */
	{ 0, 0, 0 },        /* 3: bottom right inner */
	{ 0, 0, 0 },        /* 4: bottom right outer*/
	{ 6, '\n', VK_ENTER },         /* 5: upper left bottom */
	{ 5, '\n', VK_ENTER },         /* 6: upper left top */
	{ 0, 0, 0 },        /* 7: top back left */
	{ 0, 0, 0 },        /* 8: top back middle */
	{ 0, 0, 0 }         /* 9: top back right */
};

static int btnkbd_init(ipl_device_t **dev);
static void btnkbd_reset ( ipl_device_t *dev);
static int btnkbd_getEvent ( ipl_device_t *dev, ipl_event_t* evt );

ipl_driver_t r39xxButtons = { btnkbd_init, btnkbd_getEvent, btnkbd_reset };
ipl_device_t r39xxDevice = { 10, -1, 0, NULL, &r39xxButtons, 0 };

/* these defines are used by the generic button functions */
#define KBD_FD   btnFd
#define	KBD_DEV	 "/dev/r39xxbuttons"
#define KBD_IPL  r39xxDevice
#define COMBO_DELAY 200000L  /* in micros */

static inline int isKeyPressed ( unsigned char b )
{
	return ((b & 0x80) != 0);
}

static inline int getKeyId ( unsigned char b )
{
	return (b & ~0x80);
}


#include "btnkbd.c" /* this includes the generic button kbd funcs */
