/*
 * btn_sa11xx.c - iPAQ button pseudo-keyboard device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 08/12/00 Peter Mehlitz
 */


#include "ipl.h"
#include "vk_constants.h"

#include <fcntl.h>

static int btnFd = -1;


/*************************************************************************************
 * this defines our button->key mapping (should eventually be in a separate header)
 */

#define MAX_MOD 2
#define MAX_BTN 10   /* number of buttons plus 1 (there's no button '0') */

static short kChars [MAX_BTN] [MAX_MOD] = {
	/* kChar[mod1] kChar[mod2] */
	{ 0, 0 },                 /* 0: no button */
	{ 0, 0 },                 /* 1: left upper */
	{ 0, 0 },                 /* 2: bottom left outer */
	{ 0, 0 },                 /* 3: bottom left inner */
	{ 0, 0 },                 /* 4: bottom right inner */
	{ 0, 0 },                 /* 5: bottom right outer */
	{ 0, 0 },                 /* 6: bottom center up */
	{ 0, 0 },                 /* 7: bottom center right */
	{ 0, 0 },                 /* 8: bottom center left */
	{ 0, 0 }                  /* 9: bottom center down */
}; 

static short kCodes [MAX_BTN] [MAX_MOD] = {
	/* kCodde[mod1] kCode[mod2] */
	{ 0, 0 },                 /* 0: no button */
	{ VK_ENTER, 0 },          /* 1: left upper */
	{ VK_F1, 0 },             /* 2: bottom left outer */
	{ VK_F2, 0 },             /* 3: bottom left inner */
	{ 0, 0 },                 /* 4: bottom right inner */
	{ VK_ESCAPE, 0 },         /* 5: bottom right outer */
	{ VK_UP, 0 },             /* 6: bottom center up */
	{ VK_RIGHT, 0 },          /* 7: bottom center right */
	{ VK_LEFT, 0 },           /* 8: bottom center left */
	{ VK_DOWN, 0 }            /* 9: bottom center down */
};

/*
 * the current iPAQ button driver (kernel) doesn't report nested button events
 */
static char kMod [MAX_BTN] = {
	/* MOD index */
	0,                        /* 0: no button */
	0,                        /* 1: left upper */
	0,                        /* 2: bottom left outer */
	0,                        /* 3: bottom left inner */
	0,                        /* 4: bottom right inner */
	0,                        /* 5: bottom right outer */
	0,                        /* 6: bottom center up */
	0,                        /* 7: bottom center right */
	0,                        /* 8: bottom center left */
	0                         /* 9: bottom center down */
};

static char kCombo [MAX_BTN] [3] = {
	/* combo button, kChar, kCode */
	{ 0, 0, 0 },              /* 0: no button */
	{ 0, 0, 0 },              /* 1: bottom left outer */
	{ 0, 0, 0 },              /* 2: bottom left inner */
	{ 0, 0, 0 },              /* 3: bottom right inner */
	{ 0, 0, 0 },              /* 4: bottom right outer */
	{ 0, 0, 0 },              /* 5: upper left top */
	{ 0, 0, 0 },              /* 6: upper left bottom */
	{ 0, 0, 0 },              /* 7: top back left */
	{ 0, 0, 0 },              /* 8: top back middle */
	{ 0, 0, 0 }               /* 9: top back right */
};

static int btnkbd_init(ipl_device_t **dev);
static void btnkbd_reset ( ipl_device_t *dev );
static int btnkbd_getEvent ( ipl_device_t *dev, ipl_event_t* evt );

ipl_driver_t sa11xButtons = { btnkbd_init, btnkbd_getEvent, btnkbd_reset };
ipl_device_t sa11xDevice = { 10, -1, 0, NULL, &sa11xButtons, 0 };

/* these defines are used by the generic button functions */
#define KBD_FD   btnFd
#define	KBD_DEV	 "/dev/h3600_key"
#define KBD_IPL  sa11xDevice
#define COMBO_DELAY 200000L  /* in micros */

static inline int isKeyPressed ( unsigned char b )
{
	return ((b & 0x80) == 0);
}

static inline int getKeyId ( unsigned char b )
{
	return (b & ~0x80);
}

#include "btnkbd.c" /* this includes the generic button kbd funcs */

