/*
 * mou_gpm.c - gpm based mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "ipl.h"

#include <stdlib.h>
#include <gpm.h>

static Gpm_Connect conn;
static int x, y;
static int xmax, ymax;
static int threshold = 3, accel = 4;

#define ABS(_a) (((_a) > 0) ? (_a) : -(_a))

extern ipl_device_t gpmDevice;

static int gpm_init(ipl_device_t **dev)
{
	char* env;

	gpm_zerobased    = 0;
	conn.eventMask   = ~0;  /* get all events */
	conn.defaultMask = ~GPM_HARD;
	conn.minMod      = 0;
	conn.maxMod      = ~0;

	if ( Gpm_Open( &conn, 0) == -1 ) {
		fprintf(stderr, "ipl: error opening gpm, check permissions on tty devices\n");
		*dev = NULL;
		return 0;
	}

	ipl_xPointer = x = ipl_width / 2;
	ipl_yPointer = y = ipl_height / 2;
	
	xmax = ipl_width-1;
	ymax = ipl_height-1;

	if ( (env = getenv("IPL_MOUSE")) ){
		sscanf( env, "%d %d", &threshold, &accel);
	}

	gpmDevice.fd = gpm_fd;

	*dev = &gpmDevice;
	return 1;
}

static
void gpm_reset ()
{
	if ( gpm_fd != -1 ){
		Gpm_Close();
	}
}

static
int gpm_getEvent(ipl_device_t *dev, ipl_event_t* evt )
{
	Gpm_Event gevt;

	if ( Gpm_GetEvent( &gevt) > 0 ) {
		if ( ABS( gevt.dx) >= threshold ){
			gevt.dx *= accel;
		}
		if ( ABS( gevt.dy) >= threshold ){
			gevt.dy *= accel;
		}

		x += gevt.dx;
		if ( x > xmax )			x = xmax;
		else if ( x < 0 )		x = 0;

		y += gevt.dy;
		if ( y > ymax )			y = ymax;
		else if ( y < 0 )		y = 0;
	
		if ( gevt.type & (GPM_MOVE | GPM_DRAG) ){
			evt->type = EVT_MOUSE_MOVE;
		}
		else if ( gevt.type & GPM_DOWN ){
			evt->type = EVT_BUTTON_DOWN;
		}
		else if ( gevt.type &  GPM_UP ){
			evt->type = EVT_BUTTON_UP;
		}
		else {
			return 0;
		}

		evt->buttons = 0;
		if ( gevt.buttons & GPM_B_LEFT )   evt->buttons |= IPL_BTN_LEFT;
		if ( gevt.buttons & GPM_B_MIDDLE ) evt->buttons |= IPL_BTN_MIDDLE;
		if ( gevt.buttons & GPM_B_RIGHT )  evt->buttons |= IPL_BTN_RIGHT;

		evt->x = x;
		evt->y = y;

		evt->mods = gevt.modifiers;
		evt->keyChar = evt->keyCode = 0;
		
		gettimeofday( &(evt->tv), 0);

		ipl_xPointer = evt->x;
		ipl_yPointer = evt->y;
		
		return 1;
	}

	return 0;
}

ipl_driver_t gpmMouse = { gpm_init, gpm_getEvent, gpm_reset };
ipl_device_t gpmDevice = { 10, -1, 0, NULL, &gpmMouse, 0 };
