/*
 * mou_r39xx.c - r39xx touchscreen pseudo-mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 00/00/00 Peter Mehlitz
 */


#include "ipl.h"

#include <fcntl.h>

#include "tpconfig.c"          /* includes the generic config file reader */


/* this is the data we get from the device driver */
typedef struct {
	unsigned short state;    /* 0:up, 1:down */ 
	unsigned short u;
	unsigned short v;     /* touchpad coords, need to be adjusted to screen coords */
} tp_packet_t;


static int mouFd = -1;
static tp_packet_t buf;

#define	TPDEV	"/dev/r39xxtpanel"

ipl_device_t r39xxDevice; 

static int r39xx_init(ipl_device_t **dev)
{
	if ( mouFd == -1 ){
		if ( (mouFd = open(TPDEV, O_RDONLY)) == -1 ){
			fprintf("ipl: Error opening %s\n", TPDEV);
			*dev = NULL;
			return 0;
		}
	}

	readTsConfig( &ipl_cfg);
	
	if ( ipl_cfg.xoff == TS_UNDEF )   ipl_cfg.xoff   = -60;
	if ( ipl_cfg.yoff == TS_UNDEF )   ipl_cfg.yoff   = -50;
	if ( ipl_cfg.xscale == TS_UNDEF ) ipl_cfg.xscale = 188; /* 1024-based */
	if ( ipl_cfg.yscale == TS_UNDEF ) ipl_cfg.yscale = 252; /* 1024-based */

	r39xxDevice.fd = mouFd;	
	*dev = &r39xxDevice;
	return 1;
}

static
void r39xx_reset (ipl_device_t *dev)
{
	if ( mouFd != -1 ){
		close( mouFd);
		mouFd = -1;
	}
}

static
int r39xx_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
	if ( read( mouFd, &buf, sizeof(buf)) < sizeof(buf) ){
		return 0;
	}

	switch ( buf.state ){
	case 0: evt->type = EVT_BUTTON_UP;   break;
	case 1: evt->type = EVT_BUTTON_DOWN; break;
	case 2: evt->type = EVT_MOUSE_MOVE;  break;
	}

	evt->x = (((int)buf.u + ipl_cfg.xoff) * ipl_cfg.xscale) >> 10; /* / 1024 */
	evt->y = (((int)buf.v + ipl_cfg.yoff) * ipl_cfg.yscale) >> 10; /* / 1024 */
	
	evt->mods    = 0;   /* nothing */
	evt->buttons = IPL_BTN_LEFT;   /* there is just one */
	evt->keyChar = evt->keyCode = 0;

	gettimeofday( &(evt->tv), 0);  /* maybe from the driver */

	return 1;
}

ipl_driver_t r39xxMouse = { r39xx_init, r39xx_getEvent, r39xx_reset };
ipl_device_t r39xxDevice = { 10, -1, 0, NULL, &r39xxMouse, 0 };
