/*
 * mou_sa11x.c - iPAQ touchscreen pseudo-mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 * 
 * 07/20/00 Peter Mehlitz
 */

#include "ipl.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "tpconfig.c"          /* includes the generic config file reader */

#define TSDEV "/dev/h3600_ts"
#define THRESH_X 0
#define THRESH_Y 0

ipl_device_t sa11xMice;

typedef struct {
	short pressure; /* iPAQ: 0: up, 1: down,  (otherwise 0..xxxx) */
	short xpos;        /* -1 for UP */
	short ypos;        /* -1 for UP */
	short pad;
} ts_sample;

static int state;
static int x, y;

static int mouFd = -1;

int rawCoordinates;

static int callibrate = 1;

static inline
int abs ( int v )
{
	return (v > 0) ? v : -v;
}



/***************************************************************************
 * exported funcs
 */

static int sa11x_init(ipl_device_t **dev)
{
        if ( mouFd == -1 ) {
                if ( (mouFd = open( TSDEV, O_RDONLY)) == -1 ){
			fprintf(stderr, "ipl: Error opening %s\n", TSDEV);
			*dev = NULL;
			return 0;
                }
                readTsConfig( &ipl_cfg);
	}
	sa11xMice.fd = mouFd;
	*dev = &sa11xMice;
	return 1;
}

static
void sa11x_reset (ipl_device_t *dev)
{
        if ( mouFd != -1 ){
		close( mouFd);
		mouFd = -1;
	}
}

static
int sa11x_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
        ts_sample ts;
	int n, temp;
	static int u=0, v=0;   /* it likes to loose coords on pen_up */
	static int newstate;
	
#ifdef NEVER
	fd_set  rs;
	struct timeval t;

	do {
#endif
		ts.xpos = ts.ypos = ts.pressure = 0;

		if ( (n=read( mouFd, &ts, sizeof(ts))) < sizeof(ts) ){
			printf( "read: not enough data %d/%d\n", n, sizeof(ts));
		}

		if ( ts.pressure ){ /* otherwise xpos, ypos are -1 */
			u = ts.xpos;
			v = ts.ypos; 
		}

		newstate = (ts.pressure>0);

#ifdef NEVER		
		FD_ZERO( &rs);
		FD_SET( mouFd, &rs);
		t.tv_sec = 0;
		t.tv_usec = 0;

		/* we always want to get the first press */
	} while ( (newstate == state) && _ipl_select( mouFd+1, &rs, 0, 0, &t) && 
			 (abs( u - x) <= THRESH_X) || (abs( v - y) <= THRESH_Y) );
#endif

	if ( (state != newstate) ||
		 (abs( u - x) > THRESH_X) || (abs( v - y) > THRESH_Y) ){
	
	    if ( (state == newstate) && ((x != u) || (y != v)) ) {
			evt->type = EVT_MOUSE_MOVE;
		}
	    else {
			evt->type = (newstate) ? EVT_BUTTON_DOWN : EVT_BUTTON_UP;
		}

	    x = u;
		y = v;
		state = newstate;

		if ( callibrate ){

			if ( ipl_cfg.swap == 1 ) {
				temp = x;
				x = y;
				y = temp;
			}
				
			if ( ipl_cfg.xoff != TS_UNDEF )   x += ipl_cfg.xoff;
			if ( ipl_cfg.xscale != TS_UNDEF ) x = (x * ipl_cfg.xscale) >> 10;  /* / 1024 */
			
			if ( ipl_cfg.yoff != TS_UNDEF )   y += ipl_cfg.yoff;
			if ( ipl_cfg.yscale != TS_UNDEF ) y = (y * ipl_cfg.yscale) >> 10;  /* / 1024 */
		}
		
		ROTATE(evt->x, evt->y, x, y);

		evt->buttons = IPL_BTN_LEFT;
		evt->mods = 0;
		evt->keyChar = evt->keyCode = 0;

		gettimeofday( &(evt->tv), 0);
		return 1;
	}
	return 0;
}

ipl_driver_t sa11xMouse = { sa11x_init, sa11x_getEvent, sa11x_reset };
ipl_device_t sa11xMice = { 10, -1, 0, NULL, &sa11xMouse, 0 };
