/*
 * mou_yopy.c - Yopy touchscreen pseudo-mouse device
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 * 
 * 04/26/01 "Crazy" James Simmons 
 */

#include "ipl.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "tpconfig.c"          /* includes the generic config file reader */

#define TSDEV "/dev/ts"
#define THRESH_X 0
#define THRESH_Y 0

/* 
 * Bit 31:      button state
 * Bit 20-10:	Y position
 * Bit 10-0:	X position 	 
 */

typedef struct ts_sample {
	unsigned long data; 
} ts_sample;

ipl_device_t yopyDevice;

static int x, y;

static int mouFd = -1;

int rawCoordinates;

static int callibrate;

static inline
int abs ( int v )
{
	return (v > 0) ? v : -v;
}

/***************************************************************************
 * exported funcs
 */

static int yopy_init(ipl_device_t **dev)
{
    if ( mouFd == -1 ) {
	  if ( (mouFd = open( TSDEV, O_RDONLY)) == -1 ){
	    *dev = NULL;
	    return 0;
	  }
	  callibrate = readTsConfig( &ipl_cfg);
	}
	yopyDevice.fd = mouFd;	
	*dev = &yopyDevice;
	return 1;
}

static
void yopy_reset (ipl_device_t *dev)
{
  	if ( mouFd != -1 ) {
		close( mouFd);
		mouFd = -1;
	}
}

static
int yopy_getEvent ( ipl_device_t *dev, ipl_event_t* evt )
{
	unsigned long values;
	static int pending = 0;
	ts_sample ts;
	int state, n;
	
	ts.data = 0;

	if ((n=read( mouFd, &ts, sizeof(ts))) < sizeof(ts)) {
		fprintf(stderr, "read: not enough data %d/%d\n", n, sizeof(ts));
		return 0;
	}
	values = ts.data;
	state = (((values >> 31) & 0x1) ? EVT_BUTTON_DOWN : EVT_BUTTON_UP);

	x = ((values & 0x3ff) >> 2);
	y = (((values >> 10) & 0x3ff) >> 2);

	if ((state == EVT_BUTTON_DOWN) & pending) {
		evt->type = EVT_MOUSE_MOVE;

		if ( callibrate ){
			if ( ipl_cfg.xoff != TS_UNDEF )   x += ipl_cfg.xoff;
			if ( ipl_cfg.xscale != TS_UNDEF ) x = (x * ipl_cfg.xscale) >> 10;  
			
			if ( ipl_cfg.yoff != TS_UNDEF )   y += ipl_cfg.yoff;
			if ( ipl_cfg.yscale != TS_UNDEF ) y = (y * ipl_cfg.yscale) >> 10; 
		}
		
		ROTATE(evt->x, evt->y, x, y);
		evt->buttons = IPL_BTN_LEFT;
	} else {
		if (!pending) {
			evt->type = EVT_BUTTON_DOWN;
			pending = 1;
		} else {		
			evt->type = EVT_BUTTON_UP;
			pending = 0;
		}
	}

	if ( callibrate ){
		if (ipl_cfg.xoff != TS_UNDEF)   x += ipl_cfg.xoff;
		if (ipl_cfg.xscale != TS_UNDEF) x = (x * ipl_cfg.xscale) >> 10;  
		
		if (ipl_cfg.yoff != TS_UNDEF)   y += ipl_cfg.yoff;
		if (ipl_cfg.yscale != TS_UNDEF) y = (y * ipl_cfg.yscale) >> 10; 
	}
		
	ROTATE(evt->x, evt->y, x, y);

	evt->buttons = IPL_BTN_LEFT;
	evt->mods = 0;
	evt->keyChar = evt->keyCode = 0;
	gettimeofday( &(evt->tv), 0);
	return 1;
}
ipl_driver_t yopyMouse = { yopy_init, yopy_getEvent, yopy_reset };
ipl_device_t yopyDevice = { 10, -1, 0, NULL, &yopyMouse, 0 };
