/*
 * vt_switch.c - support for Linux console vt switching
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "ipl.h"

#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <linux/vt.h>

/* VT switch console states */
#define IPL_VT_ACTIVE	0
#define IPL_VT_INACTIVE 1
#define IPL_VT_REL_REQ  2
#define IPL_VT_ACQ_REQ  3

static int tty = -1;
static struct vt_mode vt_mode;
static int vt_state = IPL_VT_ACTIVE;

static void vt_switch_signal(int signal)
{
	ipl_event_t evt;

//	printf("vt_switch_signal %d\n", signal);
	if (signal == SIGUSR1) {
		// release
		vt_state = IPL_VT_REL_REQ;
	} 
	if (signal == SIGUSR2) {
		// acquire
		vt_state = IPL_VT_ACQ_REQ;
	}

//	printf("vt_switch_signal sent %d\n", vt_state);
	evt.x = evt.y = 0;
    	evt.buttons = 0;
    	evt.mods = 0;
    	evt.keyChar = evt.keyCode = 0;
    	gettimeofday( &(evt.tv), 0);
	if (vt_state == IPL_VT_REL_REQ)
	  evt.type = EVT_VT_REL_REQ;
	else
	  evt.type = EVT_VT_ACQ_REQ;
	if (!ipl_queueEvent( &evt ))
		fprintf(stderr, "Unable to queue VT switch event\n");
}

void ipl_vtSwitchRelease()
{
	ioctl(tty, VT_RELDISP, 1);
	vt_state = IPL_VT_INACTIVE;
//	printf("ipl_vtSwitchRelease\n");
}

void ipl_vtSwitchAcquire()
{
	ioctl(tty, VT_RELDISP, VT_ACKACQ);
	vt_state = IPL_VT_ACTIVE;
//	printf("ipl_vtSwitchAcquire\n");
}

static int vt_switch_init(ipl_device_t **dev)
{
	struct sigaction act, old;

	*dev = NULL;
//	printf("vtSwitchInit\n");
	tty = STDIN_FILENO;

	memset(&act, 0, sizeof(act));
	act.sa_handler = vt_switch_signal;
	sigemptyset(&act.sa_mask);
	sigaction(SIGUSR1, &act, &old);
	sigaction(SIGUSR2, &act, &old);

	if (ioctl(tty, VT_GETMODE, &vt_mode) == -1) {
		fprintf(stderr, "Error setting up Linux VT switching\n");
		return 0;
	}
//	vt_mode.mode = VT_AUTO;
	vt_mode.mode = VT_PROCESS;
	vt_mode.waitv = 0;
	vt_mode.relsig = SIGUSR1;
	vt_mode.acqsig = SIGUSR2;
	if (ioctl(tty, VT_SETMODE, &vt_mode) == -1) {
		fprintf(stderr, "Error setting up Linux VT switching\n");
		return 0;
	}
	return 1;
}

ipl_driver_t vtSwitchGen = { vt_switch_init, NULL, NULL };  
