/*
 * main.c
 * Kick off program.
 *
 * Copyright (c) 1996, 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "vxWorks.h"
#include "time.h"

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "native.h"
#include "constants.h"
#include "support.h"
#include "threads.h"

/* HACK - we cannot use the internal VM malloc system at this point so
 * turn it off.  We should fix this properly.
 */
#undef  malloc

JNIIMPORT char* engine_name;
JNIIMPORT char* engine_version;

static int java_version_major = 1;
static int java_version_minor = 1;

#include "jni.h"

JavaVMInitArgs vmargs;
JNIEnv* env;
JavaVM* vm;
static int isJar = 0;

static int options(char**);
static void usage(void);
static size_t parseSize(char*);
static int parseSizeOption(char**, char*, size_t*);
static void handleErrors(void);

/* Kaffe specific environment */
#define KAFFEHOME       "KAFFEHOME"
#define CLASSPATH1      "KAFFECLASSPATH"
#define LIBRARYHOME1    "KAFFELIBRARYPATH"

/* Fall back on standard environment */
#define CLASSPATH2      "CLASSPATH"

/* Undefine the THREADSTACKSIZE to remove any limit */
/* #undef	THREADSTACKSIZE */

/*
 * MAIN
 */
int
KAFFE(int argc, char* argv[])
{
	jarray args;
	jclass cls;
	jclass mcls;
	jmethodID mmth;
	jobject str;
	int i;
	int farg;
	char* cp;
	char* exec;

	vmargs.version = (java_version_major<<16)|java_version_minor;
	JNI_GetDefaultJavaVMInitArgs(&vmargs);

	cp = getenv(CLASSPATH1);
	if (cp == 0) {
		cp = getenv(CLASSPATH2);
#if defined(DEFAULT_CLASSPATH)
		if (cp == 0) {
			cp = DEFAULT_CLASSPATH;
		}
#endif
	}
	vmargs.classpath = cp;

	cp = getenv(LIBRARYHOME1);
#if defined(DEFAULT_LIBRARYHOME)
	if (cp == 0) {
		cp = DEFAULT_LIBRARYHOME;
	}
#endif
	vmargs.libraryhome = cp;

	cp = getenv(KAFFEHOME);
	if (cp == 0) {
#if defined(DEFAULT_KAFFEHOME)
		cp = DEFAULT_KAFFEHOME;
#else
		fprintf(stderr, "%s is not set!\n", KAFFEHOME);
		exit(1);
#endif
	}
	vmargs.classhome = cp;

	/* Process program options */
	farg = options(argv);
	argc = argc - farg;

	/* Get the class name to start with */
	if (argv[farg] == 0) {
		usage();
		exit(1);
	}

	/* Executable is a JAR? Use the JAR launcher */
	if (isJar != 0) {
		exec = "kaffe.jar.ExecJar";
	}
	else {
		exec = argv[farg];
		farg++;
		argc--;
	}

	/* Initialise */
	JNI_CreateJavaVM(&vm, &env, &vmargs);

	mcls = (*env)->FindClass(env, exec);
	handleErrors();

	mmth = (*env)->GetStaticMethodID(env, mcls, "main", "([Ljava/lang/String;)V");
	handleErrors();

	/* Build an array of strings as the arguments */
	cls = (*env)->FindClass(env, "java/lang/String");
	handleErrors();
	args = (*env)->NewObjectArray(env, argc, cls, 0);
	handleErrors();
	for (i = 0; i < argc; i++) {
		str = (*env)->NewStringUTF(env, argv[farg+i]);
		handleErrors();
		(*env)->SetObjectArrayElement(env, args, i, str);
		handleErrors();
	}

	/* Call method, check for errors and then exit */
	(*env)->CallStaticVoidMethod(env, mcls, mmth, args);
	handleErrors();

	(*vm)->DetachCurrentThread(vm);

	return (0);
}



/* pjo... 23 feb 2000 */
int kaffe(char * commandString)
{
/* Parse all of the string passed in as an argument. */
/* Note that 's' is modified by strtok_r...  */
    int i = 0 ;
    char ** p = NULL ;
    char * s = "kaffe" ;
    char * last = NULL ;

/* s starts out with our first word ("kaffe"). */
	while (1)
	{
		p = (char**) realloc(p, sizeof(char*) * (i + 2));
		if (!p)
			break ;

		p[i++] = s ;
		p[i] = NULL ;

/* Get the next word. */
		s = strtok_r(commandString, " \t\r\n", &last) ;
		commandString = NULL ;

		if (!s)		/* No next word. */
			break ;
	}

	if (! p)
	{	fprintf(stderr, "Can't allocate memory for command-line parameters.\n") ;
		return 1;
	}

/* It appears as though the real time clock gets initialized the first
	time that clock_gettime() is called...  which can make
	System.currentTimeMillis() work really strangely the first time you
	call it.
*/
	{	struct timespec tp ;
		clock_gettime(CLOCK_REALTIME, &tp) ;
	}

#if 0
/* This hack allows us to take over standard input.  If we don't do this,
	the shell has a higher priority than us and therefore will always
	get the standard input.  The delay is there to let the rest of the system
	boot (specifically to get the actual console shell to work as opposed to
	the boot startup-script shell)... we can't lower it's priority until
	it's started.
*/
	{	extern int volatile shellTaskId ;
		extern int kaffeThreadPriorityLow ;
		taskDelay(sysClkRateGet()) ;
		taskDeleteForce(shellTaskId) ;
	}
#endif

#if 1
/* Don't exeucte with taskSpawn...   let's just lower
	our own priority appropriately.
*/
	{	extern int kaffeThreadPriorityLow ;
		taskPrioritySet(taskIdSelf(), kaffeThreadPriorityLow) ;
	}
#endif

/* Go ahead and run our JVM. */
    return KAFFE(i, p) ;
}


static
void
handleErrors(void)
{
	if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
		(*vm)->DetachCurrentThread(vm);
	}
}

/*
 * Process program's flags.
 */
static
int
options(char** argv)
{
	int i;
	int sz;
	userProperty* prop;
	int ret;
	size_t val;

	for (i = 1; argv[i] != 0; i++) {
		if (argv[i][0] != '-') {
			break;
		}

		if (strcmp(argv[i], "-help") == 0) {
			usage();
			exit(0);
		}
		else if (strcmp(argv[i], "-version") == 0) {
			fprintf(stderr, "Kaffe Virtual Machine\n");
			fprintf(stderr, "Copyright (c) 1996, 1997\nTransvirtual Technologies, Inc.  All rights reserved\n");
			fprintf(stderr, "Engine: %s   Version: %s   Java Version: %d.%d\n", engine_name, engine_version, java_version_major, java_version_minor);
			exit(0);
		}
		else if (strcmp(argv[i], "-classpath") == 0) {
			i++;
			if (argv[i] == 0) {
				fprintf(stderr, "Error: No path found for -classpath option.\n");
				exit(1);
			}
			vmargs.classpath = argv[i];
		}
		else if (strcmp(argv[i], "-addclasspath") == 0) {
			i++;
			if (argv[i] == 0) {
				fprintf(stderr, "Error: No path found for -addclasspath option.\n");
				exit(1);
			}
			vmargs.classpath = argv[i];
		}
		else if ((ret = parseSizeOption(&argv[i], "-ss", &val)) != 0) {
			i += ret - 1;
#if defined(THREADSTACKSIZE)
			if (val < THREADSTACKSIZE) {
				fprintf(stderr, "Warning: Attempt to set stack size smaller than %d - ignored.\n", THREADSTACKSIZE);
			}
			else
#endif
			{
				vmargs.nativeStackSize = val;
			}
		}
		else if ((ret = parseSizeOption(&argv[i], "-mx", &val)) != 0) {
			i += ret - 1;
			vmargs.maxHeapSize = val;
		}
		else if ((ret = parseSizeOption(&argv[i], "-ms", &val)) != 0) {
			i += ret - 1;
			vmargs.allocHeapSize = val;
		}
#if 0
		else if (strcmp(argv[i], "-nopreempt") == 0) {
			vmargs.threadpreempt = 0;
		}
#endif
		else if (strcmp(argv[i], "-jar") == 0) {
			isJar = 1;
		}
		else if (strcmp(argv[i], "-verify") == 0) {
			vmargs.verifyMode = 3;
		}
		else if (strcmp(argv[i], "-verifyremote") == 0) {
			vmargs.verifyMode = 2;
		}
		else if (strcmp(argv[i], "-noverify") == 0) {
			vmargs.verifyMode = 0;
		}
		else if (strcmp(argv[i], "-verbosegc") == 0) {
			vmargs.enableVerboseGC = 1;
		}
		else if (strcmp(argv[i], "-verbosejit") == 0) {
			vmargs.enableVerboseJIT = 1;
		}
		else if (strcmp(argv[i], "-verbosemem") == 0) {
			vmargs.enableVerboseGC = 2;
		}
		else if (strcmp(argv[i], "-verbosecall") == 0) {
			vmargs.enableVerboseCall = 1;
		}
		else if (strcmp(argv[i], "-redirect") == 0) {
			vmargs.enableRedirect = 1;
		}
		else if (strcmp(argv[i], "-verbose") == 0 || strcmp(argv[i], "-v") == 0) {
			vmargs.enableVerboseClassloading = 1;
		}
		else if (argv[i][1] ==  'D') {
			/* Set a property */
			prop = malloc(sizeof(userProperty));
			assert(prop != 0);
			prop->next = userProperties;
			userProperties = prop;
			for (sz = 2; argv[i][sz] != 0; sz++) {
				if (argv[i][sz] == '=') {
					argv[i][sz] = 0;
					sz++;
					break;
				}
			}
			prop->key = &argv[i][2];
			prop->value = &argv[i][sz];
		}
		/* The following options are not supported and will be
		 * ignored for compatibility purposes.
		 */
		else if (strcmp(argv[i], "-noasyncgc") == 0 ||
		   strcmp(argv[i], "-cs") == 0 ||
		   strcmp(argv[i], "-checksource") == 0 ||
		   strcmp(argv[i], "-prof") == 0) {
		}
		else if (strcmp(argv[i], "-ms") == 0 ||
		   strcmp(argv[i], "-oss") == 0) {
			i++;
		}
		else {
			fprintf(stderr, "Unknown flag: %s\n", argv[i]);
		}
	}

	/* Return first no-flag argument */
	return (i);
}

/*
 * Print usage message.
 */
static
void
usage(void)
{
	fprintf(stderr, "usage: kaffe [-options] class\n");
	fprintf(stderr, "Options are:\n");
	fprintf(stderr, "	-help			Print this message\n");
	fprintf(stderr, "	-version		Print version number\n");
	fprintf(stderr, "	-ss <size>		Maximum native stack size\n");
	fprintf(stderr, "	-mx <size> 		Maximum heap size\n");
	fprintf(stderr, "	-ms <size> 		Initial heap size\n");
	fprintf(stderr, "	-classpath <path>	Set classpath\n");
	fprintf(stderr, "	-addclasspath <path>	Add to classpath\n");
	fprintf(stderr, "	-verify *		Verify all bytecode\n");
	fprintf(stderr, "	-verifyremote *		Verify bytecode loaded from network\n");
	fprintf(stderr, "	-noverify		Do not verify any bytecode\n");
	fprintf(stderr, "	-D<property>=<value>	Set a property\n");
	fprintf(stderr, "	-verbosegc		Print message during garbage collection\n");
	fprintf(stderr, "	-v, -verbose		Be verbose\n");
	fprintf(stderr, "	-verbosejit		Print message during JIT code generation\n");
	fprintf(stderr, "	-verbosemem		Print detailed memory allocation statistics\n");
	fprintf(stderr, "	-verbosecall		Print detailed call flow information\n");
	fprintf(stderr, "	-debug * 		Trace method calls\n");
	fprintf(stderr, "	-noasyncgc *		Do not garbage collect asynchronously\n");
	fprintf(stderr, "	-cs, -checksource *	Check source against class files\n");
	fprintf(stderr, "	-oss <size> *		Maximum java stack size\n");
	fprintf(stderr, "	-nopreempt 		Disable preemption\n");
	fprintf(stderr, "	-jar	 		Executable is a JAR\n");
	fprintf(stderr, "	-prof *			?\n");
	fprintf(stderr, "  * Option currently ignored.\n");
}

static
size_t
parseSize(char* arg)
{
	size_t sz;
	char* narg;

	sz = strtol(arg, &narg, 0);
	switch (narg[0]) {
	case 'b': case 'B':
		break;

	case 'k': case 'K':
		sz *= 1024;
		break;

	case 'm': case 'M':
		sz *= 1024 * 1024;
		break;
	}

	return (sz);
}

static
int
parseSizeOption(char** args, char* opt, size_t* val)
{
	unsigned olen;

	olen = strlen(opt);
	if (strncmp(args[0], opt, olen) != 0) {
		return (0);
	}
	else if (strlen(args[0]) != olen) {
		*val = parseSize(&args[0][olen]);
		return (1);
	}
	else if (args[1] == 0) {
		fprintf(stderr, "Bad argument to option: %s\n", opt);
		return (0);
	}
	else {
		*val = parseSize(args[1]);
		return (2);
	}
}
