/*
 * main.c
 * Kick off program.
 *
 * Copyright (c) 1996, 1997, 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "native.h"
#include "constants.h"
#include "support.h"
#include "threads.h"


/* HACK - we cannot use the internal VM malloc system at this point so
 * turn it off.  We should fix this properly.
 */

#undef	malloc
#include "jni.h"
#define MSGOUT(msg)				printf(msg)
#define MSGOUT1(msg,a1)			printf(msg,a1)
#define MSGOUT2(msg,a1,a2)		printf(msg,a1,a2)
#define MSGOUT3(msg,a1,a2,a3)	printf(msg,a1,a2,a3)
#define MSGOUT4(msg,a1,a2,a3,a4)	printf(msg,a1,a2,a3,a4)
//fprintf(stderr,msg)

static int java_version_major = 1;
static int java_version_minor = 1;
static int isJar = 0;

JavaVMInitArgs vmargs;
JNIEnv* env;
JavaVM* vm;

extern void _exit(int);
__declspec ( dllimport ) HINSTANCE appInstance;
__declspec ( dllimport ) char* engine_name;
__declspec ( dllimport ) char* engine_version;

static
void
handleErrors(void)
{
	if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
		(*vm)->DetachCurrentThread(vm);
	}
}

static
jint
_wince_vfprintf(FILE* dummy, const char* mess, va_list args)

{
//	char buffer[255];
//	sprintf( buffer, mess, args);
//	printf( "%s", buffer);

	TCHAR* m;
	TCHAR buffer[255];

	m = getUnicode(mess);
	wvsprintf(buffer, m, args);
	freeUnicode(m);
	OutputDebugString(buffer);

	return (0);
}

static
void
_wince_exit(jint code)

{
	_exit(code);
}


static
void
_wince_abort(void)
{
	_wince_exit(-1);
}

/*
 * Print usage message.
 */
static
void
usage(void)
{
	MSGOUT( "usage: kaffe [-options] class\n");
	MSGOUT( "Options are:\n");
	MSGOUT( "	-help			Print this message\n");
	MSGOUT( "	-version		Print version number\n");
	MSGOUT( "	-ss <size>		Maximum native stack size\n");
	MSGOUT( "	-mx <size> 		Maximum heap size\n");
	MSGOUT( "	-ms <size> 		Initial heap size\n");
	MSGOUT( "	-classpath <path>	Set classpath\n");
	MSGOUT( "	-addclasspath <path>	Add to classpath\n");
	MSGOUT( "	-verify *		Verify all bytecode\n");
	MSGOUT( "	-verifyremote *		Verify bytecode loaded from network\n");
	MSGOUT( "	-noverify		Do not verify any bytecode\n");
	MSGOUT( "	-D<property>=<value>	Set a property\n");
	MSGOUT( "	-verbosegc		Print message during garbage collection\n");
	MSGOUT( "	-v, -verbose		Be verbose\n");
	MSGOUT( "	-verbosejit		Print message during JIT code generation\n");
	MSGOUT( "	-verbosemem		Print detailed memory allocation statistics\n");
	MSGOUT( "	-verbosecall		Print detailed call flow information\n");
	MSGOUT( "	-debug * 		Trace method calls\n");
	MSGOUT( "	-noasyncgc *		Do not garbage collect asynchronously\n");
	MSGOUT( "	-cs, -checksource *	Check source against class files\n");
	MSGOUT( "	-oss <size> *		Maximum java stack size\n");
	MSGOUT( "	-nopreempt 		Disable preemption\n");
	MSGOUT( "	-jar	 		Executable is a JAR\n");
	MSGOUT( "	-prof *			?\n");
	MSGOUT( "  * Option currently ignored.\n");
}

static
size_t
parseSize(char* arg)
{
	size_t sz;
	char* c;

	for ( c=arg; *c; c++) {
		if ( (*c >= '0') && (*c <= '9') )
			sz = sz*10 + (*c - '0');
		else
			break;
	}

	switch (*c) {
	case 'b': case 'B':
		break;

	case 'k': case 'K':
		sz *= 1024;
		break;

	case 'm': case 'M':
		sz *= 1024 * 1024;
		break;
	}

	return (sz);
}

static
int
parseSizeOption(char** args, char* opt, size_t* val)
{
	unsigned olen;

	olen = strlen(opt);
	if (strncmp(args[0], opt, olen) != 0) {
		return (0);
	}
	else if (strlen(args[0]) != olen) {
		*val = parseSize(&args[0][olen]);
		return (1);
	}
	else if (args[1] == 0) {
		MSGOUT1( "Bad argument to option: %s\n", opt);
		return (0);
	}
	else {
		*val = parseSize(args[1]);
		return (2);
	}
}

/*
 * Process program's flags.
 */
static
int
options(char** argv)
{
	int i;
	int sz;
	userProperty* prop;
	int ret;
	size_t val;

	for ( i=0; argv[i]; i++) {

		if (argv[i][0] != '-') {
			break;
		}

		if (strcmp(argv[i], "-help") == 0) {
			usage();
			exit(0);
		}
		else if (strcmp(argv[i], "-version") == 0) {
			MSGOUT( "Kaffe Virtual Machine\n");
			MSGOUT( "Copyright (c) 1996, 1997\nTransvirtual Technologies, Inc.  All rights reserved\n");
			MSGOUT4( "Engine: %s   Version: %s   Java Version: %d.%d\n", engine_name, engine_version, java_version_major, java_version_minor);
			exit(0);
		}
		else if (strcmp(argv[i], "-classpath") == 0) {
			i++;
			if (argv[i] == 0) {
				MSGOUT( "Error: No path found for -classpath option.\n");
				exit(1);
			}
			vmargs.classpath = argv[i];
		}
		else if (strcmp(argv[i], "-addclasspath") == 0) {
			i++;
			if (argv[i] == 0) {
				MSGOUT( "Error: No path found for -addclasspath option.\n");
				exit(1);
			}
			vmargs.classpath = argv[i];
		}
		else if ((ret = parseSizeOption(&argv[i], "-ss", &val)) != 0) {
			i += ret - 1;
			if (val < THREADSTACKSIZE) {
				MSGOUT1( "Warning: Attempt to set stack size smaller than %d - ignored.\n", THREADSTACKSIZE);
			}
			else {
				vmargs.nativeStackSize = val;
			}
		}
		else if ((ret = parseSizeOption(&argv[i], "-mx", &val)) != 0) {
			i += ret - 1;
			vmargs.maxHeapSize = val;
		}
		else if ((ret = parseSizeOption(&argv[i], "-ms", &val)) != 0) {
			i += ret - 1;
			vmargs.allocHeapSize = val;
		}
		else if (strcmp(argv[i], "-jar") == 0) {
			isJar = 1;
		}
		else if (strcmp(argv[i], "-verify") == 0) {
			vmargs.verifyMode = 3;
		}
		else if (strcmp(argv[i], "-verifyremote") == 0) {
			vmargs.verifyMode = 2;
		}
		else if (strcmp(argv[i], "-noverify") == 0) {
			vmargs.verifyMode = 0;
		}
		else if (strcmp(argv[i], "-verbosegc") == 0) {
			vmargs.enableVerboseGC = 1;
		}
		else if (strcmp(argv[i], "-verbosejit") == 0) {
			vmargs.enableVerboseJIT = 1;
		}
		else if (strcmp(argv[i], "-verbosemem") == 0) {
			vmargs.enableVerboseGC = 2;
		}
		else if (strcmp(argv[i], "-verbosecall") == 0) {
			vmargs.enableVerboseCall = 1;
		}
		else if (strcmp(argv[i], "-redirect") == 0) {
			vmargs.enableRedirect = 1;
		}
		else if (strcmp(argv[i], "-verbose") == 0 || strcmp(argv[i], "-v") == 0) {
			vmargs.enableVerboseClassloading = 1;
		}
		else if (argv[i][1] ==  'D') {
			/* Set a property */
			prop = malloc(sizeof(userProperty));
			assert(prop != 0);
			prop->next = userProperties;
			userProperties = prop;
			for (sz = 2; argv[i][sz] != 0; sz++) {
				if (argv[i][sz] == '=') {
					argv[i][sz] = 0;
					sz++;
					break;
				}
			}
			prop->key = &argv[i][2];
			prop->value = &argv[i][sz];
		}
		/* The following options are not supported and will be
		 * ignored for compatibility purposes.
		 */
		else if (strcmp(argv[i], "-noasyncgc") == 0 ||
		   strcmp(argv[i], "-cs") == 0 ||
		   strcmp(argv[i], "-checksource") == 0 ||
		   strcmp(argv[i], "-prof") == 0) {
		}
		else if (strcmp(argv[i], "-ms") == 0 ||
		   strcmp(argv[i], "-oss") == 0) {
			i++;
		}
		else {
			MSGOUT1( "Unknown flag: %s\n", argv[i]);
		}
	}

	/* Return first no-flag argument */
	return (i);
}

/*
 * MAIN
 */
int
WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)

{
	jarray     args;
	jclass     cls;
	jclass     mcls;
	jmethodID  mmth;
	jobject    str;

	char* c;
	int   i, argc, farg;

	char* argv[255];
	LPTSTR cl, cl1;
	char* exec;

	LoadLibrary( TEXT("\\windows\\libkaffevm.dll"));

	argc = 0;
	for ( cl1 = lpCmdLine; (char)*cl1; cl1++) {
		if ( (*cl1 == (TCHAR)' ') || (*cl1 == (TCHAR)'\t') )
			continue;
		cl = cl1;
		for ( i=0;(char)*cl1 && (*cl1 != (TCHAR)' ') && (*cl1 != (TCHAR)'\t'); cl1++, i++);
		*cl1 = (TCHAR)0;
		c = argv[argc++] = (char*)malloc(i);
		for ( ; *cl; cl++) {
			*(c++) = (char)*cl;
		}
		*c = 0;
		// MSGOUT1( "param: %s\n", argv[argc-1] );
	}
	argv[argc] = 0;

	appInstance = hInstance;

	vmargs.version = (java_version_major<<16)|java_version_minor;
	JNI_GetDefaultJavaVMInitArgs(&vmargs);

	vmargs.classpath = getRegistryValue( HKEY_CURRENT_USER, TEXT("\\Software\\Kaffe"), TEXT( "classpath"), (char*)malloc(255), 255);
	if ( *vmargs.classpath == 0) {
		strcpy( (char*)vmargs.classpath, "\\windows");
	}

	vmargs.libraryhome = "\\windows";
	/* default to 58k - 2k guard */
	vmargs.nativeStackSize = 56*1024;
	vmargs.classhome = 0;
	vmargs.vfprintf = _wince_vfprintf;
	vmargs.exit = _wince_exit;
	vmargs.abort = _wince_abort;

	/* Process program options */
	farg = options(argv);
	argc = argc - farg;
	// MSGOUT2( "%d %d\n", farg, argc);

	/* Get the class name to start with */
	if (argv[farg] == 0) {
		usage();
		exit(1);
	}

	/* Executable is a JAR? Use the JAR launcher */
	if (isJar != 0) {
		exec = "kaffe.jar.ExecJar";
	}
	else {
		exec = argv[farg];
		farg++;
		argc--;
	}

	/* Initialise */
	JNI_CreateJavaVM(&vm, &env, &vmargs);

	mcls = (*env)->FindClass(env, exec);
	handleErrors();

	mmth = (*env)->GetStaticMethodID(env, mcls, "main", "([Ljava/lang/String;)V");
	handleErrors();

	/* Build an array of strings as the arguments */
	cls = (*env)->FindClass(env, "java/lang/String");
	handleErrors();

	args = (*env)->NewObjectArray(env, argc, cls, 0);
	handleErrors();

	for (i = 0; i < argc; i++) {
		str = (*env)->NewStringUTF(env, argv[farg+i]);
		handleErrors();
		(*env)->SetObjectArrayElement(env, args, i, str);
		handleErrors();
	}

	/* Call method, check for errors and then exit */
	(*env)->CallStaticVoidMethod(env, mcls, mmth, args);
	handleErrors();

	(*vm)->DetachCurrentThread(vm);

	return (0);
}
