/*
 * main.c
 * Kick off program.
 *
 * Copyright (c) 1996, 1997, 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "native.h"
#include "constants.h"
#include "support.h"
#include "errors.h"
#include "threads.h"

/* HACK - we cannot use the internal VM malloc system at this point so
 * turn it off.  We should fix this properly.
 */
#undef	malloc
#include "jni.h"

#define	java_version_major 1
#define	java_version_minor 1

JavaVMInitArgs vmargs;
JNIEnv* env;
JavaVM* vm;
HINSTANCE appInstance;

extern void _exit(int);
extern unsigned short* getUnicode(const char*);
extern void freeUnicode(unsigned short*);

static
void
handleErrors(void)
{
	if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
		(*vm)->DetachCurrentThread(vm);
	}
}

static
jint
_wince_vfprintf(FILE* dummy, const char* mess, va_list args)
{
	TCHAR* m;
	TCHAR buffer[100];

	m = getUnicode(mess);
	wvsprintf(buffer, m, args);
	freeUnicode(m);
	OutputDebugString(buffer);
	return (0);
}

static
void
_wince_exit(jint code)
{
	_exit(code);
}

static
void
_wince_abort(void)
{
	_wince_exit(-1);
}

/*
 * MAIN
 */
int
WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow)
{
	jarray     args;
	jclass     cls;
	jclass     mcls;
	jmethodID  mmth;
	jobject    str;

	//char  buf[256];
	char* c;
	int   i, nArgs;

	//wcstombs( buf, lpCmdLine, sizeof(buf));

	//char* buf = "Calc";
	//char* buf = "kaffe.applet.AppletViewer \\demos\\poetry\\poetry.html";
	//char* buf = "WidgetsDemo";
	char* buf = "SwingSet";

	for ( c=buf, nArgs=0; *c; c++ ){
		if ( (*c == ' ') || (*c == '\t') ){
			nArgs++;
			*c = 0;
		}
	}

	appInstance = hInstance;

	vmargs.version = (java_version_major<<16)|java_version_minor;
	JNI_GetDefaultJavaVMInitArgs(&vmargs);

	vmargs.classpath = "\\Kaffe\\Klasses.jar;\\Demos\\Calc;\\Demos\\Poetry;\\Demos\\Widgets\\WidgetsDemo.jar;\\Demos\\swing\\swingall.jar;\\Demos\\swingset\\swingset.jar";
	vmargs.libraryhome = 0;
	vmargs.classhome = 0;
	vmargs.vfprintf = _wince_vfprintf;
	vmargs.exit = _wince_exit;
	vmargs.abort = _wince_abort;

	/* Initialise */
	JNI_CreateJavaVM(&vm, &env, &vmargs);

	mcls = (*env)->FindClass(env, buf);
	handleErrors();

	mmth = (*env)->GetStaticMethodID(env, mcls, "main", "([Ljava/lang/String;)V");
	handleErrors();

	/* Build an array of strings as the arguments */
	cls = (*env)->FindClass(env, "java/lang/String");
	handleErrors();
	args = (*env)->NewObjectArray(env, nArgs, cls, 0);
	handleErrors();

	for (i = 0, c=buf; i < nArgs; i++) {
		while ( *c++ );
		str = (*env)->NewStringUTF(env, c);
		handleErrors();
		(*env)->SetObjectArrayElement(env, args, i, str);
		handleErrors();
	}

	/* Call method, check for errors and then exit */
	(*env)->CallStaticVoidMethod(env, mcls, mmth, args);
	handleErrors();

	(*vm)->DetachCurrentThread(vm);

	return (0);
}
