dnl Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
dnl See the file "license.terms" for details.
dnl
dnl $tvt: tvt.m4,v 1.36 2001/11/21 22:07:09 samc Exp $

dnl -------------------------------------------------------------------------
dnl TVT_COMMON()
dnl Define some essential compile time constants.  Useful for --version.
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_COMMON],
[
  AC_ARG_WITH(
    build-id,
    [  --with-build-id=ARG     specify a build ID (e.g. date string)],
    [ BUILD_ID="$withval" ]
  )
  PACKAGE=[$1]
  VERSION=[$2]
  BUILD_OPTIONS="$ac_configure_args"
  BUILD_USER="`echo $USER`"
  BUILD_HOST="`hostname -f`"
  BUILD_ARCH="`uname -m`"
  BUILD_DATE="`date`"
  AC_SUBST(PACKAGE)
  AC_SUBST(VERSION)
  AC_SUBST(BUILD_DATE)
  AC_SUBST(BUILD_USER)
  AC_SUBST(BUILD_HOST)
  AC_SUBST(BUILD_OPTIONS)
  AC_SUBST(BUILD_ID)
  AC_DEFINE_UNQUOTED(PACKAGE, "$PACKAGE", [Name of package])
  AC_DEFINE_UNQUOTED(VERSION, "$VERSION", [Version number of package])
  AC_DEFINE_UNQUOTED(BUILD_ID, "$BUILD_ID", [Build ID string])
  AC_DEFINE_UNQUOTED(ARCHOS, "$target_cpu-$target_os", [The target architecture])
  AC_DEFINE_UNQUOTED(BUILD_OPTIONS, "$BUILD_OPTIONS", [The options passed to configure])
  AC_DEFINE_UNQUOTED(BUILD_USER,"$BUILD_USER", [The user who configured the package])
  AC_DEFINE_UNQUOTED(BUILD_HOST,"$BUILD_HOST", [The hostname of the build machine])
  AC_DEFINE_UNQUOTED(BUILD_ARCH,"$BUILD_ARCH", [The architecture of the build machine])
  AC_DEFINE_UNQUOTED(BUILD_DATE,"$BUILD_DATE", [The time of the build])
])


dnl -------------------------------------------------------------------------
dnl TVT_SECTION(SECTION_NAME)
dnl Make a new section on configure output (borrowed from php)
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_SECTION],
[
  AC_MSG_RESULT()
  AC_MSG_RESULT([--- $1 ---])
])

dnl -------------------------------------------------------------------------
dnl TVT_TOOLS()
dnl Add '--with-tools' and SUBST all the relevant tool paths.
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_TOOLS],
[
  AC_ARG_WITH(tools,[  --with-tools=DIR        specify where to find build tools])
  if test "$with_tools" != "" ; then
    CC="${with_tools}gcc"
    CXX="${with_tools}g++"
    STRIP="${with_tools}strip"
    AS="${with_tools}as"
    LD="${with_tools}ld"
    NM="${with_tools}nm"
    AR="${with_tools}ar cr"
    RANLIB="${with_tools}ranlib"
    if test -x "${with_tools}gcj" ; then
      GCJ=${with_tools}gcj
    fi
  else
    AC_CHECK_PROG(AR, ar, ar cr)
    AC_CHECK_PROG(STRIP, strip, strip)
    AC_CHECK_PROG(NM, nm, nm)
    AC_CHECK_PROG(LD, ld, ld)
    AC_CHECK_PROG(CXX, g++, g++)
  fi
  AC_PROG_CC
  AC_PROG_RANLIB
  AC_PROG_MAKE_SET
  AC_SUBST(NM)
  AC_SUBST(LD)
  AC_SUBST(AR)
  AC_SUBST(STRIP)
  AC_SUBST(CXX)
])

dnl -------------------------------------------------------------------------
dnl TVT_CHECK_LIB(LIB_NAME, SYMBOL_TO_CHECK, [VARIABLE_TO_ADD_TO])
dnl Add to lib path but don't run compile test
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_CHECK_LIB],[
AC_CHECK_LIB($1,$2,
  [
  changequote(, )dnl
  if test "$$3" = "" ; then
    $3="-l$1"
  else
    $3="-l$1 $$3"
  fi
  LIBS="-l$1 $LIBS"
  ac_tr_lib=HAVE_LIB`echo $1 | tr 'abcdefghijklmnopqrstuvwxyz\-\.' 'ABCDEFGHIJKLMNOPQRSTUVWXYZ__'`
  changequote([, ])dnl
  AC_DEFINE_UNQUOTED($ac_tr_lib)
  ]
)
])

dnl -------------------------------------------------------------------------
dnl TVT_CHECK_LIB_WITH_INCLUDE(LIB_NAME, SYMBOL_TO_CHECK, [VARIABLE_TO_ADD_TO])
dnl check lib with include
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_CHECK_LIB_WITH_INCLUDE],
[
AC_CACHE_CHECK(for $3 in -l$2,ac_cv_lib_$2_$3,
ac_save_LIBS="$LIBS"
LIBS="-l$2 $LIBS"
AC_TRY_LINK([#include <$1>], [void* f = $3;],
  $4="-l$2 $$4"
  ac_cv_lib_$2_$3="yes"
  ac_tr_lib=HAVE_LIB`echo $1 | tr 'abcdefghijklmnopqrstuvwxyz\-\.' 'ABCDEFGHIJKLMNOPQRSTUVWXYZ__'`
  AC_DEFINE_UNQUOTED($ac_tr_lib),
  ac_cv_lib_$2_$3="no"
  LIBS="$ac_save_LIBS"
))
])

dnl -------------------------------------------------------------------------
dnl TVT_JAVAC()
dnl Add --with-javac for location of java compiler.
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_JAVAC],
[
  AC_ARG_WITH([javac],
              [  --with-javac=JAVAC      specify which java compiler to use],
              [ WITHJAVAC=$withval ],
              [ WITHJAVAC=jikes  ])
  AC_CHECK_PROG(JAVAC, $WITHJAVAC, $WITHJAVAC, "")
  if test "$JAVAC" = ""; then
    AC_MSG_ERROR("java compiler \"$WITHJAVAC\" no found. Use --with-javac")
  fi
])

dnl -------------------------------------------------------------------------
dnl TVT_JAVA()
dnl Add --with-javac for location of java runtime.
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_JAVA],
[
  AC_ARG_WITH(java,
            [   --with-java=JAVA        specify which java runtime],
            [ WITHJAVA=$withval    ],
            [ WITHJAVA=kaffe       ])
  AC_CHECK_PROG(JAVA, $WITHJAVA, $WITHJAVA, "")
  if test "$JAVA" = ""; then
    AC_MSG_ERROR("java runtime \"$WITHJAVA\" no found. Use --with-java")
  fi
])

dnl -------------------------------------------------------------------------
dnl TVT_KAFFE([flag])
dnl Add --with-kaffe[=<dir>] for building kaffe libs
dnl If flag is non-null then require kaffe includes even without --with-kaffe.
dnl -------------------------------------------------------------------------
AC_DEFUN([TVT_KAFFE],
[
  KAFFE_CFLAGS=""
  AC_ARG_WITH(
    [kaffe],
    [   --with-kaffe[=DIR]        specify location of kaffe installation ],
    [ 
      case "$withval" in
        "no")  KAFFEDIR=""           ;;
        "yes") KAFFEDIR="/usr/local" ;;
         *)    KAFFEDIR="$withval"   ;;
      esac 
    ], 
    [
      if test -n "$1"; then
         KAFFEDIR="/usr/local"  
      fi
    ])
  if test "$KAFFEDIR" != ""; then
    if test -e $KAFFEDIR/include/kaffe/jni.h; then
      KAFFE_CFLAGS="-I$KAFFEDIR/include/kaffe"
    elif test -e $KAFFEDIR/include/jni.h; then
      KAFFE_CFLAGS="-I$KAFFEDIR/include"
    fi
    save_cppflags=$CPPFLAGS
    CPPFLAGS="$KAFFE_CFLAGS $CPPFLAGS"
    AC_CHECK_HEADER(jni.h,,
      [ 
        echo "*** failed to find jni.h:"
        echo "***   - $KAFFEDIR/include"
        echo "***   - $KAFFEDIR/include/kaffe"
        echo "***   - system include paths"
        echo "*** Use --with-kaffe=<dir> where <dir> is the prefix"
        echo "*** used for installing kaffe."
        AC_MSG_ERROR("failed to find jni.h.")
      ])
    CPPFLAGS=$save_cppflags
  fi 
  AC_SUBST(KAFFE_CFLAGS)
])

