/*
 * jarFile.c
 * Handle JAR input from files.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define DBG(s)

#undef	USE_MEMORY_CLASSLOADER

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "jar/jar.h"
#include "classpath.h"
#include "gc.h"

#include "tx_api.h"

#define JARSIZE (800*1024)

#define	GETUINT8(F)	((int)*(uint8*)((F)++))

classpathElement classpath[16];
char* realClassPath;
char* realKaffeHome;

static uint8* JarStart;
static uint8* JarEnd;
static char pathname[128];

/*
 * Initialise the classpath.  We do this by loading a single JAR file
 * across the serial port and setting this up as the only available
 * JAR in the classpath.
 */
static
void
CPinit(void)
{
#if defined(USE_MEMORY_CLASSLOADER)
	uint32 Offset;
	uint8 Bte;
	uint8 End [ 8];
	uint32 Index;
	struct TRNGBTS {
		uint32	Hi;
		uint32	Lo;
	} *TrngBts;
	uint8* ZipF;
	static char progress[] = "|/-\\";

	/* Allocate a data buffer to load the class file into */
	ZipF = (uint8*)gc_malloc_fixed(JARSIZE);

	TrngBts = ( struct TRNGBTS *)&End [ 0];

	for ( Index = 0; Index < 8; Index++) {
		End [ Index] = 0x00;
	}

	printf("Please upload a JAR file now ");
	printf("[searching]");
	fflush(stdout);

	for (;;) {
		while ( GetC ( 0, &Bte, TX_WAIT_FOREVER))
			;
		for ( Index = 0; Index < 7; Index++) {
			End [ Index] = End [ Index + 1];
		}
		End [ 7] = Bte;
		if ( TrngBts->Hi == 0xe7fce8fd && TrngBts->Lo == 0xe5fae6fb) {
			break;
		}
	}

	for ( Index = 0; Index < 8; Index++) {
		End [ Index] = 0x00;
	}

	printf("[loading] |");
	fflush(stdout);

	Offset = 0;
	for (;;) {
		while ( GetC ( 0, &Bte, TX_WAIT_FOREVER))
			;
		for ( Index = 0; Index < 7; Index++) {
			End [ Index] = End [ Index + 1];
		}
		End [ 7] = Bte;
		ZipF [ Offset++] = Bte;
		if ( TrngBts->Hi == 0xe7fce8fd && TrngBts->Lo == 0xe5fae6fb) {
			break;
		}
		if (Offset % 1024 == 0) {
			printf("\b%c", progress[(Offset / 1024) % 4]);
			fflush(stdout);
		}
	}

	/* Put ZIP file into classpath */
	classpath[0].type = CP_ZIPFILE;
	classpath[0].path = "JAR from serial port";

	JarStart = (void*)ZipF;
	JarEnd = (void*)(ZipF + Offset - 8);

	printf("\b\b[done]\n");
	fflush(stdout);
#else
	int i;
	char* spathname;
	char* cpathname;

	printf("Please enter CLASSPATH: ");
	fflush(stdout);
	fgets(pathname, sizeof(pathname), stdin);
	pathname[strlen(pathname)-1] = 0;	// Kill newline.
	if (pathname[0] == 0) {
		strcpy(pathname, ".");
	}
	printf("%s\n", pathname);
	fflush(stdout);

	/* Split pathname into bits and setup classpath */
	i = 0;
	for (cpathname = pathname, spathname = pathname; *cpathname != 0; cpathname++) {
		if (*cpathname == ':') {
			*cpathname = 0;
			classpath[i].type = CP_DIR;
			classpath[i].path = spathname;
			i++;
			spathname = cpathname + 1;
		}
	}
	/* If we have any left over, assign it too */
	if (*spathname != 0) {
		classpath[i].type = CP_DIR;
		classpath[i].path = spathname;
	}
#endif
}

static
int
CPadd(char* str)
{
	return (0);	/* We can't do this */
}

static
int
CPremove(char* str)
{
	return (0);	/* We can't do this */
}

static
void
JrawInit(void)
{
	/* Does nothing */
}

static
void
JrawOpen(jarFile* file, char* name)
{
	file->fp = JarStart;
}

static
void
JrawClose(jarFile* file)
{
}

static
int
JrawRead16(jarFile* file)
{
	int rtmp;
	rtmp = GETUINT8(file->fp);
	rtmp |= GETUINT8(file->fp) << 8;
	return (rtmp);
}

static
int
JrawRead32(jarFile* file)
{
	int rtmp;
	rtmp = GETUINT8(file->fp);
	rtmp |= GETUINT8(file->fp) << 8;
	rtmp |= GETUINT8(file->fp) << 16;
	rtmp |= GETUINT8(file->fp) << 24;
	return (rtmp);
}

static
int
JrawReadBytes(jarFile* file, int len, uint8* buf)
{
	int i;
	for (i = 0; i < len; i++) {
		buf[i] = GETUINT8(file->fp);
	}
	return (len);
}

static
int
JrawSeekStart(jarFile* file, int pos)
{
	file->fp = JarStart + pos;
	return (0);
}

static
int
JrawSeekHere(jarFile* file, int pos)
{
	file->fp += pos;
	return (0);
}

static
int
JrawSeekEnd(jarFile* file, int pos)
{
	file->fp = JarEnd + pos;
	return (0);
}

JarRawInterface Kaffe_JarRawInterface = {

	JrawInit,
	JrawOpen,
	JrawClose,
	JrawRead16,
	JrawRead32,
	JrawReadBytes,
	JrawSeekStart,
	JrawSeekHere,
	JrawSeekEnd,

};

ClasspathInterface Kaffe_ClasspathInterface = {

	CPinit,
	CPadd,
	CPremove,

};
