/*
 * config-hacks.h
 * Various hacks to the configuration for things that aren't
 *  automatically worked out.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __config_hacks_h
#define __config_hacks_h

/* If we have not assert.h we fake the assert marco */
#if !defined(HAVE_ASSERT_H)
#define	assert(X)
#endif

/* If we have no stdio.h we fake FILE */
#if !defined(HAVE_STDIO_H)
#define	FILE void
#endif


#if defined(__cplusplus)
#define EXTERN_C extern "C"
#else
#define EXTERN_C extern
#endif

/* If we don't have strdup then define the prototype so we can use it
 * from strings.c
 */
#if !defined(HAVE_STRDUP)
extern char* strdup(const char*);
#endif

/* Hacks to configure NeXT */
#if defined(NeXT)
#undef  HAVE_SYS_UTSNAME_H
#define HAVE_DYN_UNDERSTORE 1
#endif

/* Hack to configure AmigaOS */
#if defined(__amigaos__)
#define HAVE_DYN_UNDERSTORE 1
#endif  

/* shared library support */
#define JNICALL

/* Hack for Window's */ 
#if defined(__WIN32__) || defined(WIN32) || defined(_WIN32) || defined(_WIN32_WCE)
#undef	__WIN32__
#define __WIN32__ 1
#define	TARGET_OS_WINDOWS 1

#if defined (_USRDLL )
#define JNIIMPORT extern _declspec ( dllimport )
#define NATIVE_EXPORT extern _declspec ( dllexport )
#else
#define JNIIMPORT extern
#define NATIVE_EXPORT extern
#endif
#else	/* non Windows */
#define JNIIMPORT extern
#define NATIVE_EXPORT extern
#endif

#if !defined(JNIEXPORT)
#define JNIEXPORT JNIIMPORT
#endif

/* Hack for NetBSD */
#if defined(__NetBSD__)
#define	HAVE_DYN_UNDERSTORE 1
#define	HAVE_UNDERSCORED_C_NAMES 1
#endif

/* Hack to configure AIX */
#if defined(AIX)
#define  HAVE_SYS_SELECT_H 1
#endif

/* Hack for SMX */
#if defined(SMX)
#define	_STDERR	(0)

/* PM this is BAD! smx-smx.h includes xsmx.h includes xcalls.h includes *unconditionally* string.h
 * which is nicely encapsulated in our code because we can't link strdup. If we skip the "defused"
 * string.h include below, we get a collision between the above strdup (referring to string.c) and
 * the one from <string.h>. If we set HAVE_STRDUP, we get a unresolved linker symbol because we can't
 * link the Microsoft lib
 */
#define strdup strdup_xxx
#include <string.h>
#undef strdup

//#include <stdarg.h>
//EXTERN_C int vfprintf (FILE*,const char*, va_list);
#define _VFPRINTF 0

#define _INC_IO
#define _INC_STDIO

#define SMXMSP3 1
#define AT_MOTHERBOARD 1  // required for xnet

#endif /* SMX */

/* Hack for Solaris */
#if defined(sparc)
#define  SUN_AWT_FIX 1
#endif

/* Required to get boolean defined */
#if defined(HAVE_WINSOCK2_H)
#include <winsock2.h>
#if _MSC_VER < 1200
#define	HAVE_INT32 1
#endif
#endif

/* For C we define inline since it's only supported in C++ mode */
#if !defined(__cplusplus)
# if defined(__GNUC__)
#   define inline __inline__
# elif defined(_MSC_VER)
#   define inline __inline
# else
#   define inline 
# endif
#endif

/* WinCE hacks */
#if defined(UNDER_CE)
#define	_VFPRINTF	0
#define	_EXIT		0
#define	_ABORT		0
#define	_STDERR		0
#endif

/* uCLinux hacks */
#if defined(TARGET_OS_UCLINUX)
/*
 * These aren't defined (or not found for some reason).
 */
#define ntohs(x) (x)
#define ntohl(x) (x)
#define htonl(x) (x)
#define htons(x) (x)
#define __ntohs(x) (x)
#define __ntohl(x) (x)
#define __htonl(x) (x)
#define __htons(x) (x)
#endif

#if defined(TARGET_OS_VXWORKS)
#include "vxWorks.h"
#define asm __asm__
#endif /* TARGET_OS_VXWORKS */

/* LynxOS hacks */
#if defined(TARGET_OS_LYNX)
#define asm __asm__
#include <stdarg.h>
#define KAFFE_LYNX_VERSION 310
#endif /* TARGET_OS_LYNX */


#endif
