/*
 * config-loadlib.h
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __config_loadlib_h
#define __config_loadlib_h

#if defined(HAVE_SIGSETMASK)

#include "config-signal.h"

#define	K_SIGIO		0
#define	K_SIGALRM	0
#define	K_SIGCHLD	0
#define	K_SIGVTALRM	0

#if defined(SIGIO)
#undef	K_SIGIO
#define	K_SIGIO		sigmask(SIGIO)
#endif

#if defined(SIGALRM)
#undef	K_SIGALRM
#define	K_SIGALRM	sigmask(SIGALRM)
#endif

#if defined(SIGCHLD)
#undef	K_SIGCHLD
#define	K_SIGCHLD	sigmask(SIGCHLD)
#endif

#if defined(SIGVTALRM)
#undef	K_SIGVTALRM
#define	K_SIGVTALRM	sigmask(SIGVTALRM)
#endif

#define	ASYNCBLOCK()	sigsetmask(K_SIGIO|K_SIGALRM|K_SIGCHLD|K_SIGVTALRM)
#define	ASYNCUNBLOCK()	sigsetmask(0)
#else
#define	ASYNCBLOCK()
#define	ASYNCUNBLOCK()
#endif

/*
 * UNIX BSD & SysV style shared library interface.
 */
#if defined(HAVE_DLOPEN)
#if defined(HAVE_DLFCN_H) 
#include <dlfcn.h>
#else
void* dlopen(char*, int);
void* dlsym(void*, char*);
char* dlerror(void);
#endif
#define	LIBRARYINIT()
#define	LIBRARYLOAD(HAND, LIB)						\
	ASYNCBLOCK();							\
	HAND = dlopen(LIB, RTLD_LAZY | RTLD_GLOBAL );			\
	ASYNCUNBLOCK()
#define	LIBRARYUNLOAD(HAND)						\
	ASYNCBLOCK();							\
	dlclose(HAND);							\
	ASYNCUNBLOCK()
#define	LIBRARYFUNCTION(FUNC, SYM, LOADER)				\
	{								\
		NativeLibrary* hand;					\
		for (hand = (LOADER)->nativeLibs; hand != 0; hand = hand->next) { \
			ASYNCBLOCK();					\
			FUNC = dlsym(hand->desc, SYM);			\
			ASYNCUNBLOCK();					\
			if (FUNC != 0) {				\
				DBG( vm_lib,("Found syms for dynamic library %s\n", SYM)); 								\
				break;					\
			}						\
		}							\
	}
#define	LIBRARYHANDLE	void*
#if defined(HAVE_DLERROR)
#define	LIBRARYERROR()	dlerror()
#else
#define	LIBRARYERROR()	"No specific error support"
#endif
#endif

/* ------------------------------------------------------------------------ */

/*
 * MACH style shared library interface.
 */
#if defined(HAVE_MACH_O_RLD_H)
#include <mach-o/rld.h>
#define	LIBRARYINIT()
#define	LIBRARYLOAD(HAND, LIB)						\
	{								\
		const char* filenames[2];				\
		struct mach_header* new_header;				\
		filenames[0]=(LIB);					\
		filenames[1]=NULL;					\
		HAND = rld_load(NULL,&new_header,filenames,NULL);\
	}
#define	LIBRARYFUNCTION(FUNC, SYM)					\
	rld_lookup(NULL,SYM,&FUNC)
#define	LIBRARYHANDLE	long
#define	LIBRARYERROR()	"No specific error support"
#endif

/* ------------------------------------------------------------------------ */

/*
 * ??? style shared library interface.
 */
#if defined(HAVE_SHL_LOAD)
#if defined(HAVE_DL_H)
#include <dl.h>
#endif
#define	LIBRARYINIT()
#define	LIBRARYLOAD(HAND, LIB)						\
	HAND = shl_load(LIB, BIND_DEFERRED, 0L)
#define	LIBRARYFUNCTION(FUNC, SYM)					\
	{								\
		shl_t all_libs = 0;					\
		if (shl_findsym(&all_libs,SYM,TYPE_PROCEDURE,&FUNC) == 0) { \
			break;						\
		}							\
	}
#define	LIBRARYHANDLE	shl_t
#define	LIBRARYERROR()	"No specific error support"
#endif

/* ------------------------------------------------------------------------ */

/*
 * MS-Windows style shared library interface.
 */
#if defined(UNDER_CE)
#define	LIBRARYINIT()
#define LIBRARYLOAD(HAND, LIB)						\
	{			\
		TCHAR* uc = getUnicode(LIB);	\
		HAND = LoadLibrary(uc);	\
		freeUnicode(uc);		\
	}
#define LIBRARYFUNCTION(FUNC, SYM, LOADER)				\
	{								\
		NativeLibrary* hand;					\
		TCHAR* uc = getUnicode(SYM);	\
		for (hand = (LOADER)->nativeLibs; hand != 0; hand = hand->next) { \
			FUNC = GetProcAddress(hand->desc, uc);		\
			if (FUNC != 0) {				\
				break;					\
			}						\
		}							\
		freeUnicode(uc);	\
	}
#define	LIBRARYHANDLE	HMODULE
#define	LIBRARYERROR()	"No specific error support"

#define IS_SHAREDLIBRARY(P,B) (strcmp(&(P)[strlen(P)-4], ".dll") == 0)

#else
#if defined(__WIN32__)
#define	LIBRARYINIT()
#define LIBRARYLOAD(HAND, LIB)						\
	HAND = LoadLibrary(LIB)
#define	LIBRARYUNLOAD(HAND)
#define LIBRARYFUNCTION(FUNC, SYM, LOADER)
#define OLDLIBRARYFUNCTION(FUNC, SYM)	\
	{								\
		int i;							\
		for (i = 0; i < MAXLIBS && libHandle[i].ref > 0; i++) {	\
			FUNC = GetProcAddress(libHandle[i].desc, SYM);	\
			if (FUNC != 0) {				\
				break;					\
			}						\
		}							\
	}
#define	LIBRARYHANDLE	HMODULE
#define	LIBRARYERROR()	"No specific error support"

#define IS_SHAREDLIBRARY(P,B) (strcmp(&(P)[strlen(P)-4], ".dll") == 0)

#endif
#endif


/* ------------------------------------------------------------------------ */

/*
 * GNU style shared library interface.
 */
#if defined(HAVE_LIBDLD) && !defined(HAVE_DLFCN_H)
#define	LIBRARYINIT()							\
        {								\
                static int init_dld = 0;				\
                if (init_dld == 0) {					\
                        init_dld = 1;					\
                        if (dld_init("/usr/local/bin/kaffe") != 0) {	\
                                return (-1);				\
                        }						\
                }							\
        }
#define LIBRARYLOAD(HAND, LIB)						\
	HAND = !dld_link(LIB)
#define LIBRARYFUNCTION(FUNC, SYM)					\
	FUNC = dld_get_func(SYM)
#define	LIBRARYHANDLE	int
#define	LIBRARYERROR()	"No specific error support"
#endif

/* ------------------------------------------------------------------------ */

#if !defined(IS_SHAREDLIBRARY)
#define IS_SHAREDLIBRARY(P,B) ((B)[0] == 0x7f && (B)[1] == 'E' && (B)[2] == 'L' && (B)[3] == 'F')
#endif

/*
 * If we don't have any shared library support, we put in a stub system.
 */
#if !defined(LIBRARYINIT)
#define	LIBRARYINIT()
#endif
#if !defined(LIBRARYLOAD)
#define	LIBRARYLOAD(HAND, LIB)	 	HAND = 1
#endif
#if !defined(LIBRARYHANDLE)
#define	LIBRARYHANDLE			int
#endif
#if !defined(LIBRARYERROR)
#define	LIBRARYERROR()			"No shared library support"
#endif
#if !defined(LIBRARYFUNCTION)
#define	LIBRARYFUNCTION(FUNC, STUB)
#endif

#endif
