/*
 * i386/common.h
 * Common i386 configuration information.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __i386_h
#define __i386_h

#include "../kaffe/kaffevm/callMethodInfo.h"

/* The 386 never aligns to more than a 4 byte boundary. */
#define	ALIGNMENT_OF_SIZE(S)	((S) < 4 ? (S) : 4)

/*
 * Make a call to a native or Java (JIT) method.
 *  This assembly code should build a standard C call using the passed
 *  call information.  By its nature this is highly processor specific.
 *  This function is mandatory for both JIT and Interpreter (since stubs
 *  have now been deprecated).
 */
#if 0
#define	sysdepCallMethod(CALL)						\
  do {									\
	register callMethodInfo* scall asm("%edi");			\
	scall = (CALL);							\
	asm volatile ("							\n\
1:									\n\
		cmpl $0,%0						\n\
		je 3f							\n\
		decl %0							\n\
		cmpb $0,(%2,%0)						\n\
		je 1b							\n\
		cmpb $1,(%2,%0)						\n\
		je 2f							\n\
		pushl 4(%1,%0,8)					\n\
2:									\n\
		pushl (%1,%0,8)						\n\
		jmp 1b							\n\
3:									\n\
		call *%3						\n\
		movl %5,%%ebx						\n\
		movb %4,%%cl						\n\
		cmpb $0x46,%%cl						\n\
		jne 4f							\n\
		fstps (%%ebx)						\n\
		jmp 6f							\n\
4:									\n\
		cmpb $0x44,%%cl						\n\
		jne 5f							\n\
		fstpl (%%ebx)						\n\
		jmp 6f							\n\
5:									\n\
		movl %%eax,(%%ebx)					\n\
		cmpb $0x4a,%%cl						\n\
		jne 6f							\n\
		movl %%edx,4(%%ebx)					\n\
6:									\n\
	" :								\
	  : "r" ((scall)->nrargs),					\
	    "r" ((scall)->args),					\
	    "r" ((scall)->callsize),					\
	    "m" ((scall)->function),					\
	    "m" ((scall)->rettype),					\
	    "m" ((scall)->ret)						\
	  : "ebx", "ecx", "cc", "memory");				\
	asm volatile ("							\n\
		subl %0,%%esp						\n\
	" : : "r" ((scall)->argsize * sizeof(jint)) : "cc");		\
  } while(0)
#endif

static inline void sysdepCallMethod(callMethodInfo *call) {
  unsigned args = call->nrargs;

  /* Push all arguments into the stack, in last-to-first order.  This
     assumes that the second 32bit-word of longs and doubles is
     available as an additional int argument, as callMethod[AV]() do.  */
  while (args > 0)
    asm volatile ("pushl %0" : : "m" (call->args[--args].i) : "sp");

  switch(call->retsize) {
  case 0:
    /* Must be void.  */
    ((void (*)(void))(call->function))();
    break;

  case 1:
    if (call->rettype == 'F')
      call->ret->f = ((jfloat (*)(void))(call->function))();
    else /* Must be 32-bit or smaller int.  */
      call->ret->i = ((jint (*)(void))(call->function))();
    break;

  default:
    /* It could've been `case 2;', but then we'd get an additional cmp
     * that we don't really need.  */
    if (call->rettype == 'D')
      call->ret->d = ((jdouble (*)(void))(call->function))();
    else /* Must be jlong.  */
      call->ret->j = ((jlong (*)(void))(call->function))();
    break;
  }

  /* Adjust the stack pointer.  */
  asm volatile ("addl %0,%%esp" : :
                "r" (call->argsize * sizeof(jint)) : "cc", "sp");
}

/*
   Define the execution format for this machine.
   The stack looks something like this :-

     [high mem]
     ---------------
     epc
     ebp  <--------------\
                         |
                         |
     ---------------     |
     epc                 |
     ebp  ---------------/

     ---------------
     [low mem]
*/

typedef struct _machineStackFrame {
	void*		ebp;
	void*		epc;
} machineStackFrame;

#define	STACK_FRAME(F)		((F)->ebp)
#define	STACK_PC(F)		((F)->epc)
#define	STACK_CURRENT_FRAME()	({ void* v ; asm("movl %%ebp,%0" : "=r" (v)); v; })

/*
 * Do an atomic compare and exchange. Lock the bus before in case we are
 * executing on an SMP. The value at address 'A' is checked against value
 * 'O' and if they match the value at adress 'A' is loaded with value 'N'.
 * We return 1 if the comparison is successful, otherwise 0.
 */
#if defined(TARGET_ARCH_SMP)
 
#define	COMPARE_AND_EXCHANGE(A,O,N) \
	({ \
		char ret; \
		asm volatile(" \
			movl %3,%%eax \n\
			movl %2,%%ecx \n\
		        lock \n\
		        cmpxchgl %%ecx,%1 \n\
			sete %0" \
		  : "=q" (ret) : "m" (*(A)), "r" (N), "r" (O) : "eax", "ecx", "cc", "memory" ); \
		(ret); \
	}) /* special gcc syntax: last exp evaluated (ret) is the value of the expression */

#else

#define	COMPARE_AND_EXCHANGE(A,O,N) \
	({ \
		char ret; \
		asm volatile(" \
			movl %3,%%eax \n\
			movl %2,%%ecx \n\
			cmpxchgl %%ecx,%1 \n\
			sete %0" \
		  : "=q" (ret) : "m" (*(A)), "r" (N), "r" (O) : "eax", "ecx", "cc", "memory" ); \
		(ret); \
	}) /* special gcc syntax: last exp evaluated (ret) is the value of the expression */

#endif

/**
 * Force all register which need to be preserved to be saved.  Used in the
 * JNI code.
 */
#define	FORCE_SAVE_REGISTERS() \
	asm volatile("" : : : "ebx", "edi", "esi")

#endif
