/*
 * i386/jit.h
 * Common i386 JIT configuration information.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __i386_jit_h
#define __i386_jit_h

/**/
/* Exception handling information. */
/**/

/* Extract a local argument from given frame */
#define	FRAMEOBJECT(F)							\
	(*(Hjava_lang_Object**)((uintp)(F) + 8))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(F, H)					\
        asm volatile("movl %0,%%ebp\n\tjmp *%1" : : "g" (F), "r" (H))

/**/
/* Method dispatch.  */
/**/

#define HAVE_TRAMPOLINE

#if defined(__GNUC__)
#define	PACKED	__attribute__((packed))
#else
#define	PACKED
#endif

#if defined(_MSC_VER)
#pragma pack ( push, 1 )
#endif
typedef struct _methodTrampoline {
	unsigned char call		PACKED;
	int fixup			PACKED;
	short meth			PACKED;
	struct Hjava_lang_Class* clazz	PACKED;
} methodTrampoline;
#if defined(_MSC_VER)
#pragma pack ( pop )
#endif

extern void i386_do_fixup_trampoline(void);

#define FILL_IN_TRAMPOLINE(T,M,C)					\
	do {								\
		(T)->call = 0xe8;					\
		(T)->fixup = (uintp)i386_do_fixup_trampoline-(uintp)(T)-5;\
		(T)->meth = (short)((M) - CLASS_METHODS(C));		\
		(T)->clazz = (C);					\
	} while (0)

#define FIXUP_TRAMPOLINE_DECL	char* _data
#define FIXUP_TRAMPOLINE_INIT	clazz = *(Hjava_lang_Class**)(_data+2); \
				meth = &CLASS_METHODS(clazz)[*(short*)_data];

/**/
/* Register management information. */
/**/

/* Define the register set */
#define	REGISTER_SET							\
	{ /* eax */	0, 0, Rint|Rsubint|Rref,0, 0, 0	},		\
	{ /* ecx */	0, 0, Rint|Rsubint|Rref,0, 0, 1	},		\
	{ /* edx */	0, 0, Rint|Rsubint|Rref,0, 0, 2	},		\
	{ /* ebx */	0, 0, Rint|Rsubint|Rref,Rglobal|Rnosaveoncall, 0, 3 },\
	{ /* esp */	0, 0, Reserved,		0, 0, 4	},		\
	{ /* ebp */	0, 0, Reserved,		0, 0, 5	},		\
	{ /* esi */	0, 0, Rint|Rref,	Rglobal|Rnosaveoncall, 0, 6 },\
	{ /* edi */	0, 0, Rint|Rref,	Rglobal|Rnosaveoncall, 0, 7 },\
	{ /* f0  */	0, 0, Rfloat|Rdouble,	Rreadonce, 0, 8 },

/* Number of registers in the register set */
#define	NR_REGISTERS	9

/* Number of function globals in register set */
#define	NR_GLOBALS	3

/* We keep the floats in a stack */
#define	HAVE_FLOATING_POINT_STACK	1

/**/
/* Opcode generation. */
/**/

#define LABEL_Lframe(P,D,L)						\
	{								\
		int framesize = SLOTSIZE * (maxLocal + maxStack +	\
			maxTemp - maxArgs);				\
		*(uint32*)(P) = framesize;				\
	}

#define	EXTRA_LABELS(P,D,L)						\
	case Lframe:	LABEL_Lframe(P,D,L);	break;

/**/
/* Slot management information. */
/**/

/* Size of each slot */
#define	SLOTSIZE	4

/* Generate slot offset for an argument */
#define SLOT2ARGOFFSET(_n)	(8 + SLOTSIZE * (_n))

/* Generate slot offset for a local (non-argument) */
#define SLOT2LOCALOFFSET(N)	(-SLOTSIZE * ((N) - maxArgs + 1))

/* Generate the slot offset to the stack limit */
#define	STACK_LIMIT()		SLOT2ARGOFFSET(maxArgs)

/* Wrap up a native call for the JIT */
#define KAFFEJIT_TO_NATIVE(_m)

#if 0 /* Turned off until I can be bothered to fix the makeSig issue */

#define TRACE_METHOD_ARGS() \
        if (!METHOD_IS_STATIC(meth)) { \
                kprintf("\t%p (%s) ", *(void**)args, \
                        CLASS_CNAME(((Hjava_lang_Object*)(*(void**)args))->dtable->class)); \
                args = (void*)(((uintp)args) + SLOTSIZE); \
        } \
        if (sig->len > 0 && METHOD_IS_STATIC(meth)) kprintf("\t"); \
        for (; n<sig->len; n++) { \
		int64 lval = 0; \
                switch (sig->elem[n]->data[0]) { \
                case 'F': \
                        kprintf("%e ",*(float*)args); \
			args = (void*)(((uintp)args) + SLOTSIZE); \
                        break; \
                case 'I': case 'Z': case 'S': \
                case 'B': case 'C': \
                        kprintf("%d ",*(int*)args); \
			args = (void*)(((uintp)args) + SLOTSIZE); \
                        break; \
                case 'D': \
			lval = ((int64)((void**)args)[1] << 32) | (int64)*(void**)args; \
                        kprintf("%e ",lval); \
			args = (void*)(((uintp)args) + 2 * SLOTSIZE); \
                        break; \
                case 'J': \
			lval = ((int64)((void**)args)[1] << 32) | (int64)*(void**)args; \
                        kprintf("%lld ",lval); \
			args = (void*)(((uintp)args) + 2 * SLOTSIZE); \
                        break; \
                case 'L': case '[': \
                        kprintf("0x%p ",*(void**)args); \
			args = (void*)(((uintp)args) + 2 * SLOTSIZE); \
                        break; \
                default: \
                        ABORT(); \
                } \
        } \
        if (sig->len > 0 || !METHOD_IS_STATIC(meth)) kprintf("\n")

#endif

#if defined(__GNUC__) && 0
#define TRACE_METHOD_END() \
        asm("   pushl %%eax						\n\
		pushl %%edx						\n\
		pushl %%ebp						\n\
		fstpl tStore						\n\
		pushl tStore						\n\
		movl %%esp,%%ebp					\n\
		" : : ); \
        kprintf("E\n"); \
        asm("	movl %%ebp,%%esp					\n\
		popl tStore						\n\
		fldl tStore						\n\
		popl %%ebp						\n\
		popl %%edx						\n\
		popl %%eax						\n\
		" : : : "eax", "edx", "ebp")
#endif

#endif
