/*
 * m68k/common.h
 * Common M68K configuration information.
 *
 * Copyright (c) 1996, 1997, 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * Written by Kiyo Inaba <inaba@src.ricoh.co.jp>, 1998;
 * Based on the ports
 *      by Remi Perrot <r_perrot@mail.club-internet.fr> to m68k/linux
 * and
 *      by Alexandre Oliva <oliva@dcc.unicamp.br> to sparc
 */

#ifndef __m68k_common_h
#define __m68k_common_h

/*
 * Alignment in structure is 2 bytes packed.
 */
#define ALIGNMENT_OF_SIZE(S)    (((S>1)?2:1))

/*
 * Make a call to a native function.
 */
#define sysdepCallMethod(CALL) do {				\
	int extraargs[(CALL)->nrargs];				\
	register int d0 asm ("d0");				\
	register int d1 asm ("d1");				\
	int *res;						\
	int *args = extraargs;					\
	int argidx;						\
	for(argidx = 0; argidx < (CALL)->nrargs; ++argidx) {	\
		if ((CALL)->callsize[argidx])			\
			*args++ = (CALL)->args[argidx].i;	\
		else						\
			*args++ = (CALL)->args[argidx-1].j;	\
	}							\
	asm volatile ("jsr	%2@\n"				\
	 : "=r" (d0), "=r" (d1)					\
	 : "a" ((CALL)->function),				\
	   "r" ((CALL)->nrargs * sizeof(int))			\
	 : "cc", "memory");					\
	if ((CALL)->retsize != 0) {				\
		res = (int *)(CALL)->ret;			\
		res[1] = d1;					\
		res[0] = d0;					\
	}							\
} while (0)

/*
 * Define what the machine stack looks like for stack walking.
 */
typedef struct _machineStackFrame {
	void*           fp;
	void*           pc;
} machineStackFrame;

#define	STACK_FRAME(F)		((F)->fp)
#define	STACK_PC(F)		((F)->pc)
#define STACK_CURRENT_FRAME()   ({ void* v ; asm("move%.l %/a6,%0" : "=r" (v)); v; })

#endif
