/*
 * m68k/netbsd1/jit.h
 * NetBSD M68000 JIT configuration information.
 *
 * Copyright (c) 1996, 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Written by Marcello Balduccini <marcy@morgana.usr.dsi.unimi.it>, 1997
 * NOTE: this file is heavily based on m68k/jit.h,
 * written by Richard Henderson <rth@tamu.edu>, 1997.
 *
 * Modified by Kiyo Inaba <inaba@src.ricoh.co.jp>,
 * to use trampoline code.
 */

#ifndef __m68k_jit_h
#define __m68k_jit_h

/**/
/* Exception handling information. */
/**/

/* Extract the object argument from given frame */
#define FRAMEOBJECT(f) 		(*(Hjava_lang_Object**)((f) + 8))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(F, H)					\
	__asm__ __volatile__(						\
		"move%.l %0,%/a6\n\t"					\
		"jmp %1@"						\
		: : "g" (F), "a" (H) : "cc", "memory")

/**/
/* Method dispatch.  */
/**/

#define HAVE_TRAMPOLINE

#define	PACKED	__attribute__((packed))

typedef struct _methodTrampoline {
        unsigned short call		PACKED;
	int fixup			PACKED;
	struct _methods* meth		PACKED;
	void** where			PACKED;
} methodTrampoline;

extern void m68k_do_fixup_trampoline(void);

#define FILL_IN_TRAMPOLINE(T,M,W)                                       \
        do {                                                            \
                (T)->call = 0x4eb9;	/* jsr abs.l */			\
                (T)->fixup = (int)m68k_do_fixup_trampoline;		\
                (T)->meth = (M);                                        \
                (T)->where = (W);                                       \
        } while (0)

#define FIXUP_TRAMPOLINE_DECL   void** _data
#define FIXUP_TRAMPOLINE_INIT   meth = (Method*)_data[0];		\
				where = (void**)_data[1];

/**/
/* Register management information. */
/**/

#define	_GL_	(Rglobal|Rnosaveoncall)

/* Define the register set */

#define	REGISTER_SET							\
	{ /* d0 */	0, 0, Rint, 		0, 0, 0  },		\
	{ /* d1 */	0, 0, Rint, 		0, 0, 1  },		\
	{ /* d2 */	0, 0, Rint, 		Rnosaveoncall, 0, 2  }, \
	{ /* d3 */	0, 0, Rint, 		Rnosaveoncall, 0, 3  }, \
	{ /* d4 */	0, 0, Rint, 		_GL_, 0, 4  }, \
	{ /* d5 */	0, 0, Rint, 		_GL_, 0, 5  }, \
	{ /* d6 */	0, 0, Rint, 		_GL_, 0, 6  }, \
	{ /* d7 */	0, 0, Rint, 		_GL_, 0, 7  }, \
	{ /* a0 */	0, 0, Rref,		0, 0, 8  },		\
	{ /* a1 */	0, 0, Rref,		0, 0, 9  },		\
	{ /* a2 */	0, 0, Rref,		_GL_, 0, 10 }, \
	{ /* a3 */	0, 0, Rref,		_GL_, 0, 11 }, \
	{ /* a4 */	0, 0, Rref,		_GL_, 0, 12 }, \
	{ /* a5 */	0, 0, Rref,		_GL_, 0, 13 }, \
	{ /* fp */	0, 0, Reserved,		0, 0, 14 },		\
	{ /* sp */	0, 0, Reserved,		0, 0, 15 },		\
	{ /* fp0 */	0, 0, Rfloat|Rdouble,	0, 0, 16 },		\
	{ /* fp1 */	0, 0, Rfloat|Rdouble,	0, 0, 17 },		\
	{ /* fp2 */	0, 0, Rfloat|Rdouble,	0, 0, 18 }, \
	{ /* fp3 */	0, 0, Rfloat|Rdouble,	0, 0, 19 }, \
	{ /* fp4 */	0, 0, Rfloat|Rdouble,	0, 0, 20 }, \
	{ /* fp5 */	0, 0, Rfloat|Rdouble,	0, 0, 21 }, \
	{ /* fp6 */	0, 0, Rfloat|Rdouble,	0, 0, 22 }, \
	{ /* fp7 */	0, 0, Rfloat|Rdouble,	0, 0, 23 }, 

/* Number of registers in the register set */
#define	NR_REGISTERS	24

/* Define the registers to be saved and restore in the prologue/epilogue */
#define	SRNR		10
#define	SAVEMASK	0x3F3C
#define	RESTOREMASK	0x3CFC

/* Define the size of the frame */
#define	FRAMESIZE \
	((maxLocal - maxArgs + maxStack + maxTemp) * SLOTSIZE)

#define Llong16         	(Larchdepend+0) /* Label is 16 bits long */
#define Llnegframe         	(Larchdepend+1)

#define	LABEL_Lnegframe(P,D,L)	*(int16*)(P) = -FRAMESIZE
#define	LABEL_Llnegframe(P,D,L)	*(int32*)(P) = -FRAMESIZE
#define	LABEL_Llong16(P,V,L)	*(int16*)(P) = (V)

#define	EXTRA_LABELS(P,D,L) \
	case Llong16:		LABEL_Llong16(P,D,L);		break; \
	case Lnegframe:		LABEL_Lnegframe(P,D,L);		break; \
	case Llnegframe:	LABEL_Llnegframe(P,D,L);	break;

/* Number of function globals in register set */
#define	NR_GLOBALS	8

/* Number of arguments passed in registers - we don't have any */
#undef	NR_ARGUMENTS

/**/
/* Opcode generation. */
/**/

/* Define if generated code uses two operands rather than one */
#define	TWO_OPERAND

/**/
/* Slot management information. */
/**/

/* Size of each slot */
#define	SLOTSIZE		4

/*
 * A stack frame looks like:
 *
 *      |				|
 *	|  Arguments			|
 *      |-------------------------------|
 *	|  Return address		|
 *      |-------------------------------|
 *      |  Saved fp			|
 * fp-> |-------------------------------|
 *      |  Temporaries			|
 * sp-> |-------------------------------|
 */

/* Generate slot offset for an argument (relative to fp) */
#define SLOT2ARGOFFSET(N)    (8 + SLOTSIZE * (N))

/* Generate slot offset for a local (non-argument) (relative to fp) */
#define SLOT2LOCALOFFSET(N)	(-SLOTSIZE * ((N) - maxArgs + 1))

/* Generate the slot offset to the stack limit */
#define	STACK_LIMIT()		SLOT2ARGOFFSET(maxArgs)

/* Wrap up a native call for the JIT */
#define	KAFFEJIT_TO_NATIVE(_m)

/* On the 68040 (and 060 I presume) we need to flush the instruction
   cache before running new code.  Stupidly, Motorola made cpush a
   privledged instruction, so we have to get operating system help
   for this.  Naturally, there is no standard there.  */

#define FLUSH_040_DCACHE(beg, end) \
	__asm__ __volatile__( \
		"movem%.l %/d0-%/d7/%/a0-%/a5,%-\n\t" \
		"move%.l %0,%/a1\n\t" \
		"move%.l %1,%/d1\n\t" \
		"sub%.l %/d1,%/a1\n\t" \
		"movel  %#0x80000004,d0\n\t" \
		"trap   %#12\n\t" \
		"movem%.l %+,%/d0-%/d7/%/a0-%/a5" \
		: : "g" (beg), "g" (end))

#undef FLUSH_DCACHE
#define FLUSH_DCACHE(B,E)	FLUSH_040_DCACHE(B,E)

#endif
