/*
 * mips/common.h
 * Common MIPS configuration information.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __mips_common_h
#define __mips_common_h

#include <stddef.h>

/*
 * Make a call to a native or Java (JIT) method.
 *  This assembly code should build a standard C call using the passed
 *  call information.  By its nature this is highly processor specific.
 *  This function is mandatory for both JIT and Interpreter (since stubs
 *  have now been deprecated).
 */

#if defined(HAVE_NO_FLOATING_POINT)
asm(".equ HAVE_NO_FLOATING_POINT,1");
#endif
#if defined(WORDS_BIGENDIAN)
asm(".equ WORDS_BIGENDIAN,1");
#endif

/*
 * MIPS has a horrible calling convention.  See the MIPSpro
 * Assembly Language Programmer's Guide for details.
 */
#define	sysdepCallMethod(CALL)						\
do {									\
	callMethodInfo* scall;						\
	scall = (CALL);							\
	asm volatile ("							\n\
		.set	noreorder					\n\
		addiu	$sp,$sp,-16					\n\
		sw	%4,0($sp)					\n\
		sw	%5,4($sp)					\n\
		sw	$gp,8($sp)					\n\
		# sw	$31,12($sp)					\n\
		move	$16,$sp		# save sp			\n\
		li	$8,0		# arg_idx			\n\
		li	$9,0		# fp_idx			\n\
		li	$11,0		# found_int_arg			\n\
		sll	$12,%0,3	# reserve space for stack	\n\
		andi	$13,%0,0x0001					\n\
		beq	$13,$0,lp_pre					\n\
		nop							\n\
		addiu	$12,$12,8					\n\
lp_pre:									\n\
		subu	$sp,$sp,$12					\n\
lp_start:								\n\
		beq	%0,$0,docall					\n\
		nop							\n\
		lbu	$12,0(%2)					\n\
.ifndef HAVE_NO_FLOATING_POINT						\n\
		li	$13,0x44	# type == 'D' ?			\n\
		bne	$13,$12,arg1					\n\
		nop							\n\
		andi	$12,$8,0x0001	# double-word aligned ?		\n\
		beq	$12,$0,a01					\n\
		nop							\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a01:									\n\
		bne	$11,$0,a03	# found_int_arg ?		\n\
		nop							\n\
		li	$12,0		# fp_idx == 0 ?			\n\
		bne	$12,$9,a02					\n\
		nop							\n\
.ifdef WORDS_BIGENDIAN							\n\
		lwc1	$f12,4(%1)	# --> $f12			\n\
		lwc1	$f13,0(%1)					\n\
.else									\n\
		lwc1	$f12,0(%1)	# --> $f12			\n\
		lwc1	$f13,4(%1)					\n\
.endif									\n\
		addiu	$9,$9,1		# fp_idx += 1			\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
a02:									\n\
		li	$12,1		# fp_idx == 1 ?			\n\
		bne	$12,$9,a03					\n\
		nop							\n\
.ifdef WORDS_BIGENDIAN							\n\
		lwc1	$f14,4(%1)	# --> $f12			\n\
		lwc1	$f15,0(%1)					\n\
.else									\n\
		lwc1	$f14,0(%1)	# --> $f14			\n\
		lwc1	$f15,4(%1)					\n\
.endif									\n\
		addiu	$9,$9,1		# fp_idx += 1			\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
a03:									\n\
		li	$12,2		# arg_idx == 2 ?		\n\
		bne	$12,$8,a04					\n\
		nop							\n\
		lw	$6,0(%1)	# --> $a3, $a4			\n\
		lw	$7,4(%1)					\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
a04:									\n\
		sll	$12,$8,2	# calc stack position		\n\
		addu	$12,$12,$sp					\n\
		lw	$13,0(%1)	# --> stack			\n\
		lw	$14,4(%1)					\n\
		sw	$13,0($12)					\n\
		sw	$14,4($12)					\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
arg1:									\n\
		li	$13,0x46	# type == 'F' ?			\n\
		bne	$13,$12,arg2					\n\
		nop							\n\
		bne	$11,$0,a12	# found_int_arg ?		\n\
		nop							\n\
		li	$12,0		# fp_idx == 0 ?			\n\
		bne	$12,$9,a11					\n\
		nop							\n\
		lwc1	$f12,0(%1)	# --> $f12			\n\
		addiu	$9,$9,1		# fp_idx += 1			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a11:									\n\
		li	$12,1		# fp_idx == 1 ?			\n\
		bne	$12,$9,a12					\n\
		nop							\n\
		lwc1	$f14,0(%1)	# --> $f14			\n\
		addiu	$9,$9,1		# fp_idx += 1			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a12:									\n\
		li	$12,1						\n\
		bne	$12,$8,a13	# arg_idx == 1 ?		\n\
		nop							\n\
		lw	$5,0(%1)	# --> $a1			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a13:									\n\
		li	$12,2						\n\
		bne	$12,$8,a14	# arg_idx == 2 ?		\n\
		nop							\n\
		lw	$6,0(%1)	# --> $a2			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a14:									\n\
		li	$12,3						\n\
		bne	$12,$8,a15	# arg_idx == 3 ?		\n\
		nop							\n\
		lw	$7,0(%1)	# --> $a3			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a15:									\n\
		sll	$12,$8,2	# calc stack position		\n\
		addu	$12,$12,$sp					\n\
		lw	$13,0(%1)	# --> stack			\n\
		sw	$13,0($12)					\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
.endif									\n\
arg2:									\n\
		li	$13,0x4a	# type == 'J' ?			\n\
		bne	$13,$12,arg3					\n\
		nop							\n\
		andi	$12,$8,0x0001	# double-word aligned ?		\n\
		beq	$12,$0,a21					\n\
		li	$11,1		# fount_int_arg = 1		\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a21:									\n\
		li	$12,0		# arg_idx == 0 ?		\n\
		bne	$12,$8,a22					\n\
		nop							\n\
		lw	$4,0(%1)	# --> $a0, $a1			\n\
		lw	$5,4(%1)					\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
a22:									\n\
		li	$12,2		# arg_idx == 2 ?		\n\
		bne	$12,$8,a23					\n\
		nop							\n\
		lw	$6,0(%1)	# --> $a2, $a3			\n\
		lw	$7,4(%1)					\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
a23:									\n\
		sll	$12,$8,2	# calc stack position		\n\
		addu	$12,$12,$sp					\n\
		lw	$13,0(%1)	# --> stack			\n\
		lw	$14,4(%1)					\n\
		sw	$13,0($12)					\n\
		sw	$14,4($12)					\n\
		b	lp_end						\n\
		addiu	$8,$8,2		# arg_idx += 2			\n\
arg3:									\n\
		li	$11,1		# fount_int_arg = 1		\n\
		li	$12,0		# arg_idx == 0 ?		\n\
		bne	$12,$8,a31	# type == other			\n\
		nop							\n\
		lw	$4,0(%1)	# --> $a0			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a31:									\n\
		li	$12,1		# arg_idx == 1 ?		\n\
		bne	$12,$8,a32					\n\
		nop							\n\
		lw	$5,0(%1)	# --> $a1			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a32:									\n\
		li	$12,2		# arg_idx == 2 ?		\n\
		bne	$12,$8,a33					\n\
		nop							\n\
		lw	$6,0(%1)	# --> $a2			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a33:									\n\
		li	$12,3		# arg_idx == 3 ?		\n\
		bne	$12,$8,a34					\n\
		nop							\n\
		lw	$7,0(%1)	# --> $a3			\n\
		b	lp_end						\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
a34:									\n\
		sll	$12,$8,2	# calc stack position		\n\
		addu	$12,$12,$sp					\n\
		lw	$13,0(%1)	# --> stack			\n\
		sw	$13,0($12)					\n\
		addiu	$8,$8,1		# arg_idx += 1			\n\
lp_end:									\n\
		addiu	%1,%1,8						\n\
		addiu	%2,%2,1						\n\
		b	lp_start					\n\
		addiu	%0,%0,-1					\n\
docall:									\n\
		# move	$31,$0		# stop the backtracer cleanly	\n\
		move	$25,%3		# do call			\n\
		jal	$25						\n\
		nop							\n\
		move	$sp,$16		# restore sp			\n\
		lw	$16,0($sp)					\n\
		lw	$17,4($sp)					\n\
		lw	$gp,8($sp)					\n\
		# lw	$31,12($sp)					\n\
		addiu	$sp,$sp,16					\n\
.ifndef HAVE_NO_FLOATING_POINT						\n\
		li	$6,0x44		# rettype == 'D' ?		\n\
		bne	$16,$6,rv1					\n\
		nop							\n\
.ifdef WORDS_BIGENDIAN							\n\
		swc1	$f0,4($17)					\n\
		swc1	$f1,0($17)					\n\
.else									\n\
		swc1	$f0,0($17)					\n\
		swc1	$f1,4($17)					\n\
.endif									\n\
		b	end						\n\
		nop							\n\
rv1:									\n\
		li	$6,0x46		# rettype == 'F' ?		\n\
		bne	$16,$6,rv2					\n\
		nop							\n\
		swc1	$f0,0($17)					\n\
		b	end						\n\
		nop							\n\
.endif									\n\
rv2:									\n\
		li	$6,0x4a		# rettype == 'J' ?		\n\
		bne	$16,$6,rv3					\n\
		nop							\n\
		sw	$2,0($17)					\n\
		sw	$3,4($17)					\n\
		b	end						\n\
		nop							\n\
rv3:									\n\
		sw	$2,0($17)	# rettype == other		\n\
		nop							\n\
end:									\n\
		.set	reorder						\n\
	" :								\
	  : "r" ((scall)->nrargs),					\
	    "r" ((scall)->args),					\
	    "r" ((scall)->calltype),					\
	    "r" ((scall)->function),					\
	    "r" ((scall)->rettype),					\
	    "r" ((scall)->ret)						\
	  : "$2", "$3", "$4", "$5", "$6", "$7", 			\
		"$8", "$9", "$10", "$11", "$12", "$13", "$14",		\
		"$16", "$17", "$25");					\
} while (0)

struct _stackFrame;
extern void __mipsGetNextFrame(struct _stackFrame*);

#define	STACK_CURRENT_FRAME() \
	({ void* v ; asm("move %0,$fp" : "=r" (v)); v; })

#define	STACK_NEXT_FRAME(FM) \
	__mipsGetNextFrame(FM)

/*
 * Do an atomic compare and exchange.  The address 'A' is checked against
 * value 'O' and if they match it's exchanged with value 'N'.
 * We return '1' if the exchange is sucessful, otherwise 0.
 */
/*
 * GCC doesn't recognize LL or SC, so we need to be creative.
 */
#if !defined(HAVE_NO_COMPARE_AND_EXCHANGE) && !defined(COMPARE_AND_EXCHANGE)
#define	COMPARE_AND_EXCHANGE(A,O,N) \
({ \
	asm volatile(" \
		.set	noreorder    				\n\
		move	$9,%0    				\n\
		move	$11,%1    				\n\
		move	$10,%2    				\n\
		.word	0xc1280000	# ll	$8,0($9) 	\n\
		nop						\n\
		.word	0x150b0003	# bne	$8,$11,3 	\n\
		nop						\n\
		.word	0xe12a0000	# sc 	$10,0($9) 	\n\
		nop						\n\
		.set	reorder    				\n\
	 " : : "r" (A), "r" (O), "r" (N) : "$8","$9","$10","$11","memory"); \
	(*(A)) == (N) ? 1 : 0;					  \
})
#endif

/**
 * Force all register which need to be preserved to be saved.  Used in the
 * JNI code.
 */
#define FORCE_SAVE_REGISTERS() \
	asm volatile("" : : : "$16", "$17", "$18", "$19", "$20", "$21", "$22", "$23")

#endif
