/*
 * mips/jit.h
 * Common MIPS JIT configuration information.
 *
 * Copyright (c) 1996, 1997
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Written by Christian Krusel <chrkr@uni-paderborn.de>, 1997.
 * Modified by Michael Chen <mikey@hydra.stanford.edu>, 1998.
 * last changed 2.15.98
 */

#ifndef __mips_jit_h
#define __mips_jit_h

/**/
/* Exception handling information. */
/**/

/* Extract the object argument from given frame */
#define FRAMEOBJECT(f)							\
	(*(Hjava_lang_Object**)(f))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(F, H)					\
	asm volatile("move $fp,%0\n\tjr %1\n\tnop" : : "r" (F), "r" (H))

/**/
/* Method dispatch.  */
/**/

#define HAVE_TRAMPOLINE
struct Hjava_lang_Class;

/* The layout of this struct is know by inline assembly.  */

typedef struct _methodTrampoline {
	unsigned code[5];
	struct _methods *meth;
	struct Hjava_lang_Class* clazz;
	unsigned pad[1];
} methodTrampoline;

extern void mips_do_fixup_trampoline(void);

/* We must construct jump address since jal can not hold absolute address */
#define FILL_IN_TRAMPOLINE(t,m,c)					\
	do {								\
		uint32 pc = (unsigned int)mips_do_fixup_trampoline;	\
		(t)->code[0] = 0x001f1021;	/* addu $2,$31,$0 */	\
		(t)->code[1] = 0x3c190000 | (pc >> 16);	/* lui $25,addr(high) */ \
		(t)->code[2] = 0x37390000 | (pc & 0xffff);/* ori $25,$25,addr(low) */ \
		(t)->code[3] = 0x0320f809;	/* jalr $31,$25 */		\
		(t)->code[4] = 0x00000000;	/* nop */		\
		(t)->meth = (m);					\
		(t)->clazz = (c);					\
	} while (0)

#define FIXUP_TRAMPOLINE_DECL	Method *_meth, Hjava_lang_Class* _clazz
#define FIXUP_TRAMPOLINE_INIT	meth = _meth; \
				clazz = _clazz;

/**/
/* Register management information. */
/**/

#define	RIL	(Rint|Rref)
#define	RFD	(Rfloat|Rdouble)
#define	RG	(Rglobal|Rnosaveoncall)

/* Define the register set, bereits an mips angepasst*/
#define	REGISTER_SET							\
	{ /* i0 */	0, 0, RIL|Reserved,	0, 0, 0   },		\
	{ /* i1 */	0, 0, Reserved,		0, 0, 1   },		\
	{ /* i2 */	0, 0, RIL|Reserved,	0, 0, 2   },		\
	{ /* i3 */	0, 0, RIL|Reserved,	0, 0, 3   },		\
	{ /* i4 */	0, 0, RIL,		0, 0, 4   },		\
	{ /* i5 */	0, 0, RIL,		0, 0, 5   },		\
	{ /* i6 */	0, 0, RIL,		0, 0, 6   },		\
	{ /* i7 */	0, 0, RIL,		0, 0, 7   },		\
	{ /* i8 */	0, 0, RIL,		0, 0, 8   },		\
	{ /* i9 */	0, 0, RIL,		0, 0, 9   },		\
	{ /* i10 */	0, 0, RIL,		0, 0, 10   },		\
	{ /* i11 */	0, 0, RIL,		0, 0, 11   },		\
	{ /* i12 */	0, 0, RIL,		0, 0, 12   },		\
	{ /* i13 */	0, 0, RIL,		0, 0, 13   },		\
	{ /* i14 */	0, 0, RIL,		0, 0, 14   },		\
	{ /* i15 */	0, 0, RIL,		0, 0, 15   },		\
	{ /* i16 */	0, 0, RIL,		RG, 0, 16   },\
	{ /* i17 */	0, 0, RIL,		RG, 0, 17   },\
	{ /* i18 */	0, 0, RIL,		RG, 0, 18   },\
	{ /* i19 */	0, 0, RIL,		RG, 0, 19   },\
	{ /* i20 */	0, 0, RIL,		RG, 0, 20   },\
	{ /* i21 */	0, 0, RIL,		RG, 0, 21   },\
	{ /* i22 */	0, 0, RIL,		RG, 0, 22   },\
	{ /* i23 */	0, 0, RIL,		RG, 0, 23   },\
	{ /* i24 */	0, 0, RIL,		0, 0, 24   },		\
	{ /* i25 */	0, 0, RIL,		0, 0, 25   },		\
	{ /* i26 */	0, 0, Reserved,		0, 0, 26   },		\
	{ /* i27 */	0, 0, Reserved,		0, 0, 27   },		\
	{ /* i28/gp */	0, 0, Reserved,		0, 0, 28   },\
	{ /* i29 */	0, 0, Reserved, 	Rnosaveoncall, 0, 29   },\
	{ /* i30 */	0, 0, Reserved,		Rnosaveoncall, 0, 30   },\
	{ /* i31 */	0, 0, Reserved,		Rnosaveoncall, 0, 31   },\
	{ /* f0 */	0, 0, RFD|Reserved,	0, 0, 0   },		\
	{ /* f1 */	0, 0, Reserved,		0, 0, 1   },		\
	{ /* f2 */	0, 0, Reserved,		0, 0, 2   },		\
	{ /* f3 */	0, 0, Reserved,		0, 0, 3   },		\
	{ /* f4 */	0, 0, RFD,		0, 0, 4   },		\
	{ /* f5 */	0, 0, Reserved,		0, 0, 5   },		\
	{ /* f6 */	0, 0, RFD,		0, 0, 6   },		\
	{ /* f7 */	0, 0, Reserved,		0, 0, 7   },		\
	{ /* f8 */	0, 0, RFD,		0, 0, 8   },		\
	{ /* f9 */	0, 0, Reserved,		0, 0, 9   },		\
	{ /* f10 */	0, 0, RFD,		0, 0, 10   },		\
	{ /* f11 */	0, 0, Reserved,		0, 0, 11   },		\
	{ /* f12 */	0, 0, RFD,		0, 0, 12   },		\
	{ /* f13 */	0, 0, Reserved,		0, 0, 13   },		\
	{ /* f14 */	0, 0, RFD,		0, 0, 14   },		\
	{ /* f15 */	0, 0, Reserved,		0, 0, 15   },		\
	{ /* f16 */	0, 0, RFD,		0, 0, 16   },		\
	{ /* f17 */	0, 0, Reserved,		0, 0, 17   },		\
	{ /* f18 */	0, 0, RFD,		0, 0, 18   },		\
	{ /* f19 */	0, 0, Reserved,		0, 0, 19   },		\
	{ /* f20 */	0, 0, RFD,		0, 0, 20   },		\
	{ /* f21 */	0, 0, Reserved,		0, 0, 21   },		\
	{ /* f22 */	0, 0, RFD,		0, 0, 22   },		\
	{ /* f23 */	0, 0, Reserved,		0, 0, 23   },		\
	{ /* f24 */	0, 0, RFD,		0, 0, 24   },		\
	{ /* f25 */	0, 0, Reserved,		0, 0, 25   },		\
	{ /* f26 */	0, 0, RFD,		0, 0, 26   },		\
	{ /* f27 */	0, 0, Reserved,		0, 0, 27   },		\
	{ /* f28 */	0, 0, RFD,		0, 0, 28   },		\
	{ /* f29 */	0, 0, Reserved,		0, 0, 29   },		\
	{ /* f30 */	0, 0, RFD,		0, 0, 30   },		\
	{ /* f31 */	0, 0, Reserved,		0, 0, 31   },		\

/* Number of registers in the register set */
#define	NR_REGISTERS	64

/**/
/* We must setup the incoming argument state */
/**/
#define NR_ARGUMENTS	4

/**/
/* Number of global registers */
/**/
#define	NR_GLOBALS	8

/**/
/* Define to use the constant pool. */
/**/
#define	USE_JIT_CONSTANT_POOL	1

/**/
/* Define to use stack limit checking - give it the place to find the limit */
/**/
#define	STACK_LIMIT()	SLOT2ARGOFFSET(maxArgs)

/**/
/* Number of registers saved in the prologue. */
/**/
#define	REGISTERS_SAVED	12

/**/
/* Opcode generation. */
/**/

/* Extra label types */
#define Llong16b	(Larchdepend+0)
#define Llong16x16	(Larchdepend+1)
#define Llong26		(Larchdepend+2)
#define Lexception	(Larchdepend+3)
#define Llong16		(Larchdepend+4)

/* HACK XXX - We add in a little extra space to stop the JIT from breaking */
#define	EXTRASPACE	2

#define	FRAMESIZE \
	STACKALIGN(((maxLocal-maxArgs) + maxStack + maxTemp + \
		REGISTERS_SAVED + EXTRASPACE + max_args) * SLOTSIZE)

#define	LABEL_Lframe(P,V,L) \
	{ \
		int framesize = FRAMESIZE; \
		assert((framesize & 0xFFFFF000) == 0); \
		*(P) = (*(P) & 0xFFFF0000) | ((-framesize) & 0xFFFF); \
	}
#define	LABEL_Lnegframe(P,V,L) \
	{ \
		int framesize = FRAMESIZE; \
		assert((framesize & 0xFFFFF000) == 0); \
		*(P) = (*(P) & 0xFFFF0000) | ((framesize) & 0xFFFF); \
	}
#define	LABEL_Llong16b(P,V,L)	(P)[0]=((P)[0]&0xFFFF0000)|((((V)-4)>>2)&0xFFFF)
#define	LABEL_Llong16(P,V,L)	(P)[0]=((P)[0]&0xFFFF0000)|(((V))&0xFFFF)
#define LABEL_Llong16x16(P,V,L) (P)[1]=((P)[1]&0xFFFF0000)|((V) & 0xFFFF); \
                                (P)[0]=((P)[0]&0xFFFF0000)|(((V)>>16)&0xFFFF)
#if defined(USE_JIT_CONSTANT_POOL)
#define	LABEL_Lexception(P,V,L)	LABEL_Lframe(P,V,L); \
				LABEL_Llong16x16((P)+1,codebase,L)
#else
#define	LABEL_Lexception(P,V,L)	LABEL_Lframe(P,V,L)
#endif

#define	EXTRA_LABELS(P,D,L)						\
	case Lframe:		LABEL_Lframe(P,D,L);		break;	\
	case Lnegframe:		LABEL_Lnegframe(P,D,L);		break;	\
	case Llong16b:		LABEL_Llong16b(P,D,L);		break;	\
	case Llong16x16:	LABEL_Llong16x16(P,D,L);	break;	\
	case Lexception:	LABEL_Lexception(P,D,L);	break;	\
	case Llong16:		LABEL_Llong16(P,D,L);		break;


/**/
/* Slot management information. */
/**/

/* Size of each slot */
#define	SLOTSIZE		4

#define	FRAMEALIGN		16
#define	STACKALIGN(v)		(((v) + FRAMEALIGN - 1) & -FRAMEALIGN)

/* We push arguments low to high. */
#define	PUSHARG_FORWARDS	1

/*
 * A mips stack frame looks like:
 *
 *
 *      |  caller's frame               |
 *      |-------------------------------|
 *      |  incoming                     |
 *      |  arguments                    |
 * $fp->|-------------------------------|
 *      |  locals and                   |
 *      |  temporaries                  |
 *      |-------------------------------|
 *      |  callee-saved registers       |
 *      |                               |
 *      |-------------------------------|  -
 *      |  outgoing                     |  | max
 *      |  arguments                    |  | Push
 * $sp->|-------------------------------|  -
 *      |  callee's frame               |
 *
 */

/* We need to do some special initialization for the MIPS JIT */
struct _methods;
extern void __mipsInitJit(struct _methods*);
#define	INIT_JIT_MD(M)		__mipsInitJit(M)

/* Generate slot offset for an argument relative to the frame pointer */
#define SLOT2ARGOFFSET(N)	(argMap[N] * SLOTSIZE)

/* Generate slot offset for a local (non-argument) */
#define SLOT2LOCALOFFSET(N) \
	(-SLOTSIZE * (REGISTERS_SAVED + (N) - maxArgs + EXTRASPACE))

/* Generate slot offset for an argument push (relative to $sp) */
#define SLOT2PUSHOFFSET(N)	((N) * SLOTSIZE)

/* Wrap up a native call for the JIT */
#define KAFFEJIT_TO_NATIVE(_m)

/* On the mips we need to flush the data cache after generating new code */
#include <sys/cachectl.h>
#define FLUSH_DCACHE(start, end) _flush_cache((uintp)(start), (uintp)(end) - (uintp)(start), BCACHE)

/* Machine dependent NOP instruction, usually NULL */
#define I_NOP 		NULL

/* Patch trampoline to native code */
#define PATCH_TRAMPOLINE( _t, _tgt ) \
	(_t)->code[0] = 0x3c190000 | (((uint)(_tgt)) >> 16); /* lui $t9,high(where) */ \
	(_t)->code[1] = 0x37390000 | (((uint)(_tgt)) & 0xffff); /* ori $t9,$t9,low(where) */ \
	(_t)->code[2] = 0x03200008; /* jr $t9 */ \
	(_t)->code[3] = 0x00000000; /* nop */

/* Patch call to native code */
#define PATCH_CALL( _p, _tgt )

/* Argument register mapping */
extern int* argMap;
extern int totalMaxArgs;
extern int max_args;

extern int dbmsg;

#define TRACE_METHOD_ARGS() \
        int found_int = 0; \
        int fp_idx = 0; \
        char * p = 0; \
        if (!METHOD_IS_STATIC(meth)) { \
                kprintf("\t0x%p %s ", *(void**)args, ((Hjava_lang_Object*)(*(void**)args))->dtable->class->name); \
                args += SLOTSIZE; \
                found_int = 1; \
        } \
        if (strlen(meth->signature) > 0 && METHOD_IS_STATIC(meth)) kprintf("\t"); \
        for (p = meth->signature; *p != 0; p++) { \
                switch (*p) { \
                case '(': \
                        continue; \
                        break; \
                case ')': \
                        break; \
                case 'F': \
                        if (!found_int && fp_idx < 2) { \
                                int v = 0; \
                                if (fp_idx == 0) { \
                                        asm volatile("mfc1 %0, $f12" : : "r" (v)); \
                                } else { \
                                        asm volatile("mfc1 %0, $f14" : : "r" (v)); \
                                } \
                                kprintf("%e ",v); \
                        } else { \
                                kprintf("%e ",*(float*)args); \
                        } \
                        args += SLOTSIZE; \
                        fp_idx++; \
                        break; \
                case 'I': case 'Z': case 'S': \
                case 'B': case 'C': \
                        kprintf("%d ",*(int*)args); \
                        args += SLOTSIZE; \
                        found_int = 1; \
                        break; \
                case 'D': \
                        args += ((uintp)args & 0x4) ? SLOTSIZE : 0; \
                        if (!found_int && fp_idx < 2) { \
                                int v1 = 0; int v2 = 0; \
                                if (fp_idx == 0) { \
                                        asm volatile("mfc1 %0, $f12\n mfc1 %1, $f13\n" : : "r" (v1), "r" (v2)); \
                                } else { \
                                        asm volatile("mfc1 %0, $f14\n mfc1 %1, $f15\n" : : "r" (v1), "r" (v2)); \
                                } \
                                kprintf("%x-%x ",v1,v2); \
                        } else { \
                                kprintf("%x-%x ",*(void**)args,*(void**)(args+4)); \
                        } \
                        args += 2*SLOTSIZE; \
                        fp_idx++; \
                        break; \
                case 'J': \
                        args += ((uintp)args & 0x4) ? SLOTSIZE : 0; \
                        kprintf("%x-%x ",*(void**)args,*(void**)(args+4)); \
                        args += 2*SLOTSIZE; \
                        found_int = 1; \
                        break; \
                case '[': \
		arrayofarray: \
			p++; \
			if( *p == 'L' ){ \
				goto object; \
			} \
			else if( *p == '[' ){ \
				goto arrayofarray; \
			} \
                        kprintf("0x%p ",*(void**)args); \
                        args += SLOTSIZE; \
                        found_int = 1; \
                        break; \
                case 'L': \
		object: \
                        kprintf("0x%p ",*(void**)args); \
                        args += SLOTSIZE; \
                        found_int = 1; \
			while (*p != ';') { \
				p++; \
	  		} \
                        break; \
                default: \
                        ABORT(); \
                } \
                if( *p == ')' ){ \
			break; \
		} \
        } \
        if (strlen(meth->signature) > 0 || !METHOD_IS_STATIC(meth)) kprintf("\n")

#define TRACE_METHOD_END() \
	switch( type ){ \
	case 'I': case 'Z': case 'S': \
	case 'B': case 'C': \
		kprintf( "E %d\n", rv ); break; \
	case 'L': case '[': \
		kprintf( "E 0x%p\n", rv ); break; \
	case 'F': \
		kprintf( "E %e\n", rv ); break; \
	case 'J': \
		kprintf( "E %lld\n", rv ); break; \
	case 'D': \
		kprintf( "E %llf\n", rv ); break; \
	case 'V': \
		kprintf( "E\n" ); break; \
	}

#endif
