/*
 * sparc/common.h
 * Common SPARC configuration information.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __sparc_common_h
#define __sparc_common_h

#define	sysdepCallMethod(CALL) do {					\
  /* allocates variable arrays one word past the minimum stack frame */	\
  int extraargs[((CALL)->nrargs>6)?((CALL)->nrargs-6):0];		\
  int *res = 0;								\
  switch((CALL)->nrargs) {						\
    register int o0 asm("o0");						\
    register int o1 asm("o1");						\
    register int o2 asm("o2");						\
    register int o3 asm("o3");						\
    register int o4 asm("o4");						\
    register int o5 asm("o5");						\
    register double f0d asm("f0");					\
    register float f0f asm("f0");					\
									\
  default:								\
    {									\
      int *args = extraargs-1;						\
      int argidx = 6;							\
      for(; argidx < (CALL)->nrargs; ++argidx, ++args) {		\
	*args = (CALL)->args[argidx].i;					\
      }									\
    }									\
  case 6:								\
    o5 = (CALL)->args[5].i;						\
  case 5:								\
    o4 = (CALL)->args[4].i;						\
  case 4:								\
    o3 = (CALL)->args[3].i;						\
  case 3:								\
    o2 = (CALL)->args[2].i;						\
  case 2:								\
    o1 = (CALL)->args[1].i;						\
  case 1:								\
    o0 = (CALL)->args[0].i;						\
  case 0:								\
    asm ("call %2,0\n							\
          nop\n"							\
	: "=r" (o0), "=r" (o1)						\
	: "r" ((CALL)->function),					\
	  "0" (o0), "1" (o1), "r" (o2), "r" (o3), "r" (o4), "r" (o5)	\
	: "g1", "g2", "g3", "g4", "o7", "cc"				\
	);								\
    if ((CALL)->retsize != 0)						\
      res = &(CALL)->ret->i;						\
    switch((CALL)->retsize) {						\
    case 2:								\
      if ((CALL)->rettype == 'D')					\
        *(double*)res = f0d;						\
      else {								\
        res[1] = o1;							\
        res[0] = o0;							\
      }									\
      break;								\
    case 1:								\
      if ((CALL)->rettype == 'F')					\
        *(float*)res = f0f;						\
      else								\
        *res = o0;							\
      break;								\
    case 0:								\
      break;								\
    default:								\
      ABORT();								\
      break;								\
    }									\
    break;								\
  }									\
} while (0)

/*
 * On the sparc we need to flush the data cache after generating new code
 */
#define FLUSH_DCACHE(beg, end)                                          \
        do {                                                            \
                int *_p = (int *)(beg), *_e = (int *)(end);             \
                while (_p < _e) {                                       \
                        asm volatile("iflush %0" : : "r"(_p++));        \
                }                                                       \
        } while (0)


/*
 * Stack format.
 */

#include "../../kaffe/kaffevm/gtypes.h"

typedef struct __machineStackFrame {
	uintp   lcl[8];
	uintp   arg[6];
	void*   retbp;
	void*   retpc;
	uintp   retstruct;
	uintp   callargs[6];
} machineStackFrame;

#define	STACK_FRAME(F)		((F)->retbp)
#define	STACK_PC(F)		((F)->retpc)
#define STACK_CURRENT_FRAME()   ({ void* v ; asm("ta 3\n\tmov %%fp,%0" : "=r" (v)); v; })

/*
 * Do an atomic exchange.  The contents of address 'A' is exchanged
 * with value 'N'.
 */
#define ATOMIC_EXCHANGE(A,N) \
	asm volatile("swap [%2],%0" : "=r" (N) : "0" (N), "r" (A) : "cc", "memory" );

#if 0
#define	COMPARE_AND_EXCHANGE(A,O,N) \
	({ \
		char ret; \
		asm volatile(" \
			mov %1, %%g1 \n\
			mov %2, %%g2 \n\
			mov %3, %%g3 \n\
			!# cas [g1], g3, g2 # \n\
			.word 0xc5e05003 \n\
			cmp %%g2, %3 \n\
			!# cmove 1,g2 # \n\
			.word 0x85646001 \n\
			!# cmovne 0,g2 # \n\
			.word 0x85666000 \n\
			mov %%g2, %0 \n\
			" \
		  : "=r" (ret) : "r" (A), "r" (N), "r" (O) : "%g1", "%g2", "%g3", "cc", "memory" ); \
		(ret); \
	}) /* special gcc syntax: last exp evaluated (ret) is the value of the expression */
#endif

/**
 * Force all register which need to be preserved to be saved.  Used in the
 * JNI code.
 */
#define FORCE_SAVE_REGISTERS() \
        asm volatile("" : : : "i0", "i1", "i2", "i3", "i4", "i5", "fp", "i7", "l0", "l1", "l2", "l3", "l4", "l5", "l6", "l7" )

#endif
