/*
 * sparc/jit.h
 * Common SPARC JIT configuration information.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __sparc_jit_h
#define __sparc_jit_h

/**/
/* Exception handling information. */
/**/

/* Extract the object argument from given frame */
#define FRAMEOBJECT(f)							\
	(*(Hjava_lang_Object**)(((machineStackFrame*)f)->retbp+68))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(F, H)					\
	asm volatile("							\n\
		mov %1,%%o2						\n\
		mov %0,%%o3						\n\
		ta 3							\n\
		sub %%sp,64,%%sp					\n\
		mov %%o3,%%fp						\n\
		jmpl %%o2,%%g0						\n\
		restore	%%o1,0,%%o0					\n\
	" : : "r" (F), "r" (H) : "o2", "o3")


/**/
/* Method dispatch.  */
/**/

#define HAVE_TRAMPOLINE

/* The layout of this struct is know by inline assembly.  */

typedef struct _methodTrampoline {
	unsigned code[3];
	struct _methods *meth;
	struct Hjava_lang_Class* clazz;
} methodTrampoline;

extern void sparc_do_fixup_trampoline(void);

#define FILL_IN_TRAMPOLINE(t,m,c)					\
	do {								\
		(t)->code[0] = 0x8210000f;	/* mov %o7, %g1 */	\
		/* call sparc_do_fixup_trampoline */			\
		(t)->code[1] = ((size_t)sparc_do_fixup_trampoline	\
				- ((size_t)(t) + 4)) / 4 | 0x40000000;	\
		(t)->code[2] = 0x01000000;	/* nop */		\
		(t)->meth = (m);					\
		(t)->clazz = (c);					\
	} while (0)

#define FIXUP_TRAMPOLINE_DECL	Method *_meth, Hjava_lang_Class* _clazz
#define FIXUP_TRAMPOLINE_INIT	meth = _meth; \
				clazz = _clazz;

/**/
/* Register management information. */
/**/
#define	RIL	(Rint|Rref)
#define	RFD	(Rfloat|Rdouble)
#define	RG	(Rglobal|Rnosaveoncall)

/* Define the register set */
#define	REGISTER_SET							\
	{ /* g0 */	0, 0, Reserved,		0, 0, 0   },		\
	{ /* g1 */	0, 0, Reserved,		0, 0, 1   },		\
	{ /* g2 */	0, 0, Reserved,		0, 0, 2   },		\
	{ /* g3 */	0, 0, Reserved,		0, 0, 3   },		\
	{ /* g4 */	0, 0, Reserved,		0, 0, 4   },		\
	{ /* g5 */	0, 0, Reserved,		0, 0, 5   },		\
	{ /* g6 */	0, 0, Reserved,		0, 0, 6   },		\
	{ /* g7 */	0, 0, Reserved,		0, 0, 7   },		\
	{ /* o0 */	0, 0, Reserved|RIL, 	0, 0, 8   },		\
	{ /* o1 */	0, 0, Reserved|RIL, 	0, 0, 9   },		\
	{ /* o2 */	0, 0, Reserved,		0, 0, 10   },		\
	{ /* o3 */	0, 0, Reserved,		0, 0, 11   },		\
	{ /* o4 */	0, 0, Reserved,		0, 0, 12   },		\
	{ /* o5 */	0, 0, Reserved,		0, 0, 13   },		\
	{ /* o6 */	0, 0, Reserved,		0, 0, 14   },		\
	{ /* o7 */	0, 0, Reserved,		0, 0, 15   },		\
	{ /* l0 */	0, 0, RIL,		RG, 0, 16   },		\
	{ /* l1 */	0, 0, RIL,		RG, 0, 17   },		\
	{ /* l2 */	0, 0, RIL,		RG, 0, 18   },		\
	{ /* l3 */	0, 0, RIL,		RG, 0, 19   },		\
	{ /* l4 */	0, 0, RIL,		RG, 0, 20   },		\
	{ /* l5 */	0, 0, RIL,		RG, 0, 21   },		\
	{ /* l6 */	0, 0, RIL,		RG, 0, 22   },		\
	{ /* l7 */	0, 0, RIL,		RG, 0, 23   },		\
	{ /* i0 */	0, 0, RIL,		Rnosaveoncall, 0, 24   },		\
	{ /* i1 */	0, 0, RIL,		Rnosaveoncall, 0, 25   },		\
	{ /* i2 */	0, 0, RIL,		Rnosaveoncall, 0, 26   },		\
	{ /* i3 */	0, 0, RIL,		Rnosaveoncall, 0, 27   },		\
	{ /* i4 */	0, 0, RIL,		Rnosaveoncall, 0, 28   },		\
	{ /* i5 */	0, 0, RIL,		Rnosaveoncall, 0, 29   },		\
	{ /* i6 */	0, 0, Reserved,		0, 0, 30   },		\
	{ /* i7 */	0, 0, Reserved,		0, 0, 31   },		\
	{ /* f0 */	0, 0, RFD,		0, 0, 0   },		\
	{ /* f1 */	0, 0, Reserved,		0, 0, 1   },		\
	{ /* f2 */	0, 0, RFD,		0, 0, 2   },		\
	{ /* f3 */	0, 0, Reserved,		0, 0, 3   },		\
	{ /* f4 */	0, 0, RFD,		0, 0, 4   },		\
	{ /* f5 */	0, 0, Reserved,		0, 0, 5   },		\
	{ /* f6 */	0, 0, RFD,		0, 0, 6   },		\
	{ /* f7 */	0, 0, Reserved,		0, 0, 7   },		\
	{ /* f8 */	0, 0, RFD,		0, 0, 8   },		\
	{ /* f9 */	0, 0, Reserved,		0, 0, 9   },		\
	{ /* f10 */	0, 0, RFD,		0, 0, 10   },		\
	{ /* f11 */	0, 0, Reserved,		0, 0, 11   },		\
	{ /* f12 */	0, 0, RFD,		0, 0, 12   },		\
	{ /* f13 */	0, 0, Reserved,		0, 0, 13   },		\
	{ /* f14 */	0, 0, RFD,		0, 0, 14   },		\
	{ /* f15 */	0, 0, Reserved,		0, 0, 15   },		\
	{ /* f16 */	0, 0, RFD,		0, 0, 16   },		\
	{ /* f17 */	0, 0, Reserved,		0, 0, 17   },		\
	{ /* f18 */	0, 0, RFD,		0, 0, 18   },		\
	{ /* f19 */	0, 0, Reserved,		0, 0, 19   },		\
	{ /* f20 */	0, 0, RFD,		0, 0, 20   },		\
	{ /* f21 */	0, 0, Reserved,		0, 0, 21   },		\
	{ /* f22 */	0, 0, RFD,		0, 0, 22   },		\
	{ /* f23 */	0, 0, Reserved,		0, 0, 23   },		\
	{ /* f24 */	0, 0, RFD,		0, 0, 24   },		\
	{ /* f25 */	0, 0, Reserved,		0, 0, 25   },		\
	{ /* f26 */	0, 0, RFD,		0, 0, 26   },		\
	{ /* f27 */	0, 0, Reserved,		0, 0, 27   },		\
	{ /* f28 */	0, 0, RFD,		0, 0, 28   },		\
	{ /* f29 */	0, 0, Reserved,		0, 0, 29   },		\
	{ /* f30 */	0, 0, RFD,		0, 0, 30   },		\
	{ /* f31 */	0, 0, Reserved,		0, 0, 31   },		\

/* Number of registers in the register set */
#define	NR_REGISTERS	64

/**/
/* We must setup the incoming argument state */
/**/
#define NR_ARGUMENTS	6

/**/
/* Number of global registers */
/**/
#define	NR_GLOBALS	8

/**/
/* Opcode generation. */
/**/

/* Extra label types */
#define Llong30		(Larchdepend+0)	/* Label is 30 bits long */
#define Llong22		(Larchdepend+1)	/* Label is 22 bits long */
#define Llong22x10	(Larchdepend+2)	/* Label is split into 22 and 10 bits */

#define	LABEL_Lframe(P,V,L) \
	{ \
		int framesize = STACKALIGN( \
			STACKALIGN((maxPush < NR_ARGUMENTS ? NR_ARGUMENTS : maxPush) * SLOTSIZE) + \
			STACKALIGN(((maxLocal-maxArgs)+maxStack+maxTemp) * \
			SLOTSIZE) + WINDOWSIZE + WINDOWEXTRA); \
		assert((framesize & 0xFFFFF000) == 0); \
		*(P) = (*(P) & 0xFFFFE000) | ((-framesize) & 0x1FFF); \
	}
#define	LABEL_Llong30(P,V,L)	(P)[0] = (((P)[0] & 0xC0000000) | (((V) / 4) & 0x3FFFFFFF))
#define	LABEL_Llong22(P,V,L)	(P)[0] = (((P)[0] & 0xFFC00000) | (((V) / 4) & 0x003FFFFF))
#define	LABEL_Llong22x10(P,V,L)	(P)[0] = (((P)[0] & 0xFFC00000) | (((V) >> 10) & 0x003FFFFF)); \
				(P)[1] = (((P)[1] & 0xFFFFFC00) | ((V) & 0x000003FF));

#define EXTRA_LABELS(P,D,L)						\
	case Lframe:		LABEL_Lframe(P,D,L);		break;	\
	case Llong30:		LABEL_Llong30(P,D,L);		break;	\
	case Llong22:		LABEL_Llong22(P,D,L);		break;	\
	case Llong22x10:	LABEL_Llong22x10(P,D,L);	break;

/**/
/* Slot management information. */
/**/

/* Size of each slot */
#define	SLOTSIZE		4

/*
 * A stack frame looks like:
 * 
 * %fp->|                               |
 *      |-------------------------------|
 *      |  Locals, temps, saved floats  |
 *      |-------------------------------|
 *      |  outgoing parameters past 6   |
 *      |-------------------------------|-\
 *      |  6 words for callee to dump   | |
 *      |  register arguments           | |
 *      |-------------------------------|  > minimum stack frame
 *      |  One word struct-ret address  | |
 *      |-------------------------------| |
 *      |  16 words to save IN and      | |
 * %sp->|  LOCAL register on overflow   | |
 *      |-------------------------------|-/
 */

#define	FRAMEALIGN		8
#define	STACKALIGN(v)		(((v) + FRAMEALIGN - 1) & -FRAMEALIGN)

#define	WINDOWSIZE		(16*SLOTSIZE)
#define	INARGSIZE		(6*SLOTSIZE)
#define	WINDOWEXTRA		(5*SLOTSIZE)

/* Generate slot offset for an argument */
#define SLOT2ARGOFFSET(_n) \
	(WINDOWSIZE + SLOTSIZE + SLOTSIZE * (_n))

/* Generate slot offset for a local (non-argument) */
#define SLOT2LOCALOFFSET(_n) \
	(-(WINDOWEXTRA + SLOTSIZE * ((_n) - maxArgs) ) )

/* Generate the slot offset to the stack limit */
#define	STACK_LIMIT()	SLOT2ARGOFFSET(maxArgs)

/* Generate slot offset for an push */
#define SLOT2PUSHOFFSET(_n) \
	(WINDOWSIZE + SLOTSIZE + SLOTSIZE * (_n))

/* Wrap up a native call for the JIT */
#define KAFFEJIT_TO_NATIVE(_m)

#define I_NOP		0x01000000

#define PATCH_TRAMPOLINE( _t, _tgt ) \
	(_t)->code[0] = 0x03000000 | (((uint)(_tgt)) >> 10 ); /* sethi hi(where), %g1 */ \
	(_t)->code[1] = 0x81c06000 | (((uint)(_tgt)) & 0x3ff ); /* jmpl %g1, imm10(where), %g0 */ \
	(_t)->code[2] = 0x01000000; /* nop */

/* Patch call to native code */
#define PATCH_CALL( _p, _tgt ) \
	*(uint*)(_p)->where = 0x40000000 | (( ( (_tgt) - (_p)->where ) / 4 ) & 0x3fffffff)

#if 0
#define TRACE_METHOD_ARGS() \
	char * p = 0; \
	args += 68; /* increment past window spill area */ \
        if (!METHOD_IS_STATIC(meth)) { \
		void * obj = *(void**)args; \
                kprintf( "\t0x%p (%s) ", obj, \
                        obj == NULL ? "???" : ((Hjava_lang_Object*)(*(void**)args))->dtable->class->name ); \
                args += SLOTSIZE; \
        } \
        if( strlen(meth->signature) > 0 && METHOD_IS_STATIC(meth)) kprintf("\t"); \
        for( p = meth->signature; *p != 0; p++ ){ \
		int64 lval = 0; \
                switch( *p ){ \
                case '(': \
                        continue; \
                        break; \
                case ')': \
                        break; \
                case 'F': \
                        kprintf("%e ",*(float*)args); \
                        args += SLOTSIZE; \
                        break; \
                case 'I': case 'Z': case 'S': \
                case 'B': case 'C': \
                        kprintf("%d ",*(int*)args); \
                        args += SLOTSIZE; \
                        break; \
                case 'D': \
			lval = ((int64)((void**)args)[0] << 32) | (int64)((void**)args)[1]; \
                        kprintf("%e ",lval); \
                        kprintf("%x-%x ",*(void**)args,*(void**)(args+4)); \
                        args += 2*SLOTSIZE; \
                        break; \
                case 'J': \
			lval = ((int64)((void**)args)[0] << 32) | (int64)((void**)args)[1]; \
                        kprintf("%lld ",lval); \
                        args += 2*SLOTSIZE; \
                        break; \
                case '[': \
                arrayofarray: \
			p++; \
                        if( *p == 'L' ){ \
                                goto object; \
                        } \
                        else if( *p == '[' ){ \
                                goto arrayofarray; \
                        } \
                        kprintf("0x%p ",*(void**)args); \
                        args += SLOTSIZE; \
                        break; \
                case 'L': \
                object: \
                        while( *p != ';' ){ \
                                p++; \
                        } \
                        kprintf("0x%p ",*(void**)args); \
                        args += SLOTSIZE; \
                        break; \
                default: \
                        ABORT(); \
                } \
                if( *p == ')' ){ \
			break; \
		} \
        } \
        if( strlen(meth->signature) > 0 || !METHOD_IS_STATIC(meth)) kprintf("\n")

#define TRACE_METHOD_END() \
	switch( type ){ \
	case 'I': case 'Z': case 'S': \
	case 'B': case 'C': \
		kprintf( "E %d\n", rv ); break; \
	case 'L': case '[': \
		kprintf( "E 0x%p\n", rv ); break; \
	case 'F': \
		kprintf( "E %e\n", rv ); break; \
	case 'J': \
		kprintf( "E %lld\n", rv ); break; \
	case 'D': \
		kprintf( "E %llf\n", rv ); break; \
	case 'V': \
		kprintf( "E\n" ); break; \
	}
#endif

#endif
