/*
 * superh/common.h
 * Common Super-H configuration information.
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __superh_common_h
#define __superh_common_h

#include <stddef.h>

#define	sysdepCallMethod(CALL) do {					\
  int extraargs[((CALL)->nrargs>4)?((CALL)->nrargs-4):0];		\
  switch((CALL)->nrargs) {						\
    register int r0 asm("r4");						\
    register int r1 asm("r5");						\
    register int r2 asm("r6");						\
    register int r3 asm("r7");						\
  default:								\
    {									\
      int *args = extraargs;						\
      int argidx = 4; 							\
      if ((CALL)->callsize[3] == 2) args++;				\
      for(; argidx < (CALL)->nrargs; ++argidx) {			\
	if ((CALL)->callsize[argidx]) { 				\
	  *args++ = (CALL)->args[argidx].i;				\
	  if ((CALL)->callsize[argidx] == 2) 				\
	    *args++ = ((CALL)->args[argidx].j) >> 32;			\
	} 								\
      }									\
    }									\
  case 4:								\
    if ((CALL)->callsize[3]) {						\
      r3 = (CALL)->args[3].i;						\
      if ((CALL)->callsize[3] == 2)					\
        *extraargs = ((CALL)->args[3].j) >> 32; 			\
    }									\
  case 3:								\
    if ((CALL)->callsize[2]) {						\
      r2 = (CALL)->args[2].i;						\
      if ((CALL)->callsize[2] == 2)					\
        r3 = ((CALL)->args[2].j) >> 32; 				\
    }									\
  case 2:								\
    if ((CALL)->callsize[1]) {						\
      r1 = (CALL)->args[1].i;						\
      if ((CALL)->callsize[1] == 2)					\
        r2 = ((CALL)->args[1].j) >> 32; 				\
    }									\
  case 1:								\
    if ((CALL)->callsize[0]) {						\
      r0 = (CALL)->args[0].i;						\
      if ((CALL)->callsize[0] == 2)					\
        r1 = ((CALL)->args[0].j) >> 32; 				\
    }									\
  case 0:								\
    asm volatile ("jsr @%2 ; nop ; mov r0,r4 ; mov r1,r5\n"		\
        : "=r" (r0), "=r" (r1)						\
	: "r" ((CALL)->function),					\
	  "0" (r0), "1" (r1), "r" (r2), "r" (r3) 			\
	: "r14"				 				\
	);								\
    switch((CALL)->rettype) {                                           \
    case 'D':                                                           \
    case 'J':                                                           \
        (&(CALL)->ret->i)[1] = r1;                                      \
        /* follow through */                                            \
    default:                                                            \
        (CALL)->ret->i = r0;                                            \
    }                                                                   \
    break;								\
  }									\
} while (0)

#define	STACK_CURRENT_FRAME() \
	({ void* v ; asm volatile("mov $r14,%0" : "=r" (v)) ; v; })

struct _stackFrame;
extern void __superhGetNextFrame(struct _stackFrame*);

#define	STACK_NEXT_FRAME(FM) \
	__superhGetNextFrame(FM)

/**
 * Force all register which need to be preserved to be saved.  Used in the
 * JNI code.
 */
#define FORCE_SAVE_REGISTERS() \
	asm volatile("" : : : "r8", "r9", "r10", "r11", "r12", "r13")


/**
 * We have test and set locking.
 */
#define	TEST_AND_SET(A) \
	({ \
		int val; \
		asm volatile ("tas.b @%1; movt %0" : "=r" (val) : "r" (A) : "memory"); \
		val; \
	})

#endif
