/* superh/jit-icode.h
 * Define the instructions which are present on the Super-H.
 *
 * Copyright (c) 2001 Transvirtual Technologies, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef __superh_jit_icode_h
#define __superh_jit_icode_h

/*
 * Size of long's compared to references.
 */
#define pusharg_long_idx_inc            2

/*
 * These must be defined for any architecture.
 */
#define	HAVE_spill_int			spill_Rxx
#define	HAVE_reload_int			reload_Rxx
#define	HAVE_spill_ref			spill_Rxx
#define	HAVE_reload_ref			reload_Rxx

#define	HAVE_move_register_int		movereg_RR
#define	HAVE_move_register_ref		movereg_RR

#define	HAVE_prologue			prologue_xxC
#define	HAVE_epilogue			epilogue_xxx
#define HAVE_exception_prologue		eprologue_xxx

#define	HAVE_move_int_const_rangecheck(v)	((v) >= -128 && (v) <= 127)
#define	HAVE_move_ref_const_rangecheck(v)	((v) == 0)

#define	HAVE_move_int_const		move_RxC
#define	HAVE_move_ref_const		move_RxC
#define	HAVE_move_int			move_RxR
#define	HAVE_move_label_const		move_RxL
#define	HAVE_move_ref			move_RxR
#define	HAVE_move_any			move_RxR

#if defined(USE_JIT_INLINE_CONSTANT_POOL)
/*
 * Load constant from the constant pool directly, without intermediate
 * register.
 */
#define	HAVE_load_constpool_ref		ld_RxL
#define	HAVE_load_constpool_int		ld_RxL
#endif

#define	HAVE_add_int			add_RRR
#define	HAVE_sub_int			sub_RRR
#define	HAVE_and_int			and_RRR
#define	HAVE_or_int			or_RRR
#define	HAVE_xor_int			xor_RRR
#undef	HAVE_nor_int
#define	HAVE_ashr_int			ashr_RRR
#define	HAVE_lshr_int			lshr_RRR
#define	HAVE_lshl_int			lshl_RRR


#define	HAVE_add_ref			add_RRR
#define	HAVE_sub_ref			sub_RRR

#define	HAVE_load_int			load_RxR
#define	HAVE_store_int			store_xRR
#define	HAVE_load_ref			load_RxR
#define	HAVE_store_ref			store_xRR

#define	HAVE_pusharg_int		push_xRC
#define	HAVE_pusharg_ref		push_xRC
#undef	HAVE_pusharg_long
#define	HAVE_popargs			popargs_xxC

#undef	HAVE_cmp_int
#undef	HAVE_cmp_int_const
#undef	HAVE_cmp_ref
#undef	HAVE_cmp_ref_const

#define	HAVE_branch			branch_xCC
#define	HAVE_branch_indirect		branch_indirect_xRC
#undef	HAVE_call_ref
#define	HAVE_call			call_xRC
#define	HAVE_ret			ret_xxx
#define	HAVE_return_int			return_Rxx
#define	HAVE_return_long		returnl_Rxx
#define	HAVE_return_ref			return_Rxx
#define	HAVE_returnarg_int		returnarg_xxR
#define	HAVE_returnarg_long		returnargl_xxR
#define	HAVE_returnarg_ref		returnarg_xxR

#define	HAVE_set_label			set_label_xxC
#define	HAVE_build_key			set_word_xxC
#define	HAVE_build_code_ref		set_wordpc_xxC

/*
 * These are sometimes optional (if long operators are defined)
 */
#define	HAVE_carry_clear		carryclear_xxx
#define	HAVE_adc_int			adc_RRR
#define	HAVE_sbc_int			sbc_RRR
#undef 	HAVE_ngc_int

/*
 * These are optional but help to optimise the code generated.
 */
#define	HAVE_add_int_const_rangecheck(v)	((v) >= -128 && (v) <= 127)
#define	HAVE_add_int_const		add_RRC

#undef	HAVE_add_ref_const
#undef	HAVE_sub_int_const
#undef  HAVE_cmp_int_const
#undef  HAVE_cmp_ref_const
#define	HAVE_load_offset_int		load_RRC
#define	HAVE_load_offset_ref		load_RRC
#undef	HAVE_load_offset_routine
#define	HAVE_store_offset_int		store_xRRC
#define	HAVE_store_offset_ref		store_xRRC
#undef	HAVE_lshl_int_const

#undef	HAVE_add_ref_const_rangecheck
#undef	HAVE_sub_int_const_rangecheck
#undef	HAVE_cmp_int_const_rangecheck
#define	HAVE_load_offset_int_rangecheck(v)	(1) /*((v) >= 0 && (v) <= 60)*/
#define	HAVE_load_offset_ref_rangecheck(v)	(1) /*((v) >= 0 && (v) <= 60)*/
#undef	HAVE_load_offset_routine_rangecheck
#define	HAVE_store_offset_int_rangecheck(v)	(1) /*((v) >= 0 && (v) <= 60)*/
#define	HAVE_store_offset_ref_rangecheck(v)	(1) /*((v) >= 0 && (v) <= 60)*/
#undef	HAVE_lshl_int_const_rangecheck

#define	HAVE_load_byte			loadb_RxR
#define	HAVE_load_char			loadc_RxR
#define	HAVE_load_short			loads_RxR
#define	HAVE_store_byte			storeb_RRx
#define	HAVE_store_char			stores_RRx
#define	HAVE_store_short		stores_RRx

#define HAVE_cbranch_int		cbranch_xRRCC
#undef HAVE_cbranch_int_const
#undef HAVE_cbranch_int_const_rangecheck
#define HAVE_cbranch_ref			cbranch_xRRCC
#undef HAVE_cbranch_ref_const
#undef HAVE_cbranch_ref_const_rangecheck

#define	HAVE_make_align			makealign_xxC

/*
 * These are optional if the architecture supports them.
 */

#define	HAVE_and_int_const_rangecheck(v)
#define	HAVE_ashr_int_const_rangecheck(v)
#define	HAVE_lshr_int_const_rangecheck(v)

#undef	HAVE_move_long_const

#undef	HAVE_swap_int
#undef	HAVE_swap_any
#define	HAVE_neg_int			neg_RRR
#undef	HAVE_mul_int
#undef	HAVE_div_int
#undef	HAVE_rem_int
#undef	HAVE_mul_int_const
#undef	HAVE_and_int_const
#undef	HAVE_ashr_int_const
#undef	HAVE_lshr_int_const

#undef	HAVE_move_long_const
#undef	HAVE_move_long

#undef	HAVE_add_long
#undef	HAVE_sub_long
#undef	HAVE_mul_long
#undef	HAVE_div_long
#undef	HAVE_rem_long
#undef	HAVE_neg_long
#undef	HAVE_and_long
#undef	HAVE_or_long
#undef	HAVE_xor_long
#undef	HAVE_ashr_long
#undef	HAVE_lshl_long
#undef	HAVE_lshr_long
#undef	HAVE_add_long_const
#undef	HAVE_sub_long_const
#undef	HAVE_and_long_const
#undef	HAVE_lshl_long_const

#undef	HAVE_load_long
#undef	HAVE_store_long
#undef	HAVE_load_offset_long
#undef	HAVE_store_offset_long

#undef	HAVE_pusharg_int_const

#undef	HAVE_cmp_long

#undef	HAVE_cvt_int_long
#undef	HAVE_cvt_int_byte
#undef	HAVE_cvt_int_char
#undef	HAVE_cvt_int_short
#undef	HAVE_cvt_long_int

#define HAVE_push_frame
#define HAVE_pop_frame
#undef HAVE_get_arg_ptr

#undef	HAVE_fakecall_constpool

#if !defined(HAVE_NO_FLOATING_POINT)

#define	HAVE_spill_float
#define	HAVE_reload_float
#define	HAVE_spill_double
#define	HAVE_reload_double

#define	HAVE_move_float
#define	HAVE_move_double

#undef	HAVE_add_float
#undef	HAVE_sub_float
#undef	HAVE_mul_float
#undef	HAVE_div_float

#undef	HAVE_add_double
#undef	HAVE_sub_double
#undef	HAVE_mul_double
#undef	HAVE_div_double

#define	HAVE_load_float
#define	HAVE_store_float
#define	HAVE_load_double
#define	HAVE_store_double

#define	HAVE_pusharg_float
#define	HAVE_pusharg_double

#define	HAVE_return_float
#define	HAVE_return_double

#define	HAVE_returnarg_float
#define	HAVE_returnarg_double

#define	HAVE_cvt_int_double
#define	HAVE_cvt_float_int
#define	HAVE_cvt_double_int
#define	HAVE_cvt_int_float
#define	HAVE_cvt_float_double
#define	HAVE_cvt_double_float

#undef	HAVE_move_float_const
#undef	HAVE_move_double_const

#undef	HAVE_cmpg_float
#undef	HAVE_cmpg_double
#undef	HAVE_cmpl_float
#undef	HAVE_cmpl_double

#define	HAVE_neg_float
#define	HAVE_neg_double
#undef	HAVE_rem_float
#undef	HAVE_rem_double

#undef	HAVE_cvt_long_double
#undef	HAVE_cvt_float_long
#undef	HAVE_cvt_double_long
#undef	HAVE_cvt_long_float

#endif

#endif
