/*
 * superh/jit.h
 * Common Super-H JIT configuration information.
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __superh_jit_h
#define __superh_jit_h

/**/
/* Exception handling information. */
/**/

/* Extract the object argument from given frame */
#define FRAMEOBJECT(f)							\
	(*(Hjava_lang_Object**)(f))

/* Call the relevant exception handler (rewinding the stack as
   necessary). */
#define CALL_KAFFE_EXCEPTION(F, H)					\
	asm volatile("jmp @%1\n\tmov %0,r14" : : "r" (F), "r" (H))

/**/
/* Method dispatch.  */
/**/

#define HAVE_TRAMPOLINE
struct Hjava_lang_Class;

/* The layout of this struct is know by inline assembly.  */

typedef struct _methodTrampoline {
	unsigned short code[4];
	struct _methods* meth;
	struct Hjava_lang_Class* clazz;
	int routine;
} methodTrampoline;

extern void superh_do_fixup_trampoline(void);

/* We must construct jump address since jal can not hold absolute address */
#define FILL_IN_TRAMPOLINE(t,m,c)					\
	do {								\
		(t)->code[0] = 0x012A;	/* sts.l pr,r1 */		\
		(t)->code[1] = 0xD003;	/* mov.l @(?,pc),r0 */		\
		(t)->code[2] = 0x400B;	/* jsr @r0 */			\
		(t)->code[3] = 0x0009;	/* nop */			\
		(t)->meth = (m);					\
		(t)->clazz = (c);					\
		(t)->routine = (uint32)superh_do_fixup_trampoline;	\
	} while (0)

#define FIXUP_TRAMPOLINE_DECL	void** _data
#define FIXUP_TRAMPOLINE_INIT	meth = (Method*)_data[0]; \
				clazz = (Hjava_lang_Class*)_data[1];

/**/
/* Register management information. */
/**/

#define	RIL	(Rint|Rref)
#define	RFD	(Rfloat|Rdouble)
#define	RG	(Rglobal|Rnosaveoncall)

#define	REG_0		0	/* Return value */
#define	REG_1		1	/* Return value (64bits) */
#define	REG_4		4	/* First int arg reg */
#define	REG_5		5	/* Second int arg reg */
#define	REG_FP		14
#define	REG_SP		15

/* Define the register set, bereits an superh angepasst*/
#define	REGISTER_SET							\
	{ /* r0 */	0, 0, Reserved|RIL,	0, 0, 0   },		\
	{ /* r1 */	0, 0, RIL,		0, 0, 1   },		\
	{ /* r2 */	0, 0, RIL,		0, 0, 2   },		\
	{ /* r3 */	0, 0, RIL,		0, 0, 3   },		\
	{ /* r4 */	0, 0, RIL,		0, 0, 4   },		\
	{ /* r5 */	0, 0, RIL,		0, 0, 5   },		\
	{ /* r6 */	0, 0, RIL,		0, 0, 6   },		\
	{ /* r7 */	0, 0, RIL,		0, 0, 7   },		\
	{ /* r8 */	0, 0, RIL,		RG, 0, 8   },		\
	{ /* r9 */	0, 0, RIL,		RG, 0, 9   },		\
	{ /* r10 */	0, 0, RIL,		RG, 0, 10  },		\
	{ /* r11 */	0, 0, RIL,		RG, 0, 11  },		\
	{ /* r12 */	0, 0, RIL,		RG, 0, 12  },		\
	{ /* r13 */	0, 0, RIL,		RG, 0, 13  },		\
	{ /* r14 */	0, 0, Reserved,		0, 0, 14  },		\
	{ /* r15 */	0, 0, Reserved,		0, 0, 15  },		\
	{ /* f0 */	0, 0, RFD,		0, 0, 16  },		\
	{ /* f1 */	0, 0, Reserved,		0, 0, 1   },		\
	{ /* f2 */	0, 0, RFD,		0, 0, 2   },		\
	{ /* f3 */	0, 0, Reserved,		0, 0, 3   },		\
	{ /* f4 */	0, 0, RFD,		0, 0, 4   },		\
	{ /* f5 */	0, 0, Reserved,		0, 0, 5   },		\
	{ /* f6 */	0, 0, RFD,		0, 0, 6   },		\
	{ /* f7 */	0, 0, Reserved,		0, 0, 7   },		\
	{ /* f8 */	0, 0, RFD,		0, 0, 8   },		\
	{ /* f9 */	0, 0, Reserved,		0, 0, 9   },		\
	{ /* f10 */	0, 0, RFD,		0, 0, 10  },		\
	{ /* f11 */	0, 0, Reserved,		0, 0, 11  },		\
	{ /* f12 */	0, 0, RFD,		0, 0, 12  },		\
	{ /* f13 */	0, 0, Reserved,		0, 0, 13  },		\
	{ /* f14 */	0, 0, RFD,		0, 0, 14  },		\
	{ /* f15 */	0, 0, Reserved,		0, 0, 15  },		\
	{ /* f16 */	0, 0, RFD,		0, 0, 16  },		\

/* Number of registers in the register set */
#define	NR_REGISTERS	(16+16)

/**/
/* We must setup the incoming argument state */
/**/
#define NR_ARGUMENTS	4

/**/
/* Number of global registers */
/**/
#define	NR_GLOBALS	6

/**/
/* Define to use the inlining constant pool. */
/**/
#define	USE_JIT_INLINE_CONSTANT_POOL	1
/**/
/* We can only load PC relatifve so far, so we make sure we
 * put the constant pool down every so often.
 */
#define	MAX_INLINE_CONSTANT_POOL_DISTANCE(PC,CNST) \
	(((PC) * 4 + (CNST)) >= 1024)

/**/
/* Define to use stack limit checking - give it the place to find the limit */
/**/
#define	STACK_LIMIT()	SLOT2ARGOFFSET(maxArgs)

/**/
/* Number of registers saved in the prologue. */
/**/
#define	REGISTERS_SAVED	6

/**/
/* Opcode generation. */
/**/

/* Extra label types */
#define Llongx8x8x8x8	(Larchdepend+0)
#define Ldisp8		(Larchdepend+1)
#define Lrelative8	(Larchdepend+2)

/*
 * Frame macros take tree args:
 *  P: where to store the value
 *  V: the value to store
 *  L: the label itself
 */
#define	LABEL_Lframe(P,V,L) {			\
		uint8* _p = (uint8*)(P);	\
		uint32 _v = FRAMESIZE;		\
		_p[0] = (uint8)(_v >> 8);	\
		_p[4] = (uint8)_v;		\
		_v = ARGSIZE;			\
		_p[10] = (uint8)(_v >> 8);	\
		_p[14] = (uint8)_v;		\
	}

#define	LABEL_Lnegframe(P,V,L) {		\
		uint8* _p = (uint8*)(P);	\
		uint32 _v = FRAMESIZE;		\
		_p[0] = (uint8)(_v >> 8);	\
		_p[4] = (uint8)_v;		\
	}

#define	LABEL_Llongx8x8x8x8(P,V,L) {		\
		uint8* _p = (uint8*)(P);	\
		uint32 _v = (uint32)(V);	\
		_p[0] = (uint8)(_v >> 24);	\
		_p[4] = (uint8)(_v >> 16);	\
		_p[8] = (uint8)(_v >> 8);	\
		_p[12] = (uint8)_v;		\
	}

#define LABEL_Ldisp8(P,V,L) {			\
		if ((V) < -256 || (V) > 254) kprintf("label=%p %d %p %p\n", (L), (V), (P), codebase); \
		assert((V) >= -256 && (V) <= 254);	\
		((uint8*)(P))[0] = (uint8)(((V) / 2) & 0xFF); \
	}

#define LABEL_Lrelative8(P,V,L) {		\
		if ((V) < 0 || (V) >= 1024) kprintf("label=%p %d %p %p\n", (L), (V), (P), codebase); \
		assert((V) >= 0 && (V) < 1024);	\
		((uint8*)(P))[0] = (uint8)(((V) / 4) & 0xFF); \
	}

#define	EXTRA_LABELS(P,D,L)						\
	case Lframe:		LABEL_Lframe(P,D,L);		break;	\
	case Lnegframe:		LABEL_Lnegframe(P,D,L);		break;	\
	case Llongx8x8x8x8:	LABEL_Llongx8x8x8x8(P,D,L);	break;	\
	case Ldisp8:		LABEL_Ldisp8(P,D,L);		break;	\
	case Lrelative8:	LABEL_Lrelative8(P,D,L);	break;	\


/**/
/* Slot management information. */
/**/

/* Size of each slot */
#define	SLOTSIZE		4
#define	FRAMEALIGN		4
#define	STACKALIGN(v)		(((v) + FRAMEALIGN - 1) & -FRAMEALIGN)

/*
 * A super-h stack frame looks like:
 *
 *      |  caller's frame               |
 *      |                               |
 *	|-------------------------------|
 *      |  incoming                     |  maxArgs < NR_ARGUMENTS ?
 *      |  args (except first 4)        |        0 : maxArgs - NR_ARGUMENTS
 *  ----|-------------------------------|----
 *      |  saved                        |  REGISTERS_SAVED
 * 	|  registers                    |
 *	|-------------------------------|
 *	|  old FP                       |  1
 *	|-------------------------------|
 *	|  return PC                    |  1
 *      |-------------------------------|
 *	|  locals and                   |  NR_ARGUMENTS + maxLocal - maxArgs
 * r14->|  temporaries                  |           + maxStack + maxTemp
 * 	|-------------------------------|
 *	|  outgoing                     |  maxPush < NR_ARGUMENTS ?
 * r15->|  arguments (except first 4)   |        0 : maxPush - NR_ARGUMENTS
 * 	|-------------------------------|----
 *      |                               |
 *      |  callee's frame               |
 *
 */

#define	PUSHARG_FORWARDS	1

#define	FRAMESIZE \
	((NR_ARGUMENTS + (maxLocal - maxArgs) + maxStack + maxTemp) * SLOTSIZE)

#define	ARGSIZE \
	((maxPush < NR_ARGUMENTS ? 0 : maxPush - NR_ARGUMENTS) * SLOTSIZE)

/* Generate slot offset for an argument relative to the frame pointer */
#define SLOT2ARGOFFSET(N) \
	((N) < NR_ARGUMENTS ? \
	  ((N) * SLOTSIZE) : \
	  (((N) + maxLocal - maxArgs + maxStack + maxTemp + REGISTERS_SAVED + 2) * SLOTSIZE))

/* Generate slot offset for a local (non-argument) */
#define SLOT2LOCALOFFSET(N) \
	(((N) - maxArgs + NR_ARGUMENTS) * SLOTSIZE)

/* Offset from stack pointer (*NOT* framepointer) of arguments not
 * in registers
 */
#define	SLOT2PUSHOFFSET(N) \
	(((N) - NR_ARGUMENTS) * SLOTSIZE)

/* Force MINTEMPS to be (MAXTEMPS-1) */
#define	MINTEMPS	(MAXTEMPS-1)

/* Wrap up a native call for the JIT */
#define KAFFEJIT_TO_NATIVE(M)

#define	FLUSH_DCACHE(S,E)

#endif
