/*
 * superh/superh.c
 * Super-H specific functions.
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "classMethod.h"
#include "objects.h"
#include "support.h"
#include "../../kaffe/kaffevm/threads.h"

/**
 * We need to stop backtracing if we hit 'main' or things go kinda
 * pearshaped.  Not terrible good if the shared library is used elsewhere!
 */
extern int main(int argc, char* argv[]);
extern void handleErrors(void);
#define	IN_MAIN(PC)	((void*)(PC) >= (void*)main && (void*)(PC) < (void*)handleErrors)

void
__superhGetNextFrame(struct _stackFrame* fm)
{
	uint16* pc;
	uint16* spc;
	uint32* ofp;
	void* nfp;
	void* npc;
	int framesize;

	if (fm->return_frame == 0) {
		spc = (uint16*)__superhGetNextFrame;
		asm("mov r14,%0" : "=r" (ofp));
	}
	else {
		spc = (uint16*)fm->return_pc;
		ofp = (uint32*)fm->return_frame;
	}

DBG(	kprintf("Looking from %p (%p)\n", spc, ofp);		)

	/* Not on stack anymore - so this is the end */
	if (!onCurrentThreadStack(ofp) || spc == 0 || IN_MAIN(spc)) {
		fm->return_pc = 0;
		fm->return_frame = 0;
DBG(		kprintf("END\n");				)
		return;
	}

	/* This may seem odd, but it is easiest to walk forward though the
	 * code looking for the epilogue and use that to work out how big
	 * the frame is and therefore the return pc and previous frame
	 * pointer.
	 * The epilogue looks like this:
	 *	add     #framesize,r14	(omitted if framesize == 0)
	 *	mov     r14,r15
	 *	lds.l   @r15+,pr
	 *	mov.l   @r15+,r?	( >= occurances)
	 *	rts
	 *	mov.l   @r15+,r?
	 */
	for (pc = spc;; pc++) {
		/* mov.l r14,r15 */
		if (pc[0] == 0x6fe3) {
			/* add #?,r14 - GCC frame */
			if ((pc[-1] & 0xFF00) == 0x7e00) {
				framesize = pc[-1] & 0xFF;
				/* jsr @r? - which means the add is in delay
				 * slot and so won't actually be relevant here
				 */
				if ((pc[-2] & 0xF0FF) == 0x400b && pc == spc) {
					framesize = 0;
				}
				/* add #?,r14 - can be more than one and it
				 * might not be the next instruction!  Look
				 * backwards over the last two instructions
				 */
				else if ((pc[-2] & 0xFF00) == 0x7e00) {
					framesize += pc[-2] & 0xFF;
				}
				else if ((pc[-3] & 0xFF00) == 0x7e00) {
					framesize += pc[-3] & 0xFF;
				}
			}
			/* mov.w (?,disp),r? ; add r?,r14 - GCC frame */
			else if ((pc[-1] & 0xFF0F) == 0x3e0c) {
				assert((pc[-2] & 0xF000) == 0x9000);
				framesize = *(uint16*)((pc[-2] & 0xFF) * 2 + (int)pc);
			}
			/* mov #?,r0 ; shll8 ; or #?,r0 - KAFFE frame */
			else if ((pc[1] & 0xFF00) == 0xE000) {
				framesize = ((pc[1] & 0xFF) << 8)|(pc[3] & 0xFF);
			}
			else {
				framesize = 0;
			}
			ofp = &ofp[framesize / 4];
			npc = (void*)ofp[0];
			nfp = (void*)ofp[1];
			break;
		}
	}

DBG(	kprintf("FOUND %p (%p) @ %p\n", npc, nfp, pc);			)

	fm->return_pc = npc;
	fm->return_frame = nfp;
}
