/*
 * Define the Java-to-filesystem call interface.
 *
 * Copyright (c) 1996, 1997, 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: jfilesystem.h,v 1.8 2001/08/01 23:40:33 samc Exp $
 */

#ifndef __jfilesystem_h
#define __jfilesystem_h

struct jdirEntry;

typedef struct FileSystemCallInterface {

	void (*init)(void);

	int	(*_open)(const char*, int);
	ssize_t	(*_read)(int, void*, size_t);
	ssize_t	(*_write)(int, const void*, size_t);
	int	(*_lseek)(int, off_t, int);
	int	(*_close)(int);
	int	(*_info)(const char*, int, void*);
	int	(*_infoFD)(int, int, void*);

	int	(*_readdir)(const char*, struct jdirEntry**);

	int	(*_mkdir)(const char*);
	int	(*_rmdir)(const char*);
	int	(*_rename)(const char*, const char*);
	int	(*_remove)(const char*);

} FileSystemCallInterface;

JNIEXPORT FileSystemCallInterface Kaffe_FileSystemCallInterface;

/* Various system call structures and flags */

/* Open modes */
#define	JOPEN_READ		1
#define	JOPEN_WRITE		2
#define	JOPEN_APPEND		(2|4)
#define	JOPEN_READWRITE		(1|2)

/* Info mode */
#define	JINFO_EXISTS		0
#define	JINFO_ISFILE		1
#define	JINFO_ISDIR		2
#define	JINFO_CANREAD		3
#define	JINFO_CANWRITE		4
#define	JINFO_LENGTH		5
#define	JINFO_LASTMODIFIED	6
#define	JINFO_AVAILABLE		7

/* Seek modes */
#define	JSEEK_SET		0
#define	JSEEK_CUR		1
#define	JSEEK_END		2

/* Directory listing */
typedef struct jdirEntry {
	struct jdirEntry*	next;
	char			name[1];
} jdirEntry;

#if !defined(NOUNIXPROTOTYPES)
/*
 * Define some 'UNIX' like macros to make things simpler.
 */

#define	open(A,B)	(*Kaffe_FileSystemCallInterface._open)(A,B)
#define	read(A,B,C)	(*Kaffe_FileSystemCallInterface._read)(A,B,C)
#define	write(A,B,C)	(*Kaffe_FileSystemCallInterface._write)(A,B,C)
#define	lseek(A,B,C)	(*Kaffe_FileSystemCallInterface._lseek)(A,B,C)
#define	close(A)	(*Kaffe_FileSystemCallInterface._close)(A)
#define	fileinfo(A,B,C)	(*Kaffe_FileSystemCallInterface._info)(A,B,C)
#define	fileinfoFD(A,B,C) (*Kaffe_FileSystemCallInterface._infoFD)(A,B,C)

#define	readdir(A,B)	(*Kaffe_FileSystemCallInterface._readdir)(A,B)

#define	mkdir(A)	(*Kaffe_FileSystemCallInterface._mkdir)(A)
#define	rmdir(A)	(*Kaffe_FileSystemCallInterface._rmdir)(A)
#define	rename(A,B)	(*Kaffe_FileSystemCallInterface._rename)(A,B)
#define	remove(A)	(*Kaffe_FileSystemCallInterface._remove)(A)

JNIEXPORT char* userDir;

#endif /* NOUNIXPROTOTYPES */

#endif /* __jfilesystem_h */
