/*
 * Define the Java-to-Network call interface.
 *
 * Copyright (c) 1996, 1997, 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: jnetwork.h,v 1.9 2001/08/01 23:40:33 samc Exp $
 */

#ifndef __jnetwork_h
#define __jnetwork_h

struct sockaddr;

typedef struct NetworkCallInterface {
	void (*init)(void);

	int	(*_socket)(int);
	int	(*_bind)(int, struct sockaddr*, int);
	int	(*_listen)(int, int);
	int	(*_connect)(int, struct sockaddr*, size_t);
	int	(*_accept)(int, struct sockaddr*, size_t*);
	ssize_t	(*_sockread)(int, void*, size_t);
	ssize_t	(*_recvfrom)(int,void*,size_t,int,struct sockaddr*,int*);
	ssize_t	(*_sockwrite)(int, const void*, size_t);
	ssize_t (*_sendto)(int,const void*,size_t,int,const struct sockaddr*,int);
	int	(*_close)(int);
	int	(*_setsockopt)(int, int, int, const void*, int);
	int	(*_getsockopt)(int, int, int, void*, int*);
	int	(*_getsockname)(int, struct sockaddr*, int*);
	int	(*_getpeername)(int, struct sockaddr*, int*);
	char*	(*_gethostname)(void);
	int*	(*_gethostbyname)(char*);
	char*	(*_gethostbyaddr)(int);
	int	(*_ioctlsocket)(int, int, int*);

	int	(*_setreadtimeout)(int, int);
	int	(*_getreadtimeout)(int);

} NetworkCallInterface;

JNIEXPORT NetworkCallInterface Kaffe_NetworkCallInterface;

#if !defined(NOUNIXPROTOTYPES)
/*
 * Define some 'UNIX' like macros to make things simpler.
 */
#define	socket(A)	(*Kaffe_NetworkCallInterface._socket)(A)
#define	bind(A,B,C)	(*Kaffe_NetworkCallInterface._bind)(A,B,C)
#define	listen(A,B)	(*Kaffe_NetworkCallInterface._listen)(A,B)
#define	connect(A,B,C)	(*Kaffe_NetworkCallInterface._connect)(A,B,C)
#define	accept(A,B,C)	(*Kaffe_NetworkCallInterface._accept)(A,B,C)
#define	sockread(A,B,C)	(*Kaffe_NetworkCallInterface._sockread)(A,B,C)
#define	recvfrom(A,B,C,D,E,F) \
			(*Kaffe_NetworkCallInterface._recvfrom)(A,B,C,D,E,F)
#define	sockwrite(A,B,C) \
			(*Kaffe_NetworkCallInterface._sockwrite)(A,B,C)
#define	sendto(A,B,C,D,E,F) \
			(*Kaffe_NetworkCallInterface._sendto)(A,B,C,D,E,F)
#define	sockclose(A)	(*Kaffe_NetworkCallInterface._close)(A)
#define	setsockopt(A,B,C,D,E) \
			(*Kaffe_NetworkCallInterface._setsockopt)(A,B,C,D,E)
#define	getsockopt(A,B,C,D,E) \
			(*Kaffe_NetworkCallInterface._getsockopt)(A,B,C,D,E)
#define	getsockname(A,B,C) \
			(*Kaffe_NetworkCallInterface._getsockname)(A,B,C)
#define	getpeername(A,B,C) \
			(*Kaffe_NetworkCallInterface._getpeername)(A,B,C)
#define	gethostbyname(A) (*Kaffe_NetworkCallInterface._gethostbyname)(A)
#define	gethostbyaddr(A) (*Kaffe_NetworkCallInterface._gethostbyaddr)(A)
#define	gethostname()	(*Kaffe_NetworkCallInterface._gethostname)()
#define	ioctlsocket(A,B,C)	(*Kaffe_NetworkCallInterface._ioctlsocket)(A,B,C)
#define	setreadtimeout(A,B) \
			(*Kaffe_NetworkCallInterface._setreadtimeout)(A,B)
#define	getreadtimeout(A) \
			(*Kaffe_NetworkCallInterface._getreadtimeout)(A)

#endif

#endif /* __jnetwork_h */
