/*
 * Java Native Interface.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * DEPRECATED.  Javasoft keeps changing the spec.
 *
 * $tvt: jvmdi.h,v 1.4 2001/07/09 22:43:00 samc Exp $
 */

#ifndef __jvmdi_h
#define	__jvmdi_h

#include <jvmdi.h>

/*
 * jvmdiError values
 */
#define JVMDI_ERROR_NONE 			1
#define JVMDI_ERROR_NULL_POINTER 		2
#define JVMDI_ERROR_OUT_OF_MEMORY 		3
#define JVMDI_ERROR_INVALID_THREAD 		4
#define JVMDI_ERROR_THREAD_SUSPENDED 		5
#define JVMDI_ERROR_VM_DEAD 			6
#define JVMDI_ERROR_THREAD_NOT_SUSPENDED 	7
#define JVMDI_ERROR_NO_MORE_FRAMES 		8
#define JVMDI_ERROR_INVALID_FRAMEID 		9
#define JVMDI_ERROR_OUT_OF_MEMORY 		10
#define JVMDI_ERROR_NATIVE_FRAME 		11
#define JVMDI_ERROR_INVALID_SLOT 		12
#define JVMDI_ERROR_TYPE_MISMATCH 		13
#define JVMDI_ERROR_INVALID_METHODID 		14
#define JVMDI_ERROR_INVALID_CLASS 		15
#define JVMDI_ERROR_INVALID_LOCATION 		16
#define JVMDI_ERROR_DUPLICATE_BREAKPOINT 	17
#define JVMDI_ERROR_ABSENT_INFORMATION 		18
#define JVMDI_ERROR_INVALID_FIELDID 		19
#define JVMDI_ERROR_INVALID_METHODID 		20

/*
 * Memory management
 */
typedef jvmdiError (*JVMDI_AllocHook)(JNIEnv *env, jlong size, jbyte** memPtr);
typedef jvmdiError (*JVMDI_DeallocHook)(JNIEnv *env, jbyte* mem);
JNIEXPORT jvmdiError JNICALL JVMDI_SetAllocationHooks(JNIEnv *env, JVMDI_AllocHook ahook, JVMDI_DeallocHook dhook);
JNIEXPORT jvmdiError JNICALL JVMDI_Allocate(JNIEnv *env, jlong size, jbyte** memPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_Deallocate(JNIEnv *env, jbyte* mem);

/*
 * Thread Execution
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetThreadStatus(JNIEnv *env, jthread thread, jint *statusPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_SuspendThread(JNIEnv *env, jthread thread); 
JNIEXPORT jvmdiError JNICALL JVMDI_ResumeThread(JNIEnv *env, jthread thread);
JNIEXPORT jvmdiError JNICALL JVMDI_SetSingleStep(JNIEnv *env, jthread thread, jboolean shouldStep);

/*
 * Thread status values
 */
#define JVMDI_THREAD_STATUS_UNKNOWN 	1
#define JVMDI_THREAD_STATUS_ZOMBIE 	2
#define JVMDI_THREAD_STATUS_RUNNING 	3
#define JVMDI_THREAD_STATUS_SLEEPING 	4
#define JVMDI_THREAD_STATUS_MONITOR 	5
#define JVMDI_THREAD_STATUS_WAIT 	6
#define JVMDI_THREAD_STATUS_SUSPENDED 	7
#define JVMDI_THREAD_STATUS_BREAK 	8

/*
 * Stack Frame Access
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetCurrentFrame(JNIEnv *env, jthread thread, jframeID *framePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetCallerFrame(JNIEnv *env, jframeID called, jframeID *framePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetFrameMethod(JNIEnv *env, jframeID frame, jclass *classPtr, jmethodID *methodPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetFrameLocation(JNIEnv *env, jframeID frame, jclass *classPtr, jmethodID *methodPtr, jlocation *locationPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_NotifyFramePop(JNIEnv *env, jframeID frame); 

/*
 * Local Variable Access
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetLocalObject(JNIEnv *env, jframeID frame, jint slot, jobject *valuePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetLocalInt(JNIEnv *env, jframeID frame, jint slot, jint *valuePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetLocalLong(JNIEnv *env, jframeID frame, jint slot, jlong *valuePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetLocalFloat(JNIEnv *env, jframeID frame, jint slot, jfloat *valuePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetLocalDouble(JNIEnv *env, jframeID frame, jint slot, jdouble *valuePtr);

JNIEXPORT jvmdiError JNICALL JVMDI_SetLocalObject(JNIEnv *env, jframeID frame, jint slot, jobject value);
JNIEXPORT jvmdiError JNICALL JVMDI_SetLocalInt(JNIEnv *env, jframeID frame, jint slot, jint value);
JNIEXPORT jvmdiError JNICALL JVMDI_SetLocalFloat(JNIEnv *env, jframeID frame, jint slot, jfloat value);
JNIEXPORT jvmdiError JNICALL JVMDI_SetLocalLong(JNIEnv *env, jframeID frame, jint slot, jlong value);
JNIEXPORT jvmdiError JNICALL JVMDI_SetLocalDouble(JNIEnv *env, jframeID frame, jint slot, jdouble value);

/*
 * Breakpoints
 */

JNIEXPORT jvmdiError JNICALL JVMDI_SetBreakpoint(JNIEnv *env, jclass clazz, jmethodID method, jlocation location);
JNIEXPORT jvmdiError JNICALL JVMDI_ClearBreakpoint(JNIEnv *env, jclass clazz, jmethodID method, jlocation location);
JNIEXPORT jvmdiError JNICALL JVMDI_ClearAllBreakpoints(JNIEnv *env);

/*
 * Class Information
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetClassName(JNIEnv *env, jclass clazz, char **namePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetSourcePath(JNIEnv *env, jclass clazz, char **sourcePathPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetClassModifiers(JNIEnv *env, jclass clazz, jint *modifiersPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetClassMethods(JNIEnv *env, jclass clazz, jint *methodCountPtr, jmethodID **methodsPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetClassFields(JNIEnv *env, jclass clazz, jint *fieldCountPtr, jfieldID **fieldsPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetImplementedInterfaces(JNIEnv *env, jclass clazz, jint *interfaceCountPtr, jclass **interfacesPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_IsInterface(JNIEnv *env, jclass clazz, jboolean *isInterfacePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_IsInterface(JNIEnv *env, jclass clazz, jboolean *isArrayPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_ClassLoader(JNIEnv *env, jclass clazz, jobject *classloaderPtr);

/*
 * Field Access
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetFieldName(JNIEnv *env, jclass clazz, jfieldID field, char **namePtr, char **signaturePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetFieldDeclaringClass(JNIEnv *env, jclass clazz, jfieldID field, jclass *declaringClassPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetFieldModifiers(JNIEnv *env, jclass clazz, jfieldID field, jint *modifiersPtr);

/*
 * Method Access
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetMethodName(JNIEnv *env, jclass clazz, jmethodID method, jstring *namePtr, jstring *signaturePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetMethodDeclaringClass(JNIEnv *env, jclass clazz, jmethodID method, jclass *declaringClassPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetMethodModifiers(JNIEnv *env, jclass clazz, jmethodID method, jint *modifiersPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetMaxStack(JNIEnv *env, jclass clazz, jmethodID method, jint *maxPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetMaxLocals(JNIEnv *env, jclass clazz, jmethodID method, jint *maxPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetArgumentsSize(JNIEnv *env, jclass clazz, jmethodID method, jint *sizePtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetLineNumberTable(JNIEnv *env, jclass clazz, jmethodID method, jint *entryCountPtr, JVMDI_line_number_entry **tablePtr);

typedef struct {
	jlocation start_location;
	jint line_number;
} JVMDI_line_number_entry;

JNIEXPORT jvmdiError JNICALL JVMDI_GetMethodLocation(JNIEnv *env, jclass clazz, jmethodID method, jlocation *startLocationPtr, jlocation *endLocationPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetLocalVariableTable(JNIEnv *env, jclass clazz, jmethodID method, jint *entryCountPtr, JVMDI_local_variable_entry **tablePtr);

typedef struct {
	jlocation start_location;   /* variable valid start_location */
	jint length;                /* upto start_location+length */ 
	char *name;                 /* name in UTF8 */
	char *signature;            /* type signature in UTF8 */
	jint slot;                  /* variable slot, see JVMDI_GetLocal*()  */
} JVMDI_local_variable_entry;

JNIEXPORT jvmdiError JNICALL JVMDI_GetExceptionHandlerTable(JNIEnv *env, jclass clazz, jmethodID method, jint *entryCountPtr, JVMDI_exception_handler_entry **tablePtr)

typedef struct {
	jlocation start_location;
	jlocation end_location;
	jlocation handler_location;
	jclass exception;           /* if null, all exceptions */
} JVMDI_exception_handler_entry;

JNIEXPORT jvmdiError JNICALL JVMDI_GetThrownExceptions(JNIEnv *env, jclass clazz, jmethodID method, jint *exceptionCountPtr, jclass **exceptionsPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetBytecodes(JNIEnv *env, jclass clazz, jmethodID method, jint *bytecodeCountPtr, jbyte **bytecodesPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_IsMethodNative(JNIEnv *env, jclass clazz, jmethodID method, jboolean *isNativePtr);

/*
 * Events
 */

typedef void (*JVMDI_EventHook)(JNIEnv *env, JVMDI_Event *event);

JNIEXPORT jvmdiError JNICALL JVMDI_SetEventHook(JNIEnv *env, JVMDI_EventHook hook);

typedef struct {
         jint kind;          /* the discriminant */

         union {
             struct { /* kind = JVMDI_EVENT_SINGLE_STEP */
                 jthread thread;
                 jclass clazz;
                 jmethodID method;
                 jlocation location;
             } single_step;
                             
             struct { /* kind = JVMDI_EVENT_BREAKPOINT */
                 jthread thread;
                 jclass clazz;
                 jmethodID method;
                 jlocation location;
             } breakpoint;

             struct { /* kind = JVMDI_EVENT_FRAME_POP */
                 jthread thread;
                 jframeID frame;
             } frame_pop;

             struct { /* kind = JVMDI_EVENT_EXCEPTION */
                 jthread thread;
                 jclass clazz;
                 jmethodID method;
                 jlocation location;
                 jobject exception;
                 jclass catch_clazz;
                 jmethodID catch_method;
                 jlocation catch_location;
             } exception;

             struct { /* kind = JVMDI_EVENT_USER_DEFINED */
                 jobject object;
                 jint key;
             } user;
                             
             struct { /* kind = JVMDI_EVENT_THREAD_END or */
                      /* JVMDI_EVENT_THREAD_START */
                 jthread thread;
             } thread_change;
                             
             struct { /* kind = JVMDI_EVENT_CLASS_LINK or */
                      /* JVMDI_EVENT_CLASS_UNLINK */
                 jclass clazz;
             } link;
                             
                      /* kind = JVMDI_EVENT_VM_DEATH */
                      /* no additional fields */             
	} u;
} JVMDI_Event;


/*
 * Miscellaneous Functions
 */
JNIEXPORT jvmdiError JNICALL JVMDI_GetLoadedClasses(JNIEnv *env, jint *classCountPtr, jclass **classesPtr);
JNIEXPORT jvmdiError JNICALL JVMDI_GetVersionNumber(JNIEnv *env, jint *versionPtr);

#endif
