/*
 * Kick off program.
 *
 * Copyright (c) 1996 - 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $Id: main.c,v 1.49 2001/08/16 00:31:49 tim Exp $
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "native.h"
#include "constants.h"
#include "support.h"
#include "threads.h"

/* HACK - we cannot use the internal VM malloc system at this point so
 * turn it off.  We should fix this properly.
 */
#undef  malloc

JNIIMPORT char* engine_name;
JNIIMPORT char* engine_version;

static int java_version_major = 1;
static int java_version_minor = 1;

#include "jni.h"

JavaVMInitArgs vmargs;
JNIEnv* env;
JavaVM* vm;
static int isJar = 0;

static int options(char**);
static void usage(void);
static size_t parseSize(char*);
static int parseSizeOption(char**, char*, size_t*);
void handleErrors(void);

/* Kaffe specific environment */
#define KAFFEHOME       "KAFFEHOME"
#define CLASSPATH1      "KAFFECLASSPATH"
#define LIBRARYHOME1    "KAFFELIBRARYPATH"

/* Fall back on standard environment */
#define CLASSPATH2      "CLASSPATH"

#if defined(DEFAULT_PROGRAM)
static char* defaultargs[] = { "kaffe", #DEFAULT_PROGRAM, 0 };
#endif

/* Let us shrink the default stacksize if we want */
#undef	THREADSTACKSIZE

/*
 * MAIN
 */
int
main(int argc, char* argv[])
{
	jarray args;
	jclass cls;
	jclass mcls;
	jmethodID mmth;
	jobject str;
	int i;
	int farg;
	char* cp;
	char* exec;

#if defined(DEFAULT_PROGRAM)
	if (argc <= 1) {
		argc = 2;
		argv = defaultargs;
	}
#endif
	
	vmargs.version = (java_version_major<<16)|java_version_minor;
	JNI_GetDefaultJavaVMInitArgs(&vmargs);
       
	cp = getenv(CLASSPATH1);
	if (cp == 0) {
		cp = getenv(CLASSPATH2);
#if defined(DEFAULT_CLASSPATH)
		if (cp == 0) {
			cp = DEFAULT_CLASSPATH;
		}
#endif
	}
	vmargs.classpath = cp;

	cp = getenv(LIBRARYHOME1);
#if defined(DEFAULT_LIBRARYHOME)
	if (cp == 0) {
		cp = DEFAULT_LIBRARYHOME;
	}
#endif
	vmargs.libraryhome = cp;

	cp = getenv(KAFFEHOME);
	if (cp == 0) {
#if defined(DEFAULT_KAFFEHOME)
		cp = DEFAULT_KAFFEHOME;
#else
		fprintf(stderr, "%s is not set!\n", KAFFEHOME);
		exit(1);
#endif
	}
	vmargs.classhome = cp;

	/* Process program options */
	farg = options(argv);
	argc = argc - farg;

	/* Get the class name to start with */
	if (argv[farg] == 0) {
		usage();
		exit(1);
	}

#if defined(DEBUG)
	/* pjo, 12 may 2000 */
	/* for GDB JIT debugging */
	/* If the KX environment points to a valid file, open it. */
#define KX "KX" /* um...  did I just do that? */
	cp = getenv(KX);
	if (cp != 0) {
		if((vmargs.kxFile = open(cp, O_RDWR)) < 0)
		    perror("open"); 
	}
#endif

	/* Executable is a JAR? Use the JAR launcher */
	if (isJar != 0) {
		exec = "kaffe.jar.ExecJar";
	}
	else {
		exec = argv[farg];
		farg++;
		argc--;
	}


        /* Initialise */
	JNI_CreateJavaVM(&vm, &env, &vmargs);

	mcls = (*env)->FindClass(env, exec);
	handleErrors();

	mmth = (*env)->GetStaticMethodID(env, mcls, "main", "([Ljava/lang/String;)V");
	handleErrors();

	/* Build an array of strings as the arguments */
	cls = (*env)->FindClass(env, "java.lang.String");
	handleErrors();
	args = (*env)->NewObjectArray(env, argc, cls, 0);
	handleErrors();
	for (i = 0; i < argc; i++) {
		str = (*env)->NewStringUTF(env, argv[farg+i]);
		handleErrors();
		(*env)->SetObjectArrayElement(env, args, i, str);
		handleErrors();
	}

	/* Call method, check for errors and then exit */
	(*env)->CallStaticVoidMethod(env, mcls, mmth, args);
	handleErrors();

	(*vm)->DetachCurrentThread(vm);

	return (0);
}

void
handleErrors(void)
{
	if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
		(*env)->ExceptionClear(env);
		/* (*vm)->DetachCurrentThread(vm); */
		exit(1);
	}
}

/*
 * Process program's flags.
 */
static
int
options(char** argv)
{
	int i;
	int sz;
	userProperty* prop;
	int ret;
	size_t val;

	for (i = 1; argv[i] != 0; i++) {
		if (argv[i][0] != '-') {
			break;
		}

		if (strcmp(argv[i], "-help") == 0) {
			usage();
			exit(0);
		}
		else if (strcmp(argv[i], "-version") == 0) {
			printf("Kaffe Virtual Machine %s", engine_version);
#if defined(BUILD_ID)
                        if (strlen(BUILD_ID) > 0)
                        printf(" (%s)", BUILD_ID);
#endif
			printf("\nCopyright (c) 1996 - 2000\n");
                        printf("Transvirtual Technologies, Inc.  All rights reserved\n");
			printf("Engine: %s   Java Version: 1.1/2\n", engine_name);
#if defined(BUILD_USER)
			printf("Compiled by  : %s\n",BUILD_USER);
#endif
#if defined(BUILD_HOST) & defined(BUILD_ARCH)
                        printf("Compiled on  : %s (%s)\n",BUILD_HOST,BUILD_ARCH);
#endif
#if defined(BUILD_DATE)
                        printf("Compile date : %s\n",BUILD_DATE);
#endif
#if defined(__GNUC__) & defined(__GNUC_MINOR__)
                        printf("GCC Version  : %d.%d\n",__GNUC__,__GNUC_MINOR__);
#endif
#if defined(BUILD_OPTIONS)
                        printf("Options      : %s\n",BUILD_OPTIONS);
#endif
			exit(0);
		}
		else if (strcmp(argv[i], "-classpath") == 0) {
			i++;
			if (argv[i] == 0) {
				fprintf(stderr, "Error: No path found for -classpath option.\n");
				exit(1);
			}
			vmargs.classpath = argv[i];
		}
		else if (strcmp(argv[i], "-addclasspath") == 0) {
			i++;
			if (argv[i] == 0) {
				fprintf(stderr, "Error: No path found for -addclasspath option.\n");
				exit(1);
			}
			vmargs.classpath = argv[i];
		}
		else if ((ret = parseSizeOption(&argv[i], "-ss", &val)) != 0) {
			i += ret - 1;
			
#if defined(UNIX_PTHREADS) 
                        if (val < PTHREAD_STACK_MIN) {
				fprintf(stderr, "Warning: Attempt to set stack size smaller than %d - ignored.\n", PTHREAD_STACK_MIN);
			}
			else
#elif defined(THREADSTACKSIZE)
			if (val < THREADSTACKSIZE) {
				fprintf(stderr, "Warning: Attempt to set stack size smaller than %d - ignored.\n", THREADSTACKSIZE);
			}
			else
#endif
                	{
				vmargs.nativeStackSize = val;
			}
		}
		else if ((ret = parseSizeOption(&argv[i], "-mx", &val)) != 0) {
			i += ret - 1;
			vmargs.maxHeapSize = val;
		}
		else if ((ret = parseSizeOption(&argv[i], "-ms", &val)) != 0) {
			i += ret - 1;
			vmargs.allocHeapSize = val;
		}
#if 0
		else if (strcmp(argv[i], "-nopreempt") == 0) {
			vmargs.threadpreempt = 0;
		}
#endif
		else if (strcmp(argv[i], "-jar") == 0) {
			isJar = 1;
		}
		else if (strcmp(argv[i], "-verify") == 0) {
			vmargs.verifyMode = 3;
		}
		else if (strcmp(argv[i], "-verifyremote") == 0) {
			vmargs.verifyMode = 2;
		}
		else if (strcmp(argv[i], "-noverify") == 0) {
			vmargs.verifyMode = 0;
		}
		else if (strcmp(argv[i], "-verbosegc") == 0) {
			vmargs.enableVerboseGC = 1;
		}
		else if (strcmp(argv[i], "-verbosejit") == 0) {
			vmargs.enableVerboseJIT = 1;
		}
		else if (strcmp(argv[i], "-verbosemem") == 0) {
			vmargs.enableVerboseGC = 2;
		}
		else if (strcmp(argv[i], "-verbosecall") == 0) {
			vmargs.enableVerboseCall = 1;
		}
		else if (strcmp(argv[i], "-redirect") == 0) {
			vmargs.enableRedirect = 1;
		}
		else if (strcmp(argv[i], "-verbose") == 0 || strcmp(argv[i], "-v") == 0) {
			vmargs.enableVerboseClassloading = 1;
		}
		else if (strcmp(argv[i], "-resources") == 0) {
			vmargs.enableVerboseClassloading = 2;
		}
		else if (argv[i][1] ==  'D') {
			/* Set a property */
			prop = malloc(sizeof(userProperty));
			assert(prop != 0);
			prop->next = userProperties;
			userProperties = prop;
			for (sz = 2; argv[i][sz] != 0; sz++) {
				if (argv[i][sz] == '=') {
					argv[i][sz] = 0;
					sz++;
					break;
				}
			}
			prop->key = &argv[i][2];
			prop->value = &argv[i][sz];
		}
		/* The following options are not supported and will be
		 * ignored for compatibility purposes.
		 */
		else if (strcmp(argv[i], "-noasyncgc") == 0 ||
		   strcmp(argv[i], "-cs") == 0 ||
		   strcmp(argv[i], "-checksource") == 0 ||
		   strcmp(argv[i], "-prof") == 0) {
		}
		else if (strcmp(argv[i], "-ms") == 0 ||
		   strcmp(argv[i], "-oss") == 0) {
			i++;
		}
		else {
			fprintf(stderr, "Unknown flag: %s\n", argv[i]);
		}
	}

	/* Return first no-flag argument */
	return (i);
}

/*
 * Print usage message.
 */
static
void
usage(void)
{
	printf("usage: kaffe [-options] class\n");
	printf("Options are:\n");
	printf("	-help			Print this message\n");
	printf("	-version		Print version number\n");
	printf("	-ss <size>		Maximum native stack size\n");
	printf("	-mx <size> 		Maximum heap size\n");
	printf("	-ms <size> 		Initial heap size\n");
	printf("	-classpath <path>	Set classpath\n");
	printf("	-addclasspath <path>	Add to classpath\n");
	printf("	-verify *		Verify all bytecode\n");
	printf("	-verifyremote *		Verify bytecode loaded from network\n");
	printf("	-noverify		Do not verify any bytecode\n");
	printf("	-D<property>=<value>	Set a property\n");
	printf("	-verbosegc		Print message during garbage collection\n");
	printf("	-v, -verbose		Be verbose\n");
	printf("	-resources		Tell me what classes and resources are loaded\n");
	printf("	-verbosejit		Print message during JIT code generation\n");
	printf("	-verbosemem		Print detailed memory allocation statistics\n");
	printf("	-verbosecall		Print detailed call flow information\n");
	printf("	-debug * 		Trace method calls\n");
	printf("	-noasyncgc *		Do not garbage collect asynchronously\n");
	printf("	-cs, -checksource *	Check source against class files\n");
	printf("	-oss <size> *		Maximum java stack size\n");
	printf("	-nopreempt *		Disable preemption\n");
	printf("	-jar	 		Executable is a JAR\n");
//	printf("	-prof *			?\n");
	printf("  * Option currently ignored.\n");
}

static
size_t
parseSize(char* arg)
{
	size_t sz;
	char* narg;

	sz = strtol(arg, &narg, 0);
	switch (narg[0]) {
	case 'b': case 'B':
		break;

	case 'k': case 'K':
		sz *= 1024;
		break;

	case 'm': case 'M':
		sz *= 1024 * 1024;
		break;
	}

	return (sz);
}

static
int
parseSizeOption(char** args, char* opt, size_t* val)
{
	unsigned olen;

	olen = strlen(opt);
	if (strncmp(args[0], opt, olen) != 0) {
		return (0);
	}
	else if (strlen(args[0]) != olen) {
		*val = parseSize(&args[0][olen]);
		return (1);
	}
	else if (args[1] == 0) {
		fprintf(stderr, "Bad argument to option: %s\n", opt);
		return (0);
	}
	else {
		*val = parseSize(args[1]);
		return (2);
	}
}
