/*
 * main.c - main func for app kaffecli (a front end for kaffe.net.KaffeServ)
 * Kick off program.
 *
 * Copyright (c) 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include <sys/time.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>

#define DEF_PORT 9455
#define DEF_HOST "localhost"

char **vArgs;
int  nArgs;

char *host = DEF_HOST;
int  port;

int  err;



void parseArgs ( int argc, char** argv ) {
  int i;

  nArgs = argc-1;
  vArgs = argv+1;

  for ( i=1; i<argc; i++ ) {
	if ( strcmp( argv[i], "-port") == 0 ) {
	  port = atoi( argv[++i]);
	  nArgs -= 2;
	  vArgs += 2;
	}
	else if ( strcmp( argv[i], "-host") == 0 ) {
	  host = argv[++i];
	  nArgs -= 2;
	  vArgs += 2;
	}
	else
	  break;
  }
}

void writeArgs ( int fd ) {
  char  cwd[256];
  int   i, len;

  if ( !getcwd( cwd, sizeof(cwd)-2) ) {
	fprintf( stderr, "error: path buffer overflow\n");
	exit(1);
  }
  else {
	len = strlen(cwd);
	write( fd, cwd, len);
  }

  for ( i=0; i<nArgs; i++) {
	write( fd, " ", 1);
	write( fd, vArgs[i], strlen( vArgs[i]));
  }

  write( fd, "\n", 1);
}


int readWrite ( int rfd, int wfd ) {
  int  i, len;
  char buf[1024];

  if ( (len = read( rfd, buf, sizeof(buf))) ) {
	write( wfd, buf, len);
  }
  	
  return len;
}


/*
 * wait for input arriving on stdin and output arriving on fd
 */
void ioLoop ( int fd ) {
  fd_set rfds, xfds;
  int ret;

  do {
    FD_ZERO( &xfds);
    FD_SET( fd, &xfds);

    FD_ZERO( &rfds);
    FD_SET( 0, &rfds);
    FD_SET( fd, &rfds);

    if ( (ret = select( fd+1, &rfds, NULL, &xfds, NULL)) ) {
      if ( FD_ISSET( fd, &rfds) ) {
      	if ( !readWrite( fd, 1) )
      	  break;                         /* end-of-file encountered */
      }

      if ( FD_ISSET( fd, &xfds) ) {
      	break;
      }
      
      if ( FD_ISSET( 0, &rfds) ) {
      	readWrite( 0, fd);
      }
  	}
  } while ( 1 );
}


int main ( int argc, char** argv ) {

  int                   fd;
  struct sockaddr_in    sa;
  struct hostent        *hp;
  struct servent        *sp;

  parseArgs( argc, argv);

  if ( port == 0 ) {
	if ( (sp = getservbyname( "kaffeserv", "tcp")) )
	  port = sp->s_port;
	else
	  port = DEF_PORT;
  }

  if ( !(hp = gethostbyname( host)) ){
    fprintf( stderr, "error: unknown host %s\n", host);
    exit( 1);
  }

  memset( (char*)&sa, 0, sizeof( sa));
  memcpy( (char*)&sa.sin_addr, (char*)hp->h_addr, hp->h_length);
  sa.sin_family = AF_INET;
  sa.sin_port = htons( port);

  if ( (fd = socket( AF_INET, SOCK_STREAM, 0)) < 0 ){
    perror( "socket");
    exit( 1);
  }
  
  if ( connect( fd, (struct sockaddr*)&sa, sizeof(sa)) < 0 ) {
    perror( "connect");
    exit( 1);
  }

  writeArgs( fd);
  
  ioLoop( fd);

  close( fd);
}
