/* -*- mode: C; mode: fold -*- */

/* includes */ /*{{{*/
#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "native.h"
#include "constants.h"
#include "support.h"
#include "threads.h"
#include "jni.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>

/*}}}*/

/* defines */ /*{{{*/
#define LAUNCH_CLASS "java.awt.AppletWindow"
#define ARGV 0
#define ARGC 0

#define CLASSPATH       "CLASSPATH"

/* Kaffe specific environment */
#define KAFFEHOME       "KAFFEHOME"
#define CLASSPATH1      "KAFFECLASSPATH"
#define LIBRARYHOME1    "KAFFELIBRARYPATH"

/* Fall back on standard environment */
#define CLASSPATH2      "CLASSPATH"


/* the socket params for launch requests */
#define PORT 9456

#define CHECK_EXCPT(_expr)    \
	_expr;                    \
	if ( handleErrors() ){    \
		State = S_ERROR;      \
		return 0;             \
	}
/*}}}*/
  
/* global data */ /*{{{*/
#define S_UNINITIALIZED 0
#define S_INITIALIZED 1
#define S_ERROR -1
int State = S_UNINITIALIZED;

pthread_t Tid = 0;

int Fd = -1;  /* the connection socket */

JNIIMPORT char* Engine_name;
JNIIMPORT char* Engine_version;

static int Java_version_major = 1;
static int Java_version_minor = 1;

JavaVMInitArgs VmArgs;
JNIEnv* Env;
JavaVM* Vm;
/*}}}*/

/****************************************************************************
 * internal funtions
 */

static int handleErrors(void) /*{{{*/
{
	if ((*Env)->ExceptionOccurred(Env)) {
		(*Env)->ExceptionDescribe(Env);
		(*Env)->ExceptionClear(Env);
		/* (*Vm)->DetachCurrentThread(Vm); */
		return 1;
	}
	
	return 0;
}
/*}}}*/

static void* runKaffe ( void* p ) /*{{{*/
{
	int argc = ARGC;
	char** argv = ARGV;
	jarray args;
	jclass cls;
	jclass mcls;
	jmethodID mmth;
	jobject str;
	int i;
	int farg;
	char* cp;
	char* exec = LAUNCH_CLASS;
	
	VmArgs.version = (Java_version_major<<16)|Java_version_minor;
	JNI_GetDefaultJavaVMInitArgs(&VmArgs);

	cp = getenv(CLASSPATH);
	if (cp == 0) {
		cp = getenv(CLASSPATH2);
#if defined(DEFAULT_CLASSPATH)
		if (cp == 0) {
			cp = DEFAULT_CLASSPATH;
		}
#endif
	}
	VmArgs.classpath = cp;

	cp = getenv(LIBRARYHOME1);
#if defined(DEFAULT_LIBRARYHOME)
	if (cp == 0) {
		cp = DEFAULT_LIBRARYHOME;
	}
#endif
	VmArgs.libraryhome = cp;

	cp = getenv(KAFFEHOME);
	if (cp == 0) {
#if defined(DEFAULT_KAFFEHOME)
		cp = DEFAULT_KAFFEHOME;
#else
		fprintf(stderr, "%s is not set!\n", KAFFEHOME);
		state = S_ERROR;
		return 0;
#endif
	}
	VmArgs.classhome = cp;

	/* Initialise */
	JNI_CreateJavaVM(&Vm, &Env, &VmArgs);

	mcls = (*Env)->FindClass(Env, exec);
	
	CHECK_EXCPT( mmth = (*Env)->GetStaticMethodID(Env, mcls,
									 "main", "([Ljava/lang/String;)V"));

	/* Build an array of strings as the arguments */
	CHECK_EXCPT( cls = (*Env)->FindClass(Env, "java.lang.String"));
	CHECK_EXCPT( args = (*Env)->NewObjectArray(Env, argc, cls, 0));
	for (i = 0; i < argc; i++) {
		CHECK_EXCPT( str = (*Env)->NewStringUTF(Env, argv[i]));
		CHECK_EXCPT( (*Env)->SetObjectArrayElement(Env, args, i, str));
	}

	/* Call method, check for errors and then exit */
	CHECK_EXCPT( (*Env)->CallStaticVoidMethod(Env, mcls, mmth, args));

	(*Vm)->DetachCurrentThread(Vm);

	return 0;
}
/*}}}*/

static int createJavaMainThread () /*{{{*/
{
	pthread_attr_t attr;
	
	pthread_attr_init( &attr);
	pthread_create( &Tid, &attr, runKaffe, 0);

	return 1;
}
/*}}}*/

static int initialize () /*{{{*/
{
	int sd;
	int port = PORT;
	struct sockaddr_in sa;
	struct hostent *hp;
	struct servent *sp;
	int addrlen;
	int opt;
	char buf[128];
	int i;

	/*
	 * we turn the socket communication sort of upside down, creating the
	 * server side here, so that we can sync on the Java side without
	 * guessing around which sync mechanism actually coexists with the
	 * clients thread system (we don't know much about)
	 */ 
	
	gethostname( buf, sizeof(buf));
	if ( !(hp = gethostbyname( buf)) ){
		fprintf( stderr, "error: can't identify local host\n");
		return 0;
	}

	memset( (char*)&sa, 0, sizeof( sa));
	memcpy( (char*)&sa.sin_addr, (char*)hp->h_addr, hp->h_length);
	sa.sin_family = AF_INET;
	sa.sin_port = htons( port);
	addrlen = sizeof( sa);
	
	if ( (sd = socket( AF_INET, SOCK_STREAM, 0)) < 0 ){
		fprintf( stderr, "error: can't create socket\n");
		return 0;
	}

	opt = 1;
	setsockopt( sd, SOL_SOCKET, SO_REUSEADDR, (void*)&opt, sizeof(opt));
	
	if ( bind( sd, (struct sockaddr*)&sa, addrlen) ){
		fprintf( stderr, "error: can't bind socket\n");
		return 0;
	}
	
	if ( listen( sd, 2) ){
		fprintf( stderr, "error: socket listen failed\n");
		return 0;
	}
	
	if ( createJavaMainThread() ){
		
		/* Now wait for the Java thread to connect. Once this is
		 * done, we should be safe to send our request
		 */
		if ( (Fd = accept( sd, (struct sockaddr*)&sa, &addrlen)) < 0 ){
			fprintf( stderr, "error: accept failed\n");
			return 0;
		}
		return 1;
	}
	else {
		return 0;
	}
}
/*}}}*/

static int sendAppletRequest ( void* wnd, int x, int y, int w, int h,
							   int argc, char** argv ) /*{{{*/
{
	char buf[256];
	int i;

 printf( "sending Applet request: %x (%d,%d,%d,%d) %s\n",
		 wnd, x, y, w, h, argv[0]);
	
	buf[0] = 0;
	sprintf( buf, "%x %d %d %d %d", wnd, x, y, w, h);
	write( Fd, buf, strlen( buf));
	
	for ( i=0; i<argc; i++ ){
 		write( Fd, " ", 1);
		write( Fd, argv[i], strlen( argv[i]));
	}
	
	write( Fd, "\n", 1);	
	
	return 1;
}
/*}}}*/



/****************************************************************************
 * Here are our exported functions
 */

int startKaffe ( void* parent, int x, int y, int w, int h,
				 int argc, char** argv ) /*{{{*/
{	
	/* we need to sync this.. */
	if ( State == S_UNINITIALIZED ) {
		if ( !initialize() ) {
			return 0;
		}
		State = S_INITIALIZED;
	}
	
	if ( State != S_ERROR ){
		if ( sendAppletRequest( parent, x, y, w, h, argc, argv) ){
			return 1;  /* success, finally */
		}
	}
	
	return 0;
}
/*}}}*/
