/* -*- mode: C; mode: fold -*- */

/* includes & defines */ /*{{{*/
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>

#define INSET 40
/*}}}*/

/* global data defs */ /*{{{*/
static Display *Dsp;
static int      Nscr;

static Window  Wnd;
static GC      Gc;

static int     Argc;
static char**  Argv;

static int Width, Height;
static int X, Y, W, H;


static long StdEvents = ExposureMask | KeyPressMask | KeyReleaseMask |
                 PointerMotionMask | 
                 ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                 EnterWindowMask | LeaveWindowMask | StructureNotifyMask |
                 FocusChangeMask | VisibilityChangeMask;
/*}}}*/


/* this is the plugin launch function */
extern int startKaffe ( void* parent, int x, int y, int w, int h,
					    int argc, char** argv );


Window createMainWindow () /*{{{*/
{
	Window wnd;
	unsigned long valueMask;
	XSetWindowAttributes attributes;
	XSizeHints sizeHints;
	XClassHint classHints;


	attributes.event_mask = StdEvents;
	attributes.background_pixel = WhitePixel( Dsp, Nscr);
	attributes.bit_gravity = ForgetGravity;
    attributes.backing_store = WhenMapped;

	valueMask = CWEventMask | CWBackPixel | CWBitGravity | CWBackingStore;

	wnd = XCreateWindow( Dsp, RootWindow( Dsp, Nscr), 0, 0, Width, Height, 0,
						 CopyFromParent, InputOutput, CopyFromParent,
						 valueMask, &attributes);

	if ( wnd ) {
		sizeHints.flags = PSize;
		classHints.res_name  = "kaffeenv";
		classHints.res_class = "KaffeEnv";
		XSetWMProperties( Dsp, wnd, 0, 0, 0, 0, &sizeHints, 0, &classHints);
	
		XStoreName( Dsp, wnd, "KaffeEnv");

		XMapRaised( Dsp, wnd);
	}

	return wnd;
}
/*}}}*/

void getGeometry () /*{{{*/
{
	int i;
	char *arg;

	/* we just center, for now (by means of a '-[0-9]+x[0-9]' opt) */ 
	
	if ( Argc ){
		arg = Argv[0];
		if ( (arg[0] == '-') && (arg[1] >= '0') && (arg[1] <= '9') ){
			arg++;
			i = sscanf( arg, "%dx%d", &W, &H);
			if ( i > 0 ){
				Width = W + 2*INSET;
				if ( i > 1 )
					Height = H + 2*INSET;
			}
			
			Argc--;
			Argv++;
		}
		else {
			H = Height - 2*INSET;
			W = Width - 2*INSET;
		}
	}

	X = INSET;
	Y = INSET;
}
/*}}}*/

void startApplet () /*{{{*/
{
	int i;
	
	printf( "++ start Applet [%d,%d - %d,%d]: ", X, Y, W, H);
	for ( i=0; i<Argc; i++ ){
		printf( "%s ", Argv[i]);
	}
	puts("");
	
	if ( !startKaffe( (void*)Wnd, X, Y, W, H, Argc, Argv) ){
		fprintf(stderr, "kaffe_startApplet failed\n");
	}
}
/*}}}*/

void processEvents () /*{{{*/
{
	XEvent evt;
	unsigned long valueMask = GCForeground;
	XGCValues values;
	KeySym keysym;
	unsigned char buf[16];
	XComposeStatus ioStatus;
	int n;
	
	while ( 1 ) {
		XNextEvent( Dsp, &evt);

		switch ( evt.type ){
		 case Expose:
			if ( !Gc ){
				values.foreground = BlackPixel( Dsp, Nscr);
				Gc = XCreateGC( Dsp, Wnd, valueMask, &values); 
			}
			
			XDrawRectangle( Dsp, Wnd, Gc, X-1, Y-1, W+1, H+1);
			break;
			
		 case KeyPress:
			n = XLookupString( &(evt.xkey), buf, sizeof( buf), 0, &ioStatus);
			keysym = XKeycodeToKeysym( Dsp, evt.xkey.keycode, 0);
			
			if ( n == 1 ){
				switch ( buf[0] ){
				 case 'q':
					printf( "++ quit kaffeenv\n");
					return;
					
				 case 's':
					startApplet();
					break;
					
				 default:
					printf( "++ kaffeenv received key: '%c'\n", buf[0]);
					break;
				}
			}

			break;
			
		 case ButtonPress:
			XSetInputFocus( Dsp, Wnd, RevertToParent, CurrentTime);
			break;
		}
	}
}
/*}}}*/


int main ( int argc, char** argv ) /*{{{*/
{
	Argc = argc;
	Argv = argv;

	/* we aren't interested in the prog name */
	Argv++;
	Argc--;
	
	if ( Argc < 2 ){
		puts( "kaffeenv - X environment for embedded AWT tests");
		puts( "usage: kaffeenv [-<width>x<height>] <classname> [<args>..]");
		return 0;
	}
	
	if ( !(Dsp = XOpenDisplay( NULL)) ){
		fprintf( stderr, "can't open display %s\n",
				 XDisplayName( NULL));
		return 1;
	}

	Nscr = DefaultScreen( Dsp);

	getGeometry();
	
	if ( !(Wnd = createMainWindow()) ){
		fprintf( stderr, "can't create main window\n");
		return 1;
	}
	
	processEvents();

	XUnmapWindow( Dsp, Wnd);
	XCloseDisplay( Dsp);

	return 0;
}
/*}}}*/
