/*
 * bag.h
 * Generic bag
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __bag_h
#define __bag_h

typedef unsigned int (*hashfunc)(void* key);
typedef int (*equalfunc)(void* bagdata, void* key);

typedef struct bag {
	void**          keys;
	void**          values;
	int*            refcounts;
	int             count;
	int             size;
	int		pressure;
	int		memtype;
	hashfunc	hashcode;
        equalfunc	equals;
} bag;

extern void  bagStart(bag* bg, int size, int pressure, hashfunc hash, equalfunc equal, int memtype);
extern void  bagFinish(bag* bg);
extern void* bagPut(bag* bg, void* key, void* value);
extern void* bagGet(bag* bg, void* key);
extern void  bagRemove(bag* bg, void* key);
extern void  bagRehash(bag* bg);

#endif
