/*
 * baseClasses.c
 * Handle base classes.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "config-signal.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "threads.h"
#include "lookup.h"
#include "slots.h"
#include "machine.h"
#include "exception.h"
#include "itypes.h"
#include "engine.h"
#include "strings.h"
#include "md.h"

#include "jdebug.h"

static int baseClassesInit;

char* init_name;
char* final_name;
char* dispatch_name;
char* constructor_name;
char* Code_name;
char* LineNumberTable_name;
char* ConstantValue_name;
char* Exceptions_name;
char* SourceFile_name;
char* LocalVariableTable_name;

char* void_sig;

static Hjava_lang_Class dummyClassClass;
Hjava_lang_Class* ClassClass = &dummyClassClass;
Hjava_lang_Class* StringClass;
Hjava_lang_Class* ObjectClass;
Hjava_lang_Class* SerialClass;
Hjava_lang_Class* CloneClass;
Hjava_lang_Class* OOMClass;
Hjava_lang_Class* SOClass;
Hjava_lang_Class* ClassLoaderClass;
Hjava_lang_Class* javaLangArrayIndexOutOfBoundsException;
Hjava_lang_Class* PrintWriterClass;
Hjava_lang_Class* ThreadClass;
Hjava_lang_Class* ThreadGroupClass;
Hjava_lang_Class* ThrowableClass;
Hjava_lang_Class* SoftcallClass;

Hjava_lang_ClassLoader systemClassLoader;

Method* loadClassMethod;
Method* printlnMethod;
Method* flushMethod;
Method* addThreadMethod;
Method* printStackTraceMethod;

/* Initialisation prototypes */
void initClasspath(void);
void initNative(void);
void initNativeThreads(void);
void initLocking(void);
void initThreads(void);
void initTypes(void);
void initFilesystem(void);
void initNetwork(void);

void finishTypes(void);

/*
 * Initialise the machine.
 */
void
initialiseKaffe(void)
{
DBG( vm, ("JVM starting up ...\n"));

	/* Machine specific initialisation first */
#if defined(INIT_MD)
	INIT_MD();
#endif

	/* The native thread system */
DBG( vm,("initNativeThreads\n"));
	initNativeThreads();

	/* The locking system */
DBG( vm,("initLocking\n"));
	initLocking();

	/* Setup the file system */
DBG( vm,("initFilesystem\n"));
	initFilesystem();

	/* Setup the network system */
DBG( vm,("initFilesystem\n"));
	initNetwork();

	/* Setup CLASSPATH */
DBG( vm,("initClasspath\n"));
	initClasspath();

	/* Init native support */
DBG( vm,("initNative\n"));
	initNative();

	/* Init the execution engine */
DBG( vm,("initEngine\n"));
	Kaffe_EngineInterface_init();

	/* Init string pools */
	initStrings();

	/* Create the initialise and finalize names and signatures. */
	init_name = addUtf("<clinit>");
	final_name = addUtf("finalize");
	dispatch_name = addUtf("dispatch");
	constructor_name = addUtf("<init>");
	Code_name = addUtf("Code");
	LineNumberTable_name = addUtf("LineNumberTable");
	ConstantValue_name = addUtf("ConstantValue");
	Exceptions_name = addUtf("Exceptions");
	SourceFile_name = addUtf("SourceFile");
	LocalVariableTable_name = addUtf("LocalVariableTable");

	/* Build void signature now */
	void_sig = addUtf("()V");

	/* Init system class loader */
	gc_add_ref(&systemClassLoader, 1+&systemClassLoader);

	/* Primitive types */
	initTypes();

DBG( vm,("initPreload\n"));
	initPreload();

DBG( vm,("initBaseClasses\n"));
	/* Read in base classes */
	initBaseClasses();

DBG( vm,("initThreads\n"));
	/* Init thread support */
	initThreads();

	/* We must process certain classes now that we might use internally */
	processClass(ClassClass, CSTATE_COMPLETE);
        CLASS_GC_TYPE( StringClass ) = GC_ALLOC_STRING;
	processClass(StringClass, CSTATE_COMPLETE);
}

/*
 * We need to use certain classes in the internal machine so we better
 * get them in now in a known way so we can refer back to them.
 * Currently we need java.lang.Object, java.lang.Class, java.lang.String
 * and java.langSystem.
 */
void
initBaseClasses(void)
{
	if (baseClassesInit != 0) {
		return;
	}
	baseClassesInit = 1;

#define	loadStaticClass(C,N) C = loadClass((N), 0, CSTATE_LINKED)

	/* The base types */
	loadStaticClass(ObjectClass, "java.lang.Object");
	loadStaticClass(SerialClass, "java.io.Serializable");
	loadStaticClass(CloneClass, "java.lang.Cloneable");
	loadStaticClass(ClassClass, "java.lang.Class");
	loadStaticClass(StringClass, "java.lang.String");
	loadStaticClass(ClassLoaderClass, "java.lang.ClassLoader");
	loadStaticClass(PrintWriterClass, "java.io.PrintWriter");
	loadStaticClass(ThreadClass, "java.lang.Thread");
	loadStaticClass(ThreadGroupClass, "java.lang.ThreadGroup");

	/* Exception handling types */
	loadStaticClass(ThrowableClass, "java.lang.Throwable");
	loadStaticClass(javaLangArrayIndexOutOfBoundsException, "java.lang.ArrayIndexOutOfBoundsException");
	loadStaticClass(OOMClass, "java.lang.OutOfMemoryError");
	loadStaticClass(SOClass, "java.lang.StackOverflowError");
	loadStaticClass(SoftcallClass, "kaffe.lang.Softcall");

	/* Methods */
	loadClassMethod = lookupClassMethod(ClassLoaderClass, "loadClass", "(Ljava/lang/String;Z)Ljava/lang/Class;");
	addThreadMethod = lookupClassMethod(ThreadGroupClass, "add", "(Ljava/lang/Thread;)V");
	printStackTraceMethod = lookupClassMethod(ThrowableClass, "printStackTrace", "()V");
}
