/*
 * classMethod.h
 * Class, method and field tables.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __classmethod_h
#define __classmethod_h

#include "gtypes.h"
#include "objects.h"
#include "constants.h"

struct _iLock;
struct Hjava_lang_Class;
struct _NativeLibrary;
struct bag;

typedef struct Hjava_lang_ClassLoader {
	Hjava_lang_Object		head;
	struct bag*			classPool;
	struct _NativeLibrary*		nativeLibs;
	void*				defaultProtectionDomain;
	struct Hjava_lang_ClassLoader*	parent;
} Hjava_lang_ClassLoader;

#ifdef DIRECT_CALL
#define	DEFINE_DISPATCH_TABLE(N,S)			\
	struct _##N {					\
		struct Hjava_lang_Class* class;		\
		uint32			 vsize;		\
		void*			 method[(S)];	\
	} N;
#else
#define	DEFINE_DISPATCH_TABLE(N,S)			\
	struct _##N {					\
		struct Hjava_lang_Class* class;		\
		uint32			 vsize;		\
		void**			 method[(S)];	\
	} N;
#endif
typedef DEFINE_DISPATCH_TABLE(dispatchTable,1);


#define __DEFINED_CLASS
typedef struct Hjava_lang_Class {

	/*
	 * Write many, read many
	 * This part of the class is changed as the class is used
	 */

	Hjava_lang_Object	head;		/* A class is an object too */
	struct _iLock*		lock;
	void*			owner;
	uint8			cstate;

	/*
	 * Write once, read many.
	 * Below here everything is constant once the class has been built
	 */

	accessFlags		accflags;
	char*			name;

	/* If non-NULL, a pointer to the superclass.
	 * However, if state < CSTATE_DOING_PREPARE, then
	 * (int) superclass is a constant pool index. */
	struct Hjava_lang_Class* superclass;

	struct _constants	constants;

	/* For regular classes, an array of the methods defined in this class.
	   For array types, used for CLASS_ELEMENT_TYPE.
	   For primitive types, used by CLASS_ARRAY_CACHE. */
	Method*			methods;
	short			nmethods;

	/* Number of methods in the dtable. */
	/* If CLASS_IS_PRIMITIVE, then the CLASS_PRIM_SIG. */
	short			msize;

	/* Pointer to array of Fields, one for each field.
	   Static fields come first. */
	Field*			fields;

	/* The size of the non-static fields, in bytes.
	   For a primitive type, the length in bytes.
	   Also used temporarily while reading fields.  */
	int			bfsize;

	/* Number of fields, including static fields. */
	short			nfields;
	/* Number of static fields. */
	short			nsfields;

	dispatchTable*		dtable;

	Hjava_lang_ClassLoader*	loader;
	struct Hjava_lang_Class** interfaces;
	short			interface_len;
	short			total_interface_len;
	short			flags;
	short			alignment;
	Method*			finalizer;	/* Structures to speedup GC */
	void*			sdata;		/* The static data */
	char*			filename;
#if defined(MEM1)
	unsigned int*		refs;
	unsigned int		nrefs; 
#elif defined(MEM2)
	unsigned int*		refs;
	unsigned int		nrefs; 
	void*			freelist;
	void*			gc_type;
#endif
#if defined( DIRECT_CALL )
        struct _patch_list*	patch;
#endif
} Hjava_lang_Class;

#define METHOD_TRANSLATED(M)		((M)->accflags & ACC_TRANSLATED)
#define METHOD_FLUSHED(M)		((M)->accflags & ACC_CACHEFLUSHED)
#define SET_METHOD_FLUSHED(M)		(M)->accflags |= ACC_CACHEFLUSHED
#define	SET_METHOD_NATIVECODE(M, S, E)	METHOD_NATIVECODE(M) = (S); \
					METHOD_NATIVECODE_END(M) = (E); \
					(M)->accflags |= ACC_TRANSLATED
#if defined(TRANSLATOR)
#define	METHOD_INDIRECTMETHOD(M)	METHOD_NATIVECODE(M)
#define	METHOD_NATIVECODE(M)		((M)->ncode)
#else
#define	METHOD_INDIRECTMETHOD(M)	(M)
#define	METHOD_NATIVECODE(M)		(*(void**)&(M)->bcode)
#endif
#define	METHOD_NATIVECODE_END(M)	((M)->ncodee)
#define	METHOD_CNAME(M)			((M)->name)

/*
 * Class flags.
 */
#define	FLAG_TYPE		0x3F

/* Class state */
#define	CSTATE_UNLOADED		0
#define CSTATE_LINKED		1
#define	CSTATE_DOING_CONSTINIT	2
#define	CSTATE_COMPLETE		3
#define	CSTATE_FAILED		4

#define	CLASS_STATE(C)		((C)->cstate)

struct bytecodeInfo;
struct _declexception;
struct _jexceptionEntry;

typedef struct _methods {
	char*			name;
	char*			signature;
	accessFlags		accflags;
	short			idx;	/* Index into class->dtable */
        void*			ncode;
        void*			ncodee;
#if defined(JIT4)
        struct bytecodeInfo*	bcode;
#endif
	struct _jexceptionEntry* exception_table;
#if !defined(HAVE_LIMITED_REFLECTION)
	struct _declexception*	exception_declared;
#endif
#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
	/* These structures not needed when optimizing */
	/* Line number information for exception reporting */
	struct _lineNumbers*	lines;
	/* Local variable information for jit debugging */
	struct _localVariables*	vars;
#endif
#if defined( DIRECT_CALL )
        struct _patch_list*	patch;
#endif
} methods;

#define	BCODE(M)		(*(bytecodeInfo**)&((M)->ncode))
#define	METHOD_BYTECODE_LEN(M)	(BCODE(M)->codelen)
#define	METHOD_BYTECODE_CODE(M)	(BCODE(M)->code)

#define	DTABLE_CLASS		0
#define	DTABLE_METHODOFFSET	(sizeof(void*) + sizeof(void*))
#define	DTABLE_METHODSIZE	(sizeof(void*))

typedef struct _fields {
	char*			name;
	char*			ftype;
	accessFlags		accflags;
	u2			bsize;		/* in bytes */
	union {
		int		boffset;	/* offset in bytes (object) */
		void*		addr;		/* address (static) */
		u2		idx;		/* constant value index */
	} info;
} fields;

#define	FIELD_CONSTANT_VALUE	0x4000
#define	FIELD_REFERENCE		0x2000

#define	FIELD_CNAME(FLD)	((FLD)->name)
#define	FIELD_TYPE_CNAME(FLD)	((FLD)->ftype)
/* Size of field FLD, in bytes. */
#define FIELD_SIZE(FLD)		((FLD)->bsize)
#define FIELD_WSIZE(FLD)	((FLD)->bsize <= sizeof(jint) ? 1 : 2)
#define FIELD_BOFFSET(FLD)	((FLD)->info.boffset)
#define FIELD_ADDRESS(FLD)	((FLD)->info.addr)
#define	FIELD_CONSTIDX(FLD)	((FLD)->info.idx)
#define	FIELD_ISREF(FLD)	(FIELD_TYPE_CNAME(FLD)[0] == 'L' || FIELD_TYPE_CNAME(FLD)[0] == '[')
#define	FIELD_IS_FINAL(FLD)	((FLD)->accflags & ACC_FINAL)

#define	CLASSMAXSIG		256

struct _Code;
struct _method_info;
struct _field_info;
struct _classFile;

#define CLASS_METHODS(CLASS)	((CLASS)->methods)
#define CLASS_NMETHODS(CLASS)	((CLASS)->nmethods)

/* An array containing all the Fields, static fields first. */
#define CLASS_FIELDS(CLASS)	((CLASS)->fields)

/* An array containing all the static Fields. */
#define CLASS_SFIELDS(CLASS)	((CLASS)->fields)

/* An array containing all the instance (non-static) Fields. */
#define CLASS_IFIELDS(CL)	(&(CL)->fields[CLASS_NSFIELDS(CL)])

/* Total number of fields (instance and static). */
#define CLASS_NFIELDS(CLASS)	((CLASS)->nfields)
/* Number of instance (non-static) fields. */
#define CLASS_NIFIELDS(CLASS)	((CLASS)->nfields - (CLASS)->nsfields)
/* Number of static fields. */
#define CLASS_NSFIELDS(CLASS)	((CLASS)->nsfields)

/* Size of a class's fields (not counting header), in bytes. */
#define CLASS_FSIZE(CLASS)	((CLASS)->bfsize)

#define OBJECT_CLASS(OBJ)	((OBJ)->dtable->class)
#define CLASS_CNAME(CL)		((CL)->name)
#define _PRIMITIVE_DTABLE	((dispatchTable*)(-1))
#define CLASS_IS_PRIMITIVE(CL)	((CL)->dtable == _PRIMITIVE_DTABLE)

/* Assuming CLASS_IS_PRIMITIVE(CL), return the 1-letter signature code. */
#define CLASS_PRIM_SIG(CL)	((CL)->msize)
#define CLASS_PRIM_NAME(CL)	(*(char**)&(CL)->fields)

#define CLASS_IS_ARRAY(CL)	(CLASS_CNAME(CL)[0] == '[')

#define	CLASS_IS_INTERFACE(CL)	((CL)->accflags & ACC_INTERFACE)
#define	CLASS_IS_ABSTRACT(CL)	((CL)->accflags & ACC_ABSTRACT)
#define	CLASS_IS_FINAL(CL)	((CL)->accflags & ACC_FINAL)

/* For an array type, the types of the elements. */
#define CLASS_ELEMENT_TYPE(ARRAYCLASS) (*(Hjava_lang_Class**)&(ARRAYCLASS)->methods)

/* Used by the lookupArray function. */
#define CLASS_ARRAY_CACHE(PRIMTYPE) (*(Hjava_lang_Class**)&(PRIMTYPE)->methods)

#if defined(MEM2)
#define SET_GC_TYPE( _c, _t ) (_c)->gc_type = (_t)
#define GC_TYPE( _c ) (_c)->gc_type
#define CLASS_GC_TYPE( _c ) (_c)->gc_type
#else
#define SET_GC_TYPE( _c, _t ) (_c)->flags |= (_t)
#define GC_TYPE( _c ) (_c)->flags
#define CLASS_GC_TYPE( _c ) (_c)->flags
#endif

#define TYPE_PRIM_SIZE(CL) ((CL)->bfsize)
#define TYPE_SIZE(CL) \
  (CLASS_IS_PRIMITIVE(CL) ? TYPE_PRIM_SIZE (CL) : PTR_TYPE_SIZE)

#define	METHOD_IS_PRIVATE(METH) ((METH)->accflags & ACC_PRIVATE)
#define	METHOD_IS_FINAL(METH) ((METH)->accflags & ACC_FINAL)
#define	METHOD_IS_STATIC(METH) ((METH)->accflags & ACC_STATIC)
#define	METHOD_IS_NATIVE(METH) ((METH)->accflags & ACC_NATIVE)
#define	METHOD_IS_CONSTRUCTOR(METH) ((METH)->accflags & ACC_CONSTRUCTOR)
#define	METHOD_IS_SYNCHRONISED(METH) ((METH)->accflags & ACC_SYNCHRONISED)

#define	METHOD_IS_INHERITABLE(METH) \
	(!(METHOD_IS_STATIC(METH) || METHOD_IS_PRIVATE(METH) || METHOD_IS_CONSTRUCTOR(METH)))

#define	CLASS_IS_PRECOMPILED(C)		((C)->sdata == (void*)1)

/*
 * 'processClass' is the core of the class initialiser and can prepare a
 * class from the cradle to the grave.
 */
JNIEXPORT int	JNICALL		processClass(Hjava_lang_Class*, int);

JNIEXPORT Hjava_lang_Class*	JNICALL loadClass(char*, Hjava_lang_ClassLoader*, int);
JNIEXPORT Hjava_lang_Class*	JNICALL loadArray(char*, Hjava_lang_ClassLoader*);

JNIEXPORT Hjava_lang_Class*	JNICALL setupClass(Hjava_lang_Class*, constIndex, constIndex, u2, Hjava_lang_ClassLoader*);
JNIEXPORT Method*		JNICALL addMethod(Hjava_lang_Class*, struct _method_info*);
JNIEXPORT Method*		JNICALL addExceptionMethod(Hjava_lang_Class*, char*, char*);
JNIEXPORT Field*        	JNICALL addField(Hjava_lang_Class*, struct _field_info*);
JNIEXPORT void			JNICALL addInterfaces(Hjava_lang_Class*, int, Hjava_lang_Class**);
JNIEXPORT void			JNICALL addSourceFilename(Hjava_lang_Class*, int);

JNIEXPORT void			JNICALL setFieldValue(Field*, u2);
JNIEXPORT Hjava_lang_Class*	JNICALL resolveFieldType(Field*, Hjava_lang_Class*);

JNIEXPORT Hjava_lang_Class*	JNICALL lookupClass(char*, Hjava_lang_ClassLoader*);
JNIEXPORT Hjava_lang_Class*	JNICALL addClass(Hjava_lang_Class*);
JNIEXPORT Hjava_lang_Class*	JNICALL lookupArray(Hjava_lang_Class*);
JNIEXPORT Hjava_lang_Class*	JNICALL lookupArrayProcess(Hjava_lang_Class*, int);
JNIEXPORT Hjava_lang_Class*	JNICALL lookupObjectArrayClass(Hjava_lang_Class*);
JNIEXPORT Field*		JNICALL lookupClassField(Hjava_lang_Class*, char*);

JNIEXPORT Hjava_lang_Class*	JNICALL getClass(constIndex, Hjava_lang_Class*, char**);

JNIEXPORT void			JNICALL countInsAndOuts(char*, short*, short*, char*);
JNIEXPORT void			JNICALL establishMethod(Method*);
JNIEXPORT Hjava_lang_Class*	JNICALL classFromSig(char**, Hjava_lang_ClassLoader*);
JNIEXPORT Hjava_lang_Class*	JNICALL classFromSigProcess(char**, Hjava_lang_ClassLoader*, int);
JNIEXPORT Hjava_lang_Class*	JNICALL getClassFromSignature(char*, Hjava_lang_ClassLoader*);
JNIEXPORT char*			JNICALL utfFromSig(char**);

JNIEXPORT void			JNICALL finishFields(Hjava_lang_Class*);
JNIEXPORT Method*		JNICALL findMethodFromPC(uintp);

JNIEXPORT void			JNICALL tidyClasses(void);
JNIEXPORT Hjava_lang_ClassLoader*	JNICALL getCurrentClassLoader(void);
JNIEXPORT void			JNICALL unloadClass(Hjava_lang_Class*);
JNIEXPORT void			JNICALL unloadClassLoader(Hjava_lang_ClassLoader*);
JNIEXPORT jboolean		JNICALL findClass(Hjava_lang_Class*);
JNIEXPORT void			JNICALL initPreload(void);
JNIEXPORT Method*		JNICALL lookupClassMethod(Hjava_lang_Class*, char*, char*);
JNIEXPORT void			JNICALL registerPrecompiledClasses(void*);

JNIEXPORT char* init_name;		/* "<clinit>" */
JNIEXPORT char* constructor_name;	/* "<init>" */
JNIEXPORT char* final_name;		/* "finalize" */
JNIEXPORT char* dispatch_name;		/* "dispatch" */
JNIEXPORT char* Code_name;		/* "Code" */
JNIEXPORT char* LineNumberTable_name;	/* "LineNumberTable" */
JNIEXPORT char* ConstantValue_name;	/* "ConstantValue" */
JNIEXPORT char* Exceptions_name;	/* "Exceptions" */
JNIEXPORT char* SourceFile_name;	/* "SourceFile" */
JNIEXPORT char* LocalVariableTable_name;/* "LocalVariableTable" */
JNIEXPORT char* void_sig;		/* "()V" */

#endif
