/*
 * cms.h
 *
 * Copyright (c) 1996-1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __cms_h
#define __cms_h

typedef struct BYTECODE_STATE_s {

	jint	version;	/* Used to distinguish different version of
				   this table */
	jref	optop;		/* Value of stack register when initializing
				   the bytecode machine */
	jref	j_sc_bottom;
	jref	j_const_pool;
	jbyte	initialized;	/* True if machine has been initialized */

	jbyte	bc_exit[2];	/* Exit bytecode */
	jbyte	bc_ret[10];	/* Return from C bytecode */

	/* Functions for opcodes that cannot be executed directly */

	jref	(*bc_new)(void);
	jref	(*bc_newarray)(jint);
	jref	(*bc_anewarray)(jint);
	jref	(*bc_multianewarray)(jint, ...);
	jint	(*bc_instanceof)(jref);
	void	(*bc_checkcast)(jref);
	void	(*bc_athrow)(jref);
	void	(*bc_monitorenter)(jref);
	void	(*bc_monitorexit)(jref);

	/* ... */

	/* Functions for exceptions */

	void	(*ex_badarrayindex)(void);
	void	(*ex_nullpointer)(void);
	void	(*ex_divzero)(void);
	void	(*ex_nosuchmethod)(void);
	void	(*ex_nosuchfield)(void);
	void	(*ex_stackoverflow)(void);

	/* ... */

	/* Others - these don't fit in anywhere */

	void	(*ot_initialize_class)(void);
	jref	(*ot_lookupmethod)(jref, jref, jref);
	void	(*ot_checkarraystore)(jref, jref);

	/* ... */

} BYTECODE_STATE_s;

typedef BYTECODE_STATE_s* BYTECODE_STATE;

extern BYTECODE_STATE engineInterface;

/*
 * CMS engine prologue and epilogue for the various 'trap' instructions.
 */

/* For the real CMS system the CMS_PC comes in the 'eax' register.  However
 * this doen't work well with GCC since it's impossible to write an asm
 * statement to put the pc into eax and then make the engine call (which
 * returns its result in eax) without it thinking eax will be clobbered and
 * so screwing up the value.  We therefore use esi for now since it's safe
 * to play with this register.
 */
/*#define	CMS_PC_REG	"eax" */
#define	CMS_PC_REG	"esi"

#define	CMS_LOAD_PC(PC)	\
	asm volatile("movl %0,%%esi" : : "r" (PC) : CMS_PC_REG)

#define	CMS_STORE_PC(PC) \
	asm volatile("movl %%esi,%0" : "=r" (PC) : : CMS_PC_REG)

#define	CMS_PROLOGUE() \
	uint8* cms_pc; \
	CMS_STORE_PC(cms_pc)

#define	CMS_EPILOGUE()

#define	CMS_PC(N)	cms_pc[(N)]

#define	CMS_CALL(F,A) \
	({ \
		CMS_LOAD_PC(code + mjbuf.pc); \
		(*engineInterface->F)A; \
	})

#endif
