/* code-analyse.h
 * Analyse a method's bytecodes.
 *
 * Copyright (c) 1996 - 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __code_analyse_h
#define __code_analyse_h

struct _methods;
struct Hjava_lang_Class;

typedef	struct _frameElement {
	uint8			type;
	uint8			used;
	uint8			modified;
} frameElement;

typedef struct _basicBlockInfo {
	struct _basicBlockInfo*	next;
	uint16			stackPointer;
	uint16			flags;
	uint16			start;
	uint16			end;
} basicBlockInfo;

typedef struct _localUse {
	int32			use;
	int32			first;
	int32			last;
	int32			write;
	uint8			type;
} localUse;

typedef struct codeinfo {
	uint16			codelen;
	uint16			stacksz;
	uint16			localsz;
	localUse*		localuse;
	basicBlockInfo*		bbinfo;
	uintp			nativePC[1];
} codeinfo;

/* -------------------------------------------------------------------- */

#define	TVOID				0x0001
#define	TINT				0x0002
#define	TLONG				0x0004
#define	TFLOAT				0x0008
#define	TDOUBLE				0x0010
#define	TOBJ				0x0020

/* -------------------------------------------------------------------- */

#define	INSN(pc)			BCODE(meth)->code[(pc)]
#define	INSNLEN(pc)			insnLen[INSN(pc)]

#define	BYTE(pc)			(int8)(BCODE(meth)->code[(pc)+0])
#define	WORD(pc)			(int16)( \
					 (BCODE(meth)->code[(pc)+0] << 8) + \
					 (BCODE(meth)->code[(pc)+1]))
#define	DWORD(pc)			(int32)( \
					 (BCODE(meth)->code[(pc)+0] << 24) + \
					 (BCODE(meth)->code[(pc)+1] << 16) + \
					 (BCODE(meth)->code[(pc)+2] << 8) + \
					 (BCODE(meth)->code[(pc)+3]))

#define	INCPC(V)			pc += (V)

#define	SET_INSNPC(P, V)		codeInfo->nativePC[(P)] = ((V)+1)
#define	INSNPC(P)			(codeInfo->nativePC[(P)]-1)

#define FRAME_SZ			(codeInfo->localsz + codeInfo->stacksz + 1)
#define	ALLOCFRAME()			allocFrame(FRAME_SZ)

/* -------------------------------------------------------------------- */

#define	FLAG_STARTOFBASICBLOCK		0x0001
#define	FLAG_STARTOFEXCEPTION		0x0002
#define	FLAG_DONEVERIFY			0x0004
#define	FLAG_NORMALFLOW			0x0008

#define	INSERT_NEW_BLOCK(PC)					\
	insertBB(codeInfo, (PC))

#define	ADVANCE_BB(PC)						\
	while ((PC) >= currBB->next->start) {			\
		currBB = currBB->next;				\
	}

#define	IS_UNREACHABLE(PC)					\
	((PC) >= currBB->end)

#define	IS_STARTOFBASICBLOCK(PC)				\
	(currBB->start == (PC) && (currBB->flags & FLAG_STARTOFBASICBLOCK) != 0)

#define	IS_STARTOFEXCEPTION(PC)					\
	(currBB->start == (PC) && (currBB->flags & FLAG_STARTOFEXCEPTION) != 0)

#define	IS_NORMALFLOW(PC)					\
	(currBB->start == (PC) && (currBB->flags & FLAG_NORMALFLOW) != 0)

#define	IS_DONEVERIFY(PC)					\
	(currBB->flags & FLAG_DONEVERIFY)

#define	IS_NEEDVERIFY(PC)					\
	(currBB->stackPointer != 0xFFFF)

#define	STACKPOINTER(PC)					\
	(currBB->stackPointer)

#define	SET_STARTOFBASICBLOCK(PC)				\
	insertBB(codeInfo, (PC))->flags |= FLAG_STARTOFBASICBLOCK

#define	SET_STARTOFEXCEPTION(PC)				\
	insertBB(codeInfo, (PC))->flags |= FLAG_STARTOFEXCEPTION

#define	SET_STACKPOINTER(PC, SP)				\
	insertBB(codeInfo, (PC))->stackPointer = (SP)

/* -------------------------------------------------------------------- */

#define	STKPUSH(L)			sp -= (L)
#define	STKPOP(L)			sp += (L)

#define	SF(S)				activeFrame[sp+(S)]
#define	LCL(L)				activeFrame[(L)]

/* -------------------------------------------------------------------- */

extern int verifyMethod(struct _methods*, struct Hjava_lang_Class*, codeinfo**);
extern void tidyVerifyMethod(codeinfo**);

extern const int8 insnLen[];
extern const int8 insnStk[];

#endif
