/*
 * code.c
 * Process a new code attribute.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "jtypes.h"
#include "file.h"
#include "classMethod.h"
#include "bytecode.h"
#include "code.h"
#include "access.h"
#include "objects.h"
#include "readClass.h"
#include "slots.h"
#include "exception.h"
#include "strings.h"

#include "jmeasure.h"
#include "jdebug.h"

void
addCode(Method* m, Hjava_lang_Class* class, uint32 len, classFile* fp)
{
	int i;
	u2 i2;
	u2 elen;
	u4 clen;
	u2 stacksz;
	u2 localsz;
	int methsync;
	char* tmp;
	char* tmp2;

	methsync = 0;
	if (METHOD_IS_SYNCHRONISED(m)) {
		methsync = 1;
	}

	readu2(&stacksz, fp);
	readu2(&localsz, fp);
	readu4(&clen, fp);
DBG( vm_code, ("Max stack = %d\n", stacksz));
DBG( vm_code, ("Max locals = %d\n", localsz));
DBG( vm_code, ("Code length = %d\n", clen));
	if (clen > 0) {
		BCODE(m) = gc_malloc_fixed_typed(sizeof(bytecodeInfo) + clen - 1 + methsync, GC_ALLOC_BYTECODE);
#if defined(JIT4)
                m->bcode = BCODE(m);
#endif                
		BCODE(m)->codelen = clen + methsync;
		readm(BCODE(m)->code, clen, sizeof(bytecode), fp);
		BCODE(m)->stacksz = stacksz;
		BCODE(m)->localsz = localsz;
	}
	else {
		BCODE(m) = 0;
	}
	readu2(&elen, fp);
DBG( vm_code, ("Exception table length = %d\n", elen));
	if (elen > 0 || methsync > 0) {
		m->exception_table = gc_malloc_fixed_typed(((elen + methsync + 1) * sizeof(jexceptionEntry)), GC_ALLOC_EXCEPTIONTABLE);
		for (i = 0; i < elen; i++) {
                        jexceptionEntry* entry = &m->exception_table[i];
			readu2(&i2, fp);
			entry->start_pc = i2;
			readu2(&i2, fp);
			entry->end_pc = i2;
			readu2(&i2, fp);
			entry->handler_pc = i2;
			readu2(&i2, fp);
			tmp = WORD2UTF(CLASS_CONSTANTS(class)->data[i2]);
			if (tmp == 0) {
				entry->catch = 0;
			}
			else {
				UTF_REPLACE(tmp2, tmp, '/', '.');
				entry->catch = addUtfCopy(tmp2, -1);
			}
		}
		/* If method is sync, then we add a dummy exception handler
		 * on the end - it is used to catch and unlock this method
		 * during exception dispatch.
		 */
		if (methsync > 0) {
			jexceptionEntry* entry = &m->exception_table[i];
			entry->start_pc = 0;
			entry->end_pc = clen;
			entry->handler_pc = clen;
			entry->catch = 0;
			i++;
			/* If we're adding a dummy exception handler we will need a single
			 * stack slot.  If the stack is 0 we better increase it to one
			 * otherwise the verifier and JIT will get upset.
			 */
			if (BCODE(m)->stacksz == 0) {
				BCODE(m)->stacksz = 1;
			}
		}
		m->exception_table[i].start_pc = (uintp)-1;
	}
	else {
		m->exception_table = 0;
	}

MSR_ADD( bcode_mem, clen);

	readAttributes(fp, class, m);
}

/*
 * Read in line numbers assocated with code.
 */
void
addLineNumbers(Method* m, uint32 len, classFile* fp)
{
#if !defined(OPTIMIZE_MEMORY) || defined(DEBUG)
	lineNumbers* lines;
	int i;
	u2 nr;
	u2 data;

	readu2(&nr, fp);
	lines = gc_malloc_fixed(sizeof(lineNumbers)+sizeof(lineNumberEntry) * nr);

	lines->length = nr;
	for (i = 0; i < nr; i++) {
                lineNumberEntry* entry = &lines->entry[i];
		readu2(&data, fp);
		entry->pc_offset = data;
		readu2(&data, fp);
		entry->line_nr = data;
	}

	/* Attach lines to method */
	m->lines = lines;
#else
	seekm(fp, len);
#endif
}

/*
 * Read in line numbers assocated with code.
 */
void
addLocalVariables(Method* m, Hjava_lang_Class* class, uint32 len, classFile* fp)
{
#if 0 && ( !defined(OPTIMIZE_MEMORY) || defined(DEBUG) )
	localVariables* vars;
	int i;
	u2 nr, data;

	readu2(&nr, fp);
	vars = gc_malloc_fixed(sizeof(localVariables)+sizeof(localVariableEntry) * nr);

	vars->length = nr;
	for (i = 0; i < nr; i++) {
                localVariableEntry* entry = &vars->entry[i];
		readu2(&data, fp);
		entry->start_pc = data;
		readu2(&data, fp);
		entry->length = data;
		readu2(&data, fp);
                entry->name = addUtf(WORD2UTF(CLASS_CONSTANTS(class)->data[data]));
		readu2(&data, fp);
                entry->type = addUtf(WORD2UTF(CLASS_CONSTANTS(class)->data[data]));
		readu2(&data, fp);
		entry->idx = data;
	}

	/* Attach lines to method */
	m->vars = vars;
#else
	seekm(fp, len);
#endif
}

/*
 * Read in exceptions declared by this method.
 */
void
addDeclaredExceptions(Method* m, uint32 len, classFile* fp)
{
#if !defined(HAVE_LIMITED_REFLECTION)
	u2 nr;
	int i;
	declexception* e;

	readu2(&nr, fp);
	if (nr == 0) {
		return;
	}

	e = gc_malloc_fixed(sizeof(declexception) + sizeof(constIndex) * (nr-1));
	m->exception_declared = e;
	m->exception_declared->length = nr;

	for (i = 0; i < nr; i++) {
		readu2(&e->entry[i], fp);
	}
#endif
}
