/*
 * findInJar.c
 * Search the CLASSPATH for the given class or property name.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	LDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "gtypes.h"
#include "file.h"
#include "exception.h"
#include "readClass.h"
#include "classMethod.h"
#include "support.h"
#include "locks.h"
#include "lerrno.h"
#include "baseClasses.h"
#include "external.h"
#include "jar/jar.h"
#include "classpath.h"
#include "md.h"
#include "paths.h"
#include "jsyscall.h"
#include "jdebug.h"
#include "jmeasure.h"
#include "jsyscall.h"

#define	MAXBUF		256

int64 timeInJar = 0;
int64 timeInJarUncompress = 0;

void initClasspath(void);
int findInJar(char*, classFile*);

/*
 * Find the named class in a directory or JAR file.
 */
jboolean
findClass(Hjava_lang_Class* class)
{
	classFile hand;
	char* cname;
	char* buf;
	int len;
	int i;
	int iLockRoot;

	cname = CLASS_CNAME(class);

	len = strlen(cname);
	buf = alloca(len + 1 + 6);
	for (i = 0; i < len; i++) {
		if (cname[i] != '.') {
			buf[i] = cname[i];
		}
		else {
			buf[i] = '/';
		}
	}
	strcpy(&buf[len], ".class");

	/* Look for the class */
DBG( vm_jar,("scanning for class: %s\n", cname))


	/* Find class in Jar file */
	if (findInJar(buf, &hand) == 0) {
		return (false);
	}
	switch (hand.type) {
	case CP_DIR:
	case CP_ZIPFILE:
		lockMutex(&class->head);
		if (class->constants.size == 0) {
			class = readClass(class, &hand, NULL);
		}
		unlockMutex(&class->head);

		if (hand.base != 0) {
			gc_free_fixed(hand.base);
		}
		if (class == 0) {
			break;
		}
		return (true);

	default:
		break;
	}

	/*
	 * Certain classes are essential.  If we don't find them then
	 * abort.
	 */
	if (strcmp(cname, "java.lang.NoClassDefFoundError") == 0 ||
	    strcmp(cname, "java.lang.Object") == 0) {
		kprintf("Cannot find essential class '%s' in class library ... aborting.\n", cname);
		ABORT();
	}

LDBG(	kprintf("NOT FOUND: %s\n", cname);		)

	return (false);
}

/*
 * Locate the given name in the CLASSPATH.
 */
int
findInJar(char* cname, classFile* hand)
{
	char buf[MAXBUF];
	int fp;
	int j;
	int i;
	jarEntry* entry;
	static iLock* jarlock;
	int iLockRoot;
	int64 tms_jar;
	int64 tme_jar;
	classpathElement* elem;

	tms_jar = 0;
	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tms_jar = currtime();
	}


	/* Look for the class */
DBG( vm_jar, ("scanning for element: %s\n", cname);)

	/* One into the jar at once */
	lockStaticMutex(&jarlock);

	for (elem = classpath; elem != 0; elem = elem->next) {
		hand->type = elem->type;
		switch (elem->type) {
		case CP_ZIPFILE:
		{
DBG( vm_jar, 		("opening JAR file: %s for %s\n", elem->path, cname))
			if (elem->u.jar == 0) {
				elem->u.jar = openJarFile(elem->path);
				if (elem->u.jar == 0) {
					break;
				}
			}

			entry = lookupJarFile(elem->u.jar, cname);
			if (entry == 0) {
				break;
			}
			hand->base = gc_malloc_fixed_typed(entry->uncompressedSize, GC_ALLOC_JAR);
			if (hand->base == 0) {
				goto error;
			}
			if (getDataJarFile(elem->u.jar, entry, hand->base) == 0) {
				gc_free_fixed(hand->base);
				postException("java.io.IOException", elem->u.jar->error);
				goto error;
			}
			hand->size = entry->uncompressedSize;
			hand->buf = hand->base;
			if (Kaffe_JavaVMArgs[0].enableVerboseClassloading == 2) {
				kprintf("%s\n", cname);
			}
			else if (Kaffe_JavaVMArgs[0].enableVerboseClassloading) {
				kprintf("Loading %s(%s)", cname, elem->path);
				if (entry->compressionMethod != COMPRESSION_STORED) {
					kprintf(" [compressed]");
				}
				kprintf("\n");
			}
			goto okay;
		}

		case CP_DIR:
		{
			jlong len;

			strcpy(buf, elem->path);
			strcat(buf, DIRSEP);
			strcat(buf, cname);
DBG( vm_jar,		("opening java file: %s for %s\n", buf, cname))
			fp = open(buf, JOPEN_READ);
			if (fp < 0) {
				break;
			}
			if (fileinfoFD(fp, JINFO_LENGTH, &len) == 0) {
				close(fp);
				break;
			}
			hand->size = (int)len;
			hand->base = gc_malloc_fixed_typed(hand->size, GC_ALLOC_JAR);
			if (hand->base == 0) {
				close(fp);
				goto error;
			}
			hand->buf = hand->base;
			for (i = 0; i < hand->size; i += j) {
				j = read(fp, hand->buf + i, hand->size - i);
				if (j < 0 && errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR) {
					close(fp);
					gc_free_fixed(hand->base);
					postException("java.io.IOException", buf);
					goto error;
				}
			}
			close(fp);
			if (Kaffe_JavaVMArgs[0].enableVerboseClassloading) {
				kprintf("Loading %s\n", cname);
			}
			goto okay;
		}

		/* NOTE - CP_SOFILE types are loaded eagerly at startup
		 * so we don't deal with them here.
		 */

		/* Ignore bad entries */
		default:
			break;
		}
	}
	/* If we call out the loop then we didn't find anything */
	hand->type = CP_INVALID;
	okay:;

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tme_jar = currtime();
		timeInJar += (tme_jar - tms_jar);
	}

	unlockStaticMutex(&jarlock);

	return (1);

	error:;

	if (Kaffe_JavaVMArgs[0].enableVerboseJIT) {
		tme_jar = currtime();
		timeInJar += (tme_jar - tms_jar);
	}

	unlockStaticMutex(&jarlock);

	return (0);
}

/*
 * Initialise class path and the JAR sub-system.
 */
void
initClasspath(void)
{
	Kaffe_JarRawInterface_init();
	Kaffe_JarInterface_init();
	(*Kaffe_ClasspathInterface.init)();
}

/*
 * Preload any shared libraries in the classpath.
 */
void
initPreload(void)
{
	classpathElement* elem;

	for (elem = classpath; elem != 0; elem = elem->next) {
		if (elem->type == CP_SOFILE) {
			loadNativeLibrary(elem->path, 0);
		}
	}
}

/*
 * Initilise the filesystem.
 */
void
initFilesystem(void)
{
	(*Kaffe_FileSystemCallInterface.init)();
}

/*
 * Initilise the network system.
 */
void
initNetwork(void)
{
	(*Kaffe_NetworkCallInterface.init)();
}
