/*
 * machine.c
 * Java virtual machine interpreter.
 *
 * Copyright (c) 1996, 1997, 1998, 1999
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	DBG(s)
#define	RDBG(s)
#define	NDBG(s)
#define	IDBG(s)
#define	CHDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "md.h"
#include "classMethod.h"
#include "callMethodInfo.h"
#include "gtypes.h"
#include "bytecode.h"
#include "slots.h"
#include "icode.h"
#include "access.h"
#include "objects.h"
#include "constants.h"
#include "gc.h"
#include "machine.h"
#include "lookup.h"
#include "soft.h"
#include "exception.h"
#include "external.h"
#include "baseClasses.h"
#include "threads.h"
#include "locks.h"
#include "checks.h"
#include "engine.h"
#include "code-analyse.h"
#include "code.h"
#include "support.h"

char* engine_name = "Interpreter";
char* engine_version = KAFFEVERSION;

Method* interpreterMethod;

extern struct JNIEnv_ Kaffe_JNIEnv;

#if defined(__GNUC__)
#define	insn_dispatch_next()	{					\
					register uintp insn;		\
					insn = code[npc];		\
					mjbuf.pc = pc = npc;		\
					npc = npc + insnLen[insn];	\
					goto *insn_dispatch[insn];	\
				}

#define	define_insn(code)	insn_dispatch_next();			\
				insn_##code##:;

#define	define_wide_insn(code)	break;					\
				case code:				\
				IDBG( printf("%03d: %s\n", pc, #code); )

#define	define_insn_alias(code)	insn_##code##:;

#else

#define	define_insn(code)	break;					\
				case code:				\
				IDBG( printf("%03d: %s\n", pc, #code); )
#define	define_insn_alias(code)	case code:				\
				IDBG( printf("%03d: %s\n", pc, #code); )
#define	define_wide_insn(code)	break;					\
				case code:				\
				IDBG( printf("%03d: %s\n", pc, #code); )

#endif

/*
 * Define CREATE_NULLPOINTER_CHECKS in md.h when your machine cannot
 * use the MMU for detecting null pointer accesses
 */
#if !defined(CREATE_NULLPOINTER_CHECKS)
#undef	check_null
#define check_null(I, S, N)
#endif

/*
 * Define CREATE_DIVZERO_CHECKS in md.h when your machine cannot
 * use exceptions to detect / 0 errors.
 */
#if !defined(CREATE_DIVZERO_CHECKS)
#undef	check_div
#undef	check_div_long
#define	check_div(I, S, N)
#define	check_div_long(I, S, N)
#endif

/* Use built-in alloca where we can */
#if defined(__GNUC__)
#undef	alloca
#define	alloca(A)	__builtin_alloca(A)
#endif

void
interpreterStart(void)
{
}

void
interpreter(methods* meth, slots* arg, slots* retval, Hjava_lang_Thread* tid)
{
#if defined(__GNUC__)
#include "dispatch.h"
#endif

	Hjava_lang_Object* mobj;
	vmException mjbuf;
	accessFlags methaccflags;

	/* If these can be kept in registers then things will go much
	 * better.
	 */
	register slots* lcl;
	register slots* sp;
	register uintp npc;

	/* Variables used directly in the machine */
	int32 idx;
	jint low;

	/* Call, field and creation information */
	callInfo cinfo;
	fieldInfo finfo;
	Hjava_lang_Class* crinfo;

	int iLockRoot;

        if (Kaffe_JavaVMArgs[0].enableVerboseCall != 0) {
		soft_trace(meth, 0);
	}

	processClass(meth->class, CSTATE_COMPLETE);

	/* If this is native, then call the real function */
	methaccflags = meth->accflags;
	if (methaccflags & (ACC_TRANSLATED|ACC_NATIVE)) {
NDBG(		printf("Call to native %s.%s%s.\n", meth->class->name->data, meth->name->data, makeStrFromSignature(meth->sig)); fflush(stdout); )
		if (methaccflags & ACC_STATIC) {
			callMethodA(meth, meth, 0, (jvalue*)arg, (jvalue*)retval);
		}
		else {
			callMethodA(meth, meth, ((jvalue*)arg)[0].l, &((jvalue*)arg)[1], (jvalue*)retval);
		}
		return;
	}

#if defined(HAVE_VERIFIER)
	/* Verify method if required */
	if ((methaccflags & ACC_VERIFIED) == 0) {
		codeinfo* mycodeInfo;
		verifyMethod(meth, &mycodeInfo);
		tidyVerifyMethod(&mycodeInfo);
	}
#endif

	/* Allocate stack space and locals. */
	lcl = alloca(sizeof(slots) * (meth->bcode->localsz + meth->bcode->stacksz));

	if (tid != 0 &&
#if defined(STACK_GROWS_UP)
			(void*)&lcl[meth->bcode->localsz+meth->bcode->stacksz] > getCurrentStackLimit()
#else
			(void*)lcl < getCurrentStackLimit()
#endif
							) {
		soft_stackoverflow();
	}

	mobj = 0;
	npc = 0;

	mjbuf.frame = STACK_CURRENT_FRAME();
	mjbuf.pc = 0;
	mjbuf.mobj = mobj;
	mjbuf.meth = meth;
	if (tid != NULL && unhand(tid)->PrivateInfo != 0) {
		mjbuf.prev = (vmException*)unhand(tid)->exceptPtr;
		unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
	}
	if (meth->exception_table != 0 || (methaccflags & ACC_SYNCHRONISED)) {
		if (setjmp(mjbuf.jbuf) != 0) {
			unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)&mjbuf;
			npc = mjbuf.pc;
			sp = &lcl[meth->bcode->localsz];
			sp->v.taddr = (void*)unhand(tid)->exceptObj;
			unhand(tid)->exceptObj = 0;
			goto restart;
		}
	}

	/* Calculate number of arguments */
	idx = 0;
	for (low = 0; low < meth->sig->len; low++) {
		switch (meth->sig->elem[low]->data[0]) {
		case 'J':
		case 'D':
			idx++;
		}
		idx++;
	}
	if ((methaccflags & ACC_STATIC) == 0) {
		idx++;
	}

	/* Copy in the arguments */
	sp = lcl;
	for (low = 0; low < idx; low++) {
		*(sp++) = *(arg++);
	}

	/* Sync. if required */
	if (methaccflags & ACC_SYNCHRONISED) { 
		if (methaccflags & ACC_STATIC) {
			mobj = &meth->class->head;
		}
		else {
			mobj = (Hjava_lang_Object*)lcl[0].v.taddr;
		}
		mjbuf.mobj = mobj;
		lockMutex(mobj);
	}       

	sp = &lcl[meth->bcode->localsz - 1];

	restart:;
	{
		register bytecode* code;
		register uintp pc;

		/* Misc machine variables */
		jlong lcc;
		jvalue tmpl;
		slots tmp[1];
		slots tmp2[1];
		slots mtable[1];
		jint high;

		code = (bytecode*)meth->bcode->code;

		/* Finally we get to actually execute the machine */
#if !defined(__GNUC__)
		for (;;) {
			assert(npc < meth->c.bcode.codelen);
			pc = npc;
			mjbuf.pc = pc;
			npc = pc + insnLen[code[pc]];

			switch (code[pc]) {
			default:
				fprintf(stderr, "Unknown bytecode %d\n", code[pc]);
				postAndThrowException("java/lang/VerifyError", CLASS_CNAME(meth->class));
				break;
#include "kaffe.def"
			}
		}
#else
		/* Use threaded code if we have GNU CC */
		insn_dispatch_next();
		insn_default:;
		fprintf(stderr, "Unknown bytecode %d\n", code[pc]);
		postAndThrowException("java/lang/VerifyError", CLASS_CNAME(meth->class));
#include "kaffe.def"
#endif
	}
	end:;

	/* Unsync. if required */
	if (mobj != 0) {
		unlockMutex(mobj);
	}       
	if (tid != NULL && unhand(tid)->PrivateInfo != 0) {
		unhand(tid)->exceptPtr = (struct Hkaffe_util_Ptr*)mjbuf.prev;
	}

RDBG(	printf("Returning from method %s%s.\n", meth->name->data, meth->signature->data); )
}

void
interpreterEnd(void)
{
}


/*
 * Dispatch an exception.
 */
/*static*/
void
EcallExceptionHandler(Method* meth, uintp handler, stackTraceInfo* frame)
{
	vmException* fm;

	fm = (vmException*)frame->arg0;

	fm->pc = handler;
	longjmp(fm->jbuf, 1);
}

/*
 * Call inter-language method
 */
void
EcallInterEngine(callMethodInfo* call)
{
	Method* meth;
	Hjava_lang_Object* sync;
	int i;
	int iLockRoot;

	meth = (Method*)call->function;
	if ((meth->accflags & ACC_NATIVE) == 0) {
		interpreter(meth, (slots*)call->args, (slots*)call->ret, Kaffe_TI_currentJava());
	}
	else {
		if (!METHOD_TRANSLATED(meth)) {
                        native(meth);
                }
		call->function = METHOD_NATIVECODE(meth);

		if (meth->accflags & ACC_SYNCHRONISED) {
			if (meth->accflags & ACC_STATIC) {
				sync = &meth->class->head;
			}
			else {
				sync = (Hjava_lang_Object*)call->args[0].l;
			}
		}
		else {
			sync = 0;
		}

		/* For a JNI method we must insert the JNI environment and
		 * possibly the class too.
		 */
		if (meth->accflags & ACC_JNI) {
			i = 1;
			if (METHOD_IS_STATIC(meth)) {
				i++;
			}
			memmove(&call->callsize[i], &call->callsize[0], sizeof(char) * call->nrargs);
			memmove(&call->calltype[i], &call->calltype[0], sizeof(char) * call->nrargs);
			memmove(&call->args[i], &call->args[0], sizeof(jvalue) * call->nrargs);
			call->callsize[0] = PTR_TYPE_SIZE / SIZEOF_INT;
			call->calltype[0] = 'L';
			call->args[0].l = (void*)&Kaffe_JNIEnv;
			call->argsize += call->callsize[0];
			call->nrargs++;

			if (METHOD_IS_STATIC(meth)) {
				call->callsize[1] = PTR_TYPE_SIZE / SIZEOF_INT;
				call->calltype[1] = 'L';
				processClass(meth->class, CSTATE_LINKED);
				call->args[1].l = (void*)meth->class;
				call->argsize += call->callsize[1];
				call->nrargs++;
			}
		}

		if (sync != 0) {
			lockMutex(sync);
		}

		/* Make the call - system dependent */
		sysdepCallMethod(call);

		if (sync != 0) {
			unlockMutex(sync);
		}
	}
}

static
void
Einit(void)
{
	/* Install the interpreter as a pseudo-method */
	interpreterMethod = gc_malloc_fixed_typed(sizeof(Method), GC_ALLOC_METHOD);
}

static
char*
EgetName(void)
{
        return ("Interpreter");
}

static
char*
EgetVersion(void)
{
        return (KAFFEVERSION);
}

static
int
EgetNumber(void)
{
        return (0);
}

static
void*
EsetupTrampoline(Method* meth, void** where)
{
	return (meth);
}

static
void
EprocessExceptionMethod(stackTraceInfo* info)
{
	vmException* except;

	except = (vmException*)unhand(Kaffe_TI_currentJava())->exceptPtr;
	for (; except != 0; except = except->prev) {
		if (except->frame == info->rframe) {
			info->meth = except->meth;
			info->object = except->mobj;
			info->arg0 = (uintp)except;
			info->pc = except->pc;
			return;
		}
	}

	/* Not found */
	info->arg0 = 0;
	info->object = 0;
	if (info->meth == interpreterMethod) {
		info->meth = UNKNOWN_METHOD;
	}
}

static
void
EsetupJNIcall(Method* xmeth, void* func)
{
	SET_METHOD_NATIVECODE(xmeth, func);
	xmeth->accflags |= ACC_JNI;
}

/*
 * Define the engine interface.
 */
EngineInterface Kaffe_EngineInterface = {

	Einit,

	EgetName,
	EgetVersion,
	EgetNumber,

	EsetupTrampoline,
	EprocessExceptionMethod,
	EcallInterEngine,
	EcallExceptionHandler,
	EsetupJNIcall,
};
