/*
 * itypes.c
 * Internal types.
 *
 * Copyright (c) 1996 - 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "classMethod.h"
#include "jtypes.h"
#include "itypes.h"
#include "baseClasses.h"
#include "objects.h"
#include "access.h"
#include "precompile.h"
#include "strings.h"

Hjava_lang_Class _Jv_intClass;
Hjava_lang_Class _Jv_longClass;
Hjava_lang_Class _Jv_booleanClass;
Hjava_lang_Class _Jv_charClass;
Hjava_lang_Class _Jv_floatClass; 
Hjava_lang_Class _Jv_doubleClass;
Hjava_lang_Class _Jv_byteClass; 
Hjava_lang_Class _Jv_shortClass;     
Hjava_lang_Class _Jv_voidClass;

Hjava_lang_Class* types[MAXTYPES];

void
initPrimClass(Hjava_lang_Class* class, char* name, char* sig, int len)
{
	class->name = addUtf(name);
	addClass(class);

	class->head.dtable = (dispatchTable*)&KaffeDispatchTable_java_lang_Class;
	class->accflags = ACC_PUBLIC;
	class->dtable = _PRIMITIVE_DTABLE;
	CLASS_STATE(class) = CSTATE_COMPLETE;
	class->constants.size = 1; /* dummy */
	CLASS_PRIM_SIG(class) = sig[0];
	CLASS_PRIM_NAME(class) = addUtf(sig);
	TYPE_PRIM_SIZE(class) = len;
}

/*
 * Intialise the internal types.
 */
void
initTypes(void)
{
	initPrimClass(&_Jv_byteClass, "byte", "B", 1);
	initPrimClass(&_Jv_shortClass, "short", "S", 2);
	initPrimClass(&_Jv_intClass, "int", "I", 4);
	initPrimClass(&_Jv_longClass, "long", "J", 8);
	initPrimClass(&_Jv_booleanClass, "boolean", "Z", 1);
	initPrimClass(&_Jv_charClass, "char", "C", 2);
	initPrimClass(&_Jv_floatClass, "float", "F", 4);
	initPrimClass(&_Jv_doubleClass, "double", "D", 8);
	initPrimClass(&_Jv_voidClass, "void", "V", 0);

	TYPE_CLASS(TYPE_Boolean) = &_Jv_booleanClass;
	TYPE_CLASS(TYPE_Char) = &_Jv_charClass;
	TYPE_CLASS(TYPE_Float) = &_Jv_floatClass;
	TYPE_CLASS(TYPE_Double) = &_Jv_doubleClass;
	TYPE_CLASS(TYPE_Byte) = &_Jv_byteClass;
	TYPE_CLASS(TYPE_Short) = &_Jv_shortClass;
	TYPE_CLASS(TYPE_Int) = &_Jv_intClass;
	TYPE_CLASS(TYPE_Long) = &_Jv_longClass;
}

Hjava_lang_Class*
getClassFromSignature(char* sig, Hjava_lang_ClassLoader* loader)
{
	return (classFromSigProcess(&sig, loader, CSTATE_LINKED));
}

Hjava_lang_Class*
classFromSig(char** strp, Hjava_lang_ClassLoader* loader)
{
	return (classFromSigProcess(strp, loader, CSTATE_LINKED));
}

Hjava_lang_Class*
classFromSigProcess(char** strp, Hjava_lang_ClassLoader* loader, int state)
{
	char* start;
	char* end;
	char* tmp;
	char* tmp2;

	switch (*(*strp)++) {
	case 'V': return (&_Jv_voidClass);
	case 'I': return (&_Jv_intClass);
	case 'Z': return (&_Jv_booleanClass);
	case 'S': return (&_Jv_shortClass);
	case 'B': return (&_Jv_byteClass);
	case 'C': return (&_Jv_charClass);
	case 'F': return (&_Jv_floatClass);
	case 'D': return (&_Jv_doubleClass);
	case 'J': return (&_Jv_longClass);
	case '[': return (lookupArrayProcess(classFromSigProcess(strp, loader, state), CSTATE_COMPLETE));
	case 'L':
		start = *strp;
		for (end = start; *end != 0 && *end != ';'; end++)
			;
		*strp = end;
		if (*end != 0) {
			(*strp)++;
		}
		tmp = alloca(end-start+1);
		strncpy(tmp, start, end-start);
		tmp[end-start] = 0;
		UTF_REPLACE(tmp2, tmp, '/', '.');
		return (loadClass(tmp2, loader, state));
	default:
		return (NULL);
	}
}

char*
utfFromSig(char** strp)
{
	char* start;
	char* end;

	/* Determine signature for element */
	start = *strp;
	end = start;
	for (;;) {
		switch (*end) {
		case 'V':
		case 'I':
		case 'Z':
		case 'S':
		case 'B':
		case 'C':
		case 'F':
		case 'D':
		case 'J':
			break;
		case '[':
			for (; *end == '['; end++)
				;
			continue;
		case 'L':
			for (; *end != 0 && *end != ';'; end++)
				;
			break;
		default:
			return (0);
		}
		break;
	}

	/* Setup so we can get next signature */
	if (*end != 0) {
		end++;
	}
	*strp = end;

	switch (*start) {
	case 'V': return (CLASS_PRIM_NAME(&_Jv_voidClass));
	case 'I': return (CLASS_PRIM_NAME(&_Jv_intClass));
	case 'Z': return (CLASS_PRIM_NAME(&_Jv_booleanClass));
	case 'S': return (CLASS_PRIM_NAME(&_Jv_shortClass));
	case 'B': return (CLASS_PRIM_NAME(&_Jv_byteClass));
	case 'C': return (CLASS_PRIM_NAME(&_Jv_charClass));
	case 'F': return (CLASS_PRIM_NAME(&_Jv_floatClass));
	case 'D': return (CLASS_PRIM_NAME(&_Jv_doubleClass));
	case 'J': return (CLASS_PRIM_NAME(&_Jv_longClass));
	case '[':
	case 'L':
		return (addUtfCopy(start, end-start));
	default:
		return (0);
	}
}
