/*
 * jarClasspath.c
 * Handle the standard Classpath representation.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define PDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "config-loadlib.h"
#include "gtypes.h"
#include "jar.h"
#include "classpath.h"
#include "gc.h"
#include "md.h"
#include "paths.h"
#include "jsyscall.h"
#include "system.h"

static void addDirClasspath(char*);
static int getClasspathType(char*);
static void appendClasspath(char*);
static int CPappend(char*);

#define	IS_ZIP(P,B) \
	((B)[0] == 'P' && (B)[1] == 'K')

classpathElement* classpath;
char* realClassPath;
char* realKaffeHome;

static
void
CPinit(void)
{
	char* cp;
	char* hm;
	int len;
	classpathElement* elem;

	cp = (char*)Kaffe_JavaVMArgs[0].classpath;
	hm = (char*)Kaffe_JavaVMArgs[0].classhome;

	if (cp != 0) {
		appendClasspath(strdup(cp));
	}

	/* Add '.' */
	CPappend(".");

	if (hm != 0) {
		addDirClasspath(hm);
	}
	else {
		hm = "/";
	}

	/* Work over classpath to work out what is what */
	len = 0;
	for (elem = classpath; elem != 0; elem = elem->next) {
		elem->type = getClasspathType(elem->path);
PDBG(		kprintf("path '%s' type %d\n", elem->path, elem->type); )
		len += strlen(elem->path) + 1;
	}

	/* Build the 'standard' classpath representation */
	realClassPath = gc_malloc_fixed_typed(len, GC_ALLOC_JAR);

	for (elem = classpath; elem != 0; elem = elem->next) {
		if (elem != classpath) {
			strcat(realClassPath, path_separator);
		}
		strcat(realClassPath, elem->path);
	}
	realKaffeHome = hm;
}

/*
 * Add an entry in the Classpath dynamically.
 */
static
int
CPappend(char* cp)
{
	classpathElement** elem;

PDBG(	kprintf("addJarToClasspath(): '%s'\n", cp);			)

	for (elem = &classpath; *elem != 0; elem = &(*elem)->next) {
		if (!strcmp(cp, (*elem)->path)) {
			return (1);
		}
	}

	*elem = gc_malloc_fixed_typed(sizeof(classpathElement), GC_ALLOC_JAR);
	(*elem)->path = strdup(cp);
	(*elem)->type = getClasspathType((*elem)->path);

PDBG(	kprintf("path '%s' type %d\n", (*elem)->path, (*elem)->type); )

	return(1);
}

static
int
CPprepend(char* cp)
{
	classpathElement* elem;

PDBG(	kprintf("addJarToClasspath(): '%s'\n", cp);			)

	for (elem = classpath; elem != 0; elem = elem->next) {
		if (!strcmp(cp, elem->path)) {
			return (1);
		}
	}

	elem = gc_malloc_fixed_typed(sizeof(classpathElement), GC_ALLOC_JAR);
	elem->path = strdup(cp);
	elem->type = getClasspathType(elem->path);

	elem->next = classpath;
	classpath = elem;

PDBG(	kprintf("path '%s' type %d\n", elem->path, elem->type); )

	return(1);
}

/*
 * Remove a classpath element.
 */
static
int
CPremove(char* cp)
{
	return (0);	/* We don't support this */
}

static
void
appendClasspath(char* cp)
{
	while (cp != 0) {
		char* start = cp;
		cp = strchr(cp, PATHSEP[0]);
		if (cp != 0) {
			*cp = 0;
			cp++;
		}
		CPappend(start);
	}
}

static
void
addDirClasspath(char* lib)
{
	int hlen;
	int len;
	char* name;
	char* path;
	jdirEntry* dirlist;

	if (readdir(lib, &dirlist) <= 0) {
		return;
	}

	/* Add the lib directory */
	CPappend(lib);

	hlen = strlen(lib);
	for (; dirlist != 0; dirlist = dirlist->next) {
		name = dirlist->name;
		len = strlen(name);
		if (strcmp(&name[len-4], ".zip") == 0 || strcmp(&name[len-4], ".jar") == 0 || strcmp(&name[len-sizeof(LIBRARYSUFFIX)+1], LIBRARYSUFFIX) == 0) {
			path = gc_malloc_fixed_typed(hlen+len+2, GC_ALLOC_JAR);
			strcpy(path, lib);
			strcat(path, file_separator);
			strcat(path, name);
			CPappend(path);
			gc_free_fixed(path);
		}
	}
}

/*
 * Work out what kind of thing this path points at.
 */
static
int
getClasspathType(char* path)
{
	int h;
	int c;
	int i;
	char buf[16];

PDBG(kprintf("getClasspathType: %s\n", path);)

	if (fileinfo(path, JINFO_ISDIR, 0) == 1) {
		return (CP_DIR);
	}
	if (fileinfo(path, JINFO_ISFILE, 0) == 0) {
		return (CP_INVALID);
	}
	/* What kind of a file ? */
	h = open(path, JOPEN_READ);
	if (h < 0) {
		return (CP_INVALID);
	}

	for (i = 0; i < sizeof(buf); i += c) {
		c = read(h, &buf[i], sizeof(buf) - i);
		if (c <= 0) {
			close(h);
			return (CP_INVALID);
		}
	}
	close(h);

	if (IS_ZIP(path, buf)) {
		return (CP_ZIPFILE);
	}
	if (IS_SHAREDLIBRARY(path, buf)) {
		return (CP_SOFILE);
	}

	return (CP_INVALID);
}

ClasspathInterface Kaffe_ClasspathInterface = {

	CPinit,
	CPappend,
	CPprepend,
	CPremove,

};
