/*
 * jarFile.c
 * Handle JAR input from files.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define PDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "../gtypes.h"
#include "jar.h"
#include "jsyscall.h"

#if defined(KAFFEH)

#undef	open
#undef	close
#undef	read
#undef	lseek
#undef	JOPEN_READ
#undef	JSEEK_SET
#undef	JSEEK_CUR
#undef	JSEEK_END
#define	JOPEN_READ	O_RDONLY
#define	JSEEK_SET	SEEK_SET
#define	JSEEK_CUR	SEEK_CUR
#define	JSEEK_END	SEEK_END
#define	gc_malloc_fixed	malloc
#define	gc_free_fixed	free

#else

#include "../gc.h"

#endif

static uint8 _buf[1024];
static int _blen = 0;
static int _bpos = 0;
static jarFile* lastjar;

#define	RESETBUFFER() \
	_blen = 0

#define	READBYTES(F, B, L) \
{ \
	int _i; \
	for (_i = 0; _i < (L); _i++) { \
		if (_bpos >= _blen) { \
			_blen = read((F), _buf, sizeof(_buf)); \
			if (_blen <= 0) { \
				(L) = _i; \
				break; \
			} \
			_bpos = 0; \
		} \
		((uint8*)(B))[_i] = _buf[_bpos++]; \
	} \
}

#define	SWITCHFILE(F) \
	if ((F) != lastjar) { \
		RESETBUFFER(); \
		lastjar = (F); \
	}

#define	INLINE static inline

INLINE
void
Kaffe_JarRawInterface_init(void)
{
	/* Does nothing. */
}

INLINE
void
Kaffe_JarRawInterface_open(jarFile* file, char* name)
{
	file->fp = open(name, JOPEN_READ);
	RESETBUFFER();
}

INLINE
void
Kaffe_JarRawInterface_close(jarFile* file)
{
	close(file->fp);
	RESETBUFFER();
}

INLINE
int
Kaffe_JarRawInterface_read16(jarFile* file)
{
	uint8 r[2];
	int l = 2;

	SWITCHFILE(file);
	READBYTES(file->fp, r, l);

	return ((((int)r[1])<<8)|(int)r[0]);
}

INLINE
int
Kaffe_JarRawInterface_read32(jarFile* file)
{
	uint8 r[4];
	int l = 4;

	SWITCHFILE(file);
	READBYTES(file->fp, r, l);

	return ((((int)r[3])<<24)|(((int)r[2])<<16)|(((int)r[1])<<8)|(int)r[0]);
}

INLINE
int
Kaffe_JarRawInterface_readBytes(jarFile* file, int len, uint8* buf)
{
	int i;
	int c;

	SWITCHFILE(file);

	for (i = 0; i < len; i += c) {
		c = len - i;
		READBYTES(file->fp, &buf[i], c);
		if (c <= 0) {
			break;
		}
	}
	return (i);
}

INLINE
int
Kaffe_JarRawInterface_seekStart(jarFile* file, int pos)
{
	RESETBUFFER();
	return (lseek(file->fp, (long)pos, JSEEK_SET));
}

INLINE
int
Kaffe_JarRawInterface_seekHere(jarFile* file, int pos)
{
	SWITCHFILE(file);
	if (_bpos + pos < _blen) {
		_bpos += pos;
		return (0);
	}
	else {
		pos -= (_blen - _bpos);
		RESETBUFFER();
		return (lseek(file->fp, (long)pos, JSEEK_CUR));
	}
}

INLINE
int
Kaffe_JarRawInterface_seekEnd(jarFile* file, int pos)
{
	RESETBUFFER();
	return (lseek(file->fp, (long)pos, JSEEK_END));
}
