/*
 * jarMmap.c
 * Handle JAR input from memory mapped files.
 *
 * Copyright (c) 1998 - 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define PDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "../gtypes.h"
#include "jar.h"
#include "jsyscall.h"
#include <unistd.h>
#include <sys/mman.h>

#if defined(KAFFEH)

#undef  open
#undef  close
#undef  read
#undef  lseek
#undef  JOPEN_READ
#undef  JSEEK_SET
#undef  JSEEK_CUR
#undef  JSEEK_END
#define JOPEN_READ      O_RDONLY
#define JSEEK_SET       SEEK_SET
#define JSEEK_CUR       SEEK_CUR
#define JSEEK_END       SEEK_END
#define gc_malloc_fixed malloc
#define gc_free_fixed   free

#else

#include "../gc.h"

#endif

#define	INLINE	static inline

INLINE
void
Kaffe_JarRawInterface_init(void)
{
	/* Does nothing. */
}

INLINE
void
Kaffe_JarRawInterface_open(jarFile* file, char* name)
{
	struct stat buf;

	file->fp = open(name, JOPEN_READ);
	fstat(file->fp, &buf);
	file->mmapLength = buf.st_size;
	file->mmapBase = mmap(0, buf.st_size, PROT_READ, MAP_SHARED, file->fp, 0);
	file->mmapPos = file->mmapBase;
}

INLINE
void
Kaffe_JarRawInterface_close(jarFile* file)
{
	munmap(file->mmapBase, file->mmapLength);
	close(file->fp);
}

INLINE
int
Kaffe_JarRawInterface_read16(jarFile* file)
{
	int v;
	uint8* r = file->mmapPos;
	file->mmapPos += 2;
	v = ((((unsigned)r[1])<<8)|(unsigned)r[0]);
	return (v);
}

INLINE
int
Kaffe_JarRawInterface_read32(jarFile* file)
{
	int v;
	uint8* r = file->mmapPos;
	file->mmapPos += 4;
	v = ((((unsigned)r[3])<<24)|(((unsigned)r[2])<<16)|(((unsigned)r[1])<<8)|(unsigned)r[0]);
	return (v);
}

INLINE
int
Kaffe_JarRawInterface_readBytes(jarFile* file, int len, uint8* buf)
{
	memcpy(buf, file->mmapPos, len);
	file->mmapPos += len;
	return (len);
}

INLINE
int
Kaffe_JarRawInterface_seekStart(jarFile* file, int pos)
{
	file->mmapPos = file->mmapBase + pos;
	return (0);
}

INLINE
int
Kaffe_JarRawInterface_seekHere(jarFile* file, int pos)
{
	file->mmapPos = file->mmapPos + pos;
	return (0);
}

INLINE
int
Kaffe_JarRawInterface_seekEnd(jarFile* file, int pos)
{
	file->mmapPos = file->mmapBase + file->mmapLength + pos;
	return (0);
}
