/*
 * basecode.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "gtypes.h"
#include "slots.h"
#include "seq.h"
#include "md.h"
#include "registers.h"
#include "icode.h"
#include "basecode.h"
#include "labels.h"
#include "codeproto.h"
#include <stdarg.h>

void
writeslot(sequence* seq, int slot, SlotInfo* data, int len)
{
	SlotData* sdata;

	NOTE_clearseq(seq, slot);
	if (data == 0) {
		seq->u[slot].s.slot = 0;
		return;
	}
	sdata = data->slot;

	NOTE_clear(data);

	/* We look for the last use of this SlotInfo and locate the sequence
	 * which used it.  We mark the sequence so we know that this was the
	 * last use of the associated SlotData so we can avoid writing back
	 * registers which are dead.  Note that we can only do this if we're
	 * not about to write to it (in which case it's still live).
	 */
	if (sdata->rseq == seq) {
		/* Skip */
	}
	else if (len == 2 && data[1].slot->rseq == seq) {
		/* Skip */
	}
	else {
		lastuse_slot(data, len);
	}

	seq->u[slot].s.slot = sdata;

	/* This slot writes a value - but this might not be used so we
	 * clear the 'refed' flag an only set it if someone reads what
	 * was written.
	 */
	seq->refed = 0;
	sdata->wseq = seq;
	if (len == 2) {
		data[1].slot->wseq = seq;
		NOTE_clear(&data[1]);
	}
	assert(sdata->rseq == 0 || sdata->rseq == seq);
}

static
inline
void
readslot(sequence* seq, int slot, SlotInfo* data, int len)
{
	SlotData* sdata;

	if (data == 0) {
		seq->u[slot].s.slot = 0;
		return;
	}

	sdata = data->slot;
	seq->u[slot].s.slot = sdata;

	/* We note on the slot which sequence used this slot.  This
	 * allows us to determine liveliness when the slot get's
	 * overwritten.
	 */
	sdata->rseq = seq;
	sdata->rseqslot = slot;
	if (len == 2) {
		data[1].slot->rseq = seq;
		data[1].slot->rseqslot = slot;
	}

	/* Note on the sequence which wrote this value that it was refed. */
	if (sdata->wseq != 0) {
		sdata->wseq->refed = 1;
	}
	if (len == 2 && data[1].slot->wseq != 0) {
		data[1].slot->wseq->refed = 1;
	}
}

void
generic_seq(ifunc f, int len, const char* desc, ...)
{
	va_list arg;
	int i;
	sequence* seq;

	seq = nextSeq(len);

	seq->func = f;

	va_start(arg, desc);
	for (i = 0; i < len; i++) {
		switch (desc[i]) {
		case 'r':		/* Slot - read */
			readslot(seq, len-i-1, va_arg(arg, SlotInfo*), 1);
			break;
		case 'w':		/* Slot - write */
			writeslot(seq, len-i-1, va_arg(arg, SlotInfo*), 1);
			break;
		case 'R':		/* Long Slot - read */
			readslot(seq, len-i-1, va_arg(arg, SlotInfo*), 2);
			break;
		case 'W':		/* Long Slot - write */
			writeslot(seq, len-i-1, va_arg(arg, SlotInfo*), 2);
			break;
		case 'i':		/* Integer */
			seq->u[len-i-1].value.i = va_arg(arg, jint);
			break;
		case 'j':		/* Long */
			seq->u[len-i-1].value.j = va_arg(arg, jlong);
			break;
		case 'd':		/* Double */
			seq->u[len-i-1].value.d = va_arg(arg, jdouble);
			break;
		default:
			ABORT();
		}
	}
	va_end(arg);
}
