/* basecode.h
 * Define the base instructions macros.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __basecode_h
#define __basecode_h

extern int argcount;

/* -------------------------------------------------------------------- */
/* Branches */

#define	ba			0	/* Always */
#define	beq			1	/* Equal */
#define	blt			2	/* Less than */
#define	ble			3	/* Less than or equal */
#define	bgt			4	/* Greater than */
#define	bge			5	/* Greater than or equal */
#define	bne			6	/* Not equal */
#define	bn			7	/* Never */
#define	bindirect		8	/* Indirect */
#define	bult			9	/* Unsigned less than */
#define	buge			10	/* Unsigned greater than or equal */
#define	bugt			11	/* Unsigned greater than */

#define branch_a(l)			branch(l, ba)

#define cbranch_int_eq(s1, s2, l)	cbranch_int(s1, s2, l, beq)
#define cbranch_int_ne(s1, s2, l)	cbranch_int(s1, s2, l, bne)
#define cbranch_int_lt(s1, s2, l)	cbranch_int(s1, s2, l, blt)
#define cbranch_int_le(s1, s2, l)	cbranch_int(s1, s2, l, ble)
#define cbranch_int_gt(s1, s2, l)	cbranch_int(s1, s2, l, bgt)
#define cbranch_int_ge(s1, s2, l)	cbranch_int(s1, s2, l, bge)
#define cbranch_int_ult(s1, s2, l)	cbranch_int(s1, s2, l, bult)
#define cbranch_int_ugt(s1, s2, l)	cbranch_int(s1, s2, l, bugt)
#define cbranch_int_uge(s1, s2, l)	cbranch_int(s1, s2, l, buge)

#define cbranch_int_const_eq(s1, s2, l) cbranch_int_const(s1, s2, l, beq)
#define cbranch_int_const_ne(s1, s2, l) cbranch_int_const(s1, s2, l, bne)
#define cbranch_int_const_lt(s1, s2, l) cbranch_int_const(s1, s2, l, blt)
#define cbranch_int_const_le(s1, s2, l) cbranch_int_const(s1, s2, l, ble)
#define cbranch_int_const_gt(s1, s2, l) cbranch_int_const(s1, s2, l, bgt)
#define cbranch_int_const_ge(s1, s2, l) cbranch_int_const(s1, s2, l, bge)
#define cbranch_int_const_ult(s1, s2, l) cbranch_int_const(s1, s2, l, bult)
#define cbranch_int_const_ugt(s1, s2, l) cbranch_int_const(s1, s2, l, bugt)
#define cbranch_int_const_uge(s1, s2, l) cbranch_int_const(s1, s2, l, buge)

#define cbranch_offset_int_eq(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, beq)
#define cbranch_offset_int_ne(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, bne)
#define cbranch_offset_int_lt(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, blt)
#define cbranch_offset_int_le(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, ble)
#define cbranch_offset_int_gt(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, bgt)
#define cbranch_offset_int_ge(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, bge)
#define cbranch_offset_int_ult(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, bult)
#define cbranch_offset_int_ugt(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, bugt)
#define cbranch_offset_int_uge(s1, s2, o, l)	cbranch_offset_int(s1, s2, o, l, buge)

#define cbranch_ref_eq(s1, s2, l)	cbranch_ref(s1, s2, l, beq)
#define cbranch_ref_ne(s1, s2, l)	cbranch_ref(s1, s2, l, bne)

#define cbranch_ref_const_eq(s1, s2, l)	cbranch_ref_const(s1, s2, l, beq)
#define cbranch_ref_const_ne(s1, s2, l)	cbranch_ref_const(s1, s2, l, bne)

/* -------------------------------------------------------------------- */
/* Basic blocks and instructions */

#define	start_instruction()	_start_instruction(pc)
#define	start_function()	prologue(meth)
#define	start_basic_block()	_start_basic_block()
#define	end_basic_block()	_end_basic_block()
#define	start_sub_block()	_start_sub_block()
#define	end_sub_block()		_end_sub_block()
#define	end_function()		epilogue(meth)
#define	start_exception_block()	_start_exception_block(stackno)
#define	sync_registers()	_syncRegisters(stackno, tmpslot)

/* -------------------------------------------------------------------- */
/* Conditional monitors */

#define	monitor_enter()		mon_enter(meth, clazz, local(0))
#define	monitor_exit()		mon_exit(meth, clazz, local(0))

/* -------------------------------------------------------------------- */
/* Instruction formats */

extern void writeslot(sequence*, int, SlotInfo*, int);
extern void generic_seq(ifunc, int, const char*, ...);

#define	GENERIC_SEQ(F,D, A...)	generic_seq((F), sizeof(D)-1, (D), A)

#define	complex_slot_slot_const(dst, src, src2, func, t)		\
	{								\
		slot_slot_const((dst), (src), (src2), (func), (t));	\
	}

#define	slot_slot_slot(dst, src, src2, func, t)				\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "rrw", (src2), (src), (dst));	\
	}

#define	lslot_lslot_lslot(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "RRW", (src2), (src), (dst));	\
	}

#define	lslot_lslot_slot(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "rRW", (src2), (src), (dst));	\
	}

#define	slot_slot_lslot(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "Rrw", (src2), (src), (dst));	\
	}

#define	slot_const_const(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		if ((t) == Tstore) {					\
			GENERIC_SEQ((func), "iir", (src2), (src), (dst));\
		}							\
		else {							\
			GENERIC_SEQ((func), "iiw", (src2), (src), (dst));\
		}							\
	}

#define	slot_slot(dst, src, func, t)					\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "rrw", (src), 0, (dst));		\
	}

#define	slot_slot_const(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		if ((t) == Tstore) {					\
			GENERIC_SEQ((func), "irr", (src2), (src), (dst));\
		}							\
		else {							\
			GENERIC_SEQ((func), "irw", (src2), (src), (dst));\
		}							\
	}

#define	slot_slot_fconst(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "drw", (src2), (src), (dst));	\
	}

#define	lslot_lslot_const(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "iRW", (src2), (src), (dst));	\
	}

#define	lslot_lslot_lconst(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "jRW", (src2), (src), (dst));	\
	}

#define	lslot_slot_lconst(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "jrW", (src2), (src), (dst));	\
	}

#define	lslot_slot_fconst(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "drW", (src2), (src), (dst));	\
	}

#define	slot_lslot_lslot(dst, src, src2, func, t)			\
	{								\
		void func(sequence*);					\
		GENERIC_SEQ((func), "RrW", (src2), (src), (dst));		\
	}

#define	slot_slot_slot_const_const(dst, src, src2, src3, src4, func, t)	\
	{								\
		void func(sequence*);					\
		if ((t) == Tstore) {					\
			GENERIC_SEQ((func), "iirrr", (src4), (src3), (src2), (src), (dst));\
		}							\
		else {							\
			GENERIC_SEQ((func), "iirrw", (src4), (src3), (src2), (src), (dst));\
		}							\
	}

#define	slot_slot_const_const_const(dst, src, src2, src3, src4, func, t) \
	{								\
		void func(sequence*);					\
		if ((t) == Tstore) {					\
			GENERIC_SEQ((func), "iiirr", (src4), (src3), (src2), (src), (dst));\
		}							\
		else {							\
			GENERIC_SEQ((func), "iiirw", (src4), (src3), (src2), (src), (dst));\
		}							\
	}

/* -------------------------------------------------------------------- */

#endif
