/* notes.c
 * Notes.
 *
 * Copyright (c) 1996-2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "gtypes.h"
#include "slots.h"
#include "seq.h"
#include "md.h"
#include "registers.h"
#include "gc.h"
#include "notes.h"

/* Sequence notes */
#define	CLASSINITSZ	8
static struct Hjava_lang_Class* lastClassInit[CLASSINITSZ];
static int nextClassInit;

/* Register notes */
#define	RNOTEHASHSZ	(MAXREG * 2)
#define	RNOTEHASH(T,V)	(((T) ^ (V)) % RNOTEHASHSZ)
#define	RNOTEFREE	(0xFF)
#define	RNOTEUNUSED	(0xFE)
typedef struct {
	uint8		regno;
	uint8		type;
	uint32		value;
} regnote;
static regnote regNotes[RNOTEHASHSZ];

uint32 wrcnter;

/* 
 * See if we have a note to say a class has been initialized.
 */
int
isClassInitialized(struct Hjava_lang_Class* cls)
{
	int i;

	for (i = 0; i < CLASSINITSZ; i++) {
		if (cls == lastClassInit[i]) {
			return (1);
		}
	}
	return (0);
}

/*
 * Set a note that a class has been initialized.
 */
void
setClassInitialized(struct Hjava_lang_Class* cls)
{
	if (!isClassInitialized(cls)) {
		lastClassInit[nextClassInit] = cls;
	//	nextClassInit = (nextClassInit + 1) % CLASSINITSZ;
		
	     	nextClassInit++;
		if (nextClassInit == CLASSINITSZ) {
			nextClassInit = 0;
		}
	}
}

/*
 * Clear class init notes.
 */
static
void
cleanClassInitialized(void)
{
	int i;

	for (i = 0; i < CLASSINITSZ; i++) {
		lastClassInit[i] = 0;
	}
	nextClassInit = 0;
}

/*
 * Setup for the beginning of a basic block.
 */
void
clearNotes(void)
{
        int i;

        for (i = 0; i < maxslot; i++) {
                NOTE_clear(&slotinfo[i]);
        }

#if defined(STACK_LIMIT)
        NOTE_clear(stack_limit);
#endif

        /* Clear any class init notes */
        cleanClassInitialized();
}

void
clearRegisterNotes(int all)
{
	int i;

	if (all != 0) {
		/* Clear any register notes */
		for (i = 0; i < RNOTEHASHSZ; i++) {
			regNotes[i].regno = RNOTEFREE;
		}
		for (i = 0; i < MAXREG; i++) {
			reginfo[i].note = NONOTE;
		}
	}
	else {
		/* Only clear notes on registers which are not
		 * saved across a function call.
		 */
		for (i = 0; i < MAXREG; i++) {
			if (calleeSave(i) == 0 && reginfo[i].note != NONOTE) {
				regNotes[reginfo[i].note].regno = RNOTEFREE;
				reginfo[i].note = NONOTE;
			}
		}
	}
}

/*
 * Set a register note.
 */
void
setRegisterNote(uint8 type, uint32 value, uint8 regno)
{
	int s;
	int i;
	regnote* rn;

	s = RNOTEHASH(type, value);
	for (i = s; i < RNOTEHASHSZ; i++) {
		rn = &regNotes[i];
		if (rn->regno >= MAXREG) {
			goto attach;
		}
	}
	for (i = 0; i < s; i++) {
		rn = &regNotes[i];
		if (rn->regno >= MAXREG) {
			goto attach;
		}
	}

	assert("No regnote space" == 0);
	return;

attach:;
	/* Setup regnote */
	rn->regno = regno;
	rn->type = type;
	rn->value = value;
	/* Attach to register */
	assert(reginfo[regno].note == NONOTE);
	reginfo[regno].note = i;
}

/*
 * Get a register note.
 */
int
getRegisterNote(uint8 type, uint32 value)
{
	int s;
	int i;
	regnote* rn;

	s = RNOTEHASH(type, value);
	for (i = s; i < RNOTEHASHSZ; i++) {
		rn = &regNotes[i];
		if (rn->regno == RNOTEFREE) {
			return (NOREG);
		}
		if (rn->regno != RNOTEUNUSED && rn->type == type && rn->value == value) {
			return (rn->regno);
		}
	}
	for (i = 0; i < s; i++) {
		rn = &regNotes[i];
		if (rn->regno == RNOTEFREE) {
			return (NOREG);
		}
		if (rn->regno != RNOTEUNUSED && rn->type == type && rn->value == value) {
			return (rn->regno);
		}
	}
	return (NOREG);
}

/*
 * Remove a register note.
 */
void
removeRegisterNote(uint8 regno)
{
	int n;

	n = reginfo[regno].note;
	if (n != NONOTE) {
		assert(regNotes[n].regno == regno);
		reginfo[regno].note = NONOTE;
		regNotes[n].regno = RNOTEUNUSED;
	}
}
