/* notes.h
 * Manage slot notes.
 *
 * Copyright (c) 1996, 1997, 1998, 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __notes_h
#define __notes_h

struct Hjava_lang_Class;
struct SlotInfo;

struct SlotNote {
	uint32	type;
	union {
		struct {
			struct Hjava_lang_Class* hinttype;
		} object;
	} u1;
	union {
		struct {
			jvalue			val;
		} constant;
		struct {
			int32			size;
			int32			hintsize;
		} array;
	} u2;
	union {
		struct {
			uint32			wrcnt;
			struct SlotInfo*	alias;
		} alias;
	} u3;
};
typedef struct SlotNote SlotNote;

extern void clearNotes(void);

extern uint32 wrcnter;

#define	NOTE_None			0x00000000
#define	NOTE_Array			0x00000001
#define	NOTE_Object			0x00000002
#define	NOTE_IntConst			0x00000004
#define	NOTE_RefConst			0x00000008
#define	NOTE_Alias			0x00000010
#define	NOTE_NotNull			0x00000020
#define	NOTE_NotZero			0x00000040

#define	NOTE_setType(S,T)		(S)->note->type |= (T)
#define	NOTE_clear(S)			(S)->note->type = NOTE_None; \
					(S)->note->u3.alias.wrcnt = ++wrcnter
#define	NOTE_isType(S,T)		(((S)->note->type & (T)) != 0)
#define	NOTE_copy(T,F)			*(T)->note = *(F)->note

/* Aliases */

#define	NOTE_setAlias(T,F)		NOTE_setType(T, NOTE_Alias); \
					(T)->note->u3.alias.wrcnt = (F)->note->u3.alias.wrcnt; \
					(T)->note->u3.alias.alias = (F)
#define	NOTE_clearAlias(S)		(S)->note->type &= ~NOTE_Alias
#define	NOTE_getAlias(S)		((S)->note->u3.alias.alias)
#define	NOTE_checkWriteCount(A,B)	((A)->note->u3.alias.wrcnt == (B)->note->u3.alias.wrcnt)
#define	NOTE_isValidAlias(S)		(NOTE_isType((S), NOTE_Alias) && \
					NOTE_checkWriteCount((S), NOTE_getAlias(S)))

/* Objects */

#define	NOTE_setObjectType(S,T)		NOTE_setType(S, NOTE_Object); \
					(S)->note->u1.object.hinttype = (T)
#define	NOTE_setObjectHintType(S,T)	NOTE_setObjectType(S,T)
#define	NOTE_getObjectHintType(S)	((S)->note->u1.object.hinttype)

/* Arrays */

#define	NOTE_setArrayTypeSize(S,T,V)	NOTE_setType(S, NOTE_Array); \
					NOTE_setObjectType(S, T); \
					(S)->note->u2.array.size = (V); \
					(S)->note->u2.array.hintsize = (V)
#define	NOTE_setArrayHintSize(S,V)	NOTE_setType(S, NOTE_Array); \
					(S)->note->u2.array.size = -1; \
					(S)->note->u2.array.hintsize = (V)
#define	NOTE_getArraySize(S)		((S)->note->u2.array.size)
#define	NOTE_getArrayHintSize(S)	((S)->note->u2.array.hintsize)
#define	NOTE_getArrayHintType(S)	NOTE_getObjectHintType(S)

/* Constants */

#define	NOTE_setIntConst(S,V)		NOTE_setType(S, NOTE_IntConst); \
					(S)->note->u2.constant.val.i = (V)
#define	NOTE_getIntConst(S)		((S)->note->u2.constant.val.i)

#define	NOTE_setRefConst(S,R)		NOTE_setType(S, NOTE_RefConst); \
					(S)->note->u2.constant.val.l = (R)
#define	NOTE_getRefConst(S)		((S)->note->u2.constant.val.l)

/* Not-null */

#define	NOTE_setNotNull(S)		NOTE_setType(S, NOTE_NotNull)
#define	NOTE_isNotNull(S)		NOTE_isType(S, NOTE_NotNull)

/* Not-zero */

#define	NOTE_setNotZero(S)		NOTE_setType(S, NOTE_NotZero)
#define	NOTE_isNotZero(S)		NOTE_isType(S, NOTE_NotZero)

/* Class initialized - a simple implementation which holds the last
 * class initialized (should be a set of classes really).
 */

struct Hjava_lang_Class;
extern void setClassInitialized(struct Hjava_lang_Class*);
extern int isClassInitialized(struct Hjava_lang_Class*);

#define	NOTE_setClassInitialized(C)	setClassInitialized(C)
#define	NOTE_isClassInitialized(C)	isClassInitialized(C)

/* Sequence notes.
 */
#define	NOTE_NrArguments		7
#define	NOTE_ArgumentMask		0x00000007

#define	NOTE_DefaultSeq			0x00000007
#define	NOTE_clearseq(S, N)		(S)->u[N].s.note = NOTE_DefaultSeq

#define	NOTE_setArgumentNr(S,N)		(S)->u[0].s.note = ((N) < NOTE_NrArguments ? (N) : 0)
#define	NOTE_getArgumentNr(S)		((S)->u[0].s.note & NOTE_ArgumentMask)

/* Register notes.
 */
extern void setRegisterNote(uint8, uint32, uint8);
extern int getRegisterNote(uint8, uint32);
extern void removeRegisterNote(uint8);
extern void clearRegisterNotes(int);

#define	NOTE_setRegisterNote(T,V,R)	setRegisterNote((T), (uint32)(V), (R))
#define	NOTE_getRegisterNote(T,V)	getRegisterNote((T), (uint32)(V))
#define	NOTE_removeRegisterNote(R)	removeRegisterNote(R)

#define	NOTE_regConstpool		0x80
#define	NOTE_regValue			0x40

#endif
