/* registers.h
 * Instruction macros to extract information.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __registers_h
#define __registers_h

/* Register types */
#define	Reserved		0x80
#define	Rglobal			0x40
#define	Rint			0x01
#define	Rlong			0x02
#define	Rfloat			0x04
#define	Rdouble			0x08
#define	Rref			0x10
#define	Rsubint			0x20

#define	Rtypemask		(Rint|Rlong|Rfloat|Rdouble|Rref|Rsubint)

/* Register flags */
#define	Rreadonce		0x01
#define	Rnosaveoncall		0x02

/* Information about the state of a particular register. */
typedef struct _kregs {
	/* The slot whose value is in this register. */
	SlotData*		slot;
	uint8			ctype;
	uint8			type;
	uint8			flags;
	uint32			used;
	uint8			regno;
	uint8			refs;
	uint8			note;
} kregs;

extern kregs reginfo[];

#define	MAXREG			NR_REGISTERS
#define	NOREG			MAXREG
#define	NONOTE			(0xFF)

/* Specify the ideal registers for default read/writes.  We can override
 * these in specific JITs to handle such things as argument registers.
 */
#define	IDEALRREG_INT		NOREG
#define	IDEALWREG_INT		NOREG
#define	IDEALRWREG_INT		NOREG
#define	IDEALRREG_REF		NOREG
#define	IDEALWREG_REF		NOREG
#define	IDEALRWREG_REF		NOREG
#define	IDEALRREG_LONG		NOREG
#define	IDEALWREG_LONG		NOREG
#define	IDEALRWREG_LONG		NOREG
#define	IDEALRREG_FLOAT		NOREG
#define	IDEALWREG_FLOAT		NOREG
#define	IDEALRWREG_FLOAT	NOREG
#define	IDEALRREG_DOUBLE	NOREG
#define	IDEALWREG_DOUBLE	NOREG
#define	IDEALRWREG_DOUBLE	NOREG
#define	IDEALRREG_SUBINT	NOREG
#define	IDEALWREG_SUBINT	NOREG
#define	IDEALRWREG_SUBINT	NOREG

/*
 * If we have stack limit support use the STACK_LIMIT macro to get
 * the slot offset.  If we don't we just return the value 0 since
 * we don't be using stack_limit anyway.
 */
#if defined(STACK_LIMIT)
#define	STACK_LIMIT_OFFSET()	STACK_LIMIT()
#else
#define	STACK_LIMIT_OFFSET()	0
#endif

#define	SLOT2FRAMEOFFSET(_s)				\
	(						\
		((_s) == stack_limit) ?			\
		 STACK_LIMIT_OFFSET() :			\
		((_s) - slotinfo < maxArgs) ? 		\
		 SLOT2ARGOFFSET((_s) - slotinfo) :	\
		 SLOT2LOCALOFFSET((_s) - slotinfo)	\
	)

struct _sequence;
void	initRegisters(void);
int	slotRegister(SlotData*, int, int, int);
void	clobberRegister(int);
int	forceRegister(SlotData*, int, int);
int	slotOffsetNoSpill(SlotData*, int);
int	slotOffset(SlotData*, int, int);
void	slot_invalidate(SlotData*);
void	spillAndUpdate(SlotData*, jboolean);
void	preloadRegister(SlotData*, int, int);
int	allocRegister(int, int);
int	allocUnusedRegister(int);
void	useRegister(int);
void	attachRegister(int, SlotData*);

#define	rreload			0
#define	rread			1
#define	rwrite			2

/* JIT2 compatibility */
#define	_slowSlotRegister(A,B,C) slotRegister(A,B,C,NOREG)
#define	_slotInRegister(S,T)	_inRegister(S,T)
#define	slotInRegister(I,T)	inRegister(I,T)
#define	slowSlotOffset(S,T,U)	slotOffset(S,T,U)
#define	_slowSlotOffset(S,T,U)	slotOffset(S,T,U)

#define	rreg_int(i)		slotRegister(seq_slot(s, i), Rint, rread, IDEALRREG_INT)
#define	wreg_int(i)		slotRegister(seq_slot(s, i), Rint, rwrite, IDEALWREG_INT)
#define	rwreg_int(i)		slotRegister(seq_slot(s, i), Rint, rread|rwrite, IDEALRWREG_INT)
#define	rslot_int(i)		slotOffset(seq_slot(s, i), Rint, rread)
#define	wslot_int(i)		slotOffset(seq_slot(s, i), Rint, rwrite)
#define	rreg_ideal_int(i,r)	slotRegister(seq_slot(s, i), Rint, rread, r)
#define	wreg_ideal_int(i,r)	slotRegister(seq_slot(s, i), Rint, rwrite, r)
#define	rwreg_ideal_int(i,r)	slotRegister(seq_slot(s, i), Rint, rread|rwrite, r)

#define	rreg_ref(i)		slotRegister(seq_slot(s, i), Rref, rread, IDEALRREG_REF)
#define	wreg_ref(i)		slotRegister(seq_slot(s, i), Rref, rwrite, IDEALWREG_REF)
#define	rwreg_ref(i)		slotRegister(seq_slot(s, i), Rref, rread|rwrite, IDEALRWREG_REF)
#define	rslot_ref(i)		slotOffset(seq_slot(s, i), Rref, rread)
#define	wslot_ref(i)		slotOffset(seq_slot(s, i), Rref, rwrite)

#define	rreg_long(i)		slotRegister(seq_slot(s, i), Rlong, rread, IDEALRREG_LONG)
#define	wreg_long(i)		slotRegister(seq_slot(s, i), Rlong, rwrite, IDEALWREG_LONG)
#define	rwreg_long(i)		slotRegister(seq_slot(s, i), Rlong, rread|rwrite, IDEALRWREG_LONG)
#define	rslot_long(i)		slotOffset(seq_slot(s, i), Rlong, rread)
#define	wslot_long(i)		slotOffset(seq_slot(s, i), Rlong, rwrite)
#define	rreg_ideal_long(i,r)	slotRegister(seq_slot(s, i), Rlong, rread, r)

#define	rreg_float(i)		slotRegister(seq_slot(s, i), Rfloat, rread, IDEALRREG_FLOAT)
#define	wreg_float(i)		slotRegister(seq_slot(s, i), Rfloat, rwrite, IDEALWREG_FLOAT)
#define	rwreg_float(i)		slotRegister(seq_slot(s, i), Rfloat, rread|rwrite, IDEALRWREG_FLOAT)
#define	rslot_float(i)		slotOffset(seq_slot(s, i), Rfloat, rread)
#define	wslot_float(i)		slotOffset(seq_slot(s, i), Rfloat, rwrite)
#define	rreg_ideal_float(i,r)	slotRegister(seq_slot(s, i), Rfloat, rread, r)

#define	rreg_double(i)		slotRegister(seq_slot(s, i), Rdouble, rread, IDEALRREG_DOUBLE)
#define	wreg_double(i)		slotRegister(seq_slot(s, i), Rdouble, rwrite, IDEALWREG_DOUBLE)
#define	rwreg_double(i)		slotRegister(seq_slot(s, i), Rdouble, rread|rwrite, IDEALRWREG_DOUBLE)
#define	rslot_double(i)		slotOffset(seq_slot(s, i), Rdouble, rread)
#define	wslot_double(i)		slotOffset(seq_slot(s, i), Rdouble, rwrite)
#define	rreg_ideal_double(i,r)	slotRegister(seq_slot(s, i), Rdouble, rread, r)

#define	rreg_subint(i)		slotRegister(seq_slot(s, i), Rsubint, rread, IDEALRREG_SUBINT)
#define	wreg_subint(i)		slotRegister(seq_slot(s, i), Rsubint, rwrite, IDEALWREG_SUBINT)
#define	rwreg_subint(i)		slotRegister(seq_slot(s, i), Rsubint, rread|rwrite, IDEALRWREG_SUBINT)
#define	rslot_subint(i)		slotOffset(seq_slot(s, i), Rsubint, rread)
#define	wslot_subint(i)		slotOffset(seq_slot(s, i), Rsubint, rwrite)
#define	rreg_ideal_subint(i,r)	slotRegister(seq_slot(s, i), Rsubint, rread, r)

#define	sreg(I)			reginfo[(seq_slot(s,I)->regno)].regno
#define	sreg_int(i)		sreg(i)
#define	sreg_ref(i)		sreg(i)
#define	sreg_long(i)		sreg(i)
#define	sreg_float(i)		sreg(i)
#define	sreg_double(i)		sreg(i)

#define	lreg_int(i)		slotRegister(seq_slot(s, i), Rint, rreload, NOREG)     
#define	lreg_ref(i)		slotRegister(seq_slot(s, i), Rref, rreload, NOREG)     
#define	lreg_long(i)		slotRegister(seq_slot(s, i), Rlong, rreload, NOREG)     
#define	lreg_float(i)		slotRegister(seq_slot(s, i), Rfloat, rreload, NOREG)     
#define	lreg_double(i)		slotRegister(seq_slot(s, i), Rdouble, rreload, NOREG)     

#define	const_int(I)		s->u[I].value.i
#define const_long(I)		s->u[I].value.l
#define const_label(I)		s->u[I].labconst
#define	const_float(I)		s->u[I].value.d
#define	const_double(I)		s->u[I].value.d

/* Reserve a register */
#define register_reserve(r)     (reginfo[r].type |= Reserved)
#define register_unreserve(r)   (reginfo[r].type &= ~Reserved)

#define	calleeSave(R)		(reginfo[R].flags & Rnosaveoncall)

/* Is a slot already in a register? */
#define	_inRegister(S,T)	((reginfo[(S)->regno].ctype & (T)) != (0))
#define	inRegister(I,T)		(_inRegister(seq_slot(s,I), T))

/* Is a register global? */
#define	isRegisterGlobal(R)	((reginfo[(R)].type & Rglobal) != 0)

extern void initGlobalRegisters(int);
extern void setGlobalRegister(int);

extern void sanityCheck(void);
extern void reload(SlotData*);

extern void HAVE_move_register_int(int, int);
extern void HAVE_move_register_ref(int, int);
extern void HAVE_move_register_long(int, int);
extern void HAVE_move_register_float(int, int);
extern void HAVE_move_register_double(int, int);

#endif
