/* seq.c
 * Pseudo instruction sequences.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "seq.h"
#include "gc.h"
#include "machine.h"

sequence* firstSeq;
sequence* lastSeq;
sequence* currSeq;
sequence* activeSeq;

/*
 * Reset the sequence list.
 */
void
initSeq(void)
{
#if !defined(OPTIMIZE_MEMORY)
	currSeq = firstSeq;
#else
	while(firstSeq != 0) {
		sequence* s;
		s = firstSeq->next;
		gc_free_fixed(firstSeq);
		firstSeq = s;
	}
	lastSeq = 0;
#endif
}

/*
 * Allocate a new sequence element.
 */
sequence*
nextSeq(int nr)
{
	sequence* ret;

#if !defined(OPTIMIZE_MEMORY)
	ret = currSeq;
	if (ret == 0) {
		int i;
		/* Allocate chunk of sequence elements */
		ret = gc_malloc_fixed_typed(ALLOCSEQNR * sizeof(sequence), GC_ALLOC_ENGINE);
		if (ret == 0) {
			longjmp(tjbuf, 1);
			return (0);
		}

		/* Attach to current chain */
		if (lastSeq == 0) {
			firstSeq = ret;
		}
		else {
			lastSeq->next = ret;
		}
		lastSeq = &ret[ALLOCSEQNR-1];

		/* Link elements into list */
		for (i = 0; i < ALLOCSEQNR-1; i++) {
			ret[i].next = &ret[i+1];
		}
		ret[ALLOCSEQNR-1].next = 0;
	}
	currSeq = ret->next;
#else
	ret = gc_malloc_fixed_typed(sizeof(sequence) - (sizeof(jvalue) * (SEQSLOTS - nr)), GC_ALLOC_ENGINE);
	if (ret == 0) {
		return (0);
	}
	if (lastSeq == 0) {
		firstSeq = ret;
	}
	else {
		lastSeq->next = ret;
	}
	lastSeq = ret;
#endif
	ret->lastuse = 0;
	ret->refed = 1;
	activeSeq = ret;
	return (ret);
}
