/* slots.c
 * Slots.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "gtypes.h"
#include "slots.h"
#include "seq.h"
#include "md.h"
#include "registers.h"
#include "gc.h"

int maxslot;
SlotInfo* basicslots;
SlotInfo* slotinfo;
SlotInfo* localinfo;
SlotInfo* tempinfo;
SlotInfo stack_limit[1];	/* Used to represent the stack limit */

static int lastnrslots = 0;

extern int maxLocal;
extern int maxArgs;
extern int maxStack;
extern int maxTemp;

/*
 * Initiate slots.
 */
void
initSlots(int islots)
{
	static SlotData* basicdatas;
	static SlotData stack_limit_data;

	int i;
	int nrslots = islots;

	/* Allocate extra slots for temps */
	nrslots += MAXTEMPS;

	/* Make sure we have enough slots space */
	if (nrslots > lastnrslots) {
		basicslots = gc_realloc_fixed(basicslots, nrslots * sizeof(SlotInfo));
		basicdatas = gc_realloc_fixed(basicdatas, nrslots * sizeof(SlotData));
		lastnrslots = nrslots;
	}
	/* Set 'maxslot' to the maximum slot usable (excluding returns) */
	maxslot = nrslots;

	/* Setup various slot pointers */
	slotinfo = &basicslots[0];
	localinfo = &slotinfo[0];
	tempinfo = &localinfo[islots];

        /* Setup all slots */
        for (i = 0; i < nrslots; i++) {
		basicslots[i].slot = &basicdatas[i];
		basicslots[i].slot->regno = NOREG;
		basicslots[i].slot->modified = 0;
		basicslots[i].slot->rnext = 0;
		basicslots[i].slot->rseq = 0;
		basicslots[i].slot->wseq = 0;
		basicslots[i].slot->offset = SLOT2FRAMEOFFSET(&basicslots[i]);
		basicslots[i].slot->global = GL_NOGLOBAL;
        }

#if defined(STACK_LIMIT)
	/* And setup the stack offset */
	stack_limit->slot = &stack_limit_data;
	stack_limit->slot->regno = NOREG;
	stack_limit->slot->modified = 0;
	stack_limit->slot->rnext = 0;
	stack_limit->slot->rseq = 0;
	stack_limit->slot->wseq = 0;
	stack_limit->slot->offset = SLOT2FRAMEOFFSET(stack_limit);
	stack_limit->slot->global = GL_NOGLOBAL;
#endif
}

/*
 * Setup for the beginning of a basic block.
 */
void
setupSlotsForBasicBlock(void)
{
	int i;
	SlotData* sdata;

	for (i = 0; i < maxslot; i++) {
		sdata = slotinfo[i].slot;
		sdata->rseq = 0;
		sdata->wseq = 0;
	}

#if defined(STACK_LIMIT)
	sdata = stack_limit->slot;
	sdata->rseq = 0;
	sdata->wseq = 0;
#endif
}

/*
 * Mark a number of slots as lastuse.
 */
void
lastuse_slot(SlotInfo* data, int nr)
{
	SlotData* sdata;

	for (; nr > 0; nr--, data++) {
		sdata = data->slot;
		if (sdata->rseq != 0 && !isGlobal(sdata)) {
			sdata->rseq->lastuse |= 1 << sdata->rseqslot;
		}
		sdata->wseq = 0;
		sdata->rseq = 0;
		NOTE_clear(data);
	}
}
