/*
 * asm.c
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "access.h"
#include "soft.h"
#include "support.h"
#include "env.h"

#if 0
static void
ASM_check_uses( instr_block * blk )
{
        expr* e;
        slot* s;
        int i;
        FILE * fd;

        fd = blk->cfg->fd;
LOG(	fprintf( fd, "non-zero uses: " ); );
        e = blk->exprs_head;
        while( e ){
LOG(	        if( e->uses.block != 0 ){
                        fprintf( fd, "\t[%d]: %d ", e->eid, e->uses.block );
                } );
                e = e->expr.next;
        }
        for( i = blk->cfg->env.method.n_locals; --i >= 0; ){
                s = (slot *)blk->locals.in[i];
LOG(	        if( s->uses.block != 0 ){
                        fprintf( fd, "\t[L%d]: %d ", i, s->uses.block );
                } );
        }
LOG(	fprintf( fd, "\n"); );
}
#endif

extern void EXPR_backup( expr *, cfg_info * );

void
ASM_load_expr( instr_block * blk, expr * e )
{
        asmFunc assm;
        cfg_info * cfg;

        cfg = blk->cfg;
        assert( e->uses.block > 0 );
LOG(    fprintf( cfg->fd, "\t%s: loading {%d}\n", SLOT_print( blk, e ), e->uses.block ); );
        if( !(e->flags & SE_generated) ){
LOG(            fprintf( cfg->fd, "\t%s: generate now!\n", SLOT_print( blk, e ) ); );
                assm = ASM_funcs[e->op];
                if( assm != NULL ){
                        (*assm)( e, blk, true );
                }
        }
	if( !(e->flags & SE_in_reg) ){
                REG_reload_expr( e, blk );
	}
        assert( e->flags & SE_in_reg );
}

static inline void
ASM_use_expr( instr_block * blk, expr * e )
{
        e->uses.block--;
        if( e->uses.block <= 0 && e->uses.other <= 0 && e->flags & SE_in_reg ){
                REG_free_expr( e, blk->cfg, true );
                if( e->flags & SE_in_frame_slot && !(e->flags & SE_slot) ){
                        TEMP_free( blk->cfg, e->loc.slot.frame );
                }
        }
}

static void
ASM_load_exprs( instr_block * blk, int cnt, ... )
{
	va_list va_p;
        expr * e;
        int n;

        n = cnt;
        va_start( va_p, cnt );
        while( --n >= 0 ){
                e = va_arg( va_p, expr * );
                e->flags |= SE_no_spill;
        }
        va_end( va_p );
        va_start( va_p, cnt );
        while( --cnt >= 0 ){
                e = va_arg( va_p, expr * );
                ASM_load_expr( blk, e );
        }
        va_end( va_p );
}

static void
ASM_use_exprs( instr_block * blk, int cnt, ... )
{
	va_list va_p;
        expr * e;

        va_start( va_p, cnt );
        while( --cnt >= 0 ){
                e = va_arg( va_p, expr * );
                e->flags &= ~SE_no_spill;
                ASM_use_expr( blk, e );
        }
        va_end( va_p );
}

static void
ASM_req_fp_reg_start( expr * e, instr_block * b, int * type )
{
        if( e->flags & SE_req_fp_reg ){
                if( e->uses.block == 1  ){
                        assert( !PRIM_FP( e->type ) );
                        *type = e->type;
                        e->type = e->type == PRIM_long ? PRIM_double : PRIM_float;
                        kprintf( "FP OPT!!!\n" );
                }
                else{
                        e->flags &= ~SE_req_fp_reg;
                }
        }
}

static void
ASM_req_fp_reg_end( expr * e, instr_block * b, int * type )
{
        if( e->flags & SE_req_fp_reg ){
                e->type = *type;
        }
}


static expr *
REG_new_temp( instr_block * b, expr * e, int t )
{
        bzero( e, sizeof(expr) );
        e->type = t;
        e->uses.block = 1;
        REG_new( e, b, 0 );
        return e;
}

#define REG_free_temp( _b, _e ) \
	ASM_use_expr( _b, _e )

static void
PREP_link_call( instr_block * b, expr * e, expr_frame * f )
{
        if( !(f->flags & CALL_throw) ){
                if( b->call.last == NULL ){
                        b->call.first = e;
                }
                else{
                        b->call.last->a[1] = e;
                }
                b->call.last = e;
        }
};

static void
ASM_opt_next_call( instr_block * b, expr * e )
{
        expr * call_e, * a_e;
        expr_frame * frame;
        int i, sz;
        arg_info * arg_p;

        for( call_e = e->a[1]; call_e != NULL; call_e = call_e->a[1] ){
                frame = call_e->val.frame;
                if( !(frame->flags & CALL_throw) && !(frame->flags & CALL_inline) ){
                        sz = frame->sz;
                        for( i = sz, arg_p = &frame->a[sz-1]; --i >= 0; arg_p-- ){
                                a_e = arg_p->expr;
                                if( a_e != NULL
                                    && arg_p->flags & SE_in_reg 
                                    && !EXPR_IS_CONST( a_e )
                                    && !(a_e->flags & SE_in_reg) 
                                    && !(a_e->flags & SE_generated) ){
                                        kprintf( "******************************************************** [%d]: opt [%d]\n", call_e->eid, a_e->eid );
                                        a_e->flags |= SE_in_reg;
                                        a_e->loc.reg = arg_p->loc.reg;
                                }
                        }
                        break;
                }
        }
};

void
ASM_catch_spill_check( expr * e, instr_block * blk )
{
        boolean must_spill;
        int i;
        instr_block * catch_b;

        if( e->flags & SE_global ){
                REG_spill_expr( e, blk, false );
                return;
        }
        must_spill = false;
        for( i = blk->catch.length; --i >= 0; ){
                catch_b = blk->catch.blocks[i];
                if( STATS_local_get( SO_load, catch_b->handler_locals, e->loc.slot.frame ) ){
                        must_spill = true;
                }
        }
        if( must_spill ){
                REG_spill_expr( e, blk, false );
        }
}

static void ASM_copy( expr *, instr_block *, boolean );
static int  ASM_frame_offset( expr *, cfg_info * );
static int  ASM_frame_size( cfg_info * );

#if defined(TGT_MIPS)
#include "asm-mips.def"
#elif defined(TGT_SPARC)
#include "asm-sparc.def"
#elif defined(TGT_ARM)
#include "asm-arm.def"
#elif defined(TGT_I386)
#include "asm-i386.def"
#endif

/*
 * Calculate internal runtime frame size
 */
static int
ASM_frame_size( cfg_info * cfg )
{
        int frame_sz, n_call_args;

#if defined(TGT_ARM)
        n_call_args = cfg->env.global.max_call_args <= 4 ? 0 : cfg->env.global.max_call_args - 4;
#elif defined(TGT_SPARC)
        n_call_args = cfg->env.global.max_call_args < 6 ? 6 : cfg->env.global.max_call_args;
#else
        n_call_args = cfg->env.global.max_call_args;
#endif
        frame_sz = RESERVED_BELOW_FP + RESERVED_ABOVE_FP 
                + cfg->env.global.cframe.max_words + 2 * cfg->env.global.max_temps 
                        + n_call_args;
#ifdef OPTIMIZE_GCSE
        frame_sz += cfg->env.global.globals.words;
#endif
LOG(    fprintf( cfg->fd, "\tFrame: %d(rsv_below_FP) + %d(mx_O) + 2 * %d(mx_T) + %d(out_A) + %d(mx_G) = %d\n",
                RESERVED_BELOW_FP, cfg->env.global.cframe.max_words,
                cfg->env.global.max_temps, n_call_args,
                cfg->env.global.globals.length, frame_sz ); );
        FRAME_ALIGN( frame_sz );

        return frame_sz;
}

/*
 * Translate a slot number into a position on runtime stack
 */
static int
ASM_frame_offset( expr * e, cfg_info * cfg )
{
        int offset, pos, base_offset;
        method_env * m_env;
        global_env * g_env;

        offset = e->loc.slot.frame;
        g_env = &cfg->env.global;
#ifdef OPTIMIZE_GCSE
        if( e->flags & SE_global ){
                m_env = cfg->env.method.source;
                pos = - RESERVED_BELOW_FP - g_env->cframe.max_words - g_env->globals.words + g_env->globals.map[offset].position;
LOG( 		fprintf( cfg->fd, "GLOBAL_offset: [%d]: 4 * ( - %d(rsv_below_FP) - %d(mx_W) - %d(mx_G) + %d(G) ) = %d (%d)\n", 
                        offset, RESERVED_BELOW_FP, g_env->cframe.max_words,
                        g_env->globals.words, offset, WORD_SZ * pos, pos ); );
        }
        else if( offset >= g_env->cframe.max_slots + g_env->globals.length ){
#else
        if( offset >= g_env->cframe.max_slots ){
#endif
                pos = - RESERVED_BELOW_FP - 2*(offset - g_env->cframe.max_slots - g_env->globals.length) - g_env->cframe.max_words - g_env->globals.words;
LOG( 		fprintf( cfg->fd, "TEMP_offset: [%d]: 4 * ( - %d(rsv_below_FP) - 2 * ( %d(T) - %d(mx_S) - %d(mx_G) ) - %d(mx_W) ) = %d (%d)\n", 
                        offset, RESERVED_BELOW_FP, offset, g_env->cframe.max_slots, 
                        g_env->globals.words, g_env->cframe.max_words,
                        WORD_SZ * pos, pos ); );
        }
        else{
                m_env = cfg->env.method.source;
#ifdef OPTIMIZE_INLINE
                while( true ){
                        if( e->eid >= m_env->first_id ){
                                break;
                        }
                        m_env = m_env->caller->env;
                }
                base_offset = m_env->cframe.offset.words;
#else
                base_offset = m_env->cframe.offset.words;
#endif
                if( m_env == cfg->env.global.root_env && offset < g_env->n_args ){
                        pos = RESERVED_ABOVE_FP + cfg->env.global.arg_map[e->loc.slot.frame];
LOG( 		        fprintf( cfg->fd, "ARG_offset: [%d]: 4 * ( %d(rsv_above_fp) + %d(A) ) = %d (%d)\n",
                                 offset,  RESERVED_ABOVE_FP, cfg->env.global.arg_map[e->loc.slot.frame],
                                 WORD_SZ * pos, pos ); );
                }
                else{
                        /* Position locals closer to top of stack */
                        pos = - RESERVED_BELOW_FP - base_offset + m_env->cframe.map[offset].position;
LOG( 		        fprintf( cfg->fd, "SLOT_offset: [%d]: D%d: 4 * ( - %d(rsv_below_FP) - %d(base_O) + %d(map[S]) ) = %d (%d)\n", 
                                 offset, m_env->inline_depth, RESERVED_BELOW_FP, 
                                 base_offset, offset, WORD_SZ * pos, pos ); );
                }
        }
        return WORD_SZ * pos;
}

static void
ASM_copy( expr * cur_e, instr_block * blk, boolean unused )
{
	expr * s_e;
        asmFunc assm;
        reg_info * cur_r;
        FILE * fd;
        cfg_info * cfg;

        cfg = blk->cfg;
        fd = cfg->fd;
        s_e = cur_e->a[0];
LOG(    fprintf( fd, "\t%s(%x) {%d,%d} -> [%d] (%x) {%d,%d}: ", SLOT_print( blk, s_e ), s_e->flags, s_e->uses.block, s_e->uses.other, cur_e->eid, cur_e->flags, cur_e->uses.block, cur_e->uses.other  ); );
        if( cur_e->uses.block == 0 && cur_e->uses.other == 0 ){
LOG(            fprintf( fd, "unused " ); );
                if( s_e->flags & SE_in_reg ){
LOG(                    fprintf( fd, "in_reg " ); );
                        cur_r = &cfg->env.block.regs.all[s_e->loc.reg.base];
                        cur_r->expr = cur_e;
                        cur_e->loc.reg.base = s_e->loc.reg.base;
                        cur_e->loc.reg.pair = s_e->loc.reg.pair;
                        cur_e->flags |= SE_in_reg;
//                        cur_e->flags |= SE_in_reg | SE_generated; 
                }
                return;
        }
        cur_e->flags |= s_e->flags & 0xfffff000;
        bcopy( &s_e->val, &cur_e->val, sizeof(constant_info) );
        if( s_e->flags & SE_generated && s_e->flags & (SE_in_frame_slot | SE_in_reg) ){
LOG(            fprintf( fd, "generated " ); );
                cur_e->flags |= SE_generated;
                /* Need to be careful if copy frame slot has been allocated  */
                if( s_e->flags & SE_slot ){
LOG(                    fprintf( fd, "slot " ); );
                        cur_e->flags |= SE_slot;
                        cur_e->loc.slot.frame = s_e->loc.slot.frame;
#ifdef OPTIMIZE_GCSE
                        if( s_e->flags & SE_global ){
                                cur_e->flags |= SE_global;
                                cur_e->type = s_e->type;
                        }
#endif
                }
                if( s_e->flags & SE_in_frame_slot ){
LOG(                    fprintf( fd, "frame_slot " ); );
                        cur_e->flags |= SE_in_frame_slot;
                        cur_e->loc.slot.frame = s_e->loc.slot.frame;
                }
                if( s_e->flags & SE_in_reg ){
LOG(                    fprintf( fd, "in_reg " ); );
			if( cur_e->flags & SE_in_reg 
			   && (cur_e->loc.reg.base != s_e->loc.reg.base 
			       || cur_e->loc.reg.pair != s_e->loc.reg.pair ) ){
	                        cur_e->op = OP_move;
                                cfg->env.global.no_free = true;
                                s_e->uses.block++; /* decremented by ASM_move */
	                        ASM_move( cur_e, blk, unused );
                                cfg->env.global.no_free = false;
                                REG_free_expr( s_e, cfg, false );
			}
			else{
                                REG_free_expr( s_e, cfg, false );
			        cur_e->flags |= SE_in_reg;
			        cur_e->loc.reg.base = s_e->loc.reg.base;
			        cur_e->loc.reg.pair = s_e->loc.reg.pair;
                                cfg->env.global.no_free = true;
			        REG_assign_expr( cur_e, cfg, REGOP_assign );
                                cfg->env.global.no_free = false;
			}
                }
                else{
			if( cur_e->flags & SE_in_reg ){
                                REG_reload_expr( cur_e, blk );
                        }
                }
                cur_e->op = s_e->op;
                cur_e->a[0] = s_e->a[0];
                cur_e->a[1] = s_e->a[1];
        }
        else{
                /* Copy flags ??? */
                cur_e->op = s_e->op;
                cur_e->a[0] = s_e->a[0];
                cur_e->a[1] = s_e->a[1];
                if( s_e->flags & SE_in_reg ){
LOG(                    fprintf( fd, "in_reg " ); );
                        if( cur_e->flags & SE_in_reg 
                            && (cur_e->loc.reg.base != s_e->loc.reg.base 
                                || cur_e->loc.reg.pair != s_e->loc.reg.pair ) ){
                                /* SOMETHING OF A HACK */
                                FAILED();
                        }
                        else{
                                cur_e->flags |= SE_in_reg;
                                cur_e->loc.reg.base = s_e->loc.reg.base;
                                cur_e->loc.reg.pair = s_e->loc.reg.pair;
                        }
                }
                assm = ASM_funcs[cur_e->op];
                if( assm != NULL ){
                        (*assm)( cur_e, blk, false );
                }
        }
LOG(    fprintf( fd, "\n" ); );
}

