/*
 * asm.h
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __asm_h
#define __asm_h

typedef void (*prepFunc)( expr *, instr_block * );
extern prepFunc PREP_funcs[];

typedef void (*asmFunc)( expr *, instr_block *, boolean );
extern asmFunc ASM_funcs[];

extern void ASM_reload( expr *, instr_block * );
extern void ASM_spill( expr *, instr_block * );
extern void ASM_load_expr( instr_block *, expr * );
extern expr * EXPR_new_move( expr *, instr_block * );
extern void EXPR_make_move( expr *, expr *, instr_block * );
extern expr * EXPR_new_move_to( expr *, instr_block *, int, int );
extern void EXPR_make_move_to( expr *, expr *, instr_block *, int, int );
extern void ASM_move( expr *, instr_block *, boolean );
extern void EXPR_remove_from_exprs_by_op( instr_block *, expr * );
extern void ASM_switch_branch( uint8 * );
extern boolean can_simplify_call( Hjava_lang_Class *, Method * );
extern void ASM_catch_spill_check( expr *, instr_block * );

#define ASM_define_vars \
        uint8 * code_p; \
        int op

#define ASM_init_vars \
        op = 0

#define EXPR_IS_CONST( _e ) \
	( ((_e)->op > OP_nop ) && ((_e)->op <= OP_const) )

#define EXPR_IS_BRANCH( _e ) \
	( ((_e)->op >= OP_cbranch_eq) && ((_e)->op <= OP_branch_indirect) )

#define EXPR_IS_CBRANCH( _e ) \
	( ((_e)->op >= OP_cbranch_eq) && ((_e)->op <= OP_cbranch_ult) )

#define EXPR_IS_CMOV( _e ) \
	( ((_e)->op >= OP_cmov_eq) && ((_e)->op <= OP_cmov_le) )

#define EXPR_SWAP( _e1, _e2 ) swap_e = (_e1); (_e1) = (_e2); (_e2) = swap_e
#define EXPR_SWAP_IF( _cond, _e1, _e2 ) if( _cond ){ EXPR_SWAP( _e1, _e2 ); }

#define PREP_FUNCTION( _name ) static void PREP_##_name( expr * cur_e, instr_block * blk )
#define ASM_FUNCTION( _name ) static void ASM_##_name( expr * cur_e, instr_block * blk, boolean force_generate )
#define LAB_FUNCTION( _name ) static void LAB_##_name( label_info * l, cfg_info * cfg )
// #define CODE_PC 	( blk->native.end - blk->cfg->env.global.native.consttable.start )
#define CODE_PC 	( blk->native.end )

#ifdef  JIT_DBG
#define PRINT_INSN( _m ) \
	LOG( fprintf _m; \
	DISASM_instruction( cfg->fd, blk, CODE_PC ) )
#else
#define PRINT_INSN( _m )
#endif

#define ASM_FAIL( _m ) \
	kprintf( "ABORT: [%d] @ B%d: %s\n", cur_e->eid, blk->bid, _m ); \
	ABORT()

#define W_CODE			(*(uint32 *)code_p)

#define BITS2LIMIT( _s ) \
	( 1 << ((_s) - 1 ) )

#define MASK_BITS( _b ) ( (1 << (_b)) - 1 )
     
#define CHECK_CONSTANT( _v, _sz ) \
	( (_v) >= -BITS2LIMIT( _sz ) && (_v) < BITS2LIMIT( _sz ) )

/*
 * These short macros make the assembler a bit more readable 
 */
/* Registers */
#define R( _e )		(_e)->loc.reg.base
#define RP( _e )	(_e)->loc.reg.base

/* Integer high/low for long */
#define IR_L( _e )	(_e)->loc.reg.pair
#define IR_H( _e )	(_e)->loc.reg.base

/* Constants */
#define C( _e )		(_e)->val.i_const
#define CI( _c )	( WORD_SZ * (_c)->index )

#define FRAME_ALIGN( _v ) ALIGN_VALUE( _v, FRAME_ALIGNMENT )

#define WORD_BIGENDIAN
#ifdef  WORD_BIGENDIAN 
#define OFFSET_L	4
#define OFFSET_H	0
#else
#define OFFSET_L	0
#define OFFSET_H	4
#endif

#endif
