/*
 * bitvector.c
 * Bitvector implementation
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "env.h"
#include "gc.h"
#include "bitvector.h"
#include "support.h"

uint **
BV_new_array( int n_elem, int n_words )
{
        int n;
        uint** array = gc_malloc_fixed( n_elem * sizeof(uint*) );
        for( n = n_elem; --n >= 0; ){
                array[n] = BV_new( n_words );
        }
        return array;
}

void
BV_free_array( uint ** array, int n_elem )
{
        int n;
        for( n = n_elem; --n >= 0; ){
                gc_free_fixed( array[n] );
        }
        gc_free_fixed( array );
}

void
BV_set_all( uint * bv, int n_blocks )
{
        int n;

        for (n = BV_words( n_blocks ); --n >= 0; bv++) {
                *bv = 0xffffffff;
        }
}

void
BV_unset_all( uint * bv, int n_blocks )
{
        int n;

        for (n = BV_words( n_blocks ); --n >= 0; bv++) {
                *bv = 0x0;
        }
}

void
BV_and( uint * dst, uint * src1, uint * src2, int n_words )
{
        int n;

        for( n = n_words; --n >= 0; dst++, src1++, src2++ ){
                *dst = *src1 & *src2;
        }
}

void
BV_or( uint * dst, uint * src1, uint * src2, int n_words )
{
        int n;

        for( n = n_words; --n >= 0; dst++, src1++, src2++ ){
                *dst = *src1 | *src2;
        }
}

void
BV_copy( uint * dst, uint * src, int n_words )
{
        int n;

        for( n = n_words; --n >= 0; dst++, src++ ){
                *dst = *src;
        }
}

boolean
BV_equal( uint * src1, uint * src2, int n_words )
{
        int n;
        boolean cmp = true;

        for( n = n_words; --n >= 0; src1++, src2++ ){
                if( *src1 != *src2 ){
                        cmp = false;
                }
        }
        return cmp;
}
