/* 
 * consts.c
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "env.h"

void
CONST_generate_all( cfg_info * cfg ){
        const_info * cur_c;
        uint8 * base_p, * code_p;
        label_info * l;
        FILE * fd;

        fd = cfg->fd;
        base_p = cfg->env.global.native.consttable.end;
        for( cur_c = cfg->env.global.constants.list; cur_c != NULL; cur_c = cur_c->next ){
                code_p = base_p + WORD_SZ * cur_c->index;
LOG(            fprintf( fd, "\t\t0x%x: constant: ", code_p - cfg->env.global.pbase ); );
                switch( cur_c->type ){
                case PRIM_short:
                case PRIM_char:
                case PRIM_byte:
                case PRIM_int:
                        assert( cur_c->aux.method == NULL );
LOG(                    fprintf( fd, "%d\n", cur_c->val.i_const ); );
                        *(jint *)code_p = cur_c->val.i_const;
                        break;
                case PRIM_ref:
                        if( cur_c->aux.method != NULL ){
                                addMethodPatch( cur_c->aux.method, code_p, PS_const );
                        }
LOG(                    fprintf( fd, "0x%p\n", cur_c->val.ref ); );
                        *(jint *)code_p = cur_c->val.i_const;
                        break;
                case PRIM_float:
LOG(                    fprintf( fd, "%f\n", cur_c->val.f_const ); );
                        *(jfloat *)code_p = cur_c->val.f_const;
                        break;
                case PRIM_long:
LOG(                    fprintf( fd, "%lld\n", cur_c->val.l_const ); );
                        *(jlong *)code_p = cur_c->val.l_const;
                        break;
                case PRIM_double:
LOG(                    fprintf( fd, "%llf\n", cur_c->val.d_const ); );
                        *(jdouble *)code_p = cur_c->val.d_const;
                        break;
                case PRIM_label:
                        l = cur_c->val.label;
                        if( l->type == LAB_expr_target ){
LOG(                            fprintf( fd, "T %p\n", LAB_get_to_addr( l ) ); );
                                *(uint8**)code_p = LAB_get_to_addr( l );
                        }
                        else if( l->type == LAB_block_target ){
LOG(                            fprintf( fd, "B %p\n", (void*)l->to.block->native.start ); );
                                *(uint8**)code_p = l->to.block->native.start;
                        }
                        else{
                                FAILED();
                        }
                        gc_free_fixed( l );
                        break;
                default:
                        FAILED();
                        break;
                }
        }
}

void
CONST_done( cfg_info * cfg ){
        const_info * cur_c, * next_c;

        for( cur_c = cfg->env.global.constants.list; cur_c != NULL; cur_c = next_c ){
                next_c = cur_c->next;
                gc_free_fixed( cur_c );
        }
}

const_info * 
CONST_lookup( cfg_info * cfg, expr * cur_e ){
        const_info * cur_c;
        boolean found;
        int type;

        found = false;
        type = cur_e->type;
        if( PRIM_TYPE( type ) != PRIM_vm_int ){
                cfg->env.global.constants.uses++;
        }
        for( cur_c = cfg->env.global.constants.list; cur_c != NULL; cur_c = cur_c->next ){
                if( type != cur_c->type ){
                        continue;
                }
                switch( type ){
                case PRIM_label:
                        found = (cur_e->val.label == cur_c->val.label) ? true : false;
                        break;
                case PRIM_ref: /* cur_e->val.ref == cur_c->val.ref */
                case PRIM_short:
                case PRIM_char:
                case PRIM_byte:
                case PRIM_int:
                        found = (cur_e->val.i_const == cur_c->val.i_const) ? true : false;
                        break;
                case PRIM_float:
                        found = (cur_e->val.f_const == cur_c->val.f_const) ? true : false;
                        break;
                case PRIM_long:
                        found = (cur_e->val.l_const == cur_c->val.l_const) ? true : false;
                        break;
                case PRIM_double:
                        found = (cur_e->val.d_const == cur_c->val.d_const) ? true : false;
                        break;
                }
                if( found ) break;
        }
        if( !found ){
                cur_c = gc_malloc_fixed( sizeof(const_info) );
                cur_c->type = cur_e->type;
                bcopy( &(cur_e->val), &(cur_c->val), sizeof(constant_info) );
                cfg->env.global.constants.max++;
                if( cur_c->type == PRIM_long || cur_c->type == PRIM_double ){
                        cfg->env.global.constants.max++;
                        DWORD_ALIGN( cfg->env.global.constants.max, 2 );
                }
                cur_c->index = - cfg->env.global.constants.max;
                SLL_PUSH( cfg->env.global.constants.list, cur_c, next );
        }
        cfg->env.global.constants.uses++;
        return cur_c;
}
