/*
 * counters.c
 * Try to use hardware cycle counters for stats
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "code.h"
#include "support.h"
#include "env.h"
#include <errno.h>

static boolean initd = false;
static boolean enabled = false;
static boolean started = false;
static unsigned long long tick_start;

#define MMAP 1
#define CYCLE_COUNTER_IS_64BIT 1

#if defined(TGT_MIPS)
#if defined(MMAP)
#include <stddef.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/syssgi.h>
#if CYCLE_COUNTER_IS_64BIT
typedef unsigned long long iotimer_t;
#else
typedef unsigned int iotimer_t;
#endif
volatile iotimer_t *iotimer_addr;
#else
#include <sys/hwperftypes.h>
#include <sys/hwperfmacros.h>
#include <sys/procfs.h>
#include <fcntl.h>
int generation;
int fd;
#endif
#elif defined(TGT_SPARC)
#include "perfmon.h"
#include <stdio.h>
#include <fcntl.h>
#endif

void 
CNTR_init( )
{
        if( !initd ){
#if defined(TGT_MIPS)
#if defined(MMAP)
                __psunsigned_t phys_addr, raddr;
                int fd, poffmask;
                unsigned int cyclesize, cycleval;
                
                initd = true;
                poffmask = getpagesize() - 1;
                syssgi( SGI_CYCLECNTR_SIZE, &cyclesize );
                phys_addr = syssgi( SGI_QUERY_CYCLECNTR, &cycleval );
                raddr = phys_addr & ~poffmask;
                fd = open("/dev/mmem", O_RDONLY);
                iotimer_addr = (volatile iotimer_t *)mmap(0, poffmask, PROT_READ,
                                                          MAP_PRIVATE, fd, (off_t)raddr);
                iotimer_addr = (iotimer_t *)((__psunsigned_t)iotimer_addr +
                                             (phys_addr & poffmask));
#else
                hwperf_profevctrarg_t evctr_args;
                int pid;
                char pfile[100];

                initd = true;
                pid = getpid();
                sprintf( pfile, "/proc/%05d", pid );
                fd = open( pfile, O_RDWR );
                bzero( &evctr_args, sizeof(hwperf_profevctrarg_t) );
                evctr_args.hwp_evctrargs.hwp_evctrl[0].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
                evctr_args.hwp_evctrargs.hwp_evctrl[0].hwperf_creg.hwp_ie = 1;
                evctr_args.hwp_evctrargs.hwp_evctrl[0].hwperf_creg.hwp_ev = 0;
                evctr_args.hwp_ovflw_sig = 0;
                if( (generation = ioctl( fd, PIOCENEVCTRS, (void *)&evctr_args )) < 0 ){
                        kprintf( "COUNTER PROBLEMS!!!\n" );
                }
#endif
#elif defined(TGT_SPARC)
                unsigned long long flags;
                int fd;

                initd = true;
                if( (fd = open( "/dev/perfmon", O_RDONLY )) == -1 ){
                        perror( "open(/dev/perfmon)" );
                        return;
                }
                flags = PCR_USER_TRACE | PCR_S0_CYCLE_CNT | PCR_S1_INSTR_CNT;
                if( ioctl( fd, PERFMON_SETPCR, &flags ) < 0 ){
                        perror( "ioctl()" );
                        return;
                }
#endif
                enabled = true;
        }
}

long long
CNTR_read( )
{
#if defined(TGT_MIPS) && !defined(MMAP)
        hwperf_cntr_t cnts;
        int mygeneration;
        int t;
#endif

        if( enabled ){
#if defined(TGT_MIPS)
#if defined(MMAP)
                return 100 * *iotimer_addr;
#else
                if( (mygeneration = ioctl( fd, PIOCGETEVCTRS, (void *)&cnts )) < 0 ){
                        kprintf( "COUNTER PROBLEMS!!!\n" );
                }
                assert( generation == mygeneration );
                return cnts.hwp_evctr[0];
#endif
#elif defined(TGT_SPARC)
                cpu_sync();
                return get_tick();
#endif
        }
        return 0L;
}

void
CNTR_begin( )
{
        CNTR_init();
        assert( !started );
        tick_start = CNTR_read();
        started = true;
}

long long
CNTR_end( )
{
        assert( started );
        started = false;
        return CNTR_read() - tick_start;
}

