/* 
 * jitstats.c
 * Method stats for various uses
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "env.h"
#include "access.h"
#include "baseClasses.h"

stats_info *
STATS_new( int n )
{
        stats_info * si;
        int words;

        words = BV_words( n );
        si = gc_malloc_fixed( sizeof(stats_info) - WORD_SZ + NUM_STATS * WORD_SZ * words );
        si->words = words;
        si->item[0] = &si->bv[0];
        si->item[1] = &si->bv[words];
        si->item[2] = &si->bv[2*words];
        si->item[3] = &si->bv[3*words];
        si->item[4] = &si->bv[4*words];
        return si;
}

static field_stats *
get_field_stats( field_list * flist, Hjava_lang_Class * c )
{
        int n, sz;
        field_stats * fs;

        for( n = flist->length; --n >= 0; ){
                fs = flist->classes[n];
                if( fs->class == c ){
                        return fs;
                }
        }
        sz = c->nsfields + c->bfsize;
        if( c->name[0] == '[' ){
                sz = 1;
        }
        fs = gc_malloc_fixed( sizeof(field_stats) );
        fs->class = c;
        fs->stats = STATS_new( sz );
        FL_add( flist, fs );
        return fs;
}

int
STATS_field_get_offset( Field * f, Hjava_lang_Class * c )
{
        int n;

        if( f->accflags & ACC_STATIC ){
                for( n = c->nsfields; --n >= 0; ){
                        if( &c->fields[n] == f ){
                                return n;
                        }
                }
        }
        else{
                return c->nsfields + f->info.boffset;
        }
        FAILED();
}

void
STATS_field_set( int type, field_list * flist, Field * f, Hjava_lang_Class * c )
{
        field_stats * fs;
        int offset;

        fs = get_field_stats( flist, c );
        offset = STATS_field_get_offset( f, c );
        BV_set( fs->stats->item[type], offset );
}

void
STATS_array_set( int type, field_list * flist, Hjava_lang_Class * elem )
{
        Hjava_lang_Class * array;
        field_stats * fs;

        if( elem == NULL ){
                elem = ObjectClass;
        }
        array = lookupArrayProcess( elem, CSTATE_LINKED );
        fs = get_field_stats( flist, array );
        BV_set( fs->stats->item[type], 0 );
}

void
STATS_fields_coalesce( field_list * dst_l, field_list * src_l )
{
        int n, i;
        field_stats * d_fs, * s_fs;
        stats_info * d_si, * s_si;

        for( n = src_l->length; --n >= 0; ){
                s_fs = src_l->classes[n];
                d_fs = get_field_stats( dst_l, s_fs->class );
                s_si = s_fs->stats;
                d_si = d_fs->stats;
                for( i = d_si->words; --i >= 0; ){
                        d_si->item[SO_load][i] |= s_si->item[SO_load][i];
                        d_si->item[SO_store][i] |= s_si->item[SO_store][i];
                }
        }
}

void
STATS_field_free( field_list * flist )
{
        int n;
        field_stats * fs;

        for( n = flist->length; --n >= 0; ){
                fs = flist->classes[n];
                STATS_free( fs->stats );
                gc_free_fixed( fs );
        }
}

void
STATS_locals_coalesce( stats_info * d, stats_info * s )
{
        int i;

        for( i = d->words; --i >= 0; ){
                d->item[SO_load][i] |= s->item[SO_load][i] | s->item[SO_iterator][i];
                d->item[SO_store][i] |= s->item[SO_store][i] | s->item[SO_iterator][i];
        }
}
