/* 
 * jitsupport.c
 * Various auxiliary routinues
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "env.h"

static int cnt_limit = 0;
static int sz_limit = 0;
static int jit3_print_id = 0;
static int jit4_print_id = 0;
int globals_limit = 8;
int inline_limit = 8;
int enable_schedule = false;
int enable_inline = false;
int enable_loopopt = false;
int enable_gcse = false;
int enable_cmov = false;

boolean
verify_check( boolean cond, FILE * fd, const char * mess, ... )
{
        va_list argptr;

        if( !cond ){
LOG(            va_start( argptr, mess );
                vfprintf( fd, mess, argptr );
                va_end( argptr ); );
                kprintf( "******** WARN ********\n" );
                va_start( argptr, mess );
                vfprintf( stderr, mess, argptr );
                va_end( argptr );
        }
        return cond;
}

boolean
verify_strict( boolean cond, FILE * fd, const char * mess, ... )
{
        va_list argptr;

        if( !cond ){
LOG(            va_start( argptr, mess );
                vfprintf( fd, mess, argptr );
                va_end( argptr ); );
                kprintf( "******** WARN ********\n" );
                va_start( argptr, mess );
                vfprintf( stderr, mess, argptr );
                va_end( argptr );
                FAILED();
        }
        return cond;
}

void *
jit4_soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
{
	Hjava_lang_Object* obj;
	jint arg;
	int i;
	int * arraydims;
	va_list ap;

        arraydims = alloca((dims+1) * sizeof(int));

	/* Extract the dimensions into an array */
	va_start(ap, dims);
	for (i = 0; i < dims; i++) {
		arg = va_arg(ap, jint);
		if (arg < 0) {
			postException("java/lang/NegativeArraySizeException", "");
			obj = 0;
			goto error;
		}
		arraydims[i] = arg;
	}
	arraydims[i] = -1;
	va_end(ap);

	/* Mmm, okay now build the array using the wonders of recursion */
        obj = newMultiArray(class, arraydims);

	error:;
	if (obj == 0) {
		throwPostedException();
	}

	/* Return the base object */
	return (obj);
}

char *
sigItemType( char ** p ){
        char * cp;
        while( true ){
                cp = *p;
                switch( *cp ){
                case '(':
                        (*p)++;
                        continue;
                        break;
                case ')':
                        return NULL;
                        break;
                case '[':
                arrayofarray:
                        (*p)++;
                        if( **p == 'L' ){
                                goto object;
                        }
                        else if( **p == '[' ){
                                goto arrayofarray;
                        }
                        (*p)++;
                        return cp;
                        break;
                case 'L':
                object:
                        while( **p != ';' ){
                                (*p)++;
                        }
                        (*p)++;
                        return cp;
                        break;
                default:
                        (*p)++;
                        return cp;
                        break;
                }
        }
}

int
sigItemCount( char * p, boolean adjust_sz ){
        int cnt = 0;
        char * tp;

        while( (tp = sigItemType( &p )) != NULL ){
                if( adjust_sz ){
                        if( *tp == 'D' || *tp == 'J' ){
                                cnt++;
                        }
                }
                cnt++;
        }
        return cnt;
}

char *
sigReturnType( char * p ){
        while( true ){
                switch( *p ){
                case ')':
                        p++;
                        return p;
                        break;
                default:
                        p++;
                        break;
                }
        }
}

char
sigToPrimType( char c ){
        switch( c ){
	case 'V':
                return PRIM_void;
		break;
	case 'L':
	case '[':
                return PRIM_ref;
		break;
	case 'I':
	case 'Z':
	case 'S':
	case 'B':
	case 'C':
                return PRIM_int;
		break;
	case 'F':
                return PRIM_float;
		break;
	case 'J':
                return PRIM_long;
		break;
	case 'D':
                return PRIM_double;
		break;
	default:	
                break;
        }
        ABORT();
        return 0;
}

int
count_bits( int num, int * b_pos ){
        int i, mask, cnt;

        cnt = 0;
        mask = 1;
        if( num <= 0 ){
                kprintf( "COUNT_BITS: %d\n", num );
        }
        for( i = 0; i < 32; i++, mask <<= 1 ){
                if( num & mask ){
                        cnt++;
                        (*b_pos) = i;
                }
        }
        return cnt;
}

boolean print_dbg = false;
int jit4_cnt;
int jit3_cnt;

static FILE * ifd = NULL;
extern FILE * pfd;

#include "jfilesystem.h"

static boolean
install_new( Method * meth ){
        char buffer[1000];
        if( ifd == NULL ){
                if( (ifd = fopen( "limit", "r+" )) == NULL ){
                        kprintf( "(%d): %s\n", errno, strerror(errno) );
                        FAILED();
                }
                fgets( buffer, 1000, ifd );
                fscanf( ifd, "%d %d %d %d %d %d %d %d %d %d %d", &cnt_limit, &sz_limit, &jit3_print_id, &jit4_print_id, &enable_inline, &inline_limit, &enable_schedule, &enable_gcse, &enable_loopopt, &globals_limit, &enable_cmov );
                fclose( ifd );
        }
/*
        if( !strcmp( meth->name->data, "fractureFrom" ) ){ // Stylepad error
                cnt_limit = 0;
                return false;
        }
*/
        if( meth->bcode->codelen < sz_limit && jit4_cnt < cnt_limit ){
                jit4_cnt++;
                print_dbg = jit4_cnt == jit4_print_id;
#ifndef INSTALL_CODE
                jit3_cnt++;
#endif
                return true;
        }
        else{
                jit3_cnt++;
                print_dbg = jit3_cnt == jit3_print_id;
        }
        return false;
}

boolean
invoke_new( Method * meth, Hjava_lang_Class * class ){
        cfg_info * cfg;
        long long cycles = 0L;

        if( install_new( meth ) ){
                cfg = CFG_new( meth, class, "out/4." );
#ifdef USE_GDB_DISASM
                DISASM_init( cfg->fd );
#endif
                COUNTER( CNTR_begin(); );
                SCAN_bytecodes( cfg );
		COUNTER_accumulate( cfg, scan );
                DFG_build( cfg );
                pfd = cfg->fd;
                CFG_free( cfg );
                cycles = cfg->env.global.time.lookup + cfg->env.global.time.scan + cfg->env.global.time.expr + cfg->env.global.time.code;
                kprintf( "J4: %4d %5lld (%4lld %4lld %4lld %4lld)  %4d %4d 0x%p %s.%s%s %dG %dI\n", 
                        jit4_cnt, cycles/1000, 
                        cfg->env.global.time.lookup/1000,
                        cfg->env.global.time.scan/1000,
                        cfg->env.global.time.expr/1000,
                        cfg->env.global.time.code/1000,
                        cfg->env.global.bytecode_len,
                        cfg->env.global.native.code.end - cfg->env.global.native.consttable.start - cfg->env.global.native.unused,
                        (void*)cfg->env.global.native.code.start,
                        CLASS_CNAME( class ), meth->name, meth->signature,
                         cfg->env.global.globals.length,
                         cfg->env.global.n_inline
                         );
#ifdef INSTALL_CODE
                return true;
#else
                return false;
#endif
        }
        return false;
}
