/*
 * lab.c
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-math.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "env.h"
#include "support.h"

#define LOOKUP_TABLE_SIZE 10 /* Arbitrary constant, based on usage in kaffe.def */
#define LABEL_LIST_APPEND( _list, _l ) \
        if( (_list) != NULL ){ \
        	DLL_INSERT_AS_PREV( (_list), (_l), label_info *, label ); \
        } \
        (_list) = (_l)

void
LAB_init( cfg_info * cfg )
{

        cfg->env.global.labels.table = (label_info **)gc_malloc_fixed( LOOKUP_TABLE_SIZE * sizeof(label_info *) );
}

void
LAB_done( cfg_info * cfg )
{

        gc_free_fixed( cfg->env.global.labels.table );
}

void
LAB_free( label_info * l, cfg_info * cfg )
{
        if( l == cfg->env.global.labels.list ){
                cfg->env.global.labels.list = l->label.next;
        }
        DLL_REMOVE( l, label_info *, label );
}

label_info *
LAB_new_block_branch( cfg_info * cfg, expr * e, instr_block * blk )
{
        label_info * l;

        l = gc_malloc_fixed( sizeof(label_info) );
        l->from.expr = e;
        l->to.block = blk;
        l->type = LAB_block_target;
	LABEL_LIST_APPEND( cfg->env.global.labels.list, l );
        return l;
}

label_info *
LAB_new_local_branch( cfg_info * cfg, int ix, expr * e, boolean expr_is_target )
{
        label_info * l;
        expr ** e_p;

        assert( ix < LOOKUP_TABLE_SIZE );
        l = cfg->env.global.labels.table[ix];
        if( l == NULL ){
                l = gc_malloc_fixed( sizeof(label_info) );
                e_p = expr_is_target ? &l->to.expr : &l->from.expr;
                l->type = LAB_expr_target;
                cfg->env.global.labels.table[ix] = l;
LOG(            fprintf( cfg->fd, "LAB_new_target(): %d\n", ix ); );
		LABEL_LIST_APPEND( cfg->env.global.labels.list, l );
        }
        else{
                e_p = expr_is_target ? &l->to.expr : &l->from.expr;
                if( *e_p != NULL ){
LOG(                    fprintf( cfg->fd, "WARN: To label [%d] overwritten\n", ix ); );
                }
                else{
                        if( expr_is_target ){
                                assert( l->from.expr != NULL );
                        }
                        else{
                                assert( l->to.expr != NULL );
                        }
                        cfg->env.global.labels.table[ix] = NULL;
                }
        }
        *e_p = e;
        return l;
}

void
LAB_free_target( cfg_info * cfg, int ix )
{
        label_info * l;

        assert( ix < LOOKUP_TABLE_SIZE );
        l = cfg->env.global.labels.table[ix];
        assert( l != NULL );
        cfg->env.global.labels.table[ix] = NULL;
}

void
LAB_new_codeblock( cfg_info * cfg, instr_block * blk, int type )
{
        label_info * l;

        l = gc_malloc_fixed( sizeof(label_info) );
        l->patch = type;
        l->where = blk;
        l->to.addr = blk->native.end;
	LABEL_LIST_APPEND( cfg->env.global.labels.list, l );
}

label_info *
LAB_new_codepatch( cfg_info * cfg, instr_block * blk, uint8 * tgt, int type )
{
        label_info * l;

        l = gc_malloc_fixed( sizeof(label_info) );
        l->patch = type;
        LAB_set_from_addr( l, blk );
        l->to.addr = tgt;
	LABEL_LIST_APPEND( cfg->env.global.labels.list, l );
        return l;
}

void 
LAB_patch_all( cfg_info * cfg )
{
        label_info * cur_l, * next_l;
        FILE * fd;
        labelFunc lf;

        fd = cfg->fd;
LOG(    fprintf( fd, "LAB_patch_all():\n" ); );
        for( cur_l = cfg->env.global.labels.list; cur_l != NULL; cur_l = next_l ){
                next_l = cur_l->label.next;
                lf = LAB_funcs[cur_l->patch];
                if( lf != NULL ){
                        (*lf)( cur_l, cfg );
                }
                if( !(cur_l->flags & LAB_const_label) ){
                        gc_free_fixed( cur_l );
                }
        }
}
