/*
 * lab.h
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __lab_h
#define __lab_h

/* Flags */
#define LAB_const_label		0x20
#define LAB_throw_label		0x40

/* Label target types */
#define LAB_expr_target		0x01
#define LAB_block_target	0x02

/* Label patch types */
#define PATCH_undef		0x00
#define PATCH_branch		0x01
#define PATCH_prologue		0x02
#define PATCH_epilogue		0x03
#define PATCH_inline_data	0x04
#define PATCH_eprologue		0x05
#define PATCH_call		0x06
#define PATCH_profile		0x07

typedef union _link_info {
	expr *			expr;
	instr_block *		block;
        const_info *		constant;
	void *			addr;
	struct _throw_info *	throw;
} link_info;

typedef struct _label_info {
	uint8			flags;
	uint8			type;
	uint8			patch;
	linked_list		label;
	link_info		from;
	link_info		to;
	instr_block *		where;
        struct {
		Hjava_lang_Class * class;
		Method * 	method;
        } aux;
} label_info;

typedef void (*labelFunc)( label_info *, cfg_info * );
extern labelFunc LAB_funcs[];

extern void LAB_init( cfg_info * );
extern void LAB_done( cfg_info * );

extern void LAB_free( label_info *, cfg_info * );
extern label_info * LAB_new_block_branch( cfg_info *, expr *, instr_block * );
extern label_info * LAB_new_local_branch( cfg_info *, int, expr *, boolean );
extern void LAB_free_target( cfg_info *, int );
extern void LAB_new_codeblock( cfg_info *, instr_block *, int );
extern label_info * LAB_new_codepatch( cfg_info *, instr_block *, uint8 *, int );

extern void LAB_patch_all( cfg_info * );
#define LAB_set_to_addr( _l, _b ) \
 	(_l)->to.addr = (uint8*)((_b)->native.end - (_b)->native.start); \
 	(_l)->where = (_b)
#define LAB_get_to_addr( _l ) \
	( (_l)->where->native.start + (uint)(_l)->to.addr )
#define LAB_set_from_addr( _l, _b ) \
 	(_l)->from.addr = (uint8*)((_b)->native.end - (_b)->native.start); \
 	(_l)->where = (_b)
#define LAB_get_from_addr( _l ) \
	( (_l)->where->native.start + (uint)(_l)->from.addr )
#endif

