/* 
 * list.c
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "gtypes.h"
#include "classMethod.h"
#include "env.h"

inline void
IL_set_limit( void * _l, int min_sz )
{
        int o_limit, n_limit;
        void ** o_items;
        item_list * il;

        il = (item_list *)_l;
        if( il->items == NULL ){
                if( min_sz == 0 ){
                        min_sz = 4;
                }
                il->items = gc_malloc_fixed( min_sz * sizeof(void*) );
                il->limit = min_sz;
        }
        else if( min_sz >= il->limit ){
                o_limit = il->limit;
                n_limit = o_limit << 1;
                o_items = il->items;
                
                il->limit = n_limit;
                il->items = gc_malloc_fixed( n_limit * sizeof(void*) );
                bcopy( o_items, il->items, o_limit * sizeof(void*) );
                gc_free_fixed( o_items );
        }
}

void
IL_add( void * _l, void * it )
{
        item_list * il;

        il = (item_list *)_l;
        IL_set_limit( _l, il->length );
	il->items[il->length++] = it;
}

void
IL_del( void * _il, void * it )
{
        int n;
        void * tmp, ** items;
        item_list * il;

        il = (item_list*)_il;
        n = il->length-1;
        tmp = il->items[n];
        if( it != tmp ) {
                items = il->items;
                for( ; --n >= 0; ){
                        if ( it == items[n]) {
                                items[n] = tmp;
                        }
                }
        }
        il->length--;
}

boolean
IL_contains( void * _il, void * it )
{
        int n;
        void ** items;
        item_list * il;

        il = (item_list*)_il;
        items = il->items;
        for( n = il->length; --n >= 0; ){
                if( items[n] == it ){
                        return true;
                }
        }
        return false;
}

/*
 * Optimized block list manager uses BV_is_set() when calling BL_contains()
 */
