/* 
 * list.h
 *
 * Copyright (c) 1999 - 2001
 *      Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
#ifndef __list_h
#define __list_h

typedef struct _item_list {
	uint16 			length;
	uint16 			limit;
	void ** 		items;
} item_list;

#define LIST_copy( _d, _s ) \
	((item_list*)_d)->items = ((item_list*)_s)->items; \
	(_d)->limit = (_s)->limit; \
	(_d)->length = (_s)->length
#define LIST_clear( _l ) \
	(_l)->length = 0
#define LIST_zero( _l ) \
	((item_list*)_l)->items = NULL; \
	(_l)->limit = 0; \
	(_l)->length = 0
#define LIST_free( _l ) \
	gc_free_fixed( ((item_list*)_l)->items )
#define LIST_pop( _l ) \
	((item_list*)_l)->items[--(_l)->length]
#define LIST_push( _l, _i ) \
	IL_add( _l, _i )

extern inline void IL_set_limit( void *, int );
extern void IL_add( void *, void * );
extern void IL_del( void *, void * );
extern boolean IL_contains( void *, void * );

typedef struct _expr_list {
	uint16 			length;
	uint16 			limit;
	struct _expr ** 	exprs;
} expr_list;

typedef struct _block_list {
	uint16 			length;
	uint16 			limit;
	struct _instr_block ** 	blocks;
} block_list;

typedef struct _instr_list {
	uint16 			length;
	uint16 			limit;
	struct _instr_info ** 	instrs;
} instr_list;

typedef struct _field_list {
	uint16 			length;
	uint16 			limit;
	struct _field_stats **	classes;
} field_list;

#define BL_add( _bl, _b ) 	IL_add( _bl, _b )
#define BL_del( _bl, _b ) 	IL_del( _bl, _b )
#define BL_contains( _bl, _b )	IL_contains( _bl, _b )

#define EL_add( _el, _e ) 	IL_add( _el, _e )
#define EL_del( _el, _e ) 	IL_del( _el, _e )
#define EL_contains( _el, _e )	IL_contains( _el, _e )

#define FL_add( _el, _e ) 	IL_add( _el, _e )
#define FL_del( _el, _e ) 	IL_del( _el, _e )
#define FL_contains( _el, _e )	IL_contains( _el, _e )

#endif
