#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "objects.h"
#include "classMethod.h"
#include "code.h"
#include "profile.h"
#include "counters.h"
#include "access.h"

static unsigned int epoch = 0;
static dep_info * deps = NULL;
static unsigned long long cycles;
static epoch_info ** epochs;

//#define FIND_DELAY
#ifdef FIND_DELAY
static unsigned int times[5] = {0,0,0,0,0};
static unsigned long long delay[5] = {0,0,0,0,0};
static unsigned long long start;
#define TIME_START() start = CNTR_read();
#define TIME_END( _i ) \
	delay[_i] += CNTR_read() - start; \
	times[_i]++
#else
#define TIME_START()
#define TIME_END( _i )
#endif
#define L( _l ) _l

static int sig2sz( char c )
{
        switch( c ){
                case 'J':
                case 'D':
                        return 8;
                case 'L':
                case '[':
                case 'I':
                case 'F':
                        return 4;
                case 'S':
                case 'C':
                        return 2;
                case 'B':
                case 'Z':
                        return 1;
                }
        ABORT();
}

int
profileArraySize( int prim_sz, int n_elem )
{
        int total_sz, array_sz;

        total_sz = array_sz = prim_sz * n_elem;
        PROFILE_ALIGN( total_sz );
        total_sz += n_elem * sizeof(profile_info);
        return total_sz - array_sz;
}

static unsigned int
pc2line( Method * m, unsigned int pc )
{
        lineNumbers * lines;
        int i, start_pc, end_pc;

        if( (lines = m->lines) != NULL ){
                start_pc = 0;
                for( i = 0; i < lines->length; i++ ){
                        end_pc = lines->entry[i].pc_offset;
                        if( pc >= start_pc && pc < end_pc ){
                                return lines->entry[i].line_nr;
                        }
                        start_pc = end_pc;
                }
        }
}

loc_info *
LOC_new( Hjava_lang_Class * c, Method * m, unsigned int pc )
{
        loc_info * l;

        l = gc_malloc_fixed( sizeof(loc_info) );
        l->class = c;
        l->method = m;
        l->line = pc2line( m, pc );
        return l;
}

dep_info *
DEP_new( Hjava_lang_Class * c, Method * m, unsigned int pc, Hjava_lang_Class * fc, Field * f, unsigned char t )
{
        dep_info * d;

        d = gc_malloc_fixed( sizeof(dep_info) );
        d->at.class = c;
        d->at.method = m;
        d->at.line = pc2line( m, pc );
        d->field.class = fc;
        d->field.field = f;
        d->type = t;
        d->next = deps;
        deps = d;
        return d;
}

inline unsigned long long
PROF_counter( int diff )
{
        cycles += CNTR_read() - diff - cycles;
        return cycles;
}

void
PROF_put( dep_info * d, profile_info * p )
{
	TIME_START();
        p->cycle = PROF_counter( 110 );
        p->epoch = epoch;
        p->from = d;
L(      kprintf( "%llu P %s.%s @ %x\n", cycles, d->field.class->name, d->field.field->name, p ); );
	TIME_END( 0 );
}

void
PROF_get( dep_info * d, profile_info * p )
{
        long long int delta;

        if( p->epoch != epoch && p->cycle != 0 ){
                TIME_START();
                d->hit++;
                delta = PROF_counter( 220 ) - p->cycle;
L(              kprintf( "%llu G %s.%s @ %x\n", cycles, d->field.class->name, d->field.field->name, p ); );
                d->delta = ( d->delta * d->hit + (unsigned int)( delta ) )/( d->hit + 1 );
                if( p->from != NULL ){
L(              	kprintf( "\t\t%llu %llu\n", p->cycle, delta ); );
                        p->from->hit++;
                }
                p->cycle = 0;
                TIME_END( 1 );
        }
}

void
PROF_arraystore( dep_info * d, Hjava_lang_Object * o, unsigned int ix )
{
        unsigned int sz, offset;
        profile_info * p;
        HArrayOfInt * a;

	TIME_START();
        a = (HArrayOfInt*)o;
        sz = sig2sz( o->dtable->class->name[1] );
        offset = 12 + sz * a->length;
        PROFILE_ALIGN( offset );
        p = (profile_info*)( (void*)o + offset );
        p = &p[ix];
        p->cycle = PROF_counter( 150 );
        p->epoch = epoch;
        p->from = d;
L(      kprintf( "%llu S %s @ %x %d\n", cycles, o->dtable->class->name, o, ix ); );
	TIME_END( 2 );
}

void
PROF_arrayload( dep_info * d, Hjava_lang_Object * o, unsigned int ix )
{
        unsigned int sz, offset;
        long long int delta;
        profile_info * p;
        HArrayOfInt * a;

	TIME_START();
        a = (HArrayOfInt*)o;
        sz = sig2sz( o->dtable->class->name[1] );
        offset = 12 + sz * a->length;
        PROFILE_ALIGN( offset );
        p = (profile_info*)( (void*)o + offset );
        p = &p[ix];
	TIME_END( 3 );
        if( p->epoch != epoch && p->cycle != 0 ){
                TIME_START();
                delta = PROF_counter( 220 ) - p->cycle;
                d->delta = ( d->delta * d->hit + (unsigned int)( delta ) )/( d->hit + 1 );
L(              kprintf( "%llu L %s @ %x %d\n", cycles, o->dtable->class->name, o, ix ); );
                d->hit++;
                if( p->from != NULL ){
L(              	kprintf( "\t\t%llu %llu\n", p->cycle, delta ); );
                        p->from->hit++;
                }
                p->cycle = 0;
                TIME_END( 4 );
        }
}

void
PROF_call( loc_info * l, void * ra )
{
        // cycles -= 36;
L(      kprintf( "%llu C %s.%s\n", cycles, l->class->name, l->method->name ); );
}

void
PROF_return()
{
        // cycles -= 45;
L(	kprintf( "%llu R\n", cycles ); );
	epoch++;
}

void
PROF_loop_preheader( loc_info * l )
{
        // cycles -= 45;
L(      kprintf( "%llu H %s.%s\n", cycles, l->class->name, l->method->name ); );
	epoch--;
}

void
PROF_loop_iteration()
{
        // cycles -= 45;
L(	kprintf( "%llu I\n", cycles ); );
	epoch++;
}

void
PROF_test()
{
#define TIMES 100000
        unsigned long long start, end;
        int i;
        dep_info d;
        profile_info p;
        Hjava_lang_Object o;

        bzero( &d, sizeof(dep_info) );
        bzero( &p, sizeof(profile_info) );
        start = CNTR_read();
        for( i = TIMES; --i >= 0; ){
                PROF_put( &d, &p );
        }
        end = CNTR_read();
        kprintf( "(%llu - %llu)/%u = %llu\n", end, start, TIMES, (end-start)/TIMES );
}

void
DEP_print()
{
        dep_info * d;
	int i;

        for( d = deps; d != NULL; d = d->next ){
                if( d->hit > 0 ){
                        kprintf( "%s.%s:%d %s.%s %u %u %c %c\n", d->at.class->name, d->at.method->name, d->at.line, 
                                d->field.class->name, d->field.field != NULL ? d->field.field->name : "",
                                d->hit, d->delta, 
                                d->type == DEP_get ? 'G' : 'P',
                                d->field.field != NULL && d->field.field->accflags & ACC_STATIC ? 'S' : 'O' );
                }
        }
        PROF_test();
#ifdef FIND_DELAY
        for( i = 5; --i >= 0; ){
	        kprintf( "[%d]: %llu/%u = %u\n", i, delay[i], times[i], (unsigned int)(delay[i]/ times[i]) );
	}
#endif
}

