/*
 * regs.h
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __regs_h
#define __regs_h

/* Register type */
#define REG_float	0x01
#define REG_double	0x02
#define REG_int		0x04
#define REG_long	0x08

/* Register role */
#define REG_temp	0x00
#define REG_persist	0x01
#define REG_arg		0x02
#define REG_reserved	0x03

/* Register status */
#define REG_assigned	0x01
#define REG_was_used	0x02
#define REG_pair	0x04
#define REG_lock	0x08

/* Register operation (for printout) */
#define REGOP_assign	0x01
#define REGOP_new	0x02
#define REGOP_free	0x04
#define REGOP_pair	0x08
#define REGOP_no_cr	0x10

#define NUM_REG_TYPES 	4

#define REG_NULL	-9

typedef struct _reg_info {
        uint8 			type;
        uint8 			role;
        uint8 			status;
        uint8 			num;
	expr *			expr;
	struct {
	        int8		prev;
	        int8		next;
	} reg;
} reg_info;

extern void TEMP_free( cfg_info *, int );
extern temp_context * TEMP_copy_context( temp_context * );
extern int TEMP_new( cfg_info *, expr * );

extern void REG_reset_context( instr_block * );
extern void REG_set_base_context( instr_block * );

extern void REG_init( cfg_info *, struct _reg_info * );
extern void REG_done( cfg_info * );
extern void REG_backup_exprs( struct _cfg_info * );

extern void REG_free_expr( expr *, cfg_info *, boolean );
extern void REG_assign_expr( expr *, cfg_info *, int );
extern void REG_new( expr *, instr_block *, int, ... );

extern boolean REG_spill_expr( expr *, instr_block *, boolean );
extern void REG_reload_expr( expr *, instr_block * );
extern void REG_spill_before_call( instr_block *, boolean, boolean );
extern void REG_spill_before_throw( instr_block * );
extern void REG_fake_spill_before_call( instr_block * );
extern void REG_fake_reload_after_call( instr_block * );

extern reg_context * REG_copy_context( reg_context * );
extern void REG_merge_inline( reg_context *, instr_block * );

extern reg_info base_regfile[];
extern reg_info regs[];
extern char* reg_names[];

#ifdef TGT_SPARC
extern reg_info leaf_regfile[];
#endif

typedef struct _temp_info {
        uint16 			offset;
	expr *			expr;
	struct {
	        int16		prev;
	        int16		next;
	} temp;
} temp_info;

extern void TEMP_reset_context( instr_block * );

#endif
