/* 
 * scanner.c
 * Parse a method's bytecodes.
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "bytecode.h"
#include "classMethod.h"
#include "lookup.h"
#include "exception.h"
#include "code.h"
#include "access.h"

#include "linkedlist.h"
#include "list.h"
#include "expr.h"
#include "cfg.h"
#include "scanner.h"
#include "env.h"
#include "itypes.h"

#define SP_POP(_o)			sp -= (_o); local_valid = false
#define SP_PUSH(_o)			sp += (_o); local_valid = false
#define SP_MOVE(_i)			sp += moveSP[(_i)]; local_valid = false

#define	INSN_LEN(_i)			iLen[_i]
#define	INSN(pc)			code[(pc)]
#define	PC_INC(_i)			pc += (_i)

#define	WORD(pc)			(int16)( \
					 (code[(pc)+0] << 8) + \
					 (code[(pc)+1]))
#define	DWORD(pc)			(int32)( \
					 (code[(pc)+0] << 24) + \
					 (code[(pc)+1] << 16) + \
					 (code[(pc)+2] << 8) + \
					 (code[(pc)+3]))

/* Instruction lengths by bytecode */
const uint8 iLen[256] = {
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	2, 3, 2, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 
	3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 0, 0, 1, 1, 1, 1, 
	1, 1, 3, 3, 3, 3, 3, 3, 3, 5, 1, 3, 2, 3, 1, 1, 
	3, 3, 1, 1, 1, 4, 3, 3, 5, 5, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
};

/* Stack changes by bytecode */
const int8 moveSP[256] = {
	 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 1, 2, 2,
	 1, 1, 1, 1, 2, 1, 2, 1, 2, 1, 1, 1, 1, 1, 2, 2,
	 2, 2, 1, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1,-1, 0,
	-1, 0,-1,-1,-1,-1,-1,-2,-1,-2,-1,-1,-1,-1,-1,-2,
	-2,-2,-2,-1,-1,-1,-1,-2,-2,-2,-2,-1,-1,-1,-1,-3,
	-4,-3,-4,-3,-3,-3,-3,-1,-2, 1, 1, 1, 2, 2, 2, 0,
	-1,-2,-1,-2,-1,-2,-1,-2,-1,-2,-1,-2,-1,-2,-1,-2,
	-1,-2,-1,-2, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-2,
	-1,-2,-1,-2, 0, 1, 0, 1,-1,-1, 0, 0, 1, 1,-1, 0,
	-1, 0, 0, 0,-3,-1,-1,-3,-3,-1,-1,-1,-1,-1,-1,-2,
	-2,-2,-2,-2,-2,-2,-2, 0, 1, 0,-1,-1,-1,-2,-1,-2,
	-1, 0, 9, 9, 9, 9, 9, 9, 9, 9, 0, 1, 0, 0, 0,-1,
	 0, 0,-1,-1, 9, 9,-1,-1, 0, 1, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 
};

static const int8 localIx[20] = {
        0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3
};

static Hjava_lang_Class * classIx[8] = {
        &_Jv_intClass, &_Jv_longClass, &_Jv_floatClass, &_Jv_doubleClass,
        NULL, &_Jv_byteClass, &_Jv_charClass, &_Jv_shortClass
};

static void
SCAN_recursive( instr_block * blk, int pc, int sp, boolean rescan )
{
	int32 tabpc;
	int32 jpc;
	int32 npc;
	int32 idx;
        uint8 insn;
	bool wide;
        cfg_info * cfg;
	callInfo call;
	fieldInfo finfo;
        Method * meth;
        Hjava_lang_Class * class;
        bytecode * code;
        boolean local_valid;
        int local_ix;

#define RESCAN() if( rescan ){ blk->sp.out = sp; return; }

        if( blk->stage == STG_scanned ){
                return;
        }
        cfg = blk->cfg;
        meth = blk->cfg->env.method.method;
        class = blk->cfg->env.method.class;
LOG(    fprintf( cfg->fd, "CFG_build():\tscan %d - %d, pc = %d, sp = %d\n", blk->pc.in, blk->pc.out, pc, sp ); );
        blk->stage = STG_scanned;
        blk->sp.in = sp;
        code = meth->bcode->code;
	wide = false;
        local_valid = false;
        local_ix = 0;
	while( pc <= blk->pc.out ){
                insn = INSN(pc);
// LOG(            fprintf( cfg->fd, "[%d]: %d\n", pc, sp); );
		switch (insn) {
		case IFEQ:	case IFNE:	case IFLT:
		case IFGE:	case IFGT:	case IFLE:
		case IF_ICMPEQ: case IF_ICMPNE:
		case IF_ICMPLT: case IF_ICMPGE:
		case IF_ICMPGT: case IF_ICMPLE:
		case IF_ACMPEQ: case IF_ACMPNE:
		case IFNULL:	case IFNONNULL:
                        SP_MOVE(insn);
                        RESCAN();
			tabpc = pc + WORD(pc+1);
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, tabpc, CFG_branch );
                        pc = npc;
                        blk->sp.out = sp;
                        cfg->env.method.flags |= METH_has_branches;
                        SCAN_recursive( blk->block.next, npc, sp, false );
                        SCAN_recursive( CFG_get_block( cfg, tabpc, true ), tabpc, sp, false );
                        return;
			break;

		case GOTO:
                        RESCAN();
			tabpc = pc + WORD(pc+1);
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, tabpc, CFG_jump );
                        pc = npc;
                        blk->sp.out = sp;
                        SCAN_recursive( CFG_get_block( cfg, tabpc, true ), tabpc, sp, false );
                        return;
			break;

		case GOTO_W:
                        RESCAN();
			tabpc = pc + DWORD(pc+1);
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, tabpc, CFG_jump );
                        pc = npc;
                        blk->sp.out = sp;
                        SCAN_recursive( CFG_get_block( cfg, tabpc, true ), tabpc, sp, false );
                        return;
			break;

		case JSR:
                        SP_PUSH(1);
                        RESCAN();
			tabpc = pc + WORD(pc+1);
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, tabpc, CFG_jsr );
                        pc = npc;
                        blk->sp.out = sp;
                        SCAN_recursive( blk->block.next, npc, sp-1, false );
                        SCAN_recursive( CFG_get_block( cfg, tabpc, true ), tabpc, sp, false );
                        return;
			break;

		case JSR_W:
                        SP_PUSH(1);
                        RESCAN();
			tabpc = pc + DWORD(pc+1);
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, tabpc, CFG_jsr );
                        pc = npc;
                        blk->sp.out = sp;
                        SCAN_recursive( blk->block.next, npc, sp-1, false );
                        SCAN_recursive( CFG_get_block( cfg, tabpc, true ), tabpc, sp, false );
                        return;
			break;

		case TABLESWITCH:
                        SP_POP(1);
                        RESCAN();
			tabpc = (pc + 4) & -4;
			idx = DWORD(tabpc+8)-DWORD(tabpc+4)+1;
                        npc = tabpc + (DWORD(tabpc+8)-DWORD(tabpc+4)+4) * 4;
			for (; idx > 0; idx--) {
                                jpc = pc+DWORD(tabpc+idx*4+8);
                                CFG_set( cfg, pc, npc, jpc, CFG_switch );
                                SCAN_recursive( CFG_get_block( cfg, jpc, true ), jpc, sp, false );
			}
                        jpc = pc+DWORD(tabpc);
                        blk = CFG_set( cfg, pc, npc, jpc, CFG_switch );
                        SCAN_recursive( CFG_get_block( cfg, jpc, true ), jpc, sp, false );
			pc = tabpc + (DWORD(tabpc+8)-DWORD(tabpc+4)+1+3) * 4;
                        blk->sp.out = sp;
                        return;
			break;

		case LOOKUPSWITCH:
                        SP_POP(1);
                        RESCAN();
			tabpc = (pc + 4) & -4;
			idx = DWORD(tabpc+4);
                        npc = tabpc + (DWORD(tabpc+4)+1) * 8;
			for (; idx > 0; idx--) {
                                jpc = pc+DWORD(tabpc+idx*8+4);
                                CFG_set( cfg, pc, npc, jpc, CFG_switch );
                                SCAN_recursive( CFG_get_block( cfg, jpc, true ), jpc, sp, false );
			}
                        jpc = pc+DWORD(tabpc);
                        blk  = CFG_set( cfg, pc, npc, jpc, CFG_switch );
                        SCAN_recursive( CFG_get_block( cfg, jpc, true ), jpc, sp, false );
			pc = tabpc + (DWORD(tabpc+4)+1) * 8;
                        blk->sp.out = sp;
                        return;
			break;

		case IRETURN:	case LRETURN:	case ARETURN:
		case FRETURN:	case DRETURN:	case RETURN:
		case ATHROW:
                        SP_MOVE(insn);
                        RESCAN();
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, 0, CFG_return );
                        pc = npc;
                        blk->sp.out = sp;
                        return;
			break;

		case RET:
                        SP_MOVE(insn);
                        RESCAN();
			npc = pc + INSN_LEN(insn);
                        blk = CFG_set( cfg, pc, npc, 0, CFG_ret );
                        pc = npc;
                        blk->sp.out = sp;
                        return;
			break;

		case WIDE:
			wide = true;
			pc = pc + INSN_LEN(insn);
			break;

		case ILOAD:	case LLOAD:	case FLOAD:
		case DLOAD:	case ALOAD:
		case ISTORE:	case LSTORE:	case FSTORE:
		case DSTORE:	case ASTORE:
#ifdef OPTIMIZE_LOOPS
                        local_ix = idx = wide ? WORD(pc+1) : INSN(pc+1);
                        STATS_local_set( insn <= ALOAD ? SO_load : SO_store, blk->stats.locals, idx );
#endif
                        PC_INC(INSN_LEN(insn));
			if (wide == true) {
				wide = false;
                                PC_INC(1);
			}
                        SP_MOVE(insn);
                        local_valid = true;
			break;

		case IINC:
#ifdef OPTIMIZE_LOOPS
                        idx = wide ? WORD(pc+1) : INSN(pc+1);
                        STATS_local_set( SO_iterator, blk->stats.locals, idx );
#endif
                        PC_INC(INSN_LEN(insn));
			if (wide == true) {
				wide = false;
                                PC_INC(2);
			}
			break;

		case GETSTATIC:
                        COUNTER_accumulate( cfg, scan );
			getField((constIndex)WORD(pc+1), class, &finfo);
                        COUNTER_accumulate( cfg, lookup );
                        STATS_field_set( SO_load, &blk->stats.fields, finfo.field, finfo.class );
			if (finfo.field == 0) {
				SP_PUSH(1);
			}
			else if (FIELD_ISREF(finfo.field)) {
				SP_PUSH(1);
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]){
			case 'I': case 'Z': case 'S': case 'B':	case 'C':
			case 'F':
				SP_PUSH(1);
				break;
			case 'J':
			case 'D':
				SP_PUSH(2);
				break;
			default:
				ABORT();
				break;
			}
			PC_INC(3);
			break;

		case PUTSTATIC:
                        COUNTER_accumulate( cfg, scan );
			getField((constIndex)WORD(pc+1), class, &finfo);
                        COUNTER_accumulate( cfg, lookup );
                        STATS_field_set( SO_store, &blk->stats.fields, finfo.field, finfo.class );
			if (finfo.field == 0) {
				SP_POP(1);
			}
			else if (FIELD_ISREF(finfo.field)) {
				SP_POP(1);
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]){
			case 'I': case 'Z': case 'S': case 'B':	case 'C':
			case 'F':
				SP_POP(1);
				break;
			case 'J':
			case 'D':
				SP_POP(2);
				break;
			default:
				ABORT();
				break;
			}
			PC_INC(3);
			break;

		case GETFIELD:
                        COUNTER_accumulate( cfg, scan );
			getField((constIndex)WORD(pc+1), class, &finfo);
                        COUNTER_accumulate( cfg, lookup );
                        STATS_field_set( SO_load, &blk->stats.fields, finfo.field, finfo.class );
                        if( local_valid ){
                                STATS_local_set( SO_null_checked, blk->stats.locals, local_ix );
                        }
			if (finfo.field == 0) {
			}
			else if (FIELD_ISREF(finfo.field)) {
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]){
			case 'I': case 'Z': case 'S': case 'B':	case 'C':
			case 'F':
				break;
			case 'J':
			case 'D':
				SP_PUSH(1);
				break;
			default:
				ABORT();
				break;
			}
			PC_INC(3);
			break;

		case PUTFIELD:
                        COUNTER_accumulate( cfg, scan );
			getField((constIndex)WORD(pc+1), class, &finfo);
                        COUNTER_accumulate( cfg, lookup );
                        STATS_field_set( SO_store, &blk->stats.fields, finfo.field, finfo.class );
                        if( local_valid ){
                                STATS_local_set( SO_null_checked, blk->stats.locals, local_ix );
                        }
			if (finfo.field == 0) {
				SP_POP(2);
			}
			else if (FIELD_ISREF(finfo.field)) {
				SP_POP(2);
			}
			else switch (FIELD_TYPE_CNAME(finfo.field)[0]){
			case 'I': case 'Z': case 'S': case 'B':	case 'C':
			case 'F':
				SP_POP(2);
				break;
			case 'J':
			case 'D':
				SP_POP(3);
				break;
			default:
				ABORT();
				break;
			}
			PC_INC(3);
			break;

		case INVOKEVIRTUAL:
		case INVOKESPECIAL:
                        /* FIX: Need to check inline conditions */
                        blk->flags |= BLK_has_java_call;
                        COUNTER_accumulate( cfg, scan );
			if (getMethodSignatureClass((constIndex)WORD(pc+1), class, (insn == INVOKEVIRTUAL ? 0 : 1), &call) == 0) {
				/* We can tolerate some errors here */
				if (!postedExceptionIs("java.lang.NoClassDefFoundError") && !postedExceptionIs("java.lang.NoSuchMethodError")) {
					FAILED();
				}
				clearPostedException();
			}
                        COUNTER_accumulate( cfg, lookup );
			SP_POP(call.in+1);
			SP_PUSH(call.out);
			PC_INC(3);
			break;

		case INVOKEINTERFACE:
                        /* FIX: Need to check inline conditions */
                        blk->flags |= BLK_has_java_call;
                        COUNTER_accumulate( cfg, scan );
			if (getMethodSignatureClass((constIndex)WORD(pc+1), class, 2, &call) == 0) {
				/* We can tolerate some errors here */
				if (!postedExceptionIs("java.lang.NoClassDefFoundError") && !postedExceptionIs("java.lang.NoSuchMethodError")) {
					FAILED();
				}
				clearPostedException();
			}
                        COUNTER_accumulate( cfg, lookup );
			SP_POP(call.in+1);
			SP_PUSH(call.out);
			PC_INC(5);
			break;

		case INVOKESTATIC:
                        /* FIX: Need to check inline conditions */
                        blk->flags |= BLK_has_java_call;
                        COUNTER_accumulate( cfg, scan );
			if (getMethodSignatureClass((constIndex)WORD(pc+1), class, 0, &call) == 0) {
				/* We can tolerate some errors here */
				if (!postedExceptionIs("java.lang.NoClassDefFoundError") && !postedExceptionIs("java.lang.NoSuchMethodError")) {
					FAILED();
				}
				clearPostedException();
			}
                        COUNTER_accumulate( cfg, lookup );
			SP_POP(call.in);
			SP_PUSH(call.out);
			PC_INC(3);
			break;

		case MULTIANEWARRAY:
			SP_POP(INSN(pc+3) - 1);
			PC_INC(4);
			break;

#ifdef OPTIMIZE_LOOPS
                case IALOAD: case LALOAD: case FALOAD: case DALOAD:
                case AALOAD: case BALOAD: case CALOAD: case SALOAD:
                        STATS_array_set( SO_load, &blk->stats.fields, classIx[insn-IALOAD] );
			PC_INC(INSN_LEN(insn));
                        SP_MOVE(insn);
                        break;
                case IASTORE: case LASTORE: case FASTORE: case DASTORE:
                case AASTORE: case BASTORE: case CASTORE: case SASTORE:
                        STATS_array_set( SO_store, &blk->stats.fields, classIx[insn-IASTORE] );
			PC_INC(INSN_LEN(insn));
                        SP_MOVE(insn);
                        break;
#endif
		default:
#ifdef OPTIMIZE_LOOPS
                        if( insn >= ILOAD_0 && insn <= ALOAD_3 ){
                                local_ix = localIx[insn-ILOAD_0];
                                STATS_local_set( SO_load, blk->stats.locals, local_ix );
                        }
                        else if( insn >= ISTORE_0 && insn <= ASTORE_3 ){
                                local_ix = localIx[insn-ISTORE_0];
                                STATS_local_set( SO_store, blk->stats.locals, local_ix );
                        }
#endif
			PC_INC(INSN_LEN(insn));
                        SP_MOVE(insn);
                        local_valid = true;
			break;
		}
	}
        blk->sp.out = sp;
        if( blk->block.next != NULL && !rescan ){
                SCAN_recursive( blk->block.next, pc, sp, false );
        }
}

#ifdef OPTIMIZE_LOOPS
#if 1
static void
SCAN_add_succ( instr_block * cur_b, loop_info * loop_i, block_list * scanned, block_list * stack, boolean loop_add )
{
        int i;
        instr_block * succ_b;

        for( i = cur_b->succ.length; --i >= 0; ){
                succ_b = cur_b->succ.blocks[i];
                if( loop_add && succ_b->bid <= cur_b->bid ){ /* A bit unsafe */
                        continue;
                }
                if( BL_contains( &loop_i->blocks, succ_b ) ){
                        if( !BL_contains( scanned, succ_b )
                            && !BL_contains( stack, succ_b ) ){
                                BL_add( stack, succ_b );
TRACE(                          kprintf( "ADD: B%d\n", succ_b->bid ); );
                        }
                }
                else{
                        if( !BL_contains( &loop_i->exits, cur_b ) ){
                                BL_add( &loop_i->exits, cur_b );
                        }
                }
        }
        for( i = cur_b->catch.length; --i >= 0; ){
                succ_b = cur_b->catch.blocks[i];
                if( loop_add && succ_b->bid <= cur_b->bid ){ /* A bit unsafe */
                        continue;
                }
                if( BL_contains( &loop_i->blocks, succ_b ) ){
                        if( !BL_contains( scanned, succ_b )
                            && !BL_contains( stack, succ_b ) ){
                                BL_add( stack, succ_b );
TRACE(                          kprintf( "ADD: B%d\n", succ_b->bid ); );
                        }
                }
                else{
                        if( !BL_contains( &loop_i->exits, cur_b ) ){
                                BL_add( &loop_i->exits, cur_b );
                        }
                }
        }
}

static void
LOOP_find_loads( loop_info * loop_i, cfg_info * cfg, int localsz )
{
        stats_info * si;
        block_list scanned, stack;
        instr_block * cur_b, * entry_b, * pred_b;
        int i, j;
        boolean scanned_pred, first, progress, can_hack;

        entry_b = loop_i->header->succ.blocks[0];
        bzero( &scanned, sizeof(block_list) );
        bzero( &stack, sizeof(block_list) );
        BL_add( &scanned, entry_b );
        si = entry_b->stats.locals;
        BV_copy( si->item[SO_loop_must_load], si->item[SO_null_checked], si->words );
        SCAN_add_succ( entry_b, loop_i, &scanned, &stack, false );
TRACE(  kprintf( "LOOP_STATS: B%d:\n", loop_i->header->bid ); );
        can_hack = false;
        while( stack.length != 0 ){
                progress = false;
                for( i = stack.length; --i >= 0; ){
                        cur_b = stack.blocks[i];
                        if( cur_b->flags & BLK_loop_entry ){
                                progress = true;
                                BL_del( &stack, cur_b );
TRACE(                          kprintf( "DONE: B%d\n", cur_b->bid ); );
                                BL_add( &scanned, cur_b );
                                si = cur_b->stats.locals;
                                for( j = cur_b->pred.length; --j >= 0; ){
                                        pred_b = cur_b->pred.blocks[j];
                                        if( pred_b->flags & BLK_loop_header ){
                                                BV_copy( si->item[SO_loop_must_load], pred_b->stats.locals->item[SO_null_checked], si->words );
                                                break;
                                        }
                                }
                                BV_or( si->item[SO_loop_must_load], si->item[SO_loop_must_load], si->item[SO_null_checked], si->words );
                                SCAN_add_succ( cur_b, loop_i, &scanned, &stack, true );
                                i = stack.length;
                        }
                        else{
                                scanned_pred = true;
                                for( j = cur_b->pred.length; --j >= 0; ){
                                        pred_b = cur_b->pred.blocks[j];
                                        assert( BL_contains( &loop_i->blocks, pred_b ) );
                                        if( !BL_contains( &scanned, pred_b ) ){
                                                scanned_pred = false;
                                                break;
                                        }
                                }
                                if( can_hack ){
                                        scanned_pred = true;
                                        can_hack = false;
                                        kprintf( "HACKED SCAN!\n" );
                                }
                                if( scanned_pred ){
                                        progress = true;
                                        BL_del( &stack, cur_b );
TRACE(                                  kprintf( "DONE: B%d\n", cur_b->bid ); );
                                        BL_add( &scanned, cur_b );
                                        si = cur_b->stats.locals;
                                        first = true;
                                        for( j = cur_b->pred.length; --j >= 0; ){
                                                pred_b = cur_b->pred.blocks[j];
                                                if( !BL_contains( &scanned, pred_b ) ){
                                                        continue;
                                                }
                                                if( first ){
                                                        BV_copy( si->item[SO_loop_must_load], pred_b->stats.locals->item[SO_loop_must_load], si->words );
                                                        first = false;
                                                }
                                                else{
                                                        BV_and( si->item[SO_loop_must_load], si->item[SO_loop_must_load], pred_b->stats.locals->item[SO_loop_must_load], si->words );
                                                }
                                        }
                                        BV_or( si->item[SO_loop_must_load], si->item[SO_loop_must_load], si->item[SO_null_checked], si->words );
                                        SCAN_add_succ( cur_b, loop_i, &scanned, &stack, false );
                                        i = stack.length;
                                }
                        }
                }
                if( !progress ){
                        can_hack = true;
                }
        }
        si = entry_b->stats.locals;
        first = true;
        for( j = entry_b->pred.length; --j >= 0; ){
                pred_b = entry_b->pred.blocks[j];
                if( BL_contains( &loop_i->blocks, pred_b ) ){
                        if( first ){
                                BV_copy( si->item[SO_loop_must_load], pred_b->stats.locals->item[SO_loop_must_load], si->words );
                                first = false;
                        }
                        else{
                                BV_and( si->item[SO_loop_must_load], si->item[SO_loop_must_load], pred_b->stats.locals->item[SO_loop_must_load], si->words );
                        }
                }
        }
        BV_or( si->item[SO_loop_must_load], si->item[SO_loop_must_load], si->item[SO_null_checked], si->words );
        si = loop_i->locals;
        first = true;
        for( i = loop_i->exits.length; --i >= 0; ){
                cur_b = loop_i->exits.blocks[i];
                if( first ){
                        BV_copy( si->item[SO_loop_must_load], cur_b->stats.locals->item[SO_loop_must_load], si->words );
                        first = false;
                }
                else{
                        BV_and( si->item[SO_loop_must_load], si->item[SO_loop_must_load], cur_b->stats.locals->item[SO_loop_must_load], si->words );
                }
        }
        gc_free_fixed( scanned.blocks );
        gc_free_fixed( stack.blocks );
}

static void
LOOP_find_iterators( loop_info * loop_i, cfg_info * cfg )
{
        instr_block * pred_b, * in_b, * cur_b;
        stats_info * si;
        int n;

        in_b = loop_i->in;
        if( in_b->pred.length == 2 ){
                pred_b = in_b->pred.blocks[0]->flags & BLK_loop_header ? in_b->pred.blocks[1] : in_b->pred.blocks[0];
        }
        else{
                pred_b = NULL;
                kprintf( "FAIL\n" );
                // FAILED();
                return;
        }
        si = loop_i->locals;
        for( n = loop_i->blocks.length; --n >= 0; ){
                cur_b = loop_i->blocks.blocks[n];
                if( BV_is_set( pred_b->dom, cur_b->bid ) ){
                        BV_or( si->item[SO_iterator], si->item[SO_iterator], cur_b->stats.locals->item[SO_iterator], si->words );
                }
        }
#if 0
        for( n = si->words; --n >= 0; ){
                if( si->item[SO_iterator][n] != 0 ){
                        kprintf( "ITERATOR\n" );
                }
        }
#endif
}
#endif

static instr_block *
LOOP_set_singular_header( loop_info * loop_i )
{
        int n;
        instr_block * in_b, * new_b, * tmp_b, * min_b;
        boolean inserted;

        in_b = loop_i->blocks.blocks[0];
        inserted = false;
LOG(    fprintf( in_b->cfg->fd, "SCAN_set_new_loop_header(): B%d\n", in_b->bid ); );

        new_b = BLK_new( in_b->cfg, NULL, NULL );
        new_b->flags |= BLK_inserted_header;
        for( n = in_b->pred.length; --n >= 0; ){
                tmp_b = in_b->pred.blocks[n];
                if( !BL_contains( &loop_i->blocks, tmp_b ) ){
                        if( BL_contains( &loop_i->blocks, tmp_b->block.next ) ){
                                DLL_INSERT_AS_NEXT( tmp_b, new_b, instr_block *, block );
                                tmp_b->flags |= BLK_has_next;
                                new_b->loop = tmp_b->loop;
                                new_b->pc.in = new_b->pc.out = tmp_b->pc.out;
                                inserted = true;
                        }
                        ARC_del( tmp_b, in_b, ARC_normal );
                        ARC_add( tmp_b, new_b, ARC_normal );
                }
        }
        if( !inserted ){
                min_b = loop_i->blocks.blocks[0];
                for( n = loop_i->blocks.length; --n >= 0; ){
                        tmp_b = loop_i->blocks.blocks[n];
                        if( tmp_b->bid < min_b->bid ){
                                min_b = tmp_b;
                        }
                }
                new_b->loop = loop_i->header->loop;
                new_b->pc.in = new_b->pc.out = ((instr_block*)min_b->block.prev)->pc.in;
                DLL_INSERT_AS_PREV( min_b, new_b, instr_block *, block );
        }
        ARC_add( new_b, in_b, ARC_normal );
        in_b->loop->header = new_b;
        return new_b;
}

static void
LOOP_add_block( loop_info * loop_i, instr_block * blk, int localsz )
{
        BL_add( &loop_i->blocks, blk );
        /* Find iterators NOW */
        STATS_locals_coalesce( loop_i->locals, blk->stats.locals );
        STATS_fields_coalesce( &loop_i->fields, &blk->stats.fields );
        if( blk->flags & BLK_has_java_call ){
                loop_i->flags |= LOOP_has_java_call;
        }
        if( blk->loop != NULL ){
                if( BL_contains( &blk->loop->blocks, loop_i->blocks.blocks[0] ) ){
                        blk->loop = loop_i;
                }
        }
        else{
                blk->loop = loop_i;
        }
}

static boolean
SCAN_blocks_for_loops( cfg_info * cfg )
{
        instr_block * cur_b, * tmp_b, * header_b, * entry_b;
        block_list loops, stack;
        int n, m, localsz;
        loop_info * loop_i;
        boolean set_singular_header, redo;

        CFG_dominators( cfg );
        redo = false;
        cur_b = cfg->env.method.root;
        localsz = cfg->env.method.n_locals;
        bzero( &loops, sizeof(block_list) );

        /* From Muchnick, Compiler Design & Implemenation, sec. 7.4 */
        /* Find loops */
        while( cur_b != NULL ){
                for( n = cur_b->succ.length; --n >= 0; ){
                        tmp_b = cur_b->succ.blocks[n];
                        if( BV_is_set( cur_b->dom, tmp_b->bid ) 
                           && !BL_contains( &loops, tmp_b ) ){
                                BL_add( &loops, tmp_b );
                        }
                }
                cur_b = cur_b->block.next;
        }
        /* Find blocks in loops */
        for( n = loops.length; --n >= 0; ){
                cur_b = loops.blocks[n];
                bzero( &stack, sizeof(block_list) );
                set_singular_header = false;
                loop_i = gc_malloc_fixed( sizeof(loop_info) );
                loop_i->locals = STATS_new( localsz );
                loop_i->in = cur_b;

                LOOP_add_block( loop_i, cur_b, localsz );
                if( cur_b->succ.length == 1 ){
                        loop_i->flags |= LOOP_infinite;
                }
                loop_i->cond = cur_b;
TRACE( 		kprintf( "LOOP: { B%d: ", cur_b->bid ); );
                for( m = cur_b->pred.length; --m >= 0; ){
                        tmp_b = cur_b->pred.blocks[m];
                        if( tmp_b->dom != NULL && BV_is_set( tmp_b->dom, cur_b->bid ) ){
                                if( tmp_b != cur_b ){
                                        LOOP_add_block( loop_i, tmp_b, localsz );
                                        BL_add( &stack, tmp_b );
TRACE(                                  kprintf( "B%d. ", tmp_b->bid ); );
                                }
                        }
                        else{
                                if( loop_i->header != NULL
                                    || tmp_b->succ.length != 1 ){
                                        set_singular_header = true;
                                }
                                loop_i->header = tmp_b;
TRACE( 	                        kprintf( "HB%d. ", tmp_b->bid ); );
                        }
                }
                while( stack.length > 0 ){
                        stack.length--;
                        cur_b = stack.blocks[stack.length];
                        for( m = cur_b->pred.length; --m >= 0; ){
                                tmp_b = cur_b->pred.blocks[m];
                                if( !BL_contains( &loop_i->blocks, tmp_b ) ){
                                        LOOP_add_block( loop_i, tmp_b, localsz );
                                        BL_add( &stack, tmp_b );
TRACE(                                  kprintf( "B%d, ", tmp_b->bid ); );
                                }
                        }
                }
                gc_free_fixed( stack.blocks );
                if( set_singular_header ){
                        loop_i->header = LOOP_set_singular_header( loop_i );
                        CFG_order_block_ids( cfg );
                        for( ; n < loops.length; n++ ){
                                cur_b = loops.blocks[n];
                                loop_i = cur_b->loop;
                                for( m = loop_i->blocks.length; --m >= 0; ){
                                        loop_i->blocks.blocks[m]->loop = NULL;
                                }
                                header_b = loop_i->header;
                                header_b->flags &= ~BLK_loop_header;
                                header_b->succ.blocks[0]->flags &= ~BLK_loop_entry;
                                STATS_free( loop_i->locals );
                                STATS_field_free( &loop_i->fields );
                                gc_free_fixed( loop_i->blocks.blocks );
//                               gc_free_fixed( loop_i );
                        }
                       redo = true;
TRACE(                 kprintf( " KILLED }\n" ); );
                       break;
                }
                header_b = loop_i->header;
                header_b->flags |= BLK_loop_header;
                entry_b = header_b->succ.blocks[0];
                entry_b->flags |= BLK_loop_entry;
                if( header_b->block.next == entry_b ){
                        loop_i->flags |= LOOP_do_while;
TRACE(                 kprintf( " DO-WHILE " ); );
                }
TRACE(          kprintf( "}\n" ); );
        }
        if( !redo ){
                for( n = loops.length; --n >= 0; ){
                        loop_i = loops.blocks[n]->loop;
                        LOOP_find_loads( loop_i, cfg, localsz );
                        LOOP_find_iterators( loop_i, cfg );
                        for( m = 0; loop_i != NULL && loop_i->header != NULL ; m++ ){
                                loop_i = loop_i->header->loop;
                        }
                        if( m > cfg->env.global.max_loop_depth ){
                                cfg->env.global.max_loop_depth = m;
                        }
                }
        }
        gc_free_fixed( loops.blocks );
        return redo;
}
#endif

static void
SCAN_exception_handler( block_list * scanned_bl, stats_info * locals, instr_block * blk, int localsz )
{
        instr_block * succ_b;
        int n;

        STATS_locals_coalesce( locals, blk->stats.locals );
        for( n = blk->succ.length; --n >= 0; ){
                succ_b = blk->succ.blocks[n];
                if( !BL_contains( scanned_bl, succ_b ) ){
                        BL_add( scanned_bl, succ_b );
                        SCAN_exception_handler( scanned_bl, locals, succ_b, localsz );
                }
        }
        for( n = blk->catch.length; --n >= 0; ){
                succ_b = blk->catch.blocks[n];
                if( !BL_contains( scanned_bl, succ_b ) ){
                        BL_add( scanned_bl, succ_b );
                        SCAN_exception_handler( scanned_bl, locals, succ_b, localsz );
                }
        }
}

/*
 * Quick scanner: identify basic blocks, calculate stack pointer
 */
void
SCAN_bytecodes( cfg_info * cfg )
{
	int32 n, pc, localsz;
        instr_block * blk;
        block_list * bl, scanned_bl;
        Method * method;
        jexceptionEntry * eptr;

        method = cfg->env.method.method;
        SCAN_recursive( cfg->env.method.root, 0, 0, false );
        if( method->exception_table != NULL ){
                if( cfg->env.method.inline_depth != 0 ){
                        kprintf( "CFG_build(): nested handlers\n" );
LOG(                    fprintf( cfg->fd, "CFG_build(): nested handlers\n" ); );
                }
                for( eptr = &method->exception_table[0]; eptr->start_pc != (uintp)-1; eptr++ ){
                        CFG_exception( cfg, eptr );
                }
        }
        if( method->exception_table != 0 ){
                for( eptr = &method->exception_table[0]; eptr->start_pc != (uintp)-1; eptr++ ){
                        pc = eptr->handler_pc;
                        SCAN_recursive( CFG_get_block( cfg, pc, true), pc, 1, false );
                }
        }

        bl = &cfg->env.method.rescan.blocks;
        for( n = bl->length; --n >= 0; ){
                blk = bl->blocks[n];
                blk->stage = 0;
LOG(            fprintf( cfg->fd, "CFG_build(): rescan\n" ); );
                bzero( blk->stats.locals->bv, NUM_STATS * WORD_SZ * blk->stats.locals->words );
                SCAN_recursive( blk, blk->pc.in, blk->sp.in, true );
        }
        bl->length = 0;

        CFG_order_block_ids( cfg );

#ifdef OPTIMIZE_LOOPS
        if( cfg->env.method.has_loops ){
                while( SCAN_blocks_for_loops( cfg ) ){}
        }
#endif
        for( n = cfg->catch.length; --n >= 0; ){
                blk = cfg->catch.blocks[n];
                bzero( &scanned_bl, sizeof(block_list) );
                localsz = ADJUST_LOCALSZ( method->bcode->localsz );
                blk->handler_locals = STATS_new( localsz );
                BL_add( &scanned_bl, blk );
                SCAN_exception_handler( &scanned_bl, blk->handler_locals, blk, localsz );
                gc_free_fixed( scanned_bl.blocks );
        }
}


