/*
 * schedule.h
 *
 * Copyright (c) 1996, 1997
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __schedule_h
#define __schedule_h

#define SCH_INSERT( _list, _i ) \
        if( _list != NULL ){ \
                DLL_INSERT_AS_PREV( _list, _i, instr_info *, instr ); \
        } \
        _list = _i

#define SCH_REMOVE( _list, _i ) \
	if( _list == _i ){ \
		_list = (_i)->instr.next; \
	} \
	DLL_REMOVE( _i, instr_info *, instr )

#define SCH_2DST 	1
#ifdef  SCH_2DST
#define N_SCH_ARGS	4
#else
#define N_SCH_ARGS	3
#endif
#ifdef OPTIMIZE_SCHEDULE
#define SCH_define_vars \
	cfg_info * cfg; \
	instr_info ** regs, * new_i, * a_i; \
	instr_list * roots

#define SCH_init_vars \
	cfg = blk->cfg; \
	regs = cfg->env.global.schedule.regs; \
	roots = &cfg->env.global.schedule.roots; \
	new_i = a_i = NULL

#define SCH_assign_reg( _r, _i, _ix ) \
	a_i = new_i->a[_ix] = regs[_r]; \
	a_i->flags |= SCH_overwritten; \
	a_i->uses++; \
	regs[_r] = _i

#define SCH_new_( _delay ) \
	new_i = SCH_new(); \
	new_i->addr = blk->native.end; \
	new_i->delay.node = (_delay)

#define SCH_0a_base( _delay ) \
	SCH_new_( _delay ); \
	SCH_print_instr( new_i, blk )

#define SCH_1a_base( _delay, _a ) \
	SCH_new_( _delay ); \
	a_i = new_i->a[0] = regs[_a]; \
	a_i->uses++; \
	SCH_print_instr( new_i, blk )

#define SCH_2a_base( _delay, _a1, _a2 ) \
	SCH_new_( _delay ); \
	a_i = new_i->a[0] = regs[_a1]; \
	a_i->uses++; \
	a_i = new_i->a[1] = regs[_a2]; \
	a_i->uses++; \
	SCH_print_instr( new_i, blk )

#define SCH_0d_0a( _delay ) \
	SCH_0a_base( _delay ); \
	IL_add( roots, new_i )
#define SCH_1d_0a( _delay, _d ) \
	SCH_0a_base( _delay ); \
	SCH_assign_reg( _d, new_i, 2 )
#ifdef  SCH_2DST
#define SCH_2d_0a( _delay, _d1, _d2 ) \
	SCH_0a_base( _delay ); \
	SCH_assign_reg( _d1, new_i, 3 ); \
	SCH_assign_reg( _d2, new_i, 2 )
#endif

#define SCH_0d_1a( _delay, _a ) \
	SCH_1a_base( _delay, _a ); \
	IL_add( roots, new_i )
#define SCH_1d_1a( _delay, _d, _a ) \
	SCH_1a_base( _delay, _a ); \
	SCH_assign_reg( _d, new_i, 2 )
#ifdef  SCH_2DST
#define SCH_2d_1a( _delay, _d1, _d2, _a ) \
	SCH_1a_base( _delay, _a ); \
	SCH_assign_reg( _d1, new_i, 3 ); \
	SCH_assign_reg( _d2, new_i, 2 )
#endif

#define SCH_0d_2a( _delay, _a1, _a2 ) \
	SCH_2a_base( _delay, _a1, _a2 ); \
	IL_add( roots, new_i )
#define SCH_1d_2a( _delay, _d, _a1, _a2 ) \
	SCH_2a_base( _delay, _a1, _a2 ); \
	SCH_assign_reg( _d, new_i, 2 )
#ifdef  SCH_2DST
#define SCH_2d_2a( _delay, _d1, _d2, _a1, _a2 ) \
	SCH_2a_base( _delay, _a1, _a2 ); \
	SCH_assign_reg( _d1, new_i, 3 ); \
	SCH_assign_reg( _d2, new_i, 2 )
#endif

#define SCH_set_last_branch( _e ) \
	cfg->env.global.schedule.last_branch = new_i; \
	cfg->env.global.schedule.branch_expr = (_e); \
	new_i->flags |= SCH_last_branch
#define SCH_set_last_jump() \
	SCH_set_last_branch( NULL )
#define SCH_set_last_call( _l ) \
	cfg->env.global.schedule.last_branch = new_i; \
	cfg->env.global.schedule.call_label = (_l); \
	new_i->flags |= SCH_last_branch | SCH_last_call
#define SCH_end_block( _blk ) SCH_end_block_( _blk )
#define SCH_init( _cfg ) SCH_init_( _cfg )
#define SCH_start_block( _blk ) (_blk)->cfg->env.global.schedule.code.start = (_blk)->native.end;
#else
#define SCH_define_vars \
	cfg_info * cfg
#define SCH_init_vars \
	cfg = blk->cfg
#define SCH_0d_0a( _delay )
#define SCH_1d_0a( _delay, _d )
#define SCH_2d_0a( _delay, _d1, _d2 )
#define SCH_0d_1a( _delay, _a )
#define SCH_1d_1a( _delay, _d, _a )
#define SCH_2d_1a( _delay, _d1, _d2, _a )
#define SCH_0d_2a( _delay, _a1, _a2 )
#define SCH_1d_2a( _delay, _d, _a1, _a2 )
#define SCH_2d_2a( _delay, _d1, _d2, _a1, _a2 )

#define SCH_set_last_branch( _e )
#define SCH_set_last_jump()
#define SCH_set_last_call( _l )
#define SCH_end_block( _blk )
#define SCH_init( _cfg )
#define SCH_start_block( _blk )
#endif

#define SCH_overwritten		0x1	/* was rd of an instruction */
#define SCH_base_reg		0x2	/* incoming register */
#define SCH_sorted		0x4
#define SCH_generated		0x8	/* was scheduled */
#define SCH_last_branch		0x10
#define SCH_last_call		0x40
#ifdef  SCH_2DST
#define SCH_2d_inserted		0x20
#endif

typedef struct _instr_info {
        uint16 			iid; /* for debugging */
        uint16 			flags;
        uint16 			uses;
        struct {
                uint16 		total;
                uint16 		node;
        } delay;
        uint8 *			addr;
        struct _instr_info *	a[N_SCH_ARGS];
        linked_list		instr;
} instr_info;

#define SCH_print_instr( _i, _blk )
extern void SCH_print_instr_( FILE *, instr_info *, instr_block *, boolean );
extern void SCH_end_block_( instr_block * );

extern void SCH_init_( cfg_info * );
extern void SCH_reset( cfg_info * );

#ifdef OPTIMIZE_MEMORY
#define SCH_new() 	gc_malloc_fixed( sizeof(instr_info) )
#define SCH_free( _i )  gc_free_fixed( _i )
/* SCH_free() WILL BREAK HERE IF A FIX IS NOT MADE */
#else
extern void SCH_free( instr_info * );
extern instr_info * SCH_new();
#endif


#endif
