/*
 * throw.c
 *
 * Copyright (c) 1999 - 2001 
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "env.h"

void
TRW_done( cfg_info * cfg )
{
        throw_info * t, * next;

        for( t = cfg->env.global.throw.list; t != NULL; t = next ){
                next = t->next;
                gc_free_fixed( t );
        }
}

void
TRW_generate_all( cfg_info * cfg )
{
        throw_info * t;

LOG(    fprintf( cfg->fd, "TRW_generate_all():\n" ); );
        for( t = cfg->env.global.throw.list; t != NULL; t = t->next ){
                assert( cfg->env.global.native.code.end + t->length < cfg->env.global.native.max );
LOG(            fprintf( cfg->fd, "\t\tthrow: %db @ %p\n", t->length, cfg->env.global.native.code.end ) );
                bcopy( t->buffer, cfg->env.global.native.code.end, t->length );
                t->where = cfg->env.global.native.code.end;
                cfg->env.global.native.code.end += t->length;
        }
}

static void
TRW_assign( throw_info * t, instr_block * blk, label_info * l )
{
        uint8 * from;

        from = LAB_get_from_addr( l );
LOG(    fprintf( blk->cfg->fd, "TRW_assign(): buffer %p - %p\n", 
                 from + 2*WORD_SZ, LAB_get_to_addr( l ) ); );
        blk->native.end = from + 2*WORD_SZ;
        ASM_switch_branch( from );
        l->to.throw = t;
}

throw_info *
TRW_new( cfg_info * cfg, instr_block * blk, label_info * l )
{
        throw_info * t;
        uint8 * start, * end;

        start = LAB_get_from_addr( l ) + 2*WORD_SZ;
        end = LAB_get_to_addr( l );
        for( t = cfg->env.global.throw.list; t != NULL; t = t->next ){
                if( !bcmp( t->buffer, start, end-start) ){
                        TRW_assign( t, blk, l );
                        return t;
                }
        }
        t = gc_malloc_fixed( sizeof(throw_info) + end-start );
        t->length = end-start;
        bcopy( start, t->buffer, t->length );
        SLL_PUSH( cfg->env.global.throw.list, t, next );
        TRW_assign( t, blk, l );
        return t;
}
