/*
 * jmeasuredef.h - this is the (variable) jmeasure configuration header
 * 
 * NOTE: this is automatically included from jmeasure.h
 * 
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

/*********************************************************************
 * define the log buffer (parameter has to be a const)
 * '0' disables logging
 */
MSR_DEFINE_LOG(1000);


/*********************************************************************
 * these are the state vars used for conditional channel measurement
 */
MSR_DEFINE_STATE( in_jit, 0);
MSR_DEFINE_STATE( in_cls_load, 0);


/*********************************************************************
 * This is the list of measure channels which can be specified in
 * the environment var MEASURE. Wrap it in MSR_BEGIN_DEFINE_CHANNELS,
 * MSR_END_DEFINE_CHANNELS macros
 */

MSR_BEGIN_DEFINE_CHANNELS

MSR_DEFINE_TIMER( total_time, "total run time");
MSR_DEFINE_TIMER( jit_time, "time spent in jit translation");
MSR_DEFINE_COUNTER( bcode_mem, "bytes", "total bytecode allocated");
MSR_DEFINE_COUNTER( ncode_mem, "bytes", "total native code allocated");

MSR_DEFINE_COUNTER( gc_alloc, "times", "gc_malloc calls");
MSR_DEFINE_COUNTER( gc_alloc_mem, "bytes", "gc allocated memory");
MSR_DEFINE_TIMER( gc_alloc_time, "gc allocation time");

MSR_DEFINE_COUNTER( gc_free, "times", "gc_free calls");
MSR_DEFINE_COUNTER( gc_free_mem, "bytes", "gc freed memory");
MSR_DEFINE_TIMER( gc_free_time, "gc deallocation time");

MSR_DEFINE_COUNTER( gc_finalize, "times", "object finalize calls");
MSR_DEFINE_TIMER( gc_finalize_time, "gc finalize time");

MSR_DEFINE_TIMER( gc_writeref, "gc writeref time");

MSR_DEFINE_TIMER( gc_sweep_time, "gc sweep time");

MSR_DEFINE_COUNTER( gc_jit_alloc, "times", "gc_malloc calls during JIT compilation");
MSR_DEFINE_COUNTER( gc_jit_free, "times", "gc_free during JIT compilation");
MSR_DEFINE_COUNTER( gc_jit_alloc_mem, "bytes", "mem allocated during JIT compilation");
MSR_DEFINE_COUNTER( gc_jit_free_mem, "bytes", "mem deallocated during JIT compilation");

MSR_DEFINE_COUNTER( gc_cls_alloc, "allocations", "number of allocations during class load");
MSR_DEFINE_COUNTER( gc_cls_free, "deallocations", "number of deallocations during class load");
MSR_DEFINE_COUNTER( gc_cls_alloc_mem, "bytes", "mem allocated during class load");
MSR_DEFINE_COUNTER( gc_cls_free_mem, "bytes", "mem deallocation during class load");

MSR_DEFINE_TIMER( code_analyse, "code analyser");

MSR_DEFINE_COUNTER( lock, "times", "number of locks");
MSR_DEFINE_COUNTER( lock_slow, "times", "number of slow locks");

MSR_DEFINE_COUNTER( thd_query, "times", "number of current thread queries");
MSR_DEFINE_COUNTER( thd_start, "times", "number of thread starts");
MSR_DEFINE_COUNTER( thd_exit, "times", "number of thread exits");

MSR_DEFINE_COUNTER( intrp_insn, "instructions", "number of executed bytecode instructions");
MSR_DEFINE_COUNTER( intrp_new, "allocations", "number of 'new' calls");
MSR_DEFINE_COUNTER( intrp_newa, "allocations", "number of arrays");
MSR_DEFINE_COUNTER( intrp_anewa, "allocations", "number of reference arrays");
MSR_DEFINE_COUNTER( intrp_new_mem, "bytes", "allocated new memory");
MSR_DEFINE_COUNTER( intrp_newa_mem, "bytes", "allocated newa memory");
MSR_DEFINE_COUNTER( intrp_anewa_mem, "bytes", "allocated anewa memory");

MSR_DEFINE_COUNTER( newObj, "allocations", "number of Object allocations");
MSR_DEFINE_COUNTER( newCls, "allocations", "number of Class allocations");
MSR_DEFINE_COUNTER( newArr, "allocations", "number of Array allocations");
MSR_DEFINE_COUNTER( newObj_mem, "bytes", "allocated Objects");
MSR_DEFINE_COUNTER( newCls_mem, "bytes", "allocated Classes");
MSR_DEFINE_COUNTER( newArr_mem, "bytes", "allocated Arrays");

MSR_DEFINE_COUNTER( cls_load, "times", "class loads");
MSR_DEFINE_TIMER( cls_load_time, "class load time");
MSR_DEFINE_TIMER( cls_proc_time, "class process time");

MSR_END_DEFINE_CHANNELS

/*********************************************************************
 * This is the measure init func, which is called via the
 * MEASURE_INIT() macro (just from the module that defines MEASURE_MAIN).
 */

#if defined(MEASURE_SETUP)

void
msr_setup()
{
  /* timers to start automatically */
  MSR_START_GLOBAL( total_time);
}

void
msr_finish()
{
  MSR_STOP_GLOBAL( total_time);
  
  MSR_FINISH();
}

#endif /* MEASURE_SETUP */
