/*
 * locks.h
 * Manage locking system.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __locks_h
#define __locks_h

#include "md.h"

struct _iLock;

#define	LOCKOBJECT			struct _iLock**
#define	lockMutex(O)			_lockMutex(&(O)->lock, &iLockRoot)
#define	unlockMutex(O)			_unlockMutex(&(O)->lock, &iLockRoot)
#define	waitCond(O,T)			_waitCond(&(O)->lock, (T))
#define	signalCond(O)			_signalCond(&(O)->lock)
#define	broadcastCond(O)		_broadcastCond(&(O)->lock)
#define	holdMutex(O)			_holdMutex(&(O)->lock)

#define	lockStaticMutex(THING)		_lockMutex((THING), &iLockRoot)
#define	unlockStaticMutex(THING)	_unlockMutex((THING), &iLockRoot)
#define	waitStaticCond(THING, TIME)	_waitCond((THING), (TIME))
#define	signalStaticCond(THING)		_signalCond((THING))
#define	broadcastStaticCond(THING)	_broadcastCond((THING))

struct Hjava_lang_Thread;
struct Hjava_lang_Object;

typedef struct _iLock {
	void*				holder; 
	struct Hjava_lang_Thread*	mux;
	struct Hjava_lang_Thread*	cv;
} iLock;

#define	LOCKINPROGRESS	((iLock*)-1)
#define	LOCKFREE	((iLock*)0)

JNIEXPORT void	_lockMutex(LOCKOBJECT, void*);
JNIEXPORT void	_unlockMutex(LOCKOBJECT, void*);
JNIEXPORT jboolean _waitCond(LOCKOBJECT, jlong);
JNIEXPORT void	_signalCond(LOCKOBJECT);
JNIEXPORT void	_broadcastCond(LOCKOBJECT);
JNIEXPORT void	lockObject(struct Hjava_lang_Object*);
JNIEXPORT void	unlockObject(struct Hjava_lang_Object*);
JNIEXPORT void* _releaseLock(iLock**);
JNIEXPORT void 	_acquireLock(iLock**, void*);

/*
 * Unblock an object and call a function. When the function returns reclaim
 * the object.
 */
#define	UNBLOCK_LOCK_EXECUTE(LK, FUNC) { \
		void* st = _releaseLock(LK); \
		FUNC; \
		_acquireLock((LK), st); \
	}
#define	UNBLOCK_EXECUTE(ENV, OBJ, FUNC) \
	UNBLOCK_LOCK_EXECUTE(&((Hjava_lang_Object*)(OBJ))->lock, FUNC)

/*
 * Used to convert POSIX locks to semaphores.
 */
typedef struct {
	void*	mux;
	void*	cv;
	int	count;

	/* 'thd' can be used to link a sem2posixLock to a certain native
	 * thread (e.g. in pthreads, to store stack boundaries)
	 */
	void* thd;
} sem2posixLock;

/*
 * Lock interface.  Either define the lock/unlock/wait/signal/broadcast
 * set of operations (for POSIX style locking) or else define the
 * semget/semput operations for semaphore style locking.
 */
typedef struct LockInterface {

	void	  (*lock)(sem2posixLock*);
	void	  (*unlock)(sem2posixLock*);
	jboolean  (*wait)(sem2posixLock*, jlong);
	void	  (*signal)(sem2posixLock*);

	jboolean  (*semget)(void*, jlong);
	void	  (*semput)(void*);

} LockInterface;

EXTERN_C LockInterface Kaffe_LockInterface;


/* these should be really turned over to inlines, here */
#define Kaffe_LI_lock(_s2pLock)     (*Kaffe_LockInterface.lock)(_s2pLock)
#define Kaffe_LI_unlock(_s2pLock)   (*Kaffe_LockInterface.unlock)(_s2pLock)
#define Kaffe_LI_wait(_s2pLock,_to) (*Kaffe_LockInterface.wait)(_s2pLock,_to)
#define Kaffe_LI_signal(_s2pLock)   (*Kaffe_LockInterface.signal)(_s2pLock)

/* map these to our generic get/put functions (used to map sem ops to
 * posix mutexes/convars */
#define Kaffe_LI_semget(_sem,_to)   _SemGet(_sem,_to)
#define Kaffe_LI_semput(_sem)       _SemPut(_sem)


/*
 * Inject the LockInterface implementation header, which can be used
 * to turn the runtime interface into a compile-time interface (by means
 * of overloading generic macros)
 */
#include "lock-impl.h" 

#endif
