/* 
 * collect.h
 * The incremental garbage collector.
 *
 * Copyright (c) 1999 - 2001
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __collect_h
#define __collect_h

#define GC_LOCK()		(*Kaffe_LockInterface.spinon)(0);
#define	GC_UNLOCK()		(*Kaffe_LockInterface.spinoff)(0);

#define STATUS_MASK		0x1
#define STATUS_FINAL		1

#define COLOR_MASK		0x2
#if defined(GC_EXP)
#define COLOR_BLACK		black_val
#define COLOR_WHITE		white_val
#else
#define COLOR_BLACK		0
#define COLOR_WHITE		2
#endif

#if defined(GC_INCREMENTAL)
#if defined(GC_EXP)
#define GET_NEXTGREY(_o) ((_o)->prev)
#define REMOVE_OBJ(_o) \
	if ((_o)->next != 0) (_o)->next->prev = (_o)->prev; \
	if ((_o)->prev != 0) (_o)->prev->next = (_o)->next

#define INSERT_AS_NEXT_OBJ(_list,_o) \
	(_o)->prev = (_list); \
	if ((_list)->next != 0) (_list)->next->prev = (_o); \
	(_o)->next = (_list)->next; \
	(_list)->next = (_o)

#define INSERT_AS_PREV_OBJ(_list,_o) \
	(_o)->next = (_list); \
	if ((_list)->prev != 0) (_list)->prev->next = (_o); \
	(_o)->prev = (_list)->prev; \
	(_list)->prev = (_o)

#define INSERT_GREYLIST(_o) \
        MARK_COLOR(_o, COLOR_BLACK); \
	gc_grey_cnt++; \
        (_o)->prev = grey_list; \
        grey_list = (_o)

#define INSERT_OBJ(_list, _o) \
	INSERT_AS_NEXT_OBJ(&(_list), (_o))

#define GET_COLOR(_o) ((_o)->color)
#define GET_NEXT(_o) ((_o)->next)

#define SET_COLOR(_o, _c) \
        (_o)->color = (_c)
#define MARK_COLOR(_o, _c) \
        (_o)->color = ((_o)->color & STATUS_MASK) | (_c)
#define IS_COLOR(_o, _c) \
        (((_o)->color & ~STATUS_MASK) == (_c))

#define MARK_FINAL(_o) \
        (_o)->color = ((_o)->color & ~STATUS_MASK) | STATUS_FINAL
#define IS_FINAL(_o) \
        (((_o)->color & STATUS_MASK) == STATUS_FINAL)
#else
#define GET_NEXTGREY(_o) ((_o)->u.nextgrey)
#define REMOVE_OBJ(_o)

#define INSERT_GREYLIST(_o) \
	gc_grey_cnt++; \
        (_o)->u.nextgrey = grey_list; \
        grey_list = (_o)

#define INSERT_OBJ(_list, _o) \
        (heap_obj*)(_o)->next = (_list); \
        (_list) = (_o)

#define GET_COLOR(_o) ((_o)->u.color)
#define GET_NEXT(_o) ((_o)->next)

#define SET_COLOR(_o, _c) \
        (_o)->u.color = (_c)
#define IS_COLOR(_o, _c) \
        (((_o)->u.color & COLOR_MASK) == (_c))

#define MARK_FINAL(_o) \
        (_o)->u.color = ((_o)->u.color & ~STATUS_MASK) | STATUS_FINAL
#define IS_FINAL(_o) \
        (((_o)->u.color & STATUS_MASK) == STATUS_FINAL)
#endif
#else

#define GET_COLOR(_o) ((_o)->u.color)
#define GET_NEXT(_o) ((_o)->next)
#define GET_NEXTGREY(_o) ((_o)->u.nextgrey)
#define REMOVE_OBJ(_o) 

#define SET_COLOR(_o, _c) \
        (_o)->u.color = (_c)
#define IS_COLOR(_o, _c) \
        (((_o)->u.color & ~STATUS_MASK) == (_c))

#define MARK_FINAL(_o) \
        (_o)->u.color = ((_o)->u.color & ~STATUS_MASK) | STATUS_FINAL
#define IS_FINAL(_o) \
        (((_o)->u.color & STATUS_MASK) == STATUS_FINAL)

#define INSERT_GREYLIST(_o) \
        (_o)->u.nextgrey = grey_list; \
        grey_list = (_o)

#define INSERT_OBJ(_list, _o) \
        (heap_obj*)(_o)->next = (_list); \
        (_list) = (_o)
#endif

#define NOTIFY_COLLECTOR() \
	if (garbageman && \
		(gc_current_alloc > (int32)gc_heap_allocation_size)) { \
		unlockStaticMutex(&gc_lock); \
		invokeGC(0); \
		lockStaticMutex(&gc_lock); \
	}

#if defined(GC_EXP)
extern unsigned int black_val;
extern unsigned int white_val;
extern heap_obj black_list;
extern heap_obj white_list;
extern heap_obj free_list;
extern heap_obj finalize_list;
extern heap_obj finalize_class_list;
#else
extern struct _heap_obj* white_list;
extern struct _heap_obj* finalize_list;
extern struct _heap_obj* finalize_class_list;
#endif
extern struct _heap_obj* grey_list;
extern int urgentState;
extern int gc_grey_cnt;
extern int in_verify;

#endif
