/* 
 * gc-funcs.c
 *
 * Copyright (c) 1999 - 2001
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "locks.h"
#include "gc.h"
#include "gc-env.h"
#include "access.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "threads.h"
#include "exception.h" 
#include "java_lang_ref_Reference.h"
#include "support.h"
#include "strings.h"
#include "bag.h"

// #define DBG(_t, _m) printf _m ; fflush(stdout)
#define DBG(_t, _m) 
// #define CDBG(s) kprintf s;
#define	CDBG(s)
// #define FDBG(s)	kprintf s;
#define FDBG(s)
#define walkNull 0	/* walkNull() does nothing */

extern void Kaffe_CallVoidMethodA(JNIEnv*, jobject, jmethodID, jvalue*);
extern void Kaffe_ExceptionClear(JNIEnv*);
extern Hjava_lang_Class* ArrayInterfaces[];

gcTable gc_table[] = {
/* GC_ALLOC_NORMAL */		{ walkConservative, 	0 },
/* GC_ALLOC_ESTACK */		{ walkNull,	    	0 },
/* GC_ALLOC_FIXED */		{ walkNull,    		0 },
/* GC_ALLOC_NORMALOBJECT */	{ walkObject,		finalizeObject   },
/* GC_ALLOC_PRIMARRAY */	{ walkNull,	    	0 },
/* GC_ALLOC_REFARRAY */		{ walkRefArray,     	0 },
/* GC_ALLOC_CLASSOBJECT */	{ walkClass,	    	finalizeClass    },
/* GC_ALLOC_THREAD */		{ walkThread,		finalizeObject   },
/* GC_ALLOC_STATICDATA */	{ walkConservative, 	0 },
/* GC_ALLOC_BAG */		{ walkBag,		0 },
/* GC_ALLOC_JREFERENCE */	{ walkJReference,   	finalizeObject   },
/* GC_ALLOC_CLASSLOADER */	{ walkConservative,  	0 },
/* GC_ALLOC_NORMALOBJECT_NOFINALIZE */		{ walkObject,		0 },
/* GC_ALLOC_STRING */		{ walkObject,		finalizeString }
};

extern iLock classHashLock;
Hjava_lang_ref_Reference* refQueue;
Hjava_lang_ref_Reference* refDispatchQueue;

#if defined(DEBUG) || defined (GC_STATS)
char* gcType (gcTable* gctable) {
        int idx = ((void*)gctable - (void*)gc_table) / sizeof(gcTable);
        switch (idx) {
        case 0: return "Nor";
        case 1: return "Stk";
        case 2: return "Fix";
        case 3: return "Obj";
        case 4: return "PAr";
        case 5: return "RAr";
        case 6: return "Cls";
        case 7: return "Thd";
        case 8: return "Sta";
        case 9: return "CPl";
        case 10: return "Ref";
        case 11: return "Ldr";
        case 12: return "NFn";
        }
        return "???";
}
#endif

void finalizeObject(void* base)
{
        Method* final = OBJECT_CLASS((Hjava_lang_Object*)base)->finalizer;

FDBG( 	( "fObj %x %s\n", base, CLASS_CNAME( OBJECT_CLASS( (Hjava_lang_Object*)base ) ) ) );

	if (final) {
                /* Use JNI to make finalizer call to protect GC against exceptions */
                Kaffe_CallVoidMethodA(0, (jobject)base, (jmethodID)final, 0);
                Kaffe_ExceptionClear(0);
        }
};

void finalizeString(void* base)
{
        Hjava_lang_String * str = (Hjava_lang_String*)base;

FDBG( 	( "fStr %x %s\n", base, CLASS_CNAME( OBJECT_CLASS( str ) ) ) );

        uninternJavaString(str);
};

void finalizeClass(void* base)
{
	Hjava_lang_Class* class = (Hjava_lang_Class*)base;
        Method* meth;
        int ix;

FDBG(	("fCls %x %s\n", class, CLASS_CNAME(class) ));

	if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0) {
		kprintf("<GC: Unloading %s>\n", CLASS_CNAME(class));
	}

	/* Unload class */
	unloadClass(class);

        /*  Free fixed objects associated with the class */
        gc_free_fixed(class->constants.data);

	meth = CLASS_METHODS(class);
	for (ix = CLASS_NMETHODS(class); --ix >= 0; meth++) {
		/* gc_free_fixed(meth->sig); - this might be shared */
                gc_free_fixed(meth->ncode);
                gc_free_fixed(meth->lines);
                gc_free_fixed(meth->exception_table);
        }
        gc_free_fixed(class->methods);

        gc_free_fixed(class->fields);
	if (class->interfaces != ArrayInterfaces) {
		gc_free_fixed(class->interfaces);
	}
        gc_free_fixed(class->refs);
        memset(class,0,sizeof(Hjava_lang_Class));
};

void
markObject(void* base)
{
	/*
	 * Check that the memory 'mem' is in fact the
	 * beginning of an object.  If not we just return.
	 */
        page_blk* header = MEM2PAGE_HDR(base);
        heap_obj* obj;

        obj = MEM2HEAP_OBJ(base);
        if (base < (void*)gc_heap_min || base >= (void*)gc_heap_max ||
            ((uint32)obj - (uint32)header) < sizeof(page_blk) ||
            ((uint)obj & 0x3)) return;
        if (IS_FIXED_PAGE(header)) return;
        if (IS_COLOR(obj, COLOR_BLACK) ){
CDBG(	("Blac %x\n", base ));
        }
        if (IS_COLOR(obj, COLOR_WHITE) ){
CDBG(	("Whit %x\n", base ));
        }
        if (IS_FINAL(obj) ){
CDBG(	("iFin %x\n", base ));
        }
        if (IS_COLOR(obj, COLOR_WHITE) && !IS_FINAL(obj)) {
                if (IS_BLOB_PAGE(header)) {
                        if (MEM2HEAP_BLOB(base)->magic != GC_MAGIC) return;
                } else {
                        if (obj->gctable == 0) return;
                        if ( (uint32)((void*)obj - ((void*)header + sizeof(page_blk))) % 
                             (PAGE_OBJSIZE(header) + sizeof(heap_obj)) )
                                return;
                }
                if (header->magic != GC_MAGIC) return;

                REMOVE_OBJ(obj);
                INSERT_GREYLIST(obj);
CDBG(	("Mark %x\n", base ));
        }
}

void
walkMemory( void * mem )
{
        _walkMemory( MEM2HEAP_OBJ( mem ) );
}

void
_walkMemory( heap_obj * obj )
{
        void * base;
        page_blk * header;
        walkFunc walk;
        uint32 size;

        base = HEAP_OBJ2MEM(obj);
        header = MEM2PAGE_HDR(base);
        if (base < (void*)gc_heap_min || base >= (void*)gc_heap_max ||
            ((uint32)obj - (uint32)header) < sizeof(page_blk) ) return;
        size = IS_BLOB_PAGE(header) ? 
                MEM2HEAP_BLOB(base)->size :
                PAGE_OBJSIZE(header);

CDBG(   ("wMem %x\n",base));

        
        walk = obj->gctable->walk;
        if( walk != NULL ){
                (*walk)(base,size);
        }
}

void
walkConservative(void* base, uint32 size) 
{
        int8 * addr;
        void * mem;

CDBG(   ("wCon %x (%d)\n",base,size));
        if (!size) return;

        for (addr = ((int8*)base) + (size & -ALIGNMENTOF_VOIDP) - sizeof(void*); (void*)addr >= base; addr -= ALIGNMENTOF_VOIDP) {
                mem = *(void**)addr;
                if( mem != NULL ){
CDBG(   ("     -> %x\n", mem ) );
                        markObject( mem );
                }
        }
};

void 
walkObject(void* base, uint32 size) 
{
#if 1
        int n;
	unsigned int* refs;
        Hjava_lang_Class* cls;
	void* mem;

        cls = OBJECT_CLASS((Hjava_lang_Object*)base);
CDBG(   ("wObj %x %s\n", base, CLASS_CNAME(cls) ) );

	markObject( ((Hjava_lang_Object*)base)->dtable->class);

        n = cls->nrefs;
        refs = cls->refs;
        while (--n >= 0) {
                mem = *(void**)((uintp)base+refs[n]);
                if (mem != 0) {
CDBG(   ("     -> %x\n", mem ) );
                        markObject(mem);
                }
        }
#else
	walkConservative(base, size);
#endif
}

void
walkThread(void* base, uint32 size) 
{
CDBG(   ("wThd %x (%d)\n",base,size));

	walkConservative(base, size);
        Kaffe_TI_GcWalkThread( (Hjava_lang_Thread*)base );
};

#ifndef walkNull
void
walkNull(void* base, uint32 size)
{
CDBG(   ("wNul %x\n",base));
	markObject(((Hjava_lang_Object*)base)->dtable);
}
#endif

void 
walkClass(void* base, uint32 size) 
{
	Hjava_lang_Class* class;
        struct _constants* pool;
	int ix;
	Method* meth;
	Field* fld;
#if defined(GC_SAFE)
        jexceptionEntry* eentry;
#endif

	class = (Hjava_lang_Class*)base;
CDBG(   ( "wCls %x %s\n", base, CLASS_CNAME(class) ) );
        if( class->loader != 0 ) {
                kprintf( "LOADER: %x\n" );
        }

        /* If this class has no dtable, its in the proces of being created
         * and doesn't yet contains anything interesting.
         */
        if (class->head.dtable == 0) {
                return;
        } 

        markObject( CLASS_CNAME(class) );
        /* Mark strings referenced in the constant pool */
	pool = CLASS_CONSTANTS (class);
        if (pool->tags)
                for (ix = pool->size; --ix >= 0 ;) {
#if !defined(GC_SAFE)
                        char type = pool->tags[ix];
                        if (type == CONSTANT_Utf8 ||
                            type == CONSTANT_String ||
                            type == CONSTANT_ResolvedString ||
                            type == CONSTANT_Class ||
                            type == CONSTANT_ResolvedClass)
#endif
                        markObject((void*)pool->data[ix]);
                }

#if defined(GC_SAFE)
        /* We might want to mark the superclass... */
	if (class->sdata && class->superclass) {
		markObject(class->superclass);
        }
#endif

        /* Handle methods */
	meth = CLASS_METHODS(class);
	for (ix = CLASS_NMETHODS(class); --ix >= 0; meth++) {
		markObject(meth->name);
		markObject(meth->signature);
#if defined(GC_SAFE)
                /* We might want to mark exception handler classes... */
                if (meth->exception_table) {
			for (eentry = meth->exception_table; eentry->start_pc != (uintp)-1; eentry++) {
                                markObject((void*)eentry->catch);
			}
		}
#endif
	}

        /* Handle fields */
	fld = CLASS_FIELDS(class);
	for (ix = CLASS_NFIELDS(class); --ix >= 0; fld++) {
		markObject(fld->name);

#if defined(GC_SAFE)
                /* We might want to mark field type classes... */
                markObject(fld->ftype);
#endif
	}

#if defined(GC_SAFE)
        /* We might want to mark the class loader... */
        markObject(class->loader);

        /* We might want to mark the class interfaces... */
	for (ix = class->total_interface_len; --ix >= 0;) {
                markObject(class->interfaces[ix]);
        }
#endif
	if (!CLASS_IS_PRIMITIVE(class)) {
		markObject(class->dtable);
        }

	/* Mark the static data */
	markObject(class->sdata);
}

void 
walkRefArray(void* base, uint32 size) 
{
	Hjava_lang_Object* arr;
	int i;
	Hjava_lang_Object** ptr;

	arr = (Hjava_lang_Object*)base;
CDBG(   ("wRef %x %s (%d elem)\n", base, arr->dtable->class->name, ARRAY_SIZE(arr)));
	markObject(arr->dtable->class);

	ptr = ARRAY_DATA( HArrayOfObject, arr );
	for (i = ARRAY_SIZE(arr); --i>= 0; ) {
		Hjava_lang_Object* el = *ptr++;
		if (el != NULL) {
CDBG(   ("     -> %x\n", el ) );
			markObject(el);
		}
	}
}

void 
walkJReference(void* base, uint32 size) 
{
	Hjava_lang_ref_Reference* obj;

CDBG(   ("wJR  %x (%d)\n",base,size));
	obj = (Hjava_lang_ref_Reference*)base;

	markObject(obj->base.dtable->class);

	/* Queue reference for later examination by GC */
	if (unhand(obj)->obj != 0) {
		unhand(obj)->next = refQueue;
		refQueue = obj;
	}

	/* Only mark the queue, not the object referenced */
	markObject(unhand(obj)->que);

	/* Walk the rest of the object */
	walkConservative(obj+1, size-sizeof(Hjava_lang_ref_Reference));
}

void 
walkBag(void* base, uint32 size) 
{
        bag * bg;

CDBG(   ("wBag %x (%d)\n",base,size));
	bg = (bag*)base;
	walkConservative(bg->keys, (2 * sizeof(void*) + sizeof(int*)) * bg->size);
}

void
markReferences(void)
{
	Hjava_lang_ref_Reference* ref;
	Hjava_lang_ref_Reference* nref;
        heap_obj* obj;

	for (ref = refQueue; ref != 0; ref = nref) {
		nref = unhand(ref)->next;

		if (unhand(ref)->obj == 0) {
			continue;
		}

		obj = MEM2HEAP_OBJ(unhand(ref)->obj);

		/* If object isn't white then it's not interesting */
                if (!IS_COLOR(obj, COLOR_WHITE)) {
			continue;
                }

		/* Soft references are dispatched before finalization but only
		 * when memory is low.
		 */
		if (unhand(ref)->type == java_lang_ref_Reference_SOFT) {
			if (urgentState != 0) {
				unhand(ref)->obj = 0;
			}
			else {
				markObject(unhand(ref)->obj);
				continue;
			}
		}
		/* Weak references are dispatched before finalization */
		else if (unhand(ref)->type == java_lang_ref_Reference_WEAK) {
			unhand(ref)->obj = 0;
		}
		/* Recycle references are dispatch and make the object
		 * reachable again.
		 */
		else if (unhand(ref)->type == java_lang_ref_Reference_RECYCLE) {
			markObject(unhand(ref)->obj);
		}
		/* Phantom references are dispatched after finalization */
		else if (unhand(ref)->type == java_lang_ref_Reference_PHANTOM) {
                        if (IS_FINAL(obj)) {
				unhand(ref)->obj = 0;
			}
			else {
				continue;
			}
		}
		/* Unknown reference type */
		else {
			continue;
		}

		/* Queue reference for dispatch */
		unhand(ref)->next = refDispatchQueue;
		refDispatchQueue = ref;
	}
}
