/* 
 * page.h
 * The heap manager.
 *
 * Copyright (c) 1999 - 2001
 *	Stanford University.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __page_h
#define	__page_h

typedef struct _base_obj {
	struct _base_obj*	next;
#if defined(GC_64B)
        uint32			pad;
#endif
} base_obj;

typedef struct _heap_obj {
	struct _heap_obj*	next;
#if defined(GC_EXP)
	struct _heap_obj*	prev;
	uint32			color;
#else
#if defined(GC_64B)
        uint32			pad;
#endif
        union {
		struct _heap_obj*	nextgrey;
		uint32			color;
	} u;
#endif
        gcTable*		gctable;
} heap_obj;

typedef struct _base_blob {
	uint32			size;
        uint32 			magic;
	struct _base_blob*	next;
#if defined(GC_64B)
        uint32			pad;
#endif
} base_blob;

typedef struct _heap_blob {
	uint32			size;
        uint32 			magic;
	struct _heap_blob*	next;
#if defined(GC_EXP)
	struct _heap_obj*	prev;
	uint32			color;
#else
#if defined(GC_64B)
        uint32			pad;
#endif
        union {
		struct _heap_obj*	nextgrey;
		uint32			color;
	} u;
#endif
        gcTable*		gctable;
} heap_blob;

#if defined(GC_64B)
#define	CHECK_ALIGNMENT( _m )	assert( ((uint32)(_m) & 0x7) == 0 )
#else
#define	CHECK_ALIGNMENT( _m )	
#endif

#define OBJ2MEM(_m) 		((void*)(_m) + sizeof(base_obj))
#define BLOB2MEM(_m) 		((void*)(_m) + sizeof(base_blob))
#define HEAP_OBJ2MEM(_m) 	((void*)(_m) + sizeof(heap_obj))
#define HEAP_BLOB2MEM(_m) 	((void*)(_m) + sizeof(heap_blob))

#define MEM2BLOB(_m)		((base_blob*)((_m) - sizeof(base_blob)))
#define MEM2OBJ(_m)		((base_obj*)((_m) - sizeof(base_obj)))
#define MEM2HEAP_BLOB(_m)	((heap_blob*)((_m) - sizeof(heap_blob)))
#define MEM2HEAP_OBJ(_m)	((heap_obj*)((_m) - sizeof(heap_obj)))
#define MEM2PAGE_HDR(_m) \
	(page_blk*)((uintp)(_m) & ~(page_size-1))

#define PAGE_OBJSIZE_MASK	0x1fff
#define PAGE_BLOB_FLAG		0x2000
#define PAGE_FIXED_FLAG		0x4000
#define PAGE_ALLOC_FLAG		0x8000

#define PAGE_BLK_OFFSET(_p,_n) \
	(page_blk*)((void*)(_p) + page_size * (_n))

#define IS_ALLOCATED_PAGE(_p)	((_p)->info & PAGE_ALLOC_FLAG)
#define SET_ALLOCATED_PAGE(_p)	(_p)->info |= PAGE_ALLOC_FLAG

#define SET_PAGE_MAGIC(_p)	(_p)->magic = GC_MAGIC

#define IS_FIXED_PAGE(_p)	((_p)->info & PAGE_FIXED_FLAG)
#define SET_FIXED_PAGE(_p)	(_p)->info |= PAGE_FIXED_FLAG

#define IS_BLOB_PAGE(_p)	((_p)->info & PAGE_BLOB_FLAG)
#define SET_BLOB_PAGE(_p)	(_p)->info |= PAGE_BLOB_FLAG

#define PAGE_OBJSIZE(_p)	((_p)->info & PAGE_OBJSIZE_MASK)
#define SET_PAGE_OBJSIZE(_p,_sz)	(_p)->info |= (PAGE_OBJSIZE_MASK & (_sz))

#define GC_MAGIC		0xD0DECADE

JNIEXPORT uint32 page_size;
JNIEXPORT uint32 collected;
JNIEXPORT int32  gc_current_alloc;

JNIEXPORT size_t gc_heap_allocation_size;
JNIEXPORT int gc_panic_level;

JNIEXPORT uintp gc_heap_min;
JNIEXPORT uintp gc_heap_max;

JNIEXPORT void JNICALL breakpt();
JNIEXPORT void* JNICALL malloc_blob(int, int, struct _gcTable*);
JNIEXPORT void* JNICALL fixed_malloc_obj(struct _page_blk**);
JNIEXPORT void  JNICALL free_freelists(void);

#endif
