/*
 * gc.h
 * This interfaces the VM to the Hans-J. Boehm Incremental Garbage
 * Collector (version 6.0 alpha 7).
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __gc2_h
#define __gc2_h

#define	GC_LINUX_THREADS	1
#define	_REENTRANT		1

#include "boehm/include/gc.h"

#define	GC_ALLOC_NORMAL		0
#define	GC_ALLOC_ESTACK		1
#define	GC_ALLOC_FIXED		2
#define	GC_ALLOC_NORMALOBJECT	3
#define	GC_ALLOC_PRIMARRAY	4
#define	GC_ALLOC_REFARRAY	5
#define	GC_ALLOC_CLASSOBJECT	6
#define	GC_ALLOC_THREAD		7
#define	GC_ALLOC_METHOD		8
#define	GC_ALLOC_FIELD		9
#define	GC_ALLOC_STATICDATA	10
#define	GC_ALLOC_DISPATCHTABLE	11
#define	GC_ALLOC_BYTECODE	12
#define	GC_ALLOC_EXCEPTIONTABLE	13
#define	GC_ALLOC_CONSTANT	14
#define	GC_ALLOC_UTF8CONST	15
#define	GC_ALLOC_INTERFACE	16
#define	GC_ALLOC_JITCODE	17
#define	GC_ALLOC_LOCK		18
#define	GC_ALLOC_THREADCTX	19
#define	GC_ALLOC_REF		20
#define	GC_ALLOC_CLASSPOOL	21
#define	GC_ALLOC_JREFERENCE	22
#define	GC_ALLOC_CLASSLOADER	23
#define	GC_ALLOC_ANALYSE	24
#define	GC_ALLOC_INFLATE	25
#define	GC_ALLOC_STRINGS	26
#define	GC_ALLOC_ENGINE		27
#define	GC_ALLOC_JAR		28
#define	GC_ALLOC_AWT		29
#define	GC_ALLOC_SIGNATURE	30
#define	GC_ALLOC_NORMALOBJECT_NOFINALIZE 31

extern void* gc_realloc_fixed(void*, size_t);
extern void* gc_malloc(size_t, int);
extern void* gc_malloc_fixed(size_t);
extern void* gc_add_ref(void*);
extern void gc_rm_ref(void*);

#define	gc_calloc(A,B,C)		gc_malloc((A) * (B), (C))
#define	gc_malloc_fixed_typed(A,B)	gc_malloc((A), (B))
#define	gc_calloc_fixed(A,B)		gc_malloc_fixed((A) * (B))
#define	gc_calloc_fixed_typed(A,B,C)	gc_calloc((A), (B), (C))
#define	gc_free(A)			GC_FREE(A)
#define	gc_free_fixed(A)		gc_free(A)

#define	invokeGC(F)			GC_gcollect()
#define	invokeFinalizer()		GC_invoke_finalizers()

#define	gc_register(A)

#define	heapTotal()			GC_get_heap_size()
#define	heapFree()			GC_get_free_bytes()

/* ??? */
#define	MIN_HEAPSIZE	0
#define	MAX_HEAPSIZE	0
#define	ALLOC_HEAPSIZE	0

/*
 * We don't need these
 */
#define	GC_WRITE(A,B)
#define	GC_WRITE_SAFE(A,B)

#endif
