/*
 * patchcall.c
 * Patch various places to improve performance
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "classMethod.h"
#include "access.h"
#include "md.h"
#include "patch.h"
#include "/usr/include/strings.h"

#define PATCH_LIST_SIZE( _n ) \
	( sizeof(patch_list) + ( ((_n) - 1) * sizeof(patch_info) ) )

#ifdef DIRECT_CALL
static void
_checkList( patch_list ** l_p )
{
        patch_list * l;

        l = *l_p;
        if( l == NULL ){
                l = *l_p = gc_malloc_fixed( PATCH_LIST_SIZE( 10 ) );
                l->max = 10;
        }
        else if( l->length + 1 >= l->max ){
                l = gc_malloc_fixed( PATCH_LIST_SIZE( (*l_p)->max + 10 ) );
                bcopy( *l_p, l, PATCH_LIST_SIZE( (*l_p)->max ) );
                l->max += 10;
                gc_free_fixed( *l_p );
                *l_p = l;
        }
}

void 
addMethodPatch( Method * m, void * where, int type )
{
        patch_list * l;
        patch_info * p;

	if( METHOD_TRANSLATED( m ) ) {
                return;
        }
        _checkList( &m->patch );
        l = m->patch;
        p = &l->patch[l->length];
        p->where = where;
        assert( type == PS_addr || type == PS_call || type == PS_const );
        p->type = type;
        l->length++;
        // kprintf ("madd: %d %x <-- %x %s %s\n", l->length, p->where, m, m->name, m->signature );
}

void 
addClassPatch( Hjava_lang_Class * c, void * where, int type )
{
        patch_list * l;
        patch_info * p;
        
        _checkList( &c->patch );
        l = c->patch;
        p = &l->patch[l->length];
        p->where = where;
        assert( type == PS_initclass );
        p->type = type;
        l->length++;
        // kprintf( "cadd: %x %s\n", p->where, c->name );
}

#define PATCH_NOP( _p ) \
	*(uint*)(_p)->where = I_NOP

#define PATCH_CONST( _p, _tgt ) \
	*(void**)(_p)->where = (_tgt)

static void 
_finishPatches( patch_list ** l_p, void * tgt )
{
        patch_list * l;
        patch_info * p;
        int i;
        methodTrampoline * tramp;

        l = *l_p;
        if( l == NULL ){
                return;
        }
        // kprintf( "patch: n=%d with %x\n", l->length, tgt );
        for( i = l->length; --i >= 0; ){
                p = &l->patch[i];
                switch( p->type ){
                case PS_call:
                        // kprintf( "cpatch: %x\n", p->where );
                        PATCH_CALL( p, tgt );
#if defined(FLUSH_DCACHE)
                        FLUSH_DCACHE(p->where, p->where+4);
#endif
                        break;
                case PS_addr:
                        /*
                         * This is a dtable entry; We replace with new taget
                         * Force trampoline to point to the new method too
                         */
                        // kprintf( "mpatch: %x\n", p->where );
                        tramp = *(void**)p->where;
                        *(void**)p->where = tgt;
                        PATCH_TRAMPOLINE( tramp, tgt );
#if defined(FLUSH_DCACHE)
                        FLUSH_DCACHE( tramp, tramp+1 );
#endif
                        break;
                case PS_const:
                        // kprintf( "vpatch: %x\n", p->where );
                        PATCH_CONST( p, tgt );
                        break;
                case PS_initclass:
                        /*
                         * This is a _Jv_InitClass call
                         * We replace with a nop to prevent further calls to it
                         */
                        // kprintf( "ipatch: %x\n", p->where );
			PATCH_NOP( p );
#if defined(FLUSH_DCACHE)
                        FLUSH_DCACHE(p->where, p->where+4);
#endif
                        break;
                default:
                        kprintf( "unexpected: %x\n", p->where );
                        assert(0);
                        break;
                }
        }
        gc_free_fixed( l );
        *l_p = NULL;
}

void
finishMethodPatches( Method * m, Hjava_lang_Class * c, void * t )
{
        // kprintf ("mpatch: %x %s.%s %s\n", m, c, m->name, m->signature );
	_finishPatches( &m->patch, t );
}

void
finishClassPatches( Hjava_lang_Class * c )
{
        // kprintf( "cpatch: %s\n", c->name );
	_finishPatches( &c->patch, 0 );
}
#endif
