/*
 * net-impl.c - pthread-based NetworkCallInterface implementation
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-net.h"
#include "gtypes.h"
#include "gc.h"

/* we need this for the interface type */
#define	NOUNIXPROTOTYPES
#include "jnetwork.h"

#include <unistd.h>
#include <netdb.h>


/* pjo, 6 apr 2000 */
typedef int socklen_t ;
#include <sys/ioctl.h>


/***********************************************************************
 * Interface functions
 */

static
void
Ninit (void)
{
  /* nothing to do */
}

static
int
Nsocket ( int stream )
{
  if (stream == 0) {
	return socket(AF_INET, SOCK_DGRAM, 0);
  }
  else {
	return socket(AF_INET, SOCK_STREAM, 0);
  }
}

static
char*
Ngethostname (void)
{
  char* buf = gc_malloc_fixed(64);

  if ( gethostname(buf, 64) == 0 ) {
	return (buf);
  }
  else {
	return ("localhost");
  }
}

static
int*
Ngethostbyname ( char* name )
{
  struct hostent  *ent;
  int             *addrs;

  if ( !(ent = gethostbyname(name)) ){
	return (0);
  }

  addrs = (int*)gc_malloc_fixed(sizeof(int) * 2);
  addrs[0] = *(int*)ent->h_addr;
  addrs[1] = 0;

  return (addrs);
}

static
char*
Ngethostbyaddr ( int addr )
{
  struct hostent  *ent;

  if ( !(ent = gethostbyaddr((char*)&addr, 4, AF_INET)) ){
	return (0);
  }

  return (ent->h_name);
}


/*
 * ? what value are we supposed to return here ?
 */
static
int
NgetReadTimeout ( int fd )
{
  int             stat;
  struct timeval  tv;
  socklen_t       optlen = sizeof(tv);

  if ( (stat = getsockopt( fd, SOL_SOCKET, SO_RCVTIMEO, &tv, &optlen)) ){
	return 0; /* error */
  }

  return (tv.tv_sec * 1000 + tv.tv_usec);
}


/***********************************************************************
 * create and initialize the NetworkCallInterface
 */


NetworkCallInterface Kaffe_NetworkCallInterface = {
  Ninit,
  Nsocket,
  (int(*)(int,struct sockaddr*,int))     bind,
  (int(*)(int,int))                      listen,
  (int(*)(int,struct sockaddr*,size_t))  connect,
  (int(*)(int,struct sockaddr*,size_t*))   accept, 
  read,  
  (ssize_t(*)(int,void*,size_t,int,struct sockaddr*,int*)) recvfrom,
  write, 
  (ssize_t(*)(int,const void*,size_t,int,const struct sockaddr*,int)) sendto, 
  close,
  (int(*)(int,int,int,const void*,int))  setsockopt, 
  (int(*)(int,int,int,void*,int*))       getsockopt,
  (int(*)(int,struct sockaddr*, int*))   getsockname, 
  (int(*)(int,struct sockaddr*,int*))    getpeername,

  Ngethostname,
  Ngethostbyname,
  Ngethostbyaddr,
  (int(*)(int,int,int*))ioctl,

  0,  /* int setreadtimeout(int,int) */
  NgetReadTimeout   /* int getreadtimeout(int) */
};
