/*
 * thread-impl.c - pthread based ThreadInterface implementation
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

//New detach stuff commit not tested


#define PTD 0

// ALWAYS TRUE ????
#ifdef __Lynx__
#define _POSIX_THREADS_CALLS
#endif

#include "config.h"
#include "config-std.h"
#include "config-signal.h"

#include <pthread.h>
#include <semaphore.h>
#include <sched.h>

#include "lerrno.h"

#include "java_lang_Throwable.h"
#include "gtypes.h"
#include "threads.h"
#include "gc.h"
#include "jni.h"
#include "locks.h"

#include "jdebug.h"

/* these are required for handling exceptions */
#include "engine.h"
#include "exception.h"

#include "thread-impl.h"
#include "thread-internal.h"

/* Some systems need special setups for the exception handling */
#if !defined(EXCEPTIONSTART)
#define EXCEPTIONSTART()
#endif
#if !defined(EXCEPTIONEND)
#define EXCEPTIONEND()
#endif

/* for debugging, needed to compile even if DEBUG is not defined */
char stat_act[]   = { ' ', 'a' };
char stat_susp[]  = { ' ', 's', ' ', 'r', ' ', ' ', ' ', ' ', ' ', ' ', ' ' };
char stat_block[] = { ' ', 'T', 'm', ' ', 'c', ' ', ' ', ' ', 't', ' ', ' ' };


/***********************************************************************
 * typedefs & defines
 */

/* check if the following is always true */

#ifndef __Lynx__
#include <bits/local_lim.h>
#endif /* __Lynx__ */

 

/*
 * Now it starts to get hackish - we have to pick some signals
 * for suspend/resume (enter/exitCritSect) which don't interfere
 * with pthread implementations. Note that we can't rely on when
 * a suspend signal is delivered, and it's therefore not safe
 * to mulitplex a sinle signal for both suspend & resume purposes
 */
#if !defined(__SIGRTMIN) || (__SIGRTMAX - __SIGRTMIN < 3)
#define          SIG_SUSPEND   SIGURG
#define          SIG_RESUME    SIGTSTP
#define          SIG_DUMP      SIGXCPU
/* included part on lynx */

/*
 * Sneak these signal in from the thread library.
 */
#define		 PSIG_RESTART	SIGUSR1
#define		 PSIG_CANCEL	SIGUSR2

#else
#define          SIG_SUSPEND   SIGUSR1
#define          SIG_RESUME    SIGUSR2
#define          SIG_DUMP      SIGXCPU

/*
 * Sneak these signal in from the thread library.
 */
#define		 PSIG_RESTART	(__SIGRTMIN)
#define		 PSIG_CANCEL	(__SIGRTMIN+1)

#endif


/***********************************************************************
 * global data
 */
 
/* our upper create limit, to ensure we don't blow the system */
#define MAX_SYS_THREADS     _POSIX_THREAD_THREADS_MAX - 1

/* Stack alignment */
#define	STACK_ALIGN	0x1000

 
 

/* We keep a list of all active threads, so that we can enumerate them */
nativeThread     *activeThreads;

/* The notorious first thread, which has to be handled differently because
 * it isn't created explicitly */
nativeThread     *firstThread;

/* Number of active non-daemon threads (the last terminating nonDaemon
 * causes the process to shut down */
int              nonDaemons;

/* Number of system threads (either running (activeThreads) or
 * blocked (pool). We need this to implement our own barrier, since
 * many kernel thread systems don't behave graceful on exceeding their limit */ 
int              nSysThreads;


int priorities[java_lang_Thread_MAX_PRIORITY+1+1+1]; 

/* thread-specific-data key to retrieve 'nativeData' */
pthread_key_t    ntKey;

/* our lock to protect list manipulation/iteration */
iLock     *tLock; 

/* level of critical sections (0 = none) */
int              critSection;

/* helper semaphore to signal completion of critical section enter/exit */
sem_t            critSem;

/* an optional deadlock watchdog thread (not in the activeThread list),
 * activated by DEBUG topic vm_thread  or vm_thread_deadlock_detect */

pthread_t        deadlockWatchdog;


/* last queried current thread (to optionally speed up currentXX queries) */
nativeThread     *lastCur;


/* tDump support: was a signal received to dump the threading system's state
   since we cannot call it directlty from a sig handler cause it calls mutex-lock and others  
   so watchdog checks for this setting in a loop*/

short pleaseDump = 0;  


#define DEADLOCK_OR_DUMPREQUEST_DETECT_INTERVAL  2    /* how long do we pause before checking for another 
                                                         deadlock or SIG_DUMP signal */
					 
#include "objects.h"
static void tInitLock ( nativeThread*  );
static void tSetLock ( nativeThread*  );

/*Determine scheduling policy */
    				

#if defined(HAVE_REALTIME_FIFO) || defined(HAVE_REALTIME_RR)
        
	
#ifdef 	HAVE_REALTIME_FIFO
	
  unsigned long SchedulePolicy = SCHED_FIFO;
#else
        
  unsigned long	SchedulePolicy = SCHED_RR;
#endif


		 
#else
       
   unsigned long SchedulePolicy = SCHED_OTHER; 
#endif       
	


static void computeStack(nativeThread *);
static void tDispose (nativeThread*);


#define LOCKSLOT  int iLockRoot

#define LOCK(_nt)                \
  _nt->blockState |= BS_THREAD;  \
  lockStaticMutex( &tLock)

#define UNLOCK(_nt)              \
  unlockStaticMutex( &tLock);    \
  _nt->blockState &= ~BS_THREAD
  
  
#define LOCKFORDUMP   lockStaticMutex( &tLock)
#define UNLOCKFORDUMP  unlockStaticMutex( &tLock)  



// errno is not thread-specific so need it as a local var on each stack


/***********************************************************************
 * internal functions
 */

/*
 * get and store the invariant part of a stack address
 */
void
tPrepareStackLookup ( nativeThread* nt )
{
	int i;
	unsigned long a = (unsigned long) nt->stackMin +1;
	unsigned long b = (unsigned long) nt->stackMax -1;
	unsigned long a0 = a;

	for ( i=0; a != b; a>>=1, b>>=1, i++ );

	nt->stackMask = (0xffffffff >> i) << i;
	nt->stackBase = a0 & nt->stackMask;
}


/*
 * On demand debug signal to dump the state of the threading system (requested
 * by a external "kill -s <SIG_DUMP> <proc-id>", where proc-id is ANY thread of the VM (so this
 * excludes pthreads' manager thread)
 */

void
dump_signal_handler ( int sig ) 
{
 /* must set a flag to be checked by watchdog since Linux  COMMENT TO CHANGE  pthreads don't yet
    conform to the standard, since signals are only directed to the PID aimed, not 
    any who does not block. Must also resuspend if was called while suspended,
    since sigsuspend returns when any signal is received */

  pleaseDump = 1;
 
}



/*
 * SIGSEGV handler (mapping system exception to Java exceptions)
 */
void
segv_signal_handler ( EXCEPTIONPROTO ) 
{

  EXCEPTIONSTART();

  EXCEPTIONFIXRETURN();

  postAndThrowException("java/lang/NullPointerException", 0);

  EXCEPTIONEND();
}


/*
 * dump a thread list
 */
void
tDumpList (nativeThread* list)
{
	int             i;
	char            a;
	nativeThread    *t;

	for ( t=list, i=0; t; t=t->next, i++ ){
		/* the first one? */
		a = (t == firstThread) ? '1' : ' ';

		kprintf ("%4d: %c %c%c%c   %p [tid: %4d, java: %p]  "
			 "lock: [mux: %d cv: %d] stack: [%d..%d..%d]\n",  /* put this all ps */
			 i, a,
			 stat_act[t->active], stat_susp[t->suspendState], 
			 stat_block[t->blockState],
			 t, t->tid, t->thread,
			 &t->mux, &t->cv,
			 t->stackMin, t->stackCur, t->stackMax);
		}
	


	
}


/*
 * dump the state of the threading system
 */
void
tDump (void)
{
	DBG_ACTION_OR( vm_thread,
	               vm_thread_deadlock_detect, {
		LOCKSLOT;
		void             *lock;
	        void             *holder = 0; /* take care of warnings */
		void             *mux = 0;
		void             *muxNat = 0;
		void             *cv = 0;
		void             *cvNat = 0;
  
  
  	        lock   = tLock;
		
		if(lock != LOCKFREE){
	        	holder = tLock->holder;
			mux    = tLock->mux;
			muxNat = tLock->mux ? unhand(tLock->mux)->PrivateInfo : 0;
			cv     = tLock->cv;
			cvNat  = tLock->cv ? unhand(tLock->cv)->PrivateInfo : 0;
                }

  
                /* want a snapshot so lock all thread lists */  
       
                LOCKFORDUMP;/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */


		       
		kprintf("======================== thread dump =========================\n");
		
		kprintf("state:  nonDaemons: %d, critSection: %d\n",
						 nonDaemons, critSection);
	        if(lock != LOCKFREE){
	                kprintf ("tLock:       %p [holder: %p, mux: %p (native: %p), cv: %p (native: %p)]\n",
	                                       lock, holder, mux, muxNat, cv, cvNat);
                }
		else{
			kprintf("tLock: currently free\n");
		}
		
		kprintf("active threads:\n");
		
		tDumpList(activeThreads);

		
			
		kprintf ("====================== end thread dump =======================\n");
		
	
		UNLOCKFORDUMP;/* ------------------------------------------------------ tLock */
	
	});
}



static
void*
tWatchdogRun (void* p)
{
      	nativeThread *t;
	int life;
	size_t           ss;
	int errno;

	
	nativeThread* nt = (nativeThread*)p;

	
        /* need the next coming stuff to be able to get locks later */
	
        if((errno = pthread_setspecific( ntKey, nt)) != 0)
		PerrorAbort("pthread_setspecific", errno);


/* get the stack boundaries (no way of knowing if setstacksize succeeded to set
	   exactly what we asked) */
#if defined(_POSIX_THREAD_ATTR_STACKSIZE)
	if((errno = pthread_attr_getstacksize( &nt->attr, &ss)) != 0)
		PerrorAbort("pthread_attr_getstacksize", errno);
	

	
#endif


#if defined(STACK_GROWS_UP)
	nt->stackMin  = (void*) ((uintp)&nt & -STACK_ALIGN);
	nt->stackMax  = (void*) ((uintp) nt->stackMin + ss);
#else
	nt->stackMax  = (void*) (((uintp)&nt + STACK_ALIGN - 1) & -STACK_ALIGN);
	nt->stackMin  = (void*) ((uintp) nt->stackMax - ss);
#endif		
	
	

	 
	tPrepareStackLookup( nt);

	
	
			
	/* main loop */						
							
	while ( nonDaemons ) {
	
	        sleep(DEADLOCK_OR_DUMPREQUEST_DETECT_INTERVAL) ;
		
		
			
DBG_ACTION( vm_thread_deadlock_detect, {

		life = 0;
		
	
		for ( t=activeThreads; t != NULL; t = t->next ){
			/*
			 * if we have a single thread that is not blocked at all, is in a
			 * timeout wait, and is not suspended, we are still safe (even though
			 * the timeout value might effectively be a deadlock). Also, holding
			   a mutex is fine. 
			 */
						
			if ( (!t->blockState || (t->blockState == BS_CV_TO) 
			     || (t->blockState == BS_THREAD)) && !t->suspendState ){
				life = 1;
				break;
			}
		}
		

		if ( !life ) {
			DBG( vm_thread_deadlock_detect, ("DEADLOCK\n"));
			tDump();
			ABORT();
		}
	        
 })		
 
 
				
DBG_ACTION( vm_thread,	{

                /* atomic test and set so miss less signals 
		   This  stuff means:
		   if(pleaseDump) pleaseDump = 0;

		*/
                if(COMPARE_AND_EXCHANGE(&pleaseDump, 1, 0)){
			tDump();
                }
		
	  })	
			  	
	
		 
	} /* while */
	

	return 0;
}



void
tStartDeadlockWatchdog ()
{
 	struct sched_param sp;
	
	nativeThread* nt;
        
	extern Hjava_lang_Class* ThreadClass;
	int errno;

	

	/* must "fake" that this is like another thread so that can get locks later */
	
	
	nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
	assert(nt != 0);

	nt->tid = deadlockWatchdog;
      	
      
      
        nt->thread = (Hjava_lang_Thread*)newObject(ThreadClass);
	assert(nt->thread != 0);

        unhand(nt->thread)->sem = gc_malloc_fixed_typed(sizeof(sem2posixLock), GC_ALLOC_THREADCTX);

        nt->suspendState = 0;

	 
		
	/* init our cv and mux fields for locking */
	tInitLock( nt);
	tSetLock( nt);



        /* end of "fake" stuff */

        sp.sched_priority = priorities[0];  /* looow */

  	        
	if((errno = pthread_attr_init( &nt->attr)) != 0)
		PerrorAbort("pthread_attr_init", errno);
       	
	if((errno = pthread_attr_setschedpolicy(&nt->attr, SchedulePolicy)) !=0)
		PerrorAbort("pthread_attr_setschedpolicy", errno);

        if((errno = pthread_attr_setschedparam(&nt->attr, &sp)) != 0)
		PerrorAbort("pthread_attr_setschedparam", errno);

#if defined(_POSIX_THREAD_ATTR_STACKSIZE)

	if((errno = pthread_attr_setstacksize(&nt->attr, PTHREAD_STACK_MIN)) != 0)
		PerrorAbort("pthread_attr_setstacksize", errno);

#endif

        if((errno = pthread_attr_setdetachstate( &nt->attr, PTHREAD_CREATE_DETACHED)) != 0)
		PerrorAbort("pthread_attr_setdetachstate");


	if((errno = pthread_create( &deadlockWatchdog, &nt->attr, tWatchdogRun, nt)) != 0)
		PerrorAbort("pthread_create", errno);
	
	
	
		
}



/***********************************************************************
 * thread system initialization
 */

/*
 * static init of signal handlers
 */
void
tInitSignalHandlers ()
{
  struct sigaction  sigSegv;
  unsigned int flags = 0;

#if defined(SA_RESTART)
  flags |= SA_RESTART;
#endif


  
DBG_ACTION( vm_thread, {
        
	                struct sigaction sigDump;
			
		        sigDump.sa_flags = flags;
	                sigDump.sa_handler = dump_signal_handler;
	                sigemptyset( &sigDump.sa_mask);
	                if(sigaction( SIG_DUMP, &sigDump, NULL) < 0){
		        	perror("sigaction");
				ABORT();
			}	

	  });



  sigSegv.sa_flags = flags;
#if defined(SA_SIGINFO)
  sigSegv.sa_flags |= SA_SIGINFO;
#endif
#if defined(SA_NOMASK)
  sigSegv.sa_flags |= SA_NOMASK;
#endif
  sigSegv.sa_handler = (void(*)(int)) segv_signal_handler;
  sigemptyset( &sigSegv.sa_mask);
  if(sigaction(SIGSEGV, &sigSegv, NULL) < 0)
  	perror("sigaction"), ABORT();
}



/*
 * static init set up of Java-to-pthread priority mapping (pthread prioritiy levels
 * are implementation dependent)
 */
static
void
tMapPriorities ()
{
	int     d, min, max, n, i;
	float   r;
	

	if((min = sched_get_priority_min( SchedulePolicy)) <0)
		perror("sched_get_priority_min"), ABORT();
	if((max = sched_get_priority_max( SchedulePolicy)) < 0)
		perror("sched_get_priority_max"), ABORT();
		
        d = max - min;
	n = sizeof( priorities) / sizeof(priorities[0]);
	r = (float)d / (float)n;
	

	for ( i=0; i<n; i++ ) {
		priorities[i] = (int)(i*r + 0.5) + 1;//man page says 0 is a bad value
	}


}


/*
 * per-native thread init of condvars, mutexes and semaphores
 */
static
void
tInitLock ( nativeThread* nt )
{
  pthread_condattr_t    cvAttr;
  pthread_mutexattr_t   muxAttr;
  int errno;

  /* init a process private condvar */
  if((errno = pthread_condattr_init( &cvAttr)) != 0)
  	PerrorAbort("pthread_condattr_init", errno);
	
  
#if defined(_POSIX_THREAD_PROCESS_SHARED)
  if((errno = pthread_condattr_setpshared( &cvAttr, PTHREAD_PROCESS_PRIVATE)) != 0)
  	PerrorAbort("pthread_condattr_setpshared", errno);
#endif
  
  if((errno = pthread_cond_init( &nt->cv, &cvAttr))!= 0)
  	PerrorAbort("pthread_cond_init", errno);


  /* init a process private mutex. We deal with priority inversion
   * by means of inheriting the highest priority of any thread who
   * requests the mutex
   */
  if((errno = pthread_mutexattr_init( &muxAttr)) != 0)
  	PerrorAbort("pthread_mutexattr_init", errno);
	
  
#if defined(_POSIX_THREAD_PROCESS_SHARED)
  if((errno = pthread_mutexattr_setpshared( &muxAttr, PTHREAD_PROCESS_PRIVATE)) != 0)
  	PerrorAbort("pthread_mutexattr_setpshared", errno);
#endif


#if defined(_POSIX_THREAD_PRIO_INHERIT)
    if((errno = pthread_mutexattr_setprotocol( &muxAttr, PTHREAD_PRIO_INHERIT)) != 0)
  	PerrorAbort("pthread_mutexattr_setprotocol", errno);
#endif

 
  if((errno = pthread_mutex_init( &nt->mux, &muxAttr)) != 0)
  	PerrorAbort("pthread_mutex_init", errno);

 

  /* init a non-shared (process-exclusive) semaphore with value '0' */
  if(sem_init( &nt->sem, 0, 0) < 0)
  	perror("sem_init"), ABORT();
}

/*
 * init the Java thread locking facilities (sem) from the native thread
 * condvar and mutex
 */
static
void
tSetLock ( nativeThread* nt )
{
  sem2posixLock *lk;

  lk = (sem2posixLock*)unhand(nt->thread)->sem;
  lk->cv  = &nt->cv;
  lk->mux = &nt->mux;
  lk->thd = nt;
}

/*
 * We must have a certain amount of credible thread information setup
 * as soon as possible.
 */
static
void
tSetupFirstNative(void)
{
  nativeThread* nt;
  int errno;

  /*
   * We need to have a native thread context available as soon as possible.
   */
  nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
  nt->tid = pthread_self();
  if((errno = pthread_setspecific( ntKey, nt)) != 0)
  	PerrorAbort("pthread_setspecific", errno); 
  nt->stackMin  = (void*)0;
  nt->stackMax  = (void*)-1;
  nt->stackCur = (void*) 1 ;
  nt->stackBase = 0;
  nt->stackMask = 0;
  nt->thread = (Hjava_lang_Thread*)1;	/* Non-zero dummy */
  lastCur = nt;
}

/*
 * The global, one-time initialization goes here. This is an
 * alternative to scattered pthread_once() calls
 */
void
Tinit(void)
{
  int errno;

  if((errno = pthread_key_create( &ntKey, NULL)) != 0)
  	PerrorAbort("pthread_create", errno);


  if(sem_init( &critSem, 0, 0) < 0)
  	perror("sem_init"), ABORT();

  tMapPriorities();
  tInitSignalHandlers();
  
  
  tSetupFirstNative();
  
   
  DBG( vm_thread, ("initialized\n"));

  
}






/***********************************************************************
 * thread creation
 */

void
Tattach ( Hjava_lang_Thread* thread )
{
  int              oldCancelType, oldCancelState;
  struct sched_param   sp;
  LOCKSLOT;
  int errno;
  nativeThread* nt = GET_CURRENT_THREAD( &nt);

 
  if (firstThread != 0) {
    /* If this thread is already attached then ignore */
    if (nt != 0) {
      return;
    }
    nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
  }
  assert(nt != 0);

  /* Make sure we know who we are.*/
  nt->tid = pthread_self();

  /* we can't use nt->attr, because it wasn't used to create this thread */
  if((errno = pthread_attr_init( &nt->attr)) != 0)
	PerrorAbort("pthread_attr_init", errno);

  nt->thread = thread;
  nt->suspendState = 0;
  nt->func   = NULL;
  nt->next   = NULL;

  /* Get stack boundaries. Note that this is just an approximation
   * which should cover all gc-relevant stack locations
   */

#if defined(STACK_GROWS_UP)
	nt->stackMin  = (void*) ((uintp)&nt & -STACK_ALIGN);
	nt->stackMax  = (void*) ((uintp) nt->stackMin + threadStackSize);
#else
	nt->stackMax  = (void*) (((uintp)&nt + STACK_ALIGN - 1) & -STACK_ALIGN);
	nt->stackMin  = (void*) ((uintp) nt->stackMax - threadStackSize);
#endif

  tPrepareStackLookup( nt);
  lastCur = nt;

  DBG( vm_thread, TMSG_SHORT( "attach ", nt));

  /* init our cv and mux fields for locking */
  tInitLock( nt);
  tSetLock( nt);

  /* Link native and Java thread objects
   * We already are executing in the right thread, so we can set the specific
   * data straight away
   */
  if((errno = pthread_setspecific( ntKey, nt)) != 0)
	PerrorAbort("pthread_setspecific", errno);
 
  unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;
	
  if((errno = pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldCancelState)) != 0)
		PerrorAbort("pthread_setcancelstate", errno);
  if((errno = pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldCancelType)) != 0)
	PerrorAbort("pthread_setcanceltype", errno);

  if((errno = pthread_detach( nt->tid)) != 0)
		PerrorAbort("pthread_detach");

  
  /* Link the new one into the activeThreads list. We lock until
   * the newly created thread is set up correctly (i.e. is walkable)
   */
  LOCK( nt); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

  nt->active = 1;
  nt->next = activeThreads;
  activeThreads = nt;
  
  nonDaemons++;

  
  if (firstThread == 0) {
    firstThread = nt;
    
    DBG_ACTION_OR( vm_thread,
		   vm_thread_deadlock_detect, { tStartDeadlockWatchdog(); });
		
    
  }
   
  nSysThreads++;
  
  sp.sched_priority = priorities[unhand(nt->thread)->priority];
  if((errno = pthread_setschedparam( nt->tid, SchedulePolicy, &sp)) != 0)
	PerrorAbort("pthread_setschedparam", errno);

  

  UNLOCK( nt); /* ---------------------------------------------------- tLock */
}


/*
 * This is our thread function wrapper, which we need because of two
 * reasons. First, there is no way to set thread specific data from
 * outside the current thread (and it should be propperly set before we
 * enter the *real* thread func). Second, this is a better place to handle
 * the recycle looping than Texit (would be strange to loop in there)
 */
static
void*
tRun ( void* p )
{
  LOCKSLOT;
  nativeThread     *cur = (nativeThread*)p;
  nativeThread     *t;
  size_t           ss;
  int              oldCancelType, oldCancelState;
  struct sched_param   sp;
  int errno;
  
      /* get the stack boundaries */
  
#if defined(_POSIX_THREAD_ATTR_STACKSIZE)

  if((errno = pthread_attr_getstacksize(&cur->attr, &ss)) != 0)
	PerrorAbort("pthread_attr_getstacksize", errno);
#endif
    
#if defined(STACK_GROWS_UP)
	cur->stackMin  = (void*) ((uintp)&cur & -STACK_ALIGN);
	cur->stackMax = (void*) ((unsigned long)cur->stackMin + ss);
#else
	cur->stackMax  = (void*)(((uintp)&cur + STACK_ALIGN - 1) & -STACK_ALIGN);
	cur->stackMin = (void*) ((unsigned long)cur->stackMax - ss);
#endif
 
  tPrepareStackLookup( cur);

  if((errno = pthread_setspecific( ntKey, cur)) != 0)
	PerrorAbort("pthread_setspecific", errno);

  if((errno = pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldCancelState)) != 0)
		PerrorAbort("pthread_setcancelstate", errno);

  if((errno = pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldCancelType)) != 0)
	PerrorAbort("pthread_setcanceltype", errno);



  /* We set this here explicity since the creater doesn't always get
   * this set in time.
   */
  cur->tid = pthread_self();/*pthread_self never fails */
 
  /* we are reasonably operational now, flag our creator that it's safe to give
   * up the thread lock */
  if(sem_post( &cur->sem) < 0)
  	perror("sem_post"), ABORT();

  DBG( vm_thread, TMSG_LONG( "calling user func of: ", cur));

  /* Now call our thread function, which happens to be firstStartThread(),
   * which will call TExit before it returns */
  cur->func();
 
  // No cache

  DBG( vm_thread, TMSG_LONG( "exiting user func of: ", cur));

  LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

  /* remove from active list */
  if ( cur == activeThreads ){
  	activeThreads = cur->next;
  }
  else {
	for ( t=activeThreads; t->next && (t->next != cur); t=t->next );
		assert( t->next != 0 );
		t->next = cur->next;
  }

  /* unlink Java and native thread */
  unhand(cur->thread)->PrivateInfo = 0;
  cur->thread = 0;
  cur->suspendState = 0;
  
  nSysThreads--;

  UNLOCK( cur); /* ---------------------------------------------------- tLock */

  tDispose( cur);

  return 0;
}


/*
 * Create a new native thread for a given Java Thread object. Note
 * that we are called from Thread.start(), i.e. we are already allowed to
 * call func() (which happens to be threads.c:firstStartThread)
 *
 * Our call graph looks like this:
 *
 *     Thread.start
 *       startThread
 *         Kaffe_ThreadInterface.create (firstStartThread) == Tcreate
 *           pthread_create(..tRun..)
 *                                                 creator-thread
 *   --------------------------------------------------------------
 *                                                 created-thread
 *     tRun
 *       firstStartThread
 *         Thread.run
 *         exitThread
 *           Thread.finish
 *           Kaffe_ThreadInterface.exit == Texit
 */

void
Tcreate ( Hjava_lang_Thread* thread, void* func )
{
  	LOCKSLOT;
  	nativeThread         *nt;
  	struct sched_param   sp;
	int errno;
        nativeThread* cur = GET_CURRENT_THREAD( &cur); 
  	

  	/* if we are the first one, it's seriously broken */
  	assert( activeThreads != 0 );

  	sp.sched_priority = priorities[unhand(thread)->priority];
  
        LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */


	if ( !unhand(thread)->daemon )
		nonDaemons++;

	UNLOCK( cur); /* ---------------------------------------------------- tLock */


        if ( nSysThreads++ > MAX_SYS_THREADS ){		
		/* fail, we exceeded our physical thread limit*/
		DBG( vm_thread, ( "too many threads (%d)\n", nSysThreads));
		postAndThrowException("java/lang/InternalError", "Attempt to create more threads than the system can handle");
		return;
	}


	nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);

       
	if((errno = pthread_attr_init( &nt->attr)) != 0)
		PerrorAbort("pthread_attr_init", errno);
        
	if((errno = pthread_attr_setschedpolicy( &nt->attr, SchedulePolicy)) != 0)
		PerrorAbort("pthread_attr_setschedpolicy", errno);
			
	if((errno = pthread_attr_setschedparam( &nt->attr, &sp)) != 0)
		PerrorAbort("pthread_attr_setschedparam", errno);


#if defined(_POSIX_THREAD_ATTR_STACKSIZE)
	
	if((errno = pthread_attr_setstacksize(&nt->attr, threadStackSize)) != 0)
		PerrorAbort("pthread_attr_setstacksize", errno);

#endif

        if((errno = pthread_attr_setdetachstate( &nt->attr, PTHREAD_CREATE_DETACHED)) != 0)
		PerrorAbort("pthread_attr_setdetachstate");


	if((errno = pthread_attr_setinheritsched(&nt->attr, PTHREAD_EXPLICIT_SCHED)) != 0)
		PerrorAbort("pthread_attr_setinheritsched", errno);
 


	nt->thread       = thread;
	nt->func         = func;
	nt->suspendState = 0;
	nt->stackMin     = 0;
	nt->stackMax     = 0;
	nt->stackCur     = 0;

	unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;

	DBG( vm_thread, TMSG_SHORT( "create new ", nt));

	/* init our cv and mux fields for locking */
	tInitLock( nt);
	tSetLock( nt);

	/* Link the new one into the activeThreads list. We lock until
	 * the newly created thread is set up correctly (i.e. is walkable)
	 */
	LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

	nt->active = 1;
	nt->next = activeThreads;
	activeThreads = nt;

	/* Note that we don't directly start 'func' because we (a) still need to
	 * set the thread specifics, and (b) we need a looper for our thread
	 * recycling. We create the new thread while still holding the lock, because
	 * we otherwise might have a invalid tid in the activeList. The new thread
	 * in turn doesn't need the lock until it exits
	 */
	 

	if((errno = pthread_create( &nt->tid, &nt->attr, tRun, nt)) != 0)
			PerrorAbort("pthread_create", errno);


	/* wait until the thread specific data has been set, and the new thread
	 * is in a suspendable state */
	if(sem_wait( &nt->sem) < 0)
		perror("sem_wait"), ABORT(); 

	UNLOCK( cur); /* ---------------------------------------------------- tLock */
}


/***********************************************************************
 * thread exit & cleanup
 */

/*
 * Native thread cleanup. This is just called in case a native thread
 * is not pooled
 */
static
void
tDispose ( nativeThread* nt )
{
        int errno; 

 	assert(nt->active == 0);

        if((errno = pthread_mutex_destroy( &nt->mux)) != 0)
		PerrorAbort("pthread_mutex_destroy", errno);
        if((errno = pthread_cond_destroy( &nt->cv)) != 0)
		PerrorAbort("pthread_cond_destroy", errno);
	if(sem_destroy( &nt->sem) < 0)
		perror("sem_destroy"), ABORT();

        gc_free_fixed( nt);
}

/*
 * Function to be called (by threads.c firstStartThread) when the thread leaves
 * the user thread function
 */
void
Texit ( Hjava_lang_Thread* thread )
{
  LOCKSLOT;
  nativeThread         *cur = NATIVE_THREAD( thread);
  nativeThread         *t;
  int errno;

  /*
   * We are leaving the thread user func, which means we are not
   * subject to GC, anymore (has to be marked here because the first thread
   * never goes through tRun)
   */
  cur->active = 0;

  DBG( vm_thread, TMSG_SHORT( "exit ", cur));  

  if ( !unhand(thread)->daemon ){
	/* the last non daemon should shut down the process */
	if ( --nonDaemons == 0 ) {
	  LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

	  DBG( vm_thread, ("exit on last nonDaemon\n"));

	  for ( t=activeThreads; t != NULL; t = t->next ){
		if ( t != cur ) {
			if((errno = pthread_cancel( t->tid)) != 0)
				PerrorAbort("pthread_cancel", errno);
		}
	  }

	  if ( deadlockWatchdog ){
		if((errno = pthread_cancel( deadlockWatchdog)) != 0)
			PerrorAbort("pthread_cancel", errno);
	  }

	  if ( (cur != firstThread) && (firstThread->active == 0) ) {
		/* if the firstThread has already been frozen, it's not in the pool */
		if((errno = pthread_cancel( firstThread->tid)) != 0)
			PerrorAbort("pthread_cancel", errno);

	  }

	  pthread_exit( 0);

	  /* pretty useless, but clean */
	  UNLOCK( cur); /* -------------------------------------------------- tLock */

	  /* we shouldn't get here, this is a last safeguard */
	  EXIT(0);
	}
  }

  if ( cur == firstThread ) {

	LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */

	/* if we would be the head, we would have been the last, too (and already exited) */
	assert( cur != activeThreads);

	for ( t=activeThreads; (t != NULL) && (t->next != cur); t=t->next )
		;
	assert( t != NULL);
	t->next = cur->next;

	UNLOCK( cur); /* ---------------------------------------------------- tLock */  

	/*
	 * Put us into a permanent freeze to avoid shut down of the whole process (it's
	 * not clear if this is common pthread behavior or just a implementation
	 * linux-threads "feature")
	 */
	if(sem_wait( &cur->sem) < 0)
		perror("sem_wait"), ABORT();
  }
}


/*
 * Thread is being finalized - free any held resource.
 */
void    
Tfinalize (Hjava_lang_Thread* thread)
{
  DBG_ACTION( vm_thread, {
	nativeThread         *cur = NATIVE_THREAD( thread);
	DBG( vm_thread, TMSG_SHORT( "finalize ", cur));
  });
}       


/***********************************************************************
 * scheduling and query
 */

/*
 * Change the scheduler priority of a running thread. Since we aren't
 * called directly, we can assume that 'prio' is within [MIN_PRIORITY..MAX_PRIORITY]
 */
void    
Tprio (Hjava_lang_Thread* thread, jint prio)
{
  nativeThread         *cur = NATIVE_THREAD( thread);
  struct sched_param   sp;
  int errno;

  if ( cur ) {
	sp.sched_priority = priorities[prio];

	DBG( vm_thread, ("set priority: %p [tid: %d, java: %x) to %d (%d)\n",
					 cur, cur->tid, cur->thread, prio, priorities[prio]));
	if((errno = pthread_setschedparam( cur->tid, SchedulePolicy, &sp)) != 0)
		PerrorAbort("pthread_setschedparam", errno);

  }
}            


/*
 * Yield - reschedule current native thread at end of all running
 * threads with same or higher priority
 */
void
Tyield (void)
{
  tYield();
}

/*
 * Get me the Java Thread object of the current native thread
 * This should be exported as a compile time interface (potential
 * high frequency call).
 */
static          
Hjava_lang_Thread*      
TcurrentJava (void)      
{
  return tCurrentJava();
}


/*
 * This is used by the execution engine to build stack traces (we don't
 * have to care for speed, here)
 */
void
TnextFrame (stackFrame* fm)
{
  nativeThread* nt = pthread_getspecific( ntKey);
  
  machineStackFrame* nfm;
  uintp fp;

  assert(nt != 0);
 
  if (fm->return_frame == 0) {
	nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
  }
  else {
	nfm = (machineStackFrame*)fm->return_frame;
  }

  fp = (uintp)STACK_FRAME(nfm);
  if ( (fp > (uintp) nt->stackMin) && (fp < (uintp) nt->stackMax) ){
	fm->return_pc = STACK_PC(nfm);
	fm->return_frame = (void*)fp;
  }
  else {
	fm->return_pc = 0;
	fm->return_frame = 0;
  }
}



/*
 * Put yourself in PRIO_CRITICAL so no one else can run
 */
 
void
TenterCritSect ()
{
        LOCKSLOT;
	nativeThread  *t;
        struct sched_param sp;
	nativeThread  *cur = GET_CURRENT_THREAD(&cur) ;
        int errno;
	
	DBG( vm_thread, ("enter crit section[%d] from: %p [tid:%d, java:%p)\n",
					   critSection, cur, cur->tid, cur->thread));
					   
	
				   
					   
        /* avoid calling if locking recursively */
	
	LOCK( cur); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++ tLock */
					   
	if ( ++critSection == 1 ){
	
        	
		DBG( vm_thread, ("enter crit section[%d] from: %p [tid:%d, java:%p)\n",
					 critSection, cur, cur->tid, cur->thread));
	
		for ( t=activeThreads; t; t = t->next )
			if ( (t != cur) && (t->suspendState == 0) && (t->active) ) 
				t->suspendState = SS_SUSPENDED;
				
		
	        sp.sched_priority =priorities[THREAD_MAXPRIO + 1];
			
		if((errno = pthread_setschedparam(cur->tid, SchedulePolicy, &sp)) !=0)
			PerrorAbort("pthread_setschedparam", errno);


        }
	
	UNLOCK( cur); /* ------------------------------------------------------ tLock */

	
	
	DBG( vm_thread, ( "critical section (%d) established\n", critSection))
	return ;
}


/*
 * Return to our priority, resuming normal scheduling. Just take action if this
 * is the outmost exit
 */
void
TexitCritSect ()
{
      	nativeThread  *t;
        struct sched_param sp;
	int errno;
	nativeThread  *cur = GET_CURRENT_THREAD(&cur) ;


	/* deal with recursive locks */				   
        if ( --critSection == 0 ){
	
		for ( t=activeThreads; t; t = t->next )
			if ( t->suspendState & (SS_PENDING_SUSPEND | SS_SUSPENDED) )
				t->suspendState = 0;
				
	
		sp.sched_priority = priorities[unhand(cur->thread)->priority];
		if((errno = pthread_setschedparam(cur->tid, SchedulePolicy, &sp)) !=0)
			PerrorAbort("pthread_setschedparam", errno);
        }
	
	DBG( vm_thread, ( "critical section (%d) exited\n", critSection))
	return ;
}


/*******************************************************************************
 * GC related stuff
 */

/*
 * Walk stacks of all threads, except of the current one (doesn't
 * make much sense since that's the GC itself, and it's task is to
 * get rid of garbage, not to pin it down - besides the fact that
 * its stack is a moving target).
 *
 * The call chain looks like this
 *
 *  startGC
 *    Kaffe_ThreadInterface.GcWalkThreads == TwalkThreads
 *      walkMemory
 *        gcFunctions[..].walk == walkThread
 *        
 */
void
TwalkThreads (void)
{
  nativeThread *t;

 
  DBG( vm_thread, ("start walking threads\n"));

  for ( t = activeThreads; t != NULL; t = t->next) {
	walkMemory((void*)t->thread);
  }

  DBG( vm_thread, ("end walking threads\n"));

  }

/*      
 * Walk the thread's internal context. The stack is actually "walked down"
 * [base+size..base].
 * ? how useful is this for the GC thread itself ?
 */
void
TwalkThread ( Hjava_lang_Thread* thread )
{
  nativeThread  *t;
  nativeThread  *nt = NATIVE_THREAD( thread);
  void          *base;
  long          size;

  if ( !nt || !nt->active || !nt->stackMin ){
	return;
  }

  DBG( vm_thread, TMSG_LONG( "walking ", nt));

  if ( !nt->suspendState && !nt->blockState ){
	t = GET_CURRENT_THREAD(&t);

	if ( t == nt ){
	  /* Ok, we do walk the gcMan for now, but just up to this point */
	  nt->stackCur = &t;
	}
	else {
	  /* everything else should be blocked or suspended by now */
	  DBG( vm_thread, ("walking a running thread %p\n", nt));
	  /* tDump(); */
	  /* ABORT(); */
	  return;
	}
  }
  else if ( nt->suspendState == SS_PENDING_SUSPEND ){
	DBG( vm_thread, ("pending suspend, walk whole stack\n"));
	/*
	 * Assuming the very next thing after a context switch to this thread
	 * would be calling the signal handler, we accept that case. Unfortunately,
	 * we don't have a stackCur, yet
	 */
#if defined(STACK_GROWS_UP)
	nt->stackCur = nt->stackMax;
#else
	nt->stackCur = nt->stackMin;
#endif	
  }

  if ( ((uintp) nt->stackCur < (uintp) nt->stackMin) ||
	   (((uintp) nt->stackCur > (uintp) nt->stackMax)) ) {
	DBG( vm_thread, ("inconsistent stack\n"));
	tDump();
	ABORT();
  }
  markObject( unhand(thread)->jnireferences);
  markObject( unhand(thread)->exceptObj);

#if defined(STACK_GROWS_UP)
  base = nt->stackMin;
  size = (unsigned long)nt->stackCur - (unsigned long) base;
#else
  base = nt->stackCur;
  size = (uintp)nt->stackMax - (uintp) base;
#endif

  walkConservative( base, size);
}



/*
 * Return the current stack boundaries. Note that 'low' and 'high'
 * are arithmetical, i.e. not with respect to stack usage. Since it
 * just works on the current thread, this function cannot be used
 * to GC-walk stacks, hence 'curr' is rather useless in this context
 */
void
TstackInfo ( threadStackInfo* info )
{
  nativeThread* nt = GET_CURRENT_THREAD( &nt);


  info->low  = nt->stackMin;
  info->high = nt->stackMax;
  info->curr = &nt;
}


/***********************************************************************
 * create and initialize the ThreadInterface
 */

ThreadInterface Kaffe_ThreadInterface = {
	Tinit,

	Tattach,
	Tcreate,
	Tyield,
	Tprio, 
	Texit,
	0,
	Tfinalize,

	TcurrentJava,
	0,

	TenterCritSect,
	TexitCritSect,

	TwalkThreads,
	TwalkThread,

	TnextFrame,

	TstackInfo
};


