/*
 * lock-impl.h - this is the part of the LockInterface implementation
 * we want to expose (by means of injection into the abstraction locks.h)
 *
 * Use it only for two purposes:
 *  (1) redefine generic interface function macros of locks.h in order
 *      to turn the runtime interface (completely or partly) into a
 *      compile-time interface
 *  (2) VERY RESTRICTIVE - define macros which can be used in clients to
 *      check for a certain subsystem (but JUST if there are strong reasons
 *      against using a clean interface function instead)
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __lock_impl_h
#define __lock_impl_h

#include "smx-smx.h"
#include "gtypes.h"

/* always include the general interface first (so that we overload macros in the right order) */
#include "locks.h"

static __inline
jboolean
lSemget (void* sem, jlong timeout)
{
	int tm;

	if (timeout == 0) {
		tm = SMXFOREVER;
	}
	else {
		tm = (int)JAVATIMEOUT2SMXTIMEOUT(timeout);
	}
	if (test((SCB_PTR)sem, tm) == TRUE) {
		return (true);
	}
	/* Timed-out */
	return (false);
}

/*
 * Put a semaphore.
 */
static __inline
void
lSemput (void* sem)
{
	signalx(sem);
}


/******************************************************************************************
 * Define the lock interface.
 */

#undef  Kaffe_LI_semget
#define Kaffe_LI_semget(_s2pLock,_to)     lSemget(_s2pLock,_to)

#undef  Kaffe_LI_semput
#define Kaffe_LI_semput(_s2pLock)   lSemput(_s2pLock)


#endif /* __lock_impl_h */
