/*
 * smx-libc.c
 * Implement all those missing libc functions we don't get under SMX.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define EXTERN_C extern "C"

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "support.h"
#include <stdarg.h>
#include "threads.h"
#include "smx-smx.h"

extern "C" void _assert(void*, void*, unsigned);
extern "C" void abort(void);
//extern "C" int vfprintf(FILE*, const char*, va_list);
extern "C" void exit(int);

extern "C" int _fltused;
extern "C" int sprintf (char*, const char*, ...);
extern "C" int sin (void);
extern "C" int cos (void);
extern "C" int tan (void);
extern "C" int asin (void);
extern "C" int acos (void);
extern "C" int atan (void);
extern "C" int exp (void);
extern "C" int log (void);
extern "C" int sqrt (void);
extern "C" int atan2 (void);
extern "C" int pow (void);
extern "C" double strtod (const char*, char**);


void
_assert(void* str, void* file, unsigned lineno)
{
	kprintf("'%s' in %s at line %d\n", str, file, lineno);
	ABORT();
}

void
abort(void)
{
	kprintf("Aborting ... \n");
//	for(;;);          // no stacktrace workaround
}

void
exit(int code)
{
	kprintf("Exiting ... \n");
	for(;;); 
}

//int _fltused;
int sprintf (char* a, const char* b, ...) { return -1; }
//int errno;
double strtod (const char* a, char** b) { return 0; }

extern "C" {

/*
 * We use vfprintf to print to the console, using an extremely
 * primitive line protocol.
 */

extern int tm_printf(int, char*, ...);

int
vfprintf(FILE* file, const char* str, va_list args)
{
#ifndef PEG
	int a1, a2, a3, a4, a5, a6;

	a1 = va_arg(args, int);
	a2 = va_arg(args, int);
	a3 = va_arg(args, int);
	a4 = va_arg(args, int);
	a5 = va_arg(args, int);
	a6 = va_arg(args, int);

	tm_printf(0, (char*)str, a1, a2, a3, a4, a5, a6);
#else
	/* no console at all if this is graphical */
#endif

	return (0);
}


/* these are calbacks from various external (SMX) libs, make sure they don't blow up anything */
void  aos_wr_screen_string (PFCCHAR in_string, BOOLEAN from_interrupt)
{
	int dummy = 0; /* ..to get a breakpoint location */
}

void aos_wr_interrupt_string(void)
{
	int dummy = 0;
}

void raw_mode_in_char (byte chr)
{
}

}

int __cdecl critical_error_handler (int driveno, int media_status, dword sector)
{
	return CRITICAL_ERROR_ABORT;
}
