/*
 * smx-net.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#define EXTERN_C extern "C"

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "threads.h"
#include "locks.h"
#include "gc.h"
#include "smx-smx.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"

/* to be defined in the startup (demo) module) */
extern unsigned char my_ip_address[4];
extern unsigned char gw_ip_address[4];
extern unsigned char ip_mask_address[4];
extern unsigned char dns_ip_addresses[][4];
extern int n_dns_servers;
extern int device_id;
extern int minor_number;
extern word ed_io_add;
extern int  ed_irq;
extern word ed_mem_add;
extern char* hostname;

static
void
Ninit (void)
{
	int ethIf;
    unsigned char ip_ff[4] = {0xff, 0xff, 0xff, 0xff}; 

	if (xn_rtip_init() != 0) {
		// net init failed
		return;
	}

	// init irq, ports etc.
	ed_io_address = ed_io_add;
	ed_mem_address = ed_mem_add;
	ed_irq_val = ed_irq;


	// Open the interface 
	if ( (ethIf = xn_interface_open(device_id, minor_number)) < 0 ) {
		// open interface failed
		return;
	}

	// Set the IP address and interface
	if (xn_set_ip(ethIf, my_ip_address, ip_mask_address)) {	
	    // set ip failed
		return;
	}

    // Set default gateway
    if (xn_rt_add( RT_DEFAULT,ip_ff, gw_ip_address, 1, ethIf, RT_INF)) {
	    // add gateway failed
		return;
    }
    
	// set list of DNS servers
	if ( n_dns_servers ) {
		if (xn_set_server_list((dword*)dns_ip_addresses, n_dns_servers)) {
			// setting dns server list failed
			return;
		}
	}
}

static
int
Nsocket(int stream)
{
	if (stream == 0) {
		return (socket(AF_INET, SOCK_DGRAM, 0));
	}
	else {
		return (socket(AF_INET, SOCK_STREAM, 0));
	}
}

static
int
Nbind(int fd, struct sockaddr* addr, int addrlen)
{
	struct sockaddr_in *addr_in = (struct sockaddr_in*) addr;
	if ( !addr_in->sin_addr ){
		tc_mv4( &addr_in->sin_addr, &my_ip_address, 4); 
	}

	return (bind(fd, (PCSOCKADDR)addr, addrlen));
}

static
int
Nlisten(int fd, int qlength)
{
	return (listen(fd, qlength));
}

static
int
Nconnect(int fd, struct sockaddr* addr, size_t addrlen)
{
	return (connect(fd, (PCSOCKADDR)addr, addrlen));
}

static
int
Naccept(int fd, struct sockaddr* addr, size_t* addrlen)
{
	return (accept(fd, (PCSOCKADDR)addr, (int*)addrlen));
}

static
ssize_t
Nsockread(int fd, void* buf, size_t len)
{
	return (recv(fd, (PFCCHAR)buf, len, 0));
}

static
ssize_t
Nrecvfrom(int fd, void* buf, size_t len, int flags, struct sockaddr* from, int* fromlen)
{
	return (recvfrom(fd, (PFCCHAR)buf, len, flags, from, fromlen));
}

static
ssize_t
Nsockwrite(int fd, const void* buf, size_t len)
{
	return (send(fd, (PFCCHAR)buf, len, 0));
}

static
ssize_t
Nsendto(int fd, const void* buf, size_t len, int flags, const struct sockaddr* to, int tolen)
{
	return (sendto(fd, (PFCCHAR)buf, len, flags, (struct sockaddr*)to, tolen));
}

static
int
Nclose(int fd)
{
	return (closesocket(fd));
}

static
int
Nsetsockopt(int fd, int level, int optname, const void* optval, int optlen)
{
	return (setsockopt(fd, level, optname, (char*)optval, optlen));
}

static
int
Ngetsockopt(int fd, int level, int optname, void* optval, int* optlen)
{
	return (getsockopt(fd, level, optname, (char*)optval, optlen));
}

static
int
Ngetsockname(int fd, struct sockaddr* addr, int* addrlen)
{
	return (getsockname(fd, (PSOCKADDR)addr, (PFINT)addrlen));
}

static
int
Ngetpeername(int fd, struct sockaddr* addr, int* addrlen)
{
	return (getpeername(fd, (PSOCKADDR)addr, (PFINT)addrlen));
}

static
char*
Ngethostname(void)
{
	return ( hostname );
}

static
int*
Ngethostbyname(char* name)
{
	PFHOSTENT ent;
	int* addrs;

	ent = gethostbyname(name);

	addrs = (int*)gc_malloc_fixed(sizeof(int) * 2);
	addrs[0] = (int)ent->ip_addr.s_un.s_addr;
	addrs[1] = 0;
	return (addrs);
}

static
char*
Ngethostbyaddr(int addr)
{
	PFHOSTENT ent;

	ent = gethostbyaddr((PFCHAR)&addr, 4, AF_INET);
	return (ent->h_name);
}

static
int
Nioctlsocket(int fd, int opt, int* val)
{
	return (ioctlsocket(fd, opt, (unsigned long*)val));
}

NetworkCallInterface Kaffe_NetworkCallInterface = {
  Ninit,

	Nsocket,
	Nbind,
	Nlisten,
	Nconnect,
	Naccept,
	Nsockread,
	Nrecvfrom,
	Nsockwrite,
	Nsendto,
	Nclose,
	Nsetsockopt,
	Ngetsockopt,
	Ngetsockname,
	Ngetpeername,
	Ngethostname,
	Ngethostbyname,
	Ngethostbyaddr,
	Nioctlsocket,

	0, /* setreadtimeout */
	0, /* getreadtimeout */

};
