/*
 * thread-impl.h - this is the part of the ThreadInterface implementation
 * we want to expose (by means of injection into the abstraction threads.h)
 *
 * Use it only for two purposes:
 *  (1) redefine generic interface function macros of threads.h in order
 *      to turn the runtime interface (completely or partly) into a
 *      compile-time interface
 *  (2) VERY RESTRICTIVE - define macros which can be used in clients to
 *      check for a certain subsystem (but JUST if there are strong reasons
 *      against using a clean interface function instead)
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __thread_impl_h
#define __thread_impl_h

#include "smx-smx.h"

/* always make sure we see the generic macros first */
#include "threads.h"


static __inline
void
tPrio (Hjava_lang_Thread* thread, jint prio)
{
	if (unhand(thread)->PrivateInfo != 0) {
		bump_task(NATIVE_THREAD(thread)->task, JAVAPRIO2SMXPRIO(prio));
	}
}               

static __inline
void
tYield(void)
{
	suspend(ct, 0);
}

static __inline
Hjava_lang_Thread*      
tCurrentJava(void)      
{
	return ((nativeThread*)TASK_SPECIFIC_DATA(ct))->thread;
	
}

/***********************************************************************************
 * forward decls from smx-kernel.c
 */

EXTERN_C void Tinit(void);
EXTERN_C void TcreateFirst(Hjava_lang_Thread* thread);
EXTERN_C void Tcreate(Hjava_lang_Thread* thread, void* func);
EXTERN_C void Texit(Hjava_lang_Thread* thread);
EXTERN_C void Tfinalize(Hjava_lang_Thread* thread);
EXTERN_C void TenterCritSect(void);
EXTERN_C void TexitCritSect(void);
EXTERN_C void TwalkThreads(void);
EXTERN_C void TwalkThread(Hjava_lang_Thread* thread);
EXTERN_C void TnextFrame(stackFrame* fm);
EXTERN_C void TstackInfo(threadStackInfo* info);

/***********************************************************************************
 * turn the runtime interface into a compiletime interface (isn't really
 * required for all funcs, but will eventually replace the old interface)
 */

#undef  Kaffe_TI_init
#define Kaffe_TI_init()                 Tinit()

#undef  Kaffe_TI_createFirst
#define Kaffe_TI_createFirst(_thread)   TcreateFirst(_thread)

#undef  Kaffe_TI_create
#define Kaffe_TI_create(_thread,_func)  Tcreate(_thread,_func)

#undef  Kaffe_TI_yield
#define Kaffe_TI_yield()                tYield()

#undef  Kaffe_TI_setPriority
#define Kaffe_TI_setPriority(_thread,_prio) tPrio(_thread,_prio)

#undef  Kaffe_TI_exit
#define Kaffe_TI_exit(_thread)          Texit(_thread)

#undef  Kaffe_TI_finalize
#define Kaffe_TI_finalize(_thread)      Tfinalize(_thread)

#undef  Kaffe_TI_currentJava
#define Kaffe_TI_currentJava()          tCurrentJava()

#undef  Kaffe_TI_enterCritSect
#define Kaffe_TI_enterCritSect()        TenterCritSect()

#undef  Kaffe_TI_exitCritSect
#define Kaffe_TI_exitCritSect()         TexitCritSect()

#undef  Kaffe_TI_GcWalkThreads
#define Kaffe_TI_GcWalkThreads()        TwalkThreads()

#undef  Kaffe_TI_GcWalkThread
#define Kaffe_TI_GcWalkThread(_thread)  TwalkThread(_thread)

#undef  Kaffe_TI_nextFrame
#define Kaffe_TI_nextFrame(_frame)      TnextFrame(_frame)

#undef  Kaffe_TI_stackInfo
#define Kaffe_TI_stackInfo(_stckInfo)   TstackInfo(_stckInfo)


#endif /* __thread_impl_h */
