/*
 * threadx.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "thread.h"
#include "locks.h"
#include "gc.h"
#include "threadx.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"
#include <serial-fs.h>

/*
 * Open a file.
 */
static
int
Sopen(const char* path, int mode)
{
	int fd;

	switch (mode) {
	case JOPEN_READ:
		fd = open(path, 0, 0);
		break;
	case JOPEN_WRITE:
		fd = open(path, 01101, 0666);
		break;
	case JOPEN_APPEND:
		fd = open(path, 02101, 0666);
		break;
	case JOPEN_READWRITE:
		fd = open(path, 0102, 0666);
		break;
	default:
		fd = -1;
		break;
	}

	return (fd);
}

/*
 * Get information about a file handle.
 */
static
int
SfileinfoFD(int fd, int type, void* extra)
{
	int r;
	struct fileinfo buf;

	switch (type) {
	case JINFO_LENGTH:
		r = rem_fileinfo(fd, &buf);
		if (r == 0) {
			*((jlong*)extra) = buf.length;
			return (1);
		}
		break;
	}

	return (0);
}

static
void
notSupported()
{
	printf("Operation not supported\n");
}

/*
 * Open file.
 */
int
Sopen(const char* path, int mode)
{
	int fd;

	switch (mode) {
	case JOPEN_READ:
		fd = open(path, 0, 0);
		break;
	case JOPEN_WRITE:
		fd = open(path, 01101, 0666);
		break;
	case JOPEN_APPEND:
		fd = open(path, 02101, 0666);
		break;
	case JOPEN_READWRITE:
		fd = open(path, 0102, 0666);
		break;
	default:
		fd = -1;
		break;
	}

	return (fd);
}

FileSystemCallInterface Kaffe_FileSystemCallInterface = {

	Sopen,
	Sread,
	Swrite,
	Slseek,
	Sclose,
	Sfileinfo,
	SfileinfoFD,

	Sreaddir,

	Smkdir,
	Srmdir,
	Srename,
	Sremove,

};
