/*
 * threadx.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "thread.h"
#include "locks.h"
#include "gc.h"
#include "threadx.h"
#define NOUNIXPROTOTYPES
#include "jsyscall.h"

/*
 * Do we run finalizer on exit?  Well this VM probably never exits anyhow.
 */
bool runFinalizerOnExit;

static Hjava_lang_Thread* activeThreads;
static Hjava_lang_Thread* fthread;

static
void
TcreateFirst(Hjava_lang_Thread* tid)
{
	static TX_SEMAPHORE firstCv;
	nativeThread* ptr;

	ptr = gc_malloc_fixed(sizeof(nativeThread));
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	ptr->threadx = tx_thread_identify();
	ptr->seminfo.sem = &firstCv;
	ptr->seminfo.next = 0;
	tx_semaphore_create(ptr->seminfo.sem, "cv", 0);

	/* Add thread to list of known java threads */
	ptr->next = activeThreads;
	activeThreads = tid;

	fthread = tid;
}

static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	nativeThread* ptr;
	void* sbase;

	ptr = gc_malloc_fixed(sizeof(nativeThread));
	ptr->threadx = gc_malloc_fixed(sizeof(TX_THREAD));
	sbase = gc_malloc_fixed(threadStackSize);
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	tx_thread_create(ptr->threadx,
		"thread", func, 0xDEADBABE,
		sbase, threadStackSize,
		JAVAPRIO2THREADXPRIO(unhand(tid)->priority),
		JAVAPRIO2THREADXPRIO(unhand(tid)->priority),
		TIMESLICE, /* TX_NO_TIME_SLICE, TIMESLICE */
		TX_DONT_START);

	/* Add thread to list of known java threads */
	ptr->next = activeThreads;
	activeThreads = tid;

	tx_thread_resume(ptr->threadx);
}

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{
	UINT old;

	if (unhand(tid)->PrivateInfo != 0) {
		tx_thread_priority_change(NATIVE_THREAD(tid)->threadx, JAVAPRIO2THREADXPRIO(prio), &old);
	}
}               

/*
 * Sleep current thread for given period.
 */
static
void
Tsleep(jlong period)
{
	tx_thread_sleep(JAVATIMEOUT2THREADXTIMEOUT(period));
}

/*
 * Terminate current thread.
 */
static
void
Texit(Hjava_lang_Thread* ttid)
{
	TX_THREAD* tid = NATIVE_THREAD(ttid)->threadx;
	Hjava_lang_Thread** tptr;
	nativeThread* ntid;

	/* Remove thread from active list */
	for (tptr = &activeThreads; *tptr != 0; tptr = &ntid->next) {
		ntid = NATIVE_THREAD(*tptr);
		if (ntid->threadx == tid) {
			*tptr = ntid->next;
			break;
		}
	}

	tx_thread_delete(tid);
}

/*
 * Yield processor
 */
static
void
Tyield(void)
{
	tx_thread_relinquish();
}

static
jint    
Tframes(Hjava_lang_Thread* tid)
{
	return (0);
}
 
/*
 * Thread is being finalized - free any held resource.
 */
static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
	if (tid != fthread) {
		gc_free_fixed(NATIVE_THREAD(tid)->threadx);
	}
	gc_free_fixed(NATIVE_THREAD(tid)->seminfo.sem);
	gc_free_fixed(NATIVE_THREAD(tid));
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{
	TX_THREAD* tid = tx_thread_identify();
	Hjava_lang_Thread* tptr;
	nativeThread* nptr;

	/* Find Java thread */
	for (tptr = activeThreads; tptr != 0; tptr = nptr->next) {
		nptr = NATIVE_THREAD(tptr);
		if (nptr->threadx == tid) {
			return (tptr);
		}
	}
	return (0);	/* Shouldn't happen */
}

static          
void*
TcurrentNative(void)      
{
	return (tx_thread_identify());
}

static
semaphoreThread*
TcurrentSem(void)
{
	TX_THREAD* tid = tx_thread_identify();
	Hjava_lang_Thread* tptr;
	nativeThread* nptr;

	for (tptr = activeThreads; tptr != 0; tptr = nptr->next) {
		nptr = NATIVE_THREAD(tptr);
		if (nptr->threadx == tid) {
			return (&nptr->seminfo);
		}
	}
	return (0);	/* Shouldn't happen */
}

static          
void
TwalkThreads(void)
{               
	Hjava_lang_Thread* tid;

	for (tid = activeThreads; tid != NULL; tid = NATIVE_THREAD(tid)->next) {
		walkMemory(tid);
	}
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	nativeThread* native;

	if (unhand(tid)->PrivateInfo == 0) {
		return;
	}

	native = NATIVE_THREAD(tid);
	markObject(unhand(tid)->jnireferences);
	markObject(unhand(tid)->exceptObj);
	walkConservative(native->threadx->tx_stack_start, native->threadx->tx_stack_end - native->threadx->tx_stack_start);
}

/*static*/
void
TnextFrame(stackFrame* fm)
{
	machineStackFrame* nfm;
	TX_THREAD* ptr;
	void* fp;

	if (fm->return_frame == 0) {
//printf("First frame\n");
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
//printf("Next frame\n");
		nfm = (machineStackFrame*)fm->return_frame;
	}

	ptr = tx_thread_identify();
	fp = STACK_FRAME(nfm);

//printf("FRAME 0x%x: ret_fp 0x%x ret_sp 0x%x ret_pc 0x%x\n", nfm, nfm->ret_fp, nfm->ret_sp, nfm->ret_pc);
//printf("THREAD: 0x%x -> 0x%x\n", ptr->tx_stack_start, ptr->tx_stack_end);
//{ int i; for (i = -4 ; i < 5; i++) { if (i == 0) { printf("[0x%08x] ", ((void**)nfm)[i]); } else { printf("0x%08x ", ((void**)nfm)[i]); } } printf("\n"); }

	if (fp > ptr->tx_stack_start && fp < ptr->tx_stack_end) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = fp;
	}
	else {
//printf("*** END OF STACK ***\n");
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

static
void  
LinitFirst(iLock* lk)
{
	static TX_SEMAPHORE firstMux;

	tx_semaphore_create(&firstMux, "mux", 1);
	lk->mux = &firstMux;
}               

static
void  
LinitRest(iLock* lk)
{
	TX_SEMAPHORE* mux;

	mux = gc_malloc_fixed(sizeof(TX_SEMAPHORE));
	tx_semaphore_create(mux, "mux", 1);
	lk->mux = mux;
}               

/*
 * Get a semaphore (with timeout).
 */
static
void
Lsemget(void* sem, jlong timeout)
{
	int tout;

	if (timeout == 0) {
		tout = TX_WAIT_FOREVER;
	}
	else {
		tout = JAVATIMEOUT2THREADXTIMEOUT(timeout);
	}
	tx_semaphore_get((TX_SEMAPHORE*)sem, tout);
}

/*
 * Put a semaphore.
 */
static
void
Lsemput(void* sem)
{
	tx_semaphore_put((TX_SEMAPHORE*)sem);
}

void    
Tspinon(void* arg)
{
}       
        
void            
Tspinoff(void* arg)
{
}               

void
Tspinoffall(void* arg)
{
}

static
void
Tinit(void)
{
}

void
notSupported()
{
	printf("Operation not supported\n");
}

/*
 * Define the thread interface.
 */
ThreadInterface Kaffe_ThreadInterface = {

	Tinit,
        TcreateFirst,
        Tcreate,
        Tsleep,
        Tyield,
        Tprio, 
        Texit,
        Tframes,
        Tfinalize,
        TcurrentJava,
	TcurrentNative,
	TcurrentSem,
        TwalkThreads,
        TwalkThread,
        TnextFrame

};

/*
 * Define the lock interface.
 */
LockInterface Kaffe_LockInterface = {

        LinitFirst,
        LinitRest,

        0,
        0,
        0,
        0,
        0,

	Lsemget,
	Lsemput,

        Tspinon,
        Tspinoff,

};

/*
 * Define the system call interface.
 */
KernelCallInterface Kaffe_KernelCallInterface = {

        notSupported,
        notSupported,	
        notSupported

};
