/*
 * kvxworks.c
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "thread.h"
#include "locks.h"
#include "jsyscall.h"
#include "gc.h"
#include "kvxworks.h"

/*
 * Do we run finalizer on exit?  Well this VM probably never exits anyhow.
 */
bool runFinalizerOnExit;

static Hjava_lang_Thread* activeThreads;
static SEM_ID spinlock = (SEM_ID)0xDEADDEAD;

static
void
TcreateFirst(Hjava_lang_Thread* tid)
{
	nativeThread* ptr;

	ptr = gc_malloc_fixed(sizeof(nativeThread));
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	ptr->tid = taskIdSelf();

	/* Add thread to list of known java threads */
	ptr->next = activeThreads;
	activeThreads = tid;

	ptr->stackBase = ...
	ptr->stackEnd = ...
}

static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	nativeThread* ptr;
	static unsigned long args[4];

	ptr = gc_malloc_fixed(sizeof(nativeThread));
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	/* Add thread to list of known java threads */
	ptr->next = activeThreads;
	activeThreads = ptr;

	ptr->tid = taskSpawn(
		NULL,
		JAVAPRIO2VXWORKSPRIO(unhand(tid)->priority),
		VX_FP_TASK,
		threadStackSize,
		func,
		0,0,0,0,0,0,0,0,0,0);

	ptr->stackBase = ...
	ptr->stackEnd = ...

}

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{
	if (unhand(tid)->PrivateInfo != 0) {
		taskPrioritySet(NATIVE_THREAD(tid)->tid, JAVAPRIO2VXWORKSPRIO(prio));
	}
}               

/*
 * Sleep current thread for given period.
 */
static
void
Tsleep(jlong period)
{
	taskDelay(JAVATIMEOUT2VXWORKSTIMEOUT(period));
}

/*
 * Terminate current thread.
 */
static
void
Texit(void)
{
	int tid;
	Hjava_lang_Thread** tptr;
	nativeThread* ntid;

	tid = taskIdSelf();

	taskDelete(tid);

	/* Remove thread from active list */
	for (tptr = &activeThreads; *tptr != 0; tptr = &ntid->next) {
		ntid = NATIVE_THREAD(*tptr);
		if (ntid->threadx == tid) {
			*tptr = ntid->next;
			break;
		}
	}
}

/*
 * Yield processor
 */
static
void
Tyield(void)
{
	taskDelay(0);
}

static
jint    
Tframes(Hjava_lang_Thread* tid)
{
	return (0);
}
 
/*
 * Thread is being finalized - free any held resource.
 */
static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
	gc_free_fixed(NATIVE_THREAD(tid));
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{
	int tid;
	Hjava_lang_Thread* tptr;
	nativeThread* nptr;

	tid = taskIdSelf();

	/* Find Java thread */
	for (tptr = activeThreads; tptr != 0; tptr = nptr->next) {
		nptr = NATIVE_THREAD(tptr);
		if (nptr->tid == tid) {
			return (tptr);
		}
	}
	return (0);	/* Shouldn't happen */
}

static          
void*
TcurrentNative(void)      
{
	return ((void*)taskIdSelf());
}

static          
void
TwalkThreads(void)
{               
	Hjava_lang_Thread* tid;

	for (tid = activeThreads; tid != NULL; tid = NATIVE_THREAD(tid)->next) {
		walkMemory(tid);
	}
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	nativeThread* native;

	if (unhand(tid)->PrivateInfo == 0) {
		return;
	}

	native = NATIVE_THREAD(tid);
	markObject(unhand(tid)->jnireferences);
	markObject(unhand(tid)->exceptObj);
	walkConservative(native->stackBase, native->stackEnd);
}

static
void
TnextFrame(stackFrame* fm)
{
	machineStackFrame* nfm;
	void* fp;
	nativeThread* native;

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

	fp = STACK_FRAME(nfm);
	native = NATIVE_THREAD(TcurrentJava());

	if (fp > native->stackBase && fp < native->stackEnd) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = fp;
	}
	else {
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

static
void  
Linit(iLock* lk)
{
	SEM_ID mux;
	SEM_ID cv;

	mux = semBCreate(SEM_Q_PRIORITY, SEM_FULL);
	cv = semCCreate(SEM_Q_PRIORITY, 0);

	lk->mux = (void*)mux;
	lk->cv = (void*)cv;
}               

/*
 * Get a semaphore (with timeout).
 */
static
void
Lsemget(void* sem, jlong timeout)
{
	int tout;

	if (timeout == 0) {
		tout = WAIT_FOREVER;
	}
	else {
		tout = JAVATIMEOUT2VXWORKSTIMEOUT(timeout);
	}

	semTake((SEM_ID)sem, tout);
}

/*
 * Put a semaphore.
 */
static
void
Lsemput(void* sem)
{
	semGive((SEM_ID)sem);
}

void    
Tspinon(void* arg)
{
	if (spinlock == (SEM_ID)0xDEADDEAD) {
		spinlock = semBCreate(SEM_Q_PRIORITY, SEM_FULL);
	}
	semTake(spinlock, WAIT_FOREVER);
}       
        
void            
Tspinoff(void* arg)
{
	semGive(spinlock);
}               

void
Tspinoffall(void* arg)
{
}

static
void
Tinit(void)
{
}

static
void
notSupported()
{
	printf("Operation not supported\n");
}

/*
 * Define the thread interface.
 */
ThreadInterface Kaffe_ThreadInterface = {

	Tinit,
        TcreateFirst,
        Tcreate,
        Tsleep,
        Tyield,
        Tprio, 
        Texit,
        Tframes,
        Tfinalize,
        TcurrentJava,
	TcurrentNative,
        TwalkThreads,
        TwalkThread,
        TnextFrame,

};

/*
 * Define the lock interface.
 */
LockInterface Kaffe_LockInterface = {

        Linit,
        Linit,

        0,
        0,
        0,
        0,
        0,

	Lsemget,
	Lsemput,

        Tspinon,
        Tspinoff,

};

/*
 * Define the system call interface.
 */
SystemCallInterface Kaffe_SystemCallInterface = {

	notSupported,
        notSupported,
        notSupported,	
        notSupported, 
        notSupported,
        notSupported,
        notSupported,
        notSupported,

        notSupported,

        notSupported,
        notSupported,
        notSupported,
        notSupported,

        notSupported,
        notSupported,
        notSupported, 
        notSupported,	
        notSupported,
        notSupported, 
        notSupported,	
        notSupported,	
        notSupported,
        notSupported, 
        notSupported,

        notSupported,	

        notSupported
};
