/*
 * vxworks-ti.c
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */


#define	SDBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "threads.h"
#include "locks.h"
#include "gc.h"
#include "support.h"
#if defined(HAVE_SYS_TIMES_H)
#include <sys/times.h>
#endif

#include "vxWorks.h"
#include "ioLib.h"

#include <time.h>

#if 0
extern void walkMemory(void*);
extern void markObject(void*);
extern void walkConservative(void*, int);
#endif

#include "vxworks-ti.h"

int sysClkRateGet(void) ;

#define NOUNIXPROTOTYPES
#include "jsyscall.h"

static Hjava_lang_Thread* TcurrentJava(void);

int runFinalizerOnExit;
jint nonDemonThreads;
static Hjava_lang_Thread* activeThreads;


static
void
TcreateFirst(Hjava_lang_Thread* tid)
{
	nativeThread* ptr;

	ptr = (nativeThread*)gc_malloc_fixed(sizeof(nativeThread));
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	ptr->task = (WIND_TCB*) taskIdSelf();
	unhand(tid)->sem = (struct Hkaffe_util_Ptr*) semBCreate(SEM_Q_FIFO, SEM_EMPTY) ;

/* WRS says that the Base is the first used stack location, the end is the last (hopefully
	never) used location.
*/
	ptr->stkend = (int) ptr->task->pStackBase ;
	ptr->stkbase = (int) ptr->task->pStackEnd ;

	ptr->next = activeThreads;
	activeThreads = tid;

	nonDemonThreads = 1;
}


static
int
startNewThread(void* arg)
{
	Hjava_lang_Thread* tid;
	nativeThread* ntid;

	tid = TcurrentJava();
	ntid = NATIVE_THREAD(tid);

/* WRS says that the Base is the first used stack location, the end is the last (hopefully
	never) used location.
*/
#if 0
	ntid->stkend = (int) ntid->task->pStackBase ;
	ntid->stkbase = (int) ntid->task->pStackEnd ;
#endif

	ntid->func();

	return (0);
}



int kaffeThreadPriorityLow = 100 ;

# define THREAD_PRIORITY_IDLE					(kaffeThreadPriorityLow - 0)
# define THREAD_PRIORITY_LOWEST					(kaffeThreadPriorityLow - 1)
# define THREAD_PRIORITY_BELOW_NORMAL			(kaffeThreadPriorityLow - 2)
# define THREAD_PRIORITY_NORMAL					(kaffeThreadPriorityLow - 3)
# define THREAD_PRIORITY_ABOVE_NORMAL			(kaffeThreadPriorityLow - 4)
# define THREAD_PRIORITY_HIGHEST				(kaffeThreadPriorityLow - 5)
# define THREAD_PRIORITY_TIME_CRITICAL			(kaffeThreadPriorityLow - 6)

static  
void    
Tprio(Hjava_lang_Thread* tid, jint prio)
{
	if (unhand(tid)->PrivateInfo == 0) {
		return;
	}
	switch (prio) {
	case 1:
		prio = THREAD_PRIORITY_IDLE;
		break;
	case 2:
		prio = THREAD_PRIORITY_LOWEST;
	case 3:
	case 4:
		prio = THREAD_PRIORITY_BELOW_NORMAL;
		break;
	case 5:
	default:
		prio = THREAD_PRIORITY_NORMAL;
		break;
	case 6:
	case 7:
	case 8:
		prio = THREAD_PRIORITY_ABOVE_NORMAL;
		break;
	case 9:
	case 10:
		prio = THREAD_PRIORITY_HIGHEST;
		break;
	case 11:
		prio = THREAD_PRIORITY_TIME_CRITICAL;
		break;
	}

	taskPrioritySet((int) NATIVE_THREAD(tid)->task, prio) ;
}            


static
void
Tcreate(Hjava_lang_Thread* tid, void* func)
{
	int priority,
		options ;
	char * stack ;
	STATUS r ; 
	WIND_TCB * tcb ;

	nativeThread* ptr;

	ptr = (nativeThread*)gc_malloc_fixed(sizeof(nativeThread));
	assert(ptr != NULL) ;
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)ptr;

	ptr->func = (void(*)(void))func;

/* Default flags used with thread creation */
	tcb = (WIND_TCB *) gc_malloc_fixed(sizeof(WIND_TCB)) ;
	taskPriorityGet((int) taskIdSelf(), &priority) ;
	options = VX_FP_TASK ;
	stack = (char*) gc_malloc_fixed(threadStackSize) ;
	assert(stack != NULL) ;

	ptr->stkbase = (int) stack ;
	ptr->stkend = (int) &(stack[threadStackSize]);

/* Um..  the stack address used by taskInit is the *top* of the stack--not the
	bottom...  So we add an offset to the address.
*/
	stack = &(stack[threadStackSize]);

	r = taskInit(tcb, "Kaffe", priority, options, stack, threadStackSize, startNewThread, 
		1, 2, 3, 4, 5, 6, 7, 8, 9, 10) ;
	assert(r != ERROR) ;

	ptr->task = tcb ;
	unhand(tid)->sem = (struct Hkaffe_util_Ptr*) semBCreate(SEM_Q_FIFO, SEM_EMPTY) ;

	ptr->next = activeThreads;
	activeThreads = tid;

	if (unhand(tid)->daemon == 0) {
		nonDemonThreads++;
	}

	Tprio(tid, unhand(tid)->priority);

	taskActivate((int) ptr->task);
}


/*
 * Terminate the current thread.
 */
static
void
Texit(Hjava_lang_Thread* tid)
{
#if 0
	Hjava_lang_Thread** ptr;
	nativeThread* ntid;

	if (unhand(tid)->daemon == 0) {
		nonDemonThreads--;
		if (nonDemonThreads == 0) {
			EXIT(1);
		}
	}

	ntid = NATIVE_THREAD(tid);
	for (ptr = &activeThreads; *ptr != 0; ptr = &NATIVE_THREAD(*ptr)->next) {
		if (tid == *ptr) {
			*ptr = ntid->next;
			if (ntid->task == (WIND_TCB*) taskIdSelf()) {
				/* wait for the finalizer to notify us to clean up
				 * (we can't do it right here, because we are going to kill our
				 * stack, which causes havoc to the gc/finalizer)
				 */
				taskSuspend((int) ntid->task);

				/* Ok, suicide is painless.. */
				semGive((SEM_ID) unhand(tid)->sem);
				tid->PrivateInfo = 0;
				gc_free_fixed(ntid);
/* We should free more gc_malloc_fixed memory */

				exit(0);
			}
			else {
				taskDelete((int) ntid->task);
/* We should free gc_malloc_fixed memory */
			}
			return;
		}
	}
#endif
	Hjava_lang_Thread** ptr;
	nativeThread* ntid;

	if (unhand(tid)->daemon == 0) {
		nonDemonThreads--;
		if (nonDemonThreads == 0) {
			EXIT(1);
		}
	}

/* Remove us from the active thread list. */
	ntid = NATIVE_THREAD(tid);
	ptr = &activeThreads ;
	while (ptr[0])
	{	if (ptr[0] == tid)
		{	ptr[0] = ntid->next;
			break ;
		}
		ptr = &(NATIVE_THREAD(ptr[0])->next) ;
	}

/* And we're done. */
	exit(0) ;
}


/*
 * Yield processor
 */
static
void
Tyield(void)
{
	taskDelay(0);
}

/*
 * Thread is being finalized - free any held resource.
 */
static
void    
Tfinalize(Hjava_lang_Thread* tid)
{
#if 0
	/* wake up, it's time to die */
	taskResume((int) (NATIVE_THREAD(tid)->task)) ;
# else
	void * p ;
	nativeThread * nt ;

/* Finalize is called after Exit for the thread.  Free the resources it used. */
	nt = (void*) NATIVE_THREAD(tid) ;

/* gc_free_... knows when we're freeing a thread. */
	unhand(tid)->PrivateInfo = NULL ;

/* Delete the thread's semaphore */
	semDelete((SEM_ID) unhand(tid)->sem) ;

/* Free the stack. */
	p = (void*) nt->stkbase ;
	gc_free_fixed(p) ;

/* Free the TCB. */
	p = nt->task ;
	gc_free_fixed(p) ;

/* And finally delete the nativeThread */
	gc_free_fixed(nt) ;
# endif
}       

static          
Hjava_lang_Thread*      
TcurrentJava(void)      
{
	Hjava_lang_Thread* ptr;
	WIND_TCB * ct;

	ct = (WIND_TCB*) taskIdSelf() ;
	for (ptr = activeThreads; ptr != 0; ptr = NATIVE_THREAD(ptr)->next) {
		if (NATIVE_THREAD(ptr)->task == ct) {
			return (ptr);
		}
	}
	return (0);
}

static
void
TenterCritSect(void)
{	taskLock() ;
}

static
void
TexitCritSect(void)
{	taskUnlock() ;
}


static          
void
TwalkThreads(void)
{               
	Hjava_lang_Thread* tid;

#if 0
	// printf("Walking threads\n");
#endif
	for (tid = activeThreads; tid != NULL; tid = NATIVE_THREAD(tid)->next) {
		walkMemory((void*)tid);
	}
}

/*      
 * Walk the thread's internal context.
 */
static
void
TwalkThread(Hjava_lang_Thread* tid)
{
	nativeThread* ntid;

#if 0
	// printf("walking thread %x\n", tid);
#endif
	ntid = NATIVE_THREAD(tid);
	if (ntid == 0) {
		return;
	}

	markObject(unhand(tid)->jnireferences);
	markObject(unhand(tid)->exceptObj);

#if 0
	// printf("stack %x - %x\n", ntid->stkbase, ntid->stkend);
#endif
	walkConservative((void*)ntid->stkbase, ntid->stkend - ntid->stkbase);
#if 0
	// printf("done\n");
#endif
}

static
void
TnextFrame(stackFrame* fm)
{
	machineStackFrame* nfm;
	uintp fp;
	nativeThread* ct = NATIVE_THREAD(TcurrentJava());

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

	fp = (uintp)STACK_FRAME(nfm);
#if 0
	// printf("fp = %x (%x-%x)\n", fp, ct->stkbase, ct->stkend);
#endif
	if (fp > ct->stkbase && fp < ct->stkend) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = (void*)fp;
	}
	else {
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

static
void
TstackInfo(threadStackInfo* info)
{
	nativeThread* ct = NATIVE_THREAD(TcurrentJava());

	info->low = (void*)ct->stkbase ;
	info->high = (void*)ct->stkend ;
	info->curr = info->low ;
}
              
/*
 * Get a semaphore (with timeout).
 */
static
jboolean
Lsemget(void* sem, jlong timeout)
{
#if 0
	DWORD tm;

SDBG(	kprintf("Lsemget: %x\n", sem);	)
	if (timeout == 0) {
		tm = INFINITE;
	}
	else {
		tm = (DWORD)timeout;
	}
	if (WaitForSingleObject((HANDLE)sem, tm) != WAIT_TIMEOUT) {
		return (true);
	}
	else {
		return (false);
	}
#endif

	if (timeout == 0)
		timeout = WAIT_FOREVER ;
	else
		timeout = (timeout * sysClkRateGet() + 999) / 1000 ;

	if (semTake((SEM_ID) sem, timeout) == ERROR)
		return false ;
	else
		return true ;
}

/*
 * Put a semaphore.
 */
static
void
Lsemput(void* sem)
{
SDBG(	kprintf("Lsemput: %x\n", sem);	)
	semGive((SEM_ID) sem) ;
}


/*
 * We must have a certain amount of credible thread information setup
 * as soon as possible.
 */
static
void
tSetupFirstNative(void)
{
  nativeThread* nt;

  /*
   * We need to have a native thread context available as soon as possible.
   */
  nt = gc_malloc( sizeof(nativeThread), GC_ALLOC_THREADCTX);
  nt->task = (WIND_TCB*) taskIdSelf() ;
  nt->next = NULL ;
  nt->stkbase  = (int) nt->task->pStackEnd ;
  nt->stkend  = (int) nt->task->pStackBase ;
  nt->func = NULL ;
}



static
void
Tinit(void)
{
	tSetupFirstNative() ;
}


void
notSupported()
{
	kprintf("Operation not supported\n");
}

void*
Mpagealloc(int size)
{
	void* ptr;

	size += gc_pgsize;

#undef malloc

	ptr = malloc(size);
	if (ptr == 0) {
		return (0);
	}
	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -(int)gc_pgsize);

	return (ptr);
}


static
int
Kinput(void)
{
	char ch;
#if 1
	static int inputInitialized = 0 ;
	if (! inputInitialized)
	{	inputInitialized = 1 ;
		ioctl(0, FIOSETOPTIONS, OPT_RAW) ;
		setbuf(stdin, NULL);
	}
#endif
	if (read(0, &ch, 1) != 1)
		return -1 ;

	return ch;
}


static
int
Kinputavailable(void)
{	int unread ;
	ioctl(0, FIONREAD, (int) &unread) ;
	return unread ;
}


static
void
Koutput(int chr)
{	char ch = (char) chr ;
	write(1, &ch, 1) ;
}


static
void
Kerror(int chr)
{	char ch = (char) chr ;
	write(2, &ch, 1) ;
}


static
jlong
Ktime(void)
{
	struct timespec tp ;
	jlong r ;
	clock_gettime(CLOCK_REALTIME, &tp) ;
	tp.tv_nsec /= 1000 ;	/* ns -> us */
	tp.tv_nsec /= 1000 ;	/* us -> ms */
	r = tp.tv_sec ;
	r *= 1000 ;
	r += tp.tv_nsec ;
	return r ;
}


/*
 * Convert C strings to UNICODE strings.
 */
unsigned short*
getUnicode(const char* str)
{
	int len;
	int i;
	unsigned short* nstr;

	len = strlen(str);
	nstr = gc_malloc_fixed((len+1) * sizeof(unsigned short));
	for (i = 0; i < len; i++) {
		nstr[i] = (jchar)str[i];
	}
	nstr[i] = 0;
	return (nstr);
}

void
freeUnicode(unsigned short* str)
{
	gc_free_fixed(str);
}

/*
 * Define the thread interface.
 */

ThreadInterface Kaffe_ThreadInterface = {

	Tinit,

	TcreateFirst,
	Tcreate,
	Tyield,
	Tprio, 
	Texit,
	0,
	Tfinalize,

	TcurrentJava,
	0,

	TenterCritSect,
	TexitCritSect,

	TwalkThreads,
	TwalkThread,

	TnextFrame,

	TstackInfo,

};

/*
 * Define the lock interface.
 */

LockInterface Kaffe_LockInterface = {

	0,
	0,
	0,
	0,

	Lsemget,
	Lsemput,

};

/*
 * Define the system call interface.
 */

KernelCallInterface Kaffe_KernelCallInterface = {

	0,
	0,
	0,
	Mpagealloc,
	0,
	Kinput,
	Koutput,
	Kerror,
	Ktime,
	Kinputavailable,

};



/* pjo, 16 mar 2000 */
/* VxWorks doesn't have a vsnprintf... so we'll just call vsprintf and hope
	that it doesn't overflow.
*/
int vsnprintf(char * str, size_t size, const char * format, va_list ap)
{	int r = vsprintf(str, format, ap) ;
	assert(strlen(str) >= size);
	return r;
}


