/*
 * lock-impl.h - this is the part of the LockInterface implementation
 * we want to expose (by means of injection into the abstraction locks.h)
 *
 * Use it only for two purposes:
 *  (1) redefine generic interface function macros of locks.h in order
 *      to turn the runtime interface (completely or partly) into a
 *      compile-time interface
 *  (2) VERY RESTRICTIVE - define macros which can be used in clients to
 *      check for a certain subsystem (but JUST if there are strong reasons
 *      against using a clean interface function instead)
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#ifndef __lock_impl_h
#define __lock_impl_h

/***********************************************************************
 * the public part - redefine the relevant LockInterface access macros
 */

static inline
jboolean
lSemget(void* sem, jlong timeout)
{
	DWORD tm;

	if (timeout == 0) {
		tm = INFINITE;
	}
	else {
		tm = (DWORD)timeout;
	}
	if (WaitForSingleObject((HANDLE)sem, tm) != WAIT_TIMEOUT) {
		return (true);
	}
	else {
		return (false);
	}
}

/*
 * Put a semaphore.
 */
static inline
void
lSemput(void* sem)
{
#if defined(UNDER_CE)
	SetEvent((HANDLE)sem);
#else
	ReleaseSemaphore((HANDLE)sem, 1, NULL);
#endif
}

/******************************************************************************************
 * Define the lock interface.
 */

#undef  Kaffe_LI_semget
#define Kaffe_LI_semget(_s2pLock,_to)     lSemget(_s2pLock,_to)

#undef  Kaffe_LI_semput
#define Kaffe_LI_semput(_s2pLock)   lSemput(_s2pLock)

#endif /* __lock_impl_h */
