/*
 * thread-impl.c
 *
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "threads.h"
#include "locks.h"
#include "gc.h"
#include "support.h"
#if defined(HAVE_SYS_TIMEB_H)
#include <sys/timeb.h>
#endif

extern void walkMemory(void*);
extern void markObject(void*);
extern void walkConservative(void*, uint32);

#define NOUNIXPROTOTYPES
#include "jsyscall.h"

#include "jdebug.h"

/* these are required for handling exceptions */
#include "java_lang_Throwable.h"
#include "engine.h"
#include "exception.h"

/* Some systems need special setups for the exception handling */
#if !defined(EXCEPTIONSTART)
#define EXCEPTIONSTART()
#endif
#if !defined(EXCEPTIONEND)
#define EXCEPTIONEND()
#endif

#if defined(UNDER_CE)
int errno;
#endif

int runFinalizerOnExit;
int nonDemonThreads;

static nativeThread* firstThread;
static nativeThread* activeThreads;

/* Pure guess work !!! */
#define	STACK_TOP(S)	((((uintp)(S)) + 0x1000 - 1) & -0x1000)

#define MAX_CACHED_THREADS 64
nativeThread*	cache;
int				nCached;
int             pendingExits;

#define LOCKSLOT   int iLockRoot
#define LOCK()     lockStaticMutex( &tLock)
#define UNLOCK()   unlockStaticMutex( &tLock)
static iLock	*tLock;

CRITICAL_SECTION  critSec;
int               tlsIdx;

/* map the Java priority levels to whatever the  impl gives us */
int priorities[] = {
	THREAD_PRIORITY_IDLE,           /* 0 */
	THREAD_PRIORITY_IDLE,           /* 1 */
	THREAD_PRIORITY_LOWEST,         /* 2 */
	THREAD_PRIORITY_BELOW_NORMAL,   /* 3 */
	THREAD_PRIORITY_BELOW_NORMAL,   /* 4 */
	THREAD_PRIORITY_NORMAL,         /* 5 */
	THREAD_PRIORITY_ABOVE_NORMAL,   /* 6 */
	THREAD_PRIORITY_ABOVE_NORMAL,   /* 7 */
	THREAD_PRIORITY_ABOVE_NORMAL,   /* 8 */
	THREAD_PRIORITY_HIGHEST,        /* 9 */
	THREAD_PRIORITY_HIGHEST,        /* 10 */
	THREAD_PRIORITY_TIME_CRITICAL   /* 11 */
}; 

inline int
mapPriority ( int prio )
{
	if ( prio < 0 ) {
		return priorities[0];
	}
	else if ( prio >= sizeof(priorities) ) {
		return priorities[sizeof(priorities)-1];
	}
	else {
		return priorities[prio];
	}
}

inline HANDLE
semCreate(void)
{
#if defined(UNDER_CE)
	return (CreateEvent( NULL, FALSE, FALSE, NULL));
#else
	return (CreateSemaphore( NULL, 0, 0xffff, NULL));
#endif
}

unsigned long
handle_exception( EXCEPTIONPROTO ) {
	EXCEPTIONSTART();

	EXCEPTIONFIXRETURN();

#if !defined(UNDER_CE)
	switch ( ctx->ExceptionRecord->ExceptionCode ) {
	case EXCEPTION_ACCESS_VIOLATION:
		postAndThrowException("java/lang/NullPointerException", 0);
		break;

	case EXCEPTION_DATATYPE_MISALIGNMENT:
	case EXCEPTION_BREAKPOINT:
	case EXCEPTION_SINGLE_STEP:
	case EXCEPTION_ARRAY_BOUNDS_EXCEEDED:
	case EXCEPTION_FLT_DENORMAL_OPERAND:
	case EXCEPTION_FLT_DIVIDE_BY_ZERO:
	case EXCEPTION_FLT_INEXACT_RESULT:
	case EXCEPTION_FLT_INVALID_OPERATION:
	case EXCEPTION_FLT_OVERFLOW:
	case EXCEPTION_FLT_STACK_CHECK:
	case EXCEPTION_FLT_UNDERFLOW:
	case EXCEPTION_INT_DIVIDE_BY_ZERO:
	case EXCEPTION_INT_OVERFLOW:
	case EXCEPTION_PRIV_INSTRUCTION:
	case EXCEPTION_IN_PAGE_ERROR:
	case EXCEPTION_ILLEGAL_INSTRUCTION:
	case EXCEPTION_NONCONTINUABLE_EXCEPTION:
	case EXCEPTION_STACK_OVERFLOW:
	case EXCEPTION_INVALID_DISPOSITION:
	case EXCEPTION_GUARD_PAGE:
	case EXCEPTION_INVALID_HANDLE:
	default:
		kprintf( "EXCEPTION: %x\n", ctx->ExceptionRecord->ExceptionCode );
		return (EXCEPTION_CONTINUE_SEARCH);
	}
#else
	postAndThrowException("java/lang/NullPointerException", 0);
#endif

	EXCEPTIONEND();

	return (EXCEPTION_EXECUTE_HANDLER);
}

void
Tattach (Hjava_lang_Thread* thread)
{
	nativeThread* cur;
	uintp stkp;
	char* mem;
	LOCKSLOT;

	/* Don't attach the thread twice */
	cur = GET_CURRENT_THREAD(&cur);
	if (cur != 0) {
		return;
	}
	cur = (nativeThread*)gc_malloc_fixed(sizeof(nativeThread));
	unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)cur;

	cur->tid = GetCurrentThreadId();
	cur->hand = GetCurrentThread();
	cur->thread = thread;
	cur->next = 0;

	cur->sem = semCreate();
	cur->semLock = semCreate();
	unhand(thread)->sem = cur->semLock;

	/* Calculate the stack start and end */
#ifdef __GNUC__
	asm ("movl %%esp, %0" : "=r" (stkp));
#else
	__asm
	{
		mov stkp, esp
	}
#endif

	stkp = STACK_TOP(stkp);
	cur->stkend = stkp;
	cur->stkbase = cur->stkend - threadStackSize;

	/* Touch the base of stack to get it allocated */
#if !defined(UNDER_CE)
	mem = alloca(threadStackSize);
	mem[threadStackSize - 1] = 0;
#endif

	LOCK(); /* ----------------------------------------------------- */

	if (firstThread == 0) {
		firstThread = cur;
	}
	cur->next = activeThreads;
	activeThreads = cur;

	TlsSetValue( tlsIdx, cur);

	nonDemonThreads++;

	UNLOCK(); /* --------------------------------------------------- */
}

static
DWORD
WINAPI
tRun (void* arg)
{
	LOCKSLOT;
	nativeThread  *cur = (nativeThread*)arg;
	nativeThread  *t;
	uintp stkp;

	/* Calculate the stack start and end */
#ifdef __GNUC__
	asm ("movl %%esp, %0" : "=r" (stkp));
#else
	__asm
	{
		mov stkp, esp
	}
#endif

	stkp = STACK_TOP(stkp);
	cur->stkend = stkp;
	cur->stkbase = cur->stkend - threadStackSize;

	TlsSetValue( tlsIdx, cur);

	/* notify creator that thread data is initialized */
	lSemput( cur->sem);

	while ( 1 ) {
#ifdef __GNUC__
	  cur->func();
#else
		__try {
			cur->func();
		}
		__except ( handle_exception( GetExceptionInformation() ) ) {
		}
#endif

		LOCK(); /* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

		if ( cur == activeThreads ){
			activeThreads = cur->next;
		}
		else {
			for ( t=activeThreads; t->next && (t->next != cur); t=t->next );
			assert( t->next != 0 );
			t->next = cur->next;
		}

		/* unlink Java and native thread */
		unhand(cur->thread)->PrivateInfo = 0;
		cur->thread = 0;
		
		/* link into cache list (if still within cache limits) */
		if ( ++nCached < MAX_CACHED_THREADS ){
			cur->next = cache;
			cache = cur;
			
			DBG( vm_thread, TMSG_SHORT( "cached thread ", cur));
		}
		pendingExits--;

		UNLOCK(); /* ----------------------------------------------------------------- */
		
		if ( nCached >= MAX_CACHED_THREADS ){
		  break;
		}
		
		WaitForSingleObject( cur->sem, INFINITE);
	}

	CloseHandle( cur->sem);
	CloseHandle( cur->semLock);
	CloseHandle( cur->hand);

	gc_free_fixed( cur);

	return (0);
}

void    
Tprio (Hjava_lang_Thread* thread, jint prio)
{
	if (unhand(thread)->PrivateInfo == 0) {
		return;
	}

	SetThreadPriority( NATIVE_THREAD(thread)->hand, mapPriority(prio));
}            

void
Tcreate ( Hjava_lang_Thread* thread, void* func )
{
	LOCKSLOT;
	nativeThread* nt;

	/* if we are the first one, it's seriously broken */
	assert( activeThreads != 0 );
	
	/*
	* This is a safeguard to avoid creating too many new threads
	* because of a high Tcreate call frequency from a high priority
	* thread (which doesn't give exiters a chance to aquire the lock
	* to update the cache list).
	*/
	if ( cache == 0 ) {
		while ( pendingExits && (cache == 0) ){
			Sleep(0);
		}
	}	

	LOCK(); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

	if ( cache != NULL ) {
		nt = cache;
		cache = nt->next;
		nCached--;

		nt->thread = thread;
		unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;
		unhand(thread)->sem = nt->semLock;

		SetThreadPriority( nt->hand, mapPriority(unhand(thread)->priority));

		nt->next = activeThreads;
		activeThreads = nt;

		lSemput( nt->sem);
	}
	else {
		nt = (nativeThread*)gc_malloc_fixed(sizeof(nativeThread));
		nt->sem =  semCreate();
		nt->semLock = semCreate();
		unhand(thread)->sem = nt->semLock;

		nt->thread = thread;
		nt->func = (void(*)(void))func;
		unhand(thread)->PrivateInfo = (struct Hkaffe_util_Ptr*)nt;

		nt->next = activeThreads;
		activeThreads = nt;

		nt->hand = CreateThread( NULL, threadStackSize, tRun, nt, 0, &nt->tid);
		SetThreadPriority( nt->hand, mapPriority(unhand(thread)->priority));

		WaitForSingleObject( nt->sem, INFINITE);
	}


	if (unhand(thread)->daemon == 0) {
		nonDemonThreads++;
	}

	UNLOCK(); /* -------------------------------------------------------------------- */
}

/*
 * Terminate current thread.
 */
void
Texit (Hjava_lang_Thread* thread)
{
	LOCKSLOT;
	nativeThread* cur = NATIVE_THREAD( thread);
	nativeThread* t;

	LOCK(); /* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

	if (unhand(thread)->daemon == 0) {
		nonDemonThreads--;

		/* the "kill all" condition */
		if (nonDemonThreads == 0) {
			DeleteCriticalSection( &critSec);
			exit(0);
		}
	}

	/* firstThread never makes it into the tRun cleanup processing */
	if ( cur == firstThread ){
		/* if we would be the head, we would have been the last one, too (and already exited) */
		assert( cur != activeThreads);
		
		for ( t=activeThreads; (t != NULL) && (t->next != cur); t=t->next );
		assert( t != NULL);
		t->next = 0;

		/* we don't exit because it would wipe a CE process */
		UNLOCK(); /* ------------------------------------------------------------------ */
		WaitForSingleObject( cur->sem, INFINITE);
		return;
	}
	else {
		pendingExits++;
	}

	UNLOCK(); /* ---------------------------------------------------------------------- */
}

/*
 * Yield processor
 */
void
Tyield (void)
{
	Sleep(0);
}

/*
 * Thread is being finalized - free any held resource.
 */
void    
Tfinalize (Hjava_lang_Thread* thread)
{
}       

Hjava_lang_Thread*      
TcurrentJava (void)
{
	return tCurrentJava();
}

void
TenterCritSect(void)
{
	LOCKSLOT;
	LOCK();
	EnterCriticalSection( &critSec);
	UNLOCK();
}

void
TexitCritSect(void)
{
	LeaveCriticalSection( &critSec);
}

void
TwalkThreads(void)
{          
	nativeThread* t;

	/* we don't have to lock - it's called from a critical section, anyway */

	// printf("Walking threads\n");
	for (t = activeThreads; t != NULL; t = t->next) {
		walkMemory((void*)t->thread);
	}
}

/*      
 * Walk the thread's internal context.
 */
void
TwalkThread (Hjava_lang_Thread* thread)
{
	nativeThread* cur = GET_CURRENT_THREAD(&cur);
	nativeThread* t = NATIVE_THREAD(thread);
	unsigned long sz;
	CONTEXT ctx;

	// printf("walking thread %x\n", t->hand);
	if (t == 0) {
		return;
	}

	if ( t != cur ) {
		ctx.ContextFlags = CONTEXT_CONTROL;
		GetThreadContext( t->hand, &ctx);
		t->stkcur = ctx.Esp;
		if ( (t->stkcur < t->stkbase) || (t->stkcur > t->stkend) ) {
			t->stkcur = t->stkbase;
		}
	}
	else {
		t->stkcur = (uintp)&cur;
	}

	markObject(unhand(thread)->jnireferences);
	markObject(unhand(thread)->exceptObj);


	sz = (unsigned long)t->stkend - (unsigned long)t->stkcur;
	walkConservative((void*)t->stkcur, sz);
}

void
TnextFrame (stackFrame* fm)
{
	machineStackFrame* nfm;
	uintp fp;
	nativeThread* cur = GET_CURRENT_THREAD(&cur);

	if (fm->return_frame == 0) {
		nfm = (machineStackFrame*)STACK_CURRENT_FRAME();
	}
	else {
		nfm = (machineStackFrame*)fm->return_frame;
	}

	fp = (uintp)STACK_FRAME(nfm);
	// printf("fp = %x (%x-%x)\n", fp, cur->stkbase, cur->stkend);
	if (fp > cur->stkbase && fp < cur->stkend) {
		fm->return_pc = STACK_PC(nfm);
		fm->return_frame = (void*)fp;
	}
	else {
		fm->return_pc = 0;
		fm->return_frame = 0;
	}
}

void
TstackInfo (threadStackInfo* info)
{
	nativeThread* cur = GET_CURRENT_THREAD(&cur);

	if ( cur ) {
		info->low = (void*)cur->stkbase;
		info->high = (void*)cur->stkend;
		info->curr = info->low; //cur->stkcur;
	}
	else { //fallback in case its not registered yet
		uintp stkp;
#ifdef __GNUC__
		asm ("movl %%esp, %0" : "=r" (stkp));
#else
		__asm
		{
			mov stkp, esp
		}
#endif

		stkp = STACK_TOP(stkp);
		info->high = (void*)stkp;
		info->low = (void*)(stkp - Kaffe_JavaVMArgs[0].nativeStackSize);
		info->curr = info->low;
	}
}
              
void
Tinit(void)
{
	InitializeCriticalSection( &critSec);
	tlsIdx = TlsAlloc();
}

jboolean
Lsemget(void* sem, jlong timeout)
{
	return lSemget( sem, timeout);
}

void
Lsemput(void* sem)
{
	lSemput( sem);
}


void
notSupported()
{
	kprintf("Operation not supported\n");
}

void*
Mpagealloc(int size)
{
	void* ptr;

	size += gc_pgsize;

#undef malloc

	ptr = malloc(size);
	if (ptr == 0) {
		return (0);
	}
	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -(int)gc_pgsize);

	return (ptr);
}

/***********************************************************************
 * Interface functions
 */

/*
 * Turn the provided file descriptor in something that conforms
 * to our threading system (eg. set non-blocking modes for user
 * thread systems). Since we talk kernel-threads, we don't care
 * for blocking
 */
jint
KFileDescriptor ( int fd )
{
  return fd;
}

static
int
Kinput(void)
{
	char ch;
	if ((*Kaffe_FileSystemCallInterface._read)(0, &ch ,1) == 1) {
		return (ch);
	}
	else {
		return (-1);
	}
}

static
int
Kinputavailable(void)
{
//	return (fileinfoFD(0, JINFO_AVAILABLE, 0));
	return 0;
}

static
void
Koutput(int chr)
{
//	fwrite( &chr, 1, 1, stdout);
	char ch = (char)chr;
	(*Kaffe_FileSystemCallInterface._write)(1, &ch ,1);
}

static
void
Kerror(int chr)
{
//	fwrite( &chr, 1, 1, stderr);
	char ch = (char)chr;
	(*Kaffe_FileSystemCallInterface._write)(2, &ch ,1);
}

static
jlong
Ktime(void)
{	
	jlong tme = 0;
#if defined(HAVE_SYS_TIMEB_H)
	struct _timeb tm;
	_ftime(&tm);
	tme = (((jlong)tm.time * (jlong)1000) + (jlong)tm.millitm);
#endif
	return (tme);
}


/*
 * Define the thread interface.
 */

ThreadInterface Kaffe_ThreadInterface = {

	Tinit,

	Tattach,
	Tcreate,
	Tyield,
	Tprio, 
	Texit,
	0,
	Tfinalize,

	TcurrentJava,
	0,

	TenterCritSect,
	TexitCritSect,

	TwalkThreads,
	TwalkThread,

	TnextFrame,

	TstackInfo,

};

LockInterface Kaffe_LockInterface = {

	0,
	0,
	0,
	0,

	Lsemget,
	Lsemput,

};

/*
 * Define the system call interface.
 */

KernelCallInterface Kaffe_KernelCallInterface = {

	KFileDescriptor,
	0,
	0,
	Mpagealloc,
	0,
	Kinput,
	Koutput,
	Kerror,
	Ktime,
	Kinputavailable
};
