/*
* thread-impl.h - this is the part of the ThreadInterface implementation
* we want to expose (by means of injection into the abstraction threads.h)
*
* Use it only for two purposes:
*  (1) redefine generic interface function macros of threads.h in order
*      to turn the runtime interface (completely or partly) into a
*      compile-time interface
*  (2) VERY RESTRICTIVE - define macros which can be used in clients to
*      check for a certain subsystem (but JUST if there are strong reasons
*      against using a clean interface function instead)
*
* Copyright (c) 1999
*      Transvirtual Technologies, Inc.  All rights reserved.
*
* See the file "license.terms" for information on usage and redistribution
* of this file.
*/

#ifndef __thread_impl_h
#define __thread_impl_h

/* always make sure we see the generic macros first */
#include "threads.h"

/* Get us the required internal types we need for our inline funcs. This seems
* to be the wrong order (impl including/exposing internal), but there's no way
* around that if we want any benefit from inlines/macros
*/
#include "thread-internal.h"

#include "jdebug.h"

/***********************************************************************************
* define the interface funtions we would like to provide as inlines
*/

static inline          
Hjava_lang_Thread*
tCurrentJava (void)      
{
	nativeThread *cur = GET_CURRENT_THREAD( &cur);
	return (cur ? cur->thread : 0);
}


/* these are the function prototypes we export in our interface macros */
JNIEXPORT void JNICALL Tinit (void);
JNIEXPORT void JNICALL Tyield (void);
JNIEXPORT void JNICALL TenterCritSect (void);
JNIEXPORT void JNICALL TexitCritSect (void);
JNIEXPORT void JNICALL Tattach (Hjava_lang_Thread* thread);
JNIEXPORT void JNICALL Tcreate ( Hjava_lang_Thread* thread, void* func );
JNIEXPORT void JNICALL Tprio (Hjava_lang_Thread* thread, jint prio);
JNIEXPORT void JNICALL Texit ( Hjava_lang_Thread* thread );
JNIEXPORT void JNICALL Tfinalize ( Hjava_lang_Thread* thread );
JNIEXPORT void JNICALL TnextFrame (stackFrame* fm);
JNIEXPORT void JNICALL TwalkThreads (void);
JNIEXPORT void JNICALL TwalkThread ( Hjava_lang_Thread* thread );
JNIEXPORT void JNICALL TstackInfo ( threadStackInfo* info );

JNIEXPORT int runFinalizerOnExit;

/***********************************************************************************
* turn the runtime interface into a compiletime interface (isn't really
* required for all funcs, but will eventually replace the old interface)
*/

#undef  Kaffe_TI_init
#define Kaffe_TI_init()                 Tinit()

#undef  Kaffe_TI_attach
#define Kaffe_TI_attach(_thread)   Tattach(_thread)

#undef  Kaffe_TI_create
#define Kaffe_TI_create(_thread,_func)  Tcreate(_thread,_func)

#undef  Kaffe_TI_yield
#define Kaffe_TI_yield()                Tyield()

#undef  Kaffe_TI_setPriority
#define Kaffe_TI_setPriority(_thread,_prio) Tprio(_thread,_prio)

#undef  Kaffe_TI_exit
#define Kaffe_TI_exit(_thread)          Texit(_thread)

#undef  Kaffe_TI_finalize
#define Kaffe_TI_finalize(_thread)      Tfinalize(_thread)

#undef  Kaffe_TI_currentJava
#define Kaffe_TI_currentJava()          tCurrentJava()

#undef  Kaffe_TI_enterCritSect
#define Kaffe_TI_enterCritSect()        TenterCritSect()

#undef  Kaffe_TI_exitCritSect
#define Kaffe_TI_exitCritSect()         TexitCritSect()

#undef  Kaffe_TI_GcWalkThreads
#define Kaffe_TI_GcWalkThreads()        TwalkThreads()

#undef  Kaffe_TI_GcWalkThread
#define Kaffe_TI_GcWalkThread(_thread)  TwalkThread(_thread)

#undef  Kaffe_TI_nextFrame
#define Kaffe_TI_nextFrame(_frame)      TnextFrame(_frame)

#undef  Kaffe_TI_stackInfo
#define Kaffe_TI_stackInfo(_stckInfo)   TstackInfo(_stckInfo)

#undef  Kaffe_TI_interrupt
#define Kaffe_TI_interrupt(_thread)     /* not supported yet */

#endif /* __thread_impl_h */
