/*
 */

#define	SOCKDBG(s)

#include "config.h"
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "gtypes.h"
#include "gc.h"
#define	NOUNIXPROTOTYPES
#include <jsyscall.h>

#if !defined(FD_SETSIZE)
#define FD_SETSIZE 32
#endif

#if defined(UNDER_CE)
#define WSASocket(a,b,c,d,e,f)		socket(a,b,c)
#define WSAConnect(a,b,c,d,e,f,g)	connect(a,b,c)
#define WSAAccept(a,b,c,d,e)		accept(a,b,c)

unsigned short* getUnicode(const char*);
void freeUnicode(unsigned short*);
#endif

static struct {
	int				inuse;
	union {
		HANDLE		hand;
		SOCKET		sock;
	};
	FILE			*ptr;
	jlong			timeout;	/* threads timeout on read */
} fhand[FD_SETSIZE];

static
int
allocHandle(void)
{
	int i;

	/* Skip stdin, stdout, stderr */
	for (i = 0; i < FD_SETSIZE; i++) {
		if (fhand[i].inuse == 0) {
			fhand[i].inuse = 1;
			fhand[i].ptr = 0;
			return (i);
		}
	}

	return (-1);
}

static
void
freeHandle(int fd)
{
	fhand[fd].inuse = 0;
}

static
void
Finit(void)
{
#if !defined(UNDER_CE)
	fhand[0].inuse = 1;
	fhand[0].hand = GetStdHandle(STD_INPUT_HANDLE);
	fhand[0].ptr = 0;
	fhand[1].inuse = 1;
	fhand[1].hand = GetStdHandle(STD_OUTPUT_HANDLE);
	fhand[1].ptr = 0;
	fhand[2].inuse = 1;
	fhand[2].hand = GetStdHandle(STD_ERROR_HANDLE);
	fhand[2].ptr = 0;
#else
	fhand[0].inuse = 1;
	fhand[0].ptr = stdin;
	fhand[0].hand = 0;
	fhand[1].inuse = 1;
	fhand[1].ptr = stdout;
	fhand[1].hand = 0;
	fhand[2].inuse = 1;
	fhand[2].ptr = stdout;	//use stdout for command shell usage
	fhand[2].hand = 0;
#endif
}

static
int
Fmkdir(const char* path)
{
#if !defined(UNDER_CE)
	if (CreateDirectory(path, NULL)) {
		return (0);
	}
	else {
		return (-1);
	}
#else
	TCHAR* upath = getUnicode(path);
	if (CreateDirectory(upath, NULL)) {
		freeUnicode(upath);
		return (0);
	}
	else {
		freeUnicode(upath);
		return (-1);
	}
#endif
}

static
int
Frmdir(const char* path)
{
#if !defined(UNDER_CE)
	if (RemoveDirectory(path)) {
		return (0);
	}
	else {
		return (-1);
	}
#else
	TCHAR* upath = getUnicode(path);
	if (RemoveDirectory(upath)) {
		freeUnicode(upath);
		return (0);
	}
	else {
		freeUnicode(upath);
		return (-1);
	}
#endif
}

static
int
FfileInfo(const char* path, int type, void* extra)
{
	BY_HANDLE_FILE_INFORMATION buf;
	HANDLE h;
	LARGE_INTEGER li;

#if !defined(UNDER_CE)
	DWORD attr = GetFileAttributes(path);
#else
	TCHAR* upath = getUnicode(path);
	DWORD  attr = GetFileAttributes(upath);
	freeUnicode(upath);
#endif

	if (attr == 0xFFFFFFFF) {
		return (0);
	}
	
	switch (type) {
	case JINFO_EXISTS:
		return (1);

	case JINFO_ISFILE:
		if (!(attr & FILE_ATTRIBUTE_DIRECTORY)) {
			return (1);
		}
		return (0);

	case JINFO_ISDIR:
		if (attr & FILE_ATTRIBUTE_DIRECTORY) {
			return (1);
		}
		return (0);

	case JINFO_CANREAD:
		return (1);

	case JINFO_CANWRITE:
		if (!(attr & FILE_ATTRIBUTE_READONLY)) { 
			return (1);
		}
		return (0);
	}

#if !defined(UNDER_CE)
	h = CreateFile(path, 0, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL);
#else
	upath = getUnicode(path);
	h = CreateFile(upath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL);
	freeUnicode(upath);
#endif

	if (h == INVALID_HANDLE_VALUE) {
		return (0);
	}

	GetFileInformationByHandle(h, &buf);
	CloseHandle(h);

	switch (type) {
	case JINFO_LENGTH:
		*((jlong*)extra) = buf.nFileSizeLow;
		return (1);

	case JINFO_LASTMODIFIED:
		li.LowPart = buf.ftLastWriteTime.dwLowDateTime;
		li.HighPart = buf.ftLastWriteTime.dwHighDateTime;
		*((jlong*)extra) = (jlong)li.QuadPart / (jlong)10;
		return (1);
	}

	return (0);
}

static
int
FfileInfoFD(int fd, int type, void* extra)
{
	DWORD r;

	switch (type) {
	case JINFO_LENGTH:
		r = GetFileSize(fhand[fd].hand, NULL);
		if (r != 0xFFFFFFFF) {
			*((jlong*)extra) = (jlong)r;
			return (1);
		}
		break;

	case JINFO_AVAILABLE:
		return (GetFileSize(fhand[fd].hand, NULL) - SetFilePointer(fhand[fd].hand, (LONG)0, NULL, FILE_CURRENT));

	default:
		break;
	}

	return (0);
}

static
int
Freaddir(const char* p, jdirEntry** list)
{
	WIN32_FIND_DATA data;
	char path[256];
	jdirEntry* dirlist;
	jdirEntry* mentry;
	int count;
	HANDLE handle;
#if defined(UNDER_CE)
	TCHAR *upath;
#endif

	strcpy(path, p);
	strcat(path, "\\*.*");

#if !defined(UNDER_CE)
	handle = FindFirstFile(path, &data);
#else
	upath = getUnicode(path);
	handle = FindFirstFile(upath, &data);
	freeUnicode(upath);
#endif

	if (handle == INVALID_HANDLE_VALUE) {
		*list = 0;
		return (-1);
	}

	count = 0;
	dirlist = 0;

#if !defined(UNDER_CE)
	for (;;) {
		/* We skip '.' and '..' */
		if (strcmp(".", data.cFileName) != 0 && strcmp("..", data.cFileName) != 0) {
			mentry = gc_malloc_fixed(sizeof(jdirEntry) + strlen(data.cFileName) + 1);
			assert(mentry != 0);
			strcpy(mentry->name, data.cFileName);
			mentry->next = dirlist;
			dirlist = mentry;
			count++;
		}
		if (!FindNextFile(handle, &data)) {
			FindClose(handle);
			break;
		}
	}
#else
	for (;;) {
		int i;

		/* We skip '.' and '..' */
		if ((char)data.cFileName[0] == '.') {
			if (data.cFileName[1] == 0) {
				goto skip;
			}
			if ((char)data.cFileName[1] == '.' && data.cFileName[2] == 0) {
				goto skip;
			}
		}
		for (i = 0; data.cFileName[i] != 0; i++)
			;
		mentry = gc_malloc_fixed(sizeof(jdirEntry) + i + 1);
		assert(mentry != 0);
		for (i = 0; data.cFileName[i] != 0; i++) {
			mentry->name[i] = (char)data.cFileName[i];
		}
		mentry->next = dirlist;
		dirlist = mentry;
		count++;
skip:;
		if (!FindNextFile(handle, &data)) {
			FindClose(handle);
			break;
		}
	}
#endif

	*list = dirlist;
	return (count);
}

/*
 * Threaded file open.
 */
static
int
Fopen(const char* path, int mode)
{
	int fd;
#if defined(UNDER_CE)
	TCHAR* upath;
#endif

	fd = allocHandle();
	if (fd == -1) {
		return (-1);
	}

#if !defined(UNDER_CE)
	switch (mode) {
	case JOPEN_READ:
		fhand[fd].hand = CreateFile(path, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case JOPEN_WRITE:
		fhand[fd].hand = CreateFile(path, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case JOPEN_APPEND:
		fhand[fd].hand = CreateFile(path, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case JOPEN_READWRITE:
		fhand[fd].hand = CreateFile(path, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	default:
		freeHandle(fd);
		return (-1);
	}
#else
	upath = getUnicode(path);
	switch (mode) {
	case JOPEN_READ:
		fhand[fd].hand = CreateFile(upath, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case JOPEN_WRITE:
		fhand[fd].hand = CreateFile(upath, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case JOPEN_APPEND:
		fhand[fd].hand = CreateFile(upath, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	case JOPEN_READWRITE:
		fhand[fd].hand = CreateFile(upath, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		break;
	default:
		freeUnicode(upath);
		freeHandle(fd);
		return (-1);
	}
	freeUnicode(upath);
#endif

	// printf("Opening %s on %d %x\n", path, fd, fhand[fd].hand);

	if (fhand[fd].hand == INVALID_HANDLE_VALUE) {
		freeHandle(fd);
		return (-1);
	}
	else {
		return (fd);
	}
}

/*
 * Threaded read
 */
static
ssize_t
Fread(int fd, void* buf, size_t len)
{
	ssize_t r;
	BOOL res;

	if ( fhand[fd].inuse && fhand[fd].ptr ) {
		r = fread(buf, 1, len, fhand[fd].ptr);
	}
	else {
		res = ReadFile(fhand[fd].hand, (LPVOID)buf, (DWORD)len, (LPDWORD)&r, NULL);
		if (res == 0) {
			r = -1;
		}
	}
	return (r);
}

/*
 * Threaded write
 */
static
ssize_t
Fwrite(int fd, const void* buf, size_t len)
{
	ssize_t r;
	BOOL res;

	if ( fhand[fd].inuse && fhand[fd].ptr ) {
		r = fwrite( buf, 1, len, fhand[fd].ptr );
	}
	else {
		res = WriteFile(fhand[fd].hand, (LPVOID)buf, (DWORD)len, (LPDWORD)&r, NULL);
		if (res == 0) {
			r = -1;
		}
	}
	return (r);
}

static
int
Flseek(int fd, off_t pos, int whence)
{
	DWORD ret;

	switch (whence) {
	case JSEEK_SET:
		ret = SetFilePointer(fhand[fd].hand, (LONG)pos, NULL, FILE_BEGIN);
		break;
	case JSEEK_CUR:
		ret = SetFilePointer(fhand[fd].hand, (LONG)pos, NULL, FILE_CURRENT);
		break;
	case JSEEK_END:
		ret = SetFilePointer(fhand[fd].hand, (LONG)pos, NULL, FILE_END);
		break;
	}

	// printf("seek %d %d %d -> %d\n", fd, (int)pos, whence, ret);

	return ((int)ret);
}

static
int
Fclose(int fd)
{
	CloseHandle(fhand[fd].hand);
	freeHandle(fd);
	return (0);
}

static
int
Frename(const char* from, const char* to)
{
	return (-1);
}

static
int
Fremove(const char* file)
{
	return (-1);
}

/* --------------------------------------------------------------------- */
/* Networking */

static
void
Ninit (void)
{
#if !defined(UNDER_CE)
	WORD ver;
	WSADATA data;
	ver = MAKEWORD(2,0);
	WSAStartup(ver, &data);
#endif
}

/*
 * Threaded socket create.
 */
static
int
Nsocket(int stream)
{
	int fd;

	fd = allocHandle();
	if (fd == -1) {
		return (-1);
	}

	if (stream == 0) {
		fhand[fd].sock = WSASocket(AF_INET, SOCK_DGRAM, 0, NULL, 0, 0);
	}
	else {
		fhand[fd].sock = WSASocket(AF_INET, SOCK_STREAM, 0, NULL, 0, 0);
	}

	if (fhand[fd].sock == INVALID_SOCKET) {
		freeHandle(fd);
		return (-1);
	}
	else {
		return (fd);
	}
}

/*
 * Threaded socket connect.
 */
static
int
Nconnect(int fd, struct sockaddr* addr, size_t len)
{
	int r;

	r = WSAConnect(fhand[fd].sock, (const struct sockaddr*)addr, len, NULL, NULL, NULL, NULL);

	if (r == SOCKET_ERROR) {
		return (-1);
	}
	else {
		return (0);
	}
}

/*
 * Threaded socket accept.
 */
static
int
Naccept(int fd, struct sockaddr* addr, size_t* len)
{
	int nfd;

	nfd = allocHandle();
	if (nfd == -1) {
		return (-1);
	}

	fhand[nfd].sock = WSAAccept(fhand[fd].sock, addr, len, NULL, 0);

	if (fhand[nfd].sock == INVALID_SOCKET) {
		freeHandle(nfd);
		return (-1);
	}
	return (nfd);
}

static
int
Nlisten(int fd, int len)
{
	if (listen(fhand[fd].sock, len) == SOCKET_ERROR) {
		return (-1);
	}
	else {
		return (0);
	}
}

static
int
Nbind(int fd, struct sockaddr* addr, int len)
{
	if (bind(fhand[fd].sock, addr, len) == SOCKET_ERROR) {
		return (-1);
	}
	else {
		return (0);
	}
}

static
int
Nclose(int fd)
{
	closesocket(fhand[fd].sock);
	freeHandle(fd);
	return (0);
}


/*
 * Threaded recvfrom 
 */
static
ssize_t 
Nrecvfrom(int fd, void* buf, size_t len, int flags, 
	struct sockaddr* from, int* fromlen)
{
	int ret;

#if !defined(UNDER_CE)
	ssize_t r = -1;
 	WSABUF wbuf;

	wbuf.len = (u_long)len;
	wbuf.buf = (char FAR*)buf;

	ret = WSARecvFrom(fhand[fd].sock, &wbuf, 1, &r, &flags,
		from, fromlen, NULL, NULL);
	if ( ret != SOCKET_ERROR ) {
		return r;
	}
#else
	ret = recvfrom(fhand[fd].sock, (char*)buf, len, flags, from, fromlen);
	if ( ret != SOCKET_ERROR ) {
		return ret;
	}
#endif

	return (-1);
}

/*
 * Threaded recvfrom 
 */
static
ssize_t 
Nread(int fd, void* buf, size_t len)
{
	int ret;

#if !defined(UNDER_CE)
	WSABUF wbuf;
	int flags;
	int r;

	wbuf.len = (u_long)len;
	wbuf.buf = (char FAR*)buf;
	flags = 0;

	ret = WSARecv(fhand[fd].sock, &wbuf, 1, &r, &flags, NULL, NULL);
	if ( ret != SOCKET_ERROR ) {
		return (r);
	}
#else
	ret = recv(fhand[fd].sock, (char*)buf, len, 0);
	if (ret != SOCKET_ERROR) {
		return (ret);
	}
#endif
	return (-1);
}

static
ssize_t
Nsendto(int fd, const void* buf, size_t len, int flags, const struct sockaddr* addr, int alen)
{
	int res;
#if !defined(UNDER_CE)
	ssize_t r = -1;
	WSABUF wbuf;

	wbuf.len = (u_long)len;
	wbuf.buf = (char FAR*)buf;

	res = WSASendTo(fhand[fd].sock, &wbuf, 1, &r, flags, 
		addr, alen, NULL, NULL);
	if (res != SOCKET_ERROR) {
		return (r);
	}
#else
	res = sendto(fhand[fd].sock, (char*)buf, len, flags, addr, alen);
	if (res != SOCKET_ERROR) {
		return (res);
	}
#endif

	return (-1);
}

static
ssize_t
Nwrite(int fd, const void* buf, size_t len)
{
	int res;
#if !defined(UNDER_CE)
	ssize_t r = -1;
	WSABUF wbuf;

	wbuf.len = (u_long)len;
	wbuf.buf = (char FAR*)buf;
	res = WSASend(fhand[fd].sock, &wbuf, 1, &r, 0, NULL, NULL);
	if (res != SOCKET_ERROR) {
		return (r);
	}
#else
	res = send(fhand[fd].sock, (char*)buf, len, 0);
	if (res != SOCKET_ERROR) {
		return (res);
	}
#endif
	return (-1);
}

/*
 * Set the socket timeout.
 */
static
int
Nsetreadtimeout(int fd, int time)
{
	fhand[fd].timeout = (jlong)time;
	return (1);
}

/*
 * Get the socket timeout.
 */
static
int
Ngetreadtimeout(int fd)
{
	return ((int)fhand[fd].timeout);
}

static
char*
Ngethostname(void)
{
	char* buf;

	buf = gc_malloc_fixed(64);
	if (gethostname(buf, 64) == 0) {
		return (buf);
	}
	else {
		return ("localhost");
	}
}

static
int*
Ngethostbyname(char* name)
{
	struct hostent* ent;
	int* addrs;

	ent = gethostbyname(name);
	if (ent == 0) {
		return (0);
	}
	addrs = (int*)gc_malloc_fixed(sizeof(int) * 2);
	addrs[0] = *(int*)ent->h_addr;
	addrs[1] = 0;
	return (addrs);
}

static
char*
Ngethostbyaddr(int addr)
{
	struct hostent* ent;

	ent = gethostbyaddr((char*)&addr, 4, AF_INET);
	if (ent == 0) {
		return (0);
	}
	return (ent->h_name);
}

static
int
Ngetsockname(int fd, struct sockaddr* addr, int* alen)
{
	if (getsockname(fhand[fd].sock, addr, alen) == SOCKET_ERROR) {
		return (-1);
	}
	else {
		return (0);
	}
}

static
int
Ngetpeername(int fd, struct sockaddr* addr, int* alen)
{	
	if (getpeername(fhand[fd].sock, addr, alen) == SOCKET_ERROR) {
		return (-1);
	}
	else {
		return (0);
	}
}

static
int
Nsetsockopt(int fd, int opt, int x1, const void* x2, int x3)
{
	return (-1);
}

static
int
Ngetsockopt(int fd, int opt, int x1, void* x2, int* x3)
{
	return (-1);
}

static
int
Nioctl(int fd, int opt, int* val)
{
	return (-1);
}

FileSystemCallInterface Kaffe_FileSystemCallInterface = {

	Finit,

	Fopen,
	Fread,	
	Fwrite, 
	Flseek,
	Fclose,
	FfileInfo,
	FfileInfoFD,

	Freaddir,

	Fmkdir,
	Frmdir,
#if !defined(UNDER_CE)
	rename,
	remove,
#else
	Frename,
	Fremove,
#endif

};

NetworkCallInterface Kaffe_NetworkCallInterface = {
    Ninit,

	Nsocket,
	Nbind,
	Nlisten,
	Nconnect,
	Naccept,
	Nread,	
	Nrecvfrom,

	Nwrite,
	Nsendto,
	Nclose,
	Nsetsockopt,   
	Ngetsockopt,
	Ngetsockname,
	Ngetpeername,

	Ngethostname,
	Ngethostbyname,
	Ngethostbyaddr,
	Nioctl,

	Nsetreadtimeout,
	Ngetreadtimeout,

};

/****************************************************************
system and helper func implementations
*****************************************************************/
#include <sys/utsname.h>
int
uname( struct utsname* uname) {
	OSVERSIONINFO vi;
	SYSTEM_INFO	  si;

	vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx( &vi);

	switch ( vi.dwPlatformId ) {
	case VER_PLATFORM_WIN32s:
		uname->sysname = "Windows";
		break;
	case VER_PLATFORM_WIN32_WINDOWS:
		if ( vi.dwMinorVersion == 0 ) {
			uname->sysname = "Windows 95";
		}
		else {
			uname->sysname = "Windows 98";
		}
		break;
	case VER_PLATFORM_WIN32_NT:
		uname->sysname = "Windows NT";
		break;
#if defined(UNDER_CE)
	case VER_PLATFORM_WIN32_CE:
		uname->sysname = "Windows CE";
		break;
#endif
	default:
		uname->sysname = "Unknown";
	}

	sprintf( uname->version, "%d.%d", vi.dwMajorVersion, vi.dwMinorVersion );

	GetSystemInfo( &si);

	switch ( si.wProcessorArchitecture ) {
		case PROCESSOR_ARCHITECTURE_INTEL:
			sprintf( uname->machine, "x86" );
			break;
#if !defined(__GNUC__)
		case PROCESSOR_ARCHITECTURE_MIPS: 
			sprintf( uname->machine, "mips" );
			break;
		case PROCESSOR_ARCHITECTURE_ALPHA:
			sprintf( uname->machine, "alpha" );
			break;
		case PROCESSOR_ARCHITECTURE_PPC:
			sprintf( uname->machine, "ppc" );
			break;
#if !defined(UNDER_CE) 
		case PROCESSOR_ARCHITECTURE_SHX:
			sprintf( uname->machine, "shx" );
			break;
		case PROCESSOR_ARCHITECTURE_ARM:
			sprintf( uname->machine, "arm" );
			break;
		case PROCESSOR_ARCHITECTURE_IA64:
			sprintf( uname->machine, "ia64" );
			break;
		case PROCESSOR_ARCHITECTURE_ALPHA64:
			sprintf( uname->machine, "alpha64" );
			break;
#endif
#endif
		default:
			sprintf( uname->machine, "Unknown" );
			break;
	}

	return 1;
}



/*
 * Convert C strings to UNICODE strings.
 */
#if defined(UNDER_CE)
unsigned short*
getUnicode(const char* str)
{
	int len;
	int i;
	unsigned short* nstr;

	len = strlen(str);
	nstr = (unsigned short*)gc_malloc_fixed((len+1) * sizeof(unsigned short));
	for (i = 0; i < len; i++) {
		nstr[i] = (jchar)str[i];
	}
	nstr[i] = 0;
	return (nstr);
}

void freeUnicode(unsigned short* str)
{
	gc_free_fixed(str);
}

char*
getRegistryValue( HKEY keyroot, LPCWSTR subkey, LPCWSTR key, char* val, unsigned long len) {
	HKEY hk;
	DWORD type;
	char* c;
	unsigned short buf[255];
	unsigned long i;

	if ( RegOpenKeyEx( keyroot, subkey, 0, KEY_QUERY_VALUE, &hk) == ERROR_SUCCESS ) {
		if ( RegQueryValueEx( hk, key, NULL, &type, (unsigned char*)buf, &len) == ERROR_SUCCESS ) {
			for ( i=0, c=val; i<len; i++)
				*(c++) = (char)buf[i];
			*c = 0;
			RegCloseKey( hk);
			return (val);
		}
		else {
			RegCloseKey( hk);
		}
	}

	*val = 0;
	return (val);
}

char*
getcwd(char* path, unsigned long pathlen) {
	char* c = getRegistryValue( HKEY_CURRENT_USER, TEXT("\\Software\\Kaffe"), TEXT( "working_dir"), path, pathlen);
	return ( *c ? c : "\\windows");

#ifdef NEVER
	WIN32_FIND_DATA fd;
	HANDLE h;
	SHFILEINFO fi;
	char *c;
	unsigned short *ps, *c1;
	int i;


	*(path+pathlen-1) = '\0';

	ps = (unsigned short*)gc_malloc_fixed( pathlen * sizeof(unsigned short));
	for ( i=0, c1=ps; (unsigned long)i<pathlen-4; i+=3, c1++) {
		*c1 = (jchar)'.';
		*(++c1) = (jchar)'.';
		*(++c1) = (jchar)'\\';
	}
	*(c1++) = (jchar)'.';
	*c1 = '\0';

	c = path+pathlen-1;
	c1--;

	/*
	for ( h = FindFirstFile( c1, &fd); h != (void*)0xffffffff; h = FindFirstFile( c1, &fd)) {
		for (i=0; fd.cFileName[i]; i++);
		for ( --i; i>=0; i--) {
			*(--c) = (char)fd.cFileName[i];
		}
		*(--c) = '\\';
		c1 -= 3;
		FindClose( h);
	}
*/

	while ( SHGetFileInfo( c1, FILE_ATTRIBUTE_NORMAL, &fi, sizeof(fi), SHGFI_DISPLAYNAME )) {
		for (i=0; fi.szDisplayName[i]; i++);
		for ( --i; i>=0; i--) {
			*(--c) = fi.szDisplayName[i];
		}
		*(--c) = '\\';
		c1 -= 3;
	}

	gc_free_fixed( ps);

	return c;
#endif
}

double atof( const char* string) {
	double val;
	TCHAR *t1,*t2;
	t1 = getUnicode( string);
	val = _tcstod(t1, &t2);
	freeUnicode( t1);
	return val;
}

char* strrchr ( const char* str, int c) {
	char* c1;
	for ( c1=(char*)str; *c1; c1++) {
		if ( *c1 == c ) {
			return (c1);
		}
	}
	return NULL;
}

#endif
