/*
 * thread.h
 * Thread support.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __thread_h
#define __thread_h

#include "md.h"

#include "native.h"
#include "java_lang_Thread.h"
#include "java_lang_ThreadGroup.h"

#define	NOTIMEOUT			(0)

#define THREAD_MAXPRIO  (java_lang_Thread_MAX_PRIORITY+1)

/* We keep a little stack space reserved as a red-zone */
#if !defined(STACKREDZONE)
#define	STACKREDZONE	8192
#endif

/*
 * Interface to the thread system.
 */
JNIEXPORT void	JNICALL initThreads(void);
JNIEXPORT void	JNICALL exitThread(void);
JNIEXPORT jint	JNICALL framesThread(Hjava_lang_Thread*);
JNIEXPORT void	JNICALL setPriorityThread(Hjava_lang_Thread*, jint);
JNIEXPORT void	JNICALL startThread(Hjava_lang_Thread*);
JNIEXPORT void	JNICALL destroyThread(Hjava_lang_Thread*);
JNIEXPORT void	JNICALL finalizeThread(Hjava_lang_Thread*);
JNIEXPORT void	JNICALL interruptThread(Hjava_lang_Thread*);
JNIEXPORT void*	JNICALL getCurrentStackLimit(void);
JNIEXPORT jboolean JNICALL onCurrentThreadStack(void*);
JNIEXPORT void	JNICALL attachCurrentThread(char*);

#include "gtypes.h" /* Philippe 25-8-2000, could not compile under ARM with pthreads 
                       without this; uint32 undefined */

JNIEXPORT uint32 threadStackSize;

/*
 * The thread system must be able to dump its stack as a series of
 * records of the following format.
 */
typedef struct _stackFrame {
	void*			return_pc;
	void*			return_frame;
} stackFrame;

typedef struct {
	void*			low;
	void*			high;
	void*			curr;
} threadStackInfo;

typedef struct ThreadInterface {

	void			(*init)(void);

	void			(*attach)(Hjava_lang_Thread*);
	void			(*create)(Hjava_lang_Thread*, void*);
	void			(*yield)(void);
	void			(*setPriority)(Hjava_lang_Thread*, jint);
	void			(*exit)(Hjava_lang_Thread*);
	void*			__unused0__;
	void			(*finalize)(Hjava_lang_Thread*);

	Hjava_lang_Thread*	(*currentJava)(void);
	void*			__unused1__;

	void			(*enterCritSect)(void);
	void			(*exitCritSect)(void);

	void			(*GcWalkThreads)(void);
	void			(*GcWalkThread)(Hjava_lang_Thread*);

	void			(*nextFrame)(stackFrame*);

	void			(*stackInfo)(threadStackInfo*);

} ThreadInterface;

JNIEXPORT ThreadInterface Kaffe_ThreadInterface;


/*
 * These are the generic macros to access the interface functions (use them
 * instead of the function pointers)
 */
#define Kaffe_TI_init()                 (*Kaffe_ThreadInterface.init)()
#define	Kaffe_TI_attach(_thread)	(*Kaffe_ThreadInterface.attach)(_thread)
#define Kaffe_TI_create(_thread,_func)  (*Kaffe_ThreadInterface.create)(_thread,_func)
#define Kaffe_TI_yield()                (*Kaffe_ThreadInterface.yield)()
#define Kaffe_TI_setPriority(_thread,_prio) (*Kaffe_ThreadInterface.setPriority)(_thread,_prio)
#define Kaffe_TI_exit(_thread)          (*Kaffe_ThreadInterface.exit)(_thread)
#define Kaffe_TI_finalize(_thread)      (*Kaffe_ThreadInterface.finalize)(_thread)
#define Kaffe_TI_currentJava()          (*Kaffe_ThreadInterface.currentJava)()
#define Kaffe_TI_enterCritSect()        (*Kaffe_ThreadInterface.enterCritSect)()
#define Kaffe_TI_exitCritSect()         (*Kaffe_ThreadInterface.exitCritSect)()
#define Kaffe_TI_GcWalkThreads()        (*Kaffe_ThreadInterface.GcWalkThreads)()
#define Kaffe_TI_GcWalkThread(_thread)  (*Kaffe_ThreadInterface.GcWalkThread)(_thread)
#define Kaffe_TI_nextFrame(_frame)      (*Kaffe_ThreadInterface.nextFrame)(_frame)
#define Kaffe_TI_stackInfo(_stckInfo)   (*Kaffe_ThreadInterface.stackInfo)(_stckInfo)

/*
 * Inject the ThreadInterface implementation header, which can be used
 * to turn the runtime interface into a compile-time interface (by means
 * of overloading the generic macros)
 */
#include "thread-impl.h" 


#endif
