/**
 * gra.c - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

void *
Java_java_awt_Toolkit_graInitGraphics(JNIEnv * env, jclass clazz,
                                      Graphics * gr, jobject tgt, jint tgtType, jboolean isVisible,
                                      jint xOff, jint yOff,
                                      jint xClip, jint yClip, jint wClip, jint hClip,
                                      jobject fnt, jint fg, jint bg )
{
	Drawable drw;
	XGCValues values;
	Image *img = 0;
	int alpha;
	unsigned long valueMask = GCForeground | GCBackground |
		GCFunction | GCGraphicsExposures;

	DBG(awt_gra, ("initGraphics: %x (%x, %d) %d,%d  %d,%d - %d,%d  %x %x %x\n",
				  gr, tgt, tgtType, xOff, yOff, xClip, yClip, wClip, hClip, fnt, fg, bg));

	if (tgt) {
		switch (tgtType) {
		case WINDOW:
			drw = (Drawable) tgt;
			break;                           /* Window   */
		case IMAGE:
			img = (Image*)tgt;
			drw = ((Image *) tgt)->pix;
			break;                           /* Image    */
		case GRAPHICS:
			drw = ((Graphics *) tgt)->drw;
			/* the visibility of our ref graphics is unknown to Java - we have to deduce it */
			isVisible = (((Graphics *)tgt)->x0 >= -10000);
			break;                           /* Graphics */
		default:
			fprintf(stderr, "illegal Graphics target: %d\n", tgtType);
			drw = 0;
		}
	}
	else {
		if (gr)
			drw = gr->drw;
		else {
			fprintf(stderr, "attempt to set an uninitialized graphics");
			drw = 0;
		}
	}

#if defined (DRAWABLE_ALPHA)
	alpha = getAlpha( fg);
	fg = stripAlpha( fg);
	bg = stripAlpha( bg);
#endif

	values.foreground = fg;
	values.background = bg;

    if ( fnt ){
	  values.font = ((XFontStruct *) fnt)->fid;
	  valueMask |= GCFont;
	}
	values.function = GXcopy;
	values.graphics_exposures = False;

	if (gr) {
		XChangeGC(X->dsp, gr->gc, valueMask, &values);
	}
	else {
		gr = (Graphics *) AWT_MALLOC(sizeof(Graphics));
		gr->gc = XCreateGC(X->dsp, drw, valueMask, &values);

		DBG(awt_gra, (" ->gr: %x (%x)\n", gr, gr->gc));
	}

#if defined (DRAWABLE_ALPHA)
	gr->img = img;
	gr->alpha = alpha;
#endif

	gr->fg = fg;
	gr->bg = bg;
	gr->drw = drw;
	gr->x0 = xOff;
	gr->y0 = yOff;
	gr->xor = 0;

	/*
	 * We have to take care of making sure that clip_mask
	 * in the graphics structure is kept None as much as 
	 * possible. We also have to make sure that the clip
	 * mask in GC is reset during the obect initaization.
	 */

        if (gr->clip_mask != None) { 
                XFreePixmap(X->dsp, gr->clip_mask);
                gr->clip_mask = None;
        } 
        XSetClipMask(X->dsp, gr->gc, None);

	gr->clip.x = xClip;
	gr->clip.y = yClip;
	gr->clip.width = (wClip > 0) ? wClip : 0;
	gr->clip.height = (hClip > 0) ? hClip : 0;
	XSetClipRectangles(X->dsp, gr->gc, gr->x0, gr->y0, &gr->clip, 1, Unsorted);

	if ( isVisible == JNI_FALSE ){
		/* this has to be consistent with graSetVisible */
		gr->x0 -= 100000;
	}
	
	return gr;
}

void
Java_java_awt_Toolkit_graSetVisible(JNIEnv * env, jclass clazz, Graphics * gr, jint isVisible)
{
	DBG(awt_gra, ("setVisble: %x  %d\n", gr, isVisible));
	
	/*
	 * This is rather a hack to "defuse" a Graphics object, but we don't want to
	 * add checks on every draw op to test if the target is invisible
	 */
	if (!isVisible) {
		if (gr->x0 >= -10000)
			gr->x0 -= 100000;
	}
	else {
		if (gr->x0 <= -10000)
			gr->x0 += 100000;
	}
}


void
Java_java_awt_Toolkit_graFreeGraphics(JNIEnv * env, jclass clazz, Graphics * gr)
{
	DBG(awt_gra, ("freeGraphics: %x\n", gr));

	XFreeGC(X->dsp, gr->gc);
	AWT_FREE(gr);
}


void
Java_java_awt_Toolkit_graCopyArea(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x, jint y, jint width, jint height,
                                  jint xDelta, jint yDelta)
{
	DBG(awt_gra, ("copyArea: %x, %d,%d-%d,%d  %d,%d\n", gr, x, y, width, height, xDelta, yDelta));

	/* some X servers act alleric on that (careless bitblt, again) */
	if ((width < 0) || (height < 0))
		return;

	x += gr->x0;
	y += gr->y0;
	XCopyArea(X->dsp, gr->drw, gr->drw, gr->gc, x, y, width, height, x + xDelta, y + yDelta);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graClearRect(JNIEnv * env, jclass clazz, Graphics * gr,
								   jint x, jint y, jint width, jint height, jint bg)
{
	DBG(awt_gra, ("clearRect: %x, %d,%d-%d,%d %x\n", gr, x, y, width, height, bg));
	gr->bg = bg;
	XSetForeground(X->dsp, gr->gc, gr->bg);
	XFillRectangle(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, width, height);
	XSetForeground(X->dsp, gr->gc, gr->fg);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawBytes(JNIEnv * env, jclass clazz, Graphics * gr,
								   jarray jBytes, jint offset, jint len, jint x, jint y)
{
	jboolean isCopy;
	int n;
	jbyte *jb, *jbo;

	if (!jBytes)
		return;

	n = (*env)->GetArrayLength(env, jBytes);
	jb = (*env)->GetByteArrayElements(env, jBytes, &isCopy);
	jbo = jb + offset;

	DBG(awt_gra, ("drawBytes: %x, %x,%d,%d  \"%s\"  %d,%d\n", gr,
				  jBytes, offset, len, jb, x, y));

	if (offset + len > n)
		len = n - offset;

	XDrawString(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, jbo, len);

	(*env)->ReleaseByteArrayElements(env, jBytes, jb, JNI_ABORT);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawChars(JNIEnv * env, jclass clazz, Graphics * gr,
								   jarray jChars, jint offset, jint len, jint x, jint y)
{
	jboolean isCopy;
	int n;
	jchar *jc, *jco;
	XChar2b *b;

	if (!jChars)
		return;

	n = (*env)->GetArrayLength(env, jChars);
	jc = (*env)->GetCharArrayElements(env, jChars, &isCopy);
	jco = jc + offset;

	DBG(awt_gra, ("drawChars: %x, %x,%d,%d  \"%s\"  %d,%d\n", gr, jChars, offset, len,
				  jchar2CString(X, jco, len), x, y));

	if (offset + len > n)
		len = n - offset;

#ifndef WORDS_BIGENDIAN
	n = sizeof(XChar2b) * len;
	b = (XChar2b *) getBuffer(X, n);
	swab((void *) jco, (void *) b, n);
#else
	b = (XChar2b *) jco;
#endif

	XDrawString16(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, b, len);

	(*env)->ReleaseCharArrayElements(env, jChars, jc, JNI_ABORT);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawString(JNIEnv * env, jclass clazz,
									Graphics * gr, jstring str, jint x, jint y)
{
	jboolean isCopy;
	int n, len;
	const jchar *jc;
	XChar2b *b;

	DBG(awt_gra, ("drawString: %x  \"%s\"  %d,%d\n", gr, java2CString(env, X, str), x, y));

	if (!str)
		return;

	len = (*env)->GetStringLength(env, str);
	jc = (*env)->GetStringChars(env, str, &isCopy);

#ifndef WORDS_BIGENDIAN
	n = sizeof(XChar2b) * len;
	b = (XChar2b *) getBuffer(X, n);
	swab(jc, b, n);
#else
	b = (XChar2b *) jc;
#endif

	XDrawString16(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, b, len);

	(*env)->ReleaseStringChars(env, str, jc);
	XFLUSH(X, False);
}

void
Java_java_awt_Toolkit_graDrawVertString(JNIEnv * env, jclass clazz,
									Graphics * gr, jstring str, jint x, jint y)
{
        /* not implemented yet */
}


void
Java_java_awt_Toolkit_graDrawLine(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x1, jint y1, jint x2, jint y2)
{
	DBG(awt_gra, ("drawLine: %x, %d,%d - %d,%d\n", gr, x1, y1, x2, y2));

	if ((x1 == x2) && (y1 == y2)) {      /* swing gimmick */
		XDrawPoint(X->dsp, gr->drw, gr->gc, x1 + gr->x0, y1 + gr->y0);
	}
	else {
		XDrawLine(X->dsp, gr->drw, gr->gc, x1 + gr->x0, y1 + gr->y0, x2 + gr->x0, y2 + gr->y0);
	}
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawArc(JNIEnv * env, jclass clazz, Graphics * gr,
                                 jint x, jint y, jint width, jint height,
                                 jint startAngle, jint arcAngle)
{
	DBG(awt_gra, ("drawArc: %x, %d,%d-%d,%d  %d,%d\n", gr,
				  x, y, width, height, startAngle, arcAngle));

	XDrawArc(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, width, height,
			 startAngle << 6, arcAngle << 6);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graFillArc(JNIEnv * env, jclass clazz, Graphics * gr,
                                 jint x, jint y, jint width, jint height,
                                 jint startAngle, jint arcAngle)
{
	DBG(awt_gra, ("fillArc: %x, %d,%d-%d,%d  %d,%d\n", gr,
				  x, y, width, height, startAngle, arcAngle));

	XFillArc(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, width, height,
			 startAngle << 6, arcAngle << 6);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawOval(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x, jint y, jint width, jint height)
{
	DBG(awt_gra, ("drawOval: %x, %d,%d - %d,%d\n", gr, x, y, width, height));

	XDrawArc(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, width, height, 0, 23040);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graFillOval(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x, jint y, jint width, jint height)
{
	DBG(awt_gra, ("fillOval: %x, %d,%d - %d,%d\n", gr, x, y, width, height));

	XFillArc(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, width, height, 0, 23040);
	XFLUSH(X, False);
}


int 
jarray2Points(JNIEnv * env, Toolkit * X, XPoint ** pp, int x0, int y0,
              jarray xPoints, jarray yPoints, int nPoints)
{
	register int i;
	int n;
	jboolean isCopy;
	jint *jx = (*env)->GetIntArrayElements(env, xPoints, &isCopy);
	jint *jy = (*env)->GetIntArrayElements(env, yPoints, &isCopy);

	if ((n = (*env)->GetArrayLength(env, xPoints)) < nPoints)
		nPoints = n;
	if ((n = (*env)->GetArrayLength(env, yPoints)) < nPoints)
		nPoints = n;

	*pp = getBuffer(X, (nPoints + 1) * sizeof(XPoint));

	for (i = 0; i < nPoints; i++) {
		(*pp)[i].x = jx[i] + x0;
		(*pp)[i].y = jy[i] + y0;
	}

	(*env)->ReleaseIntArrayElements(env, xPoints, jx, JNI_ABORT);
	(*env)->ReleaseIntArrayElements(env, yPoints, jy, JNI_ABORT);

	return nPoints;
}


void
Java_java_awt_Toolkit_graDrawPolygon(JNIEnv * env, jclass clazz, Graphics * gr,
									 jarray xPoints, jarray yPoints, jint nPoints)
{
	int n;
	XPoint *p;

	DBG(awt_gra, ("drawPolygon: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));

	if (!xPoints || !yPoints)
		return;

	nPoints = jarray2Points(env, X, &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);

	n = nPoints - 1;
	if ((p[0].x != p[n].x) || (p[0].y != p[n].y)) {
		p[nPoints].x = p[0].x;
		p[nPoints].y = p[0].y;
		nPoints++;
	}

	XDrawLines(X->dsp, gr->drw, gr->gc, p, nPoints, CoordModeOrigin);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawPolyline(JNIEnv * env, jclass clazz, Graphics * gr,
									  jarray xPoints, jarray yPoints, jint nPoints)
{
	XPoint *p;

	DBG(awt_gra, ("drawPolyline: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));

	if (!xPoints || !yPoints)
		return;

	nPoints = jarray2Points(env, X, &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);
	XDrawLines(X->dsp, gr->drw, gr->gc, p, nPoints, CoordModeOrigin);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graFillPolygon(JNIEnv * env, jclass clazz, Graphics * gr,
									 jarray xPoints, jarray yPoints, jint nPoints)
{
	XPoint *p;

	DBG(awt_gra, ("fillPolygon: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));

	if (!xPoints || !yPoints)
		return;

	nPoints = jarray2Points(env, X, &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);
	XFillPolygon(X->dsp, gr->drw, gr->gc, p, nPoints, Nonconvex, CoordModeOrigin);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDrawRect(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x, jint y, jint width, jint height)
{
	DBG(awt_gra, ("drawRect: %x, %d,%d - %d,%d\n", gr, x, y, width, height));

	if ((width >= 0) && (height >= 0))
		XDrawRectangle(X->dsp, gr->drw, gr->gc, x + gr->x0, y + gr->y0, width, height);
	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graFillRect(JNIEnv * env, jclass clazz, Graphics * gr,
                                  jint x, jint y, jint width, jint height)
{
	DBG(awt_gra, ("fillRect: %x, %d,%d - %d,%d\n", gr, x, y, width, height));

	x += gr->x0; y += gr->y0;

	if ((width >= 0) && (height >= 0)) {
		if (width == 1) {                  /* some swing gimmicks */
			XDrawLine(X->dsp, gr->drw, gr->gc, x, y, x, y + height - 1);
		}
		else if (height == 1) {
			XDrawLine(X->dsp, gr->drw, gr->gc, x, y, x + width - 1, y);
		}
		else {
			XFillRectangle(X->dsp, gr->drw, gr->gc, x, y, width, height);

#if defined (DRAWABLE_ALPHA)
			if ( gr->img && gr->img->alpha ){
				fillAlphaImage( gr->img->alpha, x, y, x+width-1, y+height-1, gr->alpha);
			}
#endif
		}
	}
	XFLUSH(X, False);
}



/*
 * we don't use the Xmu routines here because it would drag the whole Xt libs in
 */
void
Java_java_awt_Toolkit_graDrawRoundRect(JNIEnv * env, jclass clazz, Graphics * gr,
									   jint x, jint y, jint width, jint height,
                                       jint wArc, jint hArc)
{
	int x1, x2, y1, y2, a, b;

	DBG(awt_gra, ("drawRoundRect: %x, %d,%d - %d,%d  %d,%d\n", gr,
				  x, y, width, height, wArc, hArc));

	x += gr->x0;
	y += gr->y0;
	a = wArc / 2;
	b = hArc / 2;

	x1 = x + a;
	x2 = x + width - a;
	y2 = y + height;
	XDrawLine(X->dsp, gr->drw, gr->gc, x1, y, x2, y);
	XDrawLine(X->dsp, gr->drw, gr->gc, x1, y2, x2, y2);

	y1 = y + b;
	y2 = y + height - b;
	x2 = x + width;
	XDrawLine(X->dsp, gr->drw, gr->gc, x, y1, x, y2);
	XDrawLine(X->dsp, gr->drw, gr->gc, x2, y1, x2, y2);

	XDrawArc(X->dsp, gr->drw, gr->gc, x, y, wArc, hArc, 90 * 64, 90 * 64);

	x2 = x + width - wArc;
	XDrawArc(X->dsp, gr->drw, gr->gc, x2, y, wArc, hArc, 0, 90 * 64);

	y2 = y + height - hArc;
	XDrawArc(X->dsp, gr->drw, gr->gc, x2, y2, wArc, hArc, 0, -90 * 64);

	XDrawArc(X->dsp, gr->drw, gr->gc, x, y2, wArc, hArc, 180 * 64, 90 * 64);

	XFLUSH(X, False);
}

void
Java_java_awt_Toolkit_graFillRoundRect(JNIEnv * env, jclass clazz, Graphics * gr,
									   jint x, jint y, jint width, jint height,
                                       jint wArc, jint hArc)
{
	int x1, x2, y1, y2, a, b;

	DBG(awt_gra, ("fillRoundRect: %x, %d,%d - %d,%d  %d,%d\n", gr,
				  x, y, width, height, wArc, hArc));

	x += gr->x0;
	y += gr->y0;
	a = wArc / 2;
	b = hArc / 2;

	y1 = y + b;
	y2 = y + height - b;
	if (y2 > y1)
		XFillRectangle(X->dsp, gr->drw, gr->gc, x, y1, width, y2 - y1);

	x1 = x + a;
	x2 = x + width - a;
	if (x2 > x1) {
		XFillRectangle(X->dsp, gr->drw, gr->gc, x1, y, (x2 - x1), b);
		XFillRectangle(X->dsp, gr->drw, gr->gc, x1, y2, (x2 - x1), b);
	}

	XFillArc(X->dsp, gr->drw, gr->gc, x, y, wArc, hArc, 90 * 64, 90 * 64);

	x2 = x + width - wArc;
	XFillArc(X->dsp, gr->drw, gr->gc, x2, y, wArc, hArc, 0, 90 * 64);

	y2 = y + height - hArc;
	XFillArc(X->dsp, gr->drw, gr->gc, x2, y2, wArc, hArc, 0, -90 * 64);

	XFillArc(X->dsp, gr->drw, gr->gc, x, y2, wArc, hArc, 180 * 64, 90 * 64);

	XFLUSH(X, False);
}


void
Java_java_awt_Toolkit_graDraw3DRect(JNIEnv * env, jclass clazz, Graphics * gr,
									jint x, jint y, jint width, jint height,
                                    jboolean raised, jint pixelValue)
{
	int bright, dark;
	int xw, yh;

	DBG(awt_gra, ("draw3DRect: %x, %d,%d - %d,%d  %d %x\n", gr,
				  x, y, width, height, raised, pixelValue));

	/* we should use the same mechanism like ordinary Graphics brighter()/darker() here */
	dark   = (int) modPixelValue( X, pixelValue, DARKER);
	bright = (int) modPixelValue( X, pixelValue, BRIGHTER);

	x += gr->x0;
	y += gr->y0;
	xw = x + width;
	yh = y + height;

	XSetForeground(X->dsp, gr->gc, (raised ? bright : dark));
	XDrawLine(X->dsp, gr->drw, gr->gc, x, y, xw - 1, y);
	XDrawLine(X->dsp, gr->drw, gr->gc, x, y, x, yh);

	XSetForeground(X->dsp, gr->gc, (raised ? dark : bright));
	XDrawLine(X->dsp, gr->drw, gr->gc, x + 1, yh, xw, yh);
	XDrawLine(X->dsp, gr->drw, gr->gc, xw, y, xw, yh);

	XSetForeground(X->dsp, gr->gc, gr->fg);

	XFLUSH(X, False);
}

void
Java_java_awt_Toolkit_graFill3DRect(JNIEnv * env, jclass clazz, Graphics * gr,
									jint x, jint y, jint width, jint height,
                                    jboolean raised, jint rgb)
{
	DBG(awt_gra, ("fill3DRect: %x, %d,%d - %d,%d  %d %x\n", gr,
				  x, y, width, height, raised, pixelValue));

	XFillRectangle(X->dsp, gr->drw, gr->gc, x + gr->x0 + 1, y + gr->y0 + 1, width - 2, height - 2);

	Java_java_awt_Toolkit_graDraw3DRect(env, clazz, gr, x, y, width - 1, height - 1, raised, rgb);
}


void
Java_java_awt_Toolkit_graAddClip(JNIEnv * env, jclass clazz, Graphics * gr,
								 jint xClip, jint yClip, jint wClip, jint hClip)
{

        GC gcerase, gcdraw;
        XGCValues gcvalues;

	DBG(awt_gra, ("addClip: %x, %d,%d - %d,%d\n", gr, xClip, yClip, wClip, hClip));

	/*
	 * This is actually one of the very few places that makes use of the 
	 * graphics structure variable Pixmap to store the pervious clipmask
	 * and add the new clipping. This is because the Xlib does not allow
	 * (explicitely stated in documentation) to recover the clip mask.
	 */

        if (gr->clip_mask == None) {
                gr->clip_mask = XCreatePixmap(X->dsp, gr->drw, gr->clip.width, gr->clip.height, 1);
                gcerase = XCreateGC(X->dsp, gr->clip_mask, 0L, &gcvalues);
                XSetForeground(X->dsp, gcerase, 1L);
                XSetBackground(X->dsp, gcerase, 0L);
                XFillRectangle(X->dsp, gr->clip_mask, gcerase, 0, 0, gr->clip.width, gr->clip.height);
                XFreeGC(X->dsp,gcerase);
        }

        gcdraw = XCreateGC(X->dsp, gr->clip_mask, 0L, &gcvalues);
        XSetForeground(X->dsp, gcdraw, 0L);
        XSetBackground(X->dsp, gcdraw, 1L);
        XFillRectangle(X->dsp, gr->clip_mask, gcdraw, xClip, yClip, wClip, hClip);
        XSetClipMask(X->dsp, gr->gc, gr->clip_mask);
        XFreeGC(X->dsp,gcdraw);


}


void
Java_java_awt_Toolkit_graSetClip(JNIEnv * env, jclass clazz, Graphics * gr,
								 jint xClip, jint yClip, jint wClip, jint hClip)
{
	DBG(awt_gra, ("setClip: %x, %d,%d - %d,%d\n", gr, xClip, yClip, wClip, hClip));

	/* 
	 * This is the change about keeping clip mask in the graphics structure
	 * None as much as possible. Look to the comments in graAddClip for more
	 * comments on the clip mask.
	 */

        if (gr->clip_mask != None) { 
                XFreePixmap(X->dsp, gr->clip_mask);
                gr->clip_mask = None;
                XSetClipMask(X->dsp, gr->gc, gr->clip_mask);
        }

	gr->clip.x = xClip;
	gr->clip.y = yClip;
	gr->clip.width = (wClip > 0) ? wClip : 0;
	gr->clip.height = (hClip > 0) ? hClip : 0;
	XSetClipRectangles(X->dsp, gr->gc, gr->x0, gr->y0, &gr->clip, 1, Unsorted);
}

void
Java_java_awt_Toolkit_graSetColor(JNIEnv * env, jclass clazz, Graphics * gr, jint clr)
{
	DBG(awt_gra, ("setColor: %x, %x\n", gr, clr));

#if defined (DRAWABLE_ALPHA)
	gr->alpha = getAlpha( clr);
	clr = stripAlpha( clr);

	if ( gr->img && (gr->alpha != 0xff) && !gr->img->alpha ){
		createAlphaImage( X, gr->img);
	}
#endif

	gr->fg = clr;

	if (gr->xor)
		clr ^= gr->xclr;

	XSetForeground(X->dsp, gr->gc, clr);
}

void
Java_java_awt_Toolkit_graSetBackColor(JNIEnv * env, jclass clazz, Graphics * gr, jint clr)
{
	DBG(awt_gra, ("setBackColor: %x, %x\n", gr, clr));

#if defined(DRAWABLE_ALPHA)
	clr = stripAlpha( clr);
#endif

	gr->bg = clr;

	XSetBackground(X->dsp, gr->gc, clr);
}

void
Java_java_awt_Toolkit_graSetFont(JNIEnv * env, jclass clazz, Graphics * gr, jobject fnt)
{
	DBG(awt_gra, ("setFont: %x, %x\n", gr, fnt));

	XSetFont(X->dsp, gr->gc, ((XFontStruct *) fnt)->fid);
}

void
Java_java_awt_Toolkit_graSetOffset(JNIEnv * env, jclass clazz, Graphics * gr, jint xOff, jint yOff)
{
	DBG(awt_gra, ("setOffset: %x, %d,%d\n", gr, xOff, yOff));

	/* Note that changing the origin doesn't change the clipping rectangle
	 * (again behavior, not spec). We have to update the clip accordingly,
	 * so that we keep the same physical clip origin
	 */
	gr->clip.x += (gr->x0 - xOff);
	gr->clip.y += (gr->y0 - yOff);

	gr->x0 = xOff;
	gr->y0 = yOff;
}

void
Java_java_awt_Toolkit_graSetPaintMode(JNIEnv * env, jclass clazz, Graphics * gr)
{
	DBG(awt_gra, ("setPaintMode: %x\n", gr));

	gr->xor = 0;
	XSetForeground(X->dsp, gr->gc, gr->fg);
	XSetFunction(X->dsp, gr->gc, GXcopy);
}

void
Java_java_awt_Toolkit_graSetXORMode(JNIEnv * env, jclass clazz, Graphics * gr, jint xorClr)
{
	DBG(awt_gra, ("setXORMode: %x, %x\n", gr, xorClr));

	gr->xor = 1;
	gr->xclr = stripAlpha( xorClr);
	XSetForeground( X->dsp, gr->gc, gr->fg ^ gr->xclr);
	XSetFunction( X->dsp, gr->gc, GXxor);
}


/************************************************************************************
 * image rendering - internal (representation dependent) drawing routines
 */

static inline void
drawXImage (Toolkit* X, Graphics *gr, Image *img,
            jint srcX, jint srcY, jint dstX, jint dstY,
            jint width, jint height)
{
#if defined(HAVE_LIBXEXT)
	if (img->shmiImg) {
		XShmPutImage(X->dsp, gr->drw, gr->gc, img->xImg,
					 srcX, srcY, dstX, dstY, width, height, False);
	}
	else
#endif /* HAVE_LIBXEXT */
		{
			XPutImage(X->dsp, gr->drw, gr->gc, img->xImg,
					  srcX, srcY, dstX, dstY, width, height);
		}
}

static void
drawAlphaImage(Toolkit * X, Graphics * gr, Image * img,
               jint srcX, jint srcY, jint dstX, jint dstY,
               jint width, jint height, jint bgval)
{
	Image *dstImg = 0;
	XImage *dstXim = 0;
	int i, j, si, sj, alpha;
	unsigned long dpix, spix, bgpix = 0;
	int sr, sg, sb, dr, dg, db;
	int x, y, xw, yh, w, h;

	DBG(awt_gra, ("drawAlphaImage: %x %x (%x, %x),  %d,%d  %d,%d  %d,%d  %x\n",
				  gr, img, img->xImg, img->alpha, srcX, srcY, dstX, dstY, width, height, bgval));

	if (!img)
		return;

	/* for whatever reason X system does not allow us to read memory
	 * from outside of the clipping rectange. This forces us to perform
	 * clipping here and makes us very unhappy. The problem lies in the
	 * calls to XShmGetImage and XGetImage.
	 */

	/* pick up the max of clipping rectangle and the image */
	x = (dstX > gr->clip.x) ? dstX : gr->clip.x;
	y = (dstY > gr->clip.y) ? dstY : gr->clip.y;

	/* pick up the min of clipping rectangle and the image */
	xw = (dstX + width  < gr->clip.x + gr->clip.width ) ? dstX + width  : gr->clip.x + gr->clip.width;
	yh = (dstY + height < gr->clip.y + gr->clip.height) ? dstY + height : gr->clip.y + gr->clip.height;
	
	w = xw - x;
	h = yh - y;

	if (w <= 0 || h <= 0)
		return;


#if defined(HAVE_LIBXEXT)
	if (X->shm == USE_SHM) {
		dstImg = createImage(w, h, JNI_FALSE);
		if (dstImg->shmiImg) {
			dstXim = dstImg->xImg;
			XShmGetImage(X->dsp, gr->drw, dstXim, x, y, 0xffffffff);
		}
		else {                             /* Shm failed, backup to normal XImage */
			Java_java_awt_Toolkit_imgFreeImage(0, 0, dstImg);
			dstImg = 0;
		}
	}
#endif /* HAVE_LIBXEXT */

	if (!dstXim) {
		dstXim = XGetImage(X->dsp, gr->drw, x, y, w, h, 0xffffffff, ZPixmap);
	}

	if (bgval != -1)
		bgpix = pixelValue(X, bgval);

	if (dstXim) {
		for (j = 0, sj = srcY - dstY + y; j < dstXim->height; j++, sj++) {
			for (i = 0, si = srcX - dstX + x; i < dstXim->width; i++, si++) {
				dpix = (bgval == -1) ? XGetPixel(dstXim, i, j) : bgpix;

				spix = XGetPixel(img->xImg, si, sj);
				alpha = GetAlpha(img->alpha, si, sj);

				rgbValues(X, dpix, &dr, &dg, &db);
				rgbValues(X, spix, &sr, &sg, &sb);

				if (alpha) {
					/*
					 * ints still considered to be substantially faster than floats on most
					 * architectures (avoid div by 255), map all 0..255 -> 1..256
					 */
					alpha++;
					sr++; sg++; sb++;
					dr++; dg++; db++;


					dr = ((alpha * sr + (256 - alpha) * dr) + 128) >> 8;
					dg = ((alpha * sg + (256 - alpha) * dg) + 128) >> 8;
					db = ((alpha * sb + (256 - alpha) * db) + 128) >> 8;

					dr--; dg--; db--;

					dpix = pixelValue(X, (dr << 16) | (dg << 8) | (db));
					XPutPixel(dstXim, i, j, stripAlpha( dpix));
				}
				else {
				}
			}
		}

		if (dstImg != 0) {
#if defined(HAVE_LIBXEXT)
			XShmPutImage(X->dsp, gr->drw, gr->gc, dstXim, 0, 0, x, y, w, h, False);
			XSync(X->dsp, False);
			Java_java_awt_Toolkit_imgFreeImage(0, 0, dstImg);
#endif /* HAVE_LIBXEXT */
		}
		else {
			XPutImage(X->dsp, gr->drw, gr->gc, dstXim, 0, 0, x, y, w, h);
			XFlush(X->dsp);
			XDestroyImage(dstXim);
		}

		XFLUSH(X, True);
	}
}

/*
 * Draw a masked pixmap. This is considered to be our workhorse for
 * small, frequently drawn images. It uses a trick to avoid drawing
 * the mask *and* the pixmap: the mask is used as a clipmask, not to
 * punch out visible destination pels. But this means that we have to
 * take care of clipping issues by ourselves (explicitly). Note that
 * we don't use the gr->gc, but the global X->gcDefault, so that we
 * don't have to tinker around with the gr->gc state
 */
static void
drawMaskedPixmap(Toolkit* X, Graphics *gr, Image *img,
                 int srcX, int srcY, int dstX, int dstY,
                 int width, int height, int bgval)
{
	int cx, cy, cxw, cyh, dxw, dyh, dx, dy, d;

	dx = dstX;
	dy = dstY;

	cx = gr->x0 + gr->clip.x;
	cy = gr->y0 + gr->clip.y;
	cxw = cx + gr->clip.width;
	cyh = cy + gr->clip.height;
	dxw = dx + width;
	dyh = dy + height;

	/* check out if the destination has to be clipped */
	if ( cx > dx ){
		d = cx - dx;
		dx = cx;
		srcX += d;
		width -= d;
	}

	if ( cy > dy ){
		d = cy - dy;
		dy = cy;
		srcY += d;
		height -= d;
	}

	if ( cxw < dxw ){
		width -= (dxw - cxw);
	}

	if ( cyh < dyh ){
		height -= (dyh - cyh);
	}

	if ( (width > 0) && (height > 0) ) {
		/* use the mask as a (partial) clipmask */
		XSetClipOrigin( X->dsp, X->gcScreen, dstX, dstY);
		XSetClipMask( X->dsp, X->gcScreen, img->pixMask);
				
		/* blit the gr->gc clipped parts of the source image */
		XCopyArea(X->dsp, img->pix, gr->drw, X->gcScreen,
				  srcX, srcY, width, height, dx, dy);
	}
}

/*
 * Draw a masked XImage. We first punch out (black) the non-
 * transparent parts of the image (in the destination area), and
 * then draw the actual image data GXored (which is safe because all
 * invisible pels are produced black in the xImg). Since it requires
 * a lot of GC state changes, and two XPutImages, this really isn't
 * fast enough (which is why we use pixmaps, if we can)
 */
static void
drawMaskedXImage(Toolkit* X, Graphics *gr, Image *img,
                 jint srcX, jint srcY, jint dstX, jint dstY,
                 jint width, jint height, jint bgval)
{
	XGCValues values;

	/* save current gc values, we have to change them temporarily */
	XGetGCValues(X->dsp, gr->gc, GCFunction | GCPlaneMask | GCForeground | GCBackground,
				 &values);

	/* in case we have a specified background color, fill area */
	if (bgval != -1) {
		XSetForeground(X->dsp, gr->gc, bgval);
		XFillRectangle(X->dsp, gr->drw, gr->gc, dstX, dstY, width, height);
	}

	/* set attributes so that a subsequent image draw leaves the mask pels unmodified */
	XSetFunction(X->dsp, gr->gc, GXandInverted);
	XSetForeground(X->dsp, gr->gc, 0xffffffff);
	XSetBackground(X->dsp, gr->gc, 0);

	/* draw the mask bitmap */
#if defined(HAVE_LIBXEXT)
	if (img->shmiMask) {
		XShmPutImage(X->dsp, gr->drw, gr->gc, img->xMask,
					 srcX, srcY, dstX, dstY, width, height, False);
	}
	else
#endif /* HAVE_LIBXEXT */
		{
			XPutImage(X->dsp, gr->drw, gr->gc, img->xMask,
					  srcX, srcY, dstX, dstY, width, height);
		}

	/* restore gc values except of the function */
	XChangeGC(X->dsp, gr->gc, GCFunction | GCPlaneMask | GCForeground | GCBackground, &values);
	XSetFunction(X->dsp, gr->gc, GXor);

	/* draw the real data */
	drawXImage(X, gr, img, srcX, srcY, dstX, dstY, width, height);

	XSetFunction(X->dsp, gr->gc, values.function);
}


/************************************************************************************
 * image rendering - exported functions
 */

void Java_java_awt_Toolkit_graDrawImage( JNIEnv * env, jclass clazz, Graphics* gr, Image * img,
                                         jint srcX, jint srcY,
                                         jint dstX, jint dstY,
                                         jint width, jint height, jint bgval);


/*
 * create a offscreen buffer image suitable for rendering delta frames
 */
static inline Image*
getBufferImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img, int x, int y )
{
	Image    *buffer, *im;

	/* do we already have a buffer offscreen image? If not, create it and
	 * set it for all frames
	 */
	if ( !(buffer = img->buffer) ){
		img->buffer = Java_java_awt_Toolkit_imgCreateScreenImage(
		                               env, clazz, img->width, img->height);
		buffer = img->buffer;
		
		for ( im = img->next; im != img; im=im->next){
			im->buffer = img->buffer;
		}
		
		/* create a graphics for this buffer, so that we don't have to re-init
		 * each and every time (might be costly, and it's all invariant)
		 */
		buffer->gr = Java_java_awt_Toolkit_graInitGraphics( env, clazz,
		                                  0, buffer, IMAGE, 1, 0, 0,
		                                  0, 0, buffer->width, buffer->height,
		                                  0, 0, 1);

		/* check if the first frame is already transparent, and fill with
		 * the target background if it is
		 */

		if ( img->trans != -1 ){
			XCopyArea( X->dsp, gr->drw, buffer->pix, buffer->gr->gc,
					  x+gr->x0, y+gr->y0, buffer->width, buffer->height, 0, 0);
		}
	}

	/* draw the current frame into the buffer */
	Java_java_awt_Toolkit_graDrawImage( env, clazz,
	                                   buffer->gr, img, 0, 0,
	                                   0, 0, img->width, img->height, -1);

	return buffer;
}

void
Java_java_awt_Toolkit_graDrawImage(JNIEnv * env, jclass clazz, Graphics * gr, Image * img,
                                   jint srcX, jint srcY,
                                   jint dstX, jint dstY,
                                   jint width, jint height, jint bgval)
{
	DBG(awt_gra, ("drawImage: %x %x (%x,%x,%x %d,%d) %d,%d, %d,%d, %d,%d, %x\n",
				  gr, img, img->xImg, img->xMask, img->alpha, img->width, img->height,
				  srcX, srcY, dstX, dstY, width, height, bgval));

	if ( (img == 0) || (img->width == -1) ){ /* last defense sanity check */
		return;
	}

	if ( (img->flags & IMG_DRAWN) == 0 ){
		setImageState( X, img, IMG_DRAWN);
	}
	
	/* If this is a delta frame animation, we need to draw the new frame into
	 * a buffer, and then blit this buffer onto the screen. This results in
	 * two blits per frame, but the delta is probably small enough to be fast,
	 * and the overall costs significantly less than drawing from frame0
	 * (note that this scheme is also useful for video frames)
	 * Make sure we don't get infinite recursive because graDrawImage is also
	 * called from getBufferImage (to draw the frame into the offscreen buffer)
	 */
	if ( ((img->flags & IMG_BUFFER) || (img->buffer)) &&
	     ((img->buffer == 0) || (gr != img->buffer->gr))  ) {
		img = getBufferImage( env, clazz, gr, img, dstX, dstY);
	}

	dstX += gr->x0;
	dstY += gr->y0;

	dstX += img->left;
	dstY += img->top;

	if ( img->pix ) {
		if ( img->pixMask) {
			drawMaskedPixmap(X, gr, img, srcX, srcY, dstX, dstY, width, height, bgval);
		}
		else {
			if ( img->alpha ){
				if ( !img->xImg ){
					createXImage( X, img);
				}
#if defined(HAVE_LIBXEXT)
				if ( img->shmiImg) {
					XShmGetImage( X->dsp, img->pix, img->xImg, 0, 0, 0xffffffff);
				}
				else
#endif /* HAVE_LIBXEXT */
					{
						XGetSubImage( X->dsp, img->pix, 0, 0, img->width, img->height,
									  0xffffffff, ZPixmap, img->xImg, 0, 0);
					}
				drawAlphaImage(X, gr, img, srcX, srcY, dstX, dstY, width, height, bgval);
			}
			else {
				XCopyArea(X->dsp, img->pix, gr->drw, gr->gc,
						  srcX, srcY, width, height, dstX, dstY);
			}
		}
	}
	else if (img->xImg) {
		if (img->xMask) {
			drawMaskedXImage(X, gr, img, srcX, srcY, dstX, dstY, width, height, bgval);
		}
		else if (img->alpha) {
			drawAlphaImage(X, gr, img, srcX, srcY, dstX, dstY, width, height, bgval);
		}
		else {
			drawXImage(X, gr, img, srcX, srcY, dstX, dstY, width, height);
		}
	}

	XFLUSH(X, True);
}


void
Java_java_awt_Toolkit_graDrawImageScaled(JNIEnv * env, jclass clazz, Graphics * gr, Image * img,
										 jint dx0, jint dy0, jint dx1, jint dy1,
										 jint sx0, jint sy0, jint sx1, jint sy1, jint bgval)
{
	int iw, x0, y0, x1, y1;
	Image *tgt;
	int tmpXImg;

	DBG(awt_gra, ("drawImageScaled: %x %x (%x), %d,%d,%d,%d, %d,%d,%d,%d, %x\n",
				  gr, img, img->xImg, dx0, dy0, dx1, dy1, sx0, sy0, sx1, sy1, bgval));

	if ( (img == 0) || (img->width == -1) ){ /* last defense sanity check */
		return;
	}

	if ( (img->flags & IMG_SCALED) == 0 ){
		setImageState( X, img, IMG_SCALED);
	}

	/* screen images don't have an XImage, we have to get the data first */
	tmpXImg = (img->xImg == NULL);
	if ( tmpXImg) {
		img->xImg = XGetImage(X->dsp, img->pix, 0, 0, img->width, img->height,
							  0xffffffff, ZPixmap);
	}

	iw = img->xImg->width;

	if (dx1 > dx0) {
		x0 = dx0;
		x1 = dx1;
	}
	else {
		x0 = dx1;
		x1 = dx0;
	}
	if (dy1 > dy0) {
		y0 = dy0;
		y1 = dy1;
	}
	else {
		y0 = dy1;
		y1 = dy0;
	}

	if (sx0 < 0)
		sx0 = 0;
	if (sx1 < 0)
		sx1 = 0;
	if (sx0 >= iw)
		sx0 = iw - 1;
	if (sx1 >= iw)
		sx1 = iw - 1;

	tgt = createImage( (x1 - x0 + 1), (y1 - y0 + 1), 0);
	/* alpha image should always go first, since it would disable subsequent Shm usage */
	if (img->alpha)
		createAlphaImage(X, tgt);
	if (img->xMask)
		createXMaskImage(X, tgt);
	createXImage(X, tgt);

	/* this is just temporary, make sure it's not converted into a pixmap */
	tgt->flags |= IMG_SCALED;

	initScaledImage(X, tgt, img, dx0 - x0, dy0 - y0, dx1 - x0, dy1 - y0, sx0, sy0, sx1, sy1);
	Java_java_awt_Toolkit_graDrawImage(env, clazz, gr, tgt, 0, 0,
									   x0, y0, tgt->width, tgt->height, bgval);
#if defined(HAVE_LIBXEXT)
	if (tgt->shmiImg) {
		XSync(X->dsp, False);              /* since we're going to destroy tgt, process its drawing first */
	}
#endif
	Java_java_awt_Toolkit_imgFreeImage(env, clazz, tgt);

	if (tmpXImg) {
		XDestroyImage(img->xImg);
		img->xImg = NULL;
	}
}
