/*
 * minimal grlib.h implementation to facilitate image decoders
 * (imggif, imgjpeg, imgpng, ..) which can be shared with no-native-wm
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

typedef int g_contextSav_t;
typedef unsigned long g_pixel_t;

/*
 * We don't have 'common' parts of the main structs
 */
#define _(_o) (_o)

static inline void g_setScreenRotation ( char rot )
{
}

/*
 * Answer pixel value for given 8-8-8-8 ARGB
 * (remove switch for speed optimization)
 */
static inline jint g_pixelValue ( Toolkit* X, jint rgb )
{
	return pixelValue( X, rgb);
}

/*
 * Answer RGB components for given pixel value
 */
static inline void g_rgbValues ( Toolkit* X, unsigned long pixel, int* r, int* g, int* b )
{
	rgbValues( X, pixel, r, g, b);
}


/*
 * Set the foreground color of a Graphics object
 * Note that we also might have to set the active color in a
 * 'current' context, if this is the 'active' Graphics (Toolkit.common.lastGr)
 */
static inline void g_setColor ( Graphics* gr, unsigned int clr )
{
#if defined (DRAWABLE_ALPHA)
	gr->alpha = getAlpha( clr);
	clr = stripAlpha( clr);

	if ( gr->img && (gr->alpha != 0xff) && !gr->img->alpha ){
		createAlphaImage( X, gr->img);
	}
#endif

	gr->fg = clr;

	if (gr->xor)
		clr ^= gr->xclr;

	XSetForeground(X->dsp, gr->gc, clr);
}

/*
 * Answer foreground color of Graphics object
 */
static inline g_pixel_t g_getColor ( Graphics* gr )
{
	return gr->fg;
}



/*
 * Set pixel value in Image (with optional alpha value [0..255]).
 * This is called before g_completeImage(), i.e. it can use
 * temporary pixel/alpha storage
 */
static inline void g_putImagePixel ( Image* img , int x, int y,
									 g_pixel_t pix, unsigned char alpha )
{
#if defined (USE_FULL_ALPHA)
    if ( alpha == 0 ){
        if ( img->xMask ){
            pix = 0;
            XPutPixel( img->xMask, x, y, 0);
        }
		/* alpha matte is initialized with 0 */
    }
    else {
		/* we assume the image has less visible than transparent pixels */
        if ( img->alpha ){
            PutAlpha( img->alpha, x, y, alpha);
        }
    }
#else
    if ( alpha < ALPHA_VALUE_THRESHOLD ){
        if ( img->xMask ){
            pix = 0;
            XPutPixel( img->xMask, x, y, 0);
        }
    }
#endif

    XPutPixel( img->xImg, x, y, pix);
}


/*
 * Transform temporarily stored image information (both pixel color and
 * alpha values) into a representation which is suitable for drawing
 * the image. Free any temporary information.
 */
static inline void g_completeImage ( Image* img )
{
#ifdef USE_FULL_ALPHA
	if ( img->alpha ){
		/* try to reduce to a mask */
		if ( !needsFullAlpha( X, img, ALPHA_MASK_THRESHOLD) ){
			reduceAlpha( X, img, ALPHA_VALUE_THRESHOLD);
		}
	}
#endif
}


/*
 * Set alpha channel type of Image object. The supported 'transType'
 * values are NO_ALPHA, MASKED_ALPHA (transparent pixels), FULL_ALPHA.
 * After this operation, we have to be able to store pixel alpha
 * information (at least temporarily, to be later-on processed by
 * completeImage)
 */
static inline void g_setAlphaType ( Image* img, int alphaType )
{
    if ( alphaType == MASK_ALPHA ){
        if ( !img->xMask ){
            createXMaskImage( X, img);
        }
    }
	else if ( alphaType == FULL_ALPHA ){
		if ( !img->alpha ){
			createAlphaImage( X, img);
		}
	}
}

/*
 * set transparency color of the image in the native layer
 */
static inline void g_setTranspValue ( Image* img, g_pixel_t clr )
{
 
}



/*
 * Answer alpha type of image. There are three supported
 * alpha type values
 *  - NO_ALPHA     : no alpha channel at all
 *  - MASKED_ALPHA : alpha = [0,255] (transparent pixels)
 *  - FULL_ALPHA   : alpha = [0..255]
 */
static inline int g_getAlphaType ( Image* img )
{
	if ( img->xMask )
		return MASK_ALPHA;
	else if ( img->alpha )
		return FULL_ALPHA;
	else
		return NO_ALPHA;
}

/*
 * Temporarily store the current context (if the graphics driver
 * can't draw into aribtrary contexts)
 */
static inline void g_saveContext ( g_contextSav_t* cs )
{
	/* nothing to do */
}

/*
 * Restore the current context from a temporarily saved one
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_restoreContext ( g_contextSav_t* cs )
{
	/* nothing to do */
}
/*
 * Set global draw context according to the provided Image
 * (if the graphics driver can't draw into arbitrary contexts)
 */
static inline void g_setCurrentImage ( Toolkit* X, Image* img )
{
	/* nothing to do */
}
