/**
 * flash.c - MacroMedia flash interface
 * 
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

#if defined(HAVE_LIBFLASH)

#include <flash.h>

#if defined(HAVE_X)

#define	GET_PIXELS(I)	((I)->shmiImg->shmaddr)
#define	GET_SCANLINE(W)	(((W) * GET_BPP() / 8 + 3) / 4 * 4)
#define	GET_BPP()	(16)

#elif defined(FGL_H_INCLUDED)

#define	GET_PIXELS(I)	(fgl_getSlc((I)->bmp, 0, 0))
#define	GET_SCANLINE(W)	(fgl_getScanlineLength((W)) * fgl_getSlcSize())
#define	GET_BPP()	(fgl_getBitsPerPixel())

#else

#error "Unknown visual system for Flash"

#endif

typedef struct {
	FlashHandle handle;
	FlashDisplay display;
} flashWrapper;

static JNIEnv* genv;
static jmethodID urlcall;
static jmethodID swfcall;

static
void
getUrl(char* url, char* target, void* private)
{
	(*genv)->CallVoidMethod(genv, (jobject)private, urlcall, (*genv)->NewStringUTF(genv, url), (*genv)->NewStringUTF(genv, target));
}

static
void
getSwf(char* url, int level, void* private)
{
	(*genv)->CallVoidMethod(genv, (jobject)private, swfcall, (*genv)->NewStringUTF(genv, url), level);
}

jlong
Java_kaffe_awt_flash_FlashImage_flashExec0(JNIEnv* env, jobject clazz, jobject handle, jint flags, jint key, jint mouse, int x, int y)
{
	FlashEvent event;
	FlashEvent* fe;
	struct timeval tm;
	long r;
	jlong ret = -1;

	fe = &event;
	if (key != 0) {
		event.type = FeKeyPress;
		event.key = key;
	}
	else if (mouse != 0) {
		event.type = mouse;
		event.x = x;
		event.y = y;
	}
	else {
		fe = 0;
	}
	r = FlashExec(((flashWrapper*)handle)->handle, (long)flags, fe, &tm);
	if (r != 0) {
		ret = ((jlong)tm.tv_sec) * 1000;
		ret += tm.tv_usec / 1000;
	}
	return (ret);
}

jobject
Java_kaffe_awt_flash_FlashImage_flashNew0(JNIEnv* env, jobject clazz)
{
	flashWrapper* wrap;

	wrap = AWT_MALLOC(sizeof(flashWrapper));
	wrap->handle = FlashNew();

	return ((jobject)wrap);
}

jboolean
Java_kaffe_awt_flash_FlashImage_flashParse0(JNIEnv* env, jobject clazz, jobject handle, jarray data, jint offset, jint length)
{
	int r;
	jbyte* raw;

	raw = (*env)->GetByteArrayElements(env, data, 0);
	r = FlashParse(((flashWrapper*)handle)->handle, offset, raw, length);
	(*env)->ReleaseByteArrayElements(env, data, raw, JNI_ABORT);
	return (r & FLASH_PARSE_EOM ? JNI_TRUE : JNI_FALSE);
}

void
Java_kaffe_awt_flash_FlashImage_flashOpenDisplay0(JNIEnv* env, jobject clazz, jobject handle, jobject canvas, jint width, jint height)
{
	flashWrapper* wrap;
	jfieldID fld;
	Image* img;
	jclass cls;

	fld = (*env)->GetFieldID(env, (*env)->FindClass(env, "java.awt.Image"), "nativeData", "kaffe.util.Ptr");

	img = (*env)->GetObjectField(env, canvas, fld);
	Java_java_awt_Toolkit_imgFreeImage(env, 0, img);

	wrap = (flashWrapper*)handle;
	wrap->display.bpl = GET_SCANLINE(width);
	wrap->display.width = width;
	wrap->display.height = height;
	wrap->display.depth = GET_BPP();
	wrap->display.bpp = GET_BPP() / 8;
	wrap->display.flash_refresh = 0;
	wrap->display.clip_x = 0;
	wrap->display.clip_y = 0;
	wrap->display.clip_width = 0;
	wrap->display.clip_height = 0;

	img = createImage(width, height, 0);

	wrap->display.pixels = GET_PIXELS(img);

	FlashGraphicInit(wrap->handle, &wrap->display);

	(*env)->SetObjectField(env, canvas, fld, img);

	/* Register callbacks */
	genv = env;
	FlashSetGetUrlMethod(wrap->handle, getUrl, clazz);
	FlashSetGetSwfMethod(wrap->handle, getSwf, clazz);
	cls = (*env)->GetObjectClass(env, clazz);
	urlcall = (*env)->GetMethodID(env, cls, "getURLCallback", "(Ljava/lang/String;Ljava/lang/String;)V");
	swfcall = (*env)->GetMethodID(env, cls, "getSwfCallback", "(Ljava/lang/String;I)V");
}

jboolean
Java_kaffe_awt_flash_FlashImage_flashNeedRedraw0(JNIEnv* env, jobject clazz, jobject handle)
{
	if (((flashWrapper*)handle)->display.flash_refresh) {
		((flashWrapper*)handle)->display.flash_refresh = 0;
		return (JNI_TRUE);
	}
	else {
		return (JNI_FALSE);
	}
}

jint
Java_kaffe_awt_flash_FlashImage_flashWidth0(JNIEnv* env, jobject clazz, jobject handle)
{
	struct FlashInfo info;
	FlashGetInfo(((flashWrapper*)handle)->handle, &info);
	return (info.frameWidth / 20);
}

jint
Java_kaffe_awt_flash_FlashImage_flashHeight0(JNIEnv* env, jobject clazz, jobject handle)
{
	struct FlashInfo info;
	FlashGetInfo(((flashWrapper*)handle)->handle, &info);
	return (info.frameHeight / 20);
}

void
Java_kaffe_awt_flash_FlashImage_flashOpenSound0(JNIEnv* env, jobject clazz, jobject handle, jstring device)
{
	char* str;

	str = (char*)(*env)->GetStringUTFChars(env, device, 0);
	FlashSoundInit(((flashWrapper*)handle)->handle, str);
	(*env)->ReleaseStringUTFChars(env, device, str);
}

void
Java_kaffe_awt_flash_FlashImage_flashCloseDisplay0(JNIEnv* env, jobject clazz, jobject handle)
{
	flashWrapper* wrap;

	wrap = (flashWrapper*)handle;
	FlashClose(wrap->handle);

	AWT_FREE(wrap);
}

#endif
