/**
 * clr.c - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

#include "../clr-common.c"

/********************************************************************************
 * common
 */

void
initColorMapping ( JNIEnv* env, Toolkit* X )
{
  /* we query this instead of relying on a previous mode setting */
  X->nColorPlanes = bitmap_color_depth( screen);

  switch ( X->nColorPlanes ) {
  case 8:                         /* own pseudo RGB */
	X->colorMode = CM_PSEUDO;
	X->pclr = initRgb2Pseudo( env, X);
	X->maskClr = MASK_COLOR_8;
	break;
  case 15:                        /* 5-5-5 true color */
    X->colorMode = CM_TRUE;
    X->tclr = initRgb2Pixel( X);
	X->maskClr = MASK_COLOR_15;
	break;
  case 16:                        /* 5-6-5 true color */
    X->colorMode = CM_TRUE;
    X->tclr = initRgb2Pixel( X);
	X->maskClr = MASK_COLOR_16;
	break;
  case 24:                        /* 8-8-8 true color */
    X->colorMode = CM_TRUE;
    X->tclr = initRgb2Pixel( X);
	X->maskClr = MASK_COLOR_24;
	break;
  }

  X->whiteClr = pixelValue( X, 0xffffff);
  X->blackClr = pixelValue( X, 0x000000);
}

/********************************************************************************
 * specific exported functions
 */

/*
 * we need to implement brighter / darker in the native layer because the usual
 * arithmetic Rgb brightening/darkening isn't really useful if it comes to
 * PseudoColor visuals (e.g. a primitive 16 color VGA wouldn't produce any
 * usable results). Even 256 colormaps suffer from 
 */
jlong
Java_java_awt_Toolkit_clrBright ( JNIEnv* env, jclass clazz, jint rgb )
{
  unsigned r, g, b;
  jint     modRgb, modPix;

  switch ( X->colorMode )  {
  case CM_TRUE:
  case CM_PSEUDO:
	r = JRED( rgb);
	g = JGREEN( rgb);
	b = JBLUE( rgb);

	r = (r * 4) / 3;
	g = (g * 4) / 3;
	b = (b * 4) / 3;
	
	if ( r > 0xff ) r = 0xff;
	if ( g > 0xff ) g = 0xff;
	if ( b > 0xff ) b = 0xff;

	modRgb = JRGB( r, g, b);
	modPix = pixelValue( X, modRgb);
	break;
  }

  return (((jlong)modPix << 32) | modRgb);
}

jlong
Java_java_awt_Toolkit_clrDark ( JNIEnv* env, jclass clazz, jint rgb )
{
  unsigned r, g, b;
  jint     modRgb, modPix;

  switch ( X->colorMode ) {
  case CM_TRUE:
  case CM_PSEUDO:
	r = JRED( rgb);
	g = JGREEN( rgb);
	b = JBLUE( rgb);

	r = (r * 2) / 3;
	g = (g * 2) / 3;
	b = (b * 2) / 3;
	
	modRgb = JRGB( r, g, b);
	modPix = pixelValue( X, modRgb);
	break;
  }

  return (((jlong)modPix << 32) | modRgb);
}
