/**
 * evt.c - allegro event handling
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

void   initKeyboard (JNIEnv* env, Toolkit* X);
void   initMouse (JNIEnv* env, Toolkit* X);

/*
 * This is the real input event routine. We basically check the global input
 * flags and translate them into the X->event structure (so that we can generate
 * all the missing Java events from that state)
 */
__inline__ void
getNextEvent ( Toolkit*X )
{
  X->event.flags = 0;

  if ( keypressed() ){  /* do we have keyboard input? Don't block! */
	X->event.flags |= EF_NEW_EVENT | EF_KEY_PRESSED;

	X->event.key = readkey() >> 8;

	/* what  coincidence, we use the same flags (shift=1,ctrl=2,alt=4) */
	X->event.kbstat = key_shifts & 0x7;
  }

  if ( (X->event.x != mouse_x) || (X->event.x != mouse_y) ){ /* we have a motion */
	X->event.flags |= EF_NEW_EVENT | EF_MOTION;
	X->event.x = mouse_x;
	X->event.y = mouse_y;
  }

  if ( X->event.buttons != mouse_b ) { /* we have a press / release */
	X->event.flags |= EF_NEW_EVENT;

	/* This is not completely safe: we might have a press/release switch */
	if ( X->event.buttons > mouse_b )
	  X->event.flags |= EF_BUTTON_UP;
	else
	  X->event.flags |= EF_BUTTON_DOWN;

	/* what a coincidence, we use the same flags */
	X->event.buttons = mouse_b;
  }
}

/************************************************************************************
 * time to include the common stuff (which needs getNextEvent)
 */
#include "../evt-common.c"


void
initKeyboard  ( JNIEnv* env, Toolkit* X )
{
  install_keyboard();

  initKeyTab( env, X);
}


void
initMouse ( JNIEnv* env, Toolkit* X )
{
  int xspeed, yspeed;

  xspeed = get_config_int( "mouse", "xspeed", 0);
  yspeed = get_config_int( "mouse", "yspeed", 0);

  install_mouse();
  set_mouse_speed( xspeed, yspeed);
  set_mouse_sprite( X->cursors[0]->sprite);
  show_mouse( screen);
}


/*******************************************************************************
 * these remain here because of close relation to getNextEvent (i.e. in case
 * of blocking mode)
 */

jobject
Java_java_awt_Toolkit_evtPeekEvent ( JNIEnv* env, jclass clazz )
{
  return NULL;
}


/*
 * We can't do a XPeekIfEvent here because it would block until such an event
 * comes in
 */
jobject
Java_java_awt_Toolkit_evtPeekEventId ( JNIEnv* env, jclass clazz, jint id )
{
  return NULL;
}


/*
 * This is just used to wakeup the getNextEvent call if we are multithreaded,
 * and we post an event from outside the dispatcher thread (which might be blocked
 * in a native getNextEvent). This is a tribute to the fact that we actually have
 * two different queues - the native one, and the queue for self-posted events
 */
void
Java_java_awt_Toolkit_evtWakeup ( JNIEnv* env, jclass clazz )
{
}
