/**
 * gra.c - Graphics draw & state operations
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <string.h>

#include "toolkit.h"


/*******************************************************************************
 *
 */

/*
 * This is based on Allegro textout, but handles wide chars as well as
 * non-null terminated strings. If we once get into real unicode drawing,
 * this is where it should show up
 */
extern int _textmode;

void
drawText ( BITMAP *bmp, FONT *f,
		   void *str, int len, int isWideChar,
		   int x, int y, int color)
{
  FONT_PROP *fp;
  BITMAP *b;
  unsigned short *pw;
  unsigned char  *pc;
  int c, i;

  if (f->height == 8) {
	pc = (isWideChar) ? jchar2CString( X, str, len) : jbyte2CString( X, str, len);
	bmp->vtable->textout_fixed(bmp, f->dat.dat_8x8, 3, pc, x, y, color);
	return;
  }
  else if (f->height == 16) {
	pc = (isWideChar) ? jchar2CString( X, str, len) : jbyte2CString( X, str, len);
	bmp->vtable->textout_fixed(bmp, f->dat.dat_8x16, 4, pc, x, y, color);
	return;
  }

  fp = f->dat.dat_prop;

  if ( isWideChar ) {
	for ( i=0, pw=str; i<len && *pw; i++, pw++ ) {
	  c = (int)*pw - ' ';
	  if ( (c < 0) || (c >= FONT_SIZE) ) c = 0;
	  if ( (b = fp->dat[c]) ){
		bmp->vtable->draw_character( bmp, b, x, y, color);
		if ( (x += b->w) >= bmp->cr )
		  break;
	  }
	}
  }
  else {
	for ( i=0, pc=str; i<len && *pc; i++, pc++ ) {
	  c = (int)*pc - ' ';
	  if ( (c < 0) || (c >= FONT_SIZE) ) c = 0;
	  if ( (b = fp->dat[c]) ){
		bmp->vtable->draw_character( bmp, b, x, y, color);
		if ( (x += b->w) >= bmp->cr )
		  break;
	  }
	}
  }
}

/**********************************************************************************
 * with the TEXT_DRAW implementation defined, we are now ready for the common stuff
 */
#include "../gra-common.c"

void*
Java_java_awt_Toolkit_graInitGraphics ( JNIEnv* env, jclass clazz,
										Graphics* gr, jobject tgt, jint tgtType,
										jint xOff, jint yOff,
										jint xClip, jint yClip, jint wClip, jint hClip,
										jobject fnt, jint fg, jint bg, jboolean blank )
{
  if ( !gr ) {
	gr = (Graphics*) calloc( 1, sizeof(Graphics));
  }

  if ( tgt ) {
	switch ( tgtType ) {
	case WINDOW:
	  gr->gc = screen;
	  gr->tgt = tgt;
	  gr->xOff = xOff;
	  gr->yOff = yOff;
	  xOff += ((Window*)tgt)->bounds.x0;
	  yOff += ((Window*)tgt)->bounds.y0;
	  break;
	case IMAGE:     gr->gc = ((Image*)tgt)->gc;      break;
	case GRAPHICS:  gr->gc = ((Graphics*)tgt)->gc;   break;
	default: fprintf( stderr, "illegal Graphics target: %d\n", tgtType);
	}
	gr->tgtType = tgtType;
  }
  else {
	if ( gr->tgtType == WINDOW ) {
	  if ( gr->tgt ) {
		gr->xOff = xOff;
		gr->yOff = yOff;
		xOff += ((Window*)(gr->tgt))->bounds.x0;
		yOff += ((Window*)(gr->tgt))->bounds.y0;
	  }
	  else {
		fprintf( stderr, "Graphics target not initialized yet\n");
	  }
	}
  }

  gr->fg   = fg;
  gr->bg   = bg;
  gr->clr  = fg;
  gr->xclr = -1;

  gr->x0   = xOff;
  gr->y0   = yOff;


  gr->clip.x0 = xOff + xClip;
  gr->clip.y0 = yOff + yClip;
  gr->clip.x1 = gr->clip.x0 + wClip - 1;
  gr->clip.y1 = gr->clip.y0 + hClip - 1;

  if ( tgtType == WINDOW ) {
	recalcClipSegments( gr);
	gr->visTag = ((Window*)gr->tgt)->visTag;
	gr->curClip = 0;
  }
  else if ( tgtType == IMAGE ){
	gr->clipSegments = newRect( gr->clip.x0, gr->clip.y0, gr->clip.x1, gr->clip.y1, 0, 0);
	gr->curClip = 0;
  }

  if ( blank ) {
	if ( X->lastGc != gr->gc ) {
	  X->lastGc = gr->gc;
	}
	if ( gr->clipSegments ) {
	  gr->curClip = gr->clipSegments;
	  do {
		set_clip( gr->gc,
				  gr->curClip->x0, gr->curClip->y0, gr->curClip->x1, gr->curClip->y1);
		rectfill( gr->gc,
				  gr->curClip->x0, gr->curClip->y0, gr->curClip->x1, gr->curClip->y1, bg);
	  } while ( nextClipSegment( gr) );
	}
  }

  return gr;
}


void
Java_java_awt_Toolkit_graFreeGraphics ( JNIEnv* env, jclass clazz, Graphics* gr )
{
  if ( gr->clipSegments )
	freeRectList( gr->clipSegments);

  free( gr);
}

void
Java_java_awt_Toolkit_graSetColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
{
  gr->fg = clr;

  if ( gr->xclr != -1 )
	gr->clr = (clr ^ gr->xclr);
  else
	gr->clr = clr;
}

void
Java_java_awt_Toolkit_graSetBackColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
{
  gr->bg = clr;
}

void
Java_java_awt_Toolkit_graSetFont ( JNIEnv* env, jclass clazz, Graphics* gr, MFONT* fnt )
{
  gr->txo.mfont = fnt;
}

void
Java_java_awt_Toolkit_graSetPaintMode ( JNIEnv* env, jclass clazz, Graphics* gr )
{
  gr->xclr = -1;
  gr->clr  = gr->fg;

  solid_mode();
}

void
Java_java_awt_Toolkit_graSetXORMode ( JNIEnv* env, jclass clazz, Graphics* gr, jint xorClr )
{
  gr->xclr = xorClr;
  gr->clr  = (gr->fg ^ xorClr);

  xor_mode( 1);
}


/*******************************************************************************
 * image rendering
 */

void
drawAlphaImage ( Graphics* gr, Image* img,
				 jint srcX, jint srcY, jint dstX, jint dstY,
				 jint width, jint height, jint bgval )
{
  int       i, j, si, sj, di, dj, alpha;
  int       dpix, spix;
  int       sr, sg, sb, dr, dg, db;
  BITMAP    *tmpImg;

  /*
   * This takes more time to build than with direct target BITMAP manipulation,
   * but it should draw faster (once the tmpImg has been completed).
   * Revisit
   */
  if ( (tmpImg = create_bitmap( width, height)) ){
	blit( gr->gc, tmpImg, dstX, dstY, 0, 0, width, height);

	for ( j=0, sj=srcY; j<height; j++, sj++ ) {
	  for ( i=0, si=srcX; i<width; i++, si++ ) {
		dpix = getpixel( tmpImg, i, j);
		spix = getpixel( img->gc, si, sj);
		alpha = GetAlpha( img->alpha, si, sj);

		rgbValues( X, dpix, &dr, &dg, &db);
		rgbValues( X, spix, &sr, &sg, &sb);

		dr = ((alpha * sr + (255 - alpha) * dr) + 128) >> 8;
		dg = ((alpha * sg + (255 - alpha) * dg) + 128) >> 8;
		db = ((alpha * sb + (255 - alpha) * db) + 128) >> 8;

		putpixel( tmpImg, i, j, pixelValue( X, (dr << 16)|(dg << 8)|(db)));
	  }
	}
	blit( tmpImg, gr->gc, 0, 0, dstX, dstY, width, height);
	destroy_bitmap( tmpImg);
  }
}

void
Java_java_awt_Toolkit_graDrawImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
									 jint srcX, jint srcY,
									 jint dstX, jint dstY,
									 jint width, jint height, jint bgval )
{
  int     x, y, x0, y0, x1, y1;
  int     dX, dY, sX, sY, sXw, sYh;

  if ( srcX + width  > img->width  ) width  = img->width  - srcX;
  if ( srcY + height > img->height ) height = img->height - srcY;

  if ( setContext( X, gr) ) {
	x0  = gr->x0 + dstX;
	y0  = gr->y0 + dstY;
	x1  = x0 + width;
	y1  = y0 + height;
	
	do {
	  if ( x0 < gr->curClip->x0 ) {
		dX = gr->curClip->x0;
		sX = srcX + (dX - x0);
	  }
	  else {
		dX = x0;
		sX = srcX;
	  }
	  if ( y0 < gr->curClip->y0 ) {
		dY = gr->curClip->y0;
		sY = srcY + (dY - y0);
	  }
	  else {
		dY = y0;
		sY = srcY;
	  }

	  if ( x1 > gr->curClip->x1 )
		sXw = sX + (gr->curClip->x1 - dX);
	  else
		sXw = sX + (x1 - dX);
	  if ( y1 > gr->curClip->y1 )
		sYh = sY + (gr->curClip->y1 - dY);
	  else
		sYh = sY + (y1 - dY);
	  
	  if ( bgval != -1 )
		rectfill( gr->gc, x0, y0, x1, y1, bgval);

	  if ( img->masked ) {
		masked_blit( img->gc, gr->gc, sX, sY, dX, dY, (sXw - sX), (sYh - sY));
	  }
	  else if ( img->alpha ) {
		drawAlphaImage( gr, img, sX, sY, dX, dY, sXw-sX, sYh-sY, bgval);
	  }
	  else {
		blit( img->gc, gr->gc, sX, sY, dX, dY, (sXw - sX), (sYh - sY));
	  }
	} while ( nextClipSegment( gr) );
  }
}


void
Java_java_awt_Toolkit_graDrawImageScaled ( JNIEnv* env, jclass clazz, 
										   Graphics* gr, Image* img,
										   jint dx0, jint dy0, jint dx1, jint dy1,
										   jint sx0, jint sy0, jint sx1, jint sy1,
										   jint bgval )
{
  int        x0, y0, x1, y1;
  int        iw = img->width;
  Image      tgt;

  if ( dx1 > dx0 ) {
	x0 = dx0; x1 = dx1;
  }
  else {
	x0 = dx1; x1 = dx0;
  }
  if ( dy1 > dy0 ) {
	y0 = dy0; y1 = dy1;
  }
  else {
	y0 = dy1; y1 = dy0;
  }

  if ( sx0 < 0 ) sx0 = 0;
  if ( sx1 < 0 ) sx1 = 0;
  if ( sx0 >= iw ) sx0 = iw-1;
  if ( sx1 >= iw ) sx1 = iw-1;

  tgt.width = (x1 - x0 +1);
  tgt.height = (y1 - y0 +1);
  tgt.gc = create_bitmap( tgt.width, tgt.height);

  if ( img->masked ) {
	tgt.masked = 1;
  }
  else {
	tgt.masked = 0;
	tgt.alpha = (img->alpha) ? createAlphaImage( X, tgt.width, tgt.height) : 0;
  }

  initScaledImage( X, &tgt, img, dx0-x0, dy0-y0, dx1-x0, dy1-y0, sx0, sy0, sx1, sy1);

  Java_java_awt_Toolkit_graDrawImage ( env, clazz, gr, &tgt, 0, 0,
									   x0, y0, tgt.width, tgt.height, bgval);

  destroy_bitmap( tgt.gc);
  if ( tgt.alpha ) {
	free( tgt.alpha->buf);
	free( tgt.alpha);
  }
}
