/**
 * img.c - image routines
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"


/************************************************************************************
 * own auxiliary funcs
 */

Image*
createImage ( int width, int height, int withMask )
{
  Image * img = calloc( 1, sizeof( Image));
  img->trans  = -1;
  img->width = width;
  img->height = height;
  img->masked = withMask;

  if ( (width > 0) && (height > 0) ){
	img->gc = create_bitmap( width, height);
  }

  return img;
}

__inline__ void
plotTransparent ( Image *img, int col, int row )
{
  putpixel( img->gc, col, row, X->maskClr);
}


__inline__ void
setScanlineRgb ( Toolkit* X, Image* img, int col, int row, jint* scanLine )
{
  PIXEL pix = getpixel( img->gc, col, row);
  int   r, g, b;

  if ( img->masked && pix == X->maskClr ) {
	scanLine[col] = 0;               /* transparency handled via alpha channel of model */
  }
  else {
	rgbValues( X, pix, &r, &g, &b);
	scanLine[col] = (0xff000000 | (r << 16) | (g << 8) | b);
  }
}

/************************************************************************************
 * time to fold in the common image stuff
 */
#include "../img-common.c"

/************************************************************************************
 * exported native methods
 */

void
Java_java_awt_Toolkit_imgFreeImage( JNIEnv * env, jclass clazz, Image * img)
{
  if ( img->gc ){
	destroy_bitmap( img->gc);
	img->gc = 0;
  }

  if ( img->alpha ) {
	free( img->alpha->buf);
	free( img->alpha);
	img->alpha = 0;
  }

  free( img);
}
