/**
 * tlk.c - main initialization, query and termination
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include "toolkit.h"

#define	BANNER_SLEEP_TIME	2

extern void initColorMapping ( JNIEnv* env, Toolkit* X );
extern void initCursors ( Toolkit* X );
extern void initWindows ( Toolkit* X );
extern void initFonts ( Toolkit* X );

/* for banner display */
extern void Java_java_awt_Toolkit_imgFreeImage( JNIEnv * env, jclass clazz, Image * img);
extern void* Java_java_awt_Toolkit_imgCreateFromFile(JNIEnv*,jclass,jstring);


/********************************************************************************
 *
 */


/*
 * we do this native because of two reasons: (1) this should appear as fast as
 * possible (without being deferred by potential lengthy Java inits), (2) in
 * native window systems we can't draw to the "root-window" from within Java
 */
void
displayBanner ( JNIEnv* env, jclass clazz, jstring banner )
{
  Image* img;
  int    x, y;

  img = Java_java_awt_Toolkit_imgCreateFromFile(env, clazz, banner);
  if (img != 0) {
	x = (X->width - img->width) / 2;
	y = (X->height - img->height) / 2;

	/*
	 * with Allegro, we can even do something more fancy here (e.g. blending)
	 */
	BIT_BLT( img->gc, screen, 0, 0, x, y, img->width, img->height);
	Java_java_awt_Toolkit_imgFreeImage( env, clazz, img);

	sleep(BANNER_SLEEP_TIME);
  }
}

jstring
Java_java_awt_Toolkit_tlkVersion ( JNIEnv* env, jclass clazz )
{
  return (*env)->NewStringUTF( env, "allegro-1.0");
}

void
Java_java_awt_Toolkit_tlkInit ( JNIEnv* env, jclass clazz, jstring name, jstring banner )
{
  int card, width, height, depth;

  X->nBuf = 128;
  X->buf = malloc( X->nBuf);

  allegro_init();

  card = get_config_int( 0, "gfx_card", GFX_AUTODETECT);
  width = get_config_int( 0, "width", 640);
  height = get_config_int( 0, "height", 480);
  depth = get_config_int( 0, "depth", 8);

  if ( set_gfx_mode( card, width, height, 0, 0) < 0 ){
	fprintf( stderr, "graphics mode not supported: %d (%dx%d %d planes)\n",
			 card, width, height, depth);
	exit( 1);
  }
  
  set_color_depth( depth);

  X->width = screen->w;
  X->height = screen->h;

  initColorMapping( env, X);

  if ( banner )
	displayBanner( env, clazz, banner); /* we need colors here */

  initFonts( X);
  initCursors( X);
  initWindows( X);
}

void
Java_java_awt_Toolkit_tlkTerminate ( JNIEnv* env, jclass clazz )
{
  remove_mouse();
  remove_keyboard();
  allegro_exit();
}

jint
Java_java_awt_Toolkit_tlkGetResolution ( JNIEnv* env, jclass clazz )
{
  return 75;
}


jint
Java_java_awt_Toolkit_tlkGetScreenHeight ( JNIEnv* env, jclass clazz )
{
  return screen->h;
}


jint
Java_java_awt_Toolkit_tlkGetScreenWidth ( JNIEnv* env, jclass clazz )
{
  return screen->w;
}


jboolean
Java_java_awt_Toolkit_tlkIsMultiThreaded ( JNIEnv* env, jclass clazz )
{
#ifdef USE_NATIVE_THREADS
  return JNI_TRUE;
#else
  return JNI_FALSE;
#endif
}


void
Java_java_awt_Toolkit_tlkSync ( JNIEnv* env, jclass clazz )
{
}


void
Java_java_awt_Toolkit_tlkBeep ( JNIEnv* env, jclass clazz )
{ 
}
