/**
 * toolkit.h - main target lib specific awt header
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __toolkit_h
#define __toolkit_h

#include <allegro.h>

#include <stdlib.h>
#include <jni.h>

#include "../../../kaffe/kaffevm/jdebug.h"
#include "../../../kaffe/kaffevm/jmeasure.h"

/* header which includes interface macros for the target lib */
#include "grlib.def"

/* this holds all common stuff not relying on target specific types */
#include "../toolkit-common-pre.h"


/*******************************************************************************
 * target lib specific types
 */

typedef struct _Image {
  BITMAP      *gc;        /* the image data */
  int         masked;     /* image has a on/off alpha channel */
  AlphaImage  *alpha;     /* full alpha channel (for alpha != 0x00 or 0xff) */
  int         trans;      /* intermediate transparency index */
  int         width, height;
} Image;

typedef struct {
  FONT            *font;  /* basically just an array of glyph BITMAPs */
  int             height;
  int             baseline;
  int             ascent;
  int             descent;
  int             leading;
  int             maxAdvance;
  int             proportional;
} MFONT;

typedef struct {
  MFONT           *mfont;
  int             mode;
} TextOption;

typedef struct {
  BITMAP          *gc;
  void            *tgt;
  char            tgtType;

  int             fg;
  int             bg;
  int             xclr;
  int             clr;
  int             x0, y0;
  short           xOff, yOff;

  LinkedRect      clip;

  unsigned        visTag;
  LinkedRect      *clipSegments;
  LinkedRect      *curClip;

  TextOption      txo;
} Graphics;


typedef struct _Cursor {
  BITMAP          *sprite;
  short           xHot, yHot;
} Cursor;


typedef struct _Event {
  unsigned      flags;
  unsigned      kbstat;
  unsigned      buttons;     /* pressed buttons */
  int           x, y;
  unsigned      key;         /* scan code */
} Event;


typedef struct _Toolkit {
  int            width;
  int            height;

  char           *buf;
  unsigned int   nBuf;

  char           *fsPath;   /* path where to look for font files */
  int            fsOffset;  /* starting font lookup size offset */
  int            fsBigger;  /* max font lookup size increase */
  int            fsSmaller; /* max font lookup size decrease */
  MFONT          *fsDefault; /* default (BIOS) font */

  int            colorMode;
  int            nColorPlanes;
  Rgb2Pixel      *tclr;
  Rgb2Pseudo     *pclr;
  int            maskClr, blackClr, whiteClr;

  int            button;
  int            dragging;
  int            evtId;
  char           keyUp, keyDown, buttonUp, buttonDown, mouseMotion;
  int            addedKbStat, removedKbStat;
  int            lastKbStat;
  Event          event;


  int            nKeyCodes;
  jbyte          *keyCodes;
  int            nKeyChars;
  jchar          *keyChars;

  Cursor         *cursors[14];

  Graphics       *lastGr;
  BITMAP         *lastGc;

  int            nWindows;
  Window         **windows;

  Window         *top;
  Window         *bottom;
  Window         *root;

  Window         *lastWindow;
  int            srcIdx;
  int            rootIdx;

  Window         *mouseWindow;
  int            xMouse, yMouse;

  int            pendingUpdates;
  LinkedRect     *rootUpdates;
} Toolkit;


/*******************************************************************************
 * this is the global data section
 */
#ifdef MAIN

Toolkit   Tk;
Toolkit   *X = &Tk;

#else

extern Toolkit* X;

#endif


/* common things which require struct _Toolkit etc. to be defined */
#include "../toolkit-common-post.h"


/*
 *we can't use GrBlack(), GrWhite() in case we have installed our own Colormap
 */
#define BlackPixel(_X) pixelValue(_X, 0)
#define WhitePixel(_X) pixelValue(_X, 0xffffff)


static __inline__ jint
pixelValue ( Toolkit* X, jint rgb )
{
  switch ( X->colorMode ) {
  case CM_TRUE:
    return (((rgb & X->tclr->blueMask) >> X->tclr->blueShift) +
            ((rgb & X->tclr->greenMask) >> X->tclr->greenShift) +
            ((rgb & X->tclr->redMask) >> X->tclr->redShift));
  case CM_PSEUDO:
	return X->pclr->pix[(int) ((JRED(rgb)   + D16) / D8)]
	                   [(int) ((JGREEN(rgb) + D16) / D8)]
                       [(int) ((JBLUE(rgb)  + D16) / D8)];
  default:
	return 0;
  }
}

static __inline__ void
rgbValues ( Toolkit* X, unsigned long pixel, int* r, int* g, int* b )
{
  switch ( X->colorMode ) {
  case CM_TRUE:
	*r = ((pixel & X->tclr->pixRedMask)   << X->tclr->redShift)   >> 16;
	*g = ((pixel & X->tclr->pixGreenMask) << X->tclr->greenShift) >> 8;
	*b = ((pixel & X->tclr->pixBlueMask)  << X->tclr->blueShift);
	break;
  case CM_PSEUDO:
	*r = X->pclr->clrs[pixel].r << 2;
	*g = X->pclr->clrs[pixel].g << 2;
	*b = X->pclr->clrs[pixel].b << 2;
	break;
  default:	
	*r = *g = *b = 0;
	break;
  }
}


#endif
