/*
 * wnd.c - target lib specific window management (mainly cursor handling)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

#include "../cursors.h"
#include "../wnd-common.c"

Cursor* buildCursor ( Toolkit* X, int w, int h, char* shape, int xHot, int yHot, PIXEL* clrs )
{
  int     i, j, idx = 0;
  BITMAP  *bmp = create_bitmap( w, h);
  Cursor  *cur;

  for ( i=0; i<h; i++ ){
	for ( j=0; j<w; j++, idx++ ){
	  putpixel( bmp, j, i, clrs[shape[idx]]);
	}
  }

  cur = malloc( sizeof(Cursor));
  cur->sprite = bmp;
  cur->xHot = xHot;
  cur->yHot = yHot;
  
  return cur;
}


void
initCursors ( Toolkit* X )
{
  cursorColors[0] = X->maskClr;
  cursorColors[1] = X->blackClr;
  cursorColors[2] = X->whiteClr;

  /* we defer creation of other cursors until we need them */
  X->cursors[0] = buildCursor( X, cursorSize, cursorSize, cursorShapes[0],
							   cursorHotSpotX[0], cursorHotSpotY[0], cursorColors);
}

Cursor*
getCursor ( Toolkit* X, int jCursor )
{
  Cursor    *cur;
  int       i = jCursor;

  if ( i > 13 ) i = 0;
  else if ( i < 0 ) i = 0;

  if ( (cur = X->cursors[i]) == 0 ) {
	X->cursors[i] = buildCursor( X, cursorSize, cursorSize, cursorShapes[i],
								 cursorHotSpotX[i], cursorHotSpotY[i], cursorColors);
	cur = X->cursors[i];
  }

  return cur;
}


/**********************************************************************************
 * exported functions
 */

void
Java_java_awt_Toolkit_wndSetCursor ( JNIEnv* env, jclass clazz, void* wnd, jint jCursor )
{
  Cursor *curs = getCursor( X, jCursor);

  show_mouse( 0);
  set_mouse_sprite( curs->sprite);
  set_mouse_sprite_focus( curs->xHot, curs->yHot);
  show_mouse( screen);
}
