/**
 * clr-comon.c - generic (common) color support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef EXTERNAL_COLORMAPPING

unsigned char a6[] = { 0, 51, 102, 153, 204, 255 };


RgbPseudoMap*
initRgbPseudo ( JNIEnv* env, Toolkit* X )
{
	jclass   clazz;
	jfieldID fid;
	jarray   rgbRequests = 0;
	jboolean isCopy;
	jint     *jrgbs = 0;
	int      nReq = 0;
	jint     req[40];
	int      i, j, k, i2, j2, k2, pix, l;
	RgbPseudoMap  *map;


	DBG( awt_clr, ("initRgb2Pseudo\n"));

	map = AWT_MALLOC( sizeof( RgbPseudoMap));
	memset( map->pix, 0, sizeof(map->pix));
	_(map)->nClrs = 256;


	/* get the kaffe.awt.Defaults.RgbRequests rgb field */
	if ( (clazz = (*env)->FindClass( env, "kaffe.awt.Defaults")) ){
		if ( (fid = (*env)->GetStaticFieldID( env, clazz, "RgbRequests", "[I")) ){
			if ( (rgbRequests = (*env)->GetStaticObjectField( env, clazz, fid)) ){
				jrgbs = (*env)->GetIntArrayElements( env, rgbRequests, &isCopy);
				nReq = (*env)->GetArrayLength( env, rgbRequests);

				if ( nReq > 40 )  nReq = 40;  /* 256 - 216 web colors */

				memcpy( req, jrgbs, nReq * sizeof( jint));
				(*env)->ReleaseIntArrayElements( env, rgbRequests, jrgbs, JNI_ABORT);
			}
		}
	}

	/* start to build our colormap (and to populate our color cube on the fly) */
	g_initPalette( _(map));
	l = 0;

	/* allocate the 216 web colors, and init the lookup cube cells from them */
	for ( i=0, i2=0; i<6; i++, i2=i*2 ){
		for ( j=0, j2=0; j<6; j++, j2=j*2 ){
			for ( k=0, k2=0; k<6; k++, k2=k*2 ){
				pix = g_setPaletteEntry( _(map), l++, a6[i], a6[j], a6[k]);

				map->pix [i2  ][j2  ][k2] = pix;
				map->pix [i2+1][j2  ][k2] = pix;
				map->pix [i2+1][j2+1][k2] = pix;				
				map->pix [i2  ][j2+1][k2] = pix;
				
				k2++;
				map->pix [i2  ][j2  ][k2] = pix;
				map->pix [i2+1][j2  ][k2] = pix;
				map->pix [i2+1][j2+1][k2] = pix;
				map->pix [i2  ][j2+1][k2] = pix;
			}
		}
	}		


	/* there might be 40 colors left, try to distribute them most evenly among
	 * the edges and diagonals of the RGB cube
	 */
	
	if ( l + nReq < 251 ) {
		/* grays */
		map->pix[10][10][10] = g_setPaletteEntry( _(map), l++, 238, 238, 238);
		map->pix[ 8][ 8][ 8] = g_setPaletteEntry( _(map), l++, 178, 178, 178);
		map->pix[ 6][ 6][ 6] = g_setPaletteEntry( _(map), l++, 128, 128, 128);
		map->pix[ 5][ 5][ 5] = g_setPaletteEntry( _(map), l++, 116, 116, 116);
		map->pix[ 3][ 3][ 3] = g_setPaletteEntry( _(map), l++,  76,  76,  76);
		map->pix[ 1][ 1][ 1] = g_setPaletteEntry( _(map), l++,  26,  26,  26);
	}

	if ( l + nReq < 251 ) {
		/* some more reds */
		map->pix[10][0][0] = g_setPaletteEntry( _(map), l++, 230, 0, 0);
		map->pix[ 8][0][0] = g_setPaletteEntry( _(map), l++, 179, 0, 0);
		map->pix[ 6][0][0] = g_setPaletteEntry( _(map), l++, 128, 0, 0);
		map->pix[ 5][0][0] = g_setPaletteEntry( _(map), l++, 116, 0, 0);
		map->pix[ 3][0][0] = g_setPaletteEntry( _(map), l++,  76, 0, 0);
		map->pix[ 1][0][0] = g_setPaletteEntry( _(map), l++,  26, 0, 0);
	}

	if ( l + nReq < 251 ) {
		/* and greens */
		map->pix[0][10][0] = g_setPaletteEntry( _(map), l++, 0, 230, 0);
		map->pix[0][ 8][0] = g_setPaletteEntry( _(map), l++, 0, 179, 0);
		map->pix[0][ 6][0] = g_setPaletteEntry( _(map), l++, 0, 128, 0);
		map->pix[0][ 5][0] = g_setPaletteEntry( _(map), l++, 0, 116, 0);
		map->pix[0][ 3][0] = g_setPaletteEntry( _(map), l++, 0,  76, 0);
		map->pix[0][ 1][0] = g_setPaletteEntry( _(map), l++, 0,  26, 0);
	}

	if ( l + nReq < 251 ) {
		/* and blue */
		map->pix[0][0][10] = g_setPaletteEntry( _(map), l++, 0, 0, 230);
		map->pix[0][0][ 8] = g_setPaletteEntry( _(map), l++, 0, 0, 179);
		map->pix[0][0][ 6] = g_setPaletteEntry( _(map), l++, 0, 0, 128);
		map->pix[0][0][ 5] = g_setPaletteEntry( _(map), l++, 0, 0, 116);
		map->pix[0][0][ 3] = g_setPaletteEntry( _(map), l++, 0, 0,  76);
		map->pix[0][0][ 1] = g_setPaletteEntry( _(map), l++, 0, 0,  26);
	}
	
	if ( l + nReq < 254 ) {
		/* and yellow */
		map->pix[10][10][0] = g_setPaletteEntry( _(map), l++, 230, 230, 0);
		map->pix[ 8][ 8][0] = g_setPaletteEntry( _(map), l++, 179, 179, 0);
		map->pix[ 6][ 6][0] = g_setPaletteEntry( _(map), l++, 128, 128, 0);
		/* map->pix[ 0][ 5][0] = g_setPaletteEntry( _(map), l++, 116, 116, 0); */
		/* map->pix[ 3][ 3][0] = g_setPaletteEntry( _(map), l++, 76,  76, 0); */
		/* map->pix[ 1][ 1][0] = g_setPaletteEntry( _(map), l++, 26,  26, 0); */
	}

	if ( l + nReq < 256 ) {
		map->pix[10][ 9][0] = g_setPaletteEntry( _(map), l++, 238, 204, 0);
	}

	if ( l + nReq < 253 ) {
		/* and cyan */
		map->pix[0][10][10] = g_setPaletteEntry( _(map), l++, 0, 230, 230);
		map->pix[0][ 8][ 8] = g_setPaletteEntry( _(map), l++, 0, 179, 179);
		map->pix[0][ 6][ 6] = g_setPaletteEntry( _(map), l++, 0, 128, 128);
		map->pix[0][ 0][ 5] = g_setPaletteEntry( _(map), l++, 0, 116, 116);
		/* map->pix[0][ 3][ 3] = g_setPaletteEntry( _(map), l++, 0, 76,  76); */
		/* map->pix[0][ 1][ 1] = g_setPaletteEntry( _(map), l++, 0, 26,  26); */
	}

	if ( l + nReq < 255 ) {
		/* sky */
		map->pix[9][10][10] = g_setPaletteEntry( _(map), l++, 204, 220, 238);
		map->pix[9][ 9][10] = g_setPaletteEntry( _(map), l++, 204, 204, 238);
	}

	if ( l + nReq < 255 ) {
		/* flesh */
		map->pix[10][9][10] = g_setPaletteEntry( _(map), l++, 238, 204, 220);
		map->pix[10][9][ 9] = g_setPaletteEntry( _(map), l++, 234, 204, 204);
	}

	if ( l + nReq < 252 ) {
		/* magenta */
		map->pix[10][0][10] = g_setPaletteEntry( _(map), l++, 230, 0, 230);
		map->pix[ 8][0][ 8] = g_setPaletteEntry( _(map), l++, 178, 0, 178);
		map->pix[ 6][0][ 6] = g_setPaletteEntry( _(map), l++, 128, 0, 128);
		/* map->pix[ 5][0][ 5] = g_setPaletteEntry( _(map), l++, 116, 0, 116); */
		map->pix[ 3][0][ 3] = g_setPaletteEntry( _(map), l++,  76, 0,  76);
		map->pix[ 1][0][ 1] = g_setPaletteEntry( _(map), l++,  26, 0,  26);
	}


	/* set cells of explicit rgb requests (watch out, this might overwrite all
	 * previous cells, so choose them carefully
	 */
	for (j = 0; (j < nReq); j++) {
		r = JRED(req[j]);
		g = JGREEN(req[j]);
		b = JBLUE(req[j]);

		map->pix[CELL_INDEX(r)]
			    [CELL_INDEX(g)]
                [CELL_INDEX(b)] = g_setPaletteEntry( _(map), l++, r, g, b);
	}

	/* make the system palette changes effective */
	g_setPalette( _(map));

	DBG( awt_clr, (" -> map: %p\n", map));

	return map;
}


/********************************************************************************
 * the one and only True/Direct color mode
 */

RgbTrueMap*
initRgbTrue ( Toolkit* X )
{
	RgbTrueMap *map = (RgbTrueMap*) AWT_MALLOC( sizeof( RgbTrueMap));

	DBG( awt_clr, ("initRgb2Pixel\n"));

	if ( _(X)->nColorPlanes == 8 ) {
		/* 8 bit pseudo RGB mode
		 * rrrrrrrrggggggggbbbbbbbb  Java
		 * rrr-----ggg-----bb------
		 * ----------------rrrgggbb 3-3-2 native
		 */
		map->redMask    = 0xe00000;
		map->redShift   = 16;
		map->greenMask  = 0xe000;
		map->greenShift = 11;
		map->blueMask   = 0xc0;
		map->blueShift  = 6;

		map->pixRedMask   = 0xe0;
		map->pixGreenMask = 0x1c;
		map->pixBlueMask  = 0x03;
	}
	else if ( _(X)->nColorPlanes == 15 ) {
		/* 16 bit RGB mode
		 *  rrrrrrrrggggggggbbbbbbbb Java
		 *  rrrrr---ggggg---bbbbb---
		 *  ---------rrrrrgggggbbbbb 5-5-5 native
		 */
		map->redMask    = 0xf80000;
		map->redShift   = 9;
		map->greenMask  = 0xf800;
		map->greenShift = 6;
		map->blueMask   = 0xf8;
		map->blueShift  = 3;

		map->pixRedMask   = 0x7c00;
		map->pixGreenMask = 0x3e0;
		map->pixBlueMask  = 0x1f;
	}
	else if ( _(X)->nColorPlanes == 16 ) {
		/* 16 bit RGB mode
		 *  rrrrrrrrggggggggbbbbbbbb Java
		 *  rrrrr---gggggg--bbbbb---
		 *  --------rrrrrggggggbbbbb 5-6-5 native
		 */
		map->redMask    = 0xf80000;
		map->redShift   = 8;
		map->greenMask  = 0xfc00;
		map->greenShift = 5;
		map->blueMask   = 0xf8;
		map->blueShift  = 3;

		map->pixRedMask   = 0xf81f;
		map->pixGreenMask = 0x7e0;
		map->pixBlueMask  = 0x1f;
	}

	DBG( awt_clr, (" -> map: %p  R: %x,%d,  G: %x,%d  B: %x,%d\n", map,
				   map->redMask, map->redShift,
				   map->greenMask, map->greenShift,
				   map->blueMask, map->blueShift));

	return map;
}


/********************************************************************************
 * VGA color support
 */

/*
 * Since we have incredible 16 colors, we have 3 wished left (*) once we
 * have allocated the standard Java colors
 */

jint vgaRgb[] = {
	/* the standard Java colors */
	JRGB(   0,   0,   0),  /* CLR_BLACK     */
	JRGB( 255, 255, 255),  /* CLR_WHITE     */

	JRGB( 255,   0,   0),  /* CLR_RED       */
	JRGB(   0, 255,   0),  /* CLR_GREEN     */
	JRGB(   0,   0, 255),  /* CLR_BLUE      */

	JRGB( 192, 192, 192),  /* CLR_LIGHTGRAY */
	JRGB( 128, 128, 128),  /* CLR_GRAY      */
	JRGB(  64,  64,  64),  /* CLR_DARKGRAY  */

	JRGB( 255, 255,   0),  /* CLR_YELLOW    */
	JRGB( 255,   0, 255),  /* CLR_MAGENTA   */
	JRGB(   0, 255, 255),  /* CLR_CYAN      */

	JRGB( 255, 175, 175),  /* CLR_PINK      */
	JRGB( 255, 200,   0),  /* CLR_ORANGE    */

	/* our own colors */
	JRGB( 224, 224, 224),  /* CLR_GRAY1: a lighter lightGray */
	JRGB( 160, 160, 160),  /* CLR_GRAY2: a lighter gray */
	JRGB( 192,   0,   0)   /* CLR_RED1:  a darker red */
};

unsigned char modVgaPix[16][3] = {   /* normal, DARKER, BRIGHTER */
	{ CLR_BLACK, CLR_BLUE, CLR_DARKGRAY },
	{ CLR_WHITE, CLR_GRAY1, CLR_WHITE   },

	{ CLR_RED, CLR_MAGENTA, CLR_ORANGE  },
	{ CLR_GREEN, CLR_BLACK, CLR_WHITE   },
	{ CLR_BLUE, CLR_BLACK, CLR_CYAN     },

	{ CLR_LIGHTGRAY, CLR_GRAY2, CLR_GRAY1 },
	{ CLR_GRAY, CLR_DARKGRAY, CLR_GRAY2  },
	{ CLR_DARKGRAY, CLR_BLACK, CLR_GRAY2 },

	{ CLR_YELLOW, CLR_ORANGE, CLR_WHITE  },
	{ CLR_MAGENTA, CLR_BLACK, CLR_PINK   },
	{ CLR_CYAN, CLR_BLUE, CLR_WHITE },

	{ CLR_PINK, CLR_MAGENTA, CLR_WHITE },
	{ CLR_ORANGE, CLR_RED, CLR_YELLOW },

	{ CLR_GRAY1, CLR_LIGHTGRAY, CLR_WHITE },
	{ CLR_GRAY2, CLR_DARKGRAY, CLR_LIGHTGRAY },
	{ CLR_RED1, CLR_BLACK, CLR_RED }
};



RgbVgaMap*
initRgbVga ( Toolkit* X )
{
	int   i;
	RgbVgaMap *map = (RgbVgaMap*) AWT_MALLOC( sizeof( RgbVgaMap));

	DBG( awt_clr, ("initRgb2Vga\n"));

	_(map)->nClrs = 16;

	g_initPalette( _(map));
	for ( i=0; i < 16; i++ ){
		g_setPaletteEntry( _(map), i, JRED( vgaRgb[i]), JGREEN( vgaRgb[i]), JBLUE( vgaRgb[i]));
	}
	g_setPalette( _(map));

	DBG( awt_clr, (" -> map: %p\n", map));

	return map;
}

jint
getVGAPixelValue ( Toolkit* X, jint rgb, int mod )
{
	unsigned char r  = (rgb >> 16) & 0xff;
	unsigned char g  = (rgb >>  8) & 0xff;
	unsigned char b  = rgb & 0xff;

	unsigned char rr = (r >> 2) << 2;
	unsigned char gg = (g >> 2) << 2;
	unsigned char bb = (b >> 2) << 2;

	if ( rr ) {
		if ( gg ) {
			if ( bb ) {  /* red, green, blue */
				if ( gg == bb ) {
					if ( rr == gg ) { /* gray or white */
						if ( r > 240 )          return modVgaPix [CLR_WHITE] [mod];
						else if ( r > 208 )     return modVgaPix [CLR_GRAY1] [mod];
						else if ( r > 176 )     return modVgaPix [CLR_LIGHTGRAY] [mod];
						else if ( r > 144 )     return modVgaPix [CLR_GRAY2] [mod];
						else if ( r > 96 )      return modVgaPix [CLR_GRAY] [mod];
						else if ( r > 32 )      return modVgaPix [CLR_DARKGRAY] [mod];
						else                    return modVgaPix [CLR_BLACK] [mod];
					}
					else if ( r < g )           return modVgaPix [CLR_CYAN] [mod];
					else                        return modVgaPix [CLR_RED] [mod];
				}
				else if ( rr == bb ) {
					if ( g > r )                return modVgaPix [CLR_GREEN] [mod];
					else                        return modVgaPix [CLR_PINK] [mod];
				}
				else {
					if ( (r > g) && (r > b) )   return modVgaPix [CLR_RED] [mod];
					else if ( (g >> b) && (g >> r) ) return modVgaPix [CLR_GREEN] [mod];
					else                        return modVgaPix [CLR_BLUE] [mod];
				}
			}
			else {        /* red, green  */
				if ( rr > gg )                  return modVgaPix [CLR_ORANGE] [mod];
				else if ( rr == gg )            return modVgaPix [CLR_YELLOW] [mod];
				else                            return modVgaPix [CLR_GREEN] [mod];
			}
		}
		else {
			if ( bb ){
			  if ( rr < bb ) /* red, blue */    return modVgaPix [CLR_BLUE] [mod];
			  else if ( rr == bb )              return modVgaPix [CLR_MAGENTA] [mod];
			  else                              return modVgaPix [CLR_RED1] [mod];
			}
			else {       /* red */
				if ( r > 200 )                  return modVgaPix [CLR_RED] [mod];
				else                            return modVgaPix [CLR_RED1] [mod];
			}
		}
	}
	else {
		if ( gg ) {
			if ( bb )    /* green, blue */      return modVgaPix [CLR_CYAN] [mod];
			else        /* green */             return modVgaPix [CLR_GREEN] [mod];
		}
		else {
			if ( bb )    /* blue */             return modVgaPix [CLR_BLUE] [mod];
			else        /* black */             return modVgaPix [CLR_BLACK] [mod];
		}
	}
}

/********************************************************************************
 * common color system initialization
 */

void
_initColorMapping ( JNIEnv* env, Toolkit* X )
{
	_(X)->nColorPlanes = g_getBitsPerPixel( X);  

	DBG( awt_clr, ("initColorMapping: %d bpp\n", _(X)->nColorPlanes));

	if ( _(X)->nColorPlanes == 1 ) {             /* monochrome */
		_(X)->colorMode = CM_MONO;
	}
	else if ( _(X)->nColorPlanes == 4 ) {        /* VGA 16 */
		_(X)->map.vclr = initRgbVga( X);
		_(X)->colorMode = CM_VGA;
	}
	else if ( _(X)->nColorPlanes == 8 ) {        /* own pseudo RGB (256) */
		_(X)->map.pclr = initRgbPseudo( env, X);
		_(X)->colorMode = CM_PSEUDO;
	}
	else {                                       /* RGB mode */
		/* some systems have to be switched to RGB explicitly */
		g_initRgbMode( X);

		_(X)->map.tclr = initRgbTrue( X);
		_(X)->colorMode = CM_TRUE;
	}

	_(X)->black = g_pixelValue( X, 0x0);
	_(X)->white = g_pixelValue( X, 0xffffff);
}



int
modPixelValue ( Toolkit* X, int pix, int mod )
{
	jint modPix;
	int r, g, b;
	int alpha = g_getAlpha( pix);

	switch ( _(X)->colorMode ){
	case CM_VGA:
		modPix = modVgaPix[ (pix & 0xf) ][mod];
		break;
	case CM_MONO:
		modPix = pix;
		break;
	case CM_PSEUDO:
	default:
		switch ( mod ){
		case BRIGHTER: mod = 4; break;
		case DARKER:   mod = 2; break;
		default:       mod = 3;
		}
		g_rgbValues( X, pix, &r, &g, &b);
		r = r * mod / 3;
		g = g * mod / 3;
		b = b * mod / 3;
		if ( r > 0xff ) r = 0xff;
		if ( g > 0xff ) g = 0xff;
		if ( b > 0xff ) b = 0xff;
		modPix = g_pixelValue( X, JRGB( r, g, b));
	}

	return ((alpha << 24) | modPix);
}



#endif /* EXTERNAL_COLORMAPPING */

/********************************************************************************
 * exported functions
 */

jint
_clrGetPixelValue ( JNIEnv* env, jclass clazz, jint rgb )
{
	jint pix;

	/*
	 * We do this deferred to avoid class init problems with Defaults<->Color
	 */
	if ( !_(X)->colorMode ) {
		initColorMapping( env, X);
	
		/*
		 * not very intuitive, but we need to init the color mapping before we
		 * init cursors, and color init is part of the automatic Toolkit init sequence
		 * (since Color is referenced by Toolkit/Defaults), so there is no danger of
		 * running into a not-yet initialized Cursor class
		 */
		if ( !X->cursors[0] )
			initCursors( X);
	}

	pix = g_pixelValue( X, rgb);
	DBG( awt_clr, ("getPixelValue: %x %d (%x)\n", (unsigned)rgb, (int)pix, (int)pix));

	return pix;
}

void
_clrSetSystemColors ( JNIEnv* env, jclass clazz, jarray sysClrs )
{
	/* there is no native window manager, hence no native defaults (leave it up to Java) */
	DBG( awt_clr, ("setSystemColors: %p\n", sysClrs));
}

jobject
getSystemColorModel ( JNIEnv * env, jclass clazz, Toolkit* X, int trans )
{
	static jarray     rgbArray;

	CommonIdxMap *map;
	jobject    cm = 0;
	jclass     cmClazz;
	jmethodID  cmCtorId;
	unsigned   bitsPerRgb, redMask, greenMask, blueMask;
	int        i, n, r, g, b;
	jint       *rgbs;
	jboolean   isCopy;

	if ( !_(X)->colorMode ) {
		initColorMapping( env, X);
	}

	switch ( _(X)->colorMode ) {

	case CM_TRUE:
		switch ( _(X)->nColorPlanes ) {
		case 8:                         /*        rrrgggbb */
			bitsPerRgb = 8;
			redMask   = 0xe0;
			greenMask = 0x1c;
			blueMask  = 0x03;
			break;
		case 15:                        /* -rrrrrgggggbbbbb */
			bitsPerRgb = 16;
			redMask   = 0x7c00;
			greenMask = 0x03e0;
			blueMask  = 0x001f;
			break;
		case 16:                        /* rrrrrggggggbbbbb */
			bitsPerRgb = 16;
			redMask   = 0xf800;
			greenMask = 0x07e0;
			blueMask  = 0x001f;
			break;
		default:
			bitsPerRgb = 0;
			redMask = 0;
			greenMask = 0;
			blueMask = 0;
			break;
		}

		cmClazz = (*env)->FindClass( env, "java.awt.image.DirectColorModel");
		cmCtorId = (*env)->GetMethodID( env, cmClazz, "<init>", "(IIIII)V");
		cm = (*env)->NewObject( env, cmClazz, cmCtorId,
								bitsPerRgb, redMask, greenMask, blueMask, 0);
		break;

	case CM_PSEUDO:
	case CM_VGA:
		cmClazz = (*env)->FindClass( env, "java.awt.image.IndexColorModel");
		cmCtorId = (*env)->GetMethodID( env, cmClazz, "<init>", "(I[II)V");

		if ( !rgbArray ) {
			/* we need to do this just once, and subsequently share
			 * the array between all SystemColorModel based instances
			 */

			/* watch out, not all implementations have _(X)->map */
			if ( _(X)->colorMode == CM_VGA ){
				map = (CommonIdxMap*) &(_(X)->map.vclr->common);
				n = 16;
			}
			else {
				map = (CommonIdxMap*) &(_(X)->map.pclr->common);
				n = 256;
			}

			rgbArray = (*env)->NewIntArray( env, n);
			rgbs = (*env)->GetIntArrayElements( env, rgbArray, &isCopy);

			for ( i=0; i<n; i++ ){
				g_queryPaletteEntry( map, i, &r, &g, &b);
				rgbs[i] = JRGB( r, g, b) | 0xff000000;
			}

			(*env)->ReleaseIntArrayElements( env, rgbArray, rgbs, 0);
			rgbArray = (*env)->NewGlobalRef( env, rgbArray);
		}

		cm = (*env)->NewObject( env, cmClazz, cmCtorId, _(X)->nColorPlanes, rgbArray, trans);
		break;

	default:
	}

	return cm;
}

jobject
_clrGetColorModel ( JNIEnv* env, jclass clazz )
{
	return getSystemColorModel( env, clazz, X, -1);
}

/*
 * we need to implement brighter / darker in the native layer because the usual
 * arithmetic Rgb brightening/darkening isn't really useful if it comes to
 * PseudoColor visuals (e.g. a primitive 16 color VGA wouldn't produce any
 * usable results). Even 256 colormaps suffer from that
 */
jlong
_clrModBrightness ( JNIEnv* env, jclass clazz, jint pixValue, jint mod )
{
	int r, g, b;
	jlong pixRgb;
	jlong pix = g_modPixelValue( X, pixValue, mod);

	g_rgbValues( X, pix, &r, &g, &b);
	pixRgb = ((pix << 32) | (pixValue & 0xff000000) | (r << 16) | (g << 8) | b);

	DBG( awt_clr, ("clrBright: %x -> rgb: %x, pix: %x\n",
				   (unsigned)pixValue, (unsigned)(pixRgb>>32), (unsigned)(pixRgb & 0xffffffff)));

	return pixRgb;
}
