/*
 * evt-common.c - common event handling support (heavily depends on
 *                common Event type and grlib.def macros)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


jclass     MouseEvent;
jclass     KeyEvent;
jclass     PaintEvent;
jclass     GenericEvent;
jclass     VTSwitchEvent;

jmethodID  getMouseEvent;
jmethodID  getKeyEvent;
jmethodID  getPaintEvent;
jmethodID  getGenericEvent;
jmethodID  getGenericDataBuffer;
jmethodID  getVTSwitchEvent;

#define    KEY_PRESSED          401
#define    KEY_RELEASED         402

#define    MOUSE_PRESSED        501
#define    MOUSE_RELEASED       502
#define    MOUSE_MOVED          503
#define    MOUSE_ENTERED        504
#define    MOUSE_EXITED         505
#define    MOUSE_DRAGGED        506

#define    VT_SWITCH_REL_REQ    601
#define    VT_SWITCH_ACQ_REQ    602

#define    EXPOSE_PAINT         800
#define    EXPOSE_UPDATE        801

#define    GENERIC_FIRST        10001

#define    VK_SHIFT             0x10
#define    VK_CTRL              0x11
#define    VK_ALT               0x12


/*******************************************************************************
 * misc utility funcs
 */

inline
int
getSourceIdx ( Toolkit* X, AwtWnd* w )
{
	int           n;
	register int  i;

	if ( w == _(X)->lastWindow ){
		return _(X)->srcIdx;
	}
	else {
		for ( i = (unsigned long) w, n=0; n < _(X)->nWindows; i++, n++ ) {
			i %= _(X)->nWindows;
			if ( _(X)->windows[i] == w ){
				_(X)->srcIdx = i;
				_(X)->lastWindow = w;
				return _(X)->srcIdx;
			}
	  
			if ( _(X)->windows[i] == 0 )
				return -1;
		}
		return -1;
	}
}

inline
AwtWnd*
getMouseWindow ( Toolkit*X, int x, int y )
{
	AwtWnd *w;

	for ( w=_(X)->top; w; w=w->next ) {
		if ( w->mapped && (w->bounds.x0 <= x) && (w->bounds.x1 >= x)
			 && (w->bounds.y0 <= y) && (w->bounds.y1 >= y) )
			return w;
	}

	return _(X)->root;
}


void
retargetMouse ( JNIEnv* env, Toolkit* X )
{
	AwtWnd *w;

	/*
	 * we play it according to the X rules here: don't change targets during
	 * pressed/released intervals
	 */
	if ( !_(X)->button ) {
		w = getMouseWindow( X, _(X)->xMouse, _(X)->yMouse);

		if ( w != _(X)->mouseWindow ) {
			/* make sure the next event is a (artificial) MOUSE_ENTER */
			g_queueMouseMotion( env, X);
		}
	}
}


/*******************************************************************************
 * the notification (Java Event production) routines
 */

/*
 * Corresponds to a expose or paint event in native window systems. This one
 * is a no-brainer because we create them explicitly via update requests
 */
jobject
paintNotify ( JNIEnv* env, Toolkit* X )
{
	int          x, y, w, h;
	LinkedRect   *r;
	AwtWnd       *wnd;

	if ( _(X)->rootUpdates ){
		r = _(X)->rootUpdates;
		_(X)->rootUpdates = r->next;

		x = r->x0;
		y = r->y0;
		w = r->x1 - r->x0 +1;
		h = r->y1 - r->y0 +1;

		DBG( awt_evt, ("paintNotify: root (%d)  %d,%d,%d,%d\n", _(X)->rootIdx,x,y,w,h));
		
		freeRect( r);
		return (*env)->CallStaticObjectMethod( env, PaintEvent, getPaintEvent,
											   _(X)->rootIdx, EXPOSE_UPDATE,
											   x, y, w, h);
	}
	else {
		for ( wnd=_(X)->bottom; wnd; wnd=wnd->prev ) {
			if ( wnd->updateRect ) {
				x = wnd->updateRect->x0 -1;
				y = wnd->updateRect->y0 -1;
				w = wnd->updateRect->x1 - x +2;
				h = wnd->updateRect->y1 - y +2;
				r = wnd->updateRect;
				wnd->updateRect = r->next;
				freeRect( r);

				if ( wnd->mapped ) {
					if ( (_(X)->srcIdx = getSourceIdx( X, wnd)) >= 0 ) {
						DBG( awt_evt, ("paintNotify: wnd (%d)  %d,%d,%d,%d\n", _(X)->srcIdx,x,y,w,h));
						return (*env)->CallStaticObjectMethod( env, PaintEvent,
															   getPaintEvent,
															   _(X)->srcIdx,
															   EXPOSE_UPDATE,
															   x, y, w, h);
					}
					else {
						/* trouble: inconsistent window list */
					}
				}
			}
		}
	}

	/* we get here just in case there is no more updateRect */
	_(X)->pendingUpdates = 0;
	return 0;
}


jobject
keyNotify ( JNIEnv* env, Toolkit* X, int evtId )
{
	int keyCode = g_getKeyCode( X);
	int keyChar = g_getKeyChar( X);

	DBG( awt_evt, ("keyNotify: %d %d,%d \'%c\'\n", evtId, keyCode, keyChar,keyChar));

	/* we don't really care for the srcIdx, as long as it refers to a live window
	 * (it's re-mapped to the keyTgt in Java, anyway)
	 */
	return (*env)->CallStaticObjectMethod( env, KeyEvent, getKeyEvent,
										   _(X)->rootIdx, evtId, keyCode, keyChar, -1);
}

jobject
motionNotify ( JNIEnv* env, Toolkit* X )
{
	AwtWnd *wnd;
	int    id, x, y;

	x = g_getMouseX( X);
	y = g_getMouseY( X);

	/* check if this is the beginning of a drag operation */
	if ( _(X)->button && ((x != _(X)->xMouse) || (y != _(X)->yMouse)) )
		_(X)->dragging = 1;

	if ( _(X)->button ) {	
		wnd = _(X)->mouseWindow;          /* don't change target */
		id = MOUSE_MOVED;
	}
	else {
		if ( !(wnd = getMouseWindow( X, x, y)) )
			return 0;                    /* don't get caught by a initial mouse placement */


		if ( wnd != _(X)->mouseWindow ){  /* do we have a target change: move -> exit,enter */
			if ( _(X)->mouseWindow ){
				id = MOUSE_EXITED;
				wnd = _(X)->mouseWindow;
				_(X)->mouseWindow = 0;
			}
			else {
				id = MOUSE_ENTERED;
				_(X)->mouseWindow = wnd;
			}
		}
		else {
			id = MOUSE_MOVED;
		}
	}

	if ( !wnd || (_(X)->srcIdx = getSourceIdx( X, wnd)) < 0 )  /* trouble: inconsistent window list */
		return 0;

	/* in case of an exit, make sure we process the enter before anything else */
	if ( id != MOUSE_EXITED ){
		_(X)->xMouse = x;
		_(X)->yMouse = y;
	}
	x -= wnd->bounds.x0;
	y -= wnd->bounds.y0;

	DBG( awt_evt, ("motionNotify: %d  %d  %d,%d\n", _(X)->srcIdx, id, x,y));

	return (*env)->CallStaticObjectMethod( env, MouseEvent, getMouseEvent,
										   _(X)->srcIdx, id, 0, x, y);
}


jobject
buttonNotify ( JNIEnv* env, Toolkit* X, int evtId )
{
	AwtWnd  *wnd = 0;   /* just for -wall :( */
	int     button;
	int     x, y;

	x = g_getMouseX( X);
	y = g_getMouseY( X);

	if ( !_(X)->dragging ) {
		wnd = getMouseWindow( X, x, y);
		_(X)->srcIdx = getSourceIdx( X, wnd);
		
		/* This is actually quite useful in case we have a touchscreen
		 * (mouseWindow is otherwise set on enter/exits). If not set here,
		 * we don't have it avail in drag events
		 */
		_(X)->mouseWindow = wnd;
	}
	if ( _(X)->srcIdx < 0 ) {
		return 0;
	}

	if ( evtId == MOUSE_PRESSED ){
		/*
		 * the event.buttons field just contains the current state,
		 * we therefore have to remember the pressed buttons
		 */
		_(X)->button = g_getMouseBtn( X);
		button = _(X)->button;
	}
	else {
		button = _(X)->button;
		_(X)->button = 0;
		wnd = _(X)->windows[_(X)->srcIdx];
		_(X)->dragging = 0;

		/* check if we got a new target while being in press/release interval */
		retargetMouse( env, X);
	}

	if ( wnd == 0 ) {
		/* can't be, something must be seriously wrong */
		return 0;
	}

	x -= wnd->bounds.x0;
	y -= wnd->bounds.y0;

	DBG( awt_evt, ("buttonNotify: %d  %d %d  %d,%d\n", _(X)->srcIdx, evtId, button, x,y));

	return (*env)->CallStaticObjectMethod( env, MouseEvent, getMouseEvent,
										   _(X)->srcIdx, evtId, button, x, y);
}


jobject
genericNotify ( JNIEnv* env, Toolkit* X, int evtId, unsigned char* data, int dlen )
{
	jobject dataObject;
	int len;
	
	DBG( awt_evt, ("genericNotify: %d  %d,%d\n", evtId, data, dlen));

	if ( GenericEvent ){ /* we might not have it */
		dataObject = (*env)->CallStaticObjectMethod( env, GenericEvent,
													getGenericDataBuffer, evtId - EVT_GENERIC_FIRST, dlen);
		
		if ( dataObject ){
			len = (*env)->GetArrayLength( env, dataObject);
			if ( len > dlen ){
				len = dlen;
			}
		
			(*env)->SetByteArrayRegion( env, dataObject, 0, len, (jbyte*) data);
		
			return (*env)->CallStaticObjectMethod( env, GenericEvent, getGenericEvent,
												  evtId - EVT_GENERIC_FIRST, dataObject, dlen);
		}
	}

   return 0;
}


jobject
vtSwitchNotify ( JNIEnv* env, Toolkit* X, int evtId )
{
	DBG( awt_evt, ("vtSwitchNotify: %d\n", evtId));

	return (*env)->CallStaticObjectMethod( env, VTSwitchEvent,
	      getVTSwitchEvent, evtId);
}



/*******************************************************************************
 * common exported event methods
 */

jobject
_evtInit ( JNIEnv* env, jclass clazz )
{
	jclass    cls;
	jarray    sources;

	DBG( awt_evt, ("evtInit\n"));

	/* cache event classes and methods */
	MouseEvent = (*env)->NewGlobalRef( env, (*env)->FindClass( env, "java.awt.MouseEvt"));
	getMouseEvent  = (*env)->GetStaticMethodID( env, MouseEvent, "getEvent",
												"(IIIII)Ljava/awt/MouseEvt;");

	KeyEvent = (*env)->NewGlobalRef( env, (*env)->FindClass( env, "java.awt.KeyEvt"));
	getKeyEvent    = (*env)->GetStaticMethodID( env, KeyEvent, "getEvent",
												"(IIIII)Ljava/awt/KeyEvt;");

	PaintEvent = (*env)->NewGlobalRef( env, (*env)->FindClass( env, "java.awt.PaintEvt"));
	getPaintEvent  = (*env)->GetStaticMethodID( env, PaintEvent, "getEvent",
												"(IIIIII)Ljava/awt/PaintEvt;");

	GenericEvent = (*env)->NewGlobalRef( env, (*env)->FindClass( env, "kaffe.awt.GenericEvent"));
	getGenericEvent  = (*env)->GetStaticMethodID( env, GenericEvent, "getEvent",
												"(I[BI)Lkaffe/awt/GenericEvent;");
	getGenericDataBuffer  = (*env)->GetStaticMethodID( env, GenericEvent, "getDataBuffer",
												"(II)[B");

	VTSwitchEvent = (*env)->NewGlobalRef( env, (*env)->FindClass( env, "kaffe.awt.VTSwitchEvent"));
	getVTSwitchEvent    = (*env)->GetStaticMethodID( env, VTSwitchEvent, "getEvent",
												"(I)Lkaffe/awt/VTSwitchEvent;");

	
	/* Init the native input subsystems. This is done deferred (rather than
	 * from tlkInit) because it might involve additional threads
	 */
	g_initKeyboard( X);
	g_initMouse( X);

	/*
	 * Create the Window-to-Component dispatch table. This is not strictly needed
	 * if we are not looking for paint optimizations, but it is better if we can
	 * send some events (e.g. paints) directly to toplevels instead of the RootWindow
	 */

	_(X)->nWindows = 47;
	_(X)->windows = (AwtWnd**) AWT_CALLOC( _(X)->nWindows, sizeof(AwtWnd*));

	_(X)->button = 0;
	_(X)->dragging = 0;

	cls = (*env)->FindClass( env, "java.awt.Component");
	sources = (*env)->NewObjectArray( env, _(X)->nWindows, cls, NULL);

	DBG( awt_evt, (" -> init callbacks: %p,%p,%p  sources: %p\n",
				   getMouseEvent,getKeyEvent,getPaintEvent, sources));

	return sources;
}


jint
_evtRegisterSource ( JNIEnv* env, jclass clazz, AwtWnd* wnd )
{
	register int i;
	int      n;

	/*
	 * we don't use a double hashing here because collisions are very unlikely
	 * (Windows are distinct addresses), so it does not make sense to
	 * hash them again - we just could make it worse
	 */
	for ( i = (unsigned long)wnd, n=0; n < _(X)->nWindows; i++, n++ ) {
		i %= _(X)->nWindows;
		if ( !_(X)->windows[i] || ((long) _(X)->windows[i] == -1) ) {
			_(X)->windows[i] = wnd;

			if ( wnd == _(X)->root )
				_(X)->rootIdx = i;

			DBG( awt_evt, ("registerSource: %p  %d\n", wnd, i));

			return i;
		}
	}
	
	DBG( awt_evt, ("window table out of space: %d", _(X)->nWindows));
	return -1;
}


jint
_evtUnregisterSource ( JNIEnv* env, jclass clazz, AwtWnd* wnd )
{
	int i = getSourceIdx( X, wnd);

	if ( i >= 0 ) {
		_(X)->windows[i] = (AwtWnd*) (long) -1;
		DBG( awt_evt, ("registerSource: unknown window!! %p\n", wnd));
	}

	DBG( awt_evt, ("registerSource: %p  %d\n", wnd, i));

	return i;
}

/*****************************************************************************
 * callbacks to ipl VT switching routines
 */

JNIEXPORT void JNICALL
Java_kaffe_awt_VTSwitchEvent_ackRelease(JNIEnv* env, jobject o)
{
#ifdef VT_SWITCH
  ipl_vtSwitchRelease();
#endif
}

JNIEXPORT void JNICALL
Java_kaffe_awt_VTSwitchEvent_ackAcquire(JNIEnv* env, jobject o)
{
#ifdef VT_SWITCH
  ipl_vtSwitchAcquire();
#endif
}
