/**
 * clr.c - platform specific color management (FGL version)
 *
 * It doesn't make a lot of sense to rely on clr-common.c, because the color
 * setup is hardcoded in the fgl driver in use (it has to be there because
 * fgl can be used as a autark lib). This is a tradeoff between decreased
 * runtime overhead (we don't need colormodes which aren't supported by
 * the fgl visual) and source redundancy (esp. for the p8 fgl driver)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"


/******************************************************************************
 * generic exported methods
 */

#define _clrGetPixelValue     EXPORT(clrGetPixelValue)
#define _clrSetSystemColors   EXPORT(clrSetSystemColors)
#define _clrModBrightness     EXPORT(clrModBrightness)
#define _clrGetColorModel     EXPORT(clrGetColorModel)

#define _initColorMapping     initColorMapping

/******************************************************************************
 * import generic color functionality
 */

/* we don't want this in clr-common.c because it's hardwired in fgld */
#define EXTERNAL_COLORMAPPING

#include "../clr-common.c"

/* we just honor the RgbRequests for PseudoColor visuals */
#if defined (IS_CMAP_MODE)
int
getRgbRequests ( JNIEnv* env, jclass clazz, int *rgbs, int max )
{
	jmethodID mid;
	jarray    rgbRequests = 0;
	int       nReq = 0;

	if ( (mid = (*env)->GetStaticMethodID( env, clazz, "getRgbRequests", "(I)[I")) ){
		if ( (rgbRequests = (*env)->CallStaticObjectMethod( env, clazz, mid, max)) ){
			(*env)->GetIntArrayRegion( env, rgbRequests, 0, max, rgbs);
			
			for ( ; nReq < max; nReq++ ){
				if ( rgbs[nReq] < 0 ) break;
			}
		}
	}

	return nReq;
}
#endif /* IS_CMAP_MODE */

void
initColorMapping ( JNIEnv* env, Toolkit* X )
{
	/* not much to do, already set up in tlkInit */

	_(X)->nColorPlanes = g_getBitsPerPixel( X);  

	DBG( awt_clr, ("initColorMapping: %d bpp\n", _(X)->nColorPlanes));

	if ( _(X)->nColorPlanes == 4 ) {         /* VGA 16 */
		_(X)->colorMode = CM_VGA;
	}
	else if ( _(X)->nColorPlanes == 8 ) {    /* own pseudo RGB (256) */
		_(X)->colorMode = CM_PSEUDO;
	}
	else {                                       /* RGB mode */
		_(X)->colorMode = CM_TRUE;
	}

	_(X)->black = g_pixelValue( X, 0x0);
	_(X)->white = g_pixelValue( X, 0xffffff);
}
