/**
 * evt.c - platform specific event handling (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define IPL_INLINE_EVENT
#include "ipl.h"

#include "toolkit.h"

#ifdef NEVER
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/threads.h"
#include "../../../../../kaffe/kaffevm/locks.h"
#endif


/******************************************************************************
 * generic exported methods
 */

#define _evtInit              EXPORT(evtInit)
#define _evtRegisterSource    EXPORT(evtRegisterSource)
#define _evtUnregisterSource  EXPORT(evtUnregisterSource)

/******************************************************************************
 * import generic functionality
 */

#include "../evt-common.c"


/*******************************************************************************
 * exported event functions
 */


jobject
EXPORT(evtGetNextEvent) ( JNIEnv* env, jclass clazz )
{
	int ret;

	DBG( awt_evt, ("getEvent\n"));

	if ( _(X)->pendingUpdates ) {
		DBG( awt_evt, ("   paintNotify\n"));
		return paintNotify( env, X);
	}

	DBG( awt_evt, ("   ipl_getEvent\n"));
	UNBLOCK_EXECUTE( env, clazz, {
		ret = ipl_getEvent( &X->event);
	});
	DBG( awt_evt, ("   ret=%d, type=%d\n", ret, X->event.type));

	if ( ret ){
		switch ( X->event.type ){

		case EVT_MOUSE_MOVE:
			fgl_moveCursor( X->event.x, X->event.y);
			return motionNotify( env, X);

		case EVT_BUTTON_DOWN:
		case EVT_BUTTON_UP:
			return buttonNotify( env, X, X->event.type);

		case EVT_KEY_DOWN:
		case EVT_KEY_UP:
			return keyNotify( env, X, X->event.type);

		case EVT_VT_REL_REQ:
		case EVT_VT_ACQ_REQ:
			return vtSwitchNotify( env, X, X->event.type );
			  
		default:
			if ( (X->event.type >= EVT_GENERIC_FIRST) &&
				 (X->event.type <= EVT_GENERIC_LAST) ){
				return genericNotify( env, X, X->event.type, X->event.data, X->event.dlen);
			}
			
			if ( _(X)->pendingUpdates ) {
				DBG( awt_evt, ("   paintNotify\n"));
				return paintNotify( env, X);
			}

			DBG( awt_evt, ("   unkown event\n"));
			return 0;
		}
	}

	return 0;
}


jobject
EXPORT(evtPeekEvent) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_evt, ("peekEvent\n"));
	return NULL;
}


/*
 * We can't do a XPeekIfEvent here because it would block until such an event
 * comes in
 */
jobject
EXPORT(evtPeekEventId) ( JNIEnv* env, jclass clazz, jint id )
{
	DBG( awt_evt, ("peekEventId: %d\n", id));
	return NULL;
}


/*
 * This is just used to wakeup the getNextEvent call if we have a TLK_IS_BLOCKING,
 * non- NATIVE_DISPATCHER_LOOP style AWT, and we post an event from outside the
 * dispatcher thread (which might be blocked in a native getNextEvent).
 * This is a tribute to the fact that we actually have two different queues - the
 * native one, and the queue for self-posted events
 */
void
EXPORT(evtWakeup) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_evt, ("evtWakeup\n"));

	ipl_wakeUp();
}

/*
 * This is a helper for (IS_DISPATCH_EXCLUSIVE | NATIVE_DISPATCHER_LOOP) windowing
 * systems, where we need to execute certain requests (e.g. window creation) in
 * the native event thread. Not required, here
 */
void
EXPORT(evtSendWMEvent) ( JNIEnv* env, jclass clazz, jobject wmEvt )
{
	DBG( awt_evt, ("evtSendWMEvent\n"));
	/* not required, we have a NATIVE_DISPATCHER_LOOP */
}
