/**
 * gra.c - platform specific drawing support (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <string.h>

/* get us inlined definitions for all appropriate fgl functions */
#ifndef DEBUG
#define FGL_INLINE_LINE
#define FGL_INLINE_RECT
#define FGL_INLINE_OVAL
#define FGL_INLINE_POLY
#define FGL_INLINE_FONT
#endif
#include "fgl.h"

#include "toolkit.h"

/*****************************************************************************
 * generic exported methods
 */

#define _graInitGraphics   EXPORT(graInitGraphics)
#define _graFreeGraphics   EXPORT(graFreeGraphics)

#define _graCopyArea       EXPORT(graCopyArea)
#define _graClearRect      EXPORT(graClearRect)
#define _graDrawLine       EXPORT(graDrawLine)
#define _graDrawArc        EXPORT(graDrawArc)
#define _graFillArc        EXPORT(graFillArc)
#define _graDrawOval       EXPORT(graDrawOval)
#define _graFillOval       EXPORT(graFillOval)
#define _graDrawRect       EXPORT(graDrawRect)
#define _graFillRect       EXPORT(graFillRect)
#define _graDrawPolygon    EXPORT(graDrawPolygon)
#define _graDrawPolyline   EXPORT(graDrawPolyline)
#define _graFillPolygon    EXPORT(graFillPolygon)

#define _graDrawRoundRect  EXPORT(graDrawRoundRect)
#define _graFillRoundRect  EXPORT(graFillRoundRect)
#define _graDraw3DRect     EXPORT(graDraw3DRect)
#define _graFill3DRect     EXPORT(graFill3DRect)

#define _graDrawBytes      EXPORT(graDrawBytes)
#define _graDrawChars      EXPORT(graDrawChars)
#define _graDrawString     EXPORT(graDrawString)
#define _graDrawVertString EXPORT(graDrawVertString)

#define _graAddClip        EXPORT(graAddClip)
#define _graSetClip        EXPORT(graSetClip)
#define _graSetOffset      EXPORT(graSetOffset)
#define _graSetVisible     EXPORT(graSetVisible)
#define _graSetColor       EXPORT(graSetColor)
#define _graSetBackColor   EXPORT(graSetBackColor)
#define _graSetFont        EXPORT(graSetFont)
#define _graSetPaintMode   EXPORT(graSetPaintMode)
#define _graSetXORMode     EXPORT(graSetXORMode)

#define _graDrawImage      EXPORT(graDrawImage)

/*****************************************************************************
 * import generic Graphics functionality
 */

#include "../gra-common.c"



/******************************************************************************
 * platform specific image rendering
 */


void
EXPORT(graDrawImageScaled) ( JNIEnv* env, jclass clazz, 
							 Graphics* gr, Image* img,
							 jint dx0, jint dy0, jint dx1, jint dy1,
							 jint sx0, jint sy0, jint sx1, jint sy1,
							 jint bgval )
{
  int        x0, y0, x1, y1;
  int        iw = img->common.width;
  Image      tgt;

  if ( (!img) || (_(img)->width == -1) ){ /* last defense sanity check */
	return;
  }

  DBG( awt_gra, ("drawImageScaled: %x %x, %d,%d,%d,%d, %d,%d,%d,%d, %x\n",
				 gr, img, dx0, dy0, dx1, dy1, sx0, sy0, sx1, sy1, bgval));

  if ( dx1 > dx0 ) {
	x0 = dx0; x1 = dx1;
  }
  else {
	x0 = dx1; x1 = dx0;
  }
  if ( dy1 > dy0 ) {
	y0 = dy0; y1 = dy1;
  }
  else {
	y0 = dy1; y1 = dy0;
  }

  if ( sx0 < 0 ) sx0 = 0;
  if ( sx1 < 0 ) sx1 = 0;
  if ( sx0 >= iw ) sx0 = iw-1;
  if ( sx1 >= iw ) sx1 = iw-1;

  memset( (void*)&tgt, 0, sizeof(Image));
  tgt.common.width = (x1 - x0 +1);
  tgt.common.height = (y1 - y0 +1);

  g_initImage( &tgt);

  g_setAlphaType( &tgt, g_getAlphaType( img));
  initScaledImage( X, &tgt, img, dx0-x0, dy0-y0, dx1-x0, dy1-y0, sx0, sy0, sx1, sy1);

  Java_java_awt_Toolkit_graDrawImage ( env, clazz, gr, &tgt, 0, 0,
									   x0, y0, tgt.common.width, tgt.common.height,
									   bgval);


  fgl_disposeBitmap( tgt.bmp);

  if ( tgt.mask ){
	 fgl_disposeMonoBitmap( tgt.mask);
  }
  if ( tgt.alpha ) {
    fgl_disposeA8Bitmap( tgt.alpha);
  }
}

