/**
 * tlk.c - main initialization, query and termination
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include "toolkit.h"
#include "../tlkprops.h"


#include "../../../kaffe/kaffevm/gtypes.h"
#define	HAVE_BOOLEAN 1
#include "../../../kaffe/kaffevm/gc.h"
#include "../../../kaffe/kaffevm/threads.h"


#define	BANNER_SLEEP_TIME	2

extern void initColorMapping ( JNIEnv* env, Toolkit* X );
extern void initCursors ( Toolkit* X );
extern void initWindows ( Toolkit* X );
extern void initFonts ( Toolkit* X );

/* for banner display */
extern void Java_java_awt_Toolkit_imgFreeImage( JNIEnv * env, jclass clazz, Image * img);
extern void* Java_java_awt_Toolkit_imgCreateFromFile(JNIEnv*,jclass,jstring);


/********************************************************************************
 * exported functions
 */

jboolean
Java_java_awt_NativeCollector_ncInit(JNIEnv * env, jclass clazz, jstring name)
{
	/* nothing to do, we don't have a NATIVE_DISPATCHER_LOOP */
	return JNI_FALSE;
}

void
Java_java_awt_NativeCollector_ncLoop(JNIEnv * env, jclass clazz)
{
	/* nothing to do, we don't have a NATIVE_DISPATCHER_LOOP */
}

/*
 * we do this native because of two reasons: (1) this should appear as fast as
 * possible (without being deferred by potential lengthy Java inits), (2) in
 * native window systems we can't draw to the "root-window" from within Java
 */
void
EXPORT(tlkDisplayBanner) ( JNIEnv* env, jclass clazz, jstring banner )
{
	/* not yet */
}


jint 
EXPORT(tlkProperties) ( JNIEnv* env, jclass clazz )
{
	return TLK_IS_BLOCKING;
}


jboolean
EXPORT(tlkInit) ( JNIEnv* env, jclass clazz, jstring name )
{
	int nRgb = 0;
	int rgbs[MAX_RGB_REQUESTS];

#if defined (IS_CMAP_MODE)
	extern int getRgbRequests ( JNIEnv* env, jclass clazz, int *rgbs, int max );
	nRgb = getRgbRequests( env, clazz, rgbs, MAX_RGB_REQUESTS);
#endif

 	if ( !fgl_init( nRgb, rgbs, JNI_TRUE) ){
		DBG( awt, ("fgl_init failed\n"));
		return JNI_FALSE;
	}

	_(X)->nBuf = 128;
	_(X)->buf = AWT_MALLOC( _(X)->nBuf);

	_(X)->width = fgl_getScreenWidth();
	_(X)->height = fgl_getScreenHeight();

        _(X)->xoff = fgl_getScreenXoff();
        _(X)->yoff = fgl_getScreenYoff();

 	if ( !ipl_init( _(X)->width, _(X)->height) ){
		DBG( awt, ("ipl_init failed\n"));
		return JNI_FALSE;
	}

	/* Init subcomponents. We skip event (keyboard, mouse) because it
	 * might involve starting the dispatcher thread, which might be polling
	 * (i.e. slowing down the init thread). Color is deferred initialized
	 * to avoid init recursion)
	 */
	initFonts( X);
	initWindows( X);
	//initCursors( X);

	return JNI_TRUE;
}

void
EXPORT(tlkTerminate) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("terminate\n"));
}

jstring
EXPORT(tlkVersion) ( JNIEnv* env, jclass clazz )
{
	return (*env)->NewStringUTF( env, "fgl-1.0");
}

jint
EXPORT(tlkGetResolution) ( JNIEnv* env, jclass clazz )
{
	return 75;
}

jint
EXPORT(tlkGetScreenHeight) ( JNIEnv* env, jclass clazz )
{
	int h = fgl_getScreenHeight(); 

	DBG( awt_tlk, ("screenHeight: %d\n", h));
	return h;
}


jint
EXPORT(tlkGetScreenWidth) ( JNIEnv* env, jclass clazz )
{
	int w = fgl_getScreenWidth(); 

	DBG( awt_tlk, ("screenWidth: %d\n", w));
	return w;
}

jint
EXPORT(tlkGetScreenXoff) ( JNIEnv* env, jclass clazz )
{
	int x = fgl_getScreenXoff(); 

	DBG( awt_tlk, ("screenXoff: %d\n", x));
	return x;
}

jint
EXPORT(tlkGetScreenYoff) ( JNIEnv* env, jclass clazz )
{
	int y = fgl_getScreenYoff(); 

	DBG( awt_tlk, ("screenYoff: %d\n", y));
	return y;
}

void
EXPORT(tlkSync) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("sync\n"));
}


void
EXPORT(tlkFlush) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("flush\n"));
}


void
EXPORT(tlkBeep) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("beep\n"));
}

void
EXPORT(scrSetRotation) ( JNIEnv* env, jclass clazz, jint rot )
{
#ifdef LIB_FGL_0_0  
  AwtWnd *w;
  fgl_setScreenRotation ( (char)rot );
  ipl_setScreenRotation ( (char)rot );
  _(X)->width = fgl_getScreenWidth();
  _(X)->height = fgl_getScreenHeight();

  if (_(X)->root != 0) {
      _(X)->root->bounds.x1 = fgl_getScreenWidth() - 1;
      _(X)->root->bounds.y1 = fgl_getScreenHeight() - 1;

      setRect( _(X)->root->visSegments, 0, 0, _(X)->width - 1, _(X)->height - 1, 0);

      for ( w = _(X)->top ; w ; w = w->next ) {
          recalcVisibility( _(X)->top, w );
      }
  }
#endif
}

int
EXPORT(scrGetRotation) ( JNIEnv* env, jclass clazz )
{
#ifdef LIB_FGL_0_0
  return fgl_getScreenRotation();
#else
  return 0;
#endif
}
