/*
 * wnd.c - target lib specific window management (mainly cursor handling)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

/*****************************************************************************
 * use generic versions of general toplevel-related native methods
 */

#define _wndSetTitle             EXPORT(wndSetTitle)
#define _wndSetResizable         EXPORT(wndSetResizable)
#define _wndCreateFrame          EXPORT(wndCreateFrame)
#define _wndCreateWindow         EXPORT(wndCreateWindow)
#define _wndCreateDialog         EXPORT(wndCreateDialog)
#define _wndRequestFocus         EXPORT(wndRequestFocus)
#define _wndSetBounds            EXPORT(wndSetBounds)
#define _wndRepaint              EXPORT(wndRepaint)
#define _wndSetIcon              EXPORT(wndSetIcon)
#define _wndSetVisible           EXPORT(wndSetVisible)
#define _wndToBack               EXPORT(wndToBack)
#define _wndToFront              EXPORT(wndToFront)
#define _wndDestroyWindow        EXPORT(wndDestroyWindow)
#define _wndSetFrameInsets       EXPORT(wndSetFrameInsets)
#define _wndSetDialogInsets      EXPORT(wndSetDialogInsets)


/*****************************************************************************
 * include generic functionality
 */
#include "../wnd-common.c"


/*****************************************************************************
 * platform specific functions
 */

#define CURSOR_PEL_TYPE u8 

#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
#include "../cursors.h"

u32 spd_cmap[] = {
	0x000000,
	0xffff00,
	0xff0000
};


fgl_sprite_t*
getCursor ( Toolkit* X, int jCursor )
{
	fgl_spriteData_t spd;
	fgl_sprite_t  *cur;
	int       i = jCursor;

	if ( i > 13 ) i = 0;
	else if ( i < 0 ) i = 0;

	if ( (cur = X->cursors[i]) == 0 ) {
		spd.width   = cursorSize;
		spd.height  = cursorSize;
		spd.xhot    = cursorHotSpotX[i];
		spd.yhot    = cursorHotSpotY[i];
		spd.nColors = 3;
		spd.transp  = 0;
		spd.cmap    = spd_cmap;
		spd.data    = cursorShapes[i];

		cur = fgl_createSprite( &spd);

		X->cursors[i] = cur;
	}

	return cur;
}
#endif /* USE_xx_CURSOR */


void
initCursors ( Toolkit* X )
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
	fgl_setCursor( getCursor( X, 0));
#endif /* USE_xx_CURSOR */
}

void
EXPORT(wndSetCursor) ( JNIEnv* env, jclass clazz, void* wnd, jint jCursor )
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
	fgl_sprite_t *curs;
  
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	curs = getCursor( X, jCursor);

	DBG( awt_wnd, ("setCursor: %p  %d %p\n", wnd, jCursor, curs));

	fgl_setCursor( curs);
#endif /* USE_xx_CURSOR */
}

void
EXPORT(wndShowCursor) ( JNIEnv* env, jclass clazz, jboolean showIt )
{
#if defined (USE_SW_CURSOR) || defined (USE_HW_CURSOR)
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	DBG( awt_wnd, ("showCursor: %d\n", showIt));

	if ( showIt ) {
		fgl_showCursor();
	}
	else {
		fgl_hideCursor();
	}

#endif /* USE_xx_CURSOR */
}
