/**
 * fnt-common.c - common font handling
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */



/*
 * since we use nativ font objects for both Font and FontMetrics queries, we
 * need to determine when it is safe to free the native font. The alternative
 * would be to use a WeakReferences in Font/FontMetrics, but our classes have
 * to deal with environments where there are no WeakReferences.
 * The alternative would be to wrap font objects, but then the graphics lib
 * wouldn't be free to cache font objects internally.
 * Since it it's unlikely that we have programs using hundreths of resident
 * font objects, we are probably better of if we use a simple array instead
 * of a hashtable
 */


#define MAX_FONTS 40

static void*                inUseFs[MAX_FONTS];
static unsigned short       inUseCount[MAX_FONTS];


static
int registerFs ( void* fs )
{
	register int i;

	/* check if there already is an entry */
	for ( i=0; i<MAX_FONTS; i++ ){
		if ( inUseFs[i] == fs ){
			inUseCount[i]++;
			return 1;
		}
	}

	/* no, there isn't - add a new entry */
	for ( i=0; i<MAX_FONTS; i++ ){
		if ( !inUseFs[i] ){
			inUseFs[i] = fs;
			inUseCount[i] = 1;
			return 1;
		}
	}

	/* outch, out of font registry space */
	DBG( awt_fnt, ("can't register font: %p (more than %d fonts)\n",
				   fs, MAX_FONTS));
	return 0;
}


/* unregister font, return '1' if it's safe to delete, '0' otherwise */
static
int unregisterFs ( void* fs )
{
	register int i;

	for ( i=0; i<MAX_FONTS; i++ ){
		if ( inUseFs[i] == fs ){
			if ( (--inUseCount[i]) == 0 ){
				DBG( awt_fnt, ("free native font: %p\n", fs));
				inUseFs[i] = 0;
				return 1;
			}
			return 0;
		}
	}	

	return 0;
}
