/**
 * img.c - platform specific image support (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

/******************************************************************************
 * generic (common) exported methods
 */
#define _imgCreateImage         EXPORT(imgCreateImage)
#define _imgFreeImage           EXPORT(imgFreeImage)
#define _imgCreateScreenImage   EXPORT(imgCreateScreenImage)
#define _imgCreateScaledImage   EXPORT(imgCreateScaledImage)

#define _imgSetIdxPels          EXPORT(imgSetIdxPels)
#define _imgSetRGBPels          EXPORT(imgSetRGBPels)
#define _imgComplete            EXPORT(imgComplete)

#define _imgCreateFromFile      EXPORT(imgCreateFromFile)
#define _imgCreateFromData      EXPORT(imgCreateFromData)
#define _imgProduceImage        EXPORT(imgProduceImage)
#define _imgSetFrame            EXPORT(imgSetFrame)
#define _imgGetWidth            EXPORT(imgGetWidth)
#define _imgGetHeight           EXPORT(imgGetHeight)
#define _imgGetFrameCount       EXPORT(imgGetFrameCount)
#define _imgGetLoopCount        EXPORT(imgGetLoopCount)
#define _imgGetLatency          EXPORT(imgGetLatency)
#define _imgGetNextFrame        EXPORT(imgGetNextFrame)


/******************************************************************************
 * import generic functionality
 */
#include "../img-common.c"


/******************************************************************************
 * platform specific things
 */

#ifdef DEBUG
void printBin( unsigned int a )
{
	int i;
	for ( i=0; i<32; i++ ){
		putchar( ( a & 0x8000000 ) ? '1' : '0' );
		a <<= 1;
	}
}
#endif

void
createMaskContext ( Image* img )
{
	GrContext save;
	int x, y, k;

	GrSaveContext( &save);

	img->mask = GrCreateContext( _(img)->width, _(img)->height, 0, 0);
	GrSetContext( img->mask);
	GrClearContext( 0);

	/* by now, the common.mask bitfield has all visible pixel bits set
	 * (which is the inverse of our GRX mask bitmap)
	 */
	if ( _(img)->mask ) {
		MASKWORD *p = _(img)->mask->buf;
		for ( y=0; y < _(img)->height; y++ ){
			for ( x=0; x < _(img)->width; ){
				for ( k=0; (k<MASKWORDSIZE) && (x < _(img)->width); k++, x++ ){
					if ( ! (*p & imageMaskBits[k]) ){
						/* set all transparent pixels - we GrAND this with the source */
						_put_pixel_( img->mask, x, y, 0xffffffff);
					}
				}
				p++;
			}
		}
		freeImageMask( _(img)->mask);
		_(img)->mask = 0;
	}
	else if ( _(img)->alpha ) {
		unsigned int a;
		for ( y=0; y < _(img)->height; y++ ) {
			for ( x=0; x < _(img)->width; x++ ) {
				a = getAlpha( _(img)->alpha, x, y);
				if ( a < 128 ){
					_put_pixel_( img->mask, x, y, 0xffffff);
				}
			}
		}

		GrSetContext( img->gc);
		for ( y=0; y < _(img)->height; y++ ) {
			for ( x=0; x < _(img)->width; x++ ) {
				a = getAlpha( _(img)->alpha, x, y);
				if ( a < 128 ){
					_put_pixel_( img->gc, x, y, 0);
				}
			}
		}

		freeImageAlpha( _(img)->alpha);
		_(img)->alpha = 0;
	}

	GrSetContext( &save);
}
