/**
 * tlk.c - main initialization, query and termination
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include "toolkit.h"
#include "../tlkprops.h"

#define	BANNER_SLEEP_TIME	2

extern void initColorMapping ( JNIEnv* env, Toolkit* X );
extern void initCursors ( Toolkit* X );
extern void initWindows ( Toolkit* X );
extern void initFonts ( Toolkit* X );

/* for banner display */
extern void Java_java_awt_Toolkit_imgFreeImage( JNIEnv * env, jclass clazz, Image * img);
extern void* Java_java_awt_Toolkit_imgCreateFromFile(JNIEnv*,jclass,jstring);


/********************************************************************************
 * exported functions
 */

jboolean
Java_java_awt_NativeCollector_ncInit(JNIEnv * env, jclass clazz, jstring name)
{
	/* nothing to do, we don't have a NATIVE_DISPATCHER_LOOP */
	return JNI_FALSE;
}

void
Java_java_awt_NativeCollector_ncLoop(JNIEnv * env, jclass clazz)
{
	/* nothing to do, we don't have a NATIVE_DISPATCHER_LOOP */
}

/*
 * we do this native because of two reasons: (1) this should appear as fast as
 * possible (without being deferred by potential lengthy Java inits), (2) in
 * native window systems we can't draw to the "root-window" from within Java
 */
void
EXPORT(tlkDisplayBanner) ( JNIEnv* env, jclass clazz, jstring banner )
{
	Image* img;
	int    x, y;

	img = Java_java_awt_Toolkit_imgCreateFromFile(env, clazz, banner);
	if (img != 0) {

		DBG( awt_tlk, ("displayBanner for: %d\n", BANNER_SLEEP_TIME));

		x = (_(X)->width  - _(img)->width)  / 2;
		y = (_(X)->height - _(img)->height) / 2;

		GrBitBlt( (GrContext*)GrScreenContext(), 0, 0, img->gc, 
				  x, y, _(img)->width, _(img)->height, GrWRITE);
		Java_java_awt_Toolkit_imgFreeImage( env, clazz, img);

		sleep(BANNER_SLEEP_TIME);
	}
}


jint 
EXPORT(tlkProperties) ( JNIEnv* env, jclass clazz )
{
	jint    props = 0;

	/*
	 * We are not blocking, aren't dispatch-exclusive and don't need flushing. And,
	 * of course, we don't have a native dispatcher loop.
	 */

	return props;
}


jboolean
EXPORT(tlkInit) ( JNIEnv* env, jclass clazz, jstring name )
{
#if defined(SIGALRM)
	void*	      alarm_hand;
#endif
#if defined(SIGVTALRM)
	void*	      vtalarm_hand;
#endif

	_(X)->nBuf = 128;
	_(X)->buf = AWT_MALLOC( _(X)->nBuf);

	/* Save the alarm signal handlers so we can but it back after GRX has
	 * been setup.  We need this for SVGALIB which screwed it up.
	 */
#if defined(SIGALRM)
	alarm_hand = signal(SIGALRM, SIG_IGN);
#endif
#if defined(SIGVTALRM)
	vtalarm_hand = signal(SIGVTALRM, SIG_IGN);
#endif

	/*
	 * We use the standard GRX mode specification, which is done via the GRX20DRV env var:
	 *    SET GRX20DRV=<driver> gw <width> gh <height> nc <colors>
	 *    (e.g. export GRX20DRV="VESA gw 1024 gh 768 gc 32K"
	 *      or  export GRX20DRV="xwin gw 640 gh 480")
	 *
	 *  Available drivers are for
	 *    DOS  : stdvga, stdega, et4000, cl5426, mach64, ati28800, VESA,
	 *           memory
	 *    Linux: svgalib, memory
	 *    X11  : xwin, memory
	 */
	GrSetMode( GR_default_graphics);
	_(X)->width = GrScreenX();
	_(X)->height = GrScreenY();

	/* Init subcomponents. We skip event (keyboard, mouse) because it
	 * might involve starting the dispatcher thread, which might be polling
	 * (i.e. slowing down the init thread). Color is deferred initialized
	 * to avoid init recursion)
	 */
	initFonts( X);
	initWindows( X);

#if defined(SIGALRM)
	signal(SIGALRM, alarm_hand);
#endif
#if defined(SIGVTALRM)
	signal(SIGVTALRM, vtalarm_hand);
#endif

	return JNI_TRUE;
}

void
EXPORT(tlkTerminate) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("terminate\n"));
}

jstring
EXPORT(tlkVersion) ( JNIEnv* env, jclass clazz )
{
	return (*env)->NewStringUTF( env, "grx-1.0");
}

jint
EXPORT(tlkGetResolution) ( JNIEnv* env, jclass clazz )
{
	return 75;
}

jint
EXPORT(tlkGetScreenHeight) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("screenHeight: %d\n", GrScreenY()));

	return GrScreenY();
}


jint
EXPORT(tlkGetScreenWidth) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("screenWidth: %d\n", GrScreenX()));

	return GrScreenX();
}


void
EXPORT(tlkSync) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("sync\n"));
}


void
EXPORT(tlkFlush) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("flush\n"));
}


void
EXPORT(tlkBeep) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("beep\n"));
}
