/**
 * toolkit.h - main target lib specific awt header
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: toolkit.h,v 1.24 2001/07/23 23:03:11 samc Exp $
 */

#ifndef __toolkit_h
#define __toolkit_h

#include "grx20.h"

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-signal.h"
#include "../../../kaffe/kaffevm/jdebug.h"
#include "../../../kaffe/kaffevm/jmeasure.h"
#include "jfilesystem.h"

#include <jni.h>

/* to get the UNBLOCK_EXECUTE macro */
#include "jlocks.h"

/* this just brings in generic types (to be used in toolkit-common-pre.h) */
#include "grlib.h"

/* this holds all common stuff not relying on target specific types
 *(outside of grlib.def)
 */
#include "../toolkit-common.h"

#define USE_FULL_ALPHA


/*******************************************************************************
 * target lib specific types
 */

typedef struct _Image {
	CommonImage      common;

	GrContext        *gc;      /* holds the image data */
	GrContext        *mask;    /* transparent mask (for icons, transp. images) */
} Image;


typedef struct _Graphics {
	CommonGraphics  common;

	GrContext       *gc;

	int             fg, bg;
	int             xclr;
	int             clr;
	int             alpha;

	GrTextOption    txo;
} Graphics;

#define IMAGE_TGT(_gr)  ((Image*)((_gr)->common.tgt))


typedef struct _Toolkit {
	CommonToolkit  common;

	char           *fsPath;   /* path where to look for font files */
	int            fsOffset;  /* starting font lookup size offset */
	int            fsBigger;  /* max font lookup size increase */
	int            fsSmaller; /* max font lookup size decrease */

	int            evtId;
	char           keyUp, keyDown, buttonUp, buttonDown, mouseMotion;
	int            addedKbStat, removedKbStat;
	int            lastKbStat;
	GrMouseEvent   event;

	int            nKeyCodes;
	jbyte          *keyCodes;
	int            nKeyChars;
	jchar          *keyChars;

	GrCursor       *cursors[14];

	Graphics       *lastGr;
	GrContext      *lastGc;
} Toolkit;


/*******************************************************************************
 * this is the global data section
 */
#ifdef MAIN  /*****************************************************************/

Toolkit   Tk;
Toolkit   *X = &Tk;

#else        /*****************************************************************/

extern Toolkit* X;

#endif       /*****************************************************************/


/* now bring in the functions from grlib.h */
#define GRLIB_FUNCS
#include "grlib.h"

#endif
