/*
 * wnd.c - target lib specific window management (mainly cursor handling)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

/*****************************************************************************
 * use generic versions of general toplevel-related native methods
 */

#define _wndSetTitle             EXPORT(wndSetTitle)
#define _wndSetResizable         EXPORT(wndSetResizable)
#define _wndCreateFrame          EXPORT(wndCreateFrame)
#define _wndCreateWindow         EXPORT(wndCreateWindow)
#define _wndCreateDialog         EXPORT(wndCreateDialog)
#define _wndRequestFocus         EXPORT(wndRequestFocus)
#define _wndSetBounds            EXPORT(wndSetBounds)
#define _wndRepaint              EXPORT(wndRepaint)
#define _wndSetIcon              EXPORT(wndSetIcon)
#define _wndSetVisible           EXPORT(wndSetVisible)
#define _wndToBack               EXPORT(wndToBack)
#define _wndToFront              EXPORT(wndToFront)
#define _wndDestroyWindow        EXPORT(wndDestroyWindow)
#define _wndSetFrameInsets       EXPORT(wndSetFrameInsets)
#define _wndSetDialogInsets      EXPORT(wndSetDialogInsets)


/*****************************************************************************
 * include generic functionality
 */
#include "../wnd-common.c"


/*****************************************************************************
 * platform specific functions
 */

#include "../cursors.h"

void
initCursors ( Toolkit* X )
{
	cursorColors[0] = 2;
	cursorColors[1] = _(X)->white;
	cursorColors[2] = _(X)->black;

	X->cursors[0] = GrBuildCursor( cursorShapes[0],
								   cursorSize, cursorSize, cursorSize,
								   cursorHotSpotX[0], cursorHotSpotY[0],
								   (GrColor*)cursorColors);
}

GrCursor*
getCursor ( Toolkit* X, int jCursor )
{
	GrCursor  *cur;
	int       i = jCursor;

	if ( i > 13 ) i = 0;
	else if ( i < 0 ) i = 0;

	if ( (cur = X->cursors[i]) == 0 ) {
		X->cursors[i] = GrBuildCursor( cursorShapes[i],
									   cursorSize, cursorSize, cursorSize,
									   cursorHotSpotX[i], cursorHotSpotY[i],
									   (GrColor*)cursorColors);
		cur = X->cursors[i];
	}

	return cur;
}


void
EXPORT(wndSetCursor) ( JNIEnv* env, jclass clazz, void* wnd, jint jCursor )
{
	GrCursor *curs;
  
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	curs = getCursor( X, jCursor);

	DBG( awt_wnd, ("setCursor: %p  %d %p\n", wnd, jCursor, curs));

	GrMouseEraseCursor();
	GrMouseSetCursor( curs);
	GrMouseDisplayCursor();
}


void
EXPORT(wndShowCursor) ( JNIEnv* env, jclass clazz, jboolean showIt )
{
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	DBG( awt_wnd, ("showCursor: %d\n", showIt));

	if ( showIt ) {
		GrMouseDisplayCursor();
	}
	else {
		GrMouseEraseCursor();
	}
}
