/**
 * evt.c - MGL event handling
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.h"

/******************************************************************************
 * generic exported methods
 */

#define _evtInit              EXPORT(evtInit)
#define _evtRegisterSource    EXPORT(evtRegisterSource)
#define _evtUnregisterSource  EXPORT(evtUnregisterSource)


#include "keytab.h"
#include "../evt-common.c"


/******************************************************************************
 * exported event aquisition/post functions
 */

jobject
EXPORT(evtGetNextEvent) ( JNIEnv* env, jclass clazz )
{
	jobject jEvt = NULL;

	DBG( awt_evt, ("getNextEvent\n"));
  
	if ( X->preFetched || EVT_getNext(&X->event, (EVT_KEYEVT | EVT_MOUSEEVT)) ){
		switch ( X->event.what ){
		case EVT_KEYDOWN:
		case EVT_KEYREPEAT:
			jEvt = keyNotify( env, X, KEY_PRESSED);
			break;
		case EVT_KEYUP:
			jEvt = keyNotify( env, X, KEY_RELEASED);
			break;
		case EVT_MOUSEDOWN:
			jEvt = buttonNotify( env, X, MOUSE_PRESSED);
			break;
		case EVT_MOUSEUP:
			jEvt = buttonNotify( env, X, MOUSE_RELEASED);
			break;
		case EVT_MOUSEMOVE:
			jEvt = motionNotify( env, X);
			break;
		}

		X->preFetched = 0;
	}
	else {
		/* process paint events after all input events */
		if ( _(X)->pendingUpdates ) {
			return paintNotify( env, X);
		}
		else {
			return 0;
		}
	}

	return jEvt;
}


jobject
EXPORT(evtPeekEvent) ( JNIEnv* env, jclass clazz )
{
  DBG( awt_evt, ("peekEvent\n"));

  return NULL;
}


/*
 * We can't do a XPeekIfEvent here because it would block until such an event
 * comes in
 */
jobject
EXPORT(evtPeekEventId) ( JNIEnv* env, jclass clazz, jint id )
{
  DBG( awt_evt, ("peekEventId: %d\n", id));

  return NULL;
}


/*
 * This is just used to wakeup the getNextEvent call if we are multithreaded,
 * and we post an event from outside the dispatcher thread (which might be blocked
 * in a native getNextEvent). This is a tribute to the fact that we actually have
 * two different queues - the native one, and the queue for self-posted events
 */
void
EXPORT(evtWakeup) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_evt, ("evtWakeup\n"));
}

/*
 * This is a helper for (IS_DISPATCH_EXCLUSIVE | NATIVE_DISPATCHER_LOOP) windowing
 * systems, where we need to execute certain requests (e.g. window creation) to
 * the native event thread. Not required here
 */
void
EXPORT(evtSendWMEvent) ( JNIEnv* env, jclass clazz, jobject wmEvt )
{
}
