/**
 * tlk.c - main initialization, query and termination
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include "toolkit.h"
#include "../tlkprops.h"

#if defined(DEBUG)
#include "modenames.h"
#endif

#define	BANNER_SLEEP_TIME	2

extern void initColorMapping ( JNIEnv* env, Toolkit* X );
extern void initCursors ( Toolkit* X );
extern void initWindows ( Toolkit* X );
extern void initFonts ( Toolkit* X );

/* for banner display */
extern void Java_java_awt_Toolkit_imgFreeImage( JNIEnv * env, jclass clazz, Image * img);
extern void* Java_java_awt_Toolkit_imgCreateFromFile(JNIEnv*,jclass,jstring);


/********************************************************************************
 * exported functions
 */

jboolean
Java_java_awt_NativeCollector_ncInit(JNIEnv * env, jclass clazz, jstring name)
{
	/* nothing to do, we don't have a NATIVE_DISPATCHER_LOOP */
	return JNI_FALSE;
}

void
Java_java_awt_NativeCollector_ncLoop(JNIEnv * env, jclass clazz)
{
	/* nothing to do, we don't have a NATIVE_DISPATCHER_LOOP */
}

/*
 * we do this native because of two reasons: (1) this should appear as fast as
 * possible (without being deferred by potential lengthy Java inits), (2) in
 * native window systems we can't draw to the "root-window" from within Java
 */
void
Java_java_awt_Toolkit_tlkDisplayBanner ( JNIEnv* env, jclass clazz, jstring banner )
{
}


jint 
Java_java_awt_Toolkit_tlkProperties ( JNIEnv* env, jclass clazz )
{
	jint    props = 0;

	/*
	 * We are not blocking, aren't dispatch-exclusive and don't need flushing. And,
	 * of course, we don't have a native dispatcher loop.
	 */

	return props;
}


void
terminateMGL ()
{
	MGL_exit();
}

jboolean
Java_java_awt_Toolkit_tlkInit ( JNIEnv* env, jclass clazz, jstring name )
{
	int mode = grDETECT;
	int driver = grDETECT;
	char* mglPath = getenv( "MGL_ROOT");
	char* mglMode = getenv( "MGL_DRV");
	char  drv[50];
	int   xres, yres, depth, xr, yr, d;
	uchar* modes;

	/* use this as a sync point for remote debugging ***********************/
	atexit( terminateMGL); /* need this to restore console mode (on signal 1) */
	//puts( "hit <CR> to start..");
	//getchar();
	/* end sync, now under realm of debugger *******************************/

	_(X)->nBuf = 128;
	_(X)->buf = AWT_MALLOC( _(X)->nBuf);

	/* init driver (dev and demo version) */
	MGL_unregisterAllDrivers();

#ifdef NEVER
	if ( mglMode && *mglMode ) {
		if ( sscanf( mglMode, "%[^_]_%dx%dx%d", drv, &xres, &yres, &depth) < 4 ){
			DBG( awt, ("illegal mode request: %s\n", mglMode));
			return JNI_FALSE;
		}
		DBG( awt, ("requested mode: %s, %dx%d, %d bpp\n", drv, xres, yres, depth));

		if ( strchr( drv, 'V') ){
			MGL_registerDriver( MGL_VGA4NAME, VGA4_driver);
			driver = grVGA;
		}

		if ( strchr( drv, 'S') ) {
			MGL_registerDriver( MGL_SVGA4NAME, SVGA4_driver);
			MGL_registerDriver( MGL_SVGA8NAME, SVGA8_driver);
			MGL_registerDriver( MGL_SVGA16NAME, SVGA16_driver);

			// MGL_registerDriver( MGL_SVGA24NAME, SVGA24_driver);
			// MGL_registerDriver( MGL_SVGA32NAME, SVGA32_driver);
			driver = grSVGA;
		}

		if ( strchr( drv, 'L') ){
			MGL_registerDriver( MGL_LINEAR8NAME, LINEAR8_driver);
			MGL_registerDriver( MGL_LINEAR16NAME, LINEAR16_driver);
			MGL_registerDriver( MGL_LINEAR24NAME, LINEAR24_driver);
			MGL_registerDriver( MGL_LINEAR32NAME, LINEAR32_driver);
			driver = grVESA;
		}

		if ( strchr( drv, 'A') ){
			MGL_registerDriver( MGL_ACCEL8NAME, ACCEL8_driver);
			MGL_registerDriver( MGL_ACCEL16NAME, ACCEL16_driver);
			MGL_registerDriver( MGL_ACCEL24NAME, ACCEL24_driver);
			MGL_registerDriver( MGL_ACCEL32NAME, ACCEL32_driver);
			driver = grACCEL;
		}

		mode = grDETECT;
		MGL_detectGraph( &driver, &mode);
		if ( mode == grNONE ) {
			/* trouble, no suitable HW/driver detected */
			DBG( awt, ("MGL_detectGraph failed"));
			return JNI_FALSE;
		}

		modes = MGL_availableModes();
		while ( *modes != 0xff ) {
			DBG( awt, ("detected mode %d : %s\n", *modes, modeNames[*modes]));

			MGL_modeResolution( *modes, &xr, &yr, &d);
			if ( (xres == xr) && (yres == yr) && (depth == d) ){
				DBG( awt, ("matching mode\n"));
				mode = *modes;
				break;
			}
			modes++;
		}
	}
#else
	mode = grSVGA_640x480x256;
	driver = grSVGA;
	MGL_registerAllDispDrivers( 1, 1, 1);
#endif

	/* fallback, check all available drivers */
	if ( mode == grDETECT ) {
		MGL_registerAllDispDrivers( 1, 1, 1);
	}

	/* we need them for offscreens and bitmaps */
	MGL_registerAllMemDrivers();

	if ( !MGL_init( &driver, &mode, mglPath) ){
		DBG( awt, ("MGL_init failed: %d, %d: %s\n", driver, mode,
				   MGL_errorMsg( MGL_result())));
		return JNI_FALSE;
	}

	if ( !(X->gc = MGL_createDisplayDC( 0)) ){
		DBG( awt, ("MGL_createDC failed: %d, %d: %s\n", driver, mode,
				   MGL_errorMsg( MGL_result())));
		return JNI_FALSE;
	}


	_(X)->width = MGL_sizex( X->gc)+1;
	_(X)->height = MGL_sizey( X->gc)+1;

	DBG( awt, ("init driver: %d, mode: %d, size: %dx%d\n",
			   driver, mode, _(X)->width, _(X)->height));

	initFonts( X);
	initWindows( X);

	return JNI_TRUE;
}

void
Java_java_awt_Toolkit_tlkTerminate ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("terminate\n"));
	MGL_exit();
}

jstring
Java_java_awt_Toolkit_tlkVersion ( JNIEnv* env, jclass clazz )
{
	return (*env)->NewStringUTF( env, "mgl-1.0");
}

jint
Java_java_awt_Toolkit_tlkGetResolution ( JNIEnv* env, jclass clazz )
{
	return 75;
}

jint
Java_java_awt_Toolkit_tlkGetScreenHeight ( JNIEnv* env, jclass clazz )
{
	jint h = MGL_sizey( X->gc)+1;

	DBG( awt_tlk, ("screenHeight: %d\n", h));

	return h;
}


jint
Java_java_awt_Toolkit_tlkGetScreenWidth ( JNIEnv* env, jclass clazz )
{
	jint w = MGL_sizex( X->gc)+1;

	DBG( awt_tlk, ("screenWidth: %d\n", w));

	return w;
}


void
Java_java_awt_Toolkit_tlkSync ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("sync\n"));
}


void
Java_java_awt_Toolkit_tlkFlush ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("flush\n"));
}


void
Java_java_awt_Toolkit_tlkBeep ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("beep\n"));
}
