/**
 * toolkit.h - main target lib specific awt header
 *
 * Copyright (c) 1998
 *		Transvirtual Technologies, Inc.	 All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: toolkit.h,v 1.7 2001/07/23 23:03:11 samc Exp $
 */

#ifndef __toolkit_h
#define __toolkit_h

#include "mgraph.h"

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-signal.h"
#include "../../../kaffe/kaffevm/jdebug.h"
#include "../../../kaffe/kaffevm/jmeasure.h"
#include "jfilesystem.h"

#include <jni.h>


/* this just brings in generic types (to be used in toolkit-common-pre.h) */
#include "grlib.h"

/* this holds all common stuff not relying on target specific types
 *(outside of grlib.def)
 */
#include "../toolkit-common.h"


/*******************************************************************************
 * target lib specific types
 */

typedef struct _Image {
	CommonImage		 common;

	/* image pixel colors */
	MGLDC			 *gc;		 /* for drawable (screen) images */
	bitmap_t		 *bmp;		 /* for immutable (external) bitmaps */

	bitmap_t		 *mask;		 /* (optional) monochrome mask bitmap */
} Image;


typedef struct {
	CommonGraphics	common;

	MGLDC			*gc;

	font_t          *fnt;

	int				fg, bg;
	int				xclr;
	int				clr;
} Graphics;


typedef struct _Toolkit {
	CommonToolkit  common;

	palette_t	   *pal;	    /* display palette */
	pixel_format_t *pf;		    /* dislay pixel format */

	int			   evtId;
	event_t        event;
	int            preFetched;  /* don't call EVT_getNext */

	int			   nKeyCodes;
	jbyte		   *keyCodes;

	cursor_t	   *cursors[14];

	MGLDC          *gc;         /* screen context */

	Graphics	   *lastGr;
	MGLDC		   *lastGc;
} Toolkit;


/*******************************************************************************
 * this is the global data section
 */
#ifdef MAIN	 /*****************************************************************/

Toolkit	  Tk;
Toolkit	  *X = &Tk;

#else		 /*****************************************************************/

extern Toolkit* X;

#endif		 /*****************************************************************/


/*************************************************************************************
 * commonly used function prototypes
 */

void		 drawAlphaImage		 ( Graphics* gr, Image* img,
								   jint srcX, jint srcY, jint dstX, jint dstY,
								   jint width, jint height, jint bgval );
void		 retargetMouse		 ( JNIEnv* env, Toolkit* X );
void		 initCursors		 ( Toolkit* X);
void		 initScaledImage	 ( Toolkit* X, Image *tgt, Image *src,
								   int dx0, int dy0, int dx1, int dy1,
								   int sx0, int sy0, int sx1, int sy1 );
Image*		 createImage		 ( int width, int height, int isDrawable );
ImageAlpha*	 createImageAlpha	 ( Toolkit* X, int width, int height );
void		 freeImageAlpha		 ( ImageAlpha* img );
ImageMask*	 createImageMask	 ( Toolkit* X, int width, int height );
void		 freeImageMask		 ( ImageMask* img );
int			 needsFullAlpha		 ( ImageAlpha *aimg, double threshold );
void		 reduceAlpha		 ( Image* img, ImageAlpha* aimg, int threshold );
void		 Java_java_awt_Toolkit_imgFreeImage( JNIEnv* env, jclass clazz, Image* img);

void		 Java_java_awt_Toolkit_graFreeGraphics( JNIEnv* env, jclass clazz, Graphics* gr);
												
void         updateWindow        ( JNIEnv* env, Toolkit* X, AwtWnd*, LinkedRect* );

void         initColorMapping    ( JNIEnv* env, Toolkit* X);
jlong        Java_java_awt_Toolkit_clrBright( JNIEnv* env, jclass clazz, jint rgb );
jlong        Java_java_awt_Toolkit_clrDark  ( JNIEnv* env, jclass clazz, jint rgb );
jint         getVGAPixelValue    ( Toolkit* X, jint rgb, int mod );


/* now bring in the functions from grlib.h */
#define GRLIB_FUNCS
#include "grlib.h"

#endif
