/*
 * wnd.c - target lib specific window management (mainly cursor handling)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

/*****************************************************************************
 * use generic versions of general toplevel-related native methods
 */

#define _wndSetTitle             EXPORT(wndSetTitle)
#define _wndSetResizable         EXPORT(wndSetResizable)
#define _wndCreateFrame          EXPORT(wndCreateFrame)
#define _wndCreateWindow         EXPORT(wndCreateWindow)
#define _wndCreateDialog         EXPORT(wndCreateDialog)
#define _wndRequestFocus         EXPORT(wndRequestFocus)
#define _wndSetBounds            EXPORT(wndSetBounds)
#define _wndRepaint              EXPORT(wndRepaint)
#define _wndSetIcon              EXPORT(wndSetIcon)
#define _wndSetVisible           EXPORT(wndSetVisible)
#define _wndToBack               EXPORT(wndToBack)
#define _wndToFront              EXPORT(wndToFront)
#define _wndDestroyWindow        EXPORT(wndDestroyWindow)
#define _wndSetFrameInsets       EXPORT(wndSetFrameInsets)
#define _wndSetDialogInsets      EXPORT(wndSetDialogInsets)


/*****************************************************************************
 * include generic functionality
 */
#include "../wnd-common.c"


/*****************************************************************************
 * platform specific functions
 */

char* cursorName[] = {
	"arrow.cur",      /* defaultCursor */
	"move.cur",       /* crossHairCursor */
	"ibeam.cur",      /* textCursor */
	"wait.cur",       /* waitCursor */
	"sizenesw.cur",   /* swCursor */
	"sizenwse.cur",   /* seCursor */
	"sizenwse.cur",   /* nwCursor */
	"sizenesw.cur",   /* neCursor */
	"sizens.cur",     /* nCursor */
	"sizens.cur",     /* sCursor */
	"sizewe.cur",     /* wCursor */
	"sizewe.cur",     /* eCursor */
	"noentry.cur",    /* handCursor ?? Hmm we have to get one for this */
	"move.cur"        /* moveCursor */
};

void
initCursors ( Toolkit* X )
{
	X->cursors[0] = MGL_loadCursor( cursorName[0]);
}

cursor_t*
getCursor ( Toolkit* X, int jCursor )
{
	cursor_t  *cur;
	int       i = jCursor;

	if ( i > 13 ) i = 0;
	else if ( i < 0 ) i = 0;

	if ( (cur = X->cursors[i]) == 0 ) {
		X->cursors[i] = MGL_loadCursor( cursorName[i]);
		cur = X->cursors[i];
	}

	return cur;
}


void
EXPORT(wndSetCursor) ( JNIEnv* env, jclass clazz, void* wnd, jint jCursor )
{
	cursor_t *curs;
  
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	curs = getCursor( X, jCursor);
	MS_setCursor( curs);

	DBG( awt_wnd, ("setCursor: %p  %d %p\n", wnd, jCursor, curs));

	
}
