/**
 * gra.c - platform specific drawing support (GRX)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include <string.h>

#include "toolkit.h"

/*****************************************************************************
 * generic exported methods
 */

#define _graInitGraphics   EXPORT(graInitGraphics)
#define _graFreeGraphics   EXPORT(graFreeGraphics)

#define _graCopyArea       EXPORT(graCopyArea)
#define _graClearRect      EXPORT(graClearRect)
#define _graDrawLine       EXPORT(graDrawLine)
#define _graDrawArc        EXPORT(graDrawArc)
#define _graFillArc        EXPORT(graFillArc)
#define _graDrawOval       EXPORT(graDrawOval)
#define _graFillOval       EXPORT(graFillOval)
#define _graDrawRect       EXPORT(graDrawRect)
#define _graFillRect       EXPORT(graFillRect)
#define _graDrawPolygon    EXPORT(graDrawPolygon)
#define _graDrawPolyline   EXPORT(graDrawPolyline)
#define _graFillPolygon    EXPORT(graFillPolygon)

#define _graDrawRoundRect  EXPORT(graDrawRoundRect)
#define _graFillRoundRect  EXPORT(graFillRoundRect)
#define _graDraw3DRect     EXPORT(graDraw3DRect)
#define _graFill3DRect     EXPORT(graFill3DRect)

#define _graDrawBytes      EXPORT(graDrawBytes)
#define _graDrawChars      EXPORT(graDrawChars)
#define _graDrawString     EXPORT(graDrawString)

#define _graAddClip        EXPORT(graAddClip)
#define _graSetClip        EXPORT(graSetClip)
#define _graSetOffset      EXPORT(graSetOffset)
#define _graSetVisible     EXPORT(graSetVisible)
#define _graSetColor       EXPORT(graSetColor)
#define _graSetBackColor   EXPORT(graSetBackColor)
#define _graSetFont        EXPORT(graSetFont)
#define _graSetPaintMode   EXPORT(graSetPaintMode)
#define _graSetXORMode     EXPORT(graSetXORMode)

#define _graDrawImage      EXPORT(graDrawImage)

/*****************************************************************************
 * import generic Graphics functionality
 */

#include "../gra-common.c"



/******************************************************************************
 * platform specific image rendering
 */

void
drawAlphaImage ( Graphics* gr, Image* img,
				 jint srcX, jint srcY, jint dstX, jint dstY,
				 jint width, jint height, jint bgval )
{
	int       di, dj, si, sj, alpha;
	UGL_COLOR dpix, spix;
	int       sr, sg, sb, dr, dg, db;

	int       dimax = dstX + width;
	int       djmax = dstY + height;
	unsigned char ai;

	DBG( awt_gra, ("drawAlphaImage: %p %p (%p),  %d,%d  %d,%d  %d,%d  %x\n",
				   gr, img, img->common.alpha,
				   (int)srcX, (int)srcY,
				   (int)dstX, (int)dstY,
				   (int)width, (int)height, (unsigned)bgval));

	_setBmpGc( img);

	for ( dj=dstY, sj=srcY; dj<djmax; dj++, sj++ ) {
		for ( di=dstX, si=srcX; di<dimax; di++, si++ ) {

			uglPixelGet( gr->gc, di, dj, &dpix);
			uglPixelGet( X->bmpGc, si, sj, &spix);
			alpha = getAlpha( _(img)->alpha, si, sj);

			g_rgbValues( X, dpix, &dr, &dg, &db);
			g_rgbValues( X, spix, &sr, &sg, &sb);

			if ( alpha ) {
				/*
				 * ints still considered to be substantially faster than floats on most
				 * architectures (avoid div by 255), map all 0..255 -> 1..256
				 */
				alpha++;
				dr++; dg++; db++;
				sr++; sg++; sb++;

				ai = 256 - alpha;

				dr = ((alpha * sr  +  ai * dr) + 128) >> 8;
				dg = ((alpha * sg  +  ai * dg) + 128) >> 8;
				db = ((alpha * sb  +  ai * db) + 128) >> 8;

				dr--; dg--; db--;

				dpix = g_pixelValue( X, (dr << 16)|(dg << 8)|(db));
				uglPixelSet( gr->gc, di, dj, dpix);
			}
		}
	}
}


void
EXPORT(graDrawImageScaled) ( JNIEnv* env, jclass clazz, 
							 Graphics* gr, Image* img,
							 jint dx0, jint dy0, jint dx1, jint dy1,
							 jint sx0, jint sy0, jint sx1, jint sy1,
							 jint bgval )
{
	DBG( awt_gra, ("drawImageScaled: %p %p, %d,%d,%d,%d, %d,%d,%d,%d, %x\n",
				   gr, img, (int)dx0, (int)dy0, (int)dx1, (int)dy1, 
				   (int)sx0, (int)sy0, (int)sx1, (int)sy1, (unsigned)bgval));


	if ( (!img) || (_(img)->width == -1) ){ /* last defense sanity check */
		return;
	}

	if ( setContext( X, gr) ) {
		dx0 += _(gr)->x0;
		dx1 += _(gr)->x0;
		dy0 += _(gr)->y0;
		dy1 += _(gr)->y0;

		do {
			/* we could be more efficient by turning clipping into
			 * cropping, but we would have to maintain scale ratios
			 * (and it's unlikely that apps use scaled image drawing
			 * w/o doing some sort of viewport management by themselves)
			 */
			uglBitmapStretchBlt ( gr->gc, img->bmp,
								  sx0, sy0, sx1, sy1, 
								  UGL_DEFAULT_ID,
								  dx0, dy0, dx1, dy1);
		} while ( nextClipSegment( gr) );
	}
}
