/**
 * tlk.c - main initialization, query and termination
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include <time.h>

#include "toolkit.h"
#include "../tlkprops.h"

#define	BANNER_SLEEP_TIME	2

extern void initColorMapping ( JNIEnv* env, Toolkit* X );
extern void initCursors ( Toolkit* X );
extern void initWindows ( Toolkit* X );
extern void initFonts ( Toolkit* X );

/* for banner display */
extern void Java_java_awt_Toolkit_imgFreeImage( JNIEnv * env, jclass clazz, Image * img);
extern void* Java_java_awt_Toolkit_imgCreateFromFile(JNIEnv*,jclass,jstring);



jboolean
initUGL ( JNIEnv* env, jclass clazz )
{
	UGL_MODE_INFO mi;
	UGL_EVENT_HANDLER_ID ehid;

	if ( uglInitialize() == UGL_STATUS_ERROR ){
		DBG( awt, ("failed to initialize UGL\n"));
		return JNI_FALSE;		
	}

	if ( uglDriverFind( UGL_DISPLAY_TYPE, 0, (UGL_UINT32*)&(X->did)) == UGL_STATUS_ERROR ){
		DBG( awt, ("failed to lookup device id\n"));
		return JNI_FALSE;
	}

	if ( uglDriverFind( UGL_FONT_ENGINE_TYPE, 0, (UGL_UINT32*)&(X->fid)) == UGL_STATUS_ERROR ){
		DBG( awt, ("failed to lookup font driver\n"));
		return JNI_FALSE;
	}

	if ( uglDriverFind( UGL_EVENT_SERVICE_TYPE, 0, (UGL_UINT32*)&(X->eid)) == UGL_STATUS_ERROR ){
		DBG( awt, ("failed to lookup input service\n"));
		return JNI_FALSE;
	}
	else {
		X->qid = uglEventQCreate( X->eid, 32);
	}

	/* Java is a inherently multi-threaded environment. In order to avoid
	 * pointer lags because of looping mouse events through Java (with several
	 * potential context switches), we delegate the pointer drawing to the
	 * UGL input driver
	 */
    if ( uglMediaInit() == UGL_STATUS_ERROR ){
        DBG( awt, ("failed to init pointer callback\n"));
        return JNI_FALSE;
    }
    else {
        uglEventHandlerGet( X->eid, &ehid);
        uglEventCallbackAdd( ehid, UGL_EVENT_TYPE_POINTER,
                             UGL_EVENT_CATEGORY_INPUT, uglCallbackCursorMove);
    }
	
	_(X)->nBuf = 128;
	_(X)->buf = AWT_MALLOC( _(X)->nBuf);

	if ( (X->gc = uglGcCreate( X->did)) == NULL ){
		DBG( awt, ("failed to create display Gc\n"));
		return JNI_FALSE;		
	}

	uglDefaultBitmapGet( X->gc, &(X->screenBmp));

	if ( (X->bmpGc = uglGcCreate( X->did)) == NULL ){
		DBG( awt, ("failed to create bitmap Gc\n"));
		return JNI_FALSE;		
	}

	if ( uglInfo( X->did, UGL_MODE_INFO_REQ, &mi) == UGL_STATUS_ERROR){
		DBG( awt, ("failed to obtain mode info\n"));
		return JNI_FALSE;		
	}

	_(X)->width = mi.width;
	_(X)->height = mi.height;

	initFonts( X);
	initWindows( X);

	return JNI_TRUE;
}

/********************************************************************************
 * exported functions
 */


/*
 * This is the first native lib function to be called in a blocked, native
 * dispatched AWT. All upfront initialization goes in here
 */
jboolean
Java_java_awt_NativeCollector_ncInit(JNIEnv * env, jclass clazz, jstring name)
{
#if defined (USE_NATIVE_DISPATCHER_LOOP)
	return initUGL( env, clazz);
#else
	return JNI_TRUE;  /* shouldn't be called, then */
#endif
}
	
	
/*
 * input acquisition loop
 *      void  Java_java_awt_NativeCollector_ncLoop ( JNIEnv * env, jclass clazz )
 * resides in evt.c, since it uses a lot of event functions/defines
 */

/*
 * we do this native because of two reasons: (1) this should appear as fast as
 * possible (without being deferred by potential lengthy Java inits), (2) in
 * native window systems we can't draw to the "root-window" from within Java
 */
void
EXPORT(tlkDisplayBanner) ( JNIEnv* env, jclass clazz, jstring banner )
{
#ifdef NEVER
	Image* img;
	int    x, y;
	struct timespec ts;

	img = Java_java_awt_Toolkit_imgCreateFromFile(env, clazz, banner);
	if (img != 0) {

		DBG( awt_tlk, ("displayBanner for: %d\n", BANNER_SLEEP_TIME));

		x = (_(X)->width  - _(img)->width)  / 2;
		y = (_(X)->height - _(img)->height) / 2;

		uglBitmapBlt( X->gc, img->bmp, 
					  0, 0, _(img)->width, _(img)->height,
					  UGL_DEFAULT_ID, x, y);
		Java_java_awt_Toolkit_imgFreeImage( env, clazz, img);

		ts.tv_sec = BANNER_SLEEP_TIME;
		ts.tv_nsec = 0;
		nanosleep( &ts, NULL);
	}
#endif
}


jint 
EXPORT(tlkProperties) ( JNIEnv* env, jclass clazz )
{
#if defined (USE_NATIVE_DISPATCHER_LOOP)
    return TLK_NATIVE_DISPATCHER_LOOP | TLK_IS_BLOCKING;
#else
    return TLK_IS_BLOCKING;
#endif
}


jboolean
EXPORT(tlkInit) ( JNIEnv* env, jclass clazz, jstring name )
{
	/* nothing to do, here (it's all in ncInit) */
#if defined (USE_NATIVE_DISPATCHER_LOOP)
	return JNI_TRUE;  /* called, but everything has been set by ncInit, already */
#else
	return initUGL( env, clazz);
#endif
}

void
EXPORT(tlkTerminate) ( JNIEnv* env, jclass clazz )
{
	UGL_EVENT event;
	extern int stopEventLoop;

	DBG( awt_tlk, ("terminate\n"));
	stopEventLoop = 1;

	/* we use an artificial terminate event to make sure the
	 * AWT-event and driver tasks really terminate. The real
	 * termination is done from within ncLoop in respond to this event
	 */
	event.header.type = TERMINATE_EVENT;
	event.header.category = UGL_EVENT_CATEGORY_INPUT;
	uglEventPost( X->eid, X->qid, &event, sizeof( event));
}

jstring
EXPORT(tlkVersion) ( JNIEnv* env, jclass clazz )
{
	return (*env)->NewStringUTF( env, "ugl-1.0");
}

jint
EXPORT(tlkGetResolution) ( JNIEnv* env, jclass clazz )
{
	return 75;
}

jint
EXPORT(tlkGetScreenHeight) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("screenHeight: %d\n", _(X)->height));

	return _(X)->height;
}


jint
EXPORT(tlkGetScreenWidth) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("screenWidth: %d\n", _(X)->width));

	return _(X)->width;
}


void
EXPORT(tlkSync) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("sync\n"));
}


void
EXPORT(tlkFlush) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("flush\n"));
}


void
EXPORT(tlkBeep) ( JNIEnv* env, jclass clazz )
{
	DBG( awt_tlk, ("beep\n"));
}
