/**
 * toolkit.h - main target lib specific awt header
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: toolkit.h,v 1.29 2001/07/23 23:03:11 samc Exp $
 */

#ifndef __toolkit_h
#define __toolkit_h

#define USE_UGL_COLOR

#define TERMINATE_EVENT  0x8000000
#define WAKEUP_EVENT      0x8100000

#include "ugl/ugl.h"
#include "ugl/uglinput.h"
#include "ugl/ugldib.h"
#include "ugl/uglucode.h"
#include "ugl/uglmedia.h"

/* we don't use full alpha channels yet, because we can't display them efficiently */
#undef USE_FULL_ALPHA

#if defined(NO_KAFFE)

/* don't forget to define the required image decoders:
 * HAVE_GIF_LIB_H  +  HAVE_LIBJPEG
 * HAVE_JPEGLIB_H  +  HAVE_LIBUNGIF
 * HAVE_TIFFIO_H   +  HAVE_LIBTIFF
 * HAVE_PNG_H      +  HAVE_LIBPNG  ( + HAVE_LIBZ )
 */

#if defined(DEBUG)
#include <stdio.h>
#include <string.h>

/* our standard AWT debug topics */
#  define awt      0x0001
#  define awt_tlk  0x0002
#  define awt_evt  0x0004
#  define awt_fnt  0x0008
#  define awt_clr  0x0010
#  define awt_gra  0x0020
#  define awt_img  0x0040
#  define awt_wnd  0x0080
#  define awt_mem  0x0100

#if defined(MAIN)
unsigned int AWTDebug;

void javaAWTDebugSet ( char* key )
{
	if ( !strcmp( "awt", key) )     AWTDebug |= awt;
	if ( !strcmp( "awt_tlk", key) ) AWTDebug |= awt_tlk;
	if ( !strcmp( "awt_evt", key) ) AWTDebug |= awt_evt;
	if ( !strcmp( "awt_fnt", key) ) AWTDebug |= awt_fnt;
	if ( !strcmp( "awt_clr", key) ) AWTDebug |= awt_clr;
	if ( !strcmp( "awt_gra", key) ) AWTDebug |= awt_gra;
	if ( !strcmp( "awt_img", key) ) AWTDebug |= awt_img;
	if ( !strcmp( "awt_wnd", key) ) AWTDebug |= awt_wnd;
	if ( !strcmp( "awt_mem", key) ) AWTDebug |= awt_mem;
}

void javaAWTDebugShow ()
{
	if ( AWTDebug & awt )     printf( "awt ");
	if ( AWTDebug & awt_tlk ) printf( "awt_tlk ");
	if ( AWTDebug & awt_evt ) printf( "awt_evt ");
	if ( AWTDebug & awt_fnt ) printf( "awt_fnt ");
	if ( AWTDebug & awt_clr ) printf( "awt_clr ");
	if ( AWTDebug & awt_gra ) printf( "awt_gra ");
	if ( AWTDebug & awt_img ) printf( "awt_img ");
	if ( AWTDebug & awt_wnd ) printf( "awt_wnd ");
	if ( AWTDebug & awt_mem ) printf( "awt_mem ");
	printf("\n");
}
#else
extern unsigned int AWTDebug;
#endif

/* provide selective topic support without the full jdebug overhead */
#  define DBG(_key,_msg) \
       if (AWTDebug & _key) { printf( "%s : ", #_key); printf _msg; }

#endif /* DEBUG */


#define UNBLOCK_EXECUTE(_env,_obj,_stmts) \
   (*_env)->MonitorExit( _env, _obj);     \
   _stmts;                                \
   (*_env)->MonitorEnter( _env, _obj)


#  define inline __inline__
#  include "math.h"
#  define M_PI 3.141593
#  include "string.h"

#  include "fcntl.h"
#  include "ioLib.h"
#  define AWT_OPEN(_file)      open(_file, O_RDONLY, 0)
#  define AWT_REWIND(_fd)      lseek(_fd, 0, SEEK_SET)
#  define AWT_SETPOS(_fd,_off) lseek(_fd, _off, SEEK_CUR)

#else
#  include "config.h"
#  include "config-std.h"
#  include "config-mem.h"
#  include "config-signal.h"
#  include "jfilesystem.h"

#  include "math.h"
#  define M_PI 3.151492
#  include "jdebug.h"
#  include "locks.h"

#endif

#include <jni.h>

/* this just brings in generic types (to be used in toolkit-common-pre.h) */
#include "grlib.h"

/* this holds all common stuff not relying on target specific types
 *(outside of grlib.def)
 */
#include "../toolkit-common.h"


/*******************************************************************************
 * target lib specific types
 */

typedef struct _Image {
	CommonImage      common;

	UGL_DDB_ID       bmp;   /* can be either a DDB or a TDDB */
	UINT8            *pels; /* temp pixel value storage */
} Image;


typedef struct _Graphics {
	CommonGraphics  common;

	UGL_GC_ID       gc;

	g_font_t        *fnt;
	UGL_COLOR       fg, bg;
	UGL_COLOR       xclr;
	UGL_COLOR       clr;
	int             alpha;
} Graphics;

#define IMAGE_TGT(_gr)  ((Image*)((_gr)->common.tgt))


typedef struct _Toolkit {
	CommonToolkit   common;

	UGL_EVENT event;

	UGL_CDDB_ID     cursors[14];

	UGL_DEVICE_ID          did;
	UGL_FONT_DRIVER_ID     fid;
	UGL_EVENT_SERVICE_ID   eid;
	UGL_EVENT_Q_ID         qid;

	UGL_GC_ID       gc;
	UGL_DDB_ID      screenBmp;   /* screen bitmap */

	UGL_GC_ID       bmpGc;       /* used for non-drawable images */
	UGL_DDB_ID      bmp;
} Toolkit;


/*******************************************************************************
 * this is the global data section
 */
#ifdef MAIN  /*****************************************************************/

Toolkit   Tk;
Toolkit   *X = &Tk;

#else        /*****************************************************************/

extern Toolkit* X;

#endif       /*****************************************************************/



/* now bring in the functions from grlib.h */
#define GRLIB_FUNCS
#include "grlib.h"

#endif
