/*
 * wnd.c - target lib specific window management (mainly cursor handling)
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.h"

/*****************************************************************************
 * use generic versions of general toplevel-related native methods
 */

#define _wndSetTitle             EXPORT(wndSetTitle)
#define _wndSetResizable         EXPORT(wndSetResizable)
#define _wndCreateFrame          EXPORT(wndCreateFrame)
#define _wndCreateWindow         EXPORT(wndCreateWindow)
#define _wndCreateDialog         EXPORT(wndCreateDialog)
#define _wndRequestFocus         EXPORT(wndRequestFocus)
#define _wndSetBounds            EXPORT(wndSetBounds)
#define _wndRepaint              EXPORT(wndRepaint)
#define _wndSetIcon              EXPORT(wndSetIcon)
#define _wndSetVisible           EXPORT(wndSetVisible)
#define _wndToBack               EXPORT(wndToBack)
#define _wndToFront              EXPORT(wndToFront)
#define _wndDestroyWindow        EXPORT(wndDestroyWindow)
#define _wndSetFrameInsets       EXPORT(wndSetFrameInsets)
#define _wndSetDialogInsets      EXPORT(wndSetDialogInsets)


/*****************************************************************************
 * include generic functionality
 */
#include "../wnd-common.c"


/*****************************************************************************
 * platform specific functions
 */

#define TRANSP_IDX  UGL_CURSOR_COLOR_TRANSPARENT
#include "../cursors.h"

UGL_CDDB_ID
getCursor ( Toolkit* X, int jCursor )
{
	UGL_CDDB_ID cur;
	UGL_CDIB  cdib;
	int       i = jCursor;

	if ( i > 13 ) i = 0;
	else if ( i < 0 ) i = 0;

	if ( (cur = X->cursors[i]) == 0 ) {
		cdib.width  = cursorSize;
		cdib.height = cursorSize;
		cdib.stride = cursorSize;
		cdib.hotSpot.x = cursorHotSpotX[i];
		cdib.hotSpot.y = cursorHotSpotY[i];
		cdib.clutSize = sizeof(cursorColors) / sizeof(cursorColors[0]);
		cdib.pClut = (UGL_ARGB*)cursorColors;
		cdib.pImage = cursorShapes[i];

		X->cursors[i] = uglCursorBitmapCreate( X->did, &cdib);
		DBG( awt_wnd, ("getCursor %d -> %p\n", i, X->cursors[i]));

		cur = X->cursors[i];
	}

	return cur;
}

void
initCursors ( Toolkit* X )
{
	cursorColors[1] = UGL_MAKE_ARGB( 0xff, 0xff, 0xff, 0xff);
	cursorColors[2] = UGL_MAKE_ARGB( 0xff, 0, 0, 0);

	getCursor( X, 0);
}


void
EXPORT(wndSetCursor) ( JNIEnv* env, jclass clazz, void* wnd, jint jCursor )
{
	UGL_CDDB_ID curs;	
  
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	curs = getCursor( X, jCursor);

	DBG( awt_wnd, ("setCursor: %p  %d %p\n", wnd, (int)jCursor, curs));

	uglCursorOff( X->did);
	uglCursorImageSet( X->did, curs);
	uglCursorOn( X->did);
}


void
EXPORT(wndShowCursor) ( JNIEnv* env, jclass clazz, jboolean showIt )
{
	if ( !X->cursors[0] ) {
		/*
		 * just a safety line, should already intialized by now (via Toolkit->Defaults->Color)
		 */
		initCursors( X);
	}

	DBG( awt_wnd, ("showCursor: %d\n", showIt));

	if ( showIt ) {
		uglCursorOn( X->did);
	}
	else {
		uglCursorOff( X->did);
	}
}
