/**
 * nplugin.c - Netscape plugin interface
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */                                                                             
             
#include "toolkit.h"
#include "./npapi.h"
#include "./npupp.h"                                                            
#include <dlfcn.h>
#include "jfilesystem.h" 

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

#define ERR_CHECK(res,msg) \
if ( res ) { DBG( awt_plg, (msg " error: %d\n", res)); }  

extern void* _getWindow( NPWindow* wnd);

typedef char* (*mime_func)(void);

typedef struct _PlugPlay {
  NPP npp;
  NPMIMEType mime;
  NPStream *cur_stream;
  void *dl_handle;
  NPPluginFuncs nppFuncs;
  NPError(*np_up)(NPNetscapeFuncs *npnfuncs, NPPluginFuncs *nppfucs);
  NPError(*np_down)(void);
  char *(*np_mime)(void);
  JNIEnv *genv;
} PlugPlay;


static jclass playercls;
static jmethodID requestdata;
static jmethodID geturl;
static jmethodID invalidate;
static jmethodID repaint;
static jmethodID statusupdate;
static jmethodID closestream;
static jobject playerobj;

/*******************************************
 * NPN callback funcs ( called from plugin )
 *******************************************/
NPError
NPN_GetValue( NPP instance, NPNVariable variable, void *value) {
  PlugPlay *pp;
  
  DBG( awt_plg, ("NPN_GetValue: var: %d value: %p\n", variable, value)); 
 
  pp = (PlugPlay*)instance->ndata;
 
#ifdef NEVER
  switch (variable)
    {
    case NPNVxDisplay:
      *(void**)value = XtDisplay( pp->canvas);
      return NPERR_NO_ERROR;
    case NPNVxtAppContext:
      *(void**)value = XtDisplayToApplicationContext( XtDisplay( pp->canvas));
      return NPERR_NO_ERROR;
    case NPNVjavascriptEnabledBool:
      (bool)(*value) = JNI_FALSE;
      return NPERR_NO_ERROR;
    case NPNVasdEnabledBool:
      (bool)(*value) = JNI_FALSE;
      return NPERR_NO_ERROR;
    case NPNVOfflineBool:
      (bool)(*value) = JNI_FALSE;
      return NPERR_NO_ERROR;
    default:
      return NPERR_INVALID_PARAM;
    }
#endif
  
  return NPERR_NO_ERROR;
}
 
void
NPN_Version(int* plugin_major, int* plugin_minor,
            int* netscape_major, int* netscape_minor){
  *netscape_major = NP_VERSION_MAJOR;
  *netscape_minor = NP_VERSION_MINOR;
  DBG( awt_plg, ("NPN_Version\n"));
}
 
NPError
NPN_GetURLNotify(NPP instance, const char* url,
                 const char* target, void* notifyData){
  DBG( awt_plg, ("NPN_URLGetNotify\n"));
  return NPERR_NO_ERROR;
}                                                                             


NPError
NPN_GetURL(NPP instance, const char* url, const char* target){
  PlugPlay *pp = (PlugPlay*)instance->ndata;
  DBG( awt_plg, ("NPN_GetURL: %s %s\n", url, target));
  (*pp->genv)->CallVoidMethod( pp->genv, playerobj, geturl,  
			       (*pp->genv)->NewStringUTF( pp->genv,url));
  return NPERR_NO_ERROR;
}
 
NPError
NPN_PostURLNotify(NPP instance, const char* url,
                  const char* target, uint32 len,
                  const char* buf, NPBool file,
                  void* notifyData){
  DBG( awt_plg, ("NPN_PostURLNotify\n"));
  return NPERR_NO_ERROR;
}
 
 
NPError
NPN_PostURL(NPP instance, const char* url,
            const char* target, uint32 len,
            const char* buf, NPBool file){
  DBG( awt_plg, ("NPN_PostURL\n"));
  return NPERR_NO_ERROR;
}
 
 
NPError
NPN_RequestRead(NPStream* stream, NPByteRange* rangeList){
  NPByteRange *range;
  PlugPlay *pp = (PlugPlay*)stream->ndata;

  for ( range = rangeList; range; range = range->next) {
    DBG( awt_plg, ("NPN_RequestRead offset: %d length: %d\n", range->offset, range->length));
    (*pp->genv)->CallVoidMethod( pp->genv, playerobj, requestdata, range->offset, range->length);
  }
  
  return NPERR_NO_ERROR;
}
                                                                              
NPError
NPN_NewStream(NPP instance, NPMIMEType type,
              const char* target, NPStream** stream){
  DBG( awt_plg, ("NPN_NewStream\n"));
  return NPERR_NO_ERROR;
}
 
int32
NPN_Write(NPP instance, NPStream* stream, int32 len, void* buffer){
  DBG( awt_plg, ("NPN_write\n"));
  return 0;
}
 
NPError
NPN_DestroyStream(NPP instance, NPStream* stream, NPReason reason){
  PlugPlay *pp = (PlugPlay*)instance->ndata;
  DBG( awt_plg, ("NPN_DestroyStream\n"));
  (*pp->genv)->CallVoidMethod( pp->genv, playerobj, closestream);  
  return NPERR_NO_ERROR;
}
 
void
NPN_Status(NPP instance, const char* message){
  PlugPlay *pp = (PlugPlay*)instance->ndata;
  DBG( awt_plg, ("NPN_Status: %s\n", message));
  (*pp->genv)->CallVoidMethod( pp->genv, playerobj, statusupdate,  
			       (*pp->genv)->NewStringUTF( pp->genv,message));
}
 
const char*
NPN_UserAgent(NPP instance){
  DBG( awt_plg, ("NPN_UserAgent\n"));
  return "UserAgent";
}
 
void*
NPN_MemAlloc(uint32 size){
  DBG( awt_plg, ("NPN_MemAlloc: %d\n", size));
  return AWT_MALLOC( size);
}

void
NPN_MemFree(void* ptr){
  DBG( awt_plg, ("NPN_MemFree: %p\n", ptr));
  AWT_FREE( ptr);
}
 
uint32
NPN_MemFlush(uint32 size){
  DBG( awt_plg, ("NPN_MemFlush: %s\n", size));
  return 0;
}
 
void
NPN_ReloadPlugins(NPBool reloadPages){
  DBG( awt_plg, ("NPN_ReloadPlugins\n"));
}
 
JRIEnv*
NPN_GetJavaEnv(void){
  DBG( awt_plg, ("NPN_GetJavaEnv\n"));
  return NULL;
}
 
jref
NPN_GetJavaPeer(NPP instance){
  DBG( awt_plg, ("NPN_GetJavaPeer\n"));
  return NULL;
}
 
NPError
NPN_SetValue(NPP instance, NPPVariable variable, void *value){
  DBG( awt_plg, ("NPN_SetValue %d %p\n", variable, value));
  return NPERR_NO_ERROR;
}
                                                                              
void
NPN_ForceRedraw(NPP instance){
  PlugPlay *pp = (PlugPlay*)instance->ndata;
  DBG( awt_plg, ("NPN_ForceRedraw\n"));
  (*pp->genv)->CallVoidMethod( pp->genv, playerobj, repaint);
}
 
void
NPN_InvalidateRect(NPP instance, NPRect *invalidRect){
  PlugPlay *pp = (PlugPlay*)instance->ndata;
  DBG( awt_plg, ("NPN_InvalidateRect\n"));
  (*pp->genv)->CallVoidMethod( pp->genv, playerobj, invalidate);
}
 
void
NPN_InvalidateRegion(NPP instance, NPRegion invalidRegion){
  PlugPlay *pp = (PlugPlay*)instance->ndata;
  DBG( awt_plg, ("NPN_InvalidateRegion\n"));
  (*pp->genv)->CallVoidMethod( pp->genv, playerobj, invalidate);
}
 
                                                                              
/********************
 * helper funcs
 *********************/
static void
get_jmethods( JNIEnv *env, jobject player, PlugPlay *pp) {
  pp->genv = env;

  playercls = (*env)->GetObjectClass( env, player);
  playerobj = (*env)->NewGlobalRef( env, player);
  requestdata = (*env)->GetMethodID( env, playercls, "requestData", "(II)V");
  statusupdate = (*env)->GetMethodID( env, playercls, "statusUpdate", "(Ljava/lang/String;)V");
  geturl = (*env)->GetMethodID( env, playercls, "getURL", "(Ljava/lang/String;)V");
  invalidate = (*env)->GetMethodID( env, playercls, "invalidate", "()V");
  repaint = (*env)->GetMethodID( env, playercls, "repaint", "()V");
  closestream = (*env)->GetMethodID( env, playercls, "closeStream", "()V");
}
 
static int 
load_lib( PlugPlay* pp, char* pathname) {
  void* handle;
  char* err;
  char* funcs[] = {"NP_Initialize","NP_Shutdown","NP_GetMIMEDescription" };

  handle = dlopen( pathname, RTLD_LAZY);
  pp->dl_handle = handle;
 
  if ( handle == 0 ) {
    DBG( awt_plg, ("failed to open plugin %s: %s\n", pathname, dlerror()));
    return 1;
  }
 
  pp->np_up = (NPError(*)(NPNetscapeFuncs*, NPPluginFuncs*))dlsym( handle, funcs[0]); 
  if ( (err = dlerror()) != NULL ) {
    DBG( awt_plg, ("cannot find %s: %s\n", funcs[0], err));
    return 2;
  }
 
  pp->np_down = (NPError(*)(void))dlsym( handle, funcs[1]);
  if ( (err = dlerror()) != NULL ) {
    DBG( awt_plg, ("cannot find %s: %s\n", funcs[1], err));
    return 3;
  }
 
  pp->np_mime = (char*(*)(void))dlsym( handle, funcs[2]);      
  if ( (err = dlerror()) != NULL ) {
    DBG ( awt_plg, ("cannot find %s: %s\n", funcs[2], err));
    return 4;
  }
 
  return 0;
}                                                                           

static int 
init_callbacks( PlugPlay* pp) {
  NPNetscapeFuncs nsfuncs;

  //init NPN funcs ( called by plugin )
  nsfuncs.version = (NP_VERSION_MAJOR << 8) + NP_VERSION_MINOR;
  nsfuncs.size = sizeof(NPNetscapeFuncs);
 
  nsfuncs.posturl       = NPN_PostURL;
  nsfuncs.geturl        = NPN_GetURL;
  nsfuncs.requestread   = NPN_RequestRead;
  nsfuncs.newstream     = NPN_NewStream;
  nsfuncs.write         = NPN_Write;
  nsfuncs.destroystream = NPN_DestroyStream;
  nsfuncs.status        = NPN_Status;
  nsfuncs.uagent        = NPN_UserAgent;
 
  nsfuncs.memalloc  = NPN_MemAlloc;
  nsfuncs.memfree   = NPN_MemFree;
  nsfuncs.memflush  = NPN_MemFlush;
 
  nsfuncs.reloadplugins = NPN_ReloadPlugins;
  nsfuncs.getJavaEnv    = NPN_GetJavaEnv;
  nsfuncs.getJavaPeer   = NPN_GetJavaPeer;
  nsfuncs.geturlnotify  = NPN_GetURLNotify;
  nsfuncs.posturlnotify = NPN_PostURLNotify;
  nsfuncs.getvalue      = NPN_GetValue;
 
  nsfuncs.setvalue         = NPN_SetValue;
  nsfuncs.forceredraw      = NPN_ForceRedraw;
  nsfuncs.invalidaterect   = NPN_InvalidateRect;
  nsfuncs.invalidateregion = NPN_InvalidateRegion;

  pp->nppFuncs.size = sizeof( NPPluginFuncs);

  //load function tables
  return (*pp->np_up)( &nsfuncs, &(pp->nppFuncs));
}      

static int
get_mime( char* pathname, char* mime) {
   PlugPlay pp;

   if ( load_lib( &pp, pathname )) 
     {
	if ( pp.dl_handle )
	  dlclose( pp.dl_handle);
	return 0;
     }

   strcpy( mime, (*pp.np_mime)() );
   dlclose( pp.dl_handle);

   return 1;
}                                                                           

static int 
init( JNIEnv *env, jobject player, PlugPlay* pp, char *clib) {
  NPSavedData sd;
  char *argn[] = {};
  char *argv[] = {};
  int ret = 0;

  fgl_PServerInit();

  pp->npp = AWT_MALLOC( sizeof( NPP_t));
  pp->npp->ndata = pp;

  pp->mime = NULL;
  pp->cur_stream = NULL;

  get_jmethods( env, player, pp);

  ret |= load_lib( pp, clib);
  ret |= init_callbacks( pp);
   
  sd.buf = (void*)Kaffe_FileSystemCallInterface._write;

  //create plugin instance
  ret |= (*pp->nppFuncs.newp)( pp->mime, pp->npp, NP_FULL, 1, argn, argv, &sd);
	
  return ret;
}      

static int
write_stream( PlugPlay *pp, char* buffer, int offset, int length) {
  int ba, bs, bw, off, len;
 
  off = offset;
  len = length;

  while( len > 0 ) {
    ba = (*pp->nppFuncs.writeready)( pp->npp, pp->cur_stream );
    if ( ba < len ) {
      bs = ba;
    }
    else {
      bs = len;
    }
    bw = (*pp->nppFuncs.write)( pp->npp, pp->cur_stream, off, bs, buffer);

    if ( bw < 1 ) {
      return bw;
    }

    //DBG( awt_plg, ("written stream: %d max: %d off: %d len: %d total: %d\n", bw, ba, off, bs, length));

    len -= bw;
    off += bw;
  }
 

  return (off - offset);
}
 
                                          
/******************************************************
 * exported funcs
 ******************************************************/
 
JNIEXPORT PlugPlay* JNICALL
Java_kaffe_awt_player_PluginPlayer_open0( JNIEnv *env, jclass clazz, jobject player, jstring lib)
{
  jbyte *strlib;
  PlugPlay *pp;
  int ret;
   
  strlib = (jbyte*)(*env)->GetStringUTFChars( env, lib, 0);

  pp = AWT_MALLOC (sizeof( PlugPlay));
  ret = init( env, player, pp, strlib);

  DBG( awt_plg, ("open0 %s (%d)\n", strlib, ret));

  (*env)->ReleaseStringUTFChars( env, lib, strlib);
   
  if ( ret ) {
    AWT_FREE( pp);
    return NULL;
  }
  else {
    return pp;
  }
} 

JNIEXPORT jboolean JNICALL
Java_kaffe_awt_player_PluginPlayer_start0( JNIEnv *env, jclass clazz, PlugPlay *pp, jstring source, jboolean asFile)
{
  jbyte *str;
  NPStream *ps;
  uint16 stype;
 
  stype = asFile ? NP_ASFILEONLY : NP_SEEK;
 
  ps = AWT_MALLOC( sizeof( NPStream));

  str = (jbyte*)(*env)->GetStringUTFChars( env, source, 0);
  ps->url = strdup(str);
  (*env)->ReleaseStringUTFChars( env, source, str);

  ps->end = 0;
  ps->lastmodified = 0;
  ps->notifyData = NULL;
  ps->ndata = pp;
 
  pp->cur_stream = ps;

  ERR_CHECK((*pp->nppFuncs.newstream)( pp->npp, pp->mime, pp->cur_stream, !asFile, &stype), "NPP_NewStream");   
  
  //stype should be adjusted by plugin to match needs
  if ( stype == NP_ASFILEONLY ) {
    (*pp->nppFuncs.asfile)( pp->npp, pp->cur_stream, ps->url);
    return JNI_FALSE;
  }
  else if ( stype == NP_SEEK ) {
    //plugin will request data
    return JNI_FALSE;
  }
 
  //NP_NORMAL (data has to be streamed)
  return JNI_TRUE;
}

JNIEXPORT jint JNICALL
Java_kaffe_awt_player_PluginPlayer_write0( JNIEnv *env, jclass clazz, PlugPlay *pp, jarray buffer, jint offset, jint length )
{
  char *buf;
  jboolean isCopy;
  int written = 0;
  
  buf = (*env)->GetByteArrayElements( env, buffer, &isCopy);
  written = write_stream( pp, buf, offset, length);
  (*env)->ReleaseByteArrayElements( env, buffer, buf, JNI_ABORT);

  return (written);
}

JNIEXPORT void JNICALL
Java_kaffe_awt_player_PluginPlayer_stop0( JNIEnv *env, jclass clazz, PlugPlay *pp)
{
  if ( pp->cur_stream ) {
    (*pp->nppFuncs.destroystream)( pp->npp, pp->cur_stream, NPRES_USER_BREAK); 
    AWT_FREE( pp->cur_stream);
    pp->cur_stream = NULL;
  }
}

JNIEXPORT void JNICALL
Java_kaffe_awt_player_PluginPlayer_close0( JNIEnv *env, jclass clazz, PlugPlay *pp)
{
  (*pp->nppFuncs.destroy)(pp->npp, NULL);
  (*pp->np_down)();

  dlclose( pp->dl_handle);
  AWT_FREE( pp->npp);
  AWT_FREE( pp);
}


JNIEXPORT void JNICALL
Java_kaffe_awt_player_PluginPlayer_setBounds0( JNIEnv *env, jclass clazz, PlugPlay *pp, jint x, jint y, jint width, jint height)
{
  NPWindow win;
  NPRect clip;
 
  win.x = x;
  win.y = y;
  win.width = width;
  win.height = height;
 
  clip.top = x;
  clip.left = y;
  clip.right = width;
  clip.bottom = height;
 
  win.clipRect = clip;
 
  win.window = _getWindow( &win);

  ERR_CHECK((*pp->nppFuncs.setwindow)( pp->npp, &win), "NPP_SetWindow");       
}

JNIEXPORT jint JNICALL
Java_kaffe_awt_player_PluginPlayer_handleEvent0( JNIEnv *env, jclass clazz, PlugPlay *pp, jint event, jint p1, jint p2)
{
  NPEvent evt;
 
  evt.p1 = p1;
  evt.p2 = p2;
  evt.event = event;

  return (*pp->nppFuncs.event)( pp->npp, &evt);       
}

JNIEXPORT jstring JNICALL
Java_kaffe_awt_player_PluginPlayer_checkPlugin0( JNIEnv *env, jclass clazz, jstring lib)
{
   jobject mimestr;
   jbyte *libpath;
   char  mime[255];
   
   libpath = (jbyte*)(*env)->GetStringUTFChars( env, lib, 0);
   
   if ( get_mime( libpath, mime) ) {
      mimestr =  (*env)->NewStringUTF( env, mime);	
   }
   else {
      mimestr = NULL;
   }
   
   (*env)->ReleaseStringUTFChars( env, lib, libpath);
   
   return mimestr;
}

