/**
 * clr.c - color management
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"

/********************************************************************************
 * VGA color support
*/


/*
 * Since we have incredible 16 colors, we have 3 wished left (*) once we
 * have allocated the standard Java colors
 */
#define TO4(a) (a ? ((a+1)/4)-1 : 0)
#define TO8(a) (a ? ((a+1)*4)-1 : 0)

UCHAR	palette [] = {
	TO4(   0), TO4(   0), TO4(   0),  // black
	TO4( 192), TO4(   0), TO4(   0),  // a darker red (*)
	TO4( 224), TO4( 224), TO4( 224),  // a lighter lightGray (*), was PEG dark green
	TO4( 255), TO4( 200), TO4(   0),  // orange					, was PEG brown
	TO4(  64), TO4(  64), TO4(  64),  // darkGray				, was PEG dark blue
	TO4( 255), TO4( 175), TO4( 175),  // pink					, was PEG magenta
	TO4( 160), TO4( 160), TO4( 160),  // a lighter gray (*) , was PEG dark cyan
	TO4( 192), TO4( 192), TO4( 192),  // lightGray
	TO4( 128), TO4( 128), TO4( 128),  // gray              , = PEG darkgray
	TO4( 255), TO4(   0), TO4(   0),  // red               , = PEG light red
	TO4(   0), TO4( 255), TO4(   0),  // green             , = PEG light green
	TO4( 255), TO4( 255), TO4(   0),  // yellow
	TO4(   0), TO4(   0), TO4( 255),  // blue              , = PEG light blue
	TO4( 255), TO4(   0), TO4( 255),  // magenta           , = PEG light magenta
	TO4(   0), TO4( 255), TO4( 255),  // cyan              , = PEG light cyan
	TO4( 255), TO4( 255), TO4( 255)   // white
};


Rgb2Vga*
initRgb2Vga ( Toolkit* X )
{
	int i, k, r, g, b;
	Rgb2Vga *map = (Rgb2Vga*) AWT_MALLOC( sizeof( Rgb2Vga));
	
	// PEG needs them packed (as a UCHAR array)
	X->screen->SetPalette( 0, 16, palette);
	
	DBG( awt_clr, ("initRgb2Vga\n"));
	
	// we need jint values, so unpack the palette
	for ( i=0, k=0; i<16; i++ ) {
		r = TO8( palette[k]); k++;
		g = TO8( palette[k]); k++;
		b = TO8( palette[k]); k++;
		map->clrs[i] = JRGB( r, g, b);
	}
	
	return map;
}


#define MODPIX(_c,_d,_b,_m)   ((_m>0) ? _b : ((_m<0) ? _d : _c))

jint
getVGAPixelValue ( Toolkit* X, jint rgb, int mod )
{
	jint pix = 0;
	
	unsigned char r  = JRED( rgb);
	unsigned char g  = JGREEN( rgb);
	unsigned char b  = JBLUE( rgb);
	
	unsigned char rr = (r >> 2) << 2;
	unsigned char gg = (g >> 2) << 2;
	unsigned char bb = (b >> 2) << 2;
	
	if ( rr ) {
		if ( gg ) {
			if ( bb ) {  /* red, green, blue */
				if ( gg == bb ) {
					if ( rr == gg ) { /* gray or white */
						if ( r > 240 )            pix = MODPIX( CLR_WHITE, CLR_GRAY1, CLR_WHITE, mod);
						else if ( r > 208 )       pix = MODPIX( CLR_GRAY1, CLR_LIGHTGRAY, CLR_WHITE, mod);
						else if ( r > 176 )       pix = MODPIX( CLR_LIGHTGRAY, CLR_GRAY2, CLR_GRAY1, mod);
						else if ( r > 144 )       pix = MODPIX( CLR_GRAY2, CLR_DARKGRAY, CLR_LIGHTGRAY, mod);
						else if ( r > 96 )        pix = MODPIX( CLR_GRAY, CLR_DARKGRAY, CLR_GRAY2, mod);
						else if ( r > 32 )        pix = MODPIX( CLR_DARKGRAY, CLR_BLACK, CLR_GRAY2, mod);
						else                      pix = MODPIX( CLR_BLACK, CLR_BLUE, CLR_DARKGRAY, mod);
					}
					else if ( r < g )           pix = MODPIX( CLR_CYAN, CLR_BLUE, CLR_WHITE, mod);
					else                        pix = MODPIX( CLR_RED, CLR_MAGENTA, CLR_ORANGE, mod);
				}
				else if ( rr == bb ) {
					if ( g > r )                pix = MODPIX( CLR_GREEN, CLR_BLACK, CLR_WHITE, mod);
					else                        pix = MODPIX( CLR_PINK, CLR_MAGENTA, CLR_WHITE, mod);
				}
				else {
					if ( (r > g) && (r > b) )   pix = MODPIX( CLR_RED, CLR_MAGENTA, CLR_ORANGE, mod);
					else if ( (g >> b) && (g >> r) ) pix = MODPIX( CLR_GREEN, CLR_BLACK, CLR_WHITE, mod);
					else                        pix = MODPIX( CLR_BLUE, CLR_BLACK, CLR_CYAN, mod);
				}
			}
			else        /* red, green  */   pix = MODPIX( CLR_YELLOW, CLR_ORANGE, CLR_WHITE, mod);
		}
		else {
			if ( bb )    /* red, blue */    pix = MODPIX( CLR_MAGENTA, CLR_BLACK, CLR_PINK, mod);
			else {       /* red */
				if ( r > 200 )                pix = MODPIX( CLR_RED, CLR_MAGENTA, CLR_ORANGE, mod);
				else                          pix = MODPIX( CLR_RED1, CLR_BLACK, CLR_RED, mod);
			}
		}
	}
	else {
		if ( gg ) {
			if ( bb )    /* green, blue */  pix = MODPIX( CLR_CYAN, CLR_BLUE, CLR_WHITE, mod);
			else        /* green */         pix = MODPIX( CLR_GREEN, CLR_BLACK, CLR_WHITE, mod);
		}
		else {
			if ( bb )    /* blue */         pix = MODPIX( CLR_BLUE, CLR_BLACK, CLR_CYAN, mod);
			else        /* black */         pix = MODPIX( CLR_BLACK, CLR_BLUE, CLR_DARKGRAY, mod);
		}
	}
	
	return pix;
}


void
initColorMapping ( JNIEnv* env, Toolkit* X )
{
	unsigned int nc = PEG_NUM_COLORS;
	
	switch ( nc ) {
	case 16:
		X->colorMode = CM_VGA;
		X->vclr = initRgb2Vga( X);
		break;
		
	default:
		// nothing else supported yet, bail out
		DBG( awt_clr, ("colormode not supported\n"));
		exit( 0);
	}
}

extern "C" {

/*
 * we need to implement brighter / darker in the native layer because the usual
 * arithmetic Rgb brightening/darkening isn't really useful if it comes to
 * PseudoColor visuals (e.g. a primitive 16 color VGA wouldn't produce any
 * usable results). Even 256 colormaps suffer from 
 */
jlong _cdecl
Java_java_awt_Toolkit_clrBright ( JNIEnv* env, jclass clazz, jint rgb )
{
	jint     modRgb, modPix;
	
	DBG( awt_clr, ("bright: %x\n", rgb));
	
	switch ( X->colorMode )  {
	case CM_VGA:
		modPix = getVGAPixelValue( X, rgb, 1);
		modRgb = X->vclr->clrs[modPix];
		break;
	}
	
	return (((jlong)modPix << 32) | modRgb);
}


jlong _cdecl
Java_java_awt_Toolkit_clrDark ( JNIEnv* env, jclass clazz, jint rgb )
{
	jint     modRgb, modPix;
	
	DBG( awt_clr, ("dark: %x\n", rgb));
	
	switch ( X->colorMode ) {
	case CM_VGA:
		modPix = getVGAPixelValue( X, rgb, -1);
		modRgb = X->vclr->clrs[modPix];
		break;
	}
	
	return (((jlong)modPix << 32) | modRgb);
}


/********************************************************************************
 * common exported functions
 */

jint _cdecl
Java_java_awt_Toolkit_clrGetPixelValue ( JNIEnv* env, jclass clazz, jint rgb )
{
	jint pix;
	
	/*
	* We do this deferred to avoid class init problems with Defaults<->Color
	*/
	if ( !X->colorMode ) {
		initColorMapping( env, X);
	}
	
	pix = pixelValue( X, rgb);
	DBG( awt_clr, ("getPixelValue: %x %d (%x)\n", rgb, pix, pix));
	
	return pix;
}

void _cdecl
Java_java_awt_Toolkit_clrSetSystemColors ( JNIEnv* env, jclass clazz, jarray sysClrs )
{
	/* there is no native window manager, hence no native defaults (leave it up to Java) */
	DBG( awt_clr, ("setSystemColors: %p\n", sysClrs));
}


jobject
Java_java_awt_Toolkit_clrGetColorModel ( JNIEnv* env, jclass clazz )
{
	jobject    cm = 0;
	jclass     cmClazz;
	jmethodID  cmCtorId;
	jarray     pal;

	if ( !X->colorMode ){
		initColorMapping( env, X);
	}

	switch ( X->colorMode ){
	case CM_VGA:
		pal      = env->NewIntArray( 16);
		env->SetIntArrayRegion( pal, 0, 15, X->vclr->clrs);

		cmClazz  = env->FindClass( "java.awt.IndexColorModel");
		cmCtorId = env->GetMethodID( cmClazz, "<init>", "(I[II)V");
		cm       = env->NewObject( cmClazz, cmCtorId, 4, pal, -1);
		break;
	default:
		break;
	}

	return cm;
}

}
