/**
 * evt.cpp PegWindow related event handling
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"

jclass     ComponentEvent;
jclass     MouseEvent;
jclass     FocusEvent;
jclass     WindowEvent;
jclass     KeyEvent;
jclass     PaintEvent;
jclass     WMEvent;

jmethodID  getComponentEvent;
jmethodID  getMouseEvent;
jmethodID  getFocusEvent;
jmethodID  getWindowEvent;
jmethodID  getKeyEvent;
jmethodID  getPaintEvent;
jmethodID  getWMEvent;

#define    COMPONENT_RESIZED    101

#define    WINDOW_CLOSING       201
#define    WINDOW_CLOSED        202
#define    WINDOW_ICONIFIED     203
#define    WINDOW_DEICONIFIED   204

#define    KEY_PRESSED          401
#define    KEY_RELEASED         402

#define    MOUSE_PRESSED        501
#define    MOUSE_RELEASED       502
#define    MOUSE_MOVED          503
#define    MOUSE_ENTERED        504
#define    MOUSE_EXITED         505

#define    PAINT                800
#define    UPDATE               801

#define    FOCUS_GAINED        1004
#define    FOCUS_LOST          1005


void
JPegDialog::Draw ()
{
	PegRect  update;
	int  paintIt = FALSE;

	// we need that in order to enable ClipRect (which works only inside of a draw block)
	pt->BeginDraw();

	// get intersection of clip rect and invalid area BEFORE super Draw() kicks in
	update.Set( mClip.wLeft, mClip.wTop, mClip.wRight, mClip.wBottom);
	paintIt = Screen()->ClipRect( update, mClip);

	PegDialog::Draw();

	if ( paintIt ) {
	  JniEnv->CallStaticObjectMethod( PaintEvent, getPaintEvent, srcIdx, UPDATE,
                                    (jint)(update.wLeft - mReal.wLeft),
                                    (jint)(update.wTop - mReal.wTop),
                                    (jint)update.Width(), (jint)update.Height());
	}

	pt->EndDraw();
}

void
JPegDialog::Resize ( PegRect newSize )
{
	PegDialog::Resize( newSize);

	tgtClip = mReal;
	JniEnv->CallStaticObjectMethod( ComponentEvent, getComponentEvent, srcIdx,
		COMPONENT_RESIZED, (jint)mReal.wLeft, (jint)mReal.wTop, (jint)mReal.Width(), (jint)mReal.Height());
}


void
JPegDecoWindow::Draw ()
{
	PegRect  update;
	int  paintIt = FALSE;

	// we need that in order to enable ClipRect (which works only inside of a draw block)
	pt->BeginDraw();

	// get intersection of clip rect and invalid area BEFORE super Draw() kicks in
	update.Set( mClip.wLeft, mClip.wTop, mClip.wRight, mClip.wBottom);
	paintIt = Screen()->ClipRect( update, mClip);

	PegDecoratedWindow::Draw();

	if ( paintIt ) {
	  JniEnv->CallStaticObjectMethod( PaintEvent, getPaintEvent, srcIdx, UPDATE,
                                    (jint)(update.wLeft - mReal.wLeft),
                                    (jint)(update.wTop - mReal.wTop),
                                    (jint)update.Width(), (jint)update.Height());
	}

	pt->EndDraw();
}

void
JPegDecoWindow::Resize ( PegRect newSize )
{
	PegDecoratedWindow::Resize( newSize);

	tgtClip = mReal;
	JniEnv->CallStaticObjectMethod( ComponentEvent, getComponentEvent, srcIdx,
		COMPONENT_RESIZED, (jint)mReal.wLeft, (jint)mReal.wTop, (jint)mReal.Width(), (jint)mReal.Height());
}

void
JPegWindow::Draw ()
{
	PegRect  update;
	int  paintIt = FALSE;

	// we need that in order to enable ClipRect (which works only inside of a draw block)
	pt->BeginDraw();

	// get intersection of clip rect and invalid area BEFORE super Draw() kicks in
	update.Set( mClip.wLeft, mClip.wTop, mClip.wRight, mClip.wBottom);
	paintIt = Screen()->ClipRect( update, mClip);

	PegWindow::Draw();

	if ( paintIt ) {
	  JniEnv->CallStaticObjectMethod( PaintEvent, getPaintEvent, srcIdx, UPDATE,
                                    (jint)(update.wLeft - mReal.wLeft),
                                    (jint)(update.wTop - mReal.wTop),
                                    (jint)update.Width(), (jint)update.Height());
	}

	pt->EndDraw();
}

void
JPegWindow::Resize ( PegRect newSize )
{
	PegWindow::Resize( newSize);
	tgtClip = mReal;
	JniEnv->CallStaticObjectMethod( ComponentEvent, getComponentEvent, srcIdx,
		COMPONENT_RESIZED, (jint)mReal.wLeft, (jint)mReal.wTop, (jint)mReal.Width(), (jint)mReal.Height());
}

// some Java key codes we can map
#define VK_LEFT        0x25
#define VK_RIGHT       0x27
#define VK_UP          0x26
#define VK_DOWN        0x28
#define VK_HOME        0x24
#define VK_END         0x23
#define VK_PAGE_UP     0x21
#define VK_PAGE_DOWN   0x22
#define VK_DELETE      0x7f
#define VK_INSERT      0x9b
#define VK_ENTER       '\n'
#define VK_BACK_SPACE  '\b'
#define VK_TAB         '\t'
#define VK_SHIFT       0x10
#define VK_CONTROL     0x11
#define VK_ALT         0x12
#define VK_CAPS_LOCK   0x14
#define VK_NUM_LOCK    0x90
#define VK_SCROLL_LOCK 0x91
#define VK_META        0x9D
#define VK_F1          0x70
#define	VK_F2          0x71
#define VK_F3          0x72
#define VK_F4          0x73
#define VK_F5          0x74
#define VK_F6          0x75
#define VK_F7          0x76
#define VK_F8          0x77
#define VK_F9          0x78
#define VK_F10         0x79
#define VK_F11         0x7A
#define VK_F12         0x7B

void
MappedWindow::emitKeyEvent ( const PegMessage& msg, int id )
{
	jint kchar = msg.iData;
	jint kcode;
	int  mods = msg.lData & 0xff;

	switch ( kchar ) {
	case PK_BACKSPACE:       kcode = VK_BACK_SPACE;   kchar = '\b';  break;
	case PK_TAB:             kcode = VK_TAB;          kchar = '\t';  break;
	case PK_CR:              kcode = VK_ENTER;        kchar = '\n';  break;
	case PK_DELETE:          kcode = VK_DELETE;       kchar = 0x7f;  break;
	case PK_INSERT:          kcode = VK_INSERT;       kchar = 0;     break;
	case PK_HOME:            kcode = VK_HOME;         kchar = 0;     break;
	case PK_END:             kcode = VK_END;          kchar = 0;     break;
	case PK_PGUP:            kcode = VK_PAGE_UP;      kchar = 0;     break;
	case PK_PGDN:            kcode = VK_PAGE_DOWN;    kchar = 0;     break;
	case PK_LNUP:            kcode = VK_UP;           kchar = 0;     break;
	case PK_LNDN:            kcode = VK_DOWN;         kchar = 0;     break;
	case PK_LEFT:            kcode = VK_LEFT;         kchar = 0;     break;
	case PK_RIGHT:           kcode = VK_RIGHT;        kchar = 0;     break;
  case PK_ALT:             kcode = VK_ALT;          kchar = 0;     break;
  case PK_CTRL:            kcode = VK_CONTROL;      kchar = 0;     break;
  case PK_SHIFT:           kcode = VK_SHIFT;        kchar = 0;     break;
  case PK_CAPS:            kcode = VK_CAPS_LOCK;    kchar = 0;     break;
  case PK_NUMLOCK:         kcode = VK_NUM_LOCK;     kchar = 0;     break;
  case PK_SCROLLLOCK:      kcode = VK_SCROLL_LOCK;  kchar = 0;     break;
  case PK_F1:              kcode = VK_F1;           kchar = 0;     break;
  case PK_F2:              kcode = VK_F2;           kchar = 0;     break;
  case PK_F3:              kcode = VK_F3;           kchar = 0;     break;
  case PK_F4:              kcode = VK_F4;           kchar = 0;     break;
  case PK_F5:              kcode = VK_F5;           kchar = 0;     break;
  case PK_F6:              kcode = VK_F6;           kchar = 0;     break;
  case PK_F7:              kcode = VK_F7;           kchar = 0;     break;
  case PK_F8:              kcode = VK_F8;           kchar = 0;     break;
  case PK_F9:              kcode = VK_F9;           kchar = 0;     break;
  case PK_F10:             kcode = VK_F10;          kchar = 0;     break;
  case PK_F11:             kcode = VK_F11;          kchar = 0;     break;
  case PK_F12:             kcode = VK_F12;          kchar = 0;     break;

	default:
		kcode = (kchar >= 'a' && kchar <= 'z' ) ? kchar ^ 0x20 : kchar;
	}

	JniEnv->CallStaticObjectMethod( KeyEvent, getKeyEvent, srcIdx, id, kcode, kchar, -1);
}


void
MappedWindow::emitJavaEvents ( const PegMessage& msg )
{
	switch ( msg.wType ) {
	case PM_SHOW:
		break;

	case PM_HIDE:
		break;

	case PM_DRAW:
		// we don't handle this here because somebody might call Draw() directly, without
		// taking a message detour (default action is just to call Draw())
		break;

	case PM_POINTER_ENTER:
		JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
			MOUSE_ENTERED, 0,
			(jint)(msg.Point.x - pt->mReal.wLeft),
			(jint)(msg.Point.y - pt->mReal.wTop));
		break;

	case PM_POINTER_EXIT:
		JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
			MOUSE_EXITED, 0,
			(jint)(msg.Point.x - pt->mReal.wLeft),
			(jint)(msg.Point.y - pt->mReal.wTop));
		break;
		
	case PM_POINTER_MOVE:
		// don't interfere with deco mouse handling
		int inDeco;
		switch ( pegType ) {
		case PEG_TYPE_DECO:
			inDeco = prt.deco->mClient.Contains( msg.Point) ? FALSE : TRUE; break;
		case PEG_TYPE_DIALOG:
			inDeco = prt.dialog->mClient.Contains( msg.Point) ? FALSE : TRUE; break;
		default:
			inDeco = 0;  // windows don"t have deco
		}

		if ( !inDeco ) {
			JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
				MOUSE_MOVED, 0,
				(jint)(msg.Point.x - pt->mReal.wLeft),
				(jint)(msg.Point.y - pt->mReal.wTop));
		}
		else {
			if ( X->pCursor != PPT_NORMAL ) {
				// restore pointer to system default
				X->pCursor = PPT_NORMAL;
				pt->Screen()->SetPointerType( X->pCursor);
			}
		}
		break;

	case PM_LBUTTONDOWN:
		JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
			MOUSE_PRESSED, 1,
			(jint)(msg.Point.x - pt->mReal.wLeft),
			(jint)(msg.Point.y - pt->mReal.wTop));
		break;
		
	case PM_LBUTTONUP:
		JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
			MOUSE_RELEASED, 1,
			(jint)(msg.Point.x - pt->mReal.wLeft),
			(jint)(msg.Point.y - pt->mReal.wTop));
			break;

	case PM_RBUTTONDOWN:
		JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
			MOUSE_PRESSED, 2,
			(jint)(msg.Point.x - pt->mReal.wLeft),
			(jint)(msg.Point.y - pt->mReal.wTop));
		break;
		
	case PM_RBUTTONUP:
		JniEnv->CallStaticObjectMethod( MouseEvent, getMouseEvent, srcIdx,
			MOUSE_RELEASED, 2,
			(jint)(msg.Point.x - pt->mReal.wLeft),
			(jint)(msg.Point.y - pt->mReal.wTop));
		break;

	case PM_CURRENT:
		if ( this != X->current ){
			JniEnv->CallStaticObjectMethod( FocusEvent, getFocusEvent, srcIdx,
				FOCUS_GAINED, JNI_FALSE);
			X->current = this;
		}
		break;

	case PM_NONCURRENT:
		if ( this == X->current )
			X->current = 0;
		JniEnv->CallStaticObjectMethod( FocusEvent, getFocusEvent, srcIdx,
			FOCUS_LOST, JNI_FALSE);
		break;

	case PM_KEY:
		emitKeyEvent( msg, KEY_PRESSED);
		break;

	case PM_KEY_RELEASE:
		emitKeyEvent( msg, KEY_RELEASED);
		break;

	case PM_MAXIMIZE:
		break;
		
	case PM_MINIMIZE:
		JniEnv->CallStaticObjectMethod( WindowEvent, getWindowEvent, srcIdx, WINDOW_ICONIFIED);
		break;
		
	case PM_RESTORE:
		JniEnv->CallStaticObjectMethod( WindowEvent, getWindowEvent, srcIdx, WINDOW_DEICONIFIED);
		break;
		
	case PM_CLOSE:
		JniEnv->CallStaticObjectMethod( WindowEvent, getWindowEvent, srcIdx, WINDOW_CLOSING);
		break;
	}
}

SIGNED
JPegWindow::Message ( const PegMessage &msg )
{
  SIGNED ret;

	// we can't determine to pass upwards on a return value of emitJavaEvents()
	// because some Java events rely on state which was changed by the super
	// Message(), and we have to explicitly specify the super class
	if ( msg.wType != PM_CLOSE )
		ret = PegWindow::Message( msg);

	emitJavaEvents( msg);

	return ret;
}

SIGNED
JPegDecoWindow::Message ( const PegMessage &msg )
{
  SIGNED ret;

	if ( msg.wType != PM_CLOSE )
		ret = PegDecoratedWindow::Message( msg);

  emitJavaEvents( msg);

	return ret;
}

SIGNED
JPegDialog::Message ( const PegMessage &msg )
{
  SIGNED ret;

	if ( msg.wType != PM_CLOSE )
		ret = PegDialog::Message( msg);

  emitJavaEvents( msg);

	return ret;
}

/*************************************************************************************
 * exported functions
 */

extern "C" {

jobject _cdecl
Java_java_awt_Toolkit_evtGetNextEvent ( JNIEnv* env, jclass clazz )
{
	/* we don't need that one here, just a dummy for the linker */
	return 0;
}

jobject _cdecl
Java_java_awt_Toolkit_evtPeekEvent ( JNIEnv* env, jclass clazz )
{
	/* we don't need that one here, just a dummy for the linker */
	return 0;
}

jobject _cdecl
Java_java_awt_Toolkit_evtPeekEventId ( JNIEnv* env, jclass clazz, jint id )
{
	/* we don't need that one here, just a dummy for the linker */
	return 0;
}

void _cdecl
Java_java_awt_Toolkit_evtWakeup ( JNIEnv* env, jclass clazz )
{
	/* we don't need that one here, just a dummy for the linker */
}

jarray _cdecl
Java_java_awt_Toolkit_evtInit ( JNIEnv* env, jclass clazz )
{
  jclass Component;
	jarray jsources;

  ComponentEvent = env->FindClass( "java.awt.ComponentEvt");
  MouseEvent     = env->FindClass( "java.awt.MouseEvt");
  FocusEvent     = env->FindClass( "java.awt.FocusEvt");
  WindowEvent    = env->FindClass( "java.awt.WindowEvt");
  KeyEvent       = env->FindClass( "java.awt.KeyEvt");
  PaintEvent     = env->FindClass( "java.awt.PaintEvt");
  WMEvent        = env->FindClass( "java.awt.WMEvent");

  getComponentEvent = env->GetStaticMethodID( ComponentEvent, "getEvent", 
												 "(IIIIII)Ljava/awt/ComponentEvt;");
  getMouseEvent     = env->GetStaticMethodID( MouseEvent, "getEvent",
												 "(IIIII)Ljava/awt/MouseEvt;");
  getFocusEvent     = env->GetStaticMethodID( FocusEvent, "getEvent",
												 "(IIZ)Ljava/awt/FocusEvt;");
  getWindowEvent    = env->GetStaticMethodID( WindowEvent, "getEvent",
												 "(II)Ljava/awt/WindowEvt;");
  getKeyEvent       = env->GetStaticMethodID( KeyEvent, "getEvent",
												 "(IIIII)Ljava/awt/KeyEvt;");
  getPaintEvent     = env->GetStaticMethodID( PaintEvent, "getEvent",
												 "(IIIIII)Ljava/awt/PaintEvt;");
  getWMEvent        = env->GetStaticMethodID( WMEvent, "getEvent",
												 "(II)Ljava/awt/WMEvent;");

  X->nWindows = 47;
  X->windows = (MappedWindow**) AWT_CALLOC( X->nWindows, sizeof(MappedWindow*));

  Component = env->FindClass( "java.awt.Component");
  jsources = env->NewObjectArray( X->nWindows, Component, NULL);

	return jsources;
}


jint _cdecl
Java_java_awt_Toolkit_evtRegisterSource ( JNIEnv* env, jclass clazz, MappedWindow* wnd )
{
  register int i;
  int      n;
	
  for ( i = (unsigned long)wnd, n=0; n < X->nWindows; i++, n++ ) {
		i %= X->nWindows;
		if ( ((long)(void*)X->windows[i]) <= 0 ) {
			X->windows[i] = wnd;
			wnd->srcIdx = i;
			return i;
		}
  }
	
  DBG( awt_evt, ("window table out of space: %d\n", X->nWindows));
  return -1;
}


jint _cdecl
Java_java_awt_Toolkit_evtUnregisterSource ( JNIEnv* env, jclass clazz, MappedWindow* wnd )
{
	int srcIdx = wnd->srcIdx;

	// This is really dangerous!! Destroy does (among other things) a delete of the
	// PegThing. If this isn't the allocated address (e.g. if we change the inheritance
	// order of JPegWindow parents), we get into trouble. But we can't do the delete by
	// ourselves, since this would cause a double delete (or we would have to do the
	// Destroy work)
	wnd->pt->Presentation()->Destroy( wnd->pt);

  if ( srcIdx >= 0 ) {
		X->windows[srcIdx] = (MappedWindow*)(void*) -1;
		return srcIdx;
	}
	else {
		return -1;
	}
}

/*
 * Helper for (IS_DISPATCH_EXCLUSIVE | NATIVE_DISPATCHER_LOOP ) windowing systems.
 * Not required here
 */
void _cdecl
Java_java_awt_Toolkit_evtSendWMEvent ( JNIEnv* env, jclass clazz, jobject wmEvt )
{
}

}
