/**
 * fnt.cpp - PEG font support
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#include "toolkit.hpp"

/*******************************************************************************
 * Font support
 */

extern "C" {

PegFont* _cdecl
Java_java_awt_Toolkit_fntInitFont ( JNIEnv* env, jclass clazz, jstring jSpec, jint style, jint size )
{
	// just two fonts (SysFont, MenuFont) to choose from
  return &SysFont;
}

void _cdecl
Java_java_awt_Toolkit_fntFreeFont ( JNIEnv* env, jclass clazz, PegFont* fnt )
{
	// Rather not, we only have SysFont and MenuFont (statics)
}


/*******************************************************************************
 * FontMetrics support
 */

PegFont* _cdecl
Java_java_awt_Toolkit_fntInitFontMetrics ( JNIEnv* env, jclass clazz, PegFont* fnt )
{
  return fnt; // font and fontmetrics are the same in PEG
}

void _cdecl
Java_java_awt_Toolkit_fntFreeFontMetrics ( JNIEnv* env, jclass clazz, PegFont* fs )
{
}

jint _cdecl
Java_java_awt_Toolkit_fntGetAscent ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return fs->uAscent;
}

jint _cdecl
Java_java_awt_Toolkit_fntGetDescent ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return fs->uDescent;
}

jint _cdecl
Java_java_awt_Toolkit_fntGetFixedWidth ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return 0;  // no fixed fonts yet
}

jint _cdecl
Java_java_awt_Toolkit_fntGetHeight ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return fs->uHeight;
}

jint _cdecl
Java_java_awt_Toolkit_fntGetLeading ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return 0;  // no leading (interline spacing) for PEG fonts
}

jint _cdecl
Java_java_awt_Toolkit_fntGetMaxAdvance ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return 0;  // not yet, should go into PegFont
}

jint _cdecl
Java_java_awt_Toolkit_fntGetMaxAscent ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return fs->uAscent;
}

jint _cdecl
Java_java_awt_Toolkit_fntGetMaxDescent ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return fs->uDescent;
}

jboolean _cdecl
Java_java_awt_Toolkit_fntIsWideFont ( JNIEnv* env, jclass clazz, PegFont* fs )
{
  return 0; // no wide (2-byte) fonts in Peg
}

jobject _cdecl
Java_java_awt_Toolkit_fntGetWidths ( JNIEnv* env, jclass clazz, PegFont* fs )
{
	char      buf[2];
	int       n = (fs->wLastChar > 256) ? 256 : fs->wLastChar;
  jarray    widths;
  jint      *jw;
  jboolean  isCopy;
  register  int i;

	buf[1] = 0;

  widths = env->NewIntArray( 256);
  jw = env->GetIntArrayElements( widths, &isCopy);

	for ( i=fs->wFirstChar; i < n; i++ ){
		buf[0] = i;
		jw[i] = X->screen->TextWidth( buf, fs);
	}

  env->ReleaseIntArrayElements( widths, jw, 0);

  return widths;
}


jint _cdecl
Java_java_awt_Toolkit_fntBytesWidth ( JNIEnv* env, jclass clazz, PegFont* fs,
									  jarray jBytes, jint off, jint len )
{
  jboolean  isCopy;
  jbyte     *jb = env->GetByteArrayElements( jBytes, &isCopy);
	jbyte     *jbo = jb + off;
  int       n = env->GetArrayLength( jBytes);
  int       w;
	char      *buf;

  if ( off+len > n ) len = n - off;

	buf = jbyte2CString( X, jbo, len);
	w = X->screen->TextWidth( buf, fs);

  env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);
  return w;
}

jint _cdecl
Java_java_awt_Toolkit_fntCharWidth ( JNIEnv* env, jclass clazz, PegFont* fs, jchar jChar )
{
	// pWidths seems not to be supported by PEG yet, and we don't have wide fonts

	char buf[2];
	buf[0] = (jChar & 0xff);
	buf[1] = 0;

	return X->screen->TextWidth( buf, fs);
}

jint _cdecl
Java_java_awt_Toolkit_fntCharsWidth ( JNIEnv* env, jclass clazz, PegFont* fs,
									  jarray jChars, jint off, jint len )
{
  jboolean  isCopy;
  jchar    *jc = env->GetCharArrayElements( jChars, &isCopy);
	jchar    *jco = jc + off;
  int      n = env->GetArrayLength( jChars);
  int      w;
	char     *buf;

  if ( off+len > n ) len = n - off;

	buf = jchar2CString( X, jco, len);
	w = X->screen->TextWidth( buf, fs);

  env->ReleaseCharArrayElements( jChars, jc, JNI_ABORT);
  return w;
}

jint _cdecl
Java_java_awt_Toolkit_fntStringWidth ( JNIEnv* env, jclass clazz, PegFont* fs, jstring jStr )
{
  jboolean isCopy;
  const jchar    *jc = env->GetStringChars( jStr, &isCopy);
  int      len = env->GetStringLength( jStr);
  int      w;
	char     *buf;

	buf = jchar2CString( X, (jchar*)jc, len);
	w = X->screen->TextWidth(buf, fs);

  env->ReleaseStringChars( jStr, jc);
  return w;
}

}