/**
 * tlk.c - native toolkit management
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */


#define MAIN

#include "toolkit.hpp"
#include "tlkprops.h"

extern "C" void PegTaskMain(void);

/********************************************************************************
 * auxiliary functions
 */

void
setupDebugConsole ()
{

}

extern "C" {
void
SendErrorToPeg ( char* msg )
{
	char *breakOnMe = msg;
	// will eventually print to the debug console..
}
}


/********************************************************************************
 *
 */

PegScreen *PScreen;

void
PegAppInitialize ( PegPresentationManager* presman )
{
	jmethodID initToolkit;

	X->screen = presman->Screen();
	X->presman = presman;

	// That's all for now, no color init (which is deferred, anyway), no
	// fonts (since we just have two of them), no window init (that's up to PEG).
	// We just have to inform the Java layer that it might start to query our
	// native characteristics

   //initToolkit = JniEnv->GetStaticMethodID( InitClass, "initialize", "()V");
	//JniEnv->CallStaticVoidMethod( InitClass, initToolkit);
}


extern "C" {

/*
 * Note that this is the *only* native method which is not in class Toolkit.
 * It is not allowed to be there because it otherwise would cause a deadlock
 * in Toolkit.clinit
 */
jboolean __cdecl
Java_java_awt_NativeCollector_ncInit ( JNIEnv* env, jclass clazz, jstring name )
{
  DBG( awt, ("init\n"));

  X->nBuf = 128;
  X->buf = (char*)AWT_MALLOC( X->nBuf);
	
  JniEnv = env;
  AWTError = env->FindClass( "java.awt.AWTError");
  InitClass = clazz;
	
  DBG_ACTION(awt, setupDebugConsole());


  return JNI_TRUE;  // just to please the compiler, if we got here, everything has finished
}

void
Java_java_awt_NativeCollector_ncLoop ( JNIEnv* env, jclass clazz )
{
  // watch out! this never returns until we shut down PEG (sits inside Execute)
  // initializations requiring the PegScreen and PegPresentationManager have to
  // be done from PegAppInitialize (which is called by PegTaskMain).
  // This implies that tlkInit does not have to be a synchronized method!
  PegTaskMain();  
}

jboolean __cdecl
Java_java_awt_Toolkit_tlkInit ( JNIEnv* env, jclass clazz, jstring name )
{
  // nothing to be done here, it's all in the ncInit
  return JNI_TRUE;
}

jstring __cdecl
Java_java_awt_Toolkit_tlkVersion ( JNIEnv* env, jclass clazz )
{
  return env->NewStringUTF( "PEG-1.0");
}

jint __cdecl
Java_java_awt_Toolkit_tlkProperties ( JNIEnv* env, jclass clazz )
{
	return (TLK_NATIVE_DISPATCHER_LOOP | TLK_IS_BLOCKING);
}

void __cdecl
Java_java_awt_Toolkit_tlkTerminate ( JNIEnv* env, jclass clazz )
{
  
}


jint __cdecl
Java_java_awt_Toolkit_tlkGetResolution ( JNIEnv* env, jclass clazz )
{
  /*
	 * Just a crude guess of the screen resolution in pixels per inch
   */
  return 75;
}


jint __cdecl
Java_java_awt_Toolkit_tlkGetScreenHeight ( JNIEnv* env, jclass clazz )
{
  return X->screen->GetYRes();
}


jint __cdecl
Java_java_awt_Toolkit_tlkGetScreenWidth ( JNIEnv* env, jclass clazz )
{
  return X->screen->GetXRes();
}

void __cdecl
Java_java_awt_Toolkit_tlkSync ( JNIEnv* env, jclass clazz )
{
  /* nothing to do for PEG, all output is sync */
}

void __cdecl
Java_java_awt_Toolkit_tlkFlush ( JNIEnv* env, jclass clazz )
{
	/* nothing to do for PEG, all output is sync */
}


void __cdecl
Java_java_awt_Toolkit_tlkBeep ( JNIEnv* env, jclass clazz )
{
	int dummy = 1;
  /* sorry, no beep available */
}

void __cdecl
Java_java_awt_Toolkit_tlkDisplayBanner ( JNIEnv* env, jclass clazz, jstring banner )
{
}

}
