/**
 * wnd.c - native toplevel window related functions
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "toolkit.hpp"

extern "C" {

void __cdecl
Java_java_awt_Toolkit_wndSetTitle ( JNIEnv* env, jclass clazz, MappedWindow* wnd, jstring jTitle )
{
	char *buf;
	PegDecoratedWindow* dw = 0;
	
	if ( wnd->pegType == PEG_TYPE_DIALOG )
		dw = (PegDecoratedWindow*) wnd->prt.dialog;
	else if ( wnd->pegType == PEG_TYPE_DECO )
		dw = wnd->prt.deco;
	
	if ( dw && jTitle ) {
		buf = java2CString( env, X, jTitle);
		
		DBG( awt_wnd, ("setTitle: %x, \"%s\"\n", wnd, buf));
		
		dw->SetTitle( buf);		
	}
}


void __cdecl
Java_java_awt_Toolkit_wndSetResizable ( JNIEnv* env, jclass clazz,
																			 MappedWindow* wnd, jboolean isResizable,
																			 int x, int y, int width, int height )
{
	DBG( awt_wnd, ("setResizable: %x, %d, %d,%d,%d,%d\n", wnd, isResizable, x,y,width,height));
	
	if ( !isResizable ) {
		wnd->pt->RemoveStatus( PSF_MOVEABLE|PSF_SIZEABLE);
	}
	else {
		wnd->pt->AddStatus( PSF_MOVEABLE|PSF_SIZEABLE);
	}
}


MappedWindow* __cdecl
Java_java_awt_Toolkit_wndCreateFrame ( JNIEnv* env, jclass clazz, jstring jTitle,
									   jint x, jint y, jint width, jint height,
									   jint jCursor, jint clrBack, jboolean isResizable )
{
	PegRect   bounds;
	PegTitle  *ptitle;
	JPegDecoWindow *wnd;
	char *buf;
	
	DBG( awt_wnd, ("createFrame( %p, %d,%d,%d,%d,..)\n", jTitle,x,y,width,height));
	
	bounds.Set( x, y, x+width, y+height);
	wnd = new JPegDecoWindow( bounds);
	
	buf = java2CString( env, X, jTitle);
	ptitle = new PegTitle( buf, TF_SYSBUTTON|TF_MINMAXBUTTON|TF_CLOSEBUTTON|TT_COPY);
	
	// adapt PegTitle colors
	ptitle->muColors[PCI_NORMAL]   = CLR_GRAY1;
	ptitle->muColors[PCI_SELECTED] = CLR_BLUE;
	ptitle->muColors[PCI_NTEXT]    = CLR_BLACK;
	ptitle->muColors[PCI_STEXT]    = CLR_YELLOW;
	
	wnd->Add( ptitle);
	
	if ( !isResizable )
		wnd->RemoveStatus( PSF_SIZEABLE);
	
	wnd->tgtClip = wnd->pt->mClip;  // save it for future graphics output
	
	return wnd;
}


MappedWindow* __cdecl
Java_java_awt_Toolkit_wndCreateWindow ( JNIEnv* env, jclass clazz, void* owner,
										jint x, jint y, jint width, jint height,
										jint jCursor, jint clrBack )
{
	PegRect bounds;
	JPegWindow *wnd;
	
	DBG( awt_wnd, ("createWindow( %p, %d,%d,%d,%d,..)\n", owner,x,y,width,height));
	
	bounds.Set( x, y, x+width, y+height);
	wnd = new JPegWindow( bounds);
	wnd->tgtClip = wnd->pt->mClip;
	
	return wnd;
}


MappedWindow* __cdecl
Java_java_awt_Toolkit_wndCreateDialog ( JNIEnv* env, jclass clazz, void* owner, jstring jTitle,
										jint x, jint y, jint width, jint height,
										jint jCursor, jint clrBack, jboolean isResizable )
{
	PegRect bounds;
	JPegDialog *wnd;
	char *buf;
	
	DBG( awt_wnd, ("createDialog( %p,%p, %d,%d,%d,%d,..)\n", owner,jTitle,x,y,width,height));
	
	bounds.Set( x, y, x+width, y+height);
	wnd = new JPegDialog( bounds);
	
	buf = java2CString( env, X, jTitle);
	wnd->Add(new PegTitle( buf, TF_CLOSEBUTTON|TT_COPY));
	
	wnd->tgtClip = wnd->pt->mClip;
	
	return wnd;
}


void __cdecl
Java_java_awt_Toolkit_wndDestroyWindow ( JNIEnv* env, jclass clazz, MappedWindow* wnd )
{
  DBG( awt_wnd, ("destroy window: %x\n", wnd));

	wnd->isVisible = FALSE;
	// we leave the real destroy up to PEG (during unregisterWindow), to make sure we
	// don't delete anything while PEG is still using it
	wnd->pt->Presentation()->Remove( wnd->pt);	
}


void __cdecl
Java_java_awt_Toolkit_wndRequestFocus ( JNIEnv* env, jclass clazz, MappedWindow* wnd )
{
  DBG( awt_wnd, ("request focus: %x\n", wnd));

	if ( wnd != X->current ) {
		wnd->pt->Presentation()->MoveFocusTree( wnd->pt);	
	}
}

void __cdecl
Java_java_awt_Toolkit_wndSetFrameInsets ( JNIEnv* env, jclass clazz,
										  jint top, jint left, jint bottom, jint right )
{
	// we are not interested in guesses, we know the real values
	jclass     cls = env->FindClass( "java.awt.Frame");
	jmethodID  setDecoInsets = env->GetStaticMethodID( cls, "setDecoInsets","(IIIII)V");
	
	top = PEG_FRAME_WIDTH + (DEFAULT_FONT)->uHeight + 2;
	left = right = bottom = PEG_FRAME_WIDTH;
	
	X->frameInsets.top  = top;
	X->frameInsets.left = left;
	X->frameInsets.bottom = bottom;
	X->frameInsets.right = right;
	X->frameInsets.guess = 0;
	
	env->CallStaticVoidMethod( cls, setDecoInsets,
		top, left, bottom, right, (jint) -1);
}

void __cdecl
Java_java_awt_Toolkit_wndSetDialogInsets ( JNIEnv* env, jclass clazz,
										  jint top, jint left, jint bottom, jint right )
{
	jclass     cls = env->FindClass( "java.awt.Dialog");
	jmethodID  setDecoInsets = env->GetStaticMethodID( cls, "setDecoInsets","(IIIII)V");
	
	top = PEG_FRAME_WIDTH + (DEFAULT_FONT)->uHeight + 2;
	left = right = bottom = PEG_FRAME_WIDTH;
	
	X->dialogInsets.top  = top;
	X->dialogInsets.left = left;
	X->dialogInsets.bottom = bottom;
	X->dialogInsets.right = right;
	X->dialogInsets.guess = 0;
	
	env->CallStaticVoidMethod( cls, setDecoInsets,
		top, left, bottom, right, (jint) -1);
}

void __cdecl
Java_java_awt_Toolkit_wndSetBounds ( JNIEnv* env, jclass clazz, MappedWindow* wnd,
									 jint x, jint y, jint width, jint height,
									 jboolean isResizable )
{
	PegRect r;
	
	DBG( awt_wnd, ("setBounds: %x %d,%d,%d,%d\n", wnd, x, y, width, height));
	
	if ( width < 0 )  width = 1;
	if ( height < 0 ) height = 1;
	
	r.Set( x, y, x+width, y+height);
	
	wnd->pt->Resize( r);	
}

void __cdecl
Java_java_awt_Toolkit_wndRepaint ( JNIEnv* env, jclass clazz, MappedWindow* wnd,
								   jint x, jint y, jint width, jint height )
{
	PegRect r;
	
	DBG( awt_wnd, ("wndRepaint: %x %d,%d,%d,%d\n", wnd, x, y, width, height));
	
	r.Set( x, y, x+width, y+height);
	
	wnd->pt->Invalidate( r);
}

void __cdecl
Java_java_awt_Toolkit_wndSetIcon ( JNIEnv* env, jclass clazz, void* wnd, void* img )
{
}


void __cdecl
Java_java_awt_Toolkit_wndSetVisible ( JNIEnv* env, jclass clazz, MappedWindow* wnd, jboolean showIt )
{
	DBG( awt_wnd, ("setVisible: %x %d\n", wnd, showIt));
	
	if ( showIt ){
		wnd->isVisible = 1;
		wnd->pt->Presentation()->Add( wnd->pt);
	}
	else {
		wnd->isVisible = 0;
		wnd->pt->Presentation()->Remove( wnd->pt);
	}
}


void __cdecl
Java_java_awt_Toolkit_wndToBack ( JNIEnv* env, jclass clazz, PegWindow* wnd )
{
	DBG( awt_wnd, ("toBack: %x\n", wnd));
}


void __cdecl
Java_java_awt_Toolkit_wndToFront ( JNIEnv* env, jclass clazz, PegWindow* wnd )
{
	DBG( awt_wnd, ("toFront: %x\n", wnd));
}


static UCHAR cursors[14] =
{
	PPT_NORMAL,     /*  0: DEFAULT_CURSOR     */
	PPT_NORMAL,     /*  1: CROSSHAIR_CURSOR   */
	PPT_IBEAM,      /*  2: TEXT_CURSOR        */
	PPT_NORMAL,     /*  3: WAIT_CURSOR        */
	PPT_NESW_SIZE,  /*  4: SW_RESIZE_CURSOR   */
	PPT_NWSE_SIZE,  /*  5: SE_RESIZE_CURSOR   */
	PPT_NWSE_SIZE,  /*  6: NW_RESIZE_CURSOR   */
	PPT_NESW_SIZE,  /*  7: NE_RESIZE_CURSOR   */
	PPT_VSIZE,      /*  8: N_RESIZE_CURSOR    */
	PPT_VSIZE,      /*  9: S_RESIZE_CURSOR    */
	PPT_HSIZE,      /* 10: W_RESIZE_CURSOR    */
	PPT_HSIZE,      /* 11: E_RESIZE_CURSOR    */
	PPT_HAND,       /* 12: HAND_CURSOR        */
	PPT_NORMAL      /* 13: MOVE_CURSOR        */
};

void __cdecl
Java_java_awt_Toolkit_wndSetCursor ( JNIEnv* env, jclass clazz, MappedWindow* wnd, jint jCursor )
{
  DBG( awt_wnd, ("setCursor: %x, %d\n", wnd, jCursor));

	if ( (jCursor > sizeof(cursors)) || (jCursor < 0) )
		jCursor = 0;

	X->pCursor = cursors[jCursor];

	wnd->pt->Screen()->SetPointerType( X->pCursor);	
}

}
