/**
* cbd.cpp - clipboard handling
*
* Copyright (c) 1999
*      Transvirtual Technologies, Inc.  All rights reserved.
*
* See the file "license.terms" for information on usage and redistribution 
* of this file. 
*/

#include "toolkit.hpp"


jclass     NativeClipboard;
jmethodID  lostOwnership;
jmethodID  createTransferable;
jmethodID  getNativeData;

unsigned int JAVA_OBJECT;

void 
selectionClear ( JNIEnv* env)
{
	env->CallStaticVoidMethod( NativeClipboard, lostOwnership);
}
 
void
selectionRequest ( JNIEnv* env, unsigned int target )
{
	TCHAR      *mime;
	jstring    jMimeType;
	jarray     jData;
	jbyte      *data;
	jboolean   isCopy;
	int        len;
	HANDLE		cbdData;
	
	if ( target == JAVA_OBJECT )
		mime = TEXT("application/x-java-serialized-object");
	else if ( target == CF_TEXT )
		mime = TEXT("text/plain; charset=unicode");
	else
		return;
	
	jMimeType = env->NewStringUTF( winString2UTF(X, mime, _tcslen(mime)));
	jData = env->CallStaticObjectMethod( NativeClipboard, getNativeData, jMimeType);
	
	if ( jData ) {
		data = env->GetByteArrayElements( jData, &isCopy);
		len = env->GetArrayLength( jData);

		OpenClipboard( NULL);
		EmptyClipboard();
		cbdData = LocalAlloc( LPTR, len);
		memcpy( cbdData, data, len);
		SetClipboardData( target, cbdData);
		CloseClipboard();

		env->ReleaseByteArrayElements( jData, data, JNI_ABORT);
	}
	

}



/*****************************************************************************************
* exported functions
*/
#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

extern "C" {
	JNIEXPORT jobject JNICALL
		Java_java_awt_Toolkit_cbdInitClipboard ( JNIEnv* env, jclass clazz )
	{
		NativeClipboard = env->FindClass( "java.awt.NativeClipboard");
		lostOwnership   = env->GetStaticMethodID( NativeClipboard,
			"lostOwnership", "()V");
		getNativeData = env->GetStaticMethodID( NativeClipboard,
			"getNativeData", "(Ljava/lang/String;)[B");
		createTransferable = env->GetStaticMethodID( NativeClipboard,
			"createTransferable",
			"(Ljava/lang/String;[B)Ljava/awt/datatransfer/Transferable;");
		
		JAVA_OBJECT  = RegisterClipboardFormat( TEXT("application/x-java-serialized-object"));
		return 0;
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_cbdFreeClipboard ( JNIEnv* env, jclass clazz, ClipBoard* cbd )
	{
	}
	
	JNIEXPORT jboolean JNICALL
		Java_java_awt_Toolkit_cbdSetOwner ( JNIEnv* env, jclass clazz, ClipBoard* cbd )
	{
		return JNI_TRUE;
	}
	
	
	/*
	* this is the incoming handler
	* there is no notion of a type request from Java, we therefore have to settle on
	* a stack of suitable formats (with most specific type on top, "fallback" type at bottom)
	*/
	
	JNIEXPORT jobject JNICALL
		Java_java_awt_Toolkit_cbdGetContents ( JNIEnv* env, jclass clazz, ClipBoard* cbd )
	{
		HANDLE		clbd;
		TCHAR       *mime;
		int			len;
		jarray      jdata;
		jstring     jMimeType;

		OpenClipboard( NULL);

		if ( (clbd = GetClipboardData( JAVA_OBJECT)) ) {
			mime = TEXT("application/x-java-serialized-object");
		}
		else if ( (clbd = GetClipboardData( CF_TEXT)) ){
			mime = TEXT("text/plain; charset=unicode");
		}

		CloseClipboard();

		if ( clbd ) {
			len = LocalSize( clbd);
			jMimeType = env->NewStringUTF( winString2UTF(X, mime, _tcslen(mime)));
			jdata = env->NewByteArray( len);
			env->SetByteArrayRegion( jdata, 0, len, (jbyte*)clbd);
			LocalFree( clbd);
			
			return (env->CallStaticObjectMethod( NativeClipboard, createTransferable,
				jMimeType, jdata));
		}
		else {
			return 0;
		}
	}
	
}
