/**
* fnt.cpp - 
*
* Copyright (c) 1999
*      Transvirtual Technologies, Inc.  All rights reserved.
*
* See the file "license.terms" for information on usage and redistribution 
* of this file. 
*/


#include <string.h>
#include "toolkit.hpp"

/*******************************************************************************
* Font support
*/
extern "C" {

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT
	
	JNIEXPORT void* JNICALL
		Java_java_awt_Toolkit_fntInitFont ( JNIEnv* env, jclass clazz, jstring jSpec, jint style, jint size )
	{
		LOGFONT fnt;
		HFONT	fh;
		Font    *pFnt;
		TCHAR    *spec = java2WinString( env, X, jSpec);
		
		_tcscpy( fnt.lfFaceName, spec);

		fnt.lfHeight = -size;
		fnt.lfWidth = 0; 
		fnt.lfEscapement = 0; 
		fnt.lfOrientation = 0; 
		fnt.lfWeight = ( style & 1) ? FW_BOLD : FW_NORMAL; 
		fnt.lfItalic = ( style & 2) ? 1 : 0; 
		fnt.lfUnderline = 0; 
		fnt.lfStrikeOut = 0; 
		fnt.lfCharSet = 0; //ANSI_CHARSET; 
		fnt.lfOutPrecision = 0; //OUT_TT_PRECIS; 
		fnt.lfClipPrecision = 0; 
		fnt.lfQuality = 0; 
		fnt.lfPitchAndFamily = 0; 
		
		fh =  CreateFontIndirect( &fnt);
		
		pFnt = new Font();
		pFnt->fnt = fh;
		pFnt->hasMetric = false;
		
		return pFnt;
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_fntFreeFont ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		DeleteObject( pFont->fnt);
		delete pFont;
	}
	
	/*******************************************************************************
	* FontMetrics support
	*/
	
	JNIEXPORT void* JNICALL
		Java_java_awt_Toolkit_fntInitFontMetrics ( JNIEnv* env, jclass clazz, Font* pFont, jobject metrics )
	{
		jclass	  cls = env->GetObjectClass(metrics);
		jfieldID  fHeight = env->GetFieldID( cls, "height", "I");
		jfieldID  fDescent = env->GetFieldID( cls, "descent", "I");
		jfieldID  fMaxDescent = env->GetFieldID( cls, "maxDescent", "I");
		jfieldID  fAscent = env->GetFieldID( cls, "ascent", "I");
		jfieldID  fLeading = env->GetFieldID( cls, "leading", "I");
		jfieldID  fMaxAscent = env->GetFieldID( cls, "maxAscent", "I");
		jfieldID  fMaxAdvance = env->GetFieldID( cls, "maxAdvance", "I");
		jfieldID  fFixedWidth = env->GetFieldID( cls, "fixedWidth", "I");
		jfieldID  fIsWideFont = env->GetFieldID( cls, "isWideFont", "B");

		if ( ! pFont->hasMetric ) {
			SelectObject( X->display, pFont->fnt);
			GetTextMetrics( X->display, &(pFont->fm));
			pFont->hasMetric = true;
		}

		env->SetIntField( metrics, fHeight, pFont->fm.tmHeight );
		env->SetIntField( metrics, fDescent, pFont->fm.tmDescent );
		env->SetIntField( metrics, fMaxDescent, pFont->fm.tmDescent );
		env->SetIntField( metrics, fAscent, pFont->fm.tmAscent );
		env->SetIntField( metrics, fLeading, pFont->fm.tmInternalLeading );
		env->SetIntField( metrics, fMaxAscent, pFont->fm.tmAscent );
		env->SetIntField( metrics, fMaxAdvance, pFont->fm.tmOverhang );
		env->SetIntField( metrics, fFixedWidth, 0 );
		env->SetBooleanField( metrics, fIsWideFont, FALSE );
	
		return pFont;
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_fntFreeFontMetrics ( JNIEnv* env, jclass clazz, Font* pFont )
	{
	}
	
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetAscent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmAscent;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetDescent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmDescent;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetFixedWidth ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return 0;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetHeight ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmHeight;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetLeading ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmInternalLeading;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetMaxAdvance ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmOverhang;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetMaxAscent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmAscent;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntGetMaxDescent ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		return pFont->fm.tmDescent;
	}
	
	JNIEXPORT jboolean JNICALL
		Java_java_awt_Toolkit_fntIsWideFont ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		//return ((pFont->fm.tmPitchAndFamily & 0xf0) == TMPF_VECTOR);
		return FALSE;
	}
	
	JNIEXPORT jobject JNICALL
		Java_java_awt_Toolkit_fntGetWidths ( JNIEnv* env, jclass clazz, Font* pFont )
	{
		jarray    widths;
		jint      *jw;
		jboolean isCopy;
		
		widths = env->NewIntArray( 256);
		jw = env->GetIntArrayElements( widths, &isCopy);
		
#if !defined(UNDER_CE)
		SelectObject( X->display, pFont->fnt);
		GetCharWidth( X->display, 0, 255, jw);
#else
		int i, afw = 2 * pFont->fm.tmAveCharWidth;
		for ( i=0; i<256; i++) {
			jw[i] = afw;
		}
#endif
		
		env->ReleaseIntArrayElements( widths, jw, 0);
		return widths;
	}
	
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntBytesWidth ( JNIEnv* env, jclass clazz, Font* pFont,
		jarray jBytes, jint off, jint len )
	{
		SIZE sz;
		jboolean  isCopy;
		jbyte    *jb = env->GetByteArrayElements( jBytes, &isCopy);
		int       n = env->GetArrayLength( jBytes);
		TCHAR	 *cs;
		
		if ( off+len > n ) len = n - off;
		
		cs = jchar2WinString( X, (jchar*)jb+off, len);
		SelectObject( X->display, pFont->fnt);
		GetTextExtentExPoint( X->display, cs, len, 0, NULL, NULL, &sz);
		
		env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);
		return sz.cx;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntCharWidth ( JNIEnv* env, jclass clazz, Font* pFont, jchar jChar )
	{
		SIZE sz;
		SelectObject( X->display, pFont->fnt);
		GetTextExtentExPoint( X->display, (TCHAR*)&jChar, 1, 0, NULL, NULL, &sz);
		return sz.cx;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntCharsWidth ( JNIEnv* env, jclass clazz, Font* pFont,
		jarray jChars, jint off, jint len )
	{
		SIZE      sz;
		jboolean  isCopy;
		jchar    *jc = env->GetCharArrayElements( jChars, &isCopy);
		int      n = env->GetArrayLength( jChars);
		TCHAR	 *cs;
		
		if ( off+len > n ) len = n - off;
		
		cs = jchar2WinString( X, jc+off, len);
		SelectObject( X->display, pFont->fnt);
		GetTextExtentExPoint( X->display, cs, len, 0, NULL, NULL, &sz);
		
		env->ReleaseCharArrayElements( jChars, jc, JNI_ABORT);
		return sz.cx;
	}
	
	JNIEXPORT jint JNICALL
		Java_java_awt_Toolkit_fntStringWidth ( JNIEnv* env, jclass clazz, Font* pFont, jstring jStr )
	{
		SIZE sz;
		TCHAR *jc = java2WinString( env, X, jStr);
		
		SelectObject( X->display, pFont->fnt);
		GetTextExtentExPoint( X->display, jc, _tcslen(jc), 0, NULL, NULL, &sz);
		
		return sz.cx;
	}
	
}