/**
* gra.cpp - 
*
* Copyright (c) 1999
*      Transvirtual Technologies, Inc.  All rights reserved.
*
* See the file "license.terms" for information on usage and redistribution 
* of this file. 
*/

#include "math.h"
#include "toolkit.hpp"

#define  WINDOW   0
#define  IMAGE    1
#define  GRAPHICS 2

extern "C" {
	jlong Java_java_awt_Toolkit_clrBright ( JNIEnv*,jclass, jint jrgb);
	jlong Java_java_awt_Toolkit_clrDark ( JNIEnv*,jclass, jint jrgb);
}

struct Graphics;
static Graphics* currentShared;


/*
* helper macros to defer DC attribute changes between
* shared DCs ( images ), to actual painting ( consistency )
*/
#define DC_ATTRIB(expr)		if ( true /*! sharedDC*/ ) { expr } \
							else { currentShared = 0; }

#define DC_ATTRIB_GR(expr)	if ( true /*! gr->sharedDC*/ ) { expr } \
							else { currentShared = 0; }

#define BEGIN_PAINT(gra)	if ( gra != currentShared ) { \
								if ( gra->sharedDC ) { \
									if ( gra->clipRgn ) SelectClipRgn( gra->dc, gra->clipRgn); \
									if ( gra->brush) SelectObject( gra->dc, gra->brush); \
									if ( gra->pen)	 SelectObject( gra->dc, gra->pen); \
									if ( gra->font)	 SelectObject( gra->dc, gra->font); \
									SetTextColor( gra->dc, gra->xor ? (gra->fgClr ^ gra->xclr) : gra->fgClr); \
									SetROP2( gra->dc, gra->xor ? R2_XORPEN : R2_COPYPEN); \
									currentShared = gra; \
								} \
							}

#if defined(UNDER_CE)
#define END_PAINT(GRA)
#else
#define END_PAINT(GRA)		GdiFlush()
#endif


void _cdecl getArcPoints( int x, int y, int width, int height,
						 int startAngle, int arcAngle, POINT& ps, POINT& pe) {
	
	int maxr, xc, yc;
	double rad, sa, ea;
	
	rad = -3.1415926535 / 180;
	maxr = (int) (max( width, height)+1) / 2;
	sa = rad * startAngle;
	ea = rad * (startAngle+arcAngle);
	xc = x + (width+1)/2;
	yc = y + (height+1)/2;
	
	ps.x = (int)( cos(sa) * maxr) + xc;
	ps.y = (int)( sin(sa) * maxr) + yc;
	pe.x = (int)( cos(ea) * maxr) + xc;
	pe.y = (int)( sin(ea) * maxr) + yc;
	
//	printf( "x:%d y:%d w:%d h:%d x1:%d y1:%d x2:%d y2:%d\n",
//			x, y, width, height, ps.x, ps.y, pe.x, pe.y);
}

int __cdecl jarray2Points ( POINT** pp, int x0, int y0,
						   jarray xPoints, jarray yPoints, int nPoints )
{
	register int i;
	int      n;
	jboolean isCopy;
	jint     *jx = JniEnv->GetIntArrayElements( xPoints, &isCopy);
	jint     *jy = JniEnv->GetIntArrayElements( yPoints, &isCopy);
	
	if ( (n = JniEnv->GetArrayLength( xPoints)) < nPoints ) nPoints = n;
	if ( (n = JniEnv->GetArrayLength( yPoints)) < nPoints ) nPoints = n;
	
	*pp = (POINT*)getBuffer( X, (nPoints+1)*sizeof(POINT));
	
	for ( i=0; i<nPoints; i++){
		(*pp)[i].x = jx[i] + x0;
		(*pp)[i].y = jy[i] + y0;
	}
	
	JniEnv->ReleaseIntArrayElements( xPoints, jx, JNI_ABORT);
	JniEnv->ReleaseIntArrayElements( yPoints, jy, JNI_ABORT);
	
	return nPoints;
}


HBRUSH HB_0 = (HBRUSH)GetStockObject( NULL_BRUSH);
HPEN   HP_0 = (HPEN)GetStockObject( NULL_PEN);

struct Graphics {
	HDC       dc;
	KWnd	  *wnd;

	HBRUSH	  brush;
	HPEN	  pen;
	Font	  *font;
	HRGN	  clipRgn;

	COLORREF  fgClr;
	COLORREF  bgClr;
	COLORREF  xclr;
	bool      xor;
	int		  x0;
	int		  y0;

	bool	  sharedDC;
	RECT	  clip;
	
	void init ( HDC hdc, KWnd* pWnd, int cfg, int cbg, Font* fnt, bool shared) {
		dc = hdc;
		wnd = pWnd;
		brush = 0;
		pen = 0;
		clipRgn = 0;
		sharedDC = shared;

		if ( shared ) {
			currentShared = 0;
		}

#if !defined(UNDER_CE)
		SetTextAlign( dc, TA_BASELINE | TA_LEFT);
#endif
		SetBkMode( dc, TRANSPARENT);

		setPaintMode();		
		setForeground( cfg);
		setBackground( cbg);
		setFont( fnt);
	}
	
	void setXORMode( int xorClr) {
		xor = true;
		xclr = xorClr;
		setForeground( fgClr);
		DC_ATTRIB (
			SetROP2( dc, R2_XORPEN);
		)
	}
	
	void setPaintMode() {
		xor = false;
		setForeground( fgClr);
		DC_ATTRIB (
			SetROP2( dc, R2_COPYPEN);
		)
	}
	
	void setFont( Font* fnt) {
		font = fnt;
		DC_ATTRIB (
			SelectObject( dc, fnt->fnt);
		)
	}
	
	bool setBackground ( int irgb) {
		if ( irgb < 0 )
			return false;
		
		bgClr = (COLORREF)irgb;
		return true;
	}
	
	bool setForeground ( int irgb) {
		COLORREF xc = (COLORREF)(xor ? (irgb ^ xclr) : irgb);

		if ( irgb < 0 )
			return false;
		
		fgClr = (COLORREF)irgb;

		if ( brush ) DeleteObject( brush);
		if ( pen )	 DeleteObject( pen);

		brush = CreateSolidBrush( xc);
		pen = CreatePen( PS_SOLID, 0, xc);

		DC_ATTRIB (
			SetTextColor( dc, xc);
			SelectObject( dc, brush);
			SelectObject( dc, pen);
		)

		return true;
	}
	
	void drawArc( int x, int y, int w, int h, int startAngle, int sweepAngle, bool fill ) {

		BEGIN_PAINT( this);

		if ( sweepAngle >= 360 ) {
			if ( fill ) {
				SelectObject( dc, HP_0);
				Ellipse( dc, x+x0, y+y0, x+x0+w+1, y+y0+h+1);
				DC_ATTRIB (
					SelectObject( dc, pen);
				)
			}
			else {
				SelectObject( dc, HB_0);
				Ellipse( dc, x+x0, y+y0, x+x0+w, y+y0+h);
				DC_ATTRIB (
					SelectObject( dc, brush);
				)
			}
		}
		else {
		#if !defined(UNDER_CE)
			POINT p1, p2;

			getArcPoints( x+x0, y+y0, w, h, startAngle, sweepAngle, p1, p2);
			if ( fill ) {
				SelectObject( dc, HP_0);
				Pie( dc, x+x0, y+y0, x+w+x0+1, y+h+y0+1, p1.x, p1.y, p2.x, p2.y);
				DC_ATTRIB (
					SelectObject( dc, pen);
				)
			}
			else {
				Arc( dc, x+x0, y+y0, x+w+x0, y+h+y0, p1.x, p1.y, p2.x, p2.y);
			}
		#endif
		}

		END_PAINT( this);
	}

	boolean fillRect( int x, int y, int w, int h, int clr) {
		HBRUSH hbr = 0;

		if ( clr < 0 )
			return false;

		BEGIN_PAINT(this);

		if ( (COLORREF)clr != fgClr ) {
			hbr = CreateSolidBrush( (COLORREF)clr );
			SelectObject( dc, hbr);
		}
		
		if ( X->osi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ) {
			// strange repaint() behaviour under NT ???
			PatBlt( dc, x+x0, y+y0, w, h, PATCOPY );
		}
		else {
			SelectObject( dc, HP_0);
			Rectangle( dc, x+x0, y+y0, x+x0+w+1, y+y0+h+1);
			DC_ATTRIB (
				SelectObject( dc, pen);
			)
		}
		
		if ( hbr != 0) {
			DC_ATTRIB (
				SelectObject( dc, brush);
			)
			DeleteObject( hbr);
		}
		
		END_PAINT(this);
		return true;
	}
	
	void drawPel( int x, int y) {
		BEGIN_PAINT( this);
		PIXEL( dc, x+x0, y+y0, fgClr);
		END_PAINT( this);
	}

	void drawRect( int x, int y, int w, int h ) {
		POINT pts[5];

		pts[0].x = x+x0;
		pts[0].y = y+y0;
		pts[1].x = x+w+x0;
		pts[1].y = y+y0;
		pts[2].x = x+w+x0;
		pts[2].y = y+h+y0;
		pts[3].x = x+x0;
		pts[3].y = y+h+y0;
		pts[4].x = x+x0;
		pts[4].y = y+y0;

		BEGIN_PAINT(this);
		Polyline(dc, pts, 5);
		END_PAINT(this);
	}
	
	void drawLine ( int x1, int y1, int x2, int y2) {
		POINT pts[2];

		if ( (y1 != y2) || (x1 == x2) ) {
			if ( y1 < y2 )	y2++;
			else			y2--;
		}
		if ( (x1 != x2) || (y1 == y2) ) {
			if ( x1 < x2 )	x2++;
			else			x2--;
		}

		pts[0].x = x1+x0;
		pts[0].y = y1+y0;
		pts[1].x = x2+x0;
		pts[1].y = y2+y0;

		BEGIN_PAINT(this);
		Polyline(dc, pts, 2);
		END_PAINT(this);
	}
	
	void drawText( int x, int y, TCHAR* txt, int len) {
#if defined(UNDER_CE)
	/* The 'y' value is the base of the character but WinCE things is the top - we deduce 10
	* to adjust it (we should be doing this based on the font size.
	*/
		//y -= 10;
		if ( ! font->hasMetric ) {
			GetTextMetrics( dc, &(font->fm));
			font->hasMetric = true;
		}
		y -= font->fm.tmAscent;
#endif
		BEGIN_PAINT(this);
		ExtTextOut( dc, x+x0, y+y0, 0, NULL, txt, len, NULL);
		END_PAINT(this);
	}
	
	void translate( int x, int y) {
	/*
	* we use explicit translation, do not change
	* viewport here
	*/
		x0 = wnd ? (x - wnd->insets.left) : x;
		y0 = wnd ? (y - wnd->insets.top) : y;
	}
	
	void cleanUp() {
		if ( brush ) { 
			SelectObject( dc, HB_0 );
			DeleteObject( brush);
			brush = 0;
		}
		if ( pen ) {
			SelectObject( dc, HP_0 );
			DeleteObject( pen);
			pen = 0;
		}
		if ( clipRgn ) {
			DeleteObject( clipRgn);
			clipRgn = 0;
		}
		if ( wnd ) {
			ReleaseDC( wnd->hwnd, dc);
			wnd = 0;
		}
		if ( sharedDC ) {
			sharedDC = false;	
			currentShared = 0;
		}
		
	}
	
	void setClip( int x, int y, int w, int h ) {
		x += x0;
		y += y0;

		if ( clipRgn && (clip.left == x) && (clip.top == y) &&
			(clip.right == x+w) && (clip.bottom == y+h ) ) {
			return;
		}

		clip.left=x;	clip.top=y;
		clip.right=x+w; clip.bottom=y+h;
		
		if ( clipRgn ) {
			DeleteObject( clipRgn);
		}

		clipRgn = CreateRectRgn( x, y, x+w, y+h);

		DC_ATTRIB (
			SelectClipRgn( dc, clipRgn);
		)
	}
	
};


/*********************************************************
*	exported functions
**********************************************************/

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

extern "C" {
	JNIEXPORT void* JNICALL
		Java_java_awt_Toolkit_graInitGraphics ( JNIEnv* env, jclass clazz,
		Graphics* gr, jobject tgt, jint tgtType,
		jint xOff, jint yOff,
		jint xClip, jint yClip, jint wClip, jint hClip,
		jobject fnt, jint fg, jint bg )
	{
		KWnd *wnd = 0;
		HDC hdc = 0;
		Image *img = 0;
		bool shared = false;
		Graphics *gr1 = 0;
		
		//printf( "%d,%d   %d,%d,%d,%d  %x %x %d\n", xOff, yOff, xClip, yClip, wClip, hClip, gr, tgt, tgtType);
		
		if ( tgt ) {
			if ( gr ) {
				gr->cleanUp();
			}
			
			switch ( tgtType) {
			case WINDOW:
				wnd = (KWnd*)tgt;
				hdc = GetDC( wnd->hwnd);
				break;
			case IMAGE:
				img = (Image*)tgt;
				hdc = img->dc;
				break;
			case GRAPHICS:
				gr1 = (Graphics*)tgt;
				wnd = gr1->wnd;
				if ( wnd ) {
					hdc = GetDC( wnd->hwnd);
				}
				else {
					// caution: going to share an image dc
					hdc = gr1->dc;
					gr1->sharedDC = true;
					shared = true;
				}
				break;
			}
		}
		else if ( gr ) {
			hdc = gr->dc;
			wnd = gr->wnd;
		}
		
		if ( ! gr) {
			gr = new Graphics();
		}
		
		gr->init( hdc, wnd, fg, bg, (Font*)fnt, shared);
		gr->translate( xOff, yOff);
		gr->setClip( xClip, yClip, wClip, hClip);
		
		return gr;
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFreeGraphics ( JNIEnv* env, jclass clazz, Graphics* gr )
	{
		DBG( awt_gra, ("freeGraphics: %x\n", gr));
		
		gr->cleanUp();
		delete gr;
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graCopyArea ( JNIEnv* env, jclass clazz, Graphics* gr, 
		jint x, jint y, jint width, jint height,
		jint xDelta, jint yDelta )
	{
		DBG( awt_gra, ("copyArea: %x, %d,%d-%d,%d  %d,%d\n", gr, x,y,width,height, xDelta,yDelta));
		BEGIN_PAINT(gr);
		BitBlt( gr->dc, x+xDelta+gr->x0, y+yDelta+gr->y0, width, height, gr->dc, x+gr->x0, y+gr->y0, SRCCOPY);
		END_PAINT(gr);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graClearRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( awt_gra, ("clearRect: %x, %d,%d-%d,%d\n", gr, x,y,width,height));
		gr->fillRect( x, y, width, height, (int)gr->bgClr);	
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawBytes ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray jBytes, jint offset, jint len, jint x, jint y )
	{
		jboolean isCopy;
		int      n;
		jbyte    *jb, *jbo;
		TCHAR	 *cs;
		
		if ( !jBytes ) return;
		
		n   = env->GetArrayLength( jBytes);
		jb = env->GetByteArrayElements( jBytes, &isCopy);
		jbo = jb + offset;
		
		if ( offset+len > n )
			len = n - offset;
		
		cs = jbyte2WinString( X, jbo, len);
		
		DBG( awt_gra, ("drawBytes: %x, %x,%d,%d  \"%s\"  %d,%d\n", gr, jBytes,offset,len, cs, x,y));
		gr->drawText( x, y, cs, len);
		
		env->ReleaseByteArrayElements( jBytes, jb, JNI_ABORT);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawChars ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray jChars, jint offset, jint len, jint x, jint y )
	{
		jboolean isCopy;
		int      n;
		jchar    *jc, *jco;
		TCHAR	 *cs;
		
		if ( !jChars ) return;
		
		n   = env->GetArrayLength( jChars);
		jc = env->GetCharArrayElements( jChars, &isCopy);
		jco = jc + offset;
		
		if ( offset+len > n )
			len = n - offset;
		
		cs = jchar2WinString( X, jco, len);
		
		DBG( awt_gra, ("drawChars: %x, %x,%d,%d  \"%s\"  %d,%d\n", gr, jChars,offset,len, cs, x,y));
		gr->drawText( x, y, cs, len);
		
		env->ReleaseCharArrayElements( jChars, jc, JNI_ABORT);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawString ( JNIEnv* env, jclass clazz,
		Graphics* gr, jstring str, jint x, jint y )
	{
		TCHAR  *jc = java2WinString( env, X, str);
		
		DBG( awt_gra, ("drawString: %x  \"%s\"  %d,%d\n", gr, jc, x,y));
		gr->drawText( x, y, jc, _tcslen( jc));
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawLine ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x1, jint y1, jint x2, jint y2 )
	{
		DBG( awt_gra, ("drawLine: %x, %d,%d - %d,%d\n", gr, x1,y1, x2,y2));
		gr->drawLine( x1, y1, x2, y2);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawArc ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jint startAngle, jint sweepAngle )
	{
		gr->drawArc( x, y, width, height, startAngle, sweepAngle, false);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFillArc ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jint startAngle, jint sweepAngle )
	{
		gr->drawArc( x, y, width, height, startAngle, sweepAngle, true);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawOval ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( awt_gra, ("drawOval: %x, %d,%d - %d,%d\n", gr, x,y,width,height));
		gr->drawArc( x, y, width, height, 0, 360, false);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFillOval ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( awt_gra, ("fillOval: %x, %d,%d - %d,%d\n", gr, x,y,width,height));
		gr->drawArc( x, y, width, height, 0, 360, true);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray xPoints, jarray yPoints, jint nPoints )
	{
		int n;
		POINT *p;
		
		DBG( awt_gra, ("drawPolygon: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));
		
		if ( !xPoints || !yPoints ) return;
		
		nPoints = jarray2Points( &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);
		
		n = nPoints-1;
		if ( (p[0].x != p[n].x) || (p[0].y != p[n].y) ) {
			p[nPoints].x = p[0].x;
			p[nPoints].y = p[0].y;
			nPoints++;
		}
		
		BEGIN_PAINT(gr);
		Polyline( gr->dc, p, nPoints);
		END_PAINT(gr);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawPolyline ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray xPoints, jarray yPoints, jint nPoints )
	{
		POINT *p;
		DBG( awt_gra, ("drawPolyline: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));
		
		if ( !xPoints || !yPoints ) return;
		
		nPoints = jarray2Points( &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);
		
		BEGIN_PAINT(gr);
		Polyline( gr->dc, p, nPoints);
		END_PAINT(gr);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFillPolygon ( JNIEnv* env, jclass clazz, Graphics* gr,
		jarray xPoints, jarray yPoints, jint nPoints )
	{
		int n;
		POINT *p;
		
		DBG( awt_gra, ("fillPolygon: %x, %x,%x  %d\n", gr, xPoints, yPoints, nPoints));
		
		if ( !xPoints || !yPoints ) return;
		
		nPoints = jarray2Points( &p, gr->x0, gr->y0, xPoints, yPoints, nPoints);
		
		n = nPoints-1;
		if ( (p[0].x != p[n].x) || (p[0].y != p[n].y) ) {
			p[nPoints].x = p[0].x;
			p[nPoints].y = p[0].y;
			nPoints++;
		}
		
		BEGIN_PAINT(gr);
		Polygon( gr->dc, p, nPoints);
		END_PAINT(gr);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( awt_gra, ("drawRect: %x, %d,%d - %d,%d\n", gr, x,y,width,height));
		
		if ( (width == 0) && (height == 0) ) {
			gr->drawPel( x, y);
		}
		else if ( (width >= 0) && (height >= 0) ){
			gr->drawRect( x, y, width, height);
		}
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFillRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height )
	{
		DBG( awt_gra, ("fillRect: %x, %d,%d - %d,%d\n", gr, x,y,width,height));
		gr->fillRect( x, y, width, height, (int)gr->fgClr);
	}
	
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height, 
		jint wArc, jint hArc )
	{
		
		DBG( awt_gra, ("drawRoundRect: %x, %d,%d - %d,%d  %d,%d\n", gr,
			x,y,width,height, wArc, hArc));
		
		BEGIN_PAINT(gr);
		SelectObject ( gr->dc, HB_0 );
		RoundRect( gr->dc, x+gr->x0, y+gr->y0, x+width+gr->x0, y+height+gr->y0, wArc, hArc);
		END_PAINT(gr);

		DC_ATTRIB_GR (
			SelectObject ( gr->dc, gr->brush );
		)
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFillRoundRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height, 
		jint wArc, jint hArc )
	{
		
		DBG( awt_gra, ("fillRoundRect: %x, %d,%d - %d,%d  %d,%d\n", gr,
			x,y,width,height, wArc, hArc));
		
		BEGIN_PAINT(gr);
		SelectObject ( gr->dc, HP_0 );
		RoundRect( gr->dc, x+gr->x0, y+gr->y0, x+width+gr->x0, y+height+gr->y0, wArc, hArc);
		END_PAINT(gr);

		DC_ATTRIB_GR (
			SelectObject ( gr->dc, gr->pen );
		)

	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDraw3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jboolean raised, jint javargb )
	{
		int     bright, dark;
		HPEN	pb, pd;
		int     xw, yh;
		
		DBG( awt_gra, ("draw3DRect: %x, %d,%d - %d,%d  %d %x\n", gr,
			x,y,width,height, raised, javargb));
		
		/* we should use the same mechanism like ordinary Graphics brighter()/darker() here */
		dark   = (int)ColorJ2N((int)Java_java_awt_Toolkit_clrDark( env, clazz, javargb));
		bright = (int)ColorJ2N((int)Java_java_awt_Toolkit_clrBright( env, clazz, javargb));
		pd = CreatePen( PS_SOLID, 0, dark);
		pb = CreatePen( PS_SOLID, 0, bright);
		
		xw = x + width;
		yh = y + height;
		
		SelectObject( gr->dc, raised ? pb : pd);
		gr->drawLine( x, y, xw-1, y);
		gr->drawLine( x, y, x, yh);
		
		SelectObject( gr->dc, raised ? pd : pb);
		gr->drawLine( x+1, yh, xw, yh);
		gr->drawLine( xw, y, xw, yh);
		
		DeleteObject( pb);
		DeleteObject( pd);

		DC_ATTRIB_GR (
			SelectObject( gr->dc, gr->pen);
		)
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graFill3DRect ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint x, jint y, jint width, jint height,
		jboolean raised, jint javargb )
	{
		
		DBG( awt_gra, ("fill3DRect: %x, %d,%d - %d,%d  %d %x\n", gr,
			x,y,width,height, raised, javargb));	
		gr->fillRect( x+1, y+1, width-2, height-2, ColorJ2N(javargb));
		Java_java_awt_Toolkit_graDraw3DRect( env, clazz, gr, x, y, width-1, height-1, raised, javargb);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetClip ( JNIEnv* env, jclass clazz, Graphics* gr,
		jint xClip, jint yClip, jint wClip, jint hClip )
	{
		DBG( awt_gra, ("setClip: %x, %d,%d - %d,%d\n", gr, xClip, yClip, wClip, hClip));
		gr->setClip( xClip, yClip, wClip, hClip);	  
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
	{
		DBG( awt_gra, ("setColor: %x, %x\n", gr, clr));
		gr->setForeground( clr);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetBackColor ( JNIEnv* env, jclass clazz, Graphics* gr, jint clr )
	{
		DBG( awt_gra, ("setBackColor: %x, %x\n", gr, clr));
		gr->setBackground( clr);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetFont ( JNIEnv* env, jclass clazz, Graphics* gr, Font* fnt )
	{
		DBG( awt_gra, ("setFont: %x, %x\n", gr, fnt));
		gr->setFont( fnt);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetOffset ( JNIEnv* env, jclass clazz, Graphics* gr, jint xOff, jint yOff )
	{
		DBG( awt_gra, ("setOffset: %x, %d,%d\n", gr, xOff, yOff));
		gr->translate( xOff, yOff);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetPaintMode ( JNIEnv* env, jclass clazz, Graphics* gr )
	{
		DBG( awt_gra, ("setPaintMode: %x\n", gr));
		gr->setPaintMode();		
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetXORMode ( JNIEnv* env, jclass clazz, Graphics* gr, jint xorClr )
	{
		DBG( awt_gra, ("setXORMode: %x, %x\n", gr, xorClr));
		gr->setXORMode( xorClr);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graSetVisible ( JNIEnv* env, jclass clazz, Graphics* gr, jint isVisible )
	{
		DBG( awt_gra, ("setVisble: %x  %d\n", gr, isVisible));
		
		/*
		* This is rather a hack to "defuse" a Graphics object, but we don't want to
		* add checks on every draw op to test if the target is invisible
		*/
		
		if ( !isVisible ) {
			if ( gr->x0 >= -10000 ) 
				gr->x0 -= 100000;
		}
		else {
			if ( gr->x0 <= -10000 )
				gr->x0 += 100000;
		}
	}
	
	
	/************************************************************************************
	* image rendering
	*/
	
	void __cdecl
		drawAlphaImage ( Graphics* gr, Image* img,
		jint srcX, jint srcY, jint dstX, jint dstY,
		jint width, jint height, jint bgval )
	{
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawImage ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
		jint srcX, jint srcY,
		jint dstX, jint dstY,
		jint width, jint height, jint bgval )
	{
		if ( img ) {
			BEGIN_PAINT(gr);
			if ( img->mask) {
				// gr->fillRect( dstX, dstY, width, height, bgval);
				if ( X->osi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ) {
					BitBlt( gr->dc, dstX+gr->x0, dstY+gr->y0, width, height, img->dcMask, srcX, srcY, SRCAND);
					BitBlt( gr->dc, dstX+gr->x0, dstY+gr->y0, width, height, img->dc, srcX, srcY, SRCINVERT);
				}
				else {
					MaskBlt( gr->dc, dstX+gr->x0, dstY+gr->y0, width, height, img->dc, srcX, srcY,
						img->mask, 0, 0, MAKEROP4( SRCPAINT, SRCCOPY));
				}
			}
			else {
				BitBlt( gr->dc, dstX+gr->x0, dstY+gr->y0, width, height, img->dc, srcX, srcY, SRCCOPY);
			}
			END_PAINT(gr);
		}
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_graDrawImageScaled ( JNIEnv* env, jclass clazz, Graphics* gr, Image* img,
		jint dx0, jint dy0, jint dx1, jint dy1,
		jint sx0, jint sy0, jint sx1, jint sy1, jint bgval )
	{
		Image* tgt;
		int iw, x0, y0, x1, y1;
		
		iw = img->width;
		
		if ( dx1 > dx0 ) {
			x0 = dx0; x1 = dx1;
		}
		else {
			x0 = dx1; x1 = dx0;
		}
		if ( dy1 > dy0 ) {
			y0 = dy0; y1 = dy1;
		}
		else {
			y0 = dy1; y1 = dy0;
		}
		
		if ( sx0 < 0 ) sx0 = 0;
		if ( sx1 < 0 ) sx1 = 0;
		if ( sx0 >= iw ) sx0 = iw-1;
		if ( sx1 >= iw ) sx1 = iw-1;
		
		tgt = createImage( (x1-x0+1), (y1-y0+1) );
		StretchBlt( tgt->dc, 0, 0, tgt->width, tgt->height,
			img->dc, sx0, sy0, sx1-sx0, sy1-sy0, SRCCOPY);
		
		if ( img->mask) {
			createMaskImage( X, tgt);
			StretchBlt( tgt->dcMask, 0, 0, tgt->width, tgt->height,
				img->dcMask, sx0, sy0, sx1-sx0, sy1-sy0, SRCCOPY);
		}
		
		
		Java_java_awt_Toolkit_graDrawImage ( env, clazz, gr, tgt, 0, 0,
									   x0, y0, tgt->width, tgt->height, bgval);
		freeImage( X, tgt);		
	}
}
