/**
* wnd.cpp - native toplevel window related functions
*
* Copyright (c) 1999
*      Transvirtual Technologies, Inc.  All rights reserved.
*
* See the file "license.terms" for information on usage and redistribution 
* of this file. 
*/

#include "toolkit.hpp"
#if defined(UNDER_CE)
#include <commctrl.h>
extern "C" HINSTANCE appInstance;
#endif

//called by tlkInit
void initWndInsets() {
	DecoInset* pin;
	RECT rt = { 0,0,200,200 };

	AdjustWindowRectEx( &rt, WS_OVERLAPPEDWINDOW, false, 0);
	pin = &(X->frameInsets);
	pin->left = -rt.left;
	pin->top = -rt.top;
	pin->right = rt.right-200;
	pin->bottom = rt.bottom-200;
	pin->guess = false;

	pin = &(X->dialogInsets);
	pin->left = -rt.left;
	pin->top = -rt.top;
	pin->right = rt.right-200;
	pin->bottom = rt.bottom-200;
	pin->guess = false;
}

class KWindow:public KWnd {
public:
	bool forceResize;

	KWindow( HWND parent, TCHAR* title, int style, int x, int y, int width, int height) {
		DecoInset* pin;
		RECT rect;

		switch ( style ) {
		case WS_OVERLAPPEDWINDOW:
			pin = &(X->frameInsets);
			break;
		default:
			pin = 0;
		}

		if ( pin ) {
			insets.left = pin->left;
			insets.top = pin->top;
			insets.right = pin->right;
			insets.bottom = pin->bottom;
		}

		style |= WS_CLIPCHILDREN;
		forceResize = false;

#if defined(UNDER_CE)
		hbar = NULL;

		// HACK: CE cannot create Menubar for 0 sized frames
		// default lange enough (this is gonna be java resized anyway )
		if ( width <= 0 )	width = 100;
		if ( height <= 0 )	height = 50;
#endif

		hwnd = CreateWindow( TEXT("window"), title, style,
					x, y, width, height,
					parent, (HMENU)0, 0, 0);

		/* hack to update java bounds on window manager corrections
		* ( e.g. too small ), because POSCHANGE flags do not
		* indicate moving or sizing here
		*/
		GetWindowRect( hwnd, &rect);
		if ( (rect.left != x) || (rect.top != y) || (rect.right != x+width) || (rect.bottom != y+height) ) {
			forceResize = true;
		}
		getDefWndProc( true);
	
	}

	bool KWindow::OnPaint( int x, int y, int width, int height){
		PAINTSTRUCT ps;
		HDC hdc;
				
		//map to frame extends
		width += insets.left + insets.right;
		height += insets.bottom + insets.top;

		hdc = BeginPaint( hwnd, &ps);

		if ( ps.fErase ) {
			if ( X->osi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ) {
				// reset clip region (win95 workaround )
				SelectClipRgn( hdc, NULL );
				ps.rcPaint.left = x;
				ps.rcPaint.top = y;
				ps.rcPaint.right = x+width;
				ps.rcPaint.bottom = y+height;
			}
			FillRect( hdc, &ps.rcPaint, brush);
		}
		
		EndPaint( hwnd, &ps);

		JniEnv->CallStaticObjectMethod( PaintEvent, getPaintEvent,
			srcIdx, ps.fErase ? UPDATE : PAINT, x, y, width, height );

		return true;
	}

	bool KWindow::OnActivate( int flags, int min, HWND handle ) {
		//handled by java		
		return false;
	}

	bool KWindow::OnCommand( int cmd) {
		HMENU hmenu = GetMenu( hwnd);
#if defined(UNDER_CE)
		if ( !hmenu && hbar )
			hmenu = CommandBar_GetMenu( hbar, 0);
#endif
		if ( hmenu )
			processMenuAction( hmenu, cmd);
		return false;
	}

	bool KWindow::OnPosChanged( int x, int y, int width, int height, int flags) {
		bool sized = (flags & SWP_NOSIZE) == 0;
		bool moved = (flags & SWP_NOMOVE) == 0;
#if !defined(UNDER_CE)
		bool update = (flags & SWP_NOREDRAW) == 0;
#else
		bool update = true;
#endif

// kprintf( "%x %d,%d %dx%d (%x %d %d %d)\n", hwnd, x, y, width, height, flags, sized, moved, update );
		if ( forceResize || (update && (sized || moved)) ) {
			JniEnv->CallStaticObjectMethod( ComponentEvent, getComponentEvent,
				srcIdx, sized ? COMPONENT_RESIZED : COMPONENT_MOVED,
				x, y, width, height);
		}

		forceResize = false;
		return true;
	}
	
	
	bool KWindow::OnClose() {
		JniEnv->CallStaticObjectMethod( WindowEvent, getWindowEvent,
			srcIdx, WINDOW_CLOSING);
		return true;
	}
		
	bool KWindow::OnSetCursor() {
		SetCursor( cursor);				
		return true;
	}

	void KWindow::setBounds(int x, int y, int width, int height) {
		SetWindowPos( hwnd, HWND_TOP, x, y, width, height, SWP_NOZORDER ); //| SWP_NOACTIVATE);
	}

	void KWindow::setVisible( jboolean b) {
		ShowWindow( hwnd, b ? SW_SHOW : SW_HIDE);
		if ( b ) {
			SetActiveWindow( hwnd);
		}
	}

	bool KWindow::updateInsets() {
	/*
	* called by Menu assign to detect appropriate
	* client insets
		*/
		RECT rcl, rclc;
		
		GetClientRect( hwnd, &rclc);
		
		rcl.left = rclc.left;
		rcl.top  = rclc.top;
		rcl.right = rclc.right;
		rcl.bottom = rclc.bottom;

#if !defined(UNDER_CE)
		AdjustWindowRectEx( &rclc, GetWindowLong( hwnd, GWL_STYLE), (GetMenu( hwnd) != NULL), 0);
#else
		AdjustWindowRectEx( &rclc, GetWindowLong( hwnd, GWL_STYLE), 0, 0);
//		if ( hbar )
//			rclc.top += CommandBar_Height( hbar);
#endif

		insets.left = rcl.left-rclc.left;
		insets.top = rcl.top-rclc.top,
		insets.right = rclc.right-rcl.right,
		insets.bottom = rclc.bottom-rcl.bottom;
		
		return true;
	}
	
};

/********************
* exported functions
*********************/

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

extern "C" {
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetTitle ( JNIEnv* env, jclass clazz, KWindow* wnd, jstring jTitle )
	{
		wnd->setText( jTitle);
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetResizable ( JNIEnv* env, jclass clazz, KWindow* wnd, jboolean isResizable,
		int x, int y, int width, int height )
	{
		int style;
		
		DBG( awt_wnd, ("setResizable: %x, %d, %d,%d,%d,%d\n", wnd, isResizable, x,y,width,height));
		
		style = GetWindowLong( wnd->hwnd, GWL_STYLE);
		
		if ( isResizable) {
			style &= ~WS_BORDER;
			style |= WS_THICKFRAME;
		}
		else {
			style |= WS_BORDER;
			style &= ~WS_THICKFRAME;
		}
		
		SetWindowLong( wnd->hwnd, GWL_STYLE, style);
	}
	
	JNIEXPORT void* JNICALL
		Java_java_awt_Toolkit_wndCreateFrame ( JNIEnv* env, jclass clazz, jstring jTitle,
		jint x, jint y, jint width, jint height,
		jint jCursor, jint clrBack, jboolean isResizable )
	{		
		DBG( awt_wnd, ("createFrame( %p, %d,%d,%d,%d,..)\n", jTitle,x,y,width,height));
		
		KWindow* wnd;
		TCHAR* title;

		title = java2WinString( env, X, jTitle);
		wnd = new KWindow( NULL, title, WS_OVERLAPPEDWINDOW, x, y, width, height);
		DBG( awt_wnd, (" -> %x\n", wnd->hwnd));
		
		if ( !isResizable ) {
			Java_java_awt_Toolkit_wndSetResizable( env, clazz,
				wnd, false, x, y, width, height);			
		}
		
		SetActiveWindow( wnd->hwnd);
		return wnd;
		
	}
	
	
	JNIEXPORT void* JNICALL
		Java_java_awt_Toolkit_wndCreateWindow ( JNIEnv* env, jclass clazz, KWindow* owner,
		jint x, jint y, jint width, jint height,
		jint jCursor, jint clrBack )
	{
		KWindow* wnd;
		
		DBG( awt_wnd, ("createWindow( %p, %d,%d,%d,%d,..)\n", owner,x,y,width,height));
		
		wnd = new KWindow( owner ? owner->hwnd : NULL, TEXT(""), WS_POPUP, x, y, width, height);
		
		DBG( awt_wnd, (" -> %x\n", wnd->hwnd));
		
		return wnd;
	}
	
	JNIEXPORT void* JNICALL
		Java_java_awt_Toolkit_wndCreateDialog ( JNIEnv* env, jclass clazz, KWindow* owner, jstring jTitle,
		jint x, jint y, jint width, jint height,
		jint jCursor, jint clrBack, jboolean isResizable )
	{
		DBG( awt_wnd, ("createDialog( %p,%p, %d,%d,%d,%d,..)\n", owner,jTitle,x,y,width,height));
		
		KWindow* wnd;
		TCHAR* title;
				
		title = java2WinString( env, X, jTitle);
		wnd = new KWindow( owner->hwnd, title, WS_OVERLAPPEDWINDOW, x, y, width, height);
		
		DBG( awt_wnd, (" -> %x\n", wnd->hwnd));
				
		if ( !isResizable ) {			
			Java_java_awt_Toolkit_wndSetResizable( env, clazz,				
				wnd, false, x, y, width, height);						
		}
				
		SetActiveWindow( wnd->hwnd);
		return wnd;
		
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndDestroyWindow ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		wnd->destroy();
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndRequestFocus ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( awt_wnd, ("request focus: %x\n", wnd));
		wnd->setFocus();
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetFrameInsets ( JNIEnv* env, jclass clazz,
		jint top, jint left, jint bottom, jint right )
	{
		// we already know the exact insets
		jclass wndCls = JniEnv->FindClass( "java.awt.Frame");
		jmethodID setDecoInsets = JniEnv->GetStaticMethodID( wndCls, "setDecoInsets", "(IIIII)V");
		DecoInset *pin = &(X->frameInsets);
		JniEnv->CallStaticVoidMethod( wndCls, setDecoInsets, pin->top, pin->left, 
									pin->bottom, pin->right, -1);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetDialogInsets ( JNIEnv* env, jclass clazz,
		jint top, jint left, jint bottom, jint right )
	{
		// we already know the exact insets
		jclass wndCls = JniEnv->FindClass( "java.awt.Dialog");
		jmethodID setDecoInsets = JniEnv->GetStaticMethodID( wndCls, "setDecoInsets", "(IIIII)V");
		DecoInset *pin = &(X->dialogInsets);
		JniEnv->CallStaticVoidMethod( wndCls, setDecoInsets, pin->top, pin->left, 
									pin->bottom, pin->right, -1);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetBounds ( JNIEnv* env, jclass clazz, KWindow* wnd,
		jint x, jint y, jint width, jint height,
		jboolean isResizable )
	{		
		DBG( awt_wnd, ("setBounds: %x %d,%d,%d,%d\n", wnd, x, y, width, height));
				
		if ( width < 0 )  width = 1;		
		if ( height < 0 ) height = 1;
		
		wnd->setBounds( x, y, width, height);
		
		if ( !isResizable ) {
			Java_java_awt_Toolkit_wndSetResizable( env, clazz,
				wnd, false, x, y, width, height);
		}
		
	}
	
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndRepaint ( JNIEnv* env, jclass clazz, KWindow* wnd,
		jint x, jint y, jint width, jint height )
	{
		DBG( awt_wnd, ("wndRepaint: %x %d,%d,%d,%d\n", wnd, x, y, width, height));
		
		RECT rcl;
		
		rcl.left = x;
		rcl.top = y;
		rcl.right = x + width;
		rcl.bottom = y + height;
		
		InvalidateRect( wnd->hwnd, &rcl, true);
		UpdateWindow( wnd->hwnd);
	}
	
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetIcon ( JNIEnv* env, jclass clazz, KWindow* wnd, void* img )
	{
	}
	
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetVisible ( JNIEnv* env, jclass clazz, KWindow* wnd, jboolean showIt )
	{
		DBG( awt_wnd, ("setVisible: %x %d\n", wnd, showIt));
		wnd->setVisible( showIt);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndToBack ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( awt_wnd, ("toBack: %x\n", wnd));
		SetWindowPos( wnd->hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndToFront ( JNIEnv* env, jclass clazz, KWindow* wnd )
	{
		DBG( awt_wnd, ("toFront: %x\n", wnd));
		SetWindowPos( wnd->hwnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	
	JNIEXPORT void JNICALL
		Java_java_awt_Toolkit_wndSetCursor ( JNIEnv* env, jclass clazz, KWindow* wnd, jint jCursor )
	{
		DBG( awt_wnd, ("setCursor: %x, %d\n", wnd, jCursor));
		wnd->setCursor( getCursor( jCursor));
	}
	
}
