/*
 * kaffe.comm.SerialPort.c
 *
 * Copyright (c) 1996 - 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include <jni.h>
#include <termios.h>
#include "jsyscall.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static jfieldID filedesc;
static jfieldID devname;
static jclass ioexception;

JNIEXPORT void JNICALL
Java_kaffe_comm_SerialPort_init(JNIEnv* env, jclass serial)
{
	filedesc = (*env)->GetFieldID(env, serial, "filedesc", "I");
	devname = (*env)->GetFieldID(env, serial, "name", "java.lang.String");
	ioexception = (*env)->FindClass(env, "java.io.IOException");
}

JNIEXPORT void JNICALL
Java_kaffe_comm_SerialPort_open(JNIEnv* env, jobject port)
{
	int fd;
	const char* cname;
	struct termios ctrl;
	int r;
	jstring name;

	name = (*env)->GetObjectField(env, port, devname);
	cname = (*env)->GetStringUTFChars(env, name, 0);
	fd = open(cname, JOPEN_READWRITE /* O_RDWR */);
	(*env)->ReleaseStringUTFChars(env, name, cname);

	if (fd < 0) {
		(*env)->ThrowNew(env, ioexception, "failed to open comm port");
	}
	else {
		r = tcgetattr(fd, &ctrl);
		if (r < 0) perror("");

#if 0
		ctrl.c_iflag = 0;
		ctrl.c_oflag = 0;
		ctrl.c_cflag = CS8|CREAD|CRTSCTS;
		ctrl.c_lflag = 0;
		cfsetispeed(&ctrl, B9600);
		cfsetospeed(&ctrl, B9600);
#else
		ctrl.c_lflag &= ~(ECHO|ICANON);
#endif

		r = tcsetattr(fd, TCSANOW, &ctrl);
		if (r < 0) perror("");

		(*env)->SetIntField(env, port, filedesc, fd);
	}
}

JNIEXPORT void JNICALL
Java_kaffe_comm_SerialPort_close(JNIEnv* env, jobject port)
{
	int fd;

	fd = (*env)->GetIntField(env, port, filedesc);
	if (fd != -1) {
		close(fd);
		(*env)->SetIntField(env, port, filedesc, -1);
	}
}

JNIEXPORT jint JNICALL
Java_kaffe_comm_SerialPort_readBytes(JNIEnv* env, jobject port, jarray bytes, jint start, jint len)
{
	int fd;
	jbyte* buf;
	jboolean iscopy;
	int alen;

	fd = (*env)->GetIntField(env, port, filedesc);
	buf = (*env)->GetByteArrayElements(env, bytes, &iscopy);
	/* Bounds checking done in java code */
	alen = read(fd, &buf[start], len);
	(*env)->ReleaseByteArrayElements(env, bytes, buf, JNI_COMMIT);
	return (alen);
}

JNIEXPORT jint JNICALL
Java_kaffe_comm_SerialPort_writeBytes(JNIEnv* env, jobject port, jarray bytes, jint start, jint len)
{
	int fd;
	jbyte* buf;
	jboolean iscopy;
	int alen;

	fd = (*env)->GetIntField(env, port, filedesc);
	buf = (*env)->GetByteArrayElements(env, bytes, &iscopy);
	/* Bounds checking done in java code */
	alen = write(fd, &buf[start], len);
	(*env)->ReleaseByteArrayElements(env, bytes, buf, JNI_ABORT);
	return (alen);
}
