/*
 * kaffe.comm.SoundPort.c
 *
 * Copyright (c) 1996 - 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include <jni.h>
#include "jsyscall.h"

#include <sys/soundcard.h>

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static jfieldID filedesc;
static jfieldID devname;
static jclass ioexception;

JNIEXPORT void JNICALL
Java_kaffe_comm_SoundPort_init(JNIEnv* env, jclass serial)
{
	filedesc = (*env)->GetFieldID(env, serial, "filedesc", "I");
	devname = (*env)->GetFieldID(env, serial, "name", "java.lang.String");
	ioexception = (*env)->FindClass(env, "java.io.IOException");
}

JNIEXPORT void JNICALL
Java_kaffe_comm_SoundPort_open(JNIEnv* env, jobject port)
{
	int fd;
	const char* cname;
	int temp;
	jstring name;

	name = (*env)->GetObjectField(env, port, devname);
	cname = (*env)->GetStringUTFChars(env, name, 0);
	fd = open(cname, JOPEN_READWRITE /* O_RDWR */);
	(*env)->ReleaseStringUTFChars(env, name, cname);

	if (fd < 0) {
		(*env)->ThrowNew(env, ioexception, "failed to open sound port");
	}
	else {
		/* Init the sound system */
		ioctl(fd, SNDCTL_DSP_RESET);
		temp = AFMT_S16_NE;
		ioctl(fd, SNDCTL_DSP_SETFMT, &temp);
		temp = 1;
		ioctl(fd, SNDCTL_DSP_STEREO, &temp);
		temp = 44100;
		ioctl(fd, SNDCTL_DSP_SPEED, &temp);

		(*env)->SetIntField(env, port, filedesc, fd);
	}
}

JNIEXPORT void JNICALL
Java_kaffe_comm_SoundPort_close(JNIEnv* env, jobject port)
{
	int fd;

	fd = (*env)->GetIntField(env, port, filedesc);
	if (fd != -1) {
		close(fd);
		(*env)->SetIntField(env, port, filedesc, -1);
	}
}

JNIEXPORT jint JNICALL
Java_kaffe_comm_SoundPort_writeBytes(JNIEnv* env, jobject port, jarray bytes, jint start, jint len)
{
	int fd;
	jbyte* buf;
	jboolean iscopy;
	int alen;

	fd = (*env)->GetIntField(env, port, filedesc);
	buf = (*env)->GetByteArrayElements(env, bytes, &iscopy);
	/* Bounds checking done in java code */
	alen = write(fd, &buf[start], len);
	(*env)->ReleaseByteArrayElements(env, bytes, buf, JNI_ABORT);
	return (alen);
}

JNIEXPORT void JNICALL
Java_kaffe_comm_SoundPort_setChannels(JNIEnv* env, jobject port, jint chans)
{
	int fd;
	fd = (*env)->GetIntField(env, port, filedesc);
	ioctl(fd, SNDCTL_DSP_CHANNELS, &chans);
}

JNIEXPORT void JNICALL
Java_kaffe_comm_SoundPort_setSpeed(JNIEnv* env, jobject port, jint speed)
{
	int fd;
	fd = (*env)->GetIntField(env, port, filedesc);
	ioctl(fd, SNDCTL_DSP_SPEED, &speed);
}
