/* 
 * Interfaces to libcanna translation libraries.
 *
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * $tvt: CharTables.c,v 1.3 2001/07/23 23:03:12 samc Exp $
 */
#include <jni.h>

#include "codetables/jis0208.h"
#include "codetables/jis0201.h"
#include "codetables/jis0212.h"

JNIEXPORT jint JNICALL
Java_kaffe_io_CharTables_JIS0208ToChar (JNIEnv* env, jclass cls, jint index)
{
  if (index >= 0x1e80 || index < 0) return -1;
  return __jis0208_to_ucs[index];
}

JNIEXPORT jint JNICALL
Java_kaffe_io_CharTables_JIS0201ToChar (JNIEnv* env, jclass cls, jint index)
{
  if (index >= 256 || index < 0) return -1;
  return jis0201_to_ucs[index];
}

JNIEXPORT jint JNICALL
Java_kaffe_io_CharTables_JIS0212ToChar (JNIEnv* env, jclass cls, jint index)
{
  uint32_t rtn = 0;
  const struct jisx0212_idx *rp = __jisx0212_to_ucs_idx;

  if (index >= 256 || index < 0) return -1;

  while (index > rp->end)
    {
      rp++;
    }

  if (index >= rp->start)
    {
      rtn = __jisx0212_to_ucs[rp->idx + index - rp->start];
    }

  if (rtn == 0) rtn = -1;
  
  return rtn;
}

JNIEXPORT jbyteArray JNICALL
Java_kaffe_io_CharTables_CharToJIS0208 (JNIEnv* env, jclass cls, jint ch)
{
  jbyteArray rtn = 0;
  const char * cp;

  if (ch < 0x100)
    {
      cp = __jisx0208_from_ucs4_lat1[ch];
    }
  else if (ch >= 0x391 && ch <= 0x451)
    {
      cp = __jisx0208_from_ucs4_greek[ch - 0x391];
    }
  else
    {
      const struct jisx0208_ucs_idx *rp = __jisx0208_from_ucs_idx;
      if (ch >= 0xffff)
        {
          return 0;
        }
      while (ch > rp->end)
        {
          ++rp;
        }
      if (ch >= rp->start)
        {
          cp = __jisx0208_from_ucs_tab[rp->idx + ch - rp->start];
        }
      else
        {
          return 0;
        }
    }

  if (cp[0] == '\0')
    {
      return 0;
    }

  rtn = (*env)->NewByteArray (env, 2);
  (*env)->SetByteArrayRegion(env, rtn, 0, 2, (jbyte *)cp);

  return rtn;
}

JNIEXPORT jbyteArray JNICALL
Java_kaffe_io_CharTables_CharToJIS0212 (JNIEnv* env, jclass cls, jint ch)
{
  jbyteArray rtn = 0;
  const struct jisx0212_idx *rp = __jisx0212_from_ucs_idx;
  const char *cp;

  if (ch >= 0xffff) { return 0; }

  while (ch > rp->end) { ++rp; }

  if (ch >= rp->start)
    {
      cp = __jisx0212_from_ucs[rp->idx + ch - rp->start];
    }
  else { return 0; }

  if (cp[0] == '\0') { return 0; }

  rtn = (*env)->NewByteArray (env, 2);
  (*env)->SetByteArrayRegion(env, rtn, 0, 2, (jbyte *)cp);

  return rtn;
}
