/*
 * kaffe.io.FileIO.c
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "files.h"
#include "jsyscall.h"
#include "../../../kaffe/kaffevm/support.h"
#include "jni.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static jclass ioexception;

JNIEXPORT void JNICALL
Java_kaffe_io_FileIO_init0(JNIEnv* env, jclass fileio)
{
	ioexception = (*env)->FindClass(env, "java.io.IOException");
}

/*
 * Open a file.
 */
JNIEXPORT jint JNICALL
Java_kaffe_io_FileIO_open0(JNIEnv* env, jobject fileio, jstring filename, jint mode)
{
	jbyte* str;
	int fd;

	str = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
	fd = open(str, mode);
	(*env)->ReleaseStringUTFChars(env, filename, str);
	return (fd);
}

/*
 * Close a file.
 */
JNIEXPORT void JNICALL
Java_kaffe_io_FileIO_close0(JNIEnv* env, jobject fileio, jint fd)
{
	close(fd);
}

/*
 * Read bytes from a file.
 */
JNIEXPORT jint JNICALL
Java_kaffe_io_FileIO_read0(JNIEnv* env, jobject fileio, jint fd, jarray buffer, jint offset, jint length)
{
	char* buf;
	int s;

	buf = (*env)->GetByteArrayElements(env, buffer, 0);
	s = read(fd, buf+offset, length);
	(*env)->ReleaseByteArrayElements(env, buffer, buf, JNI_COMMIT);

	if (s < 0) {
		(*env)->ThrowNew(env, ioexception, SYS_ERROR);
	}
	return (s);
}

/*
 * Write bytes to a file.
 */
JNIEXPORT jint JNICALL
Java_kaffe_io_FileIO_write0(JNIEnv* env, jobject fileio, jint fd, jarray buffer, jint offset, jint length)
{
	char* buf;
	int s;

	buf = (*env)->GetByteArrayElements(env, buffer, 0);
	s = write(fd, buf+offset, length);
	(*env)->ReleaseByteArrayElements(env, buffer, buf, JNI_ABORT);

	if (s < 0) {
		(*env)->ThrowNew(env, ioexception, SYS_ERROR);
	}
	return (s);
}

/*
 * Seek on a file
 */
JNIEXPORT jlong JNICALL
Java_kaffe_io_FileIO_seek0(JNIEnv* env, jobject fileio, jint fd, jlong amount, jint mode)
{
	return ((jlong)lseek(fd, (off_t)amount, mode));
}

/*
 * How many bytes can we read without blocking?
 */
JNIEXPORT jint JNICALL
Java_kaffe_io_FileIO_available0(JNIEnv* env, jobject fileio, jint fd)
{
	return (fileinfoFD(fd, JINFO_AVAILABLE, 0));
}

/*
 * How long?
 */
JNIEXPORT jlong JNICALL
Java_kaffe_io_FileIO_length0(JNIEnv* env, jobject fileio, jint fd)
{
	jlong len;
	if (fileinfoFD(fd, JINFO_LENGTH, &len)) {
		return (len);
	}
	else {
		return (-1);
	}
}


/*
 * USE THIS WITH CARE - not all systems will support it, and the flag values
 * will vary. Used in a wrong way, it leads to OS dependend Java code and
 * even can lock Kaffe.
 * This is a workaround for the 'fifo write interrupt' problem, where the writer gets
 * interrupted before it has written out the whole buffer, ending up in a lot
 * of incomplete writes (RTSP streaming + jthreads -> non-blocked,async IO)
 */
JNIEXPORT jint JNICALL
Java_kaffe_io_FileIO_setMode0(JNIEnv* env, jobject fileio, jint fd, jint mode)
{
#ifdef HAVE_FCNTL
	return (fcntl(fd,F_SETFL,mode));
#endif
	
	return 0;
}
