/*
 * kaffe.io.FileName.c
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: FileName.c,v 1.5 2001/10/08 18:51:13 samc Exp $
 */

#include <jni.h>

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "files.h"
#include "jsyscall.h"
#include "../../../kaffe/kaffevm/support.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

JNIEXPORT jboolean JNICALL
Java_kaffe_io_FileName_mkdir0(JNIEnv* env, jobject fileio, jstring dirname)
{
	jbyte* str;
	int r;

	str = (jbyte*)(*env)->GetStringUTFChars(env, dirname, 0);
	r = mkdir(str);
	(*env)->ReleaseStringUTFChars(env, dirname, str);

	return (r < 0 ? JNI_FALSE : JNI_TRUE);
}

JNIEXPORT jboolean JNICALL
Java_kaffe_io_FileName_rmdir0(JNIEnv* env, jobject fileio, jstring dirname)
{
	jbyte* str;
	int r;

	str = (jbyte*)(*env)->GetStringUTFChars(env, dirname, 0);
	r = rmdir(str);
	(*env)->ReleaseStringUTFChars(env, dirname, str);

	return (r < 0 ? JNI_FALSE : JNI_TRUE);
}

JNIEXPORT jboolean JNICALL
Java_kaffe_io_FileName_delete0(JNIEnv* env, jobject fileio, jstring filename)
{
	jbyte* str;
	int r;

	str = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
	r = remove(str);
	(*env)->ReleaseStringUTFChars(env, filename, str);

	return (r < 0 ? JNI_FALSE : JNI_TRUE);
}

JNIEXPORT jboolean JNICALL
Java_kaffe_io_FileName_rename0(JNIEnv* env, jobject fileio, jstring from, jstring to)
{
	jbyte* fstr;
	jbyte* tstr;
	int r;

	fstr = (jbyte*)(*env)->GetStringUTFChars(env, from, 0);
	tstr = (jbyte*)(*env)->GetStringUTFChars(env, to, 0);

	r = rename(from, to);

	(*env)->ReleaseStringUTFChars(env, from, fstr);
	(*env)->ReleaseStringUTFChars(env, to, tstr);

	return (r < 0 ? JNI_FALSE : JNI_TRUE);
}

JNIEXPORT jlong JNICALL
Java_kaffe_io_FileName_lastModified0(JNIEnv* env, jobject fileio, jstring filename)
{
	jbyte* str;
	jlong tm;
	int r;

	str = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
	r = fileinfo(str, JINFO_LASTMODIFIED, &tm);
	(*env)->ReleaseStringUTFChars(env, filename, str);

	if (r == 1) {
		return (tm);
	}
	else {
		return (0);
	}
}

JNIEXPORT jlong JNICALL
Java_kaffe_io_FileName_length0(JNIEnv* env, jobject fileio, jstring filename)
{
	jbyte* str;
	jlong tm;
	int r;

	str = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
	r = fileinfo(str, JINFO_LENGTH, &tm);
	(*env)->ReleaseStringUTFChars(env, filename, str);

	if (r == 1) {
		return (tm);
	}
	else {
		return (0);
	}
}

JNIEXPORT jboolean JNICALL
Java_kaffe_io_FileName_checkMode0(JNIEnv* env, jobject fileio, jstring filename, jint mode)
{
	jbyte* str;
	jboolean r;

	str = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
	r = fileinfo(str, mode, 0);
	(*env)->ReleaseStringUTFChars(env, filename, str);

	return (r);
}

JNIEXPORT jarray JNICALL /* [Ljava.lang.String; */
Java_kaffe_io_FileName_readdir0(JNIEnv* env, jobject fileio, jstring filename)
{
	jbyte* path;
	jdirEntry* dirlist;
	jdirEntry* mentry;
	jarray array;
	int count;
	int i;

	path = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
	count = readdir(path, &dirlist);
	(*env)->ReleaseStringUTFChars(env, filename, path);
	if (count < 0) {
		return (0);
	}

	array = (*env)->NewObjectArray(env, count, (*env)->FindClass(env, "java.lang.String"), 0);
	for (i = 0; i < count; i++) {
		mentry = dirlist;
		dirlist = mentry->next;
		(*env)->SetObjectArrayElement(env, array, i, (*env)->NewStringUTF(env, mentry->name));
		__kaffe_free(mentry);
	}

	return (array);
}

JNIEXPORT jboolean JNICALL
Java_kaffe_io_FileName_createNewFile0(JNIEnv* env, jobject fileio, jstring filename)
{
        jbyte* str;
        int r;

        str = (jbyte*)(*env)->GetStringUTFChars(env, filename, 0);
        r = open (str, JOPEN_READWRITE);
        if (r < 0) {
                return JNI_FALSE;
        }
        r = close (r);
	return (r < 0 ? JNI_FALSE : JNI_TRUE);
}
