/*
 * kaffe.media.TVCtrl.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "files.h"
#include "jsyscall.h"
#include "defs.h"
#include "jni.h"
#include "../../../kaffe/kaffevm/support.h"

#undef HAVE_STRUCT_V4L2_WINDOW

#if (defined HAVE_LINUX_VIDEODEV_H)
#include <linux/videodev.h>
#endif

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

//could break device poll implementation
#if defined(select)
#undef select
#endif

typedef struct _tvctlr {
    int handle;
} TVCtrl;

static TVCtrl* curCtrl = 0;

void dumpChannels(TVCtrl* ctrl)
{
#if (defined HAVE_LINUX_VIDEODEV_H)
    struct video_capability vd;
    struct video_channel vc[10];
    int n;
	
    ioctl(ctrl->handle,VIDIOCGCAP,&vd);
    for(n=0; n<vd.channels; n++) {
	vc[n].channel=n;
	if(ioctl(ctrl->handle, VIDIOCGCHAN, &vc[n])<0) {
	    perror("ioctl(VIDIOCGCHAN)");
	    return;
	}
	fprintf(stderr, "vc[%d].channel=%d\n", n, vc[n].channel);
	fprintf(stderr, "vc[%d].name=\"%s\"\n", n, vc[n].name);
	fprintf(stderr, "vc[%d].tuners=%d\n", n, vc[n].tuners);
	fprintf(stderr, "vc[%d].flags=0x%08x", n, vc[n].flags);
	if(vc[n].flags&VIDEO_VC_TUNER) fprintf(stderr, " TUNER");
	if(vc[n].flags&VIDEO_VC_AUDIO) fprintf(stderr, " AUDIO");
	fprintf(stderr, "\n");
	fprintf(stderr, "vc[%d].type=0x%08x", n, vc[n].type);
	if(vc[n].type&VIDEO_TYPE_TV) fprintf(stderr, " TV");
	if(vc[n].type&VIDEO_TYPE_CAMERA) fprintf(stderr, " CAMERA");
	fprintf(stderr, "\n");
	fprintf(stderr, "vc[%d].norm=%d\n", n, vc[n].norm);
	fprintf(stderr, "\n");    
    }
#endif    
}

void mute(TVCtrl* ctrl, int mute) {
#if (defined HAVE_LINUX_VIDEODEV_H)
#if (defined HAVE_STRUCT_V4L2_WINDOW)
#else
    struct video_audio va;

    ioctl(ctrl->handle,VIDIOCGAUDIO,&va);

    if ( mute ) {
	va.flags |= VIDEO_AUDIO_MUTE;
    }
    else {
	va.flags &= ~VIDEO_AUDIO_MUTE;
    }
   
    ioctl(ctrl->handle,VIDIOCSAUDIO,&va);
#endif
#endif
}


JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_close( JNIEnv* env, jclass clazz, TVCtrl *ctrl ) {
#if (defined HAVE_LINUX_VIDEODEV_H)
    mute( ctrl, 1);
    close( ctrl->handle);
    free( ctrl);
#endif
}

JNIEXPORT void* JNICALL
Java_kaffe_media_TVCtrl_open( JNIEnv *env, jclass clazz, jstring dev)
{
#if (defined HAVE_LINUX_VIDEODEV_H)
    int fd;
    TVCtrl *ctrl;
    const char* str;
     

    if ( system("v4l-conf") ) {
	//      fprintf( stderr, "failed to run v4l-conf framebuffer configuration\n");
	return NULL;
    }
	
    str = (*env)->GetStringUTFChars(env, dev, 0);
    fd = open(str,O_RDWR);
    (*env)->ReleaseStringUTFChars(env, dev, str);

    if ( fd < 0 ){
	//      fprintf( stderr, "failed to open video device: %s\n", str);
	return NULL;
   }
  
    ctrl = calloc( 1, sizeof( TVCtrl));
    ctrl->handle = fd;

    curCtrl = ctrl;
    mute( ctrl, 0);
    
    return ctrl;
#else
    return NULL;
#endif
}

JNIEXPORT jint JNICALL
Java_kaffe_media_TVCtrl_getPictureWidth( JNIEnv* env, jclass clazz, TVCtrl *ctrl )
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    //currently does NOT blocks on producer
    struct v4l2_format fmt;
    ioctl(ctrl->handle, VIDIOC_G_FMT, &fmt);
  
    return fmt.fmt.pix.width;
#else
    struct video_capture cap;
    ioctl(ctrl->handle, VIDIOCGCAPTURE, &cap);
   
    return cap.width;
#endif

#endif
    return -1;
}

JNIEXPORT jint JNICALL
Java_kaffe_media_TVCtrl_getPictureHeight( JNIEnv* env, jclass clazz, TVCtrl *ctrl )
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_format fmt;
    ioctl(ctrl->handle, VIDIOC_G_FMT, &fmt);
    return fmt.fmt.pix.height;
#else
    struct video_capture cap;
    ioctl(ctrl->handle, VIDIOCGCAPTURE, &cap);
    return cap.height;
#endif

#endif
    return -1;
}


JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setTuner ( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jstring name )
{
#if (defined HAVE_LINUX_VIDEODEV_H)
    const char* str;

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_tuner tuner;
    if ( name ) {
	str = (*env)->GetStringUTFChars(env, name, 0);
	strcpy( tuner.name, str);
	(*env)->ReleaseStringUTFChars(env, name, str);
    }
    else {
	*tuner.name = 0;
    }
    ioctl(ctrl->handle, VIDIOC_S_TUNER, &tuner);
#else
    struct video_tuner tuner;
    if ( name ) {
	str = (*env)->GetStringUTFChars(env, name, 0);
	strcpy( tuner.name, str);
	(*env)->ReleaseStringUTFChars(env, name, str);
    }
    else {
	*tuner.name = 0;
    }
    ioctl(ctrl->handle, VIDIOCSTUNER, &tuner);
#endif
     
#endif
}


JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setInput ( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint input, jint norm, jint channels )
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
#else
    struct video_channel chan;
    chan.channel = input;
    ioctl(ctrl->handle, VIDIOCGCHAN, &chan);
    chan.norm = norm; /* 0:PAL, 1:NTSC, 2:SECAM */
    ioctl(ctrl->handle, VIDIOCSCHAN, &chan);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_mute ( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint state )
{
#if (defined HAVE_LINUX_VIDEODEV_H)
    mute( ctrl, state);
#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setPreview( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint state ) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    ioctl(ctrl->handle, VIDIOC_PREVIEW, &state);
#else
    ioctl(ctrl->handle, VIDIOCCAPTURE, &state);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setChromaKey( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint color) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_window win;
    ioctl( ctrl->handle, VIDIOC_G_WIN, &win);
    win.chromakey = color;
    ioctl( ctrl->handle, VIDIOC_S_WIN, &win);
#else
    struct video_window win;
    ioctl( ctrl->handle, VIDIOCGWIN, &win);
    win.chromakey = color;
    ioctl( ctrl->handle, VIDIOCSWIN, &win);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setBounds( JNIEnv* env, jclass clazz, TVCtrl *ctrl,jint x, jint y, jint width, jint height)
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_window win;
    ioctl( ctrl->handle, VIDIOC_G_WIN, &win);
   
    win.x = x;
    win.y = y;
    win.width = width;
    win.height = height;
   
    ioctl( ctrl->handle, VIDIOC_S_WIN, &win);
#else
    struct video_window win;
    ioctl( ctrl->handle, VIDIOCGWIN, &win);

    win.clipcount = 0;
    win.flags = 0;
    win.x = x;
    win.y = y;
    win.height = height;
    win.width = width;
   
    ioctl( ctrl->handle, VIDIOCSWIN, &win);
#endif

#endif
}

JNIEXPORT jint JNICALL
Java_kaffe_media_TVCtrl_getBrightness( JNIEnv* env, jclass clazz, TVCtrl *ctrl ) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)
	
#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_control pict;
    pict.id = V4L2_CID_BRIGHTNESS;
    ioctl( ctrl->handle, VIDIOC_G_CTRL, &pict);
    return pict.value;
#else
    struct video_picture pict;
    ioctl( ctrl->handle, VIDIOCGPICT, &pict);
    return pict.brightness;
#endif
	
#else
    return 0;

#endif   
}

JNIEXPORT jint JNICALL
Java_kaffe_media_TVCtrl_getVolume( JNIEnv* env, jclass clazz, TVCtrl *ctrl) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    return 0;
#else
    struct video_audio va;
    ioctl( ctrl->handle, VIDIOCGAUDIO, &va);
    return va.volume;
#endif

#else
    return 0;

#endif
}

JNIEXPORT jint JNICALL
Java_kaffe_media_TVCtrl_getContrast( JNIEnv* env, jclass clazz, TVCtrl *ctrl ) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)
	
#if (defined HAVE_STRUCT_V4l2_WINDOW)
    struct v4l2_control pict;
    pict.id = V4L2_CID_CONTRAST;
    ioctl( ctrl->handle, VIDIOC_G_CTRL, &pict);
    return pict.value;
#else
    struct video_picture pict;
    ioctl( ctrl->handle, VIDIOCGPICT, &pict);
    return pict.contrast;
#endif

#else
    return 0;

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setBrightness( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint brightness ) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_control pict;
    pict.value = brightness;
    pict.id = V4L2_CID_BRIGHTNESS;
    ioctl( ctrl->handle, VIDIOC_S_CTRL, &pict);
#else
    struct video_picture pict;
    ioctl( ctrl->handle, VIDIOCGPICT, &pict);
    pict.brightness = brightness;
    ioctl( ctrl->handle, VIDIOCSPICT, &pict);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setVolume( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint volume ) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)

#else
    struct video_audio va;
    ioctl( ctrl->handle, VIDIOCGAUDIO, &va);
    va.volume = volume;
    ioctl( ctrl->handle, VIDIOCSAUDIO, &va);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setContrast( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint contrast )
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
    struct v4l2_control pict;
    pict.value = contrast;
    pict.id = V4L2_CID_CONTRAST;
    ioctl( ctrl->handle, VIDIOC_S_CTRL, &pict);
#else
    struct video_picture pict;
    ioctl( ctrl->handle, VIDIOCGPICT, &pict);
    pict.contrast=contrast;
    ioctl( ctrl->handle, VIDIOCSPICT, &pict);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setFrequency( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint freq )
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
#else
    unsigned long f1, f2;
    ioctl(ctrl->handle,VIDIOCGFREQ,&f1);
    f2 = (unsigned long)freq;
    ioctl(ctrl->handle,VIDIOCSFREQ,&f2);
#endif

#endif
}

JNIEXPORT jint JNICALL
Java_kaffe_media_TVCtrl_getFrequency( JNIEnv* env, jclass clazz, TVCtrl *ctrl)
{
#if (defined HAVE_LINUX_VIDEODEV_H)

#if (defined HAVE_STRUCT_V4L2_WINDOW)
#else
    unsigned long uFreq;
    ioctl(ctrl->handle,VIDIOCGFREQ,&uFreq);
    return ((jint)uFreq);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_setRate( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint rate) { 
#if (defined HAVE_LINUX_VIDEODEV_H)
   unsigned long tm = 10000L * rate;
#if (defined HAVE_STRUCT_V4L2_WINDOW)
    ioctl( ctrl->handle, VIDIOC_S_FREQ, &tm);
#else 
    ioctl( ctrl->handle, VIDIOCSFREQ, &tm);
#endif

#endif
}

JNIEXPORT void JNICALL
Java_kaffe_media_TVCtrl_nextFrame( JNIEnv* env, jclass clazz, TVCtrl *ctrl, jint timeout) 
{
#if (defined HAVE_LINUX_VIDEODEV_H)

    fd_set rfds;
    struct timeval tv;
    int ret;
   
    FD_ZERO(&rfds);
    FD_SET(ctrl->handle, &rfds);
    
    tv.tv_sec = 0;
    tv.tv_usec = 1000 * timeout;
	
    ret = select( ctrl->handle+1, &rfds, NULL, NULL, &tv);

#endif
}

