/* Copyright (c) 1998 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * $Id: VideoCtrl.h,v 1.1 2001/10/19 20:22:19 joerg Exp $ 
 */

#define VSM_ENCODED      0x01
#define VSM_BOUNDS       0x02
#define VSM_SUSPEND_MODE 0x04
#define VSM_CHROMA       0x08
#define VSM_BRIGHTNESS   0x10
#define VSM_CONTRAST     0x20
#define VSM_VOLUME       0x40 

#define UNLOCK_SMEM( smem)    \
  smem->locked = 0;

#define LOCK_SMEM( smem)      \
  while( smem->locked ) {     \
    usleep( 100);             \
  }                           \
  smem->locked = 1;

#define CHECK_MOD_SET( state, smem, flag) \
{                                         \
  LOCK_SMEM( smem);                       \
  state = ((smem->modifiers & flag) != 0);\
  UNLOCK_SMEM( smem);                     \
}

#define SET_MOD( smem, flag) \
{                            \
  LOCK_SMEM( smem);          \
  smem->modifiers |= flag;   \
  UNLOCK_SMEM( smem);        \
}

#define UNSET_MOD( smem, flag) \
{                              \
  LOCK_SMEM( smem);            \
  smem->modifiers &= ~flag;    \
  UNLOCK_SMEM( smem);          \
}

struct video_cliprect {
   unsigned int x,y;
   unsigned int width,height;
   struct video_cliprect *next;
};

typedef struct _video_smem {
   //read-write lock
   unsigned char locked;
   
   //frame production format
   unsigned int encoded_width;
   unsigned int encoded_height;
   
   //render bounds
   unsigned int x;
   unsigned int y;
   unsigned int width;
   unsigned int height;

   //control settings
   unsigned int chromakey;
   unsigned int brightness;
   unsigned int contrast;
   unsigned int volume;

   //state
   unsigned char modifiers;
   unsigned char suspended;
   unsigned char stop_production;
   unsigned char rendering;

   //clipping
   struct video_cliprect clip;
   unsigned int clipcount;
   unsigned int clipflags;
}video_smem;


