/*
 * java.lang.Class.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/soft.h"
#include "../../../kaffe/kaffevm/engine.h"
#include "../../../kaffe/kaffevm/exception.h"
#include <jni.h>
#include "defs.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

Hjava_lang_Class* javaLangVoidClass;
Hjava_lang_Class* javaLangBooleanClass;
Hjava_lang_Class* javaLangByteClass;
Hjava_lang_Class* javaLangCharacterClass;
Hjava_lang_Class* javaLangShortClass;
Hjava_lang_Class* javaLangIntegerClass;
Hjava_lang_Class* javaLangLongClass;
Hjava_lang_Class* javaLangFloatClass;
Hjava_lang_Class* javaLangDoubleClass;

/*
 * Convert class to string name.
 */
JNIEXPORT jstring JNICALL
Java_java_lang_Class_getName(JNIEnv* env, jclass c)
{
	int len;
	char* name;
	int i;
	char* ptr;

	ptr = CLASS_CNAME((Hjava_lang_Class*)c);
	len = strlen(ptr);
	name = alloca(len+2);
	for (i = 0; i < len; i++) {
		char ch = *ptr++;
		if (ch == '/') {
			ch = '.';
		}
		name[i] = ch;
	}
	name[i] = 0;

	if (CLASS_IS_ARRAY((Hjava_lang_Class*)c) && !CLASS_IS_PRIMITIVE(CLASS_ELEMENT_TYPE((Hjava_lang_Class*)c))) {
		name[i++] = ';';
	}

	return ((*env)->NewStringUTF(env, name));
}

/*
 * Create a new instance of the derived class.
 */
JNIEXPORT jobject JNICALL
Java_java_lang_Class_newInstance0(JNIEnv* env, jclass clazz)
{
	return (*env)->NewObject(env, clazz, (*env)->GetMethodID(env, clazz, "<init>", "()V"));
}

/*
 * Return super class.
 */
JNIEXPORT jclass JNICALL
Java_java_lang_Class_getSuperclass(JNIEnv* env, jclass clazz)
{
	return ((*env)->GetSuperclass(env, clazz));
}

JNIEXPORT jarray JNICALL /* [Ljava.lang.Class; */
Java_java_lang_Class_getInterfaces(JNIEnv* env, jclass clazz)
{
	int i;
	int nr;
	jarray obj;

	nr = ((Hjava_lang_Class*)clazz)->interface_len;
	obj = (*env)->NewObjectArray(env, nr, (*env)->FindClass(env, "java.lang.Class"), 0);
	for (i = 0; i < nr; i++) {
		(*env)->SetObjectArrayElement(env, obj, i, (jobject)((Hjava_lang_Class*)clazz)->interfaces[i]);
	}

	return (obj);
}

/*
 * Return the class loader which loaded me.
 */
JNIEXPORT jobject JNICALL
Java_java_lang_Class_getClassLoader0(JNIEnv* env, jclass clazz)
{
	return (((Hjava_lang_Class*)clazz)->loader);
}

/*
 * Is the class an interface?
 */
JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isInterface(JNIEnv* env, jclass clazz)
{
	return (CLASS_IS_INTERFACE((Hjava_lang_Class*)clazz));
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isPrimitive(JNIEnv* env, jclass clazz)
{
	return (CLASS_IS_PRIMITIVE((Hjava_lang_Class*)clazz));
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isArray(JNIEnv* env, jclass clazz)
{
	return (CLASS_IS_ARRAY((Hjava_lang_Class*)clazz));
}

JNIEXPORT jclass JNICALL
Java_java_lang_Class_getComponentType(JNIEnv* env, jclass clazz)
{
	if (Java_java_lang_Class_isArray(env, clazz)) {
		return (CLASS_ELEMENT_TYPE((Hjava_lang_Class*)clazz));
	}
	else {
		return (0);
	}
}

JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isAssignableFrom(JNIEnv* env, jclass this, jclass cls)
{
	return (instanceof((Hjava_lang_Class*)this, (Hjava_lang_Class*)cls));
}

/*
 * Get primitive class from class name (JDK 1.1)
 */
JNIEXPORT jclass JNICALL
Java_java_lang_Class_getPrimitiveClass(JNIEnv* env, jclass clazz, jstring name)
{
	const jchar* chrs;
	jchar one;
	jchar two;

#define loadStaticClass(C,N) C = loadClass((N), 0, CSTATE_LINKED)

	chrs = (*env)->GetStringChars(env, name, 0);
	one = chrs[0];
	two = chrs[1];
	(*env)->ReleaseStringChars(env, name, chrs);

	switch (one) {
	case 'b':
		switch (two) {
		case 'y':
			loadStaticClass(javaLangByteClass, "java.lang.Byte");
			return (&_Jv_byteClass);
		case 'o':
			loadStaticClass(javaLangBooleanClass, "java.lang.Boolean");
			return (&_Jv_booleanClass);
		}
		break;
	case 'c':
		loadStaticClass(javaLangCharacterClass, "java.lang.Character");
		return (&_Jv_charClass);
	case 'd':
		loadStaticClass(javaLangDoubleClass, "java.lang.Double");
		return (&_Jv_doubleClass);
	case 'f':
		loadStaticClass(javaLangFloatClass, "java.lang.Float");
		return (&_Jv_floatClass);
	case 'i':
		loadStaticClass(javaLangIntegerClass, "java.lang.Integer");
		return (&_Jv_intClass);
	case 'l':
		loadStaticClass(javaLangLongClass, "java.lang.Long");
		return (&_Jv_longClass);
	case 's':
		loadStaticClass(javaLangShortClass, "java.lang.Short");
		return (&_Jv_shortClass);
	case 'v':
		loadStaticClass(javaLangVoidClass, "java.lang.Void");
		return (&_Jv_voidClass);
	}

	return(0);
}

/*
 * Is object instance of this class?
 */
JNIEXPORT jboolean JNICALL
Java_java_lang_Class_isInstance(JNIEnv* env, jclass this, jobject obj)
{      
	return (_Jv_IsInstanceOf((Hjava_lang_Object*)obj, (Hjava_lang_Class*)this));
}

JNIEXPORT jint JNICALL
Java_java_lang_Class_getModifiers(JNIEnv* env, jclass this)
{
	return (((Hjava_lang_Class*)this)->accflags & ACC_MASK);
}

JNIEXPORT jarray JNICALL
Java_java_lang_Class_getSigners(JNIEnv* env, jclass this)
{
	(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), "getSigners not implemented");
	return (0);
}

JNIEXPORT void JNICALL
Java_java_lang_Class_setSigners(JNIEnv* env, jclass this, jarray sigs)
{
	(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), "setSigners not implemented");
}
