/*
 * java.lang.ClassLoader.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/file.h"
#include "../../../kaffe/kaffevm/readClass.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/locks.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/baseClasses.h"
#include "../../../kaffe/kaffevm/exception.h"
#include <jni.h>
#include "defs.h"

#include "../../../kaffe/kaffevm/jdebug.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

/*
 * Translate an array of bytes into a class.
 */
JNIEXPORT jclass JNICALL
Java_java_lang_ClassLoader_defineClass0(JNIEnv* env, jclass ldr, jobject loader, jstring name, jarray data, jint offset, jint length)
{
	jclass clazz;
	jbyte* base;
	int blen;
	char* cname;

	cname = 0;
	if (name != 0) {
		cname = (char*)(*env)->GetStringUTFChars(env, name, 0);
	}

	base = (*env)->GetByteArrayElements(env, data, 0);
	blen = (*env)->GetArrayLength(env, data);

	clazz = (*env)->DefineClass(env, loader, base, blen);
	if (clazz == 0) {
		goto error;
	}

	/* 
	 * If a name was given, but the name we found in the class file
	 * was a different one, complain.
	 */
	if (cname != 0) {
		int i;
		char* classname = CLASS_CNAME((Hjava_lang_Class*)clazz);
		int len = strlen(cname);
		if (len != strlen(classname)) {
			goto error;
		}
		for (i = 0; i < len; i++) {
			if (cname[i] != classname[i] && (classname[i] != '/' || cname[i] != '.')) {
				(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.ClassFormatError"), "Wrong name");
				goto error;
			}
		}
	}
#if 0
	/*
	 * While it is not necessary that one be able to actually *use*
	 * the returned class object at this point, it is mandatory that
	 * the returned clazz object is a functional Class object.
	 *
	 * The following call will make sure that the returned class object
	 * has its dispatch table set.
	 *
	 * Presumably, it shouldn't be necessary here, but is at the
	 * moment - XXX
	 */
	processClass((Hjava_lang_Class*)clazz, CSTATE_LINKED);
#endif
	return (clazz);

	error:;
	if (cname != 0) {
		(*env)->ReleaseStringUTFChars(env, name, cname);
	}
	(*env)->ReleaseByteArrayElements(env, data, base, JNI_ABORT);
	return (0);
}

/*
 * Resolve classes reference by this class.
 */
JNIEXPORT void JNICALL
Java_java_lang_ClassLoader_resolveClass0(JNIEnv* env, jobject this, jclass class)
{
	processClass((Hjava_lang_Class*)class, CSTATE_COMPLETE);
}

/*
 * Find a loaded class.
 */
JNIEXPORT jclass JNICALL
Java_java_lang_ClassLoader_findLoadedClass0(JNIEnv* env, jobject this, jstring name)
{
	Hjava_lang_Class* cls;
	char* cname;

	cname = (char*)(*env)->GetStringUTFChars(env, name, 0);
	cls = lookupClass(cname, this);
	(*env)->ReleaseStringUTFChars(env, name, cname);

	return (cls);
}

/**
 * Class loader being finalized - remove references to any classes
 * and libraries
 */
JNIEXPORT void JNICALL
Java_java_lang_ClassLoader_finalize0(JNIEnv* env, jobject this)
{
	unloadClassLoader((Hjava_lang_ClassLoader*)this);
}
