/*
 * java.lang.Double.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include <math.h>
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/locks.h"
#include "../../../kaffe/kaffevm/lerrno.h"
#include "../../../kaffe/kaffevm/fp.h"
#include "defs.h"
#include <jni.h>
#include "md.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT
/*
 * Convert double to jlong.
 */
JNIEXPORT jlong JNICALL
Java_java_lang_Double_doubleToRawLongBits(JNIEnv* env, jclass dbl, double val)
{
	return(doubleToLong(val));
}

/*
 * Convert jlong to double.
 */
JNIEXPORT double JNICALL
Java_java_lang_Double_longBitsToDouble(JNIEnv* env, jclass dbl, jlong val)
{
	return(longToDouble(val));
}

/*
 * Convert a "normal" double to a string with the supplied precision.
 *
 * This assumes printf(3) conforms to ISO 9899: 1990 (``ISO C'').
 *
 * XXX We still print some values incorrectly. For example the double
 * XXX 0x400B333333333333 should be displayed as "3.4" instead of what
 * XXX we display, "3.3999999999999999".
 */
JNIEXPORT jstring JNICALL
Java_java_lang_Double_toStringWithPrecision(JNIEnv* env, jclass dbl, jdouble val, jint precision)
{ 
	char buf[MAXNUMLEN];
#if defined(HAVE_SPRINTF)
	const jlong bits = Java_java_lang_Double_doubleToRawLongBits(env, dbl, val);
	char* s;
	int k;

	/* Deal with negative numbers manually so negative zero is "-0.0" */
	s = buf;
	if (bits & DSIGNBIT) {
		val = -val;
		*s++ = '-';
	}

	/* Print in normal or 'scientific' form according to value */
	if (val == 0.0 || (val >= 1.0e-3 && val < 1.0e7)) {
		static const double powTen[] = {
		  1.0e0, 1.0e1, 1.0e2, 1.0e3, 1.0e4, 1.0e5, 1.0e6
		};

		/* Account for precision digits ahead of the decimal point */
		for (k = 6; k >= 0 && val < powTen[k]; k--);
		precision -= k + 1;

		/* Print in decimal notation */
		sprintf(s, "%.*f", (int) precision, (double) val);

		/* Remove trailing zeroes after the decimal point */
		for (k = strlen(buf) - 1; buf[k] == '0' && buf[k - 1] != '.'; k--) {
			buf[k] = '\0';
		}
	} else {
		char* t;
		char* eptr;
		char* eval;

		/* Print in exponential notation */
		sprintf(s, "%.*E", (int) precision - 1, (double) val);

		/* Find the exponent */
		eptr = strchr(buf, 'E');
		if (eptr != 0) {
			/* Remove a '+' sign, but leave a '-' sign */
			switch (eptr[1]) {
			case '-':
				eval = &eptr[2];
				break;
			case '+':		/* remove '+' */
				memmove(eptr + 1, eptr + 2, strlen(eptr + 1));
				/* fall through */
			default:		/* shouldn't ever happen */
				eval = &eptr[1];
				break;
			}

			/* Remove leading zeroes in the exponent, if any */
			for (t = eval; t[0] == '0' && t[1] != '\0'; t++)
				;
			memmove(eval, t, strlen(t) + 1);

			/* Remove trailing zeroes after the decimal point */
			for (t = eptr - 1; *t == '0' && t[-1] != '.'; t--)
				;
			memmove(t + 1, eptr, strlen(eptr) + 1);
		}
	}
#else
	strcpy(buf, "NotImplemented");
#endif

	return ((*env)->NewStringUTF(env, buf));
}

/*
 * Convert string to double object.
 */
JNIEXPORT double JNICALL
Java_java_lang_Double_valueOf0(JNIEnv* env, jclass dbl, jstring str)
{
	double value;
	char* endbuf;
	char* msg;
	int negate;
	char* buf;
	const char* obuf;

	obuf = (*env)->GetStringUTFChars(env, str, 0);
	buf = (char*)obuf;
	if (buf[0] == '-') {
		negate = 1;
		buf++;
	}
	else {
		negate = 0;
	}

	msg = "bad float/double format";
#if defined(HAVE_STRTOD)
	errno = 0;
	value = strtod(buf, &endbuf);
        while (*endbuf != 0) {
                switch (*endbuf) {
                case ' ':
                case '\t':
                case '\n':
                case '\r':      /* Ignore whitespace */
                case 'F':
                case 'f':       /* Ignore float suffix. */
                case 'D':
                case 'd':       /* Ignore double suffix. */
                        endbuf++;
                        break;
                default:
			goto error;
                        break;
                }
        }
	if (endbuf == buf) {
		msg = "empty string";
		goto error;
	}
	if (errno == ERANGE) {
		if (value == HUGE_VAL || value == -HUGE_VAL) {
			msg = "overflow";
			goto error;
		}
		if (value == 0.0) {
			msg = "underflow";
			goto error;
		}
	}
#elif defined(HAVE_ATOF)
	/* Fall back on old atof - no error checking */
	value = atof(buf);
#else
	value = 0;
#endif
	if (negate) {
		value = -value;
	}

	(*env)->ReleaseStringUTFChars(env, str, obuf);

	return (value);

	error:;
	(*env)->ReleaseStringUTFChars(env, str, obuf);
	(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.NumberFormatException"), msg);
	return (0);
}
