/*
 * java.lang.reflect.Field.c
 *
 * Copyright (c) 1996,97 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.co.uk>
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/soft.h"
#include "../../../kaffe/kaffevm/baseClasses.h"
#include <jni.h>
#include "defs.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static
void*
getFieldAddress(JNIEnv* env, jobject field, jobject object, jclass fcls, jint slot)
{
	Field* fld;

        fld = &CLASS_FIELDS((Hjava_lang_Class*)fcls)[slot];

        if (slot < CLASS_NSFIELDS((Hjava_lang_Class*)fcls)) {
                return (FIELD_ADDRESS(fld));
        }
        else if (object == 0) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.NullPointerException"), "");
		return (0);
	}
	else if  (!_Jv_IsInstanceOf(object, (Hjava_lang_Class*)fcls)) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.IllegalArgumentException"), "");
		return (0);
	}
	return ((void*)(((char*)(object)) + FIELD_BOFFSET(fld)));
}

JNIEXPORT jint JNICALL
Java_java_lang_reflect_Field_getModifiers0(JNIEnv* env, jclass field, jclass cls, jint slot)
{
	return (CLASS_FIELDS((Hjava_lang_Class*)cls)[slot].accflags & ACC_MASK);
}


JNIEXPORT jobject JNICALL
Java_java_lang_reflect_Field_getObject0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jobject*)getFieldAddress(env, field, obj, clazz, slot));
}


JNIEXPORT jboolean JNICALL
Java_java_lang_reflect_Field_getBoolean0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jboolean*)getFieldAddress(env, field, obj, clazz, slot));
}


JNIEXPORT jbyte JNICALL
Java_java_lang_reflect_Field_getByte0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jbyte*)getFieldAddress(env, field, obj, clazz, slot));
}

JNIEXPORT jchar JNICALL
Java_java_lang_reflect_Field_getChar0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jchar*)getFieldAddress(env, field, obj, clazz, slot));
}


JNIEXPORT jshort JNICALL
Java_java_lang_reflect_Field_getShort0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jshort*)getFieldAddress(env, field, obj, clazz, slot));
}

JNIEXPORT jint JNICALL
Java_java_lang_reflect_Field_getInt0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jint*)getFieldAddress(env, field, obj, clazz, slot));
}

JNIEXPORT jlong JNICALL
Java_java_lang_reflect_Field_getLong0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jlong*)getFieldAddress(env, field, obj, clazz, slot));
}

JNIEXPORT jfloat JNICALL
Java_java_lang_reflect_Field_getFloat0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jfloat*)getFieldAddress(env, field, obj, clazz, slot));
}

JNIEXPORT jdouble JNICALL
Java_java_lang_reflect_Field_getDouble0(JNIEnv* env, jobject field, jobject obj, jclass clazz, jint slot)
{
	return (*(jdouble*)getFieldAddress(env, field, obj, clazz, slot));
}


JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setBoolean(JNIEnv* env, jobject field, jobject obj, jboolean val, jclass clazz, jint slot)
{
	*(jboolean*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setByte(JNIEnv* env, jobject field, jobject obj, jbyte val, jclass clazz, jint slot)
{
	*(jbyte*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setChar(JNIEnv* env, jobject field, jobject obj, jchar val, jclass clazz, jint slot)
{
	*(jchar*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setShort(JNIEnv* env, jobject field, jobject obj, jshort val, jclass clazz, jint slot)
{
	*(jshort*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setInt(JNIEnv* env, jobject field, jobject obj, jint val, jclass clazz, jint slot)
{
	*(jint*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setLong(JNIEnv* env, jobject field, jobject obj, jlong val, jclass clazz, jint slot)
{
	*(jlong*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setFloat(JNIEnv* env, jobject field, jobject obj, jfloat val, jclass clazz, jint slot)
{
	*(jfloat*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_setDouble(JNIEnv* env, jobject field, jobject obj, jdouble val, jclass clazz, jint slot)
{
	*(jdouble*)getFieldAddress(env, field, obj, clazz, slot) = val;
}

JNIEXPORT void JNICALL
Java_java_lang_reflect_Field_set(JNIEnv* env, jobject field, jobject obj, jobject val, jclass clazz, jint slot)
{
	*(jobject*)getFieldAddress(env, field, obj, clazz, slot) = val;
}
