/*
 * java.lang.reflect.Method.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/exception.h"
#include "../../../kaffe/kaffevm/baseClasses.h"
#include "../../../kaffe/kaffevm/threads.h"
#include <jni.h>
#include "defs.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static jclass Zclass;
static jclass Bclass;
static jclass Sclass;
static jclass Cclass;
static jclass Iclass;
static jclass Jclass;
static jclass Fclass;
static jclass Dclass;
static jfieldID Zvalue;
static jfieldID Bvalue;
static jfieldID Svalue;
static jfieldID Cvalue;
static jfieldID Ivalue;
static jfieldID Jvalue;
static jfieldID Fvalue;
static jfieldID Dvalue;
static jmethodID Zmeth;
static jmethodID Bmeth;
static jmethodID Smeth;
static jmethodID Cmeth;
static jmethodID Imeth;
static jmethodID Jmeth;
static jmethodID Fmeth;
static jmethodID Dmeth;

JNIEXPORT void JNICALL
Java_java_lang_reflect_Method_init0(JNIEnv* env, jclass clazz)
{
	Zclass = (*env)->FindClass(env, "java.lang.Boolean");
	Zvalue = (*env)->GetFieldID(env, Zclass, "value", "Z");
	Zmeth = (*env)->GetMethodID(env, Zclass, "<init>", "(Z)V");

	Bclass = (*env)->FindClass(env, "java.lang.Byte");
	Bvalue = (*env)->GetFieldID(env, Bclass, "value", "B");
	Bmeth = (*env)->GetMethodID(env, Bclass, "<init>", "(B)V");

	Sclass = (*env)->FindClass(env, "java.lang.Short");
	Svalue = (*env)->GetFieldID(env, Sclass, "value", "S");
	Smeth = (*env)->GetMethodID(env, Sclass, "<init>", "(S)V");

	Cclass = (*env)->FindClass(env, "java.lang.Character");
	Cvalue = (*env)->GetFieldID(env, Cclass, "value", "C");
	Cmeth = (*env)->GetMethodID(env, Cclass, "<init>", "(C)V");

	Iclass = (*env)->FindClass(env, "java.lang.Integer");
	Ivalue = (*env)->GetFieldID(env, Iclass, "value", "I");
	Imeth = (*env)->GetMethodID(env, Iclass, "<init>", "(I)V");

	Jclass = (*env)->FindClass(env, "java.lang.Long");
	Jvalue = (*env)->GetFieldID(env, Jclass, "value", "J");
	Jmeth = (*env)->GetMethodID(env, Jclass, "<init>", "(J)V");

	Fclass = (*env)->FindClass(env, "java.lang.Float");
	Fvalue = (*env)->GetFieldID(env, Fclass, "value", "F");
	Fmeth = (*env)->GetMethodID(env, Fclass, "<init>", "(F)V");

	Dclass = (*env)->FindClass(env, "java.lang.Double");
	Dvalue = (*env)->GetFieldID(env, Dclass, "value", "D");
	Dmeth = (*env)->GetMethodID(env, Dclass, "<init>", "(D)V");
}

JNIEXPORT jint JNICALL
Java_java_lang_reflect_Method_getModifiers0(JNIEnv* env, jclass method, jclass clazz, jint slot)
{
	return (CLASS_METHODS((Hjava_lang_Class*)clazz)[slot].accflags & ACC_MASK);
}

JNIEXPORT jobject JNICALL
Java_java_lang_reflect_Method_invoke0(JNIEnv* env, jobject method, jobject obj, jarray argobj, jclass clazz, jint slot, jarray paramtypes, jclass retclazz)
{
	jobject arg;
	jclass argc;
	Method* meth;
	jvalue ret;
	int len;
	int i;
	jvalue* args;

	len = argobj == 0 ? 0 : (*env)->GetArrayLength(env, argobj);
	args = alloca(len * sizeof(jvalue));
	for (i = len - 1; i >= 0; i--) {
		arg = (*env)->GetObjectArrayElement(env, argobj, i);
		argc = (*env)->GetObjectArrayElement(env, paramtypes, i);
		if (!CLASS_IS_PRIMITIVE((Hjava_lang_Class*)argc)) {
			args[i].l = arg;
		}
		else switch (CLASS_PRIM_SIG((Hjava_lang_Class*)argc)) {
		case 'I':
			args[i].i = (*env)->GetIntField(env, arg, Ivalue);
			break;
		case 'Z':
			args[i].z = (*env)->GetBooleanField(env, arg, Zvalue);
			break;
		case 'S':
			args[i].s = (*env)->GetShortField(env, arg, Svalue);
			break;
		case 'B':
			args[i].b = (*env)->GetByteField(env, arg, Bvalue);
			break;
		case 'C':
			args[i].c = (*env)->GetCharField(env, arg, Cvalue);
			break;
		case 'F':
			args[i].f = (*env)->GetFloatField(env, arg, Fvalue);
			break;
		case 'D':
			args[i].d = (*env)->GetDoubleField(env, arg, Dvalue);
			break;
		case 'J':
			args[i].j = (*env)->GetLongField(env, arg, Jvalue);
			break;
		}
	}

	/* Get the method we want to call */
	meth = &CLASS_METHODS((Hjava_lang_Class*)clazz)[slot];

	/* Deal with constructors here */
	if (METHOD_IS_CONSTRUCTOR(meth)) {
		obj = (*env)->NewObjectA(env, clazz, (jmethodID)meth, args);
		if ((*env)->ExceptionOccurred(env) != 0) {
			/* Cant do this easily in JNI, so use kaffe mechanism */
			postExceptionInException("java.lang.reflect/InvocationTargetException");
		}
                return (obj);
	}

	/* Call the method.  Again, we don't use JNI here because it would
	 * be too cumbersome - so go use the kaffe meachnisms.
	 */
#ifdef DIRECT_CALL
	callAndCatchMethodA(meth, (METHOD_IS_STATIC(meth) ? METHOD_INDIRECTMETHOD(meth) : ((Hjava_lang_Object*)obj)->dtable->method[meth->idx]), obj, args, &ret);
#else
	callAndCatchMethodA(meth, (METHOD_IS_STATIC(meth) ? METHOD_INDIRECTMETHOD(meth) : *((Hjava_lang_Object*)obj)->dtable->method[meth->idx]), obj, args, &ret);
#endif

	if ((*env)->ExceptionOccurred(env) != 0) {
		/* Cant do this easily in JNI, so use kaffe mechanism */
		postExceptionInException("java.lang.reflect/InvocationTargetException");
		return (0);
	}
	else if (!CLASS_IS_PRIMITIVE((Hjava_lang_Class*)retclazz)) {
		return (ret.l);
	}
	else switch (CLASS_PRIM_SIG((Hjava_lang_Class*)retclazz)) {
	case 'I':
		return ((*env)->NewObject(env, Iclass, Imeth, ret.i));
		break;
	case 'Z':
		return ((*env)->NewObject(env, Zclass, Zmeth, ret.z));
		break;
	case 'S':
		return ((*env)->NewObject(env, Sclass, Smeth, ret.s));
		break;
	case 'B':
		return ((*env)->NewObject(env, Bclass, Bmeth, ret.b));
		break;
	case 'C':
		return ((*env)->NewObject(env, Cclass, Cmeth, ret.c));
		break;
	case 'F':
		return ((*env)->NewObject(env, Fclass, Fmeth, ret.f));
		break;
	case 'D':
		return ((*env)->NewObject(env, Dclass, Dmeth, ret.d));
		break;
	case 'J':
		return ((*env)->NewObject(env, Jclass, Jmeth, ret.j));
		break;
	}
	return (0);
}
