/*
 * java.lang.Object.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/locks.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/soft.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include <jni.h>

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

/*
 * Generate object hash code.
 */
JNIEXPORT jint JNICALL
Java_java_lang_Object_hashCode(JNIEnv* env, jobject o)
{
	/* Hash code is object's address */
	return ((jint)(jword)o);
}

/*
 * Return class object for this object.
 */
JNIEXPORT jclass JNICALL
Java_java_lang_Object_getClass(JNIEnv* env, jobject o)
{
	return ((*env)->GetObjectClass(env, o));
}

/*
 * Notify threads waiting here.
 */
JNIEXPORT void JNICALL
Java_java_lang_Object_notifyAll(JNIEnv* env, jobject o)
{
	broadcastCond((struct Hjava_lang_Object*)o);
}

/*
 * Notify a thread waiting here.
 */
JNIEXPORT void JNICALL
Java_java_lang_Object_notify(JNIEnv* env, jobject o)
{
	signalCond((struct Hjava_lang_Object*)o);
}

/*
 * Clone me.
 */
JNIEXPORT jobject JNICALL
Java_java_lang_Object_clone(JNIEnv* env, jobject ob)
{
	Hjava_lang_Object* obj;
	Hjava_lang_Object* o;
	Hjava_lang_Class* class;
	static Hjava_lang_Class* cloneClass;

	/* * Lookup cloneable class interface if we don't have it  */
	if (cloneClass == 0) {
		cloneClass = loadClass("java.lang.Cloneable", 0, CSTATE_COMPLETE);
	}

	o = (Hjava_lang_Object*)ob;
	class = OBJECT_CLASS(o);

	if (!CLASS_IS_ARRAY(class)) {
		/* Check class is cloneable and throw exception if it isn't */
		if (_Jv_IsInstanceOf(o, cloneClass) == 0) {
			(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.CloneNotSupportedException"), CLASS_CNAME(class));
			return (0);
		}
		/* Clone an object */
		obj = newObject(class);
		memcpy(OBJECT_DATA(obj), OBJECT_DATA(o), CLASS_FSIZE(class) - sizeof(Hjava_lang_Object));
#if defined(MEM2)
                {
                        int n;
                        unsigned int* refs;
                        void * mem;

                        n = class->nrefs;
                        refs = class->refs;
                        while (--n >= 0) {
                                mem = *(void**)((uintp)obj+refs[n]);
                                GC_WRITE( obj, mem );
                        }
                }
#endif
	}
	else {
		/* Clone an array */
		class = CLASS_ELEMENT_TYPE(class);
		obj = newArray(class, ARRAY_SIZE(o));
		if (class == &_Jv_intClass || class == &_Jv_floatClass) {
			memcpy(ARRAY_DATA(HArrayOfInt, obj), ARRAY_DATA(HArrayOfInt, o), ARRAY_SIZE(o) * TYPE_SIZE(class));
		}
		else if (class == &_Jv_longClass || class == &_Jv_doubleClass) {
			memcpy(ARRAY_DATA(HArrayOfLong, obj), ARRAY_DATA(HArrayOfLong, o), ARRAY_SIZE(o) * TYPE_SIZE(class));
		}
		else if (class == &_Jv_booleanClass || class == &_Jv_byteClass) {
			memcpy(ARRAY_DATA(HArrayOfByte, obj), ARRAY_DATA(HArrayOfByte, o), ARRAY_SIZE(o) * TYPE_SIZE(class));
		}
		else if (class == &_Jv_charClass || class == &_Jv_shortClass) {
			memcpy(ARRAY_DATA(HArrayOfShort, obj), ARRAY_DATA(HArrayOfShort, o), ARRAY_SIZE(o) * TYPE_SIZE(class));
		}
		else {
			memcpy(ARRAY_DATA(HArrayOfObject, obj), ARRAY_DATA(HArrayOfObject, o), ARRAY_SIZE(o) * TYPE_SIZE(class));
#if defined(MEM2)
                        {
                                int n;
                                Hjava_lang_Object** array;
                                
                                array = ARRAY_DATA(HArrayOfObject, obj);
                                n = ARRAY_SIZE(obj);
                                while (--n >= 0) {
                                        GC_WRITE( obj, array[n] );
                                }
                        }
#endif
		}
	}
	return (obj);
}

/*
 * Wait for this object to be notified.
 */
JNIEXPORT void JNICALL
Java_java_lang_Object_wait0(JNIEnv* env, jobject o, jlong timeout)
{
	waitCond((Hjava_lang_Object*)o, timeout);
}
