/*
 * java.lang.Runtime.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"

#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/external.h"
#include "../../../kaffe/kaffevm/gc.h"
#include "../../../include/system.h"
#include <native.h>
#include "files.h"
#include "defs.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

JNIIMPORT int runFinalizerOnExit;

#define	PROCESS	"kaffe.lang.UNIXProcess"
/* #define	PROCESS	"kaffe.lang.DOSProcess" */

/*
 * Load in a library file.
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_loadFileInternal(JNIEnv* env, jobject this, jstring name, jobject classloader)
{
	char* lib;
	char* error;

	lib = (char*)(*env)->GetStringUTFChars(env, name, 0);
	error = loadNativeLibrary(lib, (struct Hjava_lang_ClassLoader*)classloader);
	(*env)->ReleaseStringUTFChars(env, name, lib);
	if (error != 0) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.UnsatisfiedLinkError"), error);
	}
}

/*
 * Exit - is this just a thread or the whole thing?
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_exitInternal(JNIEnv* env, jobject this, jint v)
{
	EXIT(v);
}

/*
 * Exec another program.
 */
JNIEXPORT jobject JNICALL
Java_java_lang_Runtime_execInternal(JNIEnv* env, jobject this, jarray argv, jarray arge)
{
#if defined(PROCESS)
	static jclass pclass;
	static jmethodID pconst;

	if (pclass == 0) {
		pclass = (*env)->FindClass(env, PROCESS);
		pconst = (*env)->GetMethodID(env, pclass, "<init>", "([Ljava/lang/String;[Ljava/lang/String;)V");
	}

	return ((*env)->NewObject(env, pclass, pconst, argv, arge));
#else
	return (0);
#endif
}

/*
 * Free memory.
 */
JNIEXPORT jlong JNICALL
Java_java_lang_Runtime_freeMemory(JNIEnv* env, jobject this)
{
	return (heapFree());
}

/*
 * Total memory.
 */
JNIEXPORT jlong JNICALL
Java_java_lang_Runtime_totalMemory(JNIEnv* env, jobject this)
{
	return (heapTotal());
}

/*
 * Run the garbage collector.
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_gc(JNIEnv* env, jobject this)
{
	invokeGC(0);
}

/*
 * Run any pending finialized methods.
 *  Finalising is part of the garbage collection system - so just run that.
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_runFinalization(JNIEnv* env, jobject this)
{
	invokeGC(0);
	invokeFinalizer();
}

/*
 * Enable/disable tracing of instructions.
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_traceInstructions(JNIEnv* env, jobject this, jboolean on)
{
	if (on == JNI_TRUE) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.RuntimeException"), "Cannot trace instructions");
	}
}

/*
 * Enable/disable tracing of method calls.
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_traceMethodCalls(JNIEnv* env, jobject this, jboolean on)
{
	if (on == JNI_TRUE) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.RuntimeException"), "Cannot trace method calls");
	}
}

/*
 * Inform the runtime that it must run the finalizer when it exits.
 */
JNIEXPORT void JNICALL
Java_java_lang_Runtime_runFinalizersOnExit(JNIEnv* env, jobject this, jboolean on)
{
	runFinalizerOnExit = on;
}
