/*
 * java.lang.String.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include <jni.h>
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/strings.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

JNIEXPORT jstring JNICALL
Java_java_lang_String_intern0(JNIEnv* env, jstring str)
{
	return ((jstring)internJavaString((Hjava_lang_String*)str));
}

JNIEXPORT void JNICALL
Java_java_lang_String_unintern0(JNIEnv* env, jstring str)
{
	uninternJavaString((Hjava_lang_String*)str);
}

JNIEXPORT jint JNICALL
Java_java_lang_String_indexOf(JNIEnv* env, jstring str, jstring pat, jint offset)
{
  const jchar *a;
  const jchar *p;

  int   n;
  int   m;
  int   m2;
  jint  i, k;
  unsigned char  bs[256];
  int *ibs;

  if ( !pat || !str ) {
    return -1;
  }

  if ( offset < 0 ) {
	offset = 0;
  }

  n = (*env)->GetStringLength(env, str);

  /* If offset is greated than base lengt, we won't find anything */
  if (offset > n) {
	return (-1);
  }

  m = (*env)->GetStringLength(env, pat);
  m2 = m * 2;

  /* If pattern is 0 we'll find it af the offset */
  if (m == 0) {
	return (offset);
  }

  /* If body, starting at the offset, is less than the length of the
   * pattern, we'll never find it, so -1
   */
  if (n - offset < m) {
	return (-1);
  }

  a = (*env)->GetStringChars(env, str, 0);
  p = (*env)->GetStringChars(env, pat, 0);

  if ( (m < 3) || (n < 128) || (m > 256) ) {
	/*
	 * If the pattern is too small, iterating the searched array would be
	 * cheaper than to set up the badmatch table, or the pattern length exceeds
	 * what we can store in the badmatch table, we revert to plain old brute force
	 */
	k = n - m+1;
	for ( i=offset; i<k; i++ ) {
	  if ( memcmp( &a[i], p, m2) == 0 ) {
		(*env)->ReleaseStringChars(env, str, a);
		(*env)->ReleaseStringChars(env, pat, p);
		return (i);		
	  }
	}
  }
  else {
	/*
	 * Otherwise we use the Quick search variant of Boyer-Moore for 
	 * O(n/m) lookup. Note that the (char) cast in the badmatch table access
	 * is the justification for turning this into a native method
	 */

	/* set up the badmatch table */
	k = (m << 24) | (m<<16) | (m<<8) | m;
	for ( i=0, ibs = (int*)bs; i<64; i++ ) {
	  ibs[i] = k; /* the default jump: m */
	}
	for ( i=0; i<m; i++ ) {
	  bs[ (unsigned char)p[i] ] = m-i;
	}

	k= n - m+1;
	for ( i=offset; i <= k; ) {
	  if ( memcmp( &a[i], p, m2) == 0 ) {
		(*env)->ReleaseStringChars(env, str, a);
		(*env)->ReleaseStringChars(env, pat, p);
		return i;
	  }
	  i += bs[ (unsigned char)a[i+m] ];
	}
  }

  (*env)->ReleaseStringChars(env, str, a);
  (*env)->ReleaseStringChars(env, pat, p);
  return -1;
}
