/*
 * java.lang.System.c
 *
 * Copyright (c) 1996, 1997, 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#if defined(HAVE_SYS_UTSNAME_H)
#include <sys/utsname.h>
#endif
#if defined(HAVE_PWD_H)
#include <pwd.h>
#endif
#if defined(HAVE_TIME_H)
#include <time.h>
#endif
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/baseClasses.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/soft.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../include/system.h"
#include "defs.h"
#include <jni.h>
#include "md.h"
#include "../../../include/jsyscall.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT
#include "jfilesystem.h"

#if defined(HAVE_TZNAME)
extern char* tzname[2];
#endif

static char cwdpath[MAXPATHLEN];

extern jint java_lang_Object_hashCode(struct Hjava_lang_Object*);
JNIIMPORT userProperty* userProperties;
JNIIMPORT Hjava_lang_Class* SystemClass;
JNIIMPORT char* realClassPath;
JNIIMPORT char* realKaffeHome;

JNIIMPORT char* userDir;

static
void
setProp(JNIEnv* env, jobject properties, char* key, char* value)
{
	static jmethodID setmeth;

	if (setmeth == 0) {
		setmeth = (*env)->GetMethodID(env, (*env)->GetObjectClass(env, properties), "setProperty", "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Object;");
	}

	(*env)->CallVoidMethod(env, properties, setmeth, (*env)->NewStringUTF(env, key), (*env)->NewStringUTF(env, value));
}

/*
 * Initialise system properties to their defaults.
 */
JNIEXPORT jobject JNICALL
Java_java_lang_System_initProperties(JNIEnv* env, jclass system, jobject p)
{
	char* jhome;
	char* cpath;
	userProperty* prop;
#if defined(HAVE_SYS_UTSNAME_H)
	struct utsname sys;
#endif
#if defined(HAVE_PWD_H)
	struct passwd* pw;
#endif
#if defined(HAVE_TM_ZONE) && (defined(HAVE_TZSET) || defined(HAVE_LOCALTIME))
	struct tm* tzone;
	time_t tme;
#endif

	/* Add the default properties:
	 * java.version		Java version number
	 * java.vendor          Java vendor specific string
	 * java.vendor.url      Java vendor URL
	 * java.class.version   Java class version number
	 * java.class.path      Java classpath
	 * java.home            Java installation directory
	 * java.compiler        Java JIT compiler
	 *
	 * os.name              Operating System Name
	 * os.arch              Operating System Architecture
	 * os.version           Operating System Version
	 *
	 * file.separator       File separator ("/" on Unix)
	 * path.separator       Path separator (":" on Unix)
	 * line.separator       Line separator ("\n" on Unix)
	 *
	 * user.name            User account name
	 * user.home            User home directory
	 * user.dir             User's current working directory
	 * user.language        User's language locale code
	 * user.region          User's country locale code
	 * user.timezone	User's timezone
	 *
	 * file.encoding	Character encoding for locale
	 * file.encoding.pkg	Character encoding package
	 */

	setProp(env, p, "java.version", java_version);
	setProp(env, p, "java.vendor", kaffe_vendor);
	setProp(env, p, "java.vendor.url", kaffe_vendor_url);
	setProp(env, p, "java.compiler", "kaffe.jit");

	jhome = realKaffeHome;
	setProp(env, p, "java.home", jhome);

	setProp(env, p, "java.class.version", kaffe_class_version);

	cpath = realClassPath;
	if (cpath == 0) {
		cpath = ".";
	}
	setProp(env, p, "java.class.path", cpath);

	setProp(env, p, "file.separator", file_separator);
	setProp(env, p, "path.separator", path_separator);
	setProp(env, p, "line.separator", line_separator);

#if defined(HAVE_GETCWD)
	userDir = getcwd(cwdpath, MAXPATHLEN);
#elif defined(HAVE_GETWD)
	userDir = getwd(cwdpath);
#else
	userDir = 0;	/* Cannot get current directory */
#endif
	if (userDir == 0) {
		userDir = ".";
	}
	setProp(env, p, "user.dir", userDir);

#if defined(HAVE_SYS_UTSNAME_H) && defined(HAVE_UNAME)
	/* Setup system properties */
	if (uname(&sys) >= 0) {
		setProp(env, p, "os.name", sys.sysname);
		setProp(env, p, "os.arch", sys.machine);
		setProp(env, p, "os.version", sys.release);
	}
	else
#endif
	{
		setProp(env, p, "os.name", "Unknown");
		setProp(env, p, "os.arch", "Unknown");
		setProp(env, p, "os.version", "Unknown");
	}
#if defined(HAVE_PWD_H) && defined(HAVE_GETUID)
	/* Setup user properties */
	pw = getpwuid(getuid());
	if (pw != 0) {
		setProp(env, p, "user.name", pw->pw_name);
		setProp(env, p, "user.home", pw->pw_dir);
	}
	else
#endif
	{
		setProp(env, p, "user.name", "Unknown");
		setProp(env, p, "user.home", "Unknown");
	}

	/* We should try to work this stuff out really - XXX */
	setProp(env, p, "user.language", "EN");
	setProp(env, p, "user.region", "US");

#if defined(HAVE_TZNAME) && defined(HAVE_TZSET)
	tzset();
	setProp(env, p, "user.timezone", tzname[0]);
#elif defined(HAVE_TM_ZONE) && defined(HAVE_LOCALTIME)
	tme = 0; /* We assume at the epoch we have the standard timezone */
	tzone = localtime(&tme);
	setProp(env, p, "user.timezone", (char*)tzone->tm_zone);
#else
	/* If we don't know they set to GMT */
	setProp(env, p, "user.timezone", "GMT");
#endif

	setProp(env, p, "file.encoding.pkg", "kaffe.io");
	setProp(env, p, "file.encoding", "8859_1");

	/* Where to put temporary files */
	setProp(env, p, "java.io.tmpdir", "/tmp");

	/* Set the default compiler to kjc */
	setProp(env, p, "kaffe.compiler", "kjc");

	/* Now process user defined properties */
	for (prop = userProperties; prop != 0; prop = prop->next) {
		setProp(env, p, prop->key, prop->value);
	}

	return (p);
}

/*
 * Return current time.
 */
JNIEXPORT jlong JNICALL
Java_java_lang_System_currentTimeMillis(JNIEnv* env, jclass system)
{
	return (currtime());
}

JNIEXPORT jint JNICALL
Java_java_lang_System_identityHashCode(JNIEnv* env, jobject o)
{
       return ((jint)o);
}

/*
 * Set the stdin stream.
 */
JNIEXPORT void JNICALL
Java_java_lang_System_setIn0(JNIEnv *env, jclass system, jobject stream)
{
        jfieldID in = (*env)->GetStaticFieldID(env, system, "in", "java.io.InputStream");
        (*env)->SetStaticObjectField(env, system, in, stream);
}

/*
 * Set the stdout stream.
 */
JNIEXPORT void JNICALL
Java_java_lang_System_setOut0(JNIEnv *env, jclass system, jobject stream)
{
        jfieldID out = (*env)->GetStaticFieldID(env, system, "out", "java.io.PrintStream");
        (*env)->SetStaticObjectField(env, system, out, stream);
}

/*
 * Set the error stream.
 */
JNIEXPORT void JNICALL
Java_java_lang_System_setErr0(JNIEnv *env, jclass system, jobject stream)
{
        jfieldID err = (*env)->GetStaticFieldID(env, system, "err", "java.io.PrintStream");
        (*env)->SetStaticObjectField(env, system, err, stream);
}
