/*
 * kaffe.lang.SystemClassLoader.c
 *
 * Copyright (c) 1996, 1997, 1998, 1999
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/file.h"
#include "../../../kaffe/kaffevm/access.h"
#include "../../../kaffe/kaffevm/constants.h"
#include "../../../kaffe/kaffevm/objects.h"
#include "../../../kaffe/kaffevm/classMethod.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/soft.h"
#include "../../../kaffe/kaffevm/engine.h"
#include "../../../kaffe/kaffevm/exception.h"
#include <jni.h>
#include "defs.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

JNIEXPORT jarray JNICALL Java_kaffe_lang_SystemClassLoader_getSystemResourceAsBytes0(JNIEnv*, jclass, jstring);

extern int findInJar(char*, classFile*);

/*
 * Load a class from the system.
 */
JNIEXPORT jclass JNICALL
Java_kaffe_lang_SystemClassLoader_loadClass(JNIEnv* env, jobject loader, jstring name, jboolean init)
{
	jclass cls;
	Hjava_lang_Class* entry;
	char* cname;

	cls = 0;
	cname = (char*)(*env)->GetStringUTFChars(env, name, 0);
	if (cname[0] != '[') {
		cls = (*env)->FindClass(env, cname);
	}

	if (cls == 0) {
		/* We upgrade an error to an exception if we failed to
		 * load the class but haven't yet failed to inialize it.
		 */
		if ((*env)->GetObjectClass(env, (*env)->ExceptionOccurred(env)) == (*env)->FindClass(env, "java.lang.NoClassDefFoundError")) {
			entry = lookupClass(cname, 0);
			if (entry == 0 || CLASS_STATE(entry) != CSTATE_FAILED) {
				(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.ClassNotFoundException"), cname);
				cname = 0;
			}
		}
	}
	else if (init) {
		processClass((Hjava_lang_Class*)cls, CSTATE_COMPLETE);
	}
	if (cname != 0) {
		(*env)->ReleaseStringUTFChars(env, name, cname);
	}
	return (cls);
}

/*
 * Generate an array of classes representing the classes on the current
 * threads stack.
 */
JNIEXPORT jarray JNICALL
Java_kaffe_lang_SystemClassLoader_getClassStack0(JNIEnv* env, jclass sysload)
{
	jarray array;
	stackTraceInfo* sinfo;
	int i;
	int c;

	sinfo = buildStackTrace();

	/* Count frames in stack trace */
	c = 0;
	for (i = 0; sinfo[i].meth != ENDOFTRACE; i++) {
		if (sinfo[i].meth == 0) {
			findFrameMethod(&sinfo[i]);
		}
		if (sinfo[i].meth != UNKNOWN_METHOD && sinfo[i].meth->name != 0) {
			c++;
		}
	}

	array = (*env)->NewObjectArray(env, c, (*env)->FindClass(env, "java.lang.Class"), 0);

	c = 0;
	for (i = 0; sinfo[i].meth != ENDOFTRACE; i++) {
		if (sinfo[i].meth != UNKNOWN_METHOD && sinfo[i].meth->name != 0) {
			(*env)->SetObjectArrayElement(env, array, c, (jobject)sinfo[i].clazz);
			c++;
		}
	}

	freeStackTrace(sinfo);

	return (array);
}

/*
 * Find the first class loaded by a class loader.
 */
JNIEXPORT jclass JNICALL
Java_kaffe_lang_SystemClassLoader_getCurrentClass0(JNIEnv* env, jclass sysload)
{
	stackTraceInfo* sinfo;
	int i;

	sinfo = buildStackTrace();

	for (i = 0; sinfo[i].meth != ENDOFTRACE; i++) {
		if (sinfo[i].meth == 0) {
			findFrameMethod(&sinfo[i]);
		}
		if (sinfo[i].meth != UNKNOWN_METHOD && sinfo[i].clazz != 0 && sinfo[i].clazz->loader != 0) {
			jclass ret = (jclass)sinfo[i].clazz;
			freeStackTrace(sinfo);
			return (ret);
		}
	}
	freeStackTrace(sinfo);
	return (0);
}

/*
 * Find the depth of the first class loaded by a class loader.
 */
JNIEXPORT jint JNICALL
Java_kaffe_lang_SystemClassLoader_getCurrentClassLoaderDepth0(JNIEnv* env, jclass sysload)
{
	stackTraceInfo* sinfo;
	int i;

	sinfo = buildStackTrace();

	for (i = 0; sinfo[i].meth != ENDOFTRACE; i++) {
		if (sinfo[i].meth == 0) {
			findFrameMethod(&sinfo[i]);
		}
		if (sinfo[i].meth != UNKNOWN_METHOD && sinfo[i].clazz != 0 && sinfo[i].clazz->loader != 0) {
			freeStackTrace(sinfo);
			return (i);
		}
	}
	freeStackTrace(sinfo);
	return (-1);
}

/*
 * Find the caller of the method which called the caller of this method
 * (yes I realize this is confusing).  We do this by counting frames
 * XXX THIS IS VERY BAD SINCE IF WE CHANGE CALLING CONVENTION THIS MIGHT
 *     STOP WORKING
 * XXX
 */
JNIEXPORT jobject JNICALL
Java_kaffe_lang_SystemClassLoader_getCallerClassLoader0(JNIEnv* env, jclass sysload, jint depth)
{
	stackTraceInfo* sinfo;
	int i;

	sinfo = buildStackTrace();

	for (i = 0; sinfo[i].meth != ENDOFTRACE; i++) {
		if (sinfo[i].pc >= (uintp)&Java_kaffe_lang_SystemClassLoader_getCallerClassLoader0 && sinfo[i].pc <= (uintp)&Java_kaffe_lang_SystemClassLoader_getSystemResourceAsBytes0) {
			for (i += depth; sinfo[i].meth != ENDOFTRACE; i++) {
				findFrameMethod(&sinfo[i]);
				if (sinfo[i].meth != UNKNOWN_METHOD && sinfo[i].clazz != 0) {
					jobject ret = (jobject)sinfo[i].clazz->loader;
					freeStackTrace(sinfo);
					return (ret);
				}
			}
			break;
		}
	}
	freeStackTrace(sinfo);
	return (0);
}

/*
 * Locate the requested resource in the current Jar files and create a
 *  byte array containing it.
 */
JNIEXPORT jarray JNICALL
Java_kaffe_lang_SystemClassLoader_getSystemResourceAsBytes0(JNIEnv* env, jclass sysload, jstring name)
{
	const char* cname;
	classFile hand;
	jarray data;
	int ret;

	cname = (*env)->GetStringUTFChars(env, name, 0);
	ret = findInJar((char*)cname, &hand);
	(*env)->ReleaseStringUTFChars(env, name, cname);

	if (ret == 0 || hand.type == 0) {
		return (0);
	}

	/* Copy data from returned buffer into Java byte array.  Be nice
	 * to avoid this copy but we cannot for the moment.
	 */
	data = (*env)->NewByteArray(env, hand.size);
	if (data != 0) {
		(*env)->SetByteArrayRegion(env, data, 0, hand.size, hand.buf);
	}
	gc_free_fixed(hand.base);

	return (data);
}

/*
 * Construct a library name.
 */
JNIEXPORT jstring JNICALL
Java_kaffe_lang_SystemClassLoader_findLibrary(JNIEnv* env, jobject this, jstring name)
{
	int len;
	char* lib;
	const char* nm;

	nm = (*env)->GetStringUTFChars(env, name, 0);

	len = strlen(nm) + strlen(LIBRARYPREFIX) + strlen(LIBRARYSUFFIX) + 1;
	lib = alloca(len);

	strcpy(lib, LIBRARYPREFIX);
	strcat(lib, nm);
	strcat(lib, LIBRARYSUFFIX);

	(*env)->ReleaseStringUTFChars(env, name, nm);

	return ((*env)->NewStringUTF(env, lib));
}
