/*
 * kaffe.lang.UNIXProcess.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "files.h"
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/locks.h"
#include "../../../kaffe/kaffevm/gc.h"
#include "../../../kaffe/kaffevm/exception.h"
#include <jsyscall.h>
#include <jni.h>

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

JNIEXPORT jint JNICALL
Java_kaffe_lang_UNIXProcess_execWait0(JNIEnv* env, jobject this)
{
	int status;
	status = -1;
#if defined(HAVE_WAITPID)
	waitpid(-1, &status, 0);
#endif
	return (status);
}

JNIEXPORT jint JNICALL
Java_kaffe_lang_UNIXProcess_forkAndExec0(JNIEnv* env, jobject this, jarray args, jarray envs)
{
#if (defined(HAVE_FORK) || defined(HAVE_VFORK)) && defined(HAVE_EXECVE)
	int pid;
	int in[2];
	int out[2];
	int err[2];
	char** argv;
	char** arge;
	int i;
	const char* path;
	int arglen;
	int envlen;
	char* syspath;
	char* p;
	char pbuf[256]; /* XXX */
	jobject jpath;
	jclass uc;

	arglen = (args == 0 ? 0 : (*env)->GetArrayLength(env, args));
	envlen = (envs == 0 ? 0 : (*env)->GetArrayLength(env, envs));

	if (arglen < 1) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.io.IOException"), "No such file");
		return (-1);
	}

	syspath = getenv("PATH");

	jpath = (*env)->GetObjectArrayElement(env, args, 0);
	path = (*env)->GetStringUTFChars(env, jpath, 0);

	/* If path is absolute, check for executable */
	if (path[0] == '/' || syspath == 0) {
		strcpy(pbuf, path);
		i = access(pbuf, X_OK);
	}
	/* Otherwise search path for executable */
	else {
		i = -1;
		for (p = syspath; *p != 0; p++) {
			if (*p == ':') {
				strncpy(pbuf, syspath, p-syspath);
				strcpy(&pbuf[p-syspath], "/");
				strcat(pbuf, path);
				syspath = p+1;
				i = access(pbuf, X_OK);
				if (i >= 0) {
					break;	/* Found it! */
				}
			}
		}
	}
	(*env)->ReleaseStringUTFChars(env, jpath, path);

	if (i < 0) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.io.IOException"), "program not found");
		return (-1);
	}


	/* Create the pipes to communicate with the child */
	pipe(in);
	pipe(out);
	pipe(err);

#if defined(HAVE_FORK)
	pid = fork();
#else
	pid = vfork();
#endif
	switch (pid) {
	case 0:
		/* Child */
		dup2(in[0], 0);
		dup2(out[1], 1);
		dup2(err[1], 2);

		close(in[0]);
		close(in[1]);
		close(out[0]);
		close(out[1]);
		close(err[0]);
		close(err[1]);
		break;

	case -1:
		/* Error */
		close(in[0]);
		close(in[1]);
		close(out[0]);
		close(out[1]);
		close(err[0]);
		close(err[1]);
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.io.IOException"), "fork failed");
		return (-1);

	default:
		/* Parent */
		uc = (*env)->GetObjectClass(env, this);
		(*env)->SetIntField(env, this, (*env)->GetFieldID(env, uc, "stdin_fd", "I"), fixfd(in[1]));
		(*env)->SetIntField(env, this, (*env)->GetFieldID(env, uc, "stdout_fd", "I"), fixfd(out[0]));
		(*env)->SetIntField(env, this, (*env)->GetFieldID(env, uc, "stderr_fd", "I"), fixfd(err[0]));
		close(in[0]);
		close(out[1]);
		close(err[1]);
		return (pid);
	}

	/* Child: execute new program ... */

	/* Build arguments and environment */
	argv = gc_calloc_fixed(arglen + 1, sizeof(char*));
	for (i = 0; i < arglen; i++) {
		argv[i] = (char*)(*env)->GetStringUTFChars(env, (*env)->GetObjectArrayElement(env, args, i), 0);
	}
	arge = gc_calloc_fixed(envlen + 1, sizeof(char*));
	for (i = 0; i < envlen; i++) {
		arge[i] = (char*)(*env)->GetStringUTFChars(env, (*env)->GetObjectArrayElement(env, envs, i), 0);
	}
	/* Execute program */
	execve(pbuf, argv, arge);
	exit(-1);
#else
	unimp("fork() or execve() not provided");
#endif
	return (-1);
}

JNIEXPORT void JNICALL
Java_kaffe_lang_UNIXProcess_destroy0(JNIEnv* env, jobject this, jint pid)
{
#if defined(HAVE_KILL)
	kill(pid, SIGTERM);
#else
	unimp("kill() not provided");
#endif
}
