/*
 * kaffe.net.DNS.c
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: DNS.c,v 1.4 2001/07/23 23:03:12 samc Exp $
 */

#include <jni.h>

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-net.h"
#include "files.h"
#include "jsyscall.h"
#include "jmalloc.h"
#include "../../../kaffe/kaffevm/support.h"
#include "nets.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static JNIEnv* genv;
static jmethodID callbackM;

#if defined(ASYNC_DNS)
#define	QUERYFLAGS	(adns_qf_search)
#endif

static
void
callback(void* context, char* hostname, int ip)
{
	jstring host;

	host = 0;
	if (hostname != 0) {
		host = (*genv)->NewStringUTF(genv, hostname);
	}
	(*genv)->CallObjectMethod(genv, (jobject)context, callbackM, host, ntohl(ip));
}

JNIEXPORT void JNICALL
Java_kaffe_net_DNS_init0(JNIEnv* env, jclass dns)
{
	genv = env;
	callbackM = (*env)->GetMethodID(env, dns, "callback", "(Ljava/lang/String;I)V");
#if defined(ASYNC_DNS)
	adns_vec = callback;
#endif
}

JNIEXPORT void JNICALL
Java_kaffe_net_DNS_getHostByAddr0(JNIEnv* env, jobject dns, jint address)
{
#if !defined(ASYNC_DNS)
	callback(dns, gethostbyaddr(htonl(address)), 0);
#else
	struct sockaddr_in addr;
	adns_query query;

#if defined(BSD44)
	addr.sin_len = sizeof(addr);
#endif
	addr.sin_family = AF_INET;
	SET_ADDRESS(addr.sin_addr, htonl(address));

	adns_submit_reverse(adns, (struct sockaddr*)&addr, adns_r_ptr, QUERYFLAGS, dns, &query);
#endif
}

JNIEXPORT void JNICALL
Java_kaffe_net_DNS_getAddrByHost0(JNIEnv* env, jobject dns, jstring hostname)
{
#if !defined(ASYNC_DNS)
	char* name;
	int* entries;

	name = (char*)(*env)->GetStringUTFChars(env, hostname, 0);
	entries = gethostbyname(name);
	(*env)->ReleaseStringUTFChars(env, hostname, name);

	callback(dns, 0, (entries == 0 ? 0 : entries[0]));
	if (entries != 0) {
		free(entries);
	}
#else
	char* name;
	adns_query query;

	name = (char*)(*env)->GetStringUTFChars(env, hostname, 0);

	adns_submit(adns, name, adns_r_a, QUERYFLAGS, dns, &query);

	(*env)->ReleaseStringUTFChars(env, hostname, name);
#endif

}

JNIEXPORT jstring JNICALL
Java_kaffe_net_DNS_getLocalHostname(JNIEnv* env, jclass dns)
{
        char* hostname;

        hostname = gethostname();
	if (hostname != 0) {
		return ((*env)->NewStringUTF(env, hostname));
	}
	else {
		return (0);
	}
}
