/*
 * kaffe.net.NetIO.c
 *
 * Copyright (c) 2001
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * $tvt: NetIO.c,v 1.8 2001/08/05 08:01:59 tim Exp $
 */

#include <jni.h>

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-net.h"
#include "files.h"
#include "jsyscall.h"
#include "../../../kaffe/kaffevm/support.h"
#include "nets.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static jclass socketexception;
static jclass dnsclass;
static jclass pktclass;
static jfieldID dnsaddr;
static jfieldID pktport;

JNIEXPORT void JNICALL
Java_kaffe_net_NetIO_init0(JNIEnv* env, jclass netio)
{
	socketexception = (*env)->FindClass(env, "java.net.SocketException");
	dnsclass = (*env)->FindClass(env, "kaffe.net.DNS");
	pktclass = (*env)->FindClass(env, "java.net.DatagramPacket");
	dnsaddr = (*env)->GetFieldID(env, dnsclass, "address", "I");
	pktport = (*env)->GetFieldID(env, pktclass, "port", "I");
}

/*
 * Create a socket.
 */
JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_socket0(JNIEnv* env, jobject netio, jboolean stream)
{
	int fd = socket(stream);
#if defined(SOL_SOCKET) && defined(SO_BROADCAST)
	/* Turn on broadcast support for datagram sockets */
	if (fd >= 0 && !stream) {
		int brd = 1;
		setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &brd, sizeof(brd));
	}
#endif
	return (fd);
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_bind0(JNIEnv* env, jobject netio, jint fd, jint address, jint port)
{
	int on = 1;
	struct sockaddr_in addr;
	int r;

	/* Allow rebinding to a socket - ignore errors */
	(void)setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on));

#if defined(BSD44)
        addr.sin_len = sizeof(addr);
#endif
	addr.sin_family = AF_INET;
        addr.sin_port = htons((short)port);
	SET_ADDRESS(addr.sin_addr, htonl(address));

	r = bind(fd, (struct sockaddr*)&addr, sizeof(addr));
	return (r);
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_listen0(JNIEnv* env, jobject netio, jint fd, jint count)
{
	return (listen(fd, count));
}

JNIEXPORT void JNICALL
Java_kaffe_net_NetIO_connect0(JNIEnv* env, jobject netio, jint fd, jint address, jint port)
{
	struct sockaddr_in addr;

#if defined(BSD44)
	addr.sin_len = sizeof(addr);
#endif
	addr.sin_family = AF_INET;
	addr.sin_port = htons((short)port);
	SET_ADDRESS(addr.sin_addr, htonl(address));

	if (connect(fd, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		(*env)->ThrowNew(env, socketexception, SYS_ERROR);
	}
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_accept0(JNIEnv* env, jobject netio, jint parentfd, jint address, jint port)
{
	struct sockaddr_in addr;
	int alen;

#if defined(BSD44)
	addr.sin_len = sizeof(addr);
#endif
	addr.sin_family = AF_INET;
	addr.sin_port = htons((short)port);
	SET_ADDRESS(addr.sin_addr, htonl(address));
	alen = sizeof(addr);

	return (accept(parentfd, (struct sockaddr*)&addr, &alen));
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_getSocketInfo0(JNIEnv* env, jobject netio, jint fd, jint item)
{
	struct sockaddr_in addr;
	int alen;
	int r;

	alen = sizeof(addr);
	if (item < 2) {	/* LOCAL */
		r = getsockname(fd, (struct sockaddr*)&addr, &alen);
	}
	else { /* REMOTE */
		r = getpeername(fd, (struct sockaddr*)&addr, &alen);
	}

	if (r < 0) {
		return (-1);
	}
	else if (item == 0 || item == 2) {
		return (htonl(GET_ADDRESS(addr.sin_addr)));
	}
	else {
		return (ntohs(addr.sin_port));
	}
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_setSocketOption0(JNIEnv* env, jobject netio, jint fd, jint option, jint value)
{
	int r;

	switch (option) {
#if defined(SO_SNDBUF)
	case 0x1001:	/* SO_SNDBUF */
		r = setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &value, sizeof(value));
		break;
#endif
#if defined(SO_RCVBUF)
	case 0x1002:	/* SO_RCVBUF */
		r = setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &value, sizeof(value));
		break;
#endif
#if defined(SO_LINGER)
	case 0x0080:	/* SO_LINGER */
		{
			struct linger linger;
			if (value > 0) {
				linger.l_onoff = 1;
				linger.l_linger = value;
			}
			else {
				linger.l_onoff = 0;
				linger.l_linger = 0;
			}
			r = setsockopt(fd, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger));
		}
		break;
#endif
#if defined(SO_REUSEADDR)
	case 0x0004:	/* SO_REUSEADDR */
		r = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(value));
		break;
#endif
#if defined(TCP_NODELAY)
	case 0x0001:	/* TCP_NODELAY */
		r = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &value, sizeof(value));
		break;
#endif
	case 0x1006:	/* SO_TIMEOUT */
		r = setreadtimeout(fd, value);
		break;
#if defined(IP_MULTICAST_IF)
	case 0x0010:	/* IP_MULTICAST_IF */
		{
			struct sockaddr_in addr;
#if defined(BSD44)
			addr.sin_len = sizeof(addr);
#endif
			addr.sin_family = AF_INET;
			SET_ADDRESS(addr.sin_addr, htonl(value));
			r = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF, &addr, sizeof(addr));
		}
		break;
#endif
#if defined(IP_ADD_MEMBERSHIP)
	case 0x0023:
		{
			struct ip_mreq ipm;

			SET_ADDRESS(ipm.imr_multiaddr, htonl(value));
			ipm.imr_interface.s_addr = htonl(INADDR_ANY);
			r = setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &ipm, sizeof(ipm));
		}
		break;
#endif
#if defined(IP_DROP_MEMBERSHIP)
	case 0x0024:
		{
			struct ip_mreq ipm;

			SET_ADDRESS(ipm.imr_multiaddr, htonl(value));
			ipm.imr_interface.s_addr = htonl(INADDR_ANY);
			r = setsockopt(fd, IPPROTO_IP, IP_DROP_MEMBERSHIP, &ipm, sizeof(ipm));
		}
		break;
#endif
#if defined(IP_MULTICAST_TTL)
	case 0x0021:
		r = setsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, &value, sizeof(value));
		break;
#endif
	default:
		r = -1;
		break;
	}

	return (r);
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_getSocketOption0(JNIEnv* env, jobject netio, jint fd, jint option)
{
	int r;
	int vsize;
	int value;

	vsize = sizeof(value);

	switch (option) {
#if defined(SO_SNDBUF)
	case 0x1001:	/* SO_SNDBUF */
		r = getsockopt(fd, SOL_SOCKET, SO_SNDBUF, &value, &vsize);
		break;
#endif
#if defined(SO_RCVBUF)
	case 0x1002:	/* SO_RCVBUF */
		r = getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &value, &vsize);
		break;
#endif
#if defined(SO_LINGER)
	case 0x0080:	/* SO_LINGER */
		{
			struct linger linger;
			vsize = sizeof(linger);
			r = getsockopt(fd, SOL_SOCKET, SO_LINGER, &linger, &vsize);
			value = (linger.l_onoff != 0 ? linger.l_linger : 0);
		}
		break;
#endif
#if defined(SO_REUSEADDR)
	case 0x0004:	/* SO_REUSEADDR */
		r = getsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &value, &vsize);
		break;
#endif
#if defined(TCP_NODELAY)
	case 0x0001:	/* TCP_NODELAY */
		r = getsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &value, &vsize);
		break;
#endif
	case 0x1006:	/* SO_TIMEOUT */
		r = getreadtimeout(fd);
		value = r;
		break;
#if defined(IP_MULTICAST_IF)
	case 0x0010:	/* IP_MULTICAST_IF */
		{
			struct sockaddr_in addr;
			vsize = sizeof(addr);
			r = getsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF, &addr, &vsize);
			value = htonl(GET_ADDRESS(addr.sin_addr));
		}
		break;
#endif
#if defined(IP_MULTICAST_TTL)
	case 0x0021:
		r = getsockopt(fd, IPPROTO_IP, IP_MULTICAST_TTL, &value, &vsize);
		break;
#endif
	default:
		r = -1;
		break;
	}

	if (r < 0) {
		return (r);
	}
	else {
		return (value);
	}
}

JNIEXPORT void JNICALL
Java_kaffe_net_NetIO_sendto0(JNIEnv* env, jobject netio, jint fd, jarray buffer, jint len, jint address, jint port)
{
	int r;
	struct sockaddr_in addr;
	char* buf;

#if defined(BSD44)  
	addr.sin_len = sizeof(addr);
#endif
	addr.sin_family = AF_INET;
	addr.sin_port = htons((short)port);
        SET_ADDRESS(addr.sin_addr, htonl(address));

        buf = (*env)->GetByteArrayElements(env, buffer, 0);
	r = sendto(fd, buf, len, 0, (struct sockaddr*)&addr, sizeof(addr));
        (*env)->ReleaseByteArrayElements(env, buffer, buf, JNI_ABORT);

	if (r < 0) {
		(*env)->ThrowNew(env, socketexception, SYS_ERROR);
	}
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_recvfrom0(JNIEnv* env, jobject netio, jint fd, jarray buffer, jint len, jobject dns, jobject pkt)
{
	int r;
	struct sockaddr_in addr;
	char* buf;
	int alen;

	alen = sizeof(addr);
	memset(&addr, 0, alen);
#if defined(BSD44)  
	addr.sin_len = alen;
#endif

        buf = (*env)->GetByteArrayElements(env, buffer, 0);
	r = recvfrom(fd, buf, len, 0, (struct sockaddr*)&addr, &alen);
        (*env)->ReleaseByteArrayElements(env, buffer, buf, JNI_COMMIT);

	if (r < 0) {
		(*env)->ThrowNew(env, socketexception, SYS_ERROR);
	}

	(*env)->SetIntField(env, dns, dnsaddr, GET_ADDRESS(addr.sin_addr));
	(*env)->SetIntField(env, pkt, pktport, ntohs(addr.sin_port));
	return (r);
}

JNIEXPORT jint JNICALL
Java_kaffe_net_NetIO_peek0(JNIEnv* env, jobject netio, jint fd, jobject dns)
{
	int r;
	struct sockaddr_in addr;
	size_t alen;

	alen = sizeof(addr);
	r = recvfrom(fd, 0, 0, MSG_PEEK, (struct sockaddr*)&addr, &alen);
	if (r < 0) {
		(*env)->ThrowNew(env, socketexception, SYS_ERROR);
	}
	(*env)->SetIntField(env, dns, dnsaddr, GET_ADDRESS(addr.sin_addr));
	return (ntohs(addr.sin_port));
}
