/*
 * java.util.zip.Deflater.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include <jni.h>
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/gc.h"
#include "../../../kaffe/kaffevm/exception.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT
#include "Deflater.h"

#if defined(HAVE_LIBZ) && defined(HAVE_ZLIB_H)
#define	IMPLEMENTED 1
#include <zlib.h>
#else
typedef void* voidpf;
typedef unsigned int uInt;
#endif

static
voidpf
dalloc(voidpf opaque, uInt items, uInt size)
{
	return (gc_malloc_fixed(items * size));
}

static
void
dfree(voidpf opaque, voidpf address)
{
	gc_free_fixed(address);
}

#define	WSIZE		0x8000
#define	WSIZEBITS	15

#define GET_STREAM(this)        (*(z_stream**)&unhand(this)->strm)

JNIEXPORT void JNICALL
Java_java_util_zip_Deflater_setDictionary(JNIEnv* env, struct Hjava_util_zip_Deflater* this, HArrayOfByte* buf, jint from, jint len)
{
#if defined(IMPLEMENTED)
	int r;
        z_stream* dstream;

        dstream = GET_STREAM(this);

	r = deflateSetDictionary (dstream, &unhand(buf)->body[from], len);
	if (r < 0) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), dstream->msg ? dstream->msg : "unknown error");
	}
#else
	unimp("java_util_zip_Deflater");
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Deflater_deflate(JNIEnv* env, struct Hjava_util_zip_Deflater* this, HArrayOfByte* buf, jint off, jint len)
{
#if defined(IMPLEMENTED)
	int r;
	int ilen;
        z_stream* dstream;

        dstream = GET_STREAM(this);

	ilen = unhand(this)->len;

	dstream->next_in = &unhand(unhand(this)->buf)->body[unhand(this)->off];
	dstream->avail_in = ilen;

	dstream->next_out = &unhand(buf)->body[off];
	dstream->avail_out = len;

	r = deflate(dstream, unhand(this)->finish ? Z_FINISH : Z_NO_FLUSH);

DBG(	printf("Deflate: in %d left %d out %d status %d\n", ilen, dstream->avail_in, len - dstream->avail_out, r);	)

	switch (r) {
	case Z_OK:
		break;

	case Z_STREAM_END:
		unhand(this)->finished = 1;
		break;

	default:
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), dstream->msg ? dstream->msg : "unknown error");
		return (-1);
	}

	unhand(this)->off += (ilen - dstream->avail_in);
	unhand(this)->len = dstream->avail_in;

	return (len - dstream->avail_out);
#else
	unimp("java_util_zip_Deflater");
	return (-1);
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Deflater_getAdler(JNIEnv* env, struct Hjava_util_zip_Deflater* this)
{
#if defined(IMPLEMENTED)
	return (GET_STREAM(this)->adler);
#else
	unimp("java_util_zip_Deflater");
	return (-1);
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Deflater_getTotalIn(JNIEnv* env, struct Hjava_util_zip_Deflater* this)
{
#if defined(IMPLEMENTED)
	return (GET_STREAM(this)->total_in);
#else
	unimp("java_util_zip_Deflater");
	return (0);
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Deflater_getTotalOut(JNIEnv* env, struct Hjava_util_zip_Deflater* this)
{
#if defined(IMPLEMENTED)
	return (GET_STREAM(this)->total_out);
#else
	unimp("java_util_zip_Deflater");
	return (0);
#endif
}

JNIEXPORT void JNICALL
Java_java_util_zip_Deflater_reset(JNIEnv* env, struct Hjava_util_zip_Deflater* this)
{
#if defined(IMPLEMENTED)
	deflateReset(GET_STREAM(this));

	unhand(this)->finish = 0;
	unhand(this)->finished = 0;
#else
	unimp("java_util_zip_Deflater");
#endif
}

JNIEXPORT void JNICALL
Java_java_util_zip_Deflater_end(JNIEnv* env, struct Hjava_util_zip_Deflater* this)
{
#if defined(IMPLEMENTED)
	z_stream* dstream;

	dstream = GET_STREAM(this);
	deflateEnd(dstream);
	gc_free_fixed(dstream);
#else
	unimp("java_util_zip_Deflater");
#endif
}

JNIEXPORT void JNICALL
Java_java_util_zip_Deflater_init(JNIEnv* env, struct Hjava_util_zip_Deflater* this, jbool val)
{
#if defined(IMPLEMENTED)
	int r;
        z_stream* dstream;

	dstream = gc_malloc_fixed(sizeof(z_stream));
	if (dstream == 0) {
		throwPostedException();
	}
	dstream->next_in = 0;
	dstream->zalloc = dalloc;
	dstream->zfree = dfree;
	dstream->opaque = 0;

	r = deflateInit2(dstream, Z_DEFAULT_COMPRESSION, Z_DEFLATED, (val ? -WSIZEBITS : WSIZEBITS), 9, Z_DEFAULT_STRATEGY);
	if (r != Z_OK) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), dstream->msg ? dstream->msg : "unknown error");
		return;
	}

	GET_STREAM(this) = dstream;
#else
	unimp("java_util_zip_Deflater");
#endif
}
