/*
 * java.util.zip.Inflater.c
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include <native.h>
#include "../../../kaffe/kaffevm/support.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include "../../../kaffe/kaffevm/gc.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT
#include "Inflater.h"

#if defined(HAVE_LIBZ) && defined(HAVE_ZLIB_H)
#define	IMPLEMENTED	1
#include <zlib.h>
#endif

#define	WSIZEBITS	15

#define	GET_STREAM(this)	(*(z_stream**)&unhand(this)->strm)

JNIEXPORT void JNICALL
Java_java_util_zip_Inflater_setDictionary(JNIEnv* env, struct Hjava_util_zip_Inflater* this, HArrayOfByte* buf, jint from, jint len)
{
#if defined(IMPLEMENTED)
	int r;
	z_stream* dstream;

	dstream = GET_STREAM(this);
	r = inflateSetDictionary (dstream, &unhand(buf)->body[from], len);
	if (r < 0) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), dstream->msg ? dstream->msg : "unknown error");
	}
#else
	unimp("java_util_zip_Inflate");
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Inflater_inflate(JNIEnv* env, struct Hjava_util_zip_Inflater* this, HArrayOfByte* buf, jint off, jint len)
{
#if defined(IMPLEMENTED)
	int r;
	int ilen;
	z_stream* dstream;

	dstream = GET_STREAM(this);

	ilen = unhand(this)->len;

	/* No data - force us to get some more */
	if (ilen == 0) {
		return (0);
	}

	dstream->next_in = &unhand(unhand(this)->buf)->body[unhand(this)->off];
	dstream->avail_in = ilen;

	dstream->next_out = &unhand(buf)->body[off];
	dstream->avail_out = len;

	r = inflate(dstream, Z_PARTIAL_FLUSH);

	switch (r) {
	case Z_OK:
		break;

	case Z_STREAM_END:
		unhand(this)->finished = 1;
		break;

	case Z_NEED_DICT:
		unhand(this)->needsDictionary = 1;
		break;

	default:
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), dstream->msg ? dstream->msg : "unknown error");
		return (-1);
	}

	unhand(this)->off += (ilen - dstream->avail_in);
	unhand(this)->len = dstream->avail_in;

	return (len - dstream->avail_out);
#else
	unimp("java_util_zip_Inflate");
	return (-1);
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Inflater_getAdler(JNIEnv* env, struct Hjava_util_zip_Inflater* this)
{
#if defined(IMPLEMENTED)
	return (GET_STREAM(this)->adler);
#else
	unimp("java_util_zip_Inflate");
	return (-1);
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Inflater_getTotalIn(JNIEnv* env, struct Hjava_util_zip_Inflater* this)
{
#if defined(IMPLEMENTED)
	return (GET_STREAM(this)->total_in);
#else
	unimp("java_util_zip_Inflate");
	return (0);
#endif
}

JNIEXPORT jint JNICALL
Java_java_util_zip_Inflater_getTotalOut(JNIEnv* env, struct Hjava_util_zip_Inflater* this)
{
#if defined(IMPLEMENTED)
	return (GET_STREAM(this)->total_out);
#else
	unimp("java_util_zip_Inflate");
	return (0);
#endif
}

JNIEXPORT void JNICALL
Java_java_util_zip_Inflater_reset(JNIEnv* env, struct Hjava_util_zip_Inflater* this)
{
#if defined(IMPLEMENTED)
	inflateReset(GET_STREAM(this));

	unhand(this)->finished = 0;
	unhand(this)->len = 0;
#else
	unimp("java_util_zip_Inflate");
#endif
}

JNIEXPORT void JNICALL
Java_java_util_zip_Inflater_end(JNIEnv* env, struct Hjava_util_zip_Inflater* this)
{
#if defined(IMPLEMENTED)
	z_stream* dstream;

	dstream = GET_STREAM(this);
	inflateEnd(dstream);
	gc_free_fixed(dstream);
#else
	unimp("java_util_zip_Inflate");
#endif
}

JNIEXPORT void JNICALL
Java_java_util_zip_Inflater_init(JNIEnv* env, struct Hjava_util_zip_Inflater* this, jbool val)
{
#if defined(IMPLEMENTED)
	int r;
	z_stream* dstream;

	dstream = gc_malloc_fixed(sizeof(z_stream));
	dstream->next_out = 0;
	dstream->zalloc = 0;
	dstream->zfree = 0; 
	dstream->opaque = 0;

	r = inflateInit2(dstream, val ? -WSIZEBITS : WSIZEBITS);
	if (r != Z_OK) {
		(*env)->ThrowNew(env, (*env)->FindClass(env, "java.lang.Error"), dstream->msg ? dstream->msg : "unknown error");
		return;
	}
	GET_STREAM(this) = dstream;
#else
	unimp("java_util_zip_Inflate");
#endif
}
