/*
 * java.util.zip.ZipFile.c
 *
 * Copyright (c) 1996, 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "../../../kaffe/kaffevm/gtypes.h"
#include <jni.h>
#include "../../../kaffe/kaffevm/jar/jar.h"
#include "../../../kaffe/kaffevm/itypes.h"
#include "../../../kaffe/kaffevm/gc.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

static
jobject
makeZipEntry(JNIEnv* env, jarEntry* entry)
{
	static jclass zeclass;
	static jmethodID zeinit;
	static jfieldID zename;
	static jfieldID zesize;
	static jfieldID zemethod;
	static jfieldID zecsize;
	static jfieldID zeoffset;
	jobject zentry;

	if (zeclass == 0) {
		zeclass = (*env)->FindClass(env, "java.util.zip.ZipEntry");
		zeinit = (*env)->GetMethodID(env, zeclass, "<init>", "()V");
		zename = (*env)->GetFieldID(env, zeclass, "name", "java.lang.String");
		zesize = (*env)->GetFieldID(env, zeclass, "size", "J");
		zemethod = (*env)->GetFieldID(env, zeclass, "method", "I");
		zecsize = (*env)->GetFieldID(env, zeclass, "csize", "J");
		zeoffset = (*env)->GetFieldID(env, zeclass, "offset", "J");
	}

	zentry = (*env)->NewObject(env, zeclass, zeinit);

	(*env)->SetObjectField(env, zentry, zename, (*env)->NewStringUTF(env, entry->fileName));
	(*env)->SetLongField(env, zentry, zesize, entry->uncompressedSize);
	(*env)->SetIntField(env, zentry, zemethod, entry->compressionMethod);
	(*env)->SetLongField(env, zentry, zecsize, entry->compressedSize);
	(*env)->SetLongField(env, zentry, zeoffset, entry->dataPos);

	return (zentry);
}

JNIEXPORT jobject JNICALL
Java_java_util_zip_ZipFile_openZipFile0(JNIEnv* env, jclass zp, jstring fname)
{
	jarFile* zip;
	char* str;

	str = (char*)(*env)->GetStringUTFChars(env, fname, 0);
	zip = openJarFile(str);
	(*env)->ReleaseStringUTFChars(env, fname, str);
	return ((jobject)zip);
}

JNIEXPORT void JNICALL
Java_java_util_zip_ZipFile_closeZipFile0(JNIEnv* env, jclass zp, jobject zip)
{
	closeJarFile((jarFile*)zip);
}

JNIEXPORT jobject JNICALL
Java_java_util_zip_ZipFile_getZipEntry0(JNIEnv* env, jclass zp, jobject zip, jstring zname)
{
	jarEntry* entry;
	char* str;

	str = (char*)(*env)->GetStringUTFChars(env, zname, 0);
	entry = lookupJarFile((jarFile*)zip, str);
	(*env)->ReleaseStringUTFChars(env, zname, str);
	if (entry == 0) {
		return (0);
	}
	return (makeZipEntry(env, entry));
}

JNIEXPORT jarray JNICALL
Java_java_util_zip_ZipFile_getZipData0(JNIEnv* env, jclass zp, jobject zip, jlong size, jint method, jlong csize, jlong offset)
{
	jarEntry entry;
	jarray array;
	jbyte* body;

	entry.fileName = "";
	entry.uncompressedSize = size;
	entry.compressionMethod = method;
	entry.compressedSize = csize;
	entry.dataPos = (jint)offset;

	array = (*env)->NewByteArray(env, size);
	if (array != 0) {
		body = (*env)->GetByteArrayElements(env, array, 0);
		if (getDataJarFile((jarFile*)zip, &entry, body) != 0) {
			(*env)->ReleaseByteArrayElements(env, array, body, JNI_COMMIT);
		}
		else {
			array = 0;
		}
	}
	return (array);
}

JNIEXPORT jobject JNICALL
Java_java_util_zip_ZipFile_getZipEntries0(JNIEnv* env, jclass zp, jobject zip)
{
	static jclass vecc;
	static jmethodID pmeth;
	static jmethodID imeth;
	jobject vec;
	jarFile* zfile;
	jarEntry* entry;
	int n;

	if (vecc == 0) {
		vecc = (*env)->FindClass(env, "java.util.Vector");
		pmeth = (*env)->GetMethodID(env, vecc, "addElement", "(Ljava/lang/Object;)V");
		imeth = (*env)->GetMethodID(env, vecc, "<init>", "(I)V");
	}

	zfile = (jarFile*)zip;
	vec = (*env)->NewObject(env, vecc, imeth, zfile->count);
	for (n = 0; n < JARFILE_HASHTABLE_SIZE; n++) {
		for (entry = zfile->head[n]; entry != NULL; entry = entry->next) {
			(*env)->CallVoidMethod(env, vec, pmeth, makeZipEntry(env, entry));
		}
	}
	return (vec);
}
