package java.awt;

import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Vector;
import kaffe.awt.Defaults;

/**
 * Copyright (c) 1998
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */
public class BarMenu
  extends Component
  implements ShortcutConsumer
{
	static int Dx = 5;
	Menu selection;
	MenuBar mb;
	PopupMenu current;
	WindowListener wa;

public BarMenu ( MenuBar mb) {
	this.mb = mb;
	
	setForeground( Defaults.MenuTxtClr);
	setBackground( Defaults.MenuBgClr);
	setFont( Defaults.MenuFont);

	flags |= IS_MOUSE_AWARE;
}

public void addNotify () {
	if ( (flags & IS_ADD_NOTIFIED) == 0 ) {
		super.addNotify();

		mb.owner = getToplevel();
		mb.parent = this;
		mb.addNotify();
		mb.propagateOldEvents( ((flags & IS_OLD_EVENT) != 0));

		registerShortcuts();
	}
}

void disposeCurrent () {
	if ( current != null){
		unlinkCurrent();
		current.dispose();
		current = null;
	}
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), false);
}

public Graphics getGraphics () {
	int u = x;
	int v = y;
	
	if ( (Toolkit.flags & Toolkit.EXTERNAL_DECO) != 0 ){
		u -= parent.insets.left;
		v -= parent.insets.top;
	}

	// we can't use Component.getGraphics() here, because this is clipped on the
	// insetted parents. Since Frame insets include the BarMenu, we need to bypass
	// this (to avoid getting clipped away)
	return NativeGraphics.getGraphics( parent, ((Window)parent).nativeData,
	                                   NativeGraphics.TGT_TYPE_WINDOW,
	                                   (parent.flags & Component.IS_VISIBLE) != 0,
	                                   u, v, 0, 0, width, height,
	                                   parent.fgClr, parent.bgClr, parent.font);
}

int getX ( Menu m) {
	int sz = mb.menus.size();
	int x0 = 0;
	
	for ( int i=0; i<sz; i++) {
		Menu cm = (Menu)mb.menus.elementAt( i);
		if ( cm == m)
			return x0;
		x0 += cm.getWidth()+2*Dx;
	}
	
	return -1;		
}

public void handleShortcut( MenuShortcut ms) {
	switch ( ms.keyCode) {
		case KeyEvent.VK_LEFT:
			selectPrev();
			break;
		case KeyEvent.VK_RIGHT:
			selectNext();
			break;
		case KeyEvent.VK_ENTER:
			processSelection();
			break;
	}
}

public boolean isFocusTraversable () {
	return false;
//	return super.isFocusTraversable();
}

Menu menuAt ( int x) {
	int sz = mb.menus.size();
	int x0 = Dx;
	
	for ( int i=0; i<sz; i++) {
		Menu m = (Menu)mb.menus.elementAt( i);
		int mw = m.getWidth();
		x0 += mw+Dx;
		if ( x0 > x) {
			return m;
		}
		x0 += Dx;
	}
	
	return null;
}

boolean openSelection () {
	if ( (selection == null) || ((selection.getItemCount() == 0)) )
		return false;
	
	disposeCurrent();
	current = new PopupMenu( selection);
	current.show( this, getX( selection), height);

	MenuShortcut s1 = new MenuShortcut( this, KeyEvent.VK_LEFT, 0);
	MenuShortcut s2 = new MenuShortcut( this, KeyEvent.VK_RIGHT, 0);
	ShortcutHandler.addShortcut( s1, current.wnd, this);
	ShortcutHandler.addShortcut( s2, current.wnd, this);
	current.wnd.addWindowListener( wa);

	return true;		
}

public void paint ( Graphics g) {
	int sz = mb.menus.size();
	int dx2 = 2 * Dx;
	int x0 = 1;
	int y0 = 1;
	int mh = height - 2;

	g.setColor( getBackground());
	g.fill3DRect( 0, 0, width, height, true);
	
	for ( int i=0; i<sz; i++) {
		Menu m = (Menu)mb.menus.elementAt( i);
		int mw = m.getWidth() +dx2;

		m.paint( g, x0, Dx, y0, mw, mh, bgClr, fgClr, selection == m);
		x0 += mw;
	}
}

void paintMenu ( Graphics g, Menu m) {
	int sz = mb.menus.size();
	int x0 = 1;
	int dx2 = 2 * Dx;
	int y0 = 1;
	int mh = height - 2;
	
	for ( int i=0; i<sz; i++) {
		Menu cm = (Menu)mb.menus.elementAt( i);
		int mw = cm.getWidth() + dx2;

		if ( m == cm) {
			cm.paint( g, x0, Dx, y0, mw, mh, bgClr, fgClr, selection == cm);
			return;
		}
		x0 += mw;
	}
		
}

void processMotion( MouseEvent e) {
	switch ( e.id) {
		case MouseEvent.MOUSE_MOVED:
			if ( (kaffe.awt.AWT.keyTgt == e.getSource()) || (current != null) )
				selectMenu( menuAt( e.getX() ));
			break;
	}

	super.processMotion( e);
}

void processMouse( MouseEvent e) {
	switch (e.id) {
		case MouseEvent.MOUSE_EXITED:
			if ( current == null )
				selectMenu( null);
			break;
		case MouseEvent.MOUSE_PRESSED:
			requestFocus();
			selectMenu( menuAt( e.getX() ));
			break;
		case MouseEvent.MOUSE_RELEASED:
			processSelection();
			break;
	}
	
	super.processMouse( e);
}

void processSelection () {
	if ( selection != null) {
		if ( !openSelection() ) {
			selection.handleShortcut( null);
			selectMenu( null);
		}
	}	
}

void registerShortcuts () {
	MenuShortcut s1 = new MenuShortcut( this, KeyEvent.VK_LEFT, 0);
	MenuShortcut s2 = new MenuShortcut( this, KeyEvent.VK_RIGHT, 0);
	MenuShortcut s3 = new MenuShortcut( this, KeyEvent.VK_ENTER, 0);
	ShortcutHandler.addShortcut( s1, this, this);
	ShortcutHandler.addShortcut( s2, this, this);
	ShortcutHandler.addShortcut( s3, this, this);
}

public void removeNotify () {
	super.addNotify();
	mb.removeNotify();
	
	ShortcutHandler.removeShortcuts( this);
}

void selectMenu ( Menu m) {
	Menu ms = selection;
	
	if ( m == selection )
		return;
	
	Graphics g = getGraphics();

	if ( m == null ) {
		disposeCurrent();
		selection = null;
		paintMenu( g, ms);
	}
	else if ( selection == null ) {
		selection = m;
		paintMenu( g, m);
		openSelection();
	}
	else {
		selection = m;
		paintMenu( g, ms);
		paintMenu( g, m);
		disposeCurrent();
		openSelection();
	}
	g.dispose();
}

void selectNext () {
	try {
		if ( selection == null)
			selectMenu( (Menu)mb.menus.firstElement());
		else {
			int nIdx = mb.menus.indexOf( selection) + 1;
			selectMenu( (Menu)mb.menus.elementAt( nIdx));
		}
	}
	catch ( Exception e) {}
}

void selectPrev () {
	try {
		if ( selection == null)
			selectMenu( (Menu)mb.menus.lastElement());
		else {
			int nIdx = mb.menus.indexOf( selection) - 1;
			selectMenu( (Menu)mb.menus.elementAt( nIdx));
		}
	}
	catch ( Exception e) {}
}

void setMenus ( MenuBar mb) {
	this.mb = mb;
	if ( isShowing() )
		repaint();
}

void unlinkCurrent () {
	if ( (current != null) && (current.wnd != null ) ) {
		ShortcutHandler.removeShortcuts( current.wnd);
		current.wnd.removeWindowListener( wa);
	}
}
}
