package java.awt;


/**
 * Canvas - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author J. Mehlitz
 */
public class Canvas
  extends Component
{
	final private static long serialVersionUID = -2284879212465893870L;
	static int named;

public Canvas() {
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "canvas" + (named++);
}

public Graphics getGraphics () {
	if ( (flags & IS_ADD_NOTIFIED) != 0 ){
		NativeGraphics g = NativeGraphics.getClippedGraphics( null, this,
		                                                      0, 0, 0, 0,
		                                                      width, height);		
		if ( g != null ){
			// FIX ME - check for parent state (overlapping IS_NATIVE_LIKES)
			g.clipSiblings( this, (IS_SHOWING | IS_NATIVE_LIKE));

			linkGraphics( g);
		}

		return g;
	}
	else {
		return null;
	}
}

public boolean isFocusTraversable () {
	// for some obscure reason, Canvases are not focusTraversable by default
	return false;
}

public void paint ( Graphics g ) {
	// for some strange reason, the JDK Canvas paint() seems to clear the background
	
	g.clearRect( 0, 0, width, height); 
}

//public void paint( Graphics g) {
//	// Canvas is a nativeLike Component, i.e. its background would
//	// normally be blanked by the native window system
//	g.clearRect( 0, 0, width, height);
//}
void processPaintEvent ( int id, int ux, int uy, int uw, int uh ) {
	NativeGraphics g = NativeGraphics.getClippedGraphics( null, this, 0,0,
	                                                      ux, uy, uw, uh);
	if ( g != null ){
	
		// FIX ME should first check for parent state (overlapping IS_NATIVE_LIKES)
		g.clipSiblings( this, (IS_SHOWING | IS_NATIVE_LIKE));
	
		if ( id == PaintEvt.UPDATE ) {
			update( g);
		}
		else {
			// the rudimentary doc on PAINT says the background can be assumed to be already
			// cleared, which for the JDK is probably done in the native toolkit
			g.clearRect( 0, 0, width, height);

			paint( g);
		}
		g.dispose();
	}
}
}
