package java.awt;

import java.awt.event.ComponentEvent;
import java.awt.event.PaintEvent;
import java.awt.event.WindowEvent;
import kaffe.util.Condition;

/**
 *
 * Copyright (c) 1998
 *   Transvirtual Technologies Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 * @author P.C.Mehlitz
 */
class EventDispatchThread
  extends Thread
{
	boolean stop;
	EventQueue queue;
	int counter;

EventDispatchThread ( ThreadGroup tg, EventQueue queue ) {
	super( tg, "AWT-EventQueue-0" ); // some apps depend on this JDK thread name

	this.queue = queue;
	
	// while we dispatch, we don't want the process to exit (regardless of threadgroup)
	setDaemon( false);
	
	// mote that the JDK uses NORM_PRIORITY, and we shouldn't change that (or we risk
	// different event orders)
}

public void run () {
	AWTEvent e;

	while ( !stop ) {
		// the inner loop protects us from being disrupted by
		// an exception (we should continue to dispatch as long as possible)
		try {
			while ( !stop ) {			
				if ( (e = queue.getNextEvent()) != null ){
					counter++;
					
					e.dispatch();
				}
			}
		}
		catch ( SecurityException sx ) {
			if ( "system_exit".equals( sx.getMessage()) ) {
				// this is from our KaffeServer SecurityManager, ignore
			}
			else {
				Toolkit.tlkBeep();
				sx.printStackTrace( System.err);
			}
		}
		catch ( Throwable x ) {
			Toolkit.tlkBeep();
			x.printStackTrace( System.err);
		}
	}
}

void runUntil ( Condition cond ) {
	AWTEvent e;

	// check if we are outside of the dispatcher thread (to prevent 
	// race conditions)
	if ( Thread.currentThread() != Toolkit.eventThread ) {
		e = WMEvent.getEvent( cond, WMEvent.WM_DISPATCH_MODAL);
		Toolkit.eventQueue.postEvent( e);
		synchronized ( e ) {
			while ( !e.consumed ) {
				try { e.wait(); } catch ( InterruptedException x ) {}
			}
		}
	}
	else {
		while ( !stop ) {
			// the inner loop protects us from being disrupted by
			// an exception (we should continue to dispatch as long as possible)
			try {
				while ( !stop ) {			
					if ( (e = queue.getNextEvent()) != null ){
						counter++;

						// treelock sync has to go into sys calls (like computeMouseTarget)						
						e.dispatch();

						if ( cond.evaluate() ){
							return;
						}
					}
				}
			}
			catch ( SecurityException sx ) {
				if ( "system_exit".equals( sx.getMessage()) ) {
					// this is from our KaffeServer SecurityManager, ignore
				}
				else {
					sx.printStackTrace( System.err);
				}
			}
			catch ( Throwable x ) {
				x.printStackTrace( System.err);
			}
		}
	}
}

public void stopDispatching () {
	stop = true;
}
}
