package java.awt;

import java.awt.peer.FontPeer;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.util.Hashtable;
import java.util.Properties;
import kaffe.awt.Defaults;
import kaffe.util.Ptr;

/**
 * XXX: implement serial form! 
 */
public class Font
  implements Serializable
{
	Ptr nativeData;
	protected String name;
	protected int style;
	protected int size;
	final public static int PLAIN = 0;
	final public static int BOLD = 1;
	final public static int ITALIC = 2;
	final private static long serialVersionUID = -4206021311591459213L;
	static Properties fontMap;

static {
	InputStream is = Font.class.getResourceAsStream( "font.properties");
	
	fontMap = new Properties();
	
	if ( is != null ){
		try {
			fontMap.load( is);
		}
		catch ( IOException x ) {}
	}
}

public Font ( String fntName, int fntStyle, int fntSize ) {
	Object v;
	String spec;
	int i;

	name = (fntName != null) ? fntName : "Default";
	style = fntStyle;
	size = fntSize;

	// we map the Java font name to a spec string by means of a font.properties
	// resource	
	if ( (spec = (String) fontMap.get( name)) == null ){
		spec = (String) fontMap.get( normalize( name));
		if ( spec == null ){
			spec = name;
		}
	}

	nativeData = Toolkit.fntInitFont( spec, style, size + Defaults.FontSizeOffset);
	if ( nativeData == null ) {
		throw new AWTError( "native Font init failed: " + this);
	}
}

/**
 * @rework
 */
public static Font decode ( String fntSpec ) {
	Font     fnt;
	String   name;
	int      style = PLAIN;
	int      size = 12;
	int      i, n, l = 0;
	char     c;
	
	name = fntSpec;

	if ( (i = fntSpec.indexOf( '-')) >= 0 ) {  // format : <name>[-<style>[-<size>]]
		name = fntSpec.substring( 0, i);
		
		i++;
		if ( fntSpec.regionMatches( true, i, "plain-", 0, 6) ){
			l = 6;
		}
		else if ( fntSpec.regionMatches( true, i, "bold-", 0, 5) ){
			style = BOLD;
			l = 5;
		}
		else if ( fntSpec.regionMatches( true, i, "italic-", 0, 7) ){
			style = ITALIC;
			l = 7;
		}
		else if ( fntSpec.regionMatches( true, i, "bolditalic-", 0, 11) ) {
			style = BOLD | ITALIC;
			l = 11;
		}

		if ( l > 0 ) {
			i += l;
			size = 0;
			for ( n = fntSpec.length(); i < n; i++ ) {
				c = fntSpec.charAt( i);
				if ( c >= '0' && c <= '9' )
					size = size*10 + (c - '0');
				else
					break;
			}
		}
	}
	
	fnt = new Font( name, style, size);
	return fnt;
}

String encode () {
	String s;
	
	if ( style == PLAIN ){
		s = "-plain-";
	}
	else if ( style == ITALIC ){
		s = "-italic-";
	}
	else if ( style == BOLD ) {
		s = "-bold-";
	}
	else {
		s = "-bolditalic-";
	}
	
	return (name + s + size);
}

public boolean equals ( Object o ) {
	if ( o instanceof Font ) {
		Font fnt = (Font) o;
		if ( !fnt.name.equals( name) ) return false;
		if ( fnt.style != style ) return false;
		if ( fnt.size != size ) return false;
		return true;
	}

	return false;	
}

protected void finalize () throws Throwable {
  if ( nativeData != null ) {
    Toolkit.fntFreeFont( nativeData);
    nativeData = null;
  }
  super.finalize();
}

public String getFamily() {
	return System.getProperty( ("awt.font." + name.toLowerCase()), name);
}

public static Font getFont ( String key ) {
	return getFont( key, null);
}

public static Font getFont ( String key, Font defFont ) {
	String fSpec;

	if ( (fSpec = System.getProperty( key)) != null )
		return decode( fSpec);

	return defFont;	
}

public String getName() {
	return name;
}

public FontPeer getPeer(){
	return null;
}

public int getSize() {
	return size;
}

public int getStyle() {
	return style;
}

public boolean isBold () {
	return ((style & BOLD) != 0);
}

public boolean isItalic () {
	return ((style & ITALIC) != 0);
}

public boolean isPlain() {
	return (style == 0);
}

static String normalize ( String name ) {
	// This is slow, but we don't want to clutter the properties with all
	// possible combinations. Use the right name!
	
	if ( "Monospaced".equalsIgnoreCase( name) ) return "Monospaced";
	if ( "SansSerif".equalsIgnoreCase( name) ) return "SansSerif";
	if ( "TimesRoman".equalsIgnoreCase( name) ) return "TimesRoman";
	if ( "sans serif".equalsIgnoreCase( name) ) return "SansSerif";
	if ( "times roman".equalsIgnoreCase( name) ) return "TimesRoman";
	if ( "Default".equalsIgnoreCase( name) ) return "Default";
	if ( "Courier".equalsIgnoreCase( name) ) return "Courier";
	if ( "Helvetica".equalsIgnoreCase( name) ) return "Helvetica";
	
	return name;
}

public String toString() {
	String	s = "";

	if ( style == 0 )
		s = "plain";
	else {
		if ( (style & BOLD) != 0 )   s = "bold";
		if ( (style & ITALIC) != 0 ) s += "italic";
	}

	return getClass().getName() + " [" + name + ',' +	s + ',' + size + ']';
}
}
