package java.awt;

import java.awt.event.KeyEvent;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;

class KeyEvt
  extends KeyEvent
{
	static KeyEvt cache;

KeyEvt ( Component src, int evtId, long time, int mods, int kCode, char kChar ) {
	super( src, evtId, time, mods, kCode, kChar);
}

protected void dispatch () {
	if ( AWT.keyTgt != null )           // do we have a focus window?
		source = AWT.keyTgt;
	else {
		if ( source == null )
			return;
		AWT.keyTgt = (Component)source;   // hmm, just a defense line
	}

	if ( id == KEY_PRESSED ) {
		accelHint = true;
		if ( keyChar == 0 ){
			// update modifiers for function keys
			switch ( keyCode ) {
			case VK_SHIFT:        inputModifier |= SHIFT_MASK; break;
			case VK_CONTROL:      inputModifier |= CTRL_MASK;  break;
			case VK_ALT:          inputModifier |= ALT_MASK;   break;
			case VK_META:         inputModifier |= META_MASK;
			}
		}

		// we interpret the setEnabled spec in a way that we don't emit events for disabled comps
		if ( (AWT.keyTgt.eventMask & AWTEvent.DISABLED_MASK) == 0 ) {
			AWT.keyTgt.process( this);

			if ( !consumed && !ShortcutHandler.handle( this) ) {
				if ( keyChar != 0 ) {         // printable key
					if ( AWT.keyTgt != null ) {     // maybe a fast finger pulled the keyTgt under our feet
						id = KEY_TYPED;
						keyCode = 0;
						AWT.keyTgt.process( this);
					}
				}
			}
		}	
	}
	else if ( id == KEY_RELEASED ) {	
		accelHint = false;

		if ( (AWT.keyTgt.eventMask & AWTEvent.DISABLED_MASK) == 0 ) {
			AWT.keyTgt.process( this);
		}

		if ( keyChar == 0 ) {
			switch ( keyCode ) {
			case VK_SHIFT:        inputModifier &= ~SHIFT_MASK; break;
			case VK_CONTROL:      inputModifier &= ~CTRL_MASK;  break;
			case VK_ALT:          inputModifier &= ~ALT_MASK;   break;
			case VK_META:         inputModifier &= ~META_MASK;
			}
		}
	}

	if ( (Defaults.RecycleEvents & AWTEvent.KEY_EVENT_MASK) != 0 )
		recycle();
}

boolean equalsKeyType ( KeyEvent e ){
	if ( e instanceof KeyEvt ){
		KeyEvt ke = (KeyEvt) e;
		if ( id != ke.id ) return false;
		if ( keyCode != ke.keyCode ) return false;
		if ( keyChar != ke.keyChar ) return false;
		if ( modifiers != ke.modifiers ) return false;
	}
	else {
		if ( id != e.getID() ) return false;
		if ( keyCode != e.getKeyCode() ) return false;
		if ( keyChar != e.getKeyChar() ) return false;
		if ( modifiers != e.getModifiers() ) return false;
	}
	
	return true;
}

static synchronized KeyEvt getEvent ( Component source, String param ) {
	KeyEvt    e;
	long      when = System.currentTimeMillis();
	
	if ( cache == null ) {
		e = new KeyEvt( source, 0, 0L, 0, 0, (char)0);
	}
	else {
		e = cache;
		cache = (KeyEvt) e.next;
		
		e.source     = source;
		e.consumed	 = false;
		e.next = null;
	}
	
	e.initFromParam( param);

	return e;
}

static synchronized KeyEvt getEvent ( Component source, int id, int keyCode, int keyChar, int modifier ) {
	KeyEvt    e;
	long      when = System.currentTimeMillis();
	
	if ( cache == null ) {
		e = new KeyEvt( source, id, when, modifier, keyCode, (char)keyChar);
	}
	else {
		e = cache;
		cache = (KeyEvt) e.next;
		e.next = null;
	
		e.id         = id;
		e.source     = source;
		e.when       = when;
		e.modifiers  = modifier;
		e.keyCode    = keyCode;
		e.keyChar    = (char)keyChar;
		e.consumed	 = false;
	}

	return e;
}

static synchronized KeyEvt getEvent ( int srcIdx, int id, int keyCode, int keyChar, int modifier ) {
	KeyEvt    e;
	Component source = sources[srcIdx];
	long      when = System.currentTimeMillis();
	
	// Check for modifier keystrokes which have been "eaten" by the native window system.
	// Unfortunately, this can happen if the window manager temporarily grabs the keyboard
	// (e.g. fvwm2 during an initial window positioning)
	if ( (modifier == 0) && (inputModifier != 0) )
		inputModifier = 0;
	
	if ( cache == null ) {
		e = new KeyEvt( source, id, when, inputModifier, keyCode, (char)keyChar);
	}
	else {
		e = cache;
		cache = (KeyEvt) e.next;
		e.next = null;
	
		e.id         = id;
		e.source     = source;
		e.when       = when;
		e.modifiers  = inputModifier;
		e.keyCode    = keyCode;
		e.keyChar    = (char)keyChar;
		e.consumed	 = false;
	}

	if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
		// this is not used as a direct return value for EventQueue.getNextEvent(), 
		// it has to be Java-queued by the native layer
		Toolkit.eventQueue.postEvent( e);
	}

	return e;
}

boolean initFromParam  ( String param ) {
	// <keyAction> ",keyCode=" <keyCode>
	//   (("," <keyText>) | (",keyChar='" <keyChar> "'"))
	//     [",modifiers=" {<mod>} ]		
		
	int kid=0, kcode=0, kmods=0;
	char kchar = 0;
	StringTokenizer st = new StringTokenizer( param);
	String t;
		
	if ( (t = nextToken( st, ", ")) != null ){
		if ( "KEY_PRESSED".equalsIgnoreCase( t) )       kid = KeyEvent.KEY_PRESSED;
		else if ( "KEY_RELEASED".equalsIgnoreCase( t) ) kid = KeyEvent.KEY_RELEASED;
		else if ( "KEY_TYPED".equalsIgnoreCase( t) )    kid = KeyEvent.KEY_TYPED;
		else                                            kid = 0; // error
		
		if ( (t = nextToken( st, ",= ")) != null ){
			if ( "keyCode".equalsIgnoreCase(t) ){
				if ( (t=nextToken( st, ",= ")) != null ){
					try {
						kcode = Integer.parseInt( t);
					}
					catch ( NumberFormatException x ){
						kcode = 0;
					}
				}
				
				if ( (t = nextToken( st, ",=' ")) != null ){
					if ( "keyChar".equalsIgnoreCase( t) ){
						// Hmm, hopefully that's not still UTF8 here
						if ( (t = nextToken( st, "='")) != null ){
							kchar = t.charAt(0);
						}
					}
					else {
						// skip keyText
					}
					
					if ( (t = nextToken( st, ",=' ")) != null ){
						if ( "modifiers".equalsIgnoreCase( t) ){
							while ( (t = nextToken( st, ",=+ ")) != null ){
								if ( "Meta".equalsIgnoreCase( t) )       kmods |= KeyEvent.META_MASK;
								else if ( "Ctrl".equalsIgnoreCase( t) )  kmods |= KeyEvent.CTRL_MASK;
								else if ( "Alt".equalsIgnoreCase( t) )   kmods |= KeyEvent.ALT_MASK;
								else if ( "Shift".equalsIgnoreCase( t) ) kmods |= KeyEvent.SHIFT_MASK;
							}
						}
					}
				}
			}
		}
	}
		
	id = kid;
	keyCode = kcode;
	keyChar = kchar;
	modifiers = kmods;
	
	return true;
}

private String nextToken ( StringTokenizer st, String delim ) {
	String tok = null;
		
	try {
		if ( st.hasMoreTokens() ){
			tok = st.nextToken( delim);
		}
	}
	catch ( NoSuchElementException x ){
	}

	return tok;
}

protected void recycle () {
	synchronized ( KeyEvt.class ) {
		source = null;

		next  = cache;
		cache = this;
	}
}
}
