package java.awt;


/**
 *
 * Copyright (c) 1998
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 */
public class Panel
  extends Container
{
	final private static long serialVersionUID = -2728009084054400034L;
	private static LayoutManager defaultLayout = new FlowLayout();
	static int named;

public Panel() {
	this( defaultLayout );
}

public Panel( LayoutManager layout) {
	setLayout( layout);
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "panel" + (named++);
}

public Graphics getGraphics () {
	if ( (flags & IS_ADD_NOTIFIED) != 0 ){
		NativeGraphics g = NativeGraphics.getClippedGraphics( null, this,
		                                                      0, 0, 0, 0,
		                                                      width, height);
		if ( g != null ) {
			// FIX ME should first check for parent state (overlapping IS_NATIVE_LIKES)
			g.clipSiblings( this, (IS_SHOWING | IS_NATIVE_LIKE));
			g.clipChildren( this, (IS_SHOWING | IS_NATIVE_LIKE));

			linkGraphics( g);
		}
		return g;
	}
	else {
		return null;
	}
}

public boolean isFocusTraversable () {
	// for some obscure reason, Panels are not focusTraversable by default
	return false;
}

void processPaintEvent ( int id, int ux, int uy, int uw, int uh ) {
	NativeGraphics g = NativeGraphics.getClippedGraphics( null, this, 0,0,
	                                                      ux, uy, uw, uh);
	if ( g != null ){
		// FIX ME should first check for parent state (overlapping IS_NATIVE_LIKES)
		g.clipSiblings( this, (IS_SHOWING | IS_NATIVE_LIKE));
		g.clipChildren( this, (IS_SHOWING | IS_NATIVE_LIKE));

		emitPaintEvents( ux, uy, uw, uh);

		if ( id == PaintEvt.UPDATE ) {
			update( g);
		}
		else {
			// Another anoying anomaly: on some window systems, the
			// background of native Components is automatically blanked (by the
			// native window system or native peer), which is simulated here.
			// Note that we don't do that for UPDATE events, since the update() spec
			// explicitly states "you can assume that the background is not cleared..",
			// and we take that literally! This also shows up in Canvas
			g.clearRect( ux, uy, uw, uh);

			paint( g);
		}
		g.dispose();
	}
}
}
