package java.awt;

import java.awt.image.ImageConsumer;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;

class PiggybackProducer
  extends GenericProducer
  implements ImageObserver
{
	NativeImageProducer refProducer;
	int updateCount;

PiggybackProducer ( NativeImageProducer ref ) {
	refProducer = ref;
}

synchronized void imageAvailable () {
	// get back into 'startProduction' thread
	notify();
}

public boolean imageUpdate ( Image img, int infoflags, int x, int y, int width, int height ) {
	if ( (infoflags & FRAMEBITS) != 0 ){
		synchronized ( this ) {
			updateCount++;
			notify();
		}
	}
	
	return (consumer != null);
}

public void startProduction ( ImageConsumer ic ) {
	Image img = refProducer.img;

	addConsumer( ic);

	if ( img.nativeData == null ) {
		// this will eventually callback 'imageAvailable()'
		if ( refProducer.addDependent( this) ) {
			if ( (img.flags & img.PRODUCING) == 0 ){
				// refProducer not yet queued, kick it off synchronously (saves us a loader thread)
				img.produceSync();
			}
			else {
				// it's in the works, wait for it
				if ( ic instanceof ImageLoader ) {
					ImageLoader ldr = (ImageLoader) ic;
					// watch out, we might be the last active ImageLoader
					ldr.suspendActive();
					waitForImage( img);
					ldr.resumeActive();
				}
				else {
					waitForImage( img);
				}
			}	
			refProducer.removeDependent( this);
		}
	}

	if ( img.nativeData != null ) {
		// Ok, we finally have a ref Image, produce it
		Toolkit.imgProduceImage( this, img.nativeData, true);
	}

	if ( (img.flags & FRAMEBITS) != 0 ){
		// oh my - it's a filtered animation, make sure it's starting to
		// spin around in the animation loop
		img.loadImage( -1, -1, this);
		
		while ( consumer != null ) {
			// we have to do produce in the right thread (which is the current one), so
			// we have to wait until our own imageUpdate (which is executed in the
			// source production thread) triggers us

			synchronized ( this ) {
				try {
					wait();
				}
				catch ( InterruptedException x ) {}
			}
			
			if ( updateCount > 0 ){
				updateCount--;
				Toolkit.imgProduceImage( this, img.nativeData, false);
			}
		}
		
		img.removeObserver( this);
	}

	removeConsumer( ic);
}

synchronized void waitForImage ( Image img ) {
	while ( (img.flags & (ImageObserver.ALLBITS | ImageObserver.FRAMEBITS | ImageObserver.ABORT)) == 0 ){
		try {
			wait();
		}
		catch ( InterruptedException x ) {
		}
	}
}
}
