package java.awt;

import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.util.Hashtable;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;

/**
 * ShortcutHandler - 
 *
 * Copyright (c) 1998
 *    Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 */
class ShortcutHandler
{
	static MenuShortcut[] codeTable = new MenuShortcut[160];
	static Hashtable tgtTable = new Hashtable();
	static boolean nextOne;
	static KeyEvt focusNextKey;
	static KeyEvt focusPrevKey;

static {
	focusNextKey = KeyEvt.getEvent( Window.dummy, Defaults.FocusNextSpec);
	focusPrevKey = KeyEvt.getEvent( Window.dummy, Defaults.FocusPrevSpec);
}

static void addShortcut( MenuShortcut ms, Component owner, ShortcutConsumer consumer ) {

	ms.consumer = consumer;
	ms.next = null;
		
	MenuShortcut msr = (MenuShortcut)tgtTable.get( owner);
	
	if ( msr == null)
		tgtTable.put( owner, ms);
	else {
		for ( ; msr.next != null; msr = msr.next);
		msr.next = ms;
	}
}

static void buildCodeTable( Component tgt) {
	int i, ctl = codeTable.length;
	MenuShortcut me, mt;

	for ( i=0; i<ctl; i++)
		codeTable[i] = null;

	if ( tgt == null )
		return;
		
	//collect codes from all window components
	for ( Component c = tgt; c != null; c = c.parent) {
		for ( me = (MenuShortcut)tgtTable.get( c); me != null; me = me.next) {
			me.nextMod = null;
			mt = codeTable[ me.keyCode];
			if ( mt == null) {
				codeTable[me.keyCode] = me;
			}
			else {
				for (; mt.nextMod != null; mt = mt.nextMod);
				mt.nextMod = me;
			}
		}
	}
}

static Component focusNext ( Container root ) {
	Component c;

	synchronized ( Component.treeLock ){
		nextOne = (AWT.keyTgt == null);
		if ( (c = traverse( root, true)) == null ){
			nextOne = true;
			c = traverse( root, true);
		}
	}

	return c;
}

static Component focusPrev ( Container root ) {
	Component c;

	synchronized ( Component.treeLock ){
		nextOne = (AWT.keyTgt == null);
		if ( (c = traverse( root, false)) == null ){
			nextOne = true;
			c = traverse( root, false);
		}
	}

	return c;
}

static boolean handle( KeyEvent e) {
	Component c, cNext = null;
	Container co;	
	int em = e.getModifiers();
	int cc = e.getKeyCode();
	
	if ( (em & KeyEvent.CTRL_MASK) > 0 ) {
		if ( (cc >= 1) && (cc <= 26) )
			cc += 64;
	}
	
	if ( cc >= codeTable.length )
		return false;

	MenuShortcut ms;		
	for ( ms = codeTable[cc]; ms != null; ms = ms.nextMod) {
		if ( ms.mods == em ) {
			ms.process();
			return true;
		}
	}
			
	c = (Component) e.getSource();
	co = (Container) c.getToplevel();
	
	if ( focusNextKey.equalsKeyType( e) ){
		cNext = focusPrev( co);
		if ( (cNext != null) && (cNext != c) ) {	
			cNext.requestFocus();
		}
		return true;
	}
	if ( focusPrevKey.equalsKeyType( e) ){
		cNext = focusNext( co);
		if ( (cNext != null) && (cNext != c) ) {
			cNext.requestFocus();
		}
		return true;
	}
	
	// some day, we might add other focus-switch hotkeys (e.g. menubar, 1st, next-in-group etc.)

	// default focus tab
	if ( e.getKeyChar() == '\t' ) {
		cNext = e.isShiftDown() ? focusPrev( co) : focusNext( co);
		if ( (cNext != null) && (cNext != c) ){
			cNext.requestFocus();
		}
		return true;
	}

	return false;
}

static void removeFromOwner( Component owner, MenuShortcut msr) {
	MenuShortcut ms, msp = null;
	
	ms = (MenuShortcut)tgtTable.get( owner);
	for (; ms != null; ms = ms.next) {
		if ( msr == ms ) {
			if ( msp == null ) {
				if (msr.next == null ) {
					tgtTable.remove( owner);
				}
				else {
					tgtTable.put( owner, msr.next);
				}
			}
			else {
				msp.next = msr.next;
			}
			return;
		}
		msp = ms;
	}
}

static void removeFromOwner( Component owner, ShortcutConsumer consumer) {
	MenuShortcut ms, msp = null;
	
	ms = (MenuShortcut)tgtTable.get( owner);
	for (; ms != null; ms = ms.next) {
		if ( ms.consumer == consumer ) {
			if ( msp == null ) {
				if (ms.next == null ) {
					tgtTable.remove( owner);
					return;
				}
				else {
					tgtTable.put( owner, ms.next);
				}
			}
			else {
				msp.next = ms.next;
			}
		}
		msp = ms;
	}
}

static void removeShortcuts( Component owner) {
	tgtTable.remove( owner);
}

static int size() {
	return tgtTable.size();
}

static Component traverse ( Component c, boolean isForward ) {

	if ( c instanceof Container ) {
		Container p = (Container) c;
		int i;

		if ( isForward ){
			if ( nextOne && c.isFocusTraversable() ){
				return c;
			}
			nextOne |= (c == AWT.keyTgt);

			for ( i=0; i<p.nChildren; i++ ){
				if ( (c = traverse( p.children[i], isForward)) != null ){
					return c;
				}
			}
		}
		else {
			for ( i=p.nChildren-1; i>=0; i-- ){
				if ( (c = traverse( p.children[i], isForward)) != null ){
					return c;
				}
			}
			if ( nextOne && p.isFocusTraversable() ){
				return p;
			}
			nextOne |= (p == AWT.keyTgt);		
		}
	}
	else {
		if ( nextOne && c.isFocusTraversable() ){
			return c;
		}
		nextOne |= (c == AWT.keyTgt);
	}
	
	return null;
}
}
