package java.awt;

import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.Transferable;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.IndexColorModel;
import java.awt.peer.ComponentPeer;
import java.awt.peer.LightweightPeer;
import java.awt.peer.WindowPeer;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.AccessController;
import java.util.Properties;
import java.util.Vector;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.awt.ToolkitExtension;
import kaffe.awt.theme.ComponentDeco;
import kaffe.io.NullOutputStream;
import kaffe.util.Condition;
import kaffe.util.Ptr;
import kaffe.util.Timer;
import kaffe.util.TimerClient;
import kaffe.util.log.LogClient;
import kaffe.util.log.LogStream;

/**
 * Toolkit - used to be an abstract factory for peers, but since we don't have
 * peers, it is just a simple anchor for central mechanisms (like System-
 * EventQueue etc.) and a wrapper for all native methods. Of course, it is a
 * singleton.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P.C.Mehlitz
 */
class FlushThread
  extends Thread
{
	boolean stop;
	int flushInterval;

FlushThread ( int interval ) {
	super( "AWT-Flusher");
	
	flushInterval = interval;
	
	//setPriority( Thread.MIN_PRIORITY + 1);
	setPriority( Toolkit.eventThread.getPriority());
}

public void run () {
	while ( !stop ) {
		Toolkit.tlkFlush();
		try {
			Thread.sleep( flushInterval);
		}
		catch ( Exception x ) {}
	}
}

void stopFlushing () {
	stop = true;
}
}

class NativeCollector
  extends Thread
{
	boolean initialized;

NativeCollector () {
	super( "AWT-native");
}

native static boolean ncInit ( String displayName );

native static void ncLoop ();

public void run () {
	try {
		// most of the native init probably goes in, here (even though we
		// still have tlkInit() called, but in the wrong thread)
		if ( !ncInit( System.getProperty( "awt.display")) ) {
			throw new AWTError( "native layer init failed");
		}
		
		// wake up the Toolkit.clinit
		initialized = true;
		synchronized ( this ){
			notifyAll();
		}
	}
	catch ( Throwable x ) {
		x.printStackTrace();
	}
		
	// This does never return until we shut down. Encapsulate in a
	// own exception handler so that we don't get thrown out by a
	// native caused exception
	while ( true ){
		try {
			ncLoop();
		}
		catch ( Throwable x ) {
			x.printStackTrace();
		}
	}
}
}

public class Toolkit
{
	static Toolkit singleton;
	static Dimension screenSize;
	static Point screenOffset;
	static int resolution;
	public static EventQueue eventQueue;
	static EventDispatchThread eventThread;
	static NativeClipboard clipboard;
	static ColorModel colorModel;
	static LightweightPeer lightweightPeer = new java.awt.LightweightPeer();
	static WindowPeer windowPeer = new java.awt.WindowPeer();
	static FlushThread flushThread;
	static NativeCollector collectorThread;
	static int flags;
	final static int FAILED = -1;
	final static int IS_BLOCKING = 1;
	final static int IS_DISPATCH_EXCLUSIVE = 2;
	final static int NEEDS_FLUSH = 4;
	final static int NATIVE_DISPATCHER_LOOP = 8;
	final static int EXTERNAL_DECO = 16;
	final static AWTPermission accessEventQueue = new AWTPermission( "accessEventQueue");
	final static AWTPermission accessClipboard = new AWTPermission( "accessClipboard");

public class Extension
  implements ToolkitExtension, TimerClient
{
	int checkCounter;

public void disableRedraw ( Component c ) {
	if ( (c.flags & Component.IS_SHOWING) == Component.IS_SHOWING ) {
		c.flags &= ~Component.IS_PARENT_SHOWING;
		c.propagateParentShowing( true);
	}
}

public void dispatchModalUntil ( Condition cond ){
	eventThread.runUntil( cond);
}

public void drawBackground ( Graphics g, int cx, int cy, int cw, int ch,
                      int x, int y, int w, int h,
                      int xOff, int yOff, Image img, Color clr, int fillMode ) {
	if ( g instanceof NativeGraphics ) {
		NativeGraphics ng = (NativeGraphics) g;
		thmDrawBackground( ng.nativeData, cx, cy, cw, ch, x, y, w, h,
	                     xOff, yOff,
	                     (img != null) ? img.nativeData : null,
	                     (clr != null) ? clr.nativeValue : -1,
	                     fillMode);
	}
}

public void drawBorder ( Graphics g,
                  int x, int y, int w, int h, int alignment,
                  Image itl, Image itm, Image itr,
                  Image iml, Image imr,
                  Image ibl, Image ibm, Image ibr ) {
	if ( g instanceof NativeGraphics ) {
		NativeGraphics ng = (NativeGraphics) g;

		thmDrawBorder ( ng.nativeData,
		                x, y, w, h, alignment,
	                  (itl != null) ? itl.nativeData : null,
	                  (itm != null) ? itm.nativeData : null,
	                  (itr != null) ? itr.nativeData : null,
	                  (iml != null) ? iml.nativeData : null,
	                  (imr != null) ? imr.nativeData : null,
	                  (ibl != null) ? ibl.nativeData : null,
	                  (ibm != null) ? ibm.nativeData : null,
	                  (ibr != null) ? ibr.nativeData : null);
	}
}

public void drawVertString ( Graphics g, String s, int x, int y ) {
	if ( g instanceof NativeGraphics ) {
		NativeGraphics ng = (NativeGraphics) g;
		graDrawVertString( ng.nativeData, s, x, y);
	}
}

public void dump(String prefix, Component c) {
	c.dump(prefix);
}

public void enableRedraw ( Component c ) {
	Component parent = c.parent;

	if ( (parent != null) && ((parent.flags & Component.IS_SHOWING) == Component.IS_SHOWING) ){
		c.flags |= Component.IS_PARENT_SHOWING;
		c.propagateParentShowing( false);	
	}
}

public Component getComponentUnderMouse () {
	Vector v = AWT.windows;
	int ws = v.size();
	int x, y;
	Point mp;
	Component tgt = null;
	
	mp = AWT.mouseTgt.getLocationOnScreen();
	mp.x += AWT.mousePos.x;
	mp.y += AWT.mousePos.y;
	
	for ( int i=0; i<ws; i++ ) {
		Container top = (Container)v.elementAt(i);
		if ( top.isShowing() ) {
			if ( (top.x < mp.x) && (top.x + top.width > mp.x ) &&
				 (top.y < mp.y) && (top.y + top.height > mp.y ) ) {
				x = mp.x - top.x;
				y = mp.y - top.y;
				tgt = MouseEvt.computeMouseTarget( top, x, y);
				if ( tgt != null ){
					break;	
				}
			}
		}
	}
	
	return tgt;
}

public ComponentDeco getDeco ( Component c ) {
	return c.deco;
}

public int getHeight( Component c) {
	return c.height;
}

public Insets getInsets ( Component c ) {
	return c.insets;
}

public Graphics getLinkedGraphics ( Component c ) {

	if ( (c.flags & Component.IS_ADD_NOTIFIED) != 0 ){
		NativeGraphics g = NativeGraphics.getClippedGraphics( null, c,
		                                                      0, 0, 0, 0,
		                                                      c.width, c.height);		
		if ( g != null ){
			c.linkGraphics( g);
		}

		return g;
	}
	else {
		return null;
	}
}

public Point getMouseLocationOnScreen () {
	Point p = new Point( AWT.mousePos.x, AWT.mousePos.y);
	Point pt = AWT.mouseTgt.getLocationOnScreen();
	
	pt.x += p.x;
	pt.y += p.y;
	
	return pt;
}

public Ptr getNativeData ( Object o ) {
	if ( o instanceof NativeGraphics ){
		return ((NativeGraphics)o).nativeData;
	}
	else if ( o instanceof Window ){
		return ((Window)o).nativeData;
	}
	else if ( o instanceof Image ){
		return ((Image)o).nativeData;
	}
	else if ( o instanceof Font ){
		return ((Font)o).nativeData;
	}

	return null;
}

public ComponentDeco getReferenceDeco ( Component c ) {
	return c.getReferenceDeco();
}

public Point getScreenOffset() {
	return screenOffset;
}

public int getWidth( Component c) {
	return c.width;
}

public int getX ( Component c ) {
	return c.x;
}

public int getY ( Component c ) {
	return c.y;
}

public void paintBackground ( Component c, Graphics g ) {
	c.paintBackground( g);
}

public boolean processEvents ( long timeout ) {
	return eventQueue.processEvents( timeout);
}

public void processPaintEvents() {
	eventQueue.processPaintEvents();
}

public void resetDeco(Component c) {
	c.resetDeco();
}

public void setDeco(Component c, ComponentDeco d) {
	c.setDeco(d);
}

public void setMovable ( Window wnd, boolean isMovable ) {
	if ( isMovable ) {
		wnd.flags |= Component.IS_MOVABLE;
	}
	else {
		wnd.flags &= ~Component.IS_MOVABLE;
	}
}

public void setOwner ( Window ownee, Window owner ) {
	ownee.setOwner( owner);
}

public void showCursor ( boolean showIt ) {
	wndShowCursor( showIt);
}

public void startRespawnEventThread ( int timeout ) {
	if ( eventThread == null ){
		startDispatch();
	}

	checkCounter = eventThread.counter;
	eventQueue.postEvent( WMEvent.getEvent( Window.dummy, WMEvent.WM_EMPTY_EVENT));
	Timer.getDefaultTimer().addClient( this, timeout, timeout);
}

public void stopAnimation ( Image img ) {
	img.stopAnimation();
}

public void stopProduction ( Image img ) {
	img.stopProduction();
}

public void stopRespawnEventThread () {
	checkCounter = 0;
	Timer.getDefaultTimer().removeClient( this);
}

public void temporaryRemoveAll ( Container c ) {
	// USE WITH EXTREME CARE - this is just a helper for the pocketlinux
	// XHTMLRenderer! You have to make sure that all children subsequently
	// get added again, or you have a bunch of orphant zombies walking around.
	// This should only be called from within a disableRedraw context, or otherwise
	// there is nothing that prevents these zombies to draw - the parent chain is
	// just broken downwards.
	// This method just exists to prevent a addNotify/removeNotify jitter caused
	// by subsequent layout phases which have to reset child lists, which is
	// harmfull for all Components with expensive initialization/cleanup
	// (like FlashPlayer)
	
	if ( c.nChildren > 0 ) {
		c.invalidate();
	
		c.children = new Component[ c.children.length ];
		c.nChildren = 0;
		
		//c.repaint();  // not required if in a propper disableRedraw context
	}
}

public void timerExpired ( Timer timer ) {
	if ( eventThread != null ){
		if ( checkCounter == eventThread.counter ){
			ThreadGroup tg = eventThread.getThreadGroup();
			if ( tg == null ) {
				// Hmm, just a fallback in case the thing died in the meantime
				tg = Thread.currentThread().getThreadGroup();
			}
			else {
				eventThread.stopDispatching();
			}
			eventThread = new EventDispatchThread( tg, eventQueue);
			eventThread.start();
		}
		else {
			// Ok, it's still alive. Schedule the next test
			checkCounter = eventThread.counter;
			eventQueue.postEvent( WMEvent.getEvent( Window.dummy, WMEvent.WM_EMPTY_EVENT));
		}
	}
	else {
		// nothig to check, anymore
		stopRespawnEventThread();
	}
}

public void setScreenRotation ( int r ) {
        scrSetRotation ( r );       
	screenSize = new Dimension( tlkGetScreenWidth(), tlkGetScreenHeight());
	screenOffset = new Point( tlkGetScreenXoff(), tlkGetScreenYoff());
      
        RootWindow root = RootWindow.getDefaultRootWindow();
	if (root != null) {
		root.width  = screenSize.width;
		root.height = screenSize.height;
       
		root.propagateReshape();
		root.repaintAll();
	}
}

public int getScreenRotation () {
        return scrGetRotation ();
}

public void postFakeNativeEvent ( AWTEvent e ) {
	if ( e instanceof KeyEvent ) {
		Toolkit.eventQueue.postEvent ( new KeyEvt( (Component) e.getSource(),
							   e.getID(),
							   ((KeyEvent) e).getWhen(),
							   ((KeyEvent) e).getModifiers(),
							   ((KeyEvent) e).getKeyCode(),
							   ((KeyEvent) e).getKeyChar()) );
	}
	else if ( e instanceof MouseEvent ) { 
		Toolkit.eventQueue.postEvent ( new MouseEvt( (Component) e.getSource(),
							     e.getID(),
							     ((MouseEvent) e).getWhen(),
							     ((MouseEvent) e).getModifiers(),
							     ((MouseEvent) e).getX(),
							     ((MouseEvent) e).getY(),
							     ((MouseEvent) e).getClickCount(),
							     ((MouseEvent) e).isPopupTrigger()) );
	}
}
}

static {
	try {
		System.loadLibrary( "awt");
		flags = tlkProperties();

		if ( (flags & NATIVE_DISPATCHER_LOOP) != 0 ) {
			// Not much we can do here, we have to delegate the native init
			// to a own thread, which never returns (since it hosts the native
			// dispatcher loop)
			collectorThread = new NativeCollector();
			collectorThread.start();

			// Now it becomes dangerous. We can't sync on Toolkit.class, because we are
			// still in it's clinit, and calling back from collectorThread would end up in
			// a deadlock. However, we do have to wait until the collectorThread is set up
			// correctly, before we can proceed with our own init
			try {
				synchronized ( collectorThread ) {
					while ( !collectorThread.initialized )
						collectorThread.wait();
				}
			}
			catch ( Exception x ) {
				x.printStackTrace();
			}
		}

		// let's do the common Toolkit init (note that tlkInit has to be thread-aware
		// in case of a NATIVE_DISPATCHER_LOOP)
		if ( !tlkInit( System.getProperty( "awt.display")) ) {
			throw new AWTError( "native layer initialization failed");
		}
		initToolkit();
	}
	catch (Error e) {
		e.printStackTrace();
		throw e;
	}
}

protected Toolkit () {
}

public void beep () {
	tlkBeep();
}

native static synchronized void cbdFreeClipboard ( Ptr cbdData );

native static synchronized Transferable cbdGetContents( Ptr cbdData);

native static synchronized Ptr cbdInitClipboard ();

native static synchronized boolean cbdSetOwner ( Ptr cbdData );

public int checkImage(Image image, int width, int height, ImageObserver observer) {
	return (image.checkImage( width, height, observer, false));
}

native static synchronized ColorModel clrGetColorModel();

native static synchronized int clrGetPixelValue ( int rgb );

native static synchronized long clrModBrightness ( int pixValue, int mod );

native static synchronized int clrSetSystemColors ( int[] sysClrs );

public Image createImage ( ImageProducer producer ) {
	return new Image( producer);
}

public Image createImage ( byte[] imageData ) {
	return createImage( imageData, 0, imageData.length);
}

public Image createImage ( byte[] imagedata, int imageoffset, int imagelength ) {
	return new Image( imagedata, imageoffset, imagelength);
}

ComponentPeer createLightweight ( Component c ) {
	// WARNING! this is just a dummy to enable checks like
	// "..getPeer() != null.. or ..peer instanceof LightweightPeer..
  // see createWindow()
	return lightweightPeer;
}

static void createNative ( Component c ) {
	WMEvent e = null;

//	synchronized ( Toolkit.class ) {
		// even if this could be done in a central location, we defer this
		// as much as possible because it might involve polling (for non-threaded
		// AWTs), slowing down the startup time
		if ( eventThread == null ) {
			startDispatch();
		}
//	}

	// do we need some kind of a context switch ?
	if ( (flags & IS_DISPATCH_EXCLUSIVE) != 0 ){
		if ( (flags & NATIVE_DISPATCHER_LOOP) != 0 ){
			if ( Thread.currentThread() != collectorThread ){
				// this is beyond our capabilities (there is no Java message entry we can call
				// in the native collector), we have to revert to some native mechanism
				e =  WMEvent.getEvent( c, WMEvent.WM_CREATE);
				evtSendWMEvent( e);
			}
		}
		else {
			if ( Thread.currentThread() != eventThread ){
				// we can force the context switch by ourselves, no need to go native
				e =  WMEvent.getEvent( c, WMEvent.WM_CREATE);
				eventQueue.postEvent( e);
			}
		}
			
		// Ok, we have a request out there, wait for it to be served
		if ( e != null ) {
			// we should check for nativeData because the event might
			// already be processed (depending on the thread system)
			while ( c.getNativeData() == null ) {
				synchronized ( e ) {
					try { e.wait(); } catch ( InterruptedException x ) {}
				} 
			}
				
			return;
		}
	}
	else {
		// no need to switch threads, go native right away
		c.createNative();
	}
}

protected WindowPeer createWindow ( Window w ) {
	// WARNING! this is just a dummy to enable checks like
	// "..getPeer() != null.. or ..peer instanceof LightweightPeer..
	// it is NOT a real peer support. The peer field just exists to
	// check if a Component already passed its addNotify()/removeNotify().
	// This most probably will be removed once the "isLightweightComponent()"
	// method gets official (1.2?)

	return windowPeer;
}

static void destroyNative ( Component c ) {
	WMEvent e = null;

	// do we need some kind of a context switch ?
	if ( (flags & IS_DISPATCH_EXCLUSIVE) != 0 ){
		if ( (flags & NATIVE_DISPATCHER_LOOP) != 0 ){
			if ( Thread.currentThread() != collectorThread ){
				// this is beyond our capabilities (there is no Java message entry we can call
				// in the native collector), we have to revert to some native mechanism
				e =  WMEvent.getEvent( c, WMEvent.WM_DESTROY);
				evtSendWMEvent( e);
			}
		}
		else {
			if ( Thread.currentThread() != eventThread ){
				// we can force the context switch by ourselves, no need to go native
				e =  WMEvent.getEvent( c, WMEvent.WM_DESTROY);
				eventQueue.postEvent( e);
			}
		}
			
		// Ok, we have a request out there, wait for it to be served
		if ( e != null ) {
			// we should check for nativeData because the event might
			// already be processed (depending on the thread system)
			while ( c.getNativeData() != null ) {
				synchronized ( e ) {
					try { e.wait(); } catch ( InterruptedException x ) {}
				} 
			}
				
			return;
		}
	}
	else {
		// no need to switch threads, go native right away
		c.destroyNative();
	}
}

native static synchronized AWTEvent evtGetNextEvent ();

native static synchronized Component[] evtInit ();

native static synchronized AWTEvent evtPeekEvent ();

native static synchronized AWTEvent evtPeekEventId ( int eventId );

native static synchronized int evtRegisterSource ( Ptr wndData );

native static synchronized void evtSendWMEvent ( WMEvent e );

native static synchronized int evtUnregisterSource ( Ptr wndData );

native static synchronized void evtWakeup ();

native static synchronized int fntBytesWidth ( Ptr fmData, byte[] data, int off, int len );

native static synchronized int fntCharWidth ( Ptr fmData, char c );

native static synchronized int fntCharsWidth ( Ptr fmData, char[] data, int off, int len );

native static synchronized void fntFreeFont ( Ptr fontData );

native static synchronized void fntFreeFontMetrics ( Ptr fmData );

native static synchronized int fntGetAscent ( Ptr fmData);

native static synchronized int fntGetDescent ( Ptr fmData);

native static synchronized int fntGetFixedWidth ( Ptr fmData);

native static synchronized int fntGetHeight ( Ptr fmData);

native static synchronized int fntGetLeading ( Ptr fmData);

native static synchronized int fntGetMaxAdvance ( Ptr fmData);

native static synchronized int fntGetMaxAscent ( Ptr fmData);

native static synchronized int fntGetMaxDescent ( Ptr fmData);

native static synchronized int[] fntGetWidths ( Ptr fmData );

native static synchronized Ptr fntInitFont ( String fntSpec, int style, int size );

native static synchronized Ptr fntInitFontMetrics ( Ptr fontData, FontMetrics metrics );

native static synchronized boolean fntIsWideFont ( Ptr fmData);

native static synchronized int fntStringWidth ( Ptr fmData, String s );

native static synchronized void scrSetRotation ( int r );

native static synchronized int scrGetRotation ();

public ColorModel getColorModel () {
	if ( colorModel == null ){
		colorModel = clrGetColorModel();
	}

	return colorModel;
}

public static Toolkit getDefaultToolkit () {
	if ( singleton == null ){
		singleton = new Toolkit();
	}
	return singleton;
}

protected ToolkitExtension getExtension () {
	return new Extension();
}

public String[] getFontList () {
	String[] list = { "Default", "Monospaced",
	                  "SansSerif", "Serif",
			              "Dialog", "DialogInput", "ZapfDingbats" };
	return list;
}

public FontMetrics getFontMetrics ( Font font ) {
	return FontMetrics.getFontMetrics( font);
}

public Image getImage ( String filename ) {
	File f = new File( filename);

	// Hmm, that's a inconsistency with getImage(URL). The doc isn't very
	// helpful, here (class doc says nothing, book tells us it should return
	// null in case it's not there - but that is not known a priori for URLs).
	// JDK never returns a null object, so we don't do, either
	if (f.exists()) {
		return (new Image(f));
	}
	else {
		try {
			return (new Image(new URL(filename)));
		}
		catch (MalformedURLException _) {
		}
		return (Image.getUnknownImage());
	}
}

public Image getImage ( URL url ) {
	// how can we return 'null' before we start to produce this thing (which
	// isn't supposed to happen here)? We are NOT going to open the URLConnection
	// twice just to check if it is there
	return new Image( url);
}

public int getMenuShortcutKeyMask() {
	return InputEvent.CTRL_MASK;
}

public PrintJob getPrintJob ( Frame frame, String jobtitle, Properties props ) {
	return new PSPrintJob( frame, jobtitle, props);
}

public static String getProperty ( String key, String defaultValue ) {
	return null;
}

static int[] getRgbRequests ( int max ){
	int[] arr = null;
	int n=0;
	int x=0, b;
	boolean skip=false;

	InputStream is = Toolkit.class.getResourceAsStream( "rgbs");

	if ( is != null ){
		arr = new int[max];
	
		try {
			while ( (b = is.read()) > 0 ){
				if ( b < ' ' ){
					arr[n++] = x;
					if ( n >= max ) break;
					x = 0;
					skip = false;
					continue;
				}
				else if ( b == ' ' ){
					continue;
				}
				else if ( b == '#' ){
					skip = true;
				}
				else if ( b <= '9' ){
					b = (b - '0') & 0xf;
				}
				else if ( b >= 'a' ){
					b = (b - 'a' + 10) & 0xf;
				}
				else {
					b = (b - 'A' + 10) & 0xf;
				}
			
				if ( !skip ){
					x <<= 4;
					x += b;
				}
			}
			
			is.close();
		}
		catch ( IOException _ ) {
			System.err.println( "error reading rgbs");
		}
	}

	if ( n < max ){
		arr[n] = -1;
	}

	return arr;
}

public int getScreenResolution () {
	return resolution;
}

public Dimension getScreenSize () {
	return screenSize;
}

public Clipboard getSystemClipboard () {
	AccessController.checkPermission( accessClipboard);

	if ( clipboard == null )
		clipboard = new NativeClipboard();
		
	return clipboard;
}

public EventQueue getSystemEventQueue () {
	AccessController.checkPermission( accessEventQueue);

	return eventQueue;
}

native static synchronized void graAddClip ( Ptr grData, int xClip, int yClip, int wClip, int hClip );

native static synchronized void graClearRect ( Ptr grData, int x, int y, int width, int height, int pixelValue );

native static synchronized void graCopyArea ( Ptr grData, int x, int y, int width, int height,
					   int xDelta, int yDelta );

native static synchronized void graDraw3DRect ( Ptr grData, int x, int y, int width, int height,
		     boolean raised, int pixValue );

native static synchronized void graDrawArc ( Ptr grData, int x, int y, int width, int height, int startAngle, int arcAngle );

native static synchronized void graDrawBytes ( Ptr grData, byte[] data, int offset, int length, int x, int y );

native static synchronized void graDrawChars ( Ptr grData, char[] data, int offset, int length, int x, int y );

native static synchronized void graDrawImage( Ptr grData, Ptr imgData, int src_x, int src_y, int dst_x, int dst_y, int w, int h, int bgval );

native static synchronized void graDrawImageScaled ( Ptr grData, Ptr imgData,
			  int dx0, int dy0, int dx1, int dy1,
			  int sx0, int sy0, int sx1, int sy1, int bgval );

native static synchronized void graDrawLine ( Ptr grData, int x1, int y1, int x2, int y2 );

native static synchronized void graDrawOval ( Ptr grData, int x, int y, int width, int height );

native static synchronized void graDrawPolygon ( Ptr grData, int[] xPoints, int[] yPoints, int nPoints );

native static synchronized void graDrawPolyline ( Ptr grData, int[] xPoints, int[] yPoints, int nPoints );

native static synchronized void graDrawRect ( Ptr grData, int x, int y, int width, int height );

native static synchronized void graDrawRoundRect ( Ptr grData, int x, int y, int width, int height,
						int arcWidth, int arcHeight );

native static synchronized void graDrawString ( Ptr grData, String s, int x, int y );

native static synchronized void graDrawVertString ( Ptr grData, String s, int x, int y );

native static synchronized void graFill3DRect ( Ptr grData, int x, int y, int width, int height,
		     boolean raised, int pixValue );

native static synchronized void graFillArc ( Ptr grData, int x, int y, int width, int height, int startAngle, int arcAngle );

native static synchronized void graFillOval ( Ptr grData, int x, int y, int width, int height );

native static synchronized void graFillPolygon ( Ptr grData, int[] xPoints, int[] yPoints, int nPoints );

native static synchronized void graFillRect ( Ptr grData, int x, int y, int width, int height );

native static synchronized void graFillRoundRect ( Ptr grData, int x, int y, int width, int height,
						int arcWidth, int arcHeight );

native static synchronized void graFreeGraphics ( Ptr grData );

native static synchronized Ptr graInitGraphics ( Ptr grData, Ptr tgtData, int tgtType, boolean isShowing,
					      int xOffset, int yOffset,
					      int xClip, int yClip, int wClip, int hClip,
					      Ptr fontData, int fgClr, int bgClr );

native static synchronized void graSetBackColor ( Ptr grData, int pixelValue );

native static synchronized void graSetClip ( Ptr grData, int xClip, int yClip, int wClip, int hClip );

native static synchronized void graSetColor ( Ptr grData, int pixelValue );

native static synchronized void graSetFont ( Ptr grData, Ptr fontData );

native static synchronized void graSetOffset ( Ptr grData, int xOffset, int yOffset );

native static synchronized void graSetPaintMode ( Ptr grData );

native static synchronized void graSetVisible ( Ptr grData, boolean isVisible );

native static synchronized void graSetXORMode ( Ptr grData, int xClr );

native static synchronized void imgComplete ( Ptr imgData, int status );

native static synchronized Ptr imgCreateFromData( NativeImageProducer prod, byte[] buf, int off, int len);

native static synchronized Ptr imgCreateFromFile( NativeImageProducer prod, String gifPath);

native static synchronized Ptr imgCreateImage( int w, int h);

native static synchronized Ptr imgCreateScaledImage( Ptr imgData, int w, int h);

native static synchronized Ptr imgCreateScreenImage( int w, int h);

native static synchronized void imgFreeImage ( Ptr imgData );

native static int imgGetFrameCount ( Ptr imgData );

native static int imgGetHeight( Ptr imgData);

native static int imgGetLatency ( Ptr imgData );

native static int imgGetLoopCount ( Ptr imgData );

native static Ptr imgGetNextFrame ( Ptr imgData );

native static int imgGetWidth( Ptr imgData);

native static synchronized void imgProduceImage( ImageProducer prod, Ptr imgData, boolean firstFrame);

native static synchronized Ptr imgSetFrame( Ptr imgData, int frame);

native static synchronized void imgSetIdxPels( Ptr imgData, int x, int y, int w, int h, int[] rgbs, byte[] pels, int trans, int off, int scans);

native static synchronized void imgSetRGBPels( Ptr imgData, int x, int y, int w, int h, int[] rgbs, boolean hasAlpha, int off, int scans);

static synchronized void initToolkit () {
	// this is called when the native layer has been initialized, and it is safe
	// to query native settings / rely on native functionality

	screenSize = new Dimension( tlkGetScreenWidth(), tlkGetScreenHeight());
	screenOffset = new Point( tlkGetScreenXoff(), tlkGetScreenYoff());
	resolution = tlkGetResolution();

	// we do this here to keep the getDefaultToolkit() method as simple
	// as possible (since it might be called frequently). This is a
	// deviation from the normal Singleton (which initializes the singleton
	// instance upon request)
	singleton = getDefaultToolkit();
	
	// overriding the eventQueue class might come in handy for things like
	// monitoring/record/replay, but make sure it's derived fromjava.awt.EventQueue
	if ( Defaults.EventQueueClass != null ){
		try {
			eventQueue = (EventQueue) Class.forName(Defaults.EventQueueClass).newInstance();
		}
		catch ( Exception xx ){
			System.err.println( "user EventQueue class does not load: " +
			                      Defaults.EventQueueClass);
		}	
	}
	if ( eventQueue == null ) {
		eventQueue = new EventQueue();
	}

/**
	if ( Defaults.ConsoleClass != null ){
		// since we have to defer the ConsoleWindow until the native Toolkit is propperly
		// initialized, it seems to be a good idea to defuse any output to the standard streams
		// (which might cause SEGFAULTS on some systems (e.g. DOS)
		System.setOut( new PrintStream( NullOutputStream.singleton));
		System.setErr( System.out);
	}
**/
}

protected void loadSystemColors ( int[] sysColors ) {
	clrSetSystemColors( sysColors);
}

public boolean prepareImage ( Image image, int width, int height, ImageObserver observer ) {
	return (image.loadImage( width, height, observer));
}

static void redirectStreams () {
	try {
		LogClient lv = (LogClient) Class.forName( Defaults.ConsoleClass).newInstance();
		LogStream ls = new LogStream( 30, lv);
		lv.enable();
	
		System.setOut( new PrintStream( ls) );
		System.setErr( System.out);
		
		System.out.println( "Java console enabled");
	}
	catch ( Exception x ) {
		System.err.println( "unable to redirect out, err");
		x.printStackTrace();
	}
}

static synchronized void startDispatch () {
	if ( eventThread == null ) {
		eventThread = new EventDispatchThread( Thread.currentThread().getThreadGroup(), eventQueue);
		eventThread.start();
		
		// we defer the Console creation / output redirection up to this point, since we otherwise
		// might get all sort of trouble because of a incompletely initialized native layer / Toolkit
		if ( Defaults.ConsoleClass != null )
			redirectStreams();
	}

	if ( ((flags & NEEDS_FLUSH) != 0) && (flushThread == null) ){
		flushThread = new FlushThread( Defaults.GraFlushRate);
		flushThread.start();
	}
}

static synchronized void stopDispatch () {
	if ( eventThread != null ) {
		eventThread.stopDispatching();
		eventThread = null;
	}
	
	if ( flushThread != null ){
		flushThread.stopFlushing();
		flushThread = null;
	}
}

public void sync () {
	tlkSync();
}

static void terminate () {
	if ( clipboard != null )
		clipboard.dispose();

	stopDispatch();
	tlkTerminate();
}

native static synchronized void thmDrawBackground ( Ptr grData,
                         int cx, int cy, int cw, int ch,
                         int x, int y, int w, int h,
                         int xOff, int yOff, Ptr imgData, int pixValue, int fillMode );

native static synchronized void thmDrawBorder ( Ptr grData,
                     int x, int y, int w, int h, int alignment,
                     Ptr itlData, Ptr itmData, Ptr itrData,
                     Ptr imlData, Ptr imrData,
                     Ptr iblData, Ptr ibmData, Ptr ibrData );

native static synchronized void tlkBeep ();

native static synchronized void tlkDisplayBanner ( String banner );

native static synchronized void tlkFlush ();

native static synchronized int tlkGetResolution ();

native static synchronized int tlkGetScreenHeight ();

native static synchronized int tlkGetScreenWidth ();

native static synchronized int tlkGetScreenXoff ();

native static synchronized int tlkGetScreenYoff ();

native static synchronized boolean tlkInit ( String displayName );

native static synchronized int tlkProperties();

native static synchronized void tlkSync ();

native static synchronized void tlkTerminate ();

native static synchronized String tlkVersion ();

native static synchronized Ptr wndCreateDialog ( Ptr ownerData, String title, int x, int y, int width, int height,
		      int cursorType, int bgColor, boolean isResizable );

native static synchronized Ptr wndCreateEmbedded ( long parent, int x, int y, int width, int height,
		                    int cursorType, int bgColor );

native static synchronized Ptr wndCreateFrame ( String title, int x, int y, int width, int height,
		     int cursorType, int bgColor, boolean isResizable );

native static synchronized Ptr wndCreateWindow ( Ptr ownerData, int x, int y, int width, int height,
		      int cursorType, int bgColor );

native static synchronized void wndDestroyWindow ( Ptr wndData );

native static synchronized void wndRepaint ( Ptr wndData, int x, int y, int width, int height );

native static synchronized void wndRequestFocus ( Ptr wndData );

native static synchronized void wndSetBounds ( Ptr wndData, int x, int y, int width, int height, boolean isResizable );

native static synchronized void wndSetCursor ( Ptr wndData, int cursorType );

native static synchronized int wndSetDialogInsets ( int top, int left, int bottom, int right);

native static synchronized int wndSetFrameInsets ( int top, int left, int bottom, int right);

native static synchronized void wndSetIcon ( Ptr wndData, Ptr iconData );

native static synchronized void wndSetResizable ( Ptr wndData, boolean isResizable,
		       int x, int y, int width, int height );

native static synchronized void wndSetTitle ( Ptr wndData, String title );

native static synchronized void wndSetVisible ( Ptr wndData, boolean showIt );

native static synchronized void wndShowCursor ( boolean showIt );

native static synchronized void wndToBack ( Ptr wndData );

native static synchronized void wndToFront ( Ptr wndData );
}
