package java.awt;

import kaffe.util.Condition;

/**
 * WMEvent - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P. Mehlitz
 */
class WMEvent
  extends AWTEvent
{
	final static int WM_CREATE = 1901;
	final static int WM_DESTROY = 1902;
	final static int WM_DISPATCH_MODAL = 1903;
	final static int WM_KILLED = 1904;
	final static int WM_IMAGE_QUEUE = 1905;
	final static int WM_EMPTY_EVENT = 1906;
	final static int WM_SET_FOCUS = 1907;
	static WMEvent cache;

WMEvent ( Object source, int id ) {
	super( source, id);
}

protected void dispatch () {
	Component c;

	switch ( id ) {
		
	case WM_CREATE:
		if ( source instanceof Component ){
			c = (Component) source;
			c.createNative();
			synchronized (this) {
				this.notifyAll();
			}
		}
		break;
		
	case WM_DESTROY:
		if ( source instanceof Component ){
			c = (Component) source;
			c.destroyNative();
			synchronized (this) {
				this.notifyAll();
			}
		}
		break;
		
	case WM_DISPATCH_MODAL:
		if ( source instanceof Condition ){
			Toolkit.eventThread.runUntil( (Condition)source);
			synchronized (this) {
				// we need to flag if our client (EventDispatchThread.run) needs to wait
			    consumed = true;
			  	this.notifyAll();
			}
		}
		break;
			
	case WM_KILLED:
		if ( source instanceof Component ){
			c = (Component) source;
			// we got an external kill on one of our windows, clean up so
			// that we don't get dangling windows
			c.cleanUpNative();
		}
		break;
	
	case WM_IMAGE_QUEUE:
		// the source doesn't matter at all for this
		try {
			Thread.sleep( 100); // give notifying ImageLoaders some time to get back
		}
		catch ( Exception x ) { }
		ImageLoader.checkQueue( false);
		break;
		
	case WM_EMPTY_EVENT:
		// this event has no direct side effects, it just travels through the
		// dispatcher (e.g. to check if it is still alive)
		break;
	
	case WM_SET_FOCUS:
		c = (Component) source;
		c.requestFocus();
		break;
	}
	
	recycle();
}

static synchronized WMEvent getEvent ( Object source, int id ) {
	if ( cache == null ){
		return new WMEvent( source, id);
	}
	else {
		WMEvent e = cache;
		cache = (WMEvent) e.next;
		e.next = null;
		
		e.id = id;
		e.source = source;

		//may be set by DISPATCH_MODAL
		e.consumed = false;

		return e;
	}
}

static synchronized WMEvent getEvent ( int srcIdx, int id ) {
	WMEvent e;
	Window  source = (Window) sources[srcIdx];

	if ( source == null ) return null;

	if ( cache == null ){
		e = new WMEvent( source, id);
	}
	else {
		e = cache;
		cache = (WMEvent) e.next;
		e.next = null;
		
		e.id = id;
		e.source = source;
		//may be set by DISPATCH_MODAL
		e.consumed = false;
	}

	if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
		// this is not used as a direct return value for EventQueue.getNextEvent(), 
		// it has to be Java-queued by the native layer
		Toolkit.eventQueue.postEvent( e);
	}
	
	return e;
}

protected void recycle () {
	synchronized ( WMEvent.class ) {
		source = null;

		next = cache;	
		cache = this;
	}
}
}
