/**
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P. Mehlitz
 */

package java.awt.image;

import kaffe.awt.JavaColorModel;

abstract public class RGBImageFilter
  extends ImageFilter
{
	protected ColorModel origmodel;
	protected ColorModel newmodel;
	protected boolean canFilterIndexColorModel;

public RGBImageFilter() {
}

public IndexColorModel filterIndexColorModel( IndexColorModel icm) {
	int i, n = icm.getMapSize();	
	int[] newRgbs = new int[n];

	for ( i=0; i<n; i++) {
		newRgbs[i] = filterRGB( -1, -1, icm.rgbs[i] );
	}

	return new IndexColorModel( icm.getPixelSize(), newRgbs, icm.getTransparentPixel());
}

abstract public int filterRGB( int x, int y, int rgb );

public void filterRGBPixels( int x, int y, int w, int h, int[] pels, int off, int scan) {
	int u = x, v = y;
	int uw = u + w;
	int vh = v + h;

	// Hmm, the spec says "Pixel(m,n)" maps to "pels[ (n * scan) + m + off]", but what is
	// 'n'? The JDK obviously thinks of this as [0..h], NOT [y..y+h]
	int i = off;	
	int dl = scan - w;
	
	for ( ;v < vh; v++, i += dl) {
		for ( ;u < uw; u++, i++) {
			pels[i] = filterRGB( u, v, pels[i]);
		}
	}

	consumer.setPixels( x, y, w, h, JavaColorModel.getSingleton(), pels, off, scan);
}

public void setColorModel ( ColorModel model ) {
	if ( (model instanceof IndexColorModel) && canFilterIndexColorModel ) {
		substituteColorModel( model, filterIndexColorModel( (IndexColorModel) model));
	}
	
	consumer.setColorModel( JavaColorModel.getSingleton());
}

public void setPixels( int x, int y, int w, int h, ColorModel cm, byte[] pels, int off, int scan) {
	if ( cm == origmodel ) {
		consumer.setPixels( x, y, w, h, newmodel, pels, off, scan);
	}
	else {
		int[] fPels = new int[pels.length];
			
		int u = x, v = y;
		int uw = u + w;
		int vh = v + h;
		int i = off;	
		int dl = scan - w;
	
		for ( ;v < vh; v++, i += dl ) {
			for ( ;u < uw; u++, i++ ) {
				fPels[i] = cm.getRGB( pels[i]);
			}
		}
		filterRGBPixels( x, y, w, h, fPels, 0, w);
	}
}

public void setPixels( int x, int y, int w, int h, ColorModel cm, int[] pels, int off, int scan) {
	if ( cm == origmodel ) {
		consumer.setPixels( x, y, w, h, newmodel, pels, off, scan);
	}
	else {
		if ( cm instanceof JavaColorModel ){
			// Hmm, if we get this from a NativeSelfProducer, there is no need to convert
			// colors - but are we allowed to filter "in situ" if it is not? (probably yes,
			// nobody else knows about a JavaColorModel)
			filterRGBPixels( x, y, w, h, pels, off, scan);
		}
		else {
			int[] fPels = new int[pels.length];
			
			int u = x, v = y;
			int uw = u + w;
			int vh = v + h;
			int i = off;	
			int dl = scan - w;
	
			for ( ;v < vh; v++, i += dl) {
				for ( ;u < uw; u++, i++) {
					fPels[i] = cm.getRGB( pels[i]);
				}
			}
			filterRGBPixels( x, y, w, h, fPels, 0, w);
		}
	}
}

public void substituteColorModel( ColorModel oldcm, ColorModel newcm) {
	origmodel = oldcm;
	newmodel  = newcm;	
}
}
