package java.awt.image;

import java.util.Hashtable;

/**
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 */
public class ReplicateScaleFilter
  extends ImageFilter
{
	protected int srcWidth;
	protected int srcHeight;
	protected int destWidth;
	protected int destHeight;
	protected int[] srcrows;
	protected int[] srccols;
	protected Object outpixbuf;
	private int[][] inpixbuf;

public ReplicateScaleFilter ( int width, int height) {
	destWidth = width;
	destHeight = height;
}

public void setDimensions ( int width, int height) {
	srcWidth = width;
	srcHeight = height;

	inpixbuf = new int[srcWidth][srcHeight];
	outpixbuf = new int[destHeight * destWidth];
	
	consumer.setDimensions( destWidth, destHeight);
}

public void setPixels ( int x, int y, int w, int h, ColorModel cm, byte[] pels, int off, int scan) {

	int u = x, v = y;
	int uw = u + w;
	int vh = v + h;
	int i = off;	
	int dl = scan - w;

	for ( ;v < vh; v++, i += dl) {
		for ( ;u < uw; u++, i++) {
			inpixbuf[u][v] = cm.getRGB( pels[i]);
		}
	}
	
	//we are done, scale source buffer and set consumer pixels
	if ( vh == srcHeight) {
		ImageScaler.getScaledImage( inpixbuf, (int[])outpixbuf, srcWidth,
							  0, 0, destWidth-1, destHeight-1,
							  0, 0, srcWidth-1, srcHeight-1 );
							
		consumer.setPixels( 0, 0, destWidth, destHeight, cm, (int[])outpixbuf, 0, destWidth);

		//make sure we get rid of the buffers as soon as possible
		inpixbuf = null;
		outpixbuf = null;
	}
}

public void setPixels ( int x, int y, int w, int h, ColorModel cm, int[] pels, int off, int scan) {
	
	int u = x, v = y;
	int uw = u + w;
	int vh = v + h;
	int i = off;	
	int dl = scan - w;

	for ( ;v < vh; v++, i += dl) {
		for ( ;u < uw; u++, i++) {
			inpixbuf[u][v] = pels[i];
		}
	}
	
	//we are done, scale source buffer and set consumer pixels
	if ( vh == srcHeight) {
		ImageScaler.getScaledImage( inpixbuf, (int[])outpixbuf, srcWidth,
							  0, 0, destWidth-1, destHeight-1,
							  0, 0, srcWidth-1, srcHeight-1 );
							
		consumer.setPixels( 0, 0, destWidth, destHeight, cm, (int[])outpixbuf, 0, destWidth);

		//make sure we get rid of the buffers as soon as possible
		inpixbuf = null;
		outpixbuf = null;
	}
}

public void setProperties ( Hashtable props) {
	if ( (srcWidth > 0) && (srcHeight > 0) ) {
		if ( props == null ) {
			props = new Hashtable(3);
		}
		props.put( "scaled", 	Double.toString( (double)destWidth/srcWidth) + "x" + Double.toString((double)destHeight/srcHeight));
	}
	
	super.setProperties( props);
}
}

class ImageScaler
{
public static void getScaledImage ( int[][] srcbuf, int[]dstbuf, 
				      int scan, int dx0, int dy0, int dx1, 
				      int dy1, int sx0, int sy0, int sx1, 
				      int sy1 )
{
    double   xScale, yScale, sX, sY, sxDelta, syDelta;
    int      dx, dy, dxInc, dyInc, sx, sy;
    int      pix;

    dxInc = (dx1 > dx0) ? 1 : -1;
    dyInc = (dy1 > dy0) ? 1 : -1;
   
    dx1 += dxInc;
    dy1 += dyInc;
   
    xScale = (double) (dx1 - dx0) / (double) (sx1 - sx0 +1);
    yScale = (double) (dy1 - dy0) / (double) (sy1 - sy0 +1);

    for ( dy=dy0; dy != dy1; dy += dyInc ) {
	sY = sy0 + (dy - dy0) / yScale;
	sy = (int) sY;
	syDelta = (sy < sy1) ? sY - sy : 0;
	      
	for ( dx=dx0; dx != dx1; dx += dxInc ) {
	    sX = sx0 + (dx - dx0) / xScale;
	    sx = (int) sX;
	    sxDelta = (sx < sx1) ? sX - sx : 0;
				         
	    pix = getScaledPixel( srcbuf, scan, sx, sy, sxDelta, syDelta );
	    dstbuf[dy*(dx1-dx0)+dx] = 0xff000000 | pix;
	}
    }
   
}

public static void getScaledImage ( int[][] srcbuf, int[][] dstbuf, 
				      int scan, int dx0, int dy0, int dx1, 
				      int dy1, int sx0, int sy0, int sx1, 
				      int sy1 )
{
    double   xScale, yScale, sX, sY, sxDelta, syDelta;
    int      dx, dy, dxInc, dyInc, sx, sy;
    int      pix;

    dxInc = (dx1 > dx0) ? 1 : -1;
    dyInc = (dy1 > dy0) ? 1 : -1;
   
    dx1 += dxInc;
    dy1 += dyInc;
   
    xScale = (double) (dx1 - dx0) / (double) (sx1 - sx0 +1);
    yScale = (double) (dy1 - dy0) / (double) (sy1 - sy0 +1);

    for ( dy=dy0; dy != dy1; dy += dyInc ) {
	sY = sy0 + (dy - dy0) / yScale;
	sy = (int) sY;
	syDelta = (sy < sy1) ? sY - sy : 0;
	      
	for ( dx=dx0; dx != dx1; dx += dxInc ) {
	    sX = sx0 + (dx - dx0) / xScale;
	    sx = (int) sX;
	    sxDelta = (sx < sx1) ? sX - sx : 0;
				         
	    pix = getScaledPixel( srcbuf, scan, sx, sy, sxDelta, syDelta );
	    dstbuf[dy][dx] = 0xff000000 | pix;
	}
    }
   
}

private static int getScaledPixel( int[][] srcbuf, int scan, int x, int y, 
				  double dx, double dy )
{
    int    ul, ur, ll, lr;
    char   a0, a1, a2, a3;
    int    ulR, urR, llR, lrR, ulG, urG, llG, lrG, ulB, urB, llB, lrB, r, g, b;
    int    xi = (dx != 0) ? x+1 : x;
    int    yi = (dy != 0) ? y+1 : y;
   
    if ( xi >= scan )  xi = x;
    if ( yi >= srcbuf[0].length ) yi = y;

    a0 = a1 = a2 = a3 = 0;  

    ul = srcbuf[x][y];
    ur = srcbuf[xi][y];
    ll = srcbuf[x][yi];
    lr = srcbuf[xi][yi];
      
    if ( (ul == ur) && (ll == ul) && (lr == ll) ) {
	r = (ul & 0xff0000) >> 16;
	g = (ul & 0x00ff00) >> 8;
	b = (ul & 0x0000ff); 
    }
    else {
	ulR = (ul & 0xff0000) >> 16;
	ulG = (ul & 0x00ff00) >> 8;
	ulB = (ul & 0x0000ff); 

	urR = (ur & 0xff0000) >> 16;
	urG = (ur & 0x00ff00) >> 8;
	urB = (ur & 0x0000ff); 

	llR = (ll & 0xff0000) >> 16;
	llG = (ll & 0x00ff00) >> 8;
	llB = (ll & 0x0000ff); 

	lrR = (lr & 0xff0000) >> 16;
	lrG = (lr & 0x00ff00) >> 8;
	lrB = (lr & 0x0000ff); 
	      
	r = interpolate( ulR, urR, llR, lrR, dx, dy);
	g = interpolate( ulG, urG, llG, lrG, dx, dy);
	b = interpolate( ulB, urB, llB, lrB, dx, dy);
    }
   
    return ((r << 16) | (g << 8) | b);
}

private static int interpolate ( int ul, int ur, int ll, int lr, 
				 double dx, double dy )
{
    double u = ul + (double)(ur - ul) * dx;
    double l = ll + (double)(lr - ll) * dx;
   
    return (int) (u + (l - u) * dy  + 0.5);
}
}
