/**
 * ASControlBox -
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *      
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *      
 * @author P. Mehlitz
 */

package java.awt;

import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Calendar;
import java.util.Date;
import java.util.Vector;
import kaffe.awt.Defaults;
import kaffe.awt.GoodStuff;
import kaffe.awt.ImageDict;
import kaffe.util.Timer;
import kaffe.util.TimerClient;

class ASControlBox
  extends Window
  implements TimerClient, MouseListener, ControlBox, WindowListener, ComponentListener
{
	RootWindow root;
	int iconSize;
	int sepHeight;
	int menuHeight;
	int nPages;
	int pgHeight;
	float pgScale;
	GoodStuff goodStuff;
	static Calendar calendar;
	static Date now;
	static Font fixed;
	static FontMetrics fmFixed;
	static String[] MonthName = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static {
	fixed = new Font( "SansSerif", Font.PLAIN, 10);
	fmFixed = Toolkit.getDefaultToolkit().getFontMetrics( fixed);

	calendar = Calendar.getInstance();
	now = new Date();
}

public ASControlBox () {
	this.root = RootWindow.singleton;
	bgClr = Color.lightGray;
	page = 0xffff;
	
	iconSize = ( root.width < 1024 ) ? 32 : 48;
	width  = iconSize + 6;
	sepHeight = fmFixed.getHeight();
	
	menuHeight = sepHeight;

	pgHeight = (root.height * width) / root.width;
	nPages = 4;
	pgScale = (float) width / (float) root.width;

	goodStuff = GoodStuff.getDefaultGoodStuff();

	height = nPages*pgHeight + (width * (goodStuff.size() + 1)) + 2*sepHeight + menuHeight;

	setBounds( root.width - width, 0, width, height);
	setVisible( true);

	Timer.getDefaultTimer().addClient( this, 1000, 60000);
	addMouseListener( this);
}

public void added ( Window c ) {
	c.addWindowListener( this);
	c.addComponentListener( this);

	updatePage( root.currentPage());
}

public void componentHidden ( ComponentEvent evt ) {
	updatePage( root.currentPage());
}

public void componentMoved ( ComponentEvent evt ){
	updatePage( root.currentPage());
}

public void componentResized ( ComponentEvent evt ) {
	updatePage( root.currentPage());
}

public void componentShown ( ComponentEvent evt ){
	updatePage( root.currentPage());
}

void drawClock ( Graphics g ) {
	int hour, min, mnth, day;
	String time, date;

	now.setTime( System.currentTimeMillis());
	calendar.setTime( now);

	hour = calendar.get( Calendar.HOUR_OF_DAY);
	min = calendar.get( Calendar.MINUTE);
	mnth = calendar.get( Calendar.MONTH);
	day = calendar.get( Calendar.DAY_OF_MONTH);	
	
	time = ((hour < 10) ? "0" : "") + hour + ((min < 10) ? ":0" : ":") + min;
	date  = MonthName[mnth] + ' ' + day;
	
	drawClock( g, time, date);	
}

void drawClock ( Graphics g, String time, String date ) {
	int d = 3;
	int x, y;
	int w = width - 2*d - 1;
	int h = fmFixed.getHeight() + 4;

	g.setColor( bgClr);
	g.fill3DRect( 0, 0, width, width, true);
	g.draw3DRect( d, d, w, h, false);
	g.setColor( Color.black);
	d++;
	g.fillRect( d, d, w-1, h-1);

	g.setColor( Color.green);
	g.setFont( fixed);

	x = d + (w-2 - fmFixed.stringWidth( time)) / 2;
	y = d + 1 + fmFixed.getAscent();
	g.drawString( time, x, y);
	
	h += d;
	x = (width - fmFixed.stringWidth( date)) / 2;
	y = h + ((width - h - fmFixed.getHeight()) / 2) + fmFixed.getAscent();
	g.setColor( Color.black);
	g.drawString( date, x, y);
}

void drawGoodStuff ( Graphics g, int idx ) {
	int y = (pgHeight * nPages) + (idx * width) + width + menuHeight + 2*sepHeight;
	int w, h;
	Image icon;
		
	g.setColor( bgClr);
	
	g.setClip( 0, y, width, width);
	g.fill3DRect( 0, y, width, width, true);

	icon = goodStuff.getItemImage( idx);
	w = icon.getWidth( this);
	h = icon.getHeight( this);

	g.drawImage( icon, (width - w) / 2, y + (width - h) / 2, this);
}

void drawMenu ( Graphics g ) {
	String sel;
	int     w;

	g.setColor( bgClr);
	g.fill3DRect( 0, width, width, menuHeight, true);
	
	sel = "exit";
	w = fmFixed.stringWidth( sel);
	g.setColor( fgClr);
	g.drawString( sel, (width-w)/2, width + fmFixed.getAscent());
}

void drawPage ( Graphics g, int n ) {
	int   pg = 1<<(n-1);
	Color bg = (pg == root.page) ? Defaults.BtnClr : bgClr;
	int	y  = width + (n-1)*pgHeight + sepHeight + menuHeight;
	int yh = y + pgHeight;
	int i, u, v, w, h, vh, uw;
	Window c;

  g.setColor( bg);
	g.fill3DRect( 0, y, width, pgHeight, true);

	for ( i=root.nChildren-1; i>=0; i-- ) {
		c = (Window) root.children[i];
		if ( ((c.page & pg) != 0) && !root.minimized.contains( c) ) {
			if ( (c.page == 0xffff) && (pg != root.page) ) continue;
		
			u = (int) (c.x * pgScale);
			v = y + (int) (c.y * pgScale);
			uw = u + (int) (c.width * pgScale);
			vh = v + (int) (c.height * pgScale);

			if ( u < 1 ) u = 1;
			if ( uw >= width ) uw = width-1;
			if ( v <= y ) v = y+1;
			if ( vh >= yh ) vh = yh-1;
			w = uw - u;
			h = vh - v;

			g.setColor( (pg==root.page) ? Color.yellow : Color.white);
			g.fillRect( u, v, w, h);

			g.setColor( Color.blue);
			g.drawRect( u, v, w, h);
		}
	}
}

String getGenericName () {
	// this is not a standard Java component, but since it's a Window, getting its
	// generic name woul alter the name count. You bet -there's an application out
	// there which relies on the numbering
		
	return "controlbox";
}

public boolean imageUpdate( Image img, int flags, int x, int y, int w, int h) {
	int n = goodStuff.size();
	for ( int i=0; i<n; i++) {
		if ( goodStuff.getItemImage( i) == img ) {
			if ( ( flags & ALLBITS) > 0 ) {
				Graphics g = getGraphics();
				drawGoodStuff( g, i);
				g.dispose();
			}
			return (( flags & ( ALLBITS | ABORT)) == 0);
		}
	}
	return super.imageUpdate( img, flags, x, y, w, h);
}

public void mouseClicked ( MouseEvent evt ) {
	int y = evt.getY();
	int yMn = width;
	int yPg = width + menuHeight + sepHeight;
	int yGs = yPg + sepHeight + nPages*pgHeight;

	if ( y > yGs ) {
		int gsIdx = (y - yGs) / width;
		goodStuff.trigger( gsIdx);
	}
	else if ( y > yPg ) {
		int pgIdx = 1+ (y - yPg) / pgHeight;  // 1-based
		root.showPage( pgIdx);
	}
	else if ( y > yMn ){
		System.exit( 0); // temp!! implement full menu
	}
}

public void mouseEntered ( MouseEvent evt ) {
}

public void mouseExited ( MouseEvent evt ) {
}

public void mousePressed ( MouseEvent evt ) {
}

public void mouseReleased ( MouseEvent evt ) {
}

public void paint ( Graphics g ) {
	int i, u, v, n=goodStuff.size();

	drawClock( g);
	
	drawMenu( g);
	
	g.setColor( Color.black);
	u = (width - fmFixed.stringWidth( "pages")) / 2;
	v = width + menuHeight + fmFixed.getAscent();
	g.drawString( "pages", u, v);

	u = (width - fmFixed.stringWidth( "apps")) / 2;
	v += nPages*pgHeight + sepHeight;
	g.drawString( "apps", u, v);

	for ( i=1; i<=nPages; i++ ) {
		drawPage( g, i);
	}
	
	for ( i=0; i<n; i++ ) {
		drawGoodStuff( g, i);
	}
}

public void removed ( Window c ) {
	updatePage( root.currentPage());
}

public void switchPage ( int oldPage, int newPage ) {
	updatePage( oldPage);
	updatePage( newPage);
}

public void timerExpired ( Timer timer ) {
	Graphics g = getGraphics();

	if ( g != null ) {
		drawClock( g);	
		g.dispose();
	}
}

public void toBack ( Component c ) {
	updatePage( root.currentPage());
}

public void toFront ( Component c ) {
	updatePage( root.currentPage());
}

public void updatePage ( int pageNum ) {
	Graphics g = getGraphics();
	
	if ( g != null ){
		drawPage( g, pageNum);
		g.dispose();
	}
}

public void windowActivated ( WindowEvent evt ){
	updatePage( root.currentPage());
}

public void windowClosed ( WindowEvent evt ){
}

public void windowClosing ( WindowEvent evt ){
}

public void windowDeactivated ( WindowEvent evt ){
	updatePage( root.currentPage());
}

public void windowDeiconified ( WindowEvent evt ){
}

public void windowIconified ( WindowEvent evt ){
}

public void windowOpened ( WindowEvent evt ){
}
}
