package java.awt;

import java.awt.event.WindowEvent;
import kaffe.awt.Defaults;
import kaffe.util.Condition;
import kaffe.util.Ptr;

/**
 * Dialog - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P. Mehlitz
 */
public class Dialog
  extends Frame
{
	static Insets dialogInsets;
	static int named;

static {
	// no guessing, we rule the game	
	Insets in = Defaults.DialogInsets;
	
	dialogInsets = new Insets( in.top, in.left, in.bottom, in.right);
}

public Dialog ( Frame owner ) {
	this( owner, null, false);
}

public Dialog ( Frame owner, String title ) {
	this( owner, title, false);
}

public Dialog ( Frame owner, String title, boolean isModal) {
	super( owner, title);
	
	if ( isModal )
		flags |= IS_MODAL;
}

public Dialog ( Frame owner, boolean isModal ) {
	this( owner, null, isModal);
}

void createNative () {
	// insets seem to be set by the JDK during addNotify
	// (no need to create fresh objects since they are insets()-copied anyway <sigh>)
	insets = dialogInsets;

	nativeData = Toolkit.wndCreateDialog( (owner != null) ? owner.nativeData : null,
										 title, x, y, width, height,
	                                     cursor.type, bgClr.nativeValue,
	                                     ((flags & IS_RESIZABLE) != 0));
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "dialog" + (named++);
}

public void hide () {
	if ( (flags & IS_MODAL) != 0 ){
		dispose();
	}
	else {
		super.hide();
	}
}

public boolean isModal() {
	return ((flags & IS_MODAL) != 0);
}

boolean isSystemModal (){
	return ((flags & IS_MODAL) != 0) && (owner == null);
}

void process ( WindowEvent e ) {
	super.process( e);

	// some windowing systems don't support global dialogs - we have to mimic
	// this in terms of never let this one get obscured (otherwise we have a perfect
	// UI lockout, Java wise)
	if ( e.getID() == WindowEvent.WINDOW_DEACTIVATED ){
		if ( isSystemModal() ){
			toFront();
			//requestFocus();
			Toolkit.eventQueue.postEvent( WMEvent.getEvent( this, WMEvent.WM_SET_FOCUS));
		}
	}
}

public void setModal ( boolean isModal ) {
	if ( isModal ) 
		flags |= IS_MODAL;
	else
		flags &= ~IS_MODAL;
}

public void show () {
	// DEP - should be in setVisible()
	super.show ();
	
	if ( (flags & IS_MODAL) != 0 ) {
		if ( owner != null ) {
			// We don't use native modal dialogs, so let's at least temporarily
			// disable the owner
			owner.propagateTempEnabled( false);
		}
		else {
			// modal and no owner is interpreted as system modal
			AWTEvent.tempEnableAll( false, this);
		}
		
		// Hrmm, can't implement this directly because we can't add public
		// methods to Dialog
		Toolkit.eventThread.runUntil( new Condition() {
			public boolean evaluate (){
				return ((Dialog.this.flags & Component.IS_ADD_NOTIFIED) == 0);
			}
		});
		
		if ( owner != null ){
			owner.propagateTempEnabled( true);
		}
		else {
			// was system modal, get all back (except of the ones which were tempDisabled)
			AWTEvent.tempEnableAll( true, this);
		}
	}
}
}
