package java.awt;

import java.awt.event.FocusEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.PaintEvent;
import java.awt.event.WindowEvent;
import java.util.Enumeration;
import java.util.Vector;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.awt.OpaqueComponent;
import kaffe.awt.theme.ComponentDeco;
import kaffe.awt.theme.FrameDeco;
import kaffe.lang.Application;
import kaffe.lang.ApplicationResource;
import kaffe.util.Ptr;

/**
 * Frame - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P.C.Mehlitz
 */
public class Frame
  extends Window
  implements OpaqueComponent, ApplicationResource
{
	String title;
	Image icon;
	BarMenu bMenu;
	int lastLoc;
	Rectangle restore;
	boolean isActive;
	Insets innerInsets;
	static Cursor[] decoCursor;
	static int mh = Defaults.MenuBarHeight;
	static int named;

static {
	// we need fast color and font access for deco management

	decoCursor = new Cursor[13];
	decoCursor[0] = Cursor.getDefaultCursor();
	decoCursor[1] = Cursor.getPredefinedCursor( Cursor.N_RESIZE_CURSOR);
	decoCursor[2] = Cursor.getPredefinedCursor( Cursor.E_RESIZE_CURSOR);
	decoCursor[3] = Cursor.getPredefinedCursor( Cursor.NE_RESIZE_CURSOR);
	decoCursor[4] = Cursor.getPredefinedCursor( Cursor.S_RESIZE_CURSOR);
	decoCursor[5] = Cursor.getDefaultCursor();
	decoCursor[6] = Cursor.getPredefinedCursor( Cursor.SE_RESIZE_CURSOR);
	decoCursor[7] = Cursor.getDefaultCursor();
	decoCursor[8] = Cursor.getPredefinedCursor( Cursor.W_RESIZE_CURSOR);
	decoCursor[9] = Cursor.getPredefinedCursor( Cursor.NW_RESIZE_CURSOR);
	decoCursor[10] = Cursor.getDefaultCursor();
	decoCursor[11] = Cursor.getDefaultCursor();
	decoCursor[12] = Cursor.getPredefinedCursor( Cursor.SW_RESIZE_CURSOR);
}

public Frame () {
	this( null, null);
}

Frame ( Frame owner, String title ) {
	super( owner);

	flags |= IS_RESIZABLE | IS_MOVABLE;
	this.title = (title == null) ? "" : title;

	innerInsets = new Insets( insets.top, insets.left, insets.bottom, insets.right);
}

public Frame ( String title ) {
	this( null, title);
}

boolean canBeMovedOrSized ( int borderLoc ) {
	if ( !isActive ) return false;
	
	if ( borderLoc == 0 ) {
		if ( (flags & IS_MOVABLE) == 0 ) return false; 
	}
	else {
		if ( (flags & IS_RESIZABLE) == 0 ) return false;
	}

	return true;
}

public int countComponents() {
	// DEP - should be in getComponentCount()

	if ( bMenu == null )
		return nChildren;
	
	return nChildren-1;
}

void createNative () {
	Application.addResource(this);

	// Insets seem to be set by the JDK during addNotify
	// (no need to create fresh objects since they are insets()-copied anyway <sigh>).
	// Note that the global inset objects might be changed if the native layer
	// has to compute the real deco offsets during window creation

	nativeData = Toolkit.wndCreateFrame( title, x, y, width, height,
	                               cursor.type, bgClr.nativeValue,
	                               ((flags & IS_RESIZABLE) != 0));
}

public void dispose () {
	super.dispose();
	Application.removeResource(this);
}

public void freeResource() {
	dispose();
}

public Component getComponent( int index) {
	if ( children == null )
		return null;
	if ( bMenu == null )
		return children[index];
	
	return children[index+1];
}

public Component[] getComponents() {
	if ( bMenu == null )
		return super.getComponents();
	
	Component ca[] = new Component[nChildren-1];
	
	if ( ca.length > 0 )
		System.arraycopy( children, 1, ca, 0, nChildren-1);
	
	return ca;
}

int getDecoFlags () {
	int flags = 0;
	
	if ( isActive ) flags |= FrameDeco.ACTIVE;
	if ( owner != null ) flags |= FrameDeco.OWNED;
	if ( ownees != null ) flags |= FrameDeco.OWNEES;
	// others to follow..
	
	return flags;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "frame" + (named++);
}

public Graphics getGraphics () {
	if ( nativeData != null ) {
		// don't overpaint the frame decoration
		return NativeGraphics.getGraphics( this, nativeData,
		                                   NativeGraphics.TGT_TYPE_WINDOW,
		                                   ((flags & IS_VISIBLE) != 0),
		                                   0, 0,
		                                   innerInsets.left, innerInsets.top,  // insets because of bMenu
		                                   width - ( innerInsets.left + innerInsets.right),
					       height - ( innerInsets.top + innerInsets.bottom),
		                                   fgClr, bgClr, font);
	}
	else {
		return null;
	}
}

public Image getIconImage() {
	return icon;
}

public Insets getInsets() {
   return innerInsets;
}

public MenuBar getMenuBar () {
	return (bMenu != null) ? bMenu.mb : null;
}

public String getTitle () {
	return title;
}

public void invalidate() {
	super.invalidate();
	if ( AWT.themeChanged) {
		innerInsets.top = insets.top;
		if ( bMenu != null )
			innerInsets.top += mh; 
	}
}

boolean isPopup () {
	return false;
}

public boolean isResizable () {
	return ((flags & IS_RESIZABLE) != 0);
}

public void layout () {
	super.layout();

	if ( bMenu != null){
		bMenu.setBounds( insets.left, insets.top,
		                 width-(insets.left+insets.right), mh);
  	}
}

public void paint ( Graphics g ) {
//synchronized ( treeLock ) {
		int il = insets.left;
		int it = insets.top + ((bMenu != null) ? mh : 0);
		int w = width - (il + insets.right);
		int h = height - (it + insets.bottom);

		int n = (bMenu == null ) ? 0 : 1;

		// we should treat the bMenu special because (a) it avoids flicker, and (b)
		// some careless apps have components painting over the menu (we have to clip)
		if ( ((flags & IS_IN_UPDATE) == 0) && (bMenu != null) ) {
			// otherwise the menu has already been drawn by update()
			g.paintChild( bMenu, false);
			g.clipRect( il, it, w, h);
		}

		for ( int i=nChildren-1; i>=n; i-- ) {
			Component c = children[i];

			if ( (c.flags & IS_VISIBLE) != 0 ){
				g.paintChild( c, (flags & IS_IN_UPDATE) != 0);
			}
		}
//}
}

void process ( WindowEvent e ) {

	switch ( e.id ) {
	case WindowEvent.WINDOW_DEACTIVATED:
		if ( FocusEvt.newActiveWindow.owner != this ){
			isActive = false;
			redrawDeco();
		}
		break;
	case WindowEvent.WINDOW_ACTIVATED:
		if ( !isActive ){
			isActive = true;

			// maybe we are already on top but were not activated
			if ( RootWindow.singleton.getComponent( 0) == this ){
				redrawDeco();
			}
			else {
				toFront();
			}
		}
		break;
	}

	super.process( e);
}

void processMotion ( MouseEvent e ) {
	int x = e.getX();
	int y = e.getY();
	int loc;

	if ( RootWindow.singleton.isResizing ){
		RootWindow.singleton.resize( this, x, y);
		return;
	}

	if ( e.id == MouseEvent.MOUSE_DRAGGED ) {
	  if ( RootWindow.singleton.preparedForResize(this) ){
			RootWindow.singleton.resize( this, x, y);
			return;	
		}
	}
	else if ( (e.id == MouseEvent.MOUSE_MOVED) && (deco instanceof FrameDeco) ) {
		FrameDeco fdeco = (FrameDeco)deco;
		loc = fdeco.getBorder( this.width, this.height, x, y, 0);
		if ( loc != lastLoc ) {
			lastLoc = loc;
			if ( (flags & IS_RESIZABLE) != 0 ){
				Toolkit.wndSetCursor( nativeData, decoCursor[loc].type);
			}
		}
		
		return;
	}

	if ( lastLoc > 0 ){
		lastLoc = 0;
		Toolkit.wndSetCursor( nativeData, cursor.type);
	}

	super.processMotion( e);
}

void processMouse ( MouseEvent e ) {
	int x  = e.getX();
	int y  = e.getY();
	int id = e.id;

	switch ( id ) {
	case MouseEvent.MOUSE_EXITED:
		lastLoc = 0;
		break;
	case MouseEvent.MOUSE_PRESSED:
		if ( !isActive ){
			AWTEvent.sendEvent( FocusEvt.getEvent( this, FocusEvent.FOCUS_GAINED, false), true);
		}
		break;
	case MouseEvent.MOUSE_RELEASED:
		if ( RootWindow.singleton.isResizing ) {
			RootWindow.singleton.endResize( this);
			return;
		}		
	}

	if ( (x < insets.left) || (y < insets.top) ||
	     (x > width-insets.right) || (y > height-insets.bottom) ){ 
		switch ( id ) {
		case MouseEvent.MOUSE_PRESSED:
			lastLoc = ((FrameDeco)deco).getBorder( this.width, this.height, x, y, flags);
			if ( canBeMovedOrSized( lastLoc) ) {			
				RootWindow.singleton.prepareForResize( this, x, y);
			}
			return;

		case MouseEvent.MOUSE_CLICKED:
			int btn = ((FrameDeco)deco).getTitlebarButton( this.width, this.height, x, y, (flags & IS_RESIZABLE));
			switch ( btn ){
			case FrameDeco.TBB_CLOSE:
				AWTEvent.sendEvent( WindowEvt.getEvent( this, WindowEvent.WINDOW_CLOSING), false);
				break;
			case FrameDeco.TBB_MINIMIZE:
				if ( restore != null ) {
					restore();
				}
				else {
					restore = new Rectangle( this.x, this.y, this.width, this.height);
					RootWindow.singleton.minimize( this);					
				}
				break;
			case FrameDeco.TBB_MAXIMIZE:
				if ( restore != null ) {
					restore();
				}
				else {
					restore = new Rectangle( this.x, this.y, this.width, this.height);
					RootWindow.singleton.maximize( this);
				}
				break;
			}
			return;
		}
	}

	super.processMouse( e);
}

void processPaintEvent ( int id, int ux, int uy, int uw, int uh ) {
	NativeGraphics g = NativeGraphics.getGraphics( this, nativeData,
	                                                NativeGraphics.TGT_TYPE_WINDOW,
	                                                ((flags & IS_VISIBLE) != 0),
	                                                0,0, ux, uy, uw, uh,
	                                                fgClr, bgClr, font);
	if ( g != null ){
		deco.paintBorder( this, g, 0, 0, width, height, getDecoFlags());

		// Restore state (possibly changed by redrawDeco). Note that the clipping just needs
		// to be set in case we don't have a bMenu (otherwise it's set by update/paint, anyway),
		g.setColor( fgClr);
		g.setFont( font);
		
		if ( bMenu == null ){			
			g.clipRect( insets.left, insets.top,
			            width- (insets.left + insets.right),
			            height- (insets.top + insets.bottom));
		}

		if ( id == PaintEvent.UPDATE ){
			update( g);
		}
		else {
			paint( g);
		}

		g.dispose();
	}
}

protected void processWindowEvent ( WindowEvent event ) {
	if ( wndListener != null ) {
		switch ( event.id ) {
		case WindowEvent.WINDOW_OPENED:
			wndListener.windowOpened( event);
			break;
		case WindowEvent.WINDOW_CLOSING:
			wndListener.windowClosing( event);
			break;
		case WindowEvent.WINDOW_CLOSED:
			wndListener.windowClosed( event);
			break;
		case WindowEvent.WINDOW_ICONIFIED:
			wndListener.windowIconified( event);
			break;
		case WindowEvent.WINDOW_DEICONIFIED:
			wndListener.windowDeiconified( event);
			break;
		case WindowEvent.WINDOW_ACTIVATED:
			wndListener.windowActivated( event);
			break;
		case WindowEvent.WINDOW_DEACTIVATED:
			wndListener.windowDeactivated( event);
			break;
		}
	}
}

void redrawDeco () {
	Graphics g = NativeGraphics.getGraphics( this, nativeData,
		                                   NativeGraphics.TGT_TYPE_WINDOW,
		                                   true,
		                                   0, 0, 0, 0, width, height,
		                                   fgClr, bgClr, font);
	if ( g != null ) {
		deco.paintBorder( this, g, 0, 0, width, height, getDecoFlags());
		g.dispose();
	}
}

public void remove ( MenuComponent mc ) {
	if ( bMenu != null ) {
		if ( bMenu.mb == mc ) {
			removeMenuBar();
		}
	}
}

private void removeMenuBar () {
	if ( bMenu != null ) {
		BarMenu b = bMenu;
		bMenu = null;
		
		nChildren--;
		System.arraycopy( children, 1, children, 0, nChildren);
	
		innerInsets.top = insets.top;
	
		if ( (b.flags & IS_ADD_NOTIFIED) != 0 ){
			b.removeNotify();
			doLayout();
		}
	}
}

void restore () {
	if ( restore != null ) {
		setBounds( restore.x, restore.y, restore.width, restore.height);
		restore = null;
	}
	
	if ( (flags & IS_VISIBLE) == 0 )
		setVisible( true);

	toFront();
	requestFocus();		
}

public void setCursor(int cursorType) {
	setCursor(Cursor.getPredefinedCursor(cursorType));
}

public void setIconImage ( Image icon ) {
	this.icon = icon;
}

public void setMenuBar ( MenuBar mb ) {
	if (mb == null) {
		removeMenuBar();
	}
	else if ( bMenu != null ) {
		bMenu.setMenus( mb);
	}
	else {
		bMenu = new BarMenu( mb);
	
		// add BarMenu to children at position 0 ( without LayoutManager registration )
		if ( children == null )
			children= new Component[3];
		else if ( nChildren == children.length) {
			Component[] old = children;
			children = new Component[nChildren*2];
			System.arraycopy( old, 0, children, 1, nChildren);
		}
		else
			System.arraycopy( children, 0, children, 1, nChildren);
		
		children[0] = bMenu;
		nChildren++;
		bMenu.parent = this;

		innerInsets.top = insets.top + mh;

		if ( (flags & IS_VALID) != 0 )
			invalidate();
				
		// inherit parent attributes (if not overriden by child)
		if ( (flags & (IS_PARENT_SHOWING | IS_VISIBLE)) == (IS_PARENT_SHOWING | IS_VISIBLE) ){
			bMenu.flags |= IS_PARENT_SHOWING;
			bMenu.propagateParentShowing( false);
		}
		if ( (bMenu.flags & IS_BG_COLORED) == 0 ){
			Color bgd = bMenu.deco.getBackground();
			bMenu.propagateBgClr( (bgd != null) ? bgd : bgClr);
		}
		if ( (bMenu.flags & IS_FG_COLORED) == 0 ){
			Color fgd = bMenu.deco.getForeground();
			bMenu.propagateFgClr( (fgd != null) ? fgd : fgClr);
		}
		if ( (bMenu.flags & IS_FONTIFIED) == 0 ){
			Font fd = bMenu.deco.getFont();
			bMenu.propagateFont( (fd != null) ? fd : font);
		}
		if ( (flags & IS_ADD_NOTIFIED) != 0 ) {
			bMenu.addNotify();
			doLayout();
		}
		
	}
}

public void setResizable ( boolean isResizable ) {
	if ( isResizable )
		flags |= IS_RESIZABLE;
	else
		flags &= ~IS_RESIZABLE;
}

public void setTitle ( String newTitle ) {
	title = newTitle;

	redrawDeco();
}

void setVisibleImpl ( boolean showIt ) {
	if ( showIt )
		isActive = true;

	super.setVisibleImpl( showIt);
}

public void update ( Graphics g ) {
	int il = insets.left;
	int it = insets.top + ((bMenu != null) ? mh : 0);
	int w = width - (il + insets.right);
	int h = height - (it + insets.bottom);

	// we should treat the bMenu special because (a) it avoids flicker, and (b)
	// some careless apps have components painting over the menu (we have to clip)
	if ( bMenu != null ) {
		g.paintChild( bMenu, false);
		g.clipRect( il, it, w, h);
	}

	paintBackground( g, il, it, w, h);
	paint( g);
}
}
