package java.awt;

import java.awt.event.ContainerEvent;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.PaintEvent;
import java.awt.image.ImageObserver;
import java.util.Vector;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.awt.theme.FrameDeco;
import kaffe.awt.GoodStuff;
import kaffe.awt.ImageDict;
import kaffe.awt.VTSwitchEvent;
import kaffe.awt.VTSwitchEventHandler;

/**
 * RootWindow & MinBox -
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *      
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *      
 * @author P. Mehlitz
 */
class RootWindow
  extends Window
  implements KeyListener, MouseListener, VTSwitchEventHandler
{
	Graphics gr;
	boolean isResizing;
	Vector minimized = new Vector( 5);
	MinBox minBox;
	Image wallPaper;
	ControlBox controlBox;
	int page;
	static RootWindow singleton;
	static Point sizeOffset = new Point();
	static Rectangle sizeRect;

static {
	try {
		Class rootCls = Class.forName( Defaults.RootWindowClass);
		singleton = (RootWindow) rootCls.newInstance();
	}
	catch ( Exception x ) {
		x.printStackTrace();
		singleton = new RootWindow();
	}
}

RootWindow () {
	int ret;
	Dimension d = Toolkit.getDefaultToolkit().getScreenSize();

	flags |= IS_VISIBLE;
	fgClr = Defaults.WndForeground;
	bgClr = Defaults.RootWindowClr;
	font  = Defaults.WndFont;
	
	width  = d.width;
	height = d.height;

	if ( Defaults.RootWindowImg != null ) {
		ImageDict id = ImageDict.getDefaultDict();
		Class ac = null;
		
		ac = GoodStuff.getDefaultGoodStuff().getAppCls();
		if ( ac == null ) ac = getClass();
		
		wallPaper = id.putImage( null, Defaults.RootWindowImg, ac.getName(), this, true, false );
		if ( wallPaper != null ){
			ret = Toolkit.getDefaultToolkit().checkImage( wallPaper, 0, 0, this);
			if ( (ret & ImageObserver.ABORT) != 0 )
				wallPaper = null;
		}
	}

	AWTEvent.root = this;
	singleton = this;
	page = 1;

	addKeyListener( this);
	addMouseListener( this);

	addNotify();

	loadBoxes();

	VTSwitchEvent.setHandler( this );
}

public Component add ( Component child, int index ) {
	addImpl( child, null, index);

	if ( controlBox != null )
		controlBox.added( (Window)child);
	
	return child;
}

protected void addImpl(Component child, Object constraints, int index ) {

	if (index < -1 || index > nChildren) {
		throw new IllegalArgumentException("bad index: " + index);
	}

	if ( child.parent != null ) {
		child.parent.remove(child);
	}

	if ( children == null ) {
		children= new Component[3];
	}
	else if ( nChildren == children.length ) {
		Component[] old = children;
		children = new Component[ nChildren * 2];
		System.arraycopy( old, 0, children, 0, nChildren);
	}

	if (index < 0 || nChildren == 0 || index == nChildren) {  // append
		children[nChildren] = child;
	}
	else if (index < nChildren) {     // insert at index
		System.arraycopy( children, index, children, index+1, nChildren - index);
		children[index] = child;
	}

	nChildren++;
	child.parent = this;

	if ( (cntrListener != null) || (eventMask & AWTEvent.CONTAINER_EVENT_MASK) != 0 ){
		AWTEvent.sendEvent( ContainerEvt.getEvent( this,
		                                           ContainerEvent.COMPONENT_ADDED, child), false);
	}
}

public void addNotify () {
	if ( nativeData == null ) {
		createNative();
		flags |= IS_ADD_NOTIFIED;

		// Two reasons for doing a sync paint (instead of just doing a repaint):
		// (a) we should draw as fast as possible (because the RootWindow should
		// have a "per se" feeling), (b) slightly more subtle: the background
		// painting might get confused if other windows pop up (e.g. a Console)
		// before the initial paint is completed (we don't do any subsequent
		// child clipping, but rely on the native layer to just repaint unobscured
		// root parts).
		Graphics g = getGraphics();
		if ( g != null ) {
			paint( g);
			g.dispose();
		}
		
		AWTEvent.registerSource( this, nativeData);
	}
}

int currentPage () {
	return getPageNum( page);
}

void dumpChildren () {
	for ( int i=0; i<nChildren; i++ ) {
		System.out.println( "[" + i + "]: " + children[i]);
	}
}

void endResize ( Frame c ) {
	if ( isResizing ) {
		isResizing = false;
	
		if ( c.lastLoc == 0 )
			setCursor( Frame.decoCursor[0]);
	
		gr.drawRect( sizeRect.x, sizeRect.y, sizeRect.width, sizeRect.height);
		
		gr.dispose();
		gr = null;

		c.setBounds( sizeRect.x, sizeRect.y, sizeRect.width+1, sizeRect.height+1);
	}
	
	sizeRect = null;
}

static RootWindow getDefaultRootWindow () {
	return singleton;
}

String getGenericName () {
	// this is not a standard Java component, but since it's a Window, getting its
	// generic name woul alter the name count. You bet -there's an application out
	// there which relies on the numbering
		
	return "rootwindow";
}

static int getPageNum ( int page ) {
	int i, n;
	
	for ( i=1, n=1; (page & n) == 0; n<<=1, i++ );
	return i;
}

public void keyPressed ( KeyEvent evt ){
}

public void keyReleased ( KeyEvent evt ){
}

public void keyTyped ( KeyEvent evt ){
	char c = evt.getKeyChar();
	
	if ( c == '?' )
		dumpChildren();
	else if ( c == 'u' )
		repaintAll();
	else if ( (c >= '1') && (c <= '9') )
		showPage( c - '0');
}

public void layout () {
	// don't do nothing
}

void loadBoxes () {
	String clsName;
	Class cls;

	try {
		if ( (clsName = GoodStuff.getDefaultGoodStuff().getControlBox()) == null )
			clsName = Defaults.ControlBoxClass;

		if ( clsName != null ) {
			cls = Class.forName( clsName);
			controlBox = (ControlBox) cls.newInstance();
			
			if ( controlBox instanceof MinBox ){
				minBox = (MinBox)controlBox;
				return;
			}
		}

		if ( (clsName = GoodStuff.getDefaultGoodStuff().getMinBox()) == null )
			clsName = Defaults.MinBoxClass;

		if ( clsName != null ) {
			cls = Class.forName( clsName);
			minBox = (MinBox) cls.newInstance();
		}
	}
	catch ( Exception xx ) {
		System.err.println( "could not create RootWindow modules: " + xx);
	}
}

public static void main ( String[] arg ) {
	RootWindow r = getDefaultRootWindow();
}

void maximize ( Window w ) {
	w.setBounds( 0, 0, width, height);
}

void minimize ( Window w ) {
	int i;

	if ( minBox == null ){
		return;  // there won't be a way to restore the window
	}

	w.setVisible( false);
	
	for ( i=0; i<nChildren; i++ ) { // set focus on next visible window
		if ( ((children[i].flags & IS_VISIBLE) != 0) && (children[i] instanceof Frame) ){
			AWTEvent.sendEvent( FocusEvt.getEvent( children[i], FocusEvt.FOCUS_GAINED, false), false);
			break;
		}
	}

	if ( i==nChildren ) {            // no visible child, focus yourself
		AWTEvent.sendEvent( FocusEvt.getEvent( this, FocusEvt.FOCUS_GAINED, false), false);
	}
	
	minimized.addElement( w);

	if ( minBox != null )
		minBox.updateContents();
}

public void mouseClicked ( MouseEvent evt ) {
}

public void mouseEntered ( MouseEvent evt ) {
}

public void mouseExited ( MouseEvent evt ) {
}

public void mousePressed ( MouseEvent evt ) {
		AWTEvent.sendEvent( FocusEvt.getEvent( this, FocusEvt.FOCUS_GAINED, false), false);
}

public void mouseReleased ( MouseEvent evt ) {
}

public void paint ( Graphics g ) {
	// this is a really speed relevant method, even small delays disturb the
	// "background" feeling of the RootWindow

	int       x0, y0, x, y;
	int       w, h, cxw, cyh;
	int       cx = g.getClipX();
	int       cy = g.getClipY();
	int       cw = g.getClipWidth();
	int       ch = g.getClipHeight();

	if ( (wallPaper != null) && 
	     ((w = wallPaper.getWidth(this)) > 0) &&
	     ((h = wallPaper.getHeight(this)) > 0) ) {
		// draw relevant wallpaper tiles (and nothing else, to speed up)

		cxw = cx + cw;
		cyh = cy + ch;

		if ( (w > 0) && (h > 0) ) {
			x0 = (cx / w) * w;
			y0 = (cy / h) * h;

			for ( y=y0; y < cyh; y += h ) {
				for ( x=x0; x < cxw; x += w ){
					g.drawImage( wallPaper, x, y, this);
				}
			}
		}
	}
	else {
		g.clearRect( cx, cy, cw, ch);
	}
}

void prepareForResize ( Frame c, int x, int y ) {
	sizeOffset.setLocation( x, y);
	sizeRect = new Rectangle( c.x, c.y, c.width-1, c.height-1);
}

boolean preparedForResize ( Frame c ) {
	return (sizeRect != null);
}

public void processVTSwitchEvent ( VTSwitchEvent e ){
  if (e.getID() == VTSwitchEvent.VT_REL_REQ) {
        // Doesn't seem to be sufficient to shut things down
        // We need to be able to short-circuit all drawing to 
        // to the screen 

        // setVisible (false);

        e.ackRelease();
  }
  if (e.getID() == VTSwitchEvent.VT_ACQ_REQ) {
        // setVisible (true); 

        e.ackAcquire();

      	repaintAll();
  }
}

void redrawMinimized ( int idx ) {
}

public void remove (int index) {
	Component c = children[index];
	int       n = nChildren - 1;

	if ( (c.flags & IS_ADD_NOTIFIED) != 0 ) // Huhh? we normally should come in from there
		c.removeNotify();

	System.arraycopy(children, index+1, children, index, n-index);
	children[n] = null;
	nChildren--;

	if ( nChildren == 0)
		AWTEvent.unregisterSource( this, nativeData);

	if ( (AWT.activeWindow == null) &&
	         !Toolkit.eventQueue.hasPendingEvents( null, FocusEvt.FOCUS_GAINED) ) {
		// set focus on topmost visible window
		for ( int i=0; i<nChildren; i++ ){
			if ( (children[i].flags & IS_VISIBLE) != 0 ){
				children[i].requestFocus();
				break;
			}
		}
	}

	if ( controlBox != null )
		controlBox.removed( (Window)c);
}

public void repaint ( int x, int y, int width, int height ) {
	// no need to repaint async
	Graphics g = NativeGraphics.getGraphics( this, nativeData,
		                                   NativeGraphics.TGT_TYPE_WINDOW, true,
		                                   0, 0, x, y, width, height,
		                                   fgClr, bgClr, font);
	if ( g != null ) {
		paint( g);
		g.dispose();
	}
}

void repaintAll () {
	int i;
	Component c;

	repaint( 0, 0, width, height);
	
	for ( i=nChildren-1; i>=0; i-- ) {
		c = children[i];
		if ( (c.flags & IS_VISIBLE) != 0 )
			c.repaint( 0, 0, c.width, c.height);
	}
}

void requestFocus ( Window c ) {
	c.requestFocusImpl();
}

void resize ( Frame c, int x, int y ) {
	if ( sizeRect == null )
		return;

	if ( !isResizing ) {
		isResizing = true;
		gr = NativeGraphics.getGraphics( this, nativeData,
		                                   NativeGraphics.TGT_TYPE_WINDOW, true,
		                                   0, 0, 0, 0, width, height,
		                                   fgClr, bgClr, font);		
		if ( c.lastLoc == 0 )
			Toolkit.wndSetCursor( nativeData,
				                    Cursor.getPredefinedCursor( Cursor.MOVE_CURSOR).type);
	
		gr.setColor( Color.orange);
		gr.setXORMode( bgClr);
		gr.drawRect( sizeRect.x, sizeRect.y, sizeRect.width, sizeRect.height);
	}
	else {
		gr.drawRect( sizeRect.x, sizeRect.y, sizeRect.width, sizeRect.height);
		if ( c.lastLoc == 0 ) {
			sizeRect.translate( x -sizeOffset.x, y -sizeOffset.y);
			sizeOffset.x = x; sizeOffset.y = y;
		}
		else {
			if ( (c.lastLoc & FrameDeco.NORTH) != 0 ) {
				sizeRect.y = y+c.y; sizeRect.height = c.height - y -1;
			}
			if ( (c.lastLoc & FrameDeco.EAST) != 0 ) {
				sizeRect.width = x;
			}
			if ( (c.lastLoc & FrameDeco.SOUTH) != 0 ) {
				sizeRect.height = y;
			}
			if ( (c.lastLoc & FrameDeco.WEST) != 0 ) {
				sizeRect.x = x+c.x; sizeRect.width = c.width - x -1;
			}
		}
		gr.drawRect( sizeRect.x, sizeRect.y, sizeRect.width, sizeRect.height);
	}
}

void restore ( Window w ) {
	if ( minimized.contains( w) ){
		minimized.removeElement( w);
		if ( minBox != null )
			minBox.updateContents();
	}

	if ( w != children[0] )
		w.toFront();
	w.restore();
	
	AWTEvent.sendEvent( FocusEvt.getEvent( w, FocusEvt.FOCUS_GAINED, false), false);
}

void setBounds ( Window w, int xNew, int yNew, int wNew, int hNew ) {
	w.setBoundsImpl( xNew, yNew, wNew, hNew);
}

void setVisible ( Window w, boolean showIt ) {
	w.setVisibleImpl( showIt);

	if ( w.isVisible() && (w.page != 0xffff) )
		w.page = page;
		
	// For some obscure reasons, popup windows don't automatically get the focus
	// when opened. Swing relies on this behavior for its heavyweight popup
	// keyboard interface
	if ( !w.isPopup() )
		w.requestFocus();
}

void showPage ( int pageNum ) {
	Window c;
	int lastPage = page;

	page = 1 << (pageNum -1);

	requestFocus(); // we don't want to leave an activeWindow behind us

	for ( int i=nChildren-1; i>=0; i-- ) {
		c = (Window) children[i];

		if ( (c.page & page) != 0 ){
			if ( (c.flags & IS_TEMP_HIDDEN) != 0 ){
				c.setVisible( true);
				c.flags &= ~IS_TEMP_HIDDEN;
			}
		}
		else {
			if ( c.isVisible() ){
				c.setVisible( false);
				// Maybe some windows get automatically closed when they become invisible.
				// We don't try to restore them because there might be all kinds of internal
				// states (selections etc.), we don't know about
				if ( (c.flags & IS_ADD_NOTIFIED) != 0 ){
					c.flags |= IS_TEMP_HIDDEN;
				}
			}
		}
	}

	if ( controlBox != null )
		controlBox.switchPage( getPageNum( lastPage), pageNum);
}

void toBack ( Window c ) {
	int i, n = nChildren - 1;

	for ( i=0; i<n; i++ ) {
		if ( children[i] == c ) {
			if ( c.nativeData != null )
				Toolkit.wndToBack( c.nativeData);
			System.arraycopy( children, i+1, children, i, n-i);
			children[n] = c;
		}
	}
	
	if ( controlBox != null )
		controlBox.toFront( c);
}

void toFront ( Window c ) {
	for ( int i=1; i<nChildren; i++ ) {
		if ( children[i] == c ) {
			if ( c.nativeData != null ){
				Toolkit.wndToFront( c.nativeData);
			}
			System.arraycopy( children, 0, children, 1, i);
			children[0] = c;
			
			c.raised(); // inform Window that it got raised
		}
	}
	
	if ( controlBox != null )
		controlBox.toFront( c);
}

public void update ( Graphics g ) {
	// this is resolved because we (1) are always vivible, and (2) the native
	// layer takes care of sending us update events only for unobscured areas
	paint( g);
}
}
