/**
 * WinTaskBar -
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */

package java.awt;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Calendar;
import java.util.Date;
import java.util.Vector;
import kaffe.awt.GoodStuff;
import kaffe.util.Timer;
import kaffe.util.TimerClient;

class WinTaskBar
  extends Window
  implements ActionListener, ControlBox, MinBox, MouseListener, WindowListener, TimerClient
{
	GoodStuff goodStuff;
	RootWindow root;
	static Calendar calendar;
	static Date now;
	static String[] MonthName = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	Menu startMenu;
	FontMetrics fm;
	int clockWidth;
	int pageWidth;
	int startWidth;
	String startLabel;
	int maxButtonWidth;
	Vector frames = new Vector ( 5);
	int nPages;
	int frameWidth;
	static String[] pageLabel = { "1", "2", "3", "4" };

static {
	calendar = Calendar.getInstance();
	now = new Date();
}

WinTaskBar () {
	page = 0xffff;
	root = RootWindow.singleton;
	goodStuff = GoodStuff.getDefaultGoodStuff();
	font = new Font( "SansSerif", Font.PLAIN, 11);
	fm = Toolkit.getDefaultToolkit().getFontMetrics( font);

	nPages = Math.min( 2, pageLabel.length);

	buildStartMenu();
	clockWidth = fm.stringWidth( "24:24  May 13");
	startLabel = "Start";
	startWidth = fm.stringWidth( startLabel) + 20;
	maxButtonWidth = fm.stringWidth( "W") * 15;
		
	height = fm.getHeight() * 3 / 2;
	setBounds( 0, root.height - height, root.width, height);

	setVisible( true);
	pageWidth = nPages*height;

	Timer.getDefaultTimer().addClient( this, 1000, 60000);
	addMouseListener( this);
}

public void actionPerformed ( ActionEvent e ) {
	String cmd = e.getActionCommand();
	int i, n=goodStuff.size();

	for ( i=0; i<n; i++ ) {
		if ( goodStuff.getItemLabel( i).equals( cmd) )
			goodStuff.trigger( i);
	}
	
	if ( cmd.equals( "Shutdown") ){
		Toolkit.terminate();
		System.exit( 0);
	}
}

public void added ( Window c ) {
	if ( !(c instanceof Frame) ) return;

	c.addWindowListener( this);

	frames.addElement( c);
	drawFrames();
}

void buildStartMenu () {
	int i, n = goodStuff.size();
	
	startMenu = new Menu();
	
	for ( i=0; i<n; i++ ){
		startMenu.add( goodStuff.getItemLabel( i));
	}		
	startMenu.addSeparator();
	startMenu.add( "Shutdown");
	
	startMenu.addActionListener( this);
}

void drawClock ( Graphics g ) {
	int hour, min, mnth, day;
	String time, date;

	now.setTime( System.currentTimeMillis());
	calendar.setTime( now);

	hour = calendar.get( Calendar.HOUR_OF_DAY);
	min = calendar.get( Calendar.MINUTE);
	mnth = calendar.get( Calendar.MONTH);
	day = calendar.get( Calendar.DAY_OF_MONTH);	
	
	time = ((hour < 10) ? "0" : "") + hour + ((min < 10) ? ":0" : ":") + min;
	date  = MonthName[mnth] + ' ' + day;

	drawClock ( g, time, date );
}

void drawClock ( Graphics g, String time, String date ) {
	int d = 2;
	String dtg = time + "  " + date;
	int w = fm.stringWidth( dtg);
	int u = width - clockWidth -d;
	int h = height - 2*d;
	
	g.setColor( bgClr);
	g.draw3DRect( u, d, clockWidth, h, false);
	g.setColor( Color.black);
	g.fillRect( u+1, d+1, clockWidth-1, h-1);
	
	g.setColor( Color.green);
	g.drawString( dtg, width - clockWidth - d + (clockWidth - w)/2, d + fm.getAscent());
}

void drawFrames () {
	Graphics g = getGraphics();
	if ( g != null ) {
		drawFrames( g);
		g.dispose();
	}
}

void drawFrames ( Graphics g ) {
	Frame fr;
	int i, n = frames.size();
	int u, d = 2, dd=4;
	int u1 = width - clockWidth - pageWidth;

	u = startWidth + dd;
	if ( n > 0 ) {
		frameWidth = (u1 - startWidth - 2*dd) / n;
		if ( frameWidth > maxButtonWidth )
			frameWidth = maxButtonWidth;
	
		for ( i=0; i<n; i++ ) {
			fr = (Frame) frames.elementAt( i);
			g.setColor( bgClr);
			g.fill3DRect( u, d, frameWidth-d, height-dd, !fr.isActive);
			g.setColor( fgClr);
			g.drawString( fr.getTitle(), u+d, d + fm.getAscent());
			u += frameWidth;
		}
	}

	g.setColor( bgClr);
	g.fillRect( u, d, u1 - u, height-dd);
}

void drawPages () {
	Graphics g = getGraphics();
	if ( g != null ) {
		drawPages( g);
		g.dispose();
	}
}

void drawPages ( Graphics g ) {
	int i;
	int d = 2;
	int current = root.currentPage();
	int u = width - clockWidth - pageWidth;
	int h = height - 2*d;
	Color dark = bgClr.darker();
	Color bright = bgClr.brighter();
	String s;

	for ( i=1; i<=nPages; i++ ) {
		g.setColor( bright);
		g.drawOval( u+1, d, h, h);

		g.setColor( dark);
		g.drawOval( u, d, h, h);

		s = pageLabel[i-1];
		g.setColor( (i == current) ? Color.red : fgClr);
		g.drawString( s, u + (h - fm.stringWidth(s))/2 + 1, d + fm.getAscent() );
		
		u += height;
	}
}

void drawStartMenu ( Graphics g ) {
	int d = 2;
	int w = fm.stringWidth( startLabel);
	
	g.setColor( bgClr);
	g.draw3DRect( d, d, startWidth, height -2*d-1, true);

	g.setColor( fgClr);
	g.drawString( startLabel, d + (startWidth - w) / 2, d+fm.getAscent());
}

String getGenericName () {
	// this is not a standard Java component, but since it's a Window, getting its
	// generic name woul alter the name count. You bet -there's an application out
	// there which relies on the numbering
		
	return "taskbar";
}

public void mouseClicked ( MouseEvent evt ) {
}

public void mouseEntered ( MouseEvent evt ) {
}

public void mouseExited ( MouseEvent evt ) {
}

public void mousePressed ( MouseEvent evt ) {
	int x = evt.getX();
	int i, n = frames.size();
	int d = 2;
	int u = startWidth + 2*d;
	
	if ( x < u ){
		popupStartMenu();
		return;
	}
		
	for ( i=0; i<n; i++ ) {
		u += frameWidth;
		if ( x < u ){
			root.restore( (Frame)frames.elementAt( i));
			return;
		}
	}
	
	u = width - clockWidth - pageWidth;
	if ( x < u )
		return;
		
	for ( i=1; i<=nPages; i++ ){
		u += height;
		if ( x < u ) {
			root.showPage( i);
			return;
		}
	}
}

public void mouseReleased ( MouseEvent evt ) {
}

public void paint ( Graphics g ) {
	g.setColor( Color.lightGray);
	g.fill3DRect( 0, 0, width, height, true);

	drawStartMenu( g);
	drawFrames( g);
	
	drawPages( g);
	drawClock( g);
	
}

void popupStartMenu () {
	PopupMenu p = new PopupMenu( startMenu);
	p.show( this, 0, 0);
}

public void removed ( Window c ) {
	if ( !(c instanceof Frame) ) return;
	
	frames.removeElement( c);
	drawFrames();
}

public void switchPage ( int oldPage, int newPage ) {
	drawPages();
}

public void timerExpired ( Timer timer ) {
	Graphics g = getGraphics();

	if ( g != null ) {
		drawClock( g);	
		g.dispose();
	}
}

public void toBack ( Component c ) {
	// not interested in this, we react on focus change
}

public void toFront ( Component c ) {
	// not interested in this, we react on focus change
}

public void updateContents () {
}

public void windowActivated ( WindowEvent evt ) {
	drawFrames();
}

public void windowClosed ( WindowEvent evt ) {
	drawFrames();
}

public void windowClosing ( WindowEvent evt ) {
}

public void windowDeactivated ( WindowEvent evt ) {
	drawFrames();
}

public void windowDeiconified ( WindowEvent evt ) {
	drawFrames();
}

public void windowIconified ( WindowEvent evt ) {
	drawFrames();
}

public void windowOpened ( WindowEvent evt ) {
}
}
