package java.awt;

import java.awt.BorderLayout;
import java.awt.FocusEvt;
import java.awt.event.ComponentEvent;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.PaintEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.peer.ComponentPeer;
import java.util.Enumeration;
import java.util.Vector;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.awt.OpaqueComponent;
import kaffe.awt.theme.ComponentDeco;
import kaffe.util.Ptr;

/**
 * Window - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P.C.Mehlitz
 */
public class Window
  extends Container
  implements OpaqueComponent
{
	Ptr nativeData;
	WindowListener wndListener;
	Window owner;
	int page;
	static Window dummy = new Window();
	static int named;
	Object ownees;

Window () {
	// windows aren't visible per se, but they are exposed, colored and fontified
	flags = (IS_PARENT_SHOWING | IS_BG_COLORED | IS_FG_COLORED | IS_FONTIFIED);

	// FIX ME should be queried from deco
	fgClr = Defaults.WndForeground;
	bgClr = Defaults.WndBackground;
	font  = Defaults.WndFont;

	setLayout(new BorderLayout());

	AWT.windows.addElement( this);
}

public Window ( Frame owner ) {
	this();

	this.owner = owner;
	
	if ( Defaults.StrictPJ && (owner == null) ) {
		throw new IllegalArgumentException( "null parent frame");
	}
}

public void addNotify () {
	if ( nativeData == null ) {
		// before we do anything else, make sure the RootWindow is initialized
		Window root = RootWindow.singleton;
	
		if ( owner != null ) {
			if (owner.nativeData == null ){
				owner.addNotify();
			}

			owner.addOwnee( this); // we have to take care of Z-order by ourself
		}

		Toolkit.createNative( this);

		if ( nativeData == null )
			throw new AWTError( "native create failed: " + this);
		
		AWTEvent.registerSource( this, nativeData);
		super.addNotify();

		if ( parent == null ){  // let the root window know about us
			root.add( this, 0);
			parent = null;
		}
	}
}

void addOwnee ( Window w ) {
	Vector v;

	if ( ownees == null ){
		ownees = w;
	}
	else {
		if ( ownees instanceof Vector ){
			((Vector)ownees).addElement( w);
		}
		else if ( ownees instanceof Component ){
			v = new Vector( 2);
			v.addElement( ownees);
			v.addElement( w);
			ownees = v;
		}
	}
}

public void addWindowListener ( WindowListener newListener ) {
	wndListener = AWTEventMulticaster.add( wndListener, newListener);
}

void cleanUpNative () {
	if ( nativeData != null ) {
		AWTEvent.unregisterSource( this, nativeData);
		nativeData = null;
	}
}

void createNative () {
	nativeData = Toolkit.wndCreateWindow( (owner != null) ? owner.nativeData : null,
                                  x, y, width, height,
	                                cursor.type, bgClr.nativeValue);
}

void destroyNative () {
	Toolkit.wndDestroyWindow( nativeData);
	cleanUpNative();
}

public void dispose () {
	// prevent further drawing (might cause trouble for native windowing system)
	// we don't use hide(), since this would trigger superfluous native action
	flags &= ~IS_VISIBLE;

	// get rid of ownees as well (no native WM who does this for us)
	if ( ownees != null ){
		if ( ownees instanceof Window ){
			((Window)ownees).dispose();
		}
		else if ( (ownees instanceof Vector) && !((Vector)ownees).isEmpty() ){
			for ( Enumeration e= ((Vector)ownees).elements(); e.hasMoreElements(); ){
				((Window) (e.nextElement())).dispose();
			}
		}
	}

	if ( nativeData != null ) {
		removeNotify();
	}

  AWT.windows.removeElement( this);
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

public Component getFocusOwner () {
	return ( this == AWT.activeWindow ) ? AWT.keyTgt : null;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "win" + (named++);
}

public Container getParent () {
	return owner;
}

public ComponentPeer getPeer () {
	// this is just a dummy, i.e. we share a single peer object that can be used
	// ONLY to "(getPeer() != null)" check if we already passed addNotify()
	return ((flags & IS_ADD_NOTIFIED) != 0) ? Toolkit.windowPeer : null;
}

boolean hasModalOwnee () {
	Window w;
	Vector v;
	int i;
	
	if ( ownees == null ){
		return false;
	}
	else {
		if ( ownees instanceof Window ){
			w = (Window) ownees;
			return ( (w.flags & (IS_MODAL|IS_SHOWING)) == (IS_MODAL|IS_SHOWING) );
		}
		else if ( ownees instanceof Vector ){
			v = (Vector) ownees;
			for ( i=0; i<v.size(); i++ ){
				w = (Window) v.elementAt( i);
				if ( (w.flags & (IS_MODAL|IS_SHOWING)) == (IS_MODAL|IS_SHOWING) ){
					return true;
				}
			}
		}
	}
	
	return false;
}

boolean hasOwnee( Window ownee ) {
	if ( ownees != null ){
		if ( ownees == ownee ){
			return true;
		}
		else if ( ownees instanceof Vector ){
			return ((Vector)ownees).contains( ownee);
		}
	}
	
	return false;
}

public void hide () {
	// DEP - should be in setVisible

	// give Rootwindow a chance to intervene
	RootWindow.singleton.setVisible( this, false);
}

public boolean isFocusTraversable (){
	return false;
}

boolean isPopup () {
	return (owner != null);
}

public void pack () {
	if ( nativeData == null ) {
		addNotify();
	}
	setSize( getPreferredSize());
	
	// this happens to be one of the automatic validation points
	validate();
}

void process ( FocusEvent event ) {
	super.process( event);

	if ( (event.id == FocusEvent.FOCUS_GAINED) && !isFocusTraversable() ) {
		// Set focus on first child which can handle it. We do this automatic focus forwarding
		// ONLY if there are no other pending requests, and we do it sync (we are already in the
		// thread which processed the FOCUS_GAINED) to avoid any interference with popup focus
		// transitions. This is because we might otherwise get a out-of-order focus event:
		// (popup1-lost -> owner-gained ->post forward , popup2-gained, forward-gained ->popup2-lost)
		if ( (nChildren > 0) && !Toolkit.eventQueue.hasPendingEvents( null, FocusEvt.FOCUS_GAINED) ) {
		   Component c = ShortcutHandler.focusNext( this);
		   if ( (c != null) && (c != this) ) {
	       		AWTEvent.sendEvent( FocusEvt.getEvent( c, FocusEvt.FOCUS_GAINED, false), true);
		   }
		}
	}
}

void process ( WindowEvent e ) {
	if ( (wndListener != null) || (eventMask & AWTEvent.WINDOW_EVENT_MASK) != 0){
		processEvent( e);
	}
	if ( owner != null ){
		owner.redrawDeco();
	}
}

protected void processWindowEvent ( WindowEvent event ) {
	// This is a artificial constraint - we could happily emit ACTIVATED/DEACTIVATED
	// events with out popup focus mechanism, but the JDK class docu says it just
	// handles OPENED/CLOSED.
	if ( wndListener != null ) {
		switch ( event.id ) {
		case WindowEvent.WINDOW_OPENED:
			wndListener.windowOpened( event);
			break;
		case WindowEvent.WINDOW_CLOSED:
			wndListener.windowClosed( event);
			break;
		}
	}
}

void raised () {

	// raise ownees as well (no native WM who does it for us)
	if ( ownees != null ){
		if ( ownees instanceof Window ){
			RootWindow.singleton.toFront((Window)ownees);
		}
		else if ( (ownees instanceof Vector) && !((Vector)ownees).isEmpty() ){
			for ( Enumeration e= ((Vector)ownees).elements(); e.hasMoreElements(); ){
				Window o = (Window) e.nextElement();
				RootWindow.singleton.toFront( o);
			}
		}
	}
}

void redrawDeco (){
	// nothing to do here
}

public void removeNotify () {
	if ( (flags & IS_ADD_NOTIFIED) != 0 ) {
		// use this rather than nativeData, because the sync FOCUS_LOST might get us recursive
		flags &= ~IS_ADD_NOTIFIED;

		if ( owner != null ) {
			owner.removeOwnee( this);
			// don't null 'owner' because it would break Swings assumption on
			// Window.getParent() (which never gets 'null')
		}

		// This isn't particularly nice - in order to get popup focus "forwarding"
		// (see process(KeyEvent)) working, we have to make sure we get our FOCUS_LOST
		// processed (to reset the restoreFocus). 
		if ( (AWT.activeWindow == this) && (AWT.keyTgt != null) ){
			AWTEvent.sendEvent( FocusEvt.getEvent( AWT.keyTgt,
			                                       FocusEvent.FOCUS_LOST, false), true);
		}

 		// go super *before* getting recursive by a parent.remove(idx)
		super.removeNotify();

		RootWindow.singleton.remove( this);

		Toolkit.destroyNative( this);
		
		//modal dialog dispatcher exit
		AWTEvent.sendEvent( WindowEvt.getEvent( this, WindowEvent.WINDOW_CLOSED), false);
	}
}

void removeOwnee( Window w ) {
	if ( ownees != null ){
		if ( ownees == w ){
			ownees = null;
		}
		else if ( ownees instanceof Vector ){
			// we keep it - if we got more than one once, this might happen again
			((Vector) ownees).removeElement( w);
		}
	}
}

public void removeWindowListener ( WindowListener listener ) {
	wndListener = AWTEventMulticaster.remove( wndListener, listener);
}

public void requestFocus () {
	RootWindow.singleton.requestFocus( this);
}

void requestFocusImpl () {
        if ( AWT.keyTgt != null ) {
            if ( (AWT.keyTgt == this) || (isAncestorOf(AWT.keyTgt)) ||
	          Toolkit.eventQueue.hasPendingEvents( this, FocusEvt.FOCUS_GAINED) )
	    return;
        }
   
	Toolkit.eventQueue.postEvent( FocusEvt.getEvent( this, FocusEvt.FOCUS_GAINED, false));
}

public void reshape ( int xNew, int yNew, int wNew, int hNew ) {
	// DEP - should be in setBounds

	// give RootWindow a chance to intervene
	RootWindow.singleton.setBounds( this, xNew, yNew, wNew, hNew);
}

void restore () {
	// windows can't be minimized or maximized
}

public void setBackground ( Color clr ) {
	// we can't nullify this (rermember the "bgClr != null" invariant)
	if ( clr != null ) {
		bgClr = clr;
		propagateBgClr( clr);
		
		if ( isShowing() )
			repaint();
	}
}

void setBoundsImpl ( int xNew, int yNew, int wNew, int hNew ) {
	boolean moved   = (x != xNew) || (y != yNew);
	boolean resized = (width != wNew) || (height != hNew);
	int     id = (resized) ? ComponentEvent.COMPONENT_RESIZED :
	                         ComponentEvent.COMPONENT_MOVED;

	if ( resized )
		invalidate();

	x      = xNew;
	y      = yNew;
	width  = wNew;
	height = hNew;

	if ( nativeData != null ) {
		Toolkit.wndSetBounds( nativeData, xNew, yNew, wNew, hNew,
		                      ((flags & IS_RESIZABLE) != 0));

		if ( (flags & Component.IS_VALID) == 0 ){
			// if we are already visible, we wait for the subsequent
			// expose (fake repaint by temp changing visibility)
			if ( (flags & Component.IS_VISIBLE) != 0 ) {
				flags &= ~Component.IS_PARENT_SHOWING;
				propagateParentShowing( true);
			
				validate();
			
				flags |= Component.IS_PARENT_SHOWING;
				propagateParentShowing( false);
			}
			else {
				validate();
			}
		}

		Toolkit.eventQueue.postEvent( ComponentEvt.getEvent( this, id));
		
		propagateReshape();
	}
}

public void setFont ( Font fnt ) {
	// we can't nullify this (remember the "font != null" invariant)
	if ( fnt != null ) {
		font = fnt;
		propagateFont( fnt);
		
		if ( isShowing() )
			repaint();
	}
}

public void setForeground ( Color clr ) {
	// we can't nullify this (rermember the "fgClr != null" invariant)
	if ( clr != null ) {
		fgClr = clr;
		propagateFgClr( clr);
		
		if ( isShowing() )
			repaint();
	}
}

void setNativeCursor ( Cursor cursor ) {
	if ( nativeData != null )
		Toolkit.wndSetCursor( nativeData, cursor.type);
}

void setOwner ( Window owner ) {
	this.owner = owner;
	
	if ( owner != null ){	
		owner.addOwnee( this);
	}
	else {
		owner.removeOwnee( this);
	}
}

void setVisibleImpl ( boolean showIt ) {

	if ( showIt ) {
		if ( nativeData == null )
			addNotify();

		// this happens to be one of the automatic validation points, and it should
		// cause a layout *before* we get visible
		validate();

		if ( (flags & IS_VISIBLE) != 0 ) {
			toFront();
		}
		else {
			super.show();

			if (nativeData != null)
				Toolkit.wndSetVisible( nativeData, true);

			// the spec says that WINDOW_OPENED is delivered the first time a Window
			// is shown, and JDK sends this after it got shown, so this is the place
			if ( (flags & IS_OPENED) == 0 ) {
				flags |= IS_OPENED;

				if ( (wndListener != null) || (eventMask & AWTEvent.WINDOW_EVENT_MASK) != 0 ){
					AWTEvent.sendEvent( WindowEvt.getEvent( this,
					                        WindowEvent.WINDOW_OPENED), false);
				}
			}
		}

		// Note that we don't set the focus on Windows per default. With respect to Frames,
		// this is the resposibility of the WindowManager. With respect to popups, Swing
		// heavyweights (JWindowPopup) heavily rely on Windows NOT getting the focus unless a explicit
		// requestFocus occurs (for keyboard input and the famous ComboBox "first for Solaris" jitter)
		Toolkit.eventQueue.postEvent( ComponentEvt.getEvent( this, ComponentEvt.COMPONENT_SHOWN));
	}
	else {
		super.hide();

		if (nativeData != null) {
			Toolkit.wndSetVisible( nativeData, false);
		}
		Toolkit.eventQueue.postEvent( ComponentEvt.getEvent( this, ComponentEvt.COMPONENT_HIDDEN));
	}
}

public void show () {
	// DEP - should be in setVisible

	// give Rootwindow a chance to intervene
	RootWindow.singleton.setVisible( this, true);
}

public void toBack () {
	RootWindow.singleton.toBack( this);
}

public void toFront () {
	RootWindow.singleton.toFront( this);
}

public void validate() {
	if ( AWT.themeChanged) {
		propagateBgClr( bgClr);
		propagateFgClr( fgClr);
		propagateFont( font);
	}
	super.validate();
}
}
