package java.awt;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.util.Vector;
import kaffe.awt.AWT;
import kaffe.awt.theme.ComponentDeco;
import kaffe.awt.Defaults;

/**
 * class Choice - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Choice
  extends Container
  implements ItemSelectable, ActionListener, ItemListener, KeyListener
{
	final private static long serialVersionUID = -4075310674757313071L;
	Vector items = new Vector();
	Object selection;
	ItemListener iListener;
	ChoiceWindow prompter;
	boolean open;
	Selection entry;
	int state;
	final static int HILIGHTED = 1;
	static int named;
	static ClassProperties props = new ClassProperties();

class ChoiceWindow
  extends Window
{
// We extends List to check for the selected item against what
	// we think is selected.  Again, this 'extension' is for PL to
	// allow us to use Choices as Menus.
	List list = new List() 
{
		public boolean isIndexSelected ( int index) {
			String item = this.getItem(index);
			if (item.equals(Choice.this.getSelectedItem())) {
				return (true);
			}
			else {
				return (false);
			}
		}
	}
;

public ChoiceWindow ( Frame owner ) {
	super( owner);

	//has to be reinitailized
	this.deco = this.getDeco();
	this.insets = this.deco.getDecoInsets();
	
	this.setLayout( null);
	list.removeHScroll();
	this.add( list);

	//would override deco background
	if (ChoiceWindow.this.deco.hasBackgroundImage()) {
		ChoiceWindow.this.flags &= ~IS_BG_COLORED;
	}
}

public void doLayout() {
	Insets in = this.insets;
	int db = 1;
	list.setBounds( in.left + db, in.top + db, this.width-in.left-in.right-2*db, this.height-in.top-in.bottom-2*db);
}

ComponentDeco getReferenceDeco() {
	//not initialized in the ChoiceWindow super constructor chain
	if ( Choice.this == null ) {
		return null;
	}
	return Choice.this.deco;
}

public void paint( Graphics g) {
	super.paint( g);
	
	//some simple border theme independant
	g.setColor( Color.black);
	g.drawRect( 0, 0, this.width-1, this.height-1);
}

void popUpAt( int x, int y, int width, int height) {
	int is = items.size();

	this.setBounds( x, y, width, height);
	list.setFont( Choice.this.font);

	list.removeAll();
	for ( int i=0; i<is; i++)
		list.add( (String)items.elementAt( i));

	this.setVisible( true);
	list.select( items.indexOf( selection));
	
	//Toolkit.eventQueue.postFocusEvent( FocusEvt.getEvent( prompter.list, FocusEvt.FOCUS_GAINED, true));
	prompter.list.requestFocus();
}

void process ( WindowEvent e ) {
	super.process( e);
	if ( e.id == WindowEvent.WINDOW_DEACTIVATED ) {
		//Choice button will close the open prompter
		if ( AWT.mouseTgt != Choice.this )
			closePrompt( false);
	}
}
}
class Entry
  extends TextField
{
Entry( boolean editable) {
	if ( editable ) {
		this.addActionListener( Choice.this);
		this.addKeyListener( Choice.this);
	}
	else {
		isEditable = false;
		
		//will pass mouse events to choice ( open / close prompter on click into the entry )
		this.setEnabled( false);
		
		tCursor = new TextCursor() {
			public void paint( Graphics g) {}
		};
	}
}

ClassProperties getClassProperties () {
        return (props);
}
}
class Selection
  extends Label
{
Selection() {
}

public boolean isFocusTraversable() {
	return true;
}

void process ( FocusEvent e) {
	super.process( e);
	this.paintComponentBorder();
}

public void setText( String text) {
	super.setText( text);
	this.paintComponentBorder();
}
}

public Choice () {
	setLayout( null);
	
	entry = new Selection();
	//entry = new Entry( false);
	add ( entry);
	
	enableEvents( AWTEvent.MOUSE_EVENT_MASK);
}

public void actionPerformed ( ActionEvent e) {
	select( e.getActionCommand() );
	notifyItem();
}

public synchronized void add ( String item) {
	addItem( item);
}

public synchronized void addItem ( String item) {
	insert( item, items.size() );
}

public synchronized void addItemListener ( ItemListener il) {
	iListener = AWTEventMulticaster.add( iListener, il);
}

void checkMouseAware () {
	// we are always mouse aware if not disabled
	if ( (eventMask & AWTEvent.DISABLED_MASK) == 0 ){
		flags |= IS_MOUSE_AWARE;
	}
	else {
		flags &= ~IS_MOUSE_AWARE;
	}
}

void closePrompt( boolean resetFocus) {
	if ( open == true) {
		open = false;
		prompter.list.removeItemListener( this);
		prompter.list.removeActionListener( this);
		prompter.list.removeKeyListener( this);
		
		prompter.dispose();
		prompter = null;

		//will repaint button
		paintComponentBorder();
		
		if ( resetFocus)
			entry.requestFocus();
	}
}

public void doLayout () {
	entry.setBounds( insets.left, insets.top, width-insets.left-insets.right, height-insets.top-insets.bottom);
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

int getDecoFlags() {
	int flags = 0;
	if ( (state & HILIGHTED) > 0)
		flags |= ComponentDeco.CS_CURSOR_OVER;
	if ( open == true )
		flags |= ComponentDeco.CS_ACTIVE;
		
	return flags;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "choice" + (named++);
}

public String getItem ( int index) {
	return (String)items.elementAt( index);
}

public int getItemCount () {
	return (getItems());
}

/**
 * @deprecated
 */
public int getItems() {
	return items.size();
}

public int getSelectedIndex () {
	return items.indexOf( selection);
}

public synchronized String getSelectedItem () {
	return (String)selection;
}

public synchronized Object[] getSelectedObjects () {
	Object[] oa;
	if ( selection != null) {
		oa = new Object[1];
		oa[0] = selection;
	}
	else
		oa = new Object[0];
		
	return oa;
}

public synchronized void insert ( String item, int index) {
	if ( items.size() == 0) {
		selection = item;
		entry.setText( getSelectedItem());
	}
	items.insertElementAt( item, index);
}

public boolean isFocusTraversable() {
	return false;
}

public void itemStateChanged ( ItemEvent e) {
	if ( e.getStateChange() == e.SELECTED ){
		if ( prompter.list.ip.mouseSel() ){
			select( prompter.list.getSelectedItem() );
			notifyItem();
		}
	}
}

public void keyPressed ( KeyEvent e ) {
	Object src = e.getSource();
	int cc = e.getKeyCode();
	
	if ( (open == true) && (src == prompter.list) ){
		if ( cc == e.VK_ESCAPE )
			closePrompt( true);
	}
	else {	
		if ( cc == e.VK_DOWN )
			openPrompt();
	}
}

public void keyReleased( KeyEvent e) {
}

public void keyTyped( KeyEvent e) {
}

private int maxItemWidth() {
        int w, n = items.size();
        int wmax = 0;
        FontMetrics fm = getFontMetrics( getFont());
 
        for ( int i=0; i<n; i++ ){
                if ( (w = fm.stringWidth( (String) items.elementAt(i))) > wmax ){
                        wmax = w;
                }
        }
 
        return wmax;
}

void notifyItem() {
	if ( (iListener != null) ||
	     (eventMask & AWTEvent.ITEM_EVENT_MASK) != 0 ||
	     (flags & IS_OLD_EVENT) != 0 ){
		Toolkit.eventQueue.postEvent( ItemEvt.getEvent( this, ItemEvent.ITEM_STATE_CHANGED,
		                                                selection, ItemEvent.SELECTED));
	}
}

void openPrompt() {
	if ( open == false) {
		open = true;
		Component top = getToplevel();
		prompter = new ChoiceWindow( (top instanceof Frame) ? (Frame)top : null );
		
		Point p = getLocationOnScreen();
		//list item height !!!
		int ih = getFontMetrics( getFont()).getHeight() + 1;

		int ni = items.size();
		if ( ni == 0 )     ni = 1;
		else if ( ni > 8 ) ni = 8;

		int ph = ni*ih + 2*(BORDER_WIDTH+1);
		Dimension sz = Toolkit.getDefaultToolkit().getScreenSize();

                int ox = p.x;
                int oy = p.y + height;
                int ow = Math.max( maxItemWidth()+2*(BORDER_WIDTH+1), width);
 
                if ( p.y + height + ph > sz.height ) {
                        oy = p.y - ph;
                }
                if ( p.x + ow > sz.width ) {
                        ox = sz.width - ow;
                }
 
                prompter.popUpAt( ox, oy, ow, ph);
		
		prompter.list.addItemListener( this);
		prompter.list.addActionListener( this);
		prompter.list.addKeyListener( this);
		
		repaint();
	}
}

protected String paramString() {
	return super.paramString();
}

/**
 * @deprecated
 */
public Dimension preferredSize () {
	Dimension d = entry.getPreferredSize();
	int n = items.size();
	FontMetrics fm = getFontMetrics( getFont());

	int wmax = d.width;

	// We check the selected item by hand because (in PL) it might
	// not be in the list.  In other circumstance it doesn't matter.
	String p = getSelectedItem();
	if (p != null) {
		int w = fm.stringWidth(p);
		if (w > wmax) {
			wmax = w;
		}
	}

	// if we already have a (fixed) list of items, deduce width from there
	// (this mimics behavior of braindead Motif Choices)
	if ( items.size() > 0 ) {
		int i, w;
				
		for ( i=0; i<n; i++ ){
			if ( (w = fm.stringWidth( (String) items.elementAt(i))) > wmax ){
				wmax = w;
			}
		}
	}
		
	d.width = wmax + insets.right + insets.left;
	d.height += insets.top + insets.bottom;
	
	return d;
}

void process ( ItemEvent e ) {
	if ( (iListener != null) || ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) ){
		processEvent( e);
	}
	
	if ((flags & IS_OLD_EVENT) != 0) {
		postEvent (Event.getEvent( e));
	}
}

protected void processItemEvent( ItemEvent e) {
	if ( iListener != null ){
		iListener.itemStateChanged( e);
	}
}

void processMouse ( MouseEvent e ) {
	switch ( e.getID() ) {
	case MouseEvent.MOUSE_PRESSED:
		if ( open == false )	openPrompt();
		else 	closePrompt(entry.isEnabled());
		break;
	case MouseEvent.MOUSE_ENTERED:
		state |= HILIGHTED;
		paintComponentBorder();
		break;
	case MouseEvent.MOUSE_EXITED:
		state &= ~HILIGHTED;
		paintComponentBorder();
		break;
	}
	
	super.processMouse( e);
}

public synchronized void remove ( String item) {
	items.removeElement( item);
}

public synchronized void removeAll () {
	items.removeAllElements();
}

public synchronized void removeItemListener ( ItemListener il) {
	iListener = AWTEventMulticaster.remove( iListener, il);
}

public void requestFocus () {
	entry.requestFocus();
}

public void reshape ( int x, int y, int w, int h ) {
	super.reshape( x, y, w, h);
	
	// there is no need for validation of compound IS_NATIVE_LIKES, they are no Containers
	// in JDK, so we automagically have to re-layout them
	doLayout();
//	flags |= IS_VALID;
}

public synchronized void select ( String item) {
	closePrompt( true);
	if ( items.contains( item) ) {
		selection = item;
		if ( ! entry.getText().equals( item) )
			entry.setText( getSelectedItem());
	}
}

public synchronized void select ( int index) {
	if (index >= 0 && index < items.size()) {
		select((String)items.elementAt(index));
	}
}

public void setEnabled ( boolean isEnabled ) {
//	entry.setEnabled( isEnabled);
	super.setEnabled( isEnabled);
}
}
