package java.awt;

import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.Vector;
import kaffe.awt.Defaults;
import kaffe.awt.theme.ComponentDeco;

/**
 * class RowCanvas -
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
class RowCanvas
  extends Component
  implements AdjustmentListener, FocusListener
{
	Scrollbar hScroll;
	Scrollbar vScroll;
	int rowHeight;
	Vector rows = new Vector();
	int first;
	int xOffs;
	int initVis = -1;
	NativeGraphics rgr;

public void addNotify() {
	super.addNotify();

	if ( rgr == null ) {
		rgr = NativeGraphics.getClippedGraphics( null, this, 0, 0, 0, 0, width, height);
		if ( rgr != null )
			linkGraphics( rgr);
	}
}

public void adjustmentValueChanged( AdjustmentEvent e) {
	Object src = e.getSource();

	if ( src == vScroll){
		int val = e.getValue();
		int lf = first;
		if ( val != first){
			first = val;
			shiftVertical( lf - val, false );
			vPosChange( lf - val);
		}
	}
	else if ( src == hScroll) {
		int val = -e.getValue();
		if ( val != xOffs + insets.left) {
			xOffs = val + insets.left;
			hPosChange();
		}
	}

}

public void focusGained( FocusEvent e) {
	if ( Defaults.FocusScrolls ) {
		if ( hScroll != null )
			hScroll.setBackground( Defaults.FocusClr);
		if ( vScroll != null )
			vScroll.setBackground( Defaults.FocusClr);	
	}

	redirectFocusEvent( e);
}

public void focusLost( FocusEvent e) {
	if ( Defaults.FocusScrolls ) {
		if ( hScroll != null )
			hScroll.setBackground( Defaults.BtnClr);
		if ( vScroll != null )
			vScroll.setBackground( Defaults.BtnClr);	
	}

	redirectFocusEvent( e);
}

int getRowIdx( int yPos) {
	int db = 0;
	int i, y = db;

	if ( yPos < db ) {
		for ( i = first-1; i>=0; i--) {
			y -= rowHeight;
			if ( y < yPos )
				return i;
		}
		return 0;
	}

	int rs = rows.size();
	for ( i = first; i<rs-1; i++ ){
		y += rowHeight;
		if ( y > yPos )
			break;
	}

	return i;
}

int getRowYPos( int row ) {
	return insets.top + ( row - first ) * rowHeight;
}

int getVisibleRows() {
	return (height - insets.top - insets.bottom) / rowHeight;
}

void hPosChange() {
	paint( rgr);
}

void innerLayout() {
	Insets in = parent.insets;
	int w = parent.width - in.left - in.right;;
	int h = parent.height - in.top - in.bottom;

	int sbh = (hScroll != null) ? hScroll.getPreferredSize().height : 0;
	int sbw = (vScroll != null) ? vScroll.getPreferredSize().width : 0;

	boolean vVis = (vScroll != null ) && vScroll.isSliderShowing();
	boolean hVis = (hScroll != null ) && hScroll.isSliderShowing();

	setBounds( in.left, in.top, vVis ? w-sbw : w, hVis ? h-sbh : h);

	if ( vScroll != null ) {
		if ( vScroll.isVisible() && ! vVis )
			vScroll.setVisible( false);
		else if ( !vScroll.isVisible() && vVis )
			vScroll.setVisible( true);
		vScroll.setBounds( x + width, y, sbw, height);
	}
	if ( hScroll != null) {
		if ( hScroll.isVisible() && ! hVis )
			hScroll.setVisible( false);
		else if ( !hScroll.isVisible() && hVis )
			hScroll.setVisible( true);
		hScroll.setBounds( x, y + height, width, sbh);
	}

	updateScrolls();

	if ( initVis > -1 ) {
		makeVisible( initVis);
		initVis = -1;
	}
}

void makeVisible ( int row) {
	int mv = getVisibleRows();
	int rs = rows.size();
	int my = rs - mv;

	// we need bounds to calculate visible range
	// defer until next layout
	if ( height == 0 ) {
		initVis = row;
		return;
	}

	// row already visible	
	if ( (row >= first) && (row < first + mv) )
		return;

	if ( row < 0 )
		row = 0;
	else if ( row >= rs )
		row = rs - 1;

	int lf = first;
	if ( row < first ){
		first = row;
		shiftVertical( lf - first, true);
	}
	else if ( row > first + mv - 1){
		first = row - mv + 1;
		shiftVertical( lf - first, true);
	}
}

int maxRowWidth() {
	return 0;
}

public void mouseEntered( MouseEvent e) {

	if ( (parent.mouseListener != null) && (e.getSource() == this) ){
		// give our parent a chance to intercept mouse events
		// check for mouseListeners first, it's a rare case
		redirectMouseEvent( e);
		return;   // don't do action twice, we get called recursivly by our parent
	}
}

public void mouseExited( MouseEvent e) {

	if ( (parent.mouseListener != null) && (e.getSource() == this) ){
		// give our parent a chance to intercept mouse events
		// check for mouseListeners first, it's a rare case
		redirectMouseEvent( e);
		return;   // don't do action twice, we get called recursivly by our parent
	}
}

public void paint ( Graphics g ) {
	//secure because updated by linked graphics
	NativeGraphics rgs = rgr;
	rgr = (NativeGraphics) g;
	repaintRows( first, getVisibleRows() );
	rgr = rgs;
}

public void print ( Graphics g ) {
	//secure because updated by linked graphics
	NativeGraphics rgs = rgr;
	rgr = (NativeGraphics) g;
	repaintRows( first, getVisibleRows() );
	rgr = rgs;
}

/*
void paintBackground (Graphics g, Component graphOwner, int x, int y, int width, int height) {

	// temporary solution to use parent image background
	// *** to be reworked *** 
	int cx = graphOwner.insets.left;
	int cy = graphOwner.insets.top;
	int cw = graphOwner.width - graphOwner.insets.left - graphOwner.insets.right;
	int ch = graphOwner.height - graphOwner.insets.top - graphOwner.insets.bottom;
	
	//if  background color not explicitly set try theme background first 
	if ( (this.flags & IS_BG_COLORED) == 0 ) {
		if ( deco.paintBackground( this, g, cx, cy, cw, ch, x, y, width, height, getDecoFlags())  ||
		     parent.deco.paintBackground( this, g, cx, cy, cw, ch, x, y, width, height, getDecoFlags()) )
			return;
	}
	
	g.clearRect( x, y, width, height, bgClr);
}
*/
void process ( FocusEvent e) {
	paintComponentBorder();
	super.process( e);
}

void propagateReshape () {
	if ( rgr != null ){
		updateLinkedGraphics();

		// we have our own clipping, which we have to take care of
		rgr.setClip( insets.left, insets.top,
		             width - insets.left - insets.right,
		             height - insets.top - insets.bottom);
	}
}

void rearrange() {
	//force parent layout
	parent.invalidate();
	parent.validate();
}

void redirectFocusEvent( FocusEvent e) {
	if ( parent.focusListener != null ){
		AWTEvent.setSource( e, parent);
		parent.process( e);
	}
}

void redirectKeyEvent( KeyEvent e) {
	if ( parent.keyListener != null ){
		AWTEvent.setSource( e, parent);
		parent.process( e);
	}
}

void redirectMotionEvent( MouseEvent e) {
	if ( parent.motionListener != null ){
		e.retarget( parent, x, y);
		parent.process( e);

		// be careful, the event might be used subsequently
		e.retarget( this, -x, -y);
	}
}

void redirectMouseEvent( MouseEvent e) {
	if ( parent.mouseListener != null ){
		e.retarget( parent, x, y);
		parent.process( e);

		// be careful, the event might be used subsequently
		e.retarget( this, -x, -y);
	}
}

public void removeNotify() {
	super.removeNotify();

	if ( rgr != null ) {	
		rgr.dispose();
		rgr = null;
	}
}

void repaintRow( int idx) {
}

int repaintRows( int sIdx, int len) {
	Graphics g = rgr;
	
	if ( g == null )
		return 0;
		
	int rs = rows.size();
	int y = 0;

	for ( int i=first; i<rs; i++) {
		if ( (i >= sIdx) && (i <= sIdx + len) )
			repaintRow( i);
		y += rowHeight;
		if ( y > height)
			break;
	}

	if (y < height) {
		paintBackground(  g, this, 0, y, width, height-y);
	}

	return y;
}

void setListeners() {
	// own key/mouse/focusListeners are already set in ctor

	if ( hScroll != null )
		hScroll.addAdjustmentListener( this);
	if ( vScroll != null )
		vScroll.addAdjustmentListener( this);
}

void shiftVertical ( int rows, boolean updScroll) {
	if ( (rows == 0) || (rgr == null) )
		return;

	// PM - FIX THIS. In case we are not called from adjustmentNotify (i.e. the
	// EventDispatchThread), where it is obsolete anyway, this might cause reverse scrolling
	// (see Scrollbar.setValue). We currently solve this by fiddling with the inner
	// workings of Scrollbar.
	if ( updScroll && vScroll != null) {
		vScroll.setValues( first, vScroll.vis, vScroll.min, vScroll.max, false);
	}

	//tile image background
	if ( deco.hasBackgroundImage() || parent.deco.hasBackgroundImage() ) {
		repaintRows( first, this.rows.size()-first);
		return;
	}
	
	int d = 0;
	int d2 = d*2;
	int dy = rows * rowHeight;
	int vr = getVisibleRows();

	if ( rows < 0 ){
		if ( -rows < vr )  // otherwise the new top wasn't visible
			rgr.copyArea( insets.left, insets.top - dy + d, 
		              width - insets.left - insets.right,
		              height - insets.top - insets.bottom + dy - d - 1,
		              0, dy);

		repaintRows( first + vr + rows, 1 - rows );
	}
	else {
		if ( rows < vr )   // otherwise the old top will not be visible anymore
			rgr.copyArea( insets.left, insets.top + d,
		              width - insets.left - insets.right,
		              height - insets.top - insets.bottom - dy - d - 1,
		              0, dy);

		repaintRows( first, rows-1);
	}

}

public void update ( Graphics g ){
	// we don't need background blanking
	paint( g);
}

boolean updateHScroll() {
	if ( hScroll == null )
		return false;

	boolean	v1 = hScroll.isSliderShowing();
	int mw = maxRowWidth() + 2*(BORDER_WIDTH+1);

	// check again, because maxRowWidth deletes hScroll
	// on 0 itemWidth from external render
	if ( hScroll == null )
		return false;

	int cv = hScroll.getValue();

	// entire contents going to be visible ( reset to left )
	if ( v1 && (mw <= width) )
		cv = 0;

	// notify only on left reset
	hScroll.setValues( cv, width, 0, mw, (v1 && (cv==0)) );

	return ( v1 != hScroll.isSliderShowing() );
}

void updateScrolls() {
	boolean dv1 = updateVScroll();
	boolean dv2 = updateHScroll();

	if ( dv1 || dv2 ) {
		rearrange();
	}
}

boolean updateVScroll() {
	if ( vScroll == null )
		return false;

	boolean	v1 = vScroll.isSliderShowing();
	int rs = rows.size();
	int vr = getVisibleRows();

	//entire contents going to be visible ( reset to top )
	if ( v1 && (vr >= rs) )
		first = 0;

	//no need to notify ( first already adjusted )
	vScroll.setValues( first, vr, 0, rs, false );

	return ( v1 != vScroll.isSliderShowing() );
}

void vPosChange( int steps) {
}
}
