package java.awt;

import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;
import kaffe.awt.theme.ComponentDeco;
import kaffe.awt.theme.ScrollbarDeco;
import kaffe.util.Timer;
import kaffe.util.TimerClient;

/**
 * class Scrollbar - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Scrollbar
  extends Component
  implements Adjustable, MouseListener, MouseMotionListener, TimerClient
{
	final private static long serialVersionUID = 8451667562882310543L;
	final public static int HORIZONTAL = 0;
	final public static int VERTICAL = 1;
	int min;
	int max;
	int vis;
	int val;
	int ori = -1;
	AdjustmentListener aListener;
	int blockInc;
	int unitInc = 1;
	Graphics dragGr;
	static int currentOp;
	static int dragOffs;
	static int named;
	ScrollbarDeco sbDeco;
	int jLen;
	int jOffs;
	Rectangle railArea = new Rectangle();

public Scrollbar() {
	this( VERTICAL);
}

public Scrollbar( int ori) throws IllegalArgumentException {
	this( ori, 0, 30, 0, 100);
}

public Scrollbar( int ori, int value, int visible, int min, int max) throws IllegalArgumentException{
	setOrientation( ori);

	//defer until orientation set
	getDeco();
	
	setValues( value, visible, min, max);

	addMouseListener( this);
	addMouseMotionListener( this);


}

public synchronized void addAdjustmentListener( AdjustmentListener al) {
	aListener = AWTEventMulticaster.add( aListener, al);
}

public int getBlockIncrement() {
	return (getPageIncrement());
}

ClassProperties getClassProperties () {
           return ClassAnalyzer.analyzeAll( getClass(), true);
}

int getCurrentOp ( int x, int y) {
	if ( ori == VERTICAL )
		return sbDeco.getAdjustmentCode( this, y, height, jOffs, jLen );
	else	
		return sbDeco.getAdjustmentCode( this, x, width, jOffs, jLen );
}

protected ComponentDeco getDeco() {
	ComponentDeco cd = null;
	//skip call from Component ct ( no orientation yet )
	if ( ori > -1 ) {
		cd = AWT.getTheme().getDeco( this, (ori ==VERTICAL) ? "vert" : "horz");
//		cd = AWT.getTheme().getDeco( this, "java.awt.Scrollbar@" + ((ori ==VERTICAL) ? "vert" : "horz"), null);
		sbDeco = (ScrollbarDeco)cd;
	}
	return cd;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name

	return "scrollbar" + (named++);
}

/**
 * @deprecated
 */
public int getLineIncrement() {
	return unitInc;
}

public int getMaximum() {
	return max;
}

public int getMinimum() {
	return min;
}

public int getOrientation() {
	return ori;
}

/**
 * @deprecated
 */
public int getPageIncrement() {
	return blockInc;
}

public int getUnitIncrement() {
	return (getLineIncrement());
}

public int getValue() {
	return val;
}

/**
 * @deprecated
 */
public int getVisible() {
	return vis;
}

public int getVisibleAmount() {
	return (getVisible());
}

public boolean isFocusTraversable() {
	return false;
}

boolean isSliderShowing() {
	return jLen > 0;
}

public void mouseClicked( MouseEvent e) {
	currentOp = getCurrentOp( e.getX(), e.getY() );
	switch ( currentOp) {
	case AdjustmentEvent.BLOCK_DECREMENT:
		setValue( val - blockInc);
		break;
	case AdjustmentEvent.BLOCK_INCREMENT:
		setValue( val + blockInc);
		break;
	}
}

public void mouseDragged( MouseEvent e) {
	if ( (currentOp != AdjustmentEvent.TRACK) || (dragGr == null) )
		return;

	if ( ori == VERTICAL ) {
		int y0 = e.getY() - dragOffs;
		if (y0 <= railArea.y) {
			if ( jOffs == railArea.y ) return;
			y0 = railArea.y;
		}
		else if (y0 >= railArea.y+railArea.height-jLen) {
			if ( jOffs == railArea.y+railArea.height-jLen ) return;
			y0 = railArea.y+railArea.height-jLen;
		}
		updateVertRegion( y0);
		updateValue();
	}
	else {
		int x0 = e.getX() - dragOffs;
		if (x0 <= railArea.x) {
			if ( jOffs == railArea.x ) return;
			x0 = railArea.x;
		}
		else if (x0 >= railArea.x+railArea.width-jLen) {
			if ( jOffs == railArea.x+railArea.width-jLen ) return;
			x0 = railArea.x+railArea.width-jLen;
		}
		updateHorzRegion( x0);
		updateValue();
	}
}

public void mouseEntered( MouseEvent e) {
}

public void mouseExited( MouseEvent e) {
}

public void mouseMoved( MouseEvent e) {
}

public void mousePressed( MouseEvent e) {
	currentOp = getCurrentOp( e.getX(), e.getY() );

	if ( (dragGr = NativeGraphics.getClippedGraphics( null, this, 0, 0, railArea.x, railArea.y, railArea.width, railArea.height)) != null ) {
		dragGr.setColor( fgClr);
	}

	dragOffs = ( ori == HORIZONTAL ) ? e.getX() - jOffs : e.getY() - jOffs;

	switch ( currentOp) {
	case AdjustmentEvent.UNIT_DECREMENT:
		if ( val > min ) {
			Timer.getDefaultTimer().addClient( this, 300, 50);
		}
		break;
	case AdjustmentEvent.UNIT_INCREMENT:
		if ( val < max ) {
			Timer.getDefaultTimer().addClient( this, 300, 50);
		}
		break;
	default:
	}

}

public void mouseReleased( MouseEvent e) {
	Timer.getDefaultTimer().removeClient( this);
	dragOffs = 0;

	if ( dragGr != null ){
		dragGr.dispose();
		dragGr = null;
	}

	switch( currentOp) {
	case AdjustmentEvent.UNIT_INCREMENT:
	case AdjustmentEvent.UNIT_DECREMENT:
		currentOp = 0;
	}
}

void notifyAdjust() {
	if ( (aListener != null) ||
	     (eventMask & AWTEvent.ADJUSTMENT_EVENT_MASK) != 0 ||
	     ((flags & IS_OLD_EVENT) != 0) ) {
	  // try to accumulate AdjustmentEvents, to avoid lag-behind effects if the
		// scroll-inflicted operation (paint) takes too long (it might be in the
		// application domain)

		Toolkit.eventQueue.postAdjustmentEvent( AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
		                                        this, currentOp, val);
/*
    // that would be non-accumulated, i.e. faster for single event reaction
		Toolkit.eventQueue.postEvent( AdjustmentEvt.getEvent( this,
		                                                      AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
		                                                      currentOp, val));
*/
	}
}

public void paint ( Graphics g ) {
	g.setColor( fgClr);
	if ( ori == HORIZONTAL ) {
		sbDeco.paintRail( this, g, 0, 0, width, height);
		if ( jLen > 0 )
			sbDeco.paintJewel( this, g, jOffs, 0, jLen, height, false);
	}
	else {
		sbDeco.paintRail( this, g, 0, 0, width, height);
		if ( jLen > 0 )
			sbDeco.paintJewel(  this, g, 0, jOffs, width, jLen,  false);
	}
	sbDeco.paintArrows( this, g, width, height, currentOp );
}

protected String paramString() {
	return super.paramString() + ",Value: " + val + ",Visible: " + vis + ",Min: " + min + ",Max: " + max; 
}

/**
 * @deprecated
 */
public Dimension preferredSize() {
	return sbDeco.getPreferredSize( this);
}

void process ( AdjustmentEvent e ) {
	if ( (aListener != null) || ((eventMask & AWTEvent.ADJUSTMENT_EVENT_MASK) != 0) ){
		processEvent( e);
	}

	if ( (flags & IS_OLD_EVENT) != 0 ) {
		postEvent( Event.getEvent( e));
	}
}

protected void processAdjustmentEvent ( AdjustmentEvent e) {
	if (  aListener != null ) {
		aListener.adjustmentValueChanged( e);
	}
}

public synchronized void removeAdjustmentListener( AdjustmentListener al) {
	aListener = AWTEventMulticaster.remove( aListener, al);
}

public void reshape( int x, int y, int width, int height) {
	super.reshape( x, y, width, height);
	
	railArea.setBounds( 0, 0, width, height);
	sbDeco.getRailArea( this, railArea);
	
	updateJewelBounds();
}

public synchronized void setBlockIncrement( int inc) {
	setPageIncrement(inc);
}

void setDeco(ComponentDeco deco) {
	sbDeco = (ScrollbarDeco)deco;
	super.setDeco(deco);
}

/**
 * @deprecated
 */
public void setLineIncrement(int inc) {
	unitInc = inc;
}

public synchronized void setMaximum( int max) {
	setValues( val, vis, min, max);
}

public synchronized void setMinimum( int min) {
	setValues( val, vis, min, max);
}

public synchronized void setOrientation( int ori) throws IllegalArgumentException {
	if ( (ori != HORIZONTAL) && (ori != VERTICAL) )
		throw new IllegalArgumentException();
	this.ori = ori;
}

public void setPageIncrement(int inc) {
	setValues( val, inc, min, max);
}

public void setUnitIncrement( int inc) {
	setLineIncrement(inc);
}

public synchronized void setValue( int value) {
	if ( val != value) {
		setValues( value, vis, min, max);
	}
}

public void setValues( int value, int visible, int min, int max) {
	setValues( value, visible, min, max, true);
}

synchronized void setValues( int value, int visible, int min, int max, boolean notify ) {
	// PM: not ideal - this is a workaround for clients with coexisting sync (own) and
	// async (Scrollbar initiated) updates. In case they rely on sync updates, they are
	// just interested in Scrollbar updates, but not notifications. However, this is a design
	// flaw of these clients (this is not a public API of Scrollbar !!)

	int lastVal = this.val;

	if ( (value == val) && (visible == vis) && (min == this.min) && (max == this.max) )
		return;

	// Make size adjustments.
	if (max < min)			max = min;
	if (visible > max-min)		visible = max-min;
	if ( value > max - visible )	value = max - visible;
	else if ( value < min )		value = min;

	this.min = min;
	this.max = max;
	this.val = value;
	this.vis = visible;
	this.blockInc = visible;

	update();

	if ( notify && (val != lastVal) )
		notifyAdjust();
}

public synchronized void setVisibleAmount( int vis) {
	setValues( val, vis, min, max);
}

public void timerExpired( Timer t) {
	switch ( currentOp) {
	case AdjustmentEvent.UNIT_DECREMENT:
		if ( val == min){
			currentOp = 0;
			t.removeClient( this);
		}
		else {
			setValue( val-unitInc);
		}
		break;
	case AdjustmentEvent.UNIT_INCREMENT:
		if ( val == max) {
			currentOp = 0;
			t.removeClient( this);
		}
		else {
			setValue( val+unitInc);
		}
		break;
	}	
}

void update() {
	Graphics g = null;

	if ( dragGr != null )
		g = dragGr;
	else if ( (flags & IS_VISIBLE) != 0 )
		g = getGraphics();

	if ( g != null ){
		updateJewelBounds();

		update( g);	
		if ( g != dragGr ) {
			g.dispose();
		}
	}
	else {
		updateJewelBounds();
	}
}

void updateHorzRegion ( int offs) {
	int dOffs = offs - jOffs;

	dragGr.copyArea( jOffs, 0, jLen, height, dOffs, 0);
	
	if ( dOffs > 0 ) {
		sbDeco.paintRail( this, dragGr, jOffs, 0, dOffs, height);
	}
	else if ( dOffs < 0 ) {
		sbDeco.paintRail( this, dragGr, offs+jLen, 0, -dOffs, height);
	}

	jOffs = offs;
}

void updateJewelBounds () {
	int range = max - min;

	//entire contents visible
	if ( range <= 0 || vis <= 0 || max == vis) {
		jLen = 0;
		return;
	}

	if ( ori == HORIZONTAL ) {
		jLen = sbDeco.getJewelExtent( this, width, vis, range);
//		jOffs = val * (width-jLen) / range;
		jOffs = railArea.x + (val-min) * (railArea.width-jLen) / (max - vis);
	} 
	else {
		jLen = sbDeco.getJewelExtent( this, height, vis, range);
//		jOffs = val * (height-jLen) / range;
		jOffs = railArea.y + (val-min) * (railArea.height-jLen) / (max - vis);
	}
}

void updateValue () {
	//value from slider position
	int newVal;
	
	if ( ori == HORIZONTAL ) {
		newVal = min + ( (max - vis) * (jOffs-railArea.x) ) / (railArea.width - jLen);
		//newVal = jOffs * range / ( railArea.width - jLen);
	}
	else {
		newVal = min + ( (max - vis) * (jOffs-railArea.y) ) / (railArea.height - jLen);
		//newVal = jOffs * range / ( railArea.height - jLen);
	}
	
	if ( val != newVal ) {
		val = newVal;
		notifyAdjust();
	}
}

void updateVertRegion ( int offs) {
	int dOffs = offs - jOffs;

	dragGr.copyArea( 0, jOffs, width, jLen, 0, dOffs);

	if ( dOffs > 0 ) {
		sbDeco.paintRail( this, dragGr, 0, jOffs, width, dOffs);
	}
	else if ( dOffs < 0 ) {
		sbDeco.paintRail( this, dragGr, 0, offs+jLen, width, -dOffs);
	}

	jOffs = offs;
}
}
