package java.awt;

import java.awt.TextEvt;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import kaffe.awt.Defaults;

/**
 * class TextField - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class TextField
  extends TextComponent
{
	final private static long serialVersionUID = -2966288784432217853L;
	static int xOffs = 2;
	ActionListener aListener;
	int cols;
	char echoChar;
	FontMetrics fm;
	TextCursor tCursor = new TextCursor();
	int selStart;
	int selEnd;
	TextBuffer textBuf = new TextBuffer();
	TextBuffer hiddenBuf;
	int first;
	NativeGraphics rgr;
	static int named;

public TextField() {
	this( "", 0);
}

public TextField( String text) {
	this( text, (text != null) ? text.length() : 0);
}

public TextField( String text, int cols) {
	this.cols = cols;

	cursor = Cursor.getPredefinedCursor( Cursor.TEXT_CURSOR);

	setText( text);
	setEditMenu( this);

	eventMask |= (AWTEvent.KEY_EVENT_MASK | AWTEvent.FOCUS_EVENT_MASK |
	              AWTEvent.MOUSE_EVENT_MASK  | AWTEvent.MOUSE_MOTION_EVENT_MASK);

	flags |= IS_MOUSE_AWARE;
	
}

public TextField( int cols) {
	this( "", cols);
}

public synchronized void addActionListener ( ActionListener al) {
	aListener = AWTEventMulticaster.add( aListener, al);
}

public void addNotify () {
	super.addNotify();

	if ( rgr == null ) {
//		rgr = NativeGraphics.getClippedGraphics( rgr, this, 0, 0, 0, 0, width, height);
		rgr = NativeGraphics.getClippedGraphics( rgr, this, 0, 0, insets.left, insets.top, 
												width-insets.left-insets.right,
												height-insets.top-insets.bottom);
		if ( rgr != null )
			linkGraphics( rgr);
	}
}

private void blankSelection() {
	if ( selStart != selEnd)
		setTextCursor( selStart, true, false);
}

public void deleteSelection() {
	if ( selStart != selEnd) {
		textBuf.remove( selStart, selEnd-selStart);
		if ( hiddenBuf != null )
			hiddenBuf.remove( selStart, selEnd-selStart);
		setTextCursor( selStart, true, false);
	}
}

public boolean echoCharIsSet() {
	return echoChar != 0;
}

private void extendSelectionTo( int idx, boolean fromLeft) {
	if ( idx < 0)
		idx = 0;
	else if ( idx > textBuf.len)
		idx = textBuf.len;
	if ( idx == tCursor.index)
		return;

	if ( idx < selStart)
		selStart = idx;
	else if ( idx > selEnd)
		selEnd = idx;
	else if ( fromLeft) {
		//force repaint from start
		tCursor.index = selStart;
		selStart = idx;
	}
	else
		selEnd = idx;

	setTextCursor( idx, false, true);
}

private TextBuffer getBuffer() {
	return ( (hiddenBuf != null ) ? hiddenBuf : textBuf );
}

public int getCaretPosition() {
	return tCursor.index;
}

public int getColumns() {
	return cols;
}

private int getCursorPos( TextBuffer tb, int idx) {
	if ( idx < 0 )
		idx = 0;
	else if ( idx > textBuf.len )
		idx = textBuf.len;
	return ( insets.left + xOffs + tb.getWidth( first, idx) );
}

public char getEchoChar() {
	return echoChar;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name

	return "textfield" + (named++);
}

public Dimension getMinimumSize() {
	return (minimumSize());
}

public Dimension getMinimumSize(int cols) {
	return (minimumSize(cols));
}

public Dimension getPreferredSize(int cols) {
	return (preferredSize(cols));
}

public String getSelectedText() {
	if ( selStart != selEnd) {
		return textBuf.getString( selStart, selEnd-selStart);
	}
	return (null);
}

public int getSelectionEnd() {
	return selEnd;
}

public int getSelectionStart() {
	return selStart;
}

public String getText() {
	return textBuf.toString();
}

private void makeVisible( int idx) {
	TextBuffer tb = getBuffer();
	int lf = first;

	if ( idx < 0 )
		idx = 0;
	else if ( idx > tb.len )
		idx = tb.len;

	if ( idx < first ) {
		first = idx;
	}
	else {
		int xp = tb.getPos( idx);
		if ( width > 0 ) {
			int mp = width - insets.left - insets.right - 2*xOffs - tCursor.width;;
			for (; mp < xp - tb.getPos( first); first++ );
		}
	}

	if ( first != lf ) {
		paint( rgr);
	}
}

/**
 * @deprecated
 */
public Dimension minimumSize() {
	return (minimumSize(cols));
}

/**
 * @deprecated
 */
public Dimension minimumSize(int cols) {
	int          nc = cols;
	int	     h = 3*BORDER_WIDTH;
	int	     w = 0;

	if ( fm == null ) {
		//no font set yet, could only happen if explicitely called before added to the parent
		//will be fixed by the parents layouter later on
		return new Dimension( 50, 20);
	}
	
	h += fm.getHeight();
	
	if ( nc <= 0 ) {
		if ( textBuf.len > 0 )
			w = textBuf.getWidth( 0, textBuf.len);
		else
			nc = 5;
	}
	if ( w == 0 ) {
		w = nc * fm.charWidth( 'X');
	}

	return new Dimension( w + insets.left + insets.right + 2*xOffs,
	                      h + insets.top + insets.bottom);
}

private void notifyAction(){
	if ( (aListener != null) || ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) ){
		Toolkit.eventQueue.postEvent( ActionEvt.getEvent( this, ActionEvent.ACTION_PERFORMED,
		                                                  getText(), 0));
	}
}

public void paint( Graphics g) {
	//secure because updated by linked graphics
	NativeGraphics rgs = rgr;
	rgr = (NativeGraphics) g;
	repaintFrom( 0);
	rgr = rgs;
}

private void paintInactiveCursor() {
	if ( Defaults.ShowInactiveCursor ) {
		rgr.setColor( Defaults.TextCursorInactiveClr);
		tCursor.blank( rgr);
	}
	else {
		paint( rgr);
	}
}

protected String paramString() {
	return super.paramString();
}

/**
 * @deprecated
 */
public Dimension preferredSize() {
	return (preferredSize(cols));
}

/**
 * @deprecated
 */
public Dimension preferredSize(int cols) {
	return (getMinimumSize(cols));
}

public void print( Graphics g) {
	//secure because updated by linked graphics
	NativeGraphics rgs = rgr;
	rgr = (NativeGraphics) g;
	repaintFrom( 0);
	rgr = rgs;
}

void process ( ActionEvent e ) {
	if ( (aListener != null) || ((eventMask & (AWTEvent.ACTION_EVENT_MASK|AWTEvent.DISABLED_MASK)) == AWTEvent.ACTION_EVENT_MASK) ){
		processEvent( e);
	}

	if ( (flags & IS_OLD_EVENT) > 0 ) {
		postEvent( Event.getEvent( e));
	}
}

protected void processActionEvent ( ActionEvent e) {
	if ( aListener != null )
		aListener.actionPerformed( e);
}

protected void processFocusEvent ( FocusEvent event ) {
	switch ( event.getID() ) {
		case FocusEvent.FOCUS_GAINED:
			if ( (flags & IS_MOUSE_AWARE) > 0 ) 
				tCursor.paint( rgr);
			paintComponentBorder();
			updateAWT();
			break;
		case FocusEvent.FOCUS_LOST:
			paintInactiveCursor();
			paintComponentBorder();
			break;
	}
	
	super.processFocusEvent( event);
}

protected void processKeyEvent ( KeyEvent event ) {

	int mods = event.getModifiers();
	boolean shift = event.isShiftDown();
	int code = event.getKeyCode();
	char c = event.getKeyChar();

	switch ( event.id ) {
		case KeyEvent.KEY_PRESSED:
			// do not consume unused keys for ShortcutHandler
			if ( (mods != 0) && (mods != KeyEvent.SHIFT_MASK) ) {
				return;
			}
			switch( code ){
				case KeyEvent.VK_ENTER:
					notifyAction();
					break;
				case KeyEvent.VK_ESCAPE:
					if (isEditable) {
						setText( "");
					}
					break;
				case KeyEvent.VK_HOME:
					if ( shift) {
						extendSelectionTo( 0, false);
					}
					else {
						setTextCursor( 0, true, false);
					}
					break;
				case KeyEvent.VK_END:
					if ( shift) {
						extendSelectionTo( textBuf.len, true);
					}
					else {
						setTextCursor( textBuf.len, true, false);
					}
					break;
				case KeyEvent.VK_LEFT:
					if ( shift) {
						extendSelectionTo( tCursor.index-1, false);
					}
					else {
						shiftTextCursor( -1, true);
					}
					break;
				case KeyEvent.VK_RIGHT:
					if ( shift) {
						extendSelectionTo( tCursor.index+1, true);
					}
					else {
						shiftTextCursor( 1, true);
					}
					break;
				case KeyEvent.VK_BACK_SPACE:
					if (isEditable) {
						if ( selStart != selEnd) {
							deleteSelection();
						}
						else if ( tCursor.index > 0) {
							textBuf.remove( tCursor.index-1, 1);
							if ( hiddenBuf != null ) {
								hiddenBuf.remove( tCursor.index-1, 1);
							}
							shiftTextCursor( -1, true);
						}
					}
					break;
				case KeyEvent.VK_DELETE:
					if (isEditable) {
						if ( selStart != selEnd) {
							deleteSelection();
						}
						else if ( tCursor.index < textBuf.len) {
							textBuf.remove( tCursor.index, 1);
							if ( hiddenBuf != null ) {
								hiddenBuf.remove( tCursor.index, 1);
							}
							repaintTrailing();
						}
					}
					break;
				default:
					super.processKeyEvent( event);
					return;
				}
				
				event.consume();
				break;
			
		case KeyEvent.KEY_TYPED:
			if ( ! isEditable || event.isConsumed() || ! isPrintableTyped( event) )
				return;

			if ( selStart != selEnd){
				textBuf.replace( selStart, selEnd - selStart, c);
				if ( hiddenBuf != null )
					hiddenBuf.replace( selStart, selEnd - selStart, echoChar);
				setTextCursor( selStart+1, true, false);
			}
			else {
				textBuf.insert( tCursor.index, c );
				if ( hiddenBuf != null )
					hiddenBuf.insert( tCursor.index, echoChar);
				shiftTextCursor( 1, true);
			}

			if ( (textListener != null) || (eventMask & AWTEvent.TEXT_EVENT_MASK) != 0 ) {
				Toolkit.eventQueue.postEvent( TextEvt.getEvent( this, TextEvt.TEXT_VALUE_CHANGED));		
			}
			
			break;
		case KeyEvent.KEY_RELEASED:
			break;
	}
	
	super.processKeyEvent( event);
}

void processMotion ( MouseEvent event ) {
	switch ( event.id ) {
	case MouseEvent.MOUSE_DRAGGED:
		TextBuffer tb = getBuffer();
		int cIdx = tb.getIdxFrom( first, event.getX() - xOffs - insets.left);
		extendSelectionTo( cIdx, cIdx > tCursor.index);
		break;
	}
	
	super.processMotion( event);
}

void processMouse ( MouseEvent event ) {

	switch ( event.getID() ) {
	case MouseEvent.MOUSE_PRESSED:
		if ( event.isPopupTrigger() ){
			if ( (triggerPopup( event.getX(), event.getY())) != null )
				return;
		}

		TextBuffer tb = getBuffer();
		int cIdx = tb.getIdxFrom( first, event.getX() - xOffs - insets.left);
		blankSelection();
		setTextCursor( cIdx, true, false);
		requestFocus();

		if ( event.getModifiers() == InputEvent.BUTTON2_MASK )
			pasteFromClipboard();

		break;
	}	
	
	super.processMouse( event);
}

void propagateFont( Font fnt) {
	super.propagateFont( fnt);
	fm = getFontMetrics( fnt);
	
	textBuf.setMetrics( fm, 20);
	if ( hiddenBuf != null )
		hiddenBuf.setMetrics( fm, 20);

	int h = fm.getHeight();
	TextBuffer tb = (hiddenBuf != null ) ? hiddenBuf : textBuf;
	tCursor.setPos( (height - h)/2, h);
	tCursor.setIndex( tCursor.index, getCursorPos( tb, tCursor.index) );

	if ( rgr != null )
		rgr.setFont( fnt);

	if ( (flags & IS_SHOWING) > 0 )
		repaint();
}

void propagateReshape () {

	if ( rgr != null ){
		updateLinkedGraphics();

		// we have our own clipping, which we have to take care of
		rgr.setClip( insets.left, insets.top,
		             width - insets.left - insets.right,
		             height - insets.top - insets.bottom);
	}

}

public synchronized void removeActionListener( ActionListener al) {
	aListener = AWTEventMulticaster.remove( aListener, al);
}

public void removeNotify () {
	super.removeNotify();

	if ( rgr != null ) {
		rgr.dispose();
		rgr = null;
	}
}

private void repaintFrom( int start){
	if ( (flags & IS_SHOWING) != IS_SHOWING ) return;

	TextBuffer tb = getBuffer();

	if ( start > 0 )
		start = start -1;

	if ( (selStart == selEnd) || (start > selEnd) )
		repaintRange( tb, start, tb.len, false);
	else {
		if ( selStart > start) 
			repaintRange( tb, start, selStart, false);
		if ( selEnd > start)
			repaintRange( tb, Math.max( selStart, start), selEnd, true);
		if ( textBuf.len > selEnd)
			repaintRange( tb, selEnd, textBuf.len, false);
	}

	if ( kaffe.awt.AWT.keyTgt == this) {
		tCursor.paint( rgr);
	}
	else if ( Defaults.ShowInactiveCursor )
		paintInactiveCursor();

}

private void repaintRange( TextBuffer tb, int start, int end, boolean invert) {
	int x0, x1;

	if ( rgr == null )
		return;

	if ( start < first )
		start = first;

	x0 = insets.left + xOffs + tb.getWidth( first, start);
	x1 = x0 + tb.getWidth ( start, end);

	if ( invert) {
		rgr.setColor( Defaults.TextFieldSelBgClr );
		if ( end == tb.len )
			x1 += tCursor.width;
		rgr.fill3DRect( x0, tCursor.y, x1-x0, tCursor.height+1, true);
		rgr.setColor( Defaults.TextFieldSelTxtClr);
	}
	else {
		if ( end == tb.len )
			x1 = width - insets.right;
		if ( start == first)
			x0 -= xOffs;
		paintBackground( rgr, x0, insets.top, x1-x0, height-insets.top-insets.bottom);
		rgr.setColor( fgClr );
	}

	tb.paintFrom( rgr, insets.left + xOffs, 0, height, first, start, end-start);
}

private void repaintTrailing() {
	repaintFrom( tCursor.index);
}

void replaceSelectionWith ( String s ) {
	deleteSelection();
	if ( s != null ) {
		textBuf.insert( selEnd, s);
		selStart = selEnd + s.length();
		selEnd = selStart;
		setCaretPosition( selStart);
		repaint();
	}
}

private void resetSelIdxs( int idx) {
	selStart = idx;
	selEnd = idx;
}

public void reshape( int x, int y, int width, int height) {
	if ( fm != null ) {
		int fh = fm.getHeight();
		tCursor.setPos( (height - fh)/2, fh);
	}
	super.reshape( x, y, width, height);
}

public void select( int start, int end) {
	selStart = start;
	selEnd = end;
	setTextCursor( selStart, false, true);
}

public void selectAll() {
	select( 0, textBuf.len);
}

public void setCaretPosition( int pos) {
	setTextCursor( pos, true, false);
}

public void setColumns( int cols) {
	if ( this.cols != cols)
		this.cols = cols;
}

public void setEchoChar( char echoChar) {
	setEchoCharacter(echoChar);
}

/**
 * @deprecated
 */
public void setEchoCharacter(char echoChar) {
	this.echoChar = echoChar;
	if ( echoChar != 0 ) {
		if ( hiddenBuf == null ) {
			hiddenBuf = new TextBuffer();
			if ( fm != null ) {
				hiddenBuf.setMetrics( fm, 20);
			}
		}
		hiddenBuf.set( echoChar, textBuf.len);
	}
	else
		hiddenBuf = null;

	setTextCursor( tCursor.index, false, true);
}

public void setSelectionEnd( int end) {
	select( selStart, end);
}

public void setSelectionStart( int start) {
	select( start, selEnd);
}

public void setText( String text) {
	textBuf.set( text);
	if ( hiddenBuf != null )
		hiddenBuf.set( echoChar, text.length() );
	setTextCursor( 0, true, false);
}

private void setTextCursor( int xIdx, boolean resetSel, boolean repaintAll) {
	int repIdx = 0;
	TextBuffer tb = getBuffer();

	if ( xIdx > tb.len)
		xIdx = tb.len;
	else if ( xIdx < 0)
		xIdx = 0;

	makeVisible( xIdx);

	if ( resetSel) {
		repIdx = Math.min( selStart, xIdx);
		resetSelIdxs( xIdx);
	}
	else if ( ! repaintAll )
		repIdx = Math.min( xIdx, tCursor.index);

	tCursor.setIndex( xIdx, getCursorPos( tb, xIdx) );

	repaintFrom( repIdx);
}

private void shiftTextCursor( int d, boolean resetSel) {
	if ( tCursor.index + d < 0 )
		setTextCursor( 0, resetSel, false);
	else if ( tCursor.index + d > textBuf.len )
		setTextCursor( textBuf.len, resetSel, false);
	else
		setTextCursor( tCursor.index + d, resetSel, false);
}
}
