package java.awt;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.WindowEvent;
import java.util.Vector;
import kaffe.awt.Defaults;
import kaffe.awt.WidgetTheme;

/**
 * class Choice - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Choice
  extends Container
  implements ItemSelectable, ActionListener, ItemListener, MouseListener, KeyListener, FocusListener
{
	final private static long serialVersionUID = -4075310674757313071L;
	Vector items = new Vector();
	Object selection;
	ItemListener iListener;
	ChoiceWindow prompter;
	// We take control of the drawing ourselves but still use a text
	// field to hold the data (for now at least)
	TextField entry = new TextField() {
		void paintBackground(Graphics g, int x, int y, int width, int height, int flags) {
		}
		void drawContents() {
		}
	};
	int state;
	static int named;

class ChoiceWindow
  extends Window
{
	List list = new List();

public ChoiceWindow ( Frame owner ) {
	super( owner);

	this.setLayout( null);
	list.removeHScroll();
	this.add( list);
}

public void doLayout() {
	int db = 1;
	list.setBounds( db, db, this.width-2*db, this.height-2*db);
}

void popUpAt( int x, int y, int width, int height) {
	int is = items.size();

	this.setBounds( x, y, width, height);
	list.setFont( Choice.this.font);

	String sel = getSelectedItem();
	int idx = -1;
	for ( int i=0; i < is; i++) {
		String item = (String)items.elementAt(i);
		if (item.equals(sel)) {
			idx = i;
		}
		list.add(item);
	}

	this.setVisible( true);
	if (idx >= 0) {
		list.select(idx);
	}
	list.requestFocus();
}

void process ( WindowEvent e ) {
	super.process( e);

	if ( e.id == WindowEvent.WINDOW_DEACTIVATED ) {
	//temporary light
//		closePrompt( false);
	}
}

void paintBackground (Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.chPaintPopup(Choice.this, this, g);
}

}

public Choice () {
	setLayout( null);
	WidgetTheme.chSetDefaults(this);
	add ( entry);
	
	entry.setEditable(false);
	entry.addActionListener( this);
	entry.addKeyListener( this);

	deco = WidgetTheme.chGetDeco( this);
	addMouseListener( this);
}

public void actionPerformed ( ActionEvent e) {
	select( e.getActionCommand() );
	notifyItem();
}

public synchronized void add ( String item) {
	addItem( item);
}

public synchronized void addItem ( String item) {
	insert( item, items.size() );
}

public synchronized void addItemListener ( ItemListener il) {
	iListener = AWTEventMulticaster.add( iListener, il);
}

void checkMouseAware () {
	// we are always mouse aware if not disabled
	if ( (eventMask & AWTEvent.DISABLED_MASK) == 0 ){
		flags |= IS_MOUSE_AWARE;
	}
	else {
		flags &= ~IS_MOUSE_AWARE;
	}
}

void closePrompt( boolean resetFocus) {
	if ( prompter != null) {
		prompter.list.removeItemListener( this);
		prompter.list.removeActionListener( this);
		prompter.list.removeKeyListener( this);
		prompter.list.removeFocusListener( this);

		prompter.dispose();
		prompter = null;
		paintButton();
		
		if ( resetFocus)
			entry.requestFocus();
	}
}

public void doLayout () {
	int d1 = WidgetTheme.borderOuterWidth( this);
	int d2 = WidgetTheme.borderInnerWidth( entry);
	Rectangle r = WidgetTheme.chTextBounds( this);
	entry.setBounds( r.x, r.y, r.width-2*(d1+d2), r.height);
}

public void focusGained ( FocusEvent e) {
}

public void focusLost ( FocusEvent e) {
	if ( (prompter != null) && (e.getSource() == prompter.list) ) {
		closePrompt( false);
	}
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "choice" + (named++);
}

public String getItem ( int index) {
	return (String)items.elementAt( index);
}

public int getItemCount () {
	return (getItems());
}

/**
 * @deprecated
 */
public int getItems() {
	return items.size();
}

public int getSelectedIndex () {
	return items.indexOf( selection);
}

public synchronized String getSelectedItem () {
	return (String)selection;
}

public synchronized Object[] getSelectedObjects () {
	Object[] oa;
	if ( selection != null) {
		oa = new Object[1];
		oa[0] = selection;
	}
	else
		oa = new Object[0];
		
	return oa;
}

public synchronized void insert ( String item, int index) {
	if ( items.size() == 0) {
		selection = item;
		entry.setText( item);
	}
	items.insertElementAt( item, index);
}

public void itemStateChanged ( ItemEvent e) {
	if ( e.getStateChange() == e.SELECTED ){
		if ( prompter.list.selMouse ){
			select( prompter.list.getSelectedItem() );
			notifyItem();
		}
	}
}

public void keyPressed ( KeyEvent e ) {
	Object src = e.getSource();
	int cc = e.getKeyCode();
	
	if ( (prompter != null) && (src == prompter.list) ){
		if ( cc == e.VK_ESCAPE )
			closePrompt( true);
	}
	else {	
		if ( cc == e.VK_DOWN )
			openPrompt();
	}
}

public void keyReleased( KeyEvent e) {
}

public void keyTyped( KeyEvent e) {
}

public void mouseClicked ( MouseEvent e) {
}

public void mouseEntered ( MouseEvent e) {
	state |= WidgetTheme.IS_POINTED;
	paintButton();
}

public void mouseExited ( MouseEvent e) {
	state &= ~WidgetTheme.IS_POINTED;
	paintButton();
}

public void mousePressed ( MouseEvent e) {
	if ( prompter == null)
		openPrompt();
	else
		closePrompt( true);
}

public void mouseReleased ( MouseEvent e) {
}

void notifyItem() {
	if ( (iListener != null) ||
	     (eventMask & AWTEvent.ITEM_EVENT_MASK) != 0 ||
	     (flags & IS_OLD_EVENT) != 0 ){
		Toolkit.eventQueue.postEvent( ItemEvt.getEvent( this, ItemEvent.ITEM_STATE_CHANGED,
		                                                selection, ItemEvent.SELECTED));
	}
}

void openPrompt() {
	if ( prompter == null) {
		Component top = getToplevel();
		prompter = new ChoiceWindow( (top instanceof Frame) ? (Frame)top : null );
		
		Point p = getLocationOnScreen();

		int ih = getFontMetrics(getFont()).getHeight() + 1;
		int ni = items.size();
		if (ni == 0) {
			ni = 1;
		}
		else if (ni > 8) {
			ni = 8;
		}
		int ph = ni * ih + 2 * (BORDER_WIDTH+1);

		// If we're poping up in a window then we're going to assume
		// that we must stay within it's borders
		int sh;
		if (top instanceof Window && !(top instanceof Frame)) {
			sh = top.getLocationOnScreen().y + top.getSize().height;
		}
		// Otherwise we just stay on the screen
		else {
			sh = Toolkit.getDefaultToolkit().getScreenSize().height;
		}

		// If the popup fits below up, put it there
		if (p.y + height + 1 + ph < sh) {
			prompter.popUpAt( p.x, p.y+height+1, width, ph);
		}
		// Otherwise we put it above us
		else {
			prompter.popUpAt( p.x, p.y-ph-1, width, ph);
		}
		prompter.list.addItemListener( this);
		prompter.list.addActionListener( this);
		prompter.list.addKeyListener( this);
//		prompter.list.addFocusListener( this);
		
		repaint();
	}
}

public void paint ( Graphics g) {
	super.paint( g);
	paintButton( g);
	WidgetTheme.borderPaintOuter( this, g, (AWTEvent.keyTgt == entry));
}

void paintBackground (Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.chPaintBackground( this, g, x+deco.x, y+deco.y, width-deco.width, height-deco.width, flags);
}

void paintButton() {
	if ( isShowing() ) {
		Graphics g = getGraphics();
		paintButton( g);
		g.dispose();
	}
}

void paintButton( Graphics g) {
	WidgetTheme.chPaint( this, g, deco, state);
}

protected String paramString() {
	return super.paramString();
}

/**
 * @deprecated
 */
public Dimension preferredSize () {
	int mw = 0;
	String ct = (String)selection;
	
	for ( int idx=items.size(); --idx>=0; ) {
		String cc  = (String)items.elementAt(idx);
		entry.buffer.set( cc);
		//int w = entry.tr.widthTo( entry.buffer, entry.buffer.len);
		int w = entry.getPreferredSize().width;
		if ( w > mw ){
			mw = w;
		}
	}
	entry.buffer.set( ct);
	mw += entry.deco.width;
	
	Dimension d = entry.getPreferredSize();
	if ( entry.tr.isHorizontal() ) {
		if ( mw > d.width )
			d.width = mw;
	}
	else {
		if ( mw > d.height )
			d.height = mw;
	}
	
	return WidgetTheme.chPreferredSize( this, d);
}

void process ( ItemEvent e ) {
	if ( (iListener != null) || ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) ){
		processEvent( e);
	}
	
	if ((flags & IS_OLD_EVENT) != 0) {
		postEvent (Event.getEvent( e));
	}
}

protected void processItemEvent( ItemEvent e) {
	if ( iListener != null ){
		iListener.itemStateChanged( e);
	}
}

public synchronized void remove ( String item) {
	items.removeElement( item);
}

public synchronized void removeAll () {
	items.removeAllElements();
}

public synchronized void removeItemListener ( ItemListener il) {
	iListener = AWTEventMulticaster.remove( iListener, il);
}

public void requestFocus () {
	entry.requestFocus();
}

public void reshape ( int x, int y, int w, int h ) {
	super.reshape( x, y, w, h);
	
	// there is no need for validation of compound IS_NATIVE_LIKES, they are no Containers
	// in JDK, so we automagically have to re-layout them
	doLayout();
	flags |= IS_VALID;
}

public synchronized void select ( String item) {
	if ( items.contains( item) ) {
		selection = item;
		if ( ! entry.getText().equals( item) ){
			entry.setText( item);
		}
	}
	closePrompt( true);
}

public synchronized void select ( int index) {
	if (index >= 0 && index < items.size()) {
		select((String)items.elementAt(index));
	}
}

public void setEnabled ( boolean isEnabled ) {
	super.setEnabled( isEnabled);
	entry.setEnabled( isEnabled);
}

public void setFont ( Font fnt) {
	super.setFont( fnt);
	entry.setFont( fnt);
}
}
