package java.awt;

import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.Vector;
import kaffe.awt.Defaults;
import kaffe.awt.RenderDirection;
import kaffe.awt.WidgetTheme;

/**
 * class RowCanvas -
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
class RowCanvas
  extends Component
  implements FocusListener
{
	Scrollbar hScroll;
	Scrollbar vScroll;
	int rowHeight = 10;
	Vector rows = new Vector();
	int first;
	NativeGraphics rgr;
	Renderer render;
	static TextBuffer convertBuffer = new TextBuffer();
	int innerBorder = WidgetTheme.borderInnerWidth( this);

abstract class Renderer
  implements AdjustmentListener
{
	Rectangle drawRect;
	TextBufferRender ir;

protected Renderer() {
	//use foreground & background color settings for rendering
	ir = new TextBufferRender( RowCanvas.this);
	drawRect = new Rectangle();
}

public void adjustmentValueChanged( AdjustmentEvent e) {
}

abstract void blankTrailing();

abstract Rectangle getDrawRect( TextBuffer tb, int x, int y);

abstract Rectangle getDrawRect( int idx);

abstract int getVisibleRows();

abstract int indexAt( int x, int y);

abstract void makeVisible(int x, int y);

int maxRowExtent() {
	int rs = rows.size();
	int ie, me = 0;
	TextBuffer b;
	
	for ( int i=0; i<rs; i++ ) {
		b = bufferAt(i);
		ie = ir.widthTo( b, b.len);
		if ( ie > me )
			me = ie;
	}
	
	return me;
}

int next( int steps) {
	return steps;
}

void paint( Graphics g, int y, int x, int dx, Rectangle r, boolean selected) {
	TextBuffer tb = bufferAt( y);
	ir.paint( g, tb, x, dx, r, false, selected );
}

boolean paintRawText( Graphics g, int idx) {
	Rectangle r = getDrawRect( idx);
	if ( r == null )
		return false;
		
	ir.paintRawText( g, bufferAt(idx), r );
	return true;
}

abstract void updateCursor( TextCursor tc, int x, int y);

abstract boolean updateHScroll();

abstract int updateRenderEnv( Font fnt);

abstract boolean updateVScroll();
}
class Horizontal
  extends Renderer
{
Horizontal() {
	ir.setMode( RenderDirection.getDirection() );
}

public void adjustmentValueChanged( AdjustmentEvent e) {
	Object src = e.getSource();
	int val = e.getValue();
	
	if ( src == vScroll) {
		int lf = first;
		if ( val != first){
			first = val;
			vPosChange( lf - val);
			shiftVertical( lf - val, false );
		}
	}
	else if ( src == hScroll) {
		if ( val != -ir.offset ){
			ir.offset = -val;
			hPosChange();
		}
	}
}

void blankTrailing() {
	int b = innerBorder;
	int y0 = (rows.size()-first) * rowHeight;

	RowCanvas.this.blankTrailing(b, y0+b, RowCanvas.this.width-2*b, RowCanvas.this.height-b-y0);
}

Rectangle getDrawRect( TextBuffer tb, int x, int y) {
	int b = innerBorder;
	int y0 = (y - first) * rowHeight + b;

	if ( (y0<0) || (y0>RowCanvas.this.height) || (y>=rows.size()) )
		return null;
	
	drawRect.y = y0;
	drawRect.height = rowHeight;
	
	return (ir.getDrawRect( tb, x, drawRect));
}

Rectangle getDrawRect(int index) {
	int b = innerBorder;
	int y0 = (index - first) * rowHeight + b;
	
	if ( (y0<0) || (y0>RowCanvas.this.height) || (index>=rows.size()) )
		return null;
	
	drawRect.setBounds( b, y0, RowCanvas.this.width-2*b, rowHeight);
	return drawRect;
}

int getVisibleRows() {
	int db = innerBorder;
	int th = RowCanvas.this.height;

	if ( (th == 0) || (rowHeight == 0) )
		return 0;
			
	return (th-2*db)/rowHeight;
}

int indexAt(int x, int y) {
	int y0 = rowHeight;
	int idx = first;
	
	//above visible contents ( used for select scrolling )
	if ( y < 0 ) {
		for( ; idx > 0; idx-- ) {
			y0 -= rowHeight;
			if ( y0 < y )
				break;
		}
	}
	//inside or below visible contents
	else {
		for( ; y0<RowCanvas.this.height; y0+=rowHeight ) {
			if ( y0 > y )
				break;
			idx++;
		}
	}

	return Math.min( rows.size()-1, idx);
}

void makeVisible ( int x, int y) {
	int vr = getVisibleRows();
	int nf = first;

	//no valid extends so far
	if ( vr <= 0 )
		return;
		
	if ( y < first ) {
		nf = y;
	}
	else {
		while ( y >= nf + vr ){
			nf++;
		}
	}
	
	TextBuffer tb = bufferAt( y);
	if ( x > tb.len )
		x = tb.len;
		
	int sts = ir.shiftsToShow( tb, x);
	if ( (nf != first) || (sts != 0) ) {
		first = nf;
		ir.offset -= sts;
		if ( updateHScroll() || updateVScroll() )
			rearrange();
		else
			repaint();
		
		if ( hScroll != null )
			hScroll.setValue( -ir.offset);
		if ( vScroll != null )
			vScroll.setValue( first);
	}

}

void shiftVertical ( int rows, boolean updScroll) {
	if ( (rows == 0) || (rgr == null) )
		return;

	// PM - FIX THIS. In case we are not called from adjustmentNotify (i.e. the
	// EventDispatchThread), where it is obsolete anyway, this might cause reverse scrolling
	// (see Scrollbar.setValue). We currently solve this by fiddling with the inner
	// workings of Scrollbar.
	if ( updScroll && vScroll != null) {
		vScroll.setValues( first, vScroll.vis, vScroll.min, vScroll.max, false);
	}

	int d = innerBorder;
	int d2 = d*2;
	int dy = rows * rowHeight;
	int vr = getVisibleRows();

	if ( rows < 0 ){
		if ( -rows < vr )  // otherwise the new top wasn't visible
			rgr.copyArea( d, d-dy, RowCanvas.this.width -d2, RowCanvas.this.height-d2+dy-1, 0, dy);

		repaintRows( first + vr + rows, 1 - rows );
	}
	else {
		if ( rows < vr )   // otherwise the old top will not be visible anymore
			rgr.copyArea( d, d, RowCanvas.this.width -d2, RowCanvas.this.height-d2-dy-1, 0, dy);

		repaintRows( first, rows-1);
	}
}

void updateCursor( TextCursor tc, int x, int y) {
	TextBuffer tb = bufferAt( y);
	if ( x > tb.len )
		x = tb.len;
	tc.setX( x, render.ir.positionFromIndex( tb, x));
	tc.setY( y, (y - first) * rowHeight + innerBorder);
}

boolean updateHScroll() {
	if ( hScroll == null )
		return false;

	boolean	v1 = hScroll.isSliderShowing();
	int mw = maxRowExtent() + 2*innerBorder + deco.width;
	
	// entire contents going to be visible ( reset to left )
	if ( v1 && (mw <= width) )
		ir.offset = 0;
		
	// notify only on left reset
	hScroll.setValues( -ir.offset, RowCanvas.this.width, 0, mw, (v1 && (ir.offset==0)) );
	
	return ( v1 != hScroll.isSliderShowing() );
}

int updateRenderEnv( Font fnt) {
	ir.width = RowCanvas.this.width;
	if ( fnt != null ) {
		FontMetrics fm = getFontMetrics( fnt);
		ir.height = fm.getHeight() + 1;
		ir.fm = fm;
		ir.tabWidth = 3*fm.charWidth( 'x');
	}
	return ir.height;
}

boolean updateVScroll() {
	if ( vScroll == null )
		return false;
		
	boolean	v1 = vScroll.isSliderShowing();
	int rs = rows.size();
	int vr = (RowCanvas.this.height -2*innerBorder) / rowHeight;

	//entire contents going to be visible ( reset to top )
	if ( v1 && (vr >= rs) )
		first = 0;
		
	//no need to notify ( first already adjusted )
	vScroll.setValues( first, vr, 0, rs, false );
	
	return ( v1 != vScroll.isSliderShowing() );
}
}
class Vertical
  extends Renderer
{
Vertical() {
	ir.setMode( RenderDirection.TB);
}

public void adjustmentValueChanged( AdjustmentEvent e) {
	Object src = e.getSource();
	int val = e.getValue();
	
	if ( src == hScroll){
		if ( val != first){
			first = val;
			hPosChange();
		}
	}
	else if ( src == vScroll) {
		if ( val != -ir.offset) {
			ir.offset = -val;
			vPosChange( 0);
			RowCanvas.this.repaint();
		}
	}
}

void blankTrailing() {
	int b = innerBorder;
	int x0 = RowCanvas.this.width - (rows.size()-first) * rowHeight;
	RowCanvas.this.blankTrailing(b, b, x0-2*b, RowCanvas.this.height-2*b);
}

Rectangle getDrawRect(TextBuffer tb, int x, int y) {
	int b = innerBorder;
	int x0 = RowCanvas.this.width - b - (y - first) * rowHeight;
	
	if ( (x0<b) || (x0>RowCanvas.this.width) || (y>=rows.size()) )
		return null;

	drawRect.x = x0-rowHeight;
	drawRect.width = rowHeight;
	
	return (ir.getDrawRect( tb, x, drawRect));
}

Rectangle getDrawRect(int index) {
	int b = innerBorder;
	int x0 = RowCanvas.this.width - b - (index - first) * rowHeight;
	
	if ( (x0<b) || (x0>RowCanvas.this.width) || (index>=rows.size()) )
		return null;

	drawRect.setBounds( x0-rowHeight, b, rowHeight, RowCanvas.this.height-2*b);
	return drawRect;
}

int getVisibleRows() {
	int db = innerBorder;
	return ((RowCanvas.this.width-2*db)/rowHeight);
}

int indexAt(int x, int y) {
	int x0 = RowCanvas.this.width -innerBorder- rowHeight;
	int idx = first;
	
	for( ; x0>0; x0-=rowHeight ) {
		if ( x0 < x )
			break;
		idx++;
	}
		
	return Math.min( rows.size()-1, idx);
}

void makeVisible(int x, int y) {
	int vr = getVisibleRows();
	int nf = first;
	
	//no valid extends so far
	if ( vr <= 0 )
		return;

	if ( y < first ) {
		nf = y;
	}
	else {
		while ( y >= nf + vr ){
			nf++;
		}
	}
	
	TextBuffer tb = bufferAt( y);
	if ( x > tb.len )
		x = tb.len;
		
	int sts = ir.shiftsToShow( tb, x);
	if ( (nf != first) || (sts != 0) ) {
		first = nf;
		ir.offset -= sts;

		if ( updateHScroll() || updateVScroll() )
			rearrange();
		else
			RowCanvas.this.repaint();

		if ( vScroll != null )
			vScroll.setValue( -ir.offset);
		if ( hScroll != null )
			hScroll.setValue( first);
	}
}

void updateCursor( TextCursor tc, int x, int y) {
	TextBuffer tb = bufferAt( y);
	if ( x > tb.len )
		x = tb.len;
	tc.setX( x, render.ir.positionFromIndex( tb, x));
	tc.setY( y, RowCanvas.this.width - (y - first + 1) * rowHeight - innerBorder); 
}

boolean updateHScroll() {
	if ( hScroll == null )
		return false;
		
	boolean	v1 = hScroll.isSliderShowing();
	int rs = rows.size();
	int vr = (width -2*innerBorder) / rowHeight;

	//entire contents going to be visible ( reset to top )
	if ( v1 && (vr >= rs) )
		first = 0;
		
	//no need to notify ( first already adjusted )
	hScroll.setValues( first, vr, 0, rs, false );
	
	return ( v1 != hScroll.isSliderShowing() );
}

int updateRenderEnv( Font fnt) {
	ir.height = RowCanvas.this.height;
	if ( fnt != null ) {
		FontMetrics fm = getFontMetrics( fnt);
		ir.width = fm.getHeight() + 1;
		ir.fm = fm;
		ir.tabWidth = 3*fm.charWidth( 'x');
	}
	return ir.width;
}

boolean updateVScroll() {
	if ( vScroll == null )
		return false;

	boolean	v1 = vScroll.isSliderShowing();
	int mh = maxRowExtent() + 2*innerBorder + deco.width;
	int cv = vScroll.getValue();
	
	// entire contents going to be visible ( reset to left )
	if ( v1 && (mh <= RowCanvas.this.height) )
		cv = 0;
		
	// notify only on left reset
	vScroll.setValues( cv, RowCanvas.this.height, 0, mh, (v1 && (cv==0)) );
	
	return ( v1 != vScroll.isSliderShowing() );
}
}

RowCanvas() {
	if ( RenderDirection.isHorizontal() )
		render = new Horizontal();
	else
		render = new Vertical();
}

public void addNotify() {
	super.addNotify();
	
	if ( rgr == null ) {
		setResGraphics();
	}
	
}

void blankTrailing(int x, int y, int width, int height) {
	rgr.setColor( bgClr);
	rgr.fillRect( x, y, width, height);
}

TextBuffer bufferAt( int y) {
	TextBuffer b;
	Object o = rows.elementAt( y);
	
	if ( o instanceof TextBuffer )
		b = (TextBuffer)o;
	else {
		b = convertBuffer;
		b.set( o.toString() );
	}

	return b;
}

public void focusGained( FocusEvent e) {
	paintComponentBorder();
	
	if ( Defaults.FocusScrolls ) {
		if ( hScroll != null )
			hScroll.setBackground( Defaults.FocusClr);
		if ( vScroll != null )
			vScroll.setBackground( Defaults.FocusClr);	
	}
	
	redirectFocusEvent( e);
}

public void focusLost( FocusEvent e) {
	paintComponentBorder();
	
	if ( Defaults.FocusScrolls ) {
		if ( hScroll != null )
			hScroll.setBackground( Defaults.BtnClr);
		if ( vScroll != null )
			vScroll.setBackground( Defaults.BtnClr);	
	}
	
	redirectFocusEvent( e);
}

void hPosChange() {
	repaint();
}

void innerLayout() {
	int pb = WidgetTheme.borderOuterWidth( this);
	int w = parent.width-2*pb;
	int h = parent.height-2*pb;
	int sbd = Defaults.ScrollbarWidth;

	boolean vVis = (vScroll != null ) && vScroll.isSliderShowing();
	boolean hVis = (hScroll != null ) && hScroll.isSliderShowing();

	setBounds( pb, pb, vVis ? w-sbd : w, hVis ? h-sbd : h);

	if ( vScroll != null ) {
		if ( vScroll.isVisible() && ! vVis )
			vScroll.setVisible( false);
		else if ( !vScroll.isVisible() && vVis )
			vScroll.setVisible( true);
		vScroll.setBounds( pb+w-sbd, pb, sbd, hVis ? h-sbd : h);
	}
	if ( hScroll != null) {
		if ( hScroll.isVisible() && ! hVis )
			hScroll.setVisible( false);
		else if ( !hScroll.isVisible() && hVis )
			hScroll.setVisible( true);
		hScroll.setBounds( pb, pb+h-sbd, vVis ? w-sbd : w, sbd);
	}
	
	updateScrolls();
}

void makeVisible( int y) {
	makeVisible( -render.ir.offset, y);
}

void makeVisible( int x, int y) {
	if ( y<0 )
		y = 0;
	else if ( y >= rows.size() )
		y = rows.size()-1;

	render.makeVisible( x, y);
}

public void mouseEntered( MouseEvent e) {

	if ( (parent.mouseListener != null) && (e.getSource() == this) ){
		// give our parent a chance to intercept mouse events
		// check for mouseListeners first, it's a rare case
		redirectMouseEvent( e);
		return;   // don't do action twice, we get called recursivly by our parent
	}
}

public void mouseExited( MouseEvent e) {

	if ( (parent.mouseListener != null) && (e.getSource() == this) ){
		// give our parent a chance to intercept mouse events
		// check for mouseListeners first, it's a rare case
		redirectMouseEvent( e);
		return;   // don't do action twice, we get called recursivly by our parent
	}
}

public void paint( Graphics g) {
	repaintRows( first, render.getVisibleRows() );
	WidgetTheme.borderPaintInner( this, g, (AWTEvent.keyTgt == this));
}

void paintBackground( Graphics g, int x, int y, int width, int height, int flags) {
}

void paintComponentBorder () {
	// not nice to resolve this, but our getGraphics() is more expensive

	Graphics g = super.getGraphics();
	if ( g != null ){
		WidgetTheme.borderPaintInner( this, g, (AWTEvent.keyTgt == this));
		g.dispose();
	}
	// repaint();
}

void propagateBgClr (Color c) {
//	render.bgClr = c;
	super.propagateBgClr( c);
}

void propagateFgClr (Color c) {
//	render.fgClr = c;
	super.propagateFgClr( c);
}

void propagateFont (Font fnt) {
	if ( rgr != null )
		rgr.setFont( fnt);
		
	rowHeight = render.updateRenderEnv( fnt);
	super.propagateFont( fnt);
}

void propagateReshape () {
	if ( rgr != null ){
		updateLinkedGraphics();
		
		// we have our own clipping, which we have to take care of
		int d = innerBorder;
		rgr.setClip( d, d, width-2*d, height-2*d);
	}

	rowHeight = render.updateRenderEnv( null);
	vPosChange(0);
}

void rearrange() {
	Graphics g = parent.getGraphics();
	if ( g != null ) {
		g.setColor( parent.bgClr );
		g.fillRect( 0, 0, parent.width, parent.height);
		WidgetTheme.borderPaintOuter( parent, g, (AWTEvent.keyTgt == this));
		g.dispose();
	}
	innerLayout();
}

void redirectFocusEvent( FocusEvent e) {
	if ( parent.focusListener != null ){
		AWTEvent.setSource( e, parent);
		parent.process( e);
	}
}

void redirectKeyEvent( KeyEvent e) {
	if ( parent.keyListener != null ){
		AWTEvent.setSource( e, parent);
		parent.process( e);
	}
}

void redirectMotionEvent( MouseEvent e) {
	if ( parent.motionListener != null ){
		e.retarget( parent, x, y);
		parent.process( e);

		// be careful, the event might be used subsequently
		e.retarget( this, -x, -y);
	}
}

void redirectMouseEvent( MouseEvent e) {
	if ( parent.mouseListener != null ){
		e.retarget( parent, x, y);
		parent.process( e);
		
		// be careful, the event might be used subsequently
		e.retarget( this, -x, -y);
	}
}

public void removeNotify() {
	super.removeNotify();

	if ( rgr != null ) {	
		rgr.dispose();
		rgr = null;
	}
}

void repaintRow( int idx) {
}

void repaintRows( int sIdx, int len) {
	if ( rgr == null )
		return;
		
	int rs = rows.size();

	for ( int i=first; i<rs; i++) {
		if ( (i >= sIdx) && (i <= sIdx + len) )
			repaintRow( i);
	}
	
	render.blankTrailing();
}

void setListeners() {
	if ( hScroll != null )
		hScroll.addAdjustmentListener( render);
	if ( vScroll != null )
		vScroll.addAdjustmentListener( render);
	addFocusListener( this);
}

void setResGraphics () {
	int d = innerBorder;
	rgr = NativeGraphics.getClippedGraphics( null, this, 0, 0,
                                           d, d, width-2*d, height-2*d);
	if ( rgr != null )
		linkGraphics( rgr);
}

public void update ( Graphics g ){
	// we don't need background blanking
	paint( g);
}

boolean updateHScroll() {
	return render.updateHScroll();
}

void updateScrolls() {
	boolean dv1 = updateVScroll();
	boolean dv2 = updateHScroll();
	
	if ( dv1 || dv2 ) {
		rearrange();
	}
}

boolean updateVScroll() {
	return render.updateVScroll();
}

void vPosChange( int steps) {
}
}
