package java.awt;

import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import kaffe.awt.Defaults;
import kaffe.awt.WidgetTheme;
import kaffe.util.Timer;
import kaffe.util.TimerClient;

/**
 * class Scrollbar - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Scrollbar
  extends Component
  implements Adjustable, MouseListener, MouseMotionListener, TimerClient
{
	final private static long serialVersionUID = 8451667562882310543L;
	final public static int HORIZONTAL = 0;
	final public static int VERTICAL = 1;
	int min;
	int max;
	int vis;
	int val;
	int ori;
	AdjustmentListener aListener;
	int blockInc;
	int unitInc = 1;
	int currentOp;
	Rectangle trackBounds = new Rectangle();
	int sliderOffset;
	int sliderLength;
	int trackOffset;
	Graphics trackGraphics;
	static int named;

public Scrollbar() {
	this( VERTICAL);
}

public Scrollbar( int ori) throws IllegalArgumentException {
	this( ori, 0, 30, 0, 100);
}

public Scrollbar( int ori, int value, int visible, int min, int max) throws IllegalArgumentException{
	setBackground( Defaults.BtnClr);
	setOrientation( ori);
	setValues( value, visible, min, max);

	deco = WidgetTheme.sbGetDeco( this);
	
	addMouseListener( this);
	addMouseMotionListener( this);
}

public synchronized void addAdjustmentListener( AdjustmentListener al) {
	aListener = AWTEventMulticaster.add( aListener, al);
}

public int getBlockIncrement() {
	return (getPageIncrement());
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

int getCurrentOp ( int x, int y) {
	WidgetTheme.sbGetTrackBounds( this, trackBounds);
	int op = WidgetTheme.sbGetAdjustmentArea( this, x, y, trackBounds, sliderOffset, sliderLength);
	
	if ( op == AdjustmentEvent.TRACK )
		trackOffset = (ori == HORIZONTAL) ? x - trackBounds.x - sliderOffset : y - trackBounds.y - sliderOffset;

	return op;
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "scrollbar" + (named++);
}

/**
 * @deprecated
 */
public int getLineIncrement() {
	return unitInc;
}

public int getMaximum() {
	return max;
}

public int getMinimum() {
	return min;
}

public int getOrientation() {
	return ori;
}

/**
 * @deprecated
 */
public int getPageIncrement() {
	return blockInc;
}

public int getUnitIncrement() {
	return (getLineIncrement());
}

public int getValue() {
	return val;
}

/**
 * @deprecated
 */
public int getVisible() {
	return vis;
}

public int getVisibleAmount() {
	return (getVisible());
}

public boolean isFocusTraversable() {
	return false;
}

boolean isSliderShowing() {
	return sliderLength > 0;
}

public void mouseClicked( MouseEvent e) {
}

public void mouseDragged( MouseEvent e) {
	if ( (currentOp != AdjustmentEvent.TRACK) || (trackGraphics == null) )
		return;

	int newOffset;

	if ( ori == VERTICAL ) {
		newOffset = e.getY() - trackBounds.y - trackOffset;
		if ( newOffset <= 0 ) {
			if ( sliderOffset == 0 )
				return;
			newOffset = 0;
		}
		else if ( newOffset > trackBounds.height - sliderLength ) {
			if ( sliderOffset + sliderLength == trackBounds.height )
				return;
			newOffset = trackBounds.height - sliderLength;
		}
		
		paintBackground(trackGraphics,  trackBounds.x, trackBounds.y + sliderOffset, trackBounds.width, sliderLength, 0);
	}
	else {
		newOffset = e.getX() - trackBounds.x - trackOffset;
		if ( newOffset <= 0 ) {
			if ( sliderOffset == 0 )
				return;
			newOffset = 0;
		}
		else if ( newOffset > trackBounds.width - sliderLength ) {
			if ( sliderOffset + sliderLength == trackBounds.width )
				return;
			newOffset = trackBounds.width - sliderLength;
		}
		
		paintBackground(trackGraphics, trackBounds.x + sliderOffset, trackBounds.y, sliderLength, trackBounds.height, 0);
	}
	
	sliderOffset  = newOffset;
	paintSlider( trackGraphics);
	updateValue();
}

public void mouseEntered( MouseEvent e) {
}

public void mouseExited( MouseEvent e) {
}

public void mouseMoved( MouseEvent e) {
}

public void mousePressed( MouseEvent e) {
	currentOp = getCurrentOp( e.getX(), e.getY() );
	
	switch ( currentOp) {
		case AdjustmentEvent.UNIT_DECREMENT:
		case AdjustmentEvent.UNIT_INCREMENT:
			paintButtons();
			//go at least once
			timerExpired( null);
			Timer.getDefaultTimer().addClient( this, 300, 50);
			break;
		case AdjustmentEvent.TRACK:
			if ( trackGraphics == null ) {
				trackGraphics = getGraphics();
				if ( trackGraphics != null )
					trackGraphics.setColor( bgClr);
			}
			break;
		default:
	}
					
}

public void mouseReleased( MouseEvent e) {

	switch( currentOp) {
	case AdjustmentEvent.UNIT_INCREMENT:
	case AdjustmentEvent.UNIT_DECREMENT:
		currentOp = 0; //for button painting
		Timer.getDefaultTimer().removeClient( this);
		paintButtons();
		break;
	case AdjustmentEvent.TRACK:
		trackOffset = -1;
		if ( trackGraphics != null ){
			trackGraphics.dispose();
			trackGraphics = null;
		}
		break;
	case AdjustmentEvent.BLOCK_DECREMENT:
		setValue( val - blockInc);
		break;
	case AdjustmentEvent.BLOCK_INCREMENT:
		setValue( val + blockInc);
		break;
	}
	
	currentOp = 0;
}

void notifyAdjust() {
	if ( (aListener != null) ||
	     (eventMask & AWTEvent.ADJUSTMENT_EVENT_MASK) != 0 ||
	     ((flags & IS_OLD_EVENT) != 0) ) {
		Toolkit.eventQueue.postEvent( AdjustmentEvt.getEvent( this,
		                                                      AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED,
		                                                      currentOp, val));
	}
}

public void paint ( Graphics g ) {
	paintButtons( g);
	paintSlider( g);
}

void paintBackground(Graphics g, int x, int y, int w, int h, int flags) {
	WidgetTheme.sbPaintBackground(this, g, x, y, w, h, flags);
}

void paintButtons () {
	Graphics g = getGraphics();
	if ( g != null ) {
		paintButtons( g);
		g.dispose();
	}
}

void paintButtons ( Graphics g ) {
	WidgetTheme.sbPaintDecrementButton( this, g, (currentOp == AdjustmentEvent.UNIT_DECREMENT));
	WidgetTheme.sbPaintIncrementButton( this, g, (currentOp == AdjustmentEvent.UNIT_INCREMENT));
}

void paintSlider ( Graphics g ) {
	WidgetTheme.sbPaintSlider( this, g, trackBounds, sliderOffset, sliderLength, (currentOp == AdjustmentEvent.TRACK));
}

protected String paramString() {
	return super.paramString() + ",Value: " + val + ",Visible: " + vis + ",Min: " + min + ",Max: " + max; 
}

/**
 * @deprecated
 */
public Dimension preferredSize() {
	Dimension d = new Dimension();
	//let Theme decide about text and direction preferrences		
	return( WidgetTheme.sbPreferredSize( this, d));
}

void process ( AdjustmentEvent e ) {
	if ( (aListener != null) || ((eventMask & AWTEvent.ADJUSTMENT_EVENT_MASK) != 0) ){
		processEvent( e);
	}
	
	if ( (flags & IS_OLD_EVENT) != 0 ) {
		postEvent( Event.getEvent( e));
	}
}

protected void processAdjustmentEvent ( AdjustmentEvent e) {
	if (  aListener != null ) {
		aListener.adjustmentValueChanged( e);
	}
}

public synchronized void removeAdjustmentListener( AdjustmentListener al) {
	aListener = AWTEventMulticaster.remove( aListener, al);
}

public void reshape( int x, int y, int width, int height) {
	super.reshape( x, y, width, height);
	updateSliderRect();
}

public synchronized void setBlockIncrement( int inc) {
	setPageIncrement(inc);
}

/**
 * @deprecated
 */
public void setLineIncrement(int inc) {
	unitInc = inc;
}

public synchronized void setMaximum( int max) {
	setValues( val, vis, min, max);
}

public synchronized void setMinimum( int min) {
	setValues( val, vis, min, max);
}

public synchronized void setOrientation( int ori) throws IllegalArgumentException {
	if ( (ori != HORIZONTAL) && (ori != VERTICAL) )
		throw new IllegalArgumentException();
	this.ori = ori;
}

public void setPageIncrement(int inc) {
	setValues( val, inc, min, max);
}

public void setUnitIncrement( int inc) {
	setLineIncrement(inc);
}

public synchronized void setValue( int value) {
	if ( val != value) {
		setValues( value, vis, min, max);
	}
}

public void setValues( int value, int visible, int min, int max) {
	setValues( value, visible, min, max, true);
}

synchronized void setValues( int value, int visible, int min, int max, boolean notify ) {
	// PM: not ideal - this is a workaround for clients with coexisting sync (own) and
	// async (Scrollbar initiated) updates. In case they rely on sync updates, they are
	// just interested in Scrollbar updates, but not notifications. However, this is a design
	// flaw of these clients (this is not a public API of Scrollbar !!)

	int lastVal = this.val;

	if ( (value == val) && (visible == vis) && (min == this.min) && (max == this.max) )
		return;

	// Make size adjustments.
	if (max < min)								max = min;
	if (visible > max-min)				visible = max-min;
	if ( value > max - visible )	value = max - visible;
	else if ( value < min )				value = min;
			
	this.min = min;
	this.max = max;
	this.val = value;
	this.vis = visible;
	this.blockInc = visible;
	
	update();

	if ( notify && (val != lastVal) )
		notifyAdjust();
		
}

public synchronized void setVisibleAmount( int vis) {
	setValues( val, vis, min, max);
}

public void timerExpired( Timer t) {
	switch ( currentOp) {
	case AdjustmentEvent.UNIT_DECREMENT:
		if ( val == min)
			Timer.getDefaultTimer().removeClient( this);
		else
			setValue( val-unitInc);
		break;
	case AdjustmentEvent.UNIT_INCREMENT:
		if ( val + vis == max)
			Timer.getDefaultTimer().removeClient( this);
		else
			setValue( val+unitInc);
		break;
	}	
}

void update() {
	Graphics g = null;
	
	if ( trackGraphics != null )
		g = trackGraphics;
	else if ( (flags & IS_VISIBLE) != 0 )
		g = getGraphics();
	
	if ( g != null ){
		paintBackground(g, trackBounds.x, trackBounds.y, trackBounds.width, trackBounds.height, 0);

		updateSliderRect();
		
		paint( g);	
		if ( g != trackGraphics ) {
			g.dispose();
		}
	}
	else {
		updateSliderRect();
	}
}

void updateSliderRect () {
	int delta = max - min;

	//entire contents visible
	if ( delta <= 0 || vis <= 0 || max == vis) {
		sliderLength = 0;
		return;
	}

	WidgetTheme.sbGetTrackBounds( this, trackBounds);
	
	//slider position from value
	if ( ori == HORIZONTAL ) {
		int dx = Math.max( vis * trackBounds.width / delta, 10);
		int x0 = (val - min) * ( trackBounds.width - dx ) / (max - vis);
		if ( x0 >= 0 ) {
			sliderOffset = x0;
			sliderLength = dx;
		}
		else
			sliderLength = 0;
	}
	else {
		int dy = Math.max (vis * trackBounds.height / delta, 10);
		int y0 = (val - min) * (trackBounds.height - dy) / (max - vis);
		if ( y0 >= 0 ) {
			sliderOffset = y0;
			sliderLength = dy;
		}
		else
			sliderLength = 0;
	}
	
}

void updateValue () {
	//value from slider position
	int maxLen = (ori == HORIZONTAL) ? trackBounds.width : trackBounds.height;
	int newVal = min + ( (max - vis) * sliderOffset ) / (maxLen - sliderLength);

	if ( val != newVal ) {
		val = newVal;
		notifyAdjust();
	}
}
}
