package java.awt;

import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Vector;
import kaffe.awt.Defaults;
import kaffe.awt.WidgetTheme;

/**
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author J.Mehlitz
 */
public class TextArea
  extends TextComponent
{
	final private static long serialVersionUID = 3692302836626095722L;
	final public static int SCROLLBARS_BOTH = 0;
	final public static int SCROLLBARS_VERTICAL_ONLY = 1;
	final public static int SCROLLBARS_HORIZONTAL_ONLY = 2;
	final public static int SCROLLBARS_NONE = 3;
	int crows;
	int ccols;
	TextPane tp = new TextPane();
	static int named;

class TextPane
  extends RowCanvas
  implements KeyListener, MouseListener, MouseMotionListener
{
	Point sSel = new Point();
	Point eSel = new Point();
	TextCursor tc = new TextCursor(2, Color.blue);

public TextPane () {	
	this.setCursor( Cursor.getPredefinedCursor( render.ir.getCursorShape()));
	insertLine( "", 0);
	
	this.addKeyListener( this);
	this.addMouseListener( this);
	this.addMouseMotionListener( this);
	this.addFocusListener( this);
}

synchronized void append ( String s ) {
	int         i, nOld, nNew;
	TextBuffer  tb;
	String[]    lines;

	blankCaret();	
	lines = breakLines( s);
	
	nOld = rows.size() - 1;
	tb = (TextBuffer)rows.elementAt( nOld);
	tb.append( lines[0]);
	
	for ( i=1; i<lines.length; i++ ){
		rows.addElement( new TextBuffer( lines[i]));
	}
	
	nNew = rows.size() - 1;
	repaintRows( nOld, nNew - nOld);
	setCaretPos( tb.len, nNew, true, true);
}

void backspace() {
	TextBuffer tb = bufferAt(tc.index.y);
	if ( tc.index.x >= 1 ) {
		tb.remove( tc.index.x-1, 1);
		setCaretPos( tc.index.x-1, tc.index.y, false, true);
		repaintLine( tc.index.y, tc.index.x, tb);
		updateScrolls();
	}
	else if ( cursorLeft(1, false) ){
		TextBuffer tb1 = bufferAt(tc.index.y);
		tb1.append( tb);
		rows.removeElement( tb);
		this.repaint();
		updateScrolls();
	}
}

void blankCaret() {
	TextBuffer tb;
		
	if ( rgr != null ) {
		rgr.setColor( this.bgClr );
		render.ir.blankCaret( rgr, tc);
		tb = bufferAt(tc.index.y);
		if ( tb.len > tc.index.x )
			render.ir.paint( rgr, tb, tc.index.x, 1, render.getDrawRect(tc.index.y), false, false);
	}
}

void blankTrailing(int x, int y, int width, int height) {
	paintBackground( rgr, x, y, width, height, 0);
}

String[] breakLines( String str) {
	int       i, i0 = 0, n = str.length();
	char      c;
	Vector    v;
	String[]  sbuf;
	char[]    cbuf;

	if ( (str == null) || (n == 0) )
		return new String[1];
		
	v = new Vector( n / 20);
	cbuf = str.toCharArray();

	// sometimes I wonder how long we will suffer from old DOS habits: some
	// editors still use obscure '\r\r\n' or '\r' "weak" linefeeds to mark
	// automatically wrapped lines

	for ( i=0; i<n; i++ ) {
		if ( (c=cbuf[i]) == '\n' ) {
			v.addElement( new String( cbuf, i0, i-i0));
			i0 = i+1;
		}
		else if ( c == '\r' ) {
			v.addElement( new String( cbuf, i0, i-i0));
			// skip all subsequent '\r'
			for ( i++; (i < n) && (cbuf[i] == '\r'); i++ );
			i0 = (cbuf[i] == '\n') ? i+1 : i;
		}
	}

	if ( i0 <= n ){
		v.addElement( new String( cbuf, i0, i-i0));
	}

	sbuf = new String[v.size()];
	v.copyInto( sbuf);

	return sbuf;
}

TextBuffer bufferAt( int y) {
	return (TextBuffer)rows.elementAt( y);
}

void cursorDown( int steps, boolean extend) {
	int newX;
	int newY = Math.min( tc.index.y + steps, rows.size() -1);
	
	if ( newY == tc.index.y)
		return;
		
	if (tc.index.x == ((TextBuffer)rows.elementAt( tc.index.y)).len )
		newX = ((TextBuffer)rows.elementAt( newY)).len;
	else
		newX = Math.min( tc.index.x, ((TextBuffer)rows.elementAt( newY)).len);
		
	if ( extend )
		updateSel( newX, newY, true);
	else
		setCaretPos( newX, newY, true, true);
}

void cursorEnd( boolean extend) {
	TextBuffer tb = bufferAt(tc.index.y);
	if ( extend)
		updateSel( tb.len, tc.index.y, true);
	else
		setCaretPos( tb.len, tc.index.y, true, true);
}

void cursorHome( boolean extend) {
	if ( extend)
		updateSel( 0, tc.index.y, true);
	else
		setCaretPos( 0, tc.index.y, true, true);
}

boolean cursorLeft( int steps, boolean extend) {
	int nx = -1, ny = -1;
	
	if ( tc.index.x >= steps) {
		nx = tc.index.x - steps;
		ny = tc.index.y;
	}
	else if ( tc.index.y > 0 ) {
		nx = ((TextBuffer)rows.elementAt( tc.index.y-1)).len;
		ny = tc.index.y-1;
	}

	if ( ny > -1 ) {
		if ( extend )
			updateSel( nx, ny, true);
		else
			setCaretPos( nx, ny, true, true);
		return true;
	}
	
	return false;
}

void cursorRight( int steps, boolean extend) {
	int nx = -1, ny = -1;
	TextBuffer tb = (TextBuffer)rows.elementAt( tc.index.y);
	
	if ( tb.len >= tc.index.x + steps) {
		nx = tc.index.x + steps;
		ny = tc.index.y;
	}
	else if ( tc.index.y < rows.size() - 1 ) {
		nx = 0;
		ny = tc.index.y+1;
	}

	if ( ny > -1 ) {
		if ( extend)
			updateSel( nx, ny, true);
		else
			setCaretPos( nx, ny, true, true);
	}
}

void cursorTextEnd( boolean extend) {
	int yIdx = rows.size() - 1;
	TextBuffer tb = (TextBuffer)rows.elementAt( yIdx);
	if ( extend)
		updateSel( tb.len, yIdx, true);
	else
		setCaretPos( tb.len, yIdx, true, true);
}

void cursorTextHome( boolean extend) {
	if ( extend)
		updateSel( 0, 0, true);
	else
		setCaretPos( 0, 0, true, true);
}

void cursorUp( int steps, boolean extend) {
	int newX;
	int newY = Math.max( tc.index.y - steps, 0);
	
	if ( newY == tc.index.y)
		return;
		
	if (tc.index.x == ((TextBuffer)rows.elementAt( tc.index.y)).len )
		newX = ((TextBuffer)rows.elementAt( newY)).len;
	else
		newX = Math.min( tc.index.x, ((TextBuffer)rows.elementAt( newY)).len);
		
	if ( extend )
		updateSel( newX, newY, true);
	else
		setCaretPos( newX, newY, true, true);
}

void del() {
	TextBuffer tb = bufferAt(tc.index.y);
	if ( tb.len > tc.index.x ) {
		tb.remove( tc.index.x, 1);
		repaintLine( tc.index.y, tc.index.x, tb);
		updateScrolls();
	}
	else if ( tc.index.y < rows.size() - 1 ){
		TextBuffer tb1 = bufferAt( tc.index.y+1);
		tb.append( tb1);
		rows.removeElement( tb1);
		updateScrolls();
		this.repaint();
	}
}

void deleteSel() {
	Point ps = getSelStart();
	Point pe = getSelEnd();
	TextBuffer tb = (TextBuffer)rows.elementAt( ps.y);
	int af = 1;
	
	if ( (ps.x == pe.x) && (ps.y == pe.y) )
		return;
		
	if ( ps.y == pe.y ) {
		tb.remove( ps.x, pe.x - ps.x);
		setCaretPos( ps.x, ps.y, true, true);
		repaintLine( ps.y, ps.x, tb);
		updateScrolls();
		return;
	}

	if ( ps.x > 0 )
		tb.remove( ps.x, tb.len - ps.x);
	else
		af = 0;
	tb = (TextBuffer)rows.elementAt( pe.y);
	tb.remove( 0, pe.x);
	
	for ( int i = ps.y + af; i < pe.y; i++)
		rows.removeElementAt( ps.y + af);
		
	updateScrolls();
	setCaretPos( ps.x, ps.y, false, true);
	this.repaint();
}

public void focusGained( FocusEvent e) {
	repaintCaret();
	super.focusGained( e);
}

public void focusLost( FocusEvent e) {
	repaintCaret();

	//would confuse replacement with clipboard contents via popup
  //	if ( hasSel() )
  //		copyToClipboard();

	super.focusLost( e);
}

int get1D( Point p) {
	int xt = 0;
	
	for ( int i=0; i<p.y; i++) {
		TextBuffer tb = (TextBuffer)rows.elementAt( i);
		xt += tb.len;
		xt ++;  // hjb 01-27-1999: count the virtual newline
	}
	xt += p.x;
	
	return xt;
}

Point get2D( int pos) {
	int rs = rows.size();
	int xt = 0;
	TextBuffer tb = null;

	if ( rs == 0 )
		return new Point( -1, -1);

	for ( int i=0; i<rs; i++) {
		tb = (TextBuffer)rows.elementAt( i);
		xt += tb.len;
		if (xt >= pos)
			return new Point( tb.len - xt + pos, i);
		pos --;  // hjb 01-27-1999: substract 1 for the virtual newline
	}
	
	return new Point( tb.len, rs-1);
}

Point getSelEnd() {
	if ( sSel.y > eSel.y )
		return sSel;
	if ( sSel.y == eSel.y ) {
		if ( sSel.x > eSel.x )
			return sSel;
		return eSel;
	}
	return eSel;
}

Point getSelStart() {
	if ( sSel.y < eSel.y )
		return sSel;
	if ( sSel.y == eSel.y ) {
		if ( sSel.x < eSel.x )
			return sSel;
		return eSel;
	}
	return eSel;
}

void hPosChange() {
	//maybe temporary inconsistent by layout
	if ( tc.index.y < rows.size() )
		render.updateCursor( tc, tc.index.x, tc.index.y);
	this.repaint();
}

boolean hasSel() {
	return ( (sSel.x != eSel.x) || (sSel.y != eSel.y) );
}

boolean hasSel( int row) {
	if ( hasSel() )
		return ( (sSel.y <= row) || (eSel.y >= row) );
	return false;
}

void insert( String s, boolean keepCursor) {
	if ( (s == null) || (s.length() == 0) )
		return;
		
	String[] lns = breakLines( s);
	int lnsi = 0;
	TextBuffer tb = (TextBuffer)rows.elementAt( tc.index.y);

	if ( lns.length == 0 )
		return;
		
	if ( lns.length == 1 ) {
		int ncp = keepCursor ? tc.index.x : tc.index.x + lns[0].length();
		tb.insert( tc.index.x, lns[0] );
		setCaretPos( ncp, tc.index.y, false, false);
		repaintLine( tc.index.y, 0, tb);
		return;
	}
	
	String le = tb.getString( tc.index.x, tb.len - tc.index.x);
	tb.remove( tc.index.x, tb.len - tc.index.x);
	tb.append( lns[lnsi++] );
	
	while ( lnsi < lns.length)
		tb = insertLine( lns[lnsi++], tc.index.y + lnsi - 1 );
	
	if ( ! keepCursor)	
		setCaretPos( tb.len, tc.index.y + lns.length - 1, false, false);
	tb.append( le);

	updateScrolls();
	this.repaint();
}

void insertChar( char c) {
	boolean app;
	
	TextBuffer tb = bufferAt( tc.index.y);
	app = ( tc.index.x == tb.len);

	tb.insert( tc.index.x, c);
	
	if ( !app){
		//leading problem ( repaint left cursor char also)
		int sIdx = (tc.index.x > 0 ) ? tc.index.x-1 : tc.index.x;
		repaintLine( tc.index.y, sIdx, tb);
	}
	
	cursorRight( 1, false);
}

TextBuffer insertLine( String str, int lIdx) {
	TextBuffer tb = new TextBuffer( str);
//	tb.setMetrics( fm, tabWidth);
	rows.insertElementAt( tb, lIdx);
	return tb;
}

/**
 * @deprecated
 */
void insertText(String s, int pos) {
	Point p = get2D( pos);
	setCaretPos( p.x, p.y, true, true);
	
	insert( s, false);
}

public void keyPressed( KeyEvent e) {

	if ( (TextArea.this.keyListener != null) && (e.getSource() == this) ){
		// give our parent a chance to intercept keystrokes
		// check for keyListeners first, it's a rare case
		redirectKeyEvent( e);
		return;   // don't do action twice, we get called recursivly by our parent
	}

	int code = e.getKeyCode();
	int mods = e.getModifiers();
	boolean sh = e.isShiftDown();

	//do not consume unused key for ShortcutHandler
	if ( (mods != 0) && (mods != e.SHIFT_MASK) )
		return;

	switch( code) {
		case e.VK_LEFT:
			cursorLeft( 1, sh);
			break;
		case e.VK_RIGHT:
			cursorRight( 1, sh);
			break;
		case e.VK_UP:
			cursorUp( 1, sh);
			break;
		case e.VK_DOWN:
			cursorDown( 1, sh);
			break;
		case e.VK_TAB:
			if (sh) {
				return;	//do not consume event for HotKeyHandler
			}
			if (!isEditable) { // hjb 01-27-1999: do nothing for
				break;     // TAB, ENTER, DEL and BS if not editable
			}
			insertChar( '\t' );
			break;
		case e.VK_ENTER:
		        if ( ! isEditable)
			    break;
			newline();
			break;
		case e.VK_BACK_SPACE:
		        if ( ! isEditable)
			    break;
			if ( hasSel() )
				replaceSelectionWith("");
			else
				backspace();
			break;
		case e.VK_DELETE:
		        if ( ! isEditable)
		            break;
			if ( hasSel() )
				replaceSelectionWith("");
			else
				del();
			break;
		case e.VK_HOME:
			cursorHome( sh);
			break;
		case e.VK_END:
			cursorEnd( sh);
			break;
		case e.VK_PAGE_UP:
			pageUp( sh);
			break;
		case e.VK_PAGE_DOWN:
			pageDown( sh);
			break;
		case e.VK_ESCAPE:
			resetSel( true);
			break;
		default:
		  return;
	}

	e.consumed = true;
}

public void keyReleased( KeyEvent e) {
	if ( (TextArea.this.keyListener != null) && (e.getSource() == this) ){
		// check for keyListeners first, it's a rare case
		redirectKeyEvent( e);
	}
}

public void keyTyped( KeyEvent e) {
	if ( ! isEditable || ! isPrintableTyped( e) )
		return;

	char c = e.getKeyChar();

	if ( hasSel() ) {
		Character cc = new Character( c);
		replaceSelectionWith( cc.toString() );
	}
	else {
		insertChar( c );
	}

	if ( (TextArea.this.keyListener != null) && (e.getSource() == this) ){
		// check for keyListeners first, it's a rare case
		redirectKeyEvent( e);
		return;
	}

	if ( (textListener != null) || (this.eventMask & AWTEvent.TEXT_EVENT_MASK) != 0 ) {
		Toolkit.eventQueue.postEvent( TextEvt.getEvent( TextArea.this,
		                                                TextEvt.TEXT_VALUE_CHANGED));		
	}
	
	e.consumed = true;
}

public void mouseClicked( MouseEvent e) {
	if ( TextArea.this.mouseListener != null ){
		// no need to retarget, already done by mousePressed
		TextArea.this.process( e);
	}
}

public void mouseDragged( MouseEvent e) {
	int y = render.indexAt( e.getX(), e.getY() );
	int x = render.ir.indexFromPosition( (TextBuffer)rows.elementAt(y), e.getX(), e.getY() );
	updateSel( x, y, true);
	
	if ( TextArea.this.motionListener != null ){
		// unlikely, check motionListener first
		redirectMotionEvent( e);
	}
}

public void mouseMoved( MouseEvent e) {
	if ( TextArea.this.motionListener != null ){
		// unlikely, check listener first
		redirectMotionEvent( e);
	}
}

public void mousePressed( MouseEvent e) {
	int mods = e.getModifiers();

	if ( e.isPopupTrigger() ){
		if ( (this.triggerPopup( e.getX(), e.getY())) != null )
			return;
	}

	switch ( mods) {
		case InputEvent.BUTTON1_MASK:
			tp.requestFocus();
			resetSel( true);
			int y = render.indexAt( e.getX(), e.getY() );
			int x = render.ir.indexFromPosition((TextBuffer)rows.elementAt( y), e.getX(), e.getY() );
			setCaretPos( x, y, true, true);
			break;
		case InputEvent.BUTTON2_MASK:
			pasteFromClipboard();
			break;
	}
	
	redirectMouseEvent( e);
}

public void mouseReleased( MouseEvent e) {
	redirectMouseEvent( e);
}

void newline() {
	String s = "";
	
	TextBuffer tb = bufferAt(tc.index.y);
	if ( tb.len > tc.index.x ) {
		int rl = tb.len - tc.index.x;
		s = tb.getString( tc.index.x, rl);
		tb.remove( tc.index.x, rl);
	}
	
	TextBuffer tbNew = insertLine( s, tc.index.y+1);
	tbNew.copyLevelFrom( tb);

//	updateScrolls();
	setCaretPos( tbNew.getLevel(), tc.index.y+1, true, true);
	repaintRows( tc.index.y-1, render.getVisibleRows() );
}

void pageDown( boolean extend) {
	cursorDown( render.getVisibleRows(), extend);
}

void pageUp( boolean extend) {
	cursorUp( render.getVisibleRows(), extend);
}

void paintBackground(Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.taPaintBackground( TextArea.this, rgr, x, y, width, height, (AWTEvent.keyTgt == this ? WidgetTheme.IS_ACTIVE : 0));
}

void propagateFont( Font fnt) {
	super.propagateFont(fnt);

	tc.setHeight( rowHeight-1, 0 );
	render.updateCursor( tc, tc.index.x, tc.index.y);

	this.repaint();
}

void repaintCaret() {
	if ( rgr != null ) {
		if ( AWTEvent.keyTgt == this)
			render.ir.paintCaret( rgr, tc);
		else {
			rgr.setColor( Defaults.TextCursorInactiveClr );
			render.ir.blankCaret( rgr, tc);
		}
	}
}

void repaintLine( int row, int startX, TextBuffer tb) {
	Rectangle r;
	
	if ( tb == null )
		tb = bufferAt( row);

	if ( (r = render.getDrawRect( tb, startX, row)) == null )
		return;

	// rgr.setColor( this.bgClr);
	// rgr.fillRect( r.x, r.y, r.width, r.height);
	paintBackground( rgr, r.x, r.y, r.width, r.height, 0);
	
	int ss = selXStart( row);
	int se = selXEnd( row, tb);

	if ( ss == se ) {
		render.paint( rgr, row, startX, tb.len-startX, r, false);
	}
	else {
		if ( ss > startX ) {
			render.paint( rgr, row, startX, ss-startX, r, false);
		}
		if ( se > startX ) {
			render.paint( rgr, row, ss, se-ss, r, true);
		}
		if ( tb.len > se ) {
			render.paint( rgr, row, se, tb.len-se, r, false);
		}
	}

	if ( tc.index.y == row )
		repaintCaret();
		
}

void repaintRow( int idx) {
	repaintLine( idx, 0, null);
}

void replaceRange( String s, int start, int end) {
	sSel = get2D( start);
	eSel = get2D( end);

	if ( (sSel.y > -1) && (eSel.y > -1) )
		deleteSel();
	else
		resetSel( false);
		
	insert( s, true);
}

void replaceSelectionWith( String s) {
	deleteSel();
	insert( s, false);
}

void resetSel( boolean repaint) {
	boolean se = hasSel();
		
	int y0 = Math.min( sSel.y, eSel.y);
	int y1 = Math.max( sSel.y, eSel.y);

	sSel.x = tc.index.x;
	sSel.y = tc.index.y;
	eSel.x = sSel.x;
	eSel.y = sSel.y;
		
	if ( se && repaint){
		repaintRows( y0, y1-y0);
	}
}

int selXEnd( int row, TextBuffer tb) {
	Point ps = getSelStart();
	Point pe = getSelEnd();
	
	if ( row > pe.y )
		return -1;
	if ( row < ps.y )
		return -1;
	if ( row == pe.y )
		return pe.x;
		
	return tb.len;
}

int selXStart( int row) {
	Point ps = getSelStart();
	Point pe = getSelEnd();

	if ( row > pe.y )
		return -1;
	if ( row < ps.y )
		return -1;
	if ( row == ps.y )
		return ps.x;
		
	return 0;
}

void setCaretPos( int x, int y, boolean repaint, boolean resetSel) {
	if ( resetSel )
		resetSel( repaint);
		
	if ( repaint) {
		blankCaret();
	}

	makeVisible( x, y);
	
	render.updateCursor( tc, x, y);
	repaintCaret();
		
	if ( resetSel)
		resetSel( false);
}

synchronized void setContents( String s) {
	String[] sa = breakLines( s);
	rows.removeAllElements();

	for ( int i=0; i<sa.length; i++)
		insertLine( sa[i], i);

	setCaretPos( 0, 0, false, true);
	
	updateScrolls();
	this.repaint();
}

boolean updateSel( int x, int y, boolean repaint) {
	int y0, y1;
	
	if ( (x == eSel.x) && (y == eSel.y) )
		return false;

	eSel.x = x;
	eSel.y = y;
	y0 = Math.min( sSel.y, eSel.y );
	y1 = Math.max( sSel.y, eSel.y );
			
	setCaretPos( x, y, false, false);
	if ( repaint) {
	//	repaint();
		repaintRows( y0, y1 - y0 + 1);
	}
		
	return true;
}

void vPosChange( int dy) {
	//maybe temporary inconsistent by layout
	if ( tc.index.y < rows.size() )
		render.updateCursor( tc, tc.index.x, tc.index.y);
}
}

public TextArea() {
	this( null, 10, 10, SCROLLBARS_BOTH);
}

public TextArea( String text) {
	this( text, 10, 20, SCROLLBARS_BOTH);
}

public TextArea( String text, int rows, int cols) {
	this( text, rows, cols, SCROLLBARS_BOTH);
}

public TextArea( String text, int rows, int cols, int scrolls) {
	crows = rows;
	ccols = cols;

	setLayout( null);
	WidgetTheme.taSetDefaults(this);

	if (scrolls == SCROLLBARS_VERTICAL_ONLY || scrolls == SCROLLBARS_BOTH) {
		tp.vScroll = new Scrollbar( Scrollbar.VERTICAL,0,0,0,0);
		add( tp.vScroll);
	}
	if (scrolls == SCROLLBARS_HORIZONTAL_ONLY || scrolls == SCROLLBARS_BOTH) {
		tp.hScroll = new Scrollbar( Scrollbar.HORIZONTAL,0,0,0,0);
		tp.hScroll.setUnitIncrement( tp.rowHeight);
		add( tp.hScroll);
	}

	buildMenu();
	add( tp);
	tp.setListeners();
	
	if ( text != null )
		append( text);
}

public TextArea( int rows, int cols) {
	this( null, rows, cols, SCROLLBARS_BOTH);
}

public void add( PopupMenu m) {
	tp.add( m);
}

public void append( String str) {
	tp.append( str);
}

/**
 * @deprecated as of JDK 1.1, replaced by append(String s)
 */
public void appendText(String s) {
	append(s);
}

protected void buildMenu() {
	PopupMenu p = new PopupMenu();
	p.add( new MenuItem("Cut")).setShortcut( new MenuShortcut( KeyEvent.VK_U, false) );
	p.add( new MenuItem("Copy")).setShortcut( new MenuShortcut( KeyEvent.VK_O, false) );
	p.add( new MenuItem("Paste")).setShortcut( new MenuShortcut( KeyEvent.VK_A, false) );
	p.addSeparator();
	p.add( new MenuItem("Select All")).setShortcut( new MenuShortcut( KeyEvent.VK_S, false) );
	p.addActionListener( this);
	
	tp.add( p);
}

public void doLayout() {
	tp.innerLayout();
}

public Color getBackground () {
	// some anomaly, we forward colors to our tp, so we should return its colors
	// for consistencies sake
	return tp.getBackground();
}

public int getCaretPosition() {
	Point p = new Point( tp.tc.index.x, tp.tc.index.y);
	return tp.get1D( p);
}

public int getColumns() {
	return ccols;
}

public Color getForeground () {
	// some anomaly, we forward colors to our tp, so we should return its colors
	// for consistencies sake
	return tp.getForeground();
}

Color getForegroundSelected(boolean sel) {
	return (WidgetTheme.taGetForegroundColorSelected(this, sel));
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "text" + (named++);
}

public Dimension getMinimumSize() {
	return (minimumSize());
}

public Dimension getMinimumSize( int rows, int cols) {
	return (minimumSize(rows, cols));
}

public Dimension getPreferredSize() {
	return preferredSize();
}

public Dimension getPreferredSize( int rows, int cols) {
	return preferredSize(rows, cols);
}

public int getRows() {
	return crows;
}

public int getScrollbarVisibility() {
	if ( (tp.hScroll == null) && (tp.vScroll == null) )
		return SCROLLBARS_NONE;
	if ( tp.hScroll == null )
		return SCROLLBARS_VERTICAL_ONLY;
	if ( tp.vScroll == null )
		return SCROLLBARS_HORIZONTAL_ONLY;

	return SCROLLBARS_BOTH;
}

public String getSelectedText() {
	StringBuffer sb = new StringBuffer();
	
	int y0 = Math.min( tp.sSel.y, tp.eSel.y);
	int yMax = Math.max( tp.sSel.y, tp.eSel.y);
	
	for ( int i=y0; i<=yMax; i++) {
		TextBuffer tb = (TextBuffer) tp.rows.elementAt( i);
		int x0 = tp.selXStart( i);
		int x1 = tp.selXEnd( i, tb);
		sb.append( tb.buf, x0, x1-x0 );
		if ( i < yMax )
			sb.append( " ");
	}
	
	return sb.toString();
}

public int getSelectionEnd() {
	int i1 = tp.get1D( tp.sSel);
	int i2 = tp.get1D( tp.eSel);
	
	return Math.max( i1, i2);
}

public int getSelectionStart() {
	int i1 = tp.get1D( tp.sSel);
	int i2 = tp.get1D( tp.eSel);
	
	return Math.min( i1, i2);
}

public String getText() {
	int i, imax = tp.rows.size()-1;
	StringBuffer sb = new StringBuffer( (imax+1) * 80);
	
	for ( i = 0; i <= imax; ) {
		TextBuffer tb = (TextBuffer) tp.rows.elementAt(i);
		sb.append( tb.buf, 0, tb.len);
		if ( i++ < imax )
			sb.append( '\n');
	}

	return sb.toString();
}

void hPosChange() {
	tp.hPosChange();
}

public synchronized void insert( String str, int pos) {
	tp.insertText( str, pos);
}

/**
 * @deprecated
 */
public Dimension minimumSize() {
	return (minimumSize(crows, ccols));
}

/**
 * @deprecated
 */
public Dimension minimumSize( int rows, int cols) {
	return new Dimension( cols*tp.rowHeight, rows*tp.rowHeight );
}

public void paint ( Graphics g ) {
	// we know about our childs, we don't have to blank the background,
	// so let's speed up things a little
	g.paintChild( tp, false);
	
	if ( (tp.hScroll != null) && ((tp.hScroll.flags & IS_VISIBLE) != 0) )
		g.paintChild( tp.hScroll, true);
	if ( (tp.vScroll != null) && ((tp.vScroll.flags & IS_VISIBLE) != 0) )
		g.paintChild( tp.vScroll, true);
		
	WidgetTheme.borderPaintOuter( this, g, (AWTEvent.keyTgt == tp) );
}

protected String paramString() {
	return super.paramString();
}

/**
 * @deprecated
 */
public Dimension preferredSize() {
	return preferredSize( crows, ccols);
}

/**
 * @deprecated
 */
public Dimension preferredSize( int rows, int cols) {
	return minimumSize( rows, cols);
}

protected void processKeyEvent( KeyEvent event ){
	// notify our listeners
	super.processKeyEvent( event);
	
	// and pass down (in case it wasn't redirected from there)
	if ( !event.isConsumed() ){
		tp.processKeyEvent( event);
	}
}

protected void processMouseEvent( MouseEvent event ){
	// notify our listeners
	super.processMouseEvent( event);
	
	// and pass down (in case it wasn't redirected from there)
	tp.processMouseEvent( event);
}

public void repaintRow( int row) {
	tp.repaintLine( row, 0, null);
}

public synchronized void replaceRange( String str, int start, int end) {
	replaceText(str, start, end);
}

void replaceSelectionWith ( String s ) {
	tp.replaceSelectionWith( s);
}

/**
 * @deprecated
 */
public synchronized void replaceText( String str, int start, int end) {
	tp.replaceRange( str, start, end);
}

public void requestFocus() {
	tp.requestFocus();
}

public void reshape ( int x, int y, int w, int h ) {
	super.reshape( x, y, w, h);
	
	// there is no need for validation of compound IS_NATIVE_LIKES, they are no Containers
	// in JDK, so we automagically have to re-layout them
	tp.innerLayout();
	// flags |= IS_VALID;
	validate();
}

public void select( int start, int end) {
	Point p = tp.get2D( end);
	
	tp.sSel = tp.get2D( start);
	tp.updateSel( p.x, p.y, true);
}

public void selectAll() {
	TextBuffer tb = (TextBuffer) tp.rows.lastElement();
	tp.sSel.x = 0;
	tp.sSel.y = 0;
	tp.eSel.x = 0;
	tp.eSel.y = 0;
	tp.updateSel( tb.len, tp.rows.size()-1, true);
}

public void setBackground( Color clr) {
	// we don't have a color of our own, forward this to tp
	tp.setBackground( clr);
}

public void setCaretPosition( int pos) {
	Point p = tp.get2D( pos);
	tp.setCaretPos( p.x, p.y, true, true);
}

public void setColumns( int cols) {
	ccols = cols;
}

public void setEnabled ( boolean isEnabled ) {
	super.setEnabled( isEnabled);

	tp.setEnabled( isEnabled);
}

public void setFont( Font f) {
	super.setFont( f);
	tp.setFont( f);
}

public void setForeground( Color clr) {
	// we don't have a color of our own, forward thisto tp
	tp.setForeground( clr);
}

public void setRows( int rows) {
}

public void setSelectionEnd( int end) {
	Point p = tp.get2D( end);
	tp.updateSel( p.x, p.y, true);
}

public void setSelectionStart( int start) {
	Point p = tp.get2D( start);
	tp.updateSel( p.x, p.y, true);
}

public void setText( String text) {
	tp.setContents( text);
}

public void update ( Graphics g ) {
	paint( g); // no background blanking required
}

void vPosChange( int dy) {
	tp.vPosChange( dy);
}
}
