package java.awt;


public class TextBuffer
{
	char[] buf;
	int len;
	static int initLen = 128;

public TextBuffer() {
	this( initLen);
}

public TextBuffer( String str) {
	this( initLen);
	append( str);
}

public TextBuffer( int length) {
	buf = new char[length];
}

public void append( String s) {
	insert( len, s);
}

public void append( TextBuffer tb) {
	checkCapacity( len + tb.len);
	for ( int i=0; i<tb.len; i++)
		buf[len++] = tb.buf[i];
}

public void append( char c) {
	insert( len, c);
}

void checkCapacity( int newLen) {
	if ( newLen <= buf.length)
		return;
	int nl = buf.length+initLen;
	if ( newLen > nl)
		nl = newLen;
	char[] pb = buf;
	buf = new char[nl];
	System.arraycopy( pb, 0, buf, 0, pb.length);
}

public void copyLevelFrom( TextBuffer tb){
	int orgLevel = tb.getLevel();
	int level = getLevel();
	
	for ( int i = 0; i<orgLevel - level; i++)
		insert( 0, '\t');
}

public int getLevel(){
	int level = 0;
	
	for ( int i=0; i<len; i++) {
		if ( buf[i] != '\t' )
			break;
		level++;
	}

	return level;	
}

public String getString( int start, int len){
	if ( this.len < start + len)
		len = this.len - start;
	if ( len > 0)
		return new String( buf, start, len);
	return null;
}

public void insert( int idx, String s) {
	replace( idx, 0, s);
}

public void insert( int idx, char c) {
	replace( idx, 0, c);
}

public void remove( int start, int len) {
	if ( (start > this.len) || (len <= 0) )
		return;
	int ll = this.len - start - len;
	if ( ll > 0 )
		System.arraycopy( buf, start+len, buf, start, ll);
	this.len -= len; 
}

public void replace( int start, int len, String s) {
	int sLen = ( s != null) ? s.length() : 0;
	int dl = sLen - len;

	checkCapacity( this.len+dl);
	int eIdx = start+len;
	System.arraycopy( buf, eIdx, buf, eIdx+dl, this.len-eIdx);
	
	if ( sLen > 0 ){
		s.getChars(0, sLen, buf, start);
	}
	this.len += dl; 
}

public void replace( int start, int len, char c) {
	int dl = 1 - len;

	checkCapacity( this.len+dl);
	int eIdx = start+len;
	System.arraycopy( buf, eIdx, buf, eIdx+dl, this.len-eIdx);
	
	buf[start] = c;	
	this.len += dl; 
}

public void set( String s) {
	if ( (s == null) || (s.length() == 0) ){
		len = 0;
	}
	else {
		int sLen = s.length();
		checkCapacity( sLen);
		s.getChars( 0, sLen, buf, 0);
		len = sLen;
	}
}

public void set( char c, int count) {
	checkCapacity( count);
	
	for ( len = 0; len < count; len++)
		buf[len] = c;
}

public String toString() {
	return new String( buf, 0, len);
}
}
