package java.awt;

import kaffe.awt.Defaults;
import kaffe.awt.RenderDirection;
import kaffe.awt.WidgetTheme;

class TextBufferRender
{
	FontMetrics fm;
	RenderModel model;
	int tabWidth = 20;
	int offset = 0;
	Component user;
	static int decoOffset = 4;
	int width;
	int height;

abstract class RenderModel
{
abstract boolean atEnd(TextBuffer tb, int idx);

abstract boolean atStart(TextBuffer tb, int idx);

abstract int getCursorShape();

abstract Rectangle getDrawRect( TextBuffer tb, int x, Rectangle r);

abstract int indexFromPosition ( TextBuffer tb, int xPos, int yPos );

abstract int next( int steps);

abstract void paint ( Graphics g, TextBuffer tb, int start, int len, Rectangle r, boolean rawText, boolean selected );

abstract void paintCaret ( Graphics g, TextCursor tc );

abstract int positionFromIndex ( TextBuffer tb, int index );

abstract int shiftsToShow ( TextBuffer tb, int index );

int widthTo ( TextBuffer tb, int index ) {
	int dx = 0;
	char c;
	
	for ( int i=0; i<index; i++) {
		c = tb.buf[i];
		if ( c == '\t' )
			dx = (dx + tabWidth) / tabWidth * tabWidth;
		else
			dx += fm.charWidth( c);
	}
	
	return dx;
}
}
class RenderLR
  extends RenderModel
{
boolean atEnd(TextBuffer tb, int idx) {
	return (idx == tb.len);
}

boolean atStart(TextBuffer tb, int idx) {
	return (idx == 0);
}

int getCursorShape() {
	return Cursor.TEXT_CURSOR;
}

Rectangle getDrawRect( TextBuffer tb, int x, Rectangle r) {
	int b1 = positionFromIndex( tb, x);
	
	if ( x == 0 )
		b1 -= user.deco.x + decoOffset;
		
	r.x = b1;
	r.width = width-b1;
	
	return r;
}

int indexFromPosition ( TextBuffer tb, int xPos, int yPos ) {
	int x0 = user.deco.x + offset + decoOffset;
	int idx, x=0;
	char c;
	
	for ( idx=0; idx<tb.len; idx++) {
		c = tb.buf[idx];
		if ( c == '\t' )
			x = (x + tabWidth) / tabWidth * tabWidth;
		else
			x += fm.charWidth( c);
		if ( x0+x >= xPos )
			break;
	}
	
	return (idx);
}

int next( int steps) {
	return steps;
}

void paint ( Graphics g, TextBuffer tb, int start, int len, 
			Rectangle r, boolean rawText, boolean selected ) {
	int fh = fm.getHeight();
	int x0, y0;
	
	if ( r != null )
		y0 = r.y + r.height - (height - fh) / 2;
	else
		y0 = height - (height -fh) / 2;

	x0 = positionFromIndex( tb, start);
	
	if ( ! rawText ) {
		int x1 = positionFromIndex( tb, start+len);
		// WidgetTheme.tbrPaintBackground(user, g, x0, y0-fh, x1-x0, fh, selected);
		user.paintBackground(g, x0, y0-fh, x1-x0, fh, (selected ? WidgetTheme.IS_ACTIVE : 0));
	}

	g.setColor(user.getForegroundSelected(selected));

	for ( int i=0; i<len; i++) { 
		if ( tb.buf[start+i] == '\t' ){
			g.drawChars( tb.buf, start, i, x0, y0-fm.getDescent());
			start += i+1;
			len -= i+1;
			i = -1;
			x0 = positionFromIndex( tb, start);
		}
	}

	g.drawChars( tb.buf, start, len, x0, y0-fm.getDescent());
}

void paintCaret ( Graphics g, TextCursor tc ) {
	int x0 = tc.position.x;
	int y0 = tc.position.y + tc.yOffset;
	
	g.fillRect( x0, y0+1, tc.width, tc.height);
}

int positionFromIndex ( TextBuffer tb, int index ) {
	return (user.deco.x + offset + decoOffset + this.widthTo( tb, index));
}

int shiftsToShow ( TextBuffer tb, int index ) {
	int shifts = 0;
	int bl = user.deco.x + decoOffset;
	int x0 = positionFromIndex( tb, index);

	if ( x0 < bl )
		shifts = x0-bl;
	else {
		x0 = width - user.deco.width - decoOffset - x0;
		if ( x0 <= 0 )
			shifts = -x0;
	}

	return shifts;
}
}
class RenderRL
  extends RenderModel
{
boolean atEnd(TextBuffer tb, int idx) {
	return (idx == 0);
}

boolean atStart(TextBuffer tb, int idx) {
	return (idx == tb.len);
}

int getCursorShape() {
	return Cursor.TEXT_CURSOR;
}

Rectangle getDrawRect( TextBuffer tb, int x, Rectangle r) {
	int br =  user.deco.width - user.deco.x;
	int b1 = positionFromIndex( tb, x);
	
	if ( x == 0 )
		b1 += br + decoOffset;
		
	r.x = br;
	r.width = b1;
	
	return r;
}

int indexFromPosition ( TextBuffer tb, int xPos, int yPos ) {
	int x0 = width - decoOffset - user.deco.width - user.deco.x - offset;
	int idx, x=0;
	char c;
	
	for ( idx=0; idx<tb.len; idx++) {
		c = tb.buf[idx];
		if ( c == '\t' )
			x = (x + tabWidth) / tabWidth * tabWidth;
		else
			x += fm.charWidth( c);
		if ( x0-x < xPos )
			break;
	}
	
	return (idx);
}

int next( int steps) {
	return -steps;
}

void paint ( Graphics g, TextBuffer tb, int start, int len,
			Rectangle r, boolean rawText, boolean selected ) {
	int cw;
	int fh = fm.getHeight();
	int x0, y0;
	
	if ( r != null )
		y0 = r.y + r.height - (height - fh) / 2;
	else
		y0 = height - (height -fh) / 2;

	x0 = positionFromIndex( tb, start);

	if ( ! rawText ) {
		int x1 = positionFromIndex( tb, start+len);
		// WidgetTheme.tbrPaintBackground(user, g, x1, y0-fh, x0-x1, fh, selected);
		user.paintBackground(g, x1, y0-fh, x0-x1, fh, (selected ? WidgetTheme.IS_ACTIVE : 0));
	}
	
	g.setColor(user.getForegroundSelected(selected));

	cw = 0;
	for ( int i=0; i<len; i++) {
		char c = tb.buf[start+i];
		if ( c == '\t' ){
			x0 = positionFromIndex( tb, start+i+1);
			cw = 0;
		}
		else {
			cw += fm.charWidth( c );
			g.drawChars( tb.buf, start+i, 1, x0-cw, y0-fm.getDescent());
		}
	}
}

void paintCaret ( Graphics g, TextCursor tc ) {
	int x0 = tc.position.x - tc.width;
	int y0 = tc.position.y + tc.yOffset;
	
	g.fillRect( x0, y0+1, tc.width, tc.height);
}

int positionFromIndex ( TextBuffer tb, int index ) {
	return ( width - decoOffset - user.deco.width - user.deco.x - offset - this.widthTo( tb, index));
}

int shiftsToShow ( TextBuffer tb, int index ) {
	int shifts = 0;
	int br = user.deco.width - user.deco.x + decoOffset;
	int x0 = positionFromIndex( tb, index);

	if ( x0 > width - br )
		shifts = (width - br) - x0;
	else if ( x0 <= br )
		shifts = br-x0;

	return shifts;
}
}
class RenderTB
  extends RenderModel
{
boolean atEnd(TextBuffer tb, int idx) {
	return (idx == tb.len);
}

boolean atStart(TextBuffer tb, int idx) {
	return (idx == 0);
}

int getCursorShape() {
	return Cursor.DEFAULT_CURSOR;
}

Rectangle getDrawRect( TextBuffer tb, int x, Rectangle r) {
	int b1 = positionFromIndex( tb, x);
	
	if ( x == 0 )
		b1 -= user.deco.y + decoOffset;
		
	r.y = b1;
	r.height = height-b1;
	
	return r;
}

int indexFromPosition ( TextBuffer tb, int xPos, int yPos ) {
	int y0 = user.deco.y + offset;
	int fh = fm.getHeight();
	int idx, y=0;
	char c;
	
	for ( idx=0; idx<tb.len; idx++) {
		c = tb.buf[idx];
		if ( c == '\t' )
			y = (y + tabWidth) / tabWidth * tabWidth;
		else
			y += fh;
		if ( y0+y > yPos )
			break;
	}
	
	return (idx);
}

int next( int steps) {
	return steps;
}

void paint ( Graphics g, TextBuffer tb, int start, int len,
			Rectangle r, boolean rawText, boolean selected ) {
	int cw, ch, fh = fm.getHeight();
	int x0, y0;
	int dx = (r != null ) ? r.x : 0;
	
	cw = fm.charWidth('X');
	x0 = dx + (width - cw) / 2;
	y0 = positionFromIndex( tb, start);

	if ( ! rawText ) {
		int y1 = positionFromIndex( tb, start+len);
		
		if ( selected ) {
			g.setColor( Defaults.TextFieldSelBgClr);
			g.fill3DRect( x0, y0, cw, y1-y0, true);
		}
		else {
			g.setColor( (user != null ) ? user.bgClr : Defaults.TextFieldBgClr);
			g.fillRect( x0, y0, cw, y1-y0);
		}
	}

	if ( selected )
		g.setColor( Defaults.TextFieldSelTxtClr);
	else
		g.setColor( (user != null) ? user.fgClr : Defaults.TextFieldTxtClr);

	for ( int i=0; i<len; i++) {
		char c = tb.buf[start+i];
		x0 = dx + (width - fm.charWidth( c)) / 2;
		if ( c == '\t' ){
			y0 = positionFromIndex( tb, start+i+1);
		}
		else {
			y0 += fh;
			g.drawChars( tb.buf, start+i, 1, x0, y0-fm.getDescent());
		}
	}
}

void paintCaret ( Graphics g, TextCursor tc ) {
	int cw = fm.charWidth( 'X');
	int yd = (width - cw) / 2;
	
	g.fillRect( tc.position.y + yd, tc.position.x, cw, tc.width);
}

int positionFromIndex ( TextBuffer tb, int index ) {
	return ( user.deco.y + offset + decoOffset + this.widthTo( tb, index));
}

int shiftsToShow ( TextBuffer tb, int index ) {
	int shifts = 0;
	int bt = user.deco.y + decoOffset;
	int bb = user.deco.height - user.deco.y + decoOffset;
	int y0 = positionFromIndex( tb, index);

	if ( y0 > height - bb )
		shifts = y0 - (height - bb);
	else if ( y0 <= bt )
		shifts = y0 - bt;

	return shifts;
}

int widthTo ( TextBuffer tb, int index ) {
	int dy = 0;
	int ch = fm.getHeight();
	char c;
	
	for ( int i=0; i<index; i++) {
		c = tb.buf[i];
		if ( c == '\t' )
			dy = (dy + tabWidth) / tabWidth * tabWidth;
		else
			dy += ch;
	}
	
	return dy;
}
}

public TextBufferRender ( Component user) {
	setMode( RenderDirection.getDirection() );
	this.user = user;
}

boolean atEnd( TextBuffer tb, int idx) {
	return model.atEnd( tb, idx);
}

boolean atStart(TextBuffer tb, int idx) {
	return model.atStart(tb, idx);
}

void blankCaret ( Graphics g, TextCursor tc ) {
	model.paintCaret( g, tc);
}

int getCursorShape() {
	return model.getCursorShape();
}

Rectangle getDrawRect( TextBuffer tb, int x, Rectangle r) {
	return model.getDrawRect( tb, x, r);
}

int indexFromPosition ( TextBuffer tb, int xPos, int yPos ) {
	return model.indexFromPosition( tb, xPos, yPos);
}

boolean isHorizontal() {
	return ( !(model instanceof RenderTB));
}

int next( int steps) {
	return model.next( steps);
}

void paint ( Graphics g, TextBuffer tb, int start, int len,
			Rectangle r, boolean rawText, boolean selected ) {
	if ( len == 0 )
		return;
		
	if ( len == -1 )
		len = tb.len - start;

	model.paint( g, tb, start, len, r, rawText, selected );
}

void paint ( Graphics g, TextBuffer tb, int start, int len, boolean selected ) {
	paint( g, tb, start, len, null, false, selected);
}

void paintCaret ( Graphics g, TextCursor tc ) {
	Color c = g.getColor();
	g.setColor( tc.clr);
	model.paintCaret( g, tc);
	g.setColor( c);
}

void paintRawText ( Graphics g, TextBuffer tb, Rectangle r) {
	model.paint( g, tb, 0, tb.len, r, true, false );
}

int positionFromIndex ( TextBuffer tb, int index ) {
	return model.positionFromIndex( tb, index);
}

void setMode( int mode ) {
	switch ( mode ) {
	case RenderDirection.LR:
		model = new RenderLR();
		break;
	case RenderDirection.RL:
		model = new RenderRL();
		break;
	case RenderDirection.TB:
		model = new RenderTB();
		break;
	default:
		break;
	}
}

int shiftsToShow ( TextBuffer tb, int index ) {
	return model.shiftsToShow( tb, index);
}

int widthTo( TextBuffer tb, int index) {
	return model.widthTo( tb, index);
}
}
