package java.awt;

import java.awt.event.*;
import kaffe.awt.Defaults;
import kaffe.awt.WidgetTheme;

public class TextField
  extends TextComponent
  implements KeyListener, MouseListener, MouseMotionListener, FocusListener, ActionListener
{
	final private static long serialVersionUID = -2966288784432217853L;
	TextBuffer buffer;
	NativeGraphics rgr;
	int selectionExtend;
	TextCursor tc = new TextCursor( 2, Defaults.TextCursorClr);
	TextBufferRender tr;
	int cols;
	char echoChar;
	static TextBuffer echoBuffer;
	ActionListener aListener;
	static int named;

public TextField() {
	this( "", 0);
}

public TextField( String text) {
	this( text, (text != null) ? text.length() : 0);
}

public TextField( String text, int cols) {
	this.cols = cols;
	
	buffer = new TextBuffer( text);
	
	//use foreground & background color settings for rendering
	tr = new TextBufferRender( this);
	cursor = Cursor.getPredefinedCursor( tr.getCursorShape() );
	
	//border decoration insets
	deco = WidgetTheme.tfGetDeco( this);
	
	WidgetTheme.tfSetDefaults(this);
	
	addKeyListener( this);
	addMouseListener( this);
	addFocusListener( this);
	addMouseMotionListener( this);
}

public TextField(int cols) {
	this( "", cols);
}

public void addActionListener( ActionListener l) {
	aListener = AWTEventMulticaster.add( aListener, l);
}

public void addNotify() {
	super.addNotify();
	setResidentGraphics();
}

void clearSelection() {
	if ( selectionExtend != 0 ) {
		selectionExtend = 0;
		drawContents();
	}
}

public void deleteSelection( ) {
	int s = getSelectionStart();
	buffer.remove( s, Math.abs( selectionExtend));
	selectionExtend = 0;
	setCaretPosition( s);
}

TextBuffer displayBuffer() {
	if ( echoChar != 0 ) {
		echoBuffer.set( echoChar, buffer.len);
		return echoBuffer;
	}
	return buffer;
}

void drawContents() {	
	if ( rgr == null )
		return;
		
	TextBuffer b = displayBuffer();
	
	if ( selectionExtend == 0 ) {
		tr.paint ( rgr, b, 0, b.len, false);
	}
	else {
		int s1,s2;
		if ( selectionExtend > 0 ) {
			s1 = tc.index.x;
			s2 = tc.index.x + selectionExtend;
		}
		else {
			s1 = tc.index.x + selectionExtend;
			s2 = tc.index.x;
		}

		tr.paint ( rgr, b, 0, s1, false);
		tr.paint ( rgr, b, s1, s2-s1, true);
		tr.paint ( rgr, b, s2, buffer.len-s2, false);
	}
	
	if ( AWTEvent.keyTgt == this)
		tr.paintCaret( rgr, tc);
	else {
//		rgr.setColor( Defaults.TextCursorInactiveClr );
//		tr.blankCaret( rgr, tc);
	}
}

public boolean echoCharIsSet() {
	return ( echoChar != 0);
}

public void focusGained( FocusEvent e) {
	repaint();
}

public void focusLost( FocusEvent e) {
	repaint();
}

public int getCaretPosition() {
	return tc.index.x;
}

public int getColumns() {
	return cols;
}

public char getEchoChar() {
	return echoChar;
}

Color getForegroundSelected(boolean sel) {
	return (WidgetTheme.tfGetForegroundColorSelected(this, sel));
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "textfield" + (named++);
}

public Dimension getMinimumSize(int cols) {
	return minimumSize(cols);
}

public Dimension getPreferredSize(int cols) {
	return preferredSize(cols);
}

public String getSelectedText() {
	int i1 = selectionExtend > 0 ? tc.index.x : tc.index.x + selectionExtend;
	return buffer.getString( i1, Math.abs( selectionExtend));
}

public int getSelectionEnd() {
	return ( selectionExtend < 0 ? tc.index.x : tc.index.x + selectionExtend );
}

public int getSelectionStart() {
	return ( selectionExtend > 0 ? tc.index.x : tc.index.x + selectionExtend );
}

public String getText() {
	return buffer.toString();
}

int indexFromPosition( int xPos, int yPos) {
	return tr.indexFromPosition( displayBuffer(), xPos, yPos);
}

public void keyPressed( KeyEvent e) {
	boolean shift = e.isShiftDown();
	
	switch ( e.getKeyCode() ) {
	case KeyEvent.VK_LEFT:
	case KeyEvent.VK_UP:
		if ( tr.atStart( buffer, tc.index.x) )
			return;
		selectionExtend -= shift ? tr.next(-1) : selectionExtend;
		setCaretPosition( tc.index.x + tr.next(-1));
		break;
	case KeyEvent.VK_RIGHT:
	case KeyEvent.VK_DOWN:
		if ( tr.atEnd( buffer, tc.index.x) )
			return;
		selectionExtend -= shift ? tr.next(1) : selectionExtend;
		setCaretPosition( tc.index.x + tr.next(1));
		break;
	case KeyEvent.VK_HOME:
		selectionExtend -= shift ? -tc.index.x : selectionExtend;
		setCaretPosition( 0);
		break;
	case KeyEvent.VK_END:
		selectionExtend -= shift ? buffer.len - tc.index.x : selectionExtend;
		setCaretPosition( buffer.len);
		break;
	case KeyEvent.VK_DELETE:
		if ( selectionExtend != 0 )
			deleteSelection();
		else if ( tc.index.x < buffer.len ) {
			buffer.remove( tc.index.x, 1);
			setCaretPosition( tc.index.x);
		}
		break;
	case KeyEvent.VK_BACK_SPACE:
		if ( selectionExtend != 0 )
			deleteSelection();
		else if ( tc.index.x > 0 ) {
			buffer.remove( tc.index.x-1, 1);
			setCaretPosition( tc.index.x-1);
		}
		break;
	case KeyEvent.VK_ENTER:
		notifyAction();
		break;
	}
}

public void keyReleased( KeyEvent e) {
}

public void keyTyped( KeyEvent e) {
	char c = e.getKeyChar();

	if ( ! Character.isISOControl( c) )
		replaceSelectionWith( c );

	if ( (textListener != null) || (eventMask & AWTEvent.TEXT_EVENT_MASK) != 0 ) {
		Toolkit.eventQueue.postEvent( TextEvt.getEvent( this, TextEvt.TEXT_VALUE_CHANGED));
	}
	
	e.consumed = true;
}

public Dimension minimumSize() {
	return minimumSize(cols);
}

public Dimension minimumSize(int cols) {
	int	nc = cols;
	int	w = 0;

	if ( nc <= 0 ) {
		if (buffer.len > 0) {
			w = tr.widthTo(buffer, buffer.len);
		}
		else {
			nc = 5;
		}
	}
	
	Dimension d;
	if (!tr.isHorizontal()) {
		if ( w == 0 ) {
			w = nc * tr.fm.getHeight();
		}
		d = new Dimension(tr.fm.charWidth('X'), w+deco.width);
	}
	else {
		if ( w == 0 ) {
			w = nc * tr.fm.charWidth('X');
		}
		d =  new Dimension(w+deco.width, tr.fm.getHeight()+deco.height);
	}

	return WidgetTheme.tfMinimumSize(this, d);
}

public void mouseClicked( MouseEvent e) {
}

public void mouseDragged( MouseEvent e) {
	int curIdx = indexFromPosition( e.getX(), e.getY() );
	if ( curIdx != tc.index.x ) {
		selectionExtend -= (curIdx-tc.index.x);
		setCaretPosition( curIdx);
	}
}

public void mouseEntered( MouseEvent e) {
}

public void mouseExited( MouseEvent e) {
}

public void mouseMoved( MouseEvent e) {
}

public void mousePressed( MouseEvent e) {
	requestFocus();
	selectionExtend = 0;
	setCaretPosition( indexFromPosition( e.getX(), e.getY() ));
}

public void mouseReleased( MouseEvent e) {
}

void notifyAction(){
	if ( (aListener != null) || ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) ){
		Toolkit.eventQueue.postEvent( ActionEvt.getEvent( this, ActionEvent.ACTION_PERFORMED,
		                                                  getText(), 0));
	}
}

public void paint( Graphics g) {
	paintBackground(g, 0, 0, width, height, 0);
	WidgetTheme.wdgPaintBorder ( this, g, deco, width, height, (AWTEvent.keyTgt == this) );
	drawContents();
}

void paintBackground (Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.tfPaintBackground(this, g, x, y, width, height, (AWTEvent.keyTgt == this ? WidgetTheme.IS_ACTIVE : 0));
}

int positionFromIndex( int index) {
	return tr.positionFromIndex( displayBuffer(), index);
}

public Dimension preferredSize() {
	return preferredSize(cols);
}

public Dimension preferredSize(int cols) {
	return getMinimumSize( cols);
}

void process ( ActionEvent e ) {
	if ( (aListener != null) || ((eventMask & (AWTEvent.ACTION_EVENT_MASK|AWTEvent.DISABLED_MASK)) == AWTEvent.ACTION_EVENT_MASK) ){
		processEvent( e);
	}

	if ( (flags & IS_OLD_EVENT) > 0 ) {
		postEvent( Event.getEvent( e));
	}
}

protected void processActionEvent ( ActionEvent e) {
	if ( aListener != null )
		aListener.actionPerformed( e);
}

void propagateFont( Font fnt) {
	super.propagateFont( fnt);
	
	tr.fm = getFontMetrics( fnt);
	if ( tr.fm != null ) {
		int ch = tr.fm.getHeight();
		tc.setHeight( ch, (height-ch)/2);
	}
}

void propagateReshape() {
	if ( rgr != null ){
		updateLinkedGraphics();

		// we have our own clipping, which we have to take care of
		rgr.setClip( deco.x+1, deco.y+1, width-deco.width-2, height-deco.height-2);
	}

	int ch = tr.fm.getHeight();
	tc.setHeight( ch, (height-ch)/2);
	
	tr.width = width;
	tr.height = height;
}

public void removeActionListener( ActionListener l) {
	aListener = AWTEventMulticaster.remove( aListener, l);
}

void replaceSelectionWith( char c) {
	int s = getSelectionStart();
	buffer.replace( s, Math.abs( selectionExtend), c);
	selectionExtend = 0;
	setCaretPosition( s+1);
}

public void select( int start, int end) {
	selectionExtend = end-start;
	setCaretPosition( start);
}

public void selectAll() {
	select( 0, buffer.len);
}

public void setCaretPosition( int index) {
	if ( index < 0 )
		index = 0;
	else if ( index > buffer.len )
		index = buffer.len;
		
	shiftToShow( index);
	tc.setX( index, positionFromIndex( index));

	if (rgr != null) {
		paintBackground(rgr, 0, 0, width, height, 0);
	}
	drawContents();
}

public void setColumns(int cols) {
	this.cols = cols;
}

public void setEchoChar( char echoChar) {
	this.echoChar = echoChar;
	if ( (echoChar != 0) && (echoBuffer == null) )
		echoBuffer = new TextBuffer();
	setCaretPosition( tc.index.x);
}

public void setEchoCharacter( char echoChar) {
	setEchoChar( echoChar);
}

void setResidentGraphics() {
	if ( rgr != null )
		return;
		
	int d = Defaults.WidgetBorderWidth;
	rgr = NativeGraphics.getClippedGraphics( rgr, this, 0, 0,
					                         d, d, width-2*d, height-2*d);
	if ( rgr != null )
		linkGraphics( rgr);
}

public void setSelectionEnd( int index) {
	select( tc.index.x, index);
}

public void setSelectionStart( int index) {
	select( index, tc.index.x );
}

public void setText( String text) {
	buffer.set( text);
	selectionExtend = 0;
	setCaretPosition( 0);
}

int shiftToShow( int index) {
	if ( (width == 0) || (height == 0) )
		return 0;
		
	TextBuffer b = displayBuffer();
	int sts = tr.shiftsToShow( b, index);
	tr.offset -= sts;

	return sts;
}
}
