/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.io;

import java.lang.String;
import kaffe.io.FileIO;

public class FileInputStream
	extends InputStream {

private FileIO fd;

public FileInputStream(File file) throws FileNotFoundException {
	this(file.getPath());
}

public FileInputStream(FileDescriptor fdObj) {
	System.getSecurityManager().checkRead(fdObj);
	fd = (FileIO)fdObj;
}

public FileInputStream(String name) throws FileNotFoundException {
	fd = new FileIO(name, FileIO.READ);
}

public int available() throws IOException {
	return (fd.available());
}

public void close() throws IOException {
	if (fd != null) {
		fd.close();
	}
}

protected void finalize() throws IOException {
	close();
	try {
		super.finalize();
	}
	catch(Throwable e){
		throw new IOException(e.getMessage());
	}
}

final public FileDescriptor getFD() throws IOException {
	return (fd);
}

public int read() throws IOException {
	return (fd.read());
}

public int read(byte b[]) throws IOException {
	return (read(b, 0, b.length));
}

public int read(byte b[], int off, int len) throws IOException {
	return (fd.read(b, off, len));
}

public long skip(long n) throws IOException {
	return (fd.skip(n));
}

}
