/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.io;

import java.lang.String;
import kaffe.io.FileIO;

public class FileOutputStream
	extends OutputStream {

private FileIO fd;

public FileOutputStream(File file) throws IOException {
	this(file.getPath());
}

public FileOutputStream(FileDescriptor fdObj) {
	System.getSecurityManager().checkWrite(fdObj);
	fd = (FileIO)fdObj;
}

public FileOutputStream(String name) throws FileNotFoundException {
	this(name, false);
}

public FileOutputStream(String name, boolean append) throws FileNotFoundException {
	if (append == false) {
		fd = new FileIO(name, FileIO.WRITE);
	}
	else {
		fd = new FileIO(name, FileIO.WRITE|FileIO.APPEND);
	}
}

public void close() throws IOException {
	fd.close();
}

protected void finalize() throws IOException {
	close();
	try {
		super.finalize();
	}
	catch(Throwable e){
		throw new IOException(e.getMessage());
	}
}

final public FileDescriptor getFD()  throws IOException {
	return (fd);
}

public void write(byte b[]) throws IOException {
	write(b, 0, b.length);
}

public void write(byte b[], int off, int len) throws IOException {
	fd.writeFully(b, off, len);
}

public void write(int b) throws IOException {
	int l = fd.write(b);
	if (l != 1) {
		throw new IOException("write failed: expected length=1, actual="+l);
	}
}

}
