/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.io;

import java.lang.String;
import kaffe.io.ByteToCharFilter;

public class InputStreamReader
  extends Reader
{
	final private static int BUFDEFAULT = 128;
	private ByteToCharFilter encoding;

public InputStreamReader(InputStream in) {	
	if (in == null) {
		throw new NullPointerException();
	}
	encoding = ByteToCharFilter.getDefault(in);
}

public InputStreamReader(InputStream in, String enc) throws UnsupportedEncodingException {
	if (in == null) {
		throw new NullPointerException();
	}
	encoding = ByteToCharFilter.getConverter(enc, in);
}

public void close() throws IOException {
	if (encoding != null) {
		encoding.close();
		encoding = null;
	}
}

public String getEncoding() {
	return (encoding.toString());
}

public int read ( char cbuf[], int off, int len ) throws IOException {
	return (encoding.read(cbuf, off, len));
}

public boolean ready() throws IOException {
	return (encoding.available() > 0 ? true : false);
}
}
