package java.io;

import java.lang.String;
import kaffe.io.CharToByteFilter;

/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class OutputStreamWriter
  extends Writer
{
	private final static int BUFDEFAULT = 1024;
	private final static int MINMARGIN = 32;
	private CharToByteFilter encoding;
	private boolean closed = false;

public OutputStreamWriter(OutputStream out) {
	encoding = CharToByteFilter.getDefault(out);
}

public OutputStreamWriter(OutputStream out, String enc) throws UnsupportedEncodingException
{
	encoding = CharToByteFilter.getConverter(enc, out);
}

public void close() throws IOException
{
	if (closed == false) {
		flush();
		encoding.close();
		encoding = null;
		closed = true;
	}
}

public void flush() throws IOException
{
	synchronized (lock) {
		if (closed == true) {
			throw new IOException("stream closed");
		}
	}
	encoding.flush();
}

public String getEncoding() {
	return (encoding.toString());
}

public void write(String str, int off, int len) throws IOException
{
	write(str.toCharArray(), off, len);
}

public void write(char cbuf[], int off, int len) throws IOException
{
	if (encoding == null) {
		throw new IOException("stream closed");
	}
	encoding.write(cbuf, off, len);
}

public void write(int c) throws IOException
{
	write (new char[] { (char)c }, 0, 1);
}
}
