/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * $tvt: Runtime.java,v 1.19 2001/08/19 07:41:13 tim Exp $
 */

package java.lang;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.StringTokenizer;
import kaffe.lang.SystemClassLoader;

public class Runtime
{
/**
 *  This is not part of the public interface.
 */
public static interface MemoryAdvice {

// These should match those in kaffe.lang.MemoryAdvice
public static final int GREEN = 0;
public static final int YELLOW = 1;
public static final int ORANGE = 2;
public static final int RED = 3;

}

private static Runtime currentRuntime = new Runtime();
private static kaffe.lang.MemoryAdvice advice;

private Runtime () {
}

public Process exec(String command) throws IOException {
	return exec(command, null);
}

public Process exec(String command, String envp[]) throws IOException {
	StringTokenizer tokenizer=new StringTokenizer(command);

	int count = tokenizer.countTokens();
	String cmdarray[] = new String[count];

	for (int pos = 0; pos < count; pos++) {
		cmdarray[pos] = tokenizer.nextToken();
	}

	return exec(cmdarray, envp);
}

public Process exec(String cmdarray[]) throws IOException {
	return exec(cmdarray, null);
}

public Process exec(String cmdarray[], String envp[]) throws IOException {
	System.getSecurityManager().checkExec(cmdarray[0]);
	return execInternal(cmdarray, envp);
}

native private Process execInternal(String cmdarray[], String envp[]) throws IOException;

public void exit(int status) {
	System.getSecurityManager().checkExit(status);
	// Handle application extensions - if this thread is part of an
	// application then we exit that rather than the whole thing.
	if (kaffe.lang.Application.exit(status) == false) {
		exitInternal(status);
	}
	// kaffe.lang.Application.exit does not destroy the thread
	// that invoked exit().  We stop that thread now.
	Thread.currentThread().destroy();
}

native private void exitInternal(int status);

native public long freeMemory();

native public void gc();

/**
 * @deprecated
 */
public InputStream getLocalizedInputStream(InputStream in) {
	return (in);
}

/**
 * @deprecated
 */
public OutputStream getLocalizedOutputStream(OutputStream out) {
	return (out);
}

public static Runtime getRuntime() {
	return currentRuntime;		
}

public void load(String filename) {
	load(filename, SystemClassLoader.getCallerClassLoader(0));
}

public void loadLibrary(String libname) {
	loadLibrary(libname, SystemClassLoader.getCallerClassLoader(0));
}

void loadLibrary(String libname, ClassLoader loader) {
	String name = loader.findLibrary(libname);
	if (name == null) {
		name = SystemClassLoader.getClassLoader().findLibrary(libname);
	}
	load(name, loader);
}

synchronized void load(String filename, ClassLoader loader) {
	System.getSecurityManager().checkLink(filename);
	if (loader == SystemClassLoader.getClassLoader()) {
		loader = null;
	}
	loadFileInternal(filename, loader);
}

native private String loadFileInternal(String filename, ClassLoader loader) throws UnsatisfiedLinkError;

kaffe.lang.MemoryAdvice getAdvice() {
	if (advice == null) {
		advice = kaffe.lang.MemoryAdvice.getInstance();
	}
	return (advice);
}

int getMemoryAdvice() {
	return (getAdvice().getColor());
}

int waitForMemoryAdvice(int level) throws InterruptedException {
	return (getAdvice().waitForOtherColor(level));
}

native public void runFinalization();

native public static void runFinalizersOnExit(boolean value);

native public long totalMemory();

native public void traceInstructions(boolean on);

native public void traceMethodCalls(boolean on);
}
