/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.security.AccessController;
import java.util.Properties;
import kaffe.lang.SystemClassLoader;

final public class System
{
	static SecurityManager defaultSecurityManager;
	final private static RuntimePermission setio = new RuntimePermission("setIO");
	final public static InputStream in;
	final public static PrintStream out;
	final public static PrintStream err;
	private static Properties props;
	private static SecurityManager security;

// When trying to debug Java code that gets executed early on during
// JVM initialization, eg, before System.err is initialized, debugging
// println() statements don't work. In these cases, the following routines
// are very handy. Simply uncomment the following two lines to enable them.

/****
public static native void debug(String s);	// print s to stderr, then \n
public static native void debugE(Throwable t);	// print stack trace to stderr
****/
static {
	// Initialise the I/O
	in = new BufferedInputStream(new FileInputStream(FileDescriptor.in), 128);
	out = new PrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.out), 128), true);
	err = new PrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.err), 128), true);	
}

private System() {
}

public static void arraycopy(Object src, int src_position, Object dst, int dst_position, int length) {

	if (dst instanceof Object[]) {
		Object[] d = (Object[])dst;
		Object[] s = (Object[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof int[]) {
		int[] d = (int[])dst;
		int[] s = (int[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof float[]) {
		float[] d = (float[])dst;
		float[] s = (float[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof long[]) {
		long[] d = (long[])dst;
		long[] s = (long[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof double[]) {
		double[] d = (double[])dst;
		double[] s = (double[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof boolean[]) {
		boolean[] d = (boolean[])dst;
		boolean[] s = (boolean[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof byte[]) {
		byte[] d = (byte[])dst;
		byte[] s = (byte[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof char[]) {
		char[] d = (char[])dst;
		char[] s = (char[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else if (dst instanceof short[]) {
		short[] d = (short[])dst;
		short[] s = (short[])src;
		if (dst_position < src_position) {
			for (int i = 0; i < length; i++) {
				d[dst_position+i] = s[src_position+i];
			}
		}
		else {
			for (int i = length - 1; i >= 0; i--) {
				d[dst_position+i] = s[src_position+i];
			}
		}
	}
	else {
		throw new ArrayStoreException("not arrays");
	}
}

private static void checkPropertyAccess() {
	getSecurityManager().checkPropertiesAccess();
}

native public static long currentTimeMillis();

public static void exit (int status) {
	Runtime.getRuntime().exit(status);
}

public static void gc() {
	Runtime.getRuntime().gc();
}

public static Properties getProperties() {
	checkPropertyAccess();
	if (props == null) {
		props = initProperties(new Properties());
	}
	return (props);
}

public static String getProperty(String key) {
	return getProperties().getProperty(key);
}

public static String getProperty(String key, String def) {
	return getProperties().getProperty(key, def);
}

public static SecurityManager getSecurityManager() {
	if (security == null) {
		defaultSecurityManager = new NullSecurityManager();
		security = defaultSecurityManager;
	}
	return security;
}

public static void setSecurityManager(SecurityManager s) {
	if (security == null || security == defaultSecurityManager) {
		if (s != null) {
			security = s;
		}
	}
	else {
		throw new SecurityException();
	}
}

public static String getenv(String name) {
	throw new Error("System.getenv no longer supported");
}

native public static int identityHashCode(Object x);

native private static Properties initProperties(Properties props);

public static void load(String filename) {
	Runtime.getRuntime().load(filename, SystemClassLoader.getCallerClassLoader(0));
}

public static void loadLibrary(String libname) {
	Runtime.getRuntime().loadLibrary(libname, SystemClassLoader.getCallerClassLoader(0));
}

public static void runFinalization() {
	Runtime.getRuntime().runFinalization();
}

public static void runFinalizersOnExit(boolean value) {
	Runtime.getRuntime().runFinalizersOnExit(value);
}

public static void setErr(PrintStream err) {
	AccessController.checkPermission(setio);
	setErr0(err);
}

native private static void setErr0(PrintStream err);

public static void setIn(InputStream in) {
	AccessController.checkPermission(setio);
	setIn0(in);
}

native private static void setIn0(InputStream in);

public static void setOut(PrintStream out) {
	AccessController.checkPermission(setio);
	setOut0(out);
}

native private static void setOut0(PrintStream out);

public static void setProperties(Properties prps) {
	checkPropertyAccess();
	props = prps;
}

}
