/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.lang.reflect;

import java.lang.String;
import java.lang.Class;

public final class Field
	implements Member {

private Class clazz;
private int slot;
private String name;
private Class type;

private Field() {
}
    
public String getName() {
	return (name);
}

public Class getDeclaringClass() {
	return (clazz);
}

public Class getType() {
	return (type);
}

public int hashCode() {
	return (clazz.getName().hashCode() ^ name.hashCode());
}

public boolean equals(Object obj) {
	// Quick test for identity
	if ((Object)this == obj) {
		return (true);
	}

	if (!(obj instanceof Field)) {
		return (false);
	}

	Field fobj = (Field)obj;
	if (clazz == fobj.clazz && type == fobj.type && name.equals(fobj.name)) {
		return (true);
	}
	return (false);
}

public Object get(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Double.TYPE) {
		return (new Double(getDouble0(obj, clazz, slot)));
	}
	else if (type == Float.TYPE) {
		return (new Float(getFloat0(obj, clazz, slot)));
	}
	else if (type == Long.TYPE) {
		return (new Long(getLong0(obj, clazz, slot)));
	}
	else if (type == Integer.TYPE) {
		return (new Integer(getInt0(obj, clazz, slot)));
	}
	else if (type == Short.TYPE) {
		return (new Short(getShort0(obj, clazz, slot)));
	}
	else if (type == Byte.TYPE) {
		return (new Byte(getByte0(obj, clazz, slot)));
	}
	else if (type == Character.TYPE) {
		return (new Character(getChar0(obj, clazz, slot)));
	}
	else {
		return (getObject0(obj, clazz, slot));
	}
}

public boolean getBoolean(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Boolean.TYPE) {
		return (getBoolean0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public byte getByte(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Byte.TYPE) {
		return (getByte0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public char getChar(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Character.TYPE) {
		return (getChar0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public double getDouble(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Double.TYPE) {
		return (getDouble0(obj, clazz, slot));
	}
	else if (type == Float.TYPE) {
		return ((double)getFloat0(obj, clazz, slot));
	}
	else if (type == Long.TYPE) {
		return ((double)getLong0(obj, clazz, slot));
	}
	else if (type == Integer.TYPE) {
		return ((double)getInt0(obj, clazz, slot));
	}
	else if (type == Short.TYPE) {
		return ((double)getShort0(obj, clazz, slot));
	}
	else if (type == Byte.TYPE) {
		return ((double)getByte0(obj, clazz, slot));
	}
	else if (type == Character.TYPE) {
		return ((double)getChar0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public float getFloat(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Float.TYPE) {
		return (getFloat0(obj, clazz, slot));
	}
	else if (type == Long.TYPE) {
		return ((float)getLong0(obj, clazz, slot));
	}
	else if (type == Integer.TYPE) {
		return ((float)getInt0(obj, clazz, slot));
	}
	else if (type == Short.TYPE) {
		return ((float)getShort0(obj, clazz, slot));
	}
	else if (type == Byte.TYPE) {
		return ((float)getByte0(obj, clazz, slot));
	}
	else if (type == Character.TYPE) {
		return ((float)getChar0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public int getInt(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Integer.TYPE) {
		return (getInt0(obj, clazz, slot));
	}
	else if (type == Short.TYPE) {
		return ((int)getShort0(obj, clazz, slot));
	}
	else if (type == Byte.TYPE) {
		return ((int)getByte0(obj, clazz, slot));
	}
	else if (type == Character.TYPE) {
		return ((int)getChar0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public long getLong(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Long.TYPE) {
		return (getLong0(obj, clazz, slot));
	}
	else if (type == Integer.TYPE) {
		return ((long)getInt0(obj, clazz, slot));
	}
	else if (type == Short.TYPE) {
		return ((long)getShort0(obj, clazz, slot));
	}
	else if (type == Byte.TYPE) {
		return ((long)getByte0(obj, clazz, slot));
	}
	else if (type == Character.TYPE) {
		return ((long)getChar0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public short getShort(Object obj) throws IllegalArgumentException, IllegalAccessException {
	if (type == Short.TYPE) {
		return (getShort0(obj, clazz, slot));
	}
	else if (type == Integer.TYPE) {
		return ((short)getInt0(obj, clazz, slot));
	}
	else if (type == Long.TYPE) {
		return ((short)getLong0(obj, clazz, slot));
	}
	else if (type == Float.TYPE) {
		return ((short)getFloat0(obj, clazz, slot));
	}
	else if (type == Double.TYPE) {
		return ((short)getDouble0(obj, clazz, slot));
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void set(Object obj, Object value) throws IllegalArgumentException, IllegalAccessException {
	if (type == Boolean.TYPE) {
		setBoolean0(obj, ((Boolean)value).booleanValue(), clazz, slot);
	}
	else if (type == Byte.TYPE) {
		setByte0(obj, ((Byte)value).byteValue(), clazz, slot);
	}
	else if (type == Short.TYPE) {
		setShort0(obj, ((Short)value).shortValue(), clazz, slot);
	}
	else if (type == Character.TYPE) {
		setChar0(obj, ((Character)value).charValue(), clazz, slot);
	}
	else if (type == Integer.TYPE) {
		setInt0(obj, ((Integer)value).intValue(), clazz, slot);
	}
	else if (type == Long.TYPE) {
		setLong0(obj, ((Long)value).longValue(), clazz, slot);
	}
	else if (type == Float.TYPE) {
		setFloat0(obj, ((Float)value).floatValue(), clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, ((Double)value).doubleValue(), clazz, slot);
	}
	else {
		setObject0(obj, value, clazz, slot);
	}
}

public void setBoolean(Object obj, boolean z) throws IllegalArgumentException, IllegalAccessException {
	if (type == Boolean.TYPE) {
		setBoolean0(obj, z, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setByte(Object obj, byte b) throws IllegalArgumentException, IllegalAccessException {
	if (type == Byte.TYPE) {
		setByte0(obj, b, clazz, slot);
	}
	else if (type == Short.TYPE) {
		setShort0(obj, (short)b, clazz, slot);
	}
	else if (type == Integer.TYPE) {
		setInt0(obj, (int)b, clazz, slot);
	}
	else if (type == Long.TYPE) {
		setLong0(obj, (long)b, clazz, slot);
	}
	else if (type == Float.TYPE) {
		setFloat0(obj, (float)b, clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, (double)b, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setChar(Object obj, char c) throws IllegalArgumentException, IllegalAccessException {
	if (type == Character.TYPE) {
		setChar0(obj, c, clazz, slot);
	}
	else if (type == Integer.TYPE) {
		setInt0(obj, (int)c, clazz, slot);
	}
	else if (type == Long.TYPE) {
		setLong0(obj, (long)c, clazz, slot);
	}
	else if (type == Float.TYPE) {
		setFloat0(obj, (float)c, clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, (double)c, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setDouble(Object obj, double d) throws IllegalArgumentException, IllegalAccessException {
	if (type == Double.TYPE) {
		setDouble0(obj, d, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setFloat(Object obj, float f) throws IllegalArgumentException, IllegalAccessException {
	if (type == Float.TYPE) {
		setFloat0(obj, f, clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, (double)f, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setInt(Object obj, int i) throws IllegalArgumentException, IllegalAccessException {
	if (type == Integer.TYPE) {
		setInt0(obj, i, clazz, slot);
	}
	else if (type == Long.TYPE) {
		setLong0(obj, (long)i, clazz, slot);
	}
	else if (type == Float.TYPE) {
		setFloat0(obj, (float)i, clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, (double)i, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setLong(Object obj, long l) throws IllegalArgumentException, IllegalAccessException {
	if (type == Long.TYPE) {
		setLong0(obj, l, clazz, slot);
	}
	else if (type == Float.TYPE) {
		setFloat0(obj, (float)l, clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, (double)l, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public void setShort(Object obj, short s) throws IllegalArgumentException, IllegalAccessException {
	if (type == Short.TYPE) {
		setShort0(obj, s, clazz, slot);
	}
	else if (type == Integer.TYPE) {
		setInt0(obj, (int)s, clazz, slot);
	}
	else if (type == Long.TYPE) {
		setLong0(obj, (long)s, clazz, slot);
	}
	else if (type == Float.TYPE) {
		setFloat0(obj, (float)s, clazz, slot);
	}
	else if (type == Double.TYPE) {
		setDouble0(obj, (double)s, clazz, slot);
	}
	else {
		throw new IllegalArgumentException();
	}
}

public String toString() {
	StringBuffer str = new StringBuffer();
	int mod = getModifiers();

	if ((mod & Modifier.PUBLIC) != 0) {
		str.append("public ");
	}
	else if ((mod & Modifier.PRIVATE) != 0) {
		str.append("private ");
	}
	else if ((mod & Modifier.PROTECTED) != 0) {
		str.append("protected ");
	}

	if ((mod & Modifier.STATIC) != 0) {
		str.append("static ");
	}
	if ((mod & Modifier.FINAL) != 0) {
		str.append("final ");
	}
	if ((mod & Modifier.TRANSIENT) != 0) {
		str.append("transient ");
	}
	if ((mod & Modifier.VOLATILE) != 0) {
		str.append("volatile ");
	}

	// Type
	str.append(Method.getPrettyName(type));
	str.append(' ');

	// Class name
	str.append(clazz.getName());
	str.append(".");
	// Field name
	str.append(name);

	return (str.toString());
}

public int getModifiers() {
	return (getModifiers0(clazz, slot));
}

native static private int getModifiers0(Class clazz, int slot);
native private boolean getBoolean0(Object obj, Class clazz, int slot);
native private byte getByte0(Object obj, Class clazz, int slot);
native private char getChar0(Object obj, Class clazz, int slot);
native private short getShort0(Object obj, Class clazz, int slot);
native private int getInt0(Object obj, Class clazz, int slot);
native private long getLong0(Object obj, Class clazz, int slot);
native private float getFloat0(Object obj, Class clazz, int slot);
native private double getDouble0(Object obj, Class clazz, int slot);
native private Object getObject0(Object obj, Class clazz, int slot);

native private void setBoolean0(Object obj, boolean v, Class clazz, int slot);
native private void setByte0(Object obj, byte v, Class clazz, int slot);
native private void setChar0(Object obj, char v, Class clazz, int slot);
native private void setShort0(Object obj, short v, Class clazz, int slot);
native private void setInt0(Object obj, int v, Class clazz, int slot);
native private void setLong0(Object obj, long v, Class clazz, int slot);
native private void setFloat0(Object obj, float v, Class clazz, int slot);
native private void setDouble0(Object obj, double v, Class clazz, int slot);
native private void setObject0(Object obj, Object v, Class clazz, int slot);

}
