/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.net;

final public class DatagramPacket {

byte[] buf;
int length;
InetAddress address;
int port;

/**
 * Receive buffer.
 */
public DatagramPacket(byte ibuf[], int ilength) {
	if (ibuf == null) {
		throw new NullPointerException();
	}
	if (ilength < 0 || ilength > ibuf.length) {
		throw new IllegalArgumentException("bad length: " + ilength);
	}
	buf = ibuf;
	length = ilength;
	address = null;
	port = -1;
}

/**
 * Send buffer.
 */
public DatagramPacket(byte obuf[], int olength, InetAddress iaddr, int oport) {
	if (oport < 0 || oport > 65535) {
		throw new IllegalArgumentException("bad port: " + oport);
	}
	if (olength < 0 || olength > obuf.length) {
		throw new IllegalArgumentException("bad length: " + olength);
	}
	if (iaddr == null) {
		throw new NullPointerException();
	}
	buf = obuf;
	length = olength;
	address = iaddr;
	port = oport;
}

public synchronized InetAddress getAddress() {
	return (address);
}

public synchronized byte[] getData() {
	return (buf);
}

public synchronized int getLength() {
	return (length);
}

public synchronized int getPort() {
	return (port);
}

public synchronized void setAddress(InetAddress addr) {
	if (addr == null) {
		throw new NullPointerException();
	}
	address = addr;
}

public synchronized void setData(byte[] newbuf) {
	if (newbuf == null) {
		throw new NullPointerException();
	}
	buf = newbuf;
}

public synchronized void setLength(int newlen) {
	if (newlen <= 0 || newlen > buf.length) {
		throw new IllegalArgumentException();
	}
	length = newlen;
}

public synchronized void setPort(int newport) {
	if (newport < 0 || newport > 65535) {
		throw new IllegalArgumentException();
	}
	port = newport;
}

}
