/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.net;

import java.io.IOException;
import kaffe.net.NetIO;

/* XXX: This class should not be public!
 * It's public right now cause it's accessed from kaffe.net.
 */
public class PlainDatagramSocketImpl
	extends DatagramSocketImpl {

private int timeout;

static {
	System.loadLibrary("net");
}

protected void create() throws SocketException {
	fd.socket(false);
}

protected void close() {
	fd.close();
}

public void setOption(int option, Object data) throws SocketException {
	int value = -1;
	if (data instanceof Boolean) {
		value = ((Boolean)data).booleanValue() ? 1 : 0;
	}
	else if (data instanceof Integer) {
		value = ((Integer)data).intValue();
	}
	else if (data instanceof InetAddress) {
		value = ((InetAddress)data).getAddressAsInt();
	}

	switch (option) {
	case SO_SNDBUF:
	case SO_RCVBUF:
	case SO_LINGER:
	case SO_REUSEADDR:
	case SO_TIMEOUT:
	case IP_MULTICAST_IF:
		fd.setSocketOption(option, value);
		break;

	case SO_BINDADDR:
		throw new SocketException("Read-only socket option");

	default:
		throw new SocketException("Unknown socket option");
	}
}

public Object getOption(int option) throws SocketException {
	switch (option) {
	case SO_SNDBUF:
	case SO_RCVBUF:
	case SO_LINGER:
	case SO_REUSEADDR:
	case SO_TIMEOUT:
		return (new Integer(fd.getSocketOption(option)));

	case SO_BINDADDR:
	case IP_MULTICAST_IF:
		return (InetAddress.newInstance(fd.getSocketInfo(NetIO.LOCALADDR)));

	default:
		throw new SocketException("Unknown socket option");
	}
}

protected void finalize() throws Throwable {
	close();
	super.finalize();
}

protected void bind(int lport, InetAddress laddr) throws SocketException {
	fd.bind(laddr.getAddressAsInt(), lport);
}

protected void send(DatagramPacket p) throws IOException {
	fd.sendto(p.buf, p.length, p.address.getAddressAsInt(), p.port);
}

protected int peek(InetAddress i) throws IOException {
	i.flushFromCache();
	return (fd.peekaddr(i.dns));
}

protected void receive(DatagramPacket p) throws IOException {
	p.address.flushFromCache();
	p.length = fd.recvfrom(p, p.address.dns);
}

protected void setTTL(byte ttl) throws IOException {
	fd.setSocketOption(IP_MULTICAST_TTL, (int)ttl);
}

protected byte getTTL() throws IOException {
	return ((byte)fd.getSocketOption(IP_MULTICAST_TTL));
}

protected void join(InetAddress inetaddr) throws IOException {
	fd.setSocketOption(IP_ADD_MEMBERSHIP, inetaddr.getAddressAsInt());
}

protected void leave(InetAddress inetaddr) throws IOException {
	fd.setSocketOption(IP_DROP_MEMBERSHIP, inetaddr.getAddressAsInt());
}

}
