/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.net;

import java.io.FileDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.String;
import kaffe.net.NetIO;

/* XXX: This class should not be public!
 * It's public right now cause it's accessed from kaffe.net.
 */
public class PlainSocketImpl
  extends SocketImpl
{
	private InputStream in;
	private OutputStream out;

static {
	System.loadLibrary("net");
}

protected void accept(SocketImpl s) throws IOException {
	s.address = InetAddress.getAnyAddress();
	s.fd.accept(fd, s.address.getAddressAsInt(), s.port);
}

protected int available() throws IOException {
	return (fd.available());
}

protected void bind(InetAddress address, int lport) throws IOException {
	fd.bind(address.getAddressAsInt(), lport);
	this.address = address;
	this.localport = fd.getLocalPort();
}

protected void close() throws IOException {
		in = null;
		out = null;
	       	fd.close();
}

protected void connect(InetAddress address, int port) throws IOException {
	fd.connect(address.getAddressAsInt(), port);
}

protected void connect(String host, int port) throws IOException {
	connect(InetAddress.getByName(host), port);
}

protected void create(boolean stream) throws IOException {
	fd.socket(stream);
}

protected void finalize() throws Throwable {
	close();
	super.finalize();
}

protected synchronized InputStream getInputStream() throws IOException {
	if (in == null) {
		in = new SocketInputStream(this); 
	}
	return (in);
}

public Object getOption(int option) throws SocketException {
	switch (option) {
	case SO_SNDBUF:
	case SO_RCVBUF:
	case SO_LINGER:
	case SO_REUSEADDR:
	case SO_TIMEOUT:
		return (new Integer(fd.getSocketOption(option)));

	case TCP_NODELAY:
		return (new Boolean(fd.getSocketOption(option) != 0));

	case SO_BINDADDR:
		return (InetAddress.newInstance(fd.getSocketInfo(NetIO.LOCALADDR)));

	default:
		throw new SocketException("Unknown socket option");
	}
}

protected synchronized OutputStream getOutputStream() throws IOException {
	if (out == null) {
		out = new SocketOutputStream(this);
	}
	return (out);
}

protected int getReceiveBufferSize() throws SocketException {
	return fd.getSocketOption(SO_RCVBUF);
}

protected int getSendBufferSize() throws SocketException {
	return fd.getSocketOption(SO_SNDBUF);
}

protected int getSoLinger() throws SocketException {
	return fd.getSocketOption(SO_LINGER);
}

protected int getSoTimeout() throws SocketException {
	return ((Integer)getOption(SocketOptions.SO_TIMEOUT)).intValue();
}

protected boolean getTcpNoDelay() throws SocketException {
	return fd.getSocketOption(TCP_NODELAY) != 0;
}

protected void listen(int count) throws IOException {
	fd.listen(count);
}

protected int read(byte[] buf, int offset, int len) throws IOException {
	return (fd.read(buf, offset, len));
}

public void setOption(int option, Object data) throws SocketException {
	int value = -1;
	if (data instanceof Boolean) {
		value = ((Boolean)data).booleanValue() ? 1 : 0;
	}
	else if (data instanceof Integer) {
		value = ((Integer)data).intValue();
	}

	switch (option) {
	case SO_SNDBUF:
	case SO_RCVBUF:
	case SO_LINGER:
	case SO_REUSEADDR:
	case TCP_NODELAY:
	case SO_TIMEOUT:
		fd.setSocketOption(option, value);
		break;

	case SO_BINDADDR:
		throw new SocketException("Read-only socket option");

	default:
		throw new SocketException("Unknown socket option");
	}
}

protected void setReceiveBufferSize(int size) throws SocketException {
	fd.setSocketOption(SO_RCVBUF, size);
}

protected void setSendBufferSize(int size) throws SocketException {
	fd.setSocketOption(SO_SNDBUF, size);
}

protected void setSoLinger(boolean on, int howlong) throws SocketException {
	fd.setSocketOption(SO_LINGER, (on ? howlong : 0));
}

protected void setSoTimeout(int timeout) throws SocketException {
	setOption(SocketOptions.SO_TIMEOUT, new Integer(timeout));
}

protected void setTcpNoDelay(boolean on) throws SocketException {
	fd.setSocketOption(TCP_NODELAY, (on ? 1 : 0));
}

protected void write(byte[] buf, int offset, int len) throws IOException {
	fd.write(buf, offset, len);
}

}
