/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package java.util.zip;

public class Adler32 implements Checksum {

  static {
    System.loadLibrary("zip");
  }

  private int adler;

  public Adler32()
  {
    adler = 0;
  }

  public void update(int b)
  {
    adler = update1(adler, (byte)b);
  }

  public void update(byte b[])
  {
    update(b, 0, b.length);
  }

  public void reset()
  {
    adler = 0;
  }

  public long getValue()
  {
    return (adler);
  }

  public void update(byte[] buf, int from, int len) {
    adler = update0(adler, buf, from, len);
  }

  private native int update0(int adler, byte[] buf, int from, int len);
  private native int update1(int adler, byte val);

}
