package kaffe.awt;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Insets;
import java.awt.Toolkit;
import java.util.Hashtable;

/**
 * Defaults - adaption parameters of package java.awt. This is our way to
 * parameterize various settings like colors and fonts. We prefer a simple
 * Java class (in favor of properties) because it doesn't need extra file
 * system access, is much faster, more robust (no parsing, compile time check),
 * and can be overridden by simply extending the CLASSPATH (no jar archive
 * modification).
 *
 * Be very carefully to add additional java.awt types, since this might
 * introduce recursive class init problems. This is especially
 * true if static fields of such classes are involved (like Color), since
 * they might be still "null" (if they in turn refer to Toolkit/Defaults)
 * when referenced inside of this class.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author P.C.Mehlitz
 */
public class Defaults
{
/**
 * If this is set to a valid classname, the corresponding class will be used to
 * instanciate the event queue. Note that this class should be derived from
 * java.awt.EventQueue in order to enable things like repaint solicitation.
 *
 * Due to a JDK bug, this field has to be the first one, to avoid recursive
 * initialization problems (not relevant for kaffe)
 */
	public static String EventQueueClass = setStringValue( "awt.EventQueueClass", null );
/**
 * If ConsoleClass is set to the classname of a kaffe.util.log.LogClient, System.out, err
 * are redirected to a kaffe.util.log.LogStream, which notifies the provided LogClient
 * (usually a simple Window displaying scrolled text). If ConsoleClass is null, no
 * io redirection is performed.
 */
	public static String ConsoleClass = /* "kaffe.util.log.SimpleLogViewer" */ null;
/**
 * If set to 'true', the last Window that is closed will automatically
 * call a System.exit()
 */
	public static boolean AutoStop = setBooleanValue( "awt.AutoStop", true);
/**
 * Upper bound (in ms) between mouse button press events which will be
 * considered as a 'click' (will increase the MouseEvent clickCount)
 */
	public static int ClickInterval = 250;
/**
 * maximum x,y coordinate distance (in pixels) between successive MOUSE_PRESSED
 * events which are still considered to be a click (required for some devices
 * without exact mouse positioning).
 */
	public static int ClickDistance = 0;
/**
 * Sleep time (in ms) between native getNextEvent calls (= event polling rate).
 * Just used in case the native layer does not support blocked IO.
 */
	public static int EventPollingRate = 30;
/**
 * Delay (in ms) between background native flush calls (from the flush thread) in case
 * the native layer needs flushing (queued graphics systems like X). This is only
 * required for apps doing graphics output outside the event dispatch thread, and for
 * native layers doing blocked IO (in this case the background drawing wouldn't appear
 * until something is going on in the dispatcher thread)
 */
	public static int GraFlushRate = setIntValue( "awt.GraFlushRate", 250);
/**
 * Physical screen size in 1/10 inches (used for scaling purposes, e.g. fr printing)
 */
	public static int ScreenSize = 130;
/**
 * just a shadow from the Defaults screen width (computed by the native layer)
 */
	public static int ScreenWidth;
/**
 * screen resolution (in dpi), to be used for printing
 */
	public static int XResolution;
/**
 * screen resolution (in dpi), to be used for printing
 */
	public static int YResolution;
/**
 * Pixel height of MenuBars
 */
	public static int MenuBarHeight;
	public static int ScrollbarWidth = 14;
/**
 * Frame decoration extends (titlebar, size-borders) in pixels. For native windowing
 * systems (like X) this is just a guess, to be used for the first Frame to be opened,
 * and later on corrected (Frame.frameDeco). For standalone envionments, this is
 * an order
 */
	public static Insets FrameInsets = new Insets( 24,4,4,4);
/**
 * Dialog decoration extends (titlebar, size-borders) in pixels. For native windowing
 * systems (like X) this is just a guess, to be used for the first Dialog to be opened,
 * and later on corrected (Dialog.dialogDeco). For standalone envionments, this is
 * an order
 */
	public static Insets DialogInsets = new Insets( 24, 4, 4, 4);
/**
 * Color used to mark the focus Component (border)
 */
	public static Color FocusClr = new Color( 191, 0, 0);
/**
 * Default foreground color for Windows
 */
	public static Color WndForeground = Color.black;
/**
 * Default background color for Windows
 */
	public static Color WndBackground = Color.lightGray;
/**
 * Default font for Windows
 */
	public static Font WndFont;
/**
 * metrics of the WndFont (computed)
 */
	public static FontMetrics WndFontMetrics;
/**
 * Default font for menus
 */
	public static Font MenuFont;
/**
 * Default text color for menus
 */
	public static Color MenuTxtClr = Color.black;
/**
 * Text color of selected meu item
 */
	public static Color MenuSelTxtClr = FocusClr;
/**
 * Default background color for menus
 */
	public static Color MenuBgClr = Color.lightGray;
	public static Color MenuSelBgClr = new Color( 180, 180, 180);
/**
 * If set to 'true', menu text will be drawn "carved" (with a down-right highlight)
 */
	public static boolean MenuTxtCarved = true;
/**
 * Default text font
 */
	public static Font TextFont;
/**
 * Default TextArea background color
 */
	public static Color TextAreaBgClr = Color.white;
/**
 * Default TextArea text color
 */
	public static Color TextAreaTxtClr = Color.black;
	public static Color TextAreaRoBgClr = Color.white;
/**
 * text color of non-editable TextComponents
 */
	public static Color TextAreaRoTxtClr = Color.gray;
/**
 * Default TextArea font
 */
	public static Font TextAreaFont;
/**
 * Default TextArea selection background color
 */
	public static Color TextAreaSelBgClr = Color.lightGray;
/**
 * Default TextArea selection text color
 */
	public static Color TextAreaSelTxtClr = Color.black;
/**
 * TextArea focus cursor color
 */
	public static Color TextCursorClr = Color.blue;
/**
 * TextArea inactive (non-focus) cursor color
 */
	public static Color TextCursorInactiveClr = Color.gray;
/**
 * If set to 'true', non-focused TextAreas keep their cursor displayed (in inactive color)
 */
	public static boolean ShowInactiveCursor = false;
/**
 * If set to 'true', ScrollBars can get focus
 */
	public static boolean FocusScrolls = false;
/**
 * Default TextField background color
 */
	public static Color TextFieldBgClr = Color.white;
/**
 * Default TextField text color
 */
	public static Color TextFieldTxtClr = Color.black;
/**
 * Default TextField selection background color
 */
	public static Color TextFieldSelBgClr = Color.lightGray;
/**
 * Default TextField text color
 */
	public static Color TextFieldSelTxtClr = Color.black;
/**
 * Default TextField font
 */
	public static Font TextFieldFont;
/**
 * Default List font
 */
	public static Font ListFont;
/**
 * Default List background color
 */
	public static Color ListBgClr = new Color( 223, 223, 223);
/**
 * Default Lis text color
 */
	public static Color ListTxtClr = Color.black;
/**
 * List selection background color
 */
	public static Color ListSelBgClr = new Color( 0xa3a3a3);
/**
 * List selection text color
 */
	public static Color ListSelTxtClr = Color.black;
/**
 * List fly-over background color (item under mouse pointer)
 */
	public static Color ListFlyOverBgClr = new Color( 214, 214, 214);
/**
 * List fly-over text color (item under mouse pointer)
 */
	public static Color ListFlyOverTxtClr = Color.black;
/**
 * If set to 'true', fly-over items will be drawn with an inset border
 * (outset otherwise)
 */
	public static boolean ListFlyOverInset = false;
/**
 * fire action command on click instead of dblClick ( e.g. touchscreen devices )
 */
	public static boolean ListClickAction = false;
/**
 * Default Label text color
 */
	public static Color LabelClr = Color.black;
/**
 * Default Label font
 */
	public static Font LabelFont;
/**
 * If set to 'true', Label text will be drawn carved (with right-down highlight)
 */
	public static boolean LabelTxtCarved = false;
/**
 * Default background color for Buttons
 */
	public static Color BtnClr = new Color( 0xa3a3a3);
	public static boolean BtnAnimation = true;
/**
 * Default text color for Buttons
 */
	public static Color BtnTxtClr = Color.black;
/**
 * Button background color if button is under mouse pointer
 */
	public static Color BtnPointClr = new Color( 180, 180, 180);
/**
 * Button text color if button is under mouse pointer
 */
	public static Color BtnPointTxtClr = FocusClr;
/**
 * Default font for Buttons
 */
	public static Font BtnFont;
/**
 * If set to 'true', Button text will be drawn carved (with right-down highlight)
 */
	public static boolean BtnTxtCarved = true;
/**
 * Color for Component borders
 */
	public static Color BorderClr = new Color( 159, 159, 159);
/**
 * Color for active window borders (only used in standalone environments)
 */
	public static Color WndActiveClr = new Color( 191, 0, 0);
/**
 * Color for active window titlebar background (only used in standalone environments)
 */
	public static Color WndActiveTitleClr = Color.yellow;
/**
 * Color for inactive window borders (only for standalone environments)
 */
	public static Color WndInactiveClr = Color.lightGray;
/**
 * Color for inactive window titlebar background (only used in standalone environments)
 */
	public static Color WndInactiveTitleClr = Color.black;
/**
 * Maximum size of image files to be produced immediately (when the corrseponding
 * image is created). Images with files exceeding this size will be produced
 * asynchronously on demand
 */
	public static int FileImageThreshold = setIntValue( "awt.FileImageThreshold", 0xffff);
/**
 * Maximum size of image data buffers to be produced immediately (when the corrseponding
 * image is created). Images with data buffers exceeding this size will be produced
 * asynchronously on demand
 */
	public static int DataImageThreshold = setIntValue( "awt.DataImageThreshold", 0xffff);
/**
 * Maximum size of MemoryImageSource data to be produced immediately (when the
 * corrseponding image is created). Images with data buffers exceeding this
 * size will be produced asynchronously on demand (currently, everything != 0
 * will cause async production)
 */
	public static int MemImageSrcThreshold = setIntValue( "awt.MemImageSrcThreshold", 0xffff);
/**
 * AWT events are usually short lived objects created with high frequency (mouse motion).
 * Since creation and destruction are usually in control of the AWT innards
 * (EventDispatchThread), they are good candidates for class local caching. Mostly. Some
 * apps really store events when processing them (mostly as a quick&dirty way to save state).
 * In case we have such a beast, turn off the xx_EVENT_MASK bit for this event type in
 * 'RecycleEvents'. A value of '0' means there is no caching at all, a value of 0xffffffff
 * caches everything.
 */
	public static int RecycleEvents = 0xffffffff;
/**
 * Name of RootWindow (screen) class. Only used in standalone environments
 */
	public static String RootWindowClass = "java.awt.RootWindow";
/**
 * Background color for RootWindow (stand-alone environments only)
 */
	public static Color RootWindowClr = setColorValue( "awt.RootWindowClr", Color.blue);
/**
 * Image filename to be used to tile RootWindow background (only for standalone environments)
 */
	public static String RootWindowImg = setStringValue( "awt.RootWindowImg", null);
// "turquoise.jpg";
/**
 * Name of class to be used to control desktop pages and controls (only for
 * standalone environments)
 */
	public static String ControlBoxClass = setStringValue( "awt.ControlBox", "java.awt.WinTaskBar");
/**
 * Name of class to be used to display and control minimized windows (only for
 * standalone environments)
 */
	public static String MinBoxClass = setStringValue( "awt.MinBox", "java.awt.WinTaskBar");
/**
 * Name of the file to load as the startup banner.
 */
	public static String banner;
/**
 * Name of the file to play when we need to 'beep'
 */
	public static String beep;
	public static String UserInitClass = setStringValue( "awt.UserInitClass", null);
/**
 * The actual font size depends not just on the point size, but also the screen
 * (resolution/size). 'FontSizeOffset' can be used to adapt this (e.g. a value
 * of '-1' means a font which is requested in Java with point size X, is actually
 * created for size X-1 in the native layer)
 */
	public static int FontSizeOffset = setIntValue( "awt.FontSizeOffset", 0);
/**
 * Depending on the native font representation, font metrics objects can be quite
 * expensive (in terms of size and initialization). If an app uses them as temporary objects,
 * it therefore might be a good idea to cache them, which is done by setting 'CacheFontMetrics'
 * to 'true'. However, if an app uses a lot of them, and caches non-temporary metrics anyway,
 * it might also be a great waste of memory
 */
	public static boolean CacheFontMetrics = setBooleanValue( "awt.CacheFontMetrics", true);
/**
 * just a shadow from the Defaults screen height (computed by the native layer)
 */
	public static int ScreenHeight;
/**
 * standard border width of widgets
 */
	public static int WidgetBorderWidth = 2;
/**
 * width of choice button
 */
	public static int ChoiceBtnWidth = 15;
/**
 * the widget rendering theme ( has to extend kaffe.awt.WidgetThemeImpl )
 */
	public static String WidgetThemeClass = setStringValue( "awt.Theme", "kaffe.awt.theme.WidgetThemeKaffe");
/**
 * Strict PJ = true means we don't allow behavioral extensions
 * like Window(null) creation
 */
	public static boolean StrictPJ = setBooleanValue( "awt.StrictPJ", false);
/* This is the default image animation frame latency in ms, which is used in case
 * it is not specified in the image
 */
	public static int FrameLatency = 100;
	public static String ThemeClass = setStringValue( "awt.Theme", "kaffe.awt.ThemeKaffe");
	public static int FrameBorderWidth = 4;
	public static int SysFontHeight;
	public static String KeyHookClass = setStringValue("awt.keyHook", "java.awt.DefKeyFilter");
	public static String MouseHookClass = setStringValue("awt.mouseHook", null);
	public static String FocusHookClass = setStringValue("awt.focusHook", null);
	public static boolean SuspendAnimations = setBooleanValue("awt.SuspendAnimations", true);
	public static String FocusNextSpec = "KEY_PRESSED,keycode=117,F6";
	public static String FocusPrevSpec = "KEY_PRESSED,keycode=118,F7";

static {
	Toolkit tk = Toolkit.getDefaultToolkit();

	// these are the values we get back from the native init
	Dimension sd = tk.getScreenSize();

	ScreenWidth  = sd.width;
	ScreenHeight = sd.height;

	// we assume a 4:3 width/height ratio of the display
	XResolution = (ScreenWidth  * 100 + 50) / (8 * ScreenSize); // dpi
	YResolution = (ScreenHeight * 100 + 50) / (6 * ScreenSize); // dpi

	WndFont = new Font( "Default", Font.BOLD, 11);
	MenuFont = new Font( "Default", Font.BOLD, 11);
	TextFont = new Font( "Default", Font.BOLD, 11);
	TextAreaFont = new Font( "Monospaced", Font.PLAIN, 11);
	TextFieldFont = new Font( "Default", Font.BOLD, 11);
	ListFont = new Font( "Default", Font.BOLD, 11);
	LabelFont = new Font( "Default", Font.PLAIN, 11);
	BtnFont = new Font( "Default", Font.BOLD, 11);
	
	WndFontMetrics = tk.getFontMetrics( WndFont);

	SysFontHeight = WndFontMetrics.getHeight();
		
	if ( MenuBarHeight == 0 ){
	  // we assume this is something like the "system font", which is a synonym
	  // for the TitlebarHeight (alternatively, we could use the MenuFont itself)
		MenuBarHeight = WndFontMetrics.getHeight() + 8;
	}

	beep = System.getProperty("awt.beep", "beep.wav");

	if ( UserInitClass != null ) {
		try {
			Class asc = Class.forName( UserInitClass);
		}
		catch ( Exception x ) {
			System.err.println( "UserInitClass not initialized: " + x);
		}
	}
}

static boolean setBooleanValue ( String propKey, boolean defValue ) {
	boolean    val;
	
	String s = System.getProperty( propKey);
	if ( s != null ) {
		val = s.equalsIgnoreCase( "true");
	}
	else {	
		val = defValue;
	}
		
	return val;
}

static Color setColorValue ( String propKey, Color defValue ) {
	Color    val = defValue;
	
	String s = System.getProperty( propKey);
	if ( s != null ) {
		try {
			val = new Color( Integer.parseInt( s, 16));
		}
		catch ( NumberFormatException x ) {
			System.err.println( "malformed awt property: " + propKey);
		}
		s = null;
	}
		
	return val;
}

static int setIntValue ( String propKey, int defValue ) {
	int    val = defValue;
	
	String s = System.getProperty( propKey);
	if ( s != null ) {
		try {
			val = Integer.parseInt( s);
		}
		catch ( NumberFormatException x ) {
			System.err.println( "malformed awt property: " + propKey);
		}
		s = null;
	}
		
	return val;
}

static String setStringValue ( String propKey, String defValue ) {
	// we could do this directly via System.getProperty(key,def), but just
	// for Strings, and we want more types
	
	String val = System.getProperty( propKey);
	if ( val == null ){
		val = defValue;
	}
	if ( "null".equals( val) ){
		val = null;
	}

	return val;
}
}
