package kaffe.awt;


public class GenericEvent
  extends java.awt.AWTEvent
{
	byte[] data;
	int dataLength;
	long when;
	static GenericEvent cache;
	final static int GENERIC_FIRST = 10001;
	final static int GENERIC_LAST = 10010;
	final static int N_GENERICS = GENERIC_LAST - GENERIC_FIRST + 1;
	static GenericEventHandler[] handlers = new GenericEventHandler[N_GENERICS];

public GenericEvent ( Object source,  int id, byte[] data, int dataLen, long when ){
	super( source, id);
	
	this.when = when;
	this.data = data;
	this.dataLength = dataLen;
}

protected void dispatch () {
	GenericEventHandler src = (GenericEventHandler) source;
	if ( src != null ){
		src.processGenericEvent( this);
	}
	
	// 2DO - make this configurable (like with the standard events)
	recycle();
}

public byte[] getData () {
	return data;
}

public static byte[] getDataBuffer ( int idx, int size ) {
	if ( (idx >= 0) && (idx < N_GENERICS) ){
		GenericEventHandler handler = handlers[idx];
		if ( handler != null ){
			return handler.getDataBuffer( size);
		}
	}
	
	return null;
}

public int getDataLength () {
	return dataLength;
}

static synchronized GenericEvent getEvent ( int id, byte[] data, int dataLen ) {
	GenericEvent e;
	Object source;
	long    when = System.currentTimeMillis();
	
	if ( (id < 0) || (id >= N_GENERICS) ){
		// this is no event we can handle
		return null;
	}
	else {
		source = handlers[id];
	}

	if ( cache == null ) {
		e = new GenericEvent( source,  id, data, dataLen, when);
	}
	else {
		e = cache;
		cache = (GenericEvent) e.next;
		e.next = null;
	
		e.id         = id;
		e.source     = source;
		e.when       = when;
		
		e.data = data;
		e.dataLength = dataLen;
	}

	// has to be made accessible via AWT
/***
	if ( (Toolkit.flags & Toolkit.NATIVE_DISPATCHER_LOOP) != 0 ) {
		// this is not used as a direct return value for EventQueue.getNextEvent(), 
		// it has to be Java-queued by the native layer
		Toolkit.eventQueue.postEvent( e);
	}
***/

	return e;
}

protected void recycle () {
	synchronized ( GenericEvent.class ) {
		source = null;

		next = cache;	
		cache = this;
	}
}

public static boolean setHandler ( int idx, GenericEventHandler handler ){
	if ( (idx < 0) || (idx >= N_GENERICS) ){ // no such slot
		return false;
	}

	handlers[idx] = handler;
	return true;
}
}
