package kaffe.awt;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.Observable;
import java.util.Observer;

public class ViewSlider
  implements Runnable, ComponentListener, LayoutManager
{
	Container parent;
	Image dbImage;
	int overlay;
	int curIdx;
	Thread slider = new Thread( this);
	int degressionRate = 1;
	int raster = 3;
	int degressionDelay;
	int linearDelay;
	int mode;
	boolean active;
	static int SLIDING_TIME_REQUEST = 300;
	final public static int NO_SLIDE = 1;
	final public static int PUSH_SLIDE = 2;
	final public static int COVER_SLIDE = 3;
	final public static int COVER_SLIDE_DOWN = 4;

public ViewSlider( Container parent, int overlay) {
	this( parent, overlay, COVER_SLIDE);
}

public ViewSlider( Container parent, int overlay, int mode ) {
	this.parent = parent;
	this.overlay = overlay;
	this.mode = mode;

	curIdx = 0;
	showCurrent();

	updateImage();
	parent.addComponentListener( this);
	
	//has to set component bounds anyway, so just use it as LayoutManager
	parent.setLayout( this);

	slider.setDaemon( true);
	slider.start();
}

public void addLayoutComponent ( String name, Component child ){
}

private void animate_cover() {
	int x0, x1, y0, y1, y2, pw, ph;
	int step, delay;
	long t1;

	int width = AWT.extension.getWidth( parent);
	int height = AWT.extension.getHeight( parent);
	Insets in = parent.getInsets();

	x0 = in.left;
	x1 = width - in.right;
	y0 = in.top + overlay;
	y1 = height - in.bottom;
	pw = width-in.left-in.right;
	ph = height-in.top-in.bottom;

	Graphics g = parent.getGraphics();

	if (curIdx == 0) {
		delay = linearDelay;
		step = raster;

		t1 = System.currentTimeMillis();
		for ( int dy = step + overlay; dy < ph; dy += step ) {
			if ( dy + step  > ph )
				dy = ph;
				
			g.setClip( x0, in.top, pw, dy);
			g.drawImage( dbImage, x0, in.top + dy - ph, null);
			
			delay( delay);
		}

		calculateRasterValues( (int)(System.currentTimeMillis()-t1), y1-y0);
	}
	else {
		delay = degressionDelay;
		step = degressionSteps(y1-y0 - overlay); 

		t1 = System.currentTimeMillis();
		for( y2 = y1; y2 > overlay; ) {

			g.setClip( x0, in.top, pw, y2-in.top);
			g.copyArea( x0, in.top, pw, y2-in.top, 0, -step);

			g.setClip( x0, y2-step, pw, step);
			g.drawImage( dbImage, x0, y0, null);

			delay( delay);
			
			y2 -= step;
			if ( (step-=degressionRate) <= 0 ) {
				y2 = step;
			}
		}
		calculateDegressionValues ( (int)(System.currentTimeMillis()-t1), y1-y0);
	}

	g.dispose();

	showCurrent();	
}

private void animate_cover_down() {
	int x0, x1, y0, y1, y2, pw, ph;
	int step, delay;
	long t1;

	int width = AWT.extension.getWidth( parent);
	int height = AWT.extension.getHeight( parent);
	Insets in = parent.getInsets();

	x0 = in.left;
	x1 = width - in.right;
	y0 = in.top + overlay;
	y1 = height - in.bottom;
	pw = width-in.left-in.right;
	ph = height-in.top-in.bottom;

	Graphics g = parent.getGraphics();

	if (curIdx == 0) {
		delay = linearDelay;
		step = raster;

		t1 = System.currentTimeMillis();
		for ( int dy = step + overlay; dy < ph; dy += step ) {
			if ( dy + step  > ph ) {
				dy = ph;
			}
				
			g.drawImage( dbImage, x0, y1-dy, null);
			delay( delay);
			
		}
		calculateRasterValues( (int)(System.currentTimeMillis()-t1), y1-y0);
	}
	else {
		delay = degressionDelay;
		step = degressionSteps(y1-y0 - overlay); 

		t1 = System.currentTimeMillis();
		for ( int dy = step; dy < ph - overlay; dy += step ) {
				
			g.setClip( x0, in.top+ dy-step, pw, ph-dy + step);
			g.copyArea( x0, in.top+ dy-step, pw, ph-dy, 0, step);

			g.setClip( x0, in.top+ dy-step, pw, step);
			g.drawImage( dbImage, x0, in.top, null);
			
			delay( delay);

			step-=degressionRate;
			if ( step < 1 )
				step = 1;
			if ( dy + step > ph - overlay )
				step = ph - overlay - dy - 1;
		}
		calculateDegressionValues ( (int)(System.currentTimeMillis()-t1), y1-y0);
	}

	g.dispose();

	showCurrent();	
}

private void animate_push() {
	int x0, x1, y0, y1, y2, pw, ph;
	int step, delay;
	long t1;

	int width = AWT.extension.getWidth( parent);
	int height = AWT.extension.getHeight( parent);
	Insets in = parent.getInsets();

	x0 = in.left;
	x1 = width - in.right;
	y0 = in.top + overlay;
	y1 = height - in.bottom;
	pw = width-in.left-in.right;
	ph = height-in.top-in.bottom;

	Graphics g = parent.getGraphics();

	if (curIdx == 0) {
		delay = degressionDelay;
		step = degressionSteps(y1-y0 - overlay); 

		t1 = System.currentTimeMillis();
		for ( int dy = step + overlay; dy < ph; dy += step ) {
			if ( dy + step  > ph )
				dy = ph;
				
			g.copyArea( x0, in.top + step, pw, y1-dy, 0, -step);
			g.drawImage( dbImage, x0, y1-dy, null);
			
			delay( delay);
			
			step-=degressionRate;
			if ( step < 1 )
				step = 1;
		}
		calculateDegressionValues ( (int)(System.currentTimeMillis()-t1), y1-y0);
	}
	else {
		delay = degressionDelay;
		step = degressionSteps(y1-y0 - overlay); 

		t1 = System.currentTimeMillis();
		for ( int dy = step + overlay; dy < ph; dy += step ) {
			if ( dy + step  > ph )
				dy = ph;
				
			y2 = in.top + dy - step - overlay;
			g.copyArea( x0, y2, pw, y1-y2, 0, step);
			g.drawImage( dbImage, x0, y2 + step - ph, null);
			
			delay( delay);
			
			step-=degressionRate;
			if ( step < 1 )
				step = 1;
		}

		calculateDegressionValues ( (int)(System.currentTimeMillis()-t1), y1-y0);
	}

	g.dispose();

	showCurrent();	
}

private void calculateDegressionValues(  int dt, int height) {
	if ( dt > SLIDING_TIME_REQUEST ) {
		//too slow, increment degression, if not already delayed
		if ( degressionDelay == 0 )
			degressionRate++;
	}
	else if ( dt < SLIDING_TIME_REQUEST ){
		//too fast, increase delay
		degressionDelay+=10;
	}

	//	System.out.println( "elapsed up time:" + dt + " should:" + SLIDING_TIME_REQUEST );
	// 	System.out.println( "update degression (rate:" + degressionRate + ",delay:" + degressionDelay + ")");
}

private void calculateRasterValues( int dt, int height) {
	int maxStep = height / 4;
	int update = (raster * dt) / SLIDING_TIME_REQUEST;

	if ( update < 1 ) {
		//too fast, increase delay
		update = 1;
		linearDelay = (SLIDING_TIME_REQUEST-(raster*dt)) / height;
	}
	else if ( update > maxStep ) {
		//too slow, do at least some bigraster sliding
		update = maxStep;
	}


	raster = update;

	//	System.out.println( "elapsed down time:" + dt + " should:" + SLIDING_TIME_REQUEST );
	// 	System.out.println( "update linear (raster:" + raster + ",delay:" + linearDelay + ")");
}

public void componentHidden ( ComponentEvent evt ) {
}

public void componentMoved ( ComponentEvent evt ) {
}

public void componentResized ( ComponentEvent evt ) {
	updateImage();
}

public void componentShown ( ComponentEvent evt ) {
}

private int degressionSteps( int height) {
	int ii, ch = height;
	for ( ii=1; ch>degressionRate; ii+=degressionRate ) {
		ch-=ii;
	}
	return (ii);
}

private void delay( int ms) {
	if ( ms > 0 ) {
		try { Thread.sleep( ms); }
		catch ( Exception _) {}
	}
}

public int getView() {
	return curIdx;
}

public int getViewCount() {
	return parent.getComponentCount();
}

private void hideAll() {
	int nc = parent.getComponentCount();
	for ( int i=0; i<nc; i++ ) {
		parent.getComponent(i).setVisible( false);
	}
}

public void layoutContainer ( Container parent ){
	Insets in = parent.getInsets();
	int x = in.left;
	int y = in.top;
	int width = AWT.extension.getWidth( parent) - x - in.right;
	int height = AWT.extension.getHeight( parent) - y - in.bottom;;
	int nc = parent.getComponentCount();


	if ( curIdx == 0 ) {
		if ( nc > 0 ) {
			parent.getComponent(0).setBounds( x, y, width, height);
		}	
	}
	else if ( curIdx < nc ){
		switch ( mode) {
			case PUSH_SLIDE:
			case COVER_SLIDE_DOWN:
				parent.getComponent(0).setBounds( x, y+height-overlay, width, height);
				parent.getComponent(curIdx).setBounds( x, y, width, height-overlay);
				break;
			case NO_SLIDE:
			case COVER_SLIDE:
				parent.getComponent(0).setBounds( x, y+overlay-height, width, height);
				parent.getComponent(curIdx).setBounds( x, y+overlay, width, height-overlay);
				break;
		}
	}
}

public Dimension minimumLayoutSize ( Container parent ){
	return preferredLayoutSize( parent);
}

public Dimension preferredLayoutSize ( Container parent ){
	int nc = parent.getComponentCount();
	if ( nc > 0 )
		return parent.getComponent(0).getPreferredSize();
	return new Dimension( 100, 100);
}

public void removeLayoutComponent ( Component child ){
}

public void run() {
	while( true ) {
		synchronized ( this) {
			try { wait(); }
			catch ( InterruptedException _) {}
		}
		switch ( mode) {
		case PUSH_SLIDE:
			animate_push();
			break;
		case COVER_SLIDE:
			animate_cover();
			break;
		case COVER_SLIDE_DOWN:
			animate_cover_down();
			break;
		}
	}
}

public void setOverlay( int overlay) {
	this.overlay = overlay;
}

public boolean setView( int idx) {
	synchronized (this) {
		if ( active || (parent.getComponentCount() < idx) || (idx == curIdx) ) {
			return false;
		}
		active = true;
		curIdx = idx;
	}
	
	sizeView( curIdx);
	
	if ( mode == NO_SLIDE) {
		//not much to do, just size and make visible
		showCurrent();
	}
	else {
		Component cn = parent.getComponent( curIdx);
		cn.validate();

		if ( dbImage == null )
			updateImage();
			
		Graphics dbg = dbImage.getGraphics();
		//slides in from top with cover (over-)size, so translate for bottom alignment
		if ( (mode == PUSH_SLIDE) && (curIdx > 0) )
			dbg.translate( 0, overlay);
		cn.paint( dbg);
		dbg.dispose();

		synchronized (this) {
			notify();
		}
	}
	return true;
}

private void showCurrent() {
	Component c0 = parent.getComponent(0 );
	Component cc = parent.getComponent(curIdx );

	AWT.extension.disableRedraw( parent);
	
	hideAll();

	// Layout *must* sync treelock otherwise we risk deadlocks!!
	synchronized (parent.getTreeLock()) {
		layoutContainer( parent);
	}

	AWT.extension.enableRedraw( parent);

	if  ((curIdx > 0) && ( overlay > 0)) {
		c0.setVisible( true);
	}

	cc.setVisible( true);

	active = false;
}

private void sizeView ( int idx ){
	//prepare in-sliding view to be painted to image
	
	Insets in = parent.getInsets();
	int width = AWT.extension.getWidth( parent) - in.left - in.right;
	int height = AWT.extension.getHeight( parent) - in.top - in.bottom;

	if ( idx == 0 ) {
		parent.getComponent(idx).setSize( width, height);
	}
	else {
		parent.getComponent(idx).setSize( width, height-overlay);
	}
}

private void updateImage ( ) {
	int width = AWT.extension.getWidth( parent);
	int height = AWT.extension.getHeight( parent);
	Insets in = parent.getInsets();

	if ( dbImage != null )
		dbImage.flush();

	dbImage = parent.createImage( width-in.left-in.right, height-in.top-in.bottom); 
}

public boolean isActive() {
	return (active);
}

}
