package kaffe.awt.dialogs;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Toolkit;
import java.awt.Window;
import java.util.Vector;
import kaffe.awt.widgets.ItemList;
import kaffe.awt.widgets.ItemRender;
import kaffe.awt.widgets.ItemRenderEnv;

public class FontDialog
  extends BaseDialog
  implements ItemRender
{
	ItemList lst;
	ItemRenderEnv renderEnv = new ItemRenderEnv();

public FontDialog( String title, Window owner) {
	this( title, 5, 20, owner);
}

public FontDialog( String title, int min, int max, Window owner) {
	super( title, owner);

	lst = new ItemList( this);

	Vector v = loadFonts( min, max);
	lst.setContents( v);
	
	add( lst);
	
	lst.addActionListener( this);
}

public void drawItem( ItemRenderEnv env) {
	Font f = (Font)env.items.elementAt( env.curIdx);
	Graphics g = env.graph;
	g.setFont( f);

	FontMetrics fm = g.getFontMetrics();
	String s = f.getName() + "[" + f.getSize() + "]";

	int y0 = env.rect.y + (env.rect.height - fm.getHeight()) / 2 + fm.getAscent();
	g.drawString( s, 10, y0);
}

public int getItemHeight( ItemRenderEnv env){
	return 30;
}

public int getMaxWidth( ItemRenderEnv env){
	return 300;
}

public ItemRenderEnv getRenderEnv(){
	return renderEnv;
}

public String getSelection() {
	return lst.getSelectedItem();
}

protected Vector loadFonts( int min, int max) {
	String[] fnts = Toolkit.getDefaultToolkit().getFontList();
	Vector v = new Vector();

	for ( int i=0; i<fnts.length; i++ ) {
		for ( int sz = min; sz <= max; sz++ ) {
			v.addElement( new Font( fnts[i], Font.PLAIN, sz));
		}
	}


	return v;	
}

public static Font prompt( String title, Window owner) {
	return prompt( title, 5, 20, owner);
}

public static Font prompt( String title, int min, int max, Window owner) {
	FontDialog p = new FontDialog( title, min, max, owner);
	String s = p.process();
	
	if ( s != null ) {
		return Font.decode( s);
	}
	
	return null;
}
}
