package kaffe.awt.doc;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.util.Hashtable;
import org.w3c.dom.*;

public class ElementImpl
  extends NodeImpl
  implements Element
{
	String link;
	int fWidth;
	int fHeight;
	int border;
	Color borderClr;
	Hashtable attributes;

protected ElementImpl() {
}

/**
 * @dom
 */
public Node appendChild(Node newChild) {
	if ( newChild instanceof ElementImpl )
		((ElementImpl)newChild).setLink(link);
	return (super.appendChild(newChild));
}

public void endElement() {
}

/**
 * @dom
 */
public String getAttribute(String name) {
	return getAttributeNS( null, name);
}

/**
 * @dom
 */
public String getAttributeNS(String namespaceURI, String localName) {
	return ((attributes!=null) ? (String)attributes.get( localName) : null);
}

/**
 * @dom
 */
public Attr getAttributeNode(String name) {
	return getAttributeNodeNS( null, name);
}

public Attr getAttributeNodeNS(String namespaceURI, String localName) {
	return null;
}

/**
 *
 */
public DocumentImpl getDocument() {
	Container p;
	for ( p = getParent(); p != null; p = p.getParent() ) {
		if ( p instanceof DocumentImpl )
			return (DocumentImpl)p;
	}
	return null;
}

/**
 * @dom
 */
public NodeList getElementsByTagName(String name) {
	return getElementsByTagNameNS( null, name);
}

/**
 * @dom
 */
public NodeList getElementsByTagNameNS(String namespaceURI, String localName) {
	return null;
}

public Container getFrame() {
	Container p = getParent();
	for ( ; p.getParent() != null; p = p.getParent() );
	return p;
}

public String getLink() {
	return link;
}

public Dimension getPreferredSize() {
	int newWidth = fWidth;
	int newHeight = fHeight;

	if ( (fWidth<0) || (fHeight<0) ) {
		Dimension dp = getParent().getSize();
		if ( fWidth < 0)
			newWidth = -fWidth*dp.width/100;
		if ( fHeight < 0)
			newHeight = -fHeight*dp.height/100;
	}

	if ( (newWidth>0) && (newHeight>0) )
		return new Dimension( newWidth, newHeight);

	return super.getPreferredSize();
}

public int getRelativeHeight( Dimension parentDim) {
	if ( fHeight < 0 )
		return ( -fHeight*parentDim.height/100 );
	return ( getPreferredSize().height );
}

public int getRelativeWidth( Dimension parentDim) {
	if ( fWidth < 0 )
		return ( -fWidth*parentDim.width/100 );
	return ( getPreferredSize().width );
}

public DocumentImpl getRootDocument() {
	DocumentImpl d = null;

	for ( Container p = getParent(); p.getParent() != null; p = p.getParent() ) {
		if ( p instanceof DocumentImpl )
			d = (DocumentImpl) p;
	}

	return d;
}

/**
 * @dom
 */
public String getTagName() {
	return null;
}

public boolean hasFixedDims() {
	return ( (fWidth>0) && (fHeight>0) );
}

public boolean hasFixedHeight() {
	return ( fHeight>0 );
}

public boolean hasFixedWidth() {
	return ( fWidth>0 );
}

public boolean hasRelativeHeight() {
	return ( fHeight< 0 );
}

public boolean hasRelativeWidth() {
	return ( fWidth< 0 );
}

int initLayout( int offset, int x0, int x1) {
	return offset;
}

int intAttrValue( String key) {
	return intAttrValue( key, 0);
}

int intAttrValue( String key, int def) {
	String val = (String)attributes.get( key);
	if ( (val == null) || (val.length() == 0) )
		return (def);
	else if ( val.endsWith( "%") )
		return ( - Integer.parseInt( val.substring( 0, val.length()-1)));
	else
		return (Integer.parseInt( val));
}

/**
 * @dom
 */
public void normalize() {
}

public void paint( Graphics g) {
	super.paint( g);
	if ( border > 0 ) {
		paintBorder( g);
	}
}

void paintBorder( Graphics g) {
	Dimension d = getSize();
	g.setColor( (borderClr != null) ? borderClr : Color.black);
	for ( int i=0;i<border;i++) {
		g.drawRect( i, i, d.width-2*i-1, d.height-2*i-1);
	}
}

public int preferredHeight() {
	int prefHeight = fHeight;

	if (fHeight<0) {
		Dimension dp = getParent().getSize();
		prefHeight = -fHeight*dp.height/100;
	}

	return prefHeight;		
}

public int preferredWidth() {
	int prefWidth = fWidth;

	if (fWidth<0) {
		Dimension dp = getParent().getSize();
		prefWidth = -fWidth*dp.width/100;
	}

	return prefWidth;		
}

/**
 * @dom
 */
public void removeAttribute(String name) {
	removeAttributeNS( null, name);
}

/**
 * @dom
 */
public void removeAttributeNS(String namespaceURI, String localName) {
	if ( attributes != null )
		attributes.remove( localName);
}

/**
 * @dom
 */
public Attr removeAttributeNode(Attr oldAttr) {
	return null;
}

/**
 * @dom
 */
public void setAttribute(String name, String value) {
	setAttributeNS( null, name, value);
}

/**
 * @dom
 */
public void setAttributeNS(String namespaceURI, String localName, String value) {
	if ( localName.equalsIgnoreCase( "width") )
		fWidth = sizeValue( value);
	else if ( localName.equalsIgnoreCase( "height") )
		fHeight = sizeValue( value);
	else if ( localName.equalsIgnoreCase( "border") )
		border = sizeValue( value);
	else if ( localName.equalsIgnoreCase( "borderColor") ) {
		if( value.startsWith( "#")) {
			borderClr = new Color( (int)Long.parseLong( value.substring(1), 16));
		}
	}
	else if ( localName.equalsIgnoreCase( "font") ) {
		//format : <name>[-<style>[-<size>]
		Font fnt = Font.decode(value);
		setFont( fnt);
	}
	
	if ( attributes == null )
		attributes = new Hashtable();
			
	attributes.put( localName, value);
}

/**
 * @dom
 */
public Attr setAttributeNode(Attr newAttr) {
	return newAttr;
}

/**
 * @dom
 */
public Attr setAttributeNodeNS(Attr newAttr) {
	return newAttr;
}

protected void setBorderClr( String clr) {
	if ( (clr != null) && (clr.startsWith( "#"))) {
		borderClr = new Color( (int)Long.parseLong( clr.substring(1), 16));
	}
}

void setLink(String link) {
	if (link != null) {
		this.link = link;
		addMouseListener(LinkListener.getInstance());
		setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
	}
}

void setVisited() {
}

static int sizeValue( String dim) {
	int val = 0;
	try {
		if ( (dim == null) || (dim.length() == 0) )
			val = 0;
		else if ( dim.endsWith( "%") )
			val = - Integer.parseInt( dim.substring( 0, dim.length()-1));
		else
			val = Integer.parseInt( dim);
	}
	catch ( NumberFormatException _ ) {
		val = -1;
	}
		
	return val;
}
}
