package kaffe.awt.doc;

import com.pocketlinux.engine.Context;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import org.w3c.dom.*;

class ActionPad
  extends Component
{
	boolean pushed;
	String cmd;
	String label;
	FontMetrics fm;
	PopupMenu pops;

ActionPad() {
	this( null, null);
}

ActionPad( String cmd) {
	this( cmd, cmd);
}

ActionPad( String label, String cmd) {
	this.label = label;
	this.cmd = cmd;
	setFont( new Font("Helvetica", Font.PLAIN, 8) );
}

public void add( PopupMenu p) {
	super.add( p);
	pops = p;
	
	p.setFont( getFont() );
}

public Dimension getPreferredSize() {
	int w = (label != null) ? fm.stringWidth( label) + 5 : 0;
	int h = 4*fm.getHeight()/3;
	return new Dimension( w, h);
}

public void paint( Graphics g) {
	Dimension d = getSize();
	
	g.setColor( Color.lightGray);
	g.fill3DRect( 0, 0, d.width, d.height, !pushed);

	if ( label != null ) {
		g.setColor( Color.black);
		int x0 = (d.width-fm.stringWidth(label))/2;
		int y0 = d.height-(d.height-fm.getHeight())/2 - fm.getDescent();
		if ( pushed ) {
			x0-=1;
			y0-=1;
		}
		g.drawString( label, x0, y0);	
	}
}

public void setFont( Font fnt) {
	super.setFont( fnt);
	fm = getFontMetrics( fnt);
}
}

class HomePad
  extends ActionPad
{
public void paint( Graphics g) {
	Dimension d = getSize();
	int oo = 3;
	
	g.setColor( Color.lightGray);
	g.fill3DRect( 0, 0, d.width, d.height, !pushed);
	g.setColor( Color.black);
	g.fillOval( oo, oo, d.width-2*oo, d.height-2*oo); 
}
}

public class Navigator
  extends ElementImpl
  implements MouseListener
{
	HomePad home;
	PrevPad prev;
	NextPad next;
	VarPad vars;

public Navigator() {
	setLayout( null);
	
	add(home = new HomePad());
	home.cmd = "home";
	
	add(prev = new PrevPad());
	add(next = new NextPad());
	add(vars = new VarPad());

	home.addMouseListener( this);
	prev.addMouseListener( this);
	next.addMouseListener( this);
}

public ActionPad addActionPad( String label, String cmd) {
	return addActionPad( label, cmd, null);
}

public ActionPad addActionPad( String label, String cmd, PopupMenu p) {
	if ( ! containsLabel( label) ) {
		ActionPad ap = new ActionPad( label, cmd);
		ap.addMouseListener( this);
		vars.add( ap);
	
		if ( p != null )
			ap.add( p);
		
		return ap;
	}
	return null;
}

boolean containsLabel( String label) {
	int nc = vars.getComponentCount();
	for ( int i=0; i<nc; i++ ) {
		ActionPad ap = (ActionPad)vars.getComponent(i);
		if ( label.equals( ap.label) )
			return true;
	}
	return false;
}

public Dimension getPreferredSize() {
	Dimension d = home.getPreferredSize();
	return new Dimension( 100, d.height);
}

public void layout() {
	Dimension d = getSize();
	int w = d.height;
	
	home.setBounds( 0, 0, w, w);
	vars.setBounds( w, 0, d.width-3*w, w);
	prev.setBounds( d.width-2*w, 0, w, w);
	next.setBounds( d.width-w, 0, w, w);
}

public void mouseClicked( MouseEvent evt) {
	Object o = evt.getSource();
	
	if ( o instanceof ActionPad ) {
		ActionPad ap = (ActionPad)o;

		if ( ap.pops != null ) {
			Dimension d = ap.getSize();
			ap.pops.show( ap, 0, d.height );
		}
	}
}

public void mouseEntered( MouseEvent ect) {
}

public void mouseExited( MouseEvent evt ) {
}

public void mousePressed( MouseEvent evt) {
	Object o = evt.getSource();
	
	if ( o instanceof ActionPad ) {
		ActionPad ap = (ActionPad)o;
		ap.pushed = true;
		ap.repaint();
	}
}

public void mouseReleased( MouseEvent evt) {
	Object o = evt.getSource();
	
	if ( !( o instanceof ActionPad) )
		return;
		
	ActionPad ap = (ActionPad)o;
	ap.pushed = false;
	ap.repaint();

	if ( ap == prev ) {
		vars.offs = Math.max(0, vars.offs-1);
		vars.layout();
		vars.repaint();
	}
	else if ( ap == next ) {
		vars.offs = Math.min( vars.getComponentCount()-1, vars.offs+1);
		vars.layout();
		vars.repaint();
	}
	else if ( ap.cmd != null ){

		//document related stuff	
		//getOwnerDocument().perform(ap.cmd):

                // return to the initial page
                Context.loadInitialPage();
	}
}

public void setAttributeNS(String namespaceURI, String localName, String value) {
	if ( localName.equalsIgnoreCase( "action") )
		addActionPad( value, value);
	else
		super.setAttributeNS( namespaceURI, localName, value);
}
}

class NextPad
  extends ActionPad
{
public void paint( Graphics g) {
	Dimension d = getSize();
	int xa[] = new int[3];
	int ya[] = new int[3];
	int oo = 3;
	
	g.setColor( Color.lightGray);
	g.fill3DRect( 0, 0, d.width, d.height, !pushed);
	g.setColor( Color.black);
	
	xa[0]=d.width-oo; ya[0]=d.height/2;
	xa[1]=oo; ya[1]=d.height-oo;
	xa[2]=oo; ya[2]=oo;
	
	g.fillPolygon( xa, ya, 3); 
}
}

class PrevPad
  extends ActionPad
{
public void paint( Graphics g) {
	Dimension d = getSize();
	int xa[] = new int[3];
	int ya[] = new int[3];
	int oo = 3;
	
	g.setColor( Color.lightGray);
	g.fill3DRect( 0, 0, d.width, d.height, !pushed);
	g.setColor( Color.black);
	
	xa[0]=oo;	ya[0]=d.height/2;
	xa[1]=d.width-oo; ya[1]=oo;
	xa[2]=d.width-oo; ya[2]=d.height-oo;
	
	g.fillPolygon( xa, ya, 3); 
}
}

class VarPad
  extends Container
{
	int offs;
	int gap = 0;

VarPad() {
	setLayout( null);
}

public void layout() {
	int x0 = 0;
	int nc = getComponentCount();

	for ( int i=0; i<offs; i++ ) {
		Component c = getComponent(i);
		Dimension d = c.getPreferredSize();
		x0 -= d.width + gap;
	}
	
	Dimension d = getSize();
	for ( int i=0; i<nc; i++ ) {
		Component c = getComponent(i);
		Dimension d1 = c.getPreferredSize();
		c.setBounds( x0+gap, (d.height-d1.height)/2, d1.width, d1.height);
		x0 += d1.width + gap;
	}
}

public void paint( Graphics g) {
	Dimension d = getSize();
	g.setColor( Color.lightGray);
	g.fill3DRect( 0, 0, d.width, d.height, true);

	super.paint( g);
}
}
