package kaffe.awt.doc;

import java.awt.Component;
import java.awt.Container;
import java.util.Vector;
import org.w3c.dom.*;

public class SimpleNode
  extends Object
  implements Node, NamedNode
{
	String value;
	String name;
	Vector childs;
	Node parent;
	Component renderer;

public Node appendChild(Node newChild) throws DOMException {
	if ( childs == null ) {
		childs = new Vector();
	}
	
	childs.addElement( newChild);

	SimpleNode sn = (SimpleNode)newChild;
	sn.parent = this;
		
	//if newChild is a widget, detect AWT parent and add
	if ( sn.renderer != null ) { 
		getNextRenderer().add( sn.renderer);
	}
		
	return newChild;
}

public Node cloneNode(boolean deep) {
	try {
		return (Node)clone();
	}
	catch ( CloneNotSupportedException _) {
		return null;
	}
}

public NamedNodeMap getAttributes() {
	return null;
}

public NodeList getChildNodes() {
	return null;
}

public Node getFirstChild() {
	if ( childs.size() > 0 )
		return (Node)childs.firstElement();
	return null;
}

public Node getLastChild() {
	if ( childs.size() > 0 )
		return (Node)childs.lastElement();
	return null;
}

public String getLocalName() {
	return name;
}

public String getName() {
	return (name);
}

public String getNamespaceURI() {
	return null;
}

Container getNextRenderer() {
	for ( Node n=this; n != null; n = n.getParentNode() ) {
		if ( n instanceof SimpleNode ) {
			SimpleNode snc = (SimpleNode)n;
			if ( snc.renderer != null ) {
				return ((Container)snc.renderer);
			} 
		}
	}
	return (null);
}

public Node getNextSibling() {
	return null;
}

public String getNodeName() {
	return name;
}

public short getNodeType() {
	return 0;
}

public String getNodeValue() throws DOMException {
	return value;
}

public Document getOwnerDocument() {
	for ( Node n=this; n != null; n = n.getParentNode() ) {
		if ( n instanceof Document )
			return ((Document)n);
	}
	
	return (null);
}

public Node getParentNode() {
	return parent;
}

public String getPrefix() {
	return null;
}

public Node getPreviousSibling() {
	return null;
}

Container getRenderParent() {
	if ( renderer != null ) {
		for ( Node n=getParentNode(); n != null; n = n.getParentNode() ) {
			if ( n instanceof SimpleNode ) {
				SimpleNode snc = (SimpleNode)n;
				if ( snc.renderer != null ) {
					return ((Container)snc.renderer);
				} 
			}
		}
	}
	return (null);
}

public boolean hasChildNodes() {
	return ( childs.size() > 0);
}

public Node insertBefore(Node newChild, Node refChild) throws DOMException {
	int ri = childs.indexOf(refChild);
	
	if ( ri > -1 ) {
		childs.insertElementAt( newChild, ri);
		updateRenderTree();
		return (newChild);
	}
	
	return (null);
}

public Node removeChild(Node oldChild) throws DOMException {
	if ( childs.removeElement( oldChild) ) {
		updateRenderTree();
		return oldChild;
	}

	return (null);
}

public Node replaceChild(Node newChild, Node oldChild) throws DOMException {
	int ri = childs.indexOf(oldChild);
	
	if ( ri > -1 ) {
		childs.setElementAt( newChild, ri);
		updateRenderTree();
		return (newChild);
	}
	
	return (null);
}

public void setName(String nodeName){
	this.name = nodeName;
}

public void setNodeValue(String nodeValue) throws DOMException {
	this.value = value;
}

public void setPrefix(String prefix) throws DOMException {
}

public boolean supports(String feature, String version) {
	return false;
}

protected void updateRenderTree() {
	Container rp = getNextRenderer();
	if ( rp != null ) {
		int cs = childs.size();
		rp.removeAll();
		for ( int i=0; i<cs; i++ ) {
			SimpleNode sn = (SimpleNode)childs.elementAt(i);
			if ( sn.renderer != null) {
				rp.add( sn.renderer );
			}
		}
	}
}
}
