package kaffe.awt.doc;

import java.awt.*;
import org.w3c.dom.*;

public class Table
  extends SimpleElement
{
	TableCell currentCell;
	int cellpadding;
	int cellspacing;

class Render
  extends Container
{
Render () {
	setLayout( new TableLayout() );
}
}

public Table() {
	this( 0, 0, 0, 0, 0, null);
}

public Table( String width, String height, int cellpadding, int cellspacing) {
	this( sizeValue(width), sizeValue(height), cellpadding, cellspacing, 0, null);
}

public Table( String width, String height, int cellpadding, int cellspacing, int border) {
	this( sizeValue(width), sizeValue(height), cellpadding, cellspacing, border, null);
}

public Table( String width, String height, int cellpadding, int cellspacing, int border, String clrBorder) {
	this( sizeValue(width), sizeValue(height), cellpadding, cellspacing, border, clrBorder);
}

public Table( int width, int height, int cellpadding, int cellspacing) {
	this( width, height, cellpadding, cellspacing, 0, null);
}

public Table( int width, int height, int cellpadding, int cellspacing, int border, String clrBorder) {
	renderer = new Render()
	
	this.cellpadding = cellpadding;
	this.cellspacing = cellspacing;
	fWidth = width;
	fHeight = height;
	this.border	= border;
	setBorderClr( clrBorder);
}

public Node appendChild(Node newChild) throws DOMException {
	if ( ! (newChild instanceof TableRow) ) {
		throw new DOMException();
	}
	super.appendChild( newChild);
	((TableRow)newChild).row = childs.size()-1;
	
	return newChild;
}

int getCellHeight( TableCell tc) {
	return ((TableLayout)getLayout()).getCellHeight( this, tc);
}

int getCellWidth( TableCell tc) {
	return ((TableLayout)getLayout()).getCellWidth( this, tc);
}

int getColCount() {
	return ((TableLayout)getLayout()).widths.length;
}

int getRowCount() {
	return ((TableLayout)getLayout()).heights.length;
}

public void setAttributeNS(String namespaceURI, String localName, String value) {
	if ( localName.equalsIgnoreCase( "cellpadding") )
		cellpadding = sizeValue( value);
	else if ( localName.equalsIgnoreCase( "cellspacing") )
		cellspacing = sizeValue( value);
	else
		super.setAttributeNS( namespaceURI, localName, value);
}
}

class TableLayout
  implements LayoutManager
{
	int[] widths;
	int[] heights;
	boolean[] fixWidths;
	boolean[] fixHeights;

public void addLayoutComponent ( String name, Component child ) {
}

void dumpGrid () {
	int i;
	for ( i=0; i<widths.length; i++)
		System.out.print( "        " + widths[i] + "[" + fixWidths[i] + "]" );
	System.out.println();
	for ( i=0; i<heights.length; i++)
		System.out.println( heights[i]  + "[" + fixHeights[i] + "]");

}

int getCellHeight( Table t, TableCell tc) {
	int h = 0;
	if ( (fixHeights == null) || (! fixHeights[tc.row]) )
		return h;
	for ( int i=0; i<tc.rowspan; i++)
		h += heights[tc.row+i];
	if ( h > 0 )
		h += (tc.rowspan-1) * t.cellspacing;
	return h;
}

int getCellWidth( Table t, TableCell tc) {
	int w = 0;
	if ( (fixWidths == null) || (! fixWidths[tc.col]) )
		return w;
	for ( int i=0; i<tc.colspan; i++)
		w += widths[tc.col+i];
	if ( w > 0 )
		w += (tc.colspan-1) * t.cellspacing;
	return w;
}

void initGrid( Table t) {
	int nc = t.getComponentCount();
	int x0, y0, i;
	
	widths = new int[maxCols(t)];
	heights = new int[t.rows.size()];
	fixWidths = new boolean[widths.length];
	fixHeights = new boolean[heights.length];
	
	TableCell[][] grid = new TableCell[widths.length][heights.length];
	for ( y0=0; y0<heights.length; y0++ ) {
		for ( x0=0; x0<widths.length; x0++) {
			grid[x0][y0] = null;
		}
	}
	
	for ( i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		x0 = 0;
		try {
		for ( x0=0; (grid[x0][tc.row] != null) && (x0<widths.length); x0++);
		}
		catch ( Exception _ ) {
		System.out.println( x0 + " " + tc.row + " " + widths.length + " " + heights.length );
		}
		tc.col = x0;
		for ( int y=0; y<tc.rowspan; y++) {
			for ( int x=0; x<tc.colspan; x++) {
				grid[x+x0][y+tc.row] = tc;
			}
		}
	}

	// path 1: get maximum fixedsize cell dimensions
	for ( i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		int tpw = tc.preferredWidth();
		int tph = tc.preferredHeight();

		if ( tc.colspan == 1) {
			widths[tc.col] = Math.max( widths[tc.col], tpw);
			if ( tc.hasFixedWidth() ) {
				fixWidths[tc.col] = true;
			}
		}		
		if ( tc.rowspan == 1) {
			heights[tc.row] = Math.max( heights[tc.row], tph);
			if ( tc.hasFixedHeight() )
				fixHeights[tc.row] = true;
		}		
	}


	// path 2: get maximum preferredsize cell dimensions
	for ( i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		Dimension dc = tc.getPreferredSize();

		if ( tc.colspan == 1) {
			if ( !fixWidths[tc.col] )
				widths[tc.col] = Math.max( widths[tc.col], dc.width);
		}		
		if ( tc.rowspan == 1) {
			if ( !fixHeights[tc.row] ) {
				heights[tc.row] = Math.max( heights[tc.row], dc.height);
			}
		}		
	}


	int sumw = 0;
	int varw = 0;
	for( i=0; i<widths.length; i++) {
		if ( !fixWidths[i])
			varw++;
		sumw += widths[i];
	}

	int sumh = 0;
	int varh = 0;
	for( i=0; i<heights.length; i++) {
		if ( !fixHeights[i])
			varh++;
		sumh += heights[i];
	}
	
	// path 3: share extra size between non fixedsize cells
	Dimension dc = t.getSize();

	if ( dc.width == 0) 
		dc.width = t.preferredWidth();
	if ( dc.height == 0) 
		dc.height = t.preferredHeight();

	dc.width -= (widths.length+1)*t.cellspacing;
	dc.height -= (heights.length+1)*t.cellspacing;
	
	int pvw = (varw > 0) ? (dc.width - sumw) / varw : 0;
	int pvh = (varh > 0) ? (dc.height - sumh) / varh : 0;

	if ( pvw > 0 ) {
		for( i=0; i<widths.length; i++) {
			if (!fixWidths[i])
				widths[i] += pvw;
		}
	}
	
	if ( pvh > 0 ) {
		for( i=0; i<heights.length; i++) {
			if (!fixHeights[i])
				heights[i] += pvh;
		}
	}
}

void initGrid_1( Table t) {
	int nc = t.getComponentCount();
	int x0, y0, i;
	
	if ( widths == null ) {
		widths = new int[maxCols(t)];
		heights = new int[t.rows.size()];
		fixWidths = new boolean[widths.length];
		fixHeights = new boolean[heights.length];
	}
	else {
		for( y0=0; y0<heights.length; y0++) {
			heights[y0] = 0;
			fixHeights[y0] = false;
		}
		for( x0=0; x0<widths.length; x0++) {
			widths[x0] = 0;
			fixWidths[x0] = false;
		}
	}
	
	TableCell[][] grid = new TableCell[widths.length][heights.length];
	for ( y0=0; y0<heights.length; y0++ ) {
		for ( x0=0; x0<widths.length; x0++) {
			grid[x0][y0] = null;
		}
	}
	
	for ( i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		x0 = 0;
		try {
		for ( x0=0; (grid[x0][tc.row] != null) && (x0<widths.length); x0++);
		}
		catch ( Exception _ ) {
		System.out.println( x0 + " " + tc.row + " " + widths.length + " " + heights.length );
		}
		tc.col = x0;
		for ( int y=0; y<tc.rowspan; y++) {
			for ( int x=0; x<tc.colspan; x++) {
				grid[x+x0][y+tc.row] = tc;
			}
		}
	}

	// path 1: get maximum fixedsize cell dimensions
	for ( i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		int tpw = tc.preferredWidth();
		int tph = tc.preferredHeight();

		if ( tc.colspan == 1) {
			widths[tc.col] = Math.max( widths[tc.col], tpw);
			if ( tc.hasFixedWidth() ) {
				fixWidths[tc.col] = true;
			}
		}		
		if ( tc.rowspan == 1) {
			heights[tc.row] = Math.max( heights[tc.row], tph);
			if ( tc.hasFixedHeight() )
				fixHeights[tc.row] = true;
		}		
	}


	// path 2: get maximum preferredsize cell dimensions
	for ( i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		Dimension dc = tc.getPreferredSize();

		if ( tc.colspan == 1) {
			if ( !fixWidths[tc.col] )
				widths[tc.col] = Math.max( widths[tc.col], dc.width);
		}		
		if ( tc.rowspan == 1) {
			if ( !fixHeights[tc.row] ) {
				heights[tc.row] = Math.max( heights[tc.row], dc.height);
			}
		}		
	}


	int sumw = 0;
	int varw = 0;
	for( i=0; i<widths.length; i++) {
		if ( !fixWidths[i])
			varw++;
		sumw += widths[i];
	}

	int sumh = 0;
	int varh = 0;
	for( i=0; i<heights.length; i++) {
		if ( !fixHeights[i])
			varh++;
		sumh += heights[i];
	}
	
	// path 3: share extra size between non fixedsize cells
	Dimension dc = t.getSize();

	if ( dc.width == 0) 
		dc.width = t.preferredWidth();
	if ( dc.height == 0) 
		dc.height = t.preferredHeight();

	dc.width -= (widths.length+1)*t.cellspacing;
	dc.height -= (heights.length+1)*t.cellspacing;
	
	int pvw = (varw > 0) ? (dc.width - sumw) / varw : 0;
	int pvh = (varh > 0) ? (dc.height - sumh) / varh : 0;

	if ( pvw > 0 ) {
		for( i=0; i<widths.length; i++) {
			if (!fixWidths[i])
				widths[i] += pvw;
		}
	}
	
	if ( pvh > 0 ) {
		for( i=0; i<heights.length; i++) {
			if (!fixHeights[i])
				heights[i] += pvh;
		}
	}
}

public void layoutContainer ( Container parent ) {
	synchronized ( parent.getTreeLock() ) {
		Table t = (Table)parent;
		int cs = t.cellspacing;
		int nc = t.getComponentCount();
		int x0, y0, w, h, i1;

		initGrid( t);	
		//	dumpGrid();

		for ( int i=0; i<nc; i++) {
			TableCell tc = (TableCell)parent.getComponent( i);
			for ( i1=0, x0=cs; i1<tc.col; x0 += widths[i1++] + cs );
			for ( i1=0, y0=cs; i1<tc.row; y0 += heights[i1++] + cs);
			for ( i1=0, w=0; i1<tc.colspan; w += widths[tc.col + i1++]);
			for ( i1=0, h=0; i1<tc.rowspan; h += heights[tc.row + i1++]);

			w += cs*(tc.colspan-1);
			h += cs*(tc.rowspan-1);

			tc.setBounds( x0, y0, w, h);
			//		System.out.println( tc + " " + tc.getBounds() );
		}
	}
}

private int maxCols( Table t) {
	int nc = t.getComponentCount();
	int rowCols = 0;
	int curY = 0;
	int maxCols = 0;
	
	for ( int i=0; i<nc; i++) {
		TableCell tc = (TableCell)t.getComponent( i);
		if ( tc.row > curY ) {
			maxCols = Math.max( maxCols, rowCols);
			rowCols = 0;
			curY = tc.row;
		}
		rowCols += tc.colspan;
	}
	maxCols = Math.max( maxCols, rowCols);

	return maxCols;
}

public Dimension minimumLayoutSize ( Container parent ) {
	Dimension d = new Dimension();
	int cc = parent.getComponentCount();

	for ( int i=0; i<cc; i++) {
		Component c = parent.getComponent( i);
		Dimension dc = c.getMinimumSize();
		if ( dc.width > d.width )
			d.width = dc.width;
		if ( dc.height > d.height )
			d.height = dc.height;
	}
	
	return d;
}

public Dimension preferredLayoutSize ( Container parent ) {
	Table t = (Table)parent;
	int cs = t.cellspacing;
	Dimension d = new Dimension( cs, cs);
	
	initGrid( t);
		
	for ( int i=0; i<widths.length; i++)
		d.width += widths[i] + cs;
		
	for ( int i=0; i<heights.length; i++)
		d.height += heights[i] + cs;
	
	return d;
}

public void removeLayoutComponent ( Component child ) {
}
}
