package kaffe.awt.doc;

import java.awt.*;
import org.w3c.dom.Node;

public class TableCell
  extends SimpleElement
{
	int row;
	int col;
	int colspan;
	int rowspan;
	int halign = LEFT;
	int valign = CENTER;
	final public static int LEFT = 0;
	final public static int TOP = 0;
	final public static int CENTER = 1;
	final public static int RIGHT = 2;
	final public static int BOTTOM = 2;
	int absYOffset;

class Render
  extends Container
{
Render () {
	setLayout( new TableCellLayout() );
}

public Dimension getPreferredSize() {
	Dimension d = super.getPreferredSize();
	if ( fHeight < 0 )
		d.height += absYOffset;
	return d;
}
}

public TableCell() {
	this( 1, 1, 0, 0, -1, -1, null);
}

public TableCell( String clrBorder) {
	this( 1, 1, 0, 0, -1, -1, clrBorder);
}

public TableCell( int colspan, int rowspan) {
	this( colspan, rowspan, 0, 0, -1, -1, null);
}

public TableCell( int colspan, int rowspan, String width, String height) {
	this( colspan, rowspan, sizeValue(width), sizeValue(height), -1, -1, null );
}

public TableCell( int colspan, int rowspan, String width, String height, String clrBorder) {
	this( colspan, rowspan, sizeValue(width), sizeValue(height), -1, -1, clrBorder );
}

public TableCell( int colspan, int rowspan, int width, int height) {
	this( colspan, rowspan, width, height, -1, -1, null );
}

public TableCell( int colspan, int rowspan, int width, int height, int halign, int valign) {
	this( colspan, rowspan, width, height, halign, valign, null);
}

public TableCell( int colspan, int rowspan, int width, int height, int halign, int valign, String clrBorder) {

	renderer = new Render();
	
	this.colspan = colspan;
	this.rowspan = rowspan;
	
	if ( halign >= LEFT )
		this.halign = halign;
	if ( valign >= TOP )
		this.valign = valign;
	
//	setBorderClr( clrBorder);
	
	fWidth = width;
	fHeight = height;
}

Table getTable() {
	TableRow tr = (TableRow)parent;
	return ( ((Table)tr.parent) );
}

public int preferredHeight() {
	int ph = 0;
	
	//we got a fixed pixel size, just use it
	if ( fHeight > 0 ) {
		ph = fHeight;
	}
	else {
		Table t = getTable();
		//relative parent size
		if ( fHeight < 0 ) {
			ph = super.preferredHeight();
			ph -= t.cellspacing * (t.getRowCount()-rowspan+2) - absYOffset;
		}
		//best guess from table heights
		else {
			ph = t.getCellHeight( this);
		}
	}
	
	return ph;
}

public int preferredWidth() {
	int pw = 0;
	
	if ( fWidth > 0 ) {
		pw = fWidth;
	}
	else {
		Table t = getTable();
		if ( fWidth < 0 ) {
			pw = super.preferredWidth();
			pw -= t.cellspacing * (t.getColCount()-colspan+2);
		}
		else {
			pw = t.getCellWidth( this);
		}
	}
	
	return pw;
}

public void setAttributeNS(String namespaceURI, String localName, String value) {
	if ( localName.equalsIgnoreCase( "colspan") )
		colspan = sizeValue( value);
	else if ( localName.equalsIgnoreCase( "rowspan") )
		rowspan = sizeValue( value);
	else if ( localName.equalsIgnoreCase( "halign") ){
		halign = sizeValue( value);
		if ( halign == -1 ) {
			if ( value.equalsIgnoreCase( "center") )
				halign = CENTER;
			else if ( value.equalsIgnoreCase( "right") )
				halign = RIGHT;
			else
				halign = LEFT;
		}
	}
	else if ( localName.equalsIgnoreCase( "valign") ){
		valign = sizeValue( value);
		if ( valign == -1 ) {
			if ( value.equalsIgnoreCase( "top") )
				valign = TOP;
			else if ( value.equalsIgnoreCase( "bottom") )
				valign = BOTTOM;
			else
				valign = CENTER;
		}
	}
	else
		super.setAttributeNS( namespaceURI, localName, value);
}

public String toString() {
	return ( getClass().getName() + "[" + col + "," + row + "]["
			+ colspan + "x" + rowspan +"]");
}
}

class TableCellLayout
  implements LayoutManager
{
public void addLayoutComponent ( String name, Component child ) {
}

private int alignRow ( TableCell tc, int start, int end, int width, int y0, Dimension d ) {
	int cp = tc.getTable().cellpadding;
	int mlh = 0;
	int x0, pe;
	
	switch ( tc.halign ) {
		case tc.CENTER:
			x0 = (d.width-width)/2;
			break;
		case tc.RIGHT:
			x0 = d.width-width-cp;
			break;
		default:
			x0 = cp;
	}
	
	for ( int i=start; i<end; i++) {
		ElementImpl e = (ElementImpl)tc.getComponent( i);
		pe = e.getRelativeHeight(d);
		mlh = Math.max( mlh, pe);
	}

	for ( int i=start; i<end; i++) {
		ElementImpl e = (ElementImpl)tc.getComponent( i);
		pe = e.getRelativeWidth(d);
		e.setBounds( x0, y0, pe, mlh);
		//kaffe.lang.Application.debug( e.toString() );
		x0 += pe;
	}

	
	return mlh;
}

int getContentsHeight( TableCell tc, Dimension d ) {
	int cp = tc.getTable().cellpadding;
	int cc = tc.getComponentCount();
	int x0 = cp;
	int s = 0;
	int height=0;
	
	for ( int i=0; i<cc; i++) {
		ElementImpl e = (ElementImpl)tc.getComponent( i);
		int epw = e.getRelativeWidth(d);

		if ( x0 + epw > d.width-cp ) {
			height += maxLineHeight( tc, s, i);
			x0 = cp;
			s = i;
		}
		x0 += epw;
	}
	
	height += maxLineHeight( tc, s, cc);
	
	return height;
}

public void layoutContainer ( Container parent ) {
	synchronized ( parent.getTreeLock() ) {
		TableCell tc = (TableCell)parent;
		Dimension d = tc.getSize();
		int cp = tc.getTable().cellpadding;
		int cc = tc.getComponentCount();
		int x0 = cp;
		int s = 0;
		int y0;

		switch( tc.valign) {
		case tc.CENTER:
			y0 = Math.max( (d.height-getContentsHeight(tc,d))/2 , cp);
			break;
		case tc.BOTTOM:
			y0 = d.height-cp-getContentsHeight(tc,d);
			break;
		default:
			y0 = cp;
		}

		for ( int i=0; i<parent.getComponentCount(); i++) {
			ElementImpl e = (ElementImpl)tc.getComponent( i);
			int epw = e.getRelativeWidth( d);

			//this may modify childcount
			e.initLayout( x0, cp, d.width-cp);

			if ( x0 + epw > d.width-cp ) {
				y0 += alignRow( tc, s, i, x0-cp, y0, d );
				x0 = cp;
				s = i;
			}
			x0 += epw;
		}

		alignRow( tc, s, parent.getComponentCount(), x0-cp, y0, d);
	}
}

private int maxLineHeight ( TableCell tc, int start, int end ) {
	int mlh = 0;
	
	for ( int i=start; i<end; i++) {
		ElementImpl e = (ElementImpl)tc.getComponent( i);
		int eph = e.getPreferredSize().height;
		mlh = Math.max( mlh, eph);
	}

	return mlh;
}

private int minWidth ( TableCell tc) {
	int mlw = 0;
	int wmin = 0;
	
	for ( int i=0; i<tc.getComponentCount(); i++) {
		ElementImpl e = (ElementImpl)tc.getComponent( i);
		e.initLayout( 0, 0, 0);
		if ( e instanceof Break ) {
			wmin = Math.max( wmin, mlw+1);
			mlw = 0;
		}
		else {
			Dimension de = e.getPreferredSize();
			mlw += de.width;
		}
	}
	
	wmin = Math.max( wmin, mlw);
	
	return (wmin);
}

public Dimension minimumLayoutSize ( Container parent ) {
	return new Dimension();
}

public Dimension preferredLayoutSize ( Container parent ) {
	TableCell.Render tc = (TableCell.Render)parent;
	Table t = tc.getTable();
	int cp = t.cellpadding;
	int mlh = 0;
	int x0 = 0;
	int epw, eph;
	
	Dimension d = new Dimension( tc.preferredWidth(), tc.preferredHeight());

	if ( d.width == 0 ) {
		d.width = minWidth( tc) + 2*cp;
	}

	for ( int i=0; i<parent.getComponentCount(); i++) {
		ElementImpl e = (ElementImpl)parent.getComponent( i);

		e.initLayout( x0, cp, d.width-cp);
			
		Dimension de = e.getPreferredSize();
		epw = de.width;
		eph = de.height;

		if ( x0 + epw > d.width -2*cp ) {
			d.height += mlh;
			x0 = epw;
			mlh = eph;
		}
		else {
			x0 += epw;
			mlh = Math.max( mlh, eph);
		}
	}

	d.height += mlh + 2*cp;
//	d.height = Math.max( d.height, tc.preferredHeight());

	return d;
}

public void removeLayoutComponent ( Component child ) {
}
}
