package kaffe.awt.doc;

import java.awt.*;
import java.util.Vector;
import org.w3c.dom.*;

public class TextImpl
  extends ElementImpl
  implements Text
{
	char[] text;
	boolean visited;
	FontMetrics fm;
	Vector rows = new Vector();

class TextRow
  extends ElementImpl
{
	int start;
	int len;

TextRow( int start, int len) {
	this.start = start;
	this.len = len;

	this.setFont( TextImpl.this.getFont());
	this.setForeground( TextImpl.this.getForeground() );
	
	this.fHeight = fm.getHeight();
	this.fWidth = fm.charsWidth( text, start, len);

	this.setLink(TextImpl.this.link);
}

public void paint( Graphics g) {
	int y0 = this.fHeight - fm.getDescent();

	g.drawChars( text, start, len, 0, y0);
	if ( this.link != null ) {
		g.setColor( Color.black); //visited ? BodyAttributes.vLink() : BodyAttributes.link() );
		g.drawLine( 0, y0, this.fWidth, y0);
	}
	
//	g.setColor( Color.blue);
//	g.drawRect( 0, 0, this.fWidth-1, this.fHeight-1);
}

void setVisited() {
	TextImpl.this.setVisited();
}
}

public TextImpl() {
	this("");
}

public TextImpl( String text) {
	this.text = text.toCharArray();
	setFont( FontAttributes.getFont() );
	setForeground( FontAttributes.getColor() );
	
	fWidth = 0;
	fHeight = 0;
}

public void appendData(String arg) throws DOMException {
	int al = arg.length();
	char[] na = new char[text.length+al];
	
	System.arraycopy( text, 0, na, 0, text.length);
	arg.getChars(0, arg.length(), na, text.length);
	
	text = na;
}

public void deleteData(int offset, int count) throws DOMException {
	char[] na = new char[text.length-count];
	
	System.arraycopy( text, 0, na, 0, offset);		
	System.arraycopy( text, offset+count, na, offset, text.length-(offset+count));
	
	text = na;
}

public String getData() {
	return new String( text);
}

public int getLength() {
	return text.length;
}

public Dimension getPreferredSize() {	
	return new Dimension( 0, 0);
}

int initLayout( int offset, int x0, int x1) {
	removeRows();
	
	if ( offset >= x1-x0 )
		offset = 0;

	offset = makeRows( offset, x0, x1);
	insertRows();

	return offset;
}

public void insertData(int offset, String arg) throws DOMException {
	int al = arg.length();
	char[] na = new char[text.length+al];
	
	System.arraycopy( text, 0, na, 0, offset);
	arg.getChars(0, al, na, offset);
	System.arraycopy( text, offset, na, offset+al, text.length-offset);
	
	text = na;
}

void insertRows() {
	Container p = getParent();
	if (p instanceof AnchorImpl) {
		p = p.getParent();
	}
	int cc = p.getComponentCount();
	int idx;
	
	for ( idx=0; idx<cc;) {
		if ( p.getComponent(idx++) == this )
			break;
	}
	
	int rs = rows.size();
	for ( int i=0; i<rs; i++) {
		p.add( (TextRow)rows.elementAt( i), idx+i);
	}
	
}

private int makeRows( int offset, int x0, int x1) {
	/* break the whole thing into visible lines:
	 * x0: left cell limit
	 * x1: right cell limit
	 * offset: rendering offset from x0
	*/
	int x, s, e;
	
	if ( x1 == 0 ) {
		//no cell width to render in, guess just single line
		rows.addElement( new TextRow( 0, text.length) );
		return (offset + fm.charsWidth( text, 0, text.length));
	}
	
	x = x0 + offset;
	s = 0;
	e = -1;
	
	for ( int i=0; i<text.length; i++) {
		//dont know how to handle, guess space
		if ( text[i] == '\n' )
			text[i] = ' ';
			
		int cw = fm.charWidth( text[i]);
		if ( text[i] == ' ' || text[i] == '-'){
			e = i;
			x += cw;
			continue;
		}
		if (x+cw > x1) {
			if (e > 0) {
				//kaffe.lang.Application.debug( x + " " + cw + " " + x1 + " " + s + " " + e);
				rows.addElement( new TextRow( s, e-s+1) );
				x = x0;
				i = e;
				s = e+1;
				e = -1;
			}
			else {
				//row cannot break to fit in width
				//kaffe.lang.Application.debug( s + " " + x + " " + x1 + " " + new String(text));
				x = x0;
				i--;
			}
		}
		else {
			x += cw;
		}
	}

	if ( x > x0) {
		rows.addElement( new TextRow( s, text.length-s));
	}
	
	return x-x0;
}

public void paint( Graphics g) {	
}

void removeRows() {
	Container p = getParent();
	if (p instanceof AnchorImpl) {
		p = p.getParent();
	}
	int rs = rows.size();
	for ( int i=0; i<rs; i++) {
		p.remove( (TextRow)rows.elementAt( i));
	}
	rows.removeAllElements();
}

public void replaceData(int offset, int count, String arg) throws DOMException {
	deleteData( offset, count);
	insertData( offset, arg);
}

public void setData(String data) {
	this.text = data.toCharArray();
}

public void setFont( Font fnt) {
	super.setFont( fnt);
	fm = getFontMetrics( fnt);
}

void setVisited() {
	if ( ! visited ) {
		int rs = rows.size();
		visited = true;
		for ( int i=0; i<rs; i++ ) {
			((TextRow)rows.elementAt(i)).repaint();
		}
	}
}

public Text splitText(int offset) throws DOMException {
	return null;
}

public String substringData(int offset, int count) throws DOMException {
	return new String(text, offset, count);
}
}
