package kaffe.awt.player;

import java.awt.AWTError;
import java.awt.AWTEvent;
import java.awt.AWTEventMulticaster;
import java.awt.AWTException;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.io.File;
import java.net.URL;
import kaffe.awt.AWT;
import kaffe.awt.flash.FlashEvent;
import kaffe.util.Ptr;

public class FlashPlayer
  extends Component
  implements Runnable
{
	ActionListener aListener;
	Ptr nativeData;
	Graphics rgr;
	int flags;
	Object source;
	final static int INITIALIZED = 0x01;
	final static int WINDOW_SET = 0x02;
	final static int SOURCE_SET = 0x04;
	final static int GRAPHICS_SET = 0x08;
	final static int STARTED = 0x10;
	final static int RUNNING = 0x20;
	final static int READY_WAITING = 0x100;
	final static int PLAYING = 0x200;
	final static int SUSPENDED = 0x400;
	final static int STOPPED = 0x800;
	final static int SOURCE_CHANGED = 0x1000;
	final static int PAINTED = 0x2000;
	final static int TRANSPARENT = 0x1000000;
	final static int USE_SOUND = 0x2000000;
	final static int FLUSH_ON_STOP = 0x4000000;

public FlashPlayer () {
	this( false, false, true, true);
}

public FlashPlayer ( boolean sendInput ) {
	this( sendInput, false, true, true);
}

public FlashPlayer ( boolean sendInput, boolean isTransparent ) {
	this( sendInput, isTransparent, true, true);
}

public FlashPlayer ( boolean sendInput, boolean isTransparent, boolean enableSound, boolean flushOnStop ) {

	if ( isTransparent ) {
		flags |= TRANSPARENT;
	}

	if ( enableSound ) {
		flags |= USE_SOUND;
	}
	
	if ( flushOnStop ) {
		flags |= FLUSH_ON_STOP;
	}
	
	if ( sendInput ) {
		enableMouseEvents();
		enableMotionEvents();
		enableKeyEvents();
	}
	
	if ( ! initNative() ){
		throw new AWTError( "can't initialize native Flash object");
	}
}

public void addActionListener ( ActionListener a) {
	aListener = AWTEventMulticaster.add( aListener, a);
}

public synchronized void addNotify () {
	super.addNotify();
	
	setGraphics();	
}

public synchronized void blank () {

	if ( (flags & TRANSPARENT) == 0 ) {
		Graphics g = getGraphics();
		if ( g != null ) {
			AWT.extension.paintBackground( this, g);
			g.dispose();
		}
	}
	else {
		flaBlank( nativeData);
	}
}

public void enableKeyEvents () {
	enableEvents( AWTEvent.KEY_EVENT_MASK);
}

public void enableMotionEvents () {
	enableEvents( AWTEvent.MOUSE_MOTION_EVENT_MASK);
}

public void enableMouseEvents () {
	enableEvents( AWTEvent.MOUSE_EVENT_MASK);
}

native void flaBlank ( Ptr data );

native void flaDispose ( Ptr data );

native synchronized void flaDrawAll ( Ptr graphicsData );

native int flaGetHeight ( Ptr data );

native int flaGetWidth ( Ptr data );

native Ptr flaInit ( Object toolkitLock );

native void flaInitSound ( Ptr data );

native synchronized void flaInputEvent ( Ptr data, int type, int x, int y, int key );

native int flaLoadSource ( Ptr data );

native int flaPlay ( Ptr data );

native void flaResume ( Ptr data );

native void flaRewind ( Ptr data );

native int flaSetFileSource ( Ptr data, String fname );

native void flaSetGraphics ( Ptr data, Ptr graphicsData );

native int flaSetWindow ( Ptr data, int xOff, int yOff, int width, int height, boolean isTransparent );

native boolean flaStop ( Ptr data );

native boolean flaSuspend ( Ptr data );

void flashCallback ( String cmd, String arg ) {
	FlashEvent e = new FlashEvent( this, cmd, arg);
	getToolkit().getSystemEventQueue().postEvent( e);
}

public void flush () {
	// clean up all native data

	if ( nativeData != null ) {
		if ( (flags & RUNNING) != 0 ){
			stop();
		}
		
		flaDispose( nativeData);
		nativeData = null;
		flags &= ~(INITIALIZED | WINDOW_SET | GRAPHICS_SET | SOURCE_SET);
	}
}

public int getFlashHeight () {
	return flaGetHeight( nativeData);
}

public int getFlashWidth () {
	return flaGetWidth( nativeData);
}

public synchronized void hide () {
	// make sure the animation is suspended
	if ( (flags & RUNNING) != 0 ){
		suspend();
	}
	
	super.hide();
}

boolean initNative () {	
	// do whatever native initialization (not load or play) can be done

	if ( (flags & INITIALIZED) == 0 ){
		nativeData = flaInit( Toolkit.class);
		
		if ( nativeData != null ){
			flags |= INITIALIZED;
		}
		else {
			return false;
		}
		
		if ( (flags & USE_SOUND) != 0 ){
			flaInitSound( nativeData);
		}
	}
	
	if ( ((flags & SOURCE_SET) == 0) && (source != null) ){
		if ( source instanceof String ) {
			setSource( (String) source);
		}
		else if ( source instanceof URL ){
			setSource( (URL) source);
		}
	}
		
	if ( ((flags & WINDOW_SET) == 0) && (AWT.extension.getWidth(this) > 0) ){
		setWindow();
	}

	if ( (flags & GRAPHICS_SET) == 0 ){
		setGraphics();
	}
	
	return true;
}

public boolean isFocusTraversable () {
	return true;
}

public boolean isRunning () {
	return ((flags & RUNNING) != 0);
}

public void load () {
	// load from another thread, but don't play yet

	if ( (flags & (SOURCE_SET|SOURCE_CHANGED)) == (SOURCE_SET|SOURCE_CHANGED) ){
		Thread t = new Thread( this);
		t.setDaemon( true);
		t.start();
	}
}

public void loadSource () {
	boolean loaded = false;

	synchronized ( this ) {
		if ( (flags & (SOURCE_SET|SOURCE_CHANGED)) == (SOURCE_SET|SOURCE_CHANGED) ){
			flaLoadSource( nativeData);
			flags &= ~SOURCE_CHANGED;
			loaded = true;
		}
	}

	if ( loaded && (aListener != null) ){
		aListener.actionPerformed( new FlashEvent( this, FlashEvent.FS_LOADED));
	}
}

public synchronized void paint ( Graphics g ) {
	if ( (flags & (PLAYING|STARTED)) == STARTED ){
		flags |= PAINTED;
		start();
	}
	else {
		if ( nativeData != null ) {
			flaDrawAll( nativeData);
		}
	}
}

protected void processActionEvent ( ActionEvent event ) {
	if ( aListener != null ) {
		aListener.actionPerformed( event);
	}
}

protected void processKeyEvent ( KeyEvent event ) {
	super.processKeyEvent( event);
	
	if ( (flags & RUNNING) != 0 ) {
		flaInputEvent( nativeData, event.getID(), 0, 0, event.getKeyCode());
	}
}

protected void processMouseEvent ( MouseEvent event ) {
	super.processMouseEvent( event);

	if ( (flags & RUNNING) != 0 ) {
			flaInputEvent( nativeData, event.getID(), event.getX(), event.getY(), event.getModifiers());
	}
}

protected void processMouseMotionEvent ( MouseEvent event ) {
	super.processMouseMotionEvent( event);

	if ( (flags & RUNNING) != 0 ) {
		flaInputEvent( nativeData, event.getID(), event.getX(), event.getY(), event.getModifiers());
	}
}

public void removeActionListener ( ActionListener a) {
	aListener = AWTEventMulticaster.remove( aListener, a);
}

public synchronized void removeNotify () {

	if ( (flags & RUNNING) != 0 ){
		stop();
	}

	super.removeNotify();
	
	if ( nativeData != null ){		
		flaDispose( nativeData);
		nativeData = null;

		// reset all but the state flags
		flags &= (TRANSPARENT | USE_SOUND | FLUSH_ON_STOP);		
	}
	
	if ( rgr != null ) {
		rgr.dispose();
		rgr = null;
	}
}

public void reshape ( int x, int y, int width, int height ) {
	int oldWidth = AWT.extension.getWidth( this);
	int oldHeight = AWT.extension.getHeight( this);
	boolean isPlaying = ((flags & PLAYING) != 0);

	super.reshape( x, y, width, height);

	if ( (width != oldWidth) || (height != oldHeight) ){
		if ( isPlaying ){
			// this is the sledgehammer approach, restart the whole thing from scratch
			stop();
		}
		
		if ( (width > 0) && (height > 0) ){
			setWindow();
			
			if ( isPlaying ){
				// we will get a subsequent paint, which will (re)start us
				flags |= STARTED;
			}
		}
	}	
}

public synchronized void resume () {
	if ( ((flags & SUSPENDED) != 0) && isShowing() ) {
		flags &= ~SUSPENDED;
		flaResume( nativeData);
	}
}

public synchronized void rewind () {
	if ( nativeData != null ){
		flaRewind( nativeData);
		blank();
	}
}

public synchronized void run () {
	if ( (flags & (RUNNING|SOURCE_SET)) != SOURCE_SET ){
		// don't start twice, or go off without a source
		return;
	}
	flags |= RUNNING;

	// this is what we can do prior to getting visible
	initNative();
	if ( (flags & (SOURCE_CHANGED)) != 0 ){
		loadSource();
	}

	// maybe we just have been requested to load so far, wait until we get started
	while ( (flags & (STARTED|PAINTED)) != (STARTED|PAINTED) ){
		if ( (flags & STOPPED) != 0 ) {
			// outch, somebody stopped us while we were waiting
			flags &= ~RUNNING;
			return;
		}

		try {
			flags |= READY_WAITING;
			wait();
			flags &= ~READY_WAITING;
		}
		catch ( InterruptedException _ ) {}
	}

	// we should get here just in case we are visible and ready to run, but
	// make sure we've got propper (re)initialization
	initNative();

	if ( aListener != null ){
		aListener.actionPerformed( new FlashEvent( this, FlashEvent.FS_STARTED));
	}

	flags |= PLAYING;
	flaPlay( nativeData);
	flags &= ~(RUNNING | PLAYING);
	
	if ( aListener != null ){
		aListener.actionPerformed( new FlashEvent( this, FlashEvent.FS_STOPPED));	
	}
}

void setGraphics () {
	if ( (flags & GRAPHICS_SET) == 0){
		if ( rgr == null ){
			rgr = AWT.extension.getLinkedGraphics( this);
		}
		
		if ( ((flags & INITIALIZED) != 0) && (rgr != null) ){
			flaSetGraphics( nativeData, AWT.extension.getNativeData( rgr));
			
			flags |= GRAPHICS_SET;
		}
	}
}

public synchronized void setSource ( String fname ) {
	boolean isRunning = ((flags & RUNNING) != 0);
	int ret;

	source = fname;

	if ( (flags & INITIALIZED) == 0 ){
		// might have been flushed
		initNative();
	}
	
	if ( isRunning ) {
		stop();
	}
		
	ret = flaSetFileSource( nativeData, fname);
	if ( ret > 0 ) {
		flags |= (SOURCE_SET | SOURCE_CHANGED);
	}
		
	if ( isRunning ) {
		start();
	}
}

public synchronized void setSource ( URL url ) {
	if ("file".equals(url.getProtocol())) {
		setSource(url.getFile());
	}
	else {
		throw new kaffe.util.NotImplemented();
	}
}

public void setTransparent ( boolean isTransparent ) {

	boolean running = isRunning();

	if ( isTransparent != ((flags & TRANSPARENT) != 0) ){
		if ( (flags & INITIALIZED) != 0 ){
			if ( running ){
				stop();
			}
	
			flush();
		}

		if ( isTransparent ) {
			flags |= TRANSPARENT;
		}
		else {
			flags &= ~TRANSPARENT;
		}
	
		if ( running ){
			start();
		}
	}
}

void setWindow () {
	int ret;
	
	if ( (flags & INITIALIZED) == 0 ){
		if ( !initNative() ) return;
	}
	
	if ( (flags & WINDOW_SET) == 0 ){
		Insets in = AWT.extension.getInsets( this);
		int    w  = AWT.extension.getWidth( this);
		int    h  = AWT.extension.getHeight( this);

		int    cw = w - (in.left + in.right);
		int    ch = h - (in.top + in.bottom);

		if ( rgr != null ) {
			rgr.setClip( in.left, in.top, cw, ch);
		}

		ret = flaSetWindow( nativeData, in.left, in.top, cw, ch, ((flags & TRANSPARENT) != 0));	
		if ( ret > 0 ){
			flags |= WINDOW_SET;
		}
	}
}

public synchronized void show () {
	super.show();

	// if it was suspended, resume
	if ( (flags & SUSPENDED) != 0 ){
		flaDrawAll( nativeData);
		resume();
	}	
}

public synchronized void start () {
	if ( (flags & SUSPENDED) != 0 ){
		resume();
		return;
	}

	if ( !initNative() ) return;

	// remember state in case we can't run yet
	flags |= STARTED;
	flags &= ~STOPPED;

	if ( isShowing() ) {
		if ( (flags & (TRANSPARENT|PAINTED)) == TRANSPARENT ){
			// make sure we first get painted, or we can't save our background
			repaint();
			return;
		}

		if ( (flags & READY_WAITING) != 0 ){
			// maybe we're already loaded and just wait for being unleashed
			notify();
		}
		else {
			if (  (flags & (WINDOW_SET|SOURCE_SET)) == (WINDOW_SET|SOURCE_SET) ){		
				Thread t = new Thread( this);
				t.setDaemon( true);
				t.start();
			}
		}
	}
}

public synchronized void stop () {
	if ( (flags & (RUNNING|STOPPED)) == RUNNING ){
		flags |= STOPPED;
		if ( flaStop( nativeData) ){
			flags &= ~(RUNNING | PLAYING | PAINTED | STARTED);
			
			if ( ((flags & FLUSH_ON_STOP) != 0) ) {
				flush();
			}
			
			repaint();
		}
	}
}

public synchronized void suspend () {
	if ( (flags & RUNNING) != 0 ){
		if ( flaSuspend( nativeData) ){
			flags |= SUSPENDED;
		}	
	}
}

public String toString () {
	StringBuffer sb = new StringBuffer(100);
	
	sb.append(super.toString());
	sb.append( " [");
	
	if ( (flags & INITIALIZED) != 0 ) sb.append( "init ");
	if ( (flags & WINDOW_SET) != 0 ) sb.append( "wnd_set ");
	if ( (flags & SOURCE_SET) != 0 ) sb.append( "src_set ");
	if ( (flags & SOURCE_CHANGED) != 0 ) sb.append( "src_changed ");
	if ( (flags & GRAPHICS_SET) != 0 ) sb.append( "gr_set ");
	if ( (flags & STARTED) != 0 ) sb.append( "started ");
	if ( (flags & RUNNING) != 0 ) sb.append( "running ");
	if ( (flags & SUSPENDED) != 0 ) sb.append( "susp ");
	if ( (flags & STOPPED) != 0 ) sb.append( "stopped ");
	if ( (flags & READY_WAITING) != 0 ) sb.append( "ready_wait ");
	if ( (flags & PAINTED) != 0 ) sb.append( "painted ");
	if ( (flags & TRANSPARENT) != 0 ) sb.append( "transp ");	

	sb.append("]");
	
	return sb.toString();
}
}
