package kaffe.awt.player;

import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import kaffe.awt.AWT;
import kaffe.io.RtpMediaStream;
import kaffe.util.Ptr;

public class PluginPlayer
  extends Component
{
	int flags;
	String source;
	ComponentAdapter ca;
	Ptr nativeData;
	String lib;
	Stream currentStream;
	static Hashtable plugins = new Hashtable();
	final static int PICTURE_SET = 0x01;
	final static int SOURCE_SET = 0x02;
	final static int STARTED = 0x04;
	final static int ZOOM_PICTURE = 0x08;
	final static int PLAYING = 0x10;
	final static String PLG_SEARCH_PROP = "KNSPluginPath";

private class Stream
  implements Runnable
{
	URL url;
	Thread thread;
	boolean running;
	boolean done;
	final static int BUFFER_SIZE = 4096;

Stream( String url) {
	try {
		if ( url.startsWith( "/") ) {
			url = "file:" + url;
		}
//		System.err.println( "url: " + url);

		this.url = new URL( url);
		thread = new Thread( this);
		
		//would block ipl input otherwise ( esp. when streaming over rtsp )
		thread.setPriority( Thread.MIN_PRIORITY);

		thread.start();
	}
	catch ( MalformedURLException _) {
		_.printStackTrace();
	}
	catch ( IOException _) {
		_.printStackTrace();
	}
}

public void run() {
	int bw, br;
	InputStream stream;
	URLConnection con;
	byte[] buffer = new byte[BUFFER_SIZE];
	
	if ( ! running ) {
		//wait for a start request
		synchronized( this) {
			try {this.wait();}
			catch ( InterruptedException _) {
			}
		}
	}
	
	bw = -1;

	try {
		con = url.openConnection();
		stream = con.getInputStream();
		
		//be aware to offset header bytes 
		int off = (stream instanceof RtpMediaStream) ? RtpMediaStream.HEADER_LEN : 0;   

		do {
			br = stream.read( buffer, 0, BUFFER_SIZE);
			if ( (br > 0) && running && (nativeData != null) ) {
				bw = write0( nativeData, buffer, off, br);
			}
		} while ( running && (bw > 0) && (br > 0));
		
	 	//clean up connection	
		stream.close();
		if ( con instanceof HttpURLConnection ) {
			((HttpURLConnection)con).disconnect();
		}

	}
	catch ( Exception _ ) {
		System.err.println( "PluginPlayer cannot run " + url);
		_.printStackTrace();
	}

	if ( ! running ) {
		//this is from a stop request which blocks on it
		synchronized( this) {
			this.notify();
		}
	}
	else {
		running = false;
		done = true;
	}
	
	//System.err.println( this + " done");
}

void start(){
	if ( !running && !done) {
		running = true;
	
		synchronized( this) {
			this.notify();
		}
	}	
}

void stop(){
	try {
	if ( running ) {
		running = false;
		synchronized( this) {
			this.wait(300);
		}
	}
	}
	catch ( InterruptedException _ ) {
	}
}

public String toString() {
	return "Stream[" + url.toString() + "]";
}
}
class EvtHandler
  implements MouseListener, MouseMotionListener, KeyListener
{
EvtHandler() {
	PluginPlayer.this.addMouseMotionListener( this);
	PluginPlayer.this.addMouseListener( this);
	PluginPlayer.this.addKeyListener( this);
}

public void keyPressed( KeyEvent e) {
	transferEvent( e.getID(), e.getKeyCode(), e.getKeyChar() );
}

public void keyReleased( KeyEvent e) {
	transferEvent( e.getID(), e.getKeyCode(), e.getKeyChar() );
}

public void keyTyped( KeyEvent e) {
	transferEvent( e.getID(), e.getKeyCode(), e.getKeyChar() );
}

public void mouseClicked( MouseEvent e) {
}

public void mouseDragged( MouseEvent e) {
	transferEvent( e.getID(), e.getX(), e.getY());
}

public void mouseEntered( MouseEvent e) {
}

public void mouseExited( MouseEvent e) {
}

public void mouseMoved( MouseEvent e) {
	transferEvent( e.getID(), e.getX(), e.getY());
}

public void mousePressed( MouseEvent e) {
	transferEvent( e.getID(), e.getX(), e.getY());
}

public void mouseReleased( MouseEvent e) {
	transferEvent( e.getID(), e.getX(), e.getY());
}

void transferEvent( int event, int p1, int p2) {
	if ( nativeData != null ) {
		handleEvent0( nativeData, event, p1, p2);
	}
}
}

static {
	initPlugins();
}

public PluginPlayer () {
	flags |= ZOOM_PICTURE;
	new EvtHandler();
}

public void addNotify() {
	super.addNotify();
	watchFrame( true);
}

private void blank() {
	Graphics g = getGraphics();
	if ( g !=null ) {
		AWT.extension.paintBackground( this, g);
		g.dispose();
	}
}

native private static String checkPlugin0 (String lib);

native private void close0 (Ptr nativeData);

public void closeStream(){
	if ( currentStream != null ) {
		currentStream.stop();
		currentStream = null;
	}
}

public static Vector getSupportedMimeTypes() {
	Vector v = new Vector();
	for ( Enumeration e = plugins.keys(); e.hasMoreElements(); ){
		String key = (String)e.nextElement();
		if ( key.indexOf( '/') > -1 ) {
			v.addElement( key);
		}
	}
	return v;
}

public void getURL( String nppRequest){
	//callback from plugin to request new content
	setSource( nppRequest);
}

native private int handleEvent0 (Ptr nativeData, int event, int x, int y);

public synchronized void hide () {
	if ( (flags & PLAYING) != 0 ){
		stop();
	}

	super.hide();
}

private static void initPlugins() {
	String path = System.getProperty( PLG_SEARCH_PROP);
	
	if ( path != null ) {
		File f = new File( path);
		String[] libs = f.list( new FilenameFilter() {
			public boolean accept( File dir, String name) {
				if ( name.endsWith( ".so")) {
					String lPath = new File( dir, name).getPath();
					String mimes = checkPlugin0( lPath );
					if ( mimes != null ) {
						registerPlugin( lPath, mimes);
						return true;
					} 
				}
				return false;
			}
		});
		
	}
}

public boolean isFocusTraversable () {
	return true;
}

public boolean isRunning () {
	return ((flags & PLAYING) != 0);
}

private synchronized void loadPlugin ( String fname ) {
	boolean isRunning = ((flags & PLAYING) != 0);
	String curLib = lib;

	if ( isRunning ) {
		stop();
	}

	flags |= SOURCE_SET;
	flags &= ~(PLAYING | STARTED | PICTURE_SET);

	lib = lookupPlugin( fname);
	if ( lib == null ) {
		System.out.println( "cannot find a plugin for: " + fname);
		return;
	}
		
//	if ( (curLib == null) || !curLib.equals( lib) ){
		if ( nativeData != null ) {
			close0( nativeData);
		}
		nativeData = open0( this, lib);
//	}
	
	if ( isRunning ) {
		start();
	}
}

public boolean loadSource () {
	boolean streaming;

	synchronized ( this ) {
		if ( (flags & (SOURCE_SET|STARTED)) == SOURCE_SET ){
			if ( nativeData != null ) {
				//try to play as local file
				streaming = start0 ( nativeData, source, true );
				if ( streaming ) {
					currentStream = new Stream( source);
				}
				flags |= STARTED;
				return true;
			}
			return false;
		}
		return true;
	}
}

private static String lookupPlugin( String file ) {
	String lib = null;
	
	int idx = file.lastIndexOf( '.');
	if ( idx > 0 ) {
		String ext = file.substring( idx+1);
		lib = (String)plugins.get( ext);
	}
	
	return lib;
}

native private Ptr open0 (Object player, String lib);

public synchronized void paint ( Graphics g ) {
	AWT.extension.paintBackground( this, g);

	if ( (flags & (PLAYING|STARTED)) == STARTED ){
		start();
	}
}

private static void registerPlugin( String lib, String mimes) {
	//format is: mimeType:suffix<,suffix...>:comment<;mimes...>
	Vector keys = new Vector();
	StringTokenizer t0 = new StringTokenizer(mimes,":;");
	
	for ( int i=0; i<t0.countTokens(); i++ ) {
		String tk0 = t0.nextToken();
		switch( i%3) {
			case 0:
			case 1:
				StringTokenizer t1 = new StringTokenizer( tk0, ",");
				while( t1.hasMoreTokens() ) {
					keys.addElement( t1.nextToken() );
				}
				break;
			case 2:
				break;
		}
	}
	
	for ( int i=0; i<keys.size(); i++ ) {
		String key = ((String)keys.elementAt(i)).trim();
		System.err.println( "register plugin: " + key + ": " + lib);
		plugins.put( key, lib);
	}
}

public synchronized void removeNotify () {
	if ( (flags & PLAYING) != 0 ){
		stop();
	}
	watchFrame( false);
	super.removeNotify();
}

public void requestData( int offset, int length){
}

public synchronized void reshape ( int x, int y, int width, int height ) {
	int oldWidth = AWT.extension.getWidth( this);
	int oldHeight = AWT.extension.getHeight( this);
	boolean isPlaying = ((flags & PLAYING) != 0);
	
	super.reshape( x, y, width, height);

	if ( (width != oldWidth) || (height != oldHeight) ){
		if ( isPlaying ){
			//flags &= ~PLAYING;
		}

		if ( (width > 0) && (height > 0) ){
			flags &= ~PICTURE_SET;
			setPicture();

			if ( isPlaying ){
				// we will get a subsequent paint, which will (re)start us
				//flags |= STARTED;
			}
		}
	}
	
}

native private void setBounds0 (Ptr nativeData, int x, int y, int width, int height);

private synchronized void setPicture () {
	if ( (flags & PICTURE_SET) == 0 ){
		if ( nativeData == null ) {
			return;
		}
		Container c = null;
		for( c=getParent();(c != null) && (c.getParent() != null); c=c.getParent() );
		if ( (c == null) || ( ! (c instanceof Window) ) ) {
			return;
		}

		int   w  = AWT.extension.getWidth( this);
		int   h  = AWT.extension.getHeight( this);
		if (( w <= 0) || (h <= 0) ) {
			return;
		}

		Point p = getLocationOnScreen();

		if ( (flags & ZOOM_PICTURE) == 0) {
/*
 int wp = ctrl.getPictureWidth();
			int hp = ctrl.getPictureHeight();
			if ( wp == 0 || hp == 0 ) {
				return;
			}
			if ( wp > w ) wp = w;
			if ( hp > h )  hp = h;
			setBounds0( nativeData, p.x + (w - wp) / 2, p.y + (h - hp) / 2, wp, hp);
*/
		}
		else {
			Insets in = AWT.extension.getInsets( this);
			int cw = w - (in.left + in.right);
			int ch = h - (in.top + in.bottom);

			setBounds0( nativeData, p.x+in.left, p.y+in.top, cw, ch);
		}

		flags |= PICTURE_SET;
	}

}

public synchronized void setSource ( String url ) {
	source = url;
	loadPlugin( url);
}

public void setZoom ( boolean state) {
	boolean isRunning = ((flags & PLAYING) != 0);

	if ( isRunning ) {
		stop();
	}

	flags &= ~( PICTURE_SET | ZOOM_PICTURE);
	if ( state ) {
		flags |= ZOOM_PICTURE;
	}
	
	if ( isRunning ) {
		start();
	}
	
}

public synchronized void show () {
	super.show();
	start();
}

public synchronized void start () {
	if ( loadSource() && isShowing() ) {
		setPicture();
		
		if ( currentStream != null) {
			currentStream.start();
		}

		flags |= PLAYING;
	}

}

native private boolean start0 (Ptr nativeData, String source, boolean asFile);

public void statusUpdate( String status){
	//callback from plugin on status update
}

public synchronized void stop () {
	if ( (flags & PLAYING) != 0 ) {
		closeStream();
		if ( nativeData != null ) {
			stop0( nativeData);
			flags &= ~(PLAYING | STARTED | PICTURE_SET);
			blank();
		}
	}
}

native private void stop0 (Ptr nativeData);

public String toString () {
	StringBuffer sb = new StringBuffer(100);

	sb.append( super.toString());
	sb.append( " [");

	if ( (flags & PICTURE_SET) != 0 ) sb.append( "pict_set ");
	if ( (flags & SOURCE_SET) != 0 ) sb.append( "src_set ");
	if ( (flags & STARTED) != 0 ) sb.append( "started ");

	sb.append("]");

	return sb.toString();
}

public static void updatePlugins() {
	plugins.clear();
	initPlugins();
}

private void watchFrame( boolean on) {
	//this is needed because a Frame move will not emmit a reshape on a widget
	//so this is the only way to let the plugin rendering know about changed
	//screen coordinates

	Component c;
	for ( c=this; !(c instanceof Window) && c != null; c= c.getParent() );

	if ( c != null) {
		if ( on ) {
			if ( ca == null ) {
				ca = new ComponentAdapter() {
					public void componentMoved( ComponentEvent e) {
						if ( ((flags & PLAYING) != 0) ){
						//	flags &= ~PLAYING;
						}

						flags &= ~PICTURE_SET;
						setPicture();

						if ( ((flags & PLAYING) != 0) ){
							// we will get a subsequent paint, which will (re)start us
						//	flags |= STARTED;
						}
					}
				};
			}
	
			c.addComponentListener( ca);
		}
		else {
			if ( ca != null ) {
				c.removeComponentListener( ca);
			}
		}
	}
}

native private int write0 (Ptr nativeData, byte[] buffer, int offset, int length);
}
