package kaffe.awt.theme;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Rectangle;
import java.util.Properties;
import java.util.StringTokenizer;
import kaffe.awt.AWT;

public class ImageComponentDeco
  extends ComponentDeco
{
	Image[] cornerImgs;
	Image backImg;
	Image[] borderImgs;
	static String IMG_BG = "background";
	static String IMG_BORDER = "borders";
	static String IMG_CORNER = "corners";

public ImageComponentDeco() {
	int d = 0;
	insets = new Insets( d, d, d, d);
}

protected void calculateInsets() {
	if ( borderImgs[0] != null ) {
		insets.top = borderImgs[0].getHeight( null);
	}
	if ( borderImgs[1] != null ) {
		insets.right = borderImgs[1].getWidth( null);
	}
	if ( borderImgs[2] != null ) {
		insets.bottom = borderImgs[2].getHeight( null);
	}
	if ( borderImgs[3] != null ) {
		insets.left = borderImgs[3].getWidth( null);
	}
}

protected void calculateInsets_1() {
	if ( cornerImgs[0] != null ) {
		insets.left = cornerImgs[0].getWidth( null);
		insets.top = cornerImgs[0].getHeight( null);
	}
	if ( cornerImgs[3] != null) {
		insets.left = Math.max( insets.left, cornerImgs[3].getWidth( null));
		insets.bottom = cornerImgs[3].getHeight( null);
	}

	if ( cornerImgs[1] != null ) {
		insets.right = cornerImgs[1].getWidth( null);
		insets.top = Math.max( insets.top, cornerImgs[1].getHeight( null));
	}
	if ( cornerImgs[2] != null) {
		insets.right = Math.max( insets.right, cornerImgs[2].getWidth( null));
		insets.bottom = Math.max( insets.bottom, cornerImgs[2].getHeight( null));
	}
}

public boolean hasBackgroundImage() {
	return (backImg != null);
}

void initImages( Properties props, String clazz, String style, String path, Component c) {
	String val, nextToken;
	StringTokenizer tok;
	
	//get color and font Attributes
	initAttributes( props, clazz, style, c);
	
	if ( (val = Theme.getProperty(  props, clazz, IMG_BG, style)) != null ) {
		Image[] imgs = new Image[1];
		imgs[0] = ImageTheme.getImage( path , val, style)  ;
		ImageTheme.trackImages( imgs);
		backImg = imgs[0];
	}

	/*       corner images
	 *        0 ---- 1
	 *         |            |
	 *        3 ---- 2
	 */

	if ( (val = Theme.getProperty( props, clazz, IMG_CORNER, style)) != null ) { 
		tok = new StringTokenizer( val, ",; ");
		cornerImgs = new Image[4];
		for ( int idx = 0; tok.hasMoreTokens(); idx++ ){
			nextToken = tok.nextToken();
			if ( ! "null".equalsIgnoreCase( nextToken) ) {
				cornerImgs[idx] = ImageTheme.getImage( path , nextToken, style)  ;
			}
		}
		ImageTheme.trackImages( cornerImgs);
	}

	/*       border ( tile )images
	 *          --0--
	 *         3          1
	 *          --2-- 
	 */
	if ( (val = Theme.getProperty( props, clazz,  IMG_BORDER, style)) != null ) { 
		tok = new StringTokenizer( val, ",; ");
		borderImgs = new Image[4];
		for ( int idx = 0; tok.hasMoreTokens(); idx++ ){
			nextToken = tok.nextToken();
			if ( ! "null".equalsIgnoreCase( nextToken) ) {
				borderImgs[idx] = ImageTheme.getImage( path , nextToken, style)  ;					
			}
		}
		ImageTheme.trackImages( borderImgs);
	}

	calculateInsets();
}

public boolean paintBackground( Component o, Component c, Graphics g, int cx, int cy, int cw, int ch, int x, int y, int w, int h, int xOff, int yOff, int state) {
	if ( backImg == null ) 
		return false;
		
	AWT.extension.drawBackground( g, cx, cy, cw, ch, x, y, w, h, xOff, yOff, 
							 backImg, null, Theme.FILL_TILE);

	return true;
}

public void paintBorder( Component c, Graphics parentGr, int x, int y, int w, int h, int state) {
	AWT.extension.drawBorder( parentGr, x, y, w, h, Theme.ALIGN_OUTER, 
		cornerImgs[0], borderImgs[0], cornerImgs[1],
		borderImgs[3], borderImgs[1],
		cornerImgs[3], borderImgs[2], cornerImgs[2] );
}
}
