package kaffe.awt.theme;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Scrollbar;
import java.awt.event.AdjustmentEvent;
import java.util.Properties;
import java.util.StringTokenizer;
import kaffe.awt.AWT;

public class ImageScrollbarDeco
  extends ComponentDeco
  implements ScrollbarDeco
{
	int[] imgExts;
	Image[] imgs;
	int cmnWidth;
	static String IMG_JEWEL = "jewel";
	static String IMG_RAIL = "rail";
	static String IMG_ARROWS = "arrows";

public int getAdjustmentCode( Scrollbar s, int offs, int len, int jOffs, int jLen) {
	if ( offs < imgExts[0] )
		return AdjustmentEvent.UNIT_DECREMENT;
	else if ( offs > len - imgExts[3] )
		return AdjustmentEvent.UNIT_INCREMENT;
	else if ( offs < jOffs )
		return AdjustmentEvent.BLOCK_DECREMENT;
	else if ( offs > jOffs + jLen )
		return AdjustmentEvent.BLOCK_INCREMENT;

	return AdjustmentEvent.TRACK;
}

public int getJewelExtent( Scrollbar s, int w, int pgInc, int range) {
	return imgExts[1];
}

public Dimension getPreferredSize( Scrollbar s) {
	if ( s.getOrientation() == Scrollbar.VERTICAL )
		return new Dimension( cmnWidth, 100);
	return new Dimension( 100, cmnWidth);
}

public Rectangle getRailArea( Scrollbar s, Rectangle r) {
	if ( s.getOrientation() == s.HORIZONTAL ) {
		r.x += imgExts[0];
		r.width -= imgExts[0] + imgExts[3];
	}
	else {
		r.y += imgExts[0];
		r.height -= imgExts[0] + imgExts[3];
	}
	return r;
}

void initImages( Properties props, String clazz, String style, String path, Component c) {

	Image[] imgs = new Image[4];
	String val;
	StringTokenizer tok;
	
	//get color and font Attributes
	initAttributes( props, clazz, style, c);

	if ( (val = Theme.getProperty(  props, clazz, IMG_JEWEL, style)) != null ) {
		imgs[1] =  ImageTheme.getImage( path, val, style) ;
	}

	if ( (val = Theme.getProperty( props, clazz,  IMG_RAIL, style)) != null ) { 
		imgs[2] =  ImageTheme.getImage( path, val, style);
	}

	if ( (val = Theme.getProperty(  props, clazz,  IMG_ARROWS, style)) != null ) { 
		tok = new StringTokenizer( val, ",; ");
		for ( int idx = 0; (idx < 4) && tok.hasMoreTokens(); idx+=3 ){
			imgs[idx] = ImageTheme.getImage( path, tok.nextToken(), style );
		}
	}

	ImageTheme.trackImages( imgs);
	
	setImages( imgs, ((Scrollbar)c).getOrientation());
}

public void paintArrows( Scrollbar s, Graphics g, int w, int h, int state) {
	if ( s.getOrientation() == s.HORIZONTAL ) {
		g.drawImage( imgs[0], 0, 0, null);
		g.drawImage( imgs[3], w-imgExts[3], 0, null);
	}
	else {
		g.drawImage( imgs[0], 0, 0, null);
		g.drawImage( imgs[3], 0, h-imgExts[3],  null);
	}
}

public void paintJewel( Scrollbar s, Graphics g, int x, int y, int w, int h, boolean tracking) {
	g.drawImage( imgs[1], x, y, null);
}

public void paintRail( Scrollbar s, Graphics g, int x, int y, int w, int h) {
	if ( s.getOrientation() == s.VERTICAL ) {
		AWT.extension.drawBackground( g, 0, 0, AWT.extension.getWidth(s), 
								AWT.extension.getHeight(s),
								x, y, w, h, 0, imgExts[0], 
								imgs[2], null, Theme.FILL_TILE);
	}
	else {
		AWT.extension.drawBackground( g, 0, 0, AWT.extension.getWidth(s), 
								AWT.extension.getHeight(s), x, y, w, h, imgExts[0], 0, 
								imgs[2], null, Theme.FILL_TILE);
	}
}

private void setImages( Image[] imgs, int orientation ) {
	this.imgs = imgs;
	imgExts = new int[imgs.length];
	
	for ( int i=0; i<imgs.length; i++ ) {
		if ( orientation == Scrollbar.VERTICAL ) {
			imgExts[i] = imgs[i].getHeight( null);
			cmnWidth = Math.max( cmnWidth, imgs[i].getWidth( null));
		}
		else {
			imgExts[i] = imgs[i].getWidth( null);
			cmnWidth = Math.max( cmnWidth, imgs[i].getHeight( null));
		}
	} 
}
}
