package kaffe.awt.theme;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Scrollbar;
import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;
import kaffe.io.AccessibleBAIStream;
import kaffe.net.www.protocol.zip.ZipURLConnection;

public class ImageTheme
  extends Theme
{
	static Toolkit tlk = Toolkit.getDefaultToolkit();
	static ZipFile defZip;
	static MediaTracker tracker = new MediaTracker( null);
	static String ZIP_PROPERTY = "defaultZip";

public ComponentDeco getDeco( Component c, String clazz, String style) {
	ImageComponentDeco icd; 
	String path, val, nextToken, key;
	Image[] imgs;
	StringTokenizer tok;

	key = getPropertyKey( clazz, style);

	if ( decos != null ) {  
		ComponentDeco cd = (ComponentDeco)decos.get( key);
		if ( cd != null )
			return cd;
	}

	if ( props == null) {
		return nullDeco;
	}

	if ( decos == null )
		decos = new Hashtable();
		
	//currently just the SimpleFrameDeco supported
	if ( c instanceof java.awt.Frame ) {
		SimpleFrameDeco sfd = new SimpleFrameDeco();
		decos.put( key, sfd);
		return sfd;
	}
	
	path = getProperty( props, clazz, "imgPath", style);
	
	// no image properties, just look for color and font attributes
	if ( path == null ) {
		if ( ComponentDeco.hasAttributes( props, clazz, style) ){
			ComponentDeco cd = new ComponentDeco();
			cd.initAttributes( props, clazz, style, c);
			decos.put( key, cd);
			return cd;
		}
		else {
			return nullDeco;
		}
	}
		
	//special deco ojects
	if ( c instanceof java.awt.Scrollbar ) {
		ImageScrollbarDeco isd = new ImageScrollbarDeco();
		isd.initImages( props, clazz, style, path, c);
		decos.put( key, isd);
		return isd;
	}
	
	if ( c instanceof java.awt.Choice ) {
		icd = new ImageChoiceDeco();
	}
	else if ( c instanceof java.awt.Checkbox ) {
		icd = new ImageCheckboxDeco();
	}
	else {
		icd = new ImageComponentDeco();
	}
	
	icd.initImages( props, clazz, style, path, c);
	decos.put( key, icd);

	return icd;
}

static Image getImage( String path, String file, String style) {
	URL u;
	String name = new String();
	Image img = null;
	
//	if ( style != null )
//		name += style + "/";
	name += file;

	try {
		if ( path.equals( ZIP_PROPERTY) ) {
			img = getImage( defZip, name);
		}
		else if ( path.endsWith( ".zip") ) {
			u = new URL("zip:file:" + path + "!/");
			ZipURLConnection zc = (ZipURLConnection)u.openConnection();
			img = getImage( zc.getZipFile(), name);
		}
		else {
			u = new URL( "file:" + path + name);
			img =  tlk.getImage( u);
		}
	}
	catch (IOException _) {
		img = null;
	}

	return img;
}

static Image getImage( ZipFile zf, String entry) throws IOException{
	if ( zf != null ) {
		ZipEntry ze = zf.getEntry( entry);
		if ( ze != null ) {
			byte[] data = ((AccessibleBAIStream)zf.getInputStream(ze)).getByteArray();
			return tlk.createImage(data);
		}
	}
	
	return null;
}

public boolean load( String propertyFile) {
	super.load( propertyFile);
	if( props != null) {
		String zipPath = props.getProperty(  ZIP_PROPERTY );
		if ( zipPath != null ) {
			try {
				URL u = new URL("zip:file:" + zipPath + "!/");
				ZipURLConnection zc = (ZipURLConnection)u.openConnection();
				defZip = zc.getZipFile();
			}
			catch ( IOException _) {
				defZip = null;
			}
		}
		else {
			defZip = null;
		}
		return true;
	}
	return false;
}

static void trackImages( Image[] imgs) {
	for ( int i=0; i<imgs.length; i++ ) {
		if ( imgs[i] != null )
			tracker.addImage( imgs[i], i);
	}

	try { tracker.waitForAll(); }
	catch ( Exception _) {}

	for ( int i=0; i<imgs.length; i++ ) {
		if ( imgs[i] != null )
			tracker.removeImage( imgs[i]);
	}
}
}
