package kaffe.awt.theme;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Insets;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;

public class SimpleFrameDeco
  extends FrameDeco
{
public SimpleFrameDeco() {
	int bw = Defaults.FrameBorderWidth;
	int th = Defaults.WndFontMetrics.getHeight() * 4 / 3;

	fgClr = Defaults.WndForeground;
	bgClr = Defaults.WndBackground;
	fnt  = Defaults.WndFont;

	insets =  new Insets ( th+bw, bw, bw, bw);
}

public int getBorder ( int width, int height, int xPos, int yPos, int flags ) {
	int bw = Defaults.FrameBorderWidth;
	int d = 15;
	int border = 0;
	int u = width - xPos;
	int v = height - yPos;
	
	if ( xPos <= bw ) {
		border |= WEST;
		if ( yPos < d ) {
			return (WEST | NORTH);
		}
		else if ( v < d ){
			return (WEST | SOUTH);
		}
	}
	if ( yPos <= bw ) {
		border |= NORTH;
		if ( xPos < d ) {
			return (NORTH | WEST);
		}
		else if ( u < d ) {
			return (NORTH | EAST);
		}
	}
		
	if ( u <= bw ) {
		border |= EAST;
		if ( yPos < d ) {
			return (EAST | NORTH);
		}
		else if ( v < d ){
			return (EAST | SOUTH);
		}
	}

	if ( v <= bw ) {
		border |= SOUTH;
		if ( xPos < d ) {
			return (SOUTH | WEST);
		}
		else if ( u < d ) {
			return (SOUTH | EAST);
		}
	}

	return border;
}

public int getTitlebarButton ( int width, int height, int xPos, int yPos, int flags ) {
	int d0 = Defaults.FrameBorderWidth;
	int th = Defaults.WndFontMetrics.getHeight() * 4 / 3;
	int d1 = d0 + th;
	
	if ( (yPos > d0) && (yPos < d1) ){
		if ( (xPos > d0) && (xPos < d1) ){
			return TBB_CLOSE;
		}
		else if ( flags != 0) {
			xPos = width - xPos;
		
			if ( (xPos > d0) && (xPos < d1) ){
				xPos = th - xPos;
				if ( xPos > yPos ) return TBB_MAXIMIZE;
				else return TBB_MINIMIZE;
			}
		}
	} 
	
	return 0;
}

public void paintBorder ( Component c, Graphics g, int x, int y, int w, int h, int state ) {
	Frame fr = (Frame) c;
	Color clrb, clrd, clr, clrt;
	int bw = Defaults.FrameBorderWidth;
	int i;
	int u  = x;
	int v  = y;
	int uw = u + w - 1;
	int vh = v + h - 1;
	FontMetrics fm = Defaults.WndFontMetrics;
	int fh = fm.getHeight();
	int th = fh * 4 / 3;

	// get colors
	if ( (state & ACTIVE) != 0 ){
		clr  = Defaults.WndActiveClr;
		clrt = Defaults.WndActiveTitleClr;
		
		if ( (state & OWNED) != 0 ){
			clr = clr.brighter();
		}
	}
	else {
		clr  = Defaults.WndInactiveClr;
		clrt = Defaults.WndInactiveTitleClr;
	}
	clrd = clr.darker();
	clrb = clr.brighter();
	
	// the brighter/darker exterior edges
	g.setColor( clrb);
	g.drawLine( u, v, uw-1, v);
	g.drawLine( u, v, u, vh);
	g.setColor( clrd);
	g.drawLine( uw, v, uw, vh);
	g.drawLine( u+1, vh, uw, vh);
	
	// the border interior
	u++; v++; uw--; vh--;
	g.setColor( clr);
	for ( i=2; i<bw; i++ ) {
		g.drawRect( u, v, uw-u, vh-v);
		u++; v++; uw--; vh--;
	}
	
	// the darke/brighter interior edges
	g.setColor( clrd);
	g.drawLine( u, v, uw-1, v);
	g.drawLine( u, v, u, vh);
	g.setColor( clrb);
	g.drawLine( uw, v, uw, vh);
	g.drawLine( u+1, vh, uw, vh);
	
	u++; v++;
	
	// the titlebar
	g.setColor( clr);
	g.fill3DRect( u, v, uw-u, th, true);
	
	g.setColor( clrt);
	g.setFont( Defaults.WndFont);
	g.drawString( fr.getTitle(), x + bw + th, y + bw + (th - fh)/2 + fm.getAscent());
	
	int d = th / 3;
	int d2 = d / 2;
	
	// close button
	u  = x + bw + d;
	v  = y + bw + d;
	uw = u + d;
	vh = v + d;
	g.setColor( clrd);
	g.drawLine( u, v, uw, vh);
	g.drawLine( uw, v, u, vh);
	
	u--; uw--;
	g.setColor( clrb);
	g.drawLine( u, v, uw, vh);
	g.drawLine( uw, v, u, vh);

	if ( fr.isResizable() ) {
		uw = x + w - bw - d2;
		v = y + bw + d2 + 1;
		u = uw - d - d2;
		vh = v + d + d2;

		// max button	
		g.setColor( clrd);
		//g.drawLine( u, v, uw, v);
		g.drawLine( uw, v, uw, vh);
		g.drawLine( uw, vh, u, v);
		
		u--; uw--;
		g.setColor( clrb);
		g.drawLine( u, v, uw, v);
		g.drawLine( uw, v, uw, vh);
		g.drawLine( uw, vh, u, v); 

		// min button
		uw = x + w - bw - d + 1;
		v = y + bw + d - 1;
		u = uw - d - d2;
		vh = v + d + d2;

		g.setColor( clrd);
		g.drawLine( u, v, uw, vh);
		//g.drawLine( uw, vh, u, vh);
		g.drawLine( u, vh, u, v); 

		u--; uw--;
		g.setColor( clrb);
		g.drawLine( u, v, uw, vh);
		g.drawLine( uw, vh, u, vh);
		g.drawLine( u, vh, u, v); 
	}
}
}
