package kaffe.awt.theme;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Scrollbar;
import java.awt.event.AdjustmentEvent;
import kaffe.awt.Defaults;

public class SimpleScrollbarDeco
  extends ComponentDeco
  implements ScrollbarDeco
{
	final static int MIN_JEWEL = 10;

SimpleScrollbarDeco() {
	super( null, Defaults.BtnClr, null);
}

public int getAdjustmentCode( Scrollbar s, int offs, int len, int jOffs, int jLen) {
	if ( ( offs > jOffs) && ( offs < jOffs + jLen) )
		return AdjustmentEvent.TRACK;
	else if ( offs < Defaults.ScrollbarWidth )
		return AdjustmentEvent.UNIT_DECREMENT;
	else if ( offs > len - Defaults.ScrollbarWidth )
		return AdjustmentEvent.UNIT_INCREMENT;
	else if ( offs < jOffs )
		return AdjustmentEvent.BLOCK_DECREMENT;
	else if ( offs > jOffs + jLen )
		return AdjustmentEvent.BLOCK_INCREMENT;

	return 0;
}

public int getJewelExtent( Scrollbar s, int w, int pgInc, int range) {
	int je = w * pgInc / range;
	if ( je < MIN_JEWEL )
		je = MIN_JEWEL;
	else if ( je >= w )
		je = -1;
	return je;
}

public Dimension getPreferredSize( Scrollbar s) {
	if ( s.getOrientation() == Scrollbar.VERTICAL )
		return new Dimension( Defaults.ScrollbarWidth, 100);
	return new Dimension( 100, Defaults.ScrollbarWidth);
}

public Rectangle getRailArea( Scrollbar s, Rectangle r) {
	return r;
}

public void paintArrows( Scrollbar s, Graphics g, int w, int h, int state) {
}

public void paintJewel( Scrollbar s, Graphics g, int x, int y, int w, int h, boolean tracking) {
	int db = Defaults.WidgetBorderWidth;

	if ( s.getOrientation() == s.HORIZONTAL )
		g.fill3DRect( x, y+db, w, h-2*db, true);
	else
		g.fill3DRect( x+db, y, w-2*db, h, true);
}

public void paintRail( Scrollbar s, Graphics g, int x, int y, int w, int h) {
	g.clearRect( x, y, w, h);

	if ( s.getOrientation() == s.VERTICAL ){
		g.draw3DRect( x+w/2-1, y, 1, h-1, true);
	}
	else {
		// Hrmm, 3D rects always have equally colored vertical edges, so we have to do
		// it the hard way
		Color c = g.getColor();
		int xw = x + w -1;		
		y += h/2 -1;

		g.setColor( c.brighter());
		g.drawLine( x, y, xw, y);
		g.setColor( c.darker());
		y++;
		g.drawLine( x, y, xw, y);
		
		g.setColor( c);
	}
}
}
