package kaffe.awt.theme;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Scrollbar;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;

abstract public class Theme
{
	final public static int ALIGN_INNER = 0;
	final public static int ALIGN_CENTER = 1;
	final public static int ALIGN_OUTER = 2;
	final public static int FILL_NONE = 0;
	final public static int FILL_COLOR = 1;
	final public static int FILL_TILE = 2;
	final public static int FILL_CENTER = 4;
	Properties props;
	Hashtable decos;
	public static ComponentDeco nullDeco = new ComponentDeco();
	static String PROPERTY_KEY = "theme_properties";
	static String INHERIT_KEY = "inherit";
	static char ATTR_SEPARATOR = '@';

public ComponentDeco getDeco( Component c) {
	return getDeco( c, null);
}

public ComponentDeco getDeco( Component c, String style) {
	String clazz, val;
	Class cc;
	StringTokenizer tok;
	
	if ( props != null ) {	
		for (cc = c.getClass(); cc != null; cc = cc.getSuperclass()) {
			clazz = cc.getName();
			val = (String)props.get( clazz + ATTR_SEPARATOR + INHERIT_KEY);
			if ( val != null ) {
				for ( tok = new StringTokenizer( val, " -"); tok.hasMoreTokens(); ) {
					String nextToken = tok.nextToken();
					if ( nextToken.equals( c.getClass().getName()) ) {
						clazz = c.getClass().getName();
						return getDeco( c, clazz, style);
					}
				}
				return getDeco( c, clazz, style);
			}
		}
	}
	
	clazz = c.getClass().getName();
	return getDeco( c, clazz, style);
}

abstract public ComponentDeco getDeco( Component c, String clazz, String style);

static String getProperty( Properties props, String clazz, String attr, String style ) {
	String val, key;
	
	if ( props != null ) {
		//lookup alias
		val = props.getProperty( clazz);
		key = ((val == null ) ? clazz : val) + ATTR_SEPARATOR + attr;

		if ( style != null ) {
			val = props.getProperty( key + ATTR_SEPARATOR + style);
			if ( val != null ) {
				return val;
			}
		}
		return props.getProperty( key);
	}
	return null;
}

String  getPropertyKey(  String clazz, String style) {
	String val, key = clazz;
	
	if ( props != null ) {
		//alias lookup
		if ( (val = props.getProperty( clazz)) != null ) {
			key = val;
		}
	}

	if ( style != null ) {
		key += ATTR_SEPARATOR + style;
	}
	
	if ( props != null ) {
		if ( (val = props.getProperty( key + ATTR_SEPARATOR + "like")) != null ) {
			key = val;
			//another alias
			if ( (val = props.getProperty( key)) != null ) {
				key = val;
			}
		}
			
	}
	
	return key;
}

public boolean load( String propertyFile) {
	if ( propertyFile == null ) {
		propertyFile = System.getProperty( PROPERTY_KEY);
	}
	if ( propertyFile == null ) {
		return false;
	}

	//renew properties and cache
	props = new Properties();
	decos = new Hashtable();

	try {
		FileInputStream pi = new FileInputStream( propertyFile);
		props.load( pi);
	//	props.list( System.out);
	}
	catch ( FileNotFoundException _ ) { 
		return false;
	}
	catch ( IOException _) { 
		return false;
	}

	return true;
}
}
