package kaffe.awt.theme;

import java.awt.Button;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Label;
import java.awt.List;
import java.awt.Rectangle;
import java.awt.ScrollPane;
import java.awt.Scrollbar;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.Window;
import java.awt.event.AdjustmentEvent;
import kaffe.awt.Defaults;

public class WidgetThemeKaffe
  extends WidgetThemeImpl
{
public int borderInnerWidth ( Component c ) {
	//width of border for inner elements of compound widgets ( e.g.RowCanvas )
	return( Defaults.WidgetBorderWidth);
}

public int borderOuterWidth ( Component c ) {
	//width of outer border for compound widgets ( e.g.List )
	return( 0);
}

public void borderPaint ( Component c, Graphics g, int left, int top, int right, int bottom, int width, int height, boolean focused ) {
	int w = width - (left + right);
	int h = height - (top + bottom);

	if ( focused )
		g.setColor( Defaults.FocusClr);
	else
		g.setColor( Defaults.BorderClr);

	if (w-1 > 0 && h-1 > 0) {
		g.draw3DRect( left, top,  w-1, h-1, true);
	}
	if (w-3 > 0 && h-3 > 0) {
		g.draw3DRect( left+1, top+1, w-3, h-3, false);
	}
}

public void borderPaintInner ( Component c, Graphics g, boolean focused ) {
	//paint border for inner elements of compound widgets ( e.g. RowCanvas of List )
	Dimension d = c.getSize();
	borderPaint( c, g, 0, 0, 0, 0, d.width, d.height, focused );
}

public void borderPaintOuter ( Component c, Graphics g, boolean focused ) {
	//paint border around compound widget ( e.g. List )
}

public Rectangle btnGetDeco( Button btn) {
	int b = Defaults.WidgetBorderWidth;
	return new Rectangle( b, b, 2*b, 2*b );
}

public void btnPaint( Button btn, Graphics g, Rectangle deco, int flags) {
	int bw = Defaults.WidgetBorderWidth;
	Dimension d = btn.getSize();

	if ( (flags & WidgetTheme.NEEDS_BORDER_UPDATE) > 0 ) {
		wdgPaintBorder( btn, g, deco, d.width, d.height, (flags & WidgetTheme.HAS_FOCUS) > 0);
		g.setColor( ((flags & WidgetTheme.IS_POINTED) > 0) ? Defaults.BtnPointClr : btn.getBackground());	
		g.fill3DRect( bw, bw, d.width-2*bw, d.height-2*bw, ((flags & WidgetTheme.IS_ACTIVE) == 0));
	}

	String label = btn.getLabel();
	//nothing more to paint
	if ( label == null )
		return;
		
	Color c1 = null, c2;
	Color bg = btn.getBackground();
	int x = 0;
	int y = 0;

	if (!btn.isEnabled()) {
		c1 = null;
		c2 = bg.darker();
	}
	else if ( (flags & WidgetTheme.IS_ACTIVE) > 0){
		x--; y--;
		c1 = Color.yellow;
		c2 = Color.red;
	}
	else {
		if ( (flags & WidgetTheme.IS_POINTED) > 0) {
			c1 = Defaults.BtnTxtCarved ? bg.brighter() : null;
			c2 = Defaults.BtnPointTxtClr;
		}
		else {
			c1 = Defaults.BtnTxtCarved ? bg.brighter() : null;
			c2 = btn.getForeground();
		}
	}

	if ( c1 != null ){
		g.setColor( c1);
		RenderDirection.drawString( label, g, x+1, y+1, d.width, d.height, 1);
	}
	
	g.setColor( c2);
	RenderDirection.drawString( label, g, x, y, d.width, d.height, 1);
}

public void btnPaintBackground(Button btn, Graphics g, int x, int y, int width,
int height, int flags) {
}

public Dimension btnPreferredSize( Button btn, Dimension rawSize) {
	String label = btn.getLabel();

	//maybe an image button?
	if ( label != null ) {
		FontMetrics fm = btn.getFontMetrics( btn.getFont() );
		int dx = 40;
		int dy = 20;

		if ( RenderDirection.isHorizontal() ) {
			rawSize.width = Math.max( dx, (fm != null) ? 4*fm.stringWidth( label)/3 : 0);
			rawSize.height = Math.max( dy, (fm != null) ? 2*fm.getHeight() : 0);
		}
		else {
			rawSize.width = Math.max( dy, (fm != null) ? 2*fm.charWidth( 'X') : 0);
			rawSize.height = Math.max( dx, (fm != null) ? (label.length()+1)*fm.getHeight() : 0);
		}
	}
		
	return( rawSize);
}

public void btnSetDefaults(Button btn) {
	btn.setFont( Defaults.BtnFont);
	btn.setForeground( Defaults.BtnTxtClr);
	btn.setBackground( Defaults.BtnClr);
}

public Rectangle cbGetDeco( Checkbox cb) {
	return new Rectangle();
}

public void cbPaint( Checkbox cb, Graphics g, Rectangle deco, int flags) {
	Color c1, c2;
	FontMetrics fm = cb.getFontMetrics( cb.getFont());
	Dimension d = cb.getSize();
	String label = cb.getLabel();
	int ext = fm.getHeight();                 // this controls the button size
	int dx  = fm.charWidth( 'x');
	int x0, y0;
	
	//blank entire widget
	g.setColor( cb.getBackground() );
	g.fillRect( 0, 0, d.width, d.height);

	c1 = ((flags & WidgetTheme.IS_POINTED) > 0) ? Defaults.FocusClr : cb.getForeground();
	c2 = Color.white;

	switch ( RenderDirection.getDirection() ) {
		case RenderDirection.LR:
			x0 = dx;
			y0 = (d.height - ext) / 2;
			g.setColor( c2);
			RenderDirection.drawString( label, g, x0+ext+dx+1, 1, d.width, d.height, 0);
			g.setColor( c1);
			RenderDirection.drawString( label, g, x0+ext+dx, 0, d.width, d.height, 0);
			break;
		case RenderDirection.RL:
			x0 = d.width - dx - ext;
			y0 = (d.height - ext) / 2;
			g.setColor( c2);
			RenderDirection.drawString( label, g, 1, 1, x0-dx, d.height, 2);
			g.setColor( c1);
			RenderDirection.drawString( label, g, 0, 0, x0-dx, d.height, 2);
			break;
		default:
			x0 = (d.width - ext) / 2;
			y0 = dx;
			g.setColor( c2);
			RenderDirection.drawString( label, g, 1, y0+ext+dx+1, d.width, d.height, 0);
			g.setColor( c1);
			RenderDirection.drawString( label, g, 0, y0+ext+dx, d.width, d.height, 0);
			break;
	}
	
	// button
	g.setColor( ((flags & WidgetTheme.IS_POINTED) > 0) ? Defaults.BtnPointClr : Defaults.BtnClr);
	g.fill3DRect( x0, y0, ext, ext, true);

	//space suffix frames the widget
	if ( label.endsWith( " " ) )
		borderPaint( cb, g, 0, 0, 0, 0, d.width, d.height, ((flags & WidgetTheme.HAS_FOCUS) > 0));
	else {
		int db = Defaults.WidgetBorderWidth;
		borderPaint( cb, g, x0-db, y0-db, d.width-(x0+ext+db), d.height-(y0+ext+db), d.width, d.height, ((flags & WidgetTheme.HAS_FOCUS) > 0) );
	}

	// checkmark
	if ( ((flags & WidgetTheme.IS_ACTIVE) > 0)) {
		int xl = x0 + 2;
		int xr = x0 + ext - 5;
		int yt = y0 + 3;
		int yb = y0 + ext - 4;

		g.setColor( Color.black);
		g.drawLine( xl, yt, xr, yb);
		g.drawLine( xl, yb, xr, yt);

		xl++; xr++;
		g.setColor( Color.white);
		g.drawLine( xl, yt, xr, yb);
		g.drawLine( xl, yb, xr, yt);	
	}
}

public void cbPaintBackground( Checkbox cb, Graphics g, int x, int y, int width,
int height, int flags) {
}

public Dimension cbPreferredSize( Checkbox cb, Dimension rawSize) {
	String label = cb.getLabel();
	
	//maybe an image Checkbox?
	if ( label != null ) {
		int dx = 50;
		int dy = 20;
		FontMetrics fm = cb.getFontMetrics( cb.getFont() );
	
		if ( RenderDirection.isHorizontal() ) {
			rawSize.width = Math.max( dx, (fm != null) ? fm.stringWidth( label) + 2*fm.getHeight() : 0);
			rawSize.height = Math.max( dy, (fm != null) ? 3*fm.getHeight()/2 : 0);
		}
		else {
			rawSize.width = Math.max( dy, (fm != null) ? 2*fm.charWidth( 'X') : 0);
			rawSize.height = Math.max( dx, (fm != null) ? (label.length()+3)*fm.getHeight() : 0);
		}
	}
	
	return( rawSize);
}

public boolean chButtonRight( Choice ch) {
	return( true);
}

public Rectangle chGetDeco( Choice ch) {
	int b = Defaults.WidgetBorderWidth;
	return new Rectangle( b, b, 2*b, 2*b );
}

public boolean chOpenPrompt( Choice cb, int x, int y) {
	Dimension d = cb.getSize();
	return ( x > d.width - Defaults.ChoiceBtnWidth);
}

public void chPaint( Choice ch, Graphics g, Rectangle deco, int flags) {
	int x, y;
	int bw = Defaults.WidgetBorderWidth;
	Dimension d = ch.getSize();
	int btnWidth = Defaults.ChoiceBtnWidth;

	g.setColor( ((flags & WidgetTheme.IS_POINTED) > 0) ? Defaults.BtnPointClr : Defaults.BtnClr);
	g.fill3DRect ( d.width-btnWidth-bw, bw, btnWidth, d.height-2*bw, ((flags & WidgetTheme.IS_ACTIVE) == 0) );
	
	//draw button marker
	x = d.width - bw - btnWidth/2;
	y = d.height - (d.height-7)/2;
	g.setColor( Color.white);
	g.drawLine( x, y, x-3, y-7);
	g.drawLine( x, y, x+3, y-7);
	x--;
	g.setColor( ((flags & WidgetTheme.IS_POINTED) > 0) ? Defaults.FocusClr : Color.black);
	g.drawLine( x, y, x-3, y-7);
	g.drawLine( x, y, x+3, y-7);
	
	//draw non focus border around button
	g.setColor( Defaults.BtnClr);
	x = d.width - btnWidth - 2*bw;
	g.draw3DRect( x, 0,  d.width-x-1, d.height-1, true);
	g.draw3DRect( x+1, 1, d.width-x-3, d.height-3, false);
}

public void chPaintBackground( Choice ch, Graphics g, int x, int y, int width,
int height, int flags) {
	g.setColor (Defaults.TextFieldBgClr);
	g.fillRect( x, y, width, height);
}

public void chPaintPopup(Choice ch, Window popup, Graphics g) {
}

public Dimension chPreferredSize( Choice cb, Dimension rawSize) {
	//add additional deco space ( e.g. frame ) to the widgets raw size
	rawSize.width += Defaults.ChoiceBtnWidth + 2*Defaults.WidgetBorderWidth;
	return (rawSize);
}

public void chSetDefaults( Choice ch) {
        ch.setBackground (Color.lightGray);
}

public Rectangle chTextBounds( Choice ch) {
	Dimension d = ch.getSize();
	return new Rectangle( 0, 0, d.width-Defaults.ChoiceBtnWidth, d.height);
}

public Rectangle lblGetDeco( Label lbl) {
	return new Rectangle();
}

public void lblPaint( Label lbl, Graphics g, Rectangle deco, int flags ) {
	Dimension d = lbl.getSize();
	String txt = lbl.getText();
	int border = ((flags & WidgetTheme.NEEDS_BORDER_UPDATE) > 0) ? Defaults.WidgetBorderWidth : 0;
	
	//black entire widget
	g.setColor( lbl.getBackground() );
	g.fillRect( 0, 0, d.width, d.height);

	if ( ((flags & WidgetTheme.NEEDS_BORDER_UPDATE) > 0) )
		borderPaint( lbl, g, 0, 0, 0, 0, d.width, d.height, false);
			
	if ( Defaults.LabelTxtCarved ) {
		g.setColor( Color.white);
		RenderDirection.drawString( txt, g, border+1, border+1, d.width-2*border, d.height-2*border, lbl.getAlignment() );
	}
	
	g.setColor( lbl.getForeground() );
	RenderDirection.drawString( txt, g, border, border, d.width-2*border, d.height-2*border, lbl.getAlignment() );
}

public void lblPaintBackground( Label lbl, Graphics g, int x, int y, int width,
int height, int flags) {
}

public Dimension lblPreferredSize( Label lbl, Dimension rawSize ) {
	String label = lbl.getText();
	
	if ( label != null ) {
		int dx = 40;
		int dy = 20;
		FontMetrics fm = lbl.getFontMetrics( lbl.getFont() );
		
		if ( RenderDirection.isHorizontal() ) {
			rawSize.width = Math.max( dx, (fm != null) ? fm.stringWidth( label) : 0);
			rawSize.height = Math.max( dy, (fm != null) ? 3*fm.getHeight()/2 : 0);
		}
		else {
			rawSize.width = Math.max( dy, (fm != null) ? 3*fm.charWidth( 'X')/2 : 0);
			rawSize.height = Math.max( dx, (fm != null) ? (label.length()+1)*fm.getHeight() : 0);
		}
	}
	
	return( rawSize);
}

public void lblSetDefaults(Label lbl) {
        lbl.setFont(Defaults.LabelFont);
        lbl.setForeground(Defaults.LabelClr);
}

public void lstPaintBackground( List lst, Graphics g, int x, int y, int width,
int height, int flags) {
        if (flags == WidgetTheme.IS_ACTIVE) {
                g.setColor( Defaults.ListSelBgClr);
                g.fill3DRect( x, y, width, height, true);
                g.setColor( Defaults.ListSelTxtClr);
        }
        else if (flags == WidgetTheme.IS_POINTED) {
                g.setColor( Defaults.ListFlyOverBgClr);
                g.fill3DRect( x, y, width, height, !Defaults.ListFlyOverInset);
                g.setColor( Defaults.ListFlyOverTxtClr);
        }
        else {
                g.setColor( lst.getBackground() );
                g.fillRect( x, y, width, height);
                g.setColor( lst.getForeground() );
        }
}

public void lstSetDefaults(List lst) {
        lst.setForeground( Defaults.ListTxtClr);
        lst.setBackground( Defaults.ListBgClr);
        lst.setFont( Defaults.ListFont);
}

public int sbGetAdjustmentArea( Scrollbar sb, int x, int y, Rectangle trackBounds, int sliderOffset, int sliderLength) {
	Dimension d = sb.getSize();
	
	//get unit adjustment values from position and slider
	switch( sb.getOrientation() ) {
	case Scrollbar.VERTICAL:
		if ( y > d.height - d.width ) {
			if ( trackBounds.y + sliderOffset + sliderLength >= y )
				return AdjustmentEvent.TRACK;
			return AdjustmentEvent.UNIT_INCREMENT;
		}
		if ( y < d.width ) {
			if ( sliderOffset <= y )
				return AdjustmentEvent.TRACK;
			return AdjustmentEvent.UNIT_DECREMENT;
		}
		else {
			if ( y - trackBounds.y < sliderOffset )
				return AdjustmentEvent.BLOCK_DECREMENT;
			else if ( y - trackBounds.y > sliderOffset + sliderLength ) 
				return AdjustmentEvent.BLOCK_INCREMENT;
			else {
				return AdjustmentEvent.TRACK;
			}
		}
	case Scrollbar.HORIZONTAL:
		if ( x > d.width - d.height ){
			if ( trackBounds.x + sliderOffset + sliderLength >= x )
				return AdjustmentEvent.TRACK;
			return AdjustmentEvent.UNIT_INCREMENT;
		}
		if ( x < d.height ) {
			if ( sliderOffset <= y )
				return AdjustmentEvent.TRACK;
			return AdjustmentEvent.UNIT_DECREMENT;
		}			
		else { 
			if ( x - trackBounds.x < sliderOffset )
				return AdjustmentEvent.BLOCK_DECREMENT;
			else if ( x - trackBounds.x > sliderOffset + sliderLength ) 
				return AdjustmentEvent.BLOCK_INCREMENT;
			else {
				return AdjustmentEvent.TRACK;
			}
		}
	}

	return 0;
}

public Rectangle sbGetDeco( Scrollbar sb) {
	return new Rectangle();
}

public Rectangle sbGetTrackBounds( Scrollbar sb, Rectangle trackBounds ) {
	//get max bounds of track Rectangle ( excluding spins and deco)
	Dimension d = sb.getSize();
	int db = Defaults.WidgetBorderWidth;
	
	switch( sb.getOrientation() ) {
	case Scrollbar.HORIZONTAL:
		trackBounds.setBounds( 0, db, d.width, d.height-2*db);
		break;
	default:
		trackBounds.setBounds( db, 0, d.width-2*db, d.height);
	}
	
	return( trackBounds);
}

public void sbPaint( Scrollbar sb, Graphics g, Rectangle deco, int flags) {
	sbPaintDecrementButton( sb, g, false);
	sbPaintIncrementButton( sb, g, false);
}

public void sbPaintBackground( Scrollbar sb, Graphics g, int x, int y, int width, int height, int flags) {
        g.setColor( sb.getParent().getBackground() );
        g.fillRect( x, y, width, height);
}

public void sbPaintDecrementButton( Scrollbar sb, Graphics g, boolean active) {
}

public void sbPaintIncrementButton( Scrollbar sb, Graphics g, boolean active) {
}

public void sbPaintSlider( Scrollbar sb, Graphics g, Rectangle trackRect, int sliderOffset, int sliderLength, boolean active ) {
	int x0, y0;

	//nothing to scroll, blank entire widget
	if ( sliderLength == 0 ) {
		Dimension d = sb.getSize();
		g.setColor( sb.getParent().getBackground() );
		g.fillRect( 0, 0, d.width, d.height);
		return;
	}
		
	g.setColor( sb.getBackground() );
		
	if ( sb.getOrientation() == Scrollbar.HORIZONTAL) {
		x0 = trackRect.x;
		y0 = trackRect.y + trackRect.height/2 -1;
	
		//left beam
		if ( sliderOffset > 0 )
			g.draw3DRect( x0, y0, sliderOffset, 1, true);
		
		//slider
		x0 += sliderOffset;
		if ( sliderLength > 0 )
			g.fill3DRect( x0, trackRect.y, sliderLength, trackRect.height, true);
		
		//right beam
		x0 += sliderLength;
		if ( x0 < trackRect.x + trackRect.width )
			g.draw3DRect( x0, y0, trackRect.x + trackRect.width -x0, 1, true);
	}

	else {
		x0 = trackRect.x + trackRect.width/2 -1;
		y0 = trackRect.y;

		//top beam		
		if ( sliderOffset > 0 )
			g.draw3DRect( x0, y0, 1, sliderOffset, true);
			
		//slider
		y0 += sliderOffset;
		if ( sliderLength > 0 )
			g.fill3DRect( trackRect.x, y0, trackRect.width, sliderLength, true);
	
		//bottom beam	
		y0 += sliderLength;
		if ( y0 < trackRect.y + trackRect.height )
			g.draw3DRect( x0, y0, 1, trackRect.y + trackRect.height -y0, true);
	}
}

public Dimension sbPreferredSize( Scrollbar sb, Dimension rawSize ) {
	if( sb.getOrientation() == Scrollbar.HORIZONTAL )
		rawSize.setSize( 100, Defaults.ScrollbarWidth);
	else
		rawSize.setSize( Defaults.ScrollbarWidth, 100);
		
	return( rawSize);
}

public void spPaintBackground( ScrollPane sp, Graphics g, int x, int y, int w, int h, int flags) {
	g.fillRect(x, y, w, h);
}

public Color taGetForegroundColorSelected(TextArea c, boolean selected) {
        if (selected) {
                return (Defaults.TextFieldSelTxtClr);
        }
        else {
                return (c.getForeground());
        }
}

public void taPaintBackground( TextArea ta, Graphics g, int x, int y, int width,
int height, int flags) {
        if (flags != 0) {
                g.setColor (Defaults.TextFieldSelBgClr);
                g.fill3DRect( x, y, width, height, true);
        }
        else {
                g.setColor (ta.getBackground());
                g.fillRect( x, y, width, height);
        }
}

public void taSetDefaults(TextArea ta) {
        ta.setFont( Defaults.TextAreaFont);
        ta.setForeground( Defaults.TextAreaTxtClr);
        ta.setBackground( Defaults.TextAreaBgClr);
}

public Rectangle tfGetDeco( TextField tf) {
	int b = Defaults.WidgetBorderWidth;
	return new Rectangle( b, b, 2*b, 2*b );
}

public Color tfGetForegroundColorSelected(TextField c, boolean selected) {
        if (selected) {
                return (Defaults.TextFieldSelTxtClr);
        }
        else {
                return (c.getForeground());
        }
}

public Dimension tfMinimumSize( TextField tf, Dimension rawSize) {
	rawSize.width += 4 * Defaults.WidgetBorderWidth;
	rawSize.height += 0;
	return (rawSize);
}

public void tfPaintBackground( TextField tf, Graphics g, int x, int y, int width
, int height, int flags) {
        if (flags != 0) {
                g.setColor (Defaults.TextFieldSelBgClr);
                g.fill3DRect( x, y, width, height, true);
        }
        else {
                g.setColor (tf.getBackground());
                g.fillRect( x, y, width, height);
        }
}

public void tfSetDefaults(TextField tf) {
        tf.setFont( Defaults.TextFieldFont);
        tf.setForeground( Defaults.TextFieldTxtClr);
        tf.setBackground( Defaults.TextFieldBgClr);
}

public void wdgPaintBorder ( Component c, Graphics g, Rectangle deco, int width, int height, boolean focused ) {

	g.setColor( focused ? Defaults.FocusClr :Defaults.BorderClr);

	g.drawRect( 0, 0, width-1, height-1);
	g.drawRect( deco.x, deco.y, width-deco.width-1, height-deco.height-1);

//	borderPaintInner( c, g, focused);
}

public void wndPaintBackground(Window wnd,  Graphics g, int x, int y, int width, int height, int flags) {
}

public void wndSetDefaults(Window wnd) {
        wnd.setFont( Defaults.WndFont);
        wnd.setForeground (Defaults.WndForeground);
        wnd.setBackground (Defaults.WndBackground);
}
}
