package kaffe.awt.theme;

import java.awt.Button;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Label;
import java.awt.List;
import java.awt.Rectangle;
import java.awt.ScrollPane;
import java.awt.Scrollbar;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.Window;
import java.awt.event.AdjustmentEvent;
import kaffe.awt.Defaults;

/**
 * sample widget rendering theme
 */
public class WidgetTheme_Win32
  extends WidgetThemeKaffe
{
	static int[] xp = new int[3];
	static int[] yp = new int[3];

public int borderInnerWidth ( Component c ) {
	//width of border for inner elements of compound widgets ( e.g.RowCanvas )
	return( 0);
}

public int borderOuterWidth ( Component c ) {
	//width of outer border for compound widgets ( e.g.List )
	return( Defaults.WidgetBorderWidth);
}

public void borderPaint ( Component c, Graphics g, int left, int top, int right, int bottom, int width, int height, boolean focused ) {
	int w = width - (left + right);
	int h = height - (top + bottom);
	int bw = Defaults.WidgetBorderWidth;
	int i, x0, y0, x1, y1, ww, hh;

	g.setColor( Color.darkGray);
	x0 = left; y0 = top; x1 = left+w-1; y1 = top+h-1;
	for ( i=0; i<bw; i++ ) {
		g.drawLine( x0, y0, x1, y0);
		g.drawLine( x0, y0, x0, y1);
		x0++; y0++; x1--; y1--;
	}

	g.setColor( Color.white );
	x0 = left; y0 = top; x1 = left+w-1; y1 = top+h-1;
	for ( i=0; i<bw; i++ ) {
		if ( i == bw-1 )
			g.setColor( Color.lightGray);
		g.drawLine( x0, y1, x1, y1);
		g.drawLine( x1, y1, x1, y0);
		x0++; y0++; x1--; y1--;
	}
}

public void borderPaintInner ( Component c, Graphics g, boolean focused ) {
	//paint border for inner elements of compound widgets ( e.g. RowCanvas of List )
}

public void borderPaintOuter ( Component c, Graphics g, boolean focused ) {
	Dimension d = c.getSize();
	borderPaint( c, g, 0, 0, 0, 0, d.width, d.height, focused );
}

public Rectangle btnGetDeco( Button btn ) {
	int d = Defaults.WidgetBorderWidth;
	return new Rectangle( d, d, 2*d, 2*d);
}

public void btnPaint( Button btn, Graphics g, Rectangle deco, int flags) {
	Dimension sz = btn.size();
	int di = (flags & WidgetTheme.HAS_FOCUS) > 0 ? 1 : 0;
	int db = 4;
	
	g.setColor( btn.getBackground() );
	g.fill3DRect( di, di, sz.width-2*di, sz.height-2*di, (flags & WidgetTheme.IS_ACTIVE) == 0);

	if ( (flags & WidgetTheme.HAS_FOCUS) > 0 ) {
		g.setColor( Color.black);
		g.drawRect( 0, 0, sz.width-1, sz.height-1);
		paintFocusRect( g, db, db, sz.width-2*db-1, sz.height-2*db-1);
	}
	
	String lbl = btn.getLabel();
	if ( lbl != null ) {
		g.setColor( btn.getForeground() );
		RenderDirection.drawString( lbl, g, 0, 0, sz.width, sz.height, 1);
	}

}

public Dimension btnPreferredSize( Button btn, Dimension rawSize) {
	String label = btn.getLabel();

	//maybe an image button?
	if ( label != null ) {
		FontMetrics fm = btn.getFontMetrics( btn.getFont() );
		int dx = 40;
		int dy = 20;

		if ( RenderDirection.isHorizontal() ) {
			rawSize.width = Math.max( dx, (fm != null) ? 4*fm.stringWidth( label)/3 : 0);
			rawSize.height = Math.max( dy, (fm != null) ? 2*fm.getHeight() : 0);
		}
		else {
			rawSize.width = Math.max( dy, (fm != null) ? 2*fm.charWidth( 'X') : 0);
			rawSize.height = Math.max( dx, (fm != null) ? (label.length()+1)*fm.getHeight() : 0);
		}
	}

	return( rawSize);}

public Rectangle cbGetDeco( Checkbox cb ) {
	return new Rectangle();
}

public void cbPaint( Checkbox cb, Graphics g, Rectangle deco, int flags) {
	FontMetrics fm = cb.getFontMetrics( cb.getFont() );
	Dimension sz = cb.size();
	int bExt = Math.min( sz.height, fm.getHeight() );
	int y0 = (sz.height-bExt)/2;
	
	g.setColor( cb.getBackground() );

	// box
	if ( cb.getCheckboxGroup() == null ) {
		g.draw3DRect( 0, y0, bExt, bExt, false);
		g.setColor( Color.white);
		g.fillRect( 2, y0+2, bExt-3, bExt-3 );
	
		if ( (flags & WidgetTheme.IS_ACTIVE) > 0 ) {
			g.setColor( Color.black);
			g.drawLine( 3, y0+bExt/2, bExt/2, y0+bExt-3);
			g.drawLine( 3, y0+bExt/2-1, bExt/2, y0+bExt-4);
			g.drawLine( bExt/2, y0+bExt-3, bExt-2, y0+5);
			g.drawLine( bExt/2, y0+bExt-4, bExt-2, y0+4);
		}
	}
	// radio
	else {
		g.setColor( cb.getBackground().darker() );
		g.drawArc( 0, y0, bExt, bExt, 45, -180);
		g.setColor( Color.white);
		g.drawArc( 0, y0, bExt, bExt, 225, -180);
	
		g.setColor( Color.white);
		g.fillArc( 1, y0+1, bExt-3, bExt-3, 0, 360);

		if ( (flags & WidgetTheme.IS_ACTIVE) > 0 ) {
			g.setColor( Color.black);
			g.fillArc( bExt/2-2, y0+bExt/2-2, 4, 4, 0, 360);
		}
	}
	
	g.setColor( cb.getForeground() );
	g.drawString( cb.getLabel(), 4*bExt/3, y0+fm.getAscent() );
	
	if ( (flags & WidgetTheme.HAS_FOCUS) > 0 ) {
		g.setColor( Color.black);
		paintFocusRect( g, 4*bExt/3-2, y0, fm.stringWidth( cb.getLabel() ) + 4, fm.getHeight() );
	}
}

public Dimension cbPreferredSize( Checkbox cb, Dimension rawSize) {
	String label = cb.getLabel();
	
	//maybe an image Checkbox?
	if ( label != null ) {
		int dx = 50;
		int dy = 20;
		FontMetrics fm = cb.getFontMetrics( cb.getFont() );
	
		if ( RenderDirection.isHorizontal() ) {
			rawSize.width = Math.max( dx, (fm != null) ? fm.stringWidth( label) + 2*fm.getHeight() : 0);
			rawSize.height = Math.max( dy, (fm != null) ? 3*fm.getHeight()/2 : 0);
		}
		else {
			rawSize.width = Math.max( dy, (fm != null) ? 2*fm.charWidth( 'X') : 0);
			rawSize.height = Math.max( dx, (fm != null) ? (label.length()+3)*fm.getHeight() : 0);
		}
	}
	
	return( rawSize);
}

public Rectangle chGetDeco( Choice ch ) {
	int d = Defaults.WidgetBorderWidth;
	return new Rectangle( d, d, 2*d, 2*d);
}

public boolean chOpenPrompt( Choice ch, int x, int y) {
	return( true);
}

public void chPaint( Choice ch, Graphics g, Rectangle deco, int flags) {
 	Dimension sz = ch.size();
	int d = Defaults.WidgetBorderWidth;
	int db = Defaults.ChoiceBtnWidth;
	
	g.setColor( ch.getBackground() );
	g.fill3DRect ( sz.width-db-d, d, db, sz.height-2*d, (flags & WidgetTheme.IS_ACTIVE) == 0 );
}

public void chPaintPopup(Choice ch, Window popup, Graphics g) {
	popup.getComponent(0).paint(g);
}

public Dimension chPreferredSize( Choice cb, Dimension rawSize) {
	//add additional deco space ( e.g. frame ) to the widgets raw size
	rawSize.width += Defaults.ChoiceBtnWidth + 2*Defaults.WidgetBorderWidth;
	return (rawSize);
}

public Rectangle chTextBounds( Choice ch) {
	Dimension d = ch.getSize();
	return new Rectangle( 0, 0, d.width-Defaults.ChoiceBtnWidth, d.height);
}

public Rectangle lblGetDeco( Label lbl ) {
	return new Rectangle();
}

public void lblPaint( Label lbl, Graphics g, Rectangle deco, int flags ) {
	Dimension sz = lbl.getSize();
	String label = lbl.getText();
	
	g.setColor( lbl.getBackground() );
	g.fillRect( 0, 0, sz.width, sz.height);
	
	g.setColor( lbl.getForeground() );
	RenderDirection.drawString( label, g, 0, 0, sz.width, sz.height, lbl.getAlignment() );
}

public Dimension lblPreferredSize( Label lbl, Dimension rawSize ) {
	String label = lbl.getText();
	
	if ( label != null ) {
		int dx = 40;
		int dy = 20;
		FontMetrics fm = lbl.getFontMetrics( lbl.getFont() );
		
		if ( RenderDirection.isHorizontal() ) {
			rawSize.width = Math.max( dx, (fm != null) ? fm.stringWidth( label) : 0);
			rawSize.height = Math.max( dy, (fm != null) ? 3*fm.getHeight()/2 : 0);
		}
		else {
			rawSize.width = Math.max( dy, (fm != null) ? 3*fm.charWidth( 'X')/2 : 0);
			rawSize.height = Math.max( dx, (fm != null) ? (label.length()+1)*fm.getHeight() : 0);
		}
	}
	
	return( rawSize);
}

private static void paintFocusRect( Graphics g, int x, int y, int w, int h) {
	int dl = 1;
	
	for ( int dx=1; dx<w-dl; dx+=3*dl) {
		g.drawLine( x+dx, y, x+dx+dl, y);
		g.drawLine( x+dx, y+h, x+dx+dl, y+h);
	}
	for ( int dy=1; dy<h-dl; dy+=3*dl) {
		g.drawLine( x, y+dy, x, y+dy+dl);
		g.drawLine( x+w, y+dy, x+w, y+dy+dl);
	}
}

public int sbGetAdjustmentArea( Scrollbar sb, int x, int y, Rectangle trackBounds, int sliderOffset, int sliderLength) {
	Dimension d = sb.getSize();
	
	//get unit adjustment values from position and slider
	switch( sb.getOrientation() ) {
	case Scrollbar.VERTICAL:
		if ( y > d.height - d.width )
			return AdjustmentEvent.UNIT_INCREMENT;
		else if ( y < d.width )
			return AdjustmentEvent.UNIT_DECREMENT;
		else {
			if ( y - trackBounds.y < sliderOffset )
				return AdjustmentEvent.BLOCK_DECREMENT;
			else if ( y - trackBounds.y > sliderOffset + sliderLength ) 
				return AdjustmentEvent.BLOCK_INCREMENT;
			else {
				return AdjustmentEvent.TRACK;
			}
		}
	case Scrollbar.HORIZONTAL:
		if ( x > d.width - d.height )
			return AdjustmentEvent.UNIT_INCREMENT;
		else if ( x < d.height )
			return AdjustmentEvent.UNIT_DECREMENT;
		else { 
			if ( x - trackBounds.x < sliderOffset )
				return AdjustmentEvent.BLOCK_DECREMENT;
			else if ( x - trackBounds.x > sliderOffset + sliderLength ) 
				return AdjustmentEvent.BLOCK_INCREMENT;
			else {
				return AdjustmentEvent.TRACK;
			}
		}
	}

	return 0;
}

public Rectangle sbGetDeco( Scrollbar sb) {
	return new Rectangle();
}

public Rectangle sbGetTrackBounds( Scrollbar sb, Rectangle trackBounds ) {
	//get max bounds of track Rectangle ( excluding spins and deco)
	Dimension d = sb.getSize();
	
	switch( sb.getOrientation() ) {
	case Scrollbar.HORIZONTAL:
		trackBounds.setBounds( d.height, 0, d.width-2*d.height, d.height);
		break;
	default:
		trackBounds.setBounds( 0, d.width, d.width, d.height-2*d.width);
	}
	
	return( trackBounds);
}

public void sbPaint( Scrollbar sb, Graphics g, Rectangle deco, int flags) {
}

public void sbPaintDecrementButton( Scrollbar sb, Graphics g, boolean active) {
	Dimension d = sb.getSize();
	int db = 4;
	
	g.setColor( sb.getBackground() );
	
	if ( sb.getOrientation() == Scrollbar.HORIZONTAL ) {
		g.fill3DRect( 0, 0, d.height, d.height, !active);
		xp[0] = db; yp[0] = d.height/2;
		xp[1] = d.height-db; yp[1] = db;
		xp[2] = xp[1]; yp[2] = d.height-db;
	}
	else {
		g.fill3DRect( 0, 0, d.width, d.width, !active);
		xp[0] = d.width/2-1; yp[0] = db-1;
		xp[1] = d.width-db; yp[1] = d.width-db;
		xp[2] = db-1; yp[2] = yp[1];
	}
	
	Color c = g.getColor();
	g.setColor( Color.black);
	g.fillPolygon( xp, yp, 3);
	g.setColor( c);
}

public void sbPaintIncrementButton( Scrollbar sb, Graphics g, boolean active) {
	int db = 4;
	Dimension d = sb.getSize();
	
	if ( sb.getOrientation() == Scrollbar.HORIZONTAL ) {
		g.fill3DRect( d.width-d.height, 0, d.height, d.height, !active);
		xp[0] = d.width-db;
		xp[1] = d.width-d.height+db;
		xp[2] = xp[1];
		yp[0] = d.height/2;
		yp[1] = db;
		yp[2] = d.height-db;
	}
	else {
		g.fill3DRect( 0, d.height-d.width, d.width, d.width, !active);
		xp[0] = d.width/2;
		xp[1] = d.width-db;
		xp[2] = db;
		yp[0] = d.height-db;
		yp[1] = d.height-d.width+db;
		yp[2] = yp[1];
	}
	
	Color c = g.getColor();
	g.setColor( Color.black);
	g.fillPolygon( xp, yp, 3);
	g.setColor( c);
}

public void sbPaintSlider( Scrollbar sb, Graphics g, Rectangle trackBounds, int sliderOffset, int sliderLength, boolean active ) {
	g.setColor( sb.getBackground() );
	if ( sb.getOrientation() == Scrollbar.HORIZONTAL )
		g.fill3DRect( trackBounds.x+sliderOffset, trackBounds.y, sliderLength, trackBounds.height, true);
	else
		g.fill3DRect( trackBounds.x, trackBounds.y+sliderOffset, trackBounds.width, sliderLength, true);
}

public Dimension sbPreferredSize( Scrollbar sb, Dimension rawSize ) {
	if( sb.getOrientation() == Scrollbar.HORIZONTAL )
		rawSize.setSize( 100, Defaults.ScrollbarWidth);
	else
		rawSize.setSize( Defaults.ScrollbarWidth, 100);
		
	return( rawSize);
}

public Rectangle tfGetDeco( TextField tf) {
	int b = Defaults.WidgetBorderWidth;
	return new Rectangle( b, b, 2*b, 2*b );
}

public void wdgPaintBorder ( Component c, Graphics g, Rectangle deco, int width, int height, boolean focused ) {
	borderPaint( c, g, 0, 0, 0, 0, width, height, focused);
}
}
