package kaffe.awt.widgets;

import java.awt.AWTEventMulticaster;
import java.awt.Button;
import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.ItemSelectable;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Vector;
import kaffe.awt.GenericLayout;
import kaffe.awt.GenericLayoutConstraints;

public class DatePane
  extends Container
  implements ActionListener, ItemSelectable
{
	ItemList lst = new ItemList();
	DateModel dateModel = new DateModel();
	Button btnPrevYear = new Button( "<<");
	Button btnNextYear = new Button( ">>");
	Button btnNextMonth = new Button( ">");
	Button btnPrevMonth = new Button( "<");
	Button btnMonth = new Button( "Month");
	Button btnYear = new Button( "Year");
	ItemListener iListener;
	Date selection;

class DateModel
  implements ItemRender, ItemProcessor
{
	ItemRenderEnv renderEnv = new ItemRenderEnv();
	Date date = new Date();

DateModel() {
}

public void drawItem( ItemRenderEnv env){
	Graphics g = env.graph;
	g.setColor( Color.black);
	
	int dh = (env.curIdx > 0) ? 1 : 0;
	int w = env.rect.width-1;
	g.drawRect( env.rect.x, env.rect.y-dh, w, env.rect.height-1+dh);

	((WeekEntry)env.items.elementAt( env.curIdx)).paint( env);
}

Vector getDates() {
	Calendar c = Calendar.getInstance();

	Vector v = new Vector();
	v.addElement( new WeekHeaderEntry( null) );

	this.date = date;
	
	c.setTime( date);
	c.set( c.DATE, 1);
	int dow = c.get( Calendar.DAY_OF_WEEK) - 1;
	int month =c.get( Calendar.MONTH);
	int ws = 1;
	
	for ( int day = dow; ; day++) {
		if ( month != c.get( Calendar.MONTH) ) {
			v.addElement( new WeekEntry(  dow, ws, day-dow ));
			break;
		}
		if ( day == 7 ) {
			v.addElement( new WeekEntry(  dow, ws, day-dow));
			ws += 7 - dow;
			dow = 0;
			day = 0;
		}
		c.add( Calendar.DATE, 1);
	}

	return v;
}

public int getItemHeight( ItemRenderEnv env){
	//use default item height
	return 0;
}

public int getMaxWidth( ItemRenderEnv env){
	//no horizontal scrollbar
	env.itemWidth = 0;
	return env.itemWidth;
}

public ItemRenderEnv getRenderEnv(){
	return renderEnv;
}

public boolean process( ItemPane pane, int index, int mouXOffs){
	if ( index > 0 ) {	
		WeekEntry we = (WeekEntry)pane.getItems().elementAt( index);
		int dw = (renderEnv.rect.width + renderEnv.rect.x) / 7;
		int day = we.start + Math.min((mouXOffs / dw), 6) - we.off;
	
		if ( (day > 0) && (day < we.start + we.len) ) {
			Calendar c = Calendar.getInstance();
			c.setTime( date);
			c.set( Calendar.DATE, day);
			selection = c.getTime();
			
			ItemEvent ie = new ItemEvent( DatePane.this, ItemEvent.ITEM_STATE_CHANGED,
		                                                  	selection, ItemEvent.SELECTED);
			Toolkit.getDefaultToolkit().getSystemEventQueue().postEvent(  ie);
		}
	}
	return true;
}

void spinMonth( int num) {
	Calendar c = Calendar.getInstance();
	c.setTime( date);
	c.add( Calendar.MONTH, num);
	date = c.getTime();
}

void spinYear( int num) {
	Calendar c = Calendar.getInstance();
	c.setTime( date);
	c.add( Calendar.YEAR, num);
	date = c.getTime();
}
}

public DatePane() {
	this( new Date() );
}

public DatePane( Date d) {
	setLayout( new GenericLayout() );
	
	GenericLayoutConstraints gc = new GenericLayoutConstraints();

	gc.setXCorner( 30, 0, gc.WEST);
	gc.setYCorner( 30, 0, gc.NORTH);
	add( btnPrevYear, gc); 
	btnPrevYear.addActionListener( this); 

	gc.setXOrigin( 30, 0, gc.WEST);
	gc.setXCorner( 60, 0, gc.WEST);
	add( btnPrevMonth, gc);
	btnPrevMonth.addActionListener( this); 

	gc.setXOrigin( 60, 0, gc.WEST);
	gc.setXCorner( 0, 500, gc.WEST);
	add( btnMonth, gc); 

	gc.setXOrigin( 0, 500, gc.EAST);
	gc.setXCorner( 60, 0, gc.EAST);
	add( btnYear, gc); 

	gc.setXOrigin( 60, 0, gc.EAST);
	gc.setXCorner( 30, 0, gc.EAST);
	add( btnNextMonth, gc); 
	btnNextMonth.addActionListener( this); 

	gc.setXOrigin( 30, 0, gc.EAST);
	gc.setXCorner( 0, 0, gc.EAST);
	add( btnNextYear, gc); 
	btnNextYear.addActionListener( this); 

	gc.setXOrigin( 0, 0, gc.WEST);
	gc.setYOrigin( 35, 0, gc.NORTH);
	gc.setYCorner( 0, 0, gc.SOUTH);	
	add( lst, gc);

	lst.setHandler( dateModel);
	setDate( d);
}

public void actionPerformed( ActionEvent e) {
	Object src = e.getSource();
	
	if ( src == btnPrevYear) {
		dateModel.spinYear( -1);
	}
	else if ( src == btnPrevMonth) {
		dateModel.spinMonth( -1);
	}
	else if ( src == btnNextMonth) {
		dateModel.spinMonth( 1);
	}
	else if ( src == btnNextYear) {
		dateModel.spinYear( 1);
	}
	
	updateContents();
}

public void addItemListener ( ItemListener listener ){
	iListener = AWTEventMulticaster.add( iListener, listener);
}

public Object[] getSelectedObjects (){
	return new Object[] { selection};
}

protected void processItemEvent( ItemEvent e) {
	if ( iListener != null ) {
		iListener.itemStateChanged( e);
	}
}

public void removeItemListener ( ItemListener listener ){
	iListener = AWTEventMulticaster.remove(iListener, listener);
}

public void setDate( Date d) {
	dateModel.date = d;
	updateContents();
}

void updateContents() {
	
	Calendar c = Calendar.getInstance();
	c.setTime( dateModel.date);
	
	int month = c.get( Calendar.MONTH);
	int year = c.get( Calendar.YEAR);
	
	SimpleDateFormat sdf = new SimpleDateFormat( "MMM");
	btnMonth.setLabel( sdf.format( dateModel.date) );
	
	btnYear.setLabel( Integer.toString( year) );
	
	lst.setContents( dateModel.getDates() );
}
}

class WeekEntry
{
	int off;
	int start;
	int len;
	static Font fntSmall = new Font( "Dialog", Font.PLAIN, 5);

WeekEntry() {
}

WeekEntry( int off, int start, int len) {
	this.off = off;
	this.start = start;
	this.len = len;
//	System.out.println( off + " " +  start + " " + len);
}

void paint( ItemRenderEnv env) {
	int dw = (env.rect.width + env.rect.x) / 7;
	Graphics g = env.graph;

	g.setFont( fntSmall);
	FontMetrics fm = g.getFontMetrics();

	int x0;
	int y0 = env.rect.y +  fm.getAscent();
	
	for ( int i=0; i<7; i++ ) {
		x0 =  i*dw;
		g.setColor( Color.black);
		g.drawLine( x0, env.rect.y, x0, env.rect.y + env.rect.height - 1);
		if ( i >= off ) {
			if ( i < off + len) {
				String s = Integer.toString(start + i - off);
				if ( (i == 0) || (i == 6) ) {
					g.setColor( Color.red);
				}
				x0 += 3;
				g.drawString( s, x0, y0);
			}
		}
	}
}
}

class WeekHeaderEntry
  extends WeekEntry
{
	static String[] defDayNames = {"Sun", "Mon","Tue","Wed","Thu","Fri","Sat"};
	String[] dayNames;
	static Font fntReg = new Font("Dialog", Font.BOLD, 10);

WeekHeaderEntry( String[] dayNames) {
	if ( dayNames == null) {
		dayNames = defDayNames;
	}
	this.dayNames = dayNames;
}

void paint( ItemRenderEnv env) {
	int dw = (env.rect.width + env.rect.x) / 7;
	Graphics g = env.graph;

	g.setFont( fntReg);
	FontMetrics fm = g.getFontMetrics();
	
	int x0;
	int y0 = env.rect.y + (env.rect.height - fm.getHeight()) / 2 + fm.getAscent();
	
	for ( int i=0; i<7; i++ ) {
		x0 = i*dw;
		g.setColor( Color.black);
		g.drawLine( x0, env.rect.y, x0, env.rect.y + env.rect.height);
		x0 += (dw - fm.stringWidth( dayNames[i])) / 2;
		if ( (i == 0) || (i == 6) ) {
			g.setColor( Color.red);
		}
		g.drawString( dayNames[i], x0, y0);
	}
}
}
