package kaffe.awt.widgets;

import java.awt.AWTEventMulticaster;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.ItemSelectable;
import java.awt.Toolkit;
import java.awt.event.*;
import java.net.URL;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;

public class ImageFolder
  extends Container
  implements ItemListener, ItemSelectable
{
	ItemListener iListener;
	Color highClr;

class Entry
  extends Component
  implements MouseListener, FocusListener, KeyListener
{
	Image img;
	String title1;
	String title2;
	Object arg;

Entry( Image img, String title, Object arg ) {
	this.img = img;
	this.title1 = title;
	this.arg = arg;

	if (title != null) {
		// Split title if we need to
		FontMetrics fm = ImageFolder.this.getFontMetrics(ImageFolder.this.getFont());
		int tw = fm.stringWidth(title);
		if (tw > img.getWidth(null)) {
			int pos = title.lastIndexOf(' ', title.length() / 2);
			if (pos == -1) {
				pos = title.indexOf(' ', title.length() / 2);
			}
			if (pos != -1) {
				title1 = title.substring(0, pos);
				title2 = title.substring(pos+1);
			}
		}
	}
	this.addMouseListener( this);
	this.addFocusListener( this);
	this.addKeyListener( this);
}

public void focusGained( FocusEvent e) {
	this.repaint();
}

public void focusLost( FocusEvent e) {
	this.repaint();
}

public Dimension getPreferredSize() {
	int cx = 0;
	int cy = 0;
	if (img != null){
		cx = img.getWidth(this);
		cy = img.getHeight(this);
	}
	if (title1 != null) {
		FontMetrics fm = ImageFolder.this.getFontMetrics(ImageFolder.this.getFont());
		int tw = fm.stringWidth(title1);
		if (tw > cx) {
			cx = tw;
		}
		cy += fm.getHeight();
		if (title2 != null) {
			tw = fm.stringWidth(title2);
			if (tw > cx) {
				cx = tw;
			}
			cy += fm.getHeight();
		}
	}
	return (new Dimension(cx,cy));
}

private void highlight( Graphics g) {
	Graphics g1 = (g != null) ? g : this.getGraphics();
	g1.setColor( (highClr != null) ? highClr : Defaults.FocusClr);
	g1.drawRect(0, 0, AWT.extension.getWidth(this)-1, AWT.extension.getHeight(this)-1);
	if ( g1 != g) {
		g1.dispose();
	}
}

public boolean isFocusTraversable() {
	return (this.isShowing() && this.isEnabled());
}

public void keyPressed( KeyEvent e) {
}

public void keyReleased( KeyEvent e) {
	switch ( e.getKeyCode() ) {
		case KeyEvent.VK_ENTER:
			if ( iListener != null ) {
				ItemEvent ie = new ItemEvent( ImageFolder.this, ItemEvent.ITEM_STATE_CHANGED, arg, ItemEvent.SELECTED);
				iListener.itemStateChanged( ie);
			}
			break;
	}
}

public void keyTyped( KeyEvent e) {
}

public void mouseClicked( MouseEvent evt) {
}

public void mouseEntered( MouseEvent evt) {
}

public void mouseExited( MouseEvent evt) {
}

public void mousePressed( MouseEvent evt) {
	this.requestFocus();
}

public void mouseReleased( MouseEvent evt) {
	ItemEvent ie = new ItemEvent( ImageFolder.this, ItemEvent.ITEM_STATE_CHANGED, arg, ItemEvent.SELECTED);
	if (highClr != null) {
		highlight( null);
	}
	if ( iListener != null ) {
		iListener.itemStateChanged( ie);
	}
}

public void paint( Graphics g) {
	int cx = img != null ? img.getWidth(this) : 0;
	int cy = img != null ? img.getHeight(this) : 0;
	
	int th = 0;
	int tw1 = 0;
	int tw2 = 0;
	int ta = 0;
	if (title1 != null) {
		FontMetrics fm = ImageFolder.this.getFontMetrics( ImageFolder.this.getFont() );
		tw1 = fm.stringWidth(title1);
		th = fm.getHeight();
		ta = fm.getAscent();
		if (title2 != null) {
			tw2 = fm.stringWidth(title2);
		}
	}

	Dimension d = this.getSize();
	if ( img != null ) {
		g.drawImage( img, (d.width-cx)/2, 0, this);
	}
	if (title1 != null) {
		g.drawString( title1, (d.width-tw1) / 2, cy+ta );
	}
	if (title2 != null) {
		g.drawString( title2, (d.width-tw2) / 2, cy+th+ta );
	}
	
	if ( AWT.keyTgt == this) {
		highlight( g);
	}
}
}

public ImageFolder() {
	this( 0, 2, null);
}

public ImageFolder( int rows, int cols, Color highClr) {
	setLayout( new GridLayout( rows, cols, 0, 0));
	this.highClr = highClr;
}

public void addItem( Image img, String title, Object arg, boolean prepend) {
	Entry e = new Entry( img, title, arg);
	add(e, prepend ? 0 : -1);
}

public void addItem( String imgPath, String title, Object arg, boolean prepend) {
	Image img = (imgPath != null) ? Toolkit.getDefaultToolkit().getImage( imgPath) : null;
	addItem( img, title, arg, prepend);
}

public void addItem( URL imgURL, String title, Object arg, boolean prepend){
	Image img = (imgURL != null ) ? Toolkit.getDefaultToolkit().getImage( imgURL) : null;
	addItem( img, title, arg, prepend);
}

public void addItemListener( ItemListener l ) {
	iListener = AWTEventMulticaster.add( iListener, l);
}

public int getColumns() {
	return ((GridLayout)getLayout()).getColumns();
}

public int getMaxWidth() {
	int width = 0;
	int len = getComponentCount();
	for (int i = 0; i < len; i++) {
		Entry e = (Entry)getComponent(i);
		Dimension d = e.getPreferredSize();
		if (d.width > width) {
			width = d.width;
		}
	}
	// Add in vgap
	return (width + 2 * ((GridLayout)getLayout()).getVgap());
}

public int getRows() {
	return ((GridLayout)getLayout()).getRows();
}

public Object[] getSelectedObjects () {
	return null;
}

public void itemStateChanged( ItemEvent e) {
	if ( iListener != null ){
		iListener.itemStateChanged( e);
	}
}

public boolean removeItem(Object arg) {
	int len = getComponentCount();
	for (int i = 0; i < len; i++) {
		Entry e = (Entry)getComponent(i);
		if (e.arg == arg) {
			remove(i);
			return (true);
		}
	}
	return (false);
}

public boolean removeItem(int arg) {
	int len = getComponentCount();
	if (arg >= len) {
		return (false);
	}
	remove(arg);
	return (true);
}

public void removeItemListener( ItemListener l ) {
	iListener = AWTEventMulticaster.remove( iListener, l);
}

public void setColumns( int cols) {
	((GridLayout)getLayout()).setColumns( cols);
}

public void setRows( int rows) {
	((GridLayout)getLayout()).setRows( rows);
}
}
