package kaffe.awt.widgets;

import java.awt.Image;
import java.util.Vector;

class ItemWrapper
  implements TreeNavigator
{
	static ItemWrapper instance;

public void collapse( Object o){
	((TreeItem)o).collapse();
}

public void expand( Object o, boolean deep){
	((TreeItem)o).expand( deep);
}

public Vector getChildren( Object o){
	return ((TreeItem)o).getChildren();
}

static ItemWrapper getInstance() {
	if ( instance == null ) {
		instance = new ItemWrapper();
	}
	return instance;
}

public int getLevel( Object o){
	return ((TreeItem)o).getLevel();
}

public Object getParent( Object o){
	return ((TreeItem)o).getParent();
}

public boolean hasChildren( Object o){
	return ((TreeItem)o).hasChildren();
}

public boolean isExpanded( Object o){
	return ((TreeItem)o).isExpanded();
}
}

public class TreeHandler
  implements ItemFilter, ItemProcessor, ItemRender
{
	ItemRender render;
	Image iCollapsed;
	Image iExpanded;
	TreeNavigator nav;

public TreeHandler( ItemRender render) {
	this( render, null, null, null);
}

public TreeHandler( ItemRender render, TreeNavigator nav, Image iCollapsed, Image iExpanded) {
	this.iCollapsed = iCollapsed;
	this.iExpanded = iExpanded;
	
	if ( render == null ) {
		render = DefaultItemRender.getInstance();
	}
	
	this.nav = (nav != null) ? nav : ItemWrapper.getInstance();
	this.render = render;
}

public void drawItem( ItemRenderEnv env) {
	drawPrefix( env);
	render.drawItem( env);
}

protected void drawPrefix( ItemRenderEnv env) {
	Object o = env.items.elementAt( env.curIdx);

	int lvl = nav.getLevel( o);
	int pw = getPrefixWidth();
	
	env.xOffs = pw* lvl;
	env.xOffs += drawPrefix( env, o, pw);
}

protected int drawPrefix( ItemRenderEnv env, Object o, int w) {
	int xspc = 0;
	
	if ( nav.hasChildren(o) ) {
		Image img = nav.isExpanded(o) ? iExpanded : iCollapsed;
		if ( img != null ) {
			int ih = img.getHeight( null);
			int x0 = env.rect.x + env.xOffs;
			env.graph.drawImage( img, x0, env.rect.y + (env.rect.height - ih) / 2, null ); 
			xspc += w + 5;
		}
	}
	
	return xspc;
}

public Vector filter( Object contents){
	Vector v = null;
	
	if ( contents instanceof Vector ) {
		v =  (Vector)contents;
	}
	else {
		v = nav.getChildren( contents);
	}
	return v;
}

public int getItemHeight( ItemRenderEnv env){
	return (Math.max( getPrefixHeight(), render.getItemHeight( env)));
}

public int getMaxWidth( ItemRenderEnv env){
	return render.getMaxWidth( env);
}

protected int getPrefixHeight() {
	int mh = 0;
	if ( iCollapsed != null ) {
		mh = iCollapsed.getHeight( null);
	}
	if ( iExpanded != null ) {
		mh = Math.max( mh, iExpanded.getHeight( null) );
	}
	return mh;
}

protected int getPrefixWidth() {
	int mw = 0;
	if ( iCollapsed != null ) {
		mw = iCollapsed.getWidth( null);
	}
	if ( iExpanded != null ) {
		mw = Math.max( mw, iExpanded.getWidth( null) );
	}
	return mw;
}

public ItemRenderEnv getRenderEnv(){
	return render.getRenderEnv();
}

public boolean process( ItemPane pane, int index, int mouXOffs){
	Vector items = pane.getItems();
	Object o = items.elementAt( index);

	int pw = getPrefixWidth();
	int lvl = nav.getLevel( o);
	int x0 = lvl * pw;
	
	if ( (x0 < mouXOffs) && (x0 + pw > mouXOffs )) {
		if ( nav.hasChildren(o)) {
			if ( nav.isExpanded(o)) {
				nav.collapse(o);
				for( int idx  = index+1; idx < items.size(); ) {
					o = items.elementAt( idx);
					if ( nav.getLevel( o) > lvl ) {
						nav.collapse(o);
						items.remove( idx);
					}
					else {
						break;
					}
				}
				pane.setContents( items);
			}
			else {
				Vector si = nav.getChildren(o);
				int sis = si.size();
				nav.expand( o, false);
				for ( int idxs = 0; idxs < sis; idxs++ ) {
					items.insertElementAt( si.elementAt( idxs), ++index);
				}
				pane.setContents( items);
			}
			return true;			
		}
	}
	
	return false;
}
}
