/**
 * DataExchange.java
 *  The DataExchange forms the heart of the sync mechanism and allows
 *  data to be send and received between two machines via a serial port.
 *  For the moment at least one end acts as a server taking URL requests
 *  from the client and sending back the data.
 *
 *  TODO: Make this more generic so both ends can send/receive data.
 *        Also add some resilience to the protocol.
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.comm;

import java.io.*;
import java.net.*;
import kaffe.util.UTF8;

public class DataExchange
	implements Runnable {

private InputStream in;
private OutputStream out;

private boolean active;
private InputStream din = new InputStream() {
	public int read() throws IOException {
		if (active == false) {
			return (-1);
		}
		int val = in.read();
		if (val == 0) {
			val = in.read() - 1;
		}
		if (val == -1) {
			active = false;
		}
		if (active == false) {
			synchronized (DataExchange.this) {
				DataExchange.this.notify();
			}
		}
		return (val);
	}
  };

public DataExchange(SerialPort port) {
	in = port.getInputStream();
	out = port.getOutputStream();
}

public synchronized InputStream getData(String name) {
	while (active == true) {
		try {
			wait();
		}
		catch (InterruptedException _) {
		}
	}
	try {
		byte[] data = UTF8.encode(name);
                out.write(data.length >> 8);
                out.write(data.length & 0xFF);
                out.write(data, 0, data.length);
		active = true;
	}
	catch (IOException _) {
	}
	return (din);
}

public static void startServer(SerialPort port) {
	(new Thread(new DataExchange(port))).start();
}

/**
 * This is the server side of the transaction - it wait to read requests
 * on it's input, then fetches the data associated with the received
 * URL and sends it to the requester.
 */
public void run() {
	for (;;) {
		try {
			int l = in.read() << 8;
			l |= in.read();
			String request = UTF8.decode(in, l & 0xFFFF);
System.out.println("Fetching: " + request);
			try {
				URL u = new URL(request);
				InputStream cin = u.openStream();
				if (cin != null) {
					for (;;) {
						int val = cin.read();
						if (val == -1) {
							break;
						}
						// System.out.print((char)val);
						out.write(val);
						if (val == 0) {
							out.write(1);
						}
					}
				}
			}
			catch (IOException _) {
				_.printStackTrace();
			}
			out.write(0);
			out.write(0);
			// System.out.println();
		}
		catch (IOException e) {
			e.printStackTrace();
			break;
		}
	}
}

}
