/**
 * SerialPort.java
 *  One day this will probably get merged into the javax.comm system but for now
 *  it's a direct handle on the serial ports.
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.comm;

import java.io.*;

public class SerialPort {

static SerialPort[] ports;

private String name;
private int filedesc;
private InputStream in;
private OutputStream out;

static {
	System.loadLibrary("comm");
	init();
	// This should really come from resources
	ports = new SerialPort[1];
	ports[0] = new SerialPort("/dev/ttyS0");
}

private SerialPort(String name) {
	this.name = name;
}

public static SerialPort getInstance(int nr) throws IOException {
	if (nr < 0 || nr >= ports.length) {
		throw new IllegalArgumentException("bad port number");
	}
	SerialPort p = ports[nr];
	p.open();
	return (p);
}

public synchronized InputStream getInputStream() {
	if (in == null) {
		in = new InputStream() {

			private byte[] bt = new byte[1];

			public int read() throws IOException {
				if (readBytes(bt, 0, 1) == 1) {
					return ((int)bt[0] & 0xFF);
				}
				else {
					return (-1);
				}
			}

			public int read(byte[] buf, int start, int len) throws IOException {
				if (start < 0 || start+len > buf.length) {
					throw new IOException("bad array");
				}
				return (readBytes(buf, start, len));
			}

			public void close() {
				SerialPort.this.close();
			}
		};
	}
	return (in);
}

public synchronized OutputStream getOutputStream() {
	if (out == null) {
		out = new OutputStream() {

			private byte[] bt = new byte[1];

			public void write(int data) throws IOException {
				bt[0] = (byte)data;
				writeBytes(bt, 0, 1);
			}

			public void write(byte[] buf, int start, int len) throws IOException {
				if (start < 0 || start+len > buf.length) {
					throw new IOException("bad array");
				}
				writeBytes(buf, start, len);
			}

			public void close() {
				SerialPort.this.close();
			}
		};
	}
	return (out);
}

public void finalize() throws Throwable {
	close();
	super.finalize();
}

public native void close();

private native static void init();
private native void open() throws IOException;
private native int readBytes(byte[] data, int start, int len);
private native int writeBytes(byte[] data, int start, int len);

}
