/**
 * SoundPort.java
 *  One day this will probably get merged into the javax.comm system but for now
 *  it's a direct handle on the serial ports.
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.comm;

import java.io.*;

public class SoundPort {

static SoundPort[] ports;
static String[] formats = new String[]{ "audio/x-pcm" };

private String name;
private int filedesc;
private InputStream in;
private OutputStream out;

static {
	System.loadLibrary("comm");
	init();
	// This should really come from resources
	ports = new SoundPort[1];
	ports[0] = new SoundPort("/dev/dsp");
}

private SoundPort(String name) {
	this.name = name;
}

public static SoundPort getInstance(int nr) throws IOException {
	if (nr < 0 || nr >= ports.length) {
		throw new IllegalArgumentException("bad port number");
	}
	SoundPort p = ports[nr];
	p.open();
	return (p);
}

public synchronized OutputStream getOutputStream() {
	if (out == null) {
		out = new OutputStream() {

			private byte[] bt = new byte[1];

			public void write(int data) throws IOException {
				bt[0] = (byte)data;
				writeBytes(bt, 0, 1);
			}

			public void write(byte[] buf, int start, int len) throws IOException {
				if (start < 0 || start+len > buf.length) {
					throw new IOException("bad array");
				}
				writeBytes(buf, start, len);
			}

			public void close() {
				SoundPort.this.close();
			}
		};
	}
	return (out);
}

public String[] getOutputFormats() {
	return (formats);
}

public void finalize() throws Throwable {
	close();
	super.finalize();
}

public native void close();
public native void setChannels(int chan);
public native void setSpeed(int speed);
private native static void init();
private native void open() throws IOException;
private native int writeBytes(byte[] data, int start, int len);

}
