/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * [SBC]
 *
 * $Id somefile.java,v 1.10 2001/02/07 00:40:04 somebody Exp $ 
 */
package kaffe.io;

import java.io.IOException;
import java.io.CharConversionException;

class ByteToCharFilterEUCJP extends ByteToCharFilter 
{
  private static int JIS0212ToChar (int ch1, int ch2, int offset) 
    throws CharConversionException 
  {
    if (ch1 < 0 || ch1 < offset 
        || (ch1 - offset) < 0x22 || (ch1 - offset) > 0x6d)
      {
        throw new CharConversionException();
      }
    if (ch2 < 0 || ch2 < offset 
        || (ch2 - offset) <= 0x20 || (ch2 - offset) >= 0x7f)
      {
        throw new CharConversionException();
      }

    int idx = (ch1 - offset - 0x21) * 94 + (ch2 - offset - 0x21);
    int rtn = CharTables.JIS0212ToChar (idx);
    if (rtn == -1)
      {
        throw new CharConversionException();
      }
    return rtn;
  }

  private static int JIS0208ToChar (int ch1, int ch2, int offset)
    throws CharConversionException
  {
    if (ch1 < 0 || ch1 < offset || (ch1 - offset) <= 0x20) 
      {
        throw new CharConversionException();
      }
    if (ch2 < 0 || ch2 < offset || (ch2 - offset) <= 0x20 || (ch2 - offset) >= 0x7f)
      {
        throw new CharConversionException();
      }

    int idx = (ch1 - 0x21 - offset) * 94 + (ch2 - 0x21 - offset);

    int rtn = CharTables.JIS0208ToChar (idx);
    if (rtn == -1)
      {
        throw new CharConversionException();
      }
    return rtn;
  }

  private static int JIS0201ToChar (int ch)
    throws CharConversionException 
  {
    int rtn = CharTables.JIS0201ToChar (ch);
    if (rtn == 0 && ch != '\0')
      {
        rtn = -1;
      }
    if (rtn == -1)
      {
        throw new CharConversionException();
      }
    return rtn;
  }

  public int read(char cbuf[], int off, int len) throws IOException 
  {
    int pos = off;
    while (pos < off+len) 
      {
        /* convert EUCJP to UCS */
        int ch1 = super.read();
        if (ch1 == -1)
          {
            /* EOF */
            if (pos == off) 
              {
                return (-1);
              }
            return (pos - off);
          }
        else if (ch1 < 0x8e || (ch1 >= 0x90 && ch1 <= 0x9f))
          {
            /* single byte char */
            cbuf[pos] = (char)ch1;
            pos++;
            continue;
          }
        else if (ch1 == 0xff)
          {
            throw new CharConversionException();
          }        

        /* multi-byte char */
        int ch2 = super.read();

        /* All second bytes of a multibyte character must be >= 0xa1. */
        if (ch2 == -1)
          {
            /* EOF */
            if (pos == off) 
              {
                return (-1);
              }
            return (pos - off);
          }
        else if (ch2 < 0xa1)
          {
            throw new CharConversionException();
          }

        /* which king of JIS to we have */
        if (ch1 == 0x8e)
          {
            /* This is code set 2: half-width katakana.  */
            System.err.println ("UNTESTED: JIS X 0201: 0x" + Integer.toHexString (ch2));
            int rtn = JIS0201ToChar (ch2);
            cbuf[pos] = (char)rtn;
            pos++;
          }
        else
          {
            if (ch1 == 0x8f)
              {
                /* This is code set 3: JIS X 0212-1990.  */
                System.err.println ("UNTESTED: JIS X 0212-1990: 0x" + Integer.toHexString (ch1) + " 0x" + Integer.toHexString(ch2));
                int rtn = JIS0212ToChar (ch1, ch2, 0x80);
                cbuf[pos] = (char)rtn;
                pos++;
              }
            else
              {
                /* This is code set 1: JIS X 0208.  */
                //System.err.println ("JIS X 0208: 0x" + Integer.toHexString(ch) + " 0x" + Integer.toHexString(ch2));
                int rtn = JIS0208ToChar (ch1, ch2, 0x80);
                cbuf[pos] = (char)rtn;
                pos++;
              }
          }
      }
    /* convert UCS to UNICODE */
    return (len);
  }

}
