/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * [SBC]
 *
 * $Id somefile.java,v 1.10 2001/02/07 00:40:04 somebody Exp $ 
 */
package kaffe.io;

import java.io.IOException;
import java.io.CharConversionException;

/**
 * Convert java Strings to EUC-JP encoded byte arrays.
 * EUC-JP is the standard used by many japanese text input/output systems.
 */
class CharToByteFilterEUCJP extends CharToByteFilter 
{
  public byte[] CharToJIS0201 (int ch)
    throws CharConversionException
  {
    byte[] rtn = new byte[1];
    if (ch == 0xa5) 
      {
        rtn[0] = 0x5c;
      }
    else if (ch == 0x203e)
      {
        rtn[0] = 0x7e;
      }
    else if (ch < 0x7e && ch != 0x5c)
      {
        rtn[0] = (byte)ch;
      }
    else if (ch >= 0xff61 && ch <= 0xff9f)
      {
        rtn[0] = (byte)(ch - 0xfec0);
      }
    else
      {
        throw new CharConversionException ();
      }
    return rtn;
  }

  public byte[] CharToJIS0208 (int ch)
    throws CharConversionException
  {
    byte[] rtn = CharTables.CharToJIS0208 (ch);

    if (rtn == null)
      {
        throw new CharConversionException ();
      }

    return rtn;
  }

  public byte[] CharToJIS0212 (int ch)
    throws CharConversionException
  {
    byte[] rtn = CharTables.CharToJIS0212 (ch);

    if (rtn == null)
      {
        throw new CharConversionException ();
      }

    return rtn;
  }

  public void write(char cbuf[], int off, int len) throws IOException 
  {
    for (int i = 0; i < len; i++) 
      {
        int ch = cbuf[off+i];
        if (ch < 0x8e || (ch >= 0x90 && ch <= 0x9f))      
          {
            /* regular ASCII or C1.  */
            super.write(ch);
          }
        else if (ch == 0xa5)   
          {
            /* YEN sign => backslash  */
            super.write(0x5c);
          }
        else if (ch == 0x203e) 
          {
            /* overscore => asciitilde */      
            super.write(0x7e);
          }
        else
          {
            /* Try the JIS character sets. */
            byte[] jis;                            
            try
              {
                jis = CharToJIS0201 (ch);
                /* Yes, we can use a JIS 0201 character. */
                System.err.println ("UNTESTED: writing JIS 0201 char (using shift byte)");
                super.write (0x8e);
                super.write (jis[0]);
              }
            catch (CharConversionException e)
              {
                /* No JIS 0201 character found. */
                try
                  {
                    jis = CharToJIS0208 (ch);
                    /* Yes, we can use a JIS 0208 character, adjust it for EUC-JP. */
                    super.write (jis[0] + 0x80);
                    super.write (jis[1] + 0x80);
                  }
                catch (CharConversionException e2)
                  {
                    /* No JIS 0208 character. */
                    jis = CharToJIS0212 (ch);
                    /* It's a JIS 0212 character, adjust it for EUC-JP.  */
                    System.err.println ("UNTESTED: writing JIS 0212 char (using shift byte)");
                    super.write (0x8f);
                    super.write (jis[0] + 0x80);
                    super.write (jis[1] + 0x80);
                  }
              }
          }
      }
  }
}
