package kaffe.io;

import java.io.FileDescriptor;
import java.io.SyncFailedException;
import java.io.IOException;
import java.io.FileNotFoundException;

public class FileIO
	extends FileDescriptor {

public static final int READ = 1;
public static final int WRITE = 2;
public static final int APPEND = 4;

public static final int START = 0;
public static final int HERE = 1;
public static final int END = 2;

public static final int EXISTS = 0;
public static final int FILE = 1;
public static final int DIRECTORY = 2;
public static final int CANREAD = 3;
public static final int CANWRITE = 4;

static {
	System.loadLibrary("io");
	init0();
}

public FileIO() {
}

public FileIO(int fd) {
	this.fd = fd;
}

public FileIO(String filename, int mode) throws FileNotFoundException {
	if (filename == null) {
		throw new IllegalArgumentException("null filename");
	}
	if ((mode & READ) != 0) {
		System.getSecurityManager().checkRead(filename);
	}
	if ((mode & WRITE) != 0) {
		System.getSecurityManager().checkWrite(filename);
	}
	fd = open0(filename, mode);
	if (fd == -1) {
		throw new FileNotFoundException(filename);
	}
}

private native int open0(String filename, int mode);

public int read(byte[] buffer, int offset, int length) throws IOException {
	if (offset < 0 || offset + length > buffer.length) {
		throw new ArrayIndexOutOfBoundsException();
	}
        // make sure we return -1 when we are at the EOF even if
        // 0 bytes were requested.
        if (length == 0 && available () == 0) {
                return -1;
        }
	int r = read0(fd, buffer, offset, length);
	if (r < 0) {
		throw new IOException("read failed");
	}
	else if (r > 0 || length == 0) {
		return (r);
	}
	else {
		return (-1);
	}
}

public int read() throws IOException {
	byte b[] = new byte[1];
	if (read(b, 0, 1) == 1) {
		return (b[0] & 0xFF);
	}
	else {
		return (-1);
	}
}

private native int read0(int fd, byte[] buffer, int offset, int length) throws IOException;

		
public int setMode(int mode) throws IOException {
	// WATCH OUT - this is a dangerous method, because it can easily be used to
	// turn non-blocking in to blocking IO mode (or vice versa), which might
	// heavily conflict with the Kaffe threading mode. It is just here to
	// deal with incomplete writes, which might esp. occur in a producer/consumer
	// scenario involving fifos (e.g. the RTSP mpeg-2 player)
	return setMode0( fd, mode);
}
		
private native int setMode0(int fd, int mode) throws IOException;

public int write(byte[] buffer, int offset, int length) throws IOException {
	if (offset < 0 || offset + length > buffer.length) {
		throw new ArrayIndexOutOfBoundsException();
	}
	return (write0(fd, buffer, offset, length));
}

public void writeFully(byte[] buffer, int offset, int length) throws IOException {
	if (offset < 0 || offset + length > buffer.length) {
		throw new ArrayIndexOutOfBoundsException();
	}
	for (int i = 0; i < length; ) {
		int l = write0(fd, buffer, offset+i, length-i);
		if (l <= 0) {
			throw new IOException("write failed: expected length="+length+", actual="+i);
		}
		i += l;
	}
}

public int write(int val) throws IOException {
	return (write(new byte[] { (byte)val } , 0, 1));
}

private native int write0(int fd, byte[] buffer, int offset, int length) throws IOException;

public long skip(long amount) throws IOException {
	try {
		long oldpos = where();
		long newpos = seek(amount, HERE);
		return (newpos - oldpos);
	}
	catch (IOException _) {
	}
	// Not seekable - so we just read the data away
	byte[] buf = new byte[100];
	long res = 0;
	while (amount > 0) {
		int l = buf.length;
		if (amount < l) {
			l = (int)amount;
		}
		l = read0(fd, buf, 0, l);
		if (l == -1) {
			throw new IOException("seek failed");
		}
		if (l == 0) {
			break;
		}
		res += l;
		amount -= l;
	}
	return (res);
}

public long seek(long amount, int mode) throws IOException {
	long newpos = seek0(fd, amount, mode);
	if (newpos == -1) {
		throw new IOException("seek failed");
	}
	return (newpos);
}

private native long seek0(int fd, long amount, int mode) throws IOException;

public long where() throws IOException {
	return (seek(0L, HERE));
}

public void close() {
	if (fd >= 0) {
		int ofd = fd;
		fd = -1;
		close0(ofd);
	}
}

private native void close0(int fd);

public int available() throws IOException {
	return (available0(fd));
}

private native int available0(int fd);

public long length() {
        return (length0(fd));
}

private native long length0(int fd);

private native static void init0();

}
