/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.lang;

import java.lang.String;
import java.io.FileDescriptor;
import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import kaffe.io.FileIO;

public class UNIXProcess
  extends Process
  implements Runnable
{
	boolean isalive;
	int exit_code;
	int stdin_fd;
	int stdout_fd;
	int stderr_fd;
	int pid;
	OutputStream raw_stdin;
	InputStream raw_stdout;
	InputStream raw_stderr;

public UNIXProcess() {
}

public UNIXProcess(String argv[], String arge[]) {

	/* We first create a thread to start the new process in.  This
	 * is because on some system we can only wait for the child from
	 * it's parent (UNIX for example).  So, we have to create a thread
	 * to do the waiting.
	 */
	final String _argv[] = argv;
	final String _arge[] = arge;
	Thread sitter = new Thread() {
		public void run() {
			Thread t = this;
			try {
				pid = forkAndExec0(_argv, _arge);
				if (pid != -1) {
					synchronized(t) {
						isalive = true;
						t.notify();
					}
					exit_code = execWait0();
				}
			}
			catch (IOException e) {
			}
			synchronized(t) {
				isalive = false;
				t.notifyAll();
			}
			synchronized(UNIXProcess.this) {
				UNIXProcess.this.notifyAll();
			}
		}
	};

	/* Start the sitter then wait until it says it's child has started.
	 * We then retrieve the childs connection information.
	 */
	synchronized(sitter) {
		try {
                        sitter.start();
			sitter.wait();
		}
		catch (InterruptedException _) {
		}

		// Create streams from the file descriptors
		raw_stdin = new FileOutputStream(new FileIO(stdin_fd));
		raw_stdout = new FileInputStream(new FileIO(stdout_fd));
		raw_stderr = new FileInputStream(new FileIO(stderr_fd));
	}
}

public void destroy() {
	destroy0(pid);
}

public int exitValue() {
	if (isalive) {
		throw new IllegalThreadStateException();
	}
	return exit_code;
}

public void run() {
}

public InputStream getErrorStream() {
	return raw_stderr;
}

public InputStream getInputStream() {
	return raw_stdout;
}

public OutputStream getOutputStream() {
	return raw_stdin;
}

public int waitFor() throws InterruptedException {
	synchronized(this) {
		while (isalive == true) {
			wait();
		}
	}
	return (exit_code);
}

/**
 * Native callback.
 */
private void processDied(int status) {
	synchronized(this) {
		isalive = false;
		exit_code = status;
		notifyAll();
	}
}

native public void destroy0(int pid);
native private int forkAndExec0(Object cmd[], Object env[]) throws IOException;
native private int execWait0();

}
