/*
 * Java core library component.
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.lang.ref;

import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;

public class Recycle {

private static Stack freelist = new Stack();
private static Vector activelist = new Vector();

/**
 * We want to recycle the object we're given so that when it would be
 * freed by the GC, it get's it's recycle method called instead.
 */
public static void add(Recycleable obj) {
	RecycleReference ref;
	synchronized (freelist) {
		if (!freelist.empty()) {
			ref = (RecycleReference)freelist.pop();
		}
		else {
			ref = new RecycleReference();
		}
	}
	ref.init(obj);
	activelist.addElement(ref);
}

/**
 * Remove the recylceable object from the recyle system.
 */
public static void remove(Recycleable obj) {
	Enumeration enum = activelist.elements();
	while (enum.hasMoreElements()) {
		RecycleReference ref = (RecycleReference)enum.nextElement();
		if (ref.get() == obj) {
			activelist.removeElement(ref);
			ref.clear();
			freelist.push(ref);
			break;
		}
	}
}

/**
 * Recyle the object.
 */
static void recycle(Recycleable obj) {
	remove(obj);
	obj.recycle();
}

/**
 * When a recyle reference is the only thing referencing an object then
 * this methods will be called with the recyle reference.  We get the object
 * to be recycled, clear the reference and recyle it before recycling the
 * object itself.
 */
static void dispatch(RecycleReference ref) {
	Recycleable obj = (Recycleable)ref.get();
	activelist.removeElement(ref);
	ref.clear();
	freelist.push(ref);
	obj.recycle();
}

}
