package kaffe.media;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Window;
import kaffe.awt.AWT;
import kaffe.media.TVCtrl;

public class TVPlayer
  extends Component
{
	static TVCtrl ctrl = TVCtrl.getDefault();
	String tuner;
	int flags;
	final static int PICTURE_SET = 0x01;
	final static int TUNER_SET = 0x02;
	final static int STARTED = 0x04;
	final static int PLAYING = 0x10;
	final static int SUSPENDED = 0x20;

public TVPlayer () {
	setTuner( null);
	setBackground( TVCtrl.getDefault().getChromaKey());
}

void blank() {
	Graphics g = getGraphics();
	if ( g !=null ) {
		AWT.extension.paintBackground( this, g);
		g.dispose();
	}
}

public synchronized void hide () {
	if ( (flags & PLAYING) != 0 ){
		suspend();
	}

	super.hide();
}

public boolean isRunning () {
	return ((flags & PLAYING) != 0);
}

public synchronized void paint ( Graphics g ) {
	if ( (flags & (PLAYING|STARTED)) == STARTED ){
		start();
	}
	g.clearRect( 0, 0, AWT.extension.getWidth( this), AWT.extension.getHeight( this));
}

public synchronized void removeNotify () {
	if ( (flags & PLAYING) != 0 ){
		stop();
	}
	super.removeNotify();
}

public void reshape ( int x, int y, int width, int height ) {
	int oldWidth = AWT.extension.getWidth( this);
	int oldHeight = AWT.extension.getHeight( this);
	boolean isPlaying = ((flags & PLAYING) != 0);

	super.reshape( x, y, width, height);

	if ( (width != oldWidth) || (height != oldHeight) ){
		if ( isPlaying ){
			ctrl.setPreview( 0);
			flags &= ~PLAYING;
		}

		if ( (width > 0) && (height > 0) ){
			flags &= ~PICTURE_SET;
			setPicture();

			if ( isPlaying ){
				// we will get a subsequent paint, which will (re)start us
				flags |= STARTED;
			}
		}
	}	
}

public synchronized void resume () {
	if ( ((flags & SUSPENDED) != 0) && isShowing() ) {
		flags &= ~SUSPENDED;
		ctrl.setPreview( 1);
	}
}

public void setChromaKey ( Color c) {
	if ( c != null ) {
		setBackground( c);
		ctrl.setChromaKey( c.getRGB() );
	}
}

void setPicture () {

	if ( (flags & PICTURE_SET) == 0 ){

		// ok, this is the tribute to pocketlinux html layouting
		// where parent chain during layout phase might get broken
		Container c = null;
		for( c=getParent();(c != null) && (c.getParent() != null); c=c.getParent() );
		if ( (c == null) || ( ! (c instanceof Window) ) ) {
			return;
		}

		int   w  = AWT.extension.getWidth( this);
		int   h  = AWT.extension.getHeight( this);
		if (( w <= 0) || (h <= 0) ) {
			return;
		}

		Point p = getLocationOnScreen();
		Insets in = AWT.extension.getInsets( this);
		int cw = w - (in.left + in.right);
		int ch = h - (in.top + in.bottom);

	        ctrl.setBounds( p.x+in.left, p.y+in.top, cw, ch);

		flags |= PICTURE_SET;
	}
}

public synchronized void setTuner ( String fname ) {
	boolean isRunning = ((flags & PLAYING) != 0);

	tuner = fname;

	if ( isRunning ) {
		stop();
	}

	flags |= TUNER_SET;
	flags &= ~(PLAYING | STARTED | PICTURE_SET);

	if ( isRunning ) {
		start();
	}
}

public synchronized void show () {
	super.show();

	// if it was suspended, resume
	if ( (flags & SUSPENDED) != 0 ){
		resume();
	}	
}

public synchronized void start () {
	if ( (flags & SUSPENDED) != 0 ){
		resume();
		return;
	}
	
	if ( startTuner() && isShowing() ) {
		setPicture();
		flags |= PLAYING;
		ctrl.setPreview( 1);
	}

}

public boolean startTuner () {
	synchronized ( this ) {
		if ( (flags & (TUNER_SET|STARTED)) == TUNER_SET ){
		    ctrl.setTuner( tuner);
		    flags |= STARTED;
		}
		return true;
	}
}

public synchronized void stop () {
	if ( (flags & PLAYING) != 0 ){
		ctrl.setPreview( 0);
		ctrl.setTuner(null);
		flags &= ~(PLAYING | STARTED | PICTURE_SET);
		blank();
	}
}

public synchronized void suspend () {
	if ( (flags & PLAYING) != 0 ){
		ctrl.setPreview( 0);
		flags |= SUSPENDED;
	}
}

public String toString () {
	StringBuffer sb = new StringBuffer(100);

	sb.append( "TVPlayer ");
	sb.append( Integer.toHexString(hashCode()));
	sb.append( " [");

	if ( (flags & PICTURE_SET) != 0 ) sb.append( "pict_set ");
	if ( (flags & TUNER_SET) != 0 ) sb.append( "src_set ");
	if ( (flags & STARTED) != 0 ) sb.append( "started ");
	if ( (flags & SUSPENDED) != 0 ) sb.append( "susp ");

	sb.append("]");

	return sb.toString();
}
}
