package kaffe.media;

import kaffe.awt.AWT;
import kaffe.util.Ptr;

public class VideoCtrl
{
	Ptr nativeData;
	static VideoCtrl defCtrl;

static {
	System.loadLibrary( "media");
}

private VideoCtrl () {
}

public void close() {
	if ( nativeData != null ) {
		close0( nativeData);
		nativeData = null;
	}
}

native private void close0( Ptr nativeData);

public void finalize() throws Throwable {
	super.finalize();
	close();
}

public int getBrightness() {
	if ( nativeData != null ) {
		return getBrightness( nativeData);
	}
	return 0;
}

native private int getBrightness( Ptr nativeData);

public static VideoCtrl getDefault() {
	String decoder = System.getProperty( "video_decoder");
	
	//will return (nativeData = null) if not defined
	if ( defCtrl == null ) {
		defCtrl = new VideoCtrl();
		defCtrl.open( decoder);
	}
	return defCtrl;
}

public int getPictureHeight() {
	if ( nativeData != null ){
		return getPictureHeight( nativeData);
	}
	return -1;
}

native private int getPictureHeight( Ptr nativeData);

public int getPictureWidth() {
	if ( nativeData != null ) {
		return getPictureWidth( nativeData);
	}
	return -1;
}

native private int getPictureWidth( Ptr nativeData);

public int getVolume( ) {
	if ( nativeData != null ){
		return getVolume( nativeData);
	}
	return 0;
}

native private int getVolume( Ptr nativeData);

native private void nextFrame( Ptr nativeData, int timeout);

public void nextFrame( int timeout) {
	if ( nativeData != null ){
		nextFrame( nativeData, timeout);
	}
}

public void open( String decoder) {
	if ( nativeData == null ) {
		nativeData = open0( decoder);
	}
}

native private Ptr open0( String decoder);

public void resume() {
	if ( nativeData != null ) {
		resume( nativeData);
	}
}

native private void resume( Ptr nativeData);

native private void setBounds( Ptr nativeData, int x, int y, int w, int h);

public void setBounds( int x, int y,  int w, int h ) {
	if ( nativeData != null ) {
		int xOff = 0;//AWT.extension.getScreenOffset().x;
		int yOff = 0;//AWT.extension.getScreenOffset().y;
		setBounds( nativeData, x+xOff, y+yOff, w, h);
	}
}

native private void setBrightness( Ptr nativeData, int val);

public void setBrightness( int val ) {
	if ( nativeData != null ) {
		setBrightness( nativeData, val);
	}
}

native private void setChromaKey( Ptr nativeData, int rgb);

public void setChromaKey( int rgb) {
	if ( nativeData != null ) {
		setChromaKey( nativeData, rgb );
	}
}

native private void setVolume( Ptr nativeData, int val);

public void setVolume( int val ) {
	if ( nativeData != null ) {
		setVolume( nativeData, val);
	}
}

native private void start( Ptr nativeData, String name);

public void start( String name) {
	if ( nativeData != null ) {
		start( nativeData, name);
	}
}

public boolean start( String name, int waitExtendsTimeout) {
	if ( nativeData != null ) {
		start( nativeData, name);

		if ( waitExtendsTimeout > 0 ) {
			for ( int i=waitExtendsTimeout; i>0; i-=30 ) {
				if ( getPictureWidth( nativeData) > 0 ) {
					return true;
				}
				try{ Thread.sleep( 30); }
				catch ( Exception _) {}
			}
			return false;
		}
		return true;
	}
	return false;
}

public void stop() {
	if ( nativeData != null ) {
		stop( nativeData);
	}
}

native private void stop( Ptr nativeData);

public void suspend() {
	if ( nativeData != null ) {
		suspend( nativeData);
	}
}

native private void suspend( Ptr nativeData);

public void suspendSyncRender( ) {
	//10 fps should be a reasonable minimum
	suspendSyncRender( 100);
}

public void suspendSyncRender( int timeout) {
	if ( nativeData != null ) {
		suspend( nativeData);
		if ( timeout > 0 ) {
		   nextFrame( nativeData, timeout);
		}
	}
}
}
