package kaffe.media;

import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Window;
import kaffe.awt.AWT;
import kaffe.media.VideoCtrl;

public class VideoPlayer
  extends Component
{
	static VideoCtrl ctrl = VideoCtrl.getDefault();
	int flags;
	String source;
	final static int PICTURE_SET = 0x01;
	final static int SOURCE_SET = 0x02;
	final static int STARTED = 0x04;
	final static int ZOOM_PICTURE = 0x08;
	final static int PLAYING = 0x10;
	final static int SUSPENDED = 0x20;

public VideoPlayer () {
	this( false);
}

public VideoPlayer ( boolean zoomMode ) {
	if ( zoomMode ) {
		flags |= ZOOM_PICTURE;
	}
}

void blank() {
	Graphics g = getGraphics();
	if ( g !=null ) {
		AWT.extension.paintBackground( this, g);
		g.dispose();
	}
}

public synchronized void hide () {
	if ( (flags & PLAYING) != 0 ){
		suspend();
	}

	super.hide();
}

public boolean isFocusTraversable () {
	return true;
}

public boolean isRunning () {
	return ((flags & PLAYING) != 0);
}

public boolean loadSource () {
	synchronized ( this ) {
		if ( (flags & (SOURCE_SET|STARTED)) == SOURCE_SET ){
			if ( ctrl.start( source, 1500) ) {
				flags |= STARTED;
				return true;
			}
			return false;
		}
		return true;
	}
}

public synchronized void paint ( Graphics g ) {
	if ( (flags & (PLAYING|STARTED)) == STARTED ){
		start();
	}
}

public synchronized void removeNotify () {
	if ( (flags & PLAYING) != 0 ){
		stop();
	}
	super.removeNotify();
}

public void reshape ( int x, int y, int width, int height ) {
	int oldWidth = AWT.extension.getWidth( this);
	int oldHeight = AWT.extension.getHeight( this);
	boolean isPlaying = ((flags & PLAYING) != 0);

	super.reshape( x, y, width, height);

	if ( (width != oldWidth) || (height != oldHeight) ){
		if ( isPlaying ){
			ctrl.suspendSyncRender();
			flags &= ~PLAYING;
		}

		if ( (width > 0) && (height > 0) ){
			flags &= ~PICTURE_SET;
			setPicture();

			if ( isPlaying ){
				// we will get a subsequent paint, which will (re)start us
				flags |= STARTED;
			}
		}
	}	
}

public synchronized void resume () {
	if ( ((flags & SUSPENDED) != 0) && isShowing() ) {
		flags &= ~SUSPENDED;
		ctrl.resume();
	}
}

void setPicture () {

	if ( (flags & PICTURE_SET) == 0 ){

		// ok, this is the tribute to pocketlinux html layouting
		// where parent chain during layout phase might get broken
		Container c = null;
		for( c=getParent();(c != null) && (c.getParent() != null); c=c.getParent() );
		if ( (c == null) || ( ! (c instanceof Window) ) ) {
			return;
		}

		int   w  = AWT.extension.getWidth( this);
		int   h  = AWT.extension.getHeight( this);
		if (( w <= 0) || (h <= 0) ) {
			return;
		}

		Point p = getLocationOnScreen();

		if ( (flags & ZOOM_PICTURE) == 0) {
			int wp = ctrl.getPictureWidth();
			int hp = ctrl.getPictureHeight();
			if ( wp == 0 || hp == 0 ) {
				return;
			}
			if ( wp > w ) wp = w;
			if ( hp > h )  hp = h;
			ctrl.setBounds( p.x + (w - wp) / 2, p.y + (h - hp) / 2, wp, hp);
		}
		else {
			Insets in = AWT.extension.getInsets( this);
			int cw = w - (in.left + in.right);
			int ch = h - (in.top + in.bottom);

			ctrl.setBounds( p.x+in.left, p.y+in.top, cw, ch);
		}

		flags |= PICTURE_SET;
	}
}

public synchronized void setSource ( String fname ) {
	boolean isRunning = ((flags & PLAYING) != 0);

	source = fname;

	if ( isRunning ) {
		stop();
	}

	flags |= SOURCE_SET;
	flags &= ~(PLAYING | STARTED | PICTURE_SET);

	if ( isRunning ) {
		start();
	}
}

public void setZoom ( boolean state) {
	boolean isRunning = ((flags & PLAYING) != 0);

	if ( isRunning ) {
		stop();
	}

	flags &= ~( PICTURE_SET | ZOOM_PICTURE);
	if ( state ) {
		flags |= ZOOM_PICTURE;
	}
	
	if ( isRunning ) {
		start();
	}
	
}

public synchronized void show () {
	super.show();

	// if it was suspended, resume
	if ( (flags & SUSPENDED) != 0 ){
		resume();
	}	
}

public synchronized void start () {
	if ( (flags & SUSPENDED) != 0 ){
		resume();
		return;
	}
	
	//already running ?
	if ( ctrl.getPictureWidth() == 0 ) {
		flags &= ~(STARTED | PLAYING);
	}
	
	if ( loadSource() && isShowing() ) {
		setPicture();
		flags |= PLAYING;
		ctrl.resume();
	}

}

public synchronized void stop () {
	if ( (flags & PLAYING) != 0 ){
		ctrl.suspendSyncRender();
		ctrl.stop();
		flags &= ~(PLAYING | STARTED | PICTURE_SET);
		blank();
	}
}

public synchronized void suspend () {
	if ( (flags & PLAYING) != 0 ){
		ctrl.suspendSyncRender();
		flags |= SUSPENDED;
	}
}

public String toString () {
	StringBuffer sb = new StringBuffer(100);

	sb.append( super.toString());
	sb.append( " [");

	if ( (flags & PICTURE_SET) != 0 ) sb.append( "pict_set ");
	if ( (flags & SOURCE_SET) != 0 ) sb.append( "src_set ");
	if ( (flags & STARTED) != 0 ) sb.append( "started ");
	if ( (flags & SUSPENDED) != 0 ) sb.append( "susp ");

	sb.append("]");

	return sb.toString();
}
}
