package kaffe.net;

import java.net.UnknownHostException;

public class DNS {

private static final int TODO = 0;
private static final int OKAY = 1;
private static final int ERROR = 2;

String hostname;
int address;
int state;

static {
	System.loadLibrary("net");
	init0();
}

public DNS(String hostname) {
	this.hostname = hostname;
	state = TODO;
}

public DNS(int address) {
	this.address = address;
	state = TODO;
}

public DNS(String hostname, int address) {
	this.hostname = hostname;
	this.address = address;
	state = OKAY;
}

public void clear() {
	hostname = null;
	address = 0;
	state = TODO;
}

public boolean equals(Object obj) {
	if (obj instanceof DNS) {
		DNS that = (DNS)obj;
		if ((hostname == that.hostname || (hostname != null && hostname.equals(that.hostname))) && (address == that.address)) {
			return (true);
		}
	}
	return (false);
}

public synchronized String getHostname() throws UnknownHostException {
	if (state == TODO) {
		// No hostname - initiate lookup
		if (hostname == null) {
			getHostByAddr0(address);

			// Wait until we get an answer or an error
			while (hostname == null && state == TODO) {
				try {
					wait();
				}
				catch (InterruptedException _) {
					hostname = null;
					state = ERROR;
				}
			}
		}
	}
	if (state == ERROR) {
		throw new UnknownHostException("IP: " + addressToString(address));
	}
	return(hostname);
}

public synchronized int getAddress() throws UnknownHostException {
	if (state == TODO) {
		// No address - initiate lookup
		if (address == 0) {
			getAddrByHost0(hostname);

			// Wait until we get an answer or an error
			while (address == 0 && state == TODO) {
				try {
					wait();
				}
				catch (InterruptedException _) {
					address = 0;
					state = ERROR;
				}
			}
		}
	}
	if (state == ERROR) {
		throw new UnknownHostException("Host: " + hostname);
	}
	return (address);
}

public synchronized void callback(String hostname, int address) {
	if (hostname == null && address == 0) {
		if (this.address == 0) {
			state = ERROR;
		}
		else {
			this.hostname = addressToString(this.address);
			state = OKAY;
		}
	}
	else if (hostname != null) {
		this.hostname = hostname;
		state = OKAY;
	}
	else {
		this.address = address;
		state = OKAY;
	}
	notify();
}

public boolean isValid() {
	return (state == OKAY);
}

public static String addressToString(int addr) {
	StringBuffer buf = new StringBuffer();
	buf.append((addr >> 24) & 0xFF);
	buf.append(".");
	buf.append((addr >> 16) & 0xFF);
	buf.append(".");
	buf.append((addr >> 8) & 0xFF);
	buf.append(".");
	buf.append(addr & 0xFF);
	return (buf.toString());
}

private native static void init0();
private native void getHostByAddr0(int address);
private native void getAddrByHost0(String hostname);
public native static String getLocalHostname();

}
