package kaffe.net;

import java.lang.String;
import java.io.BufferedReader;
import java.io.FileDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * KaffeServer - server application for kaffecli client
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 */
public class KaffeServer
  extends Thread
{
	ServerSocket ListenSocket;
	boolean Log;
	int Port;
	boolean GoOn = true;
	int NConnect;
	static PrintStream SysOut = System.out;
	static InputStream SysIn = System.in;
	final static int DEF_PORT = 9455;

class Connection
  implements Runnable
{
	BufferedReader CliOut;
	PrintStream CliIn;
	Socket Client;
	String UserDir;

Connection ( Socket cli, SrvThreadGroup tg ) {
	
	Client = cli;
	
	try {
		InputStream  cis = Client.getInputStream();
		OutputStream cos = Client.getOutputStream();

		CliOut = new BufferedReader( new InputStreamReader( cis));
		CliIn = new PrintStream( cos);

		tg.setOut( CliIn);
		tg.setIn( cis);
	}
	catch ( IOException x ) {
		try {
			Client.close();
		}
		catch ( IOException xx ) {}
		System.err.println( "cannot open streams on socket");
	}
}

void callMain ( String[] args ) {
	if ( args.length < 2 )
		return;

	try {
		Class      cls = Class.forName( args[1]);
		Class[]    parTypes = { args.getClass() };
		Method     mm  = cls.getMethod( "main", parTypes);
		int        n = args.length - 2;
		String[]   sArr = new String[n];
		Object[]   mArgs = { sArr };

		if ( args.length > 1 )
			System.arraycopy( args, 2, sArr, 0, n);

		UserDir = args[0];
		if ( Log ) {
			SysOut.print( "invoke: " + cls.getName() + ".main( ");
			for ( int i=0; i<sArr.length; i++ ) {
				SysOut.print( sArr[i]);
				SysOut.print( ' ');
			}
			SysOut.println( ")");
		}
		mm.invoke( null, mArgs);
	}
	catch ( SecurityException sx ) {
		if ( "system_exit".equals( sx.getMessage()) ){ // System.exit was called
			if ( Log )
				SysOut.println( "caught System.exit()");
			return;
		}
		else {
			SysOut.println( "received Security exception in TG: "
			                     + Thread.currentThread().getThreadGroup());
			sx.printStackTrace();
		}
	}
	catch ( Exception x ) {
		x.printStackTrace();
	}
}

String[] getArgs ( String input ) {
	Vector v = new Vector( 5);
	StringTokenizer st = new StringTokenizer( input);
	String[] args;
	String s;
	int i, n;
	
	while ( st.hasMoreTokens() )
		v.addElement( st.nextToken());

	n = v.size();
	args = new String[n];
	
	for ( i=0; i<n; i++ )
		args[i] = (String)v.elementAt( i);

	return args;
}

public void run () {
	String    input;
	String[]  args;
	String[]  targets;
	
	try {
		input = CliOut.readLine();
		if ( input != null ) {
			if ( input.equals( "STOP") ){
				GoOn = false;
				return;
			}
		
			if ( Log ){
				SysOut.print( "KaffeServer request: ");
				SysOut.println( input);
			}

			args = getArgs( input);
			callMain( args);
		}
	}
	catch ( IOException x ) {
		x.printStackTrace();
	}
	
	finally {
		try {
			Client.close();
		}
		catch ( IOException x ) {}
	}
}
}
class SrvThreadGroup
  extends ThreadGroup
{
	InputStream In;
	PrintStream Out;

SrvThreadGroup ( String name ) {
	super( name);
}

void setIn ( InputStream in ) {
	In = in;
}

void setOut ( PrintStream out ) {
	Out = out;
}
}
class SrvSecMgr
  extends SecurityManager
{
SrvSecMgr () {
}

public void checkAccept ( String host, int port ) {
}

public void checkAccess ( Thread t ) {
}

public void checkAccess ( ThreadGroup tg ) {
}

public void checkAwtEventQueueAccess() {
}

public void checkConnect ( String host, int port ) {
}

public void checkConnect ( String host, int port, Object context ) {
}

public void checkCreateClassLoader() {
}

public void checkDelete ( String file ) {
}

public void checkExec ( String cmd ) {
}

public void checkExit ( int ret ) {
	throw new SecurityException( "system_exit");
}

public void checkLink ( String lib ) {
}

public void checkListen ( int port ) {
}

public void checkMemberAccess ( Class cls, int which ) {
}

public void checkMulticast ( InetAddress maddr ) {
}

public void checkMulticast ( InetAddress maddr, byte ttl ) {
}

public void checkPackageAccess ( String pkg ) {
}

public void checkPackageDefinition ( String pkg ) {
}

public void checkPrintJobAccess() {
}

public void checkPropertiesAccess() {
}

public void checkPropertyAccess ( String key ) {
}

public void checkRead ( FileDescriptor fd ) {
}

public void checkRead ( String file ) {
}

public void checkRead ( String file, Object context ) {
}

public void checkSecurityAccess ( String action ) {
}

public void checkSetFactory() {
}

public void checkSystemClipboardAccess() {
}

public boolean checkTopLevelWindow ( Object window ) {
	return true;
}

public void checkWrite ( FileDescriptor fd ) {
}

public void checkWrite ( String file ) {
}
}
class SrvPrintStream
  extends PrintStream
{
SrvPrintStream ( PrintStream defStream ) {
	super( defStream, true);
}

public void print ( Object o ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( o);
	else
		super.print( o);
}

public void print ( String s ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( s);
	else
		super.print( s);
}

public void print ( boolean b ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( b);
	else
		super.print( b);
}

public void print ( char[] ca ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( ca);
	else
		super.print( ca);
}

public void print ( double d ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( d);
	else
		super.print( d);
}

public void print ( float f ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( f);
	else
		super.print( f);
}

public void print ( int i ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( i);
	else
		super.print( i);
}

public void print ( long l ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.print( l);
	else
		super.print( l);
}

public void println () {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup ){
		((SrvThreadGroup)tg).Out.println();
	}
	else {
		super.println();
	}
}

public void println ( Object o ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( o);
	else
		super.println( o);
}

public void println ( String s ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( s);
	else
		super.println( s);
}

public void println ( boolean b ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( b);
	else
		super.println( b);
}

public void println ( char[] ca ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( ca);
	else
		super.println( ca);
}

public void println ( double d ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( d);
	else
		super.println( d);
}

public void println ( float f ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( f);
	else
		super.println( f);
}

public void println ( int i ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( i);
	else
		super.println( i);
}

public void println ( long l ) {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.println( l);
	else
		super.println( l);
}

public void write ( byte buf[], int off, int len ) {
	ThreadGroup tg;

	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.write( buf, off, len);
	else
		super.write( buf, off, len);
}

public void write ( int n ) {
	ThreadGroup tg;

	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		((SrvThreadGroup)tg).Out.write( n);
	else
		super.write( n);
}
}
class SrvInputStream
  extends InputStream
{
	InputStream StdIn;

SrvInputStream ( InputStream stdin ) {
	StdIn = stdin;
}

public int read () throws IOException {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		return ((SrvThreadGroup)tg).In.read();
	else
		return StdIn.read();
}

public int read ( byte b[], int off, int len ) throws IOException {
	ThreadGroup    tg;
	
	if ( (tg = Thread.currentThread().getThreadGroup()) instanceof SrvThreadGroup )
		return ((SrvThreadGroup)tg).In.read( b, off, len);
	else
		return StdIn.read( b, off, len);
}
}

public KaffeServer ( String args[] ) {
	setArgs( args);

	SysOut = System.out;
	System.setOut( new SrvPrintStream( SysOut));
	
	SysIn = System.in;
	System.setIn( new SrvInputStream( SysIn));

	System.setSecurityManager( new KaffeServer.SrvSecMgr());

	run();
}

protected void finalize () throws Throwable {
	if ( ListenSocket != null ) {
		ListenSocket.close();
	}
	super.finalize();
}

public static void main ( String[] args ) {
	KaffeServer server = new KaffeServer( args);
}

public void run () {
	try {
		ListenSocket = new ServerSocket( Port, 10);
	}
	catch ( IOException x ) {
		System.err.println( "could not open ServerSocket on port: " + Port);
		System.exit( 1);
	}

	try {
		if ( Log )	
			SysOut.println( "KaffeServer started on port: " + Port);

		while ( GoOn ) {
			if ( Log )
				SysOut.println( "KaffeServer ready..");
		
			Socket client = ListenSocket.accept();
			NConnect++;
			
			SrvThreadGroup tg = new SrvThreadGroup( Integer.toString( NConnect));
			Connection connection = new Connection( client, tg);			
			Thread t = new Thread( tg, connection);
			t.start();
		}
	}
	catch ( IOException x ) {
		System.err.println( "connection failed");
	}
}

void setArgs ( String[] args ) {
	int i;
	
	Port = 0;

	try {	
		for ( i=0; i<args.length; i++ ) {
			if ( args[i].equals( "-port") ) {
				Port = Integer.parseInt( args[++i]);
			}
			else if ( args[i].equals( "-log") ) {
				Log = true;
			}
		}
	}
	catch ( Exception x ) {}
	
	// we should not init this from biss.jde.Jde, since it would
	// load the whole Toolkit (and we might want to rebuild the awt itself)
	if ( Port == 0 )
		Port = DEF_PORT;
}
}
