package kaffe.net;

import java.io.InterruptedIOException;
import java.net.DatagramPacket;
import java.net.SocketException;
import kaffe.io.FileIO;

public class NetIO
	extends FileIO {

public static final int LOCALADDR = 0;
public static final int LOCALPORT = 1;
public static final int REMOTEADDR = 2;
public static final int REMOTEPORT = 3;

private int laddress;
private int lport;
private int raddress;
private int rport;

static {
	System.loadLibrary("net");
	init0();
}

public NetIO() {
}

public int getLocalPort() {
	return (lport);
}

public int getLocalAddress() {
	return (laddress);
}

public int getRemotePort() {
	return (rport);
}

public int getRemoteAddress() {
	return (raddress);
}

public void socket(boolean stream) throws SocketException {
	fd = socket0(stream);
	if (fd < 0) {
		throw new SocketException("failed to create socket");
	}
}

private native int socket0(boolean stream);

public void connect(int address, int port) throws SocketException {
	try {
		connect0(fd, address, port);
		rport = port;
		raddress = address;
		lport = getSocketInfo(LOCALPORT);
		laddress = getSocketInfo(LOCALADDR);
	}
	catch (InterruptedIOException _) {
		throw new SocketException("connect aborted");
	}
}

private native void connect0(int fd, int address, int port) throws SocketException, InterruptedIOException;

public void bind(int address, int port) throws SocketException {
	laddress = address;
	if (bind0(fd, address, port) < 0) {
		throw new SocketException("bind failed");
	}
	lport = getSocketInfo(LOCALPORT);
}

private native int bind0(int fd, int address, int port);

public void listen(int count) throws SocketException {
	if (listen0(fd, count) < 0) {
		throw new SocketException("listen failed");
	}
}

private native int listen0(int fd, int count);

public void accept(NetIO parent, int address, int port) throws SocketException {
	fd = accept0(parent.fd, address, port);
	if (fd < 0) {
		throw new SocketException("accept failed");
	}
	raddress = getSocketInfo(REMOTEADDR);
	rport = getSocketInfo(REMOTEPORT);
}

private native int accept0(int fd, int address, int port);

public int getSocketInfo(int item) {
	return (getSocketInfo0(fd, item));
}

private native int getSocketInfo0(int fd, int item);

public void setSocketOption(int option, int value) throws SocketException {
	if (setSocketOption0(fd, option, value) < 0) {
		throw new SocketException("setSocketOption failed");
	}
}

private native int setSocketOption0(int fd, int option, int value);

public int getSocketOption(int option) throws SocketException {
	int val = getSocketOption0(fd, option);
	if (val == -1) {
		throw new SocketException("getSocketOption failed");
	}
	return (val);
}

public int recvfrom(DatagramPacket pkt, DNS dns) throws SocketException {
	dns.clear();
	return (recvfrom0(fd, pkt.getData(), pkt.getLength(), dns, pkt));
}

private native int recvfrom0(int fd, byte[] buf, int len, DNS dns, DatagramPacket pkt) throws SocketException;

public void sendto(byte[] buf, int len, int addr, int port) throws SocketException {
	sendto0(fd, buf, len, addr, port);
}

private native void sendto0(int fd, byte[] buf, int len, int addr, int port) throws SocketException;

public int peekaddr(DNS dns) throws SocketException {
	dns.clear();
	return (peekaddr0(fd, dns));
}

private native int peekaddr0(int fd, DNS dns) throws SocketException;

private native int getSocketOption0(int fd, int option);

private native static void init0();

}
