/*
 * BasicURLConnection -
 *  Simple abstract URL handler.
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.net.www.protocol;

import java.awt.Toolkit;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLConnection;

abstract public class BasicURLConnection
  extends URLConnection
{
	final protected static String[] headers = {
	"content-encoding",
	"content-length",
	"content-type",
	"date",
	"expiration",
	"If-Modified-Since",
	"lastModified",
};
	protected String[] headersValue = new String[headers.length];
	final protected static int ContentEncoding = 0;
	final protected static int ContentLength = 1;
	final protected static int ContentType = 2;
	final protected static int Date = 3;
	final protected static int Expiration = 4;
	final protected static int IfModifiedSince = 5;
	final protected static int LastModified = 6;

public BasicURLConnection(URL url) {
	super(url);
}

abstract public void connect() throws IOException;

public String getHeaderField(String name) {
	try {
		connect();
		for (int i = 0; i < headers.length; i++) {
			if (headers[i].equals( name)) {
				return (getHeaderField(i));
			}
		}
	}
	catch (IOException _) {
	}
	return (null);
}

public String getHeaderField(int pos) {
	try {
		connect();
		if (pos >= 0 && pos < headersValue.length) {
			return (headersValue[pos]);
		}
	}
	catch (IOException _) {
	}
	return (null);
}

public String getHeaderFieldKey(int pos) {
	try {
		connect();
		if (pos >= 0 && pos < headers.length) {
			return (headers[pos]);
		}
	}
	catch (IOException _) {
	}
	return (null);
}

protected void setContentTypeFromName() {
	String ct = getFileNameMap().getContentTypeFor( url.getFile());
	headersValue[ContentType] = ct;
	headersValue[ContentEncoding] = ct;
}

protected void setHeaderField( String key, String value) {
	for (int i = 0; i < headers.length; i++ ) {
		if ( headers[i].equals( key) ) {
			headersValue[i] = value;
			break;
		}
	}
}
}
