/*
 * HttpURLConnection -
 *  Simple HTTP URL handler.
 *
 * Copyright (c) 1998, 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package kaffe.net.www.protocol.http;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownServiceException;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import kaffe.io.AccessibleBIStream;

public class HttpURLConnection
  extends java.net.HttpURLConnection
{
	private static String proxyHost;
	private static int proxyPort = -1;
	private Vector headerKeys = new Vector();
	private Vector headerVals = new Vector();
	private Hashtable reqprops;
	private boolean error;
	protected Socket sock;
	protected InputStream in;
	protected DataOutputStream out;
	protected ByteArrayOutputStream userout;
	protected boolean redir = getFollowRedirects();

  static {
	// How these properties are undocumented in the API doc.  We know
	// about them from www.icesoft.no's webpage
	//
	proxyHost = System.getProperty("http.proxyHost");
	if (proxyHost != null) {
		// Sun also supports a http.nonProxyHosts property to
		// avoid proxy use for local sites.  It's a regular expression
		// like so "*.pa.dec.com|*.compaq.com"
		// For now, we always use a proxy if these properties are set.
		// 
		// In JDK 1.2, the properties are read when an connection
		// is established, allowing the use of different proxies
		// during the run-time of a program.  FIXME
		String pp = System.getProperty("http.proxyPort");
		if (pp != null) {
			proxyPort = Integer.parseInt(pp);
		}
	}
	}

public HttpURLConnection(URL url) {
	super(url);
}


// Override these four functions to layer HTTP on top of another protocol
// such as SSL.
protected int defaultPort() {
  return 80;
}

protected Socket makeSocket(String host, int port) throws IOException {
  return new Socket(host, port);
}

protected String getProxyHost() {
  return proxyHost;
}

protected int getProxyPort() {
  return proxyPort;
}

public void connect() throws IOException {
	if (connected == true) {
		return;
	}
	for (;;) {
		int port;
		String host;
		if (getProxyHost() != null) {
			port = getProxyPort();
			host = getProxyHost();
		}
		else {
			port = url.getPort();
			host = url.getHost();
		}
		if (port == -1) {
			port = defaultPort();
		}

		// Make the connection.  If it fails, mark the error and
		// rethrow
		try {
		  sock = makeSocket(host, port);
		}
		catch (IOException e) {
			error = true;
			throw e;
		}

		in = new BufferedInputStream(sock.getInputStream());
		out = new DataOutputStream(sock.getOutputStream());
		//kaffe.io.SplitOutputStream split = new kaffe.io.SplitOutputStream(sock.getOutputStream(), System.out); out = new DataOutputStream(split);

		// Make the http request.
		String file;
		if (getProxyHost() != null) {
			file = url.toString();
		}
		else {
			file = url.getFile();
			if (file.equals("")) {
				file = "/";
			}
		}

		if (userout != null) {
			setRequestProperty("Content-Length",  "" + userout.size());
		}

		out.writeBytes(method + " " + file + " HTTP/1.0\r\n");

		setRequestProperties();

                for (Enumeration keys = reqprops.keys(); keys.hasMoreElements(); ) {
			Object key = keys.nextElement();
			out.writeBytes(key + ": " + reqprops.get(key) + "\r\n");
		}

		// Write any user data
		if (userout != null) {
			out.writeBytes("\r\n");
			userout.writeTo(out);
		}

		out.writeBytes("\r\n");
		out.flush();

		clearHeaders ();

		DataInputStream inp = new DataInputStream(in);
		for (;;) {
			String line = inp.readLine();
			if (line == null || line.equals("")) {
				break;
			}
			// This is the page response 
			else if (line.startsWith("HTTP")) {
				responseMessage = line;
				StringTokenizer tok = new StringTokenizer(line);
				if (tok.countTokens() >= 3) {
					tok.nextToken();
					try {
						responseCode = Integer.parseInt(tok.nextToken());
					}
					catch (NumberFormatException _) {
						responseCode = HTTP_SERVER_ERROR;
					}
				}
			}
			// Everything else is a colon seperated header and value.
			else {
				int pos = line.indexOf(':');
				if (pos > 0) {
					String key = line.substring(0, pos);
					for (pos++; Character.isWhitespace(line.charAt(pos)); pos++)
						;
					String val;
					if (!key.equalsIgnoreCase ("Set-Cookie"))
					  { // HACK to fix Tim's hack
					    // it's a mistake to trim the header in this
					    // way, but if we're going to do it
					    // it's necessary to explicitly _not_
					    // do it for the cookie field
					    int epos = line.indexOf(';', pos);;
					    if (epos == -1) {
					      epos = line.length();
					    }
					    val = line.substring(pos, epos);
					  }
					else
					  {
					    val = line.substring(pos);
					  }
					setHeaderField(key, val);
				}
			}
		}

		// Connection established
		connected = true;

		// Handle redirection
		String location = getHeaderField("location");
		if (redir == false || responseCode < HTTP_MULT_CHOICE || responseCode > HTTP_USE_PROXY || location == null) {
			break;
		}

		// Redirection - disconnect
		disconnect();
		url = new URL(location);
	}
}

public void disconnect() {
	if (connected == true) {
		try {
			sock.close();
		}
		catch (IOException _) {
		}
		sock = null;
		connected = false;
	}
}

public String getHeaderField(String key) {
	if (!error) {
		try {
			connect();
			key = key.toLowerCase();
			int pos = headerKeys.indexOf(key);
			if (pos != -1) {
				return ((String)headerVals.elementAt(pos));
			}
		}
		catch (IOException _) {
		}
	}
	return (null);
}

public String getHeaderField(int pos) {
	if (!error) {
		try {
			connect();
			if (pos >= 0 && pos < headerVals.size()) {
				return ((String)headerVals.elementAt(pos));
			}
		}
		catch (IOException _) {
		}
	}
	return (null);
}

public String getHeaderFieldKey(int pos) {
	try {
		connect();
		if (pos >= 0 && pos < headerKeys.size()) {
			return ((String)headerKeys.elementAt(pos));
		}
	}
	catch (IOException _) {
	}
	return (null);
}

  private void clearHeaders () 
  {
    headerKeys.removeAllElements ();
    headerVals.removeAllElements ();
  }


  protected void setRequestProperties()
  {
    if(url.getPort() == -1 || url.getPort() == defaultPort())
      {
	setRequestProperty("Host", url.getHost());
      }
    else
      {
	setRequestProperty("Host", url.getHost() + ":" + url.getPort());
      }
  }

public InputStream getInputStream() throws IOException {
	if (!error) {
		connect();
		return (in);
	}
	throw new IOException("connection error");
}

public OutputStream getOutputStream() throws IOException {
	if (method.equals("POST")) {
		synchronized (this) {
			if (userout == null) {
				userout = new ByteArrayOutputStream();
			}
		}
		return (userout);
	}
	else {
		throw new UnknownServiceException();
	}
}

/**
 * If we enable output, then we're doing a post.
 */
public void setDoOutput(boolean dooutput) {
	if (dooutput) {
		method = "POST";
	}
	super.setDoOutput(dooutput);
}

protected void setContentTypeFromName() {
	String ct = getFileNameMap().getContentTypeFor( url.getFile());
	setHeaderField("content-type", ct);
	setHeaderField("content-encoding", ct);
}

protected void setHeaderField( String key, String value) {
	key = key.toLowerCase();
	int pos = headerKeys.indexOf(key);
	if (pos == -1) {
		headerKeys.addElement(key);
		headerVals.addElement(value);
	}
	else {
		headerVals.setElementAt(value, pos);
	}
}

public void setRequestProperty(String key, String value) {
	if (reqprops == null) {
		reqprops = new Hashtable();
	}
	reqprops.put(key, value);
}

public void delelteRequestProperty(String key) 
  {
	if (reqprops == null) 
	  {
	    reqprops = new Hashtable();
	  }
	else 
	  {
	    reqprops.remove(key);
	  }
    
  }

/**
 * Used by the ICE browser.
 */
public void setInstanceFollowRedirects(boolean redir) {
	this.redir = redir;
}

public boolean usingProxy() {
	return (getProxyHost() != null);
}

public static void setProxy(String host, int port) {
	proxyHost = host;
	proxyPort = port;
}

}
