package kaffe.net.www.protocol.jar;

import java.awt.Toolkit;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/*
 * Java core library component.
 *
 * Copyright (c) 1999
 *      Archie L. Cobbs.  All rights reserved.
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Author: Archie L. Cobbs <archie@whistle.com>
 */
public class JarURLConnection
  extends java.net.JarURLConnection
{
	private URL jarFileURL;
	private File underlyingFile;
	private JarFile jarFile;
	private String jarEntryName;
	private ZipEntry jarEntry;

	static URL lastJarFileURL;
	static JarFile lastJarFile;
	static File lastUnderlyingFile;

public JarURLConnection(URL url) throws MalformedURLException {
	super(url);
	String file = url.getFile();
	int sep = file.lastIndexOf("!/");
	if (sep == -1) {
		throw new MalformedURLException("missing !/");
	}
	jarFileURL = new URL(file.substring(0, sep));
	jarEntryName = file.substring(sep + 2);
}

public void connect() throws IOException {

	// Already connected?
	if (connected) {
		return;
	}

	synchronized (JarURLConnection.class) {
		if (!jarFileURL.equals(lastJarFileURL)) {

			// Get the file: if inner URL is file: just use it directly
			if (jarFileURL.getProtocol().equals("file")) {
				underlyingFile = new File(jarFileURL.getFile().replace('/', File.separatorChar));
			}
			else {
				// Get the JAR file input stream
				jarFileURLConnection = jarFileURL.openConnection();
				InputStream in = jarFileURLConnection.getInputStream();

				// Save it to a temp file
				underlyingFile = File.createTempFile("jar", null);
				underlyingFile.deleteOnExit();
				OutputStream out = new FileOutputStream(underlyingFile);
				byte[] buf = new byte[1024];
				for (int r; (r = in.read(buf)) != -1; ) {
					out.write(buf, 0, r);
				}
				in.close();
				out.close();
			}
			if (lastUnderlyingFile != null && !lastJarFileURL.getProtocol().equals("file")) {
				lastUnderlyingFile.delete();
			}
			lastUnderlyingFile = underlyingFile;
			// Create JAR file out of underlying file
			lastJarFileURL = jarFileURL;
			lastJarFile = new JarFile(underlyingFile);
		}
		jarFile = lastJarFile;
	}

	// Get the entry in the file, or possibly the whole JAR file
	if (jarEntryName.equals("")) {
		jarEntry = null;
	}
	else {
		jarEntry = jarFile.getEntry(jarEntryName);
		if (jarEntry == null) {
			throw new IOException("JAR entry \""
			    + jarEntryName + "\" not found");
		}
	}
	connected = true;
}

public InputStream getInputStream() throws IOException {
	connect();
	if (jarEntry == null) {
		return (new FileInputStream(underlyingFile));
	}
	else {
		return (jarFile.getInputStream(jarEntry));
	}
}

public JarFile getJarFile() throws IOException {
	connect();
	return (jarFile);		// XXX should be read-only
}
}
