package kaffe.net.www.protocol.rtsp;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.util.Vector;

public class RtcpReceiver
  extends Thread
{
	final static int RTCP_PACKET_LENGTH = 1000;
	byte[] packetBytes;
	InetAddress rtcpAddr;
	int rtcpPort;
	DatagramSocket dgSocket;
	boolean listen = false;

public RtcpReceiver( DatagramSocket socket){
	this.dgSocket = socket;
	packetBytes = new byte[RTCP_PACKET_LENGTH];

	setPriority( MIN_PRIORITY);
}

void dumpRTCPData(byte[] bytes ) throws IOException{

	//*********Write RTCP data to files for diagnostics********************************
	byte[] rtcpBytes = bytes;
	RandomAccessFile rtcpTextFile = new RandomAccessFile( "rtcp.txt", "rw" );

	rtcpTextFile.seek( rtcpTextFile.length() );  //set pointer to end of file

	FileOutputStream fosRTCP = new FileOutputStream( rtcpTextFile.getFD() );

	//for( int x=0; x < 100; x++ ){

	//System.err.println( "RTCP PACKET " + x  );

	for( int i=0; i < rtcpBytes.length; i++ ){

		byte b = rtcpBytes[i];

		Byte byteObj = new Byte( b );
		Integer intObj = new Integer(0);

		if( i == 0)
			fosRTCP.write( new String( "##########################\n" ).getBytes() );

		if( i < 52 )
			fosRTCP.write( intObj.toHexString( byteObj.intValue() ).getBytes() );

		if( i == 0 ){
			fosRTCP.write( new String( " -->header\n" ).getBytes() );
		}else if( i == 1 )
			fosRTCP.write( new String( " -->packet type\n").getBytes());
		else if( i == 2 )
			fosRTCP.write( new String( "-").getBytes());
		else if( i == 3 )
			fosRTCP.write( new String(  " -->length\n").getBytes());
		else if( i == 4 || i == 5 || i == 6 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 7 )
			fosRTCP.write( new String(  " -->SSRC\n").getBytes());
		else if( i == 8 || i == 9 || i == 10 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 11 )
			fosRTCP.write( new String(  " -->NTP Timestamp, most sig word\n").getBytes());
		else if( i == 12 || i == 13 || i == 14 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 15 )
			fosRTCP.write( new String(  " -->NTP Timestamp, least sig word\n").getBytes());
		else if( i == 16 || i == 17 || i == 18 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 19 )
			fosRTCP.write( new String(  " -->RTP Timestamp\n").getBytes());
		else if( i == 20 || i == 21 || i == 22 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 23 )
			fosRTCP.write( new String(  " -->Sender's packet count\n").getBytes());
		else if( i == 24 || i == 25 || i == 26 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 27 )
			fosRTCP.write( new String(  " -->Sender's octet count\n").getBytes());
		else if( i == 28 || i == 29 || i == 30 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 31 )
			fosRTCP.write( new String(  " -->SSRC_1\n").getBytes());
		else if( i == 32 )
			fosRTCP.write( new String(  "-->Fraction Lost\n").getBytes());
		else if( i == 33 || i == 34 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 35 )
			fosRTCP.write( new String(  " -->Cumulative # packets lost\n").getBytes());
		else if( i == 36 || i == 37 || i == 38 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 39 )
			fosRTCP.write( new String(  " -->Extended highest sequence number received\n").getBytes());
		else if( i == 40 || i == 41 || i == 42 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 43 )
			fosRTCP.write( new String(  " -->Inter-Arrival Jitter\n").getBytes());
		else if( i == 44 || i == 45 || i == 46 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 47 )
			fosRTCP.write( new String(  " -->LSR\n").getBytes());
		else if( i == 48 || i == 49 || i == 50 )
			fosRTCP.write( new String(  "-").getBytes());
		else if( i == 51 )
			fosRTCP.write( new String(  " -->DLSR\n").getBytes());
		//else
		//fosRTCP.write( new String( "\n" ).getBytes() );
	} 
	//fosRTCP.write( rtcpBytes[x] );

	//}

	fosRTCP.close();
}

byte[] getUnfilteredBytes(){
	return this.packetBytes;
}

void listen( boolean bool ){
	listen = bool;
}

void processResponse( byte[] incoming, DatagramPacket incomingPacket ) throws IOException{

	byte[] response = new byte[ 32 ];

	response[0] = -127; //0x10000001;  //version 2; no padding; one reception report
	response[1] = -55; //0x11001001;  //packet type 201, meaning RR (Reception Report)
	response[2] = 0;   //0x00000000;
	response[3] = 7;   //0x00000001;  //length is 1

	response[4] = incoming[4];
	response[5] = incoming[5];
	response[6] = incoming[6];
	response[7] = incoming[7]; //bouncing SSRC of incoming

	response[8] = incoming[28];
	response[9] = incoming[29];
	response[10] = incoming[30];
	response[11] = incoming[31]; //bouncing SSRC_1 of incoming

	response[12] = incoming[32]; //bouncing fraction lost of incoming
	response[13] = incoming[33]; 
	response[14] = incoming[34]; 
	response[15] = incoming[35]; //bouncing cumulative #packets lost of incoming

	response[16] = incoming[36];
	response[17] = incoming[37]; 
	response[18] = incoming[38];
	response[19] = incoming[39]; //bouncing Extended hightest sequence number received

	response[20] = incoming[40];
	response[21] = incoming[41]; 
	response[22] = incoming[42];
	response[23] = incoming[43]; //bouncing inter-arrival jitter

	response[24] = incoming[44];
	response[25] = incoming[45]; 
	response[26] = incoming[46];
	response[27] = incoming[47]; //bouncing LSR

	response[28] = incoming[48];
	response[29] = incoming[49];
	response[30] = incoming[50];
	response[31] = incoming[51]; //bounce DLSR

	rtcpAddr = incomingPacket.getAddress();
	rtcpPort = incomingPacket.getPort();

	//System.err.println( "reply: " + rtcpAddr + " " + rtcpPort);
	DatagramPacket dgPacket = new DatagramPacket( response, response.length, rtcpAddr, rtcpPort );

	//send to Transmitter
	dgSocket.send( dgPacket );
}

void readStream(){

	DatagramPacket dgPacket = null;

	int counter = 0;

	while( listen ){

		try{
			//bytes = new byte[RTCP_PACKET_LENGTH];
			dgPacket = new DatagramPacket( packetBytes, RTCP_PACKET_LENGTH );

			dgSocket.receive( dgPacket );
			packetBytes = dgPacket.getData();

			//check for a zeroed out packet and discard it
			System.err.println( "RTCP INCOMING MSG TYPE: " + (256+packetBytes[1]) );
			if( packetBytes[0] == 0x00000000 ){
				System.err.println( "DISCARDING ZEROED RTCP PACKET!!!!" );
				return;
			}else{
				//dumpRTCPData( packetBytes );
				processResponse( packetBytes, dgPacket );
			}      

		}catch( IOException ioe ){
			ioe.printStackTrace();
			break;
		}

		counter++;

	}

	System.err.println( "done reading RTCP stream..." );

}

public void run(){
	readStream();
}
}
