package kaffe.net.www.protocol.rtsp;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.DatagramSocket;
import java.net.HttpURLConnection;
import java.net.Socket;
import java.net.SocketException;
import java.net.URL;
import java.net.URLConnection;
import kaffe.io.RtpMediaStream;

/*
 * RtspURLConnection -
 *  Simple RTSP URL handler.
 *
 * Copyright (c) 1998, 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class RtspURLConnection
  extends HttpURLConnection
{
	boolean connected;
	int cseq = 1;
	final static int RTP_NO_PAYLOAD_PACKET_LENGTH = 1000;
	int rtcpPort;
	int rtpPort;
	DatagramSocket rtpSocket;
	DatagramSocket rtcpSocket;
	String sessionid = "";
	String streamid = "";
	String client_port = "";
	StringBuffer sbuf = new StringBuffer();
	byte[] rbuf = new byte[RTP_NO_PAYLOAD_PACKET_LENGTH];
	RtcpReceiver rtcpReceiver;
	boolean useRtcp;
	final static boolean DEBUG = false;
	protected InputStream rtspIn;
	protected Socket rtspSocket;

public RtspURLConnection(URL url) {
	this( url, false);
}

public RtspURLConnection(URL url, boolean useRtcp) {
	super(url);

	this.useRtcp = useRtcp;
	
	int portAttempt = 9940;
	while( !establishRtpRtcpPorts( portAttempt  ) ){
		portAttempt = portAttempt + 2;
	}

}

private void closeConnection() throws IOException{
	String sr;
	int br;
	
	InputStream in = rtspSocket.getInputStream();
	OutputStream out = rtspSocket.getOutputStream();
	
	sbuf.setLength( 0);
	
	sbuf.append( "TEARDOWN " + url.toString() + " RTSP/1.0\n");
	sbuf.append( "Session: " + sessionid + "\n");
	sbuf.append( "CSeq: " + cseq + "\n\n");
	
	sr = sbuf.toString();
	dbgSent( sr);

	out.write( sbuf.toString().getBytes() );
	out.flush();
	
	cseq++;
	
	br = in.read( rbuf);
	sr = new String( rbuf, 0, br);

	dbgResponse( sr);
	
}

public void connect() throws IOException {
	if (connected == true) {
		return;
	}

	int port = url.getPort();
	String host = url.getHost();

	if ( port == -1 ) {
		port = defaultPort();
	}

	rtspSocket = new Socket(host, port);
	rtspIn = new RtpMediaStream( rtpSocket);
	initConnection();

	connected = true;
	
	if ( useRtcp ) {
		rtcpReceiver = new RtcpReceiver( rtcpSocket);
		rtcpReceiver.listen( true );
		rtcpReceiver.start();
	}
}

private void dbgResponse( String msg) {
	if ( DEBUG ) {
		System.err.println( "*** RTSP response ***");
		System.err.println( msg);
	}
}

private void dbgSent( String msg) {
	if ( DEBUG ) {
		System.err.println( "*** RTSP sent ***");
		System.err.println( msg);
	}
}

// Override these four functions to layer HTTP on top of another protocol
// such as SSL.
protected static int defaultPort() {
  return 80;
}

public void disconnect() {
	if (connected == true) {
		try {
			if ( rtcpReceiver != null ) {
				rtcpReceiver.listen( false);
			}
			closeConnection();
			rtspIn.close();
			rtspSocket.close();
		}
		catch (IOException _) {
		}
		rtspSocket = null;
		connected = false;
	}
}

private boolean establishRtpRtcpPorts( int port ){

	try{
		rtpSocket = new DatagramSocket( port );
		rtcpSocket = new DatagramSocket( port +1 );
	}catch( SocketException se ){
		return false;
	}

	rtpPort = port;
	rtcpPort = port + 1;
	
	client_port = Integer.toString(rtpPort) + "-" + Integer.toString(rtcpPort);
	
	return true;
}

public InputStream getInputStream() throws IOException {
	connect();
	return (rtspIn);
}

private void initConnection() throws IOException{
	int i1, i2, br;
	String sr;
	
	InputStream in = rtspSocket.getInputStream();
	OutputStream out = rtspSocket.getOutputStream();
	
	sbuf.setLength( 0);
	
	sbuf.append( "DESCRIBE " + url.toString() + " RTSP/1.0\n");
	sbuf.append( "Accept: application/sdp, application/mpeg, application/x-rtsp-mh\n");
	sbuf.append( "CSeq: " + cseq + "\n\n");

	sr = sbuf.toString();
	dbgSent( sr);
		
	out.write( sr.getBytes() );
	out.flush();
	
	cseq++;

	br = in.read( rbuf);
	sr = new String( rbuf, 0, br);

	dbgResponse( sr);

	i1 = sr.indexOf( "a=control:");
	if ( sr.charAt( i1+10) == '*') {
		i1 = sr.indexOf( "a=control:", i1+10);
	}
	i2 = sr.indexOf( '\n', i1);
	streamid = sr.substring( i1+10, i2-1);
	
	sbuf.setLength( 0);
	
	sbuf.append( "SETUP " + url.toString() + "/" + streamid + " RTSP/1.0\n");
	sbuf.append( "Transport: RTP/AVP/UDP;unicast;client_port=" + client_port + ";mode=play\n");
	sbuf.append( "CSeq: " + cseq + "\n\n");
	
	sr = sbuf.toString();
	dbgSent( sr);
		
	out.write( sr.getBytes() );
	out.flush();

	cseq++;

	br = in.read( rbuf);
	sr = new String( rbuf, 0, br);

	dbgResponse( sr);

	i1 = sr.indexOf( "Session: ");
	i2 = sr.indexOf( '\n', i1);
	sessionid = sr.substring( i1+9, i2-1);

	sbuf.setLength( 0);
	
	sbuf.append( "PLAY " + url.toString() + " RTSP/1.0\n");
 	sbuf.append( "CSeq: " + cseq + "\n");
 	sbuf.append( "Range: npt=0-\n");
	sbuf.append( "Session: " + sessionid + "\n\n");
	
	sr = sbuf.toString();
	dbgSent( sr);
		
	out.write( sr.getBytes() );
	out.flush();

	cseq++;

	br = in.read( rbuf);
	sr = new String( rbuf, 0, br);

	dbgResponse( sr);

}

public boolean usingProxy() {
	return false;
}
}
