package kaffe.net.www.protocol.zip;

import java.awt.Toolkit;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;

/*
 * Java core library component.
 *
 * Copyright (c) 1999
 *      Archie L. Cobbs.  All rights reserved.
 * Copyright (c) 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Author: Archie L. Cobbs <archie@whistle.com>
 */
public class ZipURLConnection extends URLConnection {

	static private ZipFile lastZipFile;
	static private URL lastZipFileURL;

	private URL zipFileURL;
	private ZipFile zipFile;
	protected String zipEntryName;
	protected ZipEntry zipEntry;

public ZipURLConnection(URL url) throws MalformedURLException {
	super(url);
	String file = url.getFile();
	int sep = file.lastIndexOf("!/");
	if (sep == -1) {
		throw new MalformedURLException("missing !/");
	}
	zipFileURL = new URL(file.substring(0, sep));
	zipEntryName = file.substring(sep + 2);
}

protected ZipFile getUnderlyingZipFile(URL url) throws IOException {
	// Get the file: if inner URL is file: just use it directly
	if (url.getProtocol().equals("file")) {
		return (new ZipFile(url.getFile().replace('/', File.separatorChar)));
	}
	else {
		synchronized (ZipURLConnection.class) {
			if (url.equals(lastZipFileURL)) {
				return (lastZipFile);
			}
		}

		InputStream in = url.openStream();
		File file = getUnderlyingFileFromStream(in);
		in.close();
		ZipFile zfile = new ZipFile(file);
		file.delete();

		synchronized (ZipURLConnection.class) {
			lastZipFile = zfile;
			lastZipFileURL = url;
		}

		return (zfile);
	}
}

protected File getUnderlyingFileFromStream(InputStream in) throws IOException {
	File file = File.createTempFile("zip", null);
	OutputStream out = new FileOutputStream(file);
	byte[] buf = new byte[1024];
	for (int r; (r = in.read(buf)) != -1; ) {
		out.write(buf, 0, r);
	}
	out.close();
	return (file);
}

public void connect() throws IOException {
	// Already connected?
	if (connected) {
		return;
	}

	// Create ZIP file out of underlying file
	zipFile = getUnderlyingZipFile(zipFileURL);

	if ("".equals(zipEntryName)) {
		zipEntry = null;
	}
	else {
		zipEntry = zipFile.getEntry(zipEntryName);
		if (zipEntry == null) {
			throw new IOException("ZIP entry '" + zipEntryName + "' not found");
		}
	}

	connected = true;
}

public InputStream getInputStream() throws IOException {
	connect();
	return (zipFile.getInputStream(zipEntry));
}

public ZipFile getZipFile() throws IOException {
	connect();
	return (zipFile);		// XXX should be read-only
}

public int getContentLength() {
	try {
		connect();
		if (zipEntry != null) {
			return ((int)zipEntry.getSize());
		}
	}
	catch (IOException _) {
	}
	return (-1);
}

}
