/*
 * EHttpURLConnection -
 *  Simple HTTP URL handler.
 *
 * Copyright (c) 1998, 1999
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package com.transvirtual.net.www.protocol.ehttp;

import com.transvirtual.net.www.protocol.ehttp.server.WebReply;
import com.transvirtual.net.www.protocol.ehttp.server.WebServer;
import java.awt.Toolkit;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLConnection;
import java.util.StringTokenizer;
import kaffe.awt.URLConnectionProducer;

public class EHttpURLConnection
  extends java.net.HttpURLConnection
{
	final private static String[] headers = {
	"content-encoding",
	"content-type",
	"content-length",
	"date",
	"expiration",
	"If-Modified-Since",
	"lastModified",
	"content-location",
	"location",
	"server"
};
	final private static int ContentEncoding = 0;
	final private static int ContentType = 1;
	private String[] headersValue = new String[headers.length];
	private WebReply rep;
	private InputStream in;
	private boolean redir = getFollowRedirects();
	private String referer;
	private static WebServer server;

static {
	server = new WebServer();
}

public EHttpURLConnection(URL url) {
	super(url);
}

public void connect() throws IOException {
	for (;;) {
		rep = server.process(this);
		in = new BufferedInputStream(rep.getInputStream());
		DataInputStream inp = new DataInputStream(in);
		for (;;) {
			String line = inp.readLine();
			if (line == null || line.equals("")) {
				break;
			}
			// This is the page response 
			else if (line.startsWith("HTTP")) {
				responseMessage = line;
				StringTokenizer tok = new StringTokenizer(line);
				if (tok.countTokens() >= 3) {
					tok.nextToken();
					try {
						responseCode = Integer.parseInt(tok.nextToken());
					}
					catch (NumberFormatException _) {
						responseCode = HTTP_SERVER_ERROR;
					}
				}
			}
			// Everything else is a colon seperated header and value.
			else {
				int pos = line.indexOf(':');
				if (pos > 0) {
					String key = line.substring(0, pos);
					for (pos++; Character.isWhitespace(line.charAt(pos)); pos++)
						;
					String val = line.substring(pos);
					setHeaderField(key, val);
				}
			}
		}

		// Handle redirection
		String location = getHeaderField("location");
		if (redir == false || responseCode < HTTP_MULT_CHOICE || responseCode > HTTP_USE_PROXY || location == null) {
			break;
		}
		url = new URL(location);
		rep.recycle();
		responseCode = HTTP_SERVER_ERROR;
	}
}

public void disconnect() {
	rep.recycle();
}

public Object getContent() throws IOException {
	if (headersValue[ContentEncoding] == null) {
		return (in);
	}
	
	// We only understand a limited number of things so far
	
	// if it's an image, return a ImageProducer
	if (headersValue[ContentEncoding].startsWith("image/")) {
		return new URLConnectionProducer( this);
	}

	// Return the input stream if we don't understand
	return (in);
}

public String getHeaderField(String name) {
	for (int i = 0; i < headers.length; i++) {
		if (headers[i].equalsIgnoreCase( name)) {
			return (getHeaderField(i));
		}
	}
	return (null);
}

public String getHeaderField(int pos) {
	if (pos < 0 || pos >= headersValue.length) {
		return (null);
	}
	return (headersValue[pos]);
}

public String getHeaderFieldKey(int pos) {
	if (pos < 0 || pos >= headers.length) {
		return (null);
	}
	return (headers[pos]);
}

public InputStream getInputStream() throws IOException {
	return (in);
}

public String getRequestProperty(String key) {
	if (key.equalsIgnoreCase("referer")) {
		return (referer);
	}
	else {
		return (null);
	}
}

protected void setContentTypeFromName() {
	String ct = getFileNameMap().getContentTypeFor( url.getFile());
	headersValue[ContentType] = ct;
	headersValue[ContentEncoding] = ct;
}

protected void setHeaderField( String key, String value) {
	for ( int i=0; i<headers.length; i++ ) {
		if ( headers[i].equalsIgnoreCase( key) ) {
			headersValue[i] = value;
			break;
		}
	}
}

/**
 * Used by the ICE browser.
 */
public void setInstanceFollowRedirects(boolean redir) {
	this.redir = redir;
}

public void setRequestProperty(String key, String value) {
	if (key.equalsIgnoreCase("referer")) {
		referer = value;
	}
}

public boolean usingProxy() {
	return (false);
}
}
