/**
 * BindingWebHandler.java
 *
 *  This provides a binding between one piece of server space and another.
 *  If a filename falls in the binding space, it is renamed to the filename
 *  with the binding name replaces by the mapping name.
 */

package com.transvirtual.net.www.protocol.ehttp.server;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

public class BindingWebHandler extends WebApplication {

private String binding;
private Hashtable map = new Hashtable();

public BindingWebHandler() {
	binding = getConfig("Root", "/binding");
	String str = getConfig("Bindings", "");
	StringTokenizer tok = new StringTokenizer(str, " \t:");
	while (tok.hasMoreTokens()) {
		String from = tok.nextToken();
		String to = tok.nextToken();
		map.put(from, to);
		// System.out.println("  Adding binding: " + from + " -> " + to);
	}
}

/**
 * Map the given request which is in the binding space to the associated
 * binding (usually a server side application).  We generate a redirect
 * for the webbrowser to follow.
 */
public void doGet(HttpServletRequest req, HttpServletResponse rep) throws ServletException {
	String name = ((WebRequest)req).getFileNextElement();
	String rest = req.getPathInfo();
	if (rest == null) {
		rest = "";
	}
	String app = null;
	if (name != null) {
		app = (String)map.get(name);
	}
	if (app == null) {
		app = (String)map.get("default");
	}
	if (app == null) {
		throw new ServletException("no app found: " + app);
	}
//System.out.println("redirect to " + app + '/' + rest);
	try {
		rep.sendRedirect(app + '/' + rest);
	}
	catch (IOException _) {
		throw new ServletException();
	}
}

/**
 * Check to see if this request is in the binding space.  If it
 * is indicate we can deal with it.
 */
public boolean accepts(WebRequest req) {
	String filename = req.getFileFirstElement();
	if (filename.equals(binding)) {
		return (true);
	}
	else {
		return (false);
	}
}

}
