package com.transvirtual.net.www.protocol.ehttp.server;

import java.net.*;
import java.io.*;
import java.text.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import kaffe.io.AccessibleBAOStream;
import kaffe.io.CharToByteConverter;

public class WebReply implements HttpServletResponse {

private static WebReply free;

private WebReply next;
private OutputStream out;
private int status;
private WebOutputStream buffer = new WebOutputStream();
private String type;
private Hashtable headers = new Hashtable();
private InputStream in;

private WebReply() {
	status = HttpURLConnection.HTTP_OK;
}

void send() {
	try {
		DataOutputStream dout = new DataOutputStream(out);

		// Write status
		dout.writeBytes("HTTP/1.1 " + Integer.toString(status) + " ");
		switch (status) {
		case HttpURLConnection.HTTP_OK:
			dout.writeBytes("OK");
			break;
		case HttpURLConnection.HTTP_MOVED_TEMP:
			dout.writeBytes("Found");
			break;
		case HttpURLConnection.HTTP_NOT_FOUND:
			dout.writeBytes("File Not Found");
			break;
		default:
			dout.writeBytes("Server Error");
			break;
		}
		dout.writeBytes("\n");
		dout.writeBytes("Date: " + DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG).format(new Date()) + "\n");
		dout.writeBytes("Server: " + WebServer.name + "\n");
		int size = buffer.size();
		if (size > 0) {
			dout.writeBytes("Content-Length: " + size + "\n");
		}
		dout.writeBytes("Connection: close\n");
		dout.writeBytes("Content-Type: " + (type == null ? "text/plain" : type) + "\n");
		writeFieldHeaders(dout);

		dout.writeBytes("\n");

		// Write data
		if (size > 0) {
			buffer.writeTo(out);
		}

		dout.flush();
	}
	catch (IOException _) {
	}
}

/**
 * This is public so the protocol class can get hold of it.
 */
public InputStream getInputStream() {
	if (in == null) {
		send();
		AccessibleBAOStream aout = (AccessibleBAOStream)out;
		in = new ByteArrayInputStream(aout.getBuffer(), 0, aout.size());
	}
	return (in);
}

static WebReply getInstance() {
	WebReply rep;
	synchronized (WebReply.class) {
		if (free == null) {
			rep = new WebReply();
		}
		else {
			rep = free;
			free = rep.next;
			rep.next = null;
		}
		rep.out = new AccessibleBAOStream(128);
	}
	return (rep);
}

private void writeFieldHeaders(DataOutputStream dout) throws IOException {
	Enumeration keys = headers.keys();
	while (keys.hasMoreElements()) {
		String key = (String)keys.nextElement();;
		dout.writeBytes(key + ": " + (String)headers.get(key));
	}
}

public void recycle() {
	out = null;
	status = 0;
	buffer.reset();
	type = null;
	headers.clear();
	in = null;

	this.next = free;
	free = this;
}

// ----------------------------------------------------------------------

public void setContentType(String type) {
	this.type = type;
}

public void setContentLength(int len) {
}

public ServletOutputStream getOutputStream() throws IOException {
	return (buffer);
}

public PrintWriter getWriter() throws IOException {
	return (new PrintWriter(buffer, true));
}

public String getCharacterEncoding() {
	return (null);
}

public void addCookie(Cookie cookie) {
	throw new kaffe.util.NotImplemented();
}

public boolean containsHeader(String name) {
	return (headers.contains(name));
}

public void setStatus(int sc, String sm) {
	setStatus(sc);
	// Discard message 'sm' ... 
}

public void setStatus(int sc) {
	status = sc;
}

public void setHeader(String name, String value) {
	headers.put(name, value);
}

public void setIntHeader(String name, int value) {
	setHeader(name, Integer.toString(value));
}

public void setDateHeader(String name, long date) {
	setHeader("Date", new Date(date).toString());
}

public void sendError(int sc, String msg) throws IOException {
	throw new kaffe.util.NotImplemented();
}

public void sendError(int sc) throws IOException {
	throw new kaffe.util.NotImplemented();
}

public void sendRedirect(String location) throws IOException {
	setStatus(SC_MOVED_TEMPORARILY);
        setHeader("location", WebServer.protocol + location);
}

public String encodeUrl(String url) {
	throw new kaffe.util.NotImplemented();
}

public String encodeRedirectUrl(String url) {
	throw new kaffe.util.NotImplemented();
}

}
