package com.transvirtual.net.www.protocol.ehttp.server;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import com.transvirtual.net.www.protocol.ehttp.EHttpURLConnection;

public class WebServer {

final public static String name = "Kaffe WebServer 0.3";
final public static String protocol = "ehttp:";

static private ResourceBundle config;

private Vector apps = new Vector();

static {
	try {
		config = ResourceBundle.getBundle("webserver");
	}
	catch (MissingResourceException _) {
	}
}

public WebServer() {
	// Get the handler list
	String str = getConfig(getClass().getName() + ".Handlers", "");
	StringTokenizer tok = new StringTokenizer(str, " \t,");
	while (tok.hasMoreTokens()) {
		String hand = tok.nextToken();
		try {
			WebApplication app = WebApplication.getInstance(hand);
			addHandler(app);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	/* If all else fails, we cann the not found handler */
	addHandler(new NotFoundWebHandler());
}

/**
 * This is the entry point from the protocol handler.
 */
public WebReply process(EHttpURLConnection conn) {
	WebRequest req = WebRequest.getInstance(conn);
	WebReply rep = WebReply.getInstance();
	try {
		WebApplication app = getInstance(req);
		String cmd = req.getMethod();
		if (cmd.equals("GET")) {
			app.doGet(req, rep);
		}
		else if (cmd.equals("POST")) {
			app.doPost(req, rep);
		}
		else {
			throw new Exception("bad method: " + cmd);
		}
	}
	catch (Exception e) {
		rep.setStatus(HttpURLConnection.HTTP_NOT_FOUND);
	}
	req.recycle();
	return (rep);
}

public static String getConfig(String key, String def) {
	if (config != null) {
		try {
			return (config.getString(key));
		}
		catch (MissingResourceException _) {
		}
	}
	return (def);
}

public WebApplication getInstance(WebRequest req) {
	// Search handler for someone who can handle this request
	int len = apps.size();
	for (int i = 0; i < len; i++) {
		WebApplication app = (WebApplication)apps.elementAt(i);
		if (app.accepts(req)) {
			return (app);
		}
	}
	// If no one can, return the default app
	return (null);
}

public void addHandler(WebApplication app) {
	apps.addElement(app);
}

public void removeHandler(WebApplication app) {
	apps.removeElement(app);
}

}
