/*
 * WebSlinger.java
 *  WebSlinger internal webserver connection.
 *
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

package com.transvirtual.net.www.protocol.webslinger;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.ByteArrayInputStream;
import java.io.InputStreamReader;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.net.URL;
import java.net.URLConnection;
import java.net.Socket;
import java.io.FileInputStream;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpSession;
import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletContext;
import javax.servlet.ServletConfig;
import kaffe.io.AccessibleBAOStream;
import kaffe.io.SplitOutputStream;

public class WebSlingerURLConnection
  extends kaffe.net.www.protocol.http.HttpURLConnection {

/*
 * Name of the WebSlinger servlet class.
 * We do this using forName to avoid needing to have WebSlinger around
 * when compiling this.
 */
final static private String WEBSLINGERSERVLET = "com.brainfood.webslinger.servlet.Servlet";

static private final boolean debug = true;
static private HttpServlet servlet;

static {
	try {
		servlet = (HttpServlet)Class.forName(WEBSLINGERSERVLET).newInstance();
		servlet.init(new Config());
	}
	catch (ClassNotFoundException e) {
		e.printStackTrace(System.out);
	}
	catch (InstantiationException e) {
		e.printStackTrace(System.out);
	}
	catch (IllegalAccessException e) {
		e.printStackTrace(System.out);
	}
	catch (ServletException e) {
		e.printStackTrace(System.out);
	}
}

public WebSlingerURLConnection(URL url) {
	super(url);
}

public void connect() throws IOException {
	if (connected == true) {
		return;
	}
	try {
		Request q = new Request();
		Response p = new Response();
		servlet.service(q, p);
		p.finish();
	}
	catch (ServletException e) {
		throw new IOException(e.toString());
	}
	connected = true;
}

public void disconnect() {
	connected = false;
}

/*
 * A WebSlinger request.
 */
class Request implements HttpServletRequest {

private BufferedReader reader;

/*
 * No POST handling - XXX
 */
public int getContentLength() {
	return (0);
}

public String getContentType() {
	return ("text/plain");
}

public String getProtocol() {
	return ("HTTP/1.1");
}

public String getScheme() {
	return ("webslinger");
}

public String getServerName() {
	return (null);
}

public int getServerPort() {
	return (0);
}

public String getRemoteAddr() {
	return (null);
}

public String getRemoteHost() {
	return (null);
}

public String getRealPath(String path) {
	throw new kaffe.util.NotImplemented();
}

public ServletInputStream getInputStream() {
	throw new kaffe.util.NotImplemented();
}

public String getParameter(String name) {
	throw new kaffe.util.NotImplemented();
}

public String[] getParameterValues(String name) {
	throw new kaffe.util.NotImplemented();
}

public Enumeration getParameterNames() {
	throw new kaffe.util.NotImplemented();
}

public Object getAttribute(String name) {
	throw new kaffe.util.NotImplemented();
}

public BufferedReader getReader() throws IOException {
	if (reader == null) {
		reader = new BufferedReader(new InputStreamReader(getInputStream()));
	}
	return (reader);
}

public String getCharacterEncoding() {
	return (null);
}

public Cookie[] getCookies() {
	throw new kaffe.util.NotImplemented();
}

public String getMethod() {
	return (WebSlingerURLConnection.this.getRequestMethod());
}

public String getRequestURI() {
	throw new kaffe.util.NotImplemented();
}

public String getServletPath() {
	throw new kaffe.util.NotImplemented();
}

public String getPathInfo() {
	throw new kaffe.util.NotImplemented();
}

public String getPathTranslated() {
	return (WebSlingerURLConnection.this.getURL().getFile());
}

public String getQueryString() {
	throw new kaffe.util.NotImplemented();
}

public String getRemoteUser() {
	return (null);
}

public String getAuthType() {
	return (null);
}

public String getHeader(String name) {
	throw new kaffe.util.NotImplemented();
}

public int getIntHeader(String name) {
	try {
		return (Integer.parseInt(getHeader(name)));
	}
	catch (NumberFormatException _) {
		return (-1);
	}
}

public long getDateHeader(String name) {
	try {
		return ((new SimpleDateFormat("EEE MMM dd HH:mm:ss zzz yyyy")).parse(getHeader(name)).getTime());
	}
	catch (ParseException _) {
		return (-1);
	}
}

public Enumeration getHeaderNames() {
	throw new kaffe.util.NotImplemented();
}

public HttpSession getSession(boolean create) {
	throw new kaffe.util.NotImplemented();
}

public String getRequestedSessionId() {
	throw new kaffe.util.NotImplemented();
}

public boolean isRequestedSessionIdValid() {
	throw new kaffe.util.NotImplemented();
}

public boolean isRequestedSessionIdFromCookie() {
	throw new kaffe.util.NotImplemented();
}

public boolean isRequestedSessionIdFromUrl() {
	throw new kaffe.util.NotImplemented();
}

};

/*
 * A WebSlinger response.
 */
class Response implements HttpServletResponse {

private AccessibleBAOStream output;
private ServletOutputStream sout;
private PrintWriter writer;

public Response() throws IOException {
	output = new AccessibleBAOStream(new byte[1024]);
}

public void setContentType(String type) {
	setHeader("content-encoding", type);
}

public void setContentLength(int len) {
        setIntHeader("content-length", len);
}

public ServletOutputStream getOutputStream() throws IOException {
	if (sout == null) {
		sout = new Output(output);
	}
	return (sout);
}

public PrintWriter getWriter() throws IOException {
	if (writer == null) {
		writer = new PrintWriter(getOutputStream());
	}
	return (writer);
}

void finish() throws IOException {
	WebSlingerURLConnection.this.in = new ByteArrayInputStream(output.getBuffer());
}

public String getCharacterEncoding() {
	return (null);
}

public void addCookie(Cookie cookie) {
	// Cookies not supported
}

public boolean containsHeader(String name) {
	if (WebSlingerURLConnection.this.getHeaderField(name) == null) {
		return (false);
	}
	else {
		return (true);
	}
}

public void setStatus(int sc, String sm) {
	setStatus(sc);
	setContentType("text/plain");
	try {
		getWriter().println(sm);
	}
	catch (IOException _) {
	}
}

public void setStatus(int sc) {
	WebSlingerURLConnection.this.responseCode = sc;
}

public void setHeader(String name, String value) {
	WebSlingerURLConnection.this.setHeaderField(name, value);
}

public void setIntHeader(String name, int value) {
	setHeader(name, Integer.toString(value));
}

public void setDateHeader(String name, long date) {
	setHeader(name, new Date(date).toString());
}

public void sendError(int sc, String msg) throws IOException {
	setStatus(sc, msg);
}

public void sendError(int sc) throws IOException {
	setStatus(sc);
}

public void sendRedirect(String location) throws IOException {
	setHeader("location", location);
	setStatus(SC_MOVED_TEMPORARILY);
}

public String encodeUrl(String url) {
	throw new kaffe.util.NotImplemented();
}

public String encodeRedirectUrl(String url) {
	throw new kaffe.util.NotImplemented();
}

};

static class Context implements ServletContext { 

public Servlet getServlet(String name) throws ServletException {
	throw new kaffe.util.NotImplemented();
}

public Enumeration getServlets() {
	throw new kaffe.util.NotImplemented();
}

public Enumeration getServletNames() {
	throw new kaffe.util.NotImplemented();
}

public void log(String msg) {
	if (debug) {
		System.err.println(msg);
	}
}

public void log(Exception exception, String msg) {
	log(exception.toString() + ": " + msg);
}

public String getRealPath(String path) {
	throw new kaffe.util.NotImplemented();
}

public String getMimeType(String file) {
	throw new kaffe.util.NotImplemented();
}

public String getServerInfo() {
	throw new kaffe.util.NotImplemented();
}

public Object getAttribute(String name) {
	throw new kaffe.util.NotImplemented();
}

};

static class Config implements ServletConfig {

public ServletContext getServletContext() {
	return (new Context());
}
public String getInitParameter(String name) {
	throw new kaffe.util.NotImplemented();
}

public Enumeration getInitParameterNames() {
	throw new kaffe.util.NotImplemented();
}

};

static class Output extends ServletOutputStream {

private OutputStream out;

public Output(OutputStream strm) {
	if (debug) {
		out = new SplitOutputStream(strm, System.err);
	}
	else {
		out = strm;
	}
}

public void write(int b) throws IOException {
	out.write(b);
}

public void write(byte b[], int off, int len) throws IOException {
	out.write(b, off, len);
}

};

}
