package com.transvirtual.webserver;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;
import kaffe.io.StreamUtils;

public class WebServer 
	implements Runnable {

final public static String name = "Kaffe Embedded WebServer 0.1";

private ResourceBundle config;
private ServerSocket ssock;
private String protocol;
private String root;

public static void main(String[] args) {
	InputStream props = null;
	if (args.length == 1) {
		try {
			props = new FileInputStream(args[0]);
		}
		catch (IOException _) {
		}
	}
	// If no properties, use default.
	if (props == null) {
		props = WebServer.class.getResourceAsStream("webserver.properties");
	}

	ResourceBundle res;
	try {
		res = new PropertyResourceBundle(props);
	}
	catch (IOException e) {
		e.printStackTrace();
		return;
	}
	WebServer server = new WebServer(res);
	server.start();
}

private WebServer(ResourceBundle props) {
	config = props;
}

private void start() {
	int port = getConfigInt("port", 8000);
	protocol = getConfigString("protocol", "file");
	int threads = getConfigInt("threads", 1);
	root = getConfigString("root", "/");

	try {
		ssock = new ServerSocket(port);
	}
	catch (IOException e) {
		e.printStackTrace();
		return;
	}
	System.err.println(name + " (" + protocol + ":" + root + ")");

	for (int i = 1; i < threads; i++) {
		(new Thread(this)).start();
	}
	run();
}

public void run() {
	for (;;) {
		Socket sock = null;
		InputStream in = null;
		OutputStream out = null;
		HttpURLConnection conn = null;
		try {
			sock = ssock.accept();
			in = sock.getInputStream();
			out = sock.getOutputStream();

			BufferedReader rin = new BufferedReader(new InputStreamReader(in));
			StringTokenizer line = new StringTokenizer(rin.readLine());
			rin.readLine();	// Read the blank line
			String method = line.nextToken();
			String file = line.nextToken();

			URL u = new URL(protocol, null, root + file);
			conn = (HttpURLConnection)u.openConnection();
			conn.setRequestMethod(method);
			InputStream uin = conn.getInputStream();
			StreamUtils.writeTo(out, uin);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		if (conn != null) {
			conn.disconnect();
		}
		if (in != null) {
			try {
				in.close();
			}
			catch (IOException _) {
			}
		}
		if (out != null) {
			try {
				out.close();
			}
			catch (IOException _) {
			}
		}
		if (sock != null) {
			try {
				sock.close();
			}
			catch (IOException _) {
			}
		}
	}
}

private int getConfigInt(String name, int def) {
	try {
		return (Integer.parseInt(config.getString(name)));
	}
	catch (NumberFormatException _) {
		_.printStackTrace();
	}
	catch (MissingResourceException _) {
		_.printStackTrace();
	}
	return (def);
}

private String getConfigString(String name, String def) {
	try {
		return (config.getString(name));
	}
	catch (MissingResourceException _) {
	}
	return (def);
}

}
