/*
 * fgl.h - fgl master header (only included part in clients)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: fgl.h,v 1.19 2001/11/28 00:08:25 damian Exp $
 */

#ifndef FGL_H_INCLUDED
#define FGL_H_INCLUDED

#define LIB_FGL_0_0

#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdio.h>

/*****************************************************************************
 */
typedef unsigned char  u8;
typedef unsigned short u16;
typedef unsigned int   u32;
typedef char  i8;
typedef short i16;
typedef int   i32;

#include <fgl/config.h>
/* strutcture definition */

typedef struct {
    i32 x;
    i32 y;
} fgl_point_t;

typedef struct _fgl_bitmap_t {  /*{{{*/
    fgl_slc_t    *surface;
    u32          width;
    u32          height;
    u32          sLength;  /* length of scanline in scanline cell units */
    u32          state;
    fgl_color_t  transp;
    u32          lastOperation;
    u32          nextOperation;
} fgl_bitmap_t;
/*}}}*/

typedef struct _fgl_a8Bitmap_t { /*{{{*/
    u8           *surface;
    u32          width;
    u32          height;
    u32          sLength; /* same as width, but we keep it for conformity */
    u32          state;
} fgl_a8Bitmap_t;
/*}}}*/

typedef struct _fgl_a4Bitmap_t { /*{{{*/
    u8           *surface;
    u32          width;
    u32          height;
    u32          sLength;  /* length of scanline in bytes */
    u32          state;
} fgl_a4Bitmap_t;

typedef struct _fgl_a2Bitmap_t { /*{{{*/
    u8           *surface;
    u32          width;
    u32          height;
    u32          sLength;  /* length of scanline in bytes */
    u32          state;
} fgl_a2Bitmap_t;

typedef struct _fgl_monoBitmap_t { /*{{{*/
    u8           *surface;
    u32          width;
    u32          height;
    u32          sLength;  /* length of scanline in bytes */
    u32          state;
} fgl_monoBitmap_t;
/*}}}*/

#define MONO_SLENGTH(_width) (((_width)+7) >> 3)
#define A2_SLENGTH(_width) (((_width)+3) >> 2)

typedef struct _fgl_spriteData_t { /*{{{*/
    u32          width, height;
    u16          xhot, yhot;
    u8           nColors;
    u8           transp;
    u32          *cmap;
    u8           *data;
    u32          state;
} fgl_spriteData_t;
/*}}}*/

typedef struct _fgl_sprite_t { /*{{{*/
    fgl_bitmap_t      *bmp;
    fgl_monoBitmap_t  *mask;
    fgl_bitmap_t      *sav;
    i16               xhot, yhot;

    i32               x0, y0, x1, y1;
    u8                isVisible;
    u32               state;
    void              *ext; /* extension to hook-up native data */
} fgl_sprite_t;
/*}}}*/

typedef struct _fgl_cmap_t { /*{{{*/
    int len;
    u8 *r;
    u8 *g;
    u8 *b;
} fgl_cmap_t;
/*}}}*/

typedef struct _fgl_font_t { /*{{{*/
    u16         nGlyphs;
    u16         uniOffset;

    /* glyph data itself */
    u8          *surface;

    /* per glyph data */
    u32         *surfOff;
    u8          *width;
    u8          *height;
    i8          *xBearing;
    i8          *yBearing;
    i8          *advance;

    /* per font info */
    u8          fontHeight;
    u8          ascend;
    u8          descend;
    int         maxAdvance;

    char        *fname;
    u16         ptSize;
    u32         nRefs;

    u32         fSize;
    void        *fAdr;
    int         fh;

    /*
     * state variable replaces the old swappedBO, aliased and rotated variables.
     *
     * bits ( 0 - 1 ) current rotation;
     * bits ( 2 - 3 ) file rotation (might be different from 0 );
     * bits ( 4 - 7 ) antialiasing depth 
     * bits ( 8     ) swappedBO
     * bits ( 9     ) rotatedSURF;
     */
    
    u32         state;
    /* cache rotated sLength */
    u8          *sLength;
    
    /* introduce extended sets */
    u8          set_number;
    struct _fgl_font_t **set;
    
} fgl_font_t;

typedef struct _fgl_fontEntry_t { /*{{{*/
    fgl_font_t *fnt;
    char        *fname;
    u16         ptSize;
    struct _fgl_fontEntry_t *next;
} fgl_fontEntry_t;
/*}}}*/

typedef enum _fgl_mode_t { /*{{{*/
    PAINT_MODE = 0,
    XOR_MODE   = 1
} fgl_mode_t;
/*}}}*/

#define MODE_BRANCH(_ctx,_func,_arg) \
  if ( (_ctx)->mode == PAINT_MODE ) {_func _arg;} else {_func##Xor _arg;}

typedef struct _fgl_context_t { /*{{{*/
    fgl_bitmap_t *bmp;
    i32          x0Clip, y0Clip, x1Clip, y1Clip;
    u32          xoff, yoff;
    fgl_color_t  clr;
    fgl_color_t  xclr;
    u8           alphar;
    u8           alphag;
    u8           alphab;
    fgl_font_t   *fnt;
    fgl_mode_t   mode;

} fgl_context_t;
/*}}}*/

typedef struct _fgl_config_t {
    u32 xoff;
    u32 yoff;
    u32 width;
    u32 height;
} fgl_config_t;

/* fgl player server structures */

#define MAX_PSRV_CLIENTS 3
#define MAX_PSRV_STATES 16

#define PCS_BOUNDS     0
#define PCS_EVENT      1
 
#define PCS_SUSPEND    5
#define PCS_CONNECTED  6
#define PCS_FLUSHING   7
 
#define PCS_VOLUME     10
#define PCS_BRIGHTNESS 11
#define PCS_CONTRAST   12
#define PCS_HUE        13
#define PCS_SATURATION 14
 
#define PCE_KEY_TYPED    400
#define PCE_KEY_PRESSED  401
#define PCE_KEY_RELEASED 402
 
#define PCE_MOUSE_PRESSED  501
#define PCE_MOUSE_RELEASED 502
#define PCE_MOUSE_MOVED    503
#define PCE_MOUSE_DRAGGED  506

typedef struct _PClientRect{
    u16  x;
    u16  y;
    u16  width;
    u16  height;
}PClientRect_t;

typedef struct _PClientEvent{
    u16  event;
    u16  p1;
    u16  p2;
}PClientEvent_t;

typedef struct _PClientRgn {
    PClientRect_t   clip[5];
    u8              nClips;
    fgl_bitmap_t    bmp;
}PClientRgn_t;

typedef struct _PClient {
    void           *srv_addr;       /* attached server memory */
    u16            src_width;
    u16            src_height;
    PClientRect_t  srv_bounds;
    PClientEvent_t event;
    u8             modifiers[MAX_PSRV_STATES];
    u16            states[MAX_PSRV_STATES];
    u16            pages;
    PClientRgn_t   *crgn;
}fgl_PClient_t;

typedef struct _PServer {
    u8            status[4];
    fgl_context_t *ctx;
    u32           reserved[10];
    fgl_PClient_t clients[MAX_PSRV_CLIENTS];
    int  cursor;
}fgl_PServer_t;




/*
 * Evans&Sutherland clip codes
 */
#define CC_LEFT    1
#define CC_RIGHT   2
#define CC_BOTTOM  4
#define CC_TOP     8

/*
 * clip rect relation
 */
#define CR_OUT     0
#define CR_IN      1
#define CR_PART    2

/* 
 * rotation codes
 */
#define ROT_0   0
#define ROT_90  1
#define ROT_180 2
#define ROT_270 3

/*
 * swapped byte order code 
 */
#define SBO 0x100
#define RSF 0x200

#define ROTATE_BITS 3
/*
 * Operation type for frame buffer direct operation
 */
#define FGL_FRAMEBUFFER_OP 0x0000
/* macros */
#define RGB_888(_r,_g,_b) (u32) ((_r << 16) | (_g << 8) | (_b))

#define A_8(_rgb)    (u8) (((_rgb) & 0xff000000) >> 24)
#define R_8(_rgb)    (u8) (((_rgb) & 0x00ff0000) >> 16)
#define G_8(_rgb)    (u8) (((_rgb) & 0x0000ff00) >> 8)
#define B_8(_rgb)    (u8) (((_rgb) & 0x000000ff))


/* global variables */
extern fgl_bitmap_t               *fgl_screenBmp;
extern fgl_context_t              *fgl_screenCtx;
extern fgl_sprite_t               *fgl_cursor;
extern fgl_sprite_t               *fgl_client_cursor;
extern fgl_PServer_t              *PServer;

extern fgl_color_t fgl_clrBlack;
extern fgl_color_t fgl_clrWhite;
extern fgl_color_t fgl_clrRed;
extern fgl_color_t fgl_clrGreen;
extern fgl_color_t fgl_clrBlue;
extern fgl_color_t fgl_clrYellow;
extern fgl_color_t fgl_clrCyan;
extern fgl_color_t fgl_clrGray;
extern fgl_color_t fgl_clrLightGray;
extern fgl_color_t fgl_clrOrange;

/* functions from module: cntx */
u8 fgl_getClipCode ( fgl_context_t *ctx, i32 x, i32 y );
void fgl_setClip ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 );
void fgl_setBmp ( fgl_context_t *ctx, fgl_bitmap_t *bmp );
fgl_context_t* fgl_getContext( fgl_bitmap_t *bmp );
void fgl_setFont ( fgl_context_t* ctx, fgl_font_t* fnt );
void fgl_setColor ( fgl_context_t* ctx, fgl_color_t clr );
void fgl_setAlpha ( fgl_context_t* ctx, u8 alpha );
void fgl_setAlphas ( fgl_context_t* ctx, u8 alphar, u8 alphag, u8 alphab );
void fgl_setXorMode ( fgl_context_t* ctx, fgl_color_t xclr );
void fgl_setPaintMode ( fgl_context_t* ctx );

/* functions from module: sprite */
void fgl_hideSprite ( fgl_context_t* ctx, fgl_sprite_t* sp );
void fgl_showSprite ( fgl_context_t* ctx, fgl_sprite_t* sp );
fgl_sprite_t* fgl_createSprite ( fgl_spriteData_t *spd );
void fgl_moveSprite ( fgl_context_t* ctx, fgl_sprite_t* sp, i32 x, i32 y );
void fgl_moveCursor ( i32 x, i32 y );
void fgl_showCursor ();
void fgl_hideCursor ();
void fgl_setCursor ( fgl_sprite_t* sp );

/* functions from module: init_fb */
int fgl_getScreenWidth ();
int fgl_getScreenHeight ();
int fgl_getScreenXoff ();
int fgl_getScreenYoff ();
void fgl_terminate ();
int fgl_init ( int nReq, u32* rgbReqs, int restore );
int fgl_initialize ();
int fgl_open (u32 x, u32 y, u32 w, u32 h);
fgl_cmap_t* fgl_allocCmap ();

/* functions from module: rot */
void fgl_setScreenRotation( u8 rot );
int fgl_getScreenRotation();

/* functions from module: cmap */
fgl_color_t fgl_getPixelValue ( u32 argb );
void fgl_getRgbValues ( fgl_color_t pix, u32* r, u32* g, u32* b );
fgl_color_t fgl_modPixelValue( fgl_color_t pix, int mod );
fgl_color_t fgl_blendPixelValues ( u32 alphar, u32 alphag, u32 alphab, fgl_color_t d, fgl_color_t s );
int fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld );
fgl_cmap_t* fgl_getCmap ();

/* functions from module: line */
void fgl_drawVLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 y1 );
void fgl_drawHLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1 );
void fgl_drawLine ( fgl_context_t *ctx,  i32 x0, i32 y0, i32 x1, i32 y1 );

/* functions from module: rect */
void fgl_drawRect ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 );
void fgl_fillRectClr ( fgl_context_t *ctx, i32 x0, i32 y0,  i32 x1, i32 y1, fgl_color_t clr );
void fgl_fillRect ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 );

/* functions from module: oval */
void fgl_drawCircle ( fgl_context_t* ctx, i32 cx, i32 cy, u32 r );
void fgl_fillCircle ( fgl_context_t* ctx, i32 cx, i32 cy, u32 r );
void fgl_drawEllipse ( fgl_context_t* ctx, i32 cx, i32 cy, u32 rx, u32 ry );
void fgl_fillEllipse ( fgl_context_t* ctx, i32 cx, i32 cy, u32 rx, u32 ry );

/* functions from module cbmp */
fgl_monoBitmap_t* fgl_createBitmapMask ( fgl_bitmap_t* bmp );
void fgl_setPixel ( fgl_context_t* ctx, u32 x, u32 y, fgl_color_t clr );
void fgl_setBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y, fgl_color_t clr );
void fgl_setBmpTransparency ( fgl_bitmap_t* bmp, fgl_color_t clr );
fgl_color_t fgl_getBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y );
void fgl_clear ( fgl_context_t* ctx );
fgl_bitmap_t* fgl_createBitmap ( u32 width, u32 height, i32 clr );
void fgl_disposeBitmap ( fgl_bitmap_t* bmp );
void fgl_moveArea ( fgl_context_t* ctx, i32 x0, i32 y0, i32 x1, i32 y1, i32 dx, i32 dy );
void fgl_bitBlt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 );
void fgl_maskBlt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, fgl_monoBitmap_t* mask );
void fgl_alphaMaskBlt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, fgl_a8Bitmap_t* mask);

/* functions from module mbmp */
void fgl_setMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row );
void fgl_unsetMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row );
int fgl_isMonoBitSet ( fgl_monoBitmap_t* bmp, u32 col, u32 row );
fgl_monoBitmap_t* fgl_createMonoBitmap ( u32 width, u32 height, int init );
void fgl_disposeMonoBitmap ( fgl_monoBitmap_t* bmp );
void fgl_monoBlt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_monoBitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, fgl_color_t clr );

/* functions from module a2bmp */
void fgl_setPixelA2 ( fgl_a2Bitmap_t* bmp, u32 col, u32 row, u8 pixel );
u8 fgl_getPixelA2 ( fgl_a2Bitmap_t* bmp, u32 col, u32 row );
fgl_a2Bitmap_t* fgl_createA2Bitmap ( u32 width, u32 height, int init );
void fgl_disposeA2Bitmap ( fgl_a2Bitmap_t* bmp );
void fgl_a2Blt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_a2Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, fgl_color_t clr );

/* functions from module a4bmp */
void fgl_setPixelA4 ( fgl_a4Bitmap_t* bmp, u32 col, u32 row, u8 pixel );
u8 fgl_getPixelA4 ( fgl_a4Bitmap_t* bmp, u32 col, u32 row );
fgl_a4Bitmap_t* fgl_createA4Bitmap ( u32 width, u32 height, int init );
void fgl_disposeA4Bitmap ( fgl_a4Bitmap_t* bmp );
void fgl_a4Blt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_a4Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, fgl_color_t clr );

/* functions from module a8bmp */
void fgl_setPixelA8 ( fgl_a8Bitmap_t* bmp, u32 col, u32 row, u8 pixel );
u8 fgl_getPixelA8 ( fgl_a8Bitmap_t* bmp, u32 col, u32 row );
fgl_a8Bitmap_t* fgl_createA8Bitmap ( u32 width, u32 height, int init );
void fgl_disposeA8Bitmap ( fgl_a8Bitmap_t* bmp );
void fgl_a8Blt ( fgl_context_t* dst, i32 u0, i32 v0, fgl_a8Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, fgl_color_t clr );

/* functions from module font */
void fgl_addFontReference ( fgl_font_t* fnt );
fgl_font_t* fgl_getFont ( char* fname, u8 ptSize );
int fgl_getMaxAdvance ( fgl_font_t* fnt );
void fgl_disposeFont ( fgl_font_t* fnt );
void fgl_drawText ( fgl_context_t* ctx, i32 x, i32 y, char* text, i32 len );
void fgl_drawWText ( fgl_context_t* ctx, i32 x, i32 y, u16* text, u32 len );
void fgl_drawWVertText ( fgl_context_t* ctx, i32 x, i32 y, u16* text, u32 len );
int fgl_getTextLength ( fgl_font_t* fnt, char* text, u32 len );
int fgl_getWTextLength ( fgl_font_t* fnt, u16* text, u32 len );
int fgl_getFixedWidth ( fgl_font_t* fnt );

/* functions from module poly */
void fgl_drawPolygon ( fgl_context_t* ctx, fgl_point_t* points, u32 nPoints );
void fgl_fillPolygonClr ( fgl_context_t* ctx, fgl_point_t* points, u32 nPoints, fgl_color_t clr );

/* functions from module locale */
int fgl_setLocale (char *locale);

/* functions from module fgld */
u8 fgl_getSlcSize ();
int fgl_getScanlineLength ( int length );
fgl_slc_t* fgl_getSlc ( fgl_bitmap_t* bmp, u32 x, u32 y );

/* functions from module psrv */
fgl_PServer_t* fgl_PServerInit();
void fgl_PServerUpdateCursor( fgl_sprite_t *sp );
fgl_PClient_t* fgl_PServerOpenConnection();
fgl_PClient_t* fgl_PClientConnect( u16 *pWidth, u16 *pHeight, int *pDepth);
void fgl_PClientUpdateSize( fgl_PClient_t *pCli, u16 width, u16 height);
void fgl_PClientDisconnect( fgl_PClient_t* pCli);
void fgl_PClientAddClipRegion( fgl_PClient_t* pCli, u16 x, u16 y, u16 width, u16 height);
int  fgl_PClientFlushFrame( fgl_PClient_t* pCli);
void fgl_PClientSetBounds( fgl_PClient_t* pCli, u16 x, u16 y, u16 w, u16 h);
void fgl_PClientHandleEvent( fgl_PClient_t* pCli, u16 event, u16 p1, u16 p2);
void fgl_PClientSetValue( fgl_PClient_t* pCli, u8 key, u16 val);
u8   fgl_PClientCheckValue( fgl_PClient_t* pCli, u8 key, u16 *pval);


/* usefull macros for functions */
#define fgl_bitBltBmp( _dst, _u0, _v0, _src ) \
        fgl_bitBlt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1)

#define fgl_maskBltBmp( _dst, _u0, _v0, _src, _mask ) \
        fgl_maskBlt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1, _mask)

#define fgl_alphaMaskBltBmp( _dst, _u0, _v0, _src, _mask ) \
        fgl_alphaMaskBlt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1, _mask)

#define fgl_monoBltBmp( _dst, _u0, _v0, _src, _clr) \
        fgl_monoBlt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1, _clr)

#define fgl_a8BltBmp( _dst, _u0, _v0, _src, _clr) \
        fgl_a8Blt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1, _clr)

#define fgl_a4BltBmp( _dst, _u0, _v0, _src, _clr) \
        fgl_a4Blt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1, _clr)

#define fgl_a2BltBmp( _dst, _u0, _v0, _src, _clr) \
        fgl_a2Blt( _dst, _u0, _v0, _src, 0, 0, (_src)->width-1, (_src)->height-1, _clr)

#define fgl_fillPolygon( _dst, _pts, _npts ) \
        fgl_fillPolygonClr( _dst, _pts, _npts, (_dst)->clr )
  
/*
 * binary font file state flags ( included for bacward compability )
 */
#define IS_FILEMSB   0x001
#define IS_ANTIALIAS 0x002
#define IS_ROTATED   0x004
#define IS_SHAREDMEM 0x008

static inline
int fgl_getBitsPerPixel()
{
  return (FGL_COLOR_DEPTH);
}
  
#endif /* FGL_H_INCLUDED */
