/*
 * a4bmp.c - fgl module to handle monochrome bitmaps (depth 1)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "a4bmp.h"

u8 _a4Bits[2] = { 0xf0, 0x0f };

void fgl_rotateA4Bitmap( fgl_a4Bitmap_t* bmp )
{
#if defined(DYNAMIC)

  u8 brot,srot;
  u8 rotation;
  u8 *surface;
  u8 pix;
  u32 slength1, slength2;
  u32 w,h,u,v,x,y;

  u32 size = ( bmp->width + 1 ) * ( bmp->height + 1 ) / 2;

  /* we can always assume that the image is rotated from its current 
  *    * to the screen orientation
  *       */
  srot = fgl_screenBmp->state & 0x3;
  brot = bmp->state & 0x3;

  if ( srot == brot ) return; /* if nothing ton do do not waste time */

  rotation = ( 4 + srot - brot ) % 4; /* calculate the counter-clockwise */

  w = bmp->width; h = bmp->height;

  surface = (u8 *) FGL_MALLOC ( size );

  /* setup the slengths for the rotation 
  *    * slength1 is the slength of the original image
  *       * and slength2 is for the rotated one.
  *          */

  slength1 = bmp->sLength;

  switch( rotation ) {
    case 1:
      slength2 = ( bmp->height + 1) / 2;
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              int off = (1 - (x % 2)) << 2 ;
              pix = ( bmp->surface[ y * slength1 + x / 2 ] & _a4Bits[ x % 2 ] ) >> off;
              u = h - y - 1;
              v = x;
              off = (1 - (u % 2)) << 2;
              surface[ v * slength2 + u / 2 ] &= ~(0x0f << off );
              surface[ v * slength2 + u / 2 ] |= (pix << off );
          }
      }
      bmp->sLength = slength2;
      break;
    case 2:
      slength2 = ( bmp->width + 1) / 2;
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              int off = (1 - (x % 2)) << 2 ;
              pix = ( bmp->surface[ y * slength1 + x / 2 ] & _a4Bits[ x % 2 ] ) >> off;
              u = w - x - 1;
              v = h - y - 1;
              off = (1 - (u % 2)) << 2;
              surface[ v * slength2 + u / 2 ] &= ~(0x0f << off );
              surface[ v * slength2 + u / 2 ] |= (pix << off );
          }
      }
      break;
    case 3:
      slength2 = ( bmp->height + 1 ) / 2;
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              int off = (1 - (x % 2)) << 2 ;
              pix = ( bmp->surface[ y * slength1 + x / 2 ] & _a4Bits[ x % 2 ] ) >> off;
              u = y;
              v = w - x - 1;
              off = (1 - (u % 2)) << 2;
              surface[ v * slength2 + u / 2 ] &= ~(0x0f << off );
              surface[ v * slength2 + u / 2 ] |= (pix << off );
          }
      }
      bmp->sLength = slength2;
      break;
    default:
      return;
  }

  FGL_FREE( bmp->surface );
  bmp->surface = surface;

  BMP_SET_ROTATION( bmp, srot );

#endif /* DYNAMIC */
}

void fgl_setPixelA4 ( fgl_a4Bitmap_t* bmp, u32 col, u32 row, u8 pixel ) 
{
  _fgl_setPixelA4 ( bmp, col, row, pixel );
}

u8 fgl_getPixelA4 ( fgl_a4Bitmap_t* bmp, u32 col, u32 row ) 
{
  return (_fgl_getPixelA4 ( bmp, col, row ));
}

fgl_a4Bitmap_t* fgl_createA4Bitmap ( u32 width, u32 height, int init ) 
{
  u32 size =  (width + 1) * (height + 1) / 2;
  fgl_a4Bitmap_t *bmp = (fgl_a4Bitmap_t*) FGL_MALLOC( sizeof(fgl_a4Bitmap_t));

  if ( !bmp )
    {
      return 0;
    }

  bmp->surface = (u8*) FGL_MALLOC(  sizeof(u8) * size);
  bmp->width = width;
  bmp->height = height;


  if ( init == -1 )
    {
      SET_ROTATION( fgl_screenBmp );
      BMP_SET_ROTATION( ((fgl_bitmap_t *)bmp), rot );

      if ( rot & 1 )
        {
          bmp->sLength = (height + 1) / 2;
        }
      else
        {
          bmp->sLength = (width + 1) / 2;
        }
    }
  else
    { 
      BMP_SET_ROTATION( ((fgl_bitmap_t *)bmp), 0 );
      bmp->sLength = (width + 1) / 2;
      memset( bmp->surface, init, size);
      fgl_rotateA4Bitmap(bmp);
    }
  return bmp;
}

void fgl_disposeA4Bitmap ( fgl_a4Bitmap_t* bmp ) 
{
  _fgl_disposeA4Bitmap ( bmp);
}

void fgl_a4Blt ( fgl_context_t* dst, i32 u0, i32 v0,
				   fgl_a4Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
				   fgl_color_t clr ) 
{
	u8 *ps;
	fgl_slc_t *pd;
	fgl_color_t pix;
	i32 u, u1, v1, sinc, dinc;
	i32 w, h;
	register u8 a, b, i;
	u8 i0, off;
	u8 blend[16] = {0,17,34,51,68,85,102,119,136,153,170,187,204,221,238,255};

	if ( x0 > x1 )  swap32( &x0, &x1);
	if ( y0 > y1 )  swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}
        
        fgl_rotateA4Bitmap( src );
        fgl_rotateCBitmap( dst->bmp );
          {
            SET_ROTATION( ((fgl_bitmap_t *)src) );
            RECT_ROTATE( rot, ((fgl_bitmap_t *)src), x0, y0, x1, y1 );
          }

          {
            SET_ROTATION( dst->bmp);
            RECT_ROTATE( rot, dst->bmp, u0, v0, u1, v1 );
          }

        w = x1 - x0;
        h = y1 - y0;
        off = x0 % 2;

	ps = src->surface + (y0 * src->sLength) + (x0 / 2);
	pd = _fgl_getSlc( dst->bmp, u0, v0);

	a = *ps;
	u = u0;
	u1 = u0 + w;
	i = i0 = 2-off;

	dinc = (dst->bmp->sLength - u1) + u0;                 /* in pixels */
	sinc = src->sLength + (x0>>1) - (x1>>1);            /* in bytes */

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( h-- >= 0 ){
	  a <<= off;

	  while ( 1 ){
	    
	    b = a >> 4;
	    if (b){
	      pix = fgl_blendPixelValues(blend[b] , blend[b], blend[b], _fgl_getSlcPixel( pd, u) , clr);
	      _fgl_setSlcPixel( pd, u, pix);
	    }

	    if ( u == u1 ){
	      break;
	    }
	    
	    if ( --i == 0 ){
	      /* next source byte */
	      i = 2;
	      a = *(++ps);
	    }
	    else {
	      a <<= 4;
	    }
	    
	    _fgl_incSlc( &pd, u);
	    u++;
	  }

	  /* next source and destination scanlines */
	  _fgl_addSlc( &pd, u, dinc);
	  ps += sinc;
	  a = *ps;
	  u = u0;
	  i = i0;
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}



