/*
 * a8bmp.c - fgl module to handle alpha bitmaps (alpha value [0..255]) 
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "a8bmp.h"

void fgl_rotateA8Bitmap( fgl_a8Bitmap_t* bmp )
{
#if defined(DYNAMIC)

  u8 brot,srot;
  u8 rotation;
  u8 *surface;

  int x, y;

  int h;
  int w = bmp->sLength;
  /* we can always assume that the image is rotated from its current 
  *    * to the screen orientation
  *       */

  srot = fgl_screenBmp->state & 0x3;
  brot = bmp->state & 0x3;

  if ( srot == brot ) return;

  rotation = ( 4 + srot - brot ) % 4; /* calculate the counter-clockwise */

  if ( brot & 1)
    {
      h = bmp->width;
      w = bmp->height;
    }
  else
    {
      h = bmp->height;
      w = bmp->width;
    }

  surface = (u8 *) FGL_MALLOC ( w * h );

  switch( rotation ) {
    case 1:
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              surface[ x * h + h - y - 1 ] = bmp->surface[ x + y * w ];
          }
      }
      bmp->sLength = h;
      break;
    case 2:
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              surface[y*w+x] = bmp->surface[(h-y-1)*w+(w-x-1)];
          }
      }
      break;
    case 3:
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              surface[ h * ( w - x - 1 ) + y] = bmp->surface[ x + y * w ];
          }
      }
      bmp->sLength = h;
      break;
    default:
      return;
  }

  FGL_FREE( bmp->surface );
  bmp->surface = surface;

  BMP_SET_ROTATION( bmp, srot );

#endif /* DYNAMIC */
}

fgl_a8Bitmap_t* fgl_createA8Bitmap ( u32 width, u32 height, int init ) 
{
  fgl_a8Bitmap_t *bmp = (fgl_a8Bitmap_t*) FGL_MALLOC( sizeof(fgl_a8Bitmap_t));

  bmp->width   = width;
  bmp->height  = height;

  bmp->surface = (u8 *) FGL_MALLOC ( width * height );

    {
      SET_ROTATION( fgl_screenBmp );
      BMP_SET_ROTATION( (fgl_bitmap_t *)bmp, rot );

      if ( rot & 1) {
          bmp->sLength = height;
      }
      else {
          bmp->sLength = width;
      }
    }

  if ( init != -1 )
    {
      memset( bmp->surface, init, (width * height));
    }

  return bmp;
}

void fgl_disposeA8Bitmap ( fgl_a8Bitmap_t* bmp ) 
{
  _fgl_disposeA8Bitmap ( bmp ); 
}

void fgl_setPixelA8 ( fgl_a8Bitmap_t* bmp,  u32 x, u32 y, u8 alpha ) 
{
  _fgl_setPixelA8 ( bmp, x, y, alpha );
}

u8 fgl_getPixelA8 ( fgl_a8Bitmap_t* bmp,  u32 x, u32 y ) 
{
  return _fgl_getPixelA8 ( bmp, x, y ); 
}

void fgl_a8Blt ( fgl_context_t* dst, i32 u0, i32 v0,
					fgl_a8Bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1, 
					fgl_color_t clr ) 
{
	i32 u1, v1, u, da, dd;
	fgl_slc_t *pde, *pd;
	u8 *pa;
	fgl_color_t pix;

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

        fgl_rotateA8Bitmap ( src );
        {
          SET_ROTATION( src );
          RECT_ROTATE( rot, (fgl_bitmap_t *)src, x0, y0, x1, y1 );
        }
        {
          SET_ROTATION( dst->bmp);
          RECT_ROTATE( rot, dst->bmp, u0, v0, u1, v1 );
        }


        pd  = _fgl_getSlc( dst->bmp, u0, v0);
        pde = _fgl_getSlc( dst->bmp, u0, v1);
        dd  = dst->bmp->sLength - u1 + u0 - 1;

	pa  = src->surface + x0 + src->sLength * y0;
	da  = src->sLength - x1 - 1 + x0;

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( pd <= pde ){
		for ( u=u0; u<=u1; u++ ){
		        pix = fgl_blendPixelValues( *pa, *pa, *pa, _fgl_getSlcPixel( pd, u), clr);
			_fgl_setSlcPixel( pd, u, pix);

			_fgl_incSlc( &pd, u);
			pa++;
		}

		pa += da;
		_fgl_addSlc( &pd, u1, dd);
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}


