/*
 * a8bmp.h - fgl module to handle alpha bitmaps (alpha value [0..255]) 
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_A8BMP_H_INCLUDED
#define FGL_A8BMP_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "rot.h"
#include "draw.h"

static inline
void _fgl_disposeA8Bitmap ( fgl_a8Bitmap_t* bmp ) 
{
  FGL_FREE( bmp->surface );
  FGL_FREE( bmp );
}

static inline
void _fgl_setPixelA8 ( fgl_a8Bitmap_t* bmp,  u32 x, u32 y, u8 alpha ) 
{
  if ( (x < bmp->width) && (y < bmp->height) )
    {
        {
          SET_ROTATION( bmp );
          POINT_ROTATE( rot, (fgl_bitmap_t *)bmp, x, y );
        }
      bmp->surface[ (y * bmp->sLength) + x ] = alpha;
  }
}

static inline
u8 _fgl_getPixelA8 ( fgl_a8Bitmap_t* bmp,  u32 x, u32 y ) 
{
  if ( (x < bmp->width) && (y < bmp->height) )
    {
        {
          SET_ROTATION( bmp );
          POINT_ROTATE( rot, (fgl_bitmap_t *)bmp, x, y );
        }
      return bmp->surface[ (y * bmp->sLength) + x ];
    }
  else 
    {
      return 0;
    }
}

#endif /* FGL_A8BMP_H_INCLUDED */
