/*
 * cbmp.c - fgl module to handle framebuffer-depth bitmaps
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "cbmp.h"

void fgl_rotateCBitmap( fgl_bitmap_t* bmp )
{
#if defined(DYNAMIC)

  u8 brot,srot;
  u8 rotation;
  fgl_slc_t *old_surface, *new_surface;

  int x, y;

  int h;
  int w;
  /* we can always assume that the image is rotated from its current 
   * to the screen orientation
   */

  if ( bmp == fgl_screenBmp )
    {
      return;
    }

  srot = fgl_screenBmp->state & 0x3;
  brot = bmp->state & 0x3;


  if ( srot == brot ) return; /* if nothing ton do do not waste time */

  rotation = ( 4 + srot - brot ) % 4; /* calculate the counter-clockwise */

  if ( brot & 1)
    {
      h = bmp->width;
      w = bmp->height;

    }
  else
    {
      h = bmp->height;
      w = bmp->width;
    }

  new_surface = (fgl_slc_t *) FGL_MALLOC ( w * h * sizeof ( fgl_slc_t ));
  old_surface = _fgl_getColorSurface( bmp );

  switch( rotation ) {
    case 1:
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              new_surface[ x * h + h - y - 1 ] = old_surface[ x + y * w ];
          }
      }
      bmp->sLength = h;
      break;
    case 2:
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              new_surface[y*w+x] = old_surface[(h-y-1)*w+(w-x-1)];
          }
      }
      bmp->sLength = w;
      break;
    case 3:
      for( x = 0 ; x < w ; x++ ) {
          for ( y = 0 ; y < h ; y++ ) {
              new_surface[ h * ( w - x - 1 ) + y] = old_surface[ x + y * w ];
          }
      }
      bmp->sLength = h;
      break;
    default:
      return;
  }
  if ( bmp->state & IS_SHAREDMEM )
    {
      _fgl_cpySlc( new_surface, 0, old_surface, 0, w * h );
      FGL_FREE( new_surface );
    }
  else
    {
      FGL_FREE( bmp->surface );
      bmp->surface = new_surface;
    }
  BMP_SET_ROTATION( bmp, srot );

#endif /* DYNAMIC */
}

fgl_monoBitmap_t* fgl_createBitmapMask ( fgl_bitmap_t* bmp ) 
{
        return  fgl_createMonoBitmap( bmp->width, bmp->height, 0xff);
}

void fgl_setBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y, fgl_color_t clr ) 
{
  fgl_context_t cx;

  fgl_setBmp( &cx, bmp );
  _fgl_setPixel( &cx, x, y, clr ); 

}

void fgl_setBmpTransparency ( fgl_bitmap_t* bmp, fgl_color_t clr ) 
{
        _fgl_setBmpTransparency ( bmp, clr );
}

fgl_color_t fgl_getBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y ) 
{
  return _fgl_getBmpPixel ( bmp, x, y );
}


void fgl_clear ( fgl_context_t* ctx ) 
{
  _fgl_clear ( ctx );
}

void fgl_setPixel ( fgl_context_t* ctx, u32 x, u32 y, fgl_color_t clr ) /*{{{*/
{
  _fgl_setBmpPixel ( ctx->bmp, x, y, clr );
}

fgl_bitmap_t* fgl_createBitmap ( u32 width, u32 height, i32 clr ) 
{
	fgl_bitmap_t *bmp = (fgl_bitmap_t*) FGL_MALLOC( sizeof(fgl_bitmap_t) );

	if ( !bmp ){
		return 0;
	}
	
	bmp->width  = width;
	bmp->height = height;
        bmp->state = 0;

          {
            SET_ROTATION( fgl_screenBmp );

            BMP_SET_ROTATION( bmp, rot );
            
            if ( rot & 1) {
	      bmp->sLength = _fgl_getScanlineLength ( height );
            }
            else {
	      bmp->sLength = _fgl_getScanlineLength ( width );
            }
          }
        
	bmp->surface = (fgl_slc_t *) FGL_MALLOC ( width * _fgl_getSlcSize() * height );

	if ( clr != -1 ) {
		_fgl_initBmp( bmp, clr);
	}

	bmp->lastOperation = FGL_FRAMEBUFFER_OP;
	bmp->nextOperation = FGL_FRAMEBUFFER_OP;
	return bmp;
}

void fgl_disposeBitmap ( fgl_bitmap_t* bmp ) 
{
  if ( !(bmp->state & IS_SHAREDMEM ))
    {
      FGL_FREE( bmp->surface );
      FGL_FREE( bmp );
    }
}

void fgl_moveArea ( fgl_context_t* ctx,
					i32 x0, i32 y0, i32 x1, i32 y1, i32 dx, i32 dy ) 
{
	i32 u0, v0, u1, v1, w;
	fgl_slc_t *p, *q, *pe;

	if ( !(dx | dy) )
		return;

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	u0 = (i32)x0 + dx;
	v0 = (i32)y0 + dy;

	if ( !clipBlt( ctx, &u0, &v0, &u1, &v1,
				   ctx->bmp->width, ctx->bmp->height, &x0, &y0, &x1, &y1) ){
		return;
	}

        {
          SET_ROTATION( ctx->bmp );
          RECT_ROTATE( rot, ctx->bmp, x0, y0, x1, y1 );
          RECT_ROTATE( rot, ctx->bmp, u0, v0, u1, v1 );
        }

#ifdef USE_ACCEL
	if ( fglaccel_moveArea && (ctx->bmp->state & IMAGE_ON_SCREEN)) {
	  fgl_beginDraw( ctx->bmp, u0, v0, u1, v1, FGL_ACCEL_MOVEAREA_OP );
	  fgl_accel_moveArea( fgl_screenCtx, x0, y0, x1, y1, dx, dy );
	  fgl_endDraw( ctx->bmp, u0, v0, u1, v1);
	  return;
	}
#endif /* USE_ACCEL */

	w = x1 - x0 + 1;

	fgl_beginDraw( ctx->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	if ( v1 > y1 ){
		/* move down, start with bottom scanline */
		p  = _fgl_getSlc( ctx->bmp, x0, y1);
		pe = _fgl_getSlc( ctx->bmp, x0, y0);
		q  = _fgl_getSlc( ctx->bmp, u0, v1);

		while ( p >= pe ){
			_fgl_cpySlc( p, x0, q, u0, w);

			_fgl_subSlc( &p, x0, ctx->bmp->sLength);
			_fgl_subSlc( &q, u0, ctx->bmp->sLength);
		}
	}
	else {
		/* start with top scanline */
		p  = _fgl_getSlc( ctx->bmp, x0, y0);
		pe = _fgl_getSlc( ctx->bmp, x0, y1);
		q  = _fgl_getSlc( ctx->bmp, u0, v0);

		/* non-overlapping scanline copy */
		if ( (v1 < y1) || (x1 < u0) || (u1 < x0) ){
			while ( p <= pe ){
				_fgl_cpySlc( p, x0, q, u0, w);

				_fgl_addSlc( &p, x0, ctx->bmp->sLength);
				_fgl_addSlc( &q, u0, ctx->bmp->sLength);
			}
		}
		else {
			while ( p <= pe ){
				_fgl_moveSlc( p, x0, q, u0, w);

				_fgl_addSlc( &p, x0, ctx->bmp->sLength);
				_fgl_addSlc( &q, u0, ctx->bmp->sLength);
			}
		}
	}

	fgl_endDraw( ctx->bmp, u0, v0, u1, v1);
}

void fgl_bitBlt ( fgl_context_t* dst, i32 u0, i32 v0,
				  fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
{
	i32 u, x, u1, v1, w;
	fgl_slc_t *ps, *pse, *pd;
	u32 ds, dd;
	fgl_color_t pix;

	if ( src == dst->bmp ){
		fgl_moveArea( dst, x0, y0, x1, y1, u0-x0, v0-y0);
		return;
	}

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

        /* this has to be done in more conservative way 
         * first we should compare the rotation of source 
         * and destination and then rotate source to same
         * rotation as the destination.
         */

        fgl_rotateCBitmap( dst->bmp );
        fgl_rotateCBitmap( src );
        {
          SET_ROTATION( src );
          RECT_ROTATE( rot, src, x0, y0, x1, y1 );
          RECT_ROTATE( rot, dst->bmp, u0, v0, u1, v1 );
        }
	
        w = x1 - x0 + 1;

	ps  = _fgl_getSlc( src, x0, y0);
	pse = _fgl_getSlc( src, x0, y1);
	pd  = _fgl_getSlc( dst->bmp, u0, v0);
	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	if ( dst->alphar == 0xff && dst->alphag == 0xff && dst->alphab == 0xff){
		/* this is a opaque blit, delegate as much as possible to the driver */
		while ( ps <= pse ){
			_fgl_cpySlc( ps, x0, pd, u0, w);
			_fgl_addSlc( &ps, x0, src->sLength);
			_fgl_addSlc( &pd, u0, dst->bmp->sLength);
		}
	}
	else {
		/* use the context's alpha value to blend this in */
		ds  = src->sLength - x1 + x0 - 1;
		dd  = dst->bmp->sLength - u1 + u0 - 1; 

		while ( ps <= pse ){
			for ( u=u0, x=x0; u<=u1; u++, x++ ){
				pix = fgl_blendPixelValues( dst->alphar, dst->alphag, 
                                                            dst->alphab,
							    _fgl_getSlcPixel( pd, u),
							    _fgl_getSlcPixel( ps, x));
				
				_fgl_setSlcPixel( pd, u, pix);

				_fgl_incSlc( &pd, u);
				_fgl_incSlc( &ps, x);
			}

			_fgl_addSlc( &pd, u1, dd);
			_fgl_addSlc( &ps, x1, ds);
		}
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}

void fgl_maskBlt ( fgl_context_t* dst, i32 u0, i32 v0,
				   fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1,
				   fgl_monoBitmap_t* mask ) /*{{{*/
{
	u8 *pm;
	fgl_slc_t *pd, *ps;
	fgl_color_t clr;
	i32 x, u, u1, v1, sinc, dinc, minc;
	i32 w, h;
	register u8 a, i;
	u8 i0, off;
	if ( (mask->width < src->width) || (mask->height < src->height) ){
		/* illegal! */
		return;
	}

	if ( x0 > x1 )  swap32( &x0, &x1);
	if ( y0 > y1 )  swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

        fgl_rotateMBitmap( mask );
        fgl_rotateCBitmap( src );
        fgl_rotateCBitmap( dst->bmp );

        {
          SET_ROTATION( src );
          RECT_ROTATE( rot, src, x0, y0, x1, y1 );
        }
        {
          SET_ROTATION( dst->bmp);
          RECT_ROTATE( rot, dst->bmp, u0, v0, u1, v1 );
        }

#ifdef USE_ACCEL
	if ( fgl_accel_maskBlt && (dst->bmp->state & IMAGE_ON_SCREEN)) {
	  fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_ACCEL_MASKBLT_OP );
	  fgl_accel_maskBlt( fgl_screenCtx, u0, v0, src, x0, y0, x1, y1 );
	  fgl_endDraw( dst->bmp, u0, v0, u1, v1);
	  return;
	}
#endif /* USE_ACCEL */

	w = x1 - x0;
	h = y1 - y0;
	off = x0 % 8;

	pm = _fgl_getMonoSurface( mask ) + (y0 * mask->sLength) + (x0 / 8);
	pd = _fgl_getSlc( dst->bmp, u0, v0);
	ps = _fgl_getSlc( src, x0, y0);

	a = *pm;
	u = u0;
	x = x0;
	i = i0 = 8-off;

	dinc = (dst->bmp->sLength - u1) + u0;               /* in pixels */
	sinc = (src->sLength - x1) + x0;                    /* in pixels */
	minc = mask->sLength + (x0>>3) - (x1>>3);           /* in bytes */
	
	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( h-- >= 0 ){
		a <<= off;

		while ( 1 ){
			if ( a & 0x80 ){
				clr = _fgl_getSlcPixel( ps, x);
				_fgl_setSlcPixel( pd, u, clr);
			}

			if ( u == u1 )
				break;

			if ( --i == 0 ){
				/* next source byte */
				i = 8;
				a = *(++pm);
			}
			else {
				a <<= 1;
			}

			_fgl_incSlc( &pd, u);
			u++;
			_fgl_incSlc( &ps, x);
			x++;
		}

		/* next source and destination scanlines */
		_fgl_addSlc( &pd, u, dinc);
		_fgl_addSlc( &ps, x, sinc);

		pm += minc;
		a = *pm;
		u = u0;
		x = x0;
		i = i0;
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}

void fgl_alphaMaskBlt ( fgl_context_t* dst, i32 u0, i32 v0,
						fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1,
						fgl_a8Bitmap_t* mask) 
{
	i32 u1, v1, u, x, da, ds, dd;
	fgl_slc_t *ps, *pse, *pd;
	u8 *pa;
	fgl_color_t pix;

	if ( (mask->width < src->width) || (mask->height < src->height) ){
		/* illegal! */
		return;
	}

	if ( x0 > x1 )		swap32( &x0, &x1);
	if ( y0 > y1 )		swap32( &y0, &y1);

	if ( !clipBlt( dst, &u0, &v0, &u1, &v1,
				   src->width, src->height, &x0, &y0, &x1, &y1) ){
		return;
	}

        fgl_rotateCBitmap( src );
        fgl_rotateA8Bitmap( mask );
        fgl_rotateCBitmap( dst->bmp );

        {
          SET_ROTATION( src );
          RECT_ROTATE( rot, src, x0, y0, x1, y1 );
        }
        {
          SET_ROTATION( dst->bmp);
          RECT_ROTATE( rot, dst->bmp, u0, v0, u1, v1 );
        }

	ps  = _fgl_getSlc( src, x0, y0);
	pse = _fgl_getSlc( src, x0, y1);
	ds  = src->sLength - x1 + x0 -1;

	pd  = _fgl_getSlc( dst->bmp, u0, v0);
	dd  = dst->bmp->sLength - u1 + u0 - 1; 

	pa  = mask->surface;
	da  = mask->sLength - x1 - 1;

	fgl_beginDraw( dst->bmp, u0, v0, u1, v1, FGL_FRAMEBUFFER_OP );

	while ( ps <= pse ){
		for ( u=u0, x=x0; u<=u1; u++, x++ ){
			pix = fgl_blendPixelValues( *pa, *pa, *pa,
						    _fgl_getSlcPixel( pd, u),
						    _fgl_getSlcPixel( ps, x));
			_fgl_setSlcPixel( pd, u, pix);

			_fgl_incSlc( &pd, u);
			_fgl_incSlc( &ps, x);
			pa++;
		}

		pa += da;
		_fgl_addSlc( &pd, u1, dd);
		_fgl_addSlc( &ps, x1, ds);
	}

	fgl_endDraw( dst->bmp, u0, v0, u1, v1);
}

void fgl_saveArea ( fgl_bitmap_t* dst, i32 u0, i32 v0,
        fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 )
{
  _fgl_saveArea ( dst, u0, v0, src, x0, y0, x1, y1 );
}


