/*
 * cbmp.h - fgl module to handle framebuffer-depth bitmaps
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_CBMP_H_INCLUDED
#define FGL_CBMP_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "rot.h"
#include "draw.h"
#include "mbmp.h"

static inline
void _fgl_setBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y, fgl_color_t clr ) 
{
	fgl_slc_t *p;

	if ( (x < bmp->width) && (y < bmp->height) ){
              {
                SET_ROTATION( bmp );
                POINT_ROTATE( rot, bmp, x, y );
              }
            p = _fgl_getSlc( bmp, x, y);
            _fgl_setSlcPixel( p, x, clr);		
        }
}

static inline
void _fgl_setBmpTransparency ( fgl_bitmap_t* bmp, fgl_color_t clr ) 
{
       bmp->transp = clr;
}

static inline
fgl_slc_t* _fgl_getColorSurface( fgl_bitmap_t* bmp )
{
  if ( bmp->state & IS_SHAREDMEM )
    {
      return (fgl_slc_t *)((u8 *)bmp + sizeof( fgl_bitmap_t ));
    }
  else
    {
      return bmp->surface;
    }
}

static inline
fgl_color_t _fgl_getBmpPixel ( fgl_bitmap_t* bmp,  u32 x, u32 y ) 
{
	fgl_slc_t    *p;

	if ( (x < bmp->width) && (y < bmp->height) ){
              {
                SET_ROTATION( bmp );
                POINT_ROTATE( rot, bmp, x, y );
              }
            p = _fgl_getSlc( bmp, x, y);
            return _fgl_getSlcPixel( p, x);
        }
        else {
            return 0;
        }
}

static inline
void _fgl_clear ( fgl_context_t* ctx ) 
{
        fgl_beginDraw( ctx->bmp, 0, 0, ctx->bmp->width-1,
                       ctx->bmp->height-1, FGL_FRAMEBUFFER_OP );  

        _fgl_initBmp( ctx->bmp, ctx->clr);

        fgl_endDraw( ctx->bmp, 0, 0, ctx->bmp->width-1,
                     ctx->bmp->height-1 );  
}

static inline
void _fgl_setPixel ( fgl_context_t* ctx, u32 x, u32 y, fgl_color_t clr ) 
{
  fgl_slc_t *p;

  if ( (x >= ctx->x0Clip) && (x <= ctx->x1Clip) &&
      (y >= ctx->y0Clip) && (y <= ctx->y1Clip) ){
        {
          SET_ROTATION( ctx->bmp );
          POINT_ROTATE( rot, ctx->bmp, x, y );
        }
      p = _fgl_getSlc( ctx->bmp, x, y);
      _fgl_setSlcPixel( p, x, clr);
  }
}

static inline
void _fgl_disposeBitmap ( fgl_bitmap_t* bmp ) 
{
	FGL_FREE( bmp);
}

static inline
void _fgl_saveArea ( fgl_bitmap_t* dst, i32 u0, i32 v0,
    fgl_bitmap_t* src, i32 x0, i32 y0, i32 x1, i32 y1 ) 
{
  fgl_context_t dstCtx;

  fgl_setBmp( &dstCtx, dst);
  fgl_setAlpha( &dstCtx, 0xff);

  fgl_bitBlt( &dstCtx, u0, v0, src, x0, y0, x1, y1);
}

#endif /* FGL_CBMP_H_INCLUDED */

