/*
 * cmap_t565.c - fgl module for pixelvalue/RGB mapping (16 bpp 565 TrueColor)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_CMAP_H_INCLUDED
#define FGL_CMAP_H_INCLUDED

#include "libfgl.h"
#include <fgl.h>

/*
 * these are our public APIs for RGB<->pixVal conversion
 */

static inline
fgl_color_t _fgl_getPixelValue ( u32 argb ) /*{{{*/
{
  return (((argb & 0xf80000) >> 8) |
      ((argb & 0xfc00) >> 5) |
      ((argb & 0xf8) >> 3));
}
/*}}}*/

static inline
void _fgl_getRgbValues ( fgl_color_t pix, u32* r, u32* g, u32* b ) 
{
  *r = (pix & 0xf81f) >> 8;
  *g = (pix & 0x07e0) >> 3;
  *b = (pix & 0x001f) << 3;
}

/*
 * This is here because we can't simply increase/dicrease channel values
 * in a seriously restricted colorspace (VGA, 4-bit grayscale etc.)
 */

static inline
fgl_color_t _fgl_modPixelValue( fgl_color_t pix, int mod ) 
{
  u32 r, g, b;
  fgl_getRgbValues( pix, &r, &g, &b);

  mod = 3 + mod;
  if ( mod < 0 ) mod = 0;

  r = (r * mod) / 3;
  g = (g * mod) / 3;
  b = (b * mod) / 3;

  if ( r > 0xff ) r = 0xff;
  if ( g > 0xff ) g = 0xff;
  if ( b > 0xff ) b = 0xff;

  return ((r << 8) & 0xf81f) | ((g << 3) & 0x7e0) | (b >> 3);
} 

static inline
fgl_color_t _fgl_blendPixelValues ( u32 alphar, u32 alphag, u32 alphab, fgl_color_t d, fgl_color_t s ) 
{
  u32 ialphar = 255 - alphar;
  u32 ialphag = 255 - alphag;
  u32 ialphab = 255 - alphab;

  u32 dr, dg, db, sr, sg, sb;

  fgl_getRgbValues( d, &dr, &dg, &db);
  fgl_getRgbValues( s, &sr, &sg, &sb);

  alphar++;
  alphag++;
  alphab++;

  dr = ((alphar * sr + ialphar * dr) + 255) >> 8;
  dg = ((alphag * sg + ialphag * dg) + 255) >> 8;
  db = ((alphab * sb + ialphab * db) + 255) >> 8;

  return fgl_getPixelValue( RGB_888(dr, dg, db));
}

/*
 * This is the color mapping setup, which is called from init_XX, and
 * therefore can't be inlined
 */

static inline
int _fgl_initRgbCmap ( int nReq, u32* rgbReq, int saveOld )
{
  fgl_clrBlack     = fgl_getPixelValue( 0);
  fgl_clrWhite     = fgl_getPixelValue( 0xffffff);
  fgl_clrRed       = fgl_getPixelValue( 0xff0000);
  fgl_clrGreen     = fgl_getPixelValue( 0x00ff00);
  fgl_clrBlue      = fgl_getPixelValue( 0x0000ff);
  fgl_clrYellow    = fgl_getPixelValue( 0xffff00);
  fgl_clrCyan      = fgl_getPixelValue( 0x00ffff);
  fgl_clrGray      = fgl_getPixelValue( 0x808080);
  fgl_clrLightGray = fgl_getPixelValue( 0xc0c0c0);
  fgl_clrOrange    = fgl_getPixelValue( 0xffc800);

  /* we don't have to set up a colormap */
  return 1;
}

static inline
fgl_cmap_t* _fgl_getCmap () 
{
  return 0;
}

#endif /* FGL_CMAP_H_INCLUDED */
