/* -*- mode: C; mode: fold -*- */
/*
 * cntx.h - fgl module to handle graphics context objects
 *          (attribute containers for color, fonts, clipping)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

/* fgl module epilog */ /*{{{*/
#ifndef FGL_CNTX_H_INCLUDED
#define FGL_CNTX_H_INCLUDED

#include "libfgl.h"
#include "../include/fgl.h"

#include "util.h"

static inline
u8 _fgl_getClipCode ( fgl_context_t *ctx, i32 x, i32 y ) /*{{{*/
{
	u8 cc = 0;


	if ( y > ctx->y1Clip )      cc = CC_BOTTOM;
	else if ( y < ctx->y0Clip ) cc = CC_TOP;

	if ( x > ctx->x1Clip )      cc |= CC_RIGHT;
	else if ( x < ctx->x0Clip ) cc |= CC_LEFT;

	return cc;
}
/*}}}*/

static inline
void _fgl_setClip ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 ) /*{{{*/
{

	if ( x0 > x1 ) swap32( &x0, &x1);
	if ( y0 > y1 ) swap32( &y0, &y1);

	if ( x0 < 0 ) x0 = 0;
	if ( y0 < 0 ) y0 = 0;
	if ( x1 < 0 ) x1 = 0;
	if ( y1 < 0 ) y1 = 0;

	if ( x0 >= (i32)ctx->bmp->width )  x0 = ctx->bmp->width-1;
	if ( x1 >= (i32)ctx->bmp->width )  x1 = ctx->bmp->width-1;
	
	if ( y0 >= (i32)ctx->bmp->height ) y0 = ctx->bmp->height-1;
	if ( y1 >= (i32)ctx->bmp->height ) y1 = ctx->bmp->height-1;

	ctx->x0Clip = x0;
	ctx->y0Clip = y0;
	ctx->x1Clip = x1;
	ctx->y1Clip = y1;
}
/*}}}*/

static inline
void _fgl_setBmp ( fgl_context_t *ctx, fgl_bitmap_t *bmp ) /*{{{*/
{
	ctx->bmp = bmp;

	ctx->x0Clip = 0;
	ctx->y0Clip = 0;
	ctx->x1Clip = bmp->width-1;
	ctx->y1Clip = bmp->height-1;
}
/*}}}*/

static inline
fgl_context_t* _fgl_getContext( fgl_bitmap_t *bmp ) /*{{{*/
{
	fgl_context_t* ctx = (fgl_context_t*)calloc( 1, sizeof(fgl_context_t));
	ctx->alphar = 0xff;
	ctx->alphag = 0xff;
	ctx->alphab = 0xff;
	_fgl_setBmp( ctx, bmp);
	
	return ctx;
}
/*}}}*/

static inline
void _fgl_setFont ( fgl_context_t* ctx, fgl_font_t* fnt ) /*{{{*/
{
	ctx->fnt = fnt;
}
/*}}}*/

static inline
void _fgl_setColor ( fgl_context_t* ctx, fgl_color_t clr ) /*{{{*/
{
	ctx->clr = clr;
}
/*}}}*/

static inline
void _fgl_setAlphas ( fgl_context_t* ctx, u8 alphar, u8 alphag, u8 alphab) /*{{{*/
{
	ctx->alphar = alphar;
	ctx->alphag = alphag;
	ctx->alphab = alphab;
}
/*}}}*/
	
static inline
void _fgl_setXorMode ( fgl_context_t* ctx, fgl_color_t xclr ) /*{{{*/
{
	if ( ctx->mode == PAINT_MODE ){
		ctx->xclr = xclr;
		ctx->clr  ^= xclr;
		ctx->mode = XOR_MODE;
	}
}
/*}}}*/

static inline
void _fgl_setPaintMode ( fgl_context_t* ctx ) /*{{{*/
{
	if ( ctx->mode == XOR_MODE ){
		ctx->clr  ^= ctx->xclr;
		ctx->mode = PAINT_MODE;
	}
}
/*}}}*/

/* fgl module epilog */ /*{{{*/
#endif /* FGL_CNTX_H_INCLUDED */
/*}}}*/
