/*
 * feng_bin.c - fgl module to implement a binary font file (*.fnt)
 *              font engine
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "feng.h"

char *fntDir;

/* static int delta[] = { 0, -1, 1, -2, 2 }; */

fgl_fontEntry_t            *_fgl_fontCache;
int                        _fgl_fontInitialized = 0;
fgl_fontEntry_t            *_fgl_setCache;
int                        _fgl_setInitialized = 0;
fgl_font_t                 *fgl_defaultFont = 0;

int                        _fgl_setCount;
char                       **_fgl_setName;


int fgl_initFontEngine () 
{
	if ( !(fntDir = getenv( "FGL_FONTS")) ){
#if defined(DEF_FONTPATH)
		fntDir = DEF_FONTPATH;
#else
		fntDir = "/usr/local/lib/fgl/fonts";
#endif
	}

	return 1;
}

fgl_font_t* fgl_loadFont ( char* fname, u8 ptSize )
{
  char buf[128];
  int fh;
  struct stat fs;
  u8 *p;
  fgl_font_t *fnt;
  u32 i;
  u8 isFileMsb ;
  static u8 isHostMsb = (__BYTE_ORDER == __BIG_ENDIAN);

  if (strcmp (fname, "default") == 0 )
    {
      sprintf( buf, "%s/%s.fnt", fntDir, fname);
    }
  else
    {
      sprintf( buf, "%s/%s_%d.fnt", fntDir, fname, ptSize);
    }
  fh = open( buf, O_RDONLY);

  if ( fh >= 0 &&  fstat( fh, &fs) != 0)
    {
      fh = -1;
    }

  /* if we still don't find a match, fall back to the default font */
  if ( fh < 0 ){
      /* we have lost, there's nothing else we can try */
      warnx("no matching font file found: %s, %d\n", fname, ptSize);
      return 0;
  }

  p = (u8*) mmap( NULL, fs.st_size, PROT_READ, (MAP_FILE | MAP_SHARED), fh, 0);
  if ( p == (void*)-1 ){
      warnx("mmap %s failed: %d\n", buf, errno);

      p = (u8*) FGL_MALLOC( fs.st_size);
      if ( read( fh, p, fs.st_size) != fs.st_size ){
          FGL_FREE( p);
          return 0;
      }

      fh = -1;
  }

  fnt = (fgl_font_t*) FGL_CALLOC( 1, sizeof(fgl_font_t));

  fnt->fh = fh;
  fnt->fSize = fs.st_size;
  fnt->fAdr = p;

  /*
   * Read font-info from mmaped data, and adapt to different
   * byte order. But watch out - the compiler might do fancy things
   * with mis-aligned short/int addresses (strongARM, for instance)
   */

  fnt->state      = 0;

  isFileMsb       = ((*p) & IS_FILEMSB); 

  if ((*p) & IS_ANTIALIAS )
    {
      fnt->state |= AAF_8;
    }
  p++;

  fnt->nGlyphs    = ((u16)p[1]<<8) + (u16)p[0];   p += sizeof(fnt->nGlyphs);
  fnt->uniOffset  = ((u16)p[1]<<8) + (u16)p[0];   p += sizeof(fnt->uniOffset);

  if ( isFileMsb != isHostMsb )
    {
      fnt->nGlyphs   = swapU16Bytes( fnt->nGlyphs);
      fnt->uniOffset = swapU16Bytes( fnt->uniOffset);
    }

  fnt->fontHeight = *p;          p += sizeof(fnt->fontHeight);
  fnt->ascend     = *p;          p += sizeof(fnt->ascend);
  fnt->descend    = *p;          p += sizeof(fnt->descend);

  fnt->surfOff    = (u32*)p;     p += sizeof(fnt->surfOff[0]) * fnt->nGlyphs;

  if ( isFileMsb != isHostMsb )
    {
      u32* newSurf = (u32*) FGL_MALLOC( sizeof(u32) * fnt->nGlyphs);	
      for ( i=0; i< fnt->nGlyphs; i++ )
        {
          newSurf[i] = swapU32Bytes( fnt->surfOff[i]);
        }
      fnt->surfOff = newSurf;
      fnt->state |= SBO;
    }

  fnt->width      = (u8*)p;        p += sizeof(fnt->width[0]) * fnt->nGlyphs;
  fnt->height     = (u8*)p;        p += sizeof(fnt->height[0]) * fnt->nGlyphs;
  fnt->xBearing   = (i8*) (u8*)p;  p += sizeof(fnt->xBearing[0]) * fnt->nGlyphs;
  fnt->yBearing   = (i8*) (u8*)p;  p += sizeof(fnt->yBearing[0]) * fnt->nGlyphs;
  fnt->advance    = (i8*) (u8*)p;  p += sizeof(fnt->advance[0]) * fnt->nGlyphs;

  fnt->surface    = (u8*)p;

  fnt->fname      = strdup( fname);
  fnt->ptSize     = ptSize; 

  fnt->maxAdvance   = -1;
  fnt->nRefs = 1;
  fnt->sLength = fnt->width;
  
  return fnt;
}

void fgl_unloadFont ( fgl_font_t* fnt ) /*{{{*/
{
	if ( fnt->fh != -1 ){
		munmap( fnt->fAdr, fnt->fSize);
		close( fnt->fh);
	}
	else {
		FGL_FREE( fnt->fAdr);
	}
	if ( fnt->state & RSF ){
		FGL_FREE( fnt->surface);
	}
	if ( fnt->state & SBO ){
		FGL_FREE( fnt->surfOff);
	}
	FGL_FREE( fnt);
}

