/* -*- mode: C; mode: fold -*- */
/*
 * fgld_16_i386.h - fgl header for i386 specific framebuffer access
 *                  accelerations for 16-bpp framebuffer formats
 *           
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */


#ifdef OPTIMIZE_SPACE

static inline
void _stoswl_ ( unsigned int val, unsigned short* dst, int numwords ) /*{{{*/
{
	int dum1, dum2;

    asm  volatile ("rep\n"
				   "stosw\n"
				   : "=D" (dum1), "=c" (dum2)
				   : "a" (val), "D" (dst), "c" (numwords)
				   : "cc");
}
/*}}}*/

static inline
void _movswl_ ( unsigned short* src, unsigned short* dst, int numwords ) /*{{{*/
{
	int dum1, dum2, dum3;

	asm  volatile ("rep\n"
				   "movsw\n"
				   : "=S" (dum1), "=D" (dum2), "=c" (dum3)
				   : "S" (src), "D" (dst), "c" (numwords)
				   : "cc");
}
/*}}}*/

#else /* !OPTIMIZE_SPACE */

/*
 * these two routines perform a "rep stosw/movsw", but by means of
 * double words (which speeds up at least by a factor of 2, depending
 * on the source/destination alignment)
 */

static inline
void _stoswl_ ( unsigned int val, unsigned short* dst, int numwords ) /*{{{*/
{
	int dum1, dum2, dum3;
	
	asm volatile (            "pushl %%ebx\n"
                                  "movl %%eax, %%ebx\n"
				  "sall $16, %%ebx\n"
				  "orl %%ebx, %%eax\n"

				  "testl $3, %%edi\n"
				  "je 0f\n"

				  "movw %%ax, (%%edi)\n"
				  "addl $2, %%edi\n"
				  "decl %%ecx\n"
				  "0:\n"
				  "movl %%ecx, %%ebx\n"
				  "sarl $1, %%ecx\n"

				  "rep\n"
				  "stosl\n"
				  
				  "testl $1, %%ebx\n"
				  "je 1f\n"

				  "movw %%ax, (%%edi)\n"
				  "1:\n"

                                  "popl %%ebx\n"

				  : "=a" (dum1), "=D" (dum2), "=c" (dum3)
				  : "a" (val), "D" (dst), "c" (numwords)
				  : "%ebx", "cc");
} 
/*}}}*/


static inline
void _movswl_ ( unsigned short* src, unsigned short* dest, int numwords ) /*{{{*/
{
	int dum1, dum2, dum3;
	
	asm  volatile (            "pushl %%ebx\n"
                                   "testl $3, %%esi\n"
				   "je 0f\n"

				   "movw (%%esi), %%bx\n"
				   "movw %%bx, (%%edi)\n"
				   "addl $2, %%esi\n"
				   "addl $2, %%edi\n"
				   "decl %%ecx\n"
				   "0:\n"
				   "movl %%ecx, %%ebx\n"
				   "sarl $1, %%ecx\n"

				   "rep\n"
				   "movsl\n"

				   "testl $1, %%ebx\n"
				   "je 1f\n"

				   "movw (%%esi), %%bx\n"
				   "movw %%bx, (%%edi)\n"
				   "1:\n"
                                        
                                   "popl %%ebx\n"
                                        
				   : "=S" (dum1), "=D" (dum2), "=c" (dum3)
				   : "S" (src), "D" (dest), "c" (numwords)
				   : "%ebx", "cc");
}
/*}}}*/

#endif /* !OPTIMIZE_SPACE */


#define HAVE_SETSLCCOLS
static inline
void _fgl_setSlcCols ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	_stoswl_( clr, p, len);
}
/*}}}*/


#define HAVE_INITBMP
static inline
void _fgl_initBmp ( fgl_bitmap_t* bmp, i32 clr )  /*{{{*/
{
	fgl_slc_t* p = bmp->surface;
	u32 len = bmp->width * bmp->height;
	_stoswl_( clr, p, len);
}
/*}}}*/


#define HAVE_CPYSLC
static inline
void _fgl_cpySlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len )  /*{{{*/
{
	_movswl_( src, dest, len);
}
/*}}}*/


#ifdef NEVER
#define HAVE_MOVESLC
static inline
void _fgl_moveSlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len )  /*{{{*/
{
	_movswl_( src, dest, len);
}
/*}}}*/
#endif
