/* -*- mode: C; mode: fold -*- */
/*
 * fgld__i386.h  - fgl header for i386 specific framebuffer access
 *                  accelerations for 8-bpp framebuffer formats
 *           
 * 
 * Copyright (c) 2000
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 *
 * $tvt: fgld_8_i386.h,v 1.3 2001/11/20 20:42:12 damian Exp $
 */

#ifdef OPTIMIZE_SPACE

static inline
void _stosbl_ ( u8 val, u8* dest, int numbytes ) /*{{{*/
{
	int dum1, dum2;

	asm volatile (
				  "rep\n"
				  "stosb\n"
				  : "=D" (dum1), "=c" (dum2)
				  : "a" (val), "D" (dest), "c" (numbytes)
				  : "cc");
}
/*}}}*/

static inline
void _movsbl_ ( u8* src, u8* dest, int numbytes ) /*{{{*/
{
	int dum1, dum2, dum3;

	asm volatile (
				  "rep\n"
				  "movsb\n"
				  : "=S" (dum1), "=D" (dum2), "=c" (dum3)
				  : "S" (src), "D" (dest), "c" (numbytes)
				  : "cc");
}
/*}}}*/

#else /* !OPTIMIZE_SPACE */

/*
 * these two routines perform a "rep stosb/movsb", but by means of
 * double words (which speeds up at least by a factor of 4, depending
 * on the source/destination alignment)
 */

static inline
void _stosbl_ ( u8 val, u8* dest, int numbytes ) /*{{{*/
{
	int dum1, dum2, dum3;
	
	__asm__ __volatile__ (
                                                  "pushl %%ebx\n"
						  "cmpl $8, %%ecx\n"
						  "jg 0f\n"
						  "rep\n"
						  "stosb\n"
						  "jmp 3f\n"

					   "0:\n"
						  /* extend our fill pattern */
						  "andl $255, %%eax\n"
						  "movl %%eax, %%edx\n"
						  "sall $8, %%edx\n"
						  "orl %%eax, %%edx\n"
						  "movl %%edx, %%eax\n"
						  "sall $16, %%eax\n"
						  "orl %%edx, %%eax\n"

						  /* fill up to 4-byte alignment */
						  "movl %%ecx, %%edx\n"
						  "movl %%edi, %%ebx\n"
						  "andl $3, %%ebx\n"
						  "cmpl $0, %%ebx\n"
						  "je 1f\n"

						  "subl %%ebx, %%edx\n"
						  "movl %%ebx, %%ecx\n"
						  "rep\n"
						  "stosb\n"

						  /* stos dwords (if more than 4 bytes left) */
					  "1:\n"
						  "movl %%edx, %%ecx\n"

						  "movl %%ecx, %%ebx\n"
						  "andl $-4, %%ebx\n"
						  "subl %%ebx, %%edx\n"

						  "sarl $2, %%ecx\n"
						  "cmpl $0, %%ecx\n"
						  "je 2f\n"
						  "rep\n"
						  "stosl\n"

						  /* fill up remaining bytes */
					  "2:\n"
						  "movl %%edx, %%ecx\n"
						  "rep\n"
						  "stosb\n"

					  "3:\n"
                                                  "popl %%ebx\n"
						  : "=a" (dum1), "=D" (dum2), "=c" (dum3)
						  : "a" (val), "D" (dest), "c" (numbytes)
						  : "%ebx", "%edx", "cc"
						  );
} 
/*}}}*/

static inline
void _movsbl_ ( u8* src, u8* dest, int numbytes ) /*{{{*/
{
	int dum1, dum2, dum3;
	
	__asm__ __volatile__ (
                                                  "pushl %%ebx\n"
						  "cmpl $8, %%ecx\n"
						  "jg 0f\n"
						  "rep\n"
						  "movsb\n"
						  "jmp 3f\n"

					  "0:\n"
						  /* copy up to 4-byte alignment */
						  "movl %%ecx, %%edx\n"
						  "movl %%edi, %%ebx\n"
						  "andl $3, %%ebx\n"
						  "cmpl $0, %%ebx\n"
						  "je 1f\n"

						  "subl %%ebx, %%edx\n"
						  "movl %%ebx, %%ecx\n"
						  "rep\n"
						  "movsb\n"

						  /* movsl dwords (if more than 4 bytes left) */
					  "1:\n"
						  "movl %%edx, %%ecx\n"

						  "movl %%ecx, %%ebx\n"
						  "andl $-4, %%ebx\n"
						  "subl %%ebx, %%edx\n"

						  "sarl $2, %%ecx\n"
						  "cmpl $0, %%ecx\n"
						  "je 2f\n"
						  "rep\n"
						  "movsl\n"

						  /* copy remaining bytes */
					  "2:\n"
						  "movl %%edx, %%ecx\n"
						  "rep\n"
						  "movsb\n"

					  "3:\n"
                                                  "popl %%ebx\n"
						  : "=S" (dum1), "=D" (dum2), "=c" (dum3)
						  : "S" (src), "D" (dest), "c" (numbytes)
						  : "%ebx", "%edx", "cc"
						  );
}
/*}}}*/

#endif /* !OPTIMIZE_SPACE */


#define HAVE_SETSLCCOLS
static inline
void _fgl_setSlcCols ( fgl_slc_t *p, u32 x, u32 len, fgl_color_t clr ) /*{{{*/
{
	_stosbl_( clr, p, len);
}
/*}}}*/


#define HAVE_INITBMP
static inline
void _fgl_initBmp ( fgl_bitmap_t* bmp, i32 clr )  /*{{{*/
{
	fgl_slc_t* p = bmp->surface;
	u32 len = bmp->sLength * bmp->height;

	_stosbl_( clr, p, len);
}
/*}}}*/


#define HAVE_CPYSLC
static inline
void _fgl_cpySlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len )  /*{{{*/
{
	_movsbl_( src, dest, len);
}
/*}}}*/


#define HAVE_MOVESLC
static inline
void _fgl_moveSlc ( fgl_slc_t *src, u32 x, fgl_slc_t *dest, u32 u, u32 len )  /*{{{*/
{
	_movsbl_( src, dest, len);
}
/*}}}*/
