/*
 * line.c - fgl module to draw lines
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 *
 * 01/01/00  Peter Mehlitz
 */

#include "line.h"

void fgl_drawHLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1 )
{
  i32 y1 = y0;

  if ( x1 < x0 ) swap32( &x0, &x1);

  if (( x0 > ctx->x1Clip ) || ( x1 < ctx->x0Clip ) ||
      ( y1 < ctx->y0Clip ) || ( y0 > ctx->y1Clip )) return;

  if ( x0 < ctx->x0Clip ) x0 = ctx->x0Clip;
  if ( x1 > ctx->x1Clip ) x1 = ctx->x1Clip;

  /* set the rotation */
    {
      SET_ROTATION( ctx->bmp )
      LINE_ROTATE( rot, ctx->bmp, x0, y0, x1, y1 )
      
      if ( rot & 1 )
        {
  if ( y1 < y0 ) swap32( &y0, &y1);
          _fgl_drawVLine( ctx, x0, y0, y1 );
        }
      else
        {
  if ( x1 < x0 ) swap32( &x0, &x1);
          _fgl_drawHLine ( ctx, x0, y0, x1 );
        }
    }
}

void fgl_drawVLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 y1 )
{
  i32 x1 = x0;
  
  if ( y1 < y0 ) swap32( &y0, &y1);

  if ( (x0 > ctx->x1Clip) || (x0 < ctx->x0Clip) ||
      (y1 < ctx->y0Clip) || (y0 > ctx->y1Clip) )
      return;

  if ( y0 < ctx->y0Clip ) y0 = ctx->y0Clip;
  if ( y1 > ctx->y1Clip ) y1 = ctx->y1Clip;
 
  /* set the rotation */
    {
      SET_ROTATION( ctx->bmp );
      LINE_ROTATE( rot, ctx->bmp, x0, y0, x1, y1 );
      if ( rot & 1 )
        {
  if ( x1 < x0 ) swap32( &x0, &x1);
          _fgl_drawHLine( ctx, x0, y0, x1 );
        }
      else
        {
  if ( y1 < y0 ) swap32( &y0, &y1);
          _fgl_drawVLine ( ctx, x0, y0, y1 );
        }
    }
}

void fgl_drawLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 )
{
  fgl_slc_t *p;

  i32 x, dx, dy, a, b, c, slinc, slinc1;

  if ( !clipLine( ctx, &x0, &y0, &x1, &y1) )
    {
      return;
    }
  
  /* set the rotation */
    { 
      SET_ROTATION( ctx->bmp );
      LINE_ROTATE( rot, ctx->bmp, x0, y0, x1, y1 );
    }

#ifdef USE_ACCEL
  if ( (fgl_accel_drawLine) && (ctx->bmp->state & IMAGE_ON_SCREEN) )
    {
      fgl_screenCtx->clr = ctx->clr;
      fgl_screenCtx->mode = ctx->mode;
      fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_ACCEL_DRAWLINE_OP );
      fgl_accel_drawLine( fgl_screenCtx, x0, y0, x1, y1);
      fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
      return ;
    }
#endif /* USE_ACCEL */

  if ( x1 < x0 )
    {
      swap32( &x0, &x1);
      swap32( &y0, &y1);
    }

  dx = (i32)x1 - (i32)x0;
  dy = (i32)y1 - (i32)y0;

  p = _fgl_getSlc( ctx->bmp, x0, y0);

  if ( dy < 0 )
    {
      swap32( &y0, &y1);
    }

  fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

  if ( dy == 0 )
    {
      MODE_BRANCH( ctx, _fgl_setSlcCols,(p, x0, dx+1, ctx->clr));
    }
  else if ( dx == 0 )
    {
      i32 inc, len;
      if ( dy > 0 )
        {
          inc = ctx->bmp->sLength; len = dy;
        }
      else
        {
          inc = - (i32)ctx->bmp->sLength; len = -dy;
        }
      MODE_BRANCH( ctx, _fgl_setSlcRows,(p, inc, x0, len+1, ctx->clr));
    }
  else if ( dx == -dy )
    {
      MODE_BRANCH( ctx, _fgl_setSlcRows,(p, -(ctx->bmp->sLength-1), x0, -dy+1, ctx->clr));
    }
  else if ( dx == dy )
    {
      MODE_BRANCH( ctx, _fgl_setSlcRows,(p, ctx->bmp->sLength+1, x0, dy+1, ctx->clr));
    }
  else {
      x = x0;
      slinc = ctx->bmp->sLength;
      if ( dy < 0 )
        {
          dy = - dy;
          slinc = -slinc;
        }
      slinc1 = slinc+1;

      if ( dx > dy )
        {
          a = 2 * dy;
          b = 2 * (dy-dx);
          c = 2 * dy - dx;


          if ( ctx->mode == PAINT_MODE )
            {
              _fgl_setSlcPixel( p, x, ctx->clr);
              while ( x < x1 )
                {
                  if ( c < 0 )
                    {
                      _fgl_incSlc( &p, x); c += a;
                    }
                  else
                    {
                      _fgl_addSlc( &p, x, slinc1); c += b;
                    }
                  x++;
                  _fgl_setSlcPixel( p, x, ctx->clr);
                }
            }
          else
            {
              _fgl_setSlcPixelXor( p, x, ctx->clr);
              while ( x <= x1 )
                {
                  if ( c < 0 )
                    {
                      _fgl_incSlc( &p, x); c += a;
                    }
                  else
                    {
                      _fgl_addSlc( &p, x, slinc1); c += b;
                    }
                  x++;
                  _fgl_setSlcPixelXor( p, x, ctx->clr);
                }
            }
        }
      else
        {
          a = 2 * dx;
          b = 2 * (dx-dy);
          c = 2 * dx - dy;

          if ( ctx->mode == PAINT_MODE ){
              _fgl_setSlcPixel( p, x, ctx->clr);
              while ( dy-- )
                {
                  if ( c < 0 )
                    {
                      _fgl_addSlc( &p, x, slinc); c += a;
                    }
                  else
                    {
                      _fgl_addSlc( &p, x, slinc1); c += b;
                      x++;
                    }
                  _fgl_setSlcPixel( p, x, ctx->clr);
                }
          }
          else
            {
              _fgl_setSlcPixelXor( p, x, ctx->clr);
              while ( dy-- )
                {
                  if ( c < 0 )
                    {
                      _fgl_addSlc( &p, x, slinc); c += a;
                    }
                  else
                    {
                      _fgl_addSlc( &p, x, slinc1); c += b;
                      x++;
                    }
                  _fgl_setSlcPixelXor( p, x, ctx->clr);
                }
            }
        }
  }

  fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}
