/*
 * line.h - fgl module to draw lines
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 *
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_LINE_H_INCLUDED
#define FGL_LINE_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "accel.h"
#include "draw.h"
#include "rot.h"
#include "util.h"

static inline
void _fgl_drawHLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1 )
{
  fgl_slc_t* p;
  
#ifdef USE_ACCEL
  if ( (fgl_accel_drawHLine) && (ctx->bmp->state & IMAGE_ON_SCREEN) )
    {
      fgl_screenCtx->clr = ctx->clr ;
      fgl_screenCtx->mode = ctx->mode;
      fgl_beginDraw( ctx->bmp, x0, y0, x1, y0, FGL_ACCEL_DRAWLINE_OP );
      fgl_accel_drawHLine( fgl_screenCtx, x0, y0, x1 );
      fgl_endDraw( ctx->bmp, x0, y0, x1, y0);
      return ;
   }
#endif /* USE_ACCEL */
  p = _fgl_getSlc( ctx->bmp, x0, y0);

  fgl_beginDraw( ctx->bmp, x0, y0, x1, y0, FGL_FRAMEBUFFER_OP );
  MODE_BRANCH( ctx, _fgl_setSlcCols, (p, x0, (x1-x0)+1, ctx->clr));
  fgl_endDraw( ctx->bmp, x0, y0, x1, y0);
}

static inline
void _fgl_drawVLine ( fgl_context_t *ctx, i32 x0, i32 y0, i32 y1 )
{
  fgl_slc_t *p;
#ifdef USE_ACCEL
  if ( (fgl_accel_drawVLine) && (ctx->bmp->state & IMAGE_ON_SCREEN) )
    {
      fgl_screenCtx->clr = ctx->clr;
      fgl_screenCtx->mode = ctx->mode;
      fgl_beginDraw( ctx->bmp, x0, y0, x0, y1, FGL_ACCEL_DRAWLINE_OP );
      fgl_accel_drawVLine( fgl_screenCtx, x0, y0, y1 );
      fgl_endDraw( ctx->bmp, x0, y0, x0, y1);
      return ;
  }
#endif /* USE_ACCEL */
  p = _fgl_getSlc( ctx->bmp, x0, y0);

  fgl_beginDraw( ctx->bmp, x0, y0, x0, y1, FGL_FRAMEBUFFER_OP );
  MODE_BRANCH( ctx, _fgl_setSlcRows, (p, ctx->bmp->sLength, x0, (y1-y0)+1, ctx->clr));
  fgl_endDraw( ctx->bmp, x0, y0, x0, y1);
}

static inline
i32 rnd ( i32 a, i32 b, i32 c )
{
	const int D = 8192;
	i32 d = (a * b) * D;
	d += (d <0) ? -D/2 : D/2;
	return (d / c) / D;
}

static inline
int clipLine ( fgl_context_t* ctx, i32* px0, i32* py0, i32* px1, i32* py1 )
{
  u8 c0, c1;
  i32 dx, dy;
	
  c0 = fgl_getClipCode( ctx, *px0, *py0);
  c1 = fgl_getClipCode( ctx, *px1, *py1);

  while ( 1 )
    {
      if ( (c0 | c1) == 0 )
        {
          return 1;
        }
      else if ( (c0 & c1) != 0 )
        {
          return 0;
        }
      else 
        {
          if ( !c0 )
            {
              swap32( px0, px1);
              swap32( py0, py1);
              swap8( &c0, &c1);
            }

          dx = (*px1 - *px0);
          dy = (*py1 - *py0);

          if ( c0 & CC_LEFT )
            {
			  *py0 += rnd( ctx->x0Clip - *px0, dy, dx);
              *px0 = ctx->x0Clip;

              c0 &= ~CC_LEFT;
              if ( *py0 > ctx->y1Clip ) c0 |= CC_BOTTOM;
              else if ( *py0 < ctx->y0Clip ) c0 |= CC_TOP;
              else c0 &= ~(CC_BOTTOM | CC_TOP);
            }
          else if ( c0 & CC_RIGHT )
            {
			  *py0 += rnd( ctx->x1Clip - *px0, dy, dx);
              *px0 = ctx->x1Clip;

              c0 &= ~CC_RIGHT;
              if ( *py0 > ctx->y1Clip ) c0 |= CC_BOTTOM;
              else if ( *py0 < ctx->y0Clip ) c0 |= CC_TOP;
              else c0 &= ~(CC_BOTTOM | CC_TOP);
          }
          else if ( c0 & CC_BOTTOM )
            {
              if ( *px0 != *px1 ) 
                {
				  *px0 += rnd( ctx->y1Clip - *py0, dx, dy);
                }
              *py0 = ctx->y1Clip;

              c0 &= ~CC_BOTTOM;
              if( *px0 > ctx->x1Clip ) c0 |= CC_RIGHT;
              else if ( *px0 < ctx->x0Clip ) c0 |= CC_LEFT;
              else c0 &= ~(CC_RIGHT | CC_LEFT);
            }
          else 
            {
              if ( *px0 != *px1 ) 
				{
                  *px0 += rnd( ctx->y0Clip - *py0, dx, dy);
                }
              *py0 = ctx->y0Clip;

              c0 &= ~CC_TOP;
              if( *px0 > ctx->x1Clip ) c0 |= CC_RIGHT;
              else if ( *px0 < ctx->x0Clip ) c0 |= CC_LEFT;
              else c0 &= ~(CC_RIGHT | CC_LEFT);
            }
        }
    }
}

#endif /* FGL_LINE_H_INCLUDED */
