/*
 * mbmp.h - fgl module to handle monochrome bitmaps (depth 1)
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_MBMP_H_INCLUDED
#define FGL_MBMP_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "rot.h"
#include "draw.h"

static inline
void _fgl_setMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) 
{
	u8 *p;
	p = bmp->surface + (row * bmp->sLength) + (col / 8);
	*p |= _monoBits[col % 8];
}

static inline
void _fgl_unsetMonoBit ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) 
{
	u8 *p;
	p = bmp->surface + (row * bmp->sLength) + (col / 8);
	*p &= ~_monoBits[col % 8];
}

static inline
int _fgl_isMonoBitSet ( fgl_monoBitmap_t* bmp, u32 col, u32 row ) 
{
	u8 *p;
	p = bmp->surface + (row * bmp->sLength) + (col / 8);
	return (*p & _monoBits[col % 8]);
}

static inline
u8* _fgl_getMonoSurface( fgl_monoBitmap_t* bmp )
{ 
  if ( bmp->state & IS_SHAREDMEM )
    {
      return (u8 *)bmp + sizeof( fgl_monoBitmap_t );
    }
  else
    {
      return bmp->surface;
    }
}

static inline
void _fgl_disposeMonoBitmap ( fgl_monoBitmap_t* bmp ) 
{
  FGL_FREE( bmp->surface );
  FGL_FREE( bmp);
}

#endif /* FGL_MBMP_H_INCLUDED */
