/*
 * oval.c - fgl module to draw and fill circles and ellipses
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "oval.h"

void fgl_drawCircle ( fgl_context_t* ctx, i32 cx, i32 cy, u32 r ) 
{
	i32 x0, y0, x1, y1;
	i32 x = 0, y = r;
	i32 a = 1 - r;
	u8  clip;

	x0 = cx - r;
	x1 = cx + r;
	y0 = cy - r;
	y1 = cy + r;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	_fgl_setPixel( ctx, cx, cy+r, ctx->clr);
	_fgl_setPixel( ctx, cx, cy-r, ctx->clr);
	_fgl_setPixel( ctx, cx+r, cy, ctx->clr);
	_fgl_setPixel( ctx, cx-r, cy, ctx->clr);

	while ( x < y ){
		x++;
		if ( a < 0 ){
			a += 2*x + 1;
		}
		else {
			y--;
			a += 2 * (x - y) + 1;
		}

		_fgl_setCirclePoints( ctx, cx, cy, x, y, clip);
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}

void fgl_fillCircle ( fgl_context_t* ctx, i32 cx, i32 cy, u32 r ) 
{
	i32 x0, y0, x1, y1;
	i32 x = 0, y = r;
	i32 a = 1 - r;
	u8  clip;

	x0 = cx - r;
	x1 = cx + r;
	y0 = cy - r;
	y1 = cy + r;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}
          {
                SET_ROTATION( ctx->bmp );
                POINT_ROTATE( rot, ctx->bmp, cx, cy );
          }
        
	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );
	
	while ( x < y ){
		if ( a < 0 ){
			a += 2*x + 1;
		}
		else {
			y--;
			a += 2 * (x - y) + 1;
		}

		_fgl_fillCirclePoints( ctx, cx, cy, x, y, clip);
		x++;
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}

void fgl_drawEllipse ( fgl_context_t* ctx, i32 cx, i32 cy, u32 rx, u32 ry ) 
{
	i32 x0, y0, x1, y1;
	i32 x, y, a, ax, ay;
	i32 rxrx, ryry,  rxrx2, ryry2;
	i8  clip;

	if ( rx == ry ) {
		fgl_drawCircle( ctx, cx, cy, rx);
		return;
	}

	x0 = cx - rx;
	x1 = cx + rx;
	y0 = cy - ry;
	y1 = cy + ry;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	rxrx = rx*rx;
	ryry = ry*ry;
	rxrx2 = 2 * rxrx;
	ryry2 = 2 * ryry;

	x = 0;   ax = 0;
	y = ry;  ay = rxrx2 * y;
	a = ( (20*(ryry - (rxrx * (i32)ry)) + 5*rxrx) + 10) / 20;

	_fgl_setPixel( ctx, cx, cy + ry, ctx->clr);
	_fgl_setPixel( ctx, cx, cy - ry, ctx->clr);

	while ( ax < ay ){
		x++;
		ax += ryry2;
		if ( a < 0 ){
			a += ryry + ax;
		}
		else {
			y--;
			ay -= rxrx2;
			a += ryry + ax - ay;
		}

		_fgl_setEllipsePoints( ctx, cx, cy, x, y, clip);
	}
	
	a = (ryry*(20*x*x +20*x + 5) + 20*(rxrx*(y-1)*(y-1) - rxrx*ryry) + 10) / 20;
	while ( y > 0 ){
		y--;
		ay -= rxrx2;
		if ( a > 0 ){
			a += rxrx - ay;
		}
		else {
			x++;
			ax += ryry2;
			a += rxrx - ay + ax;
		}

		_fgl_setEllipsePoints( ctx, cx, cy, x, y, clip);
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}

void fgl_fillEllipse ( fgl_context_t* ctx, i32 cx, i32 cy, u32 rx, u32 ry ) 
{
	i32 x0, y0, x1, y1;
	i32 x, y, a, ax, ay;
	i32 rxrx, ryry,  rxrx2, ryry2;
	i8  clip;

	if ( rx == ry ) {
		fgl_fillCircle( ctx, cx, cy, rx);
		return;
	}

	x0 = cx - rx;
	x1 = cx + rx;
	y0 = cy - ry;
	y1 = cy + ry;
	if ( (clip = clipRel( ctx, x0, y0, x1, y1)) == CR_OUT ){
		return;
	}

          {
                SET_ROTATION( ctx->bmp );
                POINT_ROTATE( rot, ctx->bmp, cx, cy );
                if ( rot & 1 ) {
                    swap32( &rx, &ry );
                }
          }
        
	fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

	rxrx = rx*rx;
	ryry = ry*ry;
	rxrx2 = 2 * rxrx;
	ryry2 = 2 * ryry;

	x = 0;   ax = 0;
	y = ry;  ay = rxrx2 * y;

	a = ( (20*(ryry - (rxrx * (i32)ry)) + 5*rxrx) + 10) / 20;
	while ( ax < ay ){
		ax += ryry2;
		if ( a < 0 ){
			a += ryry + ax;
		}
		else {
			y--;
			ay -= rxrx2;
			a += ryry + ax - ay;
		}

		_fgl_fillEllipsePoints( ctx, cx, cy, x, y, clip);
		x++;
	}

	a = (ryry*(20*x*x +20*x + 5) + 20*(rxrx*(y-1)*(y-1) - rxrx*ryry) + 10) / 20;
	while ( y > 0 ){
		y--;
		ay -= rxrx2;
		if ( a > 0 ){
			a += rxrx - ay;
		}
		else {
			x++;
			ax += ryry2;
			a += rxrx - ay + ax;
		}

		_fgl_fillEllipsePoints( ctx, cx, cy, x, y, clip);
	}

	fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}

