/*
 * oval.h - fgl module to draw and fill circles and ellipses
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_OVAL_H_INCLUDED
#define FGL_OVAL_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "cbmp.h"
#include "draw.h"
#include "rot.h"
#include "util.h"

static inline
void _fgl_setCirclePoints ( fgl_context_t* ctx, i32 cx, i32 cy, i32 dx, i32 dy, u8 clip ) 
{
  u32 x0 = cx + dx;
  u32 y0 = cy + dy;
  u32 x1 = cx - dx;
  u32 y1 = cy - dy;
  u32 x2 = cx + dy;
  u32 y2 = cy + dx;
  u32 x3 = cx - dy;
  u32 y3 = cy - dx;

  if ( clip == CR_IN ){
      _fgl_setPixel( ctx, x0, y0, ctx->clr); 
      _fgl_setPixel( ctx, x1, y0, ctx->clr); 
      _fgl_setPixel( ctx, x0, y1, ctx->clr); 
      _fgl_setPixel( ctx, x1, y1, ctx->clr); 
      _fgl_setPixel( ctx, x2, y2, ctx->clr); 
      _fgl_setPixel( ctx, x3, y2, ctx->clr); 
      _fgl_setPixel( ctx, x2, y3, ctx->clr); 
      _fgl_setPixel( ctx, x3, y3, ctx->clr); 
  }
  else {
      if ( inInterval( x0, ctx->x0Clip, ctx->x1Clip) ){
          if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x0, y0, ctx->clr);
          }
          if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x0, y1, ctx->clr);
          }
      }
      if ( inInterval( x1, ctx->x0Clip, ctx->x1Clip) ){
          if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x1, y0, ctx->clr);
          }
          if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x1, y1, ctx->clr);
          }
      }		
      if ( inInterval( x2, ctx->x0Clip, ctx->x1Clip) ){
          if ( inInterval( y2, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x2, y2, ctx->clr);
          }
          if ( inInterval( y3, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x2, y3, ctx->clr);
          }
      }		
      if ( inInterval( x3, ctx->x0Clip, ctx->x1Clip) ){
          if ( inInterval( y2, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x3, y2, ctx->clr);
          }
          if ( inInterval( y3, ctx->y0Clip, ctx->y1Clip) ){
              _fgl_setPixel( ctx, x3, y3, ctx->clr);
          }
      }		
  }
}

static inline
void _fgl_fillCirclePoints ( fgl_context_t* ctx, i32 cx, i32 cy, i32 dx, i32 dy, u8 clip ) 
{
	fgl_slc_t *p;

	i32 x0 = cx - dx;
	i32 y0 = cy - dy;
	i32 x1 = cx + dx;
	i32 y1 = cy + dy;
	i32 x2 = cx - dy;
	i32 y2 = cy - dx;
	i32 x3 = cx + dy;
	i32 y3 = cy + dx;
	i32 w1, w2;
	
	if ( clip == CR_PART ){
		x0 = (ctx->x0Clip > x0) ? ctx->x0Clip : x0;
		x2 = (ctx->x0Clip > x2) ? ctx->x0Clip : x2;
		x1 = (ctx->x1Clip < x1) ? ctx->x1Clip : x1;
		x3 = (ctx->x1Clip < x3) ? ctx->x1Clip : x3;

		w1 = x1 - x0 +1;
		w2 = x3 - x2 +1;

		if ( w1 > 0 ) {
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				p = _fgl_getSlc( ctx->bmp, x0, y0);
				_fgl_setSlcCols( p, x0, w1, ctx->clr);
			}

			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				p = _fgl_getSlc( ctx->bmp, x0, y1);
				_fgl_setSlcCols( p, x0, w1, ctx->clr);
			}
		}

		if ( w2 > 0 ) {
			if ( inInterval( y2, ctx->y0Clip, ctx->y1Clip) ){
				p = _fgl_getSlc( ctx->bmp, x2, y2);
				_fgl_setSlcCols(p, x2, w2, ctx->clr);
			}
			
			if ( inInterval( y3, ctx->y0Clip, ctx->y1Clip) ){
				p = _fgl_getSlc( ctx->bmp, x2, y3);
				_fgl_setSlcCols(p, x2, w2, ctx->clr);
			}
		}
	}
	else {
		p = _fgl_getSlc( ctx->bmp, x0, y0);
		_fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);

		p = _fgl_getSlc( ctx->bmp, x0, y1);
		_fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);

		p = _fgl_getSlc( ctx->bmp, x2, y2);
		_fgl_setSlcCols(p, x2, (x3-x2)+1, ctx->clr);

		p = _fgl_getSlc( ctx->bmp, x2, y3);
		_fgl_setSlcCols(p, x2, (x3-x2)+1, ctx->clr);
	}
}

static inline
void _fgl_setEllipsePoints ( fgl_context_t* ctx, i32 cx, i32 cy, i32 dx, i32 dy, i8 clip ) 
{

	u32 x0 = cx + dx;
	u32 y0 = cy + dy;
	u32 x1 = cx - dx;
	u32 y1 = cy - dy;

	if ( clip == CR_IN ){
		_fgl_setPixel( ctx, x0, y0, ctx->clr);
		_fgl_setPixel( ctx, x1, y0, ctx->clr);
		_fgl_setPixel( ctx, x0, y1, ctx->clr);
		_fgl_setPixel( ctx, x1, y1, ctx->clr);
	}
	else {
		if ( inInterval( x0, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				_fgl_setPixel( ctx, x0, y0, ctx->clr);
			}
			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				_fgl_setPixel( ctx, x0, y1, ctx->clr);
			}
		}
		if ( inInterval( x1, ctx->x0Clip, ctx->x1Clip) ){
			if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
				_fgl_setPixel( ctx, x1, y0, ctx->clr);
			}
			if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
				_fgl_setPixel( ctx, x1, y1, ctx->clr);
			}
		}
	}
}

static inline
void _fgl_fillEllipsePoints ( fgl_context_t* ctx, i32 cx, i32 cy, i32 dx, i32 dy, i8 clip ) 
{
	fgl_slc_t *p;

	i32 x0 = cx - dx;
	i32 y0 = cy - dy;
	i32 x1 = cx + dx;
	i32 y1 = cy + dy;

	if ( clip == CR_PART ){
		x0 = (ctx->x0Clip > x0) ? ctx->x0Clip : x0;
		x1 = (ctx->x1Clip < x1) ? ctx->x1Clip : x1;
		y0 = (ctx->y0Clip > y0) ? ctx->y0Clip : y0;
		y1 = (ctx->y1Clip < y1) ? ctx->y1Clip : y1;
		
		if ( x1 < x0 ){
			return;
		}

		if ( inInterval( y0, ctx->y0Clip, ctx->y1Clip) ){
			p = _fgl_getSlc( ctx->bmp, x0, y0);
			_fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);
		}

		if ( inInterval( y1, ctx->y0Clip, ctx->y1Clip) ){
			p = _fgl_getSlc( ctx->bmp, x0, y1);
			_fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);
		}
		
	}
	else {	
		p = _fgl_getSlc( ctx->bmp, x0, y0);
		_fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);

		p = _fgl_getSlc( ctx->bmp, x0, y1);
		_fgl_setSlcCols( p, x0, (x1-x0)+1, ctx->clr);
	}
}

#endif /* FGL_OVAL_H_INCLUDED */
