/*
 * poly.c - fgl module to draw and fill polygons
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#include "poly.h"

void fgl_drawPolygon ( fgl_context_t* ctx, fgl_point_t* points, u32 nPoints )
{
	fgl_point_t *pt;
	fgl_point_t *ptMax = (points + nPoints-1);
	i32 x0, y0;

	for ( pt = points; pt < ptMax; ){
		x0 = pt->x;
		y0 = pt->y;
		pt++;
		fgl_drawLine( ctx, x0, y0, pt->x, pt->y); 
	}
}

void fgl_fillPolygonClr ( fgl_context_t* ctx, fgl_point_t* points,
						  u32 nPoints, fgl_color_t clr )
{
	int xmin, xmax, ymin, ymax;
	int n, clip = 0;
	int y, x0, x1, dx, xlast;
	edge_t *nextYmin = 0, *active = 0, *el, *e0, *e1;
	fgl_slc_t *p;	

        /* we don't need the polygon to be closed explicitly */
	n = nPoints-1;
	if ( (points[n].x == points[0].x) && (points[n].y == points[0].y ) ){
		nPoints--;
	}

	boundingBoxPoly( points, nPoints, &xmin, &ymin, &xmax, &ymax);

	if ( (xmin < ctx->x0Clip) || (xmax > ctx->x1Clip) ||
		 (ymin < ctx->y0Clip) || (ymax > ctx->y1Clip) ){
		clip = 1;
	}

	n  = nPoints * sizeof(edge_t);
	if ( nPoints > MAX_EDGES ){
		el = (edge_t*) FGL_MALLOC( n);
	}
	else {
		el = (edge_t*) FGL_ALLOCA( n);
	}
	memset( el, 0, n);

	initEdgeList( points, nPoints, &nextYmin, el, ymin);

	fgl_beginDraw( ctx->bmp, xmin, ymin, xmax, ymax, FGL_FRAMEBUFFER_OP );

	p = _fgl_getSlc( ctx->bmp, 0, ymin);
	xlast = dx = 0;
	
	for ( y=ymin; updateActiveList( &active, &nextYmin, y); y++ ){

		dx += active->x;
		e0 = active;
		
		if ( clip ){
			if ( y >= ctx->y1Clip ) /* that's it, get out of here */
				break;

			if ( y >= ctx->y0Clip ){
				while ( 1 ){
					e1 = e0->nextActive;

					x0 = (e0->x < ctx->x0Clip) ? ctx->x0Clip : e0->x;
					x1 = (e1->x > ctx->x1Clip) ? ctx->x1Clip : e1->x;

					_fgl_addSlc( &p, xlast, dx);
					xlast = x0;
					
					if ( (x0 <= ctx->x1Clip) && (x1 >= ctx->x0Clip) ){
						MODE_BRANCH( ctx, _fgl_setSlcCols, (p, x0, (x1-x0)+1, ctx->clr));
					}

					if ( e1->nextActive ){
						e0 = e1->nextActive;
						dx = e0->x - xlast;
					}
					else {
						break;
					}
				}
			}
			else {
				dx += ctx->bmp->sLength;
			}
		}
		else {	
			while ( 1 ){
				_fgl_addSlc( &p, xlast, dx);
				xlast = e0->x;
				
				e1 = e0->nextActive;
  				MODE_BRANCH( ctx, _fgl_setSlcCols, (p, e0->x, (e1->x - e0->x)+1, ctx->clr));			
				
				if ( e1->nextActive ){
					e0 = e1->nextActive;
					dx = e0->x - xlast;
				}
				else {
					break;
				}
			}
		}
		
		dx = ctx->bmp->sLength - xlast;
	}

	fgl_endDraw( ctx->bmp, xmin, ymin, xmax, ymax);

	if ( nPoints > MAX_EDGES ){
		FGL_FREE( el);
	}
	else {
		FGL_ALLOCA_FREE( el);
	}
}

