#include "psrv.h"

#define SHMKEY "psrvsid"

#define FATAL(format, arg...)  fprintf( stderr, format, ## arg)
#define DBG(format, arg...)  //fprintf( stderr, format, ## arg)

fgl_PServer_t* PServer;
int PServerHandle;

/*******************************
 * convenience functions
 ******************************/
static int getShmem( int size, void** paddr) {
  int sid = shmget( IPC_PRIVATE, size, IPC_CREAT | 0777);
  if ( sid == -1 ) {
    FATAL("SHMEM error\n");
    return -1;
  }

  *paddr = shmat( sid, 0, 0);
  if ( *paddr == (void*)-1 ) {
    FATAL( "SHMEM error\n");
    return -1;
  }

  return sid;
}

static void registerHandle( int sid) {
  char buf[20];
  sprintf( buf, "%d", sid);
  setenv( SHMKEY, buf, 1);

  PServerHandle = sid;
}

static fgl_PServer_t* initClient() {
  //entry point for client, init fgl and and shared memory access
  fgl_init( 0, 0, 0);
  PServerHandle = atoi( getenv( SHMKEY));

  //attach shared memory to client process
  return ((fgl_PServer_t*)shmat( PServerHandle, 0, 0));
}

/********************************
 * server API
 ********************************/
fgl_PServer_t*
fgl_PServerInit() {
  int sid;
  if ( PServer == NULL ) {

    DBG( "PServer init %d %d %d %d\n", fgl_screenCtx->x0Clip,
	 fgl_screenCtx->y0Clip, fgl_screenCtx->x1Clip, fgl_screenCtx->y1Clip);

    sid = getShmem( sizeof( fgl_PServer_t), (void**)&(PServer));

    if ( sid > -1 ) {
      registerHandle( sid);
      PServer->ctx = fgl_screenCtx;
    }
  }

  PServer->cursor = 0;
  if ( fgl_cursor )
    {
      if ( fgl_cursor->isVisible )
         {
            _fgl_hideCursor();
            fgl_PServerUpdateCursor( fgl_cursor );
            _fgl_showCursor();
         }
       else
         {
            fgl_PServerUpdateCursor( fgl_cursor );
         }
    }

  return PServer;
}

void
fgl_PServerUpdateCursor( fgl_sprite_t *sp ) {
    int size;
    if ( sp && ( size = _fgl_getSpriteSize( sp ))) {
        PServer->cursor = size;
        _fgl_copySprite( sp, (void *)((char *)PServer + sizeof(fgl_PServer_t)));
        fgl_cursor = (fgl_sprite_t *)((char *)PServer + sizeof(fgl_PServer_t));
    }
}

/*** return first disconnected client entry ***/
fgl_PClient_t*
fgl_PServerOpenConnection() {
  int ci;

  if ( PServer == NULL ) {
    return NULL;
  }

  for ( ci=0; ci<MAX_PSRV_CLIENTS; ci++ ) {
    if ( PServer->clients[ci].states[PCS_CONNECTED] == 0 ) {
      DBG( "PServerOpen() %d\n", ci);
      return (&(PServer->clients[ci]));
    }
  }

  FATAL("PServerOpen() failed\n");
  return NULL;
}


/*******************************
 * client API
 *******************************/
fgl_PClient_t*
fgl_PClientConnect( u16 *pWidth, u16 *pHeight, int *depth) {
  int ci, sz;
  fgl_PServer_t *pSrv;
  fgl_PClient_t *pCli;

  pSrv = initClient();

  if ( pSrv == NULL ) {
    return NULL;
  }

  for ( ci=0; ci<MAX_PSRV_CLIENTS; ci++ ) {
    pCli = &(pSrv->clients[ci]);
    if ( pCli->states[PCS_CONNECTED] == 0) {
      pCli->srv_addr = pSrv;
      pCli->states[PCS_CONNECTED] = 1;
      DBG( "PClientConnect: %d %p\n", ci, pSrv);

      if ( *pWidth == 0 ){
	*pWidth = pCli->srv_bounds.width;
      }
      if ( *pHeight == 0 ) {
	*pHeight = pCli->srv_bounds.height;
      }

      sz = sizeof( PClientRgn_t) + (*pWidth * *pHeight * fgl_getSlcSize());
      getShmem( sz, (void**)&(pCli->crgn) );
      pCli->pages = (sz + 4095) / 4096;
      
      pCli->src_width = *pWidth;
      pCli->src_height = *pHeight;

      pCli->states[PCS_SUSPEND] = 0;

      pCli->crgn->nClips = 0;

      pCli->crgn->bmp.surface = (fgl_slc_t*)((u8*)pCli->crgn + sizeof( PClientRgn_t));
      pCli->crgn->bmp.width = *pWidth;
      pCli->crgn->bmp.height = *pHeight;
      pCli->crgn->bmp.lastOperation = FGL_FRAMEBUFFER_OP;
      pCli->crgn->bmp.nextOperation = FGL_FRAMEBUFFER_OP;
      pCli->crgn->bmp.state = 0;
      if ( pSrv->cursor )
        {
          fgl_cursor = ( fgl_sprite_t *)((char *)pSrv + sizeof( fgl_PServer_t ));
        }
      //ROT
      pCli->crgn->bmp.sLength = fgl_getScanlineLength( *pWidth);

      *depth = fgl_getBitsPerPixel();
      
      return pCli;
    }
  }

  shmdt( pSrv);

  return NULL;
}

void
fgl_PClientUpdateSize( fgl_PClient_t *pCli, u16 width, u16 height) {
  int sz, pg;

  sz = sizeof( PClientRgn_t) + (width * height * fgl_getSlcSize());
  pg = (sz + 4095) / 4096;

  if ( pg > pCli->pages ) {
    shmdt( pCli->crgn);
    getShmem( sz, (void**)&(pCli->crgn) );

    pCli->pages = pg;
    pCli->crgn->bmp.surface = (fgl_slc_t*)((u8*)pCli->crgn + 
					   sizeof( PClientRgn_t));
  }
      
  pCli->crgn->nClips = 0;

  pCli->crgn->bmp.width = width;
  pCli->crgn->bmp.height = height;
  pCli->crgn->bmp.lastOperation = FGL_FRAMEBUFFER_OP;
  pCli->crgn->bmp.nextOperation = FGL_FRAMEBUFFER_OP;
  pCli->crgn->bmp.state = 0;

  //ROT
  pCli->crgn->bmp.sLength = fgl_getScanlineLength( width);

}

void
fgl_PClientDisconnect( fgl_PClient_t* pCli) {
  void *addr;
  DBG( "PClient disconnect: %p %p\n", pCli, pCli->srv_addr);

  if ( pCli->srv_addr ) {
    pCli->states[PCS_CONNECTED] = 0;
    if ( pCli->crgn ) {
      addr = pCli->crgn;
      pCli->crgn = NULL;
      shmdt( addr);
    }
      
    addr = pCli->srv_addr;
    pCli->srv_addr = NULL;
    shmdt( addr);
  }

  fgl_terminate();
}

void fgl_PClientAddClipRegion( fgl_PClient_t* pCli, u16 x, u16 y, 
			       u16 width, u16 height){

  PClientRect_t *pCr = &(pCli->crgn->clip[pCli->crgn->nClips]);

  pCr->x = x;
  pCr->y = y;
  pCr->width = width;
  pCr->height = height;

  pCli->crgn->nClips++;
}

int 
fgl_PClientFlushFrame( fgl_PClient_t* pCli){
  int i, x, y, w, h;
  int cx0, cy0, cx1, cy1;
  fgl_PServer_t *pSrv = (fgl_PServer_t*)pCli->srv_addr;
  
  //client not initialized
  if ( ! pSrv ) {
    return 0;
  }

  //suspended by Server
  if ( pCli->states[PCS_SUSPEND] ) {
    return 0;
  }

  //not yet initialized by Server
  if ( ! pCli->srv_bounds.width || ! pCli->srv_bounds.height ) {
    return 0;
  }

  //clip to update rects
  if ( pCli->crgn->nClips > 0) {
    cx0 = pCli->crgn->bmp.width;
    cy0 = pCli->crgn->bmp.height;
    cx1 = 0;
    cy1 = 0;
    for ( i=0; i<pCli->crgn->nClips; i++ ) {
      x = pCli->crgn->clip[i].x;
      if ( x < cx0 ) cx0 = x;
      y = pCli->crgn->clip[i].y;
      if ( y < cy0 ) cy0 = y;
      w = pCli->crgn->clip[i].width;
      if ( x+w > cx1 ) cx1 = x+w;
      h = pCli->crgn->clip[i].height;
      if ( y+h > cy1 ) cy1 = y+h;
    }
  }
  else {
    cx0 = 0;
    cy0 = 0;
    cx1 = pCli->crgn->bmp.width;
    cy1 = pCli->crgn->bmp.height;
  }

  //center to canvas
  x = pCli->srv_bounds.x + 
    ( pCli->srv_bounds.width- pCli->src_width ) / 2;
  y = pCli->srv_bounds.y + 
    ( pCli->srv_bounds.height- pCli->src_height ) / 2;

  fgl_PClientSetValue( pCli, PCS_FLUSHING, 1);

  fgl_bitBlt( fgl_screenCtx,
	      x+cx0, y+cy0,
	      &(pCli->crgn->bmp),
	      cx0, cy0, cx1, cy1);

  fgl_PClientSetValue( pCli, PCS_FLUSHING, 0);

  pCli->crgn->nClips = 0;

  return 1;
}

void
fgl_PClientSetBounds ( fgl_PClient_t* pCli, u16 x, u16 y, u16 w, u16 h) {
  pCli->srv_bounds.x = x;
  pCli->srv_bounds.y = y;
  pCli->srv_bounds.width = w;
  pCli->srv_bounds.height = h;

  pCli->modifiers[PCS_BOUNDS]= 1;
}

void
fgl_PClientHandleEvent ( fgl_PClient_t* pCli, u16 event, u16 p1, u16 p2) {
  pCli->event.event = event;
  pCli->event.p1 = p1;
  pCli->event.p2 = p2;

  pCli->modifiers[PCS_EVENT]= 1;
}

void
fgl_PClientSetValue ( fgl_PClient_t* pCli, u8 key, u16 val) {
  pCli->states[key]= val;
  pCli->modifiers[key] = 1;
}

u8
fgl_PClientCheckValue ( fgl_PClient_t* pCli, u8 key, u16 *pval){
  u8 modified = pCli->modifiers[key];

  if ( pval ) {
    *pval = pCli->states[key];
    pCli->modifiers[key] = 0;
  }

  return modified;
}
