/*
 * rect.h - fgl module to draw and fill rectangles
 * 
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 01/01/00  Peter Mehlitz
 */

#ifndef FGL_RECT_H_INCLUDED
#define FGL_RECT_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "draw.h"
#include "rot.h"
#include "util.h"

static inline
void _fgl_drawRect ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1 ) 
{

  if ( x0 > x1 ) swap32( &x0, &x1);
  if ( y0 > y1 ) swap32( &y0, &y1);

  fgl_drawHLine( ctx, x0+1, y0, x1-1);
  fgl_drawHLine( ctx, x0+1, y1, x1-1);
  fgl_drawVLine( ctx, x0, y0, y1);
  fgl_drawVLine( ctx, x1, y0, y1);
}

static inline
void _fgl_fillRectClr ( fgl_context_t *ctx, i32 x0, i32 y0, i32 x1, i32 y1, fgl_color_t clr ) 
{
  int w, h;
  fgl_slc_t *p;

  if ( x0 > x1 )		swap32( &x0, &x1);
  if ( y0 > y1 )		swap32( &y0, &y1);

  if ( (x1 < ctx->x0Clip) || (x0 > ctx->x1Clip) ||
      (y1 < ctx->y0Clip) || (y0 > ctx->y1Clip) ){
      return;
  }

  if ( x0 < ctx->x0Clip ) x0 = ctx->x0Clip;
  if ( x1 > ctx->x1Clip ) x1 = ctx->x1Clip;
  if ( y0 < ctx->y0Clip ) y0 = ctx->y0Clip;
  if ( y1 > ctx->y1Clip ) y1 = ctx->y1Clip;

    {
      SET_ROTATION( ctx->bmp );
      RECT_ROTATE( rot, ctx->bmp, x0, y0, x1, y1 );
    }

#ifdef USE_ACCEL 
  if ( (fgl_accel_fillRect) && (ctx->bmp->state & IMAGE_ON_SCREEN)) {
      fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_ACCEL_FILLRECT_OP );
      fgl_accel_fillRect( fgl_screenCtx, x0, y0, x1, y1, clr );
      fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
      return ;
  }
#endif /* USE_ACCEL */
  fgl_beginDraw( ctx->bmp, x0, y0, x1, y1, FGL_FRAMEBUFFER_OP );

  p = _fgl_getSlc( ctx->bmp, x0, y0);
  w = x1 - x0+1;
  h = y1 - y0;

  do {
      _fgl_setSlcCols( p, x0, w, clr);
      _fgl_addSlc( &p, x0, ctx->bmp->sLength);
  } while ( h-- );

  fgl_endDraw( ctx->bmp, x0, y0, x1, y1);
}

#endif /* FGL_RECT_H_INCLUDED */
