/*
 * rot.c rotation macros
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 07/16/01  Damian Kowalewski
 */

#include "rot.h"
#include "util.h"

void fgl_setScreenRotation( u8 nrot )
{
    {
      SET_ROTATION( fgl_screenBmp );
      BMP_SET_ROTATION( fgl_screenBmp, nrot );

      if ( ((4 + nrot - rot)%4) & 1 ) 
        {
          swap32( &(fgl_screenBmp->width), &(fgl_screenBmp->height));
          /* we should also set new clip rectangle points */
          fgl_screenCtx->x1Clip  = fgl_screenBmp->width-1;
          fgl_screenCtx->y1Clip  = fgl_screenBmp->height-1;
          fgl_screenCtx->xoff    = 0;
          fgl_screenCtx->yoff    = 0;
        }
    }
}

int fgl_getScreenRotation()
{
   SET_ROTATION( fgl_screenBmp );
   return rot;
}

void fgl_rotateMFont( fgl_font_t *fnt )
{
#if defined(DYNAMIC)

  int i,sl,dl,p;
  u8 brot,srot;
  u8 rotation;
  u8 *surface, *sf, *df;
  u8 pix;
  u8 *slength1, *slength2, *slength3;
  u32 u,v,x,y;

  int size = 0;
  u32 *newSurfOff;

  /* we can always assume that the image is rotated from its current 
   * to the screen orientation
   */
  srot = fgl_screenBmp->state & 0x3;
  brot = fnt->state & 0x3;

  if ( srot == brot ) return; /* if nothing ton do do not waste time */

  rotation = ( 4 + srot - brot ) % 4; /* calculate the counter-clockwise */

  slength1 = ( brot & 1) ? fnt->height : fnt->width ; 
  slength3 = ( brot & 1) ? fnt->width : fnt->height ; 

  /* calculate new surface offsets and calcualte its new size */
  newSurfOff = (u32 *) FGL_MALLOC( sizeof(u32) * fnt->nGlyphs );
  if ( srot & 1 )
    {
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          newSurfOff[i] = size;
          size += fnt->width[i] * MONO_SLENGTH( fnt->height[i] );
        }
      fnt->sLength = slength2 = fnt->height;
    }
  else
    {
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          newSurfOff[i] = size;
          size += fnt->height[i] * MONO_SLENGTH( fnt->width[i] );
        }
      fnt->sLength = slength2 = fnt->width;
    }
  surface = (u8 *) FGL_MALLOC ( size );
  switch( rotation ) {
    case 1:
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          sl = MONO_SLENGTH( slength1[i] );
          dl = MONO_SLENGTH( slength2[i] );
          sf = fnt->surface + fnt->surfOff[i];
          df = surface + newSurfOff[i];
          for( x = 0 ; x < slength1[i] ; x++ )
            {
              for ( y = 0 ; y < slength3[i] ; y++ )
                {
                  pix = sf[ y *  sl + x / 8 ] & _monoBits[ x % 8 ] ;
                  u = slength3[i] - y - 1;
                  v = x;
                  p = v * dl + u / 8;
                  if ( pix ) 
                    {
                      df[p] |= _monoBits[ u % 8];
                    }
                  else
                    {
                      df[p] &= ~_monoBits[ u % 8 ];
                    }
                }
            }
        }
      break;
    case 2:
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          sl = MONO_SLENGTH( slength1[i] );
          dl = MONO_SLENGTH( slength2[i] );
          sf = fnt->surface + fnt->surfOff[i];
          df = surface + newSurfOff[i];
          for( x = 0 ; x < slength1[i] ; x++ )
            {
              for ( y = 0 ; y < slength3[i] ; y++ )
                {
                  pix = sf[ y *  sl + x / 8 ] & _monoBits[ x % 8 ] ;
                  u = slength1[i] - x - 1;
                  v = slength3[i] - y - 1;
                  p = v * dl + u / 8;
                  if ( pix ) 
                    {
                      df[p] |= _monoBits[ u % 8];
                    }
                  else
                    {
                      df[p] &= ~_monoBits[ u % 8 ];
                    }
                }
            }
        }
      break;
    case 3:
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          sl = MONO_SLENGTH( slength1[i] );
          dl = MONO_SLENGTH( slength2[i] );
          sf = fnt->surface + fnt->surfOff[i];
          df = surface + newSurfOff[i];
          for( x = 0 ; x < slength1[i] ; x++ )
            {
              for ( y = 0 ; y < slength3[i] ; y++ )
                {
                  pix = sf[ y *  sl + x / 8 ] & _monoBits[ x % 8 ] ;
                  u = y;
                  v = slength1[i] - x - 1;
                  p = v * dl + u / 8;
                  if ( pix ) 
                    {
                      df[p] |= _monoBits[ u % 8];
                    }
                  else
                    {
                      df[p] &= ~_monoBits[ u % 8 ];
                    }
                }
            }
        }
      break;
    default:
      return;
  }

  if ( fnt->state & IS_ROTATED)
    {
      FGL_FREE( fnt->surface);
    }
  
  if ( (fnt->state & IS_ROTATED) || (fnt->state & 0x100) ) 
    {
      FGL_FREE( fnt->surfOff );
    }

  fnt->surface = surface;
  fnt->surfOff = newSurfOff;

  fnt->state |= IS_ROTATED;

  BMP_SET_ROTATION( fnt, srot );
  
#endif /* DYNAMIC */
}

void fgl_rotateA8Font( fgl_font_t *fnt )
{
#if defined(DYNAMIC)

  int i;
  u8 brot,srot;
  u8 rotation;
  u8 *surface, *new_surface, *old_surface;
  u8 *slength1, *slength3;
  u32 w,h,x,y;

  int size = 0;

  /* we can always assume that the image is rotated from its current 
   * to the screen orientation
   */
  srot = fgl_screenBmp->state & 0x3;
  brot = fnt->state & 0x3;
        
  if ( srot == brot ) return; /* if nothing ton do do not waste time */

  rotation = ( 4 + srot - brot ) % 4; /* calculate the counter-clockwise */

  slength1 = ( brot & 1) ? fnt->height : fnt->width ; 
  slength3 = ( brot & 1) ? fnt->width : fnt->height ; 

  for (i = 0; i < fnt->nGlyphs ; i++ )
    {
      size += fnt->width[i] *  fnt->height[i];
    }

  if ( srot & 1 )
    {
      fnt->sLength = fnt->height;
    }
  else
    {
      fnt->sLength = fnt->width;
    }

  surface = (u8 *) FGL_MALLOC ( size );

  switch( rotation ) {
    case 1:
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          w = slength1[i];
          h = slength3[i];
          old_surface = fnt->surface + fnt->surfOff[i];
          new_surface = surface + fnt->surfOff[i];
          for( x = 0 ; x < w ; x++ ) {
              for ( y = 0 ; y < h ; y++ ) {
                  new_surface[ x * h + h - y - 1 ] = old_surface[ x + y * w ];
              }
          }
        }
      break;
    case 2:
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          w = slength1[i];
          h = slength3[i];
          old_surface = fnt->surface + fnt->surfOff[i];
          new_surface = surface + fnt->surfOff[i];
          for( x = 0 ; x < w ; x++ ) {
              for ( y = 0 ; y < h ; y++ ) {
                  new_surface[y*w+x] = old_surface[(h-y-1)*w+(w-x-1)];
              }
          }
        }
      break;
    case 3:
      for (i = 0; i < fnt->nGlyphs ; i++ )
        {
          w = slength1[i];
          h = slength3[i];
          old_surface = fnt->surface + fnt->surfOff[i];
          new_surface = surface + fnt->surfOff[i];
          for( x = 0 ; x < w ; x++ ) {
              for ( y = 0 ; y < h ; y++ ) {
                  new_surface[ h * ( w - x - 1 ) + y] = old_surface[ x + y * w ];
              }
          }
        }
      break;
    default:
      return;
  }

  if ( fnt->state & IS_ROTATED)
    {
      FGL_FREE( fnt->surface);
    }
  
  fnt->surface = surface;

  fnt->state |= IS_ROTATED;

  BMP_SET_ROTATION( fnt, srot );
  
#endif /* DYNAMIC */
}

