/*
 * rot.h rotation macros
 *
 * Copyright (c) 2000, 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 07/16/01  Damian Kowalewski
 */

#ifndef ROT_H_INCLUDED
#define ROT_H_INCLUDED

/* general includes */
#include "libfgl.h"
#include <fgl.h>

/* module specific includes */
#include "cntx.h"
/*#include "draw.h"*/
#include "heap_posix.h"


#define BMP_SET_ROTATION( _bmp, _rot ) \
        (_bmp)->state &= 0xffffffffc; \
        (_bmp)->state |= _rot;

#if defined(DYNAMIC) 
#define SET_ROTATION( _bmp ) int rot = _bmp->state & 0x3;
#else
#define SET_ROTATION( _bmp ) 
#endif

static inline
void rotate90 ( fgl_bitmap_t* bmp, u32* x, u32* y )
{
  u32 tmp  = *x;
  *x = bmp->height - *y - 1;
  *y = tmp;
}

static inline
void rotate180 ( fgl_bitmap_t* bmp, u32* x, u32* y )
{
  *x = bmp->width - *x - 1;
  *y = bmp->height - *y - 1;
  
}

static inline
void rotate270 ( fgl_bitmap_t* bmp, u32* x, u32* y )
{
  u32 tmp  = *x;
  *x = *y;
  *y = bmp->width - tmp - 1;
}


#define POINT_ROTATE( _r, _bmp, _x, _y ) \
switch( _r ) { \
  case 1: \
    rotate90( _bmp, &_x, &_y ); \
    break; \
  case 2: \
    rotate180( _bmp, &_x, &_y ); \
    break; \
  case 3: \
    rotate270( _bmp, &_x, &_y ); \
    break; \
  default: \
} 

#define LINE_ROTATE( _r, _bmp, _x0, _y0, _x1, _y1 ) \
switch( _r ) { \
  case 1: \
    rotate90( _bmp, &_x0, &_y0 ); \
    rotate90( _bmp, &_x1, &_y1 ); \
    break; \
  case 2: \
    rotate180( _bmp, &_x0, &_y0 ); \
    rotate180( _bmp, &_x1, &_y1 ); \
    break; \
  case 3: \
    rotate270( _bmp, &_x0, &_y0 ); \
    rotate270( _bmp, &_x1, &_y1 ); \
    break; \
  default: \
} 

#define RECT_ROTATE( _r, _bmp, _x0, _y0, _x1, _y1 ) \
switch( _r ) { \
  case 1: \
    rotate90( _bmp, &_x0, &_y0 ); \
    rotate90( _bmp, &_x1, &_y1 ); \
    swap32( &_x0, &_x1); \
    break; \
  case 2: \
    rotate180( _bmp, &_x0, &_y0 ); \
    rotate180( _bmp, &_x1, &_y1 ); \
    swap32( &_x0, &_x1 ); \
    swap32( &_y0, &_y1 ); \
    break; \
  case 3: \
    rotate270( _bmp, &_x0, &_y0 ); \
    rotate270( _bmp, &_x1, &_y1 ); \
    swap32( &_y0, &_y1 ); \
    break; \
  default: \
} 
/*
#include "cbmp.h"
#include "mbmp.h"
#include "a2bmp.h"
#include "a4bmp.h"
*/
#endif /* ROT_H_INCLUDED */
