/*
 * screencfg.c - screen config file reader
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "COPYING" for information on usage and redistribution 
 * of this file. 
 * 
 * 04/16/01 Damian Kowalewski
 */

#ifndef FGL_SCREENCFG_INCLUDED
#define FGL_SCREENCFG_INCLUDED

#include "libfgl.h"
#include "../include/fgl.h"

#define DEFAULT_SC_CONFIG "/etc/scconfig"
#define DEFAULT_FB     "/dev/fb0"
#define DEFAULT_VT     "/dev/tty0"

int
fgl_readConfig ( fgl_config_t* cfg )
{
	FILE* fp;
	int n;
	int val;
	char *c, key[32];
	char buf[MAX_PATH];
	char *cfgFile;

	cfg->xoff     = -1;
	cfg->yoff     = -1;
	cfg->width    = -1;
	cfg->height   = -1;

        buf[0] = '\0';
        if ( (cfgFile = getenv( "SC_CONFIG")) ){
                strncat( buf, cfgFile, MAX_PATH - 1);
        }
        else {
                strncat( buf, DEFAULT_SC_CONFIG, MAX_PATH -1);
        }

	if ( (fp = fopen( buf, "r")) ) {
		do {
			c = key;
			n = fscanf( fp, "%s %d", (char*)&key, &val);
			if ( n < 0 ) break;
			
			if ( n && key[0] == '#' ){
				while ( ((val = fgetc( fp)) != EOF) && (val != '\n') );
				continue;
			}
			
			if ( n == 2 ){
			    *c |= 0x20;
				switch ( *c ){
				case 'x':
					cfg->xoff = val;
					break;

				case 'y':
					cfg->yoff = val;
					break;

				case 'w':
					cfg->width = val;
					break;

				case 'h':
					cfg->height = val;
					break;

				}
			}
		} while ( (*c == '#') || (n == 2) );

		fclose( fp);
		return 1;
	}
	
	return 0;
}

#endif /* FGL_SCREENCFG_INCLUDED */
