package com.pocketlinux.awt.theme;

import java.awt.*;
import java.util.*;
import kaffe.awt.AWT;
import kaffe.awt.theme.*;

public class StyleComponentDeco
  extends ComponentDeco
{
public Style style;
String widgetname;
String[] filenames;
Image[] images;
Hashtable fhash;

public StyleComponentDeco(Style style, String widgetname) {
	super(null, null, null);
	insets = new Insets(0, 0, 0, 0);
	this.style = style;
	this.widgetname = widgetname;
}

public StyleComponentDeco(Style style, String widgetname, String[] filenames) {
	super(null, null, null);
	insets = new Insets(0, 0, 0, 0);

	this.style = style;
	this.widgetname = widgetname;
	this.filenames = filenames;
	images = new Image[filenames.length];
}

protected boolean doBorder(Component c, Component p, int x, int y, int w, int h) {
	// Because NativeGraphics paintChild() clips to parent
	// the border has to be drawn even if only partly visible
	if ((AWT.extension.getDeco(c) == this) && (w > 0) && (h > 0) ) {
		return (true);
	}
	else {
		return (false);
	}
}

public Font getFont(String modifier) {
        int sz = 0;
        int fstyle = style.fstyle;

	// Check cache first to avoid recreating font
	Font fnt = null;
	if (fhash != null) {
		fnt = (Font)fhash.get(modifier);
		if (fnt != null) {
			return (fnt);
		}
	}

	//format: [style][+/-value]
	int idx = modifier.indexOf('-');   
	if (idx == -1) {
		idx = modifier.indexOf('+');
	}
	if (idx == -1) {
		idx = modifier.length();
	}
	String sstyle = modifier.substring(0, idx);
	if (sstyle.equalsIgnoreCase("plain")) {
		fstyle = Font.PLAIN;
	}
	else if (sstyle.equalsIgnoreCase("bold")) {
		fstyle = Font.BOLD;
	}
	else if (sstyle.equalsIgnoreCase("italic")) {
		fstyle = Font.ITALIC;
	}
	else if (sstyle.equalsIgnoreCase("bolditalic")) {
		fstyle = Font.BOLD|Font.ITALIC;
	}

	if (idx < modifier.length()) {
		char fchar = modifier.charAt(idx);
		try {
			sz = Integer.parseInt(modifier.substring(idx+1));
			if (fchar == '-') {
				sz = -sz;
			}
		}
		catch (NumberFormatException _) {
		}
	}

	fnt = new Font(style.fname, fstyle, sz + style.fsize);
	if (fhash == null) {
		fhash = new Hashtable(4);
	}
	fhash.put(modifier, fnt);
	return (fnt);
}

public void reload() {
	int len = filenames.length;

	for (int i = 0; i < len; i++) {
		String wname = '/' + widgetname + filenames[i];
		Image img = style.theme.getImageNoWait(style.stylename + wname);
		if (img == null) {
			img = style.theme.getImageNoWait("default" + wname);
		}
		if (images[i] != null) {
			images[i].flush();
		}
		images[i] = img;
		
	}

	style.theme.waitForAll();

	for (int i = 0; i < len; i++) {
		if (images[i] != null) {
			style.theme.removeImage(images[i]);
		}
	}

	// Check the images loaded, and if not use the default one
	for (int i = 0; i < len; i++) {
		Image img = images[i];
		if (img == null || style.theme.toolkit.prepareImage(img, -1, -1, null) == false) {

			img = style.theme.getIcon(null);
			if (images[i] != null) {
				images[i].flush();
			}
			images[i] = img;
//System.out.println("Missing: " + style.stylename + ":" + widgetname + ":" + filenames[i]);
		}
	}

	// Clear font cache
	fhash = null;
}

public String toString() {
	return ("[Deco: " + style.stylename + ":" + widgetname + "]");
}
}
