package com.pocketlinux.awt.theme;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;

import kaffe.awt.theme.*;
import kaffe.awt.widgets.*;
import kaffe.awt.*;
import kaffe.io.AccessibleBAIStream;
import kaffe.net.www.protocol.zip.ZipURLConnection;

import org.w3c.dom.*;

import org.xoe.core.Constants;
import org.xoe.core.Engine;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XMLSyntaxException;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;


public class Theme
  extends kaffe.awt.theme.Theme
  implements Constants
{
	public static final String ALIAS_CURRENT_THEME = "stash:/themes/current";

	static final private String THEME_BASE = "stash:/themes/";
	static Toolkit toolkit = Toolkit.getDefaultToolkit();

	MediaTracker tracker = new MediaTracker(null);
        private String defaultThemeName;
        private String currentThemeName;
	private ZipFile defaultThemeZip;
	private ZipFile currentThemeZip;
	private Style defaultStyle;
	private Style windowStyle;
	private Style headerStyle;
	private Style displayStyle;
	private Style consoleStyle;
	private Style nodrawStyle;
	private StyleComponentDeco title;
	private StyleComponentDeco bar;
	private Hashtable iconcache = new Hashtable();
	private SimpleFrameDeco frameDeco = new SimpleFrameDeco();
	private SimpleCheckboxDeco checkboxDeco = new SimpleCheckboxDeco();

public Theme() {
	defaultStyle = new Style(this, "default");
	windowStyle = new Style(this, "window");
	headerStyle = new Style(this, "header");
	displayStyle = new Style(this, "display");
	consoleStyle = new Style(this, "console");
	nodrawStyle = new Style(this);

	title = new Deco2I(defaultStyle, "title");
	bar = new Deco3I(defaultStyle, "menubar");

	// HACK - adopt the window scrollbars for default
	defaultStyle.sbh = windowStyle.sbh;
	defaultStyle.sbv = windowStyle.sbv;
	// HACK

	// Set the default theme
	defaultThemeName = THEME_BASE + "Default-COLOR.zip";
	defaultThemeZip = getZipTheme(defaultThemeName);

	loadTheme(THEME_BASE + Engine.getConfig( THEME_PROPERTY, DEFAULT_THEME) + ".zip");
}

/**
 * For the given component 'c' which is class is called 'clazz' and which
 * has the style 'option' (currently vert and horz for scrollbars), return
 * the decoration object which knows how to render it.
 */
public StyleComponentDeco getDecoWithStyle(Component c, String style, String option) {
	Style s = defaultStyle;
	if (style != null) {
		s = getStyleByName(style);
	}

	String clazz = c.getClass().getName();

	if ("org.xoe.display.dom.xhtml.Table$Render".equals(clazz) ||
	    "org.xoe.display.dom.xhtml.TableCell$Render".equals(clazz) ||
	    "org.xoe.display.dom.xhtml.DocFrameSet$Render".equals(clazz)) {
		if (style == null) {
			return (s.tran);
		}
		else {
			return (s.win);
		}
	}

	else if (c instanceof java.awt.Button) {
		return (s.btn);
	}

	else if (c instanceof java.awt.Scrollbar) {
                if ("horz".equals(option)) {
                        return (s.sbh);
                }
                else {
                        return (s.sbv);
                }
	}
	else if (c instanceof java.awt.Checkbox) {
		return (null); // HACK - not styled yet
	}
	else if (c instanceof java.awt.Choice) {
		return (s.ch);
	}
	else if (clazz.equals("java.awt.Choice$ChoiceWindow")) {
		// No scrollbars or nice window borders in the default scheme
		// so if they get used we switch to the window style
		if (s == defaultStyle) {
			s = windowStyle;
		}
		return (s.pull);
	}

	else if ("org.xoe.display.dom.xhtml.Input$BorderTextField".equals(clazz) ||
	    "org.xoe.display.dom.xhtml.TextArea$BorderTextArea".equals(clazz)) {
		return (s.border);
	}

	else if ("org.xoe.display.dom.Render".equals(clazz) || "org.xoe.display.dom.Render".equals(c.getClass ().getSuperclass ().getName ())) {
		return (defaultStyle.win);
	}

	return (s.tran);
}

public ComponentDeco getDeco( Component c, String clazz, String option) {
	if (c instanceof java.awt.Frame) {
		return (frameDeco);
	}
	else if (c instanceof java.awt.Checkbox) {
		return (checkboxDeco);
	}
	else {
		return (getDecoWithStyle(c, "default", option));
	}
}

ComponentDeco getDecoByStyle( String style) {
	ComponentDeco d = SpecialDecoFactory.getDeco( style);
	if ( d == null ) {
		d = getStyleByName( style).win;
	}
	return (d);
}

/**
 * Use the name to locate a matching icon.  We look in the current theme
 * for this icon and, if that fails, look in a default location.  If that
 * fails we return a 'broken' icon.
 */
public Image getIcon(String iconname) {
	if (iconname == null) {
		iconname = UNKNOWN_ICON;
	}
	// First look in the cache.
	Object o = iconcache.get(iconname);
	if (o != null) {
		return ((Image)o);
	}

	// Look got icon - if we fail use the question icon
	Image i = getImageNoWait("icons/" + iconname + ".gif" );
	if (i == null) {
		i = getImageNoWait("icons/" + UNKNOWN_ICON + ".gif");
	}
	waitForAll();
	removeImage(i);

	// Cache it (if we found it)
	if (i != null) {
		iconcache.put(iconname, i);
	}
	return (i);
}

// ------------------------------------------------------------------------

/**
 * Get a single image using the media tracker.
 */
public Image getImage(String name) {
	Image i = getImageNoWait(name);
	waitForAll();
	removeImage(i);
	return (i);
}

Image getImageNoWait(String name) {
	try {
		ZipEntry ze = null;
		ZipFile zf = currentThemeZip;
		if (zf != null) {
			ze = zf.getEntry(name);
		}
		if (ze == null && defaultThemeZip != null) {
			zf = defaultThemeZip;
			ze = zf.getEntry(name);
		}
		if (ze != null) {
			byte[] data = ((AccessibleBAIStream)zf.getInputStream(ze)).getByteArray();
			Image img = toolkit.createImage(data);
			tracker.addImage(img, 0);
			return (img);
		}
	}
	catch (IOException _) {
	}
	return (null);
}

Style getStyleByName( String style) {
	if (style.equalsIgnoreCase("window")) {
		return (windowStyle);
	}
	else if (style.equalsIgnoreCase("header")) {
		return (headerStyle);
	}
	else if (style.equalsIgnoreCase("display")) {
		return (displayStyle);
	}
	else if (style.equalsIgnoreCase("console")) {
		return (consoleStyle);
	}
	else {
		return (defaultStyle);
	}
}

public static Theme getTheme() {
	return ((Theme)AWT.getTheme());
}

private static ZipFile getZipTheme(String themebase) {
	try {
		URL u = new URL("zip:"+themebase+"!/");
		ZipURLConnection zc = (ZipURLConnection)u.openConnection();
		return (zc.getZipFile());
	}
	catch (IOException _) {
		return (null);
	}
}

public void loadTheme() throws XMLSyntaxException {
	if (currentThemeZip != null ) {
		// Flush any cached icons
		iconcache.clear();

		bar.reload();
		title.reload();

		defaultStyle.reload();
		windowStyle.reload();
		headerStyle.reload();
		displayStyle.reload();
		consoleStyle.reload();
		nodrawStyle.reload();

		parseXML();

		defaultStyle.finish();
		windowStyle.finish();
		headerStyle.finish();
		displayStyle.finish();
		consoleStyle.finish();
		nodrawStyle.finish();

		bar.reload();
		title.reload();

	}
}

// ----------------------------------------------------------------------
public boolean loadTheme(String themebase) {
	currentThemeZip = getZipTheme(themebase);
        try
          {
            loadTheme ();
	    currentThemeName = themebase;
          }
        catch (XMLSyntaxException xse)
          {
            Logger.log (this, "Error in parsing theme config file. Switching to default theme.");
            Logger.log (this, xse);
	    currentThemeZip = defaultThemeZip;
            try
              {
		loadTheme ();
		currentThemeName = defaultThemeName;
              }
	    catch (XMLSyntaxException _)
	      {
		Logger.log (this, "Error in parsing default theme config file");
		Logger.log (this, _);
	      }
          }
	try
	  {
	    Stash.putAlias(ALIAS_CURRENT_THEME, currentThemeName);
	  }
	catch (IOException _)
	  {
	  }
	return (true);
}

private void parseXML() throws XMLSyntaxException {
	Document doc;
	ZipFile keyboardThemeZip;
	
	ZipEntry ze = currentThemeZip.getEntry("package.xml");
	if (ze == null) {
	  return;
	}
	try {
	  doc = DOMlet.create(currentThemeZip.getInputStream(ze), "Themes");
	  if (doc == null) {
	    return;
	  }
	}
	catch (IOException _) {
	  return;
	}

	NodeList list = XPath.findNodes (doc, "//styles/style");
	for (int i = list.getLength() - 1; i >= 0; i--) {
		Element elem = (Element)list.item(i);
		Style cs = getStyleByName(elem.getAttribute("name"));
		for (Node n = elem.getFirstChild(); n != null ; n = n.getNextSibling()) 
                  {
			cs.setProperty (n.getNodeName(), DOMUtils.getTextContent (n));
		  }
        }

	// try to determine the most suitable keyboard size from available sizes
	// read in the list of keyboards
	list = XPath.findNodes (doc, "//keyboard");
	
	if ( list.getLength() != 0 ) {
		keyboardThemeZip = currentThemeZip;

	}
	else {
		ze = defaultThemeZip.getEntry("package.xml");
                if (ze == null) {
                        return;
                }
                try {
                        doc = DOMlet.create(defaultThemeZip.getInputStream(ze), "Themes");
                        if (doc == null) {
                                return;
                        }
                }
                catch (IOException _) {
                        return;
                }
        }
}

void removeImage(Image img) {
	tracker.removeImage(img);
}

void waitForAll() {
	try {
		tracker.waitForAll();
	}
	catch (InterruptedException _) {
	}
}
}
