/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Constants.java,v 1.14 2001/12/18 06:05:50 colin Exp $
 */
package org.xoe.core;


/**
 * Core Xoe Constants.
 */
public interface Constants
{
  /**
   * XML constants
   */
  public static final String XMLNS       = "xmlns";
  public static final String XMLNS_NAMESPACE = "http://www.w3.org/2000/xmlns/";
  public static final String XML = "xml";
  public static final String XML_PREFIX_NAMESPACE = "http://www.w3.org/XML/1998/namespace";

  /**
   *  default XOE namespace string
   */
  public static final String DEFAULT_NAMESPACE = "http://xoe.org/ns/2001";

  /**
   *  xhtml namespace string
   */
  public static final String XHTML_NAMESPACE = "http://www.w3.org/1999/xhtml";

  /**
   *  wml namespace string
   */
  public static final String WML_NAMESPACE = "http://www.wapforum.org/DTD/wml_1.1.xml";

  /**
   *  The system property which specifies the stash root
   */
  public static final String STASH_DIR_PROPERTY = "stashRoot";

  /**
   *  The system property which specifies the user's home directory
   */
  public static final String HOMEDIR_PROPERTY = "user.home";

  /**
   *  The system property which specifies the user's current directory
   */
  public static final String USERDIR_PROPERTY = "user.dir";

  /**
   *  The stash protocol handle
   */
  public static final String STASH_PROTOCOL = "stash";

  /**
   *  The file protocol handle
   */
  public static final String FILE_PROTOCOL = "file";

  /**
   *  The name of the file which contains the default configuration settings
   */
  public static final String DEFAULT_BOOTSTRAP = ".xoe/bootstrap.xml";

  /**
   *  The string to be appended to each property to build the complete system property name
   */
  public static final String PROPERTY_PREFIX = "xoe.";

  /**
   * The name of the logging configuration property
   */
  public static final String LOGGING_PROPERTY = "log";

  /**
   * The name of the script configuration property
   */
  public static final String SCRIPT_PROPERTY = "script";

  /**
   * Turn scripting support on or off at compile time
   */
  public static final boolean SCRIPT_SUPPORT = true;

  /**
   * The config property which specifies the screen width (in pixels)
   */
  public static final String SCREEN_WIDTH_PROPERTY = "screenWidth";

  /**
   * The config property which specifies the screen height (in pixels)
   */
  public static final String SCREEN_HEIGHT_PROPERTY = "screenHeight";

  /**
   * The config property which specifies the screen type (color,gray)
   */
  public static final String SCREEN_TYPE_PROPERTY = "screenType";

  /**
   * The config property which specifies the number of screen colors
   */
  public static final String SCREEN_COLORS_PROPERTY = "screenColors";

  /**
   * The config property which specifies the themes directory
   */
  public static final String THEMES_DIR_PROPERTY = "themesDir";

  /**
   * The default themes directory (relative path from XMLRoot)
   */
  public static final String DEFAULT_THEMES_DIR = STASH_PROTOCOL + ":/themes";

  /**
   * The config property which specifies the initial theme
   */
  public static final String THEME_PROPERTY = "defaultTheme";

  /**
   * The default theme
   */
  public static final String DEFAULT_THEME = "Tiki-COLOR";

  //  View Element Attributes
  public static final String URL_CONTEXT_ATTR = "urlcontext";
  public static final String ONCLICK_ATTR = "onclick";
  public static final String ONLOAD_ATTR = "onload";
  public static final String ONMOUSEDOWN_ATTR = "onmousedown";
  public static final String ONMOUSEUP_ATTR = "onmouseup";
  public static final String ONMOUSEOVER_ATTR = "onmouseover";
  public static final String ONMOUSEOUT_ATTR = "onmouseout";
  public static final String ONCHANGE_ATTR = "onchange";
  public static final String ACTION_ATTR = "action";
  public static final String ONKEYPRESS_ATTR = "onkeypress";

  // Attribute constants
  public static final String HREF_ATTR = "href";
  public static final String NAME_ATTR = "name";
  public static final String VALUE_ATTR = "value";
  public static final String SRC_ATTR = "src";

  // default commands
  public final static String ADD = "_ADD";
  public final static String NEW = "_NEW";
  public final static String DELETE = "_DELETE";
  public final static String SAVE = "_SAVE";
  public final static String BACK = "_BACK";
  public final static String FORWARD = "_FORWARD";
  public final static String TOP = "_TOP";
  public final static String EDIT = "_EDIT";
  public final static String LAUNCH = "_LAUNCH";
  public final static String SELECT = "_SELECT";
  public final static String REFRESH = "_REFRESH";
  public final static String LOADTHEME = "_LOADTHEME";
  public final static String DISPLAY = "_DISPLAY";
  public final static String RESET = "_RESET";
  public final static String BEEP = "_BEEP";
  public final static String STOP = "_STOP";
  public final static String ADD_APP = "_ADD_APP";

  public final static String DATA_DOC = "_DATA_DOC";
  public final static String PARENT_FRAME = "_PARENT";

  /**
   *  The icon to use when the icon is either unknown or not found
   */
  public static final String UNKNOWN_ICON = "question";

  /**
   * The text displayed under the previous directory icon
   */
  public final static String PREVIOUS_DIR = "Previous Directory";

  /**
   * The delimiter used to specify xpath queries embedded in banner attributes
   */
  public final static String QUERY_DELIMITER = "%";

  /**
   * Designates a node where the NodeType hasn't been set
   */
  public static final short UNSET_NODE_TYPE = -1;

  /**
   * 'less than' string and character values
   */
  public static final String LESSER_THAN_KEY = "lt";
  public static final char LESSER_THAN_SYMBOL = '<';

  /**
   * 'less than' string and character values
   */
  public static final String GREATER_THAN_KEY = "gt";
  public static final char GREATER_THAN_SYMBOL = '>';

  /**
   * 'ampersand' string and character values
   */
  public static final String AMPERSAND_KEY = "amp";
  public static final char AMPERSAND_SYMBOL = '&';
  public static final String AMPERSAND_SYMBOL_STRING = "&";

  /**
   * 'double quote' string and character values
   */
  public static final String DOUBLE_QUOTE_KEY = "quot";
  public static final char DOUBLE_QUOTE_SYMBOL = '"';

  /**
   * 'apostrophe' string and character values
   */
  public static final String APOSTROPHE_KEY = "apos";
  public static final char APOSTROPHE_SYMBOL = '\'';

  /**
   * 'dot' string and character values
   */
  public static final String MID_DOT_KEY = "middot";
  public static final char MID_DOT_SYMBOL = '\u00B7';

  /**
   * 'non breaking space' string and character values
   */
  public static final String NON_BREAKING_SPACE_KEY = "nbsp";
  public static final char NON_BREAKING_SPACE_SYMBOL = '\u00A0';

  /**
   * 'copyright' string and character values
   */
  public static final String COPYRIGHT_KEY = "copy";
  public static final char COPYRIGHT_SYMBOL = '\u00A9';

  /**
   * unicode decimal symbol
   */
  public final static String UNICODE_DECIMAL_SYMBOL = "#";

  /**
   * unicode hexadecimal symbol
   */
  public final static String UNICODE_HEX_SYMBOL = "#x";

  /**
   * The config property which specifies the user's country
   */
  public final static String COUNTRY_PROPERTY = "country";

  /**
   * The config property which specifies the user's language
   */
  public final static String LANGUAGE_PROPERTY = "language";

  /**
   * The config property which specifies the default 'beep' noise
   */
  public static final String BEEP_PROPERTY = "beep";

  /**
   * the localization processing instruction
   */
  public final static String XOE_LOCALIZE = "xoe-localize";

  /**
   *  the maximum number of apps kept in the last app bar in the app launcher
   */
  public static final int NUM_LAST_APPS = 3;

  /**
   *  the character to use when combining namespace and element tag name
   */
  public static final char NAMESPACE_SEPARATOR = '^';

  /**
   *  content type resolution constants
   */
  public static final String FUNCTION_CONTENT_HANDLER = "contenthandler";

  public static final String FEATURE_MIMETYPE = "mimetype";
  public static final String FEATURE_EXTENSION = "extension";
  public static final String FEATURE_DOCTYPE_NAME_NS = "doctypeNameNS";

  public static final String EXTENSION_ALL = "*";
  public static final String EXTENSION_XML = "xml";
  public static final String EXTENSION_ZIP = "zip";
  public static final String EXTENSION_XOE = "xoe";
  public static final String EXTENSION_XOE_CONFIG = "xoe-config";
  public static final String EXTENSION_HTML = "html";
  public static final String EXTENSION_XHTML = "xhtml";
  public static final String EXTENSION_WML = "wml";
  public static final String EXTENSION_DTD = "dtd";
  public static final String EXTENSION_LNK = "lnk";

  public static final String MIMETYPE_ALL = "*";
  public static final String MIMETYPE_UNKNOWN = "unknown";
  public static final String MIMETYPE_TEXT_PLAIN = "text/plain";
  public static final String MIMETYPE_TEXT_XML = "text/xml";
  public static final String MIMETYPE_TEXT_XML_DTD = "text/xml-dtd";
  public static final String MIMETYPE_TEXT_WML = "text/vnd.wap.wml";
  public static final String MIMETYPE_TEXT_HTML = "text/html";
  public static final String MIMETYPE_TEXT_VND_TVT_LNK = "text/vnd.tvt.lnk";
  public static final String MIMETYPE_APPLICATION_ZIP = "application/zip";
  public static final String MIMETYPE_APPLICATION_XOE = "application/x-xoe";
  public static final String MIMETYPE_STASHDIR = "application/x-stashdir";

  /**
   *  installed package base location
   */
  public static final String PACKAGE_ARCHIVE_ROOT = STASH_PROTOCOL + ":/installed-packages/";

  /**
   *  doctype name ns for xoe config files
   */
  public static final String XOE_CONFIG_DOCTYPE_NAME_NS = DEFAULT_NAMESPACE
  						          + NAMESPACE_SEPARATOR
						          + EXTENSION_XOE_CONFIG;

  /**
   *  doctype name ns for xhtml
   */
  public static final String DOCTYPE_NAME_NS_XHTML = XHTML_NAMESPACE
                                                     + NAMESPACE_SEPARATOR
						     + EXTENSION_HTML;

  /**
   *  default xhtml views
   */
  public static final String DEFAULT_VIEW = "default.xhtml";
  public static final String DISPLAY_VIEW = "display.xhtml";
  public static final String NEW_ITEM_VIEW = "new.xhtml";
  public static final String EDIT_ITEM_VIEW = "edit.xhtml";

  /**
   *  no path constant
   */
  public static final String NO_PATH = "";

  public static final String ASCENDING_SORT = "a";
  public static final String DESCENDING_SORT = "d";

  // xoe attributes
  public static final String CONTEXT_ATTR = "context-node";
  public static final String XPATH_ATTR = "xpath";
  public static final String BANNER_ATTR = "banner";
  public static final String STYLE_ATTR = "style";
  public static final String SORT_ORDER_ATTR = "sort-order";
  public static final String SORT_BANNER_ATTR = "sort-banner";
  public static final String INSERT_BEFORE_ATTR = "insert-before";
  public static final String HIDE_IF_ATTR = "hide-if";
  public static final String TRIGGER_ATTR = "trigger";
  public static final String PKG_URI_ATTR = "pkg-uri";
  public static final String PKG_WORKSPACE_ATTR = "pkg-workspace";
  public static final String PKG_ARCHIVE_ATTR = "pkg-archive";

  /**
   * This is a location on disk that must exist and be R/W.
   * Use this location if *must* write data to disk (e.g. NativeInstaller.java)
   * All other locations on the filesystem may be R/O depending device.
   */
  public final static String TEMP_DIR = "/tmp/xoe/";

  public static final String USER_NAME = "default";
} // PLConstants
