/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: FileContentElement.java,v 1.2 2001/12/18 06:06:15 colin Exp $
 */
package org.xoe.core.content;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import org.w3c.dom.Document;

import org.xoe.core.util.Logger;


/**
 * The class <code>FileContentElement</code> represents a <code>ContentElement
 * </code> read from disk.
 */
class FileContentElement 
  extends ContentElement 
{
  /** 
   * The file on disk that the FileContentElement represents.
   */
  protected File file;

  FileContentElement (File f, URL u, Document owner)
    throws FileNotFoundException, IOException
  {
    this (null, f, u, owner);
  }

  /**
   * Throw an exception.  FikeContentElements do not have actual data.
   */
  void setData (byte[] d)
  {
    throw new RuntimeException ("cannot setData on FileContentElement");
  }

  /**
   * Override makeAlias.
   */
  public ContentElement makeAlias (URL u)
  {
    try
      {
        return new FileContentElement (file, u, getOwnerDocument ());
      }
    catch (FileNotFoundException e)
      {
        Logger.log (this, e);
      }
    catch (IOException e)
      {
        Logger.log (this, e);
      }
    return null;
  }

  FileContentElement (ContentType type, File f, URL u, Document owner)
    throws FileNotFoundException, IOException
  {
    super (u, owner, null);
    if (f != null)
      {
        this.file = f;
        setLength (f.length ());

        // set the node name
        setNodeName (encodeName (f.getName ()));

        // set the content type
        if (type == null)
          {
            setType (DefaultContentTypeResolver.getContentType (u, 
                                                                getNodeName (), 
                                                                new FileInputStream (f)));   
          }
        else
          {
            setType (type);   
          }
      }
    else
      {
        throw new FileNotFoundException ("file argument is null");
      }    
  }

  /** 
   * Get the location of the ContentElement's data on disc 
   */
  File getFile () 
  { 
    return file; 
  }

  /** 
   * Return an InputStream from the File.
   */
  public InputStream getInputStream ()
  {
    try
      {
        return new FileInputStream (file);
      }
    catch (FileNotFoundException fnfe)
      {
        Logger.log (this, fnfe);
        return null;
      }
  }
} 
