/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: StashDirectory.java,v 1.2 2001/12/10 22:48:43 colin Exp $
 */
package org.xoe.core.content;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Date;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.DOMException;
import org.xoe.core.util.Logger;


/**
 * StashDirectory is a simple DOM element that represents a * direcotory
 * in the stash VFS.  The fist time a directory is accessed expand ()
 * will be called which triggers the ready of the directories contents
 * from the stash location on disc.
 */
public class StashDirectory
  extends FileContentElement
{
  // has the directory been read in from disk yet?
  boolean expanded = false;
      

  // -------------------------------------------------------------------------
  // constructors
  
  StashDirectory (File f, URL u, Document owner)
    throws FileNotFoundException, IOException
  {
    super (new ContentType (MIMETYPE_STASHDIR), f, u, owner);
  }


  /*
   *  set the location of the ContentElements data on disc 
   */
  void setFile (File infile) 
  { 
    file = infile; 
  }

  /*
   * Read in the directory entries from disk and make children
   * from them.
   */
  void expand ()
  {
    expanded = true;
    if (file == null) 
      {
        return;
      }

    String filename = file.getPath();

    // read the contents of the directory, if it exists.
    String absPrefix = filename + File.separator;
    String prefix = absPrefix.substring (Stash.getPath ().length ());
    if (file.isDirectory ())
      {
        String listing[] = file.list (); 
        for (int i=0; i<listing.length; i++)
          {
            // don't add file that start with a dot, end with "~" or are CVS directories
            if (listing[i].startsWith (".") || listing[i].endsWith("~") || "CVS".equals(listing[i]))
              { 
		continue;
              }


	    try
	      {
		File child = new File (absPrefix + listing[i]);
		String encodedListing = ContentElement.encodeName (listing[i]);
		if (child.isDirectory ())
                  {
		    Stash.createDirectory (this, encodedListing);
		  }
		else if (child.isFile ())
		  {
		    URL newURL = new URL (STASH_PROTOCOL + ":" + prefix + encodedListing);
		    ContentElement elem = Stash.createFileContentElement (newURL, child);
		    Stash.put (elem);
		  }
	      }
	    catch (IOException ioe)
	      {
		throw new DOMException (ioe);
	      }
          }
      }
  }
  

  //----------------------------------------------------------------------
  // org.w3c.dom.Node

  public Node getFirstChild ()
  {
    if (!expanded) 
      { 
        expand (); 
      }
    return super.getFirstChild ();
  }

  public Node getLastChild ()
  {
    if (!expanded) 
      { 
        expand (); 
      }
    return super.getLastChild ();
  }

  public Node insertBefore (Node newChild, Node refChild) 
    throws DOMException 
  {
    if (!expanded) 
      { 
        expand (); 
      }
    lastModified = new Date ();
    return super.insertBefore (newChild, refChild);
  }

  public Node replaceChild (Node newChild, Node oldChild) 
    throws DOMException
  {
    if (!expanded) 
      { 
        expand (); 
      }
    lastModified = new Date ();
    return super.replaceChild (newChild, oldChild);
  }

  public Node removeChild (Node oldChild) 
    throws DOMException
  {
    if (!expanded) 
      { 
        expand (); 
      }
    lastModified = new Date ();
    return super.removeChild (oldChild);
  }

  public Node appendChild (Node newChild) 
    throws DOMException 
  {
    if (!expanded) 
      { 
        expand (); 
      }
    lastModified = new Date ();
    return super.appendChild (newChild);
  }

  public boolean hasChildNodes ()  
  {
    if (!expanded) 
      { 
        expand (); 
      }
    return super.hasChildNodes ();
  }

  public Node cloneNode (boolean deep)
  {
    if (!expanded) 
      { 
        expand (); 
      }
    return super.cloneNode (deep);
  }


  //----------------------------------------------------------------------
  // org.w3c.dom.NodeList

  public Node item (int index)   
  {
    if (!expanded) 
      { 
        expand (); 
      }
    return super.item (index);
  }

  public int getLength () 
  {
    if (!expanded) 
      { 
        expand (); 
      }
    return super.getLength ();
  }

}
