/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: INetwork.java,v 1.7 2001/11/05 22:26:03 samc Exp $
 */
package org.xoe.core.device;

import java.io.IOException;

import org.xoe.core.services.IService;

/**
 * Java interface for control IP networking.
 * Use IOExceptions to when network configuration fails.
 */
public interface INetwork
  extends IService
{
  // DNS subsystem
  public String   getHostname      ()                throws IOException;
  public void     setHostname      (String host)     throws IOException;
  public String   getDomainname    ()                throws IOException;
  public void     setDomainname    (String domain)   throws IOException;
  public String[] getNameservers   ()                throws IOException;
  public void     setNameservers   (String[] server) throws IOException;

  // network interface discovery
  public String[] getNetworkInterfaces () throws IOException;

  // network interface specific settings.
  public NetworkInterface getSettings (String ifaceName) throws IOException;
  public void applySettings (NetworkInterface iface) throws IOException;

  static public class NetworkInterface
  {
    public boolean active   = false;
    public boolean dhcp     = false;
    public String name      = null;
    public String transport = null;
    public String address   = null;
    public String netmask   = null;
    public String broadcast = null;
    public String gateway   = null;

    // dump interface information to a String.
    public String toString ()
    {
      StringBuffer buffer = new StringBuffer ();
      buffer.append ("iface     : " + name);
      if (active) buffer.append ("(active)" + "\n");
      else buffer.append (" (not active)" + "\n");
      buffer.append ("dhcp      : ").append (dhcp     ).append ("\n");
      buffer.append ("IP        : ").append (address  ).append ("\n");
      buffer.append ("transport : ").append (transport).append ("\n");
      buffer.append ("netmask   : ").append (netmask  ).append ("\n");
      buffer.append ("broadcast : ").append (broadcast).append ("\n");
      buffer.append ("gateway   : ").append (gateway  );
      return buffer.toString ();
    }
  }
}
