/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Cache.java,v 1.3 2001/12/13 00:46:35 jason Exp $
 */
package org.xoe.core.dom;

import java.util.Stack;

import org.xoe.core.Constants;
import org.xoe.core.util.Logger;
import org.xoe.core.util.ProviderHashtable;

/**
*Class Cache creates a new ProviderHashtable to handle DOM cache.
*/
public class Cache 
  implements Constants
{
  static ProviderHashtable caches = new ProviderHashtable();

  protected Class clazz;
  protected Stack freelist;

  public static Cache getCache (Class cls) 
  {
    return (getCache (cls, Cache.class, null));
  }

  public static Cache getCache (String name) 
  {
    return (getCache (name, Cache.class, null));
  }

  public static Cache getCache (Class cls, Class ctype, Cache def) 
  {
    Cache cache = (Cache)caches.get (cls);
    if (cache != null) 
      {
	return (cache);
      }

    try 
      {
	cache = (Cache)ctype.newInstance ();
	cache.clazz = cls;
	caches.put (cls, cache);
	return (cache);
      }
    catch (NoClassDefFoundError ndfe) 
      {
	Logger.log (Cache.class, ndfe);
      }
    catch (InstantiationException ie) 
      {
	Logger.log (Cache.class, ie);
      }
    catch (IllegalAccessException iae) 
      {
	Logger.log (Cache.class, iae);
      }
    catch (NullPointerException npe) 
      {
	Logger.log (Cache.class, "while instantiating: " + cls);
	Logger.log (Cache.class, npe);
      }

    // Not found - use the default
    if (def != null) 
      {
	caches.put (cls, def);
      }

    return (def);
    
  }

  public static Cache getCache (String name, Class ctype, Cache def) 
  {
    Cache cache = (Cache)caches.get (name);
    if (cache != null) 
      {
	return (cache);
      }

    String namespace = null;
    String classname = name;

    // try looking up the default cache associated with the current
    // namespace
    if (namespace != null)
      {
	cache = (Cache)caches.get (namespace);
	if (cache != null)
	  {
	    return cache;
	  }
      }

    // Not found - use the default
    if (def != null) 
      {
	caches.put (name, def);
      }

    return (def);
  }

  Cache () 
  {
    freelist = new Stack ();
  }

  /**
   * Get an instance of the object we're caching, using one from the cache
   * if available.
   */
  public Cacheable get () 
  {
    Cacheable node = null;
    synchronized (freelist) 
      {
	if (freelist.empty () == false) 
          {
	    node = (Cacheable)freelist.pop ();
	    node.reset ();
	    return (node);
	  }
      }

    try 
      {
	node = (Cacheable)clazz.newInstance ();
      }
    catch (InstantiationException ie) 
      {
	    Logger.log( this, ie);
      }
    catch (IllegalAccessException iae) 
      {
	    Logger.log( this, iae);
      }
    catch  (NullPointerException e)
      {
	Logger.log (this, e);
	Logger.log (this, " freelist = " + freelist);
	Logger.log (this, " clazz = " + clazz);
      }
    return (node);
  }

  /**
   * Recycle the object.
   */
  public void put (Cacheable node) 
  {
    freelist.push (node);
  }

  /**
   * Create an alias to a node cache (for tags).
   */
  public static void addCacheAlias (String tag, String cname) 
  {
    caches.put (tag, getCache (cname));
  }
/**
*Create an alias to a node cache (for tags).
*/
  public static void addCacheAlias (String tag, String cname, Object provider) 
  {
    caches.put (tag, getCache (cname), provider);
  }
}
