/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: DefaultEntityResolver.java,v 1.3 2001/11/15 19:24:50 tim Exp $
 */
package org.xoe.core.dom;

import org.xoe.core.Constants;
import org.xoe.core.util.Logger;

public class DefaultEntityResolver 
  implements EntityResolver, Constants
{
  private static DefaultEntityResolver singleton;

  public static DefaultEntityResolver getInstance ()
  {
    if (singleton == null)
      {
	singleton = new DefaultEntityResolver();
      }
           
    return (singleton);
  }

  /**
   * The defaults are hard-coded so this doesn't do anything
   */
  public void addEntity (String name, String value)
  {
  }

  public static String deEntify (String segment)
  {
    StringBuffer ret = null;

    int slen = segment.length ();
    for (int counter = 0; counter < slen; counter++)
      {
        String entity = null;
	char ch = segment.charAt(counter);
	if (ch == LESSER_THAN_SYMBOL)
	  {
	    entity = LESSER_THAN_KEY;
	  }
	else if (ch == GREATER_THAN_SYMBOL)
	  {
	    entity = GREATER_THAN_KEY;
	  }
	else if (ch == AMPERSAND_SYMBOL)
	  {
	    entity =  AMPERSAND_KEY;
	  }
	else if (ch == DOUBLE_QUOTE_SYMBOL)
	  {
	    entity = DOUBLE_QUOTE_KEY;
	  }
	else if (ch == APOSTROPHE_SYMBOL)
	  {
	    entity = APOSTROPHE_KEY;
	  }
	else if (ch == MID_DOT_SYMBOL)
	  {
	    entity = MID_DOT_KEY;
	  }
	else if (ch == NON_BREAKING_SPACE_SYMBOL)
	  {
	    entity = NON_BREAKING_SPACE_KEY;
	  }
	else if (ch == COPYRIGHT_SYMBOL)
	  {
	    entity = COPYRIGHT_KEY;
	  }

	if (entity != null)
	  {
	    // First entity - create a string buffer and copy all the
	    // characters so far into it.
	    if (ret == null)
	      {
		ret = new StringBuffer ();
		if (counter > 0)
		  {
		    ret.append (segment.substring (0, counter-1));
		  }
	      }
	    ret.append ('&');
	    ret.append (entity);
	    ret.append (';');
	  }
	else if (ret != null)
	  {
	    ret.append (ch);
	  }
      }	

    // If we created a new string return that, else return the old one
    if (ret != null)
      {
	return (ret.toString ());
      }
    else
      {
	return (segment);
      }
  }

  public String getName (String value)
  {
    switch (value.charAt (0))
    {
      case LESSER_THAN_SYMBOL:
	return LESSER_THAN_KEY;
      case GREATER_THAN_SYMBOL:
        return GREATER_THAN_KEY;
      case AMPERSAND_SYMBOL:
        return AMPERSAND_KEY;
      case DOUBLE_QUOTE_SYMBOL:
        return DOUBLE_QUOTE_KEY;
      case APOSTROPHE_SYMBOL:
        return APOSTROPHE_KEY;
      case MID_DOT_SYMBOL:
        return MID_DOT_KEY;
      case NON_BREAKING_SPACE_SYMBOL:
        return NON_BREAKING_SPACE_KEY;
      case COPYRIGHT_SYMBOL:
        return COPYRIGHT_KEY;
      default:
	return null;
      }
  }

  public static void main (String [] args)
  {
    System.out.println (deEntify (args[0]));
  }

  public String getValue (String name)
  {
    name = name.intern ();

    if (name == LESSER_THAN_KEY)
      {
        return ""+LESSER_THAN_SYMBOL;
      }

    if (name == GREATER_THAN_KEY)
      {
        return ""+GREATER_THAN_SYMBOL;
      }

    if (name == AMPERSAND_KEY)
      {
        return ""+AMPERSAND_SYMBOL;
      }

    if (name == DOUBLE_QUOTE_KEY)
      {
        return ""+DOUBLE_QUOTE_SYMBOL;
      }

    if (name == APOSTROPHE_KEY)
      {
        return ""+APOSTROPHE_SYMBOL;
      }

    if (name == MID_DOT_KEY)
      {
        return ""+MID_DOT_SYMBOL;
      }

    if (name == NON_BREAKING_SPACE_KEY)
      {
        return ""+NON_BREAKING_SPACE_SYMBOL;
      }

    if (name == COPYRIGHT_KEY)
      {
        return ""+COPYRIGHT_SYMBOL;
      }

    if (name.startsWith (UNICODE_HEX_SYMBOL))
      {
        // return the hexadecimal unicode value
        return String.valueOf ((char)Integer.parseInt (name.substring (2), 16));
      } 

    if (name.startsWith (UNICODE_DECIMAL_SYMBOL))
      {
        // parse the decimal unicode value
        return String.valueOf ((char)Integer.parseInt (name.substring (1)));
      } 
    return null;
  }

}

