/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XAttr.java,v 1.1 2001/10/21 03:56:01 colin Exp $
 */
package org.xoe.core.dom;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * XAttr
 */
public class XAttr 
  extends ParentNode 
  implements Attr 
{
  Element element;
  String namespace;
  String name;
  String localname;
  String prefix;
  Object value;
    
  protected XAttr (Document owner) 
  {
    this.owner = owner;
  }

  protected XAttr (Document owner, Element eowner, String ns, String name, String value) 
  {
    this (owner);
    setNodeName (name);
    setValue (value);
    setOwnerElement (eowner);
    namespace = (ns == null ? null : ns.intern());
  }

  /**
   * Set the NodeName and Prefix from a possibly prefixed tag name.
   */
  public void setNodeName (String nodeName)
  {
    if (nodeName != null)
      {
	name = nodeName.intern();
	int idx = name.indexOf (':');
	if (idx == -1) 
	  {
	    prefix = null;
	    localname = name;
	  }
	else 
	  {
	    localname = name.substring (idx+1).intern();
	    prefix = name.substring (0, idx).intern();
	  }
      }
    else 
      {
	name = null;
	prefix = null;
	localname = null;
      }
  }

  public String getNodeName ()
  {
    return name;
  }

  public String getLocalName ()
  {
    return localname;
  }

  public String getPrefix () 
  {
    return prefix;
  }

  public void setPrefix (String prefix) throws DOMException 
  {
    setNodeName (prefix + ':' + getLocalName ());
  }

  // ------------------------------------------------------------------------------

  /**
   * used by the factory in BaseDocument to assign the correct namespace to the attr
   */
  public String getNamespaceURI () 
  {
    return (namespace);
  }

  // -------------------------------------------------------------------------
  // org.w3c.dom.Attr

  public String getName () 
  {
    return getLocalName ();
  }
 
  public boolean getSpecified ()
  { 
    return (value != null);
  }

  public String getValue ()
  {
    if (value instanceof XText)
      {
	return (((XText)value).getData ());
      }
    else
      {
	return ((String)value);
      }
  }

  public void setValue (String nvalue)
  {
    if (nvalue == null)
      {
	nvalue = "";
      }
    if (value instanceof XText)
      {
	((XText)value).setData (nvalue);
      }
    else
      {
	value = nvalue;
      }
  }

  // ----------------------------------------------------------------
  // org.w3c.dom.Node

  public Node getFirstChild ()
  {
    createText ();
    return (super.getFirstChild ());
  }

  public Node getLastChild ()
  {
    createText ();
    return (super.getLastChild ());
  }

  public boolean hasChildNodes ()
  {
    createText ();
    return (super.hasChildNodes ());
  }

  public NodeList getChildNodes ()
  {
    createText ();
    return (super.getChildNodes ());
  }

  public Node cloneNode (boolean deep)
  {
    XAttr nnode = (XAttr)super.cloneNode (deep);
    nnode.element = this.element;
    nnode.namespace = this.namespace;
    nnode.value = this.value;
    if (nnode.value instanceof XText && deep)
      {
	nnode.value = ((XText)this.value).cloneNode (deep);
      }
    nnode.name = this.name;
    nnode.localname = this.localname;
    nnode.prefix = this.prefix;
    return nnode;
  }

  public String getNodeValue () 
    throws DOMException 
  {
    return getValue ();
  }

  public void setNodeValue (String nodeValue) 
    throws DOMException 
  {
    setValue (nodeValue);
  }

  public Element getOwnerElement ()
  {
    return element;
  }

  public Node getParentNode () 
  {
    return null;
  }

  public Node getNextSibling () 
  {
    return null;
  }

  public Node getPrevSibling () 
  {
    return null;
  }

  // ------------------------------------------------------------------------------

  public short getNodeType () 
  {
    return Node.ATTRIBUTE_NODE;
  }

  public void setOwnerElement (Element ownerelm) 
  {
    element = ownerelm;        	
  }

  private synchronized void createText ()
  {
    if (value instanceof String)
      {
	XText text = new XText ();
	text.setData ((String)value);
	value = text;
      }
  }

  public String toString () 
  {
    String s = new String (getClass().getName() 
                           + ": " 
                           + (getNamespaceURI()  == null ? getLocalName() : getNamespaceURI() 
			   + ":" 
                           + getLocalName()) 
                           + " --> " 
                           + System.identityHashCode (this));
    return (s);
  }

} 
