/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XCharacterData.java,v 1.2 2001/12/12 03:42:31 suraj Exp $
 */
package org.xoe.core.dom;

import org.w3c.dom.CharacterData;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public abstract class XCharacterData
  extends XNode
  implements CharacterData
{
  public static final char[] noText = new char[0];
  
  protected char[] text = noText;

  // -------------------------------------------------------------------------
  // implementation of org.w3c.dom.CharacterData interface

  /**
   * 
   * @param param1 <description>
   * @param param2 <description>
   * @param param3 <description>
   * @exception org.w3c.dom.DOMException <description>
   */
  public void replaceData (int offset, int count, String arg) 
    throws DOMException
  {
    deleteData (offset, count);
    insertData (offset, arg);
  }

  /**
   * deleteData deletes "count" number of characters after the offset.  if count
   * goes past the end of the string, it deletes all the characters after the offset to
   * the end
   *
   * @param param1 <description>
   * @param param2 <description>
   * @exception org.w3c.dom.DOMException <description>
   */
  public void deleteData (int offset, int count) 
    throws DOMException
  {	
    if (offset<0 || count<0 || offset>getLength ())
      {
	throw new DOMException ((short) 1, "INDEX_SIZE_ERR");
      }
    else
      {
        if ((offset+count)>getLength ())
          {
            char[] na = new char[offset];
            System.arraycopy (text, 0, na, 0, offset);	
            text=na;
          }
        else
          {
            char[] na = new char[text.length-count];
            System.arraycopy (text, 0, na, 0, offset);	
            System.arraycopy (text, offset+count, na, offset, text.length-(offset+count));
            text = na;
          }
      } 
  }

  /**
   *
   * @param param1 <description>
   * @param param2 <description>
   * @exception org.w3c.dom.DOMException <description>
   */
  public void insertData (int offset, String arg) 
    throws DOMException
  {  
    if (offset<0 || offset>getLength ())
      {
	throw new DOMException ((short) 1, "INDEX_SIZE_ERR");
      }
    else
      {
        int al = arg.length ();
        char[] na = new char[text.length+al];

        System.arraycopy (text, 0, na, 0, offset);
        arg.getChars (0, al, na, offset);
        System.arraycopy (text, offset, na, offset+al, text.length-offset);

        text = na;
      }
  }

  public void appendData (char[] arg, int start, int length) 
    throws DOMException
  {
    char[] na = new char[text.length + length];
    
    System.arraycopy (text, 0, na, 0, text.length);
    System.arraycopy (arg, start, na, text.length, length);
    
    text = na;
  }

  public Node cloneNode (boolean deep)
  {
    XCharacterData nnode = (XCharacterData) super.cloneNode (deep);
    if (this.text != null)
      {
	nnode.text = (char []) this.text.clone ();
      }
    return (nnode);
  }


  /**
   *
   * @param param1 <description>
   * @exception org.w3c.dom.DOMException <description>
   */

  public void appendData (String arg) 
    throws DOMException
  {
    appendData (arg.toCharArray (), 0, arg.length ());
  }

  /**
   * substring returns a string of length count starting at the offset.  if
   * the count is too long, substring returns the rest of the string starting
   * at count
   *
   * @param param1 <description>
   * @param param2 <description>
   * @return <description>
   * @exception org.w3c.dom.DOMException <description>
   */
  public String substringData (int offset, int count) 
    throws DOMException
  {    
    if(offset<0 || count<0 || offset>getLength ())
      {
      throw new DOMException ((short) 1, "INDEX_SIZE_ERR");
      }else
	{
	  if ((count+offset)>getLength ())
	    {
	      return new String (text, offset, (getLength ()-offset));
	    }else
	      {
		return new String (text, offset, count);
	      }
	} 
  }

  /**
   *
   * @return <description>
   */
  public int getLength ()
  {
    return text.length;
  }

  /**
   *
   * @param param1 <description>
   * @exception org.w3c.dom.DOMException <description>
   */
  public void setData (String data) 
    throws DOMException
  {
    if (data != null) 
      {
        this.text = data.toCharArray ();
      }
    else 
      {
        this.text = noText;
      }
  }

  public void setNodeValue (String nodeValue)
  {
    setData (nodeValue);
  }

  /**
   *
   * @return <description>
   * @exception org.w3c.dom.DOMException <description>
   */
  public String getData () 
    throws DOMException
  {
    return new String (text);
  }

  public String getNodeValue () 
  {
    return getData ();
  }

  public void reset ()
  {
    super.reset ();
    text = noText;
  }


  // ??? what is this ???
  public NamedNodeMap getAttributes () 
  {
    return null;
  }
}
