/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XDocumentType.java,v 1.1 2001/10/21 03:56:01 colin Exp $
 */
package org.xoe.core.dom;

import java.io.Writer;
import java.io.IOException;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;


public class XDocumentType 
  extends XNode
  implements DocumentType
{
  // DocumentType members
  private NamedNodeMap entities;
  private NamedNodeMap notations;
  private String publicId;
  private String systemId;
  private String name;
  private String internalSubset;
  
  /**
   * DTD
   */
  private EntityResolver dtd;


  public XDocumentType ()
  {
    super ();
  }

  public XDocumentType (String newName, String newSystemId, String newPublicId)
  {
    name = newName;
    publicId = newPublicId;
    systemId = newSystemId;
  }

  // ----------------------------------------------------------------------------
  // member methods

  public String getName ()
  {
    return (name);
  }

  public String getNodeName ()
  {
    return (getName ());
  }

  public EntityResolver getDTD ()
  {
    return (dtd);
  }
  
  public Node cloneNode (boolean deep)
  {
    XDocumentType nnode = (XDocumentType)super.cloneNode (deep);

    /* these are unimplemented right now
       
       nnode.entities = entities;
       nnode.notations = notations;
    */
    
    nnode.publicId = this.publicId;
    nnode.systemId = this.systemId;
    nnode.name = this.name;
    nnode.internalSubset = this.internalSubset;
//    if ( this.dtd != null)
//      {
//	nnode.dtd = (EntityResolver)this.dtd.clone ();    
//      }
    
    return nnode;
  }


  // ----------------------------------------------------------------------------
  // org.w3c.dom.DocumentType Interface

  public NamedNodeMap getEntities ()
  {
    return entities;
  }

  public NamedNodeMap getNotations ()
  {
    return notations;
  }

  public String getPublicId ()
  {
    return publicId;
  }

  public String getSystemId ()
  {
    return systemId;
  }

  public String getInternalSubset ()
  {
    return internalSubset;
  }


  // ----------------------------------------------------------------------------
  // org.w3c.dom.Node

  public short getNodeType  ()
  {
    return DOCUMENT_TYPE_NODE;
  }


  // ----------------------------------------------------------------------------
  // org.xoe.core.dom.Cacheable

  public void reset ()
  {
    super.reset ();
    entities = null;
    notations = null;
    publicId = null;
    systemId = null;
  }
}














