/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XElement.java,v 1.9 2001/11/09 21:23:11 suraj Exp $
 */
package org.xoe.core.dom;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;


public class XElement 
  extends ParentNode 
  implements Element 
{
  String namespace;
  String name;
  String localname;
  String prefix;
  ElementNamedNodeMapAdaptor attributes;

  protected XElement ()
  {
    super ();
  }

  protected XElement (Document owner)
  {
    setValues (owner, null, null);
  }

  protected XElement (Document owner, String name, String namespaceURI)
  {
    setValues (owner, name, namespaceURI);
  }

  void setValues (Document owner, String name, String namespaceURI)
  {
    this.owner = owner;
    setNodeName (name);
    namespace = namespaceURI;
  }

  /**
   * Set the NodeName and Prefix from a possibly prefixed tag name.
   */
  public void setNodeName (String nodeName)
  {
    if (nodeName != null)
      {
	name = nodeName.intern ();
	int idx = name.indexOf (':');
	if (idx == -1) 
          {
	    prefix = null;
	    localname = name;
	  }
	else 
          {
	    localname = name.substring (idx+1).intern ();
	    prefix = name.substring (0, idx).intern ();
	  }
      }
    else 
      {
        name = null;
	prefix = null;
	localname = null;
      }
  }

  public String getPrefix () 
  {
    return prefix;
  }

  public void setPrefix (String prefix) throws DOMException 
  {
    setNodeName (prefix + ':' + getLocalName ());
  }


  public String getNodeName ()
  {
    return name;
  }

  public String getLocalName ()
  {
    return localname;
  }


  public Node cloneNode (boolean deep) 
  {
    XElement nnode = (XElement)super.cloneNode (deep);
    nnode.namespace = this.namespace;
    nnode.name = this.name;
    nnode.localname = this.localname;
    nnode.prefix = this.prefix;
    if (attributes != null) 
      {
	nnode.attributes = attributes.cloneMap (nnode);
      }
    return (nnode);
  }

  // -----------------------------------------------------------------------------
  // org.w3c.dom.Element

  public boolean hasAttributes ()
  {
    return (attributes != null && attributes.getLength () > 0);
  }

  public boolean hasAttribute (String name)
  {
    return hasAttributeNS (null, name);
  }

  public boolean hasAttributeNS (String namespaceURI, String localName)
  {
    return (attributes != null && attributes.getNamedItemNS (namespaceURI, localName) != null);
  }

  public String getTagName () 
  {
    return getLocalName ();
  }

  public String getAttribute (String name) 
  {
    return getAttributeNS (null, name);
  }

  public void setAttribute (String name, String value) 
  {
    setAttributeNS (null, name, value);
  }

  public void removeAttribute (String name) 
  {
    removeAttributeNS (null, name);
  }

  public Attr getAttributeNode (String name) 
  {
    return getAttributeNodeNS (null, name);
  }

  public Attr setAttributeNode (Attr newAttr) 
    throws DOMException 
  {
    return setAttributeNodeNS (newAttr);
  }

  public Attr removeAttributeNode (Attr oldAttr) 
    throws DOMException
  {
    if (attributes != null)
      {
	XAttr item = (XAttr)oldAttr;
        if (item != null)
          {
	    return ((Attr)attributes.removeNamedItemNS (item.namespace, item.localname));
          }
	throw new DOMException(DOMException.NOT_FOUND_ERR, "No attribute found for : " + oldAttr);
      }
    throw new DOMException(DOMException.NOT_FOUND_ERR, "No attribute found for name: " + oldAttr.getName());
  }

  public NodeList getElementsByTagName (String name) 
  {
    return (null);
  }

  public void normalize () 
  {
  }

  public String getAttributeNS (String namespaceURI, String localName) 
  {
    if (attributes == null) 
      {
	return (null);
      }
    XAttr item = (XAttr)attributes.getNamedItemNS (namespaceURI, localName);
    if (item == null)
      {
	return (null);
      }

    return item.getValue ();
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    if (qualifiedName == null)
      {
	throw new DOMException (DOMException.INVALID_CHARACTER_ERR, "setAttributeNS called with a null qualifiedName");
      }
    Attr attr = null;
    String localName = qualifiedName.substring (qualifiedName.indexOf (':') + 1);
    if (attributes != null)
      {
	attr = (Attr)attributes.getNamedItemNS (namespaceURI, localName);
      }
    if (attr == null)
      {
        attr = owner.createAttributeNS (namespaceURI, qualifiedName);
      }
    else 
      {
	((XAttr) attr).setNodeName (qualifiedName);
      }
    attr.setValue(value);
    setAttributeNodeNS (attr);
  }

  public void removeAttributeNS (String namespaceURI, String localName) 
  {
    if (attributes != null) 
      {
	attributes.removeNamedItemNS (namespaceURI, localName);
      }
  }

  public Attr getAttributeNodeNS (String namespaceURI, String localName) 
  {
    if (attributes == null) 
      {
	return (null);
      }
    return ((Attr)attributes.getNamedItemNS (namespaceURI, localName));
  }

  public Attr setAttributeNodeNS (Attr newAttr) throws DOMException 
  {
    if (attributes == null)
      {
	attributes = new ElementNamedNodeMapAdaptor (this);
      }

    return ((Attr)attributes.setNamedItemNS (newAttr));
  }

  public NodeList getElementsByTagNameNS (String namespaceURI, String localName) 
  {
    return (null);
  }


  // -----------------------------------------------------------------------------
  // org.w3c.dom.Node

  public NamedNodeMap getAttributes () 
  {
    if (attributes == null)
      {
	attributes = new ElementNamedNodeMapAdaptor (this);
      }
    return (attributes);
  }

  public short getNodeType () 
  {
    return Node.ELEMENT_NODE;
  }

  public String getNamespaceURI ()
  {
    return (namespace);
  }


}

