/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: BaseInstaller.java,v 1.3 2001/11/02 23:29:09 samc Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


/**
 * PackageXMLHandler for the basic package information provisio 
 * which includes name, version and target arch.
 */
public class BaseInstaller 
  implements IPackageXMLHandler
{
  /**
   * The namespace of the provios that BaseInstaller handels.
   */
  public static final String NAMESPACE = "http://www.xoe.org/installer/base";

  XoePackage xoePackage;

  public String[] getHandlerPredependencies ()
  {
    return new String[0];
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    if (PackageDependency.NAMESPACE.equals (depEl.getAttribute ("ns")))
      {
	return new PackageDependency (depEl);
      }
    else
      {
	throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
      }
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = {IPackageXMLHandler.FUNCTIONALITY};
    return ret;
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI () };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
    throws PackageException
  {
    Vector v = new Vector ();
    PackageManager.performQueryLocally (new PackageQuery ().setName (xoePackage.getPackageURI ()), v);
    if (v.size () > 0)
      { // already a package with this name
	throw new PackageException ("Package " + xoePackage + " cannot be installed because " +
				    "a package with that name is already installed (" +
				    v.elementAt (0) + ").");
      }
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------
  
  public void setPackage (XoePackage p)
  {
    xoePackage = p;
  }

  public Vector generateRequires ()
    throws PackageDescriptionException
  {
    Vector v = new Vector ();

    // copy over any explicit requires.
    NodeList ret = XPath.findNodes (xoePackage.getSpec (), "package/requires/dep");
    for (int i = 0; i < ret.getLength (); i++)
      {
	v.addElement (Dependency.create ((Element)ret.item (i)));
      }
    return v;
  }

  public Vector generateConflicts ()
    throws PackageDescriptionException
  {
    Vector v = new Vector ();

    // copy over any explicit conflicts
    NodeList ret = XPath.findNodes (xoePackage.getSpec (), "package/conflicts/dep");
    for (int i = 0; i < ret.getLength (); i++)
      {
	v.addElement (Dependency.create ((Element)ret.item (i)));
      }
    return v;
  }

  /**
   * All packages provide at least this basic proviso.
   */
  public Vector generateProvides ()
    throws PackageDescriptionException
  {
    Vector proviso = new Vector ();

    // add the implict base proviso 
    PackageVersion ver = xoePackage.getVersion ();
    String arch        = xoePackage.getArchitecture ();
    String name        = xoePackage.getName ();

    Document doc = new XDocument ();
    DocumentFragment frag = doc.createDocumentFragment ();

    Element elem;

    elem = doc.createElement ("version");
    elem.appendChild (doc.createTextNode (ver.toString ()));
    frag.appendChild (elem);

    elem = doc.createElement ("name");
    elem.appendChild (doc.createTextNode (name));
    frag.appendChild (elem);

    if (arch != "")
      {
        elem = doc.createElement ("architecture");
        elem.appendChild (doc.createTextNode (arch));
        frag.appendChild (elem);
      }

    proviso.addElement (new Proviso (PackageDependency.NAMESPACE, name, frag));

    // copy over any explicit provisos
    NodeList ret = XPath.findNodes (xoePackage.getSpec (), "package/provides/proviso");
    for (int i = 0; i < ret.getLength (); i++)
      {
	proviso.addElement (new Proviso ((Element) ret.item (i)));
      }

    return proviso;
  }
}
