/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: BindingInstaller.java,v 1.2 2001/10/23 22:37:09 suraj Exp $
 */
package org.xoe.core.packages;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.Vector;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XNodeCache;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


public class BindingInstaller 
  extends FileIteratingInstaller
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/binding";
  public static final String BINDINGS_FILENAME = "bindings.xml";

  public String[] getHandlerPredependencies ()
  {
    return new String[] { ClassInstaller.NAMESPACE,
			    NativeInstaller.NAMESPACE,
			    ResourceInstaller.NAMESPACE};
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  protected void processFile (ZipInputStream zis, ZipEntry entry)
  {
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  { // could check to make sure .class files exist for all referenced classes
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    try
      {
	ContentElement ce = Stash.fetch (Stash.getWorkingDirectory (xoePackage) + BINDINGS_FILENAME);
	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), BINDINGS_FILENAME);
	    NodeList spaces = XPath.findNodes (doc, "bindings/space");
	    for (int i = 0; i < spaces.getLength (); i++)
	      {
		Element space = (Element) spaces.item (i);
		String tagNamespace = space.getAttribute("tagNamespace");
		Element e = (Element) XPath.findFirstNode (space, "//default");
		if (e != null) 
		  {
		    installBinding (xoePackage, tagNamespace, e.getAttribute ("class"));
		  }
		
		NodeList list = XPath.findNodes (doc, "//bind");
		int len = list.getLength ();
		for (int j = 0; j < len; j++) 
		  {
		    e = (Element) list.item (j);
		    installBinding (xoePackage, 
				    tagNamespace + "^" + e.getAttribute ("tag"), 
				    e.getAttribute ("class"));
		  }
	      }
	  }
      }
    catch (IOException ex)
      {
	throw new InstallException (ex, xoePackage,
				    ("An error occurred while searching for " +
				     BINDINGS_FILENAME));
      }
  }

  private void installBinding (XoePackage xoePackage, String binding, String cname)
  {
    try
      {
	// Pass in xoePackage as the 'provider' so we can easily remove all of this
	// package's bindings
	XNodeCache.addClassToCache (cname, ((ClassInstallerData)xoePackage.getExData (ClassInstaller.NAMESPACE)).getClassLoader ().loadClass (cname), xoePackage);
	XNodeCache.addXNodeCacheAlias (binding, cname, xoePackage);
      }
    catch (ClassNotFoundException _)
      {
	Logger.log (this, "Failed to find class: " + cname);
      }
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    XNodeCache.removeProvider (xoePackage);
  }
}

