/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ClassInstaller.java,v 1.2 2001/11/08 01:26:52 erik Exp $
 */ 
package org.xoe.core.packages;

import java.io.IOException;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.w3c.dom.Element;

import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;


public class ClassInstaller 
  extends ClassIteratingInstaller
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/class";

  public String[] getHandlerPredependencies ()
  {
    return new String[0];
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  protected void processClass (String cname)
  {
    // addProviso (new Proviso (getNamespaceURI (), cname, null));
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    ZipInputStream zis = xoePackage.getZipInputStream ();
    XoeClassLoader loader = new XoeClassLoader (xoePackage);
    ClassInstallerData data = new ClassInstallerData (loader);

    PackageDescription[] allPkgs = PackageManager.getInstalledPackages ();

    for (int i = 0; i < allPkgs.length; i++)
      {
	// try to find out if this package depends on the other one
	IPackageQuery[] deps = xoePackage.getDescription ().getDependencies ();
	for (int j = 0; j < deps.length; j++)
	  {
	    Vector v = new Vector ();
	    allPkgs[i].performQuery (deps[j], v);
	    if (v.size() > 0)
	      { // dep matched the package
		XoePackage other = PackageManager.retrievePackageLocally (allPkgs[i]);
		ClassInstallerData otherData = (ClassInstallerData)other.getExData (NAMESPACE);
                if (otherData != null) 
                  {
                    data.linkTo (otherData);
                  }
		break;
	      }
	  }
	// try to find out if the other package depends on this one
	deps = allPkgs[i].getDependencies ();
	for (int j = 0; j < deps.length; j++)
	  {
	    Vector v = new Vector ();
	    xoePackage.getDescription ().performQuery (deps[j], v);
	    if (v.size() > 0)
	      { // dep matched the package
		XoePackage other = PackageManager.retrievePackageLocally (allPkgs[i]);
		ClassInstallerData otherData = (ClassInstallerData)other.getExData (NAMESPACE);
		otherData.linkTo (data);
		break;
	      }
	  }
      }

    try 
      {
	while (true)
	  {
	    ZipEntry entry = zis.getNextEntry ();
	    if (entry == null) break;
	    if (entry.getName ().endsWith (".class"))
	      {
		byte buf[] = new byte[(int) entry.getSize ()];
		
		for (int res = 0, off = 0; res != -1; )
		  {
		    res = zis.read (buf, off, buf.length - off);
		    off += res;
		  }
		
		loader.addClass (buf, 0, buf.length);
	      }
	    zis.closeEntry ();
	  }
	zis.close ();
      }
    catch (IOException e)
      {
	throw new InstallException (e, xoePackage, ("An error occurred while performing " +
						    ".class file installation."));
      }

    xoePackage.setExData (NAMESPACE, data);
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    ((ClassInstallerData)xoePackage.getExData (NAMESPACE)).unlink ();
    xoePackage.setExData (NAMESPACE, null);
  }
}
