/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ClassIteratingInstaller.java,v 1.2 2001/11/07 01:31:18 samc Exp $
 */ 
package org.xoe.core.packages;

import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;


public abstract class ClassIteratingInstaller
  extends FileIteratingInstaller
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/class";

  private String[] m_classes = null;

  public ClassIteratingInstaller ()
  {
    super.setSuffixes (new String[] { ".class" });
  }

  // Set the names of classes/interfaces of which you
  // want the subclasses/subinterfaces/implementors.
  //
  // If you don't call this, you get passed everything.
  protected void setClasses (String[] classes)
  {
    m_classes = classes;
  }

  // implement this function to actually do the deed.
  protected abstract void processClass (String cname)
    throws PackageException;

  // override this function to change default error handling
  protected void handleException (String cname, Throwable ex)
  {
    Logger.log (this, "WARNING: Ignoring class " + cname + " due to exception in Class.forName ():");
    Logger.log (this, ex);
  }

  private boolean checkClass (String cname)
  {
    if (m_classes == null)
      {
	return true;
      }

    try
      {
        XoeClassLoader loader = ((ClassInstallerData) m_package.getExData (getNamespaceURI ())).getClassLoader ();
	Class c = loader.loadClass (cname);
	return checkClass (c);
      }
    catch (Throwable ex)
      {
	handleException (cname, ex);
	return false;
      }
  }

  private boolean checkClass (Class c)
  {
    if (m_classes == null)
      {
	return true;
      }

    String name = c.getName ();
    for (int i = 0; i < m_classes.length; i++)
      {
	if (name.equals (m_classes[i]))
	  {
	    return true;
	  }
      }

    Class[] ifaces = c.getInterfaces ();
    for (int i = 0; i < ifaces.length; i++)
      {
	if (checkClass (ifaces[i]))
	  {
	    return true;
	  }
      }

    Class sc = c.getSuperclass ();
    if (sc != null && checkClass (sc))
      {
	return true;
      }

    return false;
  }

  protected void processFile (ZipInputStream zis, ZipEntry entry)
    throws PackageException
  {
    String fileName = entry.getName ();
    String cname = fileName.substring (0, fileName.length () - ".class".length ());
    cname = cname.replace ('/', '.');
    
    if (checkClass (cname))
      {
	processClass (cname);
      }
  }
}
