/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: DefaultPackageServer.java,v 1.2 2001/11/15 20:53:18 tim Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;


/**
 * PackageServer that will look for a server of the given type at 
 * query time.  This means you can create a DefaultPackageServer of
 * any type before the server implementation has been installed on 
 * the system.
 */
class DefaultPackageServer 
  implements IPackageQueryHandler
{
  private IPackageQueryHandler m_impl = null;

  private String m_type = null;
  private String m_url = null;

  protected DefaultPackageServer (String type, String url)
  {
    m_type = type;
    m_url = url;
  }

  public void performQuery (IPackageQuery in_q, Vector v)
  {
    if (m_impl == null)
      {
	SimpleServiceQuery q = SimpleServiceQuery.alloc (IPackageServerFactory.FUNCTIONALITY, IPackageServerFactory.INTERFACE);
	
	q.addFeature (IPackageServerFactory.TYPE_FEATURE, m_type);
	
	IPackageServerFactory factory = (IPackageServerFactory) ServiceLocator.findService (q);
	SimpleServiceQuery.free (q);
	
	if(factory != null)
	  {
	    m_impl = factory.getServer(m_type, m_url);
	  }
      }

    if (m_impl != null)
      {
	m_impl.performQuery (in_q, v);
      }
    else
      {
	Logger.log (this, "Unable to query server with unknown type: {type='" + m_type + "', url='" + m_url + "'}");
      }
  }
}
