/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: FileIteratingInstaller.java,v 1.2 2001/10/26 00:26:17 samc Exp $
 */ 
package org.xoe.core.packages;

import java.io.IOException;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.w3c.dom.Document;

import org.xoe.core.util.Logger;


/**
 * Base class for installing files.
 * Implement these in non-abstract subclasses:
 * public String getNamespaceURI ();
 */
public abstract class FileIteratingInstaller
  implements IPackageXMLHandler
{
  protected XoePackage m_package;

  protected Vector m_provides  = new Vector();
  protected Vector m_requires  = new Vector();
  protected Vector m_conflicts = new Vector();

  /** 
   * The filename suffixes that the install wishes to process.
   * Setting this to null results in all files being processed.
   * Subclasses should call setSuffixes.
   */
  private String[] m_suffixes = null;

  /**
   * This function is called during package *build* to generate the package.xml
   * file.  Most subclasses will simply add provisios bases on the files
   * in the zip archive.
   */
  protected abstract void processFile (ZipInputStream zis, ZipEntry entry)
    throws PackageException;

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = { IPackageXMLHandler.FUNCTIONALITY };
    return ret;
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI() };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String toString()
  {
    return (super.toString() + "[" + getNamespaceURI() + ": " + m_package + "]");
  }

  /**
   * Set the file suffixes that the installer wishes to process.
   */
  protected void setSuffixes (String[] suffixes)
  {
    m_suffixes = suffixes;
  }

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------
  
  public void setPackage (XoePackage p)
    throws PackageException
  {
    m_package = p;

    ZipInputStream zis = m_package.getZipInputStream ();

    while (true)
      {
	try
	  {
	    ZipEntry entry = zis.getNextEntry ();
	    if (entry == null) break;
	    String fileName = entry.getName ();
	    if (m_suffixes == null)
	      {
		processFile (zis, entry);
	      }
	    else
	      {
		for (int i = 0; i < m_suffixes.length; i++)
		  {
		    if (fileName.endsWith (m_suffixes[i]))
		      {
			processFile (zis, entry);
			break;
		      }
		  }
	      }
	  }
	catch (IOException ex)
	  {
	    throw new PackageException (ex, ("An error occurred while processing the package "
					     + m_package.getName()));
	  }
      }
  }

  public Vector generateRequires ()
  {
    return m_requires;
  }

  public Vector generateConflicts ()
  {
    return m_conflicts;
  }

  public Vector generateProvides ()
  {
    return m_provides;
  }
}
