/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: InstallContext.java,v 1.2 2001/11/01 22:57:30 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Hashtable;
import java.util.Vector;

import org.xoe.core.util.Logger;


public class InstallContext 
  implements IPackageQueryHandler
{
  private Hashtable m_mapPackages = null;
  private PackageDescription m_primary = null;

  private static Hashtable m_cacheQueryResults = new Hashtable ();

  private static int m_failDepth = 0;
  private static PackageDescription m_target = null;
  private static PackageDescription m_failedPkg = null;
  private static IPackageQuery m_failure = null;
  
  public static final int MAX_DEPTH = 20;

  private static void processFailure (int level, PackageDescription pkg, IPackageQuery q)
  {
    // this goes on the assumpution that the "most specific" failure
    // will occur at the greatest depth...

    // could probably improve this somewhat, but I'm not immediately
    // sure how

    if (m_failDepth < level || m_failure == null)
      {
	m_failedPkg = pkg;
	m_failDepth = level;
	m_failure = q;
      }
  }

  public static UninstallablePackage getFailure ()
  {
    if (m_failure != null)
      {
	return new UninstallablePackage (m_target, m_failedPkg, m_failure.toXPath ());
      }
    else
      {
	return null;
      }
  }

  private InstallContext (InstallContext orig)
  {
    m_mapPackages = (Hashtable)orig.m_mapPackages.clone ();
    m_primary = orig.m_primary;
  }

  public InstallContext ()
  {
    m_mapPackages = new Hashtable ();
  }

  public Object clone ()
  {
    return new InstallContext (this);
  }

  public void setPrimary (PackageDescription desc)
  {
    m_primary = desc;
  }

  public PackageDescription getPrimary ()
  {
    return m_primary;
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  {

    PackageDescription[] descs = (PackageDescription[]) PackageUtils.getValues (m_mapPackages);

    for (int i = 0; i < descs.length; i++)
      {
	descs[i].performQuery (in_q, out_res);
      }
  }

  public PackageDescription[] getPackages ()
  {
    return (PackageDescription[]) PackageUtils.getValues (m_mapPackages);
  }

  public void addPackage (PackageDescription pkg)
  {
    if (m_primary == null)
      {
	m_primary = pkg;
      }
    m_mapPackages.put (pkg.getName (), pkg );
  }

  public boolean contains (PackageDescription pkg)
  {
    PackageDescription oldPkg = (PackageDescription)m_mapPackages.get (pkg.getName ());
    if (oldPkg != null && oldPkg.getVersion ().eq (pkg.getVersion ()))
      {
	return true;
      }
    
    return false;
  }

  public boolean conflictsWith (PackageDescription pkg)
  {
    /*    if (pkg.getConflictsInSet (this).size () > 0)
      {
	return true;
      }

    PackageDescription[] descs = (PackageDescription[]) PackageUtils.getValues (m_mapPackages);

    for (int i = 0; i < descs.length; i++)
      { // a package implements IPackageQueryHandler, and is thus a set of one
	if (descs[i].getConflictsInSet (pkg).size () > 0)
	  {
	    return true;
	  }
      }
    */
    return false;
  }

  public synchronized static Vector getBatchContexts (String[] names)
  {
    Vector ctxs = null;
    for(int i = 0; i < names.length; i++)
      {
	Vector queryResults = new Vector();
	PackageQuery q = new PackageQuery().setName(names[i]);
	PackageManager.performQuery (q, queryResults, false);
	Vector temp = new Vector();
	for (int j = 0; j < queryResults.size(); j++)
	  {
	    Vector t2 = null;
	    if (ctxs == null)
	      {
		t2 = getCandidateContexts((PackageDescription)queryResults.elementAt(j));
	      }
	    else
	      {
		t2 = new Vector();
		for (int k = 0; k < ctxs.size(); k++)
		  {
		    InstallContext c = (InstallContext)((InstallContext)ctxs.elementAt(k)).clone();
		    c.addPackage ((PackageDescription)queryResults.elementAt(j));
		    t2.addElement(c);
		  }
		t2 = getCandidateContexts ((PackageDescription)queryResults.elementAt(j), t2, 0);
	      }
	    for (int k = 0; k < t2.size(); k++)
	      {
		temp.addElement (t2.elementAt(k));
	      }
	  }
	ctxs = temp;
      }

    if (ctxs == null)
      {
	return new Vector();
      }

    // only return contexts that actually contain all the packages we're looking for
    for (int i = ctxs.size() - 1; i >= 0; i--)
      {
	for (int j = 0; j < names.length; j++)
	  {
	    if (((InstallContext)ctxs.elementAt(i)).m_mapPackages.get (names[j]) == null)
	      {
		ctxs.removeElementAt(i);
		break;
	      }
	  }
      }

    return ctxs;
  }

  public synchronized static Vector getCandidateContexts (IPackageQuery q)
  {
    Vector out_cands = new Vector ();

    Vector matches = new Vector ();

    PackageDescription.resetQueryTime ();

    PackageManager.performQueryRemotely (q, matches);

    m_cacheQueryResults.put (q.toXPath (), matches);

    for (int i = 0; i < matches.size (); i++)
      {
	PackageDescription desc = (PackageDescription)matches.elementAt (i);

	Vector temp_cands = getCandidateContexts (desc);

	for (int j = 0; j < temp_cands.size (); j++)
	  {
	    out_cands.addElement (temp_cands.elementAt (j));
	  }
      }
    resetCache ();

    PackageDescription.resetQueryTime ();

    return out_cands;
  }

  public synchronized static Vector getCandidateContexts (PackageDescription desc)
  {
    m_failDepth = 0;
    m_failure = null;
    m_failedPkg = null;
    m_target = desc;

    Vector out_cands = new Vector ();
    InstallContext cand = new InstallContext ();
    cand.addPackage (desc);
    out_cands.addElement (cand);

    // 0 is for the depth-count
    out_cands = getCandidateContexts (desc, out_cands, 0);

    return out_cands;
  }

  private static void resetCache ()
  {
    m_cacheQueryResults = new Hashtable ();
  }

  // if there are no conflicts, returns q->getMatches ().size () * in_cands.size ()
  // candidates.
  private static Vector getCandidateContexts (PackageDescription p, IPackageQuery q,
					     Vector in_cands, int level)
  {
    level++;

    Vector out_cands = new Vector ();

    for (int i = 0; i < in_cands.size (); i++)
      {
	Vector resultsInCand = new Vector ();
	//check to see if there is alreay a package that matches the query in the context
	((InstallContext)in_cands.elementAt (i)).performQuery (q, resultsInCand);

	if (resultsInCand.size () != 0)
	  { // this candidate already contains a matching package
	    // need to make sure to add it in to the out_cands vector
	    out_cands.addElement ((InstallContext)in_cands.elementAt (i));
	  }

      }
    
    if (level >= MAX_DEPTH)
      { // don't expand any further
	processFailure (level, p, q);
	return out_cands;
      }

    //this is done thru PackageManager.performQuery (); which looks locally first, and only 
    //looks remotely if it finds nothing locally.
    Vector matches = (Vector)m_cacheQueryResults.get (q.toXPath ());
    if (matches == null)
      {
	matches = new Vector ();
	PackageManager.performQuery (q, matches);
	m_cacheQueryResults.put (q.toXPath (), matches);
      }

    for (int i = 0; i < matches.size (); i++)
      {
	PackageDescription desc = (PackageDescription)matches.elementAt (i);
	Vector temp_cands = new Vector ();

	for (int j = 0; j < in_cands.size (); j++)
	  {
	    Vector resultsInCand = new Vector ();
	    

	    InstallContext cand = (InstallContext)in_cands.elementAt (j);

	    //check to see if there is alreay a package that matches the query in the context
	    cand.performQuery (q, resultsInCand);
	    if (resultsInCand.size () != 0)
	      { // this candidate is still valid... it already contains a matching package

		// do nothing
	      }
	    //and if not, make sure this package does not conflict with other packages
	    else if (!cand.conflictsWith (desc))
	      {
		InstallContext nCand = (InstallContext)cand.clone ();

		nCand.addPackage (desc);

		temp_cands.addElement (nCand);
	      }
	  }

	Vector res_cands = getCandidateContexts (desc, temp_cands, level);
	for (int j = 0; j < res_cands.size (); j++)
	  {
	    out_cands.addElement (res_cands.elementAt (j));
	  }
	
      }

    if (out_cands.size () == 0)
      {
	processFailure (level, p, q);
      }

    return out_cands;
  }

  private static Vector getCandidateContexts (PackageDescription p, Vector in_cands, int level)
  {
    level++;
    Vector out_cands = in_cands;
    for (int i = 0; i < p.getDependencies ().length; i++)
      {
	out_cands = getCandidateContexts (p, p.getDependencies ()[i], out_cands, level);
      }
    return out_cands;
  }

  public static void printUsage ()
  {
    System.err.println ("usage: InstallContext <conf file> <-q queryFile | -n pkgName>");
  }
  public static void dumpContextVector (Vector contexts)
  {
    for (int i = 0; i < contexts.size (); i++)
      {
	InstallContext ctx = (InstallContext)contexts.elementAt (i);
	
	System.err.println ("Context " + i + " (" + ctx.getPackages ().length + " packages) :");

	System.err.println (ctx.toString ());
      }    
  }

  public String toString ()
  {
    StringBuffer ret = new StringBuffer ();
    Object[] descs = getPackages ();
    
    ret.append ("Primary: " + m_primary + "\n");
    for (int j = 0; j < descs.length; j++)
      {
	if (j > 0)
	  {
	    ret.append ("\n");
	  }
	ret.append (descs[j]);
      }
    return ret.toString ();
  }
}
