/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: InstallOption.java,v 1.3 2001/10/26 03:20:51 erik Exp $
 */
package org.xoe.core.packages;

public class InstallOption
{
  private PackageDescription m_primary = null;

  private PackageActionSet m_actionSet = null;

  InstallOption (PackageDescription primary, PackageActionSet pas)
  {
    m_actionSet = pas;
    m_primary = primary;
  }

  protected PackageActionSet getActionSet ()
  {
    return m_actionSet;
  }
 
  // The primary package is the 'target' package for whom an
  // install/upgrade/uninstall was initially requested.
  // All other package changes are listed to satisfy dependencies of
  // the primary package.
  public PackageDescription getPrimary ()
  {
    return m_primary;
  }

  // List of package that will be newly installed by
  // committing this option.
  public PackageDescription[] getNewInstalls ()
  {
    return m_actionSet.getNewInstalls ();
  }

  // List of package that will be up/down-graded by
  // committing this option.
  public PackageDescription[][] getReplacements ()
  {
    return m_actionSet.getReplacements ();
  }

  // List of package that will be uninstalled by
  // committing this option.
  public PackageDescription[] getUninstalls ()
  {
    return m_actionSet.getUninstalls ();
  }

  public String toString ()
  {
    return "Primary: " + m_primary + "\n" + m_actionSet.toString ();
  }

  public static String toString(InstallOption[] opts)
  {
    if (opts == null)
      {
	return "null";
      }
    StringBuffer strb = new StringBuffer ();
    for (int i = 0; i < opts.length; i++)
      {
	strb.append ("Option " + i + ":\n");
	strb.append (opts[i].toString ());
	strb.append ("\n");
      } 
    return strb.toString ();
  }
}
