/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: InstallScriptHandler.java,v 1.4 2001/11/08 01:26:52 erik Exp $
 */
package org.xoe.core.packages;

import java.net.URL;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


public class InstallScriptHandler
  implements IPackageXMLHandler
{
  private static class InstallScriptHandlerData
  {
    private InstallScript m_script = null;

    private InstallScriptHandlerData (InstallScript script)
    {
      m_script = script;
    }

    private InstallScript getScript ()
    {
      return m_script;
    }
  }

  public static final String NAMESPACE = "http://www.xoe.org/installer/custom-scripts";
  public static final String SCRIPTS_FILENAME = "scripts.xml";
  XoePackage xoePackage;

  public String[] getHandlerPredependencies ()
  {
    return new String[] { BindingInstaller.NAMESPACE,
			  ResourceInstaller.NAMESPACE,
			  ClassInstaller.NAMESPACE,   
                          MimeTypeInstaller.NAMESPACE,
			  NativeInstaller.NAMESPACE,
			  ServiceInstaller.NAMESPACE };
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    return new String[] { IPackageXMLHandler.FUNCTIONALITY };
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI() };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    try
      {
	ContentElement ce = Stash.fetch (Stash.getWorkingDirectory (xoePackage) + SCRIPTS_FILENAME);
	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), SCRIPTS_FILENAME);

	    String cname = XPath.findString (doc, "custom-script");

	    if (cname != null)
	      {
		Class clazz = ((ClassInstallerData) xoePackage.getExData (ClassInstaller.NAMESPACE)).getClassLoader ().loadClass (cname);
		InstallScript script = (InstallScript)clazz.newInstance ();
		script.postInstall ();
		InstallScriptHandlerData data = new InstallScriptHandlerData (script);
		xoePackage.setExData (NAMESPACE, data);
	      }
	  }
      }
    catch (InstallException ex)
      { // the install script can throw this
	throw ex;
      }
    catch (Exception ex)
      { // lots of things could have happened... ClassNotFoundException, ClassCastException,
	// IOException...
	throw new InstallException (ex, xoePackage,
				    "An error occurred while processing custom install scripts");
      }
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    InstallScriptHandlerData data = (InstallScriptHandlerData) xoePackage.getExData (NAMESPACE);
    if (data == null)
      {
        return;
      }

    InstallScript script = data.getScript ();
    if (script == null)
      {
        return;
      }

    try
      {
        script.preUninstall ();
      }
    catch (Exception ex)
      {
        Logger.log (this, "Unhandled runtime exception in uninstall script for package " + xoePackage + ":");
        Logger.log (this, ex);
      }

    xoePackage.setExData (NAMESPACE, null);
  }

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------
  
  public void setPackage (XoePackage p)
  {
    xoePackage = p;
  }

  public Vector generateRequires ()
  {
    return null;
  }

  public Vector generateConflicts ()
  {
    return null;
  }

  public Vector generateProvides ()
  {
    return null;
  }
}
