/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: MimeTypeInstaller.java,v 1.3 2001/11/08 19:52:13 erik Exp $
 */
package org.xoe.core.packages;

import java.io.IOException;
import java.util.Vector;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.DefaultContentTypeResolver;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;


public class MimeTypeInstaller 
  implements IPackageXMLHandler
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/mimetype/mappings";
  public static final String MIMETYPES_FILENAME = "mimetypes.xml";
  XoePackage xoePackage;

  public String[] getHandlerPredependencies ()
  {
    return new String[] {ResourceInstaller.NAMESPACE};
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    return new String[] { IPackageXMLHandler.FUNCTIONALITY };
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI() };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    try
      {
	ContentElement ce = Stash.fetch (Stash.getWorkingDirectory (xoePackage) + MIMETYPES_FILENAME);
	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), MIMETYPES_FILENAME);
	    DefaultContentTypeResolver.getInstance ().processMimetypeFile (doc, xoePackage);
	  }
      }
    catch (IOException ex)
      {
	throw new InstallException (ex, xoePackage,
				    ("An error occurred while searching for " +
				     MIMETYPES_FILENAME));
      }
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    DefaultContentTypeResolver.getInstance ().removeProvider (xoePackage);
  }

  // ------------------------------------------------------
  // build time functions used to generate package.xml 
  // ------------------------------------------------------
  
  public void setPackage (XoePackage p)
  {
    xoePackage = p;
  }

  public Vector generateRequires ()
  {
    return null;
  }

  public Vector generateConflicts ()
  {
    return null;
  }

  public Vector generateProvides ()
  {
    return null;
  }
}

