/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageAction.java,v 1.1 2001/10/22 20:19:04 colin Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;


public class PackageAction 
  implements Cloneable, IPackageQueryHandler
{
  private String[] action_strings = { "UNKNOWN",
				      "NOT_INSTALLED",
				      "INSTALLED",
				      "SELECTED_INSTALL",
				      "SELECTED_REMOVE",
				      "SELECTED_REPLACE" };
    
  public static final int PA_UNKNOWN = 0;
  public static final int PA_NOTINSTALLED = 1;
  public static final int PA_INSTALLED = 2;
  public static final int PA_SELECTED_INSTALL = 3;
  public static final int PA_SELECTED_REMOVE = 4;
  public static final int PA_SELECTED_REPLACE = 5;
  
  // The description of the package
  private PackageDescription m_desc = null;

  // For Upgrade/Downgrade, the package we will be replaced by
  private PackageDescription m_replace = null;

  private int m_action = PA_UNKNOWN;

  PackageAction (PackageDescription desc)
  {
    m_desc = desc;
  }

  PackageAction (PackageDescription desc, int state)
  {
    m_desc = desc;
    m_action = state;
  }

  public Object clone ()
    throws CloneNotSupportedException
  {
      return super.clone ();
  }

  public void performQuery (IPackageQuery q, Vector out_res)
  {
    if(m_action == PA_INSTALLED || m_action == PA_SELECTED_INSTALL)
      {
	m_desc.performQuery(q, out_res);
      }
    else if(m_action == PA_SELECTED_REPLACE)
      {
	m_replace.performQuery(q, out_res);
      }
  }
  
  public boolean isCompleteInSet (IPackageQueryHandler qHandler)
  {
    try
      {
	checkCompleteInSet (qHandler);
      }
    catch (PackageException ex)
      {
	return false;
      }
    return true;
  }

  public void checkConflictsInSet(IPackageQueryHandler qHandler)
    throws PackageIncompleteException
  {
    if (m_action == PA_INSTALLED || m_action == PA_SELECTED_INSTALL)
      {
	Vector confPkgs = m_desc.getConflictsInSet (qHandler);
	if (confPkgs.size () > 0)
	  {
	    throw new PackageIncompleteException(m_desc, confPkgs, null);
	  }
      }
    else if (m_action == PA_SELECTED_REPLACE)
      {
	Vector confPkgs = m_replace.getConflictsInSet (qHandler);
	if (confPkgs.size() > 0)
	  {
	    throw new PackageIncompleteException (m_replace, confPkgs, null);
	  }
      }
  }

  public void checkCompleteInSet (IPackageQueryHandler qHandler)
    throws PackageIncompleteException
  {
    if(m_action == PA_INSTALLED || m_action == PA_SELECTED_INSTALL)
      {
	m_desc.checkCompleteInSet (qHandler);
      }
    else if(m_action == PA_SELECTED_REPLACE)
      {
	m_replace.checkCompleteInSet (qHandler);
      }
  }

  // in the case of a replacement, return the
  // thing being replaced.
  public PackageDescription getOriginal()
  {
    if(m_action == PA_SELECTED_REPLACE)
      {
	return m_desc;
      }

    return null;
  }

  public PackageDescription getDescription()
  {
    if(m_action == PA_SELECTED_REPLACE)
      {
	return m_replace;
      }
    
    return m_desc;
  }

  public String getName()
  {
    return m_desc.getName();
  }

  public void setAction(int action)
  {
    m_action = action;
  }

  public int getAction()
  {
    return m_action;
  }

  public void setState(int state)
  {
    m_action = state;
  }

  public void replaceWith(PackageDescription newDesc)
  {
    if(m_action == PA_SELECTED_REPLACE && m_desc.equals(newDesc))
      {
	m_replace = null;
	m_action = PA_INSTALLED;
	return;
      }
    else if(m_action == PA_NOTINSTALLED)
      {
	m_desc = newDesc;
	m_action = PA_SELECTED_INSTALL;
	return;
      }

    m_replace = newDesc;
    m_action = PA_SELECTED_REPLACE;
  }

  public String toString()
  {
    if(m_action == PA_SELECTED_REPLACE)
      {
	return m_desc + " " + action_strings[m_action] + " " + m_replace;
      }
    else
      {
	return m_desc + " " + action_strings[m_action];
      }
  }
}
