/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: PackageDependency.java,v 1.3 2001/11/01 02:12:54 samc Exp $
 */ 
package org.xoe.core.packages;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.util.DOMUtils;

public class PackageDependency 
  extends Dependency
{
  public static final String NAMESPACE = BaseInstaller.NAMESPACE + "/package";

  String  m_name    = null;
  String  m_version = null;
  String  m_minVer  = null;
  String  m_maxVer  = null;
  boolean m_minInc  = true;
  boolean m_maxInc  = true;

  public PackageDependency (Element depEl)
    throws PackageDescriptionException
  {
    super (depEl);
    m_name = depEl.getAttribute ("name");

    NodeList childNodes = depEl.getChildNodes ();

    for (int i = 0; childNodes != null && i < childNodes.getLength (); i++)
      {
	Node n = childNodes.item (i);

        if (n.getNodeType () == Node.ELEMENT_NODE)
          {
            String name = n.getNodeName ();
            if ("version".equals (name))
              {
                m_version = DOMUtils.getTextContent (n);
              }
            else if ("minimum-version".equals (name))
              {
                m_minVer = DOMUtils.getTextContent (n);
                String strMinInc = ((Element)n).getAttribute ("inclusive");
                m_minInc = strMinInc == null || "true".equals (strMinInc);
              }
            else if ("maximum-version".equals (name))
              {
                m_maxVer = DOMUtils.getTextContent (n);
                String strMaxInc = ((Element)n).getAttribute ("inclusive");
                m_maxInc = strMaxInc == null || "true".equals (strMaxInc);
              }
            else
              {
                throw new PackageDescriptionException ("unknown tag in package dependancy: " + name);
              }
          }
      }

    if (m_version != null)
      {
	setXPath (toXPath (m_name, m_version));
      }
    else
      {
	setXPath (toXPath (m_name, m_minVer, m_minInc, m_maxVer, m_maxInc));
      }
  }

  public Element getDepElement (Document ownerdoc)
  {
    Element depEl = super.getDepElement (ownerdoc);
    
    if (m_version != null)
      {
	Element versionEl = ownerdoc.createElement ("version");
	Text version = ownerdoc.createTextNode (m_version);
	versionEl.appendChild (version);
	depEl.appendChild (versionEl);
      }
    else
      {
	if (m_minVer != null)
	  {
	    Element minVersionEl = ownerdoc.createElement ("minimum-version");
	    Text minVersion = ownerdoc.createTextNode (m_minVer);
	    minVersionEl.appendChild (minVersion);
	    minVersionEl.setAttribute ("inclusive", m_minInc ? "true" : "false");
	    depEl.appendChild (minVersionEl);
	  }

	if (m_maxVer != null)
	  {
	    Element maxVersionEl = ownerdoc.createElement ("maximum-version");
	    Text maxVersion = ownerdoc.createTextNode (m_maxVer);
	    maxVersionEl.appendChild (maxVersion);
	    maxVersionEl.setAttribute ("inclusive", m_maxInc ? "true" : "false");
	    depEl.appendChild (maxVersionEl);
	  }
      }

    return depEl;
  }

  public PackageDependency (String name, String minVer, boolean minInc,
			   String maxVer, boolean maxInc)
  {
    super (NAMESPACE, name, toXPath (name, minVer, minInc, maxVer, maxInc));
    m_name = name;
    m_minVer = minVer;
    m_maxVer = maxVer;
    m_minInc = minInc;
    m_maxInc = maxInc;
  }

  public PackageDependency (String name, String version)
  {
    super (NAMESPACE, name, toXPath (name, version));
    m_name = name;
    m_version = version;
  }

  public PackageDependency (String name)
  { 
    this (name, null);
    m_name = name;
  }

  static String toXPath (String name, String exactVer)
  {
    StringBuffer result = new StringBuffer ("self::proviso[name='" + name + "']");
    if(exactVer != null)
      {
	result.append ("[eq-version(version, '" + exactVer + "')]");
      }
    return result.toString();
  }


  static String toXPath (String name, String minVer, boolean minInc,
			   String maxVer, boolean maxInc)
  {
    StringBuffer result = new StringBuffer ("self::proviso[name='" + name + "']");

    if(minVer != null)
      {
	if(minInc)
	  {
	    result.append ("[geq-version(version, '" + minVer + "')]");
	  }
	else
	  {
	    result.append ("[gt-version(version, '" + minVer + "')]");
	  }
      }

    if(maxVer != null)
      {
	if(maxInc)
	  {
	    result.append ("[leq-version(version, '" + maxVer + "')]");
	  }
	else
	  {
	    result.append ("[lt-version(version, '" + maxVer + "')]");
	  }
      }    

    return result.toString();
  }
}
