/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageIncompleteException.java,v 1.4 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

/**
 * Represents conflicts and unmet dependencies of a specific package
 */
public class PackageIncompleteException 
  extends PackageException
{
  private PackageDescription m_pkgDesc = null;
  
  private Dependency[] m_unmetDeps = new Dependency[0];
  private PackageDescription[] m_confPkgs = new PackageDescription[0];

  /**
   * Returns an array of unmet dependencies of the Incomplete Package
   */
  public Dependency[] getUnmetDependencies()
  {
    return m_unmetDeps;
  }

  /**
   * Returns an array of packages that conflict with the Incomplete Package
   */
  public PackageDescription[] getConflictingPackages()
  {
    return m_confPkgs;
  }

  /**
   * Returns the package that reported conflicts or unmet dependencies
   */
  public PackageDescription getIncompletePackage ()
  {
    return m_pkgDesc;
  }
  
  /**
   * Builds an exception out of a package description and two vectors containing
   * Dependency objects that were unmet and PackageDescriptions of packages that
   * conflicted with specified package
   */
  PackageIncompleteException (PackageDescription pkgDesc, Vector v1, Vector v2)
  {
    super (buildDescString (pkgDesc, v1, v2));

    m_pkgDesc = pkgDesc;

    Vector unmetDeps = new Vector();
    Vector confPkgs = new Vector();
    
    parseVector (v1, unmetDeps, confPkgs);
    parseVector (v2, unmetDeps, confPkgs);

    m_unmetDeps = (Dependency[]) unmetDeps.toArray (new Dependency[unmetDeps.size()]);
    m_confPkgs = (PackageDescription[]) confPkgs.toArray (new PackageDescription[confPkgs.size()]);
  }
  
  private static String buildDescString (PackageDescription pkgDesc, Vector v1, Vector v2)
  {
    Vector unmetDeps = new Vector ();
    Vector confPkgs = new Vector ();

    parseVector (v1, unmetDeps, confPkgs);
    parseVector (v2, unmetDeps, confPkgs);
 
    StringBuffer strb = new StringBuffer ();

    strb.append (pkgDesc + " could not be installed for the following reasons:\n");

    if (unmetDeps.size () > 0)
      {
	strb.append ("  The following dependencies were unmet:\n");
      }
    for (int i = 0; i < unmetDeps.size (); i++)
      {
	strb.append ("    " + unmetDeps.elementAt (i) + "\n");
      }

    if (confPkgs.size () > 0)
      {
	strb.append ("  The following installed packages conflict with " + pkgDesc + ":\n");
      }

    for (int i = 0; i < confPkgs.size (); i++)
      {
	strb.append ("    " + confPkgs.elementAt (i) + "\n");
      }

    return strb.toString ();
  }

  private static void parseVector (Vector v, Vector unmetDeps, Vector confPkgs)
  {
    if (v == null)
      {
	return;
      }
  
    for (int i = 0; i < v.size (); i++)
      {
	if (v.elementAt (i) instanceof PackageDescription)
	  {
	    confPkgs.addElement (v.elementAt (i));
	  }
	else if (v.elementAt (i) instanceof IPackageQuery)
	  {
	    //unmetDeps.addElement (v.elementAt (i));
	  }
      }
  }
}
