/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageState.java,v 1.2 2001/11/08 01:26:08 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

import org.xoe.core.util.Logger;


class PackageState 
  implements IPackageQueryHandler
{
  // package state is not to change
  public static final int PS_NO_PENDING = 1;

  // package is to be installed
  public static final int PS_INSTALL_PENDING = 2;

  // package is to be removed
  public static final int PS_REMOVE_PENDING = 4;

  // package is to replace package referenced by m_replace
  public static final int PS_NEW_REPLACE_PENDING = 8;
  
  // package is to be replaced by package referenced by m_replace
  public static final int PS_OLD_REPLACE_PENDING = 16;

  public static final int PS_NOT_INSTALLED = 0;

  public static final int PS_PRE_REGISTERED = 1;
  public static final int PS_REGISTERED = 2;

  //there is no PS_POST_REGISTERED because once that happens the package is installed
  //public static final int PS_POST_REGISTERED = 3;

  public static final int PS_INSTALLED = 3;

  public static final int PS_PRE_UNREGISTERED = 4;
  public static final int PS_UNREGISTERED = 5;

  // there is no POST_UNREGISTERED, because once that happens the
  // package is PS_NOT_INSTALLED...
  //public static final int PS_POST_UNREGISTERED = 6;

  private XoePackage m_xoePackage = null;
  private PackageState m_replace = null;

  private int m_pending = PS_NO_PENDING;
  private int m_state = PS_NOT_INSTALLED;

  public PackageState (XoePackage xoePackage)
  {
    m_xoePackage = xoePackage;
    m_pending = PS_INSTALL_PENDING;
  }

  public PackageState (XoePackage newPackage, PackageState oldState)
  {
    m_xoePackage = newPackage;
    m_replace = oldState;
    m_pending = PS_NEW_REPLACE_PENDING;
    oldState.setReplaceBy (this);
  }

  public void setReplaceBy (PackageState newState)
  {
    m_pending = PS_OLD_REPLACE_PENDING;
    m_replace = newState;
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  {
    performQuery (in_q, out_res, PS_NO_PENDING);
  }

  public void performQuery (IPackageQuery in_q, Vector out_res, int type)
  {
    if (m_pending == PS_NO_PENDING && m_state == PS_NOT_INSTALLED)
      {
	return;
      }

    if ((type & m_pending) > 0)
      {
	m_xoePackage.performQuery (in_q, out_res);
      }
  }

  public int getPending ()
  {
    return m_pending;
  }

  public int getState ()
  {
    return m_state;
  }

  public void setPending (int p)
  {
    m_pending = p;
  }

  public void setState (int s)
  {
    m_state = s;
  }

  public XoePackage getXoePackage ()
  {
    return m_xoePackage;
  }

  public void verify ()
    throws PackageException
  {
    if ((m_pending == PS_INSTALL_PENDING || m_pending == PS_NEW_REPLACE_PENDING)
	&& m_state == PS_NOT_INSTALLED)
      {
	Core.verifyPackage (m_xoePackage);
	m_state = PS_PRE_REGISTERED;
      }
    else if ((m_pending == PS_REMOVE_PENDING || m_pending == PS_OLD_REPLACE_PENDING)
	     && m_state == PS_INSTALLED)
      {
	m_state = PS_PRE_UNREGISTERED;
      }
  }

  public void backOut ()
  {
    if ((m_pending == PS_INSTALL_PENDING || m_pending == PS_NEW_REPLACE_PENDING)
	&& m_state == PS_REGISTERED)
      {
	Core.uninstallPackage (m_xoePackage);
	m_state = PS_UNREGISTERED;
      }
    else
      {
	Logger.log (this, "Can't back out " + m_xoePackage);
      }
    // Can't back out an uninstalled package
    // but this is alright, since "nothing can go wrong"
    // with uninstalling packages...
  }

  public void doCommit ()
    throws PackageException
  {
    if ((m_pending == PS_INSTALL_PENDING || m_pending == PS_NEW_REPLACE_PENDING)
	&& m_state == PS_PRE_REGISTERED)
      {
	Core.installPackage (m_xoePackage);
	m_state = PS_REGISTERED;
      }
    else if ((m_pending == PS_REMOVE_PENDING || m_pending == PS_OLD_REPLACE_PENDING)
	     && m_state == PS_PRE_UNREGISTERED)
      {
	Core.uninstallPackage (m_xoePackage);
	m_state = PS_UNREGISTERED;
      }
  }

  public void postCommit ()
  {
    if ((m_pending == PS_INSTALL_PENDING || m_pending == PS_NEW_REPLACE_PENDING)
	&& m_state == PS_REGISTERED)
      {
	m_state = PS_INSTALLED;
	m_pending = PS_NO_PENDING;
      }
    else if ((m_pending == PS_REMOVE_PENDING || m_pending == PS_OLD_REPLACE_PENDING)
	     && m_state == PS_PRE_UNREGISTERED)
      {
	m_state = PS_NOT_INSTALLED;
	m_pending = PS_NO_PENDING;
      }
  }
}
