/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageVersion.java,v 1.2 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

/**
 * Represents a package version. Package versions have the form
 * M[.m][-R] where M is the major number, m is the minor number,
 * and R is the revision. M and m must be positive integers, R
 * is an opaque string.
 */

public class PackageVersion
{
  private int major = -1;
  private int minor = -1;
  private String revision = null;

  public PackageVersion (String strVer)
  {
    if (strVer.indexOf ("-") != -1)
      {
	setRevision (strVer.substring (strVer.indexOf ("-") + 1));
	strVer = strVer.substring (0, strVer.indexOf ("-"));
      }

    if (strVer.indexOf (".") != -1)
      {
	setMinor (strVer.substring (strVer.indexOf (".") + 1));
	strVer = strVer.substring (0, strVer.indexOf ("."));
      }
    if (strVer.length () > 0)
      {
	setMajor (strVer);
      }
  }

  public void setMajor (String str)
  {
    major = Integer.parseInt (str);
  }

  public void setMinor (String str)
  {
    minor = Integer.parseInt (str);
  }

  public void setRevision (String str)
  {
    revision = str;
  }

  public int getMajor ()
  {
    if (major == -1)
      {
	return 0;
      }
    return major;
  }

  public int getMinor ()
  {
    if (minor == -1)
      {
	return 0;
      }
    return minor;
  }

  public String getRevision ()
  {
    return revision;
  }

  public String toString ()
  {
    return ((major > -1) ? ("" + major) : "") 
           + ((minor > -1) ? ("." + minor) : "") 
           + ((revision != null) ? ("-" + revision) : "");
  }

  public boolean leq (PackageVersion rhs)
  {
    return (getMajor () == rhs.getMajor () && getMinor () <= rhs.getMinor ()) || (getMajor () < rhs.getMajor ());
  }

  public boolean geq (PackageVersion rhs)
  {
    return rhs.leq (this);
  }

  public boolean lt (PackageVersion rhs)
  {
    return (getMajor () == rhs.getMajor () && getMinor () < rhs.getMinor ()) || (getMajor () < rhs.getMajor ());
  }

  public boolean gt (PackageVersion rhs)
  {
    return rhs.lt (this);
  }

  public boolean eq (PackageVersion rhs)
  { 
    return (getMajor () == rhs.getMajor () 
            && getMinor () == rhs.getMinor () 
            && ((getRevision () == null && rhs.getRevision () == null)
	       || getRevision ().equals (rhs.getRevision ()) ));
  }
}
