/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: PackageXPathFunctionProvider.java,v 1.1 2001/10/22 20:19:04 colin Exp $
 */
package org.xoe.core.packages;

import java.util.Hashtable;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.FunctionProvider;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathFunction;


public class PackageXPathFunctionProvider 
  implements FunctionProvider
{
  private Hashtable m_hash = new Hashtable ();
  public static final PackageXPathFunctionProvider singleton = new PackageXPathFunctionProvider ();

  // boolean max-version(string version, string max)
  // returns true if 'version' is less than or equal to 'max'
  class geqVersion 
    extends XPathFunction
  {
    public geqVersion ()
    {
      expected = 2;
      name = "geq-version";
    }

    public Object applyN (Object[] args)
    {
      String lhs = XPath.stringValue (args[0]);
      String rhs = XPath.stringValue (args[1]);
      
      return (new PackageVersion (lhs)).geq (new PackageVersion (rhs)) ? Boolean.TRUE : Boolean.FALSE;
    }
  }

  // boolean min-version(string version, string min)
  // returns true if 'version' is greater than or equal to 'min'
  class leqVersion 
    extends XPathFunction
  {
    public leqVersion ()
    {
      expected = 2;
      name = "leq-version";
    }

    public Object applyN (Object[] args)
    {
      String lhs = XPath.stringValue (args[0]);
      String rhs = XPath.stringValue (args[1]);

      return (new PackageVersion (lhs)).leq (new PackageVersion (rhs)) ? Boolean.TRUE : Boolean.FALSE;
    }
  }


  // boolean max-version(string version, string max)
  // returns true if 'version' is less than or equal to 'max'
  class gtVersion 
    extends XPathFunction
  {
    public gtVersion ()
    {
      expected = 2;
      name = "gt-version";
    }

    public Object applyN (Object[] args)
    {
      String lhs = XPath.stringValue (args[0]);
      String rhs = XPath.stringValue (args[1]);
      
      return (new PackageVersion (lhs)).gt (new PackageVersion (rhs)) ? Boolean.TRUE : Boolean.FALSE;
    }
  }

  // boolean min-version(string version, string min)
  // returns true if 'version' is greater than or equal to 'min'
  class ltVersion 
    extends XPathFunction
  {
    public ltVersion ()
    {
      expected = 2;
      name = "lt-version";
    }

    public Object applyN (Object[] args)
    {
      String lhs = XPath.stringValue (args[0]);
      String rhs = XPath.stringValue (args[1]);

      return (new PackageVersion (lhs)).lt (new PackageVersion (rhs)) ? Boolean.TRUE : Boolean.FALSE;
    }
  }


  // boolean exact-version(string versionA, string versionB)
  // returns true if 'versionA' is equal to 'versionB'
  class eqVersion 
    extends XPathFunction
  {
    public eqVersion()
    {
      expected = 2;
      name = "eq-version";
    }

    public Object applyN(Object[] args)
    {
      String lhs = XPath.stringValue (args[0]);
      String rhs = XPath.stringValue (args[1]);

      return (new PackageVersion (lhs)).eq (new PackageVersion (rhs)) ? Boolean.TRUE : Boolean.FALSE;
    }
  }

  // boolean exact-version(string versionA, string versionB)
  // returns false if 'versionA' is equal to 'versionB'
  class neqVersion 
    extends XPathFunction
  {
    public neqVersion ()
    {
      expected = 2;
      name = "neq-version";
    }

    public Object applyN (Object[] args)
    {
      String lhs = XPath.stringValue (args[0]);
      String rhs = XPath.stringValue (args[1]);

      return (new PackageVersion (lhs)).eq (new PackageVersion (rhs)) ? Boolean.FALSE : Boolean.TRUE;
    }
  }

  private PackageXPathFunctionProvider ()
  {
    putFunctionReference (new leqVersion ());
    putFunctionReference (new geqVersion ());
    putFunctionReference (new ltVersion ());
    putFunctionReference (new gtVersion ());
    putFunctionReference (new eqVersion ());
    putFunctionReference (new neqVersion ());
  }

  public void putFunctionReference (XPathFunction func)
  {
    m_hash.put (func.getName (), func);
  }

  public XPathFunction getFunction (String fname)
  {
    return (XPathFunction) m_hash.get (fname);
  }

  public static void main (String[] args)
    throws Exception
  {
    DOMlet doc = DOMlet.create((
				"<packages >" + 
				"  <package version='2.7' />" +
				"  <package version='2.7-ew' />" +
				"  <package version='2.8' />" +
				"  <package version='2.11' />" +
				"  <package version='2.3-ew' />" +
				"  <package version='2' />" +
				"  <package version='2.0' />" +
				"  <package version='1.7-ew' />" +
				"  <package version='.7' />" +
				"  <package version='0.7' />" +
				"</packages>"
				).toCharArray(), "PackageXPathFunctionProvider.main");

    XPathContext ctx = new XPathContext ();
    ctx.pushNode (doc.getDocumentElement ());
    ctx.setFunctionProvider (new PackageXPathFunctionProvider());
    test ("//package[leq-version(@version, '2.5')]", ctx);
    test ("//package[eq-version(@version, '.7')]", ctx);
    test ("//package[eq-version(@version, '2.7')]", ctx);
    test ("//package[geq-version(@version, '2.0')]", ctx);
    test ("//package[leq-version(@version, '2.4')][geq-version(@version, '1.5')]", ctx);
    test ("//package[leq-version(@version, '2.8')][geq-version(@version, '2')]", ctx);
    test ("//package[lt-version(@version, '2.8')][gt-version(@version, '2')]", ctx);
  }

  public static void test (String strx, XPathContext ctx)
  {
    System.err.println ("Running query: " + strx);
    NodeList nl = XPath.findNodes (ctx, strx);

    for (int i = 0;  i < nl.getLength ();  i++)
      {
	System.err.println (XMLWriter.getXML (nl.item (i)));
      }
  }
}
