/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Proviso.java,v 1.2 2001/12/13 23:51:18 erik Exp $
 */
package org.xoe.core.packages;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.XMLWriter;

/**
 * Represents a logical piece of functionality provided in a package (ie, a Service).
 * Provisos have a local name (the 'name' attribute), a namespace (the 'ns' attribute),
 * and optional metadata (any childnodes). The namespace determines how the name and
 * metadata are interpreted.
 */
public class Proviso
{
  String namespace, localName;
  Node metadata;

  public String toString()
  {
    return "Proviso: ns='" + namespace + "' name='" + localName + "'";
  }

  public Proviso (Element node)
  {
    Document doc = new XDocument();

    this.namespace = node.getAttribute ("ns") ;
    this.localName = node.getAttribute ("name");
    this.metadata = doc.createDocumentFragment ();
    NodeList nl = node.getChildNodes ();
    for (int i = 0; i < nl.getLength(); i++)
      {
	metadata.appendChild (doc.importNode (nl.item (i), true));
      }
  }

  public Proviso (String namespace, String localName, Node metadata)
  {
    this.namespace = namespace;
    this.localName = localName;
    this.metadata = metadata;
  }

  public Proviso (String namespace, String localName)
  {
    this (namespace, localName, null);
  }

  public String getNamespaceURI ()
  {
    return namespace;
  }

  public String getLocalName ()
  {
    return localName;
  }

  public Node getMetadata ()
  {
    return metadata;
  }

  public boolean equals (Object o)
  {
    if (!(o instanceof Proviso)) return false;
    
    Proviso other = (Proviso) o;

    return (namespace.equals (other.namespace)
	    && localName.equals (other.localName)
	    && ((metadata == null && other.metadata == null) 
		|| XMLWriter.getXML (metadata).equals (XMLWriter.getXML (other.metadata))));
  }
}
