/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: SearchResults.java,v 1.2 2001/10/26 03:22:06 erik Exp $
 */
package org.xoe.core.packages;

public class SearchResults
{
  InstallOption[] m_installable;
  UninstallablePackage[] m_uninstallable;
  PackageDescription[] m_installed;

  protected SearchResults (InstallOption[] installable,
			   UninstallablePackage[] uninstallable,
			   PackageDescription[] installed)
  {
    m_installable = installable;
    m_uninstallable = uninstallable;
    m_installed = installed;
  }

  // Each InstallOption represents a package that matched the query and
  // all of the other packages needed to install it.
  //
  // Pass one of these options to PackageManager.commit() to install it.
  public InstallOption[] getInstallablePackages()
  {
    return m_installable;
  }

  // Each UninstallablePackage represents a package that matched the query
  // but could not be installed for some reason.
  public UninstallablePackage[] getUninstallablePackages()
  {
    return m_uninstallable;
  }

  // Each PackageDescription represents a package that matched the query
  // and is already installed.
  public PackageDescription[] getInstalledPackages()
  {
    return m_installed;
  }

  public String toString()
  {
    StringBuffer strb = new StringBuffer();

    strb.append ("SearchResults:\n{\n  installable\n  {\n");
    strb.append (InstallOption.toString (m_installable));

    strb.append("  }\n  uninstallable\n  {\n");
    for(int i = 0; m_uninstallable != null && i < m_uninstallable.length; i++)
      {
	strb.append ("    " + m_uninstallable[i].toString() + "\n");
      }

    strb.append("  }\n  installed\n  {\n");
    for(int i = 0; m_installed != null && i < m_installed.length; i++)
      {
	strb.append("    " + m_installed[i].toString() + "\n");
      }

    strb.append("  }\n}\n");
    return strb.toString();
  }
}
