/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: ServiceDependency.java,v 1.1 2001/10/22 20:19:04 colin Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.util.XMLWriter;


public class ServiceDependency 
  extends Dependency
{
  String m_func = null;
  String m_iface = null;
  String m_features[][] = null;

  public ServiceDependency (Element depEl)
    throws PackageDescriptionException
  {
    super (depEl);

    Vector vFeats = new Vector();

    NodeList childNodes = depEl.getChildNodes ();

    for (int i = 0; childNodes != null && i < childNodes.getLength (); i++)
      {
	Node n = childNodes.item(i);

	if ("functionality".equals (n.getNodeName ()))
	  {
	    Node t = childNodes.item (i).getFirstChild ();
	    if(t.getNodeType () == Node.TEXT_NODE)
	      {
		m_func = ( (Text)t).getData ();
	      }
	  }
	else if ("implements".equals (n.getNodeName ()))
	  {
	    Node t = childNodes.item (i).getFirstChild ();
	    if(t.getNodeType () == Node.TEXT_NODE)
	      {
		m_iface = ( (Text)t).getData ();
	      }
	  }
	else if ("feature".equals (n.getNodeName ()) && n.getNodeType() == Node.ELEMENT_NODE)
	  {
	    String feat[] = {((Element)n).getAttribute ("name"), ((Element)n).getAttribute ("value")};
	    vFeats.addElement (feat);
	  }
      }

    m_features = (String[][])vFeats.toArray(new String[vFeats.size ()][]);
    if(m_func == null || m_iface == null)
      {
	throw new PackageDescriptionException ("Invalid Service Dependency:\n" + XMLWriter.getXML (depEl));
      }
    setXPath(toXPath(m_func, m_iface, m_features));
  }

  public Element getDepElement (Document ownerdoc)
  {
    Element depEl = super.getDepElement (ownerdoc);
    Element funcEl = ownerdoc.createElement ("functionality");
    Text func = ownerdoc.createTextNode (m_func);
    Element ifaceEl = ownerdoc.createElement ("implements");
    Text iface = ownerdoc.createTextNode (m_iface);
    funcEl.appendChild (func);
    ifaceEl.appendChild (iface);

    depEl.appendChild (funcEl);
    depEl.appendChild (ifaceEl);

    for(int i = 0; i < m_features.length; i++)
      {
	Element featEl = ownerdoc.createElement ("feature");
	featEl.setAttribute ("name", m_features[i][0]);
	featEl.setAttribute ("value", m_features[i][1]);
	depEl.appendChild (featEl);
      }

    return depEl;
  }

  public ServiceDependency (String functionality, String iface, String[][] features)
  {
    super (ServiceInstaller.NAMESPACE, functionality, toXPath (functionality, iface, features));
    m_func = functionality;
    m_iface = iface;
    m_features = features;
  }

  static String toXPath (String functionality, String iface, String[][] features)
  {
    StringBuffer sxpath = new StringBuffer ();

    if (functionality != null || iface != null || features != null)
      {
	if (functionality != null)
	  {
	    sxpath.append ("functionality='" + functionality + "'");
	  }
	if (iface != null)
	  {
	    if (functionality != null)
	      sxpath.append (" and ");
	    
	    sxpath.append ("implements='" + iface + "'");
	  }
	if (features != null && features.length > 0)
	  {
	    if (functionality != null || iface != null )
	      sxpath.append (" and ");
	    
	    for (int i = 0; i < features.length; i++)
	      {
		sxpath.append ("feature[@name = '" + features[i][0] + 
			               "' and @value = '" + features[i][1] + "']");
		if (i+1 < features.length)
		  sxpath.append (" and ");
	      }
	  }
      }

    return sxpath.toString ();
  }
}
