/*
 * Copyright (c) 2000 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: StashPackageQueryHandler.java,v 1.2 2001/11/14 01:16:21 samc Exp $
 */
package org.xoe.core.packages;

import java.net.URL;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.util.Vector;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.NodeList;
import org.w3c.dom.Document;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.content.StashException;
import org.xoe.core.content.StashDirectory;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;


/**
 * Simple IPackageQueryHandler that searchs against a Vector
 * of PackageDescriptions.
 */
public class StashPackageQueryHandler
  implements IPackageQueryHandler
{
  // Vector of package descriptions to query against.
  private Vector m_vPackages = new Vector ();

  public StashPackageQueryHandler ()
  {
  }

  public StashPackageQueryHandler (URL url)
    throws StashException
  {
    addPackageDirectory (url);
  }

  public void addPackage (PackageDescription desc)
  {
    m_vPackages.addElement(desc);
  }
  
  public void removePackage (PackageDescription desc)
  {
    m_vPackages.removeElement(desc);
  }
  
  /**
   * Add all the XOE files in the given stash direcotry to the 
   * QueryHandler.
   */
  public void addPackageDirectory (URL url)
    throws StashException
  {
    long stime = System.currentTimeMillis();
    ContentElement dir = Stash.fetch (url);
    if (dir == null || !(dir instanceof StashDirectory))
      {
        throw new StashException ("url does not point to a stash directory: " + url);
      }
    NodeList nl = XPath.findNodes (dir, "*[ends-with(name(),'.xoe')]");
    PackageDescription desc;
    for (int i = 0; i < nl.getLength (); i++)
      {
        ContentElement elem = (ContentElement) nl.item (i);
        try
          {
            ZipInputStream stream = new ZipInputStream (elem.getInputStream ());
            ZipEntry ze;
            while ((ze = stream.getNextEntry ()) != null)
              {
                if ("package.xml".equals (ze.getName ()))
                  {
                    byte data[] = new byte[(int)ze.getSize ()];
                    stream.read (data, 0, data.length);
                    desc = new PackageDescription (new ByteArrayInputStream (data));
                    desc.setSuggestedURL (elem.getAttribute ("url"));
                    m_vPackages.addElement (desc);
                  }
                stream.closeEntry ();
              }
          }
        catch (PackageDescriptionException e)
          {
            Logger.log (this, e);
          }
        catch (IOException e)
          {
            Logger.log (this, e);
          }
      }
    Logger.log (this, "adding directory to StashPackageQueryHandler took: " + (System.currentTimeMillis() - stime) + " ms");
  }

  /**
   * Performs query on the vector of PackageDescriptions.
   */
  public void performQuery (IPackageQuery in_q, Vector out_res)
  {
    String strx = in_q.toXPath ();

    for (int i = 0; i < m_vPackages.size (); i++)
      {
	((PackageDescription) m_vPackages.elementAt (i)).performQuery (in_q, out_res);
      }
  }

  /**
   * Debugging method: dumps all PackageDescriptions to System.err.
   */
  public void dump ()
  {
    System.err.println ("--- begin StashPackageQueryHandler dump:");
    for (int i = 0; i < m_vPackages.size (); i++)
      {
	System.err.print (i + ": ");
	PackageDescription pd = (PackageDescription) m_vPackages.elementAt (i);
        System.err.println (pd);
      }
    System.err.println ("--- end StashPackageQueryHandler dump");
  }
}
