/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: UninstallContext.java,v 1.2 2001/11/07 01:19:31 erik Exp $
 */
package org.xoe.core.packages;

import java.util.Hashtable;
import java.util.Vector;

import org.xoe.core.util.Logger;


public class UninstallContext 
  implements IPackageQueryHandler
{
  private Hashtable m_mapPackages = null;
  private PackageDescription m_primary = null;

  public static final int MAX_DEPTH = 20;

  private UninstallContext (UninstallContext orig)
  {
    m_mapPackages = (Hashtable)orig.m_mapPackages.clone ();
    m_primary = orig.m_primary;
  }

  public UninstallContext ()
  {
    m_mapPackages = new Hashtable ();
  }

  public UninstallContext (PackageDescription desc)
  {
    this ();
    addPackage (desc);
  }

  public Object clone ()
  {
    return new UninstallContext (this);
  }

  public void setPrimary (PackageDescription desc)
  {
    m_primary = desc;
  }

  public PackageDescription getPrimary ()
  {
    return m_primary;
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  {

    PackageDescription[] descs = (PackageDescription[]) PackageUtils.getValues (m_mapPackages);

    for (int i = 0; i < descs.length; i++)
      {
	descs[i].performQuery(in_q, out_res);
      }
  }

  public PackageDescription[] getPackages ()
  {
    return (PackageDescription[]) PackageUtils.getValues (m_mapPackages);
  }

  public void addPackage (PackageDescription pkg)
  {
    if (m_primary == null)
      {
	m_primary = pkg;
      }
    m_mapPackages.put (pkg.getName (), pkg );
  }

  public boolean contains (PackageDescription pkg)
  {
    PackageDescription oldPkg = (PackageDescription)m_mapPackages.get (pkg.getName ());
    if (oldPkg != null && oldPkg.getVersion ().eq (pkg.getVersion ()))
      {
	return true;
      }
    
    return false;
  }

  public static UninstallContext getCandidateContext (IPackageQuery q)
  {
    Vector v = new Vector ();
    PackageManager.performQueryLocally(q, v);
    if (v.size () > 0)
      {
	return getCandidateContext ((PackageDescription)v.elementAt (0));
      }
    return null;
  }

  public static UninstallContext getCandidateContext (PackageDescription desc)
  {
    UninstallContext ctx = new UninstallContext (desc);

    PackageDescription[] pkgs = PackageManager.getInstalledPackages ();

    while (true)
      {
	for (int i = 0; i < pkgs.length; i++)
	  {
	    try
	      {
		
		if (ctx.contains (pkgs[i]))
		  { // already marked for uninstall
		    continue;
		  }
		
		boolean restart = false;
		
		IPackageQuery[] deps = pkgs[i].getDependencies ();
		
		for (int j = 0; j < deps.length; j++)
		  { 
		    // what we want to know is if _every_ package matching the dep
		    // is part of this uninstall set
		    
		    // searches for installed packages matching query
		    Vector v1 = new Vector ();
		    PackageManager.performQueryLocally (deps[j], v1);
		    
		    // searches the uninstall set
		    Vector v2 = new Vector ();
		    ctx.performQuery (deps[j], v2);
		    
		    // since all packages in the uninstall set came from the pkg-mgr,
		    // if they're the same size they're the same set...
		    if (v1.size () == v2.size ())
		      { // gotta uninstall this one too
			
			ctx.addPackage (pkgs[i]);
			
			// restart the process (at the top of the while loop)
			restart = true;
			break;
		      }
		  }
		
		if (restart)
		  {
		    break;
		  }
	      }
	    finally
	      {
		if (i + 1 == pkgs.length)
		  {
		    return ctx;
		  }
	      }
	  }
      }
  }

  public static void dumpContextVector (Vector contexts)
  {
    for (int i = 0; i < contexts.size (); i++)
      {
	UninstallContext ctx = (UninstallContext)contexts.elementAt (i);
	
	System.err.println ("Context " + i + " (" + ctx.getPackages ().length + " packages) :");

	System.err.println (ctx.toString ());
      }    
  }

  public String toString ()
  {
    StringBuffer ret = new StringBuffer ();
    Object[] descs = getPackages ();
    
    for (int j = 0; j < descs.length; j++)
      {
	if (j > 0)
	  {
	    ret.append ("\n");
	  }
	ret.append (descs[j]);
      }
    return ret.toString ();
  }
}
