/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: UninstallablePackage.java,v 1.2 2001/11/01 22:57:10 erik Exp $
 */
package org.xoe.core.packages;

public class UninstallablePackage
{
  private PackageDescription m_target;
  private PackageDescription m_badPkg;
  private String m_reason;

  protected UninstallablePackage (PackageDescription target,
				  PackageDescription badPkg, 
                                  String reason)
  {
    m_target = target;
    m_badPkg = badPkg;
    m_reason = reason;
  }

  protected UninstallablePackage (PackageDescription target, PackageIncompleteException ex)
  {
    m_target = target;
    m_badPkg = ex.getIncompletePackage();
    StringBuffer strb = new StringBuffer ();

    Dependency[] unmetDeps = ex.getUnmetDependencies();

    
    if (unmetDeps.length > 0)
      {
	strb.append ("  The following dependencies were unmet:\n");
      }
    for (int i = 0; i < unmetDeps.length; i++)
      {
	strb.append ("    " + unmetDeps [i] + "\n");
      }

    PackageDescription[] confPkgs = ex.getConflictingPackages ();
    if (confPkgs.length > 0)
      {
	strb.append ("  " + m_target + " depends on " + m_badPkg + ", which conflicts with:\n");
      }

    for (int i = 0; i < confPkgs.length ; i++)
      {
	strb.append ("    " + confPkgs[i] + "\n");
      }

    m_reason = strb.toString();
  }

  public PackageDescription getTargetPackage ()
  {
    return m_target;
  }

  public PackageDescription getBadPackage ()
  {
    return m_badPkg;
  }

  public String getReasonString ()
  {
    return m_reason;
  }

  public String toString ()
  {
    if (m_badPkg == null || m_badPkg == m_target)
      {
	return "Could not install package " + m_target + ".\nReason: " + m_reason + "\n";
      }
    else
      {
	return "Could not install package " + m_target + " because it depends on " + 
	  m_badPkg + ". Reason: " + m_reason + "\n";
      }
  }
}

