/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XoeClassLoader.java,v 1.2 2001/11/08 20:17:39 tim Exp $
 */
package org.xoe.core.packages;

import java.util.Vector;

import org.xoe.core.util.Logger;


public class XoeClassLoader 
  extends ClassLoader
{
  private Vector loaders = new Vector ();

  protected XoePackage xoePackage;

  public String toString ()
  {
    return super.toString () + " {pkg = " + xoePackage + "}";
  }

  protected XoeClassLoader (XoePackage xoePackage)
  {
    super ();
    this.xoePackage = xoePackage;
  }

  public Class addClass (byte[] buf, int start, int len)
  {
    return (defineClass (null, buf, 0, buf.length));
  }

  public Class loadClass (String name, boolean resolve) throws ClassNotFoundException
  { 
    Class cls = null;
    boolean system = true;

    // If it looks like this might be a system class, try there first to keep things
    // fast.  If this fails we just continue
    if (name.startsWith("java.") || name.startsWith("kaffe."))
      {
	try
	  {
	    cls = findSystemClass (name);
	  }
	catch (ClassNotFoundException _)
	  {
	    system = false;
	  }
      }

    // Try XOE class loaders
    cls = findXoeClass (name);

    // Now try system
    if (cls == null)
      {
	if (system)
	  {
	    cls = findSystemClass (name);
	  }
	else
	  {
	    throw new ClassNotFoundException (name);
	  }
      }

    // Resolve and return
    if (resolve)
      {
	resolveClass(cls);
      }
    return (cls);
  }

  private Class findXoeClass (String name)
  {
    // Look in me
    Class cls = findLoadedClass (name);
    if (cls != null)
      {
	return (cls);
      }

    // Then look for it in any references loaders
    synchronized (loaders)
    {
      for (int i = loaders.size () - 1; i >= 0; i--)
	{
	  cls = ((XoeClassLoader) loaders.elementAt (i)).findXoeClass (name);
	  if (cls != null)
	    {
	      return (cls);
	    }
	}
    }
    return (null);
  }

  /**
   * Add a new classloader to the class loaders we inherit.  This allows us
   * to use classes loaded in different applications and packages.  We wrap
   * this in a reference so it will magically go away if we stop using classes
   * the the loader.
   */
  public void addClassLoader (ClassLoader newloader)
  {
    synchronized (loaders)
      {
        for (int i = loaders.size () - 1; i >= 0; i--)
	  {
      	    Object loader = loaders.elementAt (i);
            if (loader == newloader)
	      {
		// already got it...
		return;
	      }
          }
      }
    loaders.addElement (newloader);
  }

  /**
   * Remove a reference to the given loader.
   */
  public void removeClassLoader (ClassLoader newloader)
  {
    synchronized (loaders)
      {
        for (int i = loaders.size () - 1; i >= 0; i--)
          {
            Object loader = loaders.elementAt (i);
            if (loader == newloader)
              {
                loaders.removeElementAt (i);
                break;
              }
          }
      }
  }

  public XoePackage getPackage () 
  {
    return xoePackage;
  }

  void dump ()
  {
    for (int i = loaders.size () - 1; i >= 0; i--)
      {
        Logger.log (this, ((XoeClassLoader)loaders.elementAt (i)).xoePackage.getName ());
      }
  }
}
