/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: FileURLConnection.java,v 1.2 2001/10/24 20:52:26 samc Exp $
 */ 
package org.xoe.core.protocol;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringBufferInputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Arrays;

import kaffe.io.AccessibleBIStream;
import kaffe.net.www.protocol.BasicURLConnection;

import org.xoe.core.Constants;
import org.xoe.core.util.Logger;


/*
 * FileURLConnection - PocketLinux URL handler.
 */
public class FileURLConnection
  extends BasicURLConnection
  implements Constants
{
  static final String FILE_IMAGE = STASH_PROTOCOL + ":/data/images/file.gif";
  static final String DIR_IMAGE = STASH_PROTOCOL + ":/data/images/dir.gif";

  InputStream file;


  // -------------------------------------------------------------------------
  // constructors

  public FileURLConnection (URL url) 
  {
    super (url);
  }


  // -------------------------------------------------------------------------
  // org.xoe.core.protocol.FileURLConnection

  protected String buildIndex (File f) 
  { 
    String [] list = f.list ();
    Arrays.sort (list);
    String path =  url.getFile ();
    String preURL = getURLPrefix () + path;
    if (path.length () > 1)
      {
        preURL += "/";
      }

    StringBuffer index = new StringBuffer();
    index.append ("<html xmlns='")
         .append (XHTML_NAMESPACE)
         .append ("'><head><title>&nbsp;")
         .append (preURL)
         .append ("</title></head><body><b>&nbsp;Directory listing of ")
         .append (path)
         .append ("</b>");


    String parentPath = getParentPath ();
    if (parentPath != null)
      {
        index.append ("<br /><a href='")
             .append (getURLPrefix ())
             .append (parentPath)
             .append ("'>&nbsp;up to a higher level directory</a>");
      }

    for (int i=0; i<list.length; i++) 
      {
        index.append ("<br /><a href='")
             .append (preURL)
             .append (list[i])
             .append ("'> <img src='")
             .append ((new File (getFilePath () + "/" + list[i])).isDirectory () ? DIR_IMAGE : FILE_IMAGE)
             .append("' border='0' />&nbsp;")
             .append(list[i])
             .append("</a>");
      }    
    index.append("</body></html>");
    return (index.toString());
  }

  protected String getURLPrefix ()
  {
    return ("file:");
  }

  protected String getParentPath ()
  {
    String path = url.getFile ();
    int idx = path.lastIndexOf (File.separator);
    if (idx != -1)
      {
	if (path.lastIndexOf (File.separator) == (path.length ()-1))
	{
	  path = path.substring (0, path.length ()-1);
	}
	idx = path.lastIndexOf (File.separator);
	if (idx != -1)
	  {
	    return (idx == 0 ? "/" : path.substring (0, idx)); 
	  }
      }
    return (null);
  } 

  protected String getFilePath ()
  {
    return (url.getFile ());
  } 

  // -------------------------------------------------------------------------
  // kaffe.net.www.protocol.file.FileURLConnection

  public void connect () 
    throws IOException 
  {
    if (connected == true) 
      {
	return;
      }

    String fn = getFilePath ();
    if ((File.separatorChar == '\\') && (fn.indexOf('/') >= 0))
      {
	// we are on a DOS-like filesystem, replace URL slashes
	fn = fn.replace ('/', File.separatorChar);
      }

    File fl = new File (fn);
    if (fl.isDirectory ()) 
      {
        String directoryIndex = buildIndex (fl);
        headersValue[ContentType] = "text/xml";
        headersValue[ContentEncoding] = "text/xml";
        headersValue[ContentLength] = Long.toString (directoryIndex.length ());
        file = new StringBufferInputStream (directoryIndex);
      }
    else 
      {
        setContentTypeFromName ();
        headersValue[ContentLength] = Long.toString (fl.length ());
        file = new AccessibleBIStream (new FileInputStream (fl));
      }
    connected = true;
  }

  public InputStream getInputStream () 
    throws IOException 
  {
    try 
      {
	connect ();
	return file;
      }
    catch (FileNotFoundException _) 
      {
      }
    catch (IOException ioe) 
      {
        Logger.log (this, ioe);
      }
    return null;
  }

}
