/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: JarURLConnection.java,v 1.1 2001/10/21 04:10:52 colin Exp $
 */ 
package org.xoe.core.protocol;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.zip.ZipFile;

import kaffe.net.www.protocol.zip.ZipURLConnection;

import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;


public class JarURLConnection
  extends ZipURLConnection
{
  protected String type;

  private static String cachedURL;
  private static ZipFile cachedZip;

  public JarURLConnection (URL url)
    throws MalformedURLException
  {
    super(url);
  }

  protected ZipFile getUnderlyingZipFile (URL url)
    throws IOException
  {
    ContentElement elem = ContentCache.fetch (url);
    if (elem == null)
      {
	throw new IOException ("not found: " + url);
      }
    // Rewrite URL incase we had an alias involved
    String rurl = elem.getAttribute ("url");
    this.url = new URL ("jar:" + rurl + "!/" + zipEntryName);

    synchronized (JarURLConnection.class)
      {
	if (rurl.equals (cachedURL))
	  {
	    return (cachedZip);
	  }
      }

    InputStream in = elem.getInputStream ();
    File f = getUnderlyingFileFromStream (in);
    in.close ();
    ZipFile zf = new ZipFile (f);
    f.delete ();

    synchronized (JarURLConnection.class)
      {
	cachedURL = rurl;
	cachedZip = zf;
      }

    return (zf);
  }


  public void connect () 
    throws IOException 
  {
    // Already connected?
    if (connected) 
      {
	return;
      }   

    super.connect ();

    if (zipEntry != null)
      {
	type = getContentTypeFor (zipEntryName);
      }
  }


  public String getContentEncoding () 
  {
    return type;
  }

  public String getContentType () 
  {
    return type;
  }

  public static String getContentTypeFor (String name)
    throws IOException
  {
    // guess the content type
    String ctype = getFileNameMap ().getContentTypeFor (name);
    if (ctype == null)
      {
	ctype = guessContentTypeFromName (name);
	if (ctype == null)
	  {
	    // .xhtml and .dtd show up as "text/plain" after kaffe
	    // has tried to guess their content types, so here we
	    // correct that before building the content element.
	    if (name.endsWith (".xhtml")
		|| name.endsWith (".dtd"))
	      {
		ctype = "text/xml";
	      }
	    else
	      {
		ctype = "plain/text";
	      }
	  }
      }
    return ctype;
  }
}
