/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: SimpleServiceQuery.java,v 1.6 2001/11/15 20:53:19 tim Exp $
 */
package org.xoe.core.services;

import java.util.Stack;
import java.util.Vector;

public class SimpleServiceQuery 
  implements IServiceQuery
{
  private static Vector emptyVector = new Vector ();
  private static Stack freelist = new Stack ();

  private String m_func = null;
  private String m_iface = null;
  private Vector m_featureList = emptyVector;

  public String toString ()
  {
    return toXPath ();
  }

  public SimpleServiceQuery (String iface)
  {
    m_iface = iface;
  }

  public SimpleServiceQuery (String func, String iface)
  {
    m_func = func;
    m_iface = iface;
  }

  public SimpleServiceQuery ()
  {
  }

  public String toXPath ()
  {
    if (m_func == null && m_iface == null && m_featureList.size() == 0)
      {
        return null;
      }
    StringBuffer query = new StringBuffer ();
    query.append ("xoeservices:service[xoeservices:functionality");

    if (m_func != null)
      {
        query.append ("[@type='" + m_func + "']");
      }
    if (m_iface != null)
      {
        query.append ("[xoeservices:implements='" + m_iface + "']");
      }
    for (int i = 0; i < m_featureList.size (); i++)
      {
        query.append ("[xoeservices:feature[@name='"+ ((String[])m_featureList.get (i))[0] + "']" +
            "[@value='" + ((String[])m_featureList.get (i))[1] + "']]");
      }
    query.append (']');
    return query.toString ();
  }

  public SimpleServiceQuery setFunctionality (String func)
  {
    m_func = func;
    return this;
  }

  public SimpleServiceQuery setInterface (String iface)
  {
    m_iface = iface;
    return this;
  }

  public SimpleServiceQuery addFeature (String name, String value)
  {
    if (m_featureList == emptyVector)
      {
	m_featureList = new Vector ();
      }
    m_featureList.addElement (new String[] { name, value });
    return this;
  }

  public void reset ()
  {
    m_func = null;
    m_iface = null;
    m_featureList = emptyVector;
  }

  public static SimpleServiceQuery alloc ()
  {
    if (freelist.empty ())
      {
	return (new SimpleServiceQuery ());
      }
    SimpleServiceQuery q = (SimpleServiceQuery) freelist.pop ();
    q.m_func = null;
    q.m_iface = null;
    q.m_featureList = emptyVector;
    return q;
  }

  public static SimpleServiceQuery alloc (String iface)
  {
    if (freelist.empty ())
      {
	return (new SimpleServiceQuery (iface));
      }
    SimpleServiceQuery q = (SimpleServiceQuery) freelist.pop ();
    q.m_func = null;
    q.m_iface = iface;
    q.m_featureList = emptyVector;
    return q;
  }
  public static SimpleServiceQuery alloc (String func, String iface)
  {
    if (freelist.empty ())
      {
	return (new SimpleServiceQuery (func, iface));
      }
    SimpleServiceQuery q = (SimpleServiceQuery) freelist.pop ();
    q.m_func = func;
    q.m_iface = iface;
    q.m_featureList = emptyVector;
    return q;
  }

  public static void free (SimpleServiceQuery q)
  {
    freelist.push (q);
  }

}
