/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Config.java,v 1.1 2001/10/21 03:12:00 colin Exp $
 */
package org.xoe.core.util;

import java.util.HashMap;
import java.util.Vector;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.DOMUtils;


public class Config
{
  private static HashMap m_configs = new HashMap();
  private HashMap m_args = new HashMap();
  private String m_defName = null;

  public Config (URL url)
    throws IOException
  {
    this (url.openStream ());
  }

  public Config (String path)
    throws IOException
  {
    this (new File (path));
  }

  public Config (File file)
    throws IOException
  //Throws FileNotFoundException if the file doesn't exist
  //Throws IOException if the file cannot be parsed
  {
    this (new FileInputStream(file));
  }

  public Config (InputStream is)
    throws IOException
  {
    this (DOMlet.create (is, "Config(InputStream)"));
  }
  
  public Config (Document configDoc)
  {
    this (configDoc.getDocumentElement ());
  }

  public Config (Element pkgEl)
  {
    m_defName = pkgEl.getTagName ();

    NodeList nl = pkgEl.getChildNodes ();
    for(int i = 0; i < nl.getLength (); i++)
      {
	if (nl.item(i) instanceof Element)
	  {
	    addArg ((Element) nl.item (i));
	  }
      }
  }

  public static Config getConfig (String name)
  {
    return (Config) m_configs.get (name);
  }

  public static void putConfig (Config conf)
  {
    putConfig (conf.getDefaultName (), conf);
  }

  public static void putConfig (String name, Config conf)
  {
    m_configs.put (name, conf);
  }

  private String getDefaultName ()
  {
    return m_defName;
  }

  public Vector getValue (String name)
  {
    return (Vector) m_args.get (name);
  }

  public String getStringValue (String name)
  {
    Vector v = getValue (name);
    if (v == null || v.size () == 0)
      {
	return null;
      }

    Object ret = v.elementAt (0);
    return (ret instanceof Element) ? DOMUtils.getTextContent ((Element) ret) : (String) ret;
  }

  private void addArg (Element n)
  {
    String name = n.getTagName ();
    NodeList nl = n.getChildNodes ();
    Vector value = new Vector ();

    for (int i = 0; i < nl.getLength (); i++)
      {
	if (nl.item (i) instanceof Element)
	  {
	    value.addElement (nl.item (i));
	  }
      }
    
    // if there were no Elements, take the text data instead
    if (value.size () == 0)
      {
	value.addElement (DOMUtils.getTextContent (n));
      }

    Vector prev = (Vector) m_args.get (name);

    if (prev == null)
      { // no previous entry
	// stick in the new value(s)
	m_args.put (name, value);
      }
    else
      { // one or more pre-existing values (Elements and/or Strings)
	// add our new value(s)
	prev.addAll (value);
      }
  }
}
