/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Logger.java,v 1.2 2001/11/14 23:11:08 samc Exp $
 */
package org.xoe.core.util;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.StringTokenizer;

import org.w3c.dom.Document;

/**
 * Created: Thu Mar  9 11:51:15 2000
 */
public class Logger 
  implements ILogger
{
  public static long logTime = 0;
  
  private static final String ALLOW = ".allow";
  
  private static final String DENY = ".deny";

  private static final char CLASS_DELIM = ',';

  private static String [] allow;

  private static String [] deny;

  private static int isLogging;

  private static String lastCallerClassName;

  private static PrintWriter[] out = new PrintWriter[]
  {
    new PrintWriter (System.err), null
  };

  // provides the date format to use for the log function
  private final static DateFormat dformat = new SimpleDateFormat ("MMM dd HH:mm:ss.SSS");

  static
  {
    allow = split (System.getProperty (PROPERTY_PREFIX + LOGGING_PROPERTY + ALLOW), CLASS_DELIM);
    
    deny  = split (System.getProperty (PROPERTY_PREFIX + LOGGING_PROPERTY + DENY), CLASS_DELIM);
    
    isLogging = STATUS | ((Boolean.getBoolean (PROPERTY_PREFIX + LOGGING_PROPERTY) || allow != null || deny != null) ? DEBUG : 0);
  }

  public Logger () 
  {
  }

  public static void log (Object caller, Object msgObject) 
  {
    log (caller, String.valueOf (msgObject));
  }

  public static void log (Object caller, String msg) 
  {
    log (caller.toString (), msg);
  }

  public static void log (Class caller, Object msgObject) 
  {
    log (caller.getName (), msgObject.toString ());
  }

  public static void log (Class caller, String msg) 
  {
    log (caller.getName (), msg);
  }

  public static void log (String caller, Object msgObject) 
  {
    log (caller, msgObject.toString ());
  }

  public static void log (Object caller, Throwable t) 
  {
    log (caller.toString (), t);
  }
 
  public static void log (Class caller, Throwable t) 
  {
    log (caller.getName (), t);
  }

  public static void log (String caller, Throwable t) 
  {
    StringWriter sw = new StringWriter ();
    PrintWriter pw = new PrintWriter (sw);
    t.printStackTrace (pw);
    log (caller, sw.toString ());
  }

  public static void log (String caller, String msg) 
  {
    log (caller, msg, DEBUG);
  }

  public static void status (String msg)
  {
    log (null, msg, STATUS);
  }

  public static void log (String caller, String msg, int level)
  {
    log (caller, msg, level, caller);
  }

  public static void log (String caller, String msg, int level, String logCategory)
  {
    if (caller == null || (isLogging & level) == 0 || out[level-1] == null)
      {
	return;
      }

    long stime = System.currentTimeMillis ();
    try
      {
        if (msg == null)
          {
            msg = "null message";
          }

        if (logCategory != null) 
          {
            int allowMatch = 0;
out:
            if (allow != null)
              { 
                for (int i = 0 ; i < allow. length ; i++)
                  { 
                    if (logCategory. startsWith (allow[i]))
                      { 
                        allowMatch = allow[i].length ();
                        break out;
                      }
                  }
                return;
              }

            if (deny != null)
              { 
                for (int i = 0 ; i < deny. length ; i++)
                  { 
                    if (logCategory. startsWith (deny[i]) && deny[i].length () > allowMatch)
                      { 
                        return;
                      }
                  }
              }
          }

        level--;

        // is this the same class issuing another log message?
        if (caller != null && !caller.equals (lastCallerClassName)) 
          {
            Date date = new Date (System.currentTimeMillis ());
            lastCallerClassName = caller;
            out[level].println ("\n" + "[" + dformat.format(date) + "] \t" + caller);
            //	out[level].println ("\n" + caller);
          }  

        StringTokenizer tok = new StringTokenizer (msg, "\n", true);
        while (tok.hasMoreTokens ())
          {
            String t = tok.nextToken ();
            if (t.equals ("\n"))
              {
                out[level].println ("");
              }
            else
              {
                out[level].print ("\t" + t);
              }
          }
        out[level].println ("");
      }
    finally
      {
        logTime += System.currentTimeMillis () - stime;
      }
  }

  // ---------------------------------------------------------------------
  // org.xoe.core.util.ILogger

  public void register (PrintWriter writer, int level)
  {
    out[level-1] = writer;
  }

  public void unregister (PrintWriter writer, int level)
  {
    out[level-1] = null;
  }

  // ---------------------------------------------------------------------
  // org.xoe.core.services.IService

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities()
  {
    return (new String[] { ILogger.FUNCTIONALITY });
  }

  public String[] getInterfaces(String func)
  {
    if (func.equals (ILogger.FUNCTIONALITY))
      {
	return (new String[] { ILogger.INTERFACE });
      }
    return (null);
  }

  public String[] getFeatures(String func)
  {
    return (null);
  }

  public String[] getInformation()
  {
    return (null);
  }
  
  private static String [] split (String str, char delimChar) 
  {
    if (str == null) 
      { 
	return null;
      }
    
    StringTokenizer st = new StringTokenizer (str, delimChar + "", false);

    String [] ret = new String[st.countTokens ()];
      
    for (int i=0; st.hasMoreTokens (); i++)
      {
	ret[i] = st.nextToken ();
      }
    return ret;
  }

} // Logger
