/* 
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: ProviderHashtable.java,v 1.3 2001/11/29 01:55:27 erik Exp $
 */
package org.xoe.core.util;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.lang.ref.SoftReference;


public class ProviderHashtable
{
  private class Entry
  {
    Hashtable m_hashProviderToValues = new Hashtable ();
    Vector m_listProviders = new Vector ();
    Object m_currentValue = null;

    Enumeration providers ()
    {
      return m_listProviders.elements ();
    }

    void put (Object provider, Object value)
    {
      if (value == null)
	{
	  remove (provider);
	  return;
	}

      m_hashProviderToValues.put (provider, value);
      // make sure this provider goes on the end of the
      // ordered list
      m_listProviders.removeElement (provider);
      m_listProviders.addElement (provider);

      calculateCurrentValue();
    }

    Object get ()
    {
      return m_currentValue;
    }

    Object get (Object provider)
    {
      return m_hashProviderToValues.get (provider);
    }

    void remove (Object provider)
    {
      m_hashProviderToValues.remove (provider);
      m_listProviders.removeElement (provider);
      calculateCurrentValue();
    }

    void calculateCurrentValue()
    {
      // If we have an entry for the "overriding provider" use it
      // otherwise use the most recent entry.
      m_currentValue = null;
      if (m_listProviders.size () > 0)
	{
	  Object provider = m_overridingProvider != null ? m_overridingProvider.get() : null;
	  if(provider == null || !m_listProviders.contains (provider))
	    {
	      provider = m_listProviders.lastElement();
	    }
	  m_currentValue = m_hashProviderToValues.get (provider);
	}
    }

    boolean empty ()
    {
      return m_currentValue == null;
    }
  } // private class Entry

  Hashtable m_hashKeyToEntry = new Hashtable ();
  Hashtable m_hashProviderToKeys = new Hashtable ();
  SoftReference m_overridingProvider = null;

  public void setOverridingProvider(Object provider)
  {
    Object oldprovider = (m_overridingProvider != null ? m_overridingProvider.get() : null);
    m_overridingProvider = new SoftReference(provider);

    if (oldprovider != null)
      { // Force anyone with entries for the old provider to recalc their
	// current value
	HashSet s = (HashSet)m_hashProviderToKeys.get (oldprovider);
	if (s != null)
	  {
	    Iterator it = s.iterator ();
	    while (it.hasNext ())
	      {
		Object key = it.next ();
		Entry e = (Entry)m_hashKeyToEntry.get (key);
		if (e != null)
		  {
		    e.calculateCurrentValue ();
		  }
	      }
	  }
      }

    if (provider != null)
      { // Force anyone with entries for the new provider to recalc their
	// current value
	HashSet s = (HashSet)m_hashProviderToKeys.get (provider);
	if (s != null)
	  {
	    Iterator it = s.iterator ();
	    while (it.hasNext ())
	      {
		Object key = it.next ();

		Entry e = (Entry)m_hashKeyToEntry.get (key);
		if (e != null)
		  {
		    e.calculateCurrentValue ();
		  }
	      }
	  }
      }

  }

  public void put (Object key, Object value)
  {
    put (key, value, null);
  }

  public void put (Object key, Object value, Object provider)
  {
    if (value == null || key == null)
      {
	throw new NullPointerException ();
      }

    if (provider == null)
      {
	provider = this;
      }

    if (m_hashProviderToKeys.get (provider) == null)
      {
	m_hashProviderToKeys.put (provider, new HashSet ());
      }

    ((HashSet)m_hashProviderToKeys.get (provider)).add (key);
    
    if (m_hashKeyToEntry.get (key) == null)
      {
	m_hashKeyToEntry.put (key, new Entry ());
      }

    ((Entry)m_hashKeyToEntry.get (key)).put (provider, value);
  }

  public Object get (Object key)
  {
    if(key == null)
      {
	return null;
      }

    Entry e = (Entry)m_hashKeyToEntry.get (key);
    return e == null ? null : e.get ();
  }

  public void remove (Object key)
  {
    remove (key, null);
  }

  public void remove (Object key, Object provider)
  {
    if (provider == null)
      {
	provider = this;
      }

    HashSet s = (HashSet)m_hashProviderToKeys.get (provider);
    if (s != null)
      {
	s.remove (key);
	if (s.size () == 0)
	  {
	    m_hashProviderToKeys.remove (provider);
	  }
      }

    Entry e = (Entry)m_hashKeyToEntry.get (key);
    if (e != null)
      {
	e.remove (provider);
	if (e.empty ())
	  {
	    m_hashKeyToEntry.remove (key);
	  }
      }
  }

  public void removeProvider (Object provider)
  {
    if (provider == null)
      {
	provider = this;
      }

    HashSet s = (HashSet)m_hashProviderToKeys.get (provider);
    if (s != null)
      {
	Iterator it = s.iterator ();
	while (it.hasNext ())
	  {
	    Object key = it.next ();
	    Entry e = (Entry)m_hashKeyToEntry.get (key);
	    if (e != null)
	      {
		e.remove (provider);
		if (e.empty ())
		  {
		    m_hashKeyToEntry.remove (key);
		  }
	      }
	  }
      }
    m_hashProviderToKeys.remove (provider);
  }

  public void logMap ()
  {
    System.err.println (this.toString ());
    Enumeration e = m_hashKeyToEntry.keys ();
    while (e.hasMoreElements ())
      {
	Object key = e.nextElement ();
	System.err.println ("key: " + key);
	System.err.println ("current value: " + get(key));

	Enumeration e2 = ( (Entry)m_hashKeyToEntry.get (key)).providers ();
	while (e2.hasMoreElements ())
	  {
	    Object provider = e2.nextElement ();
	    System.err.println ("  provider: " + provider + " value: " + ( (Entry)m_hashKeyToEntry.get (key)).get (provider));
	  }
      }

    e = m_hashProviderToKeys.keys ();
    while (e.hasMoreElements ())
      {
	Object provider = e.nextElement ();
	System.err.println ("provider: " + provider);
	Iterator it = ( (HashSet)m_hashProviderToKeys.get (provider)).iterator ();
	while (it.hasNext ())
	  {
	    Object key = it.next ();
	    System.err.println ("  key: " + key);
	  }
      }

  }
}
