/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: XPath.java,v 1.10 2001/11/20 22:32:11 suraj Exp $
 */
package org.xoe.core.xpath;

import java.net.URL;
import java.net.MalformedURLException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.Cache;
import org.xoe.core.dom.DefaultEntityResolver;
import org.xoe.core.dom.ElementNamedNodeMapAdaptor;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.NodeConsumer;
import org.xoe.core.dom.SimpleNodeList;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;


public class XPath 
  implements XPathConstants
{
  private static Cache xpathContextCache = Cache.getCache (XPathContext.class);

  short[] code;
  short codeLen;
  Object[] values;
  short valLen;
  int princpalNodeType = Node.ELEMENT_NODE;
  public static int xpathTime;


  /* 
   * This query in text form. Added for debugging purposes
   */
  public String query;

//  NodeUtil functions from webslinger

  public static String valueOf( Node in )
  {
    if (in.getNodeType () == Node.TEXT_NODE || in.getNodeType () == Node.ATTRIBUTE_NODE) 
      {
	return ( (DefaultEntityResolver.deEntify (in.getNodeValue())));
      } 
    else 
      {
	StringBuffer result = new StringBuffer();
	valueOf(result, in);
	return result.toString();
      }
  }

  public static void valueOf( StringBuffer out, Node in )
  {
      if ( in != null ) {
//		if ( in.getNodeType() == Node.ATTRIBUTE_NODE )
//		    Logger.log (XPath.class, in.getNodeName() + " = " + in.getNodeValue() );

	  if ((in.getNodeType () == Node.TEXT_NODE) ||
	      (in.getNodeType () == Node.ATTRIBUTE_NODE)) 
	    {
	      out.append( (DefaultEntityResolver.deEntify (in.getNodeValue())));
	    } 
	  else
	      if ( ( in.getNodeType() == Node.ELEMENT_NODE ) ||
		   ( in.getNodeType() == Node.DOCUMENT_NODE ) ) {
		  NodeList children = in.getChildNodes();

		  for ( int i=0; i < children.getLength(); i++ )
		      valueOf( out, children.item(i) );
	      }
      }
  }

  public static String valueOf( NodeList in )
  {
      StringBuffer out = new StringBuffer();
      int inCount = in.getLength();

      for ( int i=0; i < inCount; i++ )
	  valueOf( out, in.item(i) );

       Logger.log (XPath.class, out.toString() );
      return out.toString();
  }

  public static String stringValue(Node node)
  {
      return valueOf (node);
  }

  public static String stringValue(double value)
  {
	  long lvalue = (long) value;
	  if ((double) lvalue == value)
		  return Long.toString(lvalue);
	  else // FIXME - should never use exponential notation!
		  return Double.toString(value);
  }

  public static String stringValue(Object value)
  {
	  if (value instanceof Number)
		  return stringValue(((Number) value).doubleValue());
	  else if (value instanceof NodeList) {
		  NodeList nodes = (NodeList) value;
		  int count = nodes.getLength();
		  return count == 0 ? "" : stringValue(nodes.item(0));
	  }
	  else
		  return value.toString();
  }

  public static double numberValue(Object value)
  {
	  if (value instanceof Number)
		  return ((Number) value).doubleValue();
	  else
		  return Double.parseDouble(stringValue(value));
  }

  public static boolean booleanValue(Object value)
  {
	  if (value instanceof Boolean)
		  return ((Boolean) value).booleanValue();
	  else if (value instanceof Number)
		  return ((Number) value).doubleValue() != 0.0;
	  else if (value instanceof NodeList)
		  return ((NodeList) value).getLength() > 0;
	  else
		  return value.toString().length() > 0;
  }

  void result(XPathContext context, Object value, int resultType)
  {
	  if (value instanceof Boolean)
		  result(context, ((Boolean) value).booleanValue(), resultType);
	  else if (value instanceof Number)
		  result(context, ((Number) value).doubleValue(), resultType);
	  else if (value instanceof String)
		  result(context, (String) value, resultType);
	  else if (value instanceof NodeList)
		  result(context, (NodeList) value, resultType);
	  else if (resultType == ANY_RESULT) {
		  context.objResult = value;
		  context.resultType = ANY_RESULT;
	  }
	  else
		  throw new Error("unhandled result code"); // FIXME
  }

  void result(XPathContext context, double value, int resultType)
  {
	  if (resultType == BOOLEAN_RESULT) {
		  context.boolResult = value != 0.0;
		  context.resultType = BOOLEAN_RESULT;
	  }
	  else if (resultType == PREDICATE_RESULT) {
		  context.boolResult = value == getContextPosition(context);
		  context.resultType = BOOLEAN_RESULT;
	  }
	  else if (resultType == STRING_RESULT) {
		  context.objResult = stringValue(value);
		  context.resultType = STRING_RESULT;
	  }
	  else if (resultType == NUMBER_RESULT || resultType == ANY_RESULT) {
		  context.numResult = value;
		  context.resultType = NUMBER_RESULT;
	  }
	  else
		  throw new Error("unhandled result code"); // FIXME
  }

  static void result(XPathContext context, String value, int resultType)
  {
	  if (resultType == ANY_RESULT || resultType == STRING_RESULT) {
		  context.objResult = value;
		  context.resultType = STRING_RESULT;
	  }
	  else if (resultType == BOOLEAN_RESULT
			   || resultType == PREDICATE_RESULT) {
		  context.boolResult = value.length() > 0;
		  context.resultType = BOOLEAN_RESULT;
	  }
	  else if (resultType == NUMBER_RESULT) {
		  context.numResult = Double.parseDouble(value);
		  context.resultType = NUMBER_RESULT;
	  }
	  else if (resultType == NODESET_RESULT) {
		  SimpleNodeList list = new SimpleNodeList();
                  Node e = context.getCurrentNode ().getOwnerDocument ().createElement("Node");
                  DOMUtils.setTextContent (e, value);
                  list.appendChild (e);
		  context.objResult = list;
		  context.resultType = NODESET_RESULT;
	  }
	  else
		  throw new Error("unhandled result code"); // FIXME
  }

  static void result(XPathContext context, NodeList value, int resultType)
  {
          if (resultType == ANY_RESULT || resultType == NODESET_RESULT) {
		  context.objResult = value;
		  context.resultType = NODESET_RESULT;
	  }
	  else if (resultType == BOOLEAN_RESULT
			   || resultType == PREDICATE_RESULT) {
		  context.boolResult = value.getLength() > 0;
		  context.resultType = BOOLEAN_RESULT;
	  }
	  else if (resultType == STRING_RESULT) {
		  context.objResult = stringValue(value);
		  context.resultType = STRING_RESULT;
	  }
	  else
		  throw new Error("unhandled result code"); // FIXME
  }

  static void result(XPathContext context, boolean value, int resultType)
  {
          if (resultType == BOOLEAN_RESULT || resultType == ANY_RESULT
		  || resultType == PREDICATE_RESULT) {
		  context.boolResult = value;
		  context.resultType = BOOLEAN_RESULT;
	  }
	  else if (resultType == NUMBER_RESULT) {
		  context.numResult = value ? 1.0 : 0.0;
		  context.resultType = NUMBER_RESULT;
	  }
	  else if (resultType == STRING_RESULT) {
		  context.objResult = value ? "true" : "false";
		  context.resultType = STRING_RESULT;
	  }
	  else
		  throw new Error("unknown result code " + resultType);   // FIXME
  }

  public Object eval(XPathContext context)
  {
	  return eval(context, codeLen - 1);
  }

  public Object eval(XPathContext context, int pc)
  {
	  eval(context, pc, ANY_RESULT);
	  switch (context.resultType) {
	  case STRING_RESULT:
	  case NODESET_RESULT:
	  case ANY_RESULT:
		  return context.objResult;
	  case BOOLEAN_RESULT:
		  return context.boolResult ? Boolean.TRUE : Boolean.FALSE;
	  case NUMBER_RESULT:
		  return new Double(context.numResult);
	  default:
		  throw new Error ("unknown result Type");
	  }
  }

  public void error(String message)
  {
	  throw new Error(message);
  }

  public void checkArgs(Object function, int argCount, int expected)
  {
	  if (argCount != expected)
		  error(function.toString() + ": wrong number of arguments ("
			    + argCount +", should be " + expected + ')');
  }

  public void checkArgs(Object function, int argCount, int min, int max)
  {
	  if (argCount < min)
		  error(function.toString() + ": too few arguments ("
			    + argCount +", should be at least " + min + ')');
	  else if (argCount > max && max >= 0)
		  error(function.toString() + ": too many arguments ("
			    + argCount +", should be at most " + max + ')');
  }

  public void call(XPathContext context,
				   Object function, Object[] args, int resultType)
  {
	  error("unknown function: "+ function);
  }

int getContextPosition(XPathContext context)
  {
	  return context.getCurrentPosition();
  }

int getContextSize(XPathContext context)
  {
	  return (context.predPc > context.stepPc
			  ? (int) (getPositionAndLast(context) >> 32)
			  : context.getCurrentSize());
  }

  private long getPositionAndLast(XPathContext context)
  {
	  Node curNode = context.getCurrentNode();
	  try {
		  context.popNode();
		  Node parent = context.getCurrentNode();
		  int pc = context.stepPc;
		  short opcode = code[pc];
		  int count = 0;
		  int position = -1;
		  int axis = ((opcode >= OP_AXIS_FIRST
					   && opcode < OP_AXIS_FIRST + COUNT_OP_AXIS)
					  ? opcode - OP_AXIS_FIRST
					  : AXIS_CHILD);
		  switch (axis) {
		  case AXIS_CHILD:
			  NodeList children = parent.getChildNodes();
			  int childCount = children.getLength();
			  for ( int i = 0; i < childCount; i++ ) {
				  try	{
					  Node child = children.item(i);
					  context.pushNode(child);
					  if (child == curNode)
						  position = count;
					  if (matchStep(context, pc, context.predPc))
						  count++;
					  }
				  finally {
					  context.popNode();
				  }
			  }
		  }
		  return ((long) count << 32) | (long) position;
	  }
	  finally {
		  context.pushNode(curNode);
	  }
  }

  /**
   * Evaluate a FunctionCall.
   * @param pc index just after last argument expression
   */
  public void evalCall(XPathContext context, Object function,
					   int argCount, int pc, int resultType)
  {
	  if (function instanceof XPathFunction) {
		  ((XPathFunction) function).apply(context, argCount, this, pc, resultType);
		  return;
	  }
	  if (function == "count") {
		  checkArgs(function, argCount, 1);
		  pc -= code[pc];
		  int saveKind = context.consumeKind;
		  int saveCount = context.consumeCount;
		  try {
			  context.consumeKind = XPathContext.CONSUME_COUNT;
			  context.consumeCount = 0;
			  evalToNodes(context, pc + code[pc], context);
			  context.resultType = NUMBER_RESULT;
			  context.numResult = context.consumeCount;
		  }
		  finally {
			  context.consumeKind = saveKind;
			  context.consumeCount = saveCount;
		  }
		  return;
	  }
	  if (function == "not") {
		  checkArgs(function, argCount, 1);
		  pc -= code[pc];
		  pc += code[pc];
		  result(context, ! evalToBoolean(context, pc), resultType);
		  return;
	  }
	  if (function == "boolean") {
		  checkArgs(function, argCount, 1);
		  pc -= code[pc];
		  pc += code[pc];
		  result(context, evalToBoolean(context, pc), resultType);
		  return;
	  }
	  if (function == "number") {
		  checkArgs(function, argCount, 1);
		  pc -= code[pc];
		  pc += code[pc];
		  result(context, evalToNumber(context, pc), resultType);
		  return;
	  }
	  if (function == "position") {
		  checkArgs(function, argCount, 0);
		  result(context, getContextPosition(context), resultType);
		  return;
	  }
	  if (function == "last") {
		  checkArgs(function, argCount, 0);
		  result(context, getContextSize(context), resultType);
		  return;
	  }
	  if (function == "string") {
		  checkArgs(function, argCount, 1);
		  pc -= code[pc];
		  pc += code[pc];
		  int value;
		  result(context, evalToString(context, pc), resultType);
		  return;
	  }
	  if (function == "true") {
		  checkArgs(function, argCount, 0);
		  result(context, true, resultType);
		  return;
	  }
	  if (function == "false") {
		  checkArgs(function, argCount, 0);
		  result(context, false, resultType);
		  return;
	  }
if (function == "starts-with") {
checkArgs(function, argCount, 2);
pc -= code[pc];
pc += code[pc];

String source = evalToString( context, pc);

int len = code[++pc];
pc += len;
String match = evalToString( context, pc);

result(context, source.startsWith( match ), resultType);
return;
}
if (function == "ends-with") {
      checkArgs(function, argCount, 2);
      pc -= code[pc];
      pc += code[pc];

      String source = evalToString( context, pc);

      int len = code[++pc];
      pc += len;
      String match = evalToString( context, pc);

      result(context, source.endsWith( match ), resultType);
      return;
}
if (function == "substring-before") {
      checkArgs(function, argCount, 2);
      pc -= code[pc];
      pc += code[pc];

      String source = evalToString( context, pc);

      int len = code[++pc];
      pc += len;
      String pattern =  evalToString( context, pc);
      String result;
      int index=source.indexOf(pattern);
      if(index == -1)
	{
	  result = "";
	}
      else
	{
	  result = source.substring(0, index);
	}

      result(context, result, resultType);
      return;
}
  if (function == "concat") {
      // subtract the length of all argument expressions, including length words
      pc -= code[pc];

      StringBuffer ret = new StringBuffer();

      for(int i = 0; i < argCount; i++)
	{ // code[pc] is the length of this (argument) expression
	  // increment to the end of the expression (the actual opcode)
	  pc += code[pc];
	  ret.append(evalToString(context, pc));
	  // increment by one to go to the length of the next argument exp.
	  pc++;
	}

      result(context, ret.toString(), resultType);
      return;
}
if (function == "translate") {
      checkArgs(function, argCount, 3);
      pc -= code[pc];
      pc += code[pc];

      String source = evalToString( context, pc);

      int len = code[++pc];
      pc += len;
      String patternA =  evalToString( context, pc);
      len = code[++pc];
      pc += len;
      String patternB =  evalToString( context, pc);

      StringBuffer ret = new StringBuffer();
      for(int i = 0; i < source.length(); i++)
	{
	  int j = patternA.indexOf(source.charAt(i));
	  if(j == -1)
	    {
	      ret.append(source.charAt(i));
	    }
	  else if(j < patternB.length())
	    {
	      ret.append(patternB.charAt(j));
	    }
	  // else append nothing!
	}

      result(context, ret.toString(), resultType);
      return;
}
if (function == "contains") {
      checkArgs(function, argCount, 2);
      pc -= code[pc];
      pc += code[pc];

      String source = evalToString( context, pc);

      int len = code[++pc];
      pc += len;
      String pattern =  evalToString( context, pc);

      result(context, source.indexOf(pattern) != -1, resultType);
      return;
}
if (function == "substring-after") {
      checkArgs(function, argCount, 2);
      pc -= code[pc];
      pc += code[pc];

      String source = evalToString( context, pc);

      int len = code[++pc];
      pc += len;
      String pattern =  evalToString( context, pc);
      String result;
      int index=source.indexOf(pattern);
      if(index == -1)
	{
	  result = "";
	}
      else
	{
	  result = source.substring(index + 1);
	}

      result(context, result, resultType);
      return;
}
if (function == "substring") {
      checkArgs(function, argCount, 2);
      pc -= code[pc];
      pc += code[pc];

      String source = evalToString( context, pc);

      int len = code[++pc];
      pc += len;
      int index = (int) evalToNumber( context, pc);
      String result = source.substring( index );

      result(context, result, resultType);
      return;
}
	  if (function == "name") {
		  checkArgs(function, argCount, 0, 1);
		  if (argCount == 0) {
			  result(context,
				     context.getCurrentNode().getNodeName(), resultType);
		  }
		  else {
			  pc -= code[pc];
			  pc += code[pc];

			  int saveKind = context.consumeKind;
			  Object saveObject = context.consumeObject;
			  try {
				  context.consumeKind = XPathContext.CONSUME_NAME;
				  context.consumeObject = null;
				  evalToNodes(context, pc, context);
				  context.resultType = STRING_RESULT;
				  if (context.consumeObject == null)
				    {
				      result(context, "", resultType);
				    }
				  else
				    {
				      result(context, context.consumeObject.toString(),
					     resultType);
				    }
			  }
			  finally {
				  context.consumeKind = saveKind;
				  context.consumeObject = saveObject;
			  }
		  }
		  return;
	  }
	  Object[] args = new Object[argCount];
	  pc -= code[pc];
	  for (int i = 0;  i < argCount;  i++) {
		  pc += code[pc];
		  args[i] = eval(context, pc);
		  pc++;
	  }
	  call(context, function, args, resultType);
  }

  public static boolean compare(Object val1, Object val2, int opcode)
  {
	  if (val1 instanceof NodeList) {
		  NodeList nodes = (NodeList) val1;
		  int count = nodes.getLength();
		  if (val2 instanceof Boolean) {
			  val1 = count > 0 ? Boolean.TRUE : Boolean.FALSE;
			  return compare(val1, val2, opcode);
		  }
		  for (int j = 0;  j < count;  j++) {
			  Node node = nodes.item(j);
			  if (compare(stringValue(node), val2, opcode))
				  return true;
		  }
		  return false;
	  }
	  else if (val2 instanceof NodeList) {
		  NodeList nodes = (NodeList) val2;
		  int count = nodes.getLength();
		  if (val1 instanceof Boolean) {
			  val2 = count > 0 ? Boolean.TRUE : Boolean.FALSE;
			  return compare(val1, val2, opcode);
		  }
		  for (int j = 0;  j < count;  j++) {
			  Node node = nodes.item(j);
			  if (compare(val1, stringValue(node), opcode))
				  return true;
		  }
		  return false;
	  }
	  else if (opcode <= OP_NEQ) { // OP_EQU or OP_NEQ
		  if (val1 instanceof Boolean || val2 instanceof Boolean) {
			  boolean bool1 = booleanValue(val1);
			  boolean bool2 = booleanValue(val2);
			  return (bool1 == bool2) == (opcode == OP_EQU);
		  }
		  else if (val1 instanceof Number || val2 instanceof Number) {
			  double num1 = numberValue(val1);
			  double num2 = numberValue(val2);
			  return (num1 == num2) == (opcode == OP_EQU);
		  }
		  else {
			  String str1 = stringValue(val1);
			  String str2 = stringValue(val2);
			  return str1.equals(str2) == (opcode == OP_EQU);
		  }
	  }
	  else {
		  double num1 = numberValue(val1);
		  double num2 = numberValue(val2);
		  switch (opcode) {
		  case OP_LSS:  return num1 <  num2;
		  case OP_LEQ:  return num1 <= num2;
		  case OP_GRT:  return num1 >  num2;
		  case OP_GEQ:  return num1 >= num2;
		  default:  return false;  // Should never happen.
		  }
	  }
  }

  public void eval(XPathContext context, int pc, int resultType)
  {
	  short opcode = code[pc];
	  double val1, val2;
	  int op2len, op1;
	  switch (opcode) {
	  case OP_AND:
	  case OP_OR:
		  op2len = code[--pc];
		  pc -= 2;
		  op1 = pc - op2len;
		  boolean bval = evalToBoolean(context, op1);
		  result(context,
			     (opcode == OP_OR) == bval ? bval
			     : evalToBoolean(context, pc),
			     resultType);
		  break;
	  case OP_EQU:
	  case OP_NEQ:
	  case OP_LSS:
	  case OP_LEQ:
	  case OP_GRT:
	  case OP_GEQ:
		  op2len = code[--pc];
		  pc -= 2;
		  op1 = pc - op2len;
		  Object obj1 = eval(context, op1);
		  Object obj2 = eval(context, pc);
		  result(context, compare(obj1, obj2, opcode), resultType);
		  break;
	  case OP_ADD:
	  case OP_SUB:
	  case OP_MUL:
	  case OP_DIV:
	  case OP_MOD:
		  op2len = code[--pc];
		  pc -= 2;
		  op1 = pc - op2len;
		  val1 = evalToNumber(context, op1);
		  val2 = evalToNumber(context, pc);
		  switch (opcode) {
		  case OP_ADD:  val1 = val1 + val2;  break;
		  case OP_SUB:  val1 = val1 - val2;  break;
		  case OP_MUL:  val1 = val1 * val2;  break;
		  case OP_DIV:  val1 = val1 / val2;  break;
		  case OP_MOD:  val1 = val1 % val2;  break;  // FIXME check result
		  }
		  result(context, val1, resultType);
		  break;
	  case OP_STRING_LITERAL:
		  result(context, (String) values[code[pc-1]], resultType);
		  break;
	  case OP_SHORT:
		  result(context, (double) code[pc-1], resultType);
		  break;
	  case OP_DOUBLE:
		  long bits = 0;
		  pc--;
		  for (int i = 4;  --i >= 0; )
			  bits = (bits << (16 * i)) | (long) (code[pc - i] & 0xffff);
		  result(context, Double.longBitsToDouble(bits), resultType);
		  break;
	  case OP_VARIABLE_REFERENCE:
		  Object name = values[code[pc-1]];
		  context.getVariableReference(name, resultType);
		  break;
	  case OP_FUNCTION_CALL:
		  Object func = values[code[--pc]];
		  int count = code[--pc];
		  --pc;
		  if(func instanceof String)
		    {
		      XPathFunction xf = context.getFunctionReference((String)func);
		      if(xf != null)
			{
			  func = xf;
			}
		    }
		  evalCall(context, func, count, pc, resultType);
		  break;
	  case OP_LOCATIONPATH:
	  case OP_EXPR_LOCATIONPATH:
	  case OP_UNION:
	    if (resultType == BOOLEAN_RESULT || resultType == PREDICATE_RESULT) 
	      {
		int saveKind = context.consumeKind;
		int saveCount = context.consumeCount;
		try {
		  context.consumeKind = XPathContext.CONSUME_BOOL;
		  context.consumeCount = 0;
		  if (opcode == OP_UNION) {
		    int numOperands = code[pc-1];
		    for (int i = 0;  i < numOperands;  i++) {
		      int partPc = (i+1 == numOperands ? pc-2-numOperands
				    : code[pc-numOperands+i] + pc - 1);
		      if (! evalToNodes(context, partPc, context))
			break;
		    }
		  }
		  else
		    evalToNodes(context, pc, context);
		  context.resultType = BOOLEAN_RESULT;
		  context.boolResult = context.consumeCount > 0;
		}
		finally {
		  context.consumeKind = saveKind;
		  context.consumeCount = saveCount;
		}
		break;
	      }
	    else 
	      {
		SimpleNodeList nodes = new SimpleNodeList();
		if (opcode == OP_UNION)
		  {
		    int unionParts = code[pc-1];
		    for (int i = 0;  i < unionParts;  i++) 
		      {
			int pathPc = i+1==unionParts ? pc - 2 - unionParts
		      : pc + code[pc-unionParts+i] - 1;
			evalToNodes(context, pathPc, nodes);
		      }
		  }
		else
		  {
		    evalToNodes(context, pc, nodes);
		  }
		result(context, nodes, resultType);
		break;
	      }
	  default:
	    throw new Error ("unhandled opcode " + opcode);
	  }
  }

  public boolean evalToNodes(XPathContext context, NodeConsumer consumer)
  {
	  context.stepPc = 0;
	  context.predPc = 0;
	  return evalToNodes(context, codeLen - 1, consumer);
  }

  /**
   * Return false iff the consumer is "satisfied" - i.e. will ignore
   * any more nodes. 
   */
  public boolean evalToNodes(XPathContext context, int pc,
						     NodeConsumer consumer)
  {
	  NodeList nodes;
	  int nodeCount;
	  short opcode = code[pc];
	  int opcodepc = pc;
	  int numSteps;
	  switch (opcode) {
	  case OP_LOCATIONPATH:
		  numSteps = code[--pc];
		  pc -= numSteps;
		  return evalPath(context, code[pc] + opcodepc, pc, consumer);
	  case OP_EXPR_LOCATIONPATH:
		  // FIXME - maybe implement this using XPathContext.consume.
		  numSteps = code[--pc];
		  pc -= numSteps;
		  int start = code[pc] + opcodepc;
		  eval(context, start - 1, NODESET_RESULT);
		  nodes = (NodeList) context.objResult;
		  nodeCount = nodes.getLength();
		  for (int i = 0; i < nodeCount; i++) {
			  try	{
				  Node node = nodes.item(i);
				  context.pushNode(node);
				  if (! evalPath(context, start, pc, consumer))
					  return false;
			  }
			  finally {
				  context.popNode();
			  }
		  }
		  return true;
// 	  case OP_UNION:
// 		  int numOperands = code[pc-1];
// 		  int[] levels = new int[numOperands];
// 		  return evalUnionStep(context, levels, 0, pc, consumer);
	  default:
		  eval(context, pc, NODESET_RESULT);
		  nodes = (NodeList) context.objResult;

		  nodeCount = nodes.getLength();
		  for (int i = 0; i < nodeCount; i++)
			  if (! consumer.consume(nodes.item(i)))
				  return false;
		  return true;
	  }

  }

  /**
   * Return false iff the consumer is "satisfied" - i.e. will ignore
   * any more nodes. 
   */
  public boolean evalPath(XPathContext context, int pc, int stop,
					   NodeConsumer consumer)
  {
	  Node node = context.getCurrentNode();
	  Node save;
	  NodeList children;
	  int count, savePos;
	  for (;;) {
		  if (pc == stop) {
			  return consumer.consume(node);
		  }
		  short opcode = code[pc];
		  if (opcode >= OP_AXIS_FIRST
			  && opcode < OP_AXIS_FIRST + COUNT_OP_AXIS)
			  context.stepPc = pc;
			  context.predPc = pc;
		  int j;
		  pc++;
		  switch (opcode) {
		  case OP_AXIS_FIRST+AXIS_ANCESTOR: {
		    int oldPNT = princpalNodeType;
		    princpalNodeType = Node.ELEMENT_NODE;
		    Node parent = node.getParentNode();
		    while(parent != null)
		      {
			try{
			  context.pushNode(parent);
			  if(! evalPath(context, pc, stop, consumer))
			    { 
			      princpalNodeType = oldPNT;
			      return false;
			    }
			}
			finally{
			  context.popNode();
			}
			parent = parent.getParentNode();
		      }
		    princpalNodeType = oldPNT;
		    return true;
		  }
		  case OP_AXIS_FIRST+AXIS_CHILD: {
			  int oldPNT = princpalNodeType;
			  princpalNodeType = Node.ELEMENT_NODE;
			  children = node.getChildNodes();
			  int childCount = children.getLength();
			  for ( int i = 0; i < childCount; i++ ) {
				  try	{
					  Node child = children.item(i);
					  context.pushNode(child);
					  /*
					  int next = matchSingleStep(context, pc, stop);
					  if (next >= 0
						  && ! evalPath(context, next, stop, consumer))
						  return false;
					  */
					  if (! evalPath(context, pc, stop, consumer))
					    {
					      princpalNodeType = oldPNT;
					      return false;
					    }
				  }
				  finally {
					  context.popNode();
				  }
			  }
			  princpalNodeType = oldPNT;
			  return true;
		  }
		  case OP_AXIS_FIRST+AXIS_DESCENDANT_OR_SELF: {
			  int oldPNT = princpalNodeType;
			  princpalNodeType = Node.ELEMENT_NODE;
			  if (! evalPath(context, pc, stop, consumer))
				  return false;
			  children = node.getChildNodes();
			  count = children.getLength();
			  // The trick here is that we re-evaluate the
			  // descendant-or-self operation (by using pc-1 in the
			  // recursive call) on each child.
			  for ( int i = 0; i < count; i++ ) {
				  try {
					  Node child = children.item(i);
					  context.pushNode(child);
					  if (! evalPath(context, pc - 1, stop, consumer))
					    {
					      princpalNodeType = oldPNT;
					      return false;
					    }
				  }
				  finally {
					  context.popNode();
				  }
			  }
			  princpalNodeType = oldPNT;
			  return true;
		  }
		  case OP_AXIS_FIRST+AXIS_DESCENDANT: {
			  int oldPNT = princpalNodeType;
			  princpalNodeType = Node.ELEMENT_NODE;
			  children = node.getChildNodes();
			  count = children.getLength();
			  // The trick here is that we re-evaluate the
			  // descendant operation (by using pc-1 in the
			  // recursive call) on each child.
			  for ( int i = 0; i < count; i++ ) {
				  try {
					  Node child = children.item(i);
					  context.pushNode(child);
					  if (! evalPath(context, pc, stop, consumer)
						  || ! evalPath(context, pc - 1, stop, consumer))
					    {	
					      princpalNodeType = oldPNT;
					      return false;
					    }
				  }
				  finally {
					  context.popNode();
				  }
			  }
			  princpalNodeType = oldPNT;
			  return true;
		  }
		  case OP_AXIS_FIRST+AXIS_ATTRIBUTE: {
			  int oldPNT = princpalNodeType;
			  princpalNodeType = Node.ATTRIBUTE_NODE;
			  ElementNamedNodeMapAdaptor attrs = (ElementNamedNodeMapAdaptor)node.getAttributes();
			  if (attrs == null)  // if this is a Text node 
			      break;

			  Document owner = node.getOwnerDocument ();
			  int attrCount = attrs.getLength();
			  boolean ret = true;
			  for ( int i = 0; i < attrCount; i++ ) {
				  Node child = attrs.item(i);
				  context.pushNode(child);
				  ret = evalPath(context, pc, stop, consumer);
				  context.popNode();
				  if (!ret) {
					  break;
				  }
			  }
			  princpalNodeType = oldPNT;
			  return ret;
		  }
		  case OP_AXIS_FIRST+AXIS_PARENT:
			  {
				  int oldPNT = princpalNodeType;
				  princpalNodeType = node.getNodeType ();
				  Node parent = node.getParentNode ();
				  if (parent != null) {
				      try {
					context.pushNode(parent);
					if (! evalPath(context, pc + 1, stop, consumer))
					  {
					    princpalNodeType = oldPNT;
					    return false;
					  }
				      }
				      finally {
					context.popNode();
				      }
				      princpalNodeType = oldPNT;
				  }
				  return true;
			  }
		  case OP_SIMPLE_NAMETEST:
			  {
				Object namespace = values[code[pc]];
				Object name = values[code[pc+1]];
				if ( (node.getNodeType () == princpalNodeType) && (node.getLocalName() == name) &&  (node.getNamespaceURI() != context.getNamespaceURI ((String) namespace)))
				  {
					Logger.log (this, " mismatch!!  xpath prefix = " + namespace + " xpath namespace = " + context.getNamespaceURI ((String) namespace));
					Logger.log (this, " node namespace  = " +  node.getNamespaceURI () + " node name " + node.getLocalName ());
					Logger.log (this, "context = " + context);
					Logger.log (this, new Throwable ());
				  }
				if  ((node.getNodeType () == princpalNodeType) &&  (node.getLocalName() == name) && node.getNamespaceURI () == context.getNamespaceURI ((String) namespace))
				  {
					pc+=2;
					continue;
				  }
				else 
				  {
					return true;
				  }
			  }
		  case OP_AXIS_FIRST+AXIS_SELF:
			  {
				  int oldPNT = princpalNodeType;
				  princpalNodeType = node.getNodeType ();
				  if (code[pc] == OP_ROOT) {
				      try {
					  context.pushNode(context.getDocumentRoot());
					  if (! evalPath(context, pc + 1, stop, consumer))
						{
						  princpalNodeType = oldPNT;
					      return false;
						}
				      }
				      finally {
					  context.popNode();
				      }
					  princpalNodeType = oldPNT;
				      return true;
				  }
			  }			  
			  /* ... else fall though ... */
		  case OP_NODE:
		    continue;
		  case OP_ANY_NAMETEST:
		    if (! (node.getNodeType () == princpalNodeType))
		      {
			return true;
		      }
			  continue;
		  case OP_NAMESPACE_NAMETEST:
		    if (! (node.getNamespaceURI () == context.getNamespaceURI ((String) values[code[pc]])))
		      {
			return true;
		      }
		    pc++;
		    continue;
		  case OP_ROOT:
			  if (node.getNodeType() != Node.DOCUMENT_NODE)
				  return true;
			  continue;
		  case OP_TEXT:
			  if (node.getNodeType() != Node.TEXT_NODE)
				  return true;
			  continue;
		  case OP_COMMENT:
			  if (node.getNodeType() != Node.COMMENT_NODE)
				  return true;
			  continue;
		  case OP_PROCESSING_INSTRUCTION:
			  if (node.getNodeType() != Node.PROCESSING_INSTRUCTION_NODE)
				  return true;
			  j = code[pc++];
			  if (j >= 0) {
				  Object name = values[j];
				  if (! node.getNodeName().equals(name))
					  return true;
			  }
			  continue;
		  case OP_PREDICATE:
			  int predPc = pc - 1;
			  int counter = code[pc++];
			  // We initialize context.predPc to context.stepPc.
			  // If context.predPc is less than the current pc, it means this
			  // is the first time for this predicate (in this step), and we
			  // need to first clear the counter.  We then increment the
			  // counter - and copy the result to context.currentPosition.
			  // The counter thus keeps track of the number of times we
			  // have evaluated this predicate - which is the same as
			  // the context position.
			  // Setting context.predPc to context.stepPc on a new step
			  // essentially forces all the counter in this step to zero,
			  // since they will be zerod the first time we evaluate the
			  // predicate.  (We do the same thing in matchSingleStep.)
			  if (context.predPc < predPc)
				  context.setCounter(counter, 0);
			  context.predPc = predPc;
			  context.incrementCounter(counter);
			  j = code[pc++];
			  pc += j;
			  boolean doContinue;
			  int savePosition = context.currentPosition;
			  try {
				  context.currentPosition = context.counters[counter];
				  doContinue = evalToPredicate(context, pc - 1);
			  } finally {
				  context.currentPosition =  savePosition;
			  }
			  if (doContinue)
				  continue;
			  else
				  return true;
		  default:
			  if (opcode < 0)
				  ; // end of Step - ignore FIXME
			  else
				  error("unknown upcode "+opcode);
		  }
	  }
  }

  public boolean evalUnionStep(XPathContext context, int levels[], int level,
							   int unionPc, NodeConsumer consumer)
  {
	  NodeList children = context.getCurrentNode().getChildNodes();
	  int childCount = children.getLength();
	  for ( int i = 0; i < childCount; i++ ) {
		  try	{
			  Node child = children.item(i);
			  context.pushNode(child);
			  if (! evalUnionStepChild(context, levels, level, unionPc, consumer))
				  return false;
		  }
		  finally {
			  context.popNode();
		  }
	  }
	  return true;
  }

  public boolean evalUnionStepChild(XPathContext context, int levels[],
								    int level, int unionPc,
								    NodeConsumer consumer)
  {
	  // levels[i] is the last level-number that succeeded for part i.
	  boolean doConsume = false;
	  boolean doRecurse = false;
	  int saveStepPc = context.stepPc;
	  int savePredPc = context.predPc;
	  int unionParts = code[unionPc-1];
	  for (int i = 0;  i < unionParts;  i++) {
		  int pathPc = i+1==unionParts ? unionPc - 2 - unionParts
			  : unionPc + code[unionPc-unionParts+i] - 1;
		  int stepCount = code[pathPc-1];
		  if (level >= stepCount)
			  continue;
		  int stepStart = code[pathPc-1-stepCount+level]+pathPc;
		  int pathEnd = pathPc - 1 - stepCount;
		  int stepEnd = level+1==stepCount ? pathEnd
			  : pathPc + code[pathPc-stepCount+level];
		  if (levels[i] == level) {
			  context.stepPc = stepStart;
			  context.predPc = stepStart;
			  stepStart = matchSingleStep(context, stepStart, stepEnd);
			  if (stepStart > 0) {
				  levels[i]++;
				  if (stepStart == pathEnd)
					  doConsume = true;
				  else
					  doRecurse = true;
			  }
		  }
	  }
	  context.stepPc = saveStepPc;
	  context.predPc = savePredPc;
	  if (doConsume) {
		  if (! consumer.consume(context.getCurrentNode()))
			  return false;
	  }

	  if (doRecurse
		  && ! evalUnionStep(context, levels, level+1, unionPc, consumer))
		  return false;
	  for (int i = 0;  i < unionParts;  i++) {
		  if (levels[i] > level)
			  levels[i] = level;
	  }
	  return true;
  }

  public final double evalToNumber(XPathContext context, int opcodepc)
  {
	  eval(context, opcodepc, NUMBER_RESULT);
	  return context.numResult;
  }

  public final boolean evalToBoolean(XPathContext context, int opcodepc)
  {
	  eval(context, opcodepc, BOOLEAN_RESULT);
	  return context.boolResult;
  }

  /**
   * Eval to a Boolean without specifying an opcode. For public use.
   */
  public final boolean evalToBoolean(XPathContext context )
  {
	  evalToBoolean(context, codeLen - 1);
	  return context.boolResult;
  }

  public final String evalToString( XPathContext context )
  {
	  context.stepPc = 0;
	  context.predPc = 0;
	  return evalToString( context, codeLen - 1);
  }

  public final String evalToString(XPathContext context, int opcodepc)
  {
	  int saveKind = context.consumeKind;
	  Object saveObject = context.consumeObject;
	  String result;
	  try {
		  context.consumeKind = XPathContext.CONSUME_STRING;
		  context.consumeObject = null;
		  evalToNodes(context, opcodepc, context);
		  context.resultType = STRING_RESULT;
		  if (context.consumeObject == null)
		    {
		      result = "";
		    }
		  else
		    {
		      result = context.consumeObject.toString();
		    }
	  }
	  finally {
		  context.consumeKind = saveKind;
		  context.consumeObject = saveObject;
	  }
	  return result;
  }

  /** 
   * Same as evalToBoolean, but a number is interpreted
   * as position()==NUM. 
   */
  public boolean evalToPredicate(XPathContext context, int opcodepc)
  {
	  eval(context, opcodepc, PREDICATE_RESULT);
	  return context.boolResult;
  }

  public final double evalToNumber(XPathContext context)
  {
	  return evalToNumber(context, codeLen-1);
  }

  private boolean matchPath(XPathContext context, int pc, int stop)
  {
	  int start = pc;
	  pc = stop;
	  if (pc == start)
		  return true;
	  pc--;
	  int stepLen = code[pc];
	  int stepEnd = pc;
	  pc += stepLen; // Note that stepLen is a negative offset.
	  context.stepPc = pc;
	  int axis = code[pc];
	  //int ancestorDepth = context.getAncestorDepth();
	  if (axis != OP_AXIS_FIRST+AXIS_DESCENDANT_OR_SELF
		  && ! matchStep(context, pc, stepEnd))
		  return false;
	  Node curNode = context.getCurrentNode();
	  try {
		  context.popNode();
		  return matchPath(context, pc, stop);
	  }
	  finally {
		  context.pushNode(curNode);
	  }
  }

  private boolean matchStep(XPathContext context, int pc, int stop)
  {
	  while (pc < stop) {
		  pc = matchSingleStep(context, pc, stop);
		  if (pc < 0)
			  return false;
	  }
	  return true;
  }

  /** 
   * Match a single Step.
   * @return -1 if not a match; pc for next step if matched so far.
   */
  int matchSingleStep(XPathContext context, int pc, int stop)
  {
	  int axis = AXIS_CHILD;
	  Node node = context.getCurrentNode();
	  for (;;) {
		  if (pc == stop) {
			  return pc;
		  }
		  short opcode = code[pc];
		  pc++;
		  if (opcode >= OP_AXIS_FIRST
			  && opcode < OP_AXIS_FIRST + COUNT_OP_AXIS) {
			  axis = opcode - OP_AXIS_FIRST;
			  if (axis == AXIS_ATTRIBUTE
				  && node.getNodeType() != Node.ATTRIBUTE_NODE)
				  return pc - 1;
			  continue;
		  }
		  int j;
		  switch (opcode) {
		  case OP_SIMPLE_NAMETEST:
			  {
				Object namespace = values[code[pc]];
				Object name = values[code[pc+1]];
				if ( node.getNodeType () == princpalNodeType && node.getLocalName () == name && node.getNamespaceURI () == context.getNamespaceURI ((String) namespace))
				  {
					pc+=2;
					continue;
				  }
				else 
				  { 
					return -1;
				  }
			  }
		  case OP_NODE:
		  case OP_ANY_NAMETEST:
			  if (! (node.getNodeType () == princpalNodeType))
				{
				  return -1;
				}
			  continue;
		  case OP_ROOT:
			  if (node.getNodeType() != Node.DOCUMENT_NODE)
				  return -1;
			  continue;
		  case OP_TEXT:
			  if (node.getNodeType() != Node.TEXT_NODE)
				  return -1;
			  continue;
		  case OP_COMMENT:
			  if (node.getNodeType() != Node.COMMENT_NODE)
				  return -1;
			  continue;
		  case OP_NAMESPACE_NAMETEST:
		    if (! (node.getNamespaceURI () == context.getNamespaceURI ((String) values[code[pc]])))
		      {
			return -1;
		      }
		    pc++;
		    continue;
		  case OP_PROCESSING_INSTRUCTION:
			  if (node.getNodeType() != Node.PROCESSING_INSTRUCTION_NODE)
				  return -1;
			  j = code[pc++];
			  if (j >= 0) {
				  Object name = values[j];
				  if (! node.getNodeName().equals(name))
					  return -1;
			  }
			  continue;
		  case OP_PREDICATE:
			  // See explanation of this code in evalPath.
			  int predPc = pc - 1;
			  int counter = code[pc++];
			  if (context.predPc < predPc)
				  context.setCounter(counter, 0);
			  context.predPc = predPc;
			  context.incrementCounter(counter);
			  j = code[pc++];
			  pc += j;
			  boolean doContinue;
			  int savePosition = context.currentPosition;
			  try {
				  context.currentPosition = context.counters[counter];
				  doContinue = evalToPredicate(context, pc - 1);
			  } finally {
				  context.currentPosition =  savePosition;
			  }
			  if (doContinue)
				  continue;
			  else
				  return -1;
		  }
	  }
  }

  public boolean match(XPathContext context, int pc)
  {
	  short opcode = code[pc];
	  switch (opcode) {
	  case OP_UNION:
		  int numOperands = code[pc-1];
		  for (int i = 0;  ; ) {
			  if (++i == numOperands)
				  return match(context, pc - 2 - numOperands);
			  else if (match(context, code[pc-numOperands+i] + pc - 1))
				  return true;
		  }
	  case OP_LOCATIONPATH:
		  int opcodepc = pc;
		  int numSteps = code[--pc];
		  pc -= numSteps;
		  return matchPath(context, code[pc] + opcodepc, pc);
	  case OP_FUNCTION_CALL:
		   // handle id(key) FIXME
	  }
	  return false;
  }

  public boolean match(XPathContext context, Node node)
  {
	  try {
		  context.pushNode(node);
		  return match(context, codeLen-1);
	  }
	  finally {
		  context.popNode();
	  }
  }

  public boolean equals (Object o)
  {
    if (this == o) return true;
    if (!(o instanceof XPath)) return false;

    XPath other = (XPath) o;
    if (codeLen != other.codeLen) return false;
    if (valLen != other.valLen) return false;

    for (int i = 0; i < codeLen; i++)
      if (code[i] != other.code[i])
	return false;

    for (int i = 0; i < valLen; i++)
      if (values[i] != other.values[i])
	return false;

    return true;
  }

  // -------------------------------------------------------------------------
  // XPath utility functions 

  /**
   * Returns the first node satisfying the xpath query
   * <code>query</code> using the node passed in as the context.
   */
  public static Node findFirstNode (Node contextNode, String query)
  {
    return (findFirstNode (contextNode, XPathParser.singleton.parse (query)));
  }

  /**
   * Returns the first node satisfying the xpath query
   * <code>query</code> using the node passed in as the context.
   */
  public static Node findFirstNode (Node contextNode, XPath query)
  {
    NodeList nodes = findNodes (contextNode, query, false);
    if (nodes.getLength () > 0)
      { 
	return (nodes.item (0));
      }

    return (null);
  }

  /**
   * Returns the first node satisfying the xpath query
   * <code>query</code> using the contect passed in.
   */
  public static Node findFirstNode (XPathContext context, String query)
  {
    return findFirstNode (context, XPathParser.singleton.parse (query));
  }

  /**
   * Returns the first node satisfying the xpath query
   * <code>query</code> using the contect passed in.
   */
  public static Node findFirstNode (XPathContext context, XPath query)
  {
    long time = System.currentTimeMillis ();
    NodeList nodes = findNodes (context, query, false);
    if (nodes.getLength () > 0)
      { 
	return nodes.item (0);
      }
    time = System.currentTimeMillis () - time;
//       Logger.log (XPath.class, "findFirstNode = #" + query.query + "# =" + time);
    xpathTime += time;
    return null;
  }

  public static Node findFirstNode (XPathContext context, Node contextNode, XPath xpath)
  {
    if (context != null)
      {
	context.resetAllbutPrefixes ();
	context.pushNode (contextNode);
	return findFirstNode (context, xpath);
      }
    else
      {
	return findFirstNode (contextNode, xpath);
      }
  }    

  public static Node findFirstNode (XPathContext context, Node contextNode, String query)
  {
    if (context != null)
      {
	context.resetAllbutPrefixes ();
	context.pushNode (contextNode);
	return findFirstNode (context, query);
      }
    else
      {
	return findFirstNode (contextNode, query);
      }
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using the node passed in as the context.
   */
  public static NodeList findNodes (Node contextNode, String query)
  {
    // We use the a single xpathParser to translate the query
    // because it is sync. so safe
    return (findNodes (contextNode, XPathParser.singleton.parse (query)));
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using the context passed in.
   */
  public static NodeList findNodes (XPathContext context, String query)
  {
    // We use the a single xpathParser to translate the query
    // because it is sync. so safe
    return (findNodes (context, XPathParser.singleton.parse (query)));
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using the node passed in as the context.
   */
  public static NodeList findNodes (Node contextNode, XPath xpath) 
  {
    return (findNodes (contextNode, xpath, true));
  }

  /**
   * Returns the nodes satisfying the xpath query <code>query</code>
   * using the context passed in as the context.
   */
  public static NodeList findNodes (XPathContext context, XPath xpath) 
  {
    return (findNodes (context, xpath, true));
  }

  /**
   * Wrapper for evalToNodes.
   */
  public static NodeList findNodes (Node contextNode, XPath xpath, boolean many) 
  {  
    long time = System.currentTimeMillis ();
    XPathContext context = (XPathContext)xpathContextCache.get ();
    SimpleNodeList result = new SimpleNodeList (many);
    context.pushNode (contextNode);
    xpath.evalToNodes (context, result);
    xpathContextCache.put (context);
    time = System.currentTimeMillis () - time;
//     Logger.log (XPath.class, "findNodes #" + xpath.query + "# =" + time);
    xpathTime += time;
    return (result);
  }

  /**
   * Wrapper for evalToNodes.
   */
  public static NodeList findNodes (XPathContext context, XPath xpath, boolean many) 
  {
    SimpleNodeList result = new SimpleNodeList (many);
    xpath.evalToNodes (context, result);
    return result;
  }

  public static NodeList findNodes (XPathContext context, Node contextNode, XPath xpath)
  {
    if (context != null)
      {
	context.resetAllbutPrefixes ();
	context.pushNode (contextNode);
	return findNodes (context, xpath);
      }
    else
      {
	return findNodes (contextNode, xpath);
      }
  }

  public static NodeList findNodes (XPathContext context, Node contextNode, String query)
  {
    if (context != null)
      {
	context.resetAllbutPrefixes ();
	context.pushNode (contextNode);
	return findNodes (context, query);
      }
    else 
      {
	return findNodes (contextNode, query);
      }
  }

  /**
   * Evaluates the xpath query <code>query</code> to a string.
   */
  public static String findString (Node contextNode, String query) 
  {
    return (findString (contextNode, XPathParser.singleton.parse (query)));
  }

  /**
   * Evaluates the xpath query <code>xpath</code> to a string.
   */
  public static String findString (Node contextNode, XPath xpath) 
  {
    long time = System.currentTimeMillis ();
    XPathContext context = (XPathContext)xpathContextCache.get ();
    context.pushNode (contextNode);
    String stringVal = xpath.evalToString (context);
    xpathContextCache.put (context);
    time = System.currentTimeMillis () -time;
//     Logger.log (XPath.class, "findString #" + xpath.query + "# =" + time);
    xpathTime += time;
    return (stringVal);
  }

  public static String findString (XPathContext context, Node contextNode, XPath xpath)
  {
    if (context != null)
      {
	context.resetAllbutPrefixes ();
	context.pushNode (contextNode);
	return xpath.evalToString (context);
      }
    else
      {
	return findString (contextNode, xpath);
      }
  }

  public static String findString  (XPathContext context, Node contextNode, String query)
  {
    if (context != null)
      {
	context.resetAllbutPrefixes ();
	context.pushNode (contextNode);
	return XPathParser.singleton.parse (query).evalToString (context);
      }
    else
      {	
	return findString (contextNode, query);
      }
  }


  /**
   * Evaluates the xpath query <code>query</code> to a boolean.
   */
  public static boolean evalToBoolean (Node contextNode, String query) 
  {
    return (evalToBoolean (contextNode, XPathParser.singleton.parse (query)));
  }

  public static boolean evalToBoolean (Node contextNode, XPath xpath) 
  {
    XPathContext context = (XPathContext)xpathContextCache.get ();
    context.pushNode (contextNode);
    boolean boolVal = xpath.evalToBoolean (context);
    xpathContextCache.put (context);
    return (boolVal);
  }    
    
  public static boolean evalToBoolean (XPathContext xpathContext, Node contextNode, String query) 
  {
    return evalToBoolean (xpathContext, contextNode, XPathParser.singleton.parse (query));
  }

  public static boolean evalToBoolean (XPathContext xpathContext, Node contextNode, XPath xpath) 
  {
    xpathContext.resetAllbutPrefixes ();
    xpathContext.pushNode (contextNode);
    return xpath.evalToBoolean (xpathContext);
  }

  /**
   * main function to test your XPaths.
   */
  public static void main (String[] args)
    throws IOException
  {
    if (args.length != 2)
      {
        System.out.println ("usage: " + XPath.class.getName () + " <url or filename> <xpath query>");
        System.exit (1);
      }

    // first create a document from the URL of File.
    Document doc = null;
    try
      {
        doc = DOMlet.create (new URL (args[0]));
      }
    catch (MalformedURLException e)
      {
        try 
          {
            doc = DOMlet.create (new FileInputStream (args[0]), args[0]);
          }
        catch (FileNotFoundException e2)
          {
            System.err.println ("Invalid URL or filename: " + args[0]);
            System.exit (1);
          }
      }

    XPath xpath = new XPath();
    XPathParser parser = new XPathParser (args[1], xpath);
    parser.parse ();

    System.out.print ("Parsed len: "+xpath.codeLen+" {");
    for (int j = 0;  j < xpath.codeLen;  j++)
      {
        System.out.print (" " + xpath.code[j]);
      }
    System.out.println ("}");

    XPathContext context = new XPathContext ();
    context.pushNode (doc);
    System.out.println ("context = " + doc);

    try 
      {
	Object result = xpath.eval (context);

	result = findNodes (doc, args[1]);
    
	if (result instanceof NodeList) 
	  {
	    NodeList output = (NodeList) result;

	    int count = output.getLength ();
	    System.out.println ();
	    System.out.println ("#matches:" + count);
	    System.out.println ();
	    for (int j = 0;  j < count;  j++)
	      {
		Node node = output.item (j);
		System.out.println (j + ": " + XMLWriter.getXML (node));

		if (args.length > 2) 
		  {
		    XPath matcher = new XPath();
		    XPathParser mparser = new XPathParser (args[2], matcher);
		    mparser.parse ();
		    System.out.println ("matches: "+matcher.match (context, node));
		  }
	      }
	  }
	else
	  {
	    System.err.println ("result: "+result);
	  }
      }
    catch (Exception e)
      {
	Logger.log (XPath.class, e);
      }
    // if we passed in a third argument, do an evalToString
    System.out.println ("evalToString () = " + xpath.evalToString (context));
  }
}
