#!/usr/bin/perl -w
#
# Check java/C source code for basic standards like
# copyright and $tvt: checksrc.pl,v 1.2 2001/11/06 21:10:11 samc Exp $ tags.
#
# Use '-n' to disable actual modifictions to your files.
#
# Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
# See the file "license.terms" for details.
#
# [SBC]
#
# $tvt: checksrc.pl,v 1.2 2001/11/06 21:10:11 samc Exp $

use strict;
use warnings;

use Getopt::Std;

my $comment =
"/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved. See the file \"COPYING\" for details.
 *
 * \$tvt\$
 */
";

our($opt_h, $opt_n);

sub print_usage
{
  print "usage : checksrc.pl [options] [<files ...>]\n";
  print "If no files are specified the all .java files under the current";
  print "working directory are scanned";
  print "";
  print "Options are:\n";
  print "\t-h\t\tprint this usage message\n";
  print "\t-n\t\tdon't actually change any files\n";
}

if (!getopts('nh'))
  {
    print ("Try \`$0 -h' for more information.\n");
    exit (-1);
  }

if ($opt_h)
  {
    print_usage ();
    exit (0);
  }

if (@ARGV == 0)
  {
    @ARGV = split ('\n',`find -name "*.java"`);
  }

my $foundcomment = 1;
my $foundtag = 0;
my $donecomment = 1;
my $line;
$^I="";

while (<>)
{
  if (eof)
    {
      close ARGV if eof; 
      $donecomment = 0;
      $donecomment = 0;
      $foundtag = 0;
    }

  if ($. == 1)
    {
      if (/^\/\*/)
        {
          $foundcomment = 1;
        }
      else
        {
          print (STDERR "$ARGV: no initial comment");
          if (!$opt_n) 
            { 
              print (STDERR ": fixing..."); 
              s/^/$comment/;
            }
          print (STDERR "\n");
        }
    }
  
  if (/\$Id.*\$/) 
    {
      print (STDERR "$ARGV:$.: found old \$Id\$ tag");
      if (!$opt_n)
        {
          print (STDERR ": fixing...");
          s/\$Id.*\$/\$tvt\$/;
        }
      print (STDERR "\n");
    }

  if (/\$tvt.*\$/) 
    { 
      $foundtag = 1; 
    }

  if ($foundcomment && ! $donecomment && /\*\//)
    {
      $donecomment = 1;
      if (!$foundtag)
        {
          print (STDERR "$ARGV:$.: no \$tvt\$ tag in initial comment");
          if (!$opt_n) 
            { 
              print (STDERR ": fixing..."); 
              s/\*\//\* \$tvt\$\n \*\//;
            } 
          print (STDERR "\n");
        }
    }

  # look for one or more spaces at the end of a line.
  if (/ +\n/) 
    { 
      print (STDERR "$ARGV:$.: found trailing white space at"); 
      if (!$opt_n) 
        { 
          print (STDERR ": fixing..."); 
          s/ +\n/\n/;
        } 
      print (STDERR "\n");
    }
  print;
}

exit 0;
