package com.pocketlinux.content;

import com.pocketlinux.dom.DOMlet;

import org.xoe.test.Test;
import org.xoe.test.AssertionFailedException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Vector;

public class IconsTest extends Test
{
  private static final String JPG_ICON_PKGA = "icon-test-pkga";
  private static final String JPG_ICON_PKGB = "icon-test-pkgb";

  protected void test()
    throws AssertionFailedException
  {
    // if we fail, we still want to clean up the content-types we install
    // so that they don't interfer with other tests
    try
      { 
	// <Setup>
	DefaultContentTypeResolver.getInstance().putMapping("jpg", "image/jpeg", this);
	DOMlet doc = new DOMlet();
	ContentElement ce = new ContentElement(new URL("stash:/some.jpg"), doc, null);
	ContentType jpgType = DefaultContentTypeResolver.getContentType ("jpg");
	ce.setType (jpgType);
	URL[] urls = new URL[] { new URL("stash://0"),new URL("stash://1"),new URL("stash://2")};
	// </Setup>
	
	IconManager.setIcon("jpg", urls[0]);
	ASSERT(IconManager.getIcon(ce) == urls[0], "IconManager.getIcon(ce) == urls[0]");
	IconManager.setIcon("jpg", urls[1]);
	ASSERT(IconManager.getIcon(ce) == urls[1], "IconManager.getIcon(ce) == urls[1]");
	ASSERT(IconManager.getIcon(jpgType) == urls[1], "IconManager.getIcon(ce) == urls[1]");

	// installs "jpg"->"pkga.gif" but we expect the user pref to override it anyway
	installPackage(JPG_ICON_PKGA);
	ASSERT(IconManager.getIcon(ce) == urls[1], "IconManager.getIcon(ce) == urls[1]");
	ASSERT(IconManager.getIcon(jpgType) == urls[1], "IconManager.getIcon(ce) == urls[1]");
	IconManager.unsetIcon("jpg");
	// now we expect to get pkga.gif
	ASSERT(IconManager.getIcon(ce).toString().endsWith("/pkga.gif"),
	       "IconManager.getIcon(ce).toString().endsWith(\"/pkga.gif\")");
	ASSERT(IconManager.getIcon(jpgType).toString().endsWith("/pkga.gif"),
	       "IconManager.getIcon(jpgType).toString().endsWith(\"/pkga.gif\")");

	// installs "jpg"->"pkgb.gif"
	installPackage(JPG_ICON_PKGB);
	ASSERT(IconManager.getIcon(ce).toString().endsWith("/pkgb.gif"),
	       "IconManager.getIcon(ce).toString().endsWith(\"/pkgb.gif\")");
	ASSERT(IconManager.getIcon(jpgType).toString().endsWith("/pkgb.gif"),
	       "IconManager.getIcon(jpgType).toString().endsWith(\"/pkgb.gif\")");

	uninstallPackage(JPG_ICON_PKGB);
	ASSERT(IconManager.getIcon(ce).toString().endsWith("/pkga.gif"),
	       "IconManager.getIcon(ce).toString().endsWith(\"/pkga.gif\")");
	ASSERT(IconManager.getIcon(jpgType).toString().endsWith("/pkga.gif"),
	       "IconManager.getIcon(jpgType).toString().endsWith(\"/pkga.gif\")");

	uninstallPackage(JPG_ICON_PKGA);
	// It should be the default icon.
	// Just make sure it's not the one we just uninstalled.
	ASSERT(!IconManager.getIcon(ce).toString().endsWith("/pkga.gif"),
	       "!IconManager.getIcon(ce).toString().endsWith(\"/pkga.gif\")");
	ASSERT(!IconManager.getIcon(jpgType).toString().endsWith("/pkga.gif"),
	       "!IconManager.getIcon(jpgType).toString().endsWith(\"/pkga.gif\")");
      }
    catch(MalformedURLException ex)
      {
	report(ex);
	ASSERT(false, "Unexpected MalformedURLException");
      }
    finally
      { // cleanup a bit
	DefaultContentTypeResolver.getInstance().removeProvider(this);	
      }
  }
}
