package org.xoe.test;

import com.pocketlinux.content.ContentElement;
import com.pocketlinux.content.Stash;
import com.pocketlinux.dom.DOMlet;
import com.pocketlinux.packages.IPackageXMLHandler;
import com.pocketlinux.packages.XoePackage;
import com.pocketlinux.packages.Proviso;
import com.pocketlinux.packages.Dependency;
import com.pocketlinux.packages.ResourceInstaller;
import com.pocketlinux.packages.ClassInstaller;
import com.pocketlinux.packages.ClassInstallerData;
import com.pocketlinux.packages.PackageDescriptionException;
import com.pocketlinux.packages.InstallException;
import com.pocketlinux.packages.ServiceInstaller;
import com.pocketlinux.xpath.XPath;
import com.pocketlinux.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.net.URL;
import java.util.Vector;

public class PackageTestInstaller implements IPackageXMLHandler
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/test";
  public static final String TESTS_FILENAME = "tests.xml";
  XoePackage xoePackage;

  public String[] getHandlerPredependencies ()
  {
    return new String[] {  ResourceInstaller.NAMESPACE,
			   ClassInstaller.NAMESPACE,
			   ServiceInstaller.NAMESPACE};
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = {IPackageXMLHandler.FUNCTION};
    return ret;
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTION))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTION))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI() };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String getNamespaceURI ()
  {
    return NAMESPACE;
  }

  public void setPackage (XoePackage p)
  {
    xoePackage = p;
  }

  public Vector generateRequires ()
  {
    return null;
  }

  public Vector generateConflicts ()
  {
    return null;
  }

  public Vector generateProvides ()
  {
    return null;
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    try
      {
	ContentElement ce = Stash.fetch ( new URL (Stash.getWorkingDirectory (xoePackage), TESTS_FILENAME));
	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), TESTS_FILENAME);

	    String cname = XPath.findString (doc, "test-suite");

	    if(cname != null)
	      {
		Class clazz = ((ClassInstallerData)xoePackage.getExData (ClassInstaller.NAMESPACE)).getClassLoader ().loadClass(cname);
		TestSuite suite = (TestSuite)clazz.newInstance();
		TestSuite.packageTests.addElement(suite);
		xoePackage.setExData(NAMESPACE, suite);
	      }
	  }
      }
    catch (Exception ex)
      { // lots of things could have happened... ClassNotFoundException, ClassCastException,
	// IOException...
	throw new InstallException (ex, xoePackage,
				    "An error occurred while looking for test suites");
      }
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    TestSuite.packageTests.removeElement(xoePackage.getExData(NAMESPACE));
  }
}

