package org.xoe.test;

import com.pocketlinux.content.Stash;
import java.util.StringTokenizer;
import java.util.Vector;

import org.xoe.core.Engine;

public class TestSuite extends Test
{
  static Vector packageTests = new Vector();

  private static final String[] TOPLEVEL = {"com", "org"};

  private static final String INDENT = "  ";

  private Vector m_vChildren = new Vector();
  private Test[] m_aChildren = null;

  protected void addTestSuite(String name)
  {
    String pname = this.getName().substring(0, this.getName().lastIndexOf('.')) + "." + name;

    try
      {
	addTest(getSuite(pname));
      }
    catch(ClassNotFoundException ex)
      {
	addTest(new UnimplementedTestSuite(pname));
      }
    catch(Exception ex)
      {
	System.err.println("WARNING: Unexpectedly failed to instantiate test suite " + pname);
	ex.printStackTrace();
      }
  }

  protected String getBanner()
  {
    if(this.getClass() == TestSuite.class)
      {
	return "Running tests...";
      }
    else
      {
	return "Test Suite: " + getName();
      }
  }

  protected void addTest(Test test)
  {
    m_vChildren.addElement(test);
    test.setParent(this);
    m_aChildren = null;
  }

  private Test[] getChildren()
  {
    if ( m_aChildren == null && m_vChildren != null)
      {
	m_aChildren = (Test[])m_vChildren.toArray(new Test[m_vChildren.size()]);
      }
    return m_aChildren;
  }

  void setParent(TestSuite parent)
  {
    super.setParent(parent);
    Test[] tests = getChildren();

    for (int i = 0; tests != null && i < tests.length; i++)
      {
	tests[i].setParent(this);
      }
  }

  final void childReport(String report)
  {
    StringBuffer strb = new StringBuffer();

    StringTokenizer tok = new StringTokenizer (report, "\n", true);
    while (tok.hasMoreTokens ())
      {
	String t = tok.nextToken ();
	if (t.equals ("\n"))
	  {
	    strb.append("\n");
	  }
	else
	  {
	    strb.append(INDENT + t);
	  }
      }
    report(strb.toString());
  }

  public void reset()
  {
    Test[] tests = getChildren();

    for (int i = 0; tests != null && i < tests.length; i++)
      {
	tests[i].reset();
      }
    super.reset();
  }

  public void test()
    throws AssertionFailedException
  {
    if (this instanceof UnimplementedTestSuite)
      {
	ASSERT ( false, "UNIMPLEMENTED: " + ((UnimplementedTestSuite)this).getOriginalName());
      }

    int nfail = 0;
    Test[] tests = getChildren();
    for (int i = 0; i < tests.length; i++)
      {
	tests[i].execute();
	if(!tests[i].getSuccess())
	  {
	    nfail++;
	  }
      }
    ASSERT ( nfail == 0, nfail + " / " + tests.length + " Subtests FAILED");
  }

  private static void printUsage()
  {
    System.err.println("Usage:");

    System.err.println(TestSuite.class.getName() + " <package to test (com|org|org.xoe|com.pocketlinux.content)> ...");
    System.err.println("    Test one or more specific packages");
    System.err.println("");
    System.err.println(TestSuite.class.getName() + " [-w]");
    System.err.println("    Test all packages");
    System.err.println("      -w: Warn about packages with unimplemented test suites");
    System.err.println("");


  }

  public static void main(String args[])
    throws Exception
  {
    Engine.bootstrap();

    TestSuite master = new TestSuite();

    if (args.length > 0)
      {
	int i = 0;
	while (i < args.length)
	  {
	    if ( args[i].equals("-p"))
	      {
		i++;
		if(i == args.length)
		  {
		    printUsage();
		    System.exit(1);
		  }

		try
		  {
		    master.installPackage(args[i]);
		  }
		catch(Exception ex)
		  {
		    System.err.println("No such package: " + args[i] + "\n");
		    printUsage();
		    System.exit(1);
		  }
		i++;
	      }
	    else
	      {
		try
		  {
		    // getSuite throws an exception if the test suite doesn't exist
		    // or can't be instantiated
		    master.addTest(getSuite(args[i]));
		  }
		catch(ClassNotFoundException ex)
		  {
		    System.err.println("No such test suite: " + args[i] + "\n");
		    printUsage();
		    System.exit(1);
		  }
		catch(Exception ex)
		  {
		    ex.printStackTrace();
		    System.exit(1);
		  }
		i++;
	      }
	  }
      }
    else
      {
	for(int i = 0; i < TOPLEVEL.length; i++)
	  {
	    try
	      {
		master.addTest(getSuite(TOPLEVEL[i]));
	      }
	    catch(Exception ex)
	      {
		System.err.println("WARNING: Unexpectedly failed to instantiate top-level test suite " + TOPLEVEL[i] + "");
		ex.printStackTrace();
		
	      }
	  }
      }

    for(int i = 0; i < packageTests.size(); i++)
      {
	master.addTest((Test)packageTests.elementAt(i));
      }
    master.execute();
  }

  private static TestSuite getSuite(String pname)
    throws InstantiationException, ClassNotFoundException, IllegalAccessException
  {
    if (pname.endsWith("."))
      {
	pname = pname.substring(0, pname.length() - 1);
      }

    int lastdot = pname.lastIndexOf('.');


    String cname = lastdot != -1 ? pname.substring(lastdot + 1) : pname;
    cname = pname + "." + Character.toUpperCase(cname.charAt(0)) + cname.substring(1) + "TestSuite";
    Class c = Class.forName(cname);
    TestSuite suite = (TestSuite)c.newInstance();
    return suite;
  }
}
