/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Main.java,v 1.17 2001/11/15 22:48:34 tim Exp $
 */
package org.xoe.applicationmanager;

import java.io.IOException;
import java.net.URL;

import org.w3c.dom.Document;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.services.IApplication;
import org.xoe.services.IIconResolver;


/**
 * ViewableDocumentBuilder for .xoe-config files.
 * Converts a .xoe-config file into a viewable document by launching
 * the specified application.
 */
public class Main
  implements IViewableDocumentBuilder, IIconResolver
{
  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    return (new String[] { FUNCTION_CONTENT_HANDLER, IIconResolver.FUNCTIONALITY });
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { IViewableDocumentBuilder.INTERFACE });
      }
    else if (func.equals (IIconResolver.FUNCTIONALITY))
      {
	return new String[] { IIconResolver.INTERFACE };
      }

    return (null);
  }

  public String[] getFeatures (String func)
  {
    if (func.equals (IViewableDocumentBuilder.FUNCTION_CONTENT_HANDLER) ||
	func.equals (IIconResolver.FUNCTIONALITY))
      {
        return (new String[] { FEATURE_DOCTYPE_NAME_NS, XOE_CONFIG_DOCTYPE_NAME_NS });
      }
    return (null);
  }

  public String[] getInformation ()
  {
    return (null);
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    ContentType type = content.getType ();
    IDocumentBuilder db = (IDocumentBuilder)type.getService (IDocumentBuilder.INTERFACE);
    if (db == null)
      {
        Logger.log (this, "getViewableDocumentBuilder: no document builder found for type " + content.getType ());
        return (null);
      }

    XDocument configDoc = db.getDocument (content);
    if (configDoc == null)
      {
        Logger.log (this, "getViewableDocumentBuilder: document builder returned null for " + content.getURL ());
        return (null);
      }

    URL configURL = configDoc.getURL ();
    if (configURL == null)
      {
        Logger.log (this, "getViewableDocumentBuilder: xoe-config with null url for " + content.getURL ());
        return (null);
      }

    try
      {
	SimpleServiceQuery query = new SimpleServiceQuery (IApplication.FUNCTIONALITY, IApplication.INTERFACE);
	query.addFeature (IApplication.FEATURE_CONFIG, configURL.toString ());
	IApplication app = (IApplication) ServiceLocator.findService (query);
	if (app != null)
	  {
            Logger.log (this, "launching: " + configURL.toString ());
	    // HACK - services are no longer initialized here!!
            // app.init (configDoc);
	    // HACK 
            XDocument thing = app.getViewableDocument ();
	    return (thing);
          }
      }
    catch (Exception e)
      {
        Logger.log (this, e);
      }

    return (null);
  }

  public URL getIconURL (ContentElement resource)
    throws IOException
  { // use the icon attribute from the xoe-config
    IDocumentBuilder db = (IDocumentBuilder)resource.getType ().getService (IDocumentBuilder.INTERFACE);
    if (db == null)
      {
        Logger.log (this, "getDocumentBuilder: no service found for type " + resource.getType ());
	return (null);
      }

    Document configDoc = db.getDocument (resource);
    String iconSrc = configDoc.getDocumentElement ().getAttribute ("icon");
    if (iconSrc == null)
      { // allow the default icon for this mimetype to be used.
	Logger.log (this, "no icon attr on " + resource.getURL ());
	return (null);
      }

    return new URL (resource.getURL (), iconSrc);
  }

  public URL getIconURL (ContentType type)
  { // allow the standard .lnk icon to be used
    return (null);
  }
}
