/**
 * AuthFactory - factory class to register and instanciate IAuth
 *               authentication objects (all static interface)
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author P.C.Mehlitz
 * @see IAuth
 * @see AuthProcess
 */

package org.xoe.auth;

import java.util.Hashtable;

public class AuthFactory
{
	static Hashtable authModules;
	
	public static boolean register ( String category, String authClassName )
	{
		try {
			Class authIfc  = Class.forName( "IAuth");
			Class authImpl = Class.forName( authClassName);
		
			if ( authIfc.isAssignableFrom( authImpl) ){
				if ( authModules == null ){
					authModules = new Hashtable();
				}
				
				authModules.put( category, authClassName);
				
				return true;
			}
		}
		catch ( ClassNotFoundException x ){
			x.printStackTrace();
		}
		
		return false;
	}
	
	public static boolean registerDefault ( String authClassName )
	{
		return register( "default", authClassName);
	}
	
	public static IAuth getInstance ( String category )
	{
		Object obj;
		String clsName;
		IAuth  auth = null;
		
		if ( authModules != null ){
			obj = authModules.get( category);
			if ( obj != null ){
				if ( obj instanceof String ){
					try {
						Class clazz = Class.forName( (String) obj);
						auth = (IAuth) clazz.newInstance();
						
						authModules.put( category, auth);
					}
					catch ( ClassNotFoundException x1 ){
						x1.printStackTrace();
					}
					catch ( InstantiationException x2 ){
						x2.printStackTrace();
					}
					catch ( IllegalAccessException x3 ){
						x3.printStackTrace();
					}
				}
				else if ( obj instanceof IAuth ){
					auth = (IAuth) obj;
				}
			}
		}
		
		return auth;
	}
	
	public static IAuth getDefaultInstance ()
	{
		return getInstance( "default");
	}
}
