#!/bin/sh
#
# Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
# See the file "license.terms" for details.
#
# Run this script to build all the packages in the xoe modules.
# Set XOE_HOST_ARCH if you want to build for a ceratin architecture.
# If you pass in a directory as an argument the resulting XOE packages
# will be put there.
#
# [SBC]
#
# $tvt: build-all.sh,v 1.27 2001/11/30 00:22:16 samc Exp $

canonical="build-all.sh"
packagedir=`dirname $0`
export xoebuildarch="i386"
export DATE=`date +%Y%m%d`
export TIME=`date +%H%M`

##
# helper functions
##
print_usage ()
{
cat >&2 <<END
Usage: $canonical [options] 

   -h        - print this help message
   -s        - stop on first error
   -f        - use fast build option
END
}

count=0
while getopts fsh opt; do
  case "$opt" in
    f) build_options="-f" ;;
    s) stop_on_error="yes" ;;
    h) print_usage; die ;;
    \?) echo "Use '-h' for more information"; die ;;
  esac
  count=$((count+1))
done
shift $count

PACKAGES="architecture \
          libaudio \
          libmad \
	  libmp3/clib \
	  libmp3/javalib \
	  libcanna/clib \
	  libcanna/javalib \
	  device-pda \
	  device-pda-emulator \
	  device-pda-ipaq \
	  device-pda-ipaq-emulator \
	  device-linux/clib \
	  device-linux/javalib \
	  http \
	  mime \
	  soap \
	  package-server \
          sync-core \
	  xml-renderer-core \
          xoe-interfaces \
	  image-core \
	  theme-default-color \
          history \
	  xhtml \
	  xhtml-services \
	  xhtml-services \
	  xhtml-plugin/clib \
	  xhtml-plugin/javalib \
          xoe-ml \
	  wml \
	  html/clib \
	  html/javalib \
	  application-manager \
	  url-viewer \
	  pdf \
	  xoe-servlet \
	  rtsp \
	  tvt-shell \
          control-panel \
          email \
	  sync \
	  package-manager \
	  tvt-web \
	  tvt-addressbook \
	  tvt-themes \
	  tvt-calculator \
	  tvt-mediaplayer \
          tvt-memo \
	  chatroom \
	  task-www" 


DEST=$1

summary ()
{
  if [ -n "$FAILED" ]; then
    echo "The following packages failed to build:"
    for package in $FAILED; do 
      echo "*** $package"
    done
  fi
  if [ -n "$PASSED" ]; then
    echo ""
    echo "The following packages built:"
    for package in $PASSED; do
      echo "*** $package"
    done
  fi  
#  exit $1
}

log()
{
  rm -f xoe-build-log
  if [ -n "$FAILED" ]; then
    for package in $FAILED; do 
      echo "Building $package for $xoebuildarch-$DATE-$TIME" >> xoe-build-log
      echo "Failure" >> xoe-build-log
    done
  fi
  if [ -n "$PASSED" ]; then
    echo ""
    for package in $PASSED; do
      echo "Building $package for $xoebuildarch-$DATE-$TIME" >> xoe-build-log
      echo "Success" >> xoe-build-log
    done
  fi  
  exit $1
}

trap "summary 1" SIGINT SIGKILL
trap "log 1" SIGINT SIGKILL

for package in $PACKAGES; do 
  time -p ../tools/xoe-buildpackage $build_options $packagedir/$package $DEST; 
  if [ $? != 0 ]; then 
    FAILED="$FAILED $package"; 
    [ "$stop_on_error" = "yes" ] && break
  else
    PASSED="$PASSED $package";
  fi
  echo ""
done; 

summary 0
log 0
