/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * $tvt: Main.java,v 1.8 2001/11/16 21:35:25 ben Exp $ 
 */
package org.xoe.chatroom;

import org.xoe.core.util.XMLWriter;
import org.xoe.display.dom.xhtml.Input;
import org.xoe.display.dom.xhtml.Select;
import org.xoe.display.dom.xhtml.TextArea;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.display.dom.ViewElement;
import org.xoe.servlet.XoeServlet;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;

import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;

import org.xoe.soap.ISoapService;
import org.xoe.soap.SoapService;
import org.xoe.soap.SoapServiceException;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.io.IOException;
import java.lang.Thread;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;


/**
 * Internet Relay Chat.
 * PocketLinux application for chatting, using standard Internet IRC servers.
 */
public class Main
  extends XoeServlet implements Runnable
{
  // commands
  public static final String SEND = "SEND";
  public static final String CLEAR = "CLEAR";
  public static final String REPLY = "REPLY";
  public static final String CANCEL = "CANCEL";
  public static final String LOGIN = "LOGIN";
  public static final String LOGOUT = "LOGOUT";
  public static final String REFRESH = "REFRESH";


  // ids on elements (used for bindings)
  public static final String MSG = "MSG";
  public static final String BOARD = "BOARD";
  public static final String MEMBERS = "MEMBERS";
  public static final String NICK = "NICK";
  public static final String USER = "USER";
  public static final String NAME = "NAME";
  public static final String SERVER = "SERVER";
  public static final String CHANNEL = "CHANNEL";

  private String m_soapURL;
  private String sUser;
  private String sServer;
  private String sChannel;
  private String sName;
  private String sNick;
  
  private ISoapService m_soapService;
  
  private Thread m_poller;
  
  private Input m_nick;
  private Input m_user;
  private Input m_name;
  private Input m_server;
  private Input m_channel;

  private Input m_message;
  private TextArea m_channelActivity;
  private TextArea m_memberList;


  public void init (ServletConfig config) throws javax.servlet.ServletException
  {
    m_soapURL = config.getInitParameter("SoapURL");
    super.init (config);
  }
  
  protected void doGet(HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    res.setContentType("text/html");
    String path = req.getPathInfo();


    if ( "".equals(path) )
      {
	XDocument doc = getRelativeDocument ("default.xhtml");
	doc.expandUsingXPathAttributes();

	m_server = (Input)doc.getElementById(SERVER);  
	m_channel = (Input)doc.getElementById(CHANNEL);
	m_user = (Input)doc.getElementById(USER);
	m_nick = (Input)doc.getElementById(NICK);
	m_name = (Input)doc.getElementById(NAME);

	res.getWriter().print(doc);
      }
    else if (STOP.equals(path)) 
      {

      }
    else if (LOGIN.equals(path))
      {
	sServer = m_server.getText();	 
	sChannel = m_channel.getText();
	sUser = m_user != null ? m_user.getText() : "";
	sNick = m_nick != null ? m_nick.getText() : "";
	sName = m_name != null ? m_name.getText() : "";
	Logger.log( this, "user=" + sUser );
	Logger.log( this, "nick=" + sNick );
	Logger.log( this, "name=" + sName );
	
	try{
	  Logger.log( this, "LOGIN called" );
	  connect();

	  XDocument doc = getRelativeDocument ("login.xhtml");	
	  res.getWriter ().print (doc);

	  //clear views
	  m_channelActivity = (TextArea)doc.getElementById(BOARD);
	  m_memberList = (TextArea)doc.getElementById(MEMBERS);
	  m_message = (Input)doc.getElementById(MSG);
	  
	  m_channelActivity.setText( "" );
	  m_memberList.setText( "" );
	  
	  //start the polling mechanism
	  m_poller = new Thread( this );
	  m_poller.start();
	}catch( SoapServiceException sse ){
	  
	  //sse.printStackTrace();
	  Logger.log( this, sse.toString() );
	  Dialog.note ("Login Failed!<br/>Reason: <b>" + sse.getMessage() + "</b>" );       	    
	  return;
	}		
      }
    else if (SEND.equals(path))
      {
	say( m_message.getText() );	    
	//clear the text panel
	clear();
      }
    else if (LOGOUT.equals(path))
      {
	//stop the polling mechanism
	m_poller.stop();	
	disconnect();	
	XDocument doc = getRelativeDocument ("default.xhtml");

	doc.expandUsingXPathAttributes();
	m_server = (Input)doc.getElementById(SERVER);  
	m_channel = (Input)doc.getElementById(CHANNEL);
	m_user = (Input)doc.getElementById(USER);
	m_nick = (Input)doc.getElementById(NICK);
	m_name = (Input)doc.getElementById(NAME);

	res.getWriter ().print (doc);
      }
  }


  private void connect () throws SoapServiceException
  {
	
    if( sNick.equals( "" ) && sUser.equals( "" ) && sName.equals( "" ) )
      throw new SoapServiceException ("Need a username");
    
    //if one of nick, user, or name is null, make it equal...
    if( sNick.equals( "" ) ){
      if( !sUser.equals( "" ) )
	sNick = sUser;
      else if( !sName.equals( "" ) )
		sNick = sName;
    }
    if( sUser.equals( "" ) ){
      if( !sNick.equals( "" ) )
	sUser = sNick;
      else if( !sName.equals( "" ) )
	sUser = sName;
    }
    if( sName.equals( "" ) ){
      if( !sUser.equals( "" ) )
	sName = sUser;
      else if( !sNick.equals( "" ) )
	sName = sNick;
    }
          
    try
      {
	connect(  sServer, 
		  6667, 
		  sChannel,
		  sNick,
		  sUser,
		  sName );
      }
    catch (Exception e)
      {
	throw new SoapServiceException (e.getMessage ());
      }
  }

  //ripped from PLIRC app
  private void connect( String serverAddress, int pPort, String channel, String nick, String user, String name )
    throws Exception
  {    
    Integer port = new Integer( pPort );
    
    Node result;
    
    Stack stack = new Stack();
    
    stack.push( name );
    stack.push( user );
    stack.push( nick );
    stack.push( channel );
    stack.push( port );
    stack.push( serverAddress );
        
    URL url = null;

    Logger.log( this, "SOAP URL= " + m_soapURL ); 
    url = new URL( m_soapURL );
    
    result = findSoapService().call("urn:irc:SoapChat",
				    "connect",
				    stack,
				    url  );
    
    String resp = XPath.findString (result, "//ns1:getResponse/return" );              	
  }

  private ISoapService findSoapService()
  {
    if (m_soapService == null)
      {
        m_soapService = (ISoapService) ServiceLocator.findImplementor (ISoapService.INTERFACE);

        if (m_soapService == null)
          {
            m_soapService = new SoapService ();
            ServiceLocator.registerService (m_soapService);
          }

      }

    return m_soapService;
  }
  
  private void clear ()
  {
    m_message.setText ("");
  }

  public void run()
  {   
    Thread myThread = Thread.currentThread();
    while ( m_poller == myThread) {
      try {
	
	refreshIt();

	m_poller.sleep(3000); //poll the soap server every 3 seconds
	
      } catch (InterruptedException e){
	// the VM doesn't want us to sleep anymore,
	// so get back to work
      }
      
    }
  }


  private void refreshIt()
  {    
    Logger.log( this, "PLIRC.refreshIt(): [" + sServer + "] [" + sChannel + "] [" + sUser + "]" );
    refreshIt ( sServer, sChannel, sUser );	
  }
  
  private void refreshIt( String serverAddress, String channel, String user )
  {
    Node result;
    
    Stack stack = new Stack();
    
    stack.push( user );
    stack.push( channel );
    stack.push( serverAddress );
    
    URL url = null;
    try{
      url = new URL( m_soapURL );
      
      
      if( findSoapService() == null )
	Logger.log( this, "soap service is null" );
      
      result = findSoapService().call("urn:irc:SoapChat",
				      "getMessageLog",
				      stack,
				      url  );
      
      Logger.log( this, XMLWriter.getXML( result ) );
      
      String resp = XPath.findString (result, "//ns1:getMessageLogResponse/return" );

      if( !resp.equals("") ){
	Logger.log( this, "SOAP EMAIL result: [" + resp + "]" );
	
	String temp = m_channelActivity.getText();
	//replace tabs with new lines (hack-around for xpath nuking of new lines)
	resp = resp.replace( '\t', '\n' );
	m_channelActivity.setText( temp + resp );
      }
      
      
      stack = new Stack();	    
      stack.push( user );
      stack.push( channel );
      stack.push( serverAddress );
      
      //refresh the member list
      result = findSoapService().call("urn:irc:SoapChat",
				      "getMemberList",
				      stack,
				      url  );

      //Logger.log( this, XMLWriter.getXML( result ) );
      
      resp = XPath.findString(result, "//ns1:getMemberListResponse/return" );
      
      if( !resp.equals("") ){
	//Logger.log( this, "SOAP EMAIL result: [" + resp + "]" );
	
	//String temp = m_memberList.getText();
	//replace tabs with new lines (hack-around for xpath nuking of new lines)
	resp = resp.replace( '\t', '\n' );
	m_memberList.setText( resp );
      }
      
      
    }catch( MalformedURLException e ){
      e.printStackTrace();
    }catch( SoapServiceException se ){
      se.printStackTrace();
    }	
  }


  private void disconnect(){
    
    disconnect(  sServer, sChannel, sUser );	
  }
  
  private void say( String message ){
    
    send( sServer, sChannel, sUser, message );
    
  }

  private void disconnect( String serverAddress, String channel, String user ){

    Node result;
    
    Stack stack = new Stack();
    
    stack.push( user );
    stack.push( channel );
    stack.push( serverAddress );
    
    URL url = null;
    try{
      url = new URL( m_soapURL );
      
      
      if( findSoapService() == null )
	Logger.log( this, "soap service is null" );
      
      result = findSoapService().call("urn:irc:SoapChat",
				      "disconnect",
				      stack,
				      url  );
      
      //Logger.log (this, XMLWriter.getXML (result));
      
      String resp =  XPath.findString (result, "//ns1:disconnectResponse/return" );
      
      //Logger.log( this, "SOAP EMAIL result: [" + resp + "]" );
      
      
    }catch( MalformedURLException e ){
      e.printStackTrace();
    }catch( SoapServiceException se ){
      se.printStackTrace();
    }	
  }


  private void send( String serverAddress, String channel, String user, String message ){
    
    Node result;
    
    Stack stack = new Stack();
    
    stack.push( message );
    stack.push( user );
    stack.push( channel );
    stack.push( serverAddress );
    
    URL url = null;
    try{
      url = new URL( m_soapURL );
      
      
      if( findSoapService() == null )
	Logger.log( this, "soap service is null" );
      
      result = findSoapService().call("urn:irc:SoapChat",
				      "send",
				      stack,
				      url  );
      
      //Logger.log( this, XMLWriter.getXML( result ) );
      
      String resp = XPath.findString (result, "//ns1:sendResponse/return");
      
      //Logger.log( this, "SOAP EMAIL result: [" + resp + "]" );
      
    }catch( MalformedURLException e ){
      e.printStackTrace();
    }catch( SoapServiceException se ){
      se.printStackTrace();
    }	
    
  }

} 
