/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Main.java,v 1.21 2001/11/15 23:17:16 samc Exp $
 */
package org.xoe.controlpanel;

import java.io.IOException;
import javax.servlet.ServletException;
import javax.servlet.ServletConfig;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.xoe.core.device.Device;
import org.xoe.core.device.IScreenRotate;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.servlet.XoeServlet;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * The control panel is a very simple XoeServlet that passes most 
 * functionality to other classes ("Capplets") based on the HttpServletRequest 
 * it is given.
 */
public class Main
  extends XoeServlet
{
  /**
   * We want history in this servlet so we can go back to the main
   * control panel page.
   */
  public void init (ServletConfig sc)
    throws ServletException
  {
    trackHistory ();
  }

  public void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    Document doc = null;
    String path = req.getPathInfo();

    Logger.log (this, "path = " + path);
    
    if (path.equals(""))
      {
	path = "index.xhtml";
        doc = getRelativeDocument (path);
      }
    // check for special paths
    else if (path.equals ("timezoneSelect"))
      {
        String timezone = req.getParameter ("timezone");
        TimeDate.displayTimezone (timezone, getDisplayedDocument ());
      }
    else if (path.equals ("interfaceSelect"))
      {
        String iface = req.getParameter ("interface");
        Network.displayData (iface, getDisplayedDocument ());
      }
    else if (path.equals ("setinterfacedata"))
      {
        Network.setData (req, getDisplayedDocument ());
      }
    else if (path.equals ("brightness"))
      {
        Power.setBacklight (getDisplayedDocument ());
      }
    else if (path.equals ("settime"))
      {
        TimeDate.setDate (req);
      }
    else if (path.equals ("setrotation"))
      {
        res.setContentType ("text/html");
        int rot = Integer.parseInt (req.getParameter ("rot"));
        IScreenRotate rotate = (IScreenRotate) ServiceLocator.findImplementor (IScreenRotate.INTERFACE);
        if (rotate == null)
          {
            Dialog.note ("no IScreenRotate service found");
            return;
          }
        rotate.setRotation (rot);
      }
    // check for "capplet" pages that need data to be filled in
    else if (path.equals ("time.xhtml"))
      {
        doc = getRelativeDocument (path);
        TimeDate.updateInfo (doc);
      }
    else if (path.equals ("power.xhtml"))
      {
        doc = getRelativeDocument (path);
        Power.updateInfo (doc);
      }
    else if (path.equals ("network.xhtml"))
      {
        doc = getRelativeDocument (path);
        Network.updateInfo (doc);
      }
    else if (path.equals ("system.xhtml"))
      {
        doc = getRelativeDocument (path);
        SystemInfo.updateInfo (doc);
      }
    else if (path.equals ("storage.xhtml"))
      {
        doc = getRelativeDocument (path);
        Storage.updateInfo (doc);
      }
    else
      {
        super.doGet (req,res);
      }

    if (doc != null)
      {
	res.getWriter ().print (doc);
      }
  }
}
