/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: Network.java,v 1.14 2001/11/16 02:40:07 samc Exp $
 */
package org.xoe.controlpanel;

import java.io.IOException;
import javax.servlet.http.HttpServletRequest;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.xoe.core.device.Device;
import org.xoe.core.device.INetwork;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.util.DOMUtils;
import org.xoe.display.dom.xhtml.Dialog;

class Network
{
  static void displayData (String ifaceName, Document doc)
  {
    INetwork network = (INetwork) ServiceLocator.findImplementor (INetwork.class.getName ());
    if (network == null)
      {
        Dialog.note ("no networking services found");
        return;
      }
    Element address = doc.getElementById ("ip");
    Element netmask = doc.getElementById ("netmask");
    Element gateway = doc.getElementById ("gateway");
    Element dhcp    = doc.getElementById ("dhcp");
    Element dns1    = doc.getElementById ("DNS1");
    Element dns2    = doc.getElementById ("DNS2");
    Element host    = doc.getElementById ("hostname");
    Element domain  = doc.getElementById ("domainname");
    try
      {
        String[] servers = network.getNameservers ();
        INetwork.NetworkInterface iface = network.getSettings (ifaceName);
        address.setAttribute ("value", iface.address);
        netmask.setAttribute ("value", iface.netmask);
        gateway.setAttribute ("value", iface.gateway);
        domain .setAttribute ("value", network.getDomainname ());
        host   .setAttribute ("value", network.getHostname ());
        if (servers.length > 0)
          {
            dns1.setAttribute ("value", servers[0]);
            if (servers.length > 1)
              {
                dns2.setAttribute ("value", servers[1]);
              }
          }
        if (iface.dhcp)
          {
            dhcp.setAttribute ("checked", "checked");
            address.setAttribute ("disabled", "disabled");
            netmask.setAttribute ("disabled", "disabled");
            gateway.setAttribute ("disabled", "disabled");
          }
        else if (dhcp.getAttribute ("checked")  != null)
          {
            dhcp.removeAttribute ("checked");
          }
      }
    catch (IOException e)
      {
        Logger.log (Network.class, e);
        Dialog.note ("error retrieving network information");
        return;
      }
  }

  static void updateInfo (Document doc)
  {
    INetwork network = (INetwork) ServiceLocator.findImplementor (INetwork.class.getName ());
    if (network == null)
      {
        Dialog.note ("no networking services found");
        return;
      }
    try
      {
        String[] devices = network.getNetworkInterfaces ();
        Element list = doc.getElementById ("interfaceList");
        if (devices.length > 0)
          {
            displayData (devices[0], doc);
          }
        for (int i = 0; i < devices.length; i++)
          {
            Element elem = doc.createElement ("option");
            elem.setAttribute ("value", devices[i]);
            elem.appendChild (doc.createTextNode (devices[i]));
            list.appendChild (elem);
          }
      }
    catch (IOException e)
      {
        Logger.log (Network.class, e);
      }
  }

  static void setData (HttpServletRequest req, Document doc)
  {
    INetwork network = (INetwork) ServiceLocator.findImplementor (INetwork.class.getName ());
    if (network == null)
      {
        Dialog.note ("no networking services found");
        return;
      }
    INetwork.NetworkInterface iface = new INetwork.NetworkInterface ();
    if (req.getParameter ("dhcp") != null)
      {
        iface.dhcp = true;
      }
    iface.name    = req.getParameter ("interface");
    iface.address = req.getParameter ("ip");
    iface.netmask = req.getParameter ("netmask");
    iface.gateway = req.getParameter ("gateway");
    try
      {
        network.applySettings (iface);
      }
    catch (IOException e)
      {
        Logger.log (Network.class, e);
        Dialog.note ("error setting network up network interface: " + e);
      }

    // handle three different permutation of the two
    // DNS servers
    String[] servers = new String[0];
    String dns1   = req.getParameter ("DNS1");
    String dns2   = req.getParameter ("DNS2");
    if (!dns1.equals (""))
      {
        if (dns2.equals (""))
          {
            servers = new String[1];
            servers[0] = dns1;
          }
        else
          {
            servers = new String[2];
            servers[0] = dns1;
            servers[1] = dns2;
          }
      }
    else if (!dns2.equals (""))
      {
        servers = new String[1];
        servers[0] = dns2;
      }
    try
      {
        network.setNameservers (servers);
      }
    catch (IOException e)
      {
        Logger.log (Network.class, e);
        Dialog.note ("error setting nameservers: " + e);
      }
    displayData (iface.name, doc);
  }
}
