/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: TimeDate.java,v 1.5 2001/11/10 00:30:31 samc Exp $
 */
package org.xoe.controlpanel;

import java.io.IOException;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import javax.servlet.http.HttpServletRequest;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.xoe.core.device.ISystem;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.display.dom.xhtml.Dialog;

class TimeDate
{
  static void setDate (HttpServletRequest req)
  {
    System.err.println ("setting date");
    ISystem system = (ISystem) ServiceLocator.findImplementor (ISystem.class.getName ());
    if (system == null)
      {
        Dialog.note ("No ISystem service found<br>Unable to set time");
        return;
      }
    String timezone = req.getParameter ("timezone");
    TimeZone.setDefault (TimeZone.getTimeZone (timezone));
    Calendar cal = Calendar.getInstance ();
    cal.set (Calendar.HOUR_OF_DAY , Integer.parseInt (req.getParameter ("hours"  )));
    cal.set (Calendar.MINUTE      , Integer.parseInt (req.getParameter ("minutes")));
    cal.set (Calendar.SECOND      , Integer.parseInt (req.getParameter ("seconds")));
    cal.set (Calendar.DAY_OF_MONTH, Integer.parseInt (req.getParameter ("day"    )));
    cal.set (Calendar.MONTH       , Integer.parseInt (req.getParameter ("month"  )));
    cal.set (Calendar.YEAR        , Integer.parseInt (req.getParameter ("year"   )));
    try
      {
        system.setTimeMillis (cal.getTime ().getTime ());
      }
    catch (IOException e)
      {
        Dialog.note ("Error setting system time");
      }
  }

  static void displayTimezone (String timezone, Document doc)
  {
    Calendar cal = new GregorianCalendar (TimeZone.getTimeZone (timezone));
    updateInfo (doc, cal);
  }

  static void updateInfo (Document doc)
  {
    updateInfo (doc, Calendar.getInstance ());
  }

  static void updateInfo (Document doc, Calendar cal)
  {
    Element timezone = doc.getElementById ("timezone");
    String[] ids = cal.getTimeZone ().getAvailableIDs ();
    String id = cal.getTimeZone ().getID ();
    if (timezone.getChildNodes ().getLength () == 0)
      {
        for (int i = 0; i < ids.length; i++)
          {
            Element option = doc.createElement ("option");
            option.setAttribute ("value", ids[i]);
            DOMUtils.setTextContent (option, ids[i]);
            timezone.appendChild (option);
            if (id.equals (ids[i]))
              {
                option.setAttribute ("selected", "selected");
              }
          }
      }

    // please excuse the long line length [SBC]
    doc.getElementById ("hours"  ).setAttribute ("value", Integer.toString (cal.get (Calendar.HOUR_OF_DAY)));
    doc.getElementById ("minutes").setAttribute ("value", Integer.toString (cal.get (Calendar.MINUTE)));
    doc.getElementById ("seconds").setAttribute ("value", Integer.toString (cal.get (Calendar.SECOND)));
    doc.getElementById ("day"    ).setAttribute ("value", Integer.toString (cal.get (Calendar.DAY_OF_MONTH)));
    doc.getElementById ("month"  ).setAttribute ("value", Integer.toString (cal.get (Calendar.MONTH)));
    doc.getElementById ("year"   ).setAttribute ("value", Integer.toString (cal.get (Calendar.YEAR)));
    //DOMUtils.setTextContent (td, devices[i]);
  }
}
