/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: DebugShell.java,v 1.4 2001/11/27 22:47:59 samc Exp $
 */
package org.xoe;

import java.util.Vector;
import java.io.IOException;

import org.xoe.core.IShell;
import org.xoe.core.Engine;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageManager;

public class DebugShell
  implements IShell
{
  private final static String [] INTERFACES = new String [] { IShell.INTERFACE };
  private final static String PROMPT = "(xoe) ";

  // ---------------------------------------------------------------------
  // IService

  public String[] getInterfaces (String func) { return INTERFACES; }
  public String[] getFeatures (String func)   { return null;       }
  public String[] getInformation ()           { return null;       }
  public String[] getFunctionalities ()       { return null;       }
  public void init (org.w3c.dom.Document configDoc) {}

  // ---------------------------------------------------------------------
  // IShell

  public void start ()
  {
    String command;
    System.out.println ("Welcome to the XOE debug shell.");
    System.out.println ("Copyright 2001 Transvirtual Technologies, Inc.");
    System.out.println ("XOE Version is " + Engine.getVersion ());
    System.out.println ("Type 'help' for a list of commands.");
    while (true)
      {
        System.out.print (PROMPT);
        Vector args = new Vector ();
        try
          {
            command = readCommand (args);
          }
        catch (IOException e)
          {
            System.err.println (e);
            break;
          }
        if (doCommand (command, args) == false)
          {
            break;
          }
      }
  }

  private boolean doCommand (String command, Vector args)
  {
    if (command.equals ("quit"))
      {
        System.out.println ("bye.");
        return false;
      }
    else if (command.equals ("help"))
      {
        System.out.println ("List of commands:\n");
        System.out.println ("list    - list installed packages");
        System.out.println ("install - install a package");
        System.out.println ("remove  - remove a package");
        System.out.println ("quit    - quit the debug shell");
        System.out.println ("help    - display this message");
      }
    else if (command.equals ("list"))
      {
        System.out.println ("Name                    Ver.    Description");
        System.out.println ("=======================-=======-============================================");
        PackageDescription[] descs = PackageManager.getInstalledPackages ();
        for (int i = 0; i < descs.length; i++)
          {
             String name    = descs[i].getName ();
             String desc    = descs[i].getTextDescription ();
             String version = descs[i].getVersion ().toString ();
             if (name.length () > 23)
               {
                 name = name.substring (0, 23);
               }
             else if (name.length () < 16)
               {
                 name += "\t";
               }
             if (version.length () > 7)
               {
                 version = version.substring (0, 7);
               }
             if (desc.length () > 32)
               {
                 desc = desc.substring (0, 32);
               }

             System.out.println (name + '\t' + version + '\t' + desc);
          }
      }
    else if (command.equals ("install"))
      {
        System.err.println ("unknown command: " + command);
      }
    else
      {
        System.err.println ("unknown command: " + command);
      }
    return true;
  }

  private String readCommand (Vector args)
    throws IOException
  {
    StringBuffer buffer = new StringBuffer ();
    char c;
    while (true)
      {
        c = (char) System.in.read ();
        if (c == '\n')
          {
            break;
          }
        buffer.append (c);
      }
    return buffer.toString ();
  }
}
