/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Network.c,v 1.3 2001/11/03 00:42:45 samc Exp $
 */
#include <jni.h>
#include <unistd.h>

#define MAX_HOSTNAME 64
#define MAX_DOMAINNAME 64

JNIEXPORT jstring JNICALL
Java_org_xoe_device_LinuxNetwork_getDomainname0 (JNIEnv* env, jobject obj)
{
  char domainname[MAX_DOMAINNAME];
  jstring string;
  if (getdomainname (domainname, MAX_HOSTNAME) == -1)
    {
      string = 0;
    }
  else
    {
      string = (*env)->NewStringUTF (env, domainname);
    }
  return string;
}

JNIEXPORT jboolean JNICALL
Java_org_xoe_device_LinuxNetwork_setDomainname0 (JNIEnv* env, jobject obj, jstring string)
{
  jboolean rtn = JNI_TRUE;
  const char * domainname = (*env)->GetStringUTFChars (env, string, 0);
  if (setdomainname (domainname, strlen (domainname)) == -1)
    {
      rtn = JNI_FALSE;
    }
  (*env)->ReleaseStringUTFChars (env, string, domainname);
  return rtn;
}

JNIEXPORT jstring JNICALL
Java_org_xoe_device_LinuxNetwork_getHostname0 (JNIEnv* env, jobject obj)
{
  char hostname[MAX_HOSTNAME];
  jstring string;
  if (gethostname(hostname, MAX_HOSTNAME) == -1)
    {
      string = 0;
    }
  else
    {
      string = (*env)->NewStringUTF (env, hostname);
    }
  return string;
}

JNIEXPORT jboolean JNICALL
Java_org_xoe_device_LinuxNetwork_setHostname0 (JNIEnv* env, jobject obj, jstring string)
{
  jboolean rtn = JNI_TRUE;
  const char * hostname = (*env)->GetStringUTFChars (env, string, 0);
  if (sethostname (hostname, strlen (hostname)) == -1)
    {
      rtn = JNI_FALSE;
    }
  (*env)->ReleaseStringUTFChars (env, string, hostname);
  return rtn;
}

#if 0

#include "util.h"

#define LINE_LENGTH 1024

/**
 * Get the names of all the network devices by parsing
 * /proc/net/dev.
 */
JNIEXPORT jstring JNICALL
Java_org_xoe_device_device_LinuxNetwork_getNetworkInterfaces0 (JNIEnv* env, jobject obj)
{
  char interfaces[LINE_LENGTH];
  jstring ifaces;
  int file = open ("/proc/net/dev", O_RDONLY);
  int nl_count = 0;
  int got_device = 0;
  int num_devices = 0;
  int len = 0;
  char c;
  if (file == -1)
    {
      return 0;
    }
  while (read (file, &c, 1) == 1)
    {
      // protect against buffer overrun.
      if (len >= LINE_LENGTH - 1)
        {
          len = LINE_LENGTH - 1;
          continue;
        }
      // a newline means that we start looking for a new device name
      if (c == '\n') 
        {
          nl_count++;
          got_device = 0;
          continue;
        }
      // The first two lines are ignored.  As is the rest of line after the ':'
      // and the initial whitespace.
      if (   nl_count < 2
          || got_device
          || c == ' ') 
        {
          continue;
        } 
      // If we read a ':' then we are done with the device name.
      if (c == ':')
        {
          got_device = 1;
          num_devices++;
        } 
      // otherwise add the character
      interfaces[len++] = c;
    }
  interfaces[len] = '\0';
  ifaces = (*env)->NewStringUTF (env, interfaces);
  return ifaces;
}

/**
 * Use ifconfig to setIP
 */
JNIEXPORT void JNICALL
Java_org_xoe_device_device_LinuxNetwork_setIP (JNIEnv* env, jobject obj, jstring device)
{
  char* argv[4];
  argv[0] = "/sbin/ifconfig";
  argv[1] = "eth0";
  argv[2] = "10.0.0.1"; 
  argv[3] = 0;
  if (fork_and_exec ("/sbin/ifconfig", argv, 0, 0, 0) != 0)
    {
      printf ("error setting IP number\n");
      return;
    }
  printf ("** done **\n");
}

/**
 * Parse ifconfig output to find IP numder
 */
JNIEXPORT jstring JNICALL
Java_org_xoe_device_device_LinuxNetwork_getIP (JNIEnv* env, jobject obj, jstring device)
{
  int std_in;
  int std_out;
  int std_err;
  char* argv[4];
  jstring ip;
  argv[0] = "/sbin/ifconfig";
  argv[1] = "eth0";
  ip = (*env)->NewStringUTF (env, "test");
  return ip;
}

#endif
