/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Power.c,v 1.4 2001/11/21 22:34:41 samc Exp $
 */
#include <jni.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/reboot.h>

#include "config.h"

#ifdef HAVE_LINUX_H3600_TS_H
#define IPAQ 1
#endif /* HAVE_LINUX_H3600_TS_H */

#ifdef IPAQ
#include <linux/h3600_ts.h>
#endif /* IPAQ */

#ifdef HAVE_POWERMGNT
/* this is the interface of our platform specific (configured) power mgnt library */
#include "jpowermgnt.h"
#endif

JNIEXPORT void JNICALL
Java_org_xoe_device_LinuxPower_setBacklight0 (JNIEnv* env, jobject device, jint intensity)
{
#ifdef IPAQ
  int fd;
  FLITE_IN bl;

  fd = open ("/dev/h3600_ts", O_RDWR);
  bl.mode = 1;
  bl.pwr = intensity > 0 ? 1 : 0;
  bl.brightness = intensity * 255 / 100;
  ioctl (fd, FLITE_ON, (void*)&bl);
  close (fd);
#endif /* IPAQ */
}

/**
 * Halt function for when the JVM is running standalone withiout the 
 * SysVInit environment.  Perhaps we could umount local filesystems
 * at this point.  However when the JVM is running standalone on an
 * embedded device we better be running a journaling FS.
 */
JNIEXPORT void JNICALL
Java_org_xoe_device_LinuxPower_halt0 (JNIEnv* env, jobject device)
{
  // under linux this will block until buffers are written.
  sync ();
  reboot (RB_POWER_OFF);
  exit (0);
}

/*********************************************************************************************
 * hint apis
 */

/*
 * we have a new application power mgnt profile - extract device values (levels)
 * and process them by whatever is available on the target platform
 * 
 * NOTE: the 'profile' format has to be the same like in com.pocketlinux.Device.java
 *       (accessing fields should go into jpowermgnt.h)
 * 
 *   xxxx.xxxx.xxxx.xxxx.xxxx.xxxx.xxxx.xxxx:xxxx.xxxx.xxxx.xxxx.xxxx.xxxx.xxxx.xxxx
 *       .    .    .    .    .    .    .    :    .    .    .    .    .    .    .^^^^ --- CPU
 *       .    .    .    .    .    .    .    :    .    .    .    .    .    .^^^^ -------- DISPLAY
 *       .    .    .    .    .    .    .    :    .    .    .    .    .^^^^ ------------- SPEAKER
 *       .    .    .    .    .    .    .    :    .    .    .    .^^^^ ------------------ reserved
 *       .    .    .    .    .    .    .    :    .    .    .^^^^ ----------------------- reserved
 *       .    .    .    .    .    .    .    :    .    .^^^^ ---------------------------- reserved
 *       .    .    .    .    .    .    .    :    .^^^^ --------------------------------- USER1_LVL
 *       .    .    .    .    .    .    .    :^^^^ -------------------------------------- USER2_LVL
 *       .    .    .    .    .    .    .^^^^ ------------------------------------------- USER3_LVL
 *       .    .    .    .    .    .^^^^ ------------------------------------------------ USER4_LVL
 *       .    .    .    .    .   ^ ------------------------------------- KEYBOARD
 *       .    .    .    .    .  ^ -------------------------------------- POINTER
 *       .    .    .    .    . ^ --------------------------------------- NETWORK
 *       .    .    .    .    .^ ---------------------------------------- reserved
 *       .    .    .    .   ^ ------------------------------------------ reserved
 *       .    .    .    .  ^-------------------------------------------- reserved
 *       .    .    .    . ^--------------------------------------------- reserved
 *       .    .    .    .^---------------------------------------------- reserved
 *       .    .    .   ^ ----------------------------------------------- USER1
 *       .    .    .  ^------------------------------------------------- USER2
 *       .    .    . ^-------------------------------------------------- USER3
 *       .    .    .^--------------------------------------------------- USER4
 *  ^^^^^.^^^^.^^^^ --------------------------------- reserved
 *                                                                       <on/off devive>
 *                                                                                       <leveled device>
 */
JNIEXPORT void JNICALL
Java_org_xoe_device_LinuxPower_activateProfile0 (JNIEnv* env, jobject device, jlong profile)
{
#if defined (HAVE_POWER_MGNT)
	//.. call configured powermgnt lib
#endif
}


/*********************************************************************************************
 * imperative APIs
 */

/*
 * halt the system (this is the end..)
 */
JNIEXPORT void JNICALL
Java_org_xoe_device_LinuxPower_halt (JNIEnv* env, jobject device)
{
#if defined (HAVE_POWER_MGNT)
	//.. call configured powermgnt lib
#endif
}

/*
 * suspend the system as soon as there is no user input for 'idle' msec (0 means right away),
 * wake up automatically after 'wakeUp' msec (0 means don't wakeup automatically)
 */
JNIEXPORT void JNICALL
Java_org_xoe_device_LinuxPower_suspend ( JNIEnv* env, jobject device, jlong idle, jlong wakeUp )
{
#if defined (HAVE_POWER_MGNT)
	//.. call configured powermgnt lib
#endif
}

/*
 * put system into standby mode as soon as there is no user input for 'idle' msec
 * (0 means right away),
 */
JNIEXPORT void JNICALL
Java_org_xoe_device_LinuxPower_standby ( JNIEnv* env, jobject device, jlong idle )
{
#if defined (HAVE_POWER_MGNT)
	//.. call configured powermgnt lib
#endif
}
