/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: util.c,v 1.1 2001/10/09 22:26:17 samc Exp $
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>

/**
 * util function for calling external programs and communicating with them.
 */
int 
fork_and_exec (char* command, char* argv[], int* std_in, int* std_out, int* std_err)
{
  int pipe_in [2];
  int pipe_out[2];
  int pipe_err[2];
  int pid;
  int rtn = -1;

  printf ("forking %s\n", command);
  if ((pipe (pipe_in) != 0)
      || (pipe (pipe_out) != 0)
      || (pipe (pipe_err) != 0))
    {
      return -1;
    }

  pid  = fork ();
  if (pid < 0) /* error */
    {
      return -1;
    }
  if (pid == 0) /* child */
    {
      close (0);
      close (1);
      close (2);
      close (pipe_in [1]);
      close (pipe_out [0]);
      close (pipe_err [0]);
      dup2 (pipe_in [0], 0);
      dup2 (pipe_out[1], 1);
      dup2 (pipe_err[1], 2);
      execvp (command, argv);
      exit (1);
    }
  wait (&rtn);
  printf ("%s returned: %d\n", command, WEXITSTATUS(rtn));
  if (std_in) *std_in  = pipe_in [1];
  if (std_out) *std_out = pipe_out [0];
  if (std_err) *std_err = pipe_err [0];
  close (pipe_in [0]);
  close (pipe_out [1]);
  close (pipe_err [1]);
  return WEXITSTATUS(rtn);
}
