/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: LinuxAudio.java,v 1.8 2001/12/14 19:53:43 ben Exp $
 */
package org.xoe.device;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileDescriptor;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;
import java.io.Reader;

import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Vector;
import javax.sound.sampled.AudioFormat;

import org.xoe.core.device.IAudio;
import org.xoe.core.util.Logger;

/**
 * Java interface to native networking functions.
 */
public final class LinuxAudio
  implements IAudio
{
  private static final String[] INTERFACES = { IAudio.class.getName () };

  FileInputStream dspIn;
  FileOutputStream dspOut;


  // ---------------------------------------------------------------------
  // IService

  public String[] getInterfaces (String func) { return INTERFACES; }
  public String[] getFeatures (String func)   { return null;       }
  public String[] getInformation ()           { return null;       }
  public String[] getFunctionalities ()       { return null;       }
  public void init (org.w3c.dom.Document configDoc)
  {
    init0 ();
  }

  // ---------------------------------------------------------------------
  // IAudio

  public int read (byte[] data)
    throws IOException
  {
    return 0;
  }

  public int write (byte[] data)
    throws IOException
  {
    return write (data, 0, data.length);
  }

  public int write (byte[] data, int offset, int length)
    throws IOException
  {
    if (this.dspOut == null)
      {
        open ();
      }
    this.dspOut.write (data, offset, length);
    return data.length;
  }

  public void open ()
    throws IOException
  {
    close ();
    this.dspOut = new FileOutputStream ("/dev/dsp");
    this.dspIn = new FileInputStream ("/dev/dsp");
  }

  public void close ()
  {
    try
      {
        if (this.dspIn  != null) 
	  { 
	    this.dspIn.close (); 
	  }
        if (this.dspOut != null) 
	  { 
	    this.dspOut.close (); 
	  }
      }
    catch (IOException e) {}
    this.dspIn = null;
    this.dspOut = null;
  }

  public void setFormat (AudioFormat format)
    throws IOException
  {
    if (this.dspOut == null)
      {
        open ();
      }
    if (!setFormat0 (this.dspOut.getFD (), format.getSampleSizeInBits (), 
                     format.getChannels (), (int) format.getSampleRate ()))
      {
        throw new IOException ("failed to set audio format");
      }
  }


  public void setVolume (int percentage)
    throws IOException
    {
      if (this.dspOut == null)
	{
	  open ();
	}      
      setVolume0 ( this.dspOut.getFD (), percentage);
    }

  public native int getVolume  ();



  private native void setVolume0 (FileDescriptor fd, int percentage);

  private native boolean setFormat0 (FileDescriptor fd, int sampleSize,
                                     int channels, int bitRate);

  private static native boolean init0 ();
}
