/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: LinuxPower.java,v 1.4 2001/11/10 00:49:14 samc Exp $
 */
package org.xoe.device;

import java.io.IOException;
import java.io.File;

import org.xoe.core.device.IPower;
import org.xoe.core.util.Logger;

/**
 * Linux implementation of IPower.
 * Currently we only support APM.  Future versions will support ACPI and
 * the linux-pm driver interface.
 */
public class LinuxPower
  implements IPower
{
  private final String[] INTERFACES = { IPower.class.getName () };

  // ---------------------------------------------------------------------
  // IService

  public String[] getInterfaces (String func) { return INTERFACES; }
  public String[] getFeatures (String func)   { return null;       }
  public String[] getFunctionalities ()       { return null;       }
  public String[] getInformation ()           { return null;       }
  public void init (org.w3c.dom.Document configDoc) {}

  // ---------------------------------------------------------------------
  // IPower

  public void setBacklight (int intensity)
  {
    setBacklight0 (intensity);
  }

  /**
   * If we are on a System then has sysvinit installed then call the
   * shutdown executable.  Otherwise we assume that the JVM is being run
   * as the init process and we need to call "sync" and "reboot" sys calls.
   */
  public void halt ()
  {
    String sysVHalt = "/sbin/ihalt";
    if (new File (sysVHalt).exists ())
      {
        try
          {
            Process proc = Runtime.getRuntime ().exec (sysVHalt);
            proc.waitFor ();
            if (proc.exitValue () != 0)
              {
                Logger.log (this, sysVHalt + " returned non-zero (" + proc.exitValue () + ")");
              }
          }
        catch (IOException e)
          {
            Logger.log (this, e);
          }
        catch (InterruptedException e)
          {
            Logger.log (this, e);
          }
      }

    // if sysVHalt fails or is not found we call our own native
    // halt function that will issue the "sync" and "reboot" sys calls.
    halt0 ();
    // we should never get here...
    Logger.log (this, "You should never see this!");
    System.exit (0);
  }

  public void standby ()
  {
  }

  public void suspend ()
  {
  }

  public long getBatteryTimeRemaining ()
  {
    return -1;
  }

  public boolean hasExternalPower ()
  {
    return true; 
  }

  private native void setBacklight0 (int intensity);
  private native void halt0 ();

  public static void main (String[] args)
    throws IOException
  {
    System.loadLibrary ("device");
    System.out.println ("warning: this function will attempt to halt your machine");
    System.out.println ("press enter to continue ...");
    int c = System.in.read ();
    new LinuxPower ().halt ();
  }
}
