/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: LinuxStorage.java,v 1.4 2001/11/09 19:37:21 samc Exp $
 */
package org.xoe.device;

import java.io.IOException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.Vector;

import org.xoe.core.device.IStorage;

/**
 * Linux implementation of IStorage.
 */
public class LinuxStorage
  implements IStorage
{
  private static final String[] INTERFACES = { IStorage.class.getName () };

  // ---------------------------------------------------------------------
  // IService

  public String[] getInterfaces (String func) { return INTERFACES; }
  public String[] getFeatures (String func)   { return null;       }
  public String[] getFunctionalities ()       { return null;       }
  public String[] getInformation ()           { return null;       }
  public void init (org.w3c.dom.Document configDoc) {}

  // ---------------------------------------------------------------------
  // IStorage

  public int getTotalRAM ()
    throws IOException
  {
    FileInputStream fis = new FileInputStream ("/proc/meminfo");
    StreamTokenizer tok = new StreamTokenizer (new InputStreamReader (fis));
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        // skip first line
        if (tok.lineno () < 2)
          {
            continue;
          }
        // skip the first two tokens (Mem:)
        tok.nextToken ();
        tok.nextToken ();
        if (tok.ttype != StreamTokenizer.TT_NUMBER)
          {
            throw new IOException ("error parsing /proc/meminfo: expected number: " + tok);
          }
        break;
      }
    return (int) (tok.nval/1024);
  }

  public int getUnusedRAM ()
  {
    return -1;
  }

  /**
   * parse "/proc/partitions" to discover the current storage devices.
   */
  public String[] getStorageDevices ()
    throws IOException
  {
    Vector vector = new Vector ();
    FileInputStream fis = new FileInputStream ("/proc/partitions");
    StreamTokenizer tok = new StreamTokenizer (new InputStreamReader (fis));
    tok.wordChars ('/', '/');
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        // skip first two lines
        if (tok.lineno () < 3)
          {
            continue;
          }
        // first three token we skip.
        tok.nextToken ();
        tok.nextToken ();
        tok.nextToken ();
        if (tok.ttype != StreamTokenizer.TT_WORD)
          {
            throw new IOException ("error parsing /proc/partitions: expected partition name: " + tok);
          }
        if (tok.sval.endsWith ("disc"))
          {
            vector.addElement (tok.sval);
          }
      }
     return (String[]) vector.toArray (new String[vector.size ()]);
  }

  private static long BLOCKSIZE = 1024;

  public long getSize (String deviceName)
    throws IOException
  {
    FileInputStream fis = new FileInputStream ("/proc/partitions");
    StreamTokenizer tok = new StreamTokenizer (new InputStreamReader (fis));
    tok.wordChars ('/', '/');
    while (tok.nextToken () != StreamTokenizer.TT_EOF)
      {
        // skip first two lines
        if (tok.lineno () < 3)
          {
            continue;
          }
        // first three token we skip.
        tok.nextToken ();
        tok.nextToken ();
        if (tok.ttype != StreamTokenizer.TT_NUMBER)
          {
            throw new IOException ("error parsing /proc/partitions: expected size: " + tok);
          }
        long size = (long) tok.nval;
        tok.nextToken ();
        if (tok.ttype != StreamTokenizer.TT_WORD)
          {
            throw new IOException ("error parsing /proc/partitions: expected partition name: " + tok);
          }
        if (tok.sval.equals (deviceName))
          {
            return size * BLOCKSIZE;
          }
      }
    throw new IOException ("failed to find storage device: " + deviceName);
  }

  /**
   * format the device.
   */
  public void format (String deviceName)
    throws IOException
  {
  }

  /**
   * get the device mount point.
   * Here is where we might want to reurn a StashDirectory.
   */
  public String getMountPoint (String deviceName)
    throws IOException
  {
    return null;
  }

  public static void main (String[] args)
    throws Exception
  {
    LinuxStorage storage = new LinuxStorage ();
    System.err.println ("RAM: " + storage.getTotalRAM ());
    String[] devices = storage.getStorageDevices ();
    System.err.println ("Storage devices:");
    for (int i = 0; i < devices.length; i++)
      {
        System.err.println (i + ": " + devices[i]);
        System.err.println ("   " + storage.getSize (devices[i]));
      }
    System.exit (0);
  }
}
