/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: Main.java,v 1.13 2001/11/17 02:29:02 ben Exp $ 
 */
package org.xoe.email;

import org.xoe.core.util.XMLWriter;
import org.xoe.core.util.DOMUtils;
import org.xoe.display.dom.xhtml.Input;
import org.xoe.display.dom.xhtml.Select;
import org.xoe.display.dom.xhtml.TextArea;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.display.dom.ViewElement;
import org.xoe.servlet.XoeServlet;
import org.xoe.soap.ISoapService;
import org.xoe.soap.SoapService;
import org.xoe.soap.SoapServiceException;

import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.StashException;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.XEntityResolver;
import org.xoe.core.dom.DefaultEntityResolver;
import org.xoe.core.dom.DOMlet;

import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;

import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;

//import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.io.IOException;
import java.lang.Thread;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

/**
 * Email Client.
 * PocketLinux application for sending and receiving email via POP.
 */
public class Main
  extends XoeServlet
{
  // commands
  public static final String SEND = "SEND";
  public static final String INBOX = "INBOX";
  public static final String SETTINGS = "SETTINGS";
  public static final String CANCEL = "CANCEL";
  public static final String SAVE = "SAVE";
  public static final String COMPOSE = "COMPOSE";


  // ids on elements (used for bindings)  
  public static final String TO = "TO";
  public static final String FROM = "FROM";
  public static final String SUBJECT = "SUBJECT";
  public static final String MESSAGE = "MESSAGE";

  public static final String USER = "USER";
  public static final String NAME = "NAME";
  public static final String SERVER = "SERVER";
  public static final String CHANNEL = "CHANNEL";  
   
  private ISoapService m_soapService;
  
  private Input m_to;
  private Input m_from;
  private Input m_subject;
  private TextArea m_message;

  private Input soapURLInput;
  private Input serverTypeInput;
  private Input serverNameInput;
  private Input portInput;
  private Input userNameInput;
  private Input passwordInput;

  private final XPathContext sendMailContext = new XPathContext ();
  private final XPathContext deleteMailContext = new XPathContext ();
  private final XPathContext getMailContext = new XPathContext ();

  private static final String DATA_DIR = "stash:/home/default/data/email/";
  //private XElement userConfig;

  public void init (ServletConfig config) throws javax.servlet.ServletException
  {    
    super.init (config);

    deleteMailContext.definePrefix ("ns1", "urn:email:DeleteMail");
    getMailContext.definePrefix ("ns1", "urn:email:GetMail");
    sendMailContext.definePrefix ("ns1", "urn:email:SendMail");
  }
  
  protected void doGet(HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    res.setContentType("text/html");
    String path = req.getPathInfo();

    if ( "".equals(path) )
      {
	XDocument doc = getRelativeDocument ("default.xhtml");
	doc.expandUsingXPathAttributes ();

	m_to = (Input)doc.getElementById(TO);  
	m_from = (Input)doc.getElementById(FROM);
	m_subject = (Input)doc.getElementById(SUBJECT);
	m_message = (TextArea)doc.getElementById(MESSAGE);      

	res.getWriter().print(doc);
      }
    else if (COMPOSE.equals(path))
      {
        XDocument doc = getRelativeDocument ("compose.xhtml");
	doc.expandUsingXPathAttributes();
	m_to = (Input)doc.getElementById(TO);
	if( m_to == null )
	  Logger.log( this, "m_to is NULL!!!" );

	m_subject = (Input)doc.getElementById(SUBJECT);
	m_message = (TextArea)doc.getElementById(MESSAGE);
	res.getWriter ().print (doc);
      }
    else if (SEND.equals(path))
      {
	Logger.log( this, "user=" + XPath.findFirstNode( getLocalUserConfig(), "//username/text()" ).getNodeValue().trim() );
	Logger.log( this, "serverName=" + XPath.findFirstNode( getLocalUserConfig(), "//servername/text()" ).getNodeValue().trim() );
	Logger.log( this, "to=" + m_to.getText() );
	Logger.log( this, "subject=" + m_subject.getText() );
	Logger.log( this, "message=" + m_message.getText() );
	
        try
          {
            send ( XPath.findFirstNode( getLocalUserConfig(), "//username/text()" ).getNodeValue().trim() + 
		   "@" + 
		   XPath.findFirstNode( getLocalUserConfig(), "//servername/text()" ).getNodeValue().trim(), 
		   m_to.getText(), 
		   m_subject.getText(), 
		   m_message.getText());
	    XDocument doc = getRelativeDocument ("default.xhtml");	
	    res.getWriter ().print (doc);
          }
        catch (MalformedURLException e)
          {
            //e.printStackTrace ();
	    Logger.log( this, e.toString() );
	    Dialog.note ("Email Failed!<br/>Reason: <b>" + e.getMessage() + "</b>" );       	    
	    return;
          }
        catch (SoapServiceException se)
          {
            //se.printStackTrace ();
	    Logger.log( this, se.toString() );
	    Dialog.note ("Email Failed!<br/>Reason: <b>" + se.getMessage() + "</b>" );       	    
	    return;
          }	
      }
    else if ("BACK".equals (path))
      {
	XDocument doc = getRelativeDocument ("default.xhtml");
	res.getWriter ().print (doc);
      }
    else if ("INBOX".equals (path))
      {
	getMail( XPath.findFirstNode( getLocalUserConfig(), "//servertype/text()" ).getNodeValue().trim(), 
		 XPath.findFirstNode( getLocalUserConfig(), "//servername/text()" ).getNodeValue().trim(), 
		 XPath.findFirstNode( getLocalUserConfig(), "//port/text()" ).getNodeValue().trim(), 
	         XPath.findFirstNode( getLocalUserConfig(), "//username/text()" ).getNodeValue().trim(), 
		 XPath.findFirstNode( getLocalUserConfig(), "//password/text()" ).getNodeValue().trim()
		 );

	XDocument doc = getRelativeDocument ("inbox.xhtml");
	res.getWriter ().print (doc);
      }
    else if ("SETTINGS".equals (path))
      {
	XDocument doc = getRelativeDocument ("settings.xhtml");
	Logger.log( this, "MY CONFIG: [" + XMLWriter.getXML( getLocalUserConfig() ) + "]" );
	DOMUtils.expandUsingXPathAttributes (doc, getLocalUserConfig(), null );	 

	doc.expandUsingXPathAttributes ();
	this.userNameInput = (Input)doc.getElementById("username"); 
	this.passwordInput = (Input)doc.getElementById("password");
	this.soapURLInput = (Input)doc.getElementById("SoapURL");
	this.serverTypeInput = (Input)doc.getElementById("servertype");
	this.serverNameInput = (Input)doc.getElementById("servername");
	this.portInput = (Input)doc.getElementById("port");

	res.getWriter ().print (doc);
      }
    else if ("SAVE".equals (path))
      {   
	Logger.log( this, "SAVING SETTINGS: username = [" + this.userNameInput.getText() + "]" );
	XDocument configDoc = (XDocument)getLocalUserConfig().getOwnerDocument();

	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//username" ) , this.userNameInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//password" ) , this.passwordInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//SoapURL" ) , this.soapURLInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//servertype" ) , this.serverTypeInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//servername" ) , this.serverNameInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//port" ) , this.portInput.getText() );

	URL userConfigURL = new URL( this.DATA_DIR + "userConfig.xml" );
	String userConfigString = XMLWriter.getXML( configDoc );

	Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );

	ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
	Stash.put( userConfigCE );

	XDocument doc = getRelativeDocument ("default.xhtml");
	res.getWriter ().print (doc);
      }


  }


  private ISoapService findSoapService ()
  {
    if (m_soapService == null)
      {
        m_soapService = (ISoapService) ServiceLocator.findImplementor (ISoapService.INTERFACE);
        if (m_soapService == null)
          {
            m_soapService = new SoapService ();
            ServiceLocator.registerService (m_soapService);
          }
      }

    return m_soapService;
  }


  public void deleteMail( String serverType, String serverAddress, String sPort, String user, String password, Integer messageNum )
  {    
    Integer port = new Integer( sPort );

    Node result;
    Stack stack = new Stack();
    stack.push( messageNum );
    stack.push( port );
    stack.push( password );
    stack.push( user );
    stack.push( serverAddress );
    stack.push( serverType );
    
    Logger.log( this, "MESSAGE #" + messageNum.intValue() + " deleted from SOAP server" );
        
    URL url = null;
    try
      {
	url = new URL( XPath.findFirstNode( getLocalUserConfig(), "//SoapURL/text()" ).getNodeValue().trim() );
	
	result = findSoapService().call("urn:email:DeleteMail",
					"delete",
					stack,
					url  );
	
	String resp = XPath.findString ( deleteMailContext, result, "//ns1:deleteResponse/return");
	
	Logger.log( this, "Raw Node Result [" + XMLWriter.getXML( result ) + "]" );
	Logger.log( this, "SOAP EMAIL result: [" + resp + "]" );
	       
	Dialog.note ("<b>" + resp + "</b>" );
      }
    catch (MalformedURLException e)
      {
        e.printStackTrace ();	
      } 
    catch (SoapServiceException se)
      {
        se.printStackTrace ();	    
      }	
  }


  public void send (String from, String to, String subject, String message)
    throws SoapServiceException, MalformedURLException
  {
    Node result;

    Stack stack = new Stack ();
    stack.push (message);
    stack.push (subject);
    stack.push (to);
    stack.push (from);
    
    URL url = null;

    System.err.println( "EMAIL Main.send (): soap url [" +
			XPath.findFirstNode( getLocalUserConfig(), "//SoapURL/text()" ).getNodeValue().trim() 
			+ "]");

    url = new URL ( XPath.findFirstNode( getLocalUserConfig(), "//SoapURL/text()" ).getNodeValue().trim() );

    if (findSoapService () == null)
      {
        Logger.log (this, "soap service is null");
      }

    result = findSoapService ().call ("urn:email:SendMail",
                                          "send",
                                          stack,
                                          url  );

    Logger.log (this, XMLWriter.getXML (result));

    String resp = XPath.findString ( sendMailContext, result, "//ns1:sendResponse/return" );

    Logger.log (this, "SOAP EMAIL result: [" + resp + "]");     

    Dialog.note ("<b>" + resp + "</b>" );
  }

  private String replaceNewLines (String oldString)
  {    
    String newline = "\n";
    String htmlLineBreak = "<br>";
    
    String cleanString = "";
    int index = oldString.indexOf (newline);
    if (index == -1) return oldString;
    while (index != -1)
      {	    
      oldString = replaceString (oldString, newline, htmlLineBreak);
      index = oldString.indexOf (newline); 
    }
    return oldString;
  }


  //replaces first instance of old with new
  public String replaceString (String text, String oldS, String newS)
  {  
    int start = text.indexOf (oldS);
    int length = oldS.length ();
    String beginning = text.substring (0, start);
    String end = text.substring (start + length);

    String replaced = beginning + newS + end;
    
    return replaced;
  }

  public Node getMail (String serverType, String serverAddress, String sPort, String user, String password, Integer messageNum)
  {    
    Integer port = new Integer (sPort);
    Node result;
    Stack stack = new Stack ();
    stack.push (messageNum);
    stack.push (port);
    stack.push (password);
    stack.push (user);
    stack.push (serverAddress);
    stack.push (serverType);
    
    Logger.log (this, "MESSAGE #" + messageNum.intValue () + " retrieved from SOAP server");
        
    URL url = null;
    try
      {
	url = new URL( XPath.findFirstNode (getLocalUserConfig (), "//SoapURL/text()" ).getNodeValue ().trim ());
	
	result = findSoapService ().call ("urn:email:GetMail",
					  "get",
					  stack,
					  url);
	
	Logger.log (this, "Raw Node Result [" + XMLWriter.getXML( result ) + "]");
		

	String resp = XPath.findFirstNode (getMailContext, result, "//ns1:getResponse/return").getFirstChild ().getNodeValue ();
	
	String entifiedResp = XEntityResolver.entify (resp, DefaultEntityResolver.getInstance());
	Logger.log (this, "Entified SOAP EMAIL result: [" + entifiedResp + "]");

	XDocument doc = DocumentCache.fetchDocument (this.DATA_DIR + "selectedMessgage.xml");		
	Node dataNode = XPath.findFirstNode (doc, "selectedMessage");
	//Node dataNode = XMLRoot.getInstance().findFirstNode( "data/soapMail/SelectedMessage" );	
	DOMUtils.removeChildNodes (dataNode);
	//Logger.log( this, "data node created" );
	// the 2nd param is printed out by LazyDOM on a parse error, pass in something appropriate
	DOMlet domlet = DOMlet.create (entifiedResp.toCharArray(), "Response");  
	//Logger.log( this, "new domlet created" );

	dataNode.appendChild (domlet);
	//Logger.log( this, "domlet appended to data node" );
		
	Logger.log (this, XMLWriter.getXML (dataNode));	    
	NodeList nodeList = XPath.findNodes (dataNode, "//To");
	Logger.log (this, "node list: " + nodeList );	    
	Node fromNode = XPath.findFirstNode (dataNode, "//From");
	Logger.log (this, "from node: " + XMLWriter.getXML (fromNode));
	Node subNode = XPath.findFirstNode (dataNode, "//Subject");
	Logger.log (this, "sub node: " + XMLWriter.getXML (subNode));
	Node dateNode = XPath.findFirstNode (dataNode, "//Date");
	Logger.log (this, "date node: " + XMLWriter.getXML (dateNode));
	Node bodyNode = XPath.findFirstNode (dataNode, "//Body");
	Logger.log (this, "body node: " + XMLWriter.getXML (bodyNode));	

	return dataNode;

      }
    catch (MalformedURLException e)
      {
        e.printStackTrace ();
        return null;
      }
    catch (SoapServiceException se)
      {
        se.printStackTrace ();
        return null;
      }	
  }

  public void getMail (String serverType, String serverAddress, String sPort, String user, String password)
  {
    Integer port = new Integer (sPort);
    
    Node result;

    Stack stack = new Stack ();
    stack.push (port);
    stack.push (password);
    stack.push (user);
    stack.push (serverAddress);
    stack.push (serverType);
    
    URL url = null;
    try
      {
	url = new URL (XPath.findFirstNode (getLocalUserConfig (), "//SoapURL/text()").getNodeValue().trim());
	
	result = findSoapService ().call ("urn:email:GetMail",
					  "get",
					  stack,
					  url);
	
	String resp = XPath.findString (getMailContext, result, "//ns1:getResponse/return");
	//Logger.log (this, "SOAP EMAIL result: [" + resp + "]");
	
	String entifiedResp = XEntityResolver.entify(resp, DefaultEntityResolver.getInstance());
	//Logger.log (this, "Entified SOAP EMAIL result: [" + entifiedResp + "]");

	//ContentElement ce = Stash.fetch( DATA_DIR );
	DOMlet incomingBox = DOMlet.create (entifiedResp.toCharArray (), "inbox");

	ContentElement myBoxCE = getInbox ();	
	DOMlet existingBox = DOMlet.create (myBoxCE.getInputStream (), myBoxCE.getURL ());
	Node newMessages = existingBox.importNode( incomingBox.getDocumentElement (),true);
	existingBox.getDocumentElement ().appendChild (newMessages);

	Logger.log (this, "inbox: [" + XMLWriter.getXML (existingBox.getDocumentElement ()) + "]");

	//saving the inbox to stash
	URL inboxURL = new URL (this.DATA_DIR + "inbox.xml");
	String myboxString = XMLWriter.getXML (existingBox);
	Logger.log (this, "saving inbox.xml = [" + myboxString + "]");
	ContentElement inboxCE = Stash.getFactory().createContentElement( inboxURL, myboxString.getBytes() );
	Stash.put (inboxCE);

      }
    catch (StashException se)
      {
        se.printStackTrace();
      }	
    catch (MalformedURLException e)
      {
        e.printStackTrace();
      }
    catch (SoapServiceException se)
      {
        se.printStackTrace();
      }
    catch (IOException ioe)
      {
        ioe.printStackTrace();
      }	

  }

  /**
   * This method will always attempt to retrieve user data from the stash
   * This data WILL NOT be cached, so any sync-ing of the stash will become
   * immediately apparent.  If for some reason the stash does not contain
   * user data, this method will look to the relative file config.xml for data
   *
   */
  private XElement getLocalUserConfig ()
  {
    XElement userConfig = null;	
    XDocument doc = DocumentCache.fetchDocument (this.DATA_DIR + "userConfig.xml");		
    if (doc != null)
      userConfig = (XElement)XPath.findFirstNode (doc, "config");
    
    if (userConfig == null)
      {
	Logger.log (this, "Main.getLocalUserConfig(): fetch from stash is null.  Retrieving from package/config.xml..." );
	XElement originalConfig = (XElement)DocumentCache.fetchDocument (getRelativeSrc("config.xml")).getDocumentElement ();	
	
	try
	  {
	    //now copy this info to the stash....?	    
	    URL userConfigURL = new URL (this.DATA_DIR + "userConfig.xml");
	    String userConfigString = XMLWriter.getXML (originalConfig);
	    //Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );	    
	    ContentElement userConfigCE = Stash.getFactory ().createContentElement (userConfigURL, userConfigString.getBytes());
	    Stash.put (userConfigCE);
	    return originalConfig;
	  }
	catch (IOException ioe)
	  {
	    Logger.log (this, ioe.toString());
	    return null;
	  }      
      }
    return userConfig;
  }



  /**
   * This method will always attempt to retrieve inbox from the stash
   * This data WILL NOT be cached, so any sync-ing of the stash will become
   * immediately apparent.  If for some reason the stash does not contain
   * an inbox, this method will look to the relative file inbox-template.xml for creating an inbox in the stash
   *
   */
  private ContentElement getInbox ()
    throws StashException
  {
    XElement inboxEl = null;	
    //XDocument doc = DocumentCache.fetchDocument (this.DATA_DIR + "inbox.xml");		
    ContentElement inboxCE = Stash.fetch (this.DATA_DIR + "inbox.xml");
    /*
    if (doc != null)
      {
	inboxEl = (XElement)XPath.findFirstNode (doc, "inbox");
      }
    */
    if (inboxCE == null)
      {
	Logger.log( this, "Main.getLocalUserConfig(): fetch from stash is null.  Retrieving from package/inbox-template.xml..." );
	XElement inboxTemplateEl = (XElement)DocumentCache.fetchDocument (getRelativeSrc("inbox-template.xml")).getDocumentElement();	
	
	try
	  {
	    //now copy this info to the stash....?	    
	    URL inboxURL = new URL (this.DATA_DIR + "inbox.xml");
	    String inboxTemplateString = XMLWriter.getXML( inboxTemplateEl );
	    //Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );	    
	    inboxCE = Stash.getFactory ().createContentElement (inboxURL, inboxTemplateString.getBytes());
	    Stash.put (inboxCE);
	    return inboxCE;
	  }
	catch (IOException ioe)
	  {
	    Logger.log (this, ioe.toString());
	    return null;
	  }      
      }
    return inboxCE;
  }




} 
