package org.xoe.history;

import java.net.URL;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XNode;
import org.xoe.core.dom.Callback;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceFactory;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.services.IHistory;


public class Main
  implements IHistory, ServiceFactory
{

  static Main instance;

  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  public String[] getFunctionalities ()
    {
      return (new String[] { IHistory.FUNCTIONALITY });
    }

  public String[] getInterfaces (String func)
  {
    if (IHistory.FUNCTIONALITY.equals (func))
	{
	  return (new String[] { IHistory.INTERFACE });
	}

    return (null);
  }

  public String[] getFeatures (String func)
  {
    return (null);
  }
  
  public String[] getInformation ()
  {
    return (null);    
  }

  public void trackHistory (final DocumentViewer dv, final boolean propagate)
  {
    // The history of documents. 
    final Vector pageHistory = new Vector();
    final Vector urlHistory = new Vector();

    // The index of the current document in the history. 
    final IntWrapper index = new IntWrapper (-1);

    if (! (dv instanceof XNode))
      {
	Logger.log (this, "Cannot manage history of a " + dv.getClass ().getName () + ". It is not a XNode.");
	return;
      }
    final XNode node = (XNode) dv;
    node.setCallback ("back", new Callback () {
	public void invoke (Node n, Object payload)
	{
	  if (pageHistory != null && index.get () > 0)
	    {
	      int i = index.dec ();
	      dv.setDocument ((Document) pageHistory.elementAt (i), (URL) urlHistory.elementAt (i));
	      Logger.log (this, "back doc:" + dv.getDocument () + " index = " + index.get ());
	    }
	  else 
	    {
	      if (propagate)
		{
		  ((XDocument) node.getOwnerDocument ()).getDocumentHolder ().getCallback ("back").invoke (n,payload);
		}
	      else
		{
		  Logger.log (this, "Can't go back. history = " + pageHistory + " index = " + index.get ());
		}
	    }
	}
      });
    node.setCallback ("forward", new Callback () {
	public void invoke (Node n, Object payload)
	{
	  if (payload == null && n == null)
	    {
	      if (pageHistory != null && index.get () + 1 < pageHistory.size () )
		{
		  int i = index.inc ();
		  dv.setDocument ((Document) pageHistory.elementAt(i), (URL) urlHistory.elementAt (i));
		  Logger.log (this, "forward doc:" + dv.getDocument () + " index = " + index.get ());
		}
	      else 
		{
		  Logger.log (this, "Can't go forward. history = " + pageHistory + " index = " + index.get ());
		}
	    }
	  else 
	    {
	      URL u = (URL) payload;
	      Document d = (Document) n;
	      dv.setDocument (d, u);

	      if (index.get ()+1 < pageHistory.size ())
		{
		  index.inc ();
		  pageHistory.setElementAt (d, index.get ());
		  urlHistory.setElementAt  (u, index.get ());
		  pageHistory.setSize (index.get ()+1);
		  urlHistory.setSize  (index.get ()+1);
		  Logger.log (this, "forward (middle) doc:" + payload + " index = " + index.get ());
		}
	      else
		{
		  index.inc ();
		  pageHistory.addElement (d);
		  urlHistory.addElement  (u);
		  Logger.log (this, "forward (top) doc:" + payload + " index = " + index.get ());
		}
	    }
	}
      });
  }

  
  public void forward (DocumentViewer dv, Document doc, URL url)
  {
    ((XNode) dv).getCallback ("forward").invoke (doc, url);
  }
  
  public void back (DocumentViewer dv)
  {
    ((XNode) dv).getCallback ("back").invoke (null, null);
  }
  
  public void forward (DocumentViewer dv)
  {
    ((XNode) dv).getCallback ("forward").invoke (null, null);
  }
  
  public void init (Document d)
  {
  }

  private class IntWrapper 
  {
    int i;
    public IntWrapper (int i)
    {
      this.i = i;
    }
    public int get () 
    {
      return i;
      }
    public int set (int i)
    {
      this.i = i;
      return i;
    }
    public int inc ()
    {
      i++; return i;
    }
    public int dec ()
    {
      i--; return i;
    }
  };

  public IService newInstance ()
  {
    return new Main ();
  }
    
  
}
