#include "platform.h"
#include "html.h"
#include <jni.h>

uint spaces =  2;           /* default indentation */
uint wraplen = 68;          /* default wrap margin */
int CharEncoding = UTF8;
char *alt_text = null;      /* default text for alt attribute */
char *doctype_str = null;   /* user specified doctype */
int tabsize = 4;

DocTypeMode doctype_mode = doctype_auto; /* see doctype property */

Bool OnlyErrors = no;       /* if true normal output is suppressed */
Bool ShowWarnings = no;    /* however errors are always shown */
Bool Quiet = yes;            /* no 'Parsing X', guessed DTD or summary */
Bool IndentContent = yes;    /* indent content of appropriate tags */
Bool SmartIndent = no;      /* does text/block level content effect indentation */
Bool HideEndTags = no;      /* suppress optional end tags */
Bool XmlTags = no;          /* treat input as XML */
Bool XmlOut = yes;           /* create output as XML */
Bool xHTML = yes;            /* output extensible HTML */
Bool XmlPi = yes;            /* add <?xml?> for XML docs */
Bool RawOut = no;           /* avoid mapping values > 127 to entities */
Bool UpperCaseTags = no;    /* output tags in upper not lower case */
Bool UpperCaseAttrs = no;   /* output attributes in upper not lower case */
Bool MakeClean = no;        /* replace presentational clutter by style rules */
Bool LogicalEmphasis = no;  /* replace i by em and b by strong */
Bool DropFontTags = no;     /* discard presentation tags */
Bool DropEmptyParas = yes;  /* discard empty p elements */
Bool FixComments = yes;     /* fix comments with adjacent hyphens */
Bool BreakBeforeBR = no;    /* o/p newline before <br> or not? */
Bool BurstSlides = no;      /* create slides on each h2 element */
Bool NumEntities = no;      /* use numeric entities */
Bool QuoteMarks = yes;       /* output " marks as &quot; */
Bool QuoteNbsp = yes;       /* output non-breaking space as entity */
Bool QuoteAmpersand = yes;  /* output naked ampersand as &amp; */
Bool WrapAttVals = no;      /* wrap within attribute values */
Bool WrapScriptlets = no;   /* wrap within JavaScript string literals */
Bool WrapSection = yes;     /* wrap within <![ ... ]> section tags */
Bool WrapAsp = yes;         /* wrap within ASP pseudo elements */
Bool WrapJste = yes;        /* wrap within JSTE pseudo elements */
Bool WrapPhp = yes;         /* wrap within PHP pseudo elements */
Bool FixBackslash = yes;    /* fix URLs by replacing \ with / */
Bool IndentAttributes = no; /* newline+indent before each attribute */
Bool XmlPIs = no;           /* if set to yes PIs must end with ?> */
Bool XmlSpace = no;         /* if set to yes adds xml:space attr as needed */
Bool EncloseBodyText = no;  /* if yes text at body is wrapped in <p>'s */
Bool EncloseBlockText = no; /* if yes text in blocks is wrapped in <p>'s */
Bool KeepFileTimes = yes;   /* if yes last modied time is preserved */
Bool Word2000 = no;         /* draconian cleaning for Word2000 */
Bool TidyMark = yes;        /* add meta element indicating tidied doc */
Bool Emacs = no;            /* if true format error output for GNU Emacs */
Bool LiteralAttribs = no;   /* if true attributes may use newlines */

static JNIEnv* genv;
static jclass inputstream;
static jclass outputstream;
static jmethodID readid;
static jmethodID writeid;

#define	GETC(F)		(*genv)->CallIntMethod(genv, (F), readid)
#define	PUTC(C,F)	(*genv)->CallVoidMethod(genv, (F), writeid, (jint)(C))

void*
MemRealloc(void* mem, uint size)
{
	return (realloc(mem, size));
}

void*
MemAlloc(uint size)
{
	return (malloc(size));
}

void
MemFree(void* mem)
{
	free(mem);
}

void
Init()
{
	InitMap();
	InitAttrs();
	InitTags();
	InitEntities();
}

/* like strdup but using MemAlloc */
char *wstrdup(char *str)
{
    char *s, *p;
    int len;

    if (str == null)
        return null;

    for (len = 0; str[len] != '\0'; ++len);

    s = (char *)MemAlloc(sizeof(char)*(1+len));
    for (p = s; (*p++ = *str++););
    return s;
}

/* like strndup but using MemAlloc */
char *wstrndup(char *str, int len)
{
    char *s, *p;

    if (str == null || len < 0)
        return null;

    s = (char *)MemAlloc(sizeof(char)*(1+len));

    p = s;

    while (len-- > 0 && (*p++ = *str++));

    *p = '\0';
    return s;
}

/* exactly same as strncpy */
void wstrncpy(char *s1, char *s2, int size)
{
    if (s1 != null && s2 != null)
    {
        if (size >= 0)
        {
            while (size--)
                *s1++ = *s2++;
        }
        else
            while ((*s1++ = *s2++));
    }
}

void wstrcpy(char *s1, char *s2)
{
    while ((*s1++ = *s2++));
}

void wstrcat(char *s1, char *s2)
{
    while (*s1)
        ++s1;

    while ((*s1++ = *s2++));
}

/* exactly same as strcmp */
int wstrcmp(char *s1, char *s2)    
{
    int c;

    while ((c = *s1) == *s2)
    {
        if (c == '\0')
            return 0;

        ++s1;
        ++s2;
    }

    return (*s1 > *s2 ? 1 : -1);
}

/* returns byte count, not char count */
int wstrlen(char *str)
{
    int len = 0;

    while(*str++)
        ++len;

    return len;
}

/*
 MS C 4.2 doesn't include strcasecmp.
 Note that tolower and toupper won't
 work on chars > 127
*/
int wstrcasecmp(char *s1, char *s2)    
{
    uint c;

    while (c = (uint)(*s1), ToLower(c) == ToLower((uint)(*s2)))
    {
        if (c == '\0')
            return 0;

        ++s1;
        ++s2;
    }

    return (*s1 > *s2 ? 1 : -1);
}

int wstrncmp(char *s1, char *s2, int n)    
{
    int c;

    while ((c = *s1) == *s2)
    {
        if (c == '\0')
            return 0;

        if (n == 0)
            return 0;

        ++s1;
        ++s2;
        --n;
    }

    if (n == 0)
        return 0;

    return (*s1 > *s2 ? 1 : -1);
}

int wstrncasecmp(char *s1, char *s2, int n)    
{
    int c;

    while (c = *s1, tolower(c) == tolower(*s2))
    {
        if (c == '\0')
            return 0;

        if (n == 0)
            return 0;

        ++s1;
        ++s2;
        --n;
    }

    if (n == 0)
        return 0;

    return (*s1 > *s2 ? 1 : -1);
}

Bool wsubstr(char *s1, char *s2)
{
    int i, len1 = wstrlen(s1), len2 = wstrlen(s2);

    for (i = 0; i <= len1 - len2; ++i)
    {
        if (wstrncasecmp(s1+i, s2, len2) == 0)
            return yes;
    }

    return no;
}

/* read char from stream */
static int ReadCharFromStream(StreamIn *in)
{
    uint n, c, i, count;

    c = GETC(in->file);

    if ((int)c == -1)
	return -1;

    /* deal with UTF-8 encoded char */

    if ((c & 0xE0) == 0xC0)  /* 110X XXXX  two bytes */
    {
        n = c & 31;
        count = 1;
    }
    else if ((c & 0xF0) == 0xE0)  /* 1110 XXXX  three bytes */
    {
        n = c & 15;
        count = 2;
    }
    else if ((c & 0xF8) == 0xF0)  /* 1111 0XXX  four bytes */
    {
        n = c & 7;
        count = 3;
    }
    else if ((c & 0xFC) == 0xF8)  /* 1111 10XX  five bytes */
    {
        n = c & 3;
        count = 4;
    }
    else if ((c & 0xFE) == 0xFC)       /* 1111 110X  six bytes */
    {
        n = c & 1;
        count = 5;
    }
    else  /* 0XXX XXXX one byte */
        return c;

    /* successor bytes should have the form 10XX XXXX */
    for (i = 1; i <= count; ++i)
    {
        c = GETC(in->file);

	if ((int)c == -1)
	    return -1;

        n = (n << 6) | (c & 0x3F);
    }

    return n;
}

int ReadChar(StreamIn *in)
{
    int c;

    if (in->pushed)
    {
        in->pushed = no;
        c =  in->c;

        if (c == '\n')
        {
            in->curcol = 1;
            in->curline++;
            return c;
        }

        in->curcol++;
        return c;
    }

    in->lastcol = in->curcol;

    if (in->tabs > 0)
    {
        in->curcol++;
        in->tabs--;
        return ' ';
    }
    
    for (;;)
    {
        c = ReadCharFromStream(in);

        if (c < 0)
            return EndOfStream;

        if (c == '\n')
        {
            in->curcol = 1;
            in->curline++;
            break;
        }

        if (c == '\t')
        {
            in->tabs = tabsize - ((in->curcol - 1) % tabsize) - 1;
            in->curcol++;
            c = ' ';
            break;
        }

        /* strip control characters, except for Esc */

        if (c == '\033')
            break;

        if (0 < c && c < 32)
            continue;

        in->curcol++;
        break;
    }

    return c;
}

int EndOfFile(StreamIn* in)
{
	int c;
	c = ReadChar(in);
	if (c == -1) {
		return (1);
	}
	else {
		UngetChar(c, in);
		return (0);
	}
}

void UngetChar(int c, StreamIn *in)
{
    in->pushed = yes;
    in->c = c;

    if (c == '\n')
        --(in->curline);

    in->curcol = in->lastcol;
}

/* For mac users, should we map Unicode back to MacRoman? */
void outc(uint c, Out *out)
{
	uint ch;

        if (c < 128)
            PUTC(c, out->fp);
        else if (c <= 0x7FF)
        {
            ch = (0xC0 | (c >> 6)); PUTC(ch, out->fp);
            ch = (0x80 | (c & 0x3F)); PUTC(ch, out->fp);
        }
        else if (c <= 0xFFFF)
        {
            ch = (0xE0 | (c >> 12)); PUTC(ch, out->fp);
            ch = (0x80 | ((c >> 6) & 0x3F)); PUTC(ch, out->fp);
            ch = (0x80 | (c & 0x3F)); PUTC(ch, out->fp);
        }
        else if (c <= 0x1FFFFF)
        {
            ch = (0xF0 | (c >> 18)); PUTC(ch, out->fp);
            ch = (0x80 | ((c >> 12) & 0x3F)); PUTC(ch, out->fp);
            ch = (0x80 | ((c >> 6) & 0x3F)); PUTC(ch, out->fp);
            ch = (0x80 | (c & 0x3F)); PUTC(ch, out->fp);
        }
        else
        {
            ch = (0xF8 | (c >> 24)); PUTC(ch, out->fp);
            ch = (0x80 | ((c >> 18) & 0x3F)); PUTC(ch, out->fp);
            ch = (0x80 | ((c >> 12) & 0x3F)); PUTC(ch, out->fp);
            ch = (0x80 | ((c >> 6) & 0x3F)); PUTC(ch, out->fp);
            ch = (0x80 | (c & 0x3F)); PUTC(ch, out->fp);
        }
}

void
ReportError(Lexer *lexer, Node *element, Node *node, uint code)
{
}

void
ReportWarning(Lexer *lexer, Node *element, Node *node, uint code)
{
}

void
ReportAttrError(Lexer *lexer, Node *element, char *attr, uint code)
{
}

void
ReportEntityError(Lexer *lexer, uint code, char* entity, int c)
{
}

/* ----------------------------------------------------------------------- */

Java_org_xoe_content_text_html_init0(JNIEnv* env, jclass tidy)
{
	Init();

	genv = env;

	inputstream = (*env)->FindClass(env, "java.io.InputStream");
	outputstream = (*env)->FindClass(env, "java.io.OutputStream");

	readid = (*env)->GetMethodID(env, inputstream, "read", "()I");
	writeid = (*env)->GetMethodID(env, outputstream, "write", "(I)V");
}

Java_org_xoe_content_text_html_parse0(JNIEnv* env, jclass tidy, jobject in, jobject out)
{
	StreamIn instream;
	Lexer* lexer;
	Node* document;
	Node* doctype;
	Out outstream;

	instream.state = FSM_ASCII;
	instream.pushed = no;
	instream.c = 0;
	instream.tabs = 0;
	instream.lastcol = 0;
	instream.curcol = 1;
	instream.curline = 1;
	instream.encoding = CharEncoding;
	instream.file = in;

	outstream.state = FSM_ASCII;
	outstream.encoding = CharEncoding;
	outstream.fp = out;

	lexer = NewLexer(&instream);
	lexer->in->lexer = lexer;
	document = ParseDocument(lexer);
	NestedEmphasis(document);
	List2BQ(document);
	BQ2Div(document);
	doctype = FindDocType(document);
	if (document->content) {
		SetXHTMLDocType(lexer, document);
	}
	FixXMLPI(lexer, document);
	PPrintXMLTree(&outstream, null, 0, lexer, document);
	PFlushLine(&outstream, 0);
	FreeNode(document);
	FreeLexer(lexer);
}
