/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: DOM0Adaptor.java,v 1.17 2001/10/26 23:29:01 philippe Exp $ 
 */
package org.xoe.parser.html;

import java.util.Stack;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.util.Logger;
import org.xoe.display.dom.xhtml.XHTMLDocument;

public final class DOM0Adaptor
  extends XHTMLDocument
  implements DocumentFactory
{
  private Stack pStack;
  private int curLevel;
  private Node curNode;
  private XHTMLDocument root; //try changing this to Node ??
 
  private Node[] mightNeedFixing;
  private int n_mightNeedFixing;

  private static String[] problematicNodes = { "form", "li", "iframe", "option", "td"}; 

  public DOM0Adaptor()
  {
    pStack = new Stack();
    mightNeedFixing = new Node[50];
    int n_mightNeedFixing = 0;
  }
  
  private void addToMightNeedFixing(Node node)
  {
    if (n_mightNeedFixing == mightNeedFixing.length) //cache this length???
      {
	Node[] temp = new Node[n_mightNeedFixing + 50];
	System.arraycopy(mightNeedFixing, 0, temp, 0, n_mightNeedFixing);  
	mightNeedFixing = temp;
      }
    mightNeedFixing[n_mightNeedFixing++] = node;
  }
 
  public void setXsRoot (XHTMLDocument root)
  {
	  //System.out.println("In setXsRoot");
    try{
    
    this.root = root;
    pStack.removeAllElements ();
    pStack.push (root);
    curLevel = 0;
    
    }
    catch(NullPointerException e){ 

       System.out.println("NullPointerException in setXsRoot");
       if(root == null){
                    System.out.println("root is null");
            }

    }
    
  }

  
  public void createElement (int level, String tag)
  {
    String lowerCaseTag = null;
    Node newElement = null;
    
    try{
    
    lowerCaseTag = tag.toLowerCase ();
    
    newElement = root.createElement (lowerCaseTag);

    int len = problematicNodes.length;
    for (int i =0; i < len; i++)
      {
	if (lowerCaseTag.equals (problematicNodes[i]))
	  {
            addToMightNeedFixing(newElement);  
            break; 
          }
      } 

    insertNode(level, newElement);


    }
    catch(NullPointerException e){ 
	    if(newElement == null){
		    System.out.println("newElement is null, lowerCaseTag is " + lowerCaseTag);
            }

            if(root == null){
                    System.out.println("root is null");
            }
            
            if(tag == null){
                    System.out.println("tag is null");
            }

            if(lowerCaseTag == null){
                    System.out.println("lowerCaseTag is null");
            }

            e.printStackTrace();
    }

  }

  
  public void createTextElement (int level, String text)
  {
    try{	  // System.out.println("text is " + text);


    insertNode(level, root.createTextNode (text));

    }
    catch(NullPointerException e){ 
       if(root == null){
                    System.out.println("root is null");
            }

    }
  }

  private void doFormCheck(Node formNode)
  {
    Node parentNode = formNode.getParentNode ();
    String parentNodeName = parentNode.getNodeName ();
          
    if (parentNodeName.equals("tr"))
      {
	Node tableNode = parentNode.getParentNode ();
	  
        if (!tableNode.getNodeName ().equals ("table"))
	  {
		  //Logger.log(this, "ERROR: parent of tr is not table, it is " + tableNode.getNodeName());
	    return;
          }                
                  
        //make form child of table
        parentNode.removeChild (formNode);
        tableNode.insertBefore (formNode, parentNode);
 
        //remove tr from under table
        tableNode.removeChild(parentNode);
                   
        //move all children
        NodeList nodelist = formNode.getChildNodes();
 
        int length = nodelist.getLength();
	for (int i = 0; i < length; i++)
	  {
	    Node childNode = nodelist.item(0); //oddity 
            formNode.removeChild(childNode); 
            parentNode.appendChild(childNode);
          }                    
 
        //place tr under form
	formNode.appendChild(parentNode);
      }
  } 

  private void doIFrameCheck(Node iframeNode)
  {
    Node parentNode = iframeNode.getParentNode ();
    String parentNodeName = parentNode.getNodeName ();
          
    if (parentNodeName.equals("font"))
      {
	Node tdNode = parentNode.getParentNode ();
	  
        if (!tdNode.getNodeName ().equals ("td"))
	  {
	    Logger.log(this, "ERROR: parent of font is not td, it is " + tdNode.getNodeName());
	    return;
          }                
                  
        NodeList nodelist = parentNode.getChildNodes();
 
        int length = nodelist.getLength();

	Node childNode;
        Node prevChildNode;

        parentNode.removeChild(childNode = nodelist.item(--length));

        tdNode.insertBefore(childNode, parentNode);        
        prevChildNode = childNode;
       
        for (int i = 0; length > 0; length--)
	  {
	    childNode = nodelist.item(length - 1); 
            parentNode.removeChild(childNode); 
            tdNode.insertBefore(childNode, prevChildNode);
            prevChildNode = childNode;
          }                    
 

        tdNode.removeChild(parentNode);       

      }
  }

  private void doLiCheck(Node liNode)
  {
    Node parentNode = liNode.getParentNode();
    String parentNodeName = parentNode.getNodeName();
          

    if(!parentNodeName.equals("ul") && !parentNodeName.equals("ol") && 
       !parentNodeName.equals("dir") /* not sure about this one */)
      {
        Node ulNode = root.createElement ("ul");
                  
        parentNode.replaceChild(ulNode, liNode);
                  
        ulNode.appendChild(liNode);
      }
  }


  private void doOptionCheck(Node optionNode)
  { 
    //assume has only one child and this child's child is the right one (and it is alone)
    Node childNode = optionNode.getFirstChild();

    if (childNode != null)
      {
        String childNodeName = childNode.getNodeName();
          
        //System.out.println("name is " + childNodeName);

        if (!childNodeName.equals("#text"))
          {
            Node textNode = childNode.getFirstChild();
            //System.out.println("name is " + textNode.getNodeName());
            childNode.removeChild(textNode);

            optionNode.replaceChild(textNode, childNode);
          }

      }
  }


  private void doTdCheck(Node tdNode)
  {
    Node parentNode = tdNode.getParentNode();
    String parentNodeName = parentNode.getNodeName();
          

    if(!parentNodeName.equals("tr"))
      {
        Node trNode = root.createElement ("tr");
                  
        parentNode.replaceChild(trNode, tdNode);
                  
        trNode.appendChild(tdNode);
      }
  }



  public void finish ()
  {
  
    pStack.removeAllElements ();
    curLevel = 0;

    // potential fixups of potential problems
    //try{

    
    for(int i =0; i < n_mightNeedFixing; i++)
     {
       //System.out.println("In finish: " + mightNeedFixing[i].getNodeName());
       Node node = mightNeedFixing[i];
       String nodeName = node.getNodeName();
       
       if(nodeName.equals("form"))
        {
         doFormCheck(node);
        } 
       else if(nodeName.equals("li"))
        {
	 doLiCheck(node);
        } 
       else if(nodeName.equals("iframe"))
        {
	 doIFrameCheck(node);
        }
       else if(nodeName.equals("option"))
        {
	 doOptionCheck(node);
        }
       else if(nodeName.equals("td"))
        {
	 doTdCheck(node);
        }         
    
     }
    



	  //}
	  //catch(Exception e){ System.out.println(e + " " + e.getMessage()); 
	  //                    e.printStackTrace();
          //}
          


          //cases from slashdot.org     

          //base before html
	  /* can't get images with this in 
             keep for when base is implemented
          BaseNode rootnode = (BaseNode) root;
	
          BaseNode htmlnode = rootnode.findFirstNode("html");

          BaseNode basenode;

          if((basenode = (BaseNode) rootnode.getFirstChild()) != htmlnode){
		  rootnode.removeChild(basenode);

		  BaseNode headnode = htmlnode.findFirstNode("head"); 

		
		  if(headnode == null){   // had a problem with tvb, stopped trying before merge
			  headnode = (BaseNode) root.createElement( "head");
			  BaseNode bodynode = htmlnode.findFirstNode("body");
			  htmlnode.insertBefore(headnode, bodynode);
		  }
                
		  headnode.appendChild(basenode);
           }
  	*/
	  
   


    /* ================================================================= */

    // Process the onLoad attributes (even though we're not necessarily
    // being put on the screen yet)
    
    /*

    if ( onLoads != null ){
	int len = onLoads.size();
	for (int i = 0; i < len; i++){
	    ((XsObject)onLoads.elementAt(i)).invoke(null);
	}
	onLoads.removeAllElements();
    }
	  
    */

  }

  

  public int getCurrentLevel()
  {
    return curLevel;
  }

  private void insertNode(int level, Node elem)
  {
    // Adjust the level
    if (level < curLevel)
      {
	for ( ; curLevel > level; curLevel-- )
	  {
	    pStack.pop ();
	  }
       }
    else if ( level > curLevel )
      {
	pStack.push (curNode);
      }

    curLevel = level;

    // Set current node to this one

    curNode = elem;

    // Append to appropriate parent
    ((Node)pStack.peek ()).appendChild (curNode);
  

    /*
                 
          // Handle special DOM0 object arrays
          if (curNode instanceof DocEmbed){
          curXs = new XHTMLXsObject(elem);
          root.putNamedArrayObject (curXs, "embeds", null);
          }
          else if (curNode instanceof DocImage){
          curXs = new XHTMLXsObject(elem);
          root.putNamedArrayObject (curXs, "images", null);
          }
          else if (curNode instanceof DocForm){
          curXs = new XHTMLXsObject(elem);
          root.putNamedArrayObject (curXs, "forms", null);
          }       
          else if (curNode instanceof DocFrame){
          curXs = new XHTMLXsObject(elem);
          }
        
     */


  }

  
  public void setAttribute (String key, String val)
  { 
    //put a null pointer check????, getting it sometimes
	  //put a try loop in the createDocumentViaBc
	 
          

	  try{

               ((Element)curNode).setAttribute (key, val);

          }
          catch(NullPointerException e){

          if(curNode == null){
		  System.out.println("CurNode is null");
                 
          }

          


          }
          

    /*
          
          if ("name".equalsIgnoreCase (key)){
          root.xsPut (val, curXs);
          }         
          else if ("id".equalsIgnoreCase (key)){
          root.xsGet ("all").xsPut (val, curXs);
          }      
          
          */


  }

  
	/*
  public void setEventAttribute (String key, XsObject action)
  {
    throw new Error("not supported");

    
    /*
          //this is currently meaningless in the sense that action does not override toString
          ((Element)curNode).setAttribute (key, action.toString ());
          //was, in pocketlinux_tvb
          //((BaseElement)curElem).setAttributeNS( null, key, "" );
           
          curXs.xsPut (key, action);
            
          if ("onLoad".equalsIgnoreCase (key)){
          if (onLoads == null){
          onLoads = new Vector ();
          }
          onLoads.addElement (action);
          }
	*//*
          
  }
  
  public void setXsRoot (XsObject root)
  {
    throw new Error("not supported");
  }
*/
  

  


}








