package org.xoe.protocol.http;

import org.xoe.core.util.Logger;

import java.net.URL;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.Vector;

public class Cookie
  extends javax.servlet.http.Cookie
{
  private URL m_origin;
  private Date m_expires;

  public static Cookie createOneCookie(URL url, String str) throws InvalidCookieException
  {
    StringTokenizer tok = new StringTokenizer(str, ";");
    if(!tok.hasMoreTokens())
      {
	throw new InvalidCookieException(str);
      }

    // a cookie is a series of name=value pairs
    // the first one is the actual Name of the cookie and its Value
    String[] nameValue = parseNameValuePair(tok.nextToken());


    if(nameValue.length < 2)
      {
	throw new InvalidCookieException(str);
      }

    Cookie cookie = new Cookie(url, nameValue[0], nameValue[1]);

    while(tok.hasMoreTokens())
      {
	nameValue = parseNameValuePair(tok.nextToken());


	if(nameValue.length == 2 && nameValue[0].equalsIgnoreCase("comment"))
	  {
	    cookie.setComment(nameValue[1]);
	  }
	else if(nameValue.length == 2 && nameValue[0].equalsIgnoreCase("domain"))
	  {
	    cookie.setDomain(nameValue[1]);
	  }
	else if(nameValue.length == 2 && nameValue[0].equalsIgnoreCase("max-age"))
	  {
	    try
	      {
		cookie.setMaxAge(Integer.parseInt(nameValue[1]));
	      }
	    catch(NumberFormatException ex)
	      {
		throw new InvalidCookieException(str);		
	      }
	  }
	else if(nameValue.length == 2 && nameValue[0].equalsIgnoreCase("expires"))
	  {
	    // from the original netscape spec
	    // in reality, this is what people use.
	    
	    // need to find an easy way to parse these...
	  }
	else if(nameValue.length == 2 && nameValue[0].equalsIgnoreCase("path"))
	  {
	    cookie.setPath(nameValue[1]);
	  }
	else if(nameValue.length == 1 && nameValue[0].equalsIgnoreCase("secure"))
	  {
	    cookie.setSecure(true);
	  }
	else if(nameValue.length == 2 && nameValue[0].equalsIgnoreCase("version"))
	  {
	    try
	      {
		cookie.setVersion(Integer.parseInt(nameValue[1]));
	      }
	    catch(NumberFormatException ex)
	      {
		throw new InvalidCookieException(str);		
	      }
	  }
	else
	  {
	    // Unknown field. Ignore.
	  }
	
      }

    return cookie;
  }

  // parse all the cookies from the Set-Cookie HTTP header
  public static Cookie[] create(URL url, String cookieHeader) throws InvalidCookieException
  {
    Vector cookies = new Vector();
    // According to the specs, multiple cookies are separated by ','
    // unfortunately we get the following cookie from www.amazon.com:
    //
    // skin=; domain=.amazon.com; path=/; expires=Wed, 01-Aug-01 12:00:00 GMT
    // 
    // (Notice the embedded ',' in the date.)
    //
    // So I checked the Mozilla source, and they allow multiple cookies
    // if they are separated with '\n'. Embedded commas are treated as part
    // of the cookie. I'm going to do what Mozilla does.
    
    StringTokenizer tok = new StringTokenizer(cookieHeader, "\n");
    while(tok.hasMoreTokens())
      {
	try
	  {
	    cookies.addElement(createOneCookie(url, tok.nextToken()));
	  }
	catch(InvalidCookieException ex)
	  {
	    Logger.log(Cookie.class, ex);
	    throw ex;
	  }
      }
    
    return (Cookie[])cookies.toArray(new Cookie[cookies.size()]);
  }

  // parse the cookies and put them in the default cookie jar
  public static void processCookies(URL url, String cookieHeader)
    throws BadCookieException, InvalidCookieException
  {
    try
      {
	processCookies(url, cookieHeader, "default");
      }
    catch(NoSuchCookieJarException ex)
      { // should always have a default cookie jar
	Logger.log(Cookie.class, ex);
      }
  }

  // parse the cookies and put them in the named jar
  public static void processCookies(URL url, String cookieHeader, String jar)
    throws BadCookieException, InvalidCookieException, NoSuchCookieJarException
  {
    Cookie[] cookies = create(url, cookieHeader);

    for(int i = 0; i < cookies.length; i++)
      {
	DefaultCookieService.getCookieJar(jar).putCookie(cookies[i]);
      }
  }

  public static String getCookieField(URL url)
  {
    try
      {
	return getCookieField(url, "default");
      }
    catch(NoSuchCookieJarException ex)
      { // should always have a default cookie jar
	Logger.log(Cookie.class, ex);
	return null;
      }
  }

  public static String getCookieField(URL url, String jar)
    throws NoSuchCookieJarException
  {
    String ret = null;
    Cookie[] cookies = DefaultCookieService.getCookieJar(jar).getCookies(url);
    if(cookies != null && cookies.length > 0)
      {
	StringBuffer strb = new StringBuffer();
	for(int i = 0; i < cookies.length; i++)
	  {
	    if(i > 0)
	      {
		strb.append("; ");
	      }
	    strb.append(cookies[i].getName() + "=" + cookies[i].getValue());
	  }
	ret = strb.toString();
      }

    return ret;
  }

  public static String trim(String str)
  {
    String ws = " \t\r\n";
    int i;
    int j;

    for(i = 0; i < str.length() && ws.indexOf(str.charAt(i)) != -1; i++)
      {}

    for(j = str.length() - 1; j >= 0 && ws.indexOf(str.charAt(j)) != -1; j--)
      {}

    // i == index of first non-whitespace
    // j == index of last non-whitespace

    try
      {
	return str.substring(i, j+1);
      }
    catch(IndexOutOfBoundsException ex)
      {
	return "";
      }
  }

  public static String[] parseNameValuePair(String string)
  {
    int i = string.indexOf("=");
    if(i == -1)
      { // it's a name without a value
	String name = trim(string);
	if(name.length() > 0)
	  {
	    return new String[] {name};
	  }
	else
	  {
	    return new String[0];
	  }
      }

    String name = trim(string.substring(0, i));
    String value = trim(string.substring(i+1));

    if(name.length() == 0)
      {
	return new String[0];
      }
    else
      {
	return new String[] { name, value };
      }
  }

  public static boolean domainMatch(String lhs, String rhs)
  {
    if(lhs == null || rhs == null)
      {
	return false;
      }

    // Host A's name domain-matches host B's if
    
    //   * both host names are IP addresses and their host name strings match
    //     exactly; or
    //
    //   * both host names are FQDN strings and their host name strings match
    //     exactly; or

    if(lhs.equals(rhs))
      {
	return true;
      }

    //   * A is a FQDN string and has the form NB, where N is a non-empty name
    //     string, B has the form .B', and B' is a FQDN string.  (So, x.y.com
    //     domain-matches .y.com but not y.com.)

    if(rhs.charAt(0) == '.' && lhs.endsWith(rhs))
      {
	return true;
      }

    return false;
  }

  public Cookie(URL url, String name, String value)
  {
    super(name, value);
    
    setOrigin(url);
  }

  public boolean matchUrl(URL url)
  {
    String domain = getDomain();
    if(domain == null)
      {
	domain = m_origin.getHost();
      }

    if(!domainMatch(url.getHost(), domain))
      {
	//System.err.println("MATCH-URL: url.getHost()=" + url.getHost() + ", domain=" + domain);
	return false;
      }

    String path = getPath();
    if(path == null)
      {
	path = m_origin.getFile();
	if(path.lastIndexOf("/") > 0)
	  {
	    path = path.substring(0, path.lastIndexOf("/"));
	  }
      }

    if(!url.getFile().startsWith(path))
      {
	//System.err.println("MATCH-URL: url.getFile()=" + url.getFile() + ", path=" + path);
	return false;
      }
    return true;
  }

  public boolean isExpired()
  {
    return false;
  }
  
  public boolean checkOrigin()
  {
    //4.3.2  Rejecting Cookies

    //   To prevent possible security or privacy violations, a user agent
    //   rejects a cookie (shall not store its information) if any of the
    //   following is true:

    if(m_origin == null)
      {
	System.err.println("CHECK_ORIGIN: origin=null");
	return false;
      }

    String originHost = m_origin.getHost();
    String originPath = m_origin.getFile();

    String domain = getDomain();
    String path = getPath();

    //   * The value for the Path attribute is not a prefix of the
    //     request-URI.

    if(path != null && !originPath.startsWith(path))
      {
	System.err.println("CHECK_ORIGIN(1): origin.getFile()='" + originPath + "', path='" + path + "'");
	return false;
      }

    //   * The value for the Domain attribute contains no embedded dots or
    //     does not start with a dot.

    if(domain != null && (domain.charAt(0) != '.' || domain.lastIndexOf('.') == 0))
      {
	System.err.println("CHECK_ORIGIN(2): domain='" + domain + "'");
	return false;
      }

    //   * The value for the request-host does not domain-match the Domain
    //     attribute.

    if(domain != null && !domainMatch(originHost, domain))
      {
	System.err.println("CHECK_ORIGIN(3): m_origin.getHost()='" + originHost + "', domain='" + domain + "'");
	return false;
      }

    //   * The request-host is a FQDN (not IP address) and has the form HD,
    //     where D is the value of the Domain attribute, and H is a string
    //     that contains one or more dots.

    /*    if(domain != null && originHost.endsWith(domain)
       && originHost.substring(0, originHost.lastIndexOf(domain)).indexOf('.') > -1)
      {
	System.err.println("CHECK_ORIGIN(4): m_origin.getHost()='" + originHost + "', domain='" + domain + "'");
	return false;
      }
    */
    return true;
  }

  public URL getOrigin()
  {
    return m_origin;
  }

  public void setOrigin(URL url)
  {
    m_origin = url;
  }
  
  public Date getExpires()
  {
    return m_expires;
  }

  public void setExpires(Date expires)
  {
    m_expires = expires;
  }

  public String toString()
  {
    return ( "{ Cookie: " +
	     ( "name='" + getName() + "'," +
	       "value='" + getValue() + "'," +
	       "comment='" + getComment() + "'," +
	       "domain='" + getDomain() + "'," +
	       "maxAge='" + getMaxAge() + "'," +
	       "path='" + getPath() + "'," +
	       "secure='" + getSecure() + "'," +
	       "version='" + getVersion() + "'," +
	       "origin='" + m_origin + "'"
	       ) +
	     "}"
	     );
  }

  public static void logCookies(Cookie[] cookies)
  {
    if(cookies == null)
      return;

    for(int i = 0; i < cookies.length; i++)
      {
	Logger.log(Cookie.class, cookies[i].toString());
      }
  }

  public static void main(String[] args)
    throws Exception
  {
    if(args.length != 2)
      {
	System.err.println("Cookie <URL> <Set-Cookie field value>");
	System.exit(1);
      }

    URL url = new URL(args[0]);

    Cookie[] cookies = create(url, args[1]);
    for(int i = 0; i < cookies.length; i++)
      {
	System.err.println(cookies[i]);
	cookies[i].checkOrigin();
      }
  }
}
