package org.xoe.protocol.http;

import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;

import java.net.URL;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import org.w3c.dom.Document;

public class DefaultCookieService
  implements ICookieService
{
  private static DefaultCookieService m_instance = new DefaultCookieService ();

  private Hashtable m_cookies = new Hashtable ();

  static
  {
    ServiceLocator.registerService (m_instance);
  }

  public static DefaultCookieService getInstance ()
  {
    return m_instance;
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    return new String[] {ICookieService.FUNCTION};
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (ICookieService.FUNCTION))
      {
	return new String[] { ICookieService.INTERFACE };
      }
    return null;
  }

  public String[] getFeatures (String func)
  {
    if (func.equals (ICookieService.FUNCTION))
      {
	return new String [] { "name", "default" };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public synchronized Cookie[] getCookies (URL url)
  {
    flushExpiredCookies ();

    Vector v = new Vector ();
    Enumeration e = m_cookies.elements ();

    while (e.hasMoreElements ())
      {
	Cookie cookie = (Cookie) e.nextElement ();

	if (cookie.matchUrl (url))
	  {
	    v.addElement (cookie);
	  }
      }

    return (Cookie[]) v.toArray (new Cookie[v.size ()]);
  }

  public synchronized Cookie[] getCookies ()
  {
    Cookie[] ret = new Cookie[m_cookies.size ()];

    Enumeration e = m_cookies.elements ();

    for (int i = 0; i < m_cookies.size (); i++)
      {
	ret[i] = (Cookie) e.nextElement ();
      }

    return ret;
  }

  public synchronized Cookie getCookie (URL url, String name)
  {
    if (name == null)
      {
	return null;
      }

    Enumeration e = m_cookies.elements ();
    while (e.hasMoreElements ())
      {
	Cookie cookie = (Cookie) e.nextElement ();
	if (cookie.matchUrl (url) && name.equals (cookie.getName ()))
	  {
	    return cookie;
	  }
      }

    return null;
  }

  public synchronized void putCookie (Cookie cookie) 
    throws BadCookieException
  {
    if (cookie.checkOrigin () == false)
      {
	throw new BadCookieException (cookie);
      }

    if (cookie.isExpired ())
      { // this is a cue to remove another cookie with the same name, domain, and path
	m_cookies.remove (cookie.getDomain () + "^" + cookie.getPath () + "^" + cookie.getName ());
      }
    else
      {
	m_cookies.put (cookie.getDomain () + "^" + cookie.getPath () + "^" + cookie.getName (), cookie);
      }
  }

  protected synchronized void flushExpiredCookies ()
  {
    Cookie[] cookies = getCookies ();
    for (int i = 0; i < cookies.length; i++)
      {
	if (cookies[i].isExpired ())
	  {
	    m_cookies.remove (cookies[i].getDomain () + "^" + cookies[i].getPath () + "^" + cookies[i].getName ());
	  }
      }
  }

  public synchronized void flushSessionCookies ()
  {
  }

  public static ICookieService getCookieJar (String name) throws NoSuchCookieJarException
  {
    if (name.equals ("default"))
      {
	return getInstance ();
      }

    SimpleServiceQuery q = new SimpleServiceQuery (ICookieService.INTERFACE);

    q.addFeature ("name", name);

    ICookieService jar = (ICookieService) ServiceLocator.findService (q);
    if (jar == null)
      {
	throw new NoSuchCookieJarException (name);
      }
    return jar;
  }

  public void logCookies ()
  {
    Cookie.logCookies (getCookies ());
  }

  public static void main (String[] args) throws Exception
  {
    m_instance = new DefaultCookieService ();
    
    testInsert ("http://www.foo.com/", "nAmE=vAlUe");
    testInsert ("http://www.foo.com/some/special/place", "holy=moly");
    testInsert ("http://www.foo.com/some/special/place", "holy=moly; path=/some/other/place");
    testGet ("http://www.foo.com/index.html");
  }

  public static void testGet (String u) throws Exception
  {
    Logger.log (Cookie.class, "Getting cookies for url '" + u + "'");
    Cookie.logCookies (m_instance.getCookies (new URL(u)));
  }

  public static void testInsert (String u, String c)
  {
    Logger.log (Cookie.class, "Inserting cookies '" + c + "' from url '" + u + "'");
    try
      {
	Cookie.processCookies (new URL(u), c);
      }
    catch (Exception ex)
      {
	Logger.log (Cookie.class, ex);
      }
    m_instance.logCookies ();
  }
}
