package org.xoe.images.content;

import java.awt.Image;
import java.awt.Toolkit;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URLConnection;

import org.w3c.dom.Document;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.IImageBuilder;
import org.xoe.display.dom.IViewableDocumentBuilder;


public class Images
  implements IViewableDocumentBuilder, IDocumentBuilder, IImageBuilder
{
  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = { "contenthandler" };
    return ret;
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals ("contenthandler"))
      {
        String[] ret = { 
                         IViewableDocumentBuilder.INTERFACE,
                         IDocumentBuilder.INTERFACE, 
                         IImageBuilder.INTERFACE,
                       };
        return ret;
      }
    
    return null;
  }

  public String[] getFeatures (String func)
  {
    if (func.equals ("contenthandler"))
      {
        return new String[] { 
                              FEATURE_MIMETYPE, "image/gif",
                              FEATURE_MIMETYPE, "image/jpeg",
                              FEATURE_MIMETYPE, "image/png",
                              FEATURE_MIMETYPE, "image/tiff",
                              FEATURE_MIMETYPE, "image/x-xbitmap",
                              FEATURE_EXTENSION, "gif",
                              FEATURE_EXTENSION, "jpeg",
                              FEATURE_EXTENSION, "jpg",
                              FEATURE_EXTENSION, "jpe",
                              FEATURE_EXTENSION, "png",
                              FEATURE_EXTENSION, "tiff",
                              FEATURE_EXTENSION, "tif",
                              FEATURE_EXTENSION, "xbm"
                            };
      }
    
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IImageBuilder

  public Image getImage (ContentElement content)
    throws IOException
  {
    InputStream in = content.getInputStream ();
    byte[] streamCopy;

    // is length valid?
    int len = content.getContentLength ();
    if (len > 0)
      {
        // valid length so use it to build the array and read all at once
        streamCopy = new byte[len];
        // Cannot guarantee read will read everything in one go
        int pos = 0;
        while (len > 0)
          {
            int actual = in.read (streamCopy, pos, len);
            if (actual == -1) break;
            pos += actual;
            len -= actual;
          }
      }
    else
      {
        // invalid length, so read from the stream until we get a -1 length back
        ByteArrayOutputStream baos = new ByteArrayOutputStream ();
        byte buf[] = new byte[1024];

        while (true)
          {
            len = in.read (buf);
            if (len == -1) break;
            baos.write (buf, 0, len);
          }

        baos.flush ();
        streamCopy = baos.toByteArray ();
      }

    Image img = Toolkit.getDefaultToolkit ().createImage (streamCopy);
    in.close ();
    return img;
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    String src = content.getURL ().toString ();
    String s = "<html xmlns='" 
               + XHTML_NAMESPACE 
               + "'><head><title>" 
               + src  
               + "</title></head><body><img src='" 
               + src 
               + "' /></body></html>";
    return (DOMlet.create (s.toCharArray (), src));
  }


  // -------------------------------------------------------------------------
  // org.xoe.core.content.IDocumentBuilder

  public XDocument getDocument (ContentElement content)
    throws IOException
  {
    return getViewableDocument (content);
  }
}
