/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Codec.java,v 1.5 2001/11/17 02:24:09 samc Exp $
 */ 
package org.xoe.audio;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Hashtable;

import javax.sound.sampled.AudioFormat;

/**
 * Base class for ICodecs.  
 * Based on a simple FilterInputStream.
 */
abstract public class Codec
  extends FilterInputStream 
  implements ICodec
{
  public static final String[] INTERFACES = { ICodec.class.getName () };

  private byte[] single = new byte[1];
  private CodecListener listeners;

  public Codec () 
  {
    super (null);
  }

  public void close () 
    throws IOException 
  {
    if (in != null) 
      {
        closeCodec ();
        in.close ();
        in = null;
      }
  }

  protected void finalize () 
    throws Throwable 
  {
    close ();
    super.finalize ();
  }

  public int read () 
    throws IOException 
  {
    synchronized (this) 
      {
        int len = read (single, 0, 1);
        if (len == 1) 
          {
            return single[0] & 0xFF;
          }
        else 
          {
            return -1;
          }
      }
  }

  public int read (byte[] b, int off, int len) 
    throws IOException 
  {
    return readCodec (b, off, len);
  }

  public void setInputStream (InputStream in) 
    throws IOException 
  {
    if (this.in == null) 
      {
        this.in = in;
        openCodec ();
      }
    else 
      {
        throw new IOException ("codec already has inputstream");
      }
  }

  /**
   * Send an event to any codec event listener.
   */
  public void codecEvent (String name, Object value) 
  {
    if (listeners != null) 
      {
        listeners.codecEvent (name, value);
      }
  }

  /**
   * Add an code listener.
   */
  public synchronized void addCodecListener (CodecListener listener) 
  {
    listeners = listener;
  }

  /**
   * Remove a codec listener.
   */
  public synchronized void removeCodecListener (CodecListener listener) 
  {
    listeners = null;
  }
}
