/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: XoeLineListener.java,v 1.3 2001/12/14 19:48:02 ben Exp $
 */ 
package org.xoe.audio;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.FileDescriptor;

import org.xoe.core.device.Device;
import org.xoe.core.device.IAudio;
import org.xoe.core.util.Logger;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineListener;

/**
 * Listener for LineEvents and play audio the the IAudio
 * service accordingly.
 */
public class XoeLineListener 
  implements LineListener
{
  static IAudio audio; 

  public void update (LineEvent event)
  {
    LineEvent.Type type = event.getType ();

    if (audio == null) { audio = Device.getAudio (); }

    if (type == SoundStreamEvent.Type.DATA)
      {
        if (audio != null)
          {
            try 
              {
                SoundStreamEvent sse = (SoundStreamEvent) event;
                audio.write (sse.getData (), sse.getOffset (), sse.getLength ());
              }
            catch (IOException e)
              {
                throw new Error (e.toString ());
              }
          }
      }
    else if (type == SoundStreamEvent.Type.TYPE)
      {
        boolean reset = false;
        open ();
        if (audio == null)
          {
            Logger.log (this, "no audio service found");
          }
        else
          {
            AudioFormat format = ((DataLine) event.getSource ()).getFormat ();
            try
              {
                audio.setFormat (format);
              }
            catch (IOException e)
              {
                throw new Error (e.toString ());
              }
          }
      }

    else if (type == SoundStreamEvent.Type.LOADING)
      {
	System.err.println ("XoeLineListener.update (): LOADING event");
      }
    else if (type == SoundStreamEvent.Type.COMMENTS)
      {
	System.err.println ("XoeLineListener.update (): COMMENTS event");
      }

    else if (type == SoundStreamEvent.Type.TIME)
      {
	System.err.println ("XoeLineListener.update (): TIME event");
      }

    else if (type == SoundStreamEvent.Type.VOLUME)
      {
	System.err.println ("XoeLineListener.update (): VOLUME event");
      }


  }

  public static void close ()
  {
    if (audio == null) { audio = Device.getAudio (); }
    if (audio == null)
      {
        Logger.log (XoeLineListener.class, "no audio service found");
      }
    else
      {
        audio.close ();
      }
  }

  public static void open ()
  {
    if (audio == null) { audio = Device.getAudio (); }
    if (audio == null)
      {
        Logger.log (XoeLineListener.class, "no audio service found");
      }
    else
      {
        try
          {
            audio.open ();
          }
        catch (IOException e)
          {
            throw new Error (e.toString ());
          }
      }
  }

  /**
   * Close the DSP explicitly rather than waiting for the GC 
   * to finalize the FileOutputStream.
   */
  public void finalize ()
  {
    close ();
  }
}
