/* 
 * JNI interfaces to libcanna translation libraries.
 *
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * $tvt: canna_jni.c,v 1.3 2001/11/29 01:44:44 samc Exp $ 
 */
#include <jni.h>
#include <string.h>
#include <canna/RK.h>

#include "config.h"
#include "euc.h"

/* these functions we steak from RKroma.o */
int RkCvtRoma (struct RkRxDic* romaji, char* dst, int maxdst, char* src, int srclen, int flags);
struct RkRxDic* RkOpenRoma (char* filename);

static struct RkRxDic* dict = 0;

#define CANNA_DICT_PATH "/var/lib/canna/dic"
#define ROMA_DICT_NAME "/usr/share/canna/default.cbp"
#define MAX_LIST_LEN 1024

/**
 * Initialize ramaji conversion.
 */
static int init_roma ()
{
  if (!(dict = RkOpenRoma (ROMA_DICT_NAME)))
    {
      fprintf (stderr, "error opening romaji dict: %s\n", ROMA_DICT_NAME);
      return 1;
    }
  return 0;
}

/**
 * Initilize canna conversion.  
 */
JNIEXPORT jboolean JNICALL
Java_org_xoe_input_canna_Canna_init (JNIEnv* env, jclass cls)
{
#ifdef HAVE_LIBRK
  if (RkwInitialize (CANNA_DICT_PATH) < 0)
#else
  if (RkInitialize (CANNA_DICT_PATH) < 0)
#endif
    {
      return JNI_FALSE;
    }

  return JNI_TRUE;
}

/**
 * Create a new Kana->Kanji context with the given EUC-JP string.
 */
JNIEXPORT jint JNICALL
Java_org_xoe_input_canna_Canna_newContext (JNIEnv* env, jclass cls)
{
  int context = -1;
#ifdef HAVE_LIBRK
  if ((context = RkwCreateContext ()) < 0) 
#else
  if ((context = RkCreateContext ()) < 0) 
#endif
    {
      return -1;
    }
#ifdef HAVE_LIBRK
  if (RkwSetDicPath(context, "canna") < 0)
    {
      return -1;
    }
#endif
  return context;
}

JNIEXPORT jboolean JNICALL
Java_org_xoe_input_canna_Canna_loadDict (JNIEnv* env, jclass cls, 
                                 jint context, jstring dictname)
{
  int rtn = -1;
  char* name;
  int len = (int) (*env)->GetStringUTFLength (env, dictname);
  if (len < 1) return -1;
  name = (char*) (*env)->GetStringUTFChars (env, dictname, 0);
#ifdef HAVE_LIBRK
  rtn = RkwMountDic (context, name, 0);
#else
  rtn = RkMountDic (context, name, len);
#endif

  (*env)->ReleaseStringUTFChars (env, dictname, name);
  return rtn < 0 ? JNI_FALSE : JNI_TRUE;
}

/**
 * Set the word to get the candidates, return number of clauses
 */
JNIEXPORT jint JNICALL
Java_org_xoe_input_canna_Canna_setWord (JNIEnv* env, jclass cls,
        jint context, jbyteArray string) 
{
  int clauses = -1;
  char * jpword;
  Ushort * wword;
  
  int len = (*env)->GetArrayLength (env, string) + 1;

  /* prepare the jpword */
  jpword = (char*) malloc (len * sizeof (char));
  memset (jpword, 0, len);
  memcpy (jpword, (*env)->GetByteArrayElements (env, string, 0), len - 1);
  /* set the word */
#ifdef HAVE_LIBRK
  wword = (Ushort*) malloc (len * sizeof (Ushort));
  euc2ushort (jpword, len, wword, len);
  printf ("%d\n", ushortstrlen (wword));
  clauses = RkwBgnBun (context, (wchar_t*)wword, ushortstrlen (wword), RK_XFER);
  printf ("clauses: %d\n", clauses);
#else
  clauses = RkBgnBun (context, (unsigned char *)jpword, len, RK_XFER);
#endif
  free (jpword);
  return clauses;
}

JNIEXPORT jint JNICALL
Java_org_xoe_input_canna_Canna_nextClause (JNIEnv* env, jclass cls, jint context)
{
#ifdef HAVE_LIBRK
  return RkwRight (context);
#else
  return RkRight (context);
#endif
}

JNIEXPORT jint JNICALL
Java_org_xoe_input_canna_Canna_previousClause (JNIEnv* env, jclass cls, jint context)
{
#ifdef HAVE_LIBRK
  return RkwLeft (context);
#else
  return RkLeft (context);
#endif
}

/**
 * Clean up everything including the connection to the canna server.
 */
JNIEXPORT void JNICALL
Java_org_xoe_input_canna_Canna_closeContext (JNIEnv* env, jclass cls, jint context)
{
#ifdef HAVE_LIBRK
  RkwEndBun (context, 0);
  RkwCloseContext (context);
#else
  RkEndBun (context, 0);
  RkCloseContext (context);
#endif
}

/**
 * Get the array of EUC-JP strings representing the current candidates
 * for the context.
 */
JNIEXPORT jbyteArray JNICALL
Java_org_xoe_input_canna_Canna_getCandidates0 (JNIEnv* env, jclass cls, jint context)
{
  jbyteArray rtn = 0;
  /* jobjectArray rtn = 0; */
  int len     = 0;
  int pos     = 0;
  char list[MAX_LIST_LEN];
#ifdef HAVE_LIBRK
  Ushort list2[MAX_LIST_LEN];
#endif
  RkStat stat; 

  memset (list, 0, MAX_LIST_LEN);
  /* get the stats */
#ifdef HAVE_LIBRK
  memset (list2, 0, MAX_LIST_LEN * sizeof (Ushort));
  if (RkwGetStat (context, &stat) == -1)
#else
  if (RkGetStat (context, &stat) == -1)
#endif
    {
      fprintf (stderr, "canna: library error (RkGetStat)\n");
      return (*env)->NewByteArray (env, 0);
    }

  if (stat.maxcand == 0)
    {
      fprintf (stderr, "canna: no candidates found\n");
      /* return (*env)->NewObjectArray (0, env, jbyteArray, 0); */
      return (*env)->NewByteArray (env, 0);
    }

#ifdef HAVE_LIBRK
  if ((len = RkwGetKanjiList (context, list2, MAX_LIST_LEN)) == -1)
#else
  if ((len = RkGetKanjiList (context, (unsigned char*) list, MAX_LIST_LEN)) == -1)
#endif 
    {
      fprintf (stderr, "canna: library error\n");
      return (*env)->NewByteArray (env, 0);
    }
  printf ("%d\n", len);

#ifdef HAVE_LIBRK
  ushort2euc (list2, MAX_LIST_LEN, list, MAX_LIST_LEN);
#endif 
  /* Len returned is number of strings.  We convert it to 
     total length of string. [FIXME] ugly */
  while (len--)
    {
      int test;
      if (pos >= MAX_LIST_LEN -1) break;
      test = strlen (list+pos);
      pos += strlen (list+pos);
      pos++;
    }
  len = pos - 1;

  if (len == MAX_LIST_LEN - 1)
    {
      fprintf (stderr, "canna: possibly truncated Kanji list\n");
    }

  rtn = (*env)->NewByteArray (env, len);
  (*env)->SetByteArrayRegion (env, rtn, 0, len, list);
  return rtn;
}

/**
 * Convert EUC-JP Ramaji to EUC-JP Hiragana.
 */
JNIEXPORT jbyteArray JNICALL
Java_org_xoe_input_canna_Canna_romaToHira0 (JNIEnv* env, jclass cls, jbyteArray in_euc)
{
  jbyteArray rtn = 0;
  int in_len;
  int out_len;
  char* hira;
  char* input;
  char* input2;

  if (!dict) init_roma ();

  input = (*env)->GetByteArrayElements (env, in_euc, 0);
  in_len = (*env)->GetArrayLength (env, in_euc )+1;
  /* 
   * this is the hack because canna lib needs \0 char at the 
   * end of the input 
   */
  input2 = (char*) malloc (in_len);

  strncpy (input2, input, in_len-1);
  input2[in_len-1] = '\0';

  out_len = 2 * in_len * sizeof (char);
  hira = (char*) malloc (out_len);
  memset (hira, 0, (out_len));
  out_len = RkCvtRoma (dict, hira, out_len, input2, in_len, 
             RK_FLUSH | RK_NFER | RK_SOKON | RK_IGNORECASE);
  if (out_len > 0)
    {
      rtn = (*env)->NewByteArray (env, out_len - 1);
      (*env)->SetByteArrayRegion (env, rtn, 0, out_len - 1, hira);
    }
  free (input2);
  free (hira);
  return rtn;
}

/**
 * Convert EUC-JP Romaji to EUC-JP Katakana.
 */
JNIEXPORT jbyteArray JNICALL
Java_org_xoe_input_canna_Canna_romaToKata0 (JNIEnv* env, jclass cls, jbyteArray in_euc)
{
  jbyteArray rtn = 0;
  unsigned char * hira;
  unsigned char * kata;
  int in_len;
  int out_len;
  int out_max;
  char* input;

  if (!dict) init_roma ();

  input = (*env)->GetByteArrayElements (env, in_euc, 0);
  in_len = strlen (input);
  out_max = 2 * in_len;
  hira = (unsigned char*) malloc (out_max);
  kata = (unsigned char*) malloc (out_max);
  memset (hira, 0, (out_max));
  memset (kata, 0, (out_max));
  /* first we get the Hiragana */
  out_len = RkCvtRoma (dict, (char*) hira, out_max, input, in_len, 
             RK_FLUSH | RK_NFER | RK_SOKON | RK_IGNORECASE);
  /* then we get the Katakana */
  out_len = RkCvtKana (kata, out_max, hira, out_len);
  if (out_len > 0)
    {
      /* printf ("canna: kata (%d): %s\n", out_len, kata); */
      rtn = (*env)->NewByteArray (env, out_len);
      (*env)->SetByteArrayRegion (env, rtn, 0, out_len, (char*) kata);
    }
  free (hira);
  free (kata);
  return rtn;
}

/**
 * Convert EUC-JP Katakana to EUC-JP Katakana.
 */
JNIEXPORT jbyteArray JNICALL
Java_org_xoe_input_canna_Canna_kataToHira0 (JNIEnv* env, jclass cls, jbyteArray in_euc)
{
  return Java_org_xoe_input_canna_Canna_romaToHira0 (env, cls, in_euc);
}

/**
 * Convert EUC-JP Katakana to EUC-JP Katakana.
 */
JNIEXPORT jbyteArray JNICALL
Java_org_xoe_input_canna_Canna_hiraToKata0 (JNIEnv* env, jclass cls, jbyteArray in_euc)
{
  jbyteArray rtn = 0;
  int in_len;
  int out_len;
  int out_max;
  char* kata;
  char* input = (*env)->GetByteArrayElements (env, in_euc, 0);
  in_len = strlen (input);
  out_max = 2 * in_len;
  kata = (char*) malloc (out_max);
  memset (kata, 0, (out_max));
  out_len = RkCvtKana ((unsigned char*) kata, out_max, (unsigned char*) input, in_len);
  if (out_len > 0)
    {
      /* printf ("canna: kata (%d): %s\n", out_len, kata); */
      rtn = (*env)->NewByteArray (env, out_len);
      (*env)->SetByteArrayRegion (env, rtn, 0, out_len, (char*) kata);
    }
  free (kata);
  return rtn;
}
