#include "libsslio.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

int exDataIndex;

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslCtx_init0(JNIEnv* env, jclass SslCtx)
{
  SSL_library_init();

  exDataIndex = SSL_get_ex_new_index(0, "XOE SSL Library", 0, 0, 0);

  // replaces SSL's read/write functions with our own
  kaffe_set_fsci(&Kaffe_FileSystemCallInterface);

  // these calls can be removed to save memory
  SSL_load_error_strings();
  ERR_load_crypto_strings();
  ERR_load_OBJ_strings();

  // This can be replaced with calls to add only a few
  // basic algorithms, to save memory
  OpenSSL_add_all_algorithms();

  // from `man RAND_seed` it appears that on linux,
  // /dev/urandom is automatically used for entropy
  // should do a #warning here on other platforms, but I'm
  // not sure what to check (#ifndef LINUX ?)
}


JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslCtx_freeCtx0(JNIEnv* env, jobject this,
						   jobject internal)
{
  SSL_CTX_free((SSL_CTX*)internal);
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslCtx_setCertificateDirectory0(JNIEnv* env, jobject this,
								   jobject internal,
								   jstring dirpath)
{
  const char *s;

  s = (*env)->GetStringUTFChars(env, dirpath, 0);
  SSL_CTX_load_verify_locations((SSL_CTX*)internal, NULL, s);
  (*env)->ReleaseStringUTFChars(env, dirpath, s);
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_SslCtx_addCertificateFile0(JNIEnv* env, jobject this,
							     jobject internal,
							     jstring certpath)
{
  const char *s;

  s = (*env)->GetStringUTFChars(env, certpath, 0);
  SSL_CTX_load_verify_locations((SSL_CTX*)internal, s, NULL);
  (*env)->ReleaseStringUTFChars(env, certpath, s);
}

JNIEXPORT jobject JNICALL
Java_com_transvirtual_protocol_ssl_SslCtx_initInstance0(JNIEnv* env, jobject this,
							jint mode)
{
  SSL_METHOD *method = 0;
  SSL_CTX *ctx = 0;

  char errorString[256];

  switch(mode)
    {
    case TLS_V1:
      method = TLSv1_method();
      break;
      
    case TLS_V1_CLIENT:
      method = TLSv1_client_method();
      break;
      
    case TLS_V1_SERVER:
      method = TLSv1_server_method();
      break;
      
    case SSL_V2:
      method = SSLv2_method();
      break;
      
    case SSL_V2_CLIENT:
      method = SSLv2_client_method();
      break;
      
    case SSL_V2_SERVER:
      method = SSLv2_server_method();
      break;
      
    case SSL_V3:
      method = SSLv3_method();
      break;
      
    case SSL_V3_CLIENT:
      method = SSLv3_client_method();
      break;
      
    case SSL_V3_SERVER:
      method = SSLv3_server_method();
      break;
      
    case SSL_V23:
      method = SSLv23_method();
      break;
      
    case SSL_V23_CLIENT:
      method = SSLv23_client_method();
      break;
      
    case SSL_V23_SERVER:
      method = SSLv23_server_method();
      break;
      
    default:
      snprintf(errorString, sizeof(errorString), "%d does not denote a valid SSL_METHOD", mode);
      throwSSLRuntimeException(env, errorString);
      return 0;
    }

  ctx = SSL_CTX_new(method);
  if(ctx == 0)
    {
      throwSSLRuntimeException(env, (char*)ERR_error_string(ERR_get_error(), NULL));
      return 0;
    }

  return (jobject)ctx;
}
