#include "libsslio.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

jobject newX509CertificateChain(JNIEnv *env, STACK_OF(X509) *chain)
{
  jclass clazz = NULL;
  jmethodID init = NULL;
  
  if(0 == (clazz = (*env)->FindClass(env, "com/transvirtual/protocol/ssl/X509CertificateChain")))
    return NULL;

  if(0 == (init = (*env)->GetMethodID(env, clazz, "<init>", "(Lkaffe/util/Ptr;I)V")))
    return NULL;
  
  return (*env)->NewObject(env, clazz, init, chain, sk_X509_num(chain));
}

jobject newX509Certificate(JNIEnv *env, X509 *cert)
{
  jclass clazz = NULL;
  jmethodID init = NULL;
  jobject ret = NULL;

  if(0 == (clazz = (*env)->FindClass(env, "com/transvirtual/protocol/ssl/X509Certificate")))
    return NULL;

  if(0 == (init = (*env)->GetMethodID(env, clazz, "<init>", "(Lkaffe/util/Ptr;)V")))
    return NULL;
  
  ret = (*env)->NewObject(env, clazz, init, cert);
  CRYPTO_add(&cert->references, 1, CRYPTO_LOCK_X509);
  return ret;
}


JNIEXPORT jobject JNICALL
Java_com_transvirtual_protocol_ssl_X509CertificateChain_item0(JNIEnv* env, jobject this,
							      jobject internal, jint i)
{
  return newX509Certificate(env, sk_X509_value(internal, i));
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_X509CertificateChain_freeCertChain0(JNIEnv* env, jobject this,
							      jobject internal)
{
  sk_X509_pop_free(internal,X509_free);
}


JNIEXPORT jobject JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getExtension0(JNIEnv* env, jobject this,
								 jobject internal, int i)
{
  jclass clazz = NULL;
  jmethodID init = NULL;
  jobject ret = NULL;
  X509_EXTENSION *ex;

  ex=X509_get_ext((X509*)internal,i);

  if(0 == (clazz = (*env)->FindClass(env, "com/transvirtual/protocol/ssl/X509Extension")))
    return NULL;

  if(0 == (init = (*env)->GetMethodID(env, clazz, "<init>", "(Lcom/transvirtual/protocol/ssl/X509Certificate;Lkaffe/util/Ptr;)V")))
    return NULL;
  
  ret = (*env)->NewObject(env, clazz, init, this, ex);

  return ret;
}

JNIEXPORT jobject JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_init0(JNIEnv* env, jobject this,
							      jobject jbuf, int len)
{
  jbyte *buf = NULL;
  BIO *bio = NULL;
  X509 *cert = NULL;

  if(NULL == (buf = (*env)->GetByteArrayElements(env, jbuf, NULL)))
    {
      goto err;
    }
  if(NULL == (bio = BIO_new_mem_buf(buf, len)))
    {
      goto err;
    }
  if(NULL == (cert = PEM_read_bio_X509_AUX(bio,NULL,NULL,NULL)))
    {
      goto err;
    }
  else
    {
      goto cleanup;
    }

 err:;
  if(!(*env)->ExceptionOccurred(env))
    {
      throwSSLRuntimeException(env, "failure in X509Certificate.init0");
    }
 cleanup:;
  if(bio != NULL)
    {
      BIO_free(bio);
    }
  if(buf != NULL)
    {
      (*env)->ReleaseByteArrayElements(env, jbuf, buf, JNI_ABORT);
    }
  if(cert != NULL)
    {
      CRYPTO_add(&cert->references, 1, CRYPTO_LOCK_X509);
    }
  return cert;
}

JNIEXPORT void JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_freeCert0(JNIEnv* env, jobject this,
							      jobject internal)
{
  X509_free(internal);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getNotBefore0(JNIEnv* env, jobject this,
							      jobject internal)
{
  return (jstring)ASN1_TIME_to_jstring(env, X509_get_notBefore((X509*)internal));
}

JNIEXPORT jlong JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getVersion0(JNIEnv* env, jobject this,
							      jobject internal)
{
  // http://www.cs.auckland.ac.nz/~pgut001/pubs/x509guide.txt
  // Version ::= INTEGER { v1(0), v2(1), v3(2) }

  return X509_get_version((X509*)internal) + 1;
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getSerialNumber0(JNIEnv* env, jobject this,
							      jobject internal)
{
  ASN1_INTEGER *bs;
  long l;
  int i;
  char* neg;
  BIO *bio;
  char buffer[1024];
  jstring ret;

  bs=X509_get_serialNumber(internal);

  if(NULL == (bio = BIO_new(BIO_s_mem())))
    {
      goto err;
    }
  
  if (bs->length <= 4)
    {
      l=ASN1_INTEGER_get(bs);

      if (l < 0)
	{
	  l= -l;
	  neg="-";
	}
      else
	{
	  neg="";
	}

      if (BIO_printf(bio,"%s%lu (%s0x%lx)",neg,l,neg,l) <= 0)
	{
	  goto err;
	}
    }
  else
    {
      if(bs->type == V_ASN1_NEG_INTEGER)
	{
	  if(BIO_printf(bio," (Negative)") <= 0)
	    goto err;
	}

      for (i=0; i<bs->length; i++)
	{
	  if (BIO_printf(bio,(i+1 == bs->length) ? "%02x" : "%02x:",bs->data[i]) <= 0)
	    {
	      goto err;
	    }
	}
    }

  if(NULL == (BIO_get_mem_string (bio, buffer, 1024)))
    {
      goto err;
    }
  else if(NULL == (ret = (*env)->NewStringUTF(env, buffer)))
    {
      goto err;
    }
  else
    {
      goto cleanup;
    }

 err:;

  if(!(*env)->ExceptionOccurred(env))
    {
      throwSSLRuntimeException(env, "failure in getSerialNumber0");
    }

 cleanup:;

  if(bio != NULL)
    {
      BIO_free(bio);
    }

  return ret;

}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getSignature0(JNIEnv* env,
								 jobject this,
								 jobject internal)
{
  BIO *bio;
  char buffer[1024];
  jstring ret;
  int i, n;
  char *s;

  if(NULL == (bio = BIO_new(BIO_s_mem())))
    {
    }
  else 
    {
      n=((X509*)internal)->signature->length;
      s=(char *)((X509*)internal)->signature->data;
      for (i=0; i<n; i++)
	{
	  if ((i%18) == 0)
	    {
	      if (BIO_write(bio,"\n",1) <= 0) goto cleanup;
	    }
	  if (BIO_printf(bio,"%02x%s",(unsigned char)s[i],
			 ((i+1) == n)?"":":") <= 0) goto cleanup;
	}

      if(!BIO_get_mem_string(bio, buffer, 1024))
	{
	}
      else
	{
	  ret = (*env)->NewStringUTF(env, buffer);
	}
    }

 cleanup:;

  if(ret == NULL && !(*env)->ExceptionOccurred(env))
    {
      throwSSLRuntimeException(env, "failure in ASN1_TIME_to_jstring");
    }

  if(bio != NULL)
    {
      BIO_free(bio);
    }

  return ret;

}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getSignatureAlgorithm0(JNIEnv* env,
									  jobject this,
									  jobject internal)
{
  int i;
  char *name;
  jstring ret;

  i = OBJ_obj2nid(((X509*)internal)->sig_alg->algorithm);
  name = (i == NID_undef) ? "UNKNOWN" : (char*)OBJ_nid2ln(i);

  ret = (*env)->NewStringUTF(env, name);

  return ret;
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getPublicKey0(JNIEnv* env, jobject this,
								 jobject internal)
{
  BIO *bio;
  char buffer[1024];
  jstring ret;
  EVP_PKEY *pkey;

  if(NULL == (bio = BIO_new(BIO_s_mem())))
    {
    }
  else if(NULL == (pkey=X509_get_pubkey((X509*)internal)))
    {
    }
  else
    {
#ifndef NO_RSA
      if (pkey->type == EVP_PKEY_RSA)
	{
	  BIO_printf(bio,"RSA Public Key: (%d bit)\n","",
		     BN_num_bits(pkey->pkey.rsa->n));
	  RSA_print(bio,pkey->pkey.rsa,16);
	}
      else
#endif
#ifndef NO_DSA
      if (pkey->type == EVP_PKEY_DSA)
	{
	  BIO_printf(bio,"DSA Public Key:\n","");
	  DSA_print(bio,pkey->pkey.dsa,16);
	}
      else
#endif
	{
	  BIO_printf(bio,"Unknown Public Key Type\n","");
	}

      if(!BIO_get_mem_string(bio, buffer, 1024))
	{
	}
      else
	{
	  ret = (*env)->NewStringUTF(env, buffer);
	}
    }

  if(ret == NULL && !(*env)->ExceptionOccurred(env))
    {
      throwSSLRuntimeException(env, "failure in getSerialNumber0");
    }

  if(pkey != NULL)
    {
      EVP_PKEY_free(pkey);
    }

  if(bio != NULL)
    {
      BIO_free(bio);
    }

  return ret;
}

JNIEXPORT jint JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getNumberOfExtensions0(JNIEnv* env,
									  jobject this,
									  jobject internal)
{
  return X509_get_ext_count(internal);
}

JNIEXPORT jboolean JNICALL
Java_com_transvirtual_protocol_ssl_X509Extension_isCritical0(JNIEnv* env,
							     jobject this,
							     jobject internal)
{
  return X509_EXTENSION_get_critical((X509_EXTENSION*)internal) ? JNI_TRUE : JNI_FALSE;
}

JNIEXPORT jint JNICALL
Java_com_transvirtual_protocol_ssl_X509Extension_getID0(JNIEnv* env,
							jobject this,
							jobject internal)
{
  X509_EXTENSION *ex = (X509_EXTENSION*)internal;

  ASN1_OBJECT *obj;

  obj = ex->object;

  return OBJ_obj2nid (obj);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Extension_id2SN(JNIEnv* env, jclass SslCtx, jint nid)
{
  const char *sn;

  if (NID_undef == nid)
    {
      return NULL;
    }

  if (NULL == (sn = OBJ_nid2sn (nid)))
    {
      return NULL;
    }
  
  return (*env)->NewStringUTF(env, sn);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Extension_id2LN(JNIEnv* env, jclass SslCtx, jint nid)
{
  const char *ln;

  if (NID_undef == nid)
    {
      return NULL;
    }

  if (NULL == (ln = OBJ_nid2sn (nid)))
    {
      return NULL;
    }
  
  return (*env)->NewStringUTF(env, ln);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Extension_getShortName0(JNIEnv* env,
							       jobject this,
							       jobject internal)
{
  X509_EXTENSION *ex = (X509_EXTENSION*)internal;

  ASN1_OBJECT *obj;
  int nid;
  const char *sn;

  obj = ex->object;

  if (NID_undef == (nid = OBJ_obj2nid (obj)))
    {
      return NULL;
    }

  if (NULL == (sn = OBJ_nid2sn (nid)))
    {
      return NULL;
    }
  
  return (*env)->NewStringUTF(env, sn);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Extension_toString0(JNIEnv* env,
							   jobject this,
							   jobject internal)
{
  BIO *bio;
  char buffer[1024];
  jstring ret;
  X509_EXTENSION *ex = (X509_EXTENSION*)internal;

  if(NULL == (bio = BIO_new(BIO_s_mem())))
    {
    }
  else if(!X509V3_EXT_print(bio, ex, 0, 16))
    {
      M_ASN1_OCTET_STRING_print(bio,ex->value);
    }

  if(bio == NULL)
    {
    }
  else if(!BIO_get_mem_string(bio, buffer, 1024))
    {
    }
  else
    {
      ret = (*env)->NewStringUTF(env, buffer);
    }

  if(ret == NULL && !(*env)->ExceptionOccurred(env))
    {
      throwSSLRuntimeException(env, "failure in ASN1_TIME_to_jstring");
    }

  if(bio != NULL)
    {
      BIO_free(bio);
    }

  return ret;
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getPublicKeyAlgorithm0(JNIEnv* env,
									  jobject this,
									  jobject internal)
{
  int i;
  char *name;
  jstring ret;

  i = OBJ_obj2nid(((X509*)internal)->cert_info->key->algor->algorithm);
  name = (i == NID_undef) ? "UNKNOWN" : (char*)OBJ_nid2ln(i);

  ret = (*env)->NewStringUTF(env, name);

  return ret;
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getSubjectName0(JNIEnv* env, jobject this,
							      jobject internal)
{
  char buf[256];
  X509_NAME_oneline(X509_get_subject_name((X509*)internal), buf, sizeof(buf));
  return (*env)->NewStringUTF(env, buf);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getIssuerName0(JNIEnv* env, jobject this,
							      jobject internal)
{
  char buf[256];
  X509_NAME_oneline(X509_get_issuer_name((X509*)internal), buf, sizeof(buf));
  return (*env)->NewStringUTF(env, buf);
}

JNIEXPORT jstring JNICALL
Java_com_transvirtual_protocol_ssl_X509Certificate_getNotAfter0(JNIEnv* env, jobject this,
							      jobject internal)
{
  return (jstring)ASN1_TIME_to_jstring(env, X509_get_notAfter((X509*)internal));
}
