package com.transvirtual.protocol.https;

import com.transvirtual.protocol.ssl.SslSocket;
import com.transvirtual.protocol.ssl.X509Certificate;
import com.transvirtual.protocol.ssl.X509Extension;

import java.io.IOException;
import java.net.Socket;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import org.xoe.protocol.http.HttpURLConnection;

import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.core.util.Logger;
import org.xoe.core.util.Wildcard;

class HttpsURLConnection
  extends HttpURLConnection
{
  private static final int NID_SUBJECT_ALT_NAME = 85;

  // everytime we put something in m_accepted, increment.
  // when it gets to 20, purge all expired entries
  private static int m_counter = 0;

  // Time, in ms, for which a user acceptance is kept.
  // (if the user allows a certificate to be used, they will not
  // be asked again for this amount of time)
  private static final long ACCEPTANCE_TIMEOUT = 1000 * 60 * 60;

  // string-concat (certificate-serial-number, "^", hostname) --> Long (system time of acceptance)
  private static Hashtable m_accepted = new Hashtable ();

  private static boolean checkAccepted (String serialNum, String hostname)
  {
    Long lasttime = (Long)m_accepted.get (serialNum + "^" + hostname);

    if (lasttime != null)
      {
	if (lasttime.longValue () + ACCEPTANCE_TIMEOUT > System.currentTimeMillis ())
	  {
	    return true;
	  }

	m_accepted.remove (serialNum + "^" + hostname);
      }

    return false;
  }

  private static void accept ( String serialNum, String hostname)
  {
    m_accepted.put (serialNum + "^" + hostname, new Long (System.currentTimeMillis ()));
    m_counter++;
    if (m_counter >= 20)
      {
	Enumeration e = m_accepted.keys ();
	while (e.hasMoreElements ())
	  {
	    String key = (String)e.nextElement ();
	    Long time = (Long)m_accepted.get (key);
	    if (time.longValue () + ACCEPTANCE_TIMEOUT < System.currentTimeMillis ())
	      {
		m_accepted.remove (key);
	      }
	  }
	m_counter = 0;
      }
  }

  public HttpsURLConnection(URL u)
  {
    super(u);
  }

  protected String getProxyHost()
  {
    return null;
  }

  protected int getProxyPort()
  {
    return -1;
  }

  protected int defaultPort()
  {
    return 443;
  }

  protected Socket makeSocket(String host, int port) throws IOException
  {
    SslSocket sock = new SslSocket(host, port);
    Vector nonMatchingNames = new Vector ();
    
    Logger.log (this, "Https connection to: " + host + ":" + port);

    X509Certificate cert = sock.getPeerCertificate ();
    int n = cert.getNumberOfExtensions ();
    for ( int i = 0; i < n; i++)
      {
	X509Extension ext = cert.getExtension (i);
	if (ext.getID () == NID_SUBJECT_ALT_NAME)
	  {
	    if (nameMatch (host, ext.toString ()))
	      {
		return sock;
	      }
	    else
	      {
		nonMatchingNames.addElement (ext.toString ());
	      }
	  }
      }

    // no matching 'subjectAltName' fields were found, try
    // the subject's most specific CN
    // subject name is a DN with format "/NA=VAL/NA=VAL/..."
    String subject = cert.getSubjectName ();
    StringTokenizer tok = new StringTokenizer (subject, "/");
    String[] pairs = new String [tok.countTokens ()];

    for (int i = 0; i < pairs.length; i++)
      {
	pairs[i] = tok.nextToken ();
      }

    for (int i = pairs.length - 1; i >= 0; i--)
      {
	if (pairs[i].startsWith ("CN="))
	  {
	    if (nameMatch (host, pairs[i].substring (3)))
	      {
		return sock;
	      }
	    else
	      { // only try the most specific CN
		nonMatchingNames.addElement (pairs[i].substring (3));
		break;
	      }
	  }
      }
    
    if (checkAccepted (cert.getSerialNumber(), host))
      {
	return sock;
      }

    // should show a dialog offering to ignore the mismatch
    Logger.log (this, ("While contacting host " + host + ", received a certificate " +
		       "valid for the following names:"));

    for (int i = 0; i < nonMatchingNames.size (); i++)
      {
	Logger.log (this, nonMatchingNames.elementAt (i));
      }

    StringBuffer strb = new StringBuffer ();
    
    strb.append ( "<html><head><title>HTTPS Warning</title></head>" );
    strb.append ( "<body><b>While connecting to host " + host + ":" + port + " " );
    strb.append (          "received a certificate valid for the following ");
    strb.append (          "non-matching host-name(s):</b><br /><br/>" );
    strb.append ("<ul>");
    for (int i = 0; i < nonMatchingNames.size (); i++)
      {
	strb.append ( "<li>" + nonMatchingNames.elementAt (i) + "</li>");
      }

    strb.append ( "</ul>" );
    strb.append ( "      <b>Do you wish to accept the certificate?</b><br/><br/>");
    strb.append ( "      <form action=''>" );
    strb.append ( "        <input type='submit' name='answer' value='yes' /> ");
    strb.append ( "        <input type='submit' name='answer' value='no' /> ");
    strb.append ( "      </form></body></html>" );
    
    if (Dialog.confirm (strb.toString ()))
      {
	accept (cert.getSerialNumber (), host);
	return sock;
      }

    throw new HostNameMismatchException (host + ":" + port, nonMatchingNames );
  }

  private static boolean nameMatch (String str, String pattern)
  {
    StringTokenizer tStr = new StringTokenizer (str, ".");
    StringTokenizer tPat = new StringTokenizer (pattern, ".");

    if (tStr.countTokens () != tPat.countTokens ())
      {
	return false;
      }
    
    while (tStr.hasMoreTokens ())
      {
	String stok = tStr.nextToken ();
	String ptok = tPat.nextToken ();

	// false is for "disable '?' matching"
	if (! Wildcard.match (stok, ptok, false ))
	  {
	    return false;
	  }
      }

    return true;
  }
}
