package com.transvirtual.protocol.ssl;

import java.net.InetAddress;

public interface ICertificateVerifier
{
  /* verify will be called once for every certificate in the chain,
     starting with the root CA. 

     'appData' is the object that was passed to
     SslCtx/SslIO.setVerifyCallback() .

     'addr' is the address of the peer.

     'depth' specifies the index of the current certificate being
     examinedin 'chain'.

     'preverify_ok' specifies whether the certificate has been
     cryptographically verified.

     If 'preverify_ok' is false, 'err' is the error number given by
     OpenSSL, and 'errString' is a human readable description of the
     error.

     To allow communication to continue, simply return. Otherwise, throw
     an instance of BadCertificateException.
  */
 
  public static final int X509_V_OK =                                       0;
  public static final int X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT =            2;
  public static final int X509_V_ERR_UNABLE_TO_GET_CRL =                    3;
  public static final int X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE =     4;
  public static final int X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE =      5;
  public static final int X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY =   6;
  public static final int X509_V_ERR_CERT_SIGNATURE_FAILURE =               7;
  public static final int X509_V_ERR_CRL_SIGNATURE_FAILURE =                8;
  public static final int X509_V_ERR_CERT_NOT_YET_VALID =                   9;
  public static final int X509_V_ERR_CERT_HAS_EXPIRED =                     10;
  public static final int X509_V_ERR_CRL_NOT_YET_VALID =                    11;
  public static final int X509_V_ERR_CRL_HAS_EXPIRED =                      12;
  public static final int X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD =       13;
  public static final int X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD =        14;
  public static final int X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD =       15;
  public static final int X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD =       16;
  public static final int X509_V_ERR_OUT_OF_MEM =                           17;
  public static final int X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT =          18;
  public static final int X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN =            19;
  public static final int X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY =    20;
  public static final int X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE =      21;
  public static final int X509_V_ERR_CERT_CHAIN_TOO_LONG =                  22;
  public static final int X509_V_ERR_CERT_REVOKED =                         23;
  public static final int X509_V_ERR_INVALID_CA =                           24;
  public static final int X509_V_ERR_PATH_LENGTH_EXCEEDED =                 25;
  public static final int X509_V_ERR_INVALID_PURPOSE =                      26;
  public static final int X509_V_ERR_CERT_UNTRUSTED =                       27;
  public static final int X509_V_ERR_CERT_REJECTED =                        28;
  /* These are 'informational' when looking for issuer cert */
  public static final int X509_V_ERR_SUBJECT_ISSUER_MISMATCH =              29;
  public static final int X509_V_ERR_AKID_SKID_MISMATCH =                   30;
  public static final int X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH =          31;
  public static final int X509_V_ERR_KEYUSAGE_NO_CERTSIGN =                 32;

  void verify(Object appData, InetAddress addr, int port,
			X509CertificateChain chain, int depth,
			boolean preverify_ok, int err, String errString)
                        throws BadCertificateException;  
}
