package com.transvirtual.protocol.ssl;

import java.util.Vector;
import java.net.InetAddress;

import kaffe.util.Ptr;

import org.xoe.core.util.Logger;

public class SslCtx
{
  public static final int TLS_V1 = 0;
  public static final int TLS_V1_CLIENT = 1;
  public static final int TLS_V1_SERVER = 2;
  public static final int SSL_V2 = 3;
  public static final int SSL_V2_CLIENT = 4;
  public static final int SSL_V2_SERVER = 5;
  public static final int SSL_V3 = 6;
  public static final int SSL_V3_CLIENT = 7;
  public static final int SSL_V3_SERVER = 8;
  public static final int SSL_V23 = 9;
  public static final int SSL_V23_CLIENT = 10;
  public static final int SSL_V23_SERVER = 11;

  private static SslCtx m_defaultContext = null;

  static
  {
    try
      {
	System.loadLibrary("sslio");
      }
    catch (Throwable t)
      {

      }
    init0();
  }
  
  private native static void init0();

  private Ptr m_internal = null;
  private int m_mode = -1;

  // each element is a Object[2] containing an ICertificateVerifier and an opaque object
  // to be passed to the verifier.
  private Vector m_verifiers = new Vector ();

  public static SslCtx getDefaultContext()
  {
    if(m_defaultContext == null)
      {
	m_defaultContext = new SslCtx();
      }
    return m_defaultContext;
  }

  public SslCtx()
  {
    this(SSL_V23);
  }
  
  public SslCtx(int mode)
  {
    m_internal = initInstance0(mode);
    m_mode = mode;
  }

  protected void finalize()
  {
    freeCtx0(m_internal);
  }

  protected void verify(InetAddress addr, int port, X509CertificateChain chain, int depth,
			boolean preverify_ok, int err, String errString)
                        throws BadCertificateException
  {
    if(m_verifiers.size () > 0)
      {
	Object[] verify_pair = (Object[])m_verifiers.elementAt (m_verifiers.size () - 1);
	ICertificateVerifier verify = ((ICertificateVerifier)verify_pair[0]);
	Object appData = verify_pair[0];
	verify.verify(appData, addr, port, chain, depth, preverify_ok, err, errString);
      }
    else
      {
	if(!preverify_ok)
	  {
	    throw new BadCertificateException(chain.item(depth), errString);
	  }
      }
  }

  public void setCertificateVerifier (ICertificateVerifier verify, Object appData)
  {
    m_verifiers.addElement ( new Object[] { verify, appData } );
  }

  public void removeCertificateVerifier (ICertificateVerifier verify)
  {
    for (int i = m_verifiers.size () - 1; i >= 0; i--)
      {
	if (((Object [])m_verifiers.elementAt (i))[0] == verify)
	  {
	    m_verifiers.removeElementAt (i);
	  }
      }
  }

  public ICertificateVerifier getCertificateVerifier ()
  {
    return (ICertificateVerifier)m_verifiers.elementAt (m_verifiers.size () - 1);
  }

  public void setCertificateDirectory(String dirpath)
  {
    setCertificateDirectory0(m_internal, dirpath);
  }

  public void addCertificateFile(String certpath)
  {
    addCertificateFile0(m_internal, certpath);
  }

  private native void addCertificateFile0(Ptr internal, String certpath);
  private native void setCertificateDirectory0(Ptr internal, String dirpath);
  private native Ptr initInstance0(int mode);
  private native void freeCtx0(Ptr internal);
}

