package com.transvirtual.protocol.ssl;

import kaffe.util.Ptr;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;

import org.xoe.core.util.Logger;

public class SslIO
{
  private SslCtx m_ctx = null;
  private Ptr m_internal = null;
  private InetAddress m_addr = null;
  private int m_port = -1;
  private ICertificateVerifier m_verify = null;
  private Object m_verifyAppData = null;

  public static void main(String args[]) throws UnknownHostException, SslException
  {
    if(args.length != 2)
      {
	System.err.println("SslIO <host> <certificate-dir>");
	System.exit(1);
      }

    SslCtx.getDefaultContext().setCertificateDirectory(args[1]);

    SslIO io = new SslIO();
    InetAddress addr = InetAddress.getByName(args[0]);
    io.connect(addr.getHostAddress(), 443);
    String request = "GET / HTTP/1.0\n\n";
    io.write(request.getBytes(), 0, request.length());
    byte buf[] = new byte[1024];
    int len = 0;
    while((len = io.read(buf, 0, 1024)) != -1)
    {
      System.err.write(buf, 0, len);
    }
    io.close();
  }

  public SslIO()
  {
    this(null);
  }

  public SslIO(SslCtx ctx)
  {
    if(ctx == null)
      {
	ctx = SslCtx.getDefaultContext();
      }
    m_ctx = ctx;

    m_internal = initInstance0(m_ctx);
  }

  protected void finalize()
  {
    freeSsl0(m_internal);
  }

  public X509Certificate getPeerCertificate ()
  {
    return getPeerCertificate0 (m_internal);
  }

  public void connect(String ip, int port) throws SslException, UnknownHostException
  {
    try
      {
	m_addr = InetAddress.getByName(ip);
	connect0(m_internal, ip, port);
	m_port = port;
      }
    catch (SslException ex)
      {
	Logger.log (this, ex);
	throw ex;
      }
    catch (UnknownHostException ex)
      {
	Logger.log (this, ex);
	throw ex;
      }
  }

  public void close() throws SslException
  {
    close0(m_internal);
  }

  public int available()
  {
    return 0;
  }

  public int read(byte[] b) throws SslException
  {
    int len = read(b, 0, b.length);

    return len;
  }
  
  public void flush()
  {
    // shouldn't need to do anything... there is no buffering here
  }

  public int read(byte[] buffer, int offset, int length) throws SslException
  {
    try
      {
    if (offset < 0 || offset + length > buffer.length)
      {
	throw new ArrayIndexOutOfBoundsException();
      }
    int r = read0(m_internal, buffer, offset, length);

    if (r > 0 || length == 0)
      {
	return (r);
      }
    else
      {
	return (-1);
      }
  }
    catch (SslException ex)
      {
	Logger.log (this, ex);
	throw ex;
      }
  }

  public int read() throws SslException
  {
    byte b[] = new byte[1];
    if (read(b, 0, 1) == 1)
      {
	return (b[0] & 0xFF);
      }
    else
      {
	return (-1);
      }
  }

  public void write(byte[] b) throws SslException
  {
    write(b, 0, b.length);
  }

  public void write(byte[] buffer, int offset, int length) throws SslException
  {
    try
      {
    if (offset < 0 || offset + length > buffer.length)
      {
	throw new ArrayIndexOutOfBoundsException();
      }
    write0(m_internal, buffer, offset, length);
      }
    catch (SslException ex)
      {
	Logger.log (this, ex);
	throw ex;
      }
  }

  public void write(int val) throws SslException
  {
    write(new byte[] { (byte)val } , 0, 1);
  }


  public void setCertificateVerifier (ICertificateVerifier verify, Object appData)
  {
    m_verify = verify;
    m_verifyAppData = appData;
  }

  public ICertificateVerifier getCertificateVerifier ()
  {
    return m_verify;
  }

  public void verifyCallback(X509CertificateChain chain, 
			      int depth, int err, String errString, boolean preverify_ok)
                              throws BadCertificateException
  {
    if(m_verify != null)
      {
	m_verify.verify(m_verifyAppData, m_addr, m_port, chain, depth, preverify_ok, err, errString);
      }
    else
      {
	m_ctx.verify(m_addr, m_port, chain, depth, preverify_ok, err, errString);
      }
  }

  private native X509Certificate getPeerCertificate0 (Ptr internal);
  private native Ptr initInstance0(SslCtx ctx);
  private native Ptr connect0(Ptr internal, String addr, int port) throws SslException;
  private native void freeSsl0(Ptr internal);
  private native void close0(Ptr internal) throws SslException;
  private native int read0(Ptr m_internal, byte[] buffer,
			   int offset, int length) throws SslException;
  private native void write0(Ptr internal, byte[] buffer,
			     int offset, int length) throws SslException;
}
