package com.transvirtual.protocol.ssl;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;

public class SslSocket
  extends Socket
{
  private InputStream m_in = new SslInputStream();
  private OutputStream m_out = new SslOutputStream();

  private InetAddress m_addr;
  private int m_port;

  private SslIO m_io;

  private class SslInputStream
    extends InputStream
  {
    public int available()
    {
      return m_io.available();
    }
    
    public void close()
    {
      SslSocket.this.close();
    }

    public int read() throws SslException
    {
      return m_io.read();
    }

    public int read(byte[] b) throws SslException
    {
      return m_io.read(b);
    }
    
    public int read(byte[] b, int off, int len) throws SslException
    {
      return m_io.read(b, off, len);
    }
  }

  private class SslOutputStream
    extends OutputStream
  {
    public void close()
    {
      SslSocket.this.close();
    }

    public void flush()
    {
      m_io.flush();
    }

    public void write(int b) throws SslException
    {
      m_io.write(b);
    }

    public void write(byte[] b) throws SslException
    {
      m_io.write(b);
    }
    
    public void write(byte[] b, int off, int len) throws SslException
    {
      m_io.write(b, off, len);
    }
  }

  public SslSocket(InetAddress address, int port) throws SslException
  {
    m_addr = address;
    m_port = port;
    m_io = new SslIO();
    try
      {
	m_io.connect(m_addr.getHostAddress(), port);
      }
    catch(UnknownHostException ex)
      {
	// we just got this IP from an InetAddress, so we don't expect to get this.
	throw new SslException("Unexpected UnknownHostException");
      }
  }
  
  public SslSocket(String host, int port) throws UnknownHostException, SslException
  {
    this(InetAddress.getByName(host), port);
  }

  public void close()
  {
    try
      {
	m_io.close();
      }
    catch(SslException ex)
      { // no need to throw on close()
	//ex.printStackTrace();
      }
  }

  public X509Certificate getPeerCertificate ()
  {
    return m_io.getPeerCertificate ();
  }

  public InetAddress getInetAddress()
  {
    return m_addr;
  }

  public InputStream getInputStream()
  {
    return m_in;
  }

  public InetAddress getLocalAddress()
  {
    try
      {
	return InetAddress.getLocalHost();
      }
    catch(UnknownHostException ex)
      {
        // can this really happen?
	ex.printStackTrace();
	return null;
      }
  }

  public int getLocalPort()
  {
    return m_port;
  }

  public OutputStream getOutputStream()
  {
    return m_out;
  }

  public int getReceiveBufferSize()
  {
    return 0;
  }

  public int getSendBufferSize()
  {
    return 0;
  }

  public int getSoLinger()
  {
    return 0;
  }

  public int getSoTimeout()
  {
    return 0;
  }

  public boolean getTcpNoDelay()
  {
    return false;
  }

  public void setReceiveBufferSize(int size)
  {
  }

  public void setSendBufferSize(int size)
  {
  }

  public void setSoLinger(boolean on, int linger)
  {
  }

  public void setSoTimeout(boolean on, int linger)
  {
  }

  public void setTcpNoDelay(boolean on)
  {
  }

  public String toString()
  {
    return "SslSocket { m_addr='" + m_addr + "', m_port='" + m_port + " }";
  }
}
