package com.transvirtual.protocol.ssl;

import kaffe.util.Ptr;

public class X509Certificate
{
  private Ptr m_internal = null;
  private X509Extension[] m_extensions = null;

  static
  {
    // force the libraries to be init'd
    SslCtx.getDefaultContext();
  }
  protected void finalize()
  {
    freeCert0(m_internal);
  }

  protected X509Certificate(Ptr internal)
  {
    m_internal = internal;
  }

  public X509Certificate (byte[] buf)
  {
    m_internal = init0(buf, buf.length);
  }

  public String toString()
  {
    StringBuffer strb = new StringBuffer();
    strb.append( "{ X509Certificate version='" + getVersion() + "'\n" + 
		 "    serialNumber='" + getSerialNumber() + "'\n" +
		 "    notBefore='" + getNotBefore() + "'\n" +
		 "    notAfter='" + getNotAfter() + "'\n" +
		 "    subjectName='" + getSubjectName() + "'\n\n" +
		 "    publicKeyAlgo='" + getPublicKeyAlgorithm() + "'\n" +
		 "    publicKey= " + getPublicKey() + "\n\n");

    int num = getNumberOfExtensions();

    for(int i = 0; i < num; i++)
      {
	getExtension (i);
	strb.append("Extension " + i + " (critical=" + m_extensions[i].isCritical () + "):\n");
	strb.append("Short Name: " + m_extensions[i].getShortName () + "\n");
	strb.append("Long Name: " + m_extensions[i].getLongName () + "\n");

	strb.append(m_extensions[i].toString ());
	strb.append("\n\n");
      }
    strb.append( "    issuerName='" + getIssuerName() + "'\n\n" +
		 "    signatureAlgorithm='" + getSignatureAlgorithm() + "'\n" +
		 "    signature=\n" + getSignature() + "\n\n");
    strb.append( "}");

    return strb.toString();
  }
  
  public long getVersion()
  {
    return getVersion0(m_internal);
  }
  
  public String getSerialNumber()
  {
    return getSerialNumber0(m_internal);
  }

  public String getNotBefore()
  {
    return getNotBefore0(m_internal);
  }

  public String getNotAfter()
  {
    return getNotAfter0(m_internal);
  }

  public String getSubjectName()
  {
    return getSubjectName0(m_internal);
  }

  public String getPublicKey()
  {
    return getPublicKey0(m_internal);
  }

  public String getPublicKeyAlgorithm()
  {
    return getPublicKeyAlgorithm0(m_internal);
  }

  public int getNumberOfExtensions()
  {
    if ( m_extensions == null )
      {
	m_extensions = new X509Extension[getNumberOfExtensions0(m_internal)];
      }

    return m_extensions.length;
  }

  public X509Extension getExtension (int i)
  {
    if (m_extensions[i] == null)
      {
	m_extensions[i] = getExtension0 (m_internal, i);
      }

    return m_extensions[i];
  }

  public String getIssuerName()
  {
    return getIssuerName0(m_internal);
  }

  public String getSignature()
  {
    return getSignature0(m_internal);
  }

  public String getSignatureAlgorithm()
  {
    return getSignatureAlgorithm0(m_internal);
  }
  
  private native Ptr init0(byte[] buf, int len);
  private native String getSubjectName0(Ptr internal);

  private native String getPublicKey0(Ptr internal);
  private native String getPublicKeyAlgorithm0(Ptr internal);

  private native int getNumberOfExtensions0(Ptr internal);
  private native X509Extension getExtension0(Ptr internal, int i);
  
  private native long getVersion0(Ptr internal);
  private native String getSerialNumber0(Ptr internal);

  private native String getNotBefore0(Ptr internal);
  private native String getNotAfter0(Ptr internal);

  private native String getIssuerName0(Ptr internal);
  private native String getSignature0(Ptr internal);
  private native String getSignatureAlgorithm0(Ptr internal);

  private native void freeCert0(Ptr internal);
}
