/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: X509CertificateInstaller.java,v 1.7 2001/11/09 19:53:28 erik Exp $
 */
package com.transvirtual.protocol.ssl;

import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.XNode;
import org.xoe.core.util.Logger;
import org.xoe.core.packages.Dependency;
import org.xoe.core.packages.FileIteratingInstaller;
import org.xoe.core.packages.PackageException;
import org.xoe.core.packages.PackageDescriptionException;
import org.xoe.core.packages.Proviso;
import org.xoe.core.packages.XoePackage;

import org.xoe.core.util.XMLWriter;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipEntry;

import org.w3c.dom.Element;
import org.w3c.dom.Text;

public class X509CertificateInstaller
  extends FileIteratingInstaller
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/x509certificate";
  private static XDocument m_doc = new XDocument ();
  private static final String tempLocation = "/tmp/";

  public String[] getHandlerPredependencies ()
  {
    return new String[0];
  }

  public Dependency buildDependency ( Element depEl )
    throws PackageDescriptionException
  {
    if ( NAMESPACE.equals (depEl.getAttribute ("ns")))
      {
	return new Dependency (depEl);
      }
    else
      {
	throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
      }
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
  {
    ZipInputStream zis = xoePackage.getZipInputStream ();

    try 
      {
	while (true)
	  {
	    ZipEntry entry = zis.getNextEntry ();
	    if (entry == null) break;
            String filename = entry.getName ();

            if (filename.endsWith(".pem") )
	      {
                byte buf[] = new byte[(int) entry.getSize ()];
                for (int res = 0, off = 0; res != -1; )
                  {
                    res = zis.read (buf, off, buf.length - off);
                    off += res;
                  }

                /* make a temporary file. */
                /* [FIXME] must we do this? */
                File file = new File(filename);
                filename = tempLocation + file.getName ();
                file = new File(filename);

                if (file.exists())
                  {
                    Logger.log (this, "Warning: file already exists: " + filename);
                  }

                /* read the data from the ZipInputStream */
                FileOutputStream ofile = new FileOutputStream (filename);
                ofile.write (buf, 0, buf.length);  
                
                /* load the file */
                Logger.log (this, "Loading X509Certificate: " + filename);
                SslCtx.getDefaultContext().addCertificateFile (filename);

                /* [FIXME] can we delete the temporary file now? 
                 */
		if(!file.delete())
		  {
		    Logger.log(this, "Warning: failed to delete temp file: " + filename);
		  }
	      }
	    zis.closeEntry ();
	  }
	zis.close ();
      }
    catch (IOException e)
      {
	Logger.log(this, "done with certs");
	return;
      }
  }

  public String getNamespaceURI()
  {
    return NAMESPACE;
  }

  public X509CertificateInstaller()
  {
    setSuffixes( new String[] { ".pem" } );
  }

  public void processFile(ZipInputStream zis, ZipEntry entry)
    throws PackageException
  {
    try
      {
	byte buf[] = new byte[(int) entry.getSize ()];
	for (int res = 0, off = 0; res != -1; )
	  {
	    res = zis.read (buf, off, buf.length - off);
	    off += res;
	  }

	X509Certificate cert = new X509Certificate(buf);
	makeProviso(cert);
      }
    catch(IOException ex)
      {
	throw new PackageException (ex,"An error occurred while processing an X509 certificate "
				    + " from package " + m_package.getPackageURI());
      }
  }

  void makeDependency (X509Certificate cert)
  {
    String subject = cert.getSubjectName();
    String issuer = cert.getIssuerName();

    // should do a comparison which takes account for the
    // fact that these are both DN's
    if(!subject.equals(issuer))
      { // depends on the CA cert (or whatever else is next up the chain)
	m_requires.addElement (new Dependency (NAMESPACE, issuer));
      }
  }

  void makeProviso (X509Certificate cert)
  {
    Element certE = m_doc.createElement ("certificate");
    Element subjectE = m_doc.createElement ("subject");
    Element issuerE = m_doc.createElement ("issuer");
    Element notBeforeE = m_doc.createElement ("notBefore");
    Element notAfterE = m_doc.createElement ("notAfter");

    certE.appendChild(subjectE);
    certE.appendChild(issuerE);
    certE.appendChild(notBeforeE);
    certE.appendChild(notAfterE);

    subjectE.appendChild(m_doc.createTextNode(cert.getSubjectName()));
    issuerE.appendChild(m_doc.createTextNode(cert.getIssuerName()));
    notBeforeE.appendChild(m_doc.createTextNode(cert.getNotBefore()));
    notAfterE.appendChild(m_doc.createTextNode(cert.getNotAfter()));

    m_provides.addElement (new Proviso (getNamespaceURI (), cert.getSubjectName(), certE));
  }
}

