/* 
 * JNI interface to libmad.
 * Resampling code taken directly from madplay (http://mad.sourceforce.net/).
 *
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: MP3Codec.c,v 1.10 2001/11/20 21:56:17 suraj Exp $
 */
#include <jni.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <mad.h>

#define	ENDIAN		0
#define	WORDSIZE	2
#define	SIGNED		1


#define FREQ_TOLERANCE 6          /* percent sampling frequency tolerance */
#define MAX_NSAMPLES   (1152 * 3) /* allow for resampled frame */

struct resample_state 
{
  mad_fixed_t ratio;
  mad_fixed_t step;
  mad_fixed_t last;
};

/**
 * All stream specific mad data.
 */
struct mad_struct
{
  struct mad_stream * stream;
  struct mad_frame  * frame;
  struct mad_synth  * synth;
  int    out_freq;      /* output frequency (Hz) */
  int    out_chan;      /* output channels (1 or 2) */
  unsigned char  *buffer;        /* internal buffer */
  unsigned int    pos;           /* offset into buffer */
  int    bsize;         /* size of buffer */
  int    framecount;    /* how many frames have we decoded so far */
  int    leftover_pcm;  /* hack....ho hum */
  mad_fixed_t *ch1, *ch2;
  int    nsamples;
  
  struct resample_state resample[2];
  mad_fixed_t (*resampled)[2][MAX_NSAMPLES];
};

static int
rate_index (unsigned int rate)
{ 
  switch (rate) 
    {
    case 48000: return 0;
    case 44100: return 1;
    case 32000: return 2;
    case 24000: return 3;
    case 22050: return 4;
    case 16000: return 5;
    case 12000: return 6;
    case 11025: return 7;
    case  8000: return 8;
    }
  return -1;
}

/**
 * Resampling table.
 */
static mad_fixed_t const 
resample_table[9][9] = {
  /* 48000 */ { MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x116a3b36) /* 1.088435374 */,
                MAD_F(0x18000000) /* 1.500000000 */,
                MAD_F(0x20000000) /* 2.000000000 */,
                MAD_F(0x22d4766c) /* 2.176870748 */,
                MAD_F(0x30000000) /* 3.000000000 */,
                MAD_F(0x40000000) /* 4.000000000 */,
                MAD_F(0x45a8ecd8) /* 4.353741497 */,
                MAD_F(0x60000000) /* 6.000000000 */ },

  /* 44100 */ { MAD_F(0x0eb33333) /* 0.918750000 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x160ccccd) /* 1.378125000 */,
                MAD_F(0x1d666666) /* 1.837500000 */,
                MAD_F(0x20000000) /* 2.000000000 */,
                MAD_F(0x2c19999a) /* 2.756250000 */,
                MAD_F(0x3acccccd) /* 3.675000000 */,
                MAD_F(0x40000000) /* 4.000000000 */,
                MAD_F(0x58333333) /* 5.512500000 */ },

  /* 32000 */ { MAD_F(0x0aaaaaab) /* 0.666666667 */,
                MAD_F(0x0b9c2779) /* 0.725623583 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x15555555) /* 1.333333333 */,
                MAD_F(0x17384ef3) /* 1.451247166 */,
                MAD_F(0x20000000) /* 2.000000000 */,
                MAD_F(0x2aaaaaab) /* 2.666666667 */,
                MAD_F(0x2e709de5) /* 2.902494331 */,
                MAD_F(0x40000000) /* 4.000000000 */ },

  /* 24000 */ { MAD_F(0x08000000) /* 0.500000000 */,
                MAD_F(0x08b51d9b) /* 0.544217687 */,
                MAD_F(0x0c000000) /* 0.750000000 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x116a3b36) /* 1.088435374 */,
                MAD_F(0x18000000) /* 1.500000000 */,
                MAD_F(0x20000000) /* 2.000000000 */,
                MAD_F(0x22d4766c) /* 2.176870748 */,
                MAD_F(0x30000000) /* 3.000000000 */ },

  /* 22050 */ { MAD_F(0x0759999a) /* 0.459375000 */,
                MAD_F(0x08000000) /* 0.500000000 */,
                MAD_F(0x0b066666) /* 0.689062500 */,
                MAD_F(0x0eb33333) /* 0.918750000 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x160ccccd) /* 1.378125000 */,
                MAD_F(0x1d666666) /* 1.837500000 */,
                MAD_F(0x20000000) /* 2.000000000 */,
                MAD_F(0x2c19999a) /* 2.756250000 */ },

  /* 16000 */ { MAD_F(0x05555555) /* 0.333333333 */,
                MAD_F(0x05ce13bd) /* 0.362811791 */,
                MAD_F(0x08000000) /* 0.500000000 */,
                MAD_F(0x0aaaaaab) /* 0.666666667 */,
                MAD_F(0x0b9c2779) /* 0.725623583 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x15555555) /* 1.333333333 */,
                MAD_F(0x17384ef3) /* 1.451247166 */,
                MAD_F(0x20000000) /* 2.000000000 */ },

  /* 12000 */ { MAD_F(0x04000000) /* 0.250000000 */,
                MAD_F(0x045a8ecd) /* 0.272108844 */,
                MAD_F(0x06000000) /* 0.375000000 */,
                MAD_F(0x08000000) /* 0.500000000 */,
                MAD_F(0x08b51d9b) /* 0.544217687 */,
                MAD_F(0x0c000000) /* 0.750000000 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x116a3b36) /* 1.088435374 */,
                MAD_F(0x18000000) /* 1.500000000 */ },

  /* 11025 */ { MAD_F(0x03accccd) /* 0.229687500 */,
                MAD_F(0x04000000) /* 0.250000000 */,
                MAD_F(0x05833333) /* 0.344531250 */,
                MAD_F(0x0759999a) /* 0.459375000 */,
                MAD_F(0x08000000) /* 0.500000000 */,
                MAD_F(0x0b066666) /* 0.689062500 */,
                MAD_F(0x0eb33333) /* 0.918750000 */,
                MAD_F(0x10000000) /* 1.000000000 */,
                MAD_F(0x160ccccd) /* 1.378125000 */ },

  /*  8000 */ { MAD_F(0x02aaaaab) /* 0.166666667 */,
                MAD_F(0x02e709de) /* 0.181405896 */,
                MAD_F(0x04000000) /* 0.250000000 */,
                MAD_F(0x05555555) /* 0.333333333 */,
                MAD_F(0x05ce13bd) /* 0.362811791 */,
                MAD_F(0x08000000) /* 0.500000000 */,
                MAD_F(0x0aaaaaab) /* 0.666666667 */,
                MAD_F(0x0b9c2779) /* 0.725623583 */,
                MAD_F(0x10000000) /* 1.000000000 */ }
};

/**
 * initialize resampling state
 */
static int 
resample_init (struct resample_state *state,
               unsigned int oldrate, unsigned int newrate)
{
  int oldi, newi;

  oldi = rate_index (oldrate);
  newi = rate_index (newrate);

  if (oldi == -1 || newi == -1) { return -1; }

  state->ratio = resample_table[oldi][newi];

  state->step = 0;
  state->last = 0;

  return 0;
}

/**
 * algorithmically change the sampling rate of a PCM sample block.
 */
static unsigned int 
resample_block (struct resample_state *state,
                unsigned int nsamples, mad_fixed_t const *old,
                mad_fixed_t *new)
{
  mad_fixed_t const *end, *begin;

  /*
   * This resampling algorithm is based on a linear interpolation, which is
   * not at all the best sounding but is relatively fast and efficient.
   *
   * A better algorithm would be one that implements a bandlimited
   * interpolation.
   */

  if (state->ratio == MAD_F_ONE) 
    {
      memcpy (new, old, nsamples * sizeof (mad_fixed_t));
      return nsamples;
    }

  end   = old + nsamples;
  begin = new;

  if (state->step < 0) 
    {
      state->step = mad_f_fracpart(-state->step);

      while (state->step < MAD_F_ONE) {
          *new++ = state->step ?
            state->last + mad_f_mul(*old - state->last, state->step) : state->last;

          state->step += state->ratio;
          if (((state->step + 0x00000080L) & 0x0fffff00L) == 0)
              state->step = (state->step + 0x00000080L) & ~0x0fffffffL;
      }

      state->step -= MAD_F_ONE;
    }

  while (end - old > 1 + mad_f_intpart(state->step)) 
    {
      old        += mad_f_intpart(state->step);
      state->step = mad_f_fracpart(state->step);

      *new++ = state->step ?
        *old + mad_f_mul(old[1] - old[0], state->step) : *old;

      state->step += state->ratio;
      if (((state->step + 0x00000080L) & 0x0fffff00L) == 0)
          state->step = (state->step + 0x00000080L) & ~0x0fffffffL;
    }

  if (end - old == 1 + mad_f_intpart(state->step)) 
    {
      state->last = end[-1];
      state->step = -state->step;
    }
  else
    {
      state->step -= mad_f_fromint(end - old);
    }

  return new - begin;
}

static inline signed int 
scale (mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/*
 * Write PCM data into output buffer.
 * @return bytes written
 */
static int
write_pcm (struct mad_struct* mad_data, char* bufout, int len)
{
  const mad_fixed_t * left_ch;
  const mad_fixed_t * right_ch;
  unsigned int nsamples;
  int i = 0;

  nsamples  = mad_data->nsamples;
  left_ch   = mad_data->ch1;
  right_ch  = mad_data->ch2;

  while (nsamples--) 
    {
      signed int sample;
      if (i > len-4)
        {
          /* printf("oops: output buffer too small\n"); */
          return -1;
        }
      /* output sample(s) in 16-bit signed little-endian PCM */
      sample      = scale(*left_ch++);
      bufout[i++] = (sample >> 0) & 0xff;
      bufout[i++] = (sample >> 8) & 0xff;

      sample      = scale(*right_ch++);
      bufout[i++] = (sample >> 0) & 0xff;
      bufout[i++] = (sample >> 8) & 0xff;
    }
  return i;
}

/*
 * Add some data to the mad buffer
 */
static inline int
mad_data_add (struct mad_struct * mad_data, char* bufin, int inlength)
{
  int   new_size      = 0;
  int   old_data_size = 0;
  char* hack_buffer   = 0;

  /* tack on the new input data if there is any */
  new_size = mad_data->bsize - mad_data->pos + inlength;
  if (mad_data->buffer)
    {
      /* copy OLD buffer into hack buffer */
      hack_buffer = (char*) malloc (new_size * sizeof (char));
      old_data_size = mad_data->bsize - mad_data->pos;
      memcpy(hack_buffer, mad_data->buffer + mad_data->pos, old_data_size);
      /* make NEW buffer */
      free(mad_data->buffer);
      mad_data->buffer = (unsigned char*) malloc (new_size * sizeof (char));
      /* copy old data */
      memcpy(mad_data->buffer, hack_buffer, old_data_size);
      /* free hack buffer */
      free(hack_buffer);
    }
  else
    {
      mad_data->buffer = (unsigned char*) malloc (new_size * sizeof (char));
    }
  /* copy in new data */
  memcpy (mad_data->buffer + old_data_size, bufin, inlength);
  mad_data->bsize = new_size;
  mad_data->pos   = 0;
  return 0;
}

/*  
 * return a string describing a MAD error
 */     
static char const *
mad_error_str (enum mad_error error)
{
  static char str[17];

  switch (error) 
    {
    case MAD_ERROR_BUFLEN:
    case MAD_ERROR_BUFPTR:
      /* these errors are handled specially and/or should not occur */
      break;

    case MAD_ERROR_NOMEM:          return "not enough memory";
    case MAD_ERROR_LOSTSYNC:       return "lost synchronization";
    case MAD_ERROR_BADLAYER:       return "reserved header layer value";
    case MAD_ERROR_BADBITRATE:     return "forbidden bitrate value";
    case MAD_ERROR_BADSAMPLERATE:  return "reserved sample frequency value";
    case MAD_ERROR_BADEMPHASIS:    return "reserved emphasis value";
    case MAD_ERROR_BADCRC:         return "CRC check failed";
    case MAD_ERROR_BADBITALLOC:    return "forbidden bit allocation value";
    case MAD_ERROR_BADSCALEFACTOR: return "bad scalefactor index";
    case MAD_ERROR_BADFRAMELEN:    return "bad frame length";
    case MAD_ERROR_BADBIGVALUES:   return "bad big_values count";
    case MAD_ERROR_BADBLOCKTYPE:   return "reserved block_type";
    case MAD_ERROR_BADSCFSI:       return "bad scalefactor selection info";
    case MAD_ERROR_BADDATAPTR:     return "bad main_data_begin pointer";
    case MAD_ERROR_BADPART3LEN:    return "bad audio data length";
    case MAD_ERROR_BADHUFFTABLE:   return "bad Huffman table select";
    case MAD_ERROR_BADHUFFDATA:    return "Huffman data overrun";
    case MAD_ERROR_BADSTEREO:      return "incompatible block_type for JS";
    }

  sprintf (str, "error 0x%04x", error);
  return str;
}

JNIEXPORT jobject JNICALL
Java_org_xoe_audio_codec_MP3Codec_openCodec0 (JNIEnv* env, jobject this, int freq, int channels)
{
  struct mad_struct * mad_data;

  /* allocate mad stuff */
  mad_data = (struct mad_struct*) malloc (sizeof (struct mad_struct));
  mad_data->synth  = (struct mad_synth *) malloc (sizeof (struct mad_synth ));
  mad_data->frame  = (struct mad_frame *) malloc (sizeof (struct mad_frame ));
  mad_data->stream = (struct mad_stream*) malloc (sizeof (struct mad_stream));
  mad_data->bsize  = 0;
  mad_data->pos    = 0;
  mad_data->buffer = 0;
  mad_data->framecount = 0;
  mad_data->leftover_pcm = 0;
  mad_data->out_freq = freq;
  mad_data->out_chan = channels;

  /* init mad stuff     */
  mad_frame_init (mad_data->frame);
  mad_stream_init (mad_data->stream);
  mad_synth_init (mad_data->synth);

  return ((jobject) mad_data);
}

JNIEXPORT jint JNICALL
Java_org_xoe_audio_codec_MP3Codec_decodeCodec0(JNIEnv* env, jobject this, 
    jarray out, jint offset, jint length, jobject native, jarray in, jint inlength)
{
  int in_freq;
  int in_chan;
  int new_nsamples;
  static int total = 0;
  int len = 0;
  int rtn = 0;
  int played = 0;

  /* mad data */
  struct mad_struct * mad_data;

  /* local input and output buffers */
  jbyte*   bufin  = 0;
  jbyte*   bufout = 0;
  jboolean iscopyin;
  jboolean iscopyout;

  mad_data = (struct mad_struct *) native;

  /* get Java output buffer */
  bufout = (*env)->GetByteArrayElements (env, out, &iscopyout);

  if (inlength > 0)
    {
      bufin  = (*env)->GetByteArrayElements (env, in,  &iscopyin);
      total += inlength;
      mad_data_add (mad_data, bufin, inlength);
      (*env)->ReleaseByteArrayElements (env, in,  bufin,  JNI_ABORT);
    }

  /* write any PCM data that din't fit last time */
  if (mad_data->leftover_pcm)
    {
      rtn = write_pcm (mad_data, bufout+offset+len, length-len);
      if (rtn == -1) printf ("MP3Codec_mad: you should never see this!!\n");
      else len += rtn;
      mad_data->leftover_pcm = 0;
    }

  /* if we have no data to decode the return -1 */
  if (mad_data->bsize == 0) 
    {
      return -1;
    }

  /* point the mad stream at the local in buffer */
  mad_data->stream->buffer = mad_data->buffer + mad_data->pos;
  mad_data->stream->bufend = mad_data->buffer + mad_data->bsize;
  mad_data->stream->this_frame = mad_data->buffer + mad_data->pos;
  mad_data->stream->next_frame = mad_data->buffer + mad_data->pos;
  mad_data->stream->sync = 1;

  /*
   * Keep decoding untill we run out of data or some other
   * unrecoverable error.
   */
  while (1)
    {
      if (mad_frame_decode (mad_data->frame, mad_data->stream) == -1) 
        {
          if (!MAD_RECOVERABLE(mad_data->stream->error)) break;
          printf("MP3Codec_mad: recoverable error at frame %d: %s\n", 
                 mad_data->framecount, 
                 mad_error_str(mad_data->stream->error));
          continue;
        }
      mad_data->framecount++;
      mad_synth_frame (mad_data->synth, mad_data->frame);
      mad_stream_sync (mad_data->stream);

      /* setup resampling if required */
      in_freq = mad_data->frame->header.samplerate;
      if (MAD_NCHANNELS (&mad_data->frame->header) == 2)
        {
          in_chan = 2;
        }
      else
        {
          in_chan = 1;
        }
      if (abs(mad_data->out_freq - in_freq >
            (long) FREQ_TOLERANCE * in_freq / 100))
        {
          printf ("resampling\n");
          if (!mad_data->resampled) 
            {
              mad_data->resampled = malloc (sizeof (*mad_data->resampled));
              if (mad_data->resampled == 0) 
                {
                  printf ("MP3Codec_mad: not enough memory to allocate resampling buffer\n");
                  continue;
                }
            }

          if (   resample_init (&mad_data->resample[0], in_freq, mad_data->out_freq) == -1 
              || resample_init (&mad_data->resample[1], in_freq, mad_data->out_freq) == -1) 
            {
              printf ("MP3Codec_mad: cannot resample %u Hz to %u Hz\n", mad_data->out_freq, in_freq);
              free (mad_data->resampled);
              mad_data->resampled = 0;
              continue;
            }
        }
      else
        {
          free (mad_data->resampled);
          mad_data->resampled = 0;
        }

      mad_data->ch1 = mad_data->synth->pcm.samples[0];
      mad_data->ch2 = mad_data->synth->pcm.samples[1];
      mad_data->nsamples = mad_data->synth->pcm.length;

      /* do the resampling */
      if (mad_data->resampled) 
        {
          new_nsamples = resample_block (&mad_data->resample[0],
                                         mad_data->nsamples, 
                                         mad_data->ch1, 
                                         (*mad_data->resampled)[0]);
          mad_data->ch1 = (*mad_data->resampled)[0];
          /* resmample right channel */
          if (mad_data->out_chan == 2 && in_chan == 2)
            {
              resample_block (&mad_data->resample[1], 
                              mad_data->nsamples, 
                              mad_data->ch2, 
                              (*mad_data->resampled)[1]);
              mad_data->ch2 = (*mad_data->resampled)[1];
            }
          mad_data->nsamples = new_nsamples;
        }

      if (mad_data->out_chan == 2 
                  && in_chan == 1) { mad_data->ch2 = mad_data->ch1; }

      if (mad_data->out_chan == 1 
                  && in_chan == 2) { mad_data->ch2 = 0; }

      rtn = write_pcm (mad_data, bufout+offset+len, length-len);
      if (rtn == -1)
        {
          mad_data->leftover_pcm = 1;
          break;
        }
      else len += rtn;
    }

  /* release Java stuff */
  (*env)->ReleaseByteArrayElements (env, out, bufout, JNI_COMMIT);

  /* nothing decoded ? */
  if (len == 0) return -1;

  /* rejig the input buffer */
  played = mad_data->stream->next_frame - mad_data->stream->buffer;
  mad_data->pos += played;

  return len; 
}

JNIEXPORT void JNICALL
Java_org_xoe_audio_codec_MP3Codec_closeCodec0 (JNIEnv* env, jobject this, jobject native)
{
  /* mad data */
  struct mad_struct * mad_data;
  mad_data = (struct mad_struct *)native;

  mad_synth_finish (mad_data->synth);
  mad_frame_finish (mad_data->frame);
  mad_stream_finish (mad_data->stream);
  free (mad_data->synth);
  free (mad_data->frame);
  free (mad_data->buffer);
  free (mad_data);
}
