/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * TODO: HTTP stream buffering
 *       UDP track information from ICE cast
 *
 * $tvt: MP3SoundStream.java,v 1.14 2001/12/14 19:51:28 ben Exp $
 */ 
package org.xoe.audio;

import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.net.URL;
import java.net.URLConnection;
import java.net.Socket;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.util.Vector;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.Control;
import javax.sound.sampled.Line;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineEvent;

import org.xoe.audio.codec.MP3Codec;
import org.xoe.audio.XoeLine;
import org.xoe.core.util.Logger;
import org.xoe.core.content.Stash;

/**
 * MP3SoundStream is a SourceDataLine that takes its input
 * from an MP3 file or stream and decodes it to the mixer.
 */
public class MP3SoundStream 
  extends XoeLine
  implements SourceDataLine
{
  static final Line.Info info = new Line.Info (MP3SoundStream.class);
  URL url;
  Socket socket;
  InputStream istream;
  int bufferSize;
  byte buffer[];
  AudioFormat format;
  Comment comments[];
  int state;
  Codec codec;

  final static int CLOSED  = 0;
  final static int STOPPED = 1;
  final static int RUNNING = 2;
  final static int PAUSED  = 3;

  private final static String genres[] = 
  { 
    "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge", 
    "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B", 
    "Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska", 
    "Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient", 
    "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical", 
    "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise",
    "AlternRock", "Bass", "Soul", "Punk", "Space", "Meditative", 
    "Instrumental Pop", "Instrumental Rock", "Ethnic", "Gothic", "Darkwave", 
    "Techno-Industrial", "Electronic", "Pop-Folk", "Eurodance", "Dream",
    "Southern Rock", "Comedy", "Cult", "Gangsta", "Top 40", "Christian Rap", 
    "Pop/Funk", "Jungle", "Native American", "Cabaret", "New Wave", 
    "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi", "Tribal", 
    "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll", 
    "Hard Rock",
    // Winamp extensions follow
    "Folk", "Folk-Rock", "National Folk", "Swing", "Fast Fusion", "Bebob", 
    "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde", "Gothic Rock",
    "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock",
    "Big Band", "Chorus", "Easy Listening", "Acoustic", "Humour", "Speech",
    "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony", 
    "Booty Bass", "Primus", "Porn Groove", "Satire", "Slow Jam", "Club", 
    "Tango", "Samba", "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", 
    "Freestyle", "Duet", "Punk Rock", "Drum Solo", "Acapella", "Euro-House",
    "Dance Hall" 
  };

  /**
   * Open URL.  
   * This happens in the contructor and also everytime
   * You start or restart the stream
   */ 
  private void openURL ()
    throws IOException
  {
    if (url.getProtocol ().equals ("http"))
      {
        comments = httpConnect ();
      }
    else
      {
        if (url.getProtocol ().equals ("file"))
          {
            comments = getIDv1 (url.getFile ());
          }
        if (url.getProtocol ().equals ("stash"))
          {
            comments = getIDv1 (Stash.getPath() + "/" + url.getFile ());
          }
        URLConnection connection = url.openConnection ();
        istream = connection.getInputStream ();
      }
  }

  /**
   * Open http (shoutcast) streaming connection.
   * Send appropriate header and read response header in comment fields.
   */
  private Comment[] httpConnect ()
    throws IOException
  {
    String response  = new String ();
    String localhost = new String ();
    Vector comments  = new Vector ();
    try
      {
         localhost = InetAddress.getLocalHost ().getHostName ();
      }
    catch (UnknownHostException e) 
      { 
         Logger.log (this, "MP3SoundStream: failed to lookup localhost : " + 
                           InetAddress.getLocalHost ().toString () + " : " + 
                           e.toString ()); 
      }
  
    /* [FIXME] get Kaffe and version number from system props */
    String header = new String ("GET " + url.getFile () + " HTTP/1.0\r\n"+
                               "Host: "  + localhost + "\r\n" +
                               "User-Agent: Kaffe/1.0.6\r\n" +
                               "\r\n");

    socket  = new Socket (url.getHost (), url.getPort ());
    istream = socket.getInputStream ();
    OutputStream ostream = socket.getOutputStream ();
    ostream.write (header.getBytes ());

    /* read initial response */
    byte c;
    while ((c = (byte)istream.read ()) != -1)
      {
        if (c == '\n') break;
        if (c != '\r') response += (char)c;
      }
    if (!response.startsWith ("ICY 200"))
      {
        if (!response.startsWith ("HTTP/1.0 200 OK"))
          {
            throw new IOException ("shoutcast error: server responded: " + response);
          }
      }
    while (true)
      {
        response = new String ();
        /* read a line from the socket */
        while ((c = (byte) istream.read ()) != -1)
          {
            if (c == '\n') break;
            if (c != '\r') response += (char)c;
          }

        /* parse the new line */
        if (response.equals ("") || response.indexOf (':') == -1) break;
        String value = response.substring (response.indexOf (':') + 1);
        String name  = response.substring (0, response.indexOf (':')).intern ();
        if (name == "icy-pub" || name == "x-audiocast-public")
          {
            comments.addElement (new Comment ("PUB", value));
          }
        else if (name == "icy-url" || name == "x-audiocast-server-url")
          {
            comments.addElement (new Comment ("URL", value));
          }
        else if (name == "icy-genre" || name == "x-audiocast-genre")
          {
            comments.addElement (new Comment ("GENRE", value));
          }
        else if (name == "icy-name" || name == "x-audiocast-name")
          {
            comments.addElement (new Comment ("TITLE", value));
          }
        else if (name == "icy-notice1" || name == "x-audiocast-description")
          {
            comments.addElement (new Comment ("COMMENT1", value));
          }
        else if (name == "icy-br" || name == "x-audiocast-bitrate")
          {
            comments.addElement (new Comment ("BITRATE", value));
          }
        else if (name == "icy-notice2")
          {
            comments.addElement (new Comment ("COMMENT2", value));
          }
        else 
          {
            Logger.log (this, "MP3SoundStream: unhandled shoutcast header: " + response);
          }
      }

    /* return any headers (comments) that were found */
    Comment commentsA[] = new Comment[comments.size ()];
    for (int i=0; i<commentsA.length; i++)
      {
        commentsA[i] = (Comment)comments.elementAt (i);
      }
    return commentsA;
  }

  /**
   * Read ID3v1 tag from MP3 file and set comments appropriatly.
   */
  private Comment[] getIDv1 (String filename)
  {
    RandomAccessFile raf;

    try  
      { 
        raf = new RandomAccessFile (filename, "r"); 
      } 
    catch (FileNotFoundException ex) 
      {
        return null; 
      }

    try 
      {
        raf.seek (raf.length () - 128);
        if (raf.readByte () != 'T') return null;
        if (raf.readByte () != 'A') return null;
        if (raf.readByte () != 'G') return null;

        Vector comments = new Vector ();

        int i;
        byte buf[] = new byte[30];
        raf.readFully (buf, 0, 30);
        for (i = 0; i < 30 && buf[i] != '\0'; i++);
        if (i != 0)
          {
            comments.addElement (new Comment ("TITLE", new String (buf, 0, i)));
          }

        buf = new byte[30];
        raf.readFully (buf, 0, 30);
        for (i = 0; i < 30 && buf[i] != '\0'; i++);
        if (i != 0)
          {
            comments.addElement (new Comment ("ARTIST", new String (buf, 0, i)));
          }

        buf = new byte[30];
        raf.readFully (buf, 0, 30);
        for (i = 0; i < 30 && buf[i] != '\0'; i++);
        if (i != 0)
          {
            comments.addElement (new Comment ("ALBUM", new String (buf, 0, i)));
          }

        buf = new byte[4];
        raf.readFully (buf, 0, 4);
        for (i = 0; i < 4 && buf[i] != '\0'; i++);
        if (i != 0)
          {
            comments.addElement (new Comment ("YEAR", new String (buf, 0, i)));
          }

        buf = new byte[30];
        raf.readFully (buf, 0, 30);
        for (i = 0; i < 30 && buf[i] != '\0'; i++);
        if (i != 0)
          {
            comments.addElement (new Comment ("DESCRIPTION", new String (buf, 0, i)));
          }

        int genre = raf.readUnsignedByte ();
        if (genre <= 125)
          {
            comments.addElement (new Comment ("GENRE", genres[genre]));
          }

        if (comments.size () == 0) return null;

        Comment commentsA[] = new Comment[comments.size ()];
        for (i = 0; i < commentsA.length; i++)
          {
            commentsA[i] = (Comment) comments.elementAt (i);
          }

        return commentsA;
      } 
    catch (IOException e) 
      { 
        Logger.log (this, e); 
        return null;
      }
    finally 
      { 
        try 
          { 
            raf.close (); 
          } 
        catch (IOException e) {} 
      }
  }

  public Comment[] getComments ()
  {
    return comments;
  }

  public int getBufferSize ()
  {
    return bufferSize;
  }

  public AudioFormat getFormat ()
  {
    return format;
  }

  public long getMicrosecondPosition()
  {
    return 0;
  }
  
  public Line.Info getLineInfo () 
  {
    return info;
  }

  public void drain () 
  {
  }

  public boolean isActive () 
  {
    return state != CLOSED;
  }

  public boolean isRunning () 
  {
    return state == RUNNING;
  }

  public void flush () 
  {
  }

  /**
   * Create stream from URL.
   */
  public void open (URL url)
    throws IOException
  {
    this.url  = url;
    openURL ();
    open ();
  }

  public void open () 
  {
    codec = new MP3Codec ();
    open (new AudioFormat (44100, 16, 2, false, false));
  }

  public void open (AudioFormat format) 
  {
    open (format, 8096);
  }

  public void open (AudioFormat format, int bufferSize) 
  {
    this.format = format;
    this.bufferSize = bufferSize;
    buffer = new byte[bufferSize];
    fireEvent (SoundStreamEvent.Type.TYPE);
  }

  public float getLevel ()
  {
    return 1.0f;
  }

  public int getFramePosition ()
  {
    return 0;
  }

  public int available ()
  {
    return 0;
  }

  public void close ()
  {
    stop ();
    // fire CLOSE event here
  }

  public boolean isOpen ()
  {
    return state != CLOSED;
  }

  /**
   * Set up codec and start playback.
   */
  public void start ()
  {
    int rc = 0;
    try
      {
        codec.setInputStream (istream);
        fireEvent (SoundStreamEvent.Type.START);
        state = RUNNING;
        do
          {
            synchronized (this) 
              {
                switch (state)
                  {
                  case STOPPED:
                    codec.close ();
                    return;
                  case PAUSED:
                    try 
                      {
                        wait ();
                      }
                    catch (InterruptedException _) { }
                  }
              }
            rc = codec.read (buffer, 0, buffer.length);
            if (rc > 0) write (buffer, 0, rc);
          } 
        while (rc > 0);
      }
    catch (IOException e)
      {
        throw new Error (e.toString ());
      }
  }


  public void pause ()
  {    
    synchronized (this) 
    { 
      state = PAUSED; 
    }
  }

  public void stop ()
  {
    synchronized (this) 
    {
      state = STOPPED;
      notify ();
    }
  }

  public void play ()
  {
    synchronized (this) 
    {
      state = RUNNING;
      notify ();
    }
  }


  public void setVolume (int level)
  {    
    synchronized (this) 
    { 
      XoeControl xoeControl = (XoeControl)getControls ()[0]; 
      xoeControl.setVolume (level);     
    }
  }

  /**
   * Writes audio data to the mixer via this source data line.
   */
  public int write (byte data[], int off, int len)
  {
    Vector v;
    LineEvent event = new SoundStreamEvent (this, data, off, len);

    synchronized (this) { v = (Vector) listeners.clone (); }

    for (int i = 0; i < v.size (); i++)
      {
        ((LineListener) v.elementAt (i)).update (event);
      }

    return len;
  }

  /**
   * Sent a LineEvent to all listenters.
   */
  protected void fireEvent (LineEvent event)
  {
    Vector v;
    synchronized (this) { v = (Vector) listeners.clone (); }
    for (int i = 0; i < v.size (); i++)
      {
        ((LineListener) v.elementAt (i)).update (event);
      }
  }

  /**
   * Sent a LineEvent to all listenters.
   */
  protected void fireEvent (LineEvent.Type type)
  {
    fireEvent (new LineEvent (this, type, AudioSystem.NOT_SPECIFIED));
  }

  /**
   * main test function.
   * pass me a url. e.g. file:/tmp/test.mp3
   */
  public static void main (String[] args)
    throws Exception
  {
    if (args.length != 1)
      {
        System.err.println ("usage: " + MP3SoundStream.class.getName () + " <url>");
        System.exit (1);
      }
    System.loadLibrary ("mp3");
    XoeLineListener xoe  = new XoeLineListener ();
    MP3SoundStream stream = new MP3SoundStream ();
    stream.open (new URL (args[0]));
    stream.addLineListener (xoe);
    stream.start ();
  }
}
