package org.xoe.packagemanager;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;

import org.xoe.display.dom.DocumentViewer;
import org.xoe.servlet.XoeServlet;

public class InstalledPackageDetail
{
  private class CallbackHelper implements Callback
  {
    String m_pname;
    DocumentViewer m_dv;

    CallbackHelper (DocumentViewer dv, String pname)
    {
      m_dv = dv;
      m_pname = pname;
    }

    public void invoke (Node n, Object payload)
    {
      PackageDescription pkg = null;
      PackageDescription[] descs = PackageManager.getInstalledPackages ();
      for (int i = 0; i < descs.length; i++)
	{
	  if (descs[i].getName ().equals (m_pname))
	    {
	      pkg = descs[i];
	      break;
	    }
	}

      Document doc = null;
      if (pkg == null)
	{
	  try
	    {
	      doc = DocumentCache.fetchDocument ( new URL (Stash.getWorkingDirectory (InstalledPackageDetail.class), "index.xhtml"));
	    }
	  catch (MalformedURLException ex)
	    {
	      Logger.log (this, ex);
	    }
	}
      else
	{
	  doc = new InstalledPackageDetail (m_dv, pkg.getName (), pkg.getVersion ().toString ()).getDocument ();
	}

      m_dv.setDocument (doc);
    }
  }

  private XDocument m_doc = null;
  private DocumentViewer m_viewer = null;
  private PackageDescription m_pkg = null;
  private URL m_wd = Stash.getWorkingDirectory (InstalledPackageDetail.class);

  public InstalledPackageDetail(DocumentViewer viewer, String name, String version)
  {
    m_viewer = viewer;
    m_pkg = PackageManager.getDescription (name, version);
    try
      {
	m_doc = DocumentCache.fetchDocument ( new URL (m_wd, "details.xhtml") );
      }
    catch (MalformedURLException ex)
      {
	Logger.log (this, ex);
      }

    Hashtable params = new Hashtable();

    params.put ("title", m_pkg.getFriendlyName() + " " + m_pkg.getVersion());
    params.put ("package-uri", m_pkg.getPackageURI());
    params.put ("vendor", m_pkg.getVendor());
    params.put ("maintainerName", m_pkg.getMaintainerName ());
    params.put ("maintainerEmail", m_pkg.getMaintainerEmail ());
    params.put ("copyrightYear", m_pkg.getCopyrightYear ());
    params.put ("copyrightHolder", m_pkg.getCopyrightHolder ());

    XoeServlet.populateFields (m_doc, params);
    Node el = XPath.findFirstNode (m_doc, "//*[@id='description']");
    Node desc = m_doc.importNode (m_pkg.getXHTMLDescription (), true);
    el.appendChild (DOMUtils.cloneChildrenAsFragment (desc));

    m_doc.setCallback ("href", new Callback ()
      {
	public void invoke (Node n, Object payload)
	{
	  if (((String)payload).startsWith ("upgrade"))
	    {
	      XDocument xdoc = (XDocument)new QueryResultsPage (m_viewer, PackageManager.search (m_pkg.getUpgradeQuery ())).getDocument ();
	      xdoc.setCallback ("success", new CallbackHelper (m_viewer, m_pkg.getName ()));
	      m_viewer.setDocument (xdoc);
	    }
	  else if (((String)payload).startsWith ("uninstall"))
	    {
	      m_viewer.setDocument (new UninstallConfirm (m_viewer, m_pkg).getDocument () );
	    }
	  else
	    {
	      ((XDocument)n.getOwnerDocument()).getDocumentHolder ().getCallback ("href").invoke (n, payload);
	    }
	}
      });

  }

  public XDocument getDocument ()
  {
    return m_doc;
  }

}
