package org.xoe.packagemanager;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xoe.core.Constants;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.util.Logger;

import org.xoe.display.dom.DocumentViewer;
import org.xoe.servlet.XoeServlet;

public class InstalledPackages
{
  private final String INSTALLED_LIST = "INSTALLED_LIST";
  private XDocument m_doc = null;
  private DocumentViewer m_viewer = null;
  private URL m_wd = Stash.getWorkingDirectory (InstalledPackageDetail.class);

  public InstalledPackages(DocumentViewer viewer)
  {
    this (viewer, PackageManager.getInstalledPackages ());
  }

  public InstalledPackages(DocumentViewer viewer, PackageDescription[] descs)
  {
    m_viewer = viewer;
    try
      {
	m_doc = DocumentCache.fetchDocument ( new URL (m_wd, "show.xhtml") );
      }
    catch (MalformedURLException ex)
      {
	Logger.log (this, ex);
      }

    Element installedList = (Element) m_doc.getElementById (INSTALLED_LIST);

    Arrays.sort (descs, new Comparator () {
	public int compare (Object o1, Object o2)
	{
	  return ((PackageDescription) o1).getName (). compareTo (((PackageDescription) o2).getName ());
	}
      });

    for (int i = 0; i < descs.length; i++)
      {
	Element anchor = (Element)installedList.appendChild (m_doc.createElementNS(Constants.XHTML_NAMESPACE, "a"));
	anchor.setAttribute ("href", "installedDetail?" + 
			     "pkgname=" + URLEncoder.encode (descs[i].getName ())  + "&" + 
			     "pkgver=" +  URLEncoder.encode (descs[i].getVersion ().toString ()));
	anchor.appendChild (m_doc.createTextNode (descs[i].getName () + " " + descs[i].getVersion ()));
	installedList.appendChild (m_doc.createElementNS (Constants.XHTML_NAMESPACE, "br"));
      }

    m_doc.setCallback ("href", new Callback ()
      {
	public void invoke (Node n, Object payload)
	{
	  if (((String)payload).startsWith ("installedDetail?"))
	    {
	      String name = XoeServlet.getURLParameter ((String)payload, "pkgname");
	      String version = XoeServlet.getURLParameter ((String)payload, "pkgver");
	      m_viewer.setDocument (new InstalledPackageDetail (m_viewer, name, version).getDocument());
	    }
	  else
	    {
	      ((XDocument)n.getOwnerDocument()).getDocumentHolder ().getCallback ("href").invoke (n, payload);
	    }
	}
      });
  }

  public XDocument getDocument()
  {
    return m_doc;
  }
}
