package org.xoe.packagemanager;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.Constants;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.InstallOption;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.packages.PackageQuery;
import org.xoe.core.packages.SearchResults;
import org.xoe.core.packages.XoePackage;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;

import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IViewableDocumentBuilder;

import org.xoe.services.IPackageSelector;

public class PackageSelector
  implements IPackageSelector, IViewableDocumentBuilder, Constants
{
  private class CallbackHelper implements Callback
  {
    DocumentViewer m_dv;

    CallbackHelper (DocumentViewer dv)
    {
      m_dv = dv;
    }

    public void invoke (Node n, Object payload)
    {
      try
	{
	  m_dv.setDocument (DocumentCache.fetchDocument ( new URL (Stash.getWorkingDirectory (InstalledPackageDetail.class), "index.xhtml")));
	}
      catch (MalformedURLException ex)
	{
	  Logger.log (this, ex);
	}
    }
  }

  private XDocument m_doc = null;
  private URL m_wd = Stash.getWorkingDirectory (InstalledPackageDetail.class);

  public Document showSearchResults (String title, String summary, SearchResults results)
  {
    try
      {
	m_doc = DocumentCache.fetchDocument ( new URL (m_wd, "frame.xhtml") );
      }
    catch (MalformedURLException ex)
      {
	Logger.log (this, ex);
      }

    DocumentViewer viewer = (DocumentViewer) m_doc.getElementById ("frame");
    viewer.setDocument (new QueryResultsPage (viewer, results).getDocument ());

    return m_doc;
  }
  
  public Document upgradePackage (String title, String summary, InstallOption[] iopts)
  {
    return null;
  }

  public XDocument getViewableDocument (ContentElement content)
  {
    try
      {
	XoePackage pkg = new XoePackage (content.getInputStream (), content.getURL ().toString ());
	PackageDescription desc = pkg.getDescription ();
	desc.setSuggestedURL (content.getURL ().toString ());
	
	SearchResults res = PackageManager.search (new PackageQuery().setName (desc.getName ()).setVersion (desc.getVersion ()));

	XDocument doc = (XDocument)showSearchResults ("Install " + desc.getPackageURI () + "?", null, res);
	doc.setCallback ("success", new Callback ()
	  {
	    public void invoke (Node n, Object payload)
	    {
	      DocumentViewer dv = (DocumentViewer) m_doc.getElementById ("frame");
	      dv.setDocument (new MessageBox (dv, "Success", "Successfully installed " + payload,
					      new String [] { "OK" }, new CallbackHelper (dv)).getDocument());
	    }
	  });
	doc.setCallback ("failure", new Callback ()
	  {
	    public void invoke (Node n, Object payload)
	    {

	      DocumentViewer dv = (DocumentViewer) m_doc.getElementById ("frame");
	      dv.setDocument (new MessageBox (dv, "Failure",
					      "Failed to install "
					      + payload,
					      new String [] { "OK" }, new CallbackHelper (dv)).getDocument());
	    }
	  });

	return doc;
      }
    catch (IOException ex)
      {
	Logger.log (this, ex);
	return null;
      }
  }

  public void init (Document doc)
  {
  }

  public String[] getFunctionalities ()
  {
    return new String[] { IPackageSelector.FUNCTIONALITY, FUNCTION_CONTENT_HANDLER };
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals ( IPackageSelector.FUNCTIONALITY ))
      { 
	return new String[] { IPackageSelector.INTERFACE };
      }
    else if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { IViewableDocumentBuilder.INTERFACE });
      }

    return new String[0];
  }

  public String[] getFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { 
	  FEATURE_EXTENSION, EXTENSION_XOE,
	    FEATURE_MIMETYPE, MIMETYPE_APPLICATION_XOE 
	    });
      }
    
    return new String[0];
  }

  public String[] getInformation ()
  {
    return new String[0];
  }
}
