package org.xoe.packagemanager;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;
import java.util.Vector;
import javax.servlet.http.HttpServletRequest;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.Constants;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.InstallOption;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageException;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.packages.PackageQuery;
import org.xoe.core.packages.SearchResults;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;

import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.ViewElement;
import org.xoe.display.dom.xhtml.Select;
import org.xoe.servlet.XoeServlet;

public class QueryResultsPage
{
  private XDocument m_doc = null;
  private DocumentViewer m_viewer = null;
  private URL m_wd = Stash.getWorkingDirectory (InstalledPackageDetail.class);
  private Vector m_uniqueNewPkgs = new Vector ();
  private SearchResults m_results = null;
  private int m_selected = 0;

  private static SearchResults doSearch (HttpServletRequest req)
  {
    String keywords = req.getParameter ("keywords");
    String name = req.getParameter ("name");
    String task = req.getParameter ("task");
    String kwCaseSensitive = req.getParameter ("kwCaseSensitive");
    String kwEverywhere = req.getParameter ("kwEverywhere");
    
    PackageQuery q = new PackageQuery ();
    q.addKeywords (keywords);
    q.setName (name);
    q.setTaskMode ("true".equals (task) ? true : false);
    q.setKeywordsSearchEverywhere ("true".equals (kwEverywhere) ? true : false);
    q.setKeywordsCaseSensitive ("true".equals (kwCaseSensitive) ? true : false);
    
    return PackageManager.search (q);
  }

  public QueryResultsPage (DocumentViewer viewer, HttpServletRequest req)
  {
    this (viewer, doSearch (req));
  }


  public QueryResultsPage (DocumentViewer viewer, SearchResults results)
  {  
    m_results = results;
    m_viewer = viewer;
    try
      {
	m_doc = DocumentCache.fetchDocument ( new URL (m_wd, "queryResults.xhtml") );
      }
    catch (MalformedURLException ex)
      {
	Logger.log (this, ex);
      }

    InstallOption[] opts = m_results.getInstallablePackages ();

    for (int i = 0; i < opts.length; i++)
      {
	boolean found = false;
	for ( int j = 0; j < m_uniqueNewPkgs.size (); j++)
	  {
	    if (((InstallOption)m_uniqueNewPkgs.elementAt (j)).getPrimary ().getPackageURI ().equals (opts[i].getPrimary ().getPackageURI ()))
	      {
		found = true;
		break;
	      }
	  }
	if (!found)
	  {
	    m_uniqueNewPkgs.addElement (opts [i]);
	  }
      }

    Hashtable params = new Hashtable();
    
    if (m_uniqueNewPkgs.size () > 1)
      {
	params.put ("summary", m_uniqueNewPkgs.size () + " available packages found");
      }
    else if (m_uniqueNewPkgs.size () == 1)
      {
	params.put ("summary", "1 available package found");
      }
    else
      {
	params.put ("summary", "No available packages found");
      }

    params.put ("numInstalled", "" + m_results.getInstalledPackages ().length);
    params.put ("numUninstallable", "" + m_results.getUninstallablePackages ().length);

    XoeServlet.populateFields (m_doc, params);

    Element e = m_doc.createElementNS (null, "");

    for (int i = 0; i < m_uniqueNewPkgs.size (); i++)
      {
	PackageDescription desc = ((InstallOption)m_uniqueNewPkgs.elementAt (i)).getPrimary ();
	
	Element c = m_doc.createElementNS (Constants.XHTML_NAMESPACE, "option");
	c.setAttribute ("value", Integer.toString (i));
	c.appendChild (m_doc.createTextNode (desc.getName () + " " + desc.getVersion ()));
	e.appendChild (c);
      }

    ((Select) m_doc.getElementById ("installedList")) .watch (e.getChildNodes ());

    if (m_uniqueNewPkgs.size () > 0)
      {
	displayOpt (0);
      }

    m_doc.setCallback ("href", new Callback ()
      {
	public void invoke (Node n, Object payload)
	{
	  if (((String)payload).startsWith ("showInstalled"))
	    {
	      m_viewer.setDocument (new InstalledPackages (m_viewer, m_results.getInstalledPackages()).getDocument());
	    }
	  else if (((String)payload).startsWith ("showUninstallable"))
	    {
	      m_viewer.setDocument (new UnavailablePackages (m_viewer, m_results.getUninstallablePackages()).getDocument());
	    }
	  else if (((String)payload).startsWith ("install"))
	    {
	      InstallOption opt = (InstallOption)m_uniqueNewPkgs.elementAt (m_selected);
	      try
		{

		  PackageManager.commit (opt);
		  Logger.log (this, "Success");
		  Callback scb = m_doc.getCallback ( "success" );
		  if (scb != null)
		    {
		      scb.invoke ( m_doc, opt.getPrimary () );
		    }
		}
	      catch (PackageException ex)
		{
		  Logger.log (this, "Failure:");
		  Logger.log (this, ex);
		  Callback fcb = m_doc.getCallback ( "failure" );
		  if (fcb != null)
		    {
		      fcb.invoke ( m_doc, opt.getPrimary () );
		    }
		}
	    }
	  else
	    {
	      ((XDocument)n.getOwnerDocument()).getDocumentHolder ().getCallback ("href").invoke (n, payload);
	    }

	}
      });

    m_doc.setCallback ("onchange", new Callback ()
      {
	public void invoke (Node n, Object payload)
	{
	  System.err.println ((String)payload);
	  if (((String)payload).startsWith ("newPackageSelected?"))
	    {
	      String num = XoeServlet.getURLParameter ((String)payload, "value");
	      displayOpt (Integer.parseInt(num));
	    }
	}
      });
    /*		public void invoke (Node n, Object payload)
		{
		  PackageDescription pkgd = (PackageDescription) payload;
		  String[] buttons = {"Search Again", "search.xhtml", "", "", "Main Menu", "index.xhtml"};
		  XDocument m = message (null, "Success", null, "Successfully installed " + pkgd.getName () + " " + pkgd.getVersion (), buttons);
		  setDisplayedDocument (m);
		}
	      });
	    doc.setCallback ("failure", new Callback () 
	      {
		public void invoke (Node n, Object payload)
		{
		  PackageDescription pkgd = (PackageDescription) payload;
		  String[] buttons = {"Back to Search Results", getBackURLString (), "", "", "Main Menu", "index.xhtml"};
		  XDocument m = message (null, "Error", null, "An error ocurred while installing " + pkgd.getName () + " " + pkgd.getVersion (), buttons);
		  setDisplayedDocument(m);
		}
	      });

	  }
    */
  }

  private void displayOpt (int n)
  {
    m_selected = n;

    InstallOption opt =  ((InstallOption)m_uniqueNewPkgs.elementAt (n));
    Element optsDetails = m_doc.getElementById ("OPTS_DETAILS");

    StringBuffer strb = new StringBuffer ("<text xmlns='" + Constants.XHTML_NAMESPACE +"'>");
    strb.append ("<b>Description:</b><br />");

    Element desc = opt.getPrimary ().getXHTMLDescription ();

    if (desc != null)
      {
	NodeList nl = desc.getChildNodes ();
	for (int i = 0; i < nl.getLength (); i++)
	  {
	    switch (nl.item (i).getNodeType ())
	      {
	      case Node.ELEMENT_NODE:
		strb.append (XMLWriter.getXML (nl.item (i)));
		break;
	      case Node.TEXT_NODE:
		strb.append (((Text)nl.item (i)).getData ());
		break;
	      }
	  }
	strb.append ("<br /><br />");
      }

    for (int i=0; i < opt.getNewInstalls ().length; i++)
      {
	if (i == 0)
	  {
	    strb.append ("<b>New packages:</b> <i>");
	  }
	else
	  {
	    strb.append (", ");
	  }

	strb.append (opt.getNewInstalls ()[i].getName () + " " +
		    opt.getNewInstalls ()[i].getVersion ());
	if (i == opt.getNewInstalls ().length - 1)
	  {
	    strb.append ("</i><br />");
	  }
      }

    for (int i=0; i < opt.getReplacements ().length; i++)
      {
	if (i == 0)
	  {
	    strb.append ("<b>Required upgrades</b>: <i>");
	  }
	else
	  {
	    strb.append (", ");
	  }

	strb.append (opt.getReplacements ()[i][0].getName () + " " +
		     opt.getReplacements ()[i][0].getVersion () + "-&gt;" +
		     opt.getReplacements ()[i][1].getVersion ());
	if (i == opt.getNewInstalls ().length)
	  {
	    strb.append ("</i><br />");
	  }
      }

    for (int i=0; i < opt.getUninstalls ().length; i++)
      {
	if (i == 0)
	  {
	    strb.append ("<b>The following packages must be uninstalled</b>: <i>");
	  }
	else
	  {
	    strb.append (", ");
	  }

	strb.append (opt.getUninstalls ()[i].getName () + " " +
		     opt.getUninstalls ()[i].getVersion ());
	if (i == opt.getUninstalls ().length)
	  {
	    strb.append ("</i><br />");
	  }
      }

    strb.append ("<table width='100%'><tr><td align='center' width='50%'><a href='install'>Install</a></td><td align='center' width='50%'><a href='xoe.Back'>Back</a></td></tr></table>");
    strb.append ("</text>");

    Document newDoc = DOMlet.create (strb.toString ().toCharArray (), "AddRemoveSoftware.doCommand");
    int numChildren = newDoc.getDocumentElement ().getChildNodes ().getLength ();

    DOMUtils.removeChildNodes (optsDetails);

    for (int i = 0; i < numChildren; i++)
      {
	optsDetails.appendChild (optsDetails.getOwnerDocument ().importNode (newDoc.getDocumentElement ().getChildNodes ().item (i), true));
      }
    ((ViewElement) optsDetails).redraw (true);
  }

  public XDocument getDocument ()
  {
    return m_doc;
  }
}
