package org.xoe.packagemanager;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;

import org.w3c.dom.Node;

import org.xoe.core.Constants;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.InstallOption;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageException;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;

import org.xoe.display.dom.DocumentViewer;
import org.xoe.servlet.XoeServlet;

public class UninstallConfirm
{
  private Callback cb = new Callback ()
    {
      public void invoke (Node n, Object payload)
      {
	m_viewer.setDocument (m_index);
      }
    };
  
  private XDocument m_doc = null;
  private XDocument m_index = null;
  private DocumentViewer m_viewer = null;
  private PackageDescription m_pkg = null;
  private InstallOption m_uninstallOption = null;

  private URL m_wd = Stash.getWorkingDirectory (UninstallConfirm.class);

  public UninstallConfirm(DocumentViewer viewer, PackageDescription pkg)
  {
    m_viewer = viewer;
    m_pkg = pkg;
    try
      {
	m_index = DocumentCache.fetchDocument ( new URL (m_wd, "index.xhtml") );
	m_doc = DocumentCache.fetchDocument ( new URL (m_wd, "uninstall.xhtml") );
      }
    catch (MalformedURLException ex)
      {
	Logger.log (this, ex);
      }

    m_uninstallOption = PackageManager.getUninstallOption(m_pkg);

    PackageDescription[] descs = m_uninstallOption.getUninstalls();
    String title = "Uninstall " + pkg.getName() + " " + pkg.getVersion() + "?";
    String summary = title;
    StringBuffer strb = new StringBuffer();

    if(descs.length > 1)
      {
	summary = "Uninstall " + pkg.getName() + " " + pkg.getVersion() + " and " +
	  (descs.length - 1) + " other packages?";

	strb.append ("<html xmlns='" + Constants.XHTML_NAMESPACE +"'><b>The following packages will be uninstalled:</b><br /><br /><ul>");
	for(int i = 0; i < descs.length; i++)
	  {
	    strb.append ("<li>" + descs[i].getName() + " " + descs[i].getVersion() + "</li>");
	  }
	strb.append ("</ul>");
	strb.append ("</html>");

	DOMlet detailDoc = DOMlet.create (strb.toString ().toCharArray (), "AddRemoveSoftware.doCommand");
	m_doc.getElementById ("details").appendChild ( DOMUtils.cloneChildrenAsFragment ( m_doc.importNode ( detailDoc.getDocumentElement (), true )));
      }

    DOMUtils.setTextContent (m_doc.getElementById ("summary"), summary);
    DOMUtils.setTextContent (m_doc.getElementById ("title"), title);

    m_doc.setCallback ("onclick", new Callback ()
      {
	public void invoke (Node n, Object payload)
	{
	  if (((String)payload).startsWith ("cancel"))
	    {
	      ((XDocument)n.getOwnerDocument()).getDocumentHolder ().getCallback ("href").invoke (n, "xoe.Back");
	    }
	  else if (((String)payload).startsWith ("continue"))
	    {
	      try
		{
		  PackageManager.commit (m_uninstallOption);
		  m_viewer.setDocument (new MessageBox (m_viewer, "Success", "Successfully uninstalled " + m_uninstallOption.getPrimary ().getName (),
						    new String [] { "OK" }, cb).getDocument());

		}
	      catch (PackageException ex)
		{
		  m_viewer.setDocument (new MessageBox (m_viewer, "Failure", "Failed to uninstall " + m_uninstallOption.getPrimary ().getName (),
						    new String [] { "OK" }, cb).getDocument());
		}	
	    }
	  else
	    {
	      ((XDocument)n.getOwnerDocument()).getDocumentHolder ().getCallback ("href").invoke (n, payload);
	    }
	}
      });


  }

  public XDocument getDocument()
  {
    return m_doc;
  }
}
