/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: XoePackageManager.java,v 1.19 2001/11/06 00:08:43 erik Exp $
 */
package org.xoe.packagemanager;

import java.io.IOException;
import java.net.URLEncoder;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Hashtable;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.dom.Callback;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.Constants;
import org.xoe.core.packages.InstallOption;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageException;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.packages.PackageQuery;
import org.xoe.core.packages.SearchResults;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;

import org.xoe.services.IPackageSelector;
import org.xoe.servlet.XoeServlet;
import org.xoe.servlet.XoeServletRequest;

public class XoePackageManager 
  extends XoeServlet
{
  private Callback cb = new Callback ()
    {
      public void invoke (Node n, Object payload)
      {
	setDisplayedDocument (getRelativeDocument ("index.xhtml"));
      }
    };
  
  private final String INSTALLED_LIST = "INSTALLED_LIST";
  private final String DETAILS = "DETAILS";

  public void init (ServletConfig sc)
    throws ServletException
  {
    trackHistory ();
  }
  
  public XoePackageManager () 
    throws ServletException
  {
    super (null);
  }

  public void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    String path = req.getPathInfo ();
    Logger.log (this, "path = " + path);
    
    if (path.equals (""))
      {
	path = "index.xhtml";
      }
    XDocument doc = getRelativeDocument (path);

    if (path.equals ("show.xhtml"))
      {
	doc = new InstalledPackages (getDocumentViewer ()).getDocument ();
      }    
    else if (path.equals ("query"))
      {
	doc = new QueryResultsPage (getDocumentViewer (), req).getDocument ();
	doc.setCallback ("success", new Callback ()
	  {
	    public void invoke (Node n, Object payload)
	    {
	      getDocumentViewer().setDocument (new MessageBox (getDocumentViewer (), "Success", "Successfully installed " + payload,
						    new String [] { "OK" }, cb).getDocument());
	    }
	  });
	doc.setCallback ("failure", new Callback ()
	  {
	    public void invoke (Node n, Object payload)
	    {

	      getDocumentViewer().setDocument (new MessageBox (getDocumentViewer(), "Failure",
							       "Failed to install "
							       + payload,
							       new String [] { "OK" }, cb).getDocument());
	    }
	  });

      }
    else if (doc == null)
      {
	super.doGet (req, res);
      }
    if (doc != null)
      {
	res.getWriter (). print (doc);
      }

  /*
    String cmd = req.getParameter ("cmd");
    if (cmd == null) 
      {
	// nothing really
      }
        else if (cmd.equals ("CONTINUE_UNINSTALL"))
      {
	PackageDescription pkgd = m_uninstallOption.getPrimary ();

	try
	  {
	    PackageManager.commit (m_uninstallOption);
	    String[] buttons = {"Back to Installed Packages", 
                                "show.xhtml", "", "", 
                                "Main Menu", "index.xhtml"};
	    doc = message (req, 
                           "Success", 
                           null, 
                           "Successfully uninstalled " + pkgd.getName () + " " + pkgd.getVersion (), 
                            buttons);
	  }
	catch (PackageException ex)
	  {
	    Logger.log (this, ex);
	    String[] buttons = {"Back to Installed Packages", 
                                "show.xhtml", "", "", "Main Menu", "index.xhtml"};
	    doc = message (req, 
                           "Error", 
                           null, 
                           "An error ocurred while uninstalling " + pkgd.getName () + " " + pkgd.getVersion (), 
                           buttons);
	  }	
      }
    else if (cmd.equals ("searchUpgrade"))
      {
	InstallOption[] opts = PackageManager.search (m_currentDisplayed.getUpgradeQuery ()).getInstallablePackages ();

	IPackageSelector ps = (IPackageSelector) (ServiceLocator.findService (PACKAGE_SELECTOR_QUERY));

	doc = (XDocument) ps.upgradePackage ("Search for upgrades to \"" + m_currentDisplayed.getName () + " " +
			  	 m_currentDisplayed.getVersion () + "\"",
			 	 "There are " + opts.length + " sets of packages for this upgrade:",
				 opts);
	doc.setCallback ("success", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      String[] buttons = {"Back to Installed Packages", getBackURLString (), 
                                  "", "", "Main Menu", "index.xhtml"};
	      XDocument m = message (null, 
                                        "Success", 
                                        null, 
                                        "Successfully upgraded " + pkgd.getName () + " to version " + pkgd.getVersion (), 
                                        buttons);
	      setDisplayedDocument (m);
	    }
	  });
	doc.setCallback ("failure", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      String[] buttons = {"Back to Search Results", getBackURLString (), 
                                  "", "Main Menu", "index.xhtml"};
	      XDocument m  = message (null, 
                                         "Error", 
                                         null, 
                                         "An error ocurred while upgrading " + pkgd.getName () + " " + pkgd.getVersion (), 
                                         buttons);
	      setDisplayedDocument (m);
	    }
	  });
			 
      }
    else if (cmd.equals ("packageDetail"))
      {
	doc = populateInstalledDetail (req, req.getParameter ("pkgname"), req.getParameter ("pkgver"));
      }
    else if (cmd.equals ("UNINSTALL"))
      {
      doc = populateUninstall (req, m_currentDisplayed);
	}*/
  }

  /*
  private void updateInstalledList (XDocument doc)
  {
    Element installedList = (Element) doc.getElementById (INSTALLED_LIST);
    PackageDescription[] descs = PackageManager.getInstalledPackages ();
    Arrays.sort (descs, new Comparator () {
	public int compare (Object o1, Object o2)
	{
	  return ((PackageDescription) o1).getName (). compareTo (((PackageDescription) o2).getName ());
	}
      });
    for (int i = 0; i < descs.length; i++)
      {
	Element anchor = (Element)installedList.appendChild (doc.createElementNS(Constants.XHTML_NAMESPACE, "a"));
	anchor.setAttribute ("href", "details?cmd=packageDetail" + "&" + 
			     "pkgname=" + URLEncoder.encode (descs[i].getName ())  + "&" + 
			     "pkgver=" +  URLEncoder.encode (descs[i].getVersion ().toString ()));
	anchor.appendChild (doc.createTextNode (descs[i].getName () + " " + descs[i].getVersion ()));
	installedList.appendChild (doc.createElementNS (Constants.XHTML_NAMESPACE, "br"));
      }
  }

  private XDocument populateInstalledDetail (HttpServletRequest req, String name, String version)
  {
    PackageDescription pkg = PackageManager.getDescription (name, version);
    m_currentDisplayed = pkg;

    String title = "Package Details";
    String summary = pkg.getName () + " " + pkg.getVersion ();

    StringBuffer strb = new StringBuffer("");
    strb.append("<b xmlns='" + Constants.XHTML_NAMESPACE +"'>Description:</b><br />");

    Element desc = pkg.getXHTMLDescription();
    if (desc != null)
      {
	NodeList nl = desc.getChildNodes();
	for(int i = 0; i < nl.getLength(); i++)
	  {
            switch (nl.item(i).getNodeType())
              {
              case Node.ELEMENT_NODE:
                strb.append (XMLWriter.getXML(nl.item(i)));
                break;
              case Node.TEXT_NODE:
                strb.append (((Text)nl.item(i)).getData());
                break;
              }
          }
        strb.append ("<br /><br />");
      }

    String[] opts = {"Uninstall", "?cmd=UNINSTALL", "Search for Update", "?cmd=searchUpgrade", "Back", getBackURLString()};
    return message (req, title, summary, strb.toString (), opts);
  }

  private XDocument populateUninstall(HttpServletRequest req, PackageDescription pkg)
  {
    m_uninstallOption = PackageManager.getUninstallOption(pkg);
    PackageDescription[] descs = m_uninstallOption.getUninstalls();
    String title = "Uninstall " + pkg.getName() + " " + pkg.getVersion() + "?";
    String summary = title;
    StringBuffer strb = new StringBuffer();

    if(descs.length > 1)
      {
	summary = "Uninstall " + pkg.getName() + " " + pkg.getVersion() + " and " +
	  (descs.length - 1) + " other packages?";

	strb.append ("<html xmlns='" + Constants.XHTML_NAMESPACE +"'><b>The following packages will be uninstalled:</b><br /><br /><ul>");
	for(int i = 0; i < descs.length; i++)
	  {
	    strb.append ("<li>" + descs[i].getName() + " " + descs[i].getVersion() + "</li>");
	  }
	strb.append ("</ul>");
	strb.append ("</html>");
      }
    else
      {
      }
    DOMlet doc = DOMlet.create (strb.toString ().toCharArray (), "AddRemoveSoftware.doCommand");
    String[] opts = {"Continue", "?cmd=CONTINUE_UNINSTALL", "Cancel", "?cmd=CANCEL_UNINSTALL", "Help", "?cmd=HELP_UNINSTALL"};
    return message (req, title, summary, doc, opts);
  }

  private XDocument message (HttpServletRequest req, String title, String summary, String detail, String[] options)
  {
    DOMlet doc = DOMlet.create (("<html xmlns='" + Constants.XHTML_NAMESPACE +"'>" + detail + "</html>").toCharArray(), "AddRemoveSoftware.doCommand");
    return message (req, title, summary, doc, options);
  }

  private XDocument message (HttpServletRequest req, String title, String summary, XDocument detail, String[] options)
  {
//     Logger.log (this, XMLWriter.getXML (detail));
//     Logger.log (this, detail.getNamespaceURI ());
    Hashtable params;
    if (req != null)
      { 
	params = (Hashtable) req.getAttribute(XoeServletRequest.PARAMS);
      }
    else
      {
	params = new Hashtable();
      }
    XDocument doc = getRelativeDocument ("message.xhtml");
    if (title == null)
      {
	title = "Message";
      }
    if (summary == null)
      {
	summary = "";
      }
    params.put ("title", title);
    params.put ("summary", summary);
    
    for (Node n = detail.getDocumentElement ().getFirstChild (); n != null ; n = n.getNextSibling ())
      {
	doc.getElementById (DETAILS).appendChild (doc.importNode (n, true));
      }

    NodeList nl = XDocument.getElementsById (doc, "a"); 
    for (int i=0; i<nl.getLength () ;i++)
      {
	Element a = (Element) nl.item (i);
	a.setAttribute ("href", options[i*2+1]);
	a.getFirstChild ().setNodeValue (options[i*2]);
      }
    populateFields(doc, params);
    Logger.log (this, XMLWriter.getXML (doc));
    return (doc);
  }
  */
}

