/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: PackageQueryResponse.java,v 1.6 2001/10/26 02:48:26 erik Exp $ 
 */
package org.xoe.packages;

import java.util.Date;
import java.util.GregorianCalendar;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Enumeration;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.ExpirationVerifierException;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageDescriptionException;
import org.xoe.core.packages.PackageUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;

class PackageQueryResponse
{
  private PackageDescription[] m_descs = null;
  private Date m_expiry = null;

  // for Soap responses
  public PackageQueryResponse (Node result)
  {
    Vector v = new Vector();

    String expiresIn = XPath.findString (result, "//ns1:searchResponse/return/packageQueryResponse/expiresIn" );

    int unit = GregorianCalendar.MINUTE;
    int number = 3;

    // default to 3 minutes
    GregorianCalendar cal = new GregorianCalendar();
    cal.setTime(new Date());
    cal.add(unit, number);

    if(expiresIn != null)
      {
	try
	  {
	    cal = PackageUtils.parseExpiration(expiresIn);
	  }
	catch(ExpirationVerifierException ex)
	  { Logger.log(this, ex); }
      }

    m_expiry = cal.getTime();
    
    Node matchNode = XPath.findFirstNode (result, "//ns1:searchResponse/return/packageQueryResponse/Match");
    XDocument doc = new XDocument ("SoapPackageQueryHandler.performQuery");
    doc.appendChild (doc.importNode (matchNode, true));
    NodeList nl = doc.getDocumentElement ().getChildNodes ();	    	   
    
    for (int i = 0; i < nl.getLength (); i++)
      {
	if (((Element)nl.item (i)).getTagName ().equals ("package"))
	  {
	    try
	      {
                Document pkgDoc = new XDocument ("PackageDescription(Element)");
                pkgDoc.appendChild (pkgDoc.importNode (nl.item(i), true));
		v.addElement (new PackageDescription (pkgDoc));
	      }
	    catch (PackageDescriptionException ex)
	      {
		// bad package from the server
                Logger.log (this, "bad package from server");
	      }
	  }
      }
    m_descs = (PackageDescription[]) PackageUtils.getValues (v);
  }

  boolean isExpired()
  {
    return (new Date()).after(m_expiry);
  }

  PackageDescription[] getDescriptions()
  {
    return m_descs;
  }

  void addTo(Vector v)
  {
    for (int i =0; m_descs != null && i < m_descs.length; i++)
      {
	v.addElement(m_descs[i]);
      }
  }
}
