/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: SoapPackageServer.java,v 1.5 2001/10/26 02:48:26 erik Exp $
 */
package org.xoe.packages;

import java.io.IOException;
import java.lang.ref.SoftReference;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;
import java.util.Vector;

import org.w3c.dom.NodeList; 
import org.w3c.dom.Text; 
import org.w3c.dom.Element;

import org.xoe.core.dom.DefaultEntityResolver;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.packages.IPackageQueryHandler;
import org.xoe.core.packages.IPackageQuery;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.soap.ISoapService;
import org.xoe.soap.SoapConstants;
import org.xoe.soap.SoapService;
import org.xoe.soap.SoapServiceException;

public class SoapPackageServer 
  implements IPackageQueryHandler
{
  private PackageQueryResponseCache m_pqrc = new PackageQueryResponseCache();
  private ISoapService m_soapService;
  private String m_soapURL;

  public SoapPackageServer (String soapURL)
  {
    m_soapURL = soapURL;
  }

  private ISoapService findSoapService ()
  {
    if (m_soapService == null)
      {
	m_soapService = (ISoapService) ServiceLocator.findImplementor (ISoapService.INTERFACE);
      }
    return m_soapService;
  }

  public void performQuery (IPackageQuery in_q, Vector out_res)
  {
    PackageQueryResponse pqr = m_pqrc.get (in_q);
    if(pqr != null)
      { // found a cached result to the same XPath that isn't expired yet
	pqr.addTo(out_res);
	return;
      }

    Stack stack = new Stack();

    stack.push(in_q.toXPath());
    //stack.push( (PackageQuery)in_q );
    
    URL url = null;
    try{
      url = new URL( m_soapURL );
      
      if( findSoapService() == null )
	Logger.log( this, "soap service is null" );
      
      // PackageQueryResponse parses the soap result
      pqr = new PackageQueryResponse(findSoapService().call
				     ("urn:xoe:XoeDB", "search", stack, url,				   
				      SoapConstants.LITERAL_XML_ENCODING_STYLE_URL));

      // add it to package query response cache
      m_pqrc.put(in_q, pqr);

      // append results to the return vector
      pqr.addTo(out_res);
    }catch( MalformedURLException e ){
      e.printStackTrace();
    }catch( SoapServiceException se ){
      System.err.println( "SoapPackageQueryHandler.performQuery(): SoapServiceException caught!" );
      se.printStackTrace();
    }	
  }

  /*  public static void main (String args[])
    throws IOException
  {
    PackageQuery q = new PackageQuery().addKeywords(args[0]).setKeywordsSearchEverywhere(true);
    String url = null;

    System.err.println("Query: " + q.toXPath());
    Vector res = new Vector();
    SoapPackageQueryHandler spqh = new SoapPackageQueryHandler(args[1]);
    spqh.m_soapService = new SoapService();
    spqh.performQuery(q, res);

    System.err.println(res.size() + " results:");
    for(int i = 0; i < res.size(); i++)
      {
	System.err.println(res.elementAt(i));
      }

    System.exit(0);
    }*/
}
