#include "config.h"
#include "video_out.h"
#include "video_out_internal.h"
#include "yuv2rgb.h"

#include <fgl.h>

LIBVO_EXTERN(fgl)

#define DBG(format, arg...)  //printf( "DEBUG: " format, ## arg)
                          
static vo_info_t vo_info = 
{
  "fgl renderer",
  "fgl",
  "Transvirtual Technologies Inc.",
  ""
};


static uint32_t image_width;
static uint32_t image_height;
static uint32_t image_format;

static int depth;
static fgl_bitmap_t *bmp;
static char* imageData;

/*
 * Initialize the display driver.
 * params:
 *   width,height: image source size
 *   d_width,d_height: size of the requested window size, just a hint
 *   fullscreen: flag, 0=windowd 1=fullscreen, just a hint
 *   title: window title, if available
 *   format: fourcc of pixel format
 * returns : zero on successful initialization, non-zero on error.
 */
static uint32_t
init(uint32_t width, uint32_t height, uint32_t d_width, uint32_t d_height, uint32_t fullscreen, char *title, uint32_t format)
{
   DBG( "vo_init: %dx%d (%dx%d) flags: %d fmt: %d\n",
	width, height, d_width, d_height, fullscreen, format);

   
   image_width = width;
   image_height = height;

   fgl_init( 0, 0, 0);
   bmp = fgl_createBitmap( width, height, -1);
   
   imageData = bmp->surface;

   depth = fgl_getBitsPerPixel();
   
   if ( format == IMGFMT_YV12 ) {
      yuv2rgb_init( depth, MODE_RGB);
   }
   return 0;

}

/*
 * Query that given pixel format is supported or not.
 * params:
 *   format: fourcc of pixel format
 * returns : 1 if supported, 0 if unsupported
 */
static uint32_t
query_format(uint32_t format) {
  DBG( "vo_query_format: %d\n", format);
  switch( format ) {
  case IMGFMT_YV12:
    return 1;
  }
  
  return 0;
}

/*
 * Return driver information.
 *   returns : read-only pointer to a vo_info_t structure.
 */
static const vo_info_t*
get_info(void)
{
  return &vo_info;
}


/*
 * Display a new RGB/BGR frame of the video to the screen.
 * params:
 *   src[0] - pointer to the image
 */
static uint32_t
draw_frame(uint8_t *src[])
{
  char *d  =imageData;
  char *s  =src[0];
 
  DBG( "vo_draw_frame\n");
  //memcpy( d,s,sbpp*image_width*image_height );

  return 0;
}


/*
 * Draw a planar YUV slice to the buffer:
 * params:
 *   src[3] = source image planes (Y,U,V)
 *   stride[3] = source image planes line widths (in bytes)
 *   w,h = width*height of area to be copied (in Y pixels)
 *   x,y = position at the destination image (in Y pixels)
 */
static uint32_t
draw_slice(uint8_t *src[], int stride[], int w,int h,int x,int y)
{
 uint8_t *dst;

 dst = imageData + ( image_width * y + x ) * ( depth/8 );
 yuv2rgb( dst,src[0],src[1],src[2],w,h,image_width*( depth/8 ),
	  stride[0],stride[1] );

 DBG( "vo_draw_slice: %d,%d %dx%d\n", x, y, w, h);

 return 0;
}

/*
 * Draws OSD to the screen buffer
 */
static void
draw_osd(void)
{
}

/*
 * Blit/Flip buffer to the screen. Must be called after each frame!
 */
static void
flip_page(void)
{
  DBG( "vo_flip_page\n");
  //fgl_PClientFlushFrame( pThis);

   fgl_bitBltBmp( fgl_screenCtx, 0, 0, bmp);
}

/*
 * This func is called after every frames to handle keyboard and
 * other events. It's called in PAUSE mode too!
 */
static void check_events(void)
{
}

/*
 * Closes driver. Should restore the original state of the system.
 */
static void
uninit(void)
{
  DBG( "vo_uninit\n");

  fgl_terminate();
}





