/////////////////////////////////////////////////////////////
// Flash Plugin and Player
// Copyright (C) 1998 Olivier Debon
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include "flash.h"

#include <fgl.h>

typedef struct {
  FlashDisplay display;
  FlashHandle  handle;
} FglContext;

FglContext ctx, *pCtx=&ctx;

struct FlashInfo fi;
char *filename;
fgl_PClient_t *pThis;

long initDisplay( FlashHandle handle, int width, int height)
{
  int depth;

  pThis = fgl_PClientConnect(width, height, &depth);

  if ( pThis ) {

    pCtx->handle = handle;;

    pCtx->display.pixels = (char*)pThis->crgn->bmp.surface;
    pCtx->display.width = width;
    pCtx->display.height = height;
    pCtx->display.bpl = fgl_getScanlineLength(width) * fgl_getSlcSize();
    pCtx->display.depth = depth;
    pCtx->display.bpp = depth / 8;

    return FlashGraphicInit(pCtx->handle, &pCtx->display);
  }

  return -1;
}

void flushFrame( unsigned short noClip)
{
  
  if ( ! noClip ) {
    fgl_PClientAddClipRegion( pThis,
			      pCtx->display.clip_x,
			      pCtx->display.clip_y,
			      pCtx->display.clip_width,
			      pCtx->display.clip_height);
  }
  //fprintf( stderr, "clip: %d,%d %dx%d\n",pCtx->display.clip_x, pCtx->display.clip_y, pCtx->display.clip_width, pCtx->display.clip_height);

  fgl_PClientFlushFrame( pThis);

}

/*
 *	This file is the entry of a very simple Flash Player
 */

int
readFile(char *filename, char **buffer, long *size)
{
	FILE *in;
	char *buf;
	long length;

	in = fopen(filename,"r");
	if (in == 0) {
		perror(filename);
		return -1;
	}
	fseek(in,0,SEEK_END);
	length = ftell(in);
	rewind(in);
	buf = malloc(length);
	fread(buf,length,1,in);
	fclose(in);

	*size = length;
	*buffer = buf;

	return length;
}

void playMovie( FlashHandle handle)
{
	struct timeval wd,now;
	long cmd, delay, wakeUp;
	unsigned short  val;
        FlashEvent fe;

	while(1) {

	  if ( fgl_PClientCheckValue( pThis, PC_STOP, &val)) {
	    if ( val ){
	      fgl_PClientDisconnect( pThis);
	      pThis = NULL;
	      return;
	    }
	  }

	  cmd = FLASH_WAKEUP;
	  wakeUp = FlashExec(pCtx->handle, cmd, 0, &wd);

	  if (pCtx->display.flash_refresh) {
	    flushFrame( fgl_PClientCheckValue( pThis, PC_BOUNDS_W, &val));
	    pCtx->display.flash_refresh = 0;
	  }

	  if ( wakeUp) {
	    //usleep( (1000000L * wd.tv_sec) + wd.tv_usec);
	    /* Recompute delay */
	    gettimeofday(&now,0);
	    delay = (wd.tv_sec-now.tv_sec)*1000 + (wd.tv_usec-now.tv_usec)/1000;
fprintf( stderr, "delay: %d\n", delay);
	     usleep( delay * 1000);
	  }


	}
}

void
showUrl(char *url, char *target, void *client_data)
{
	printf("GetURL : %s\n", url);
}

void
getSwf(char *url, int level, void *client_data)
{
	FlashHandle flashHandle;
	char *buffer;
	long size;

	flashHandle = (FlashHandle) client_data;
	printf("LoadMovie: %s @ %d\n", url, level);
	if (readFile(url, &buffer, &size) > 0) {
		FlashParse(flashHandle, level, buffer, size);
	}
}

main(int argc, char **argv)
{
	char *buffer;
	long size;
	FlashHandle flashHandle;
	int status;

	if (argc < 2) {
		fprintf(stderr,"Usage : %s <file.swf>\n", argv[0]);
		exit(1);
	}

	filename = argv[1];
	if (readFile(filename, &buffer, &size) < 0) {
		exit(2);
	}

	flashHandle = FlashNew();

	// Load level 0 movie
	do {
		status = FlashParse(flashHandle, 0, buffer, size);
	} while (status & FLASH_PARSE_NEED_DATA);

	free(buffer);

	FlashGetInfo(flashHandle, &fi);

        initDisplay(flashHandle, fi.frameWidth / 20, fi.frameHeight / 20);
	FlashSoundInit(flashHandle, "/dev/dsp");
	FlashSetGetUrlMethod(flashHandle, showUrl, 0);
	FlashSetGetSwfMethod(flashHandle, getSwf, (void*)flashHandle);

	playMovie(flashHandle);

	FlashClose(flashHandle);

	exit(0);
}
