/////////////////////////////////////////////////////////////
// Flash Plugin and Player
// Copyright (C) 1998 Olivier Debon
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include "flash.h"

#include <fgl.h>

#define BUFFER_SIZE 4096
char buffer[BUFFER_SIZE];
char *data;
int fh_rd;
int total_len;

typedef struct {
  FlashDisplay display;
  FlashHandle  handle;
} FglContext;

FglContext ctx, *pCtx=&ctx;

int noClip;
struct FlashInfo fi;
fgl_PClient_t *pThis;

/**********************
 * signal handler
 **********************/
void exit_handler() 
{
//   FlashClose(pCtx->handle);
   
   if ( pThis ) {
      fgl_PClientDisconnect( pThis);
      pThis = NULL;
   }
}

void signal_handler( int sig)
{
   exit(0);
}

void register_handler() 
{
   signal( SIGHUP, signal_handler);
   signal( SIGSEGV, signal_handler);
   signal( SIGINT, signal_handler);
   atexit( exit_handler);
}

long parseNext( FlashHandle handle) {
  int len;
  long status;

  len = read( fh_rd, buffer, BUFFER_SIZE);
  if ( data == 0 ) {
    data = malloc( len);
    memcpy( data, buffer, len);
  }
  else {
    data = realloc( data, total_len + len);
    memcpy( data+total_len, buffer, len);
  }
  total_len += len;

  status = FlashParse(handle, 0, data, total_len);  
  //fprintf( stderr, "read: %d status: %x\n", len, status);
  return status;
}

void flashInitGraphics( FlashHandle handle, int width, int height)
{
  int depth;
  u16 w, h;
   
   //init from client bounds
   w = 0;
   h = 0;
   pThis = fgl_PClientConnect( &w, &h, &depth);

  fprintf( stderr, "initGraphics: %dx%d %p %dx%d\n", w, h, pThis, pThis->srv_bounds.width, pThis->srv_bounds.height);

  if ( pThis ) {

    pCtx->handle = handle;

    pCtx->display.pixels = (char*)pThis->crgn->bmp.surface;
    pCtx->display.width = w;
    pCtx->display.height = h;
    pCtx->display.bpl = fgl_getScanlineLength(w) * fgl_getSlcSize();
    pCtx->display.depth = depth;
    pCtx->display.bpp = depth / 8;

    FlashGraphicInit(pCtx->handle, &pCtx->display);
  }

}

void flashUpdateBounds() {
  u16 nw, nh;

  noClip = 1;
/*    
  nw = pThis->srv_bounds.width;
  nh = pThis->srv_bounds.height;

  fprintf( stderr, "flashUpdateBounds: %dx%d\n", nw, nh);

  fgl_PClientUpdateSize( pThis, nw, nh);
  pCtx->display.pixels = (char*)pThis->crgn->bmp.surface;
  pCtx->display.width = nw;
  pCtx->display.height = nh;
  pCtx->display.bpl = fgl_getScanlineLength(nw) * fgl_getSlcSize();

  FlashGraphicInit(pCtx->handle, &pCtx->display);
  */
}

long handleEvent( struct timeval *pwd) 
{
   u16 val;
   FlashEvent fe;
   
   if ( ! fgl_PClientCheckValue( pThis, PCS_EVENT, &val)) {
      return 0;
   }
   
   switch ( pThis->event.event ) 
     {
      case PCE_MOUSE_PRESSED:
	fe.type = FeButtonPress;                                            
	fe.x = pThis->event.p1;
	fe.y = pThis->event.p2;
	break;
      case PCE_MOUSE_RELEASED:
	fe.type = FeButtonRelease;                                            
	fe.x = pThis->event.p1;
	fe.y = pThis->event.p2;
	break;
      case PCE_MOUSE_MOVED:
      case PCE_MOUSE_DRAGGED:
	fe.type = FeMouseMove;
	fe.x = pThis->event.p1;
	fe.y = pThis->event.p2;
	break;
      default:
	return 0;
     }
   
   //fprintf( stderr, "event: %d %d,%d\n", pThis->event.event, pThis->event.p1, pThis->event.p2);
 
   return FlashExec(pCtx->handle, FLASH_EVENT, &fe, pwd);                                  
}

void flashFlushFrame()
{
  u16 val;
  
  if ( fgl_PClientCheckValue( pThis, PCS_BOUNDS, &val)) {
     flashUpdateBounds();
  }
  
  if ( ! noClip ) {
     fgl_PClientAddClipRegion( pThis,
			      pCtx->display.clip_x,
			      pCtx->display.clip_y,
			      pCtx->display.clip_width,
			      pCtx->display.clip_height);
  }
   
  if ( fgl_PClientFlushFrame( pThis) && noClip ){
     noClip = 0;
  }
   

}

void flashPlayMovie ()
{
  struct timeval wd,now;
  long wakeUp, delay, status;

  status = FLASH_PARSE_NEED_DATA;

   while ( status & FLASH_PARSE_NEED_DATA ) {
      status = parseNext(pCtx->handle);
   }

 loop:
   wakeUp = FlashExec(pCtx->handle, FLASH_WAKEUP, 0, &wd);
   wakeUp |= handleEvent( &wd);

   if (pCtx->display.flash_refresh) {
      flashFlushFrame();
   }
   
   if ( wakeUp) {
      gettimeofday(&now,0);
      delay = (wd.tv_sec-now.tv_sec)*1000 + (wd.tv_usec-now.tv_usec)/1000;
   }
   else {
      delay = 20;
   }
   
   if ( delay > 0 ) {
      usleep( delay*1000);
   }
   
   goto loop;
   
}

void
showUrl(char *url, char *target, void *client_data)
{
   fprintf( stderr, "showURL: %s %s\n", url, target);
}

void
getSwf(char *url, int level, void *client_data)
{
   fprintf( stderr, "getSwf: %s %d\n", url, level);
}

main(int argc, char **argv)
{
  FlashHandle flashHandle;
  long status;

   fprintf( stderr, "starting up flash\n");
   if ((argc < 2) || (*argv[1] != '@')) {
    fprintf(stderr,"Usage : %s <@fh>\n", argv[0]);
    exit(1);
  }

  fh_rd = atoi( argv[1] + 1);
  //fprintf( stderr, "running flash: %d\n", fh_rd);

  register_handler();
   
  flashHandle = FlashNew();

  FlashSetGetUrlMethod(flashHandle, showUrl, 0);
  FlashSetGetSwfMethod(flashHandle, getSwf, (void*)flashHandle);
  FlashSoundInit(flashHandle, "/dev/dsp");

  // Load level 0 movie
  do {
    status = parseNext( flashHandle);
 
    if ( status & FLASH_PARSE_START ) {
      FlashGetInfo(flashHandle, &fi);
      flashInitGraphics( flashHandle, fi.frameWidth/20, fi.frameHeight/20);
    }
    if ( status & FLASH_PARSE_WAKEUP ) {
       flashPlayMovie();
       break;
    }
  } while (status & FLASH_PARSE_NEED_DATA);
  

  FlashClose(flashHandle);

  exit(0);
}
