/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ISimpleSoapServer.java,v 1.5 2001/12/06 01:00:29 ben Exp $
 */ 
package org.xoe.soap.server;


import org.xoe.core.services.IService;
import java.io.IOException;
import java.util.Hashtable;


/**
 * The ISimpleSoapServer interface represents the SimpleSoapServer service.  This interface
 * provides all that is required to run a SimpleSoapServer on a XOE device.
 *
 */
public interface ISimpleSoapServer 
  extends IService
{
  public static final String FUNCTIONALITY = "simple-soap-server";
  public static final String INTERFACE = ISimpleSoapServer.class.getName();
  public static final String STANDALONE_SYNC_CONFIG_FILE = "config-sync.xml";

  /**
   * Checks for stand alone mode
   *
   * @return boolean true if in stand alone mode, otherwise false
   */
  public boolean getStandAloneMode ();

  /** 
   * Gets the directory of where the config file lives on the file system.  This only
   * returns a valid directory if running in stand alone mode.
   *
   * @return String the working directory
   */
  public String getStandAloneWorkingDir ();

  /**
   * Gets the port number the server is currently listening on
   *
   * @return int port number
   */
  public int getCurrentPort ();


  /**
   * This starts the server to listen on the given port number
   *
   * @param int port number to listen on
   * @exception SimpleSoapServerException thrown if the server failed to start
   * @exception IOException
   */
  public void listen (int port) 
    throws SimpleSoapServerException, IOException;


  /**
   * Shuts down the server
   *   
   * @exception SimpleSoapServerException thrown if the server failed to stop
   */
  public void shutdown () 
    throws SimpleSoapServerException;
  

  /**
   * Checks to see if the server is curently running.
   *   
   * @return boolean true if running, false if stopped
   */
  public boolean isRunning ();


  /**
   * Use this method to add services that the Simple Soap Server has access to
   *   
   * @param String the id of the service
   * @param Object the service itself
   */
  public void registerService(String id, Object obj);

  /**
   * Use this method to lookup services based on an ID.
   *   
   * @param String the id of the service
   */
  public Object lookupService (String id)
    throws SimpleSoapServerException;

  /**
   * Accessor method to get a handle on the hashtable that stores available services
   *
   * @return Hashtable a hash of all available Soap services accessible through the Soap Server
   */
  public Hashtable getServiceTable ();

}
