/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: Main.java,v 1.7 2001/11/20 23:21:29 ben Exp $ 
 * $tvt: Main.java,v 1.7 2001/11/20 23:21:29 ben Exp $ 
 */
package org.xoe.soap.server;

import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.ContentElementFactory;
import org.xoe.core.content.Stash;
import org.xoe.core.content.StashDirectory;
import org.xoe.core.content.StashException;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.device.IScreen;

import org.xoe.display.dom.xhtml.Input;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.display.dom.ViewElement;
import org.xoe.servlet.XoeServlet;

import org.xoe.soap.ISoapService;
import org.xoe.soap.SoapService;
import org.xoe.soap.SoapServiceException;
import org.xoe.soap.SoapMessage;
import org.xoe.soap.SoapConstants;

//import org.xoe.mime.Multipart;
import org.xoe.mime.BodyPart;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.awt.Dimension;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;
import java.util.Hashtable;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;


/**
 * Lightweight Soap Server Administration Application
 */
public class Main
  extends XoeServlet
{

  // commands
  public static final String START = "START";
  public static final String BACK = "BACK";
  public static final String STOP = "STOP";

  private ISoapService soapService;  
  private Vector fetchableURLs;
  private XDocument userConfig;

  private Input portInput; 

  private ISimpleSoapServer simpleSoapServer; 

  public void init (ServletConfig config) 
    throws javax.servlet.ServletException
  {
    super.init (config); 
  }
  
  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {

    Logger.log( this, "simple-soap-server.Main.doGet(): called" );

    res.setContentType("text/html");
    String path = req.getPathInfo();

    if ( "".equals(path) )
      {
	Logger.log( this, "simple-soap-server.Main.doGet(): path is empty" );

	XDocument doc = null;
	if (findSimpleSoapServer ().isRunning ())
	  {
	    doc = getRelativeDocument ("running.xhtml");
	  }
	else
	  {
	    if (findSimpleSoapServer ().getStandAloneMode ())
	      {
		doc = getRelativeDocument ("default-standalone.xhtml");
	      }
	    else
	      {
		doc = getRelativeDocument ("default.xhtml");
	      }
	  }
	DOMUtils.expandUsingXPathAttributes (doc, getRegisteredServicesElement(), null );
	doc.expandUsingXPathAttributes();   

	res.getWriter().print(doc);
      }
    else if ("BACK".equals (path))
      {
	XDocument doc = null;
	if (findSimpleSoapServer ().getStandAloneMode ())
	  {
	    doc = getRelativeDocument ("default-standalone.xhtml");
	  }
	else
	  {
	    doc = getRelativeDocument ("default.xhtml");
	  }
	DOMUtils.expandUsingXPathAttributes (doc, getRegisteredServicesElement(), null );
	res.getWriter ().print (doc);
      }
    else if ("SETTINGS".equals (path))
      {
	displaySettings( res );
      }
    else if ("START".equals (path))
      {   	
	try
	  {
	    listen();
	    Logger.log( this, "Main.doGet(): in START, listen() called..." );
	    XDocument doc = getRelativeDocument ("running.xhtml");
	    DOMUtils.expandUsingXPathAttributes (doc, getRegisteredServicesElement(), null );
	    res.getWriter ().print (doc);
	  }
	catch( NumberFormatException nfe )
	  {	   
	    Dialog.note ("The server failed to start: <b><br/>" 
			 +  nfe.getMessage()
			 + "</b>");   
    	    
	    XDocument doc = null;
	    if (findSimpleSoapServer ().getStandAloneMode ())
	      {
		doc = getRelativeDocument ("default-standalone.xhtml");
	      }
	    else
	      {
		doc = getRelativeDocument ("default.xhtml");
	      }
	    res.getWriter ().print (doc);
	  }
	catch( IOException ioe )
	  {
	    Dialog.note ("The server failed to start: <b><br/>" 
			 +  ioe.getMessage()
			 + "</b>");      
 	    
	    XDocument doc = null;
	    if (findSimpleSoapServer ().getStandAloneMode ())
	      {
		doc = getRelativeDocument ("default-standalone.xhtml");
	      }
	    else
	      {
		doc = getRelativeDocument ("default.xhtml");
	      }
	    res.getWriter ().print (doc);
	  }
	catch( SimpleSoapServerException ssse )
	  {
	    Dialog.note ("The server failed to start: <b><br/>" 
			 +  ssse.getMessage()
			 + "</b>");      
 	    
	    XDocument doc = null;
	    if (findSimpleSoapServer ().getStandAloneMode ())
	      {
		doc = getRelativeDocument ("default-standalone.xhtml");
	      }
	    else
	      {
		doc = getRelativeDocument ("default.xhtml");
	      }
	    res.getWriter ().print (doc);
	  }
      }
    else if ("SAVE".equals (path))
      {
	try
	  {
	    Integer intObj = new Integer( this.portInput.getText().trim() );
	  }
	catch( NumberFormatException se )
	  {
	    //se.printStackTrace();
	    Dialog.note ("The Port <b><br/>" 
			 +  this.portInput.getText().trim()
			 + "</b><br/> must be a number.  Please try again.");       	    

	    displaySettings( res );
	    return;
	  }

	Logger.log( this, "SAVING SETTINGS: port = [" + this.portInput.getText() + "]" );
	XDocument configDoc = (XDocument)getLocalUserConfig().getOwnerDocument();

	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//port" ) , this.portInput.getText() );

	URL userConfigURL = new URL( "stash:/home/default/data/simple-soap-server/userConfig.xml" );
	String userConfigString = XMLWriter.getXML( configDoc );

	Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );

	ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
	Stash.put( userConfigCE );

	XDocument doc = null;
	if (findSimpleSoapServer ().getStandAloneMode ())
	  {
	    doc = getRelativeDocument ("default-standalone.xhtml");
	  }
	else
	  {
	    doc = getRelativeDocument ("default.xhtml");
	  }
	DOMUtils.expandUsingXPathAttributes (doc, getRegisteredServicesElement(), null );
	res.getWriter ().print (doc);
      }
    else if ("STOP".equals (path))
      { 
	try
	  {
	    findSimpleSoapServer ().shutdown ();
	  }
	catch (SimpleSoapServerException ssse)
	  {
	    ssse.printStackTrace ();
	  }
	
	Logger.log( this, "Simple Soap Server: STOP: displaying default.xhtml" );
      
	XDocument doc = null;
	if (findSimpleSoapServer ().getStandAloneMode ())
	  {
	    doc = getRelativeDocument ("default-standalone.xhtml");
	  }
	else
	  {
	    doc = getRelativeDocument ("default.xhtml");
	  }
	DOMUtils.expandUsingXPathAttributes (doc, getRegisteredServicesElement(), null );
	res.getWriter ().print (doc);
	Logger.log( this, "Simple Soap Server: STOP: displayed default.xhtml" );
      }
  }


  private void listen()
    throws NumberFormatException, IOException, SimpleSoapServerException
  {
    Logger.log( this, "Main.listen(): called" );

    String portString = XPath.findFirstNode( getLocalUserConfig(), "//port/text()" ).getNodeValue().trim();
  
    findSimpleSoapServer ().listen ( new Integer( portString ).intValue() );   
  }



  private ISimpleSoapServer findSimpleSoapServer ()
  {
    if (this.simpleSoapServer == null)
      {
	this.simpleSoapServer = (ISimpleSoapServer) ServiceLocator.findImplementor (ISimpleSoapServer.INTERFACE);
	if (this.simpleSoapServer == null)
          {
            this.simpleSoapServer = new SimpleSoapServer ();
            ServiceLocator.registerService (this.simpleSoapServer);
          }
      }
    return this.simpleSoapServer;
  }



  private void displaySettings (HttpServletResponse res) throws IOException
  {
    XDocument doc = getRelativeDocument ("settings.xhtml");
    Logger.log( this, "MY CONFIG: [" + XMLWriter.getXML( getLocalUserConfig() ) + "]" );
    DOMUtils.expandUsingXPathAttributes (doc, getLocalUserConfig(), null );	 
    
    doc.expandUsingXPathAttributes ();
    this.portInput = (Input)doc.getElementById("port"); 
    
    res.getWriter ().print (doc);    
  }

  private Element getRegisteredServicesElement ()
  {   
    Hashtable serviceTable = findSimpleSoapServer ().getServiceTable ();
    Enumeration e = serviceTable.elements ();

    XDocument doc = new XDocument ();
    Element docEl = doc.createElement ("RegisteredServices");
    doc.appendChild (docEl);
    
    Element portEl = doc.createElement ("port-listening");
    docEl.appendChild (portEl);
    DOMUtils.setTextContent (portEl, findSimpleSoapServer ().getCurrentPort () + "" );

    Element serviceEl = doc.createElement ("Service");
    docEl.appendChild (serviceEl);
    DOMUtils.setTextContent (serviceEl, e.nextElement ().getClass ().getName ());

    Element modeEl = doc.createElement ("mode");
    docEl.appendChild (modeEl);
    if (findSimpleSoapServer ().getStandAloneMode ())
      {
	DOMUtils.setTextContent (modeEl, "Simple Soap Servier is running in stand alone mode" );
      }
    else
      {
	DOMUtils.setTextContent (modeEl, "Simple Soap Servier is running in application mode" );
      }


    return docEl;
  }


  /**
   * This method will always attempt to retrieve user data from the stash
   * This data WILL NOT be cached, so any sync-ing of the stash will become
   * immediately apparent.  If for some reason the stash does not contain
   * user data, this method will look to the relative file config.xml for data
   *
   */
  private XElement getLocalUserConfig ()
  {
    XElement userConfig = null;	
    XDocument doc = DocumentCache.fetchDocument ("stash:/home/default/data/simple-soap-server/userConfig.xml");		
    if( doc != null )
      userConfig = (XElement)XPath.findFirstNode (doc, "config");
    
    if ( userConfig == null)
      {
	Logger.log( this, "Main.getLocalUserConfig(): fetch from stash is null.  Retrieving from package/config.xml..." );
	XElement originalConfig = (XElement)DocumentCache.fetchDocument (getRelativeSrc("config.xml")).getDocumentElement();	
	
	try
	  {
	    //now copy this info to the stash....?	    
	    URL userConfigURL = new URL( "stash:/home/default/data/simple-soap-server/userConfig.xml" );
	    String userConfigString = XMLWriter.getXML( originalConfig );
	    //Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );	    
	    ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
	    Stash.put( userConfigCE );
	    return originalConfig;
	  }
	catch( IOException ioe )
	  {
	    Logger.log( this, ioe.toString() );
	    return null;
	  }      
      }
    return userConfig;
  }
} 
