/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: SimpleSoapServer.java,v 1.5 2001/12/06 01:00:29 ben Exp $
 */ 
package org.xoe.soap.server;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.util.DOMUtils;

import java.io.IOException;
import java.util.Hashtable;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;


/**
  * Can be run in either STAND ALONE MODE or as a standard XOE application
  *
  */
public class SimpleSoapServer 
  implements ISimpleSoapServer
{

  private SimpleSoapServerListener listener;
  private Hashtable serviceTable;
  private boolean standAloneMode = false;
  private int DEFAULT_PORT = 6666;
  private int currentPort = 0;

  private static final String STANDALONE_BOOTSTRAP_FILE = "bootstrap-sync.xml";
  private static final String STANDALONE_BOOTSTRAPE_PROPERTY = "xoe.bootstrap";  

  /**
    * If set to run in STAND ALONE MODE, the server will be listening when initialized.
    *
    */
  public SimpleSoapServer ()
  {
    String bootstrapFile = System.getProperty (STANDALONE_BOOTSTRAPE_PROPERTY);
    System.err.println ("SimpleSoapServer (): xoe.bootstrap = [" + bootstrapFile + "]");
    if (bootstrapFile.endsWith (STANDALONE_BOOTSTRAP_FILE))
      {
	this.standAloneMode = true;
	System.err.println ("***********SimpleSoapServer (): STAND ALONE MODE IS TRUE!!!");

	//if it's in stand alone mode, then we should fire up the listener
	try
	  {
	    listen ( getStandAlonePortNum ());
	  }
	catch (SimpleSoapServerException ssse)
	  {
	    ssse.printStackTrace ();
	  }
	catch (IOException ioe)
	  {
	    ioe.printStackTrace ();
	  }
      }
  }
  

  /**
    * 
    * @return boolean True if running in STAND ALONE MODE.  False if it is running as a XOE application.
    */
  public boolean getStandAloneMode ()
  {
    return this.standAloneMode;
  }

  /**
    * 
    * @return int the port the server is listening on.
    */
  public int getCurrentPort ()
  {
    return this.currentPort;
  }


  /**
    * 
    * @return String the directory where the STAND ALONE BOOTSTRAP file is located
    */
  public String getStandAloneWorkingDir ()
  {
    String bootstrapFile = System.getProperty (STANDALONE_BOOTSTRAPE_PROPERTY);
    System.err.println ("SimpleSoapServer.getStandAloneConfigFile (): xoe.bootstrap = [" + bootstrapFile + "]");
    int index = bootstrapFile.indexOf (STANDALONE_BOOTSTRAP_FILE);
    String path = bootstrapFile.substring (0,index);
    System.err.println ("SimpleSoapServer.getStandAloneConfigFile (): path = [" + path + "]");
    return path;
  }

  /**
    * 
    * @return int the port the server is listening on while in STAND ALONE MODE
    */
  public int getStandAlonePortNum ()
    throws NumberFormatException
  {    
    String url = "file://" + getStandAloneWorkingDir () + STANDALONE_SYNC_CONFIG_FILE;
    DOMlet configDomlet = DOMlet.create (url);
    Element configEl = configDomlet.getDocumentElement ();
    System.err.println( "XoeSyncIOManager.init (): config element [" + XMLWriter.getXML (configEl) + "]");    

    Node node = configEl.getFirstChild ();
    while (node != null)
      {
	if (node.getNodeName ().equals ("port"))
	  {
	    String portNum = DOMUtils.getTextContent (node);
	    return new Integer (portNum).intValue ();
	  }	
	node = node.getNextSibling ();	
      }   
    return DEFAULT_PORT;
  }


  /**
   *
   * SimpleSoapServer interface methods
   */  
  public synchronized void listen (int port) 
    throws SimpleSoapServerException, IOException
  {    

    if (getStandAloneMode ())
      {
	port = getStandAlonePortNum ();
      }

    if (listener != null)
      {
	throw new SimpleSoapServerException ("Server not shut down!");
      }
    listener = new SimpleSoapServerListener( port );
    listener.start ();
    this.currentPort = port;
  }
  

  public synchronized void shutdown () 
    throws SimpleSoapServerException
  {    
    if (listener == null)
      {
	throw new SimpleSoapServerException ("Server already shut down!");
      }
    listener.stopMe ();
    listener = null;
  }


  public boolean isRunning ()
  {
    if (listener == null)
      {
	return false;
      }
    else
      {
	return true;
      }
  }

  public synchronized Hashtable getServiceTable ()
  {
    return this.serviceTable;
  }
 
  public synchronized void registerService (String urn, Object obj)
  {
    System.err.println( "SimpleSoapServer.registerService ():  key = [" + urn + "] value = [" + obj + "]");

    if (this.serviceTable == null)
      {
	this.serviceTable = new Hashtable ();
      }

    this.serviceTable.put (urn, obj);

  }

  
  public synchronized Object lookupService (String urn)
    throws SimpleSoapServerException
  {
    if (this.serviceTable == null)
      {
	throw new SimpleSoapServerException ("Hashtable is null");
      }

    Object obj = this.serviceTable.get (urn); 
    if (obj != null)
      {
	System.err.println( "SimpleSoapServer.lookupService ():  key = [" + urn + "] value = [" + obj + "]");
	return obj;
      }
    else
      {
	throw new SimpleSoapServerException ("Service for id:[" + urn + "] is null");
      }
    
  }




    /*
     * IService Interface Methods
     *
     *
     */

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities()
  {
    return new String[] { ISimpleSoapServer.FUNCTIONALITY };
  }

  public String[] getInterfaces(String func)
  {
    if( func.equals(ISimpleSoapServer.FUNCTIONALITY) )
      {
	return new String[] { ISimpleSoapServer.INTERFACE };
      }
    return new String[0];
  }

  public String[] getFeatures (String func)
  {
    if( func.equals(ISimpleSoapServer.FUNCTIONALITY) )
      {
	return new String[] { "server", "simple-soap-server" };
      }
    return new String[0];
  }

  public String[] getInformation ()
  {
    return new String[0];
  }

}
