package org.xoe.soap.server;

import org.xoe.soap.SoapMessage;
import org.xoe.soap.SoapDocument;
import org.xoe.soap.SoapConstants;
import org.xoe.soap.SoapParameterFactory;
import org.xoe.soap.SoapParameter;
import org.xoe.mime.BodyPart;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPath;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.services.ServiceLocator;

import java.net.ServerSocket;
import java.net.Socket;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import org.w3c.dom.NodeList;
import org.w3c.dom.Element;

public class SimpleSoapServerListener extends Thread
{

  ServerSocket serverSocket;  
  Socket incomingSocket;
  boolean listen = true;

  public SimpleSoapServerListener( int port )
    throws IOException
  {
    serverSocket = new ServerSocket( port ); 
    serverSocket.setSoTimeout (0); //set to infinite timeout
    System.err.println( "SimpleSoapServerListener: ServerSocket created at port [" + port + "]" );
  }



  public void run()    
  {
    try
      {
	while (listen)
	  {
	    //System.err.println ("SimpleSoapServerListener.run (): timeout = [" + serverSocket.getSoTimeout () + "]");
	    incomingSocket = serverSocket.accept();
	    //System.err.println( "SimpleSoapServerListener.run (): incomingSocket received!!!!" );
	    InputStream istream = incomingSocket.getInputStream ();
	    OutputStream ostream = incomingSocket.getOutputStream ();
	    SoapMessage soapMessage = new SoapMessage( istream );
	    	    
	    byte[] byteArray = executeSoapCommand (soapMessage);  
	    ostream.write (byteArray); 
	    
	    istream.close ();
	    ostream.close ();
	  }
	serverSocket.close ();
      }
    catch (InterruptedIOException iioe)
      {
	//iioe.printStackTrace ();
	try
	  {
	    serverSocket.close();
	  }
	catch (IOException ioe)
	  {
	    ioe.printStackTrace ();
	  }
      }
    catch (IOException ioe)
      {
	ioe.printStackTrace ();
      }
    finally
      {
	System.err.println ("SimpleSoapServerListener.run() finally called.");	
      }   
  }


  public void stopMe()
  {
    this.listen = false;
    this.interrupt ();	
  }


  private byte[] generateServerResponse (Object responseObj, String urn, String methodName)
  {
    SoapDocument soapDocument = new SoapDocument ();
    soapDocument.addMethodElement ( urn, methodName + "Response", SoapConstants.LITERAL_XML_ENCODING_STYLE_URL, "ns1");

    SoapParameterFactory paramFactory = SoapParameterFactory.getInstance (); 
    SoapParameter soapParameter = paramFactory.createSoapParameter (responseObj);
    soapParameter.generate (soapDocument);      
    
    String respString = soapDocument.toString ();
    System.err.println ("SimpleSoapServerListener.generateServerResponse ()[" + respString + "]");
    return respString.getBytes ();
  }


  private byte[] executeSoapCommand (SoapMessage soapMessage)
  {

    //System.err.println( "SimpleSoapServerListener.run (): new SoapMessage created!!!!" );
    DOMlet domlet = soapMessage.getMultipart ().toDOMlet ();   
    XPathContext soapContext = new XPathContext ();
    soapContext.definePrefix  (SoapConstants.ATTRIBUTE_PREFIX, SoapConstants.ENVELOPE_NS);

    //System.err.println( "SimpleSoapServerListener.executeSoapCommand () ATTRIBUTE_PREFIX = [" 
    //		+ SoapConstants.ATTRIBUTE_PREFIX +
    //		"]");
    //System.err.println( "SimpleSoapServerListener.executeSoapCommand () ENVELOPE_NS = [" 
    //		+ SoapConstants.ENVELOPE_NS +
    //		"]");
    XElement node = (XElement)XPath.findFirstNode (soapContext, domlet, SoapConstants.SOAP_BODY + "/*");

    //System.err.println  ("SoapMessage.executeSoapCommand (): command node = [" + XMLWriter.getXML (node) + "]");
    String methodName = node.getNodeName ();
    String urn = node.getNamespaceURI ();

    //System.err.println ("SoapMessage.executeSoapCommand (): methodName = [" + methodName + "]");
    //System.err.println ("SoapMessage.executeSoapCommand (): urn = [" + urn + "]");

    //extract parameters from soap message
    Stack paramStack = new Stack ();
    NodeList paramList = node.getChildNodes ();
    for (int i = 0; i < paramList.getLength (); i++)
      {
	XElement paramNode =  (XElement)paramList.item (i);
	//System.err.println  ("SoapMessage.executeSoapCommand (): param node = [" + XMLWriter.getXML (paramNode) + "]");
	
	String xsiURI = DOMUtils.lookupNSForPrefix (paramNode,"xsi",null);
	String paramType = paramNode.getAttributeNS (xsiURI, SoapConstants.TYPE);
	if (paramType.equals ("xsd:string") )
	  {
	    String paramValue = DOMUtils.getTextContent (paramNode);
	    //System.err.println ("SoapMessage.executeSoapCommand (): paramType = [" + paramType + "]");
	    //System.err.println ("SoapMessage.executeSoapCommand (): paramValue = [" + paramValue + "]");
	    paramStack.push( paramValue );
	  }
	else if (paramType.equals ("xsv:Vector")) 
	  {
	    NodeList vectorList = paramNode.getChildNodes ();
	    Vector paramVector = new Vector ();
	    for (int x = 0; x < vectorList.getLength (); x++)
	      {
		XElement itemNode = (XElement)vectorList.item (x);
		paramType = itemNode.getAttributeNS (xsiURI, SoapConstants.TYPE);
		if (paramType != null)
		  {
		    if (paramType.equals ("xsd:string") )
		      {
			String paramValue = DOMUtils.getTextContent (itemNode);
			//System.err.println ("SoapMessage.executeSoapCommand (): paramType = [" + paramType + "]");
			//System.err.println ("SoapMessage.executeSoapCommand (): paramValue = [" + paramValue + "]");
			paramVector.addElement (paramValue);
		      }
		    else
		      {
		        Exception e = new Exception ("Vector contains uknown element: " + paramType);
			return generateServerResponse (e, urn, methodName);
		      }
		  }
		else
		  {
		    String cid = itemNode.getAttributeNS (null, "href");
		    if (cid != null)
		      { 	    
			//System.err.println ("SoapMessage.executeSoapCommand (): attachment id = [" + cid + "]");

			Vector mimeAttachments = soapMessage.getMimeAttachments ();
			for (int z = 0; z < mimeAttachments.size (); z++)
			  {
			    String id = ((BodyPart)mimeAttachments.elementAt (z)).getHeader ("Content-ID");
			    //System.err.println ("SoapMessage.executeSoapCommand (): mime attachment = [" + id + "]");
			  }
			
			BodyPart bp = soapMessage.getMimeAttachmentByID (cid);
			if (bp != null)
			  {
			    paramVector.addElement (bp.getBody ());
			  }
			else
			  {
			    Exception e = 
			      new Exception ("Vector contains uknown element: " + XMLWriter.getXML(itemNode));
			    return generateServerResponse (e, urn, methodName);
			  }
		      }
		    else
		      {
			SimpleSoapServerException ssse = 
			  new SimpleSoapServerException ("Vector contains uknown element: " + XMLWriter.getXML(itemNode));
			return generateServerResponse (ssse.toString (), urn, methodName);
		      }
		  }

		
	      }
	    paramStack.push( paramVector );
	  }
      }

    Object obj = null;
    try
      {
	obj = ((ISimpleSoapServer) ServiceLocator.findImplementor (ISimpleSoapServer.INTERFACE)).lookupService (urn);
	//System.err.println ("SoapMessage.executeSoapCommand (): looked up object = [" + obj + "]");
      }
    catch( SimpleSoapServerException ssse )
      {
	ssse.printStackTrace ();
	return generateServerResponse (ssse.toString (), urn, methodName);
      }

    
    try
      {
	System.err.println ("SimpleSoapServerListener.executeSoapCommand () calling invokeIt ()");
	return generateServerResponse (invokeIt (obj, methodName, paramStack), urn, methodName);
      }
    catch (ClassNotFoundException cnfe)
      {
	cnfe.printStackTrace();
	return null;
      }
    catch (IllegalAccessException iae)
      {
	iae.printStackTrace();
	return null;
      }
    catch (InvocationTargetException ite)
      {
	ite.printStackTrace();
	return null;
      }
    catch (InstantiationException ie)
      {
	ie.printStackTrace();
	return null;
      }
    catch (NoSuchMethodException nsme)
      {
	nsme.printStackTrace();
	return null;
      }

  }


  private Object invokeIt (Object obj, String methodName, Stack paramStack)
    throws ClassNotFoundException, IllegalAccessException, InvocationTargetException, InstantiationException, NoSuchMethodException   
  {

    //System.err.println ("trying to invoke: [" + obj + "." + methodName + "()]");
    Class aClass = obj.getClass ();

    Class[] classArray = new Class[ paramStack.size () ];
    for (int i = 0; i < paramStack.size (); i++)
      {
	Object paramObj = paramStack.elementAt (i);
	classArray[i] = paramObj.getClass ();
	//System.err.println ("SimpleSoapServerListener.invokeIt (): param " + i + " [" + obj.toString () + "]");
      }

    //Method[] allMethods = aClass.getMethods ();
    //for(int i = 0; i < allMethods.length; i++)
    //{
    //System.err.println( "Method Name = [" + allMethods [i].getName () + "]");
    //}


    Method method = aClass.getMethod (methodName, classArray);
    //System.err.println ("SimpleSoapServerListener.invokeIt () invoking method: [" + methodName + "]");
    //System.err.println ("SimpleSoapServerListener.invokeIt () paramStack size: [" + paramStack.size () + "]");
    //System.err.println ("SimpleSoapServerListener.invokeIt () method parameter size: [" + method.getParameterTypes ().length + "]");
    return method.invoke (obj, paramStack.toArray ());
  }



}


