/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ISoapService.java,v 1.8 2001/12/06 00:01:51 ben Exp $
 */ 
package org.xoe.soap;

import java.io.IOException;
import org.xoe.core.services.IService;
import org.xoe.core.dom.XNode;

import java.util.Stack;
import java.util.Vector;
import java.net.URL;


/**
 * A XOE Service that provides a gateway to execute remote procedure calls via SOAP.
 *
 */
public interface ISoapService 
  extends IService
{
  public static final String FUNCTIONALITY = "soap";
  public static final String INTERFACE = ISoapService.class.getName();

  /**
   * The default remote procedure call.
   * @param urn 	the universal resource name of the SOAP service
   * @param methodName 	the name of the method
   * @param parameters 	parameter objects contained in a Stack object
   * @param url 	the address of the SOAP server providing the service
   *	
   * @exception SoapServiceException	An error specific to the Soap Service 
   *
   * @return: XNode
   * @author: Ben Gardella
   * @version: 1.0
   */  	
  public XNode call( String urn, String methodName, Stack parameters, URL url ) throws SoapServiceException;
 

  /**
   * The default remote procedure call with specific encoding style.  Typically used if the remote method
   * returns a DOM node.
   * 
   * @param urn         	the universal resource name of the SOAP service
   * @param methodName  	the name of the method
   * @param parameters  	parameter objects contained in a Stack object
   * @param url         	the address of the SOAP server providing the service
   * @param encodingStyleURL 	the namespace url to set specific encoding styles:  see Class SoapConstants 
   *
   * @exception SoapServiceException    An error specific to the Soap Service
   *
   * @return: XNode
   * @author: Ben Gardella
   * @version: 1.0
   */
  public XNode call( String urn, String methodName, Stack parameters, URL url, String encodingStyleURL ) throws SoapServiceException;


  /**
   * The default remote procedure call with returning Vector.
   * @param urn         the universal resource name of the SOAP service
   * @param methodName  the name of the method
   * @param parameters  parameter objects contained in a Stack object
   * @param url         the address of the SOAP server providing the service
   * @param vec 	A Vector object for returning arrays of simple objects from remote methods
   *
   * @exception SoapServiceException    An error specific to the Soap Service
   *
   * @return: XNode
   * @author: Ben Gardella
   * @version: 1.0
   */
  public XNode call( String urn, String methodName, Stack parameters, URL url, Vector vec ) throws SoapServiceException;   

}
