/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: PackageQuerySoapParameter.java,v 1.5 2001/11/05 02:52:36 colin Exp $
 */ 
package org.xoe.soap;

import java.util.Vector;

import org.xoe.core.dom.XElement;
import org.xoe.core.packages.PackageQuery;
import org.xoe.core.util.DOMUtils;


public class PackageQuerySoapParameter 
  implements SoapParameter
{
  PackageQuery paramObj;
  private static final String TYPE_PACKAGE_QUERY = "nsx:PackageQuery-object";

  public void setParameter( Object obj )
  {
    paramObj = (PackageQuery)obj;
  }


  public void generate( SoapDocument soapDocument )
  {
    XElement c1 = (XElement)soapDocument.createElementNS ( null, soapDocument.generateParamName() );
    c1.setAttribute( "xsi:type", TYPE_PACKAGE_QUERY );
    c1.setAttribute( "xmlns:nsx", soapDocument.getMethodID() + "-pq" ); 
    //need to set encoding style to standard for the PackageQuery object
    c1.setAttribute( SoapConstants.ATTRIBUTE_PREFIX + SoapConstants.ENCODING_STYLE , SoapConstants.STANDARD_ENCODING_STYLE_URL );
    
    //add keyword values from PackageQuery object
    XElement keywords = (XElement)soapDocument.createElementNS ( null, "keywords" );
    keywords.setAttribute( "xsi:type", "nsy:Vector" );
    keywords.setAttribute( "xmlns:nsy", "http://xml.apache.org/xml-soap" );     
    Vector kw = paramObj.getKeywords();
    if( kw == null )
      keywords.setAttribute( "xsi:null", "true" );
    else
      {
	System.err.println( "SoapMessage.addParameterElement(): keyword vector size=" + kw.size() );
	for( int i = 0; i < kw.size(); i++ )
	  {
	    String val = (String)kw.elementAt( i );
	    XElement item = (XElement)soapDocument.createElementNS( null, "item" );
	    item.setAttribute( "xsi:type", "xsd:string" );
	    DOMUtils.setTextContent (item, val);
	    keywords.appendChild( item );
	  }
      }
    c1.appendChild ( keywords );
    
    //add features values from PackageQuery object
    XElement features = (XElement)soapDocument.createElementNS ( null, "features" );
    features.setAttribute( "xsi:type", "nsz:Vector" );
    features.setAttribute( "xmlns:nsz", "http://xml.apache.org/xml-soap" );     
    Vector feat = paramObj.getFeatures();
    if( feat == null )
      features.setAttribute( "xsi:null", "true" );
    else
      {
	System.err.println( "SoapMessage.addParameterElement(): feature vector size=" + feat.size() );
	for( int i = 0; i < feat.size(); i++ )
	  {
	    String val = (String)feat.elementAt( i );
	    XElement item = (XElement)soapDocument.createElementNS( null, "item" );
	    item.setAttribute( "xsi:type", "xsd:string" );
	    DOMUtils.setTextContent (item, val);
	    features.appendChild( item );
	  }
      }
    c1.appendChild ( features );
    
    //add interface value from PackageQuery object
    XElement iface = (XElement)soapDocument.createElementNS ( null, "interface" );
    iface.setAttribute( "xsi:type", "xsd:string" );	
    String ifaceString = paramObj.getInterface();
    if( ifaceString == null )
      iface.setAttribute( "xsi:null", "true" );
    else
      {
	DOMUtils.setTextContent (iface, ifaceString);
      }
    c1.appendChild ( iface );
    
    
    //add functionality value from PackageQuery object
    XElement func = (XElement)soapDocument.createElementNS ( null, "functionality" );
    func.setAttribute( "xsi:type", "xsd:string" );	
    String funcString = paramObj.getFunctionality();
    if( funcString == null )
      func.setAttribute( "xsi:null", "true" );
    else
      {
	DOMUtils.setTextContent (func, funcString);
      }
    c1.appendChild ( func );
    
	//add package name value from PackageQuery object
    XElement pkgName = (XElement)soapDocument.createElementNS ( null, "pkgName" );
    pkgName.setAttribute( "xsi:type", "xsd:string" );	
    String pkgNameString = paramObj.getPkgName();
    if( pkgNameString == null )
      pkgName.setAttribute( "xsi:null", "true" );
    else
      {
	DOMUtils.setTextContent (pkgName, pkgNameString);
      }
    c1.appendChild ( pkgName );
    
    //add version value from PackageQuery object
    XElement ver = (XElement)soapDocument.createElementNS ( null, "ver" );
    ver.setAttribute( "xsi:type", "xsd:string" );	
    String verString = paramObj.getVer();
    if( verString == null )
	  ver.setAttribute( "xsi:null", "true" );
    else
      {
	DOMUtils.setTextContent (ver, verString);
      }
    c1.appendChild ( ver );
    
    //add min version value from PackageQuery object
    XElement minVer = (XElement)soapDocument.createElementNS ( null, "minVer" );
    minVer.setAttribute( "xsi:type", "xsd:string" );	
    String minVerString = paramObj.getMinVer();
    if( minVerString == null )
      minVer.setAttribute( "xsi:null", "true" );
    else
      {
	DOMUtils.setTextContent (minVer, minVerString);
      }
    c1.appendChild ( minVer );
    
    //add max version value from PackageQuery object
    XElement maxVer = (XElement)soapDocument.createElementNS ( null, "maxVer" );
    maxVer.setAttribute( "xsi:type", "xsd:string" );	
    String maxVerString = paramObj.getMaxVer();
    if( maxVerString == null )
      maxVer.setAttribute( "xsi:null", "true" );
    else
      {
	DOMUtils.setTextContent (maxVer, maxVerString);
      }
    c1.appendChild ( maxVer );
        
    soapDocument.getCurrentMethodElement().appendChild (c1);
  }

}
