/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: SoapService.java,v 1.8 2001/12/06 00:01:51 ben Exp $
 */ 
package org.xoe.soap;


import org.xoe.core.util.Logger;
import org.xoe.core.services.*;
import org.xoe.core.dom.XNode;

import java.net.*;
import java.io.*;
import java.util.Stack;
import java.util.Vector;

import org.w3c.dom.Document;

public class SoapService 
  implements ISoapService
{
  /*
   * SoapService interface methods
   */
  


  /**
   * The default remote procedure call.
   * @param urn 	the universal resource name of the SOAP service
   * @param methodName 	the name of the method
   * @param parameters 	parameter objects contained in a Stack object
   * @param url 	the address of the SOAP server providing the service
   *	
   * @exception SoapServiceException	An error specific to the Soap Service 
   *
   * @return: XNode
   * @author: Ben Gardella
   * @version: 1.0
   */
  public synchronized XNode call( String urn, String methodName, Stack parameters, URL url  ) 
    throws SoapServiceException
  {    
    SoapMessage soapMessage = new SoapMessage( urn, methodName, parameters, url );	
    return soapMessage.sendMessage();
  }
  


  /**
   * The default remote procedure call with specific encoding style.  Typically used if the remote method
   * returns a DOM node.
   * 
   * @param urn         	the universal resource name of the SOAP service
   * @param methodName  	the name of the method
   * @param parameters  	parameter objects contained in a Stack object
   * @param url         	the address of the SOAP server providing the service
   * @param encodingStyleURL 	the namespace url to set specific encoding styles:  see Class SoapConstants 
   *
   * @exception SoapServiceException    An error specific to the Soap Service
   *
   * @return: XNode
   * @author: Ben Gardella
   * @version: 1.0
   */
  public synchronized XNode call( String urn, String methodName, Stack parameters, URL url, String encodingStyleURL  ) 
    throws SoapServiceException
  {    
    SoapMessage soapMessage = new SoapMessage( urn, methodName, parameters, url, encodingStyleURL );	
    return soapMessage.sendMessage();
  }
  



  /**
   * The default remote procedure call with returning Vector.
   * @param urn         the universal resource name of the SOAP service
   * @param methodName  the name of the method
   * @param parameters  parameter objects contained in a Stack object
   * @param url         the address of the SOAP server providing the service
   * @param vec 	A Vector object for returning arrays of simple objects from remote methods
   *
   * @exception SoapServiceException    An error specific to the Soap Service
   *
   * @return: XNode
   * @author: Ben Gardella
   * @version: 1.0
   */
  public synchronized XNode call( String urn, String methodName, Stack parameters, URL url, Vector vec  ) 
    throws SoapServiceException
  {    
    SoapMessage soapMessage = new SoapMessage( urn, methodName, parameters, url, vec );	
    return soapMessage.sendMessage();
  }
    /*
     * IService Interface Methods
     *
     *
     */

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities()
  {
    return new String[] { ISoapService.FUNCTIONALITY };
  }

  public String[] getInterfaces(String func)
  {
    if( func.equals(ISoapService.FUNCTIONALITY) )
      {
	return new String[] { ISoapService.INTERFACE };
      }
    return new String[0];
  }

  public String[] getFeatures (String func)
  {
    if( func.equals(ISoapService.FUNCTIONALITY) )
      {
	return new String[] { "protocol", "soap" };
      }
    return new String[0];
  }

  public String[] getInformation ()
  {
    return new String[0];
  }

}
