package org.xoe.syncmanagement;


import java.util.Vector;
import java.util.Hashtable;
import java.util.Date;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.IOException;
import org.w3c.dom.Element;

import org.xoe.core.dom.XElement;
import org.xoe.core.dom.SimpleNodeList;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathParser;
import org.xoe.core.content.Stash;
import org.xoe.core.content.StashException;
import org.xoe.core.util.XMLWriter;

public class SyncManifest
{

  private Vector toSend;
  private Vector toFetch;
  private Hashtable fetchModDates;

  public SyncManifest( Element syncServerReport ) throws StashException, MalformedURLException, IOException
  {    
    init( syncServerReport );
  }


  private void init( Element syncServerReport ) throws StashException, MalformedURLException, IOException
  {    
    toSend = new Vector();
    toFetch = new Vector();
    fetchModDates = new Hashtable();

    //System.err.println( "SyncManifest.init() REPORT = [" + XMLWriter.getXML( syncServerReport ) + "]" );

    XPathContext context = new XPathContext();
    context.pushNode ( syncServerReport );
    XPath xpath = XPathParser.singleton.parse ( "//absent-on-server/*" );
    SimpleNodeList matches = new SimpleNodeList();
    xpath.evalToNodes (context, matches);
    for( int i = 0; i < matches.getLength(); i++ )
      {
	//Logger.log( this,"sendDateStamps(): matched element = " + BaseElement.toXML( (BaseElement)matches.item( i ) ) ); 
	String matchedURL = ((XElement)matches.item( i )).getAttribute( "url" );
	//Logger.log( this, "sendDateStamps(): matchedURL = " + matchedURL );	
	this.toSend.addElement( Stash.fetch( matchedURL ) );	    
      }
    xpath = XPathParser.singleton.parse ( "//newer/*" );
    matches = new SimpleNodeList();
    xpath.evalToNodes (context, matches);
    for( int i = 0; i < matches.getLength(); i++ )
      {
	String matchedURL = ((XElement)matches.item( i )).getAttribute( "url" );	
	this.toSend.addElement( Stash.fetch( matchedURL ) );
      }
    
    xpath = XPathParser.singleton.parse ( "//older/*" );
    matches = new SimpleNodeList();
    xpath.evalToNodes (context, matches);
    for( int i = 0; i < matches.getLength(); i++ )
      {
	String matchedURL = ((XElement)matches.item( i )).getAttribute( "url" );
	String modDate = ((XElement)matches.item( i )).getAttribute( "ModDate" );
	String syncDate = ((XElement)matches.item( i )).getAttribute( "SyncDate" );	
	Date date = new Date( syncDate );
	//System.err.println( "SyncManifest.init(): syncDate = [" + date.toString() + "]" );
	SyncManager.addSyncDate( new URL( matchedURL ), date );
	
	this.toFetch.addElement( matchedURL );
	this.fetchModDates.put( matchedURL, modDate );
      }

    xpath = XPathParser.singleton.parse ( "//absent-on-client/*" );
    matches = new SimpleNodeList();
    xpath.evalToNodes (context, matches);
    for( int i = 0; i < matches.getLength(); i++ )
      {
	String matchedURL = ((XElement)matches.item( i )).getAttribute( "url" );
	String modDate = ((XElement)matches.item( i )).getAttribute( "ModDate" );
	String syncDate = ((XElement)matches.item( i )).getAttribute( "SyncDate" );
	URL url = new URL( matchedURL );
	Date date = new Date( syncDate );
	//System.err.println( "SyncManifest.init(): syncDate = [" + date.toString() + "]" );
	SyncManager.addSyncDate( url, date );
	
	//ContentElement emptyCE = Stash.getFactory().createContentElement( url );
	this.toFetch.addElement( matchedURL );
	this.fetchModDates.put( matchedURL, modDate );
      }    

  }


  public Vector getSendList()
  {
    return this.toSend;
  }

  public Vector getFetchList()
  {
    return this.toFetch;
  }

  public Hashtable getFetchModDates()
  {
    return this.fetchModDates;
  }

}
