package org.xoe.soap.services.sync;

import java.util.Vector;
import java.util.Hashtable;
import java.io.IOException;
import org.w3c.dom.Element;
import org.w3c.dom.Document;



public class XoeSync 
{

  private Hashtable ioManagerCache;
  //private ISimpleSoapServer simpleSoapServer;
  public static final String SERVICE_ID = "urn:sync:XoeSync";

  public XoeSync ()
  {
    ioManagerCache = new Hashtable ();
  }

  private XoeSyncIOManager getIOManagerFromCache (String user)
  {
    System.err.println ("XoeSync.getIOManagerFromCache (): for user [" + user + "]");
    if (this.ioManagerCache == null)
      {
	System.err.println ("XoeSync.getIOManagerFromCache (): ioManagerCache is null!!!");	
      }
    System.err.println ("XoeSync.getIOManagerFromCache (): calling ioManagerCache.get ()");
    XoeSyncIOManager xoeSyncIOManager = (XoeSyncIOManager)this.ioManagerCache.get( user );
    //System.err.println ("XoeSync.getIOManagerFromCache (): manager = [" + xoeSyncIOManager.toString() + "]");
    return xoeSyncIOManager;
  }

  private void addIOManagerToCache (String user, XoeSyncIOManager ioManager)
  {
    this.ioManagerCache.put (user, ioManager);
  }

  private XoeSyncIOManager fetchIOManager( String user, String pwd ) throws XoeSyncConfigException 
  {
    System.err.println ("XoeSync.fetchIOManager (): called!!!"); 
    XoeSyncIOManager ioManager = getIOManagerFromCache( user );
    if( ioManager == null )
      {
	System.err.println ("XoeSync.fetchIOManager (): creating new XoeSyncIOManager.");
	ioManager = new XoeSyncIOManager( user, pwd );
	addIOManagerToCache( user, ioManager );
      }
    return ioManager;
  }


  /**
   * Public method exposed for SOAP-RPC; for incoming files that are being sent 
   * by client.  These files will be written to the server and an entry will be 
   * added/updated on the manifest.
   */
  public String sync( Vector dataHandlerVector, 
		      Vector urlManifest, 
		      Vector lastModManifest, 
		      Vector lastSyncManifest,
		      String user,
		      String pwd ) throws IOException, XoeSyncConfigException
  {    

    XoeSyncIOManager ioManager = fetchIOManager( user, pwd );

    if (ioManager.getManifest() == null)
      {
	throw new XoeSyncConfigException( "Sync Server Manifest is null! Sync Server not configured properly." );
      }

    if (ioManager.getManifest().getDocumentElement() == null)
      {
	throw new XoeSyncConfigException( "Sync Server Manifest does not have a DocumentElement! Please check manifest file." );
      }

    if (!ioManager.verify())
      {
	throw new XoeSyncConfigException( "Incorrect username/password." );
      }
    
    return ioManager.sync( dataHandlerVector,
			   urlManifest,
			   lastModManifest,
			   lastSyncManifest );
  }

  /**
   * Public method exposed for SOAP-RPC; evaluates the urls and datestamps of files 
   * that are being offered by client for syncing,  Method returns a node listing the status 
   * of these files.  A file is found to be in one of three possible states:
   * 1) "absent" from the server
   * 2) "newer" than the matching file on the server: client file should overwrite server file
   * 3) "older" that the matching file on the server: server file should overwrite client file
   *
   */
  public Element preSync( Vector urlIDs, 
			  Vector lastModified, 
			  Vector lastSynced,
			  String user,
			  String pwd ) throws XoeSyncConfigException
  {
    System.err.println ("XoeSync.preSync (): called!!!");

    //XoeSyncIOManager ioManager = new XoeSyncIOManager( user, pwd );
    XoeSyncIOManager ioManager = fetchIOManager( user, pwd );    

    if (ioManager.getManifest() == null)
      {
	System.err.println ("XoeSync.preSync (): ioManager.getManifest () is null"); 
	throw new XoeSyncConfigException( "Sync Server Manifest is null! Sync Server not configured properly." );
      }
    
    if (!ioManager.verify())
      {
	System.err.println ("XoeSync.preSync (): ioManager.verify () failed to verify");
	return ioManager.badPasswordElement;
	//throw new XoeSyncConfigException( "Incorrect username/password." );
      }

    return ioManager.preSync( urlIDs, lastModified, lastSynced );    
  }




  public Vector fetch( Vector urlManifest,
		       String user,
		       String pwd ) throws IOException, XoeSyncConfigException
  {

    //XoeSyncIOManager ioManager = new XoeSyncIOManager( user, pwd );
    XoeSyncIOManager ioManager = fetchIOManager( user, pwd );

    if( ioManager.getManifest() == null )
      throw new XoeSyncConfigException( "Sync Server Manifest is null! Sync Server not configured properly." );

    if( !ioManager.verify() )
      throw new XoeSyncConfigException( "Incorrect username/password/" );

    return ioManager.fetch( urlManifest );
  }



}



