package org.xoe.soap.services.sync;


import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.XNode;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.SimpleNodeList;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathParser;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.services.ServiceLocator;
import org.xoe.soap.server.ISimpleSoapServer;


import java.io.FileOutputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileWriter;
import java.io.File;
import java.util.ResourceBundle;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Date;
import java.net.URL;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;;

public class XoeSyncIOManager
{
    
  private static final String CLASS_NAME = "org.xoe.soap.services.sync.XoeSyncIOManager";
  private static final String SYNC_DIRECTORY = "SYNC_DIRECTORY";
  private static final String MANIFEST_XML = "manifest.xml";
  private static final String FILE_SEPERATOR = System.getProperty( "file.separator" );
  
  public static Element badPasswordElement = (new XDocument().createElement( "invalidPassword" ) );

  private DOMlet manifest;
  private String manifestPath;
  private String syncDirectory;
  private String usr;
  private String pwd;

  public XoeSyncIOManager( String user, String password ) throws XoeSyncConfigException
  {    
    init( user, password );
  }


  private void init( String user, String password ) throws XoeSyncConfigException
  {
    System.err.println( "XoeSyncIOManager.init (): user=[" + user + "] pass=[" + password + "]");
    this.usr = user;
    this.pwd = password;
    
    ISimpleSoapServer ssServer = (ISimpleSoapServer) ServiceLocator.findImplementor (ISimpleSoapServer.INTERFACE);
    if (ssServer == null)
      {
	throw new XoeSyncConfigException( "Cannot find SimpleSoapServer!" );
      }
    if (ssServer.getStandAloneMode ())  
      {
	DOMlet configDomlet = DOMlet.create ("file://" + ssServer.getStandAloneWorkingDir () + ssServer.STANDALONE_SYNC_CONFIG_FILE);
	Element configEl = configDomlet.getDocumentElement ();	
	System.err.println( "XoeSyncIOManager.init (): config element [" + XMLWriter.getXML (configEl) + "]");

	Node node = configEl.getFirstChild ();
	while (node != null)
	  {
	    if (node.getNodeName ().equals ("SyncDir"))
	      {
		String dirName = DOMUtils.getTextContent (node);		
		this.syncDirectory = ssServer.getStandAloneWorkingDir () + dirName;
	      }	
	    node = node.getNextSibling ();	
	  }   
      }
    else
      {
	try
	  {
	    // SYNC_DIRECTORY: the local file path that stores synced files
	    System.err.println( "XoeSyncIOManager.init (): getting config.xml url...");
	    String stringURL = Stash.getWorkingDirectory (Class.forName (CLASS_NAME)).toString ()+"/config.xml";
	    System.err.println( "XoeSyncIOManager.init (): getting config element");
	    XElement configEl = (XElement)DocumentCache.fetchDocument (stringURL).getDocumentElement ();
	    System.err.println( "XoeSyncIOManager.init (): getting text node");
	    this.syncDirectory = DOMUtils.getTextContent ( configEl.getFirstChild () );
	    System.err.println( "Sync Directory: " + this.syncDirectory );
	  }
	catch (ClassNotFoundException cnfe)
	  {
	    throw new XoeSyncConfigException( "Cannot find config.xml for : [" + CLASS_NAME + "]" );
	  }	
      }
    //create DOM document from this manifest file
    try
      {
	this.syncDirectory = this.syncDirectory + FILE_SEPERATOR + this.usr;
	manifestPath = this.syncDirectory + FILE_SEPERATOR + MANIFEST_XML;
	System.err.println( "XoeSyncIOManager.init (): creating: [" + manifestPath + "]" );
	File file = createFileHandle( manifestPath );
	System.err.println( "done creating: file handle" );
	if( !file.exists() )//if manifest file does not exist yet, write one
	  {
	    FileOutputStream fos = new FileOutputStream( file );
	    String tag = "<syncManifest password=\"" + this.pwd + "\" />";
	    fos.write( tag.getBytes() );
	    fos.close();
	  }

	FileInputStream fis = new FileInputStream( file );
	System.err.println( "done creating: file input stream" );
	System.err.println( "creating DOMLet: [" + manifestPath + "]" );
	this.manifest = DOMlet.create( fis, "syncManifest" );
	System.err.println( "done creating DOMLet: [" + manifestPath + "]" );
      }
    catch( IOException ioe )
      {       
	throw new XoeSyncConfigException( "problem reading manifest file: [" + manifestPath + "]" );
      }
  }

  public DOMlet getManifest()
  {
    return this.manifest;
  }



  public String sync( Vector inputStreamVector, 
		      Vector urlManifest, 
		      Vector lastModManifest, 
		      Vector lastSyncManifest ) throws IOException, XoeSyncConfigException 
  {    
    for( int i = 0; i < inputStreamVector.size(); i++ )
      {
	InputStream istream = (InputStream)inputStreamVector.elementAt( i );
	int available = istream.available ();
	byte[] byteArray = new byte [available];
	int read = istream.read (byteArray);
	while (read < available)
	  {
	    read = istream.read (byteArray, read, available-read);
	  }

	//ContentElement newCE = Stash.getFactory().createContentElement( new URL ( (String)urlManifest.elementAt (i) ), 
	//						istream,
	//						new Date ((String)lastSyncManifest.elementAt (i)) );

	//update file manifest
	SimpleNodeList matches = new SimpleNodeList();
	XPathContext context = new XPathContext();
	context.pushNode( this.manifest );
	XPath xpath = XPathParser.singleton.parse ( "//file[@url='" + (String)urlManifest.elementAt( i ) + "']" );
	xpath.evalToNodes (context, matches); 	
		
	if( matches.getLength() != 0 ) //new files
	  {	    
	    ((XElement)matches.item( 0 )).setAttribute( "ModDate", (String)lastModManifest.elementAt( i ) );  
	    ((XElement)matches.item( 0 )).setAttribute( "SyncDate", (String)lastSyncManifest.elementAt( i ) );
	  }
	else //a file that had been previously synced
	  {
	    XElement manifestElement = (XElement)this.manifest.createElementNS ( null, "file" );
	    manifestElement.setAttribute( "url", (String)urlManifest.elementAt( i ) );
	    manifestElement.setAttribute( "ModDate", (String)lastModManifest.elementAt( i ) );
	    manifestElement.setAttribute( "SyncDate", (String)lastSyncManifest.elementAt( i ) );
	       
	    this.manifest.getDocumentElement().appendChild( manifestElement );
	  }
	  
	//write data to file system
	String stashString = (String)urlManifest.elementAt( i );
	System.err.println( "*************XoeSyncIOManager.sync() file handle: [" + stashString + "]" );
	File file = createFileHandle( stashString );
	FileOutputStream fos = new FileOutputStream( file ); 
	//dh.writeTo( fos );  
	fos.write (byteArray);
	fos.close(); 
      }
    try
      {
	saveManifest();
      }
    catch( IOException ioe )
      {
	ioe.printStackTrace();
      }
    
    return "Files transfered.";
  }

  
  public Element preSync( Vector urlIDs, Vector lastModified, Vector lastSynced ) throws XoeSyncConfigException
  {
    System.err.println( "XoeSyncIOManager.preSync() start: [" + (new Date()).toString() + "]" );
    XDocument responseDocument = new XDocument();
    responseDocument.appendChild( responseDocument.createElement( "preSync" ) );
    XElement responseElement = (XElement)responseDocument.getDocumentElement();
    XElement responseAbsentOnServer = (XElement)responseDocument.createElementNS ( null, "absent-on-server" );
    XElement responseNewer = (XElement)responseDocument.createElementNS ( null, "newer" );
    XElement responseOlder = (XElement)responseDocument.createElementNS ( null, "older" );
    XElement responseAbsentOnClient = (XElement)responseDocument.createElementNS ( null, "absent-on-client" );
    responseElement.appendChild( responseAbsentOnServer );
    responseElement.appendChild( responseNewer );
    responseElement.appendChild( responseOlder );
    responseElement.appendChild( responseAbsentOnClient );
    Date now = new Date();

    Enumeration parentUrlIDs = getURLParents( urlIDs ); 
    SimpleNodeList missingEntries = getMissingEntries( parentUrlIDs );

    NodeList manifestNodeList = this.manifest.getDocumentElement().getChildNodes(); 
    Hashtable manifestHashtable = new Hashtable();
    for( int i = 0; i < manifestNodeList.getLength(); i++ )
      {
	XElement fileNode = (XElement)manifestNodeList.item( i );
	manifestHashtable.put( fileNode.getAttribute( "url" ), fileNode );
	//System.err.println( "XoeSyncIOManager.preSync(): manifestHashtable [" + fileNode.getAttribute( "url" ) + "]" );
      }
    
   

    for( int i = 0; i < urlIDs.size(); i++ )
      {
	String incomingID = (String)urlIDs.elementAt( i );
	//System.err.println( "XoeSyncIOManager.preSync(): searching for [" + incomingID + "]" );
	String incomingStamp = (String)lastModified.elementAt( i );       
	XElement match = (XElement)manifestHashtable.get( (String)urlIDs.elementAt( i ) );
	//System.err.println( "XoeSyncIOManager.preSync(): matched element [" + XMLWriter.getXML( match )  + "]" );


	if( match == null ) //new files
	  {
	    XElement fileElement = (XElement)responseDocument.createElementNS ( null, "file" );
	    fileElement.setAttribute( "url", (String)urlIDs.elementAt( i ) );
	    responseAbsentOnServer.appendChild( fileElement );
	    System.err.println( "XoeSyncIOManager.preSync(): new file [" + (String)urlIDs.elementAt( i ) + "]" );
	  }
	else //existing files
	  {
	    int entryIndex = missingEntries.indexOf( match );
	    if( entryIndex != -1 )
	      {
		//System.err.println( "XoeSyncIOManager: removing entry: [" + match.getAttribute( "url" ) + "]" );
		missingEntries.remove( entryIndex );
	      }

	    String modDate = match.getAttribute( "ModDate" );
	    Date serverModDate = new Date( modDate );
	    Date clientModDate = new Date( (String)lastModified.elementAt( i ) );	    
	    String syncDate = match.getAttribute( "SyncDate" );
	    Date serverSyncDate = new Date( syncDate );
	    Date clientSyncDate = new Date( (String)lastSynced.elementAt( i ) );

	    int result = compareDates( clientModDate, serverModDate, 
				       clientSyncDate, serverSyncDate );

	    if( result == 1 ) //if serverDate is older
	      {
		XElement fileElement = (XElement)responseDocument.createElementNS ( null, "file" );
		fileElement.setAttribute( "url", (String)urlIDs.elementAt( i ) );
		responseNewer.appendChild( fileElement );
	      }
	    else if( result == -1 ) //if clientDate is older
	      {
		XElement fileElement = (XElement)responseDocument.createElementNS ( null, "file" );
		fileElement.setAttribute( "url", (String)urlIDs.elementAt( i ) );
				
		//set the SyncDate attribute on both ends and send back the mod date
		fileElement.setAttribute( "SyncDate", now.toString() );
		fileElement.setAttribute( "ModDate", serverModDate.toString() );
		match.setAttribute( "SyncDate", now.toString() ); //changing the manifest

		responseOlder.appendChild( fileElement );
	      }
	      else if( result == 0 )
		System.err.println( "XoeSync.preSync(): unchanged file = [" + (String)urlIDs.elementAt( i ) + "]" );
	  }
      }
    System.err.println( "XoeSyncIOManager: Missing Entries filtered num: [" + missingEntries.size() + "]" );
    for( int i = 0; i < missingEntries.size(); i++ )
      {
	String modDate = ((XElement)missingEntries.item( i )).getAttribute( "ModDate" );
	((XElement)missingEntries.item( i )).setAttribute( "SyncDate", now.toString() ); //sets manifest
	
	XElement fileElement = (XElement)responseDocument.createElementNS ( null, "file" );
	fileElement.setAttribute( "url",  ((XElement)missingEntries.elementAt( i )).getAttribute( "url" ) );
	System.err.println( "XoeSyncIOManager: Missing Entry: [" + ((XElement)missingEntries.elementAt( i )).getAttribute( "url" ) + "]" );
	fileElement.setAttribute( "SyncDate", now.toString() );
	fileElement.setAttribute( "ModDate", modDate );
	responseAbsentOnClient.appendChild( fileElement );
      }

    try
      {
	saveManifest();
      }
    catch( IOException ioe )
      {
	ioe.printStackTrace();
      }
    
    System.err.println( "XoeSyncIOManager.preSync() end: [" + (new Date()).toString() + "]" );
    return (Element)responseElement;
  }


  public Vector fetch( Vector urlManifest ) throws IOException, XoeSyncConfigException
  {
    Vector res = new Vector();
    for( int i = 0; i < urlManifest.size(); i++ )
      {
	System.err.println ("XoeSync.fetch(): url = [" + (String)urlManifest.elementAt (i) + "]");
	URL url = new URL ((String)urlManifest.elementAt (i));
	//get from file system.
	File file = createFileHandle( (String)urlManifest.elementAt( i ) );	
	ContentElement ce = Stash.getFactory ().createContentElement (url, (InputStream)new FileInputStream (file));
	res.addElement (ce);
      }
    return res;
  }



  public boolean verify()
    {
      System.err.println( "XoeSyncIOManager.verify(): pwd = " + this.pwd );
      System.err.println( "XoeSyncIOManager.verify(): manifest pwd = " + this.manifest.getDocumentElement().getAttribute( "password" ) );
      if( this.pwd.equals( this.manifest.getDocumentElement().getAttribute( "password" ) ) )
	{
	  return true;
	}
      else
	{
	  return false;	  
	}
    }

  private SimpleNodeList getMissingEntries( Enumeration parentIDs )
  {
    SimpleNodeList matches = new SimpleNodeList();
    XPathContext context = new XPathContext();
    context.pushNode( this.manifest );
    //System.err.println( "XoeSyncIOManager.preSync() xpath start: [" + (new Date()).toString() + "]" );
    while( parentIDs.hasMoreElements() )
      {
	String el = (String)parentIDs.nextElement();
	XPath xpath = XPathParser.singleton.parse ( "//file[starts-with(@url,'" + el + "')]" );
	System.err.println( "XoeSyncIOManager.getMissingEntries() xpath: //file[starts-with(@url,'" + el + "')]" );
	xpath.evalToNodes( context, matches );   
      }
    //System.err.println( "XoeSyncIOManager.getMissingEntries() matched num: [" + matches.size() + "]" ); 
    return matches;
  }



  private Enumeration getURLParents( Vector urlIDs )
  {
    Hashtable parents = new Hashtable();

    for( int i = 0; i < urlIDs.size(); i++ )
      {	
	String original = (String)urlIDs.elementAt( i );
	String parent = original.substring( 0, original.lastIndexOf( '/' ) );
	while( parent.lastIndexOf( '/' ) > parent.indexOf( '/' ) )
	  {
	    original = parent;
	    parent = original.substring( 0, original.lastIndexOf( '/' ) );
	  }
	//System.err.println( "XoeSyncIOManager.getURLParents(): original = [" + original + "]" );
	//System.err.println( "XoeSyncIOManager.getURLParents(): parent = [" + parent + "]" );
	if( parents.get( parent ) == null )
	  parents.put( parent, original );					    	
      }
    return parents.keys();
  }


  /**
   * A file object will be made based on this file path and returned.
   *
   * This method can handle an incoming url string formed in the 
   * shape of "stash:/file path" as well as a plain filepath
   */
  private File createFileHandle( String fileDirString ) throws IOException
  {
    if( fileDirString.startsWith( "stash:" ) )
      {
	fileDirString = fileDirString.substring( fileDirString.indexOf( "stash:" ) + 6 );    
	fileDirString = this.syncDirectory + fileDirString;
	//System.err.println( "writing file: {" + fileDirString + "]" );
      }

    System.err.println( "XoeSyncIOManager.createFileHandle (): filePath = [" + fileDirString + "]" );
    //test and create directories as needed
    File file = new File ( fileDirString );
    File parent = new File( file.getParent() );
    parent.mkdirs();

    return file;
  }

  private void saveManifest() throws IOException
  {
    //System.err.println( XElement.toXML((XElement)manifest.getDocumentElement()) );
    FileWriter writer = new FileWriter( this.manifestPath );
    XMLWriter.write( this.manifest.getDocumentElement(), writer, true, XMLWriter.DEFAULT_ENCODING, true );
    writer.close();
    //System.err.println( XMLWriter.getXML((XElement)manifest.getDocumentElement()) );
  }


  /** 
   * Method that analyzes 4 date objects.  There are NINE possible permutations with only 4 
   * possible outcomes:
   *
   *  0 Do nothing
   *  1 Send client file to server
   * -1 Send server file to client
   *  2 Merge file and write result to both the server and the client( not supported yet )
   */
  private int compareDates( Date clientModDate, 
			    Date serverModDate,
			    Date clientSyncDate,
			    Date serverSyncDate )
  {

    //System.err.println( "XoeSyncIOManager.compareDates(): clientModDate = " + clientModDate.toString() );
    //System.err.println( "XoeSyncIOManager.compareDates(): serverModDate = " + serverModDate.toString() );
    //System.err.println( "XoeSyncIOManager.compareDates(): clientSyncDate = " + clientSyncDate.toString() );
    //System.err.println( "XoeSyncIOManager.compareDates(): serverSyncDate = " + serverSyncDate.toString() );

    int modResult = clientModDate.compareTo( serverModDate );
    int syncResult = clientSyncDate.compareTo( serverSyncDate );

    if( modResult == 0 && syncResult == 0 ) //files are synced without a doubt
      return 0;

    if( modResult > 0 && syncResult == 0 ) //client wins
      return 1;

    if( modResult == 0 && syncResult > 0 ) //client wins; 
      return 1;

    if( modResult > 0 && syncResult > 0 ) //client wins
      return 1;

    if( modResult < 0 && syncResult < 0 ) //server wins
      return -1;

    if( modResult < 0 && syncResult == 0 ) //server wins
      return -1;

    if( modResult == 0 && syncResult < 0 ) //probably should merge. do nothing for now
      return 0;

    if( modResult < 0 && syncResult > 0 ) //probably should merge. do nothing for now
      return 0;

    //probably should merge. do nothing for now; if the client has never synced, server wins
    if( modResult > 0 && syncResult < 0 )     
      {
	if( clientSyncDate.equals( new Date( "Mon Jan 01 00:00:01 PDT 1900" ) ) )
	  return -1;
	else
	  return 0;
      }
    return 0;
  }



}

