/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: Main.java,v 1.28 2001/11/16 21:09:45 ben Exp $ 
 * $tvt: Main.java,v 1.28 2001/11/16 21:09:45 ben Exp $ 
 */
package org.xoe.sync;

import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.ContentElementFactory;
import org.xoe.core.content.Stash;
import org.xoe.core.content.StashDirectory;
import org.xoe.core.content.StashException;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.device.IScreen;

import org.xoe.display.dom.xhtml.Input;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.display.dom.ViewElement;
import org.xoe.servlet.XoeServlet;

import org.xoe.soap.ISoapService;
import org.xoe.soap.SoapService;
import org.xoe.soap.SoapServiceException;
import org.xoe.soap.SoapMessage;
import org.xoe.soap.SoapConstants;

//import org.xoe.mime.Multipart;
import org.xoe.mime.BodyPart;

import org.xoe.syncmanagement.SyncManager;
import org.xoe.syncmanagement.SyncManifest;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import java.awt.Dimension;
import java.io.IOException;
import java.io.File;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Date;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Stack;
import java.util.Hashtable;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;


/**
 * Sync App
 * PocketLinux application for syncing userdata to a server.
 */
public class Main
  extends XoeServlet
{

  // commands
  public static final String SYNC = "SYNC";
  public static final String BACK = "BACK";
  public static final String SYNC_PATH = "SYNC_PATH";

  private ISoapService soapService;  
  private Vector fetchableURLs;
  private XDocument userConfig;

  private String user;
  private String pwd;
  private String soapURL;

  private Input userNameInput; 
  private Input passwordInput;
  private Input soapURLInput;
  private Input syncDirInput;

  private final XPathContext syncContext = new XPathContext ();
  private int counter = 0;
  private Dialog statusBar;

  public void init (ServletConfig config) 
    throws javax.servlet.ServletException
  {
    this.statusBar = buildStatusBar();
    syncContext.definePrefix ("ns1", "urn:sync:XoeSync");
    super.init (config); 
  }
  
  
  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {

    Logger.log( this, "sync.Main.doGet(): called" );

    res.setContentType("text/html");
    String path = req.getPathInfo();

    if ( "".equals(path) )
      {
	Logger.log( this, "sync.Main.doGet(): path is empty" );

	XDocument doc = getRelativeDocument ("default.xhtml");
	doc.expandUsingXPathAttributes();   

	res.getWriter().print(doc);
      }
    else if (SYNC.equals(path))
      {	
	String syncString = XPath.findFirstNode( getLocalUserConfig(), "//syncDir/text()" ).getNodeValue().trim();
	Date now = new Date();
	Logger.log( this, "THE TIME IS: " + now.toString() );
	
	Logger.log( this, "SYNC_PATH = [" + syncString + "]" );

	try
	  {	    
	    //setPercentageNode( 1 );
	    Logger.log( this, "Status Doc = [" + XMLWriter.getXML( this.statusBar.getDocument().getDocumentElement() ) + "]" );
	    //this.statusBar.setVisible( true );
	    sync( syncString );
	    this.statusBar.setVisible( false );
	  }
	catch( SoapServiceException sse )
	  {
	    this.statusBar.setVisible( false );
	    //sse.printStackTrace();
	    Logger.log( this, sse.toString() );
	    Dialog.note ("Sync Failed!<br/>Reason: <b>" + sse.getMessage() + "</b>" );       	    
	    return;
	  }
	
	Dialog.note ("Sync Complete!"); 
	
	XDocument doc = getRelativeDocument ("default.xhtml");  
	doc.expandUsingXPathAttributes();	
	res.getWriter().print(doc);
      }
    else if ("BACK".equals (path))
      {
	XDocument doc = getRelativeDocument ("default.xhtml");
	res.getWriter ().print (doc);
      }
    else if ("SETTINGS".equals (path))
      {
	displaySettings( res );
      }
    else if ("SAVE".equals (path))
      {   
	try
	  {
	    ContentElement ce = Stash.fetch( this.syncDirInput.getText().trim() );
	    if( ce == null )
	      throw new StashException( "ContentElement not found" );
	  }
	catch( StashException se )
	  {
	    //se.printStackTrace();
	    Dialog.note ("The Sync Directory <b><br/>" 
			 +  this.syncDirInput.getText().trim()
			 + "</b><br/> does not exist.  Please select another stash directory.");       	    

	    displaySettings( res );
	    return;
	  }

	Logger.log( this, "SAVING SETTINGS: username = [" + this.userNameInput.getText() + "]" );
	XDocument configDoc = (XDocument)getLocalUserConfig().getOwnerDocument();

	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//username" ) , this.userNameInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//password" ) , this.passwordInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//SoapURL" ) , this.soapURLInput.getText() );
	DOMUtils.setTextContent ( XPath.findFirstNode ( configDoc, "//syncDir" ) , this.syncDirInput.getText() );

	URL userConfigURL = new URL( "stash:/home/default/data/sync/userConfig.xml" );
	String userConfigString = XMLWriter.getXML( configDoc );

	Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );

	ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
	Stash.put( userConfigCE );

	XDocument doc = getRelativeDocument ("default.xhtml");
	res.getWriter ().print (doc);
      }


  }

  private ISoapService findSoapService ()
  {
    if (this.soapService == null)
      {
	this.soapService = (ISoapService) ServiceLocator.findImplementor (ISoapService.INTERFACE);
	if (this.soapService == null)
          {
            this.soapService = new SoapService();
            ServiceLocator.registerService (this.soapService);
          }
      }
    return this.soapService;
  }


  /**
   * The backbone method which puts all subsequent syncing actions in motion
   *
   */
  public void sync (String syncPathString) 
    throws SoapServiceException
  {
    Logger.log( this, "sync(): syncPathString = [" + syncPathString + "]" );
  
    if( ( syncPathString == null ) || ( syncPathString.equals( "" ) ) )
      {
	throw new SoapServiceException( "syncPathString is null" );
      }    

    //step one
    Vector contentEls = collectContentElements( syncPathString );

    //presync
    SyncManifest syncManifest = sendProposal( contentEls );

    //sync
    sendFiles( syncManifest );   

    //fetch 
    fetchFiles( syncManifest );
  }

  /**
   * Extracts the URLs and Last Modified Dates from the Vector of Content Elements
   * and sends theese URLs and Dates to a server via SOAP
   *
   */
  private SyncManifest sendProposal (Vector contentEls) 
    throws SoapServiceException
  {
    Logger.log( this, "sendDateStamps(): contentEls.size() = " + contentEls.size() );
    Stack stack = new Stack();

    Vector contentIDs = new Vector();
    Vector lastModifieds = new Vector();
    Vector lastSynced = new Vector();

    for( int i = 0; i < contentEls.size(); i++ )
      {
	ContentElement ce = (ContentElement)contentEls.elementAt( i );     
	String lastM = ce.getLastModified().toString();
	if( SyncManager.getSyncDate( ce.getURL() ) != null )
	  {
	    String lastS = SyncManager.getSyncDate( ce.getURL() ).toString();
	    //Logger.log( this, "Main.sendProposal(): getSyncDate() = [" + lastS + "]" );
	    lastSynced.insertElementAt( lastS, i );
	  }
	else
	  {
	    //if this file has never been synced, we indicate that by making a very old sync date.
	    //Thus, every date compared to this one will be more recent.
	    lastSynced.insertElementAt( SyncManager.BEGINNING_OF_TIME, i );
	  }
	contentIDs.insertElementAt( ce.getURL().toString(), i ); 
	lastModifieds.insertElementAt( lastM, i );
		
	incrementPercentageNode( 10, 25, i, contentEls.size() );
      }

    stack.push( XPath.findFirstNode( getLocalUserConfig(), "//password/text()" ).getNodeValue().trim() );
    stack.push( XPath.findFirstNode( getLocalUserConfig(), "//username/text()" ).getNodeValue().trim() );
    stack.push( lastSynced );
    stack.push( lastModifieds );
    stack.push( contentIDs );
  
    try
      {
	URL url = new URL ( XPath.findFirstNode( getLocalUserConfig(), "//SoapURL/text()" ).getNodeValue().trim() );
	
	if (findSoapService() == null)
          {
	    Logger.log (this, "soap service is null");
          }
       	 
	Node result = findSoapService().call("urn:sync:XoeSync",
					"preSync",
					stack,
					url,
					SoapConstants.LITERAL_XML_ENCODING_STYLE_URL );

	if( XPath.findFirstNode ( syncContext, result, "//ns1:preSyncResponse/return/invalidPassword" ) != null )
	  {
	    System.err.println( "sendProposal(): invalid password!!!" );
	    throw new SoapServiceException( "Invalid Password" );
	  }

	Element preSyncManifest =  (Element) XPath.findFirstNode ( syncContext, result, "//ns1:preSyncResponse/return" );
	SyncManifest syncManifest = new SyncManifest (preSyncManifest);
	return syncManifest;     
      }
    catch( MalformedURLException e )
      {
	e.printStackTrace();
      }
    catch( StashException se )
      {
	se.printStackTrace();
      }
    catch( IOException ioe )
      {
	ioe.printStackTrace();
      }

    return null;
  }


  private void sendFiles (SyncManifest syncManifest) 
    throws SoapServiceException
  {
    if( syncManifest.getSendList().size() == 0 )
      {
	for( int i = 26; i < 51; i++ )
	  setPercentageNode( i );
	return;
      }

    Stack stack = new Stack();
    //Logger.log( this, "sync.Main.sync(): datastring = [" + str + "]" );

    Vector urls = new Vector();
    Vector lastModDates = new Vector();
    Vector lastSynced = new Vector();
    Date now = new Date();

    for( int i = 0; i < syncManifest.getSendList().size(); i ++ )
      {
	urls.addElement( ((ContentElement)syncManifest.getSendList().elementAt( i )).getURL().toString() ); 
	lastModDates.addElement( ((ContentElement)syncManifest.getSendList().elementAt( i )).getLastModified().toString() );
	//Date now = SyncManager.getNow();
	//System.err.println( "Main.sendFiles(): now = " + now.toString() );
	lastSynced.addElement( now.toString() );
	SyncManager.addSyncDate( (ContentElement)syncManifest.getSendList().elementAt( i ), now );

	incrementPercentageNode( 25, 50, i, syncManifest.getSendList().size() );
      }

    stack.push( XPath.findFirstNode( getLocalUserConfig(), "//password/text()" ).getNodeValue().trim() );
    stack.push( XPath.findFirstNode( getLocalUserConfig(), "//username/text()" ).getNodeValue().trim() );
    stack.push( lastSynced );
    stack.push( lastModDates );
    stack.push( urls );
    stack.push( syncManifest.getSendList() );   

    Node result;
    URL url = null;
    try
      {
	url = new URL ( XPath.findFirstNode( getLocalUserConfig(), "//SoapURL/text()" ).getNodeValue().trim() );
	
	if (findSoapService() == null)
          {
	    Logger.log( this, "soap service is null");
          }
	
	result = findSoapService ().call ("urn:sync:XoeSync",
					  "sync",
					  stack,
					  url);
	
	//Logger.log (this, XMLWriter.toXML (result));
	
	String resp =  XPath.findString ( syncContext, result, "//ns1:syncResponse/ns1:return");
	
	Logger.log (this, "SOAP Sync result: [" + resp + "]");     
      }
    catch (MalformedURLException e)
      {
        e.printStackTrace();
      }
    catch (SoapServiceException se)
      {
        se.printStackTrace();
      }	
  }


  private void fetchFiles (SyncManifest syncManifest) 
    throws SoapServiceException
  {
    Logger.log( this, "fetch() called." );
    if( syncManifest.getFetchList().size() == 0 )
      {
	for( int i = 51; i < 101; i++ )
	  setPercentageNode( i );
	return;
      }

    Logger.log( this, "fetch() fetchableURLs = " + syncManifest.getFetchList().size() );

    Vector fetchableURLs = new Vector();
    for( int i = 0; i < syncManifest.getFetchList().size(); i++ )
      {
	fetchableURLs.addElement( (String)syncManifest.getFetchList().elementAt( i ) );
	incrementPercentageNode( 50, 75, i, syncManifest.getFetchList().size() );
      }

    Stack stack = new Stack();

    stack.push( XPath.findFirstNode( getLocalUserConfig(), "//password/text()" ).getNodeValue().trim() );
    stack.push( XPath.findFirstNode( getLocalUserConfig(), "//username/text()" ).getNodeValue().trim() );
    stack.push( fetchableURLs ); 

    Vector mimeAttachments = new Vector(); //Expecting some files back. They will be put here.

    Node result;
    URL url = null;
    try
      {
	url = new URL (XPath.findFirstNode( getLocalUserConfig(), "//SoapURL/text()" ).getNodeValue().trim());
	
	if (findSoapService() == null)
	  {
	    Logger.log( this, "soap service is null" );
	  }

	result = findSoapService().call("urn:sync:XoeSync",
					"fetch",
					stack,
					url,
					mimeAttachments );
	
	//Logger.log( this, XMLWriter.getXML (result) );	
	String resp =  XPath.findString ( syncContext, result, "//ns1:fetchResponse/ns1:return");
	Logger.log( this, "SOAP fetch result: [" + resp + "]" );     

	//get mime attachments and do something with them
	for( int i = 0; i < mimeAttachments.size(); i ++ )
	  {
	    BodyPart part = (BodyPart)mimeAttachments.elementAt( i );
	    ContentElement ce = Stash.fetch( (String)fetchableURLs.elementAt( i ) );
	    Date serverSyncDate = new Date( (String)syncManifest.getFetchModDates().get( (String)fetchableURLs.elementAt( i ) ) );

	    ContentElement newCE = Stash.getFactory().createContentElement( new URL ( (String)fetchableURLs.elementAt( i ) ), 
									    (InputStream)part.getBody(),
									    serverSyncDate );

	    //do I need to set the content type on the new content element????
	    //ContentType newCT = new ContentType( ce.getType().getMimeType() );
	    //newCE.setContentType( newCT );	
	       
	    //for debugging.....................
	    //if( ce != null )
	    //{
	    //Logger.log( this, "Main.fetch(): original url: [" + ce.getURL().toString() + "]" );
	    //Logger.log( this, "Main.fetch(): original last mod: [" + ce.getLastModified().toString() + "]" );
	    //}

	    //Logger.log( this, "Main.fetch(): new url: [" + newCE.getURL().toString() + "]" );
	    //Logger.log( this, "Main.fetch(): new last mod: [" + newCE.getLastModified().toString() + "]" );


	    Stash.put( newCE );
	    incrementPercentageNode( 75, 100, i, mimeAttachments.size() );
	  }	
      }
    catch (MalformedURLException e)
      {
        e.printStackTrace ();
      }
    catch (SoapServiceException se)
      {
        se.printStackTrace ();
      }
    catch (StashException se)
      {
        se.printStackTrace ();
      }	
    catch (IOException ioe)
      {
        ioe.printStackTrace ();
      }

    fetchableURLs = null; //reset the fetch list
  }



  private Vector collectContentElements (String syncPathString)
  {
    System.err.println( "collectContentElements() start!"  );
    try
      {
	ContentElement ce = Stash.fetch( syncPathString );
	Vector allContentElements = new Vector();
	this.counter = 0;
	getStashItems( ce, allContentElements );	
	return allContentElements;
      }
    catch( StashException se )
      {
	se.printStackTrace();
	return null;
      }  

  }

  private void getStashItems (ContentElement el, Vector in)
  {
    NodeList nl = (NodeList)el.getChildNodes();   

    if( nl.getLength() == 0 && !(el instanceof StashDirectory) )
      {
	in.addElement( el );
      }

      for (int i = 0; i < nl.getLength (); i++)
	{
	  ContentElement node = (ContentElement)nl.item(i);
	  //System.err.println( node );
	  if( node instanceof StashDirectory )
	    {
	      getStashItems( (ContentElement)node, in );	    	      
	    }
	  else
	    {	      
	      in.addElement( node );
	    }
	    
	  if( this.counter < 10 )
	    {
	      incrementPercentageNode( 0, 10, counter++, 10 );
	    }
	}    
  }

  /**
   * This method will always attempt to retrieve user data from the stash
   * This data WILL NOT be cached, so any sync-ing of the stash will become
   * immediately apparent.  If for some reason the stash does not contain
   * user data, this method will look to the relative file config.xml for data
   *
   */
  private XElement getLocalUserConfig ()
  {
    XElement userConfig = null;	
    XDocument doc = DocumentCache.fetchDocument ("stash:/home/default/data/sync/userConfig.xml");		
    if( doc != null )
      userConfig = (XElement)XPath.findFirstNode (doc, "config");
    
    if ( userConfig == null)
      {
	Logger.log( this, "Main.getLocalUserConfig(): fetch from stash is null.  Retrieving from package/config.xml..." );
	XElement originalConfig = (XElement)DocumentCache.fetchDocument (getRelativeSrc("config.xml")).getDocumentElement();	
	
	try
	  {
	    //now copy this info to the stash....?	    
	    URL userConfigURL = new URL( "stash:/home/default/data/sync/userConfig.xml" );
	    String userConfigString = XMLWriter.getXML( originalConfig );
	    //Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );	    
	    ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
	    Stash.put( userConfigCE );
	    return originalConfig;
	  }
	catch( IOException ioe )
	  {
	    Logger.log( this, ioe.toString() );
	    return null;
	  }      
      }
    return userConfig;
  }

  private void displaySettings( HttpServletResponse res ) throws IOException
  {
    XDocument doc = getRelativeDocument ("settings.xhtml");
    XElement settingsEl = getLocalUserConfig ();
    //Logger.log( this, "MY CONFIG: [" + XMLWriter.getXML( settingsEl ) + "]" );
    //System.err.println ("Main.displaySettings (): expanding xpath attributes");
    DOMUtils.expandUsingXPathAttributes (doc, settingsEl, null );	 
    //System.err.println ("Main.displaySettings (): expanded xpath attributes");
    doc.expandUsingXPathAttributes ();
    this.userNameInput = (Input)doc.getElementById("username"); 
    this.passwordInput = (Input)doc.getElementById("password");
    this.soapURLInput = (Input)doc.getElementById("SoapURL");
    this.syncDirInput = (Input)doc.getElementById("syncDir");
    //System.err.println ("Main.displaySettings (): printing doc to screen.");
    res.getWriter ().print (doc);    
  }


  /**
   * This builds a custom Dialog used as a status bar.
   *
   */ 
  private Dialog buildStatusBar()
  {
    XDocument statusBar = new XDocument();
    XElement htmlTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "html" ); 
    XElement bodyTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "body" );
    statusBar.appendChild( htmlTag );
    htmlTag.appendChild( bodyTag );
    
    XElement tableTitleTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "table" );
    tableTitleTag.setAttribute( "width", "100%" );      
    tableTitleTag.setAttribute( "align", "center" );
    tableTitleTag.setAttribute( "name", "title" );
    bodyTag.appendChild( tableTitleTag );
    XElement trTitleTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "tr" );
    trTitleTag.setAttribute( "height", "10" );
    trTitleTag.setAttribute( "width", "100%" ); 
    tableTitleTag.appendChild( trTitleTag );
    XElement tdTitleTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "td" );
    tdTitleTag.setAttribute( "height", "100%" );
    tdTitleTag.setAttribute( "width", "100%" );
    tdTitleTag.setAttribute( "align", "center" );   
    tdTitleTag.setAttribute( "bgcolor", "#FFFFFF" );
    trTitleTag.appendChild( tdTitleTag );
    XElement boldTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "b" );
    DOMUtils.setTextContent ( boldTag , "Sync in progress..." );
    tdTitleTag.appendChild( boldTag );

    XElement tableTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "table" );
    tableTag.setAttribute( "width", "0%" );
    tableTag.setAttribute( "name", "statusBar" );
    bodyTag.appendChild( tableTag );
    XElement trTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "tr" );
    trTag.setAttribute( "height", "10" );
    trTag.setAttribute( "width", "100%" ); 
    tableTag.appendChild( trTag );
    XElement tdTag = (XElement)statusBar.createElementNS( XHTML_NAMESPACE, "td" );
    tdTag.setAttribute( "bgcolor", "#000000" ); //black bar
    tdTag.setAttribute( "height", "100%" );
    tdTag.setAttribute( "width", "100%" );
    trTag.appendChild( tdTag );
    DOMUtils.setTextContent ( tdTag , "&nbsp;" );

    Dialog dialog = new Dialog ("Sync in progress...", false);
    dialog.setDocument (statusBar);
    dialog.pack ();
    Dimension screensize = ((IScreen) ServiceLocator.findImplementor (IScreen.INTERFACE)).getScreen ().getSize ();
    Dimension dialogsize = dialog.getSize();
    dialog.setLocation ((screensize.width - dialogsize.width) / 2, (screensize.height - dialogsize.height) / 2);
    return (dialog);
  }

  /**
   * This sets the status bar dialog in terms of a percentage.
   *
   */
  private void setPercentageNode( int num )
  {
    //Logger.log( this, "setPercentageNode(): " + num );
    ViewElement tableNode = (ViewElement)XPath.findFirstNode (this.statusBar.getDocument(), "//xhtml:table[@name='statusBar']" );
    tableNode.setAttribute( "width", num + "%" );
    
    tableNode.redrawNow ( true );
  }

  /**
   * This method is designed to be used within a loop.  Makes the status bar 
   * progress smoothly.
   *
   * @param: start int :beginning percentage
   * @param: end int :ending percentage
   * @param: loopIndex int :iterator, the only value that changes within a given loop
   * @param: loopSize int :the max value of a given loop
   *
   */
  private void incrementPercentageNode( int start, int end, int loopIndex, int loopSize )
  {
    int i = 0;
    if( loopSize < (end - start) )
      i = ((end - start)/loopSize);
    else
      i = loopSize/(end - start);

    if( (loopIndex%i) == 0 )	  
      setPercentageNode( loopIndex/i + start );
  }

} 
