package org.xoe.addressbook; 

import java.awt.Dimension;
import java.io.IOException;
import java.net.URL;
import java.net.URLEncoder;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.EntityResolver;
import org.xoe.core.dom.Loadable;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceFactory;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.xhtml.Dialog;


class Main 
  extends org.xoe.editor.data.Main
  implements ServiceFactory, IViewableDocumentBuilder
{  
  public static final String DOCTYPE_NAME = "entry";
  public static final String ADDRESS_NS = "http://xoe.org/ns/address/2001";
  public static final String ADDRESS_NS_PREFIX = "address";
  public static final String DOCTYPE_NAME_NS = ADDRESS_NS 
                                               + NAMESPACE_SEPARATOR
                                               + DOCTYPE_NAME;

  private static final XPath FIRST_NAME_PATH = XPathParser.singleton.parse ("address:name[@type='first']");
  private static final XPath LAST_NAME_PATH = XPathParser.singleton.parse ("address:name[@type='last']");

  private static final String EDITING_DETAILS = "editing_details";
  private static final String EDITING_CUSTOM = "editing_custom";
  private static final String PRIMARY_CONTACT_PATH = "address:contact[@primary='true'][1]";


  private final XPathContext addressContext = new XPathContext ();

  protected Dialog itemDetailsDialog;

  protected Element currentCategory = null;
  protected XDocument categorySelectView;
  protected XDocument categorySelectTitle;
  protected XDocument customEditView;
  protected XDocument customEditTitle;
  protected XDocument customEditMenu;
  protected Dialog categoryEditDialog;
  protected Dialog categoryRenameDialog;
  protected String currentFilter = "All";

  protected XDocument templateDoc;


  // -------------------------------------------------------------------------
  // org.xoe.core.services.ServiceFactory

  public IService newInstance() 
  {
    return (new Main ());
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    ContentType type = content.getType ();

    XDocument dataDoc = null;
    IDocumentBuilder db = (IDocumentBuilder)type.getService (IDocumentBuilder.INTERFACE);
    if (db != null)
      {
	dataDoc = db.getDocument (content);
      }
  
    return getDisplayDocument (dataDoc);
  }


  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  public void init (Document configDoc) 
    throws Exception
  {
    super.init (configDoc);
    addressContext.definePrefix (ADDRESS_NS_PREFIX, ADDRESS_NS);
    setXPathContext (addressContext);
  }

  protected String [] getServletFunctionalities ()
  {
    return (new String[] { FUNCTION_CONTENT_HANDLER });
  }

  protected String [] getServletInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { IViewableDocumentBuilder.INTERFACE });
      }

    return (null);
  }

  protected String [] getServletFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { 
                               FEATURE_DOCTYPE_NAME_NS, DOCTYPE_NAME_NS, 
                               FEATURE_EXTENSION, DOCTYPE_NAME 
                             });
      }

    return (null);
  }

  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    res.setContentType (MIMETYPE_TEXT_HTML);
    String path =req.getPathInfo();

    if ("MENU_OPTIONS".equals (path))
      {
        path = req.getParameter ("value");
      }

    if ("ITEM_CANCEL".equals (path) || NO_PATH.equals (path))
      {
        //right now cancel just goes back to the original page without saving.  if we need
        //it to do something different later, we should fill this part in.
        handleItem_select (req, res);
      }
    else if ("CATEGORY_SELECTED".equals (path))
      {
        handleCategory_selected (req, res);
      }
    else if ("CATEGORY_OK".equals (path))
      {
        handleCategory_ok (req, res);
      }
    else if ("CATEGORY_NEW".equals (path))
      {
        handleCategory_new (req, res);
      }
    else if ("CATEGORY_EDIT".equals (path))
      {
        handleCategory_edit (req, res);
      }
    else if ("CATEGORY_RENAME".equals (path))
      {
        handleCategory_rename (req, res);
      }
    else if ("CATEGORY_SAVE".equals (path))
      {
        handleCategory_save (req, res);
      }
    else if ("CATEGORY_DELETE".equals (path))
      {
        handleCategory_delete (req, res);
      }
    else if ("CATEGORY_CANCEL".equals (path))
      {
        getCategoryEditDialog ().setVisible (false);
      }
    else if ("CATEGORY_RENAME_CANCEL".equals (path))
      {
        getCategoryRenameDialog ().setVisible (false);
        currentCategory = null;
      }
    else if ("CUSTOM_RENAME".equals (path))
      {
	handleCustomFields_edit (req, res);
      }
    else if ("CUSTOM_SAVE".equals (path))
      {
	handleCustomFields_save (req, res);
      }
    else if ("CUSTOM_CANCEL".equals (path))
      {
	route (req, res);
      }
    else if ("ITEM_DETAILS".equals (path))
      {
        handleItemDialog_details (req, res);
      }
    else if ("ITEM_DETAILS_CANCEL".equals (path))
      {
	getItemDetailsDialog ().setVisible (false);
      }
    else if ("ITEM_DETAILS_SAVE".equals (path))
      {
	handleItemDialog_saveDetails (req, res);
      }
    else 
      {
        super.doGet (req, res);
      }
  }


  // -------------------------------------------------------------------------
  // org.xoe.editor.data.Main

  protected void restoreTitle ()
    throws IOException
  {
    if (EDITING_CUSTOM.equals (currentState))
      {
        setCustomEditTitle ();
      }
    else
      {
        super.restoreTitle ();
      }
  }


  protected void handleMenu ()
    throws ServletException, IOException
  {
    if (EDITING_CUSTOM.equals (currentState))
      {
        setCustomEditMenu ();
      }
    else 
      {
        super.handleMenu ();
      }
  }

  protected String getItemNamespaceURI ()
  {
    return (ADDRESS_NS);
  }

  protected String getDoctypeName ()
  {
    return (DOCTYPE_NAME);
  }

  protected void handleItem_confirmDelete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getItemDetailsDialog ().setVisible (false);
    super.handleItem_confirmDelete (req, res);
  }

  protected void handleItem_cancelDelete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    super.handleItem_cancelDelete (req, res);
    getItemDetailsDialog ().setVisible (true);
  }

  public XDocument getItemSelectView ()
    throws IOException
  {
    super.getItemSelectView ();
    Element list = (Element)XPath.findFirstNode (addressContext, itemSelectView, "//xhtml:table[@id='list'][1]");
    addItemsToList (Stash.fetch (getDataDir ()), currentFilter, list);
    return (itemSelectView);
  }

  public XDocument getItemDisplayView ()
    throws IOException
  {
//HACK!!! since we destroy this doc w/ hide-if's, we'll null it out so we fetch it fresh each time
    itemDisplayView = null;

    return (super.getItemDisplayView ());
  }

  protected boolean foundData (Node n, String value)
  {
    if (n == null || value == null)
      {
        return (false);
      }
      
    String name = n.getNodeName ();
    return (!name.equals ("type") 
            && !name.equals ("category")
            && value.length () > 0);
  }

  protected String getFileNamePrefix (XDocument file)
  {
    StringBuffer fileName = new StringBuffer ();
    fileName.append (XPath.findString (addressContext, file, LAST_NAME_PATH).replace (' ','_'));
    fileName.append (XPath.findString (addressContext, file, FIRST_NAME_PATH).replace (' ','_'));
    return (ContentElement.encodeName (fileName.toString ()));
  }

  // -------------------------------------------------------------------------
  // org.xoe.addressbook.Main

  protected void handleItemDialog_details (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    changeState (EDITING_DETAILS);
    getItemDetailsDialog ().setVisible (true);
  }

  protected void handleItemDialog_saveDetails (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getItemDetailsDialog ().setVisible (false);

    // get the new primary contact type value
    String value = req.getParameter (PRIMARY_CONTACT_PATH + "/@type");

    // unset the current primary contact
    Element e = (Element) XPath.findFirstNode (addressContext, getCurrentItem (), PRIMARY_CONTACT_PATH);
    if (e != null)
      {
	e.removeAttribute ("primary");
      }
    else
      {
	Logger.log (this, "saveDetails: current primary contact not found");
      }

    // set the new primary contact
    e = (Element) XPath.findFirstNode (addressContext, getCurrentItem (), "address:contact[@type='" + value + "']");        
    if (e != null)
      {
	e.setAttribute ("primary", "true");
      }
    else
      {
	Logger.log (this, "saveDetails: new primary contact not found for type = " + value);
      }

    // set the category
    value = req.getParameter ("address:category[1]");

    e = (Element) XPath.findFirstNode (addressContext, getCurrentItem (), "address:category[1]");
    if (e != null)
      {
	DOMUtils.setTextContent (e, value);
      }
    else
      {
	Logger.log (this, "saveDetails: category[1] not found");
      }

    saveCurrentItem ();

    // redraw category
    Loadable categorySelect = (Loadable)XPath.findFirstNode (addressContext, getItemEditView (), 
                                                             "//xhtml:select[@name='address:category[1]'][1]");
    if (categorySelect != null)
      {
        categorySelect.onLoad (addressContext, getCurrentItem ());
      }
  }

  protected void handleCategory_selected (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    String selectedCategory = req.getParameter ("address:category[1]");
    if (selectedCategory != null)
      {
	if (selectedCategory.equals ("edit"))
	  {
	    if (EDITING_DETAILS.equals (currentState))
	      {
		getItemDetailsDialog ().setVisible (false);
	      }
	    res.getWriter ().print (getCategorySelectView ());
            setCategorySelectTitle ();
	  }
	else
	  {
	    if (SELECTING_ITEM.equals (currentState))
	      {
		Logger.log (this, "doGet: filter on = " + selectedCategory);
                currentFilter = selectedCategory;
                handleItem_select (req, res);
	      }
	    else
	      {
		Logger.log (this, "doGet: ignoring because we are editing something");
	      }
	  }
      }
    else
      {
	Logger.log (this, "doGet: null selected category");
      }
  }

  protected void handleCategory_ok (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (SELECTING_ITEM.equals (currentState))
      {
	handleItem_select (req, res);
      }
    else if (EDITING_ITEM.equals (currentState))
      {
	handleItem_edit (req, res);
      }
    else if (EDITING_DETAILS.equals (currentState))
      {
	handleItem_edit (req, res);
	getItemDetailsDialog ().setVisible (true);
      }
    else
      {
	Logger.log (this, "doGet: unexpected state in OK_CATEGORY_EDIT = " + currentState);
      }
  }

  protected void handleCategory_new (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    currentCategory = getUserConfig ().createElementNS (null, "category");
    getCategoryRenameDialog ().setVisible (true);
  }

  protected void handleCategory_edit (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    String categoryName = req.getParameter ("value");
    if (categoryName == null)
      {
        throw new ServletException ("no value parameter passed to handleCategory_edit");
      }

    currentCategory = (Element)XPath.findFirstNode (addressContext, 
                                                    getUserConfig (), 
                                                    "config/categories/category[text()='" 
                                                     + categoryName 
                                                     + "'][1]");
    getCategoryEditDialog ().setVisible (true);
  }

  protected void handleCategory_rename (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getCategoryEditDialog ().setVisible (false);
    getCategoryRenameDialog ().setVisible (true);
  }

  protected void handleCategory_save (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    // remove the dialog from the screen
    getCategoryRenameDialog ().setVisible (false);
    
    // get the form value representing the new category name
    String categoryName = req.getParameter ("self::*");
    if (categoryName == null)
      {
        throw new ServletException ("no 'self::*' parameter passed to handleCategory_save");
      }

    if (categoryName.length () > 0)
      {
        if (currentCategory == null)
          {
	    throw new ServletException ("null current category in handleCategory_save");
	  }
        DOMUtils.setTextContent (currentCategory, categoryName);

        // if the current category has no parent, then it is new, so add it to the user config file
	if (currentCategory.getParentNode () == null)
	  {
            Element categories = (Element) XPath.findFirstNode (getUserConfig (), "config/categories[1]");
            if (categories != null)
              {
		categories.appendChild (currentCategory);
              }

            DOMUtils.sortChildren (categories);
	  }

	updateCategories ();
      }

    currentCategory = null;
  }

  protected void handleCategory_delete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getCategoryEditDialog ().setVisible (false);
    Element categories = (Element) XPath.findFirstNode (getUserConfig (), "config/categories[1]");
    if (categories != null)
      {
	categories.removeChild (currentCategory);
      }
    currentCategory = null;
    updateCategories ();
  }

  protected void updateCategories ()
    throws ServletException, IOException
  {
    // save the config doc
    DocumentCache.putDocument (getUserConfig ());
    // redraw category list box
    Loadable categoryList = (Loadable)XPath.findFirstNode (addressContext, 
                                                           getCategorySelectView (),  
                                                           "//xhtml:select[@id='list'][1]");
    if (categoryList != null)
      {
        categoryList.onLoad (addressContext, null);
      }
  }

  protected void handleCustomFields_edit (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    changeState (EDITING_CUSTOM);
    res.getWriter ().print (getCustomEditView ());
    setCustomEditTitle ();
  }

  protected void handleCustomFields_save (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (templateDoc != null)
      {
	// write the edits back to the config doc
	saveEdits (req, templateDoc);
	// put the modified config doc back in the cache
	DocumentCache.putDocument (templateDoc);
      }
    // route back to the previous screen
    route (req, res);
  }

  public XDocument getCategorySelectView ()
    throws IOException
  {
    if (categorySelectView == null)
      {
	categorySelectView = getPage ("category-select-view", "category-select-view.xhtml");
	if (categorySelectView == null)
	 { 
           throw new IOException ("view not found in getCategorySelectView");
         }
      }
    DOMUtils.expandUsingXPathAttributes (categorySelectView, null, addressContext);
    return (categorySelectView);
  }


  protected void setCategorySelectTitle ()
    throws IOException
  {
    if (categorySelectTitle == null)
      {
	categorySelectTitle = getPage ("category-select-title", "category-select-title.xhtml");
      }
    setTitleDocument (categorySelectTitle);
  }

  public XDocument getCustomEditView ()
    throws IOException
  {
    if (customEditView == null)
      {
	customEditView = getPage ("custom-edit-view", "custom-edit-view.xhtml");
	if (customEditView == null)
	 { 
           throw new IOException ("view not found in getCustomEditView");
         }
      }

    templateDoc = getFileFromPackageWorkspace (ITEM_TEMPLATE + getDoctypeName ());
    if (templateDoc != null)
      {
	DOMUtils.expandUsingXPathAttributes (customEditView, templateDoc.getDocumentElement (), addressContext);
      }
    return (customEditView);
  }

  public void setCustomEditTitle ()
    throws IOException
  {
    if (customEditTitle == null)
      {
	customEditTitle = getPage ("custom-edit-title", "custom-edit-title.xhtml");
      }
    setTitleDocument (customEditTitle);
  }

  public void setCustomEditMenu ()
    throws IOException
  {
    if (customEditMenu == null)
      {
	customEditMenu = getPage ("custom-edit-menu", "custom-edit-menu.xhtml");
      }
    setMenu (customEditMenu);
  }

  public Dialog getItemDetailsDialog ()
    throws IOException
  {
    if (itemDetailsDialog == null)
      {
	XDocument dialogView = getPage ("item-details-dialog", "item-details-dialog.xhtml");
	if (dialogView == null)
	 { 
           throw new IOException ("view not found in getItemDetailsDialog");
         }

        // notify us if there is an action, onclick or onchange event
	dialogView.setCallback ("action", defaultCallback);
	dialogView.setCallback ("onclick", defaultCallback);
	dialogView.setCallback ("onchange", defaultCallback);

        itemDetailsDialog = buildDialog (dialogView);
      }

    DOMUtils.expandUsingXPathAttributes (itemDetailsDialog.getDocument (), getCurrentItem (), addressContext);
    return (itemDetailsDialog);
  }

  public Dialog getCategoryEditDialog ()
    throws IOException
  {
    if (categoryEditDialog == null)
      {
	XDocument dialogView = getPage ("category-edit-dialog", "category-edit-dialog.xhtml");
	if (dialogView == null)
	 { 
           throw new IOException ("view not found in getCategoryEditDialog");
         }

        // notify us if there is an onclick event
	dialogView.setCallback ("onclick", defaultCallback);

        categoryEditDialog = buildDialog (dialogView);
      }

    DOMUtils.expandUsingXPathAttributes (categoryEditDialog.getDocument (), currentCategory, addressContext);
    return (categoryEditDialog);
  }

  public Dialog getCategoryRenameDialog ()
    throws IOException
  {
    if (categoryRenameDialog == null)
      {
	XDocument dialogView = getPage ("category-rename-dialog", "category-rename-dialog.xhtml");
	if (dialogView == null)
	 { 
           throw new IOException ("view not found in getCategoryRenameDialog");
         }

        // notify us if there is an action or onclick event
	dialogView.setCallback ("action", defaultCallback);
	dialogView.setCallback ("onclick", defaultCallback);

        categoryRenameDialog = buildDialog (dialogView);
      }

    DOMUtils.expandUsingXPathAttributes (categoryRenameDialog.getDocument (), currentCategory, addressContext);
    return (categoryRenameDialog);
  }

  public XDocument getDisplayDocument (XDocument dataDoc)
    throws IOException
  {
    if (dataDoc != null)
      {
	setCurrentItem (dataDoc.getDocumentElement ());
	setDisplayedDocument (getItemDisplayView ());
      }

    return (getViewableDocument ());
  }

  public void addItemsToList (ContentElement stashDir, String filter, Element list)
    throws IOException
  {
    // clear the list
    DOMUtils.removeChildNodes (list);     
 
    // does the directory exist?
    if (stashDir == null)
      {
        return;        
      }

    // hack: explicitly construct the document builder
    org.xoe.core.content.text.xml db = new org.xoe.core.content.text.xml ();

    if ("All".equals (currentFilter))
      {
        filter = null;
      }

    NodeList dirContents = stashDir.getChildNodes ();
    for (int i=0; i < dirContents.getLength (); i++)
      {
        ContentElement content = (ContentElement)dirContents.item (i);
        if (DOCTYPE_NAME_NS.equals (content.getType ().getXMLDoctypeNameNS ()))
          {
	    XDocument itemDoc = db.getDocument (content);
            Node item = itemDoc.getDocumentElement ();

	    Node category = null;
	    if (filter != null)
	      {
		category = XPath.findFirstNode (addressContext, 
					        item, 
					       "address:category[text()='" 
						+ filter 
						+ "']");
	      }

	    // if there is no filter or we found the category we 
            // filtered on then add the item to the list
	    if (filter == null || category != null)
	      {
		String itemURI = URLEncoder.encode (itemDoc.getURL ().toString ());
		String src = "ITEM_DISPLAY?value=" +  itemURI;

		Element tr = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "tr");

		Element td = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "td");

		Element font = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "font");
                font.setAttribute ("size", "-1");

		Element a = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "a");
		a.setAttribute ("href", src);
		String value = XPath.findString (addressContext, item, "address:name[@type='last']");
		if (value.length () > 0)
		  {
		    value += ", ";
		  }
		value += XPath.findString (addressContext, item, "address:name[@type='first']");
		// if no first or last name, set the display name to unnamed
		if (value.length () == 0)
		  {
		    value = "-Unnamed-";
		  }
		DOMUtils.setTextContent (a, value); 

                font.appendChild (a);
                td.appendChild (font);
                tr.appendChild (td);
		list.appendChild (tr);

		td = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "td");
		td.setAttribute ("align","right");

		font = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "font");
                font.setAttribute ("size", "-1");

		a = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "a");
		a.setAttribute ("href", src);

		Element primary = (Element)XPath.findFirstNode (addressContext, item, "address:contact[@primary='true'][1]");
		if (primary == null)
		  {
		    primary = (Element)XPath.findFirstNode (addressContext, item, "address:contact[1]");
		  }
		// get the contact info
		value = DOMUtils.getTextContent (primary);
		// if necessary, append the first character of the entified contact type 
		// to the contact info
		if (value != null && value.length () > 0)
		  {
		    value += "&nbsp;" 
			     + ((EntityResolver)list.getOwnerDocument ()).getValue (primary.getAttribute ("type")).charAt (0);
		  }
		DOMUtils.setTextContent (a, value); 

		font.appendChild (a);
                td.appendChild (font);
		tr.appendChild (td);

                // if there is a note, add the note icon to the row
		if (XPath.findFirstNode (addressContext, item, "xoe:note[1]") != null)
		  {
		    td = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "td");
		    td.setAttribute ("align","right");

		    a = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "a");
		    a.setAttribute ("href", "NOTE_EDIT?item=" + itemURI);

                    Element img = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "img");
                    img.setAttribute ("src", getNoteIconURL ().toString ());
                    img.setAttribute ("border", "0");

                    a.appendChild (img);
                    td.appendChild (a);                    
		    tr.appendChild (td);
                  }
	      }
          }
      }  

    // sort the addresses in the list
    DOMUtils.sortChildren (list);

    Element tr = list.getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "tr");
    tr.setAttribute ("height", "*");
    list.appendChild (tr);
  }

}
