package org.xoe.calendar; 

import java.awt.Dimension;
import java.io.IOException;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Date;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.Loadable;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceFactory;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.xhtml.Dialog;


class Main 
  extends org.xoe.editor.data.Main
  implements ServiceFactory, IViewableDocumentBuilder
{  
  public static final String DOCTYPE_NAME = "appt";
  public static final String CALENDAR_NS = "http://xoe.org/ns/calendar/2001";
  public static final String CALENDAR_NS_PREFIX = "cal";
  public static final String DOCTYPE_NAME_NS = CALENDAR_NS 
                                               + NAMESPACE_SEPARATOR
                                               + DOCTYPE_NAME;

  protected static final XPath DAY_PATH = XPathParser.singleton.parse ("cal:date/@day");
  protected static final XPath MONTH_PATH = XPathParser.singleton.parse ("cal:date/@month");
  protected static final XPath YEAR_PATH = XPathParser.singleton.parse ("cal:date/@year");

  protected static final String DAY_VIEW = "day";
  protected static final String WEEK_VIEW = "week";
  protected static final String MONTH_VIEW = "month";
  protected static final String GOTO_VIEW = "goto";

  private final XPathContext calContext = new XPathContext ();

  protected String currentView;
  protected String previousView;

  // field where the date is changed in the navigation title document 
  protected Element dateField;

  protected Element daySelectFrame;
  protected XDocument daySelectDoc;
  protected Element currentDayBox;

  protected Element lastAnchor;
  protected Element nextAnchor;

  protected int selectedHour = 0;

  // field where the title is changed in the simple title document 
  protected Element titleField;

  protected XDocument dateSelectView;
  protected Element yearField;
  protected NodeList monthBoxes;
  protected Element currentMonthBox;
  protected Element todayButton;
  protected Date previousDate;

  protected Element currentStartTime;
  protected Element currentEndTime;

  protected Dialog timeSetDialog;
  protected Element timeBeingSet;


  // -------------------------------------------------------------------------
  // org.xoe.core.services.ServiceFactory

  public IService newInstance() 
  {
    return (new Main ());
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    return (null);
  }

  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  public void init (Document configDoc) 
    throws Exception
  {
    super.init (configDoc);
    calContext.definePrefix (CALENDAR_NS_PREFIX, CALENDAR_NS);
    setXPathContext (calContext);
  }

  protected String [] getServletFunctionalities ()
  {
    return (new String[] { FUNCTION_CONTENT_HANDLER });
  }

  protected String [] getServletInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { IViewableDocumentBuilder.INTERFACE });
      }

    return (null);
  }

  protected String [] getServletFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { 
//                               FEATURE_DOCTYPE_NAME_NS, DOCTYPE_NAME_NS, 
//                               FEATURE_EXTENSION, DOCTYPE_NAME 
                             });
      }

    return (null);
  }

  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    res.setContentType (MIMETYPE_TEXT_HTML);
    String path =req.getPathInfo();

    if ("MENU_OPTIONS".equals (path))
      {
        path = req.getParameter ("value");
      }
    else if ("MENU_HIDE".equals (path))
      {
        restoreTitle ();
      }
    else if ("MENU_SHOW".equals (path))
      {
        handleMenu ();
      }

    if (path.startsWith ("GOTO"))
      {
        handleGoto (path, req, res);
      }
    else if (path.equals ("SET_START_TIME"))
      {
        timeBeingSet = currentStartTime;
	getTimeSetDialog ().setVisible (true);
      }
    else if (path.equals ("SET_END_TIME"))
      {
        timeBeingSet = currentEndTime;
	getTimeSetDialog ().setVisible (true);
      }
    else if (path.equals ("SET_NO_TIME"))
      {
	getTimeSetDialog ().setVisible (false);
	setTime ("--","--","--","--");
	res.getWriter ().print (getItemEditView ());
      }
    else if (path.equals ("SET_TIME_SAVE"))
      {
	getTimeSetDialog ().setVisible (false);
        super.saveEdits (req, timeBeingSet);

        // here we'll have to do a bunch of checks to ensure 
        // it's a legal setting
//        if (timeBeingSet == currentStartTime)
//          {
//            setEndTime (
//          }

	res.getWriter ().print (getItemEditView ());
      }
    else if (path.equals ("SET_TIME_CANCEL"))
      {
	getTimeSetDialog ().setVisible (false);
      }
    else 
      {
        super.doGet (req, res);
      }
  }


  // -------------------------------------------------------------------------
  // org.xoe.editor.data.Main

  protected void setItemSelectTitle ()
    throws IOException
  {
    setNavigationTitle ();
  }

  protected void handleItem_select (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (currentView == null)
      {
	setCurrentView (DAY_VIEW);
      }
    super.handleItem_select (req, res);
  }


  // -------------------------------------------------------------------------
  // org.xoe.calendar.Main

  protected void restoreTitle ()
    throws IOException
  {
    if (EDITING_ITEM.equals (currentState))
      {
        if (GOTO_VIEW.equals (currentView))
          {
            setSimpleTitle (getItemEditView ().getValue ("Set Date"));
          }
        else
          {
	    setItemEditTitle ();
          }
      }
    else if (EDITING_NOTE.equals (currentState))
      {
        setNoteEditTitle ();
      }
    else 
      {
        setTitle ();
      }
  }


  protected void handleMenu ()
    throws ServletException, IOException
  {
//    if (EDITING_CUSTOM.equals (currentState))
//      {
//        setCustomEditMenu ();
//      }
//    else 
//      {
        super.handleMenu ();
//      }
  }

  protected void handleGoto (String path, HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if ("GOTO_LAST_WEEK".equals (path))
      {
	CurrentDate.setLastWeek ();
      }
    else if ("GOTO_NEXT_WEEK".equals (path))
      {
	CurrentDate.setNextWeek ();
      }
    else if ("GOTO_LAST_MONTH".equals (path))
      {
	CurrentDate.setLastMonth ();
      }
    else if ("GOTO_NEXT_MONTH".equals (path))
      {
	CurrentDate.setNextMonth ();
      }
    else if ("GOTO_LAST_YEAR".equals (path))
      {
	CurrentDate.setLastYear ();
      }
    else if ("GOTO_NEXT_YEAR".equals (path))
      {
	CurrentDate.setNextYear ();
      }
    else if ("GOTO_TODAY".equals (path))
      {
	CurrentDate.setDate (new Date ());
	if (GOTO_VIEW.equals (currentView))
	  {
	    setCurrentView (previousView);
	  }
      }
    else if ("GOTO_DAY".equals (path))
      {
	CurrentDate.setCurrentDayOfWeek (Integer.valueOf (req.getParameter ("value")).intValue ());
      }
    else if ("GOTO_MONTH".equals (path))
      {
	CurrentDate.setMonth (Integer.valueOf (req.getParameter ("value")).intValue ());
      }
    else if ("GOTO_WEEK_VIEW".equals (path))
      {
	setCurrentView (WEEK_VIEW);
      }
    else if ("GOTO_MONTH_VIEW".equals (path))
      {
	setCurrentView (MONTH_VIEW);
      }
    else if ("GOTO_DATE".equals (path))
      {
	setCurrentView (GOTO_VIEW);
      }
    else if ("GOTO_CANCEL".equals (path))
      {
	if (previousDate != null)
	  {
	    CurrentDate.setDate (previousDate);
            previousDate = null;
	  }
	setCurrentView (previousView);
      }
    else
      {
	setCurrentView (DAY_VIEW);
      }

    update (req, res);
  }

  protected void setCurrentView (String newView)
  {
    previousView = currentView;
    currentView = newView;
  }

  protected void update (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (GOTO_VIEW.equals (currentView))
      {
	res.getWriter ().print (getDateSelectView ());
      }
    else
      {
	res.getWriter ().print (getItemSelectView ());
      }
    setTitle ();
  }

  protected XDocument getDateSelectView ()
    throws IOException
  {
    if (dateSelectView == null)
      {
	dateSelectView = getRelativeDocument ("date-select-view.xhtml");
	if (dateSelectView == null)
	  { 
            throw new IOException ("view not found in getDateSelectView");
          }
       
	// get the year field, where the current year is shown on the screen
        String locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:b[1]";
	yearField = (Element) XPath.findFirstNode (calContext, dateSelectView, locationPath);

	// get the nodelist representing the <td> elements for each month box
        locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:table[@id='months']/xhtml:tr/xhtml:td";
	monthBoxes = XPath.findNodes (calContext, dateSelectView, locationPath);
  
	// get the today/this month/this year button
        locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:input[2]";
	todayButton = (Element) XPath.findFirstNode (calContext, dateSelectView, locationPath);
      }

    // save the previous date in case the user presses cancel
    if (previousDate == null)
      {
	previousDate = CurrentDate.getDate ();
      }

    DOMUtils.setTextContent (yearField, Integer.toString (CurrentDate.getYear ()));

    highlightMonthBox ();

    if (DAY_VIEW.equals (previousView))
      {
        todayButton.setAttributeNS (null, "value", dateSelectView.getValue ("today"));
      }
    else if (WEEK_VIEW.equals (previousView))
      {
        todayButton.setAttributeNS (null, "value", dateSelectView.getValue ("this_week"));
      }
    else if (MONTH_VIEW.equals (previousView))
      {
        todayButton.setAttributeNS (null, "value", dateSelectView.getValue ("this_month"));
      }

    return (dateSelectView);
  }

  protected void highlightMonthBox ()
    throws IOException
  {
    if (monthBoxes != null)
      {
        Element previousMonthBox = currentMonthBox;
        currentMonthBox = (Element)monthBoxes.item (CurrentDate.getMonth ());
        // if we got a month box and it's different from the previous month box
        if (currentMonthBox != null && previousMonthBox != currentMonthBox)
          {
	    currentMonthBox.setAttributeNS (null, "bgcolor", "black");
            Element font = (Element)currentMonthBox.getFirstChild ().getFirstChild ();
            if (font != null)
              {
                font.setAttributeNS (null, "color", "white");                 
              }

            // clear the previous box
            if (previousMonthBox != null)
              {
		previousMonthBox.setAttributeNS (null, "bgcolor", "white");
		font = (Element)previousMonthBox.getFirstChild ().getFirstChild ();
		if (font != null)
		  {
		    font.removeAttributeNS (null, "color");                 
		  }
              }
          }
      }
  }

  public XDocument getItemEditView ()
    throws IOException
  {
    // HACK!!! since we destroy this doc w/ hide-if's, we'll 
    // null it out so we fetch it fresh each time
    itemEditView = null;

    return (super.getItemEditView ());
  }

  protected String getItemNamespaceURI ()
  {
    return (CALENDAR_NS);
  }

  protected String getDoctypeName ()
  {
    return (DOCTYPE_NAME);
  }

  protected void setTitle ()
    throws IOException
  {
    if (GOTO_VIEW.equals (currentView))
      {
	setSimpleTitle (getItemEditView ().getValue ("goto_date"));
      }
    else
      {
	setNavigationTitle ();
      }
  }

  protected void setSimpleTitle (String titleText)
    throws IOException
  {
    if (titleField == null)
      {
        XDocument simpleTitle = getPage ("simple-title", "simple-title.xhtml");
	if (simpleTitle == null)
	  {
            throw new IOException ("simple-title not found in setSimpleTitle");
          }

	// get the date element - where the relevant date for the view is put
        String locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:font";
	titleField = (Element)XPath.findFirstNode (calContext, simpleTitle, locationPath);
      }

    DOMUtils.setTextContent (titleField, titleText);
    setTitleDocument (titleField.getOwnerDocument ());
  }

  protected void setNavigationTitle ()
    throws IOException
  {
    if (dateField == null)
      {
	XDocument navTitle = getPage ("navigation-title", "navigation-title.xhtml");
	if (navTitle == null)
	  {
            throw new IOException ("navigation-title not found in setNavigationTitle");
          }

	// get the date element - where the relevant date for the view is put
        String locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:font/xhtml:b[1]";
	dateField = (Element)XPath.findFirstNode (calContext, navTitle, locationPath);

	// get the frame where the day select boxes will go if this is a day view
        locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:iframe[1]";
	daySelectFrame = (Element)XPath.findFirstNode (calContext, navTitle, locationPath);

	// get the "last" arrow
        locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:a[1]";
	lastAnchor = (Element)XPath.findFirstNode (calContext, navTitle, locationPath);

	// get the "next" arrow
        locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td/xhtml:a[2]";
	nextAnchor = (Element)XPath.findFirstNode (calContext, navTitle, locationPath);
      }
 
    // get the month name
    String d = CurrentDate.getMonthAsString ()
	       + "&nbsp;";

    // if it's the day view, show the day of the month
    if (DAY_VIEW.equals (currentView))
      {
	d+= Integer.toString (CurrentDate.getDayOfMonth ()) 
	    + ",&nbsp;";
      }

    // append the year
    d+= Integer.toString (CurrentDate.getYear ());
    DOMUtils.setTextContent (dateField, d);

    if (DAY_VIEW.equals (currentView))
      {
        // highlight the box for this day of the week
        highlightDayBox (CurrentDate.getDayOfWeek ());
        ((DocumentViewer)daySelectFrame).setDocument (daySelectDoc);
        
        // set arrows
        lastAnchor.setAttributeNS (null, "href", "GOTO_LAST_WEEK");
        nextAnchor.setAttributeNS (null, "href", "GOTO_NEXT_WEEK");
      }    
    else if (WEEK_VIEW.equals (currentView))
      {
        // set arrows
        lastAnchor.setAttributeNS (null, "href", "GOTO_LAST_WEEK");
        nextAnchor.setAttributeNS (null, "href", "GOTO_NEXT_WEEK");

        ((DocumentViewer)daySelectFrame).setDocument (null);
      }
    else if (MONTH_VIEW.equals (currentView))
      {
        // set arrows
        lastAnchor.setAttributeNS (null, "href", "GOTO_LAST_MONTH");
        nextAnchor.setAttributeNS (null, "href", "GOTO_NEXT_MONTH");

        ((DocumentViewer)daySelectFrame).setDocument (null);
      }

    setTitleDocument (dateField.getOwnerDocument ());
  }

  protected void highlightDayBox (int idx)
    throws IOException
  {
    if (daySelectDoc == null)
     {
       daySelectDoc = getPage ("day-select-view", "day-select-view.xhtml");
     }

    if (daySelectDoc != null)
      {
        String locationPath = "xhtml:html/xhtml:body/xhtml:table/xhtml:tr/xhtml:td[" + idx + "]";
        Element previousDayBox = currentDayBox;
        currentDayBox = (Element)XPath.findFirstNode (calContext, daySelectDoc, locationPath);
        // if we got a day box and it's different from the previous day box
        if (currentDayBox != null && previousDayBox != currentDayBox)
          {
	    currentDayBox.setAttributeNS (null, "bgcolor", "black");
            Element font = (Element)currentDayBox.getFirstChild ().getFirstChild ();
            if (font != null)
              {
                font.setAttributeNS (null, "color", "white");                 
              }

            // clear the previous box
            if (previousDayBox != null)
              {
		previousDayBox.setAttributeNS (null, "bgcolor", "white");
		font = (Element)previousDayBox.getFirstChild ().getFirstChild ();
		if (font != null)
		  {
		    font.removeAttributeNS (null, "color");                 
		  }
              }
          }
      }
  }

  public Dialog getTimeSetDialog ()
    throws IOException
  {
    if (timeSetDialog == null)
      {
	XDocument dialogView = getPage ("time-set-dialog", "time-set-dialog.xhtml");
	if (dialogView == null)
	 { 
           throw new IOException ("view not found in getTimeSetDialog");
         }

        // notify us if there is an action or onclick event
	dialogView.setCallback ("action", defaultCallback);
	dialogView.setCallback ("onclick", defaultCallback);

        timeSetDialog = buildDialog (dialogView);
      }

    DOMUtils.expandUsingXPathAttributes (timeSetDialog.getDocument (), 
                                         timeBeingSet, 
                                         calContext);
    return (timeSetDialog);
  }

  protected boolean foundData (Node n, String value)
  {
    if (n == null || value == null)
      {
        return (false);
      }
      
    return (true);
  }

  protected String getFileNamePrefix (XDocument file)
  {
    StringBuffer fileName = new StringBuffer ();
    fileName.append (XPath.findString (calContext, file, DAY_PATH).replace (' ','_'));
    fileName.append (XPath.findString (calContext, file, MONTH_PATH).replace (' ','_'));
    fileName.append (XPath.findString (calContext, file, YEAR_PATH).replace (' ','_'));
    return (ContentElement.encodeName (fileName.toString ()));
  }


  protected void createNewItem ()
    throws ServletException, IOException
  {
    super.createNewItem ();
    
    setCurrentItemDate (Integer.toString (CurrentDate.getMonth ()),
                        Integer.toString (CurrentDate.getDayOfMonth ()),
			Integer.toString (CurrentDate.getYear ()));

    currentStartTime = (Element) XPath.findFirstNode (calContext, 
                                                  getCurrentItem (), 
                                                  "cal:start");
    
    currentEndTime = (Element) XPath.findFirstNode (calContext, 
                                                getCurrentItem (), 
                                                "cal:end");
    
    if (selectedHour != 0)
      {
        setTime (Integer.toString (selectedHour), 
                 "00",
                 Integer.toString (selectedHour + 1),
		 "00");
      }
  }

  protected void setCurrentItemDate (int month, int day, int year)
  {
    setCurrentItemDate (Integer.toString (month), 
                        Integer.toString (day), 
                        Integer.toString (year));
  }

  protected void setCurrentItemDate (String month, String day, String year)
  {
    // initialize the month
    String locationPath = "cal:date/@month";
    Node n = XPath.findFirstNode (calContext, getCurrentItem (), locationPath);
    DOMUtils.setTextContent (n, month);
    
    // initialize the day
    locationPath = "cal:date/@day";
    n = XPath.findFirstNode (calContext, getCurrentItem (), locationPath);
    DOMUtils.setTextContent (n, day);
    
    // initialize the year
    locationPath = "cal:date/@year";
    n = XPath.findFirstNode (calContext, getCurrentItem (), locationPath);
    DOMUtils.setTextContent (n, year);
  }

  protected boolean saveEdits (HttpServletRequest req, Node contextNode)
    throws IOException
  {
    boolean dude = super.saveEdits (req, contextNode);
    String subject = req.getParameter ("cal:subject");
    if (subject == null || subject.length () == 0)
      {
	// add the default subject
	Node n = XPath.findFirstNode (calContext, getCurrentItem (), "cal:subject");
	DOMUtils.setTextContent (n, getItemEditView().getValue ("no_subject"));
      }
    return (dude);
  }

  protected void setTime (String startHours, String startMinutes,
                          String endHours, String endMinutes)
  {
    setStartTime (startHours, startMinutes);
    setEndTime (endHours, endMinutes);
  }

  protected void setStartTime (String startHours, String startMinutes)
  {
    // initialize the start hours
    DOMUtils.setTextContent (currentStartTime.getAttributeNode ("hours"), 
			     startHours);

    // initialize the start minutes
    DOMUtils.setTextContent (currentStartTime.getAttributeNode ("minutes"), 
			     startMinutes);
  }

  protected void setEndTime (String endHours, String endMinutes)
  {
    // initialize the end hours
    DOMUtils.setTextContent (currentEndTime.getAttributeNode ("hours"), 
			     endHours);

    // initialize the end minutes
    DOMUtils.setTextContent (currentEndTime.getAttributeNode ("minutes"), 
			     endMinutes);
   }
}
