/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: Main.java,v 1.21 2001/11/21 01:30:49 samc Exp $ 
 */
package org.xoe.mediaplayer;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.ContentCache;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.display.dom.xhtml.Select;
import org.xoe.display.dom.xhtml.Dialog;

import org.xoe.audio.MP3SoundStream;
import org.xoe.audio.XoeLineListener;
import org.xoe.audio.Comment;
import org.xoe.display.dom.xhtml.Input;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.servlet.XoeServlet;
import org.xoe.services.IApplication;

import java.io.IOException;
import java.net.URL;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * MediaPlayer.
 * PocketLinux application for playing mp3 audio (local/streaming).
 */
public class Main
  extends XoeServlet
  implements IViewableDocumentBuilder
{
  public static final int MAX_VOLUME = 256;
  public static final int DEFAULT_VOLUME = 100;

  // commands
  public static final String SELECT        = "SELECT";
  public static final String PLAY          = "PLAY";
  public static final String PAUSE         = "PAUSE";
  public static final String STOP          = "STOP";
  public static final String OPEN          = "OPEN";
  public static final String ADJUST_VOLUME = "ADJUST_VOLUME";
  public static final String START         = "START";

  // ids on elements (used for bindings)
  public static final String FILENAME   = "filename";
  public static final String COMMENT    = "comment";
  public static final String ARTIST     = "artist";
  public static final String TITLE      = "title";
  public static final String ALBUM      = "album";
  public static final String YEAR       = "year";
  public static final String GENRE      = "genre";
  public static final String VOLUME     = "volume";
  public static final String PLAYPAUSE  = "playpause";

  Input playpause;
  Input volControl;

  StreamThread streamThread;
  ContentElement currentContent = null;

  static XoeLineListener listener;

  static
  {
    listener = new XoeLineListener ();
  }

  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  /**
   * Get the tvt-mediaplayer functionalities.  That is, application and 
   * content handler for certain media types.
   */
  public String[] getServletFunctionalities ()
  {
    return new String[] { FUNCTION_CONTENT_HANDLER };
  }

  /**
   * Get the tvt-mediaplayer interfaces.
   */
  public String[] getServletInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	return new String[] { IViewableDocumentBuilder.INTERFACE };
      }
    return null;
  }

  public String[] getServletFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	return new String[] { FEATURE_MIMETYPE, "audio/mpeg", 
                              FEATURE_EXTENSION, "mp3" };
      }
    return null;
  }

  // -------------------------------------------------------------------------
  // org.xoe.core.content.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    ContentType type = content.getType ();
    Logger.log (this, type);

    // [FIXME] hack to get the servlet to start playing the currentContent
    currentContent = content;
    XDocument doc = getViewableDocument ();
    doc.getCallback ("onclick").invoke (doc.getDocumentElement (), "START");
    return doc;
  }

  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    res.setContentType ("text/html");
    String path = req.getPathInfo ();
    Logger.log (this, path);

    if ("".equals (path) || OPEN.equals (path))
      {
	Document doc = getRelativeDocument ("select.xhtml");
	DOMUtils.expandUsingXPathAttributes (doc, null, null);

	res.getWriter ().print (doc);
      }
    else if (STOP.equals (path)) 
      {
	playpause.setAttribute ("name", ">");
	playpause.setAttribute ("onclick", "START");
	streamThread.stop ();

        /*
         * [FIXME] ummm....is this the correct way 
         * to get oss to actaully close /dev/dsp?? 
         */
        //listener.close ();
      }
    else if (START.equals (path))
      {
	Document doc = getRelativeDocument ("play.xhtml");
	playpause = (Input) doc.getElementById (PLAYPAUSE);
	volControl = (Input) doc.getElementById (VOLUME);
	volControl.setValue (DEFAULT_VOLUME);
        String surl = req.getParameter ("url");
        if (surl != null)
          {
            currentContent = ContentCache.fetch (surl);
          }
        play (currentContent, doc);
        res.getWriter ().print (doc);
      }
    else if (ADJUST_VOLUME.equals (path)) 
      {
        //	streamThread.setVolume (MAX_VOLUME - volControl.getValue () );
      }
    else if (PLAY.equals (path)) 
      {
        playpause.setAttribute ("value", "||");
        playpause.setAttribute ("onclick", "PAUSE");
        streamThread.play (); 
      }
    else if (PAUSE.equals (path)) 
      {
        playpause.setAttribute ("value", ">");
        playpause.setAttribute ("onclick", "PLAY");
        streamThread.pause (); 
      }
  }

  private void play (ContentElement content, Document doc)
    throws IOException
  {
    URL url = content.getURL ();
    Logger.log (this, "playing: " + url);
    MP3SoundStream stream = new MP3SoundStream ();
    stream.open (url);

    Element fileName = doc.getElementById (FILENAME);;
    Element artist   = doc.getElementById (ARTIST);
    Element title    = doc.getElementById (TITLE);
    Element album    = doc.getElementById (ALBUM);
    Element year     = doc.getElementById (YEAR);
    Element genre    = doc.getElementById (GENRE);
    Element comment1 = doc.getElementById (COMMENT);
   
    // first reset all the data fields  
    String surl = url.toString ();
    int len = surl.length ();
    if (len > 30)
      {
        surl = "... " + surl.substring (len - 30, len);
      }
    DOMUtils.setTextContent (fileName, surl);
    DOMUtils.setTextContent (artist, "");
    DOMUtils.setTextContent (title, "");
    DOMUtils.setTextContent (album, "");
    DOMUtils.setTextContent (genre, "");

    Comment [] comms = stream.getComments ();
    if (comms != null)
      {
        Logger.log (this, "got comments in tvt-mediaplayer");
	for (int i = 0; i < comms.length; i++)
	  {
	    String name = comms[i].getName ();
	    if (name.equals ("ARTIST"))
	      {
		DOMUtils.setTextContent (artist, comms[i].getValue ().trim ());
	      }
	    if (name.equals ("TITLE"))
	      {
		DOMUtils.setTextContent (title, comms[i].getValue ().trim ());
	      }
	    if (name.equals ("ALBUM"))
	      {
		DOMUtils.setTextContent (album, comms[i].getValue ().trim ());
	      }
	    if (name.equals ("GENRE"))
	      {
		DOMUtils.setTextContent (genre, comms[i].getValue ().trim ());
	      }
	    DOMUtils.setTextContent (comment1, "");
	  }
      }
    if (streamThread != null)
      {
        streamThread.stop ();
      }
    streamThread = new StreamThread (stream);
    Thread t = new Thread (streamThread);
    t.start ();
    streamThread.play ();
  }

  class StreamThread 
    implements Runnable
  {
    MP3SoundStream stream;

    StreamThread (MP3SoundStream stream)
    {
      this.stream = stream;
    }

    public void run () 
    {
      stream.addLineListener (listener);
      stream.open ();
      stream.start ();
      stream.stop ();
    }

    public void play () 
    {
      stream.play ();
    }

    public void pause () 
    {
      stream.pause();
    }

    public void stop () 
    {
      stream.stop ();
      stream.removeLineListener (listener);
    }
  }
} 
