/* 
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "COPYING" for details.
 *
 * $tvt: Main.java,v 1.5 2001/11/04 00:09:32 joerg Exp $ 
 */
package org.xoe.pluginplayer;

import org.xoe.core.dom.XDocument;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.ContentCache;
import org.xoe.core.util.Logger;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.xpath.XPath;

import java.net.URL;
import java.io.IOException;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.util.DOMUtils;
import org.xoe.display.dom.ViewElement;   
import org.xoe.display.dom.xhtml.Input;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.servlet.XoeServlet;
import org.xoe.services.IApplication;

import org.xoe.display.dom.xhtml.Embed;

/**
 * PluginPlayer.
 * PocketLinux application for plugin players (local/streaming).
 */
public class Main
  extends XoeServlet
  implements IViewableDocumentBuilder
{
  public static final int MAX_VOLUME = 256;
  public static final int DEFAULT_VOLUME = 100;

  // commands
  public static final String SELECT        = "SELECT";
  public static final String PLAY          = "PLAY";
  public static final String PAUSE         = "PAUSE";
  public static final String STOP          = "STOP";
  public static final String START         = "START";
  public static final String OPEN          = "OPEN";
  public static final String ADJUST_VOLUME = "ADJUST_VOLUME";
  public static final String PLAYSTREAM    = "PLAYSTREAM";
  public static final String BACK          = "BACK";

  // ids on elements (used for bindings)
  public static final String TITLE      = "TITLE";
  public static final String VOLUME     = "VOLUME";
  public static final String PLAYER     = "PLAYER";

  ViewElement player;

  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  /**
   * Get the tvt-pluginplayer functionalities.  That is, application and 
   * content handler for certain media types.
   */
  public String[] getServletFunctionalities ()
  {
    return new String[] { FUNCTION_CONTENT_HANDLER };
  }

  /**
   * Get the tvt-pluginplayer interfaces.
   */
  public String[] getServletInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	return new String[] { IViewableDocumentBuilder.INTERFACE };
      }
    return null;
  }

  public String[] getServletFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	//depends on registered plugin services, so this can be just a guess
	return new String[] { FEATURE_MIMETYPE, "video/mpeg", 
                              FEATURE_EXTENSION, "mpg",
                              FEATURE_MIMETYPE,  "application/x-shockwave-flash", 
                              FEATURE_EXTENSION, "swf" };
      }
    return null;
  }

  // -------------------------------------------------------------------------
  // org.xoe.core.content.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    ContentType type = content.getType ();
    Logger.log (this, type);
    XDocument doc =  getRelativeDocument("play0.xhtml");
    play (content, doc);

    //this view doesn't have control buttons, start right away
    player.start();
 
    return doc;
  }

  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    res.setContentType ("text/html");
    String path = req.getPathInfo ();

    //System.err.println( res + " path: " + path);
 
    if ("".equals (path) || OPEN.equals (path))
      {
	Document doc = getRelativeDocument("select.xhtml");
	DOMUtils.expandUsingXPathAttributes ( doc, null, null);

	res.getWriter ().print (doc);
      }
    else if (PLAYSTREAM.equals (path))
      {
	Document doc = getRelativeDocument("play.xhtml");
	String surl = req.getParameter( "value").trim();
	res.getWriter ().print (doc);
	play (ContentCache.fetch(surl), doc);

      }
    else if (STOP.equals (path)) 
      {
	player.stop();
      }
    else if (START.equals (path)) 
      {
	player.start();
      }
     else if (BACK.equals (path))
      {
 	Document doc = getRelativeDocument ("select.xhtml");
	DOMUtils.expandUsingXPathAttributes (doc, null, null);
	res.getWriter ().print (doc);
      }
     
  }

  private void play (ContentElement content, Document doc)
  {
    String uri = content.getURL().toString();
          
    player = ((Embed) doc.getElementById (PLAYER)).getPlugin();
    player.setAttributeNS ( null, "src", uri );
  }

} 
