/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Main.java,v 1.20 2001/11/07 20:53:23 tim Exp $
 */
package org.xoe.shell;

import java.awt.Container;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.GenericServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.Engine;
import org.xoe.core.IShell;
import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.content.Stash;
import org.xoe.core.device.IScreen;
import org.xoe.core.device.Device;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.Render;
import org.xoe.servlet.XoeServlet;
import org.xoe.servlet.XoeServletConfig;
import org.xoe.services.IIconResolver;


public class Main
  extends XoeServlet
  implements IShell
{
  private final static String [] funcs = new String [] { IShell.FUNCTION_SHELL };
  private final static String [] ifaces = new String [] { IShell.INTERFACE };
  private final static int COLS = 4;

  private Document maindoc;
  private Element programs;
  private Element last;
  private Vector lastapps = new Vector ();
  private Hashtable apps = new Hashtable ();
  private boolean lastappschanged = true;

  private Document appdoc;
  private DocumentViewer program;
  private boolean active = false;

  private byte[] beepsound;

  long lastMod = 0;


  // -------------------------------------------------------------------------
  // org.xoe.core.IShell

  public void start ()
  {
    IScreen screen = (IScreen) ServiceLocator.findImplementor (IScreen.INTERFACE);
    if (screen == null)
      {
        Logger.log (this, "*** ERROR *** No screen found");
        System.exit (1);
      }
    Container win = screen.getScreen ();
    Render render = new Render ();
    win.add (render);
    render.setDocument (getViewableDocument ());
    win.validate ();
    win.repaint ();
  }


  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  public void init (ServletConfig c)
    throws ServletException
  {
    maindoc = getRelativeDocument ("index.xhtml");
    // We won't find maindoc during packaging !!!! - BUG!!!!
    if (maindoc != null)
      {
	programs = maindoc.getElementById ("programs");
	last = maindoc.getElementById ("last");
	appdoc = getRelativeDocument ("app.xhtml");
	program = (DocumentViewer)appdoc.getElementById ("program");
      }
  }

  public void doGet (HttpServletRequest req, HttpServletResponse rep)
    throws ServletException, IOException
  {
    String path = req.getPathInfo ();
    if ("".equals (path))
      {
        beep ();

        String arg = req.getParameter ("application");
      
        rep.setContentType ("text/html");
        PrintWriter out = rep.getWriter ();
      
        if ("return".equals (arg))
          {
            if (active)
              {
                out.print (appdoc);
              }
            else
              {
                buildMain ();
                out.print (maindoc);
              }
          }
        else if ("main".equals (arg) || arg == null)
          {
             buildMain ();
             out.print (maindoc);
          }
        else
          {
	    AppInfo app = (AppInfo)apps.get (arg);
	    if (app.view == null)
	      {
		ContentElement elem = ContentCache.fetch (arg);
		if (elem == null)
		  {
		    Logger.log (this, "no content for " + arg);
		    return;
	    	  }
	        IViewableDocumentBuilder builder = (IViewableDocumentBuilder)elem.getType ().getService (IViewableDocumentBuilder.INTERFACE);
	        app.view = builder.getViewableDocument (elem);
	      }
	    // Insert icon into lastapp list
	    int idx = lastapps.indexOf (app.icon);
            if (idx != 0)
              {
		if (idx != -1)
		  {
		    lastapps.removeElementAt (idx);
		  }
		else if (lastapps.size () == 3)
		  {
		    lastapps.remove (2);
		  }
		lastapps.insertElementAt (app.icon, 0);
		lastappschanged = true;
		active = true;
	      }
	    program.setDocument (app.view);
            out.print (appdoc);
          }
      }        
    else
      {
        DocumentCache.fetchDocument (path);
      }
  }

  protected String [] getServletFunctionalities ()
  {
    return (funcs);
  }

  protected String [] getServletInterfaces (String func)
  {
    if (func.equals (IShell.FUNCTION_SHELL))
      {
	return (ifaces);
      }

    return (null);
  }

  protected String [] getServletFeatures (String func)
  {
    return (null);
  }


  // -------------------------------------------------------------------------
  // org.xoe.shell.Main

  /**
   * Build the document for the main icon screen so that someone can
   * select an application to execute.
   */
  private void buildMain ()
    throws IOException
  {
    ContentElement appDir = Stash.fetch ("stash:/home/default/apps");

    if (appDir.getLastModified ().getTime () != lastMod)
      {
        apps.clear ();

	Node row = null;
        // Remove data so we can replace it
        for (;;)
          {
            row = programs.getFirstChild ();
            if (row == null)
              {
                break;
              }

            // remove all the <td>'s from the <tr>
            DOMUtils.removeChildNodes (row);

            // remove the <tr>
            programs.removeChild (row);
          }
    
        NodeList stuff = appDir.getChildNodes ();
        int len = stuff.getLength ();
        for (int i = 0; i < len; i++)
          {
            ContentElement thing = (ContentElement)stuff.item (i);

            IIconResolver ir = (IIconResolver)thing.getType ().getService (IIconResolver.INTERFACE);
            if (ir == null)
              {
                Logger.log (this, "buildMain: icon resolver not found for " + thing.getURL ());
                continue;
              }
 
            if (!MIMETYPE_STASHDIR.equals (thing.getType ().getMimetype ()))
              {
		IDocumentBuilder db = (IDocumentBuilder)thing.getType ().getService (IDocumentBuilder.INTERFACE);
		if (db == null)
		  {
		    Logger.log (this, "buildMain: document builder not found for " + thing.getURL ());
		    continue;
		  }

		Document doc = db.getDocument (thing);
		if (doc == null)
		  {
		    Logger.log (this, "buildMain: document builder returned null for " + thing.getURL ());
		    continue;
		  }

		String name = doc.getDocumentElement ().getAttribute (NAME_ATTR);
		AppInfo ai = new AppInfo (name, ir.getIconURL (thing), thing.getURL ().toString ());
		apps.put (thing.getURL ().toString (), ai);
              }
          }

        int appCount = 0;
        for (Enumeration e = apps.elements (); e.hasMoreElements (); appCount++)
          {
            if ((appCount % COLS) == 0)
              {
                row = programs.appendChild (maindoc.createElementNS(XHTML_NAMESPACE, "tr"));
              }

            AppInfo ai = (AppInfo)e.nextElement ();
            row.removeChild (ai.cell);
            row.appendChild (ai.cell);
          }

        lastMod = appDir.getLastModified ().getTime ();
      }

    if (lastappschanged)
      {
        for (;;)
          {
            Node child = last.getFirstChild ().getNextSibling ();
            if (child == null)
              {
                break;
              }
            last.removeChild (child);
          }

        for (int i = 0; i < lastapps.size (); i++)
          {
            last.appendChild ((Node)lastapps.elementAt (i));
          }

        lastappschanged = false;
      }
  }

  private class AppInfo
  {
    Element cell;
    Element icon;
    Document view;

    AppInfo (String name, URL img, String link)
    {
      cell = maindoc.createElementNS(XHTML_NAMESPACE, "td");
      cell.setAttribute ("width", "" + (100 / COLS) + "%");
      cell.setAttribute ("halign", "center");
      cell.setAttribute ("valign", "top");

      Element anchor = (Element)cell.appendChild (maindoc.createElementNS(XHTML_NAMESPACE, "a"));
      anchor.setAttribute ("href", "?application=" + URLEncoder.encode (link));

      icon = maindoc.createElementNS(XHTML_NAMESPACE, "img");
      icon.setAttribute ("src", img.toString ());
      icon.setAttribute ("height", "32");
      icon.setAttribute ("width", "32");
      icon.setAttribute ("onclick", "?application=" + URLEncoder.encode (link));
      Element fnt = maindoc.createElementNS(XHTML_NAMESPACE, "font");
      fnt.setAttribute ("color", "#00000000");
      fnt.appendChild (icon.cloneNode (true));

      anchor.appendChild (fnt);
      anchor.appendChild (maindoc.createElementNS(XHTML_NAMESPACE, "br"));
      anchor.appendChild (maindoc.createTextNode (name));
    }
  }

  private void beep ()
  {
    if (beepsound == null)
      {
        String s = Engine.getConfig (BEEP_PROPERTY, "file:/etc/beep.raw");
        beepsound = new byte[0];
        try
          {
            ContentElement elem = ContentCache.fetch (s);
            if (elem != null)
              {
                beepsound = new byte[elem.getContentLength ()];
                elem.getInputStream ().read (beepsound);
              }
          }
        catch (IOException e)
          {
            Logger.log (this, e);
          }
      }
//    Device.beep (beepsound, 0, beepsound.length);
  }

}

