package org.xoe.web; 

import java.awt.Dimension;
import java.io.IOException;
import java.net.URL;
import java.net.URLEncoder;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.EntityResolver;
import org.xoe.core.dom.Loadable;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceFactory;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IEditableDocumentBuilder;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.servlet.XoeServlet;

class Bookmarks
  extends XoeServlet
implements IViewableDocumentBuilder, IDocumentBuilder, IEditableDocumentBuilder
{  
  public static final String BOOKMARKS_PREFIX = "bookmarks";
  public static final String BOOKMARKS_NS = "http://xoe.org/ns/bookmarks/2001";
  public static final String DOCTYPE_NAME = "bookmarks";
  public static final String DOCTYPE_NAME_NS = BOOKMARKS_NS + NAMESPACE_SEPARATOR + DOCTYPE_NAME;

  public XDocument getDocument (ContentElement content)
    throws IOException
  {
    ContentType type = new ContentType ("text/xml");
    IDocumentBuilder db = (IDocumentBuilder)type.getService (IDocumentBuilder.INTERFACE);
    if (db != null)
      {
	return db.getDocument (content);
      }
    Logger.log (this, "No IDocumentBuilder for text/xml!!");
    return null;
  }

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    XDocument dataDoc = getDocument (content);
    XDocument doc = getRelativeDocument ("bookmarks.xhtml");
    DOMUtils.expandUsingRepeat (doc, dataDoc, null);
    doc.setURL (content.getURL ());
    return doc;
  }

  public XDocument getEditableDocument (ContentElement content)
    throws IOException
  {
    Bookmarks bk = new Bookmarks ();     // does "s" need to be inited with the same configDoc as "this" ?


    // set up local data
    bk.dataDoc = getDocument (content);
    bk.setCurrentCategory ();
    
    // construct frame
    XDocument doc = bk.getViewableDocument ();

    return doc;
  }

  public void save (XDocument doc)
  {}

  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  public void init (Document configDoc) 
    throws Exception
  {
    super.init (configDoc);
  }

  protected String [] getServletFunctionalities ()
  {
    return (new String[] { FUNCTION_CONTENT_HANDLER });
  }

  protected String [] getServletInterfaces (String func)
  {
    if (func.equals ( FUNCTION_CONTENT_HANDLER ))
      {
        return (new String[] { IViewableDocumentBuilder.INTERFACE, IDocumentBuilder.INTERFACE, IEditableDocumentBuilder.INTERFACE });
      }

    return (null);
  }

  protected String [] getServletFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
        return (new String[] { 
	  FEATURE_DOCTYPE_NAME_NS, DOCTYPE_NAME_NS, 
	    FEATURE_EXTENSION, DOCTYPE_NAME
	    });
      }

    return (null);
  }


  XDocument dataDoc;
  Element currentCategory;
  XoeServlet categoriesPage;

  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    String path = req.getPathInfo ();
    Logger.log (this, "path = " + path);
    grabDataFromScreen ();

    String category = req.getParameter ("category");
    if (category != null)
      {
	currentCategory = getCategoryNode (category);
      }

    if (req.getParameter ("addnew") != null)
      {

	currentCategory.appendChild (dataDoc.createElementNS (null, "link"));
      }

    if (req.getParameter ("delete") != null)
      {
	NodeList delete = XPath.findNodes (currentCategory, "link[ count(text ()) = 0 or text () = '' or count (@url) = 0 or @url= '']");
	for (int i=0; i < delete.getLength (); i++)
	  {
	    currentCategory.removeChild (delete.item(i));
	  }
      }

    if (req.getParameter ("revert") != null)
      {
	dataDoc = DocumentCache.fetchDocument (dataDoc.getURL ());
	setCurrentCategory ();
      }

    if ("".equals (path) || "redraw".equals (path) || "bookmarks".equals (path))
      {
	if (getTitleDocument () == null)
	  {
	    XDocument title = getRelativeDocument ("edit-bookmarks-title.xhtml");
	    DOMUtils.expandUsingRepeat (title, dataDoc, null);
	    // adjust select box so that the last category is selected
	    Node option = XPath.findFirstNode (title, "//xhtml:option[text () = '" + getCurrentCategoryName () + "']");
	    if (option != null)
	      {
		((Element) option).setAttribute ("selected", "selected");
	      }
	    setTitleDocument (title);
	  }
	XDocument doc = getRelativeDocument ("edit-bookmarks.xhtml");
	DOMUtils.expandUsingRepeat (doc, currentCategory, null);
	res.getWriter ().print (doc);
	createCallbacks ();
      }
    else if ("save".equals (path))
      {
	DOMUtils.setTextContent (XPath.findFirstNode (dataDoc, "//lastcategory"), getCurrentCategoryName ());
	DocumentCache.putDocument (dataDoc);
      }    
    else if ("categories".equals (path))
      {
	try 
	  {
	    if (categoriesPage == null)
	      {
		categoriesPage = new Categories (this);
		categoriesPage.init ((Document) null);
	      }
	    categoriesPage.service (req, res);
	  }
	catch (Exception e)
	  {
	    Logger.log (this, e);
	  }
      }
  }

  Element getCategoryNode (String category)
  {
    return (Element) XPath.findFirstNode (dataDoc, "*/category[@name = '" + category + "']");
  }

  void setCurrentCategory ()
  {
    String category = XPath.findString (dataDoc, "//lastcategory");
    currentCategory = getCategoryNode (category);
  }
  
  String getCurrentCategoryName ()
  {
    return currentCategory.getAttributeNS (null, "name");
  }

  void grabDataFromScreen ()
  {
    Document disp = getDisplayedDocument ();
    if (disp != null && currentCategory  != null)
      {
	NodeList names = XPath.findNodes (disp, "//*[@id = 'name']");
	NodeList urls = XPath.findNodes (disp, "//*[@id = 'url']");
	NodeList links = XPath.findNodes (currentCategory, "link");
	int length = names.getLength (); // all the lengths should be the same
	for (int i = 0; i < length ; i++)
	  {
	    Element link = (Element) links.item (i);
	    link.setAttributeNS (null, "url", ((Element) urls.item(i)).getAttributeNS (null, "value"));
	    DOMUtils.setTextContent (link, ((Element) names.item(i)).getAttributeNS (null, "value"));
	  }
      }
  }

}
