/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Main.java,v 1.27 2001/12/17 23:36:52 erik Exp $
 */ 
package org.xoe.web;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.HttpURLConnection;
import java.net.URLConnection;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.xoe.core.dom.XElement;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentElementFactory;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XNode;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IEditableDocumentBuilder;
import org.xoe.display.dom.ImageFactory;
import org.xoe.display.dom.ViewElement;
import org.xoe.display.dom.xhtml.Anchor;
import org.xoe.display.dom.xhtml.Form;
import org.xoe.services.IHistory;
import org.xoe.services.IURLViewer;
import org.xoe.servlet.XoeServlet;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;

public class Main
  extends XoeServlet
{
  private IURLViewer viewer;
  private DocumentViewer page;
  private Element addr;
  private IHistory history;
  private String homepage;
  private String currentPage;
  private XDocument index = null;
  private XDocument normalTitle, fullScreenTitle;



  private Callback webcallback = new Callback ()
   {
     public void invoke (Node node, Object payload)
     {
       Logger.log (this, "invoke: webcallback");
//        Logger.log (this, "node = " + node);
//        Logger.log (this, "payload = " + payload);
       try
         {
	   stopLoading ();
           // resolve the payload as a string url relative to the url of 
           // the document that owns the node that invoked this callback.
	   String oldURL = ((XDocument) node.getOwnerDocument ()).getURL ().toString ();
	   String newURL = (String)payload;

	   String oldProto = oldURL.substring (0, oldURL.indexOf (":"));

	   int i = newURL.indexOf (":");
	   if (i > -1 && !newURL.substring (0, i).equals (oldProto))
	     { // different protocol, don't build relative url
	       webClick (newURL);
	     }
	   else
	     {
	       webClick (new URL (((XDocument) node.getOwnerDocument ()).getURL (), newURL).toString ());
	     }
         }
       catch (MalformedURLException mue)
         {
           Logger.log (this, mue);
         }
     }
   };


  private Callback formcallback = new Callback ()
   {
     public void invoke (Node node, Object payload)
     {
       Logger.log (this, "invoke: formcallback");
       Logger.log (this, "node = " + node);
       Logger.log (this, "payload = " + payload);
       if (node instanceof Form && "POST".equalsIgnoreCase (((Form) node).getAttributeNS (null, "method")))
	 {
	   try
	     {
	       stopLoading ();
	       String u = (String)payload;
	       String body = "";
	       int idx = u.indexOf ('?');
	       if (idx != -1)
		 {
		   body = u.substring (idx);
		   u = u.substring (0, idx);
		 }
	       URL url = new URL (((XDocument) node.getOwnerDocument ()).getURL (), u);
	       URLConnection con = url.openConnection ();
	       ((HttpURLConnection) con).setRequestMethod ("POST");
	       con.setRequestProperty ("Content-Type", "application/x-www-form-urlencoded");
	       con.setDoOutput (true);
	       con.getOutputStream ().write (body.getBytes ());
	       ContentElement content = ContentCache.getFactory ().createContentElement (url, con.getInputStream ());
	       XDocument doc = viewer.getPage (content);
	       page.setDocument (doc, url);
	     }
	   catch (MalformedURLException mue)
	     {
	       Logger.log (this, mue);
	     }	   
	   catch (IOException e)
	     {
	       Logger.log (this, e);
	     }
	 }
       else 
	 {
	   webcallback.invoke (node, payload);
	 }
     }
    };


  private  Thread updatingThread;

  public void init (ServletConfig config) 
    throws ServletException
  {
    // look up the web browser service to use for getting web pages
    viewer = (IURLViewer) ServiceLocator.findImplementor (IURLViewer.INTERFACE);

    if (viewer == null)
      {
        throw new ServletException ("no web browser service found");
      }
    super.init (config);
    this.homepage = getDataDir () + "default.bookmarks";
    try 
      {
	if (Stash.fetch (this.homepage) == null)
	  {
	    XDocument doc = getRelativeDocument ("default.bookmarks");
	    doc.setURL (new URL (this.homepage));
	    DocumentCache.putDocument (doc);
	  }
      }
    catch (Exception e)
      {
	Logger.log (this, e);
      }
  }

 
  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    String path = req.getPathInfo ();
    Logger.log (this, "path = " + path);

    if (index == null)
      {
	index = getRelativeDocument ("index.xhtml");
	normalTitle = getRelativeDocument ("title.xhtml");
	fullScreenTitle = getRelativeDocument ("title-fullscreen.xhtml");
	
	this.addr = normalTitle.getElementById ("addr");
	XNode pagenode = (XNode)index.getElementById ("page");
	pagenode.setCallback ("href", webcallback);
	pagenode.setCallback ("action", formcallback);
	page = (DocumentViewer)pagenode;
	history = (IHistory) ServiceLocator.findService (historyServiceQuery);
	history.trackHistory (page, false);
      }

    if ("".equals (path))
      {
	webClick (this.homepage);
	res.setContentType ("text/html");
	res.getWriter ().print (index);
	setTitleDocument (normalTitle);
      }
    else if ("load".equals (path))
      {
        webClick (req.getParameter ("addr"));
	this.currentPage = req.getParameter ("addr");
      }
    else if ("buttons".equals (path))
      {	
	 path= req.getParameter ("cmd");
	 Logger.log (this, "path = " + path);
 	 
	 if ("back".equals (path))
	   {
	     history.back (page);
	     this.addr.setAttribute ("value", (page.getURL ().toString ()));
	   }
 	 else if ("forward".equals (path))
	   {
	     history.forward (page);
	     this.addr.setAttribute ("value", (page.getURL ().toString ()));
	   }
	 else if ("stop".equals (path))
	   {
	      stopLoading ();
	   }
	 else if ("home".equals (path))
	   {
	      webClick (this.homepage);
	   }
	 else if ("edit".equals (path))
	   {

	     String src = req.getParameter ("addr");
	     System.err.println ("IN EDIT: addr parameter = [" + src + "]");
	     
	     if (src != null)
	       {
		 edit (new URL (getCurrentURL(), src));
	       }
	     else
	       {
		 edit (getCurrentURL ());
	       }
	   }
	 else if ("settings".equals (path))
	   {	  	     
	     XDocument doc = getRelativeDocument ("options.xhtml");
	     XElement settingsEl = getLocalUserConfig ();
	     DOMUtils.expandUsingXPathAttributes (doc, settingsEl, null );	    

	     this.addr.setAttribute ("value", doc.getURL ().toString ());
	     System.err.println( "OPTIONS URL: [" + doc.getURL ().toString () + "]" );
	     history.forward (this.page, doc, doc.getURL ());

	     //webClick (getRelativeSrc ("options.xhtml"));	    
	   }
	 else if ("set_current_page".equals (path))
	   {	  
	     //System.err.println ("current page = [" + this.currentPage + "]");

	     XDocument configDoc = (XDocument)getLocalUserConfig().getOwnerDocument();
	     DOMUtils.setTextContent ( XPath.findFirstNode( configDoc, "//homepage/text()" ) , this.currentPage );
	     saveLocalUserConfig (configDoc);
	     this.homepage = this.currentPage;

	     XDocument doc = getRelativeDocument ("options.xhtml");
	     XElement settingsEl = getLocalUserConfig ();
	     DOMUtils.expandUsingXPathAttributes (doc, settingsEl, null );	    

	     this.addr.setAttribute ("value", doc.getURL ().toString ());
	     //System.err.println( "OPTIONS URL: [" + doc.getURL ().toString () + "]" );
	     history.forward (this.page, doc, doc.getURL ());	   	    	    
	   }
	 else if ("set_default_page".equals (path))
	   {	  
	     XDocument configDoc = (XDocument)getLocalUserConfig().getOwnerDocument();
	     String defaultPage = XPath.findFirstNode( configDoc, "//default/text()" ).getNodeValue().trim();
	     DOMUtils.setTextContent ( XPath.findFirstNode( configDoc, "//homepage/text()" ) , defaultPage );
	     saveLocalUserConfig (configDoc);
	     this.homepage = defaultPage;

	     XDocument doc = getRelativeDocument ("options.xhtml");
	     XElement settingsEl = getLocalUserConfig ();
	     DOMUtils.expandUsingXPathAttributes (doc, settingsEl, null );	    

	     this.addr.setAttribute ("value", doc.getURL ().toString ());
	     //System.err.println( "OPTIONS URL: [" + doc.getURL ().toString () + "]" );
	     history.forward (this.page, doc, doc.getURL ());	   	    	     
	   }


	 else if ("zoom".equals (path))
	   {
	     setTitleDocument (fullScreenTitle);
	   }
	 else if ("normal".equals (path))
	   {
	     setTitleDocument (normalTitle);
	   }
	 else if ("reload".equals (path))
	   {
	     XDocument last = (XDocument) page.getDocument ();
	     if (last != null)
	       {
		 URL url = last.getURL ();
		 if (url != null)
		   {
		     String  sUrl = url.toString ();
		     int refIdx = sUrl.lastIndexOf ('#');
		     sUrl = refIdx > -1 ? sUrl.substring( 0, refIdx) : sUrl;
		     ContentCache.clear (sUrl);
		     XDocument ndoc = viewer.getPage (sUrl);
		     page.setDocument (ndoc, page.getURL ());
		   }
	       }
	   }
	 else
	   {
	      Logger.log (this, " *** NOT IMPLEMENTED ***");
	   }
      }
  }



  private void webClick (String src)
  {
    updatingThread = Thread.currentThread ();

    XDocument last = (XDocument) page.getDocument ();
    XDocument ndoc = null;
    String ref = null;
     
    if (src.indexOf (':') == -1)  //if there is no protocol declaration, give it one.
      {
	src = "http://" +  src ;
      }
    
    int refIdx = src.lastIndexOf ('#');
    String srcNoref = refIdx > -1 ? src.substring( 0, refIdx) : src;
    try 
      {
	URL srcURL = new URL (src);
	
	if (last != null) 
	  {
	    String lastSrc = last.getURL ().toString ();
	    if ( srcNoref.equalsIgnoreCase (lastSrc))
	      {
		ndoc = last;
		history.forward (page, ndoc, srcURL); 
		this.addr.setAttribute ("value", src);
	      }
	  }
	
	if ( ndoc == null ) 
	  {    
	    ndoc = viewer.getPage (srcNoref);
	    if (ndoc != null)
	      {
		if (!ndoc.getURL ().toString ().equalsIgnoreCase (srcNoref))
		  {
		    srcURL = ndoc.getURL ();
		  }
		history.forward (page, ndoc, srcURL);
		this.addr.setAttribute ("value", srcURL.toString ());
	      }
	  }
      }
    catch (MalformedURLException e)
      {
	Logger.log (this, e);
      }
    catch (IOException ioe)
      {
	page.setDocument (ndoc);
	history.forward (page, ndoc, null);
	 
	//removed until figured out the URL NULL handler bug which
	//holds back fgl from displaying a package query document
//	addr.setAttribute ("value", ndoc.getURL ().toString ());

	Logger.log (this, "webClick: src = " + src);
	Logger.log (this, ioe);

      }
    updatingThread = null;
  }

  private URL  getCurrentURL ()
  {
    Document last = page.getDocument ();
    URL srcURL = ((XDocument) last).getURL ();
    return srcURL;
  }

  private void edit (URL srcURL)
  {
    if (srcURL == null)
      {
	Logger.log (this, "last.getURL () = null");
	return;
      }
    String src = srcURL.toString ();
    Logger.log (this, src);
    XDocument ndoc = null;
    try
      {
	if (src.indexOf (':') == -1)
	  {
	    src = getRelativeSrc (src);
          }
	
	ContentElement content = ContentCache.fetch ( new URL (src));

	// get it's type
	ContentType type = content.getType ();
	Logger.log (this, type);
	// is there a editable document builder for this type?
	IEditableDocumentBuilder edb = (IEditableDocumentBuilder)type.getService(IEditableDocumentBuilder.INTERFACE);
	if (edb == null)
	  {
	    type = new ContentType ("text/xml");
	    edb = (IEditableDocumentBuilder)type.getService(IEditableDocumentBuilder.INTERFACE);
	    Logger.log (this, "no editable document builder service found for " + src);
	  }
	page.setDocument (edb.getEditableDocument (content));
	history.forward (page, page.getDocument (), null);
      }
    catch (IOException ioe)
      {
        Logger.log (this, "webClick: src = " + src);
        Logger.log (this, ioe);
      }
  }

  public void stopLoading ()
  {
    if (updatingThread != null && updatingThread != Thread.currentThread ())
      {
	updatingThread.interrupt ();
      }
    ImageFactory.stop ();
  }


  /**
   * This method will always attempt to retrieve user data from the stash
   * This data WILL NOT be cached, so any sync-ing of the stash will become
   * immediately apparent.  If for some reason the stash does not contain
   * user data, this method will look to the relative file config.xml for data
   *
   */
  private XElement getLocalUserConfig ()
  {
    XElement userConfig = null;	
    XDocument doc = DocumentCache.fetchDocument ("stash:/home/default/data/tvt-web/userConfig.xml");		
    if( doc != null )
      userConfig = (XElement)XPath.findFirstNode (doc, "config");
    
    if ( userConfig == null)
      {
	//Logger.log( this, "Main.getLocalUserConfig(): fetch from stash is null.  Retrieving from package/config.xml..." );
	XElement originalConfig = (XElement)DocumentCache.fetchDocument (getRelativeSrc("config.xml")).getDocumentElement();	
	
	try
	  {
	    //now copy this info to the stash....?	    
	    URL userConfigURL = new URL( "stash:/home/default/data/tvt-web/userConfig.xml" );
	    String userConfigString = XMLWriter.getXML( originalConfig );
	    //Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );	    
	    ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
	    Stash.put( userConfigCE );
	    return originalConfig;
	  }
	catch( IOException ioe )
	  {
	    Logger.log( this, ioe.toString() );
	    return null;
	  }      
      }
    return userConfig;
  }


  private void saveLocalUserConfig (XDocument configDoc)
    throws IOException, MalformedURLException
    {      
      URL userConfigURL = new URL( "stash:/home/default/data/tvt-web/userConfig.xml" );
      String userConfigString = XMLWriter.getXML( configDoc );     
      //Logger.log( this, "Main.SAVE: userConfig.xml = [" + userConfigString + "]" );      
      ContentElement userConfigCE = Stash.getFactory().createContentElement( userConfigURL, userConfigString.getBytes() );
      Stash.put( userConfigCE );
    }



}
