/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Main.java,v 1.10 2001/12/14 23:39:05 suraj Exp $
 */
package org.xoe.urlviewer;

import java.awt.Component;
import java.awt.Window;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.lang.InterruptedException;
import java.net.URL;
import java.net.UnknownHostException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DefaultContentTypeResolver;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.IPackageQuery;
import org.xoe.core.packages.InstallOption;
import org.xoe.core.packages.PackageDescription;
import org.xoe.core.packages.PackageQuery;
import org.xoe.core.packages.PackageManager;
import org.xoe.core.packages.SearchResults;
import org.xoe.core.protocol.IProtocolHandlerFactory;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.services.IService;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.xhtml.XHTMLDocument;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.display.dom.xhtml.IFrame;
import org.xoe.services.IPackageSelector;
import org.xoe.services.IURLViewer;


public class Main
  implements IURLViewer
{
  public static final XDocument PAGE_NOT_FOUND_404;
  public static final ContentType NULL_TYPE = new ContentType ();

  private static final SimpleServiceQuery PACKAGE_SELECTOR_QUERY = new SimpleServiceQuery ().setInterface (IPackageSelector.INTERFACE);

  static
  {
    PAGE_NOT_FOUND_404 = DOMlet.create (("<html xmlns=\"" + XHTML_NAMESPACE + "\"><head><title>404 - Page Not Found</title></head><body><h2>404 - Page Not Found</h2></body></html>").toCharArray (), "404");
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = { IURLViewer.FUNCTIONALITY };
    return (ret);
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals(IURLViewer.FUNCTIONALITY))
      {
	String[] ret = { IURLViewer.INTERFACE };
	return ret;
      }
    return (null);
  }

  public String[] getFeatures (String func)
  {
    return (null);
  }

  public String[] getInformation ()
  {
    return (null);
  }

  public XDocument getPage (String src)
    throws IOException
  {
    // check to see if this protocol is supported
    String protocol = src.substring (0, src.indexOf(":"));

    SimpleServiceQuery query = new SimpleServiceQuery (IProtocolHandlerFactory.FUNCTIONALITY, 
                                                       IProtocolHandlerFactory.INTERFACE);
    query.addFeature (IProtocolHandlerFactory.FEATURE, protocol);
    IService fac = ServiceLocator.findService (query);

    if (fac == null) 
      {
	return getViewablePackageQuery (src, protocol);
      }

    return (getPage (new URL (src)));
  }

  /**
   *  Returns the web page specified by the url passed in.  
   *  Returns null for various non-fatal errors, i.e. host not found.
   */
  public XDocument getPage (final URL url)
    throws IOException
  {
    if (url == null)
      {
        return (null);
      }

    // fetch the content
    ContentElement content = ContentCache.fetch (url);
    return getPage (content);
  }

  public XDocument getPage (ContentElement content)
    throws IOException
  {
    URL url = content.getURL ();
    if (content == null)
      {
	Logger.log (this, "getPage: content = null");
	return PAGE_NOT_FOUND_404;
      }
    try 
      {
	// get it's type
	ContentType type = content.getType ();
	if (NULL_TYPE.equals (type))
	  {
	    try 
	      {
		url.openConnection ().getInputStream ();
		Dialog.note ("<br />Page contains no data.<br />");
	      }
	    catch (UnknownHostException uhe)
	      {
		Dialog.note ("<br />Unable to locate the server " + url.toString () + ".<br /><br />Please check the server name and try again.<br />");
	      }
	    catch (Exception e)
	      {
		Dialog.note ("<br />" + e.getMessage () + "<br />");
	      }
	    return null;
	  }

	// is there a viewable document builder for this type?
	Logger.log (this, type.getService(IViewableDocumentBuilder.INTERFACE));
	IViewableDocumentBuilder vdb = (IViewableDocumentBuilder)type.getService(IViewableDocumentBuilder.INTERFACE);
	if (vdb == null)
	  {
	    Logger.log (this, "getPage: no viewable document builder service found for " + url);
	    return getViewablePackageQuery (url, type);   
	  }
	try 
	  {
	    return (vdb.getViewableDocument (content));
	  }
	catch (java.net.SocketException e)
	  {
	    Logger.log (this, "SocketException");
	  }
      }
    catch (InterruptedIOException e)
      {
	Logger.log (this, "InterruptedIOException");
      }
    return null;
  }

  private XDocument getViewablePackageQuery (final URL url, ContentType type)
  {
    final XDocument doc;

    IPackageSelector ps = (IPackageSelector) (ServiceLocator.findService (PACKAGE_SELECTOR_QUERY));
    
    if (ps == null)
      {
	Dialog.note ("We cannot currently display content where the <b>" 
                     + type.toExternalForm () 
                     + "</b><br />Try installing a Package Selector service to locate a content handler for this type.");
	return null;
      }
	
    Logger.log (this, Thread.currentThread ());
    Logger.log (this, "Thread.interrupted = " + Thread.interrupted ());

    boolean conf = Dialog.confirm ("We cannot currently display content where the <b>" 
                                        + type.toExternalForm () 
                                        + "</b>. <br />Should we look on the server for " 
                                        + " a package that supports this content type?");
    if (conf)
      {
	IPackageQuery pq = type.getPackageQuery (IViewableDocumentBuilder.INTERFACE);
	Logger.log (this, "getPage: Trying to find a package using query (" + pq.toXPath() + ")");
	SearchResults results = PackageManager.search (pq);
	InstallOption[] options = results.getInstallablePackages ();
	Logger.log(this, "getPage: results:\n" + InstallOption.toString(options));

	if (options == null || options.length == 0)
	  {
	    Dialog.note ("There are no appropriate software packages on the server at this time.");
	    return null;
	  }

	
	XDocument viewDoc = new XHTMLDocument ();
	viewDoc.setURL (url);
	Node body = viewDoc.appendChild (viewDoc.createElement ("body"));
	final Element iframe = (Element) body.appendChild (viewDoc.createElement ("iframe"));
	iframe.setAttribute ("width", "100%");
	iframe.setAttribute ("height", "100%");


	doc = (XDocument) ps.showSearchResults("Pick a package", 
                                               "Choose a package, and then hit \"Install\"", 
					       results);

	((DocumentViewer) iframe).setDocument (doc);
	doc.setCallback ("success", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      try 
		{ 
		  XDocument m = getPage(url);
		  ((DocumentViewer) iframe).setDocument (m);
		}
	      catch  (IOException e)
		{
		  Logger.log(this, e);
		}
	    }
	  });
	doc.setCallback ("failure", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      Logger.log (this, "failure");
	    }
	  });
	return (viewDoc);
      }

    return (null);
  }

  private XDocument getViewablePackageQuery (final String url, String protocol)
  {
    final XDocument doc;

    IPackageSelector ps = (IPackageSelector) (ServiceLocator.findService (PACKAGE_SELECTOR_QUERY));
    
    if (ps == null)
      {
	Dialog.note ("We cannot currently access urls with protocol <b>" 
                     + protocol
                     + "</b>.<br />Try installing a Package Selector service to locate a protocol handler for this type.");
	return null;
      }
	
    boolean conf = Dialog.confirm ("We cannot currently display access urls with protocol <b>" 
                                        + protocol
                                        + "</b>. <br />Should we look on the server for " 
                                        + " a package that supports this protocol?");
    if (conf)
      {
	PackageQuery pq = new PackageQuery ();
	pq.setFunctionality (IProtocolHandlerFactory.FUNCTIONALITY);
	pq.setInterface (IProtocolHandlerFactory.INTERFACE);
	pq.addFeature (IProtocolHandlerFactory.FEATURE, protocol);

	Logger.log (this, "getPage: Trying to find a package using query (" + pq.toXPath() + ")");
	SearchResults results = PackageManager.search (pq);
	InstallOption[] options = results.getInstallablePackages ();
	Logger.log(this, "getPage: results:\n" + InstallOption.toString(options));

	if (options == null || options.length == 0)
	  {
	    Dialog.note ("There are no appropriate software packages on the server at this time.");
	    return null;
	  }

	
	XDocument viewDoc = new XHTMLDocument ();
	Node body = viewDoc.appendChild (viewDoc.createElement ("body"));
	final Element iframe = (Element) body.appendChild (viewDoc.createElement ("iframe"));
	iframe.setAttribute ("width", "100%");
	iframe.setAttribute ("height", "100%");


	doc = (XDocument) ps.showSearchResults("Pick a package", 
                                               "Choose a package, and then hit \"Install\"", 
					       results);

	((DocumentViewer) iframe).setDocument (doc);
	doc.setCallback ("success", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      try 
		{ 
		  XDocument m = getPage(url);
		  ((DocumentViewer) iframe).setDocument (m);
		}
	      catch  (IOException e)
		{
		  Logger.log(this, e);
		}
	    }
	  });
	doc.setCallback ("failure", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      Logger.log (this, "failure");
	    }
	  });
	return (viewDoc);
      }

    return (null);
  }
}
