/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: Tag.java,v 1.2 2001/12/07 21:10:02 damian Exp $
 */

package org.xoe.input.virtual;

class Tag 
{
  private int firstx;
  private int firsty;
  private int lastx;
  private int lasty;
  private int startx;
  private int starty;
  private byte nsdir;
  private byte ewdir;
  private int patternval;
  private final int dist;
  private final int edist;

  protected int leftx;
  protected int rightx;

  /**
   * Create a Tag handwriting pen which listens to mouse
   * events and dispatches to the current pen focus.
   */
  public Tag (int padsize) 
  {
    dist = padsize / 8;
    edist = padsize / 4;
  }

  /**
   * Start the recognition process at these coordinate.
   */
  public void start(int x, int y) 
  {
    lastx = x;
    lasty = y;
    startx = x;
    starty = y;
    firstx = x;
    firsty = y;
    nsdir = -1;
    ewdir = -1;
    patternval = 1;
    leftx = x;
    rightx = x;
  }

  /**
   * Move the pen to these coordinates.
   */
  public void move (int x, int y) 
  {
    if (x < leftx ) { leftx  = x; }
    if (x > rightx) { rightx = x; }

    int xd = x - startx;
    int yd = starty - y;

    if (Math.abs(xd) <= dist && Math.abs(yd) <= dist) 
      {
        //System.out.println("x="+x+",y="+y);
        return;
      }

    byte newdir = -1;
    if (yd > dist) 
      {
        newdir = TagPatterns.NORTH;
      }
    else if (yd < -dist) 
      {
        newdir = TagPatterns.SOUTH;
      }


    if (newdir != -1) 
      {
        if (nsdir != newdir) 
          {
            if (nsdir != -1) 
              {
                ewdir = -1;
              }
            nsdir = newdir;
            patternval = (patternval << 2) | newdir;
            lasty = y;
            //System.out.println("x="+x+",y="+y+" " + (newdir == TagPatterns.NORTH ? "N" : "S"));
          }
        starty = y;
      }

    newdir = -1;
    if (xd > dist) 
      {
        newdir = TagPatterns.EAST;
      }
    else if (xd < -dist) 
      {
        newdir = TagPatterns.WEST;
      }

    if (newdir != -1) 
      {
        if (ewdir != newdir) 
          {
            if (ewdir != -1) 
              {
                nsdir = -1;
              }
            ewdir = newdir;
            patternval = (patternval << 2) | newdir;
            lastx = x;
            //System.out.println("x="+x+",y="+y+" " + (newdir == TagPatterns.EAST ? "E" : "W"));
          }
        startx = x;
      }
  }

  /**
   * Pen ends - work out the key value and return it.
   */
  public char end(int x, int y) 
  {
    // Nothing written
    if (firstx == -1) 
      {
        return (0);
      }

    // When we release the pen we see if the starting coordinate are
    // near the ending coordinate.
    int xd = Math.abs(x - firstx);
    int yd = Math.abs(y - firsty);
    if (xd < edist && yd < edist) 
      {
        // Okay, we're very near where we started - we encode this
        // by negating the patternval.
        patternval = -patternval;
      }
    //System.out.println("xd="+xd+", yd="+yd+", pat="+patternval);

    int area = selectArea();
    char key = TagPatterns.recognizeCharacter(patternval, area);
    if (key == 0) 
      {
        key = TagPatterns.recognizeCharacter(patternval, TagPatterns.COMMON);
      }

    /* The following is a hack for backwards compatibility with the
     * current patterns.  This should be removed once the Tag pen
     * system has been retaught what to recognize.
     */
    if (key == 0 && patternval < 0) 
      {
        key = TagPatterns.recognizeCharacter(-patternval, area);
        if (key == 0) 
          {
            key = TagPatterns.recognizeCharacter(-patternval, TagPatterns.COMMON);
          }
      }

    firstx = -1;
    firsty = -1;

    return (key);
  }

  /*
   * Use the start and end coordinate to select which area of the grafitti
   * pad we're using.
   */
  protected int selectArea() 
  {
    return (TagPatterns.LOWERCASE);
  }
}
