package org.xoe.input.virtual;

import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.TextComponent;
import java.awt.Toolkit;
import java.io.InputStream;
import java.net.URL;

import kaffe.awt.AWT;
import kaffe.awt.HookInstaller;
import kaffe.awt.MouseHook;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.device.IScreen;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;

import org.xoe.services.IVirtualInput;

public class VirtualKeyboard
  extends Component
  implements IService, IVirtualInput, MouseHook
{

  private String language;
  private String locale;
  private VirtualKeymap[] keymap;
  private MouseHook mouseHook;
  private boolean attached;
  private boolean active;
  private Container screen;
  private int current;
  private Component focus;
  private Graphics gpen;
  protected Tag tpen;
  protected int last;
  protected boolean nextShift;
  protected boolean upperCase;
  private int oldCode;
  private Dimension offset = new Dimension (0, 0);
  private VirtualFilter filter;

  protected static VirtualHolder holder;

  protected static VirtualKeyboard hooked;

  VirtualKeyboard ()
  {
  }

  public void init (Document doc)
    throws Exception
  {
    XDocument xdoc = (XDocument) doc;

    NodeList list = XPath.findNodes (doc,"/virtualkeyboard[1]");
    language = ((Element)list.item (0)).getAttribute ("language").toUpperCase();
    locale = ((Element)list.item (0)).getAttribute ("locale").toUpperCase();

    Logger.log (this, "Installing virtual keyboard (" + language + "," + locale + ")");

    // get all the keymaps
    list = XPath.findNodes (doc,"/virtualkeyboard/map");
    keymap = new VirtualKeymap [list.getLength()];
    int keyId = VirtualKeymap.getNewId ();

    for (int i = 0 ; i < list.getLength() ; i++)
      { 
        keymap[i] = new VirtualKeymap ((Element)list.item(i), keyId); 
      }

    // wait for all the images to load
    VirtualKeymap.waitForImages (keyId);

    // Enable events for the keyboard component!
    enableEvents(AWTEvent.MOUSE_EVENT_MASK
        |AWTEvent.MOUSE_MOTION_EVENT_MASK | AWTEvent.KEY_EVENT_MASK );

    // Find the screen service
    IScreen is = (IScreen) ServiceLocator.findImplementor (IScreen.INTERFACE);
    if ( is != null )
      {
        screen = is.getScreen ();
      }
     
    if (holder == null)
      {
        holder = new VirtualHolder ();
        screen.add (holder, BorderLayout.SOUTH); 
      }

    // The number 60 is a calibration hack by Tim
    tpen = new Tag(60)
      {
        protected int selectArea()
          {
            if (leftx < 70)
              {
                return (TagPatterns.SYMBOLS);
              }
            else if (leftx > 168)
              {
                return (TagPatterns.NUMBERS);
              }
            else if (leftx < 168 && rightx > 168)
              {
                return (TagPatterns.UPPERCASE);
              }
            else if (leftx > 164 && rightx < 172)
              {
                return (TagPatterns.UPPERCASE);
              }
            else
              {
                return (TagPatterns.LOWERCASE);
              }
          }
      };

    if (hooked == null)
      {
        attach (this);
        hooked = this;
      }
  }

  public String[] getFunctionalities ()
  {
    return (new String[] {IVirtualInput.FUNCTIONALITY} );
  }

  public String[] getInterfaces (String functionality)
  {
    if (functionality.equals (IVirtualInput.FUNCTIONALITY))
      {
        return (new String[] {IVirtualInput.INTERFACE});
      }
    else
      {
        return null;
      }
  }

  public String[] getFeatures (String functionality)
  {
    if (functionality.equals (IVirtualInput.FUNCTIONALITY))
      {
        return (new String[] {IVirtualInput.LANGUAGE_FEATURE, getInputLanguage(),
                              IVirtualInput.LOCALE_FEATURE, getInputLocale()});
      }
    else
      {
        return null;
      }
  }

  public String[] getInformation ()
  {
    return null;
  }
  
  // IVirtualInput interface 

  public String getInputLanguage ()
  {
    return language;
  }
  
  public String getInputLocale ()
  {
    return locale;
  }

  public boolean attach ()
  {
    return attach (this);
  }

  public boolean attach (Component c)
  {
    if (screen != null && holder != null)
      { 
        holder.removeAll();
        holder.add (c); 
        // attach mouse hook
        HookInstaller.installMouseHook (this);
        attached = true;
        hooked = this;
      }
    return attached;
  }

  public void detach ()
  {
    detach (this);
  }

  public void detach (Component c)
  { 
    if (attached)
      {
        holder.remove (c);
        HookInstaller.installMouseHook (null);
        attached = false;
        if (hooked == this)
          {
            hooked = null;
          }
      }
  }

  public boolean isAttached ()
  {
    return attached;
  }

  public TextComponent getFocus ()
  {
    if (focus instanceof TextComponent)
      {
        return (TextComponent) focus;
      }
    else
      {
        return null;
      }
  }

  // Component interface
  public Dimension preferredSize ()
  {
    if (active)
      {
        Image img = keymap[current].getImage ();
        return (new Dimension (img.getWidth (null), img.getHeight (null)));
      }
    else
      {
        return (new Dimension (0, 0));
      }
  }

  public void paint (Graphics g)
  {
    if (active)
      {
        int w = AWT.extension.getWidth (this); 
        int h = AWT.extension.getHeight (this);

        Image img = keymap[current].getImage ();

        offset.width  = (w - img.getWidth (null)) / 2;
        offset.height = (h - img.getHeight (null)) / 2;
        
        g.setColor (Color.black);
        g.fillRect (0, 0, w, h);
        g.drawImage (img, offset.width, offset.height, null);
      }
  }

  public void redraw ()
  {
    if (screen != null)
      {
        Logger.log (this, "Redrawing");
        this.invalidate ();
        screen.validate ();
        screen.repaint ();
      }
  }

  // MouseHook interface
  public boolean intercept (MouseEvent evt)
  {  
    if (!active)
      {
        if (willTrigger (evt))
          {
            current = 0;
            active = true;
            redraw();
            focus = evt.getComponent ();
            focus.requestFocus ();
          }
      }
    else
      {
        if (evt.getComponent () == this)
          {
            return (dispatchEvent (evt));
          }
        else
          {
            if (willTrigger (evt))
              {
                focus = evt.getComponent ();
                focus.requestFocus ();
              }
          }
      }
    return false;
  }

  public void close ()
  {
    active = false;
    redraw ();
  }

  public boolean willTrigger (MouseEvent evt)
  {
    // Default - double click raises keyboard
    if ((evt.getClickCount () == 1) && (evt.getComponent () instanceof TextComponent))  {
        return true;
    }
    return false;
  }

  public boolean dispatchEvent (MouseEvent evt)
  {

    if (keymap[current].type == VirtualKeymap.PEN) 
      {
        return processPen (evt);
      }

    int id = evt.getID();
    if (evt.getComponent () != this || id != MouseEvent.MOUSE_RELEASED )
      {
        return false;
      }
    else
      {
        return processKeyboard (evt);
      }
  }

  protected boolean processPen (MouseEvent evt)
  {
    int id = evt.getID();
    int x = evt.getX();
    int y = evt.getY();

    switch (id)
      {
      case MouseEvent.MOUSE_PRESSED:
        oldCode = keymap[current].mapKey( x, y );
        oldCode &= 0x0000ffff;
        gpen.drawOval(x, y, 1, 1);
        tpen.start(x, y);
        break; 
      case MouseEvent.MOUSE_DRAGGED:
        gpen.drawOval (x, y, 1, 1);
        tpen.move (x, y);
        break;
      case MouseEvent.MOUSE_RELEASED:
        int ch = keymap[current].mapKey (x, y);
        ch &= 0x0000ffff;
        if (ch != 0 && ch == oldCode)  
          {
            if (ch > VirtualKeymap.COMMAND_OFFSET && ch <= VirtualKeymap.KEYMAP_OFFSET)
              {
                switchKeymap (VirtualKeymap.KEYMAP_OFFSET - ch);
              }
            else 
              {
                if (ch == VirtualKeymap.COMMAND_OFFSET - VirtualKeymap.SHIFT)
                  {
                    if (!nextShift) { nextShift = true;  }
                    else            { nextShift = false; }
                  }
                if (ch == VirtualKeymap.COMMAND_OFFSET - VirtualKeymap.CLOSE )
                  {
                    close ();
                  }
              }
          }
        else
          {
            ch = tpen.end(x, y);
            if (ch != 0) { emitKey (ch); }
          }
        gpen.drawImage (keymap[current].getImage (), offset.width, offset.height, null);
        break;
      }
    return (true);
  }
  
  protected boolean processKeyboard (MouseEvent evt)
  {
    int id = evt.getID();
    int ox = evt.getX();
    int oy = evt.getY();
    int x = ox - offset.width;
    int y = oy - offset.height;

    char ch = keymap[current].mapKey( x, y );

    boolean shift = nextShift;

    if (nextShift) { switchKeymap (last); }

    nextShift = false;
    if (ch != 0)
      {
        if (ch < 0xe000)
          {
            if (filter != null)
              {
                filter.setKeyboard (this);
                if (filter.processChar (keymap[current].mode, (char) ch))
                  {
                    return true;
                  }
              }

            // we are dealing with a character
            if (upperCase)
              {
                emitKey( Character.toUpperCase( ch ));
                upperCase = false;
              }
            else
              {
                emitKey (ch);
              }
          }
        else
          {
            upperCase = false;
            // we         are dealing with a special code 
            if ( ch > VirtualKeymap.COMMAND_OFFSET && ch <= VirtualKeymap.KEYMAP_OFFSET )
              {
                switchKeymap(VirtualKeymap.KEYMAP_OFFSET - ch);
              }
            else
              {
                if (ch == VirtualKeymap.COMMAND_OFFSET - VirtualKeymap.SHIFT)
                  {
                    if (!shift)
                      {
                        nextShift = true;
                        last = current;
                        int next = keymap[current].getShift();
                        if (next != -1)
                          {
                            switchKeymap( next );
                          }
                        else
                          {
                            upperCase = true;
                          }
                      }
                    else
                      {
                        nextShift = false;
                        switchKeymap( last );
                      }
                  }
                if (ch == VirtualKeymap.COMMAND_OFFSET - VirtualKeymap.CLOSE)
                  {
                    close ();
                  }
              }
          }
      }
    return (true);
  }

  protected void emitKey (int kchar)
  {
    if (focus == null)
      {
        return;
      }

    int kcode;
    // Deal with function keys
    if (kchar < 0)
      {
        kcode = -kchar;
        kchar = 0;
      }
    // Deal with other specials
    else switch (kchar)
      {
      case '\b':
        kcode = KeyEvent.VK_BACK_SPACE;
        break;
      case '\t':
        kcode = KeyEvent.VK_TAB;
        break;
      case '\n':
        kcode = KeyEvent.VK_ENTER;
        break;
      case 27 :
        kcode = KeyEvent.VK_ESCAPE;
        break;
      case 127 :
        kcode = KeyEvent.VK_DELETE;
        break;
      default:
        kcode = KeyEvent.VK_UNDEFINED;
        break;
      }

    int mods = 0;
    if (Character.isUpperCase((char)kchar))
      {
        mods |= KeyEvent.SHIFT_MASK;
      }

    long time = System.currentTimeMillis();

    if ( kcode == KeyEvent.VK_UNDEFINED )
      {
        focus.dispatchEvent(new KeyEvent(focus, KeyEvent.KEY_TYPED   , time, mods, kcode, (char)kchar))
          ;
      }
    else
      {
        focus.dispatchEvent(new KeyEvent(focus, KeyEvent.KEY_PRESSED , time, mods, kcode, (char)kchar))
          ;
        focus.dispatchEvent(new KeyEvent(focus, KeyEvent.KEY_TYPED   , time, mods, kcode, (char)kchar))
          ;
        focus.dispatchEvent(new KeyEvent(focus, KeyEvent.KEY_RELEASED, time, mods, kcode, (char)kchar))
          ;
      }
  }

  protected void switchKeymap (int map)
  {
    current = map;
    redraw ();
  }

  public void validate () 
  {
    super.validate ();
    if (gpen != null)
      {
        gpen.dispose ();
      } 
    gpen = getGraphics ();
  }

  public void setFilter (VirtualFilter filter)
  {
    this.filter = filter;
  }

  public void removeFilter ()
  {
    this.filter = null;
  }

}
