/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: KanjiKeyboard.java,v 1.2 2001/11/09 19:06:50 samc Exp $
 */
package org.xoe.input.virtual;

import java.awt.Toolkit;
import java.awt.Color;
import java.awt.Image;
import java.awt.Window;
import java.awt.Point;	
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Button;
import java.awt.Component;
import java.awt.Container;
import java.awt.AWTEvent;
import java.awt.Event;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.ImageObserver;
import java.awt.LayoutManager;
import java.awt.List;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.TextComponent;
import java.net.URL;
import java.util.EventObject;
import java.util.Hashtable;

import kaffe.awt.AWT;
import kaffe.awt.widgets.ImageButton;

import com.pocketlinux.awt.theme.Theme;

import org.xoe.core.services.ServiceLocator;
import org.xoe.input.canna.Canna;

class KanjiKeyboard 
  extends Container
  implements ActionListener, ItemListener, VirtualFilter 
{
  private boolean init = false;

  private TextField    word = new TextField (); 
  private List      choices = new List ();
  private ImageButton  left = new ImageButton (Theme.getTheme ().getImage ("icons/leftarrow.gif"));
  private ImageButton right = new ImageButton (Theme.getTheme ().getImage ("icons/rightarrow.gif"));
  private Button    replace = new Button ("Replace");
  private Button     cancel = new Button ("Cancel");

  private int clauseCount = 0;

  private String[][] clauses;
  private int clauseLength[];
  private int currentClause;

  private int[] supportedModes;
  private TextComponent target;

  private VirtualKeyboard vi;

  public KanjiKeyboard() 
  {
    setLayout(new KeyboardLayout(240,95));

    left.setActionCommand  ("Left");
    right.setActionCommand ("Right");

    left.addActionListener   (this);
    right.addActionListener  (this);
    replace.addActionListener(this);
    cancel.addActionListener (this);

    choices.addItemListener  (this);

    Image leftarrow  = Theme.getTheme ().getImage ("icons/leftarrow.gif");
    Image rightarrow = Theme.getTheme ().getImage ("icons/rightarrow.gif");

    add (left   , "Left");
    add (right  , "Right");
    add (word   , "Line");
    add (choices, "Text");
    add (replace, "Replace");
    add (cancel , "Cancel");

    supportedModes = new int[3];

    supportedModes[0] = VirtualKeymap.HIRAGANA;
    supportedModes[1] = VirtualKeymap.KATAKANA;
    supportedModes[2] = VirtualKeymap.KANJI;
  } 

  private void highlightClause (int clause) 
  {
    int startPos = getClausePos (clause);
    word.select (startPos, startPos + clauseLength[clause]);
  }

  private void setClause (int clause) 
  {
    choices.removeAll ();
    for (int i = 0; i < clauses[clause].length; i++) 
      {
        choices.add(clauses[clause][i]);
      }
    currentClause = clause;
    highlightClause (clause);
  }

  private int getClausePos( int clause ) 
  {
    int startPos = 0;
    for (int i = 0; i < clause; i++) 
      {
        startPos += clauseLength[i];
      }
    return startPos;
  }

  // overload VirtualKeyboard open method
  public void open () 
  {
    String word = getCurrentWord();
    if (word != null) 
      {
        setCurrentWord();
        try 
          {
            String hiragana = Canna.romaToHira( word );
            Canna canna = new Canna(hiragana);
            this.word.setText(canna.getString());
            clauseCount = canna.getClauseCount();

            clauses = new String[clauseCount][];
            clauseLength = new int[clauseCount];

            for (int i=0; i < clauseCount; i++) 
              {
                clauses[i] = canna.getCandidates();
                canna.setNextClause();
                clauseLength[i] = clauses[i][clauses[i].length - 1].length();
                replaceClause( i, clauses[i][0]);
              }

            setClause(0);

          }
        catch(Exception e) 
          {
            System.err.println(e);
          }
        vi.attach (this);
      }
  }

  public boolean willTrigger (TextComponent tc) 
  {
    return (getCurrentWord () != null);
  }

  public Dimension preferredSize () 
  {
    return (new Dimension(245, 90));
  }

  public void replaceClause(int clause, String clauseText ) 
  {
    String newText = "";
    String oldText = word.getText();
    int startPos = getClausePos (clause); 
    if (startPos > 0) 
      {
        newText += oldText.substring(0,startPos);
      }
    newText += clauseText;
    if (startPos + clauseLength[clause] < oldText.length ()) 
      {
        newText += oldText.substring (startPos + clauseLength[clause]);
      }
    clauseLength[clause] = clauseText.length ();
    word.setText (newText);
    highlightClause (currentClause);
  }

  public void itemStateChanged (ItemEvent evt) 
  {
    int item = ((Integer) evt.getItem()).intValue();
    replaceClause( currentClause, clauses[currentClause][item]);
  }

  public void setKeyboard (VirtualKeyboard vi)
  {
    this.vi = vi;
    target = vi.getFocus ();
  }

  public void close ()
  {
    vi.detach (this);
  }

  public boolean processChar (int mode, char ch)
  {
    // if mode is not supported do nothing
    if (mode != VirtualKeymap.HIRAGANA && mode != VirtualKeymap.KATAKANA && mode != VirtualKeymap.KANJI) return false;

    if (ch == '\n' || ch == '\b' || ch == '\t' || ch == 27 || ch == 127 || isDelimiter(ch))
      {
        if ( mode == VirtualKeymap.KANJI && isDelimiter(ch)) 
          {
            insertWord( "" + ch );
            open();
            return true;
          }
        else 
          {
            return false; 
          }
      }
    else 
      {
        int caret = target.getCaretPosition();
        int beginWord = getCurrentWordStart( target.getText(), caret );
        insertWord( "" + ch );
        if (beginWord != caret + 1)
          {
            String hiragana = null;
            String word = target.getText().substring( beginWord, caret + 1 );

           int wlength = word.length();
           boolean skipn = false;

           if (wlength > 0) 
             {
               if (wlength == 1)
                 {
                   if (word.charAt (wlength - 1) == 'n') { skipn = true; }
                 }
               else
                 {
                   if (word.charAt (wlength - 1) == 'n' &&
                       word.charAt (wlength - 2) != 'n' ) { skipn = true; }
                 }
             }
           if (!skipn) 
             {
               if (mode == VirtualKeymap.HIRAGANA || mode == VirtualKeymap.KANJI)
                 {
                   hiragana = Canna.romaToHira( word );
                 }
               if (mode == VirtualKeymap.KATAKANA)
                 {
                   hiragana = Canna.romaToKata( word );
                 }
             }
           if (hiragana != null)
             {
               int wordLength     = wlength = word.length();
               int hiraganaLength = hiragana.length();

               int i = hiraganaLength;

               for (; i > 0; i--)
                 {
                   if ( hiragana.regionMatches( 0, word, 0, i ))
                     {
                       break;
                     }
                 }
               String replacement = hiragana.substring( i );

               if (target instanceof TextArea) 
                 {
                   ((TextArea)target ).replaceRange (replacement, beginWord + i, 
                                                     beginWord + wordLength);
                 }
               else 
                 {
                   String temp = target.getText ();
                   temp = temp.substring (0, beginWord + i) + replacement 
                     + temp.substring (beginWord + wordLength);
                   target.setText (temp);
                 }
             }
           target.setCaretPosition (beginWord + wlength);
          }   
        return true;
      }
  }

  public void actionPerformed( ActionEvent evt ) 
  {
    String command = evt.getActionCommand(); 

    if (command.equals("Cancel")) 
      {
        close();
        target.setCaretPosition( target.getCaretPosition() + 1 );
        return;
      }
    if (command.equals("Replace")) 
      {
        replaceCurrentWord( word.getText());
        close();
        return;	
      }
    if (command.equals("Left")) 
      {
        setClause(( clauseCount + currentClause - 1) % clauseCount );
        return;
      }
    if (command.equals("Right")) 
      {
        setClause(( currentClause + 1) % clauseCount );
        return;
      }
  }


  public void replaceCurrentWord (String word)
  {
    String text = target.getText();
    int caret = target.getCaretPosition();
    int beginWord = getCurrentWordStart( text, caret );
    int endWord = getCurrentWordEnd( text, caret );
    int wordLength = word.length();

    if (target instanceof TextArea)
      {
        (( TextArea )target ).replaceRange (word, beginWord, endWord);
      }
    else
      {
        String temp = text.substring( 0, beginWord ) +
          word +
          text.substring( endWord );
        target.setText( temp );
      }
    target.setCaretPosition( beginWord + wordLength + 1);
  }

  protected int getCurrentWordStart (String text, int caret)
  {
    int beginWord = text.lastIndexOf( delimiters.charAt( 0 ), caret - 1 ) + 1;
    for ( int i = 1; i < delimiters.length() ; i ++ )
      {
        int next = text.lastIndexOf( delimiters.charAt( i ), caret - 1 ) + 1;
        if ( next > beginWord ) beginWord = next;
      }
    return beginWord;
  }

  protected int getCurrentWordEnd (String text, int caret)
  {
    int endWord = text.indexOf (delimiters.charAt (0), caret);
    for (int i = 1; i < delimiters.length (); i ++)
      {
        int next = text.indexOf (delimiters.charAt( i ), caret);
        if ( next < endWord ) endWord = next;
      }
    if (endWord == -1)
      {
        endWord = caret;
      }
    return endWord;
  }

  public String getCurrentWord ()
  {
    String text = target.getText ();
    int caret = target.getCaretPosition ();

    if (isDelimiter (text.charAt( caret - 1)))
      {
        caret--;
      }

    int beginWord = getCurrentWordStart (text, caret);
    int endWord = getCurrentWordEnd (text, caret);

    if (beginWord == endWord)
      {
        return null;
      }
    return text.substring (beginWord, endWord);
  }

  public void setCurrentWord ()
  {
    String text = target.getText();
    int caret = target.getCaretPosition();

    if ( isDelimiter( text.charAt( caret - 1 )))
      {
        caret--;
      }

    int beginWord = getCurrentWordStart (text, caret);
    int endWord = getCurrentWordEnd (text, caret);

    target.select (beginWord, endWord);
  }

  public String getCurrentWordToCaret ()
  {
    String text = target.getText();
    int caret = target.getCaretPosition();

    int beginWord = getCurrentWordStart( text, caret );

    return  text.substring( beginWord, caret );
  }

  private static String delimiters = " .,:;\t\n?!'\"[]{}`|\\/";

  protected boolean isDelimiter( char ch )
  {
    for (int i = 0; i < delimiters.length (); i++)
      {
        if (ch == delimiters.charAt (i))
          {
            return true;
          }
      }

    return false;
  }

  public void insertWord (String word)
  {
    int caret = target.getCaretPosition ();
    int beginWord = getCurrentWordStart (target.getText (), caret);

    if (target instanceof TextArea)
      {
        ((TextArea)target).insert (word , caret);
      }
    else
      {
        String temp = target.getText ();
        temp = temp.substring (0, caret)
          + word
          + temp.substring (caret);
        target.setText (temp);
      }
    target.setCaretPosition (caret + word.length ());
  }

};

/**
 *
 */
class KeyboardLayout 
  implements LayoutManager 
{
  int width;
  int height;
  Hashtable comps = new Hashtable ();	

  public KeyboardLayout (int width, int height) 
  {
    this.width = width;
    this.height = height;
  }

  public void addLayoutComponent (String name, Component comp) 
  {
    comps.put (comp, name);
  }

  public void removeLayoutComponent (Component comp) 
  {
    comps.remove (comp);
  }

  public Dimension minimumLayoutSize (Container parent) 
  {
      return new Dimension (width, height);
  }

  public Dimension preferredLayoutSize (Container parent) 
  {
    return new Dimension(width, height);
  }

  public void layoutContainer (Container parent) 
  {
    synchronized (parent.getTreeLock ()) 
      {
        int n = parent.getComponentCount();
        for (int i=0; i < n; i++) 
          {
            Component c = parent.getComponent(i);
            if ((( String)comps.get (c)).equals ("Line")) 
              {
                c.setBounds (5,5,155,20);
              }
            if ((( String)comps.get (c)).equals("Text")) 
              {
                c.setBounds (5,25,155,height-30);
              }
            if ((( String)comps.get (c)).equals("Left")) 
              {
                c.setBounds (165,5,30,30);
              }
            if ((( String)comps.get (c)).equals("Right")) 
              {
                c.setBounds (195,5,30,30);
              }
            if ((( String)comps.get (c)).equals("Replace")) 
              {
                c.setBounds (165,40,60,20);
              }
            if ((( String)comps.get (c)).equals("Cancel")) 
              {
                c.setBounds (165,65,60,20);
              }
          }
      }
  }

};
