/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: WebBrowser.java,v 1.19 2001/10/17 01:42:03 colin Exp $
 */
package com.pocketlinux.services.web;

import com.pocketlinux.content.ContentCache;
import com.pocketlinux.content.ContentElement;
import com.pocketlinux.content.ContentType;
import com.pocketlinux.content.DefaultContentTypeResolver;
import com.pocketlinux.dom.BaseDocument;
import com.pocketlinux.dom.Callback;
import com.pocketlinux.dom.DOMlet;
import com.pocketlinux.packages.IPackageQuery;
import com.pocketlinux.packages.InstallOption;
import com.pocketlinux.packages.PackageDescription;
import com.pocketlinux.packages.PackageManager;
import com.pocketlinux.services.ServiceLocator;
import com.pocketlinux.services.SimpleServiceQuery;
import com.pocketlinux.util.Logger;

import java.awt.Component;
import java.awt.Window;
import java.io.IOException;
import java.net.URL;
import java.net.UnknownHostException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.xhtml.IFrame;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.services.IPackageSelector;
import org.xoe.services.IWebBrowser;


public class WebBrowser
  implements IWebBrowser
{
  public static final BaseDocument PAGE_NOT_FOUND_404;
  public static final BaseDocument HOST_NOT_FOUND;

  private static final SimpleServiceQuery PACKAGE_SELECTOR_QUERY = new SimpleServiceQuery ().setInterface (IPackageSelector.INTERFACE);

  static
  {
    PAGE_NOT_FOUND_404 = DOMlet.create (("<html xmlns=\"" + XHTML_NAMESPACE + "\"><head><title>404 - Page Not Found</title></head><body><h2>404 - Page Not Found</h2></body></html>").toCharArray (), "404");
    HOST_NOT_FOUND = DOMlet.create (("<html xmlns=\"" + XHTML_NAMESPACE + "\"><head><title>Host Not Found</title></head><body><h2>Host Not Found</h2></body></html>").toCharArray (), "");
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = { IWebBrowser.FUNCTIONALITY };
    return (ret);
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals(IWebBrowser.FUNCTIONALITY))
      {
	String[] ret = { IWebBrowser.INTERFACE };
	return ret;
      }
    return (null);
  }

  public String[] getFeatures (String func)
  {
    return (null);
  }

  public String[] getInformation ()
  {
    return (null);
  }

  public BaseDocument getWebPage (String src)
    throws IOException
  {
    return (getWebPage (new URL (src)));
  }

  public BaseDocument getWebPage (final URL url)
    throws IOException
  {
    // fetch the content
    ContentElement content = null;
    try
      {
	content = ContentCache.fetch (url);
      }
    catch (UnknownHostException _)
      {
	return HOST_NOT_FOUND;
      }

    if (content == null)
      {
	Logger.log (this, "getWebPage: content = null");
	return PAGE_NOT_FOUND_404;
      }

    // get it's type
    ContentType type = content.getType ();

    // is there a viewable document builder for this type?
    IViewableDocumentBuilder vdb = (IViewableDocumentBuilder)type.getService(IViewableDocumentBuilder.INTERFACE);
    if (vdb == null)
      {
	Logger.log (this, "getWebPage: no viewable document builder service found for " + url);
        return getViewablePackageQuery (url, type);   
      }

    return (vdb.getViewableDocument (content));
  }

  private BaseDocument getViewablePackageQuery (final URL url, ContentType type)
  {
    final BaseDocument doc;

    IPackageSelector ps = (IPackageSelector) (ServiceLocator.findService (PACKAGE_SELECTOR_QUERY));
    
    if (ps == null)
      {
	Dialog.note ("We cannot currently display content where the <b>" 
                     + type.toExternalForm () 
                     + "</b><br />Try installing a Package Selector service to locate a content handler for this type.");
	return null;
      }
	
    boolean conf = Dialog.confirm ("We cannot currently display content where the <b>" 
                                        + type.toExternalForm () 
                                        + "</b>. <br />Should we look on the server for " 
                                        + " a package that supports this content type?");
    if (conf)
      {
	IPackageQuery pq = type.getPackageQuery (IViewableDocumentBuilder.INTERFACE);
	Logger.log (this, "getWebPage: Trying to find a package using query (" + pq.toXPath() + ")");
	InstallOption[] options = PackageManager.search (pq).getInstallablePackages ();
	Logger.log(this, "getWebPage: results:\n" + InstallOption.toString(options));

	if (options == null || options.length == 0)
	  {
	    Dialog.note ("There are no appropriate software packages on the server at this time.");
	    return null;
	  }


	doc = (BaseDocument) ps.installPackage("Pick a package", 
                                               "Choose a package, and then hit \"Install\"", 
					       options);

	doc.setCallback ("success", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      Logger.log (this, "success");
	      PackageDescription pkgd = (PackageDescription) payload;
	      try 
		{ 
		  Logger.log(this, "after confirm");
		  BaseDocument m = getWebPage(url);
		  ((IFrame) doc. getDocumentElement ()). setDocument(m);
		}
	      catch  (IOException e)
		{
		  Logger.log(this, e);
		}
	    }
	  });
	doc.setCallback ("failure", new Callback () 
	  {
	    public void invoke (Node n, Object payload)
	    {
	      PackageDescription pkgd = (PackageDescription) payload;
	      Logger.log (this, "failure");
	    }
	  });
	return (doc);
      }

    return (null);
  }
}
