package org.xoe.content.text;

import java.io.IOException;
import java.net.URL;

import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathParser;
import org.xoe.display.IconManager;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.services.IIconResolver;


public class lnk 
  extends xml
{
  /**
   * the xpath used to find the lnk src, relative to the document node
   */
  private static final XPath LNK_SRC_PATH = XPathParser.singleton.parse ("lnk/@src");

  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  public String[] getFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER) ||
	func.equals (IIconResolver.FUNCTIONALITY))
      {
	return (new String[]{
                              FEATURE_EXTENSION, EXTENSION_LNK,
                              FEATURE_MIMETYPE, MIMETYPE_TEXT_VND_TVT_LNK
	  		    });
      }
    return (null);
  }

  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    content = getRedirectedContent (content);
    if (content == null)
      {
        return (null);
      }

    IViewableDocumentBuilder vdb = (IViewableDocumentBuilder)content.getType ().getService (IViewableDocumentBuilder.INTERFACE);
    if (vdb == null)
      {
	Logger.log (this, "getViewableDocument: no viewable document builder service found for " + content.getURL ());
        return (null);
      }

    return (vdb.getViewableDocument (content));
  }

  public URL getIconURL (ContentElement resource)
    throws IOException
  { // use the icon for the file that the link represents
    return (IconManager.getIconURL (getRedirectedContent (resource)));
  }

  public URL getIcon (ContentType type)
  { // allow the standard .lnk icon to be used
    return (null);
  }

  
  private ContentElement getRedirectedContent (ContentElement content)
    throws IOException
  {
    // get the xml document builder service
    SimpleServiceQuery q = new SimpleServiceQuery (FUNCTION_CONTENT_HANDLER,
						   IDocumentBuilder.INTERFACE);

    q.addFeature (FEATURE_MIMETYPE, MIMETYPE_TEXT_XML);
    IDocumentBuilder db = (IDocumentBuilder) ServiceLocator.findService (q);
    if (db == null)
      {
        return (null);
      }

    XDocument d = db.getDocument (content);
    String lnkSrc = XPath.findString (d, LNK_SRC_PATH);
    return (ContentCache.fetch (lnkSrc));
  }
}
