/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: xml.java,v 1.8 2001/12/04 23:04:21 colin Exp $
 */
package org.xoe.content.text;

import java.io.IOException;
import java.io.StringBufferInputStream;
import java.net.URL;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.IDocumentBuilder;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.display.dom.xhtml.FormElement;
import org.xoe.display.dom.IViewableDocumentBuilder;
import org.xoe.display.dom.IEditableDocumentBuilder;
import org.xoe.display.dom.ViewElement;
import org.xoe.services.IIconResolver;


public class xml 
  implements IViewableDocumentBuilder, IEditableDocumentBuilder, IIconResolver
{
  private static final String SAVE = "edb.xml.save";

  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    return (new String[] { FUNCTION_CONTENT_HANDLER, IIconResolver.FUNCTIONALITY, IEditableDocumentBuilder.FUNCTIONALITY });
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	return (new String[] { IViewableDocumentBuilder.INTERFACE });
      }
    else if (func.equals (IIconResolver.FUNCTIONALITY))
      {
	return (new String[] { IIconResolver.INTERFACE });
      }
    else if (func.equals (IEditableDocumentBuilder.FUNCTIONALITY))
      {
	return (new String[] { IEditableDocumentBuilder.INTERFACE });
      }
    return (null);
  }

  public String[] getFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER) || func.equals (IEditableDocumentBuilder.FUNCTIONALITY))
      {
	return (new String[]{
                             FEATURE_MIMETYPE, MIMETYPE_TEXT_XML,
                             FEATURE_MIMETYPE, MIMETYPE_TEXT_XML_DTD,
			     FEATURE_EXTENSION, EXTENSION_XML,
			     FEATURE_EXTENSION, EXTENSION_DTD,
	  		    });
      }
    else if(func.equals (IIconResolver.FUNCTIONALITY))
      {
	return (new String[]{
                             FEATURE_MIMETYPE, MIMETYPE_TEXT_XML,
			     FEATURE_EXTENSION, EXTENSION_XML,
	  		    });
      }

    return (null);
  }

  public String[] getInformation ()
  {
    return (null);
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    // get the document builder service and build the xml document
    IDocumentBuilder db = (IDocumentBuilder)content.getType ().getService (IDocumentBuilder.INTERFACE);
    if (db == null)
      {
        return (null);
      }

    XDocument doc = db.getDocument (content);

    // get the src url
    URL src = content.getURL ();


    // since no specific xml editing service volunteered to handle this, it 
    // fell to us (the default xml renderer). therefore, this is some kind of 
    // non-renderable xml data - so we just display it, nicely formated.
    String s = "<html xmlns=\""  
               + XHTML_NAMESPACE 
               + "\"><head><title>" 
               + src 
               + "</title></head><body><pre></pre></body></html>";
    char[] c = s.toCharArray ();
    DOMlet wrapperView = DOMlet.create (c, src);

    // get the pre tag
    Node pre = XPath.findFirstNode (wrapperView, "xhtml:html/xhtml:body/xhtml:pre");
    DOMUtils.setTextContent (pre, XMLWriter.getXML (doc));

    return (wrapperView);
  }
 

  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IEditableDocumentBuilder

  public XDocument getEditableDocument (final ContentElement content)
    throws IOException
  {
    if (false) // read only check
      {
	// TODO a title and some kind of message that indicates that it is read-only
	return getViewableDocument (content); 
      }
    else 
      {
	IDocumentBuilder db = (IDocumentBuilder)content.getType ().getService (IDocumentBuilder.INTERFACE);
	XDocument srcDoc;
	
	if (db != null)
	  {
	    srcDoc = db.getDocument (content);
	  }
	else 
	  {
	    IViewableDocumentBuilder vdb = (IViewableDocumentBuilder)content.getType ().getService (IViewableDocumentBuilder.INTERFACE);
	    if (vdb == null)
	      {
		Logger.log (this, "Can't find either a IDocumentBuilder or IViewableDocumentBuilder for content-type " + content.getType ());
		return null;
	      }
	    srcDoc = vdb.getViewableDocument (content);
	  }

	// get the src url
	final URL srcURL = content.getURL ();

        // get the body element then convert all it's children to xml strings
        Node body = XPath.findFirstNode (srcDoc, "xhtml:html/xhtml:body");
	if (body == null)
	  {
	    body = srcDoc;
	  }

        // the string we are building
        String editableContent = "";

        for (Node child = body.getFirstChild (); child != null; child = child.getNextSibling ())
          {
	    editableContent += XMLWriter.getXML (child);
	  }
	editableContent.replace ('\'', '"');
        final String outer = "<html xmlns='" + XHTML_NAMESPACE + "'><head><title>" + content.getURL().getFile () 
                              + "</title></head><body><textarea id='edit' width='100%' height='100%' text='" 
    			      + editableContent + "' /></body></html>";

	final XDocument d = DOMlet.create (outer.toCharArray (), "edit page");

	final Callback oldCallback = d.getCallback ("action");
	d.setCallback (SAVE, new Callback () 
				 {
				   public void invoke (Node n, Object payload)
				   {
				     try 
				       {
					 String data = ((FormElement) d.getElementById ("edit")).getFormValue ();
                                         if (data != null && data.length () > 0)
                                           {
                                             String s = "<html xmlns='" 
                                                         + XHTML_NAMESPACE 
                                                         + "'><body>" 
                                                         + data 
                                                         + "</body></html>";

					     XDocument editedDoc = DOMlet.create (new StringBufferInputStream (s), "Save file");
					     editedDoc.setURL (srcURL);
					     DocumentCache.putDocument (editedDoc);
                                           }
					 d.setCallback ("action", oldCallback);
				       }
				     catch (Exception e)
				       {
                                         Logger.log (this, e);
				       }
				   }
				 });
	return (d);
      }
  }

  public void save (XDocument d)
  {
    Callback c = d.getCallback (SAVE);
    if (c != null)
      {
	c.invoke (null, null);
      }
  }

  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IIconResolver

  public URL getIconURL (ContentElement resource)
    throws IOException
  { // use the icon for the file that the link represents
    SimpleServiceQuery q = new SimpleServiceQuery (IIconResolver.FUNCTIONALITY,
						   IIconResolver.INTERFACE);

    q.addFeature (FEATURE_DOCTYPE_NAME_NS, resource.getType ().getXMLDoctypeNameNS ());

    IService[] providers = ServiceLocator.findServices (q);
    for (int i = 0; i < providers.length; i++)
      {
	URL iconURL = ((IIconResolver)providers[i]).getIconURL (resource);
	if (iconURL != null)
	  {
	    return (iconURL);
	  }
      }

    return (null);
  }

  public URL getIconURL (ContentType type)
  { 
    SimpleServiceQuery q = new SimpleServiceQuery (IIconResolver.FUNCTIONALITY,
						   IIconResolver.INTERFACE);

    q.addFeature (FEATURE_DOCTYPE_NAME_NS, type.getXMLDoctypeNameNS ());

    IService[] providers = ServiceLocator.findServices (q);
    for (int i = 0; i < providers.length; i++)
      {
	URL iconURL = ((IIconResolver)providers[i]).getIconURL (type);
	if (iconURL != null)
	  {
	    return (iconURL);
	  }
      }

    return (null);
  }

}
