/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: StashDirectoryBuilder.java,v 1.3 2001/11/02 02:32:47 colin Exp $
 */
package org.xoe.display.dom;

import java.io.IOException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.display.IconManager;


/**
 * Service that builds Documents and ViewableDocuments from 
 * ConentElements of type "application/x-stashdir"
 */
public class StashDirectoryBuilder 
  implements IViewableDocumentBuilder
{
  public StashDirectoryBuilder() {}

  // -------------------------------------------------------------------------
  // org.xoe.core.services.IService

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = { FUNCTION_CONTENT_HANDLER };
    return (ret);
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	String[] ret = { 
	                 IViewableDocumentBuilder.INTERFACE,
                       };
	return (ret);
      }
    return null;
  }

  public String[] getFeatures (String func)
  {
    if (func.equals (FUNCTION_CONTENT_HANDLER))
      {
	return (new String[]{ FEATURE_MIMETYPE, MIMETYPE_STASHDIR, });
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }


  // -------------------------------------------------------------------------
  // org.xoe.display.dom.IViewableDocumentBuilder

  public XDocument getViewableDocument (ContentElement content)
    throws IOException
  {
    // this should probably look up the xml document builder service and use that
    return (DOMlet.create (buildIndex (content).toCharArray (), content.getURL ()));
  }
 
  /**
   * Build an html string from a StashDirectory.  The resulting string
   * looks a lot like the direcotory indexs that netscape generates.
   */
  protected String buildIndex (ContentElement content)
    throws IOException
  {
    Node parent = content.getParentNode ();
    if (!(parent instanceof ContentElement)) { parent = null; }
    NodeList list = content.getChildNodes ();
    String path =  content.getURL ().toString ();

    StringBuffer index = new StringBuffer();
    index.append ("<html xmlns='")
         .append (XHTML_NAMESPACE)
         .append ("'><head><title>&nbsp;")
         .append (path)
         .append ("</title></head><body><b>&nbsp;Directory listing of ")
         .append (path)
         .append ("</b>");


    if (parent != null)
      {
        System.out.println(parent);
        index.append ("<br /><a href='")
             .append (((ContentElement)parent).getURL ().toString ())
             .append ("'>&nbsp;up to a higher level directory</a>");
      }

    for (int i=0; i<list.getLength (); i++)
      {
        ContentElement elem = (ContentElement) list.item (i);
        String filename = elem.getURL ().toString ();
        index.append ("<br /><a href='")
             .append (filename)
             .append ("'> <img src='")
	     .append (IconManager.getIconURL (elem).toString())
             .append ("' border='0' />&nbsp;")
             .append (filename.substring (content.getURL ().toString ().length ()))
             .append ("</a>");
      }
    index.append ("</body></html>");

    return (index.toString ());
  }
}
