/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "license.terms" for details.
 *
 * $tvt: IconInstaller.java,v 1.1 2001/10/22 22:51:14 colin Exp $
 */
package org.xoe.packages;

import java.io.IOException;
import java.net.URL;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentElement;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.packages.Dependency;
import org.xoe.core.packages.InstallException;
import org.xoe.core.packages.IPackageXMLHandler;
import org.xoe.core.packages.PackageDescriptionException;
import org.xoe.core.packages.Proviso;
import org.xoe.core.packages.ResourceInstaller;
import org.xoe.core.packages.XoePackage;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.display.IconManager;


public class IconInstaller 
  implements IPackageXMLHandler
{
  public static final String NAMESPACE = "http://www.xoe.org/installer/icon";
  public static final String ICONS_FILENAME = "icons.xml";

  XoePackage xoePackage;

  public String[] getHandlerPredependencies ()
  {
    return new String[] {ResourceInstaller.NAMESPACE};
  }

  public Dependency buildDependency (Element depEl)
    throws PackageDescriptionException
  {
    throw new PackageDescriptionException ("Unrecognized dependency type:\n" + XMLWriter.getXML (depEl));
  }

  public void init (Document configDoc)
  { // No configuration necessary
  }

  public String[] getFunctionalities ()
  {
    String[] ret = {IPackageXMLHandler.FUNCTIONALITY};
    return ret;
  }

  public String[] getInterfaces (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { IPackageXMLHandler.INTERFACE };
      }
    return null;
  }
  
  public String[] getFeatures (String func)
  {
    if (func.equals (IPackageXMLHandler.FUNCTIONALITY))
      {
	return new String[] { FEATURE_NAMESPACE, getNamespaceURI() };
      }
    return null;
  }

  public String[] getInformation ()
  {
    return null;
  }

  public String getNamespaceURI ()
  {
    return (NAMESPACE);
  }

  public void setPackage (XoePackage p)
  {
    xoePackage = p;
  }

  public Vector generateRequires ()
  {
    return null;
  }

  public Vector generateConflicts ()
  {
    return null;
  }

  public Vector generateProvides ()
  {
    return (null);
  }

  public void verify (XoePackage xoePackage, Proviso[] provisos)
  {
  }

  public void install (XoePackage xoePackage, Proviso[] provisos)
    throws InstallException
  {
    // Install icons
    try
      {
        URL pkgBase = Stash.getWorkingDirectory (xoePackage);
	ContentElement ce = Stash.fetch (new URL (pkgBase, ICONS_FILENAME));
	if (ce != null)
	  {
	    Document doc = DOMlet.create (ce.getInputStream (), ICONS_FILENAME);

	    NodeList list = XPath.findNodes (doc, "icons/icon");
	    int len = list.getLength ();
	    for (int i = 0; i < len; i++) 
	      {
		Element elem = (Element) list.item (i);
		String sfx = elem.getAttribute ("suffix");
		String mt = elem.getAttribute ("mimetype");
		String src = elem.getAttribute ("src");

		try
		  {
		    if(sfx != null)
		      {
			installIcon (xoePackage, sfx, src);
		      }

		    if(mt != null)
		      {
			installIcon (xoePackage, mt, src);
		      }
		  }
		catch (IOException ex)
		  {
		    Logger.log (this, "failed to install icon mapping: " + src);
		  }
	      }
	  }
      }
    catch (IOException ex)
      {
	throw new InstallException (ex, xoePackage,
				    ("An error occurred while searching for " +
				     ICONS_FILENAME));
      }
  }


  private void installIcon (XoePackage xoePackage, String key, String src)
    throws IOException
  {
    URL url = new URL (Stash.getWorkingDirectory (xoePackage), src);
    if (Stash.fetch (url) != null)
      { // only install if the file actually exists
	IconManager.setIcon (key, url, xoePackage);
      }
    else
      {
	Logger.log (this, "Ignoring icon mapping " + key + " with invalid url: " + url);
      }
  }

  public void uninstall (XoePackage xoePackage, Proviso[] provisos)
  {
    IconManager.removeProvider(xoePackage);
  }
}

