/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Applet.java,v 1.10 2001/12/10 21:06:48 joerg Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.applet.AppletContext;
import java.applet.AppletStub;
import java.applet.AudioClip;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Toolkit;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;                                                               
import java.util.Enumeration;

import kaffe.applet.AudioPlayer;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.display.dom.ViewElement;


public class Applet
  extends ViewElement
{
  java.applet.Applet applet;
  String src;

  class Render
    extends Container
    implements AppletStub, AppletContext
  {
    Render() {
      setLayout( null);
    }

    public void addNotify () 
    {
      if (getPeer () == null && applet != null) 
	{
	  applet.init ();
	  applet.start ();
	}
      super.addNotify ();
    }
   
    public void appletResize (int width, int height) 
    {
      Dimension size = getSize();
      if (size.width != width || size.height != height)
	{
	  setSize (width, height);
	}
    }

    public void doLayout () 
    {
      if (this.getComponentCount () > 0 ) 
	{
	  Dimension d = this.getSize ();
	  this.getComponent (0).setBounds (0, 0, d.width, d.height);
	}
    }
    
    public java.applet.Applet getApplet (String name) 
    {
      Applet n = (Applet) XPath.findFirstNode (getOwnerDocument (), "xhtml://applet[@name='"+name+"']");
	if (n != null) 
	  {
	    return (n.applet);
	  }
	return (null);
    }
    
    public AppletContext getAppletContext () 
    {
      return this;
    }

    public Enumeration getApplets () 
    {
      return null;
    }

    public AudioClip getAudioClip (URL url) 
    {
      return (new AudioPlayer (url));
    }

    public URL getCodeBase () 
    {
      String cb = getAttribute ("codebase");
      if (cb == null ) 
        {
	   cb = "./";
	}
      else if ( ! cb.endsWith (".jar") && ! cb.endsWith (".zip") ) 
	{
	  if ( ! cb.endsWith( "//" )) 
	    {
	      cb += "//";
	    }
	}
       
      return getURL (cb);
    }
    
    public URL getDocumentBase () 
    {
      URL base = null;
      
      try {
	String clsName = applet.getClass ().getName ();
	clsName = clsName.substring (clsName.lastIndexOf ('.')+1) + ".class";
	
	String baseStr = (applet.getClass ().getResource (clsName)).toString ();
	base = new URL( baseStr.substring (0, baseStr.lastIndexOf ('/')));
	}
	catch (MalformedURLException mfe) 
	  {
	    Logger.log (this, mfe);
	  }
      return base;
    }
    
public Image getImage (URL url) 
    {
      return (Toolkit.getDefaultToolkit ().getImage (url));
    }

    public String getParameter (String name) 
    {
      Element n = (Element) XPath.findFirstNode (Applet.this, "xhtml:param[@name='"+name+"']");
      if (n != null) 
	{
	  return (n.getAttribute("value"));
	}
      return (null);
    }
    
    public boolean isActive () 
    {
      return true;
    }

    public void setApplet (String appletCls) 
    {
      try 
	{
	  //Class ac = Class.forName (appletCls);

          URL [] ua = new URL[1];
	  ua[0] = getCodeBase();
	    
	  URLClassLoader l = new URLClassLoader( ua);

	  if ( appletCls.endsWith( ".class") ) 
	    {
	      appletCls = appletCls.substring( 0, appletCls.length() - 6);
	    }
	   
	  Class ac = l.loadClass( appletCls);

	  applet = (java.applet.Applet)ac.newInstance ();
	   
      	  //applet init may rely on bounds
	  Dimension d = getVis ().getSize ();
	  applet.setBounds (0, 0, d.width, d.height);
	  applet.setStub (this);
	  this.add (applet);
	}
      catch ( Exception e) 
	{
	  Logger.log (this, e);
	}
    }
    
    public void removeNotify () 
    {
      if (getPeer () != null && applet != null) 
	{
	  applet.stop ();
	}
      super.removeNotify();
    }
    
    public void showDocument (URL url) 
    {
      showDocument (url, null);
    }
    
    public void showDocument (URL url, String target) 
    {
      // throw new Error(); - HACK FIX
    }
    
    public void showStatus (String status) 
    {
    }
  }
  
  public Applet() 
  {
  }

  protected void makeVis () 
  {
    vis = new Render ();
    if (src != null) 
      {
	((Render)vis).setApplet (src);
      }
  }
  
  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);
    if ((localName.equalsIgnoreCase("src")) ||
	(localName.equalsIgnoreCase("code")))
      {
	src = value;
	if (vis != null) 
	  {
	    ((Render)vis).setApplet (src);
	  }
      }
    
    super.setAttributeNS(namespaceURI, localName, value);
  }

  public Node cloneNode (boolean deep) 
  {
    Applet nnode = (Applet)super.cloneNode (deep);
    nnode.src = src;
    nnode.applet = null;
    return nnode;
  }
}
