/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Dialog.java,v 1.13 2001/10/31 22:59:57 samc Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.BorderLayout;
import java.awt.Dimension;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.dom.XDocument;
import org.xoe.core.device.IScreen;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.Callback;
import org.xoe.core.services.ServiceLocator;
import org.xoe.display.dom.Render;


public class Dialog
  extends java.awt.Dialog
{
  private final static int percent = 80;

  private Render render;
  private String trigger;
  private boolean modal;

  private Callback callback = new Callback()
  {
    public void invoke (Node node, Object payload)
    {
      trigger = (String)payload;
      Dialog.this.setVisible (false);
    }
  };

  public Dialog (String title)
  {
    this (title, true); 
  }

  public Dialog (String title, boolean modalFlag)
  {
    super (null, title, true);
    modal = modalFlag;
    render = new Render ()
    {
      public Dimension preferredSize ()
      {
        Dimension screensize = ((IScreen)ServiceLocator.findImplementor (IScreen.INTERFACE)).getScreen ().getSize ();
        return (getPreferredSize (new Dimension (screensize.width * percent / 100, screensize.height * percent / 100), false));
      }
    };
    add (render, BorderLayout.CENTER);
  }

  public void setVisible (boolean flag)
  {
    if (!modal && flag && !isVisible ())
      {
	(new Thread ()
	 {
	    public void run ()
	    {
	      show ();
	    }
	 }).start ();
      }
    else
      {
	super.setVisible (flag);
      }
  }

  public void write (String xhtml)
  {
    DOMlet doc = DOMlet.create (xhtml.toCharArray (), "");
    doc.setCallback ("action", callback);
    setDocument (doc);
  }

  public void setDocument (XDocument doc)
  {
    render.setDocument (doc);
  }

  public XDocument getDocument ()
  {
    return ((XDocument)render.getDocument ());
  }

  /**
   * Wait until we get an answer to the dialog question and return it.
   */
  public String getAnswer ()
  {
    return (trigger);
  }

  public static boolean confirm (String message)
  {
    // If message doesn't look like an HTML page we wrap it up in HTML
    // NB. HACK - 'yes' and 'no' are not localized!!!
    if (message.startsWith ("<html"))
      {
	message = "<html xmlns='http://www.w3.org/1999/xhtml'><iframe width='' height=''>"
                  + message 
                  + "</iframe></html>";
      }
    else
      {
	message = "<html xmlns='http://www.w3.org/1999/xhtml'><iframe width='' height=''><center>" 
                  + message 
                  + "<br /><form action=''><input type='submit' name='answer' value='yes' /><input type='submit' name='answer' value='no' /></form></center></iframe></html>";
      }

    Dialog dialog = new Dialog ("Please confirm:");
    dialog.write (message);
    dialog.pack ();
    Dimension screensize = ((IScreen)ServiceLocator.findImplementor (IScreen.INTERFACE)).getScreen ().getSize ();
    Dimension dialogsize = dialog.getSize ();
    dialog.setLocation ((screensize.width - dialogsize.width) / 2, (screensize.height - dialogsize.height) / 2);
    dialog.setVisible (true);

    String answer = dialog.getAnswer ();
    int idx = answer.indexOf ('=');
    if (idx > -1)
      {
	answer = answer.substring (idx+1);
      }

    if ("yes".equalsIgnoreCase (answer))
      {
	dialog.hide ();
	return (true);
      }
    else
      {
	dialog.hide ();
	return (false);
      }
  }

  public static void note (String message)
  {
    // If message doesn't look like an HTML page we wrap it up in HTML
    // NB. HACK - 'yes' and 'no' are not localized!!!
    if (message.startsWith ("<html"))
      {
	message = "<html xmlns='http://www.w3.org/1999/xhtml'><iframe width='' height=''>"
                  + message 
                  + "</iframe></html>";
      }
    else
      {
	message = "<html xmlns='http://www.w3.org/1999/xhtml'><iframe width='' height=''><center>" 
                  + message 
                  + "<br /><form action=''><input type='submit' value='OK' /></form></center></iframe></html>";
      }

    Dialog dialog = new Dialog ("Note:");
    dialog.write (message);
    dialog.pack ();
    Dimension screensize = ((IScreen)ServiceLocator.findImplementor (IScreen.INTERFACE)).getScreen ().getSize ();
    Dimension dialogsize = dialog.getSize ();
    dialog.setLocation ((screensize.width - dialogsize.width) / 2, (screensize.height - dialogsize.height) / 2);
    dialog.setVisible (true);

    dialog.hide ();
  }
}
