/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Emphasis.java,v 1.12 2001/11/07 20:53:11 tim Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.Font;
import java.awt.Color;
import java.awt.Cursor;
import java.util.StringTokenizer;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.util.Logger;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;


public class Emphasis
  extends ViewElement 
{
  public void layout (LayoutCursor cur) 
  {
    Font fnt = cur.font;
    String name = fnt.getName ();
    int style = fnt.getStyle ();
    int size = fnt.getSize ();
    
    String tag = getLocalName ();
    
    if (tag.equalsIgnoreCase ("b") ||
	tag.equalsIgnoreCase ("strong")) 
      {	
	// Bold
	cur.setFont (name, style | Font.BOLD, size);
	layoutChildren (cur);
	cur.setFont(fnt);
      }
    else if (tag.equalsIgnoreCase ("i") ||
	     tag.equalsIgnoreCase ("dfn") ||
	     tag.equalsIgnoreCase ("var") ||
	     tag.equalsIgnoreCase ("em") ||
	     tag.equalsIgnoreCase ("cite")) 
      {
	// Italic
	cur.setFont (name, style | Font.ITALIC, size);
	layoutChildren (cur);
	cur.setFont (fnt);
      }
    else if (tag.equalsIgnoreCase ("blockquote") ||
	     tag.equalsIgnoreCase ("address")) 
      {	// Quote
	cur.setFont (name, style | Font.ITALIC, size);
	int w = cur.vis.getFontMetrics (cur.font).stringWidth ("    ");
	cur.leftindent += w;
	cur.width -= w;
	cur.blankline ();
	layoutChildren (cur);
	cur.blankline ();
	cur.setFont (fnt);
	cur.width += w;
	cur.leftindent -= w;
      }
    else if (tag.equalsIgnoreCase ("font")) 
      { 
	// Font - complex
	// face, size, color
	Color fgclr = cur.fgclr;
	String val = getAttribute ("size");
	if (val != null) 
	  {
	    size = cur.deffont.getSize () * 100;
	    try 
	      {
		int c;
		switch (val.charAt (0)) 
		  {
		  case '+':
		    c = Integer.parseInt (val.substring (1), 10);
		    if (c < 0 || c > 6)
		      {
			size = size * 140 / 100;
		      }
		    else
		      {
			for (; c > 0; c--)
			  {
			    size = size * 120 / 100;
			  }
		      }
		    break;
		  case '-':
		    c = Integer.parseInt (val, 10);
		    if (c < 0 || c > 6)
		      {
			size = size * 70 / 100;
		      }
		    else
		      {
			for (; c > 0; c--)
			  {
			    size = size * 80 / 100;
			  }
		      }
		    break;
		  default:
		    c = Integer.parseInt (val, 10);
		    switch (c)
		      {
			case 0:
			  size = size * 60 / 100;
			  break;
			case 1:
			  size = size * 70 / 100;
			  break;
			case 2:
			  size = size * 80 / 100;
			  break;
			case 3:
			  size = size * 100 / 100;
			  break;
			case 4:
			  size = size * 120 / 100;
			  break;
			case 5:
			  size = size * 150 / 100;
			  break;
			case 6:
			  size = size * 200 / 100;
			  break;
			default:
			  size = size * 140 / 100;
			  break;
		      }
		    break;
		  }
	      }
	    catch (NumberFormatException nfe) 
	      {
		// Bad argument - ignore
	      }
	    size /= 100;
	  }
	cur.fgclr = getColor(getAttribute ("color"), cur.fgclr);
	val = getAttribute ("face");
	if (val != null) {
	  
	  StringTokenizer tok = new StringTokenizer (val, ",");
	  // For now just pick the first one
	  if (tok.countTokens () == 0) 
	    {
	      name = val;
	    }
	  else 
	    {
	      name = tok.nextToken ();
	    }
	}
	cur.setFont (name, style, size);
	layoutChildren (cur);
	cur.setFont (fnt);
	cur.fgclr = fgclr;
      }
    else if (tag.equalsIgnoreCase ("big")) 
      { // Font+1
	cur.setFont (name, style, size * 120 / 100);
	layoutChildren (cur);
	cur.setFont (fnt);
      }
    else if (tag.equalsIgnoreCase ("small")) 
      { 
	// Font-1
	cur.setFont (name, style, size * 80 / 100);
	layoutChildren (cur);
	cur.setFont (fnt);
      }
    else if (tag.equalsIgnoreCase ("center")) 
      { 
	// Center
	int halign = cur.halign;
	cur.newline ();
	cur.halign = LayoutCursor.CENTER;
	layoutChildren (cur);
	cur.newline ();
	cur.halign = halign;
      }
    else if (tag.equalsIgnoreCase ("div")) 
      { 
	// Divide (and align)
	int halign = cur.halign;
	cur.newline ();
	cur.halign = getAlign (getAttribute ("align"), cur.halign);
	layoutChildren (cur);
	cur.newline ();
	cur.halign = halign;
      }
    else if (tag.equalsIgnoreCase ("p")) 
      { 
	// Paragraph break (and align)
	int halign = cur.halign;
	if (cur.x > cur.leftindent) 
	  {
	    cur.blankline ();
	  }
	cur.halign = getAlign (getAttribute ("align"), cur.halign);
	layoutChildren (cur);
	if (cur.x > cur.leftindent) 
	  {
	    if (getNextSibling () != null) 
	      {
		cur.blankline ();
	      }
	    else {
	      cur.newline ();
	    }
	  }
	cur.halign = halign;
      }
    else if (tag.equalsIgnoreCase ("br")) 
      { 
	// Line-break
	String type = getAttribute ("clear");
	int ftype = LayoutCursor.NONE;
	if ("left".equalsIgnoreCase (type)) 
	  {
	    ftype = LayoutCursor.LEFT;
	  }
	else if ("right".equalsIgnoreCase (type)) 
	  {
	    ftype = LayoutCursor.RIGHT;
	  }
	else if ("all".equalsIgnoreCase (type)) 
	  {
	    ftype = LayoutCursor.CENTER;
	  }
	
	if (cur.x > cur.leftindent) 
	  {
	    cur.newline ();
	  }
	else {
	  cur.blankline ();
	}
	cur.flushfloats (ftype);
	layoutChildren (cur);
	if (cur.x > cur.leftindent) 
	  {
	    cur.newline ();
	  }
		cur.flushfloats (ftype);
      }
    else if (tag.equalsIgnoreCase ("nobr")) 
      { 
	// No soft line breaks
	if (cur.nobreaks == false) 
	  {
	    cur.nobreaks = true;
	    cur.nobrx = cur.x;
	    layoutChildren (cur);
	    cur.nobreaks = false;
	    if (cur.nobrx > cur.maxwidth) 
	      {
		cur.maxwidth = cur.nobrx;
	      }
	  }
	else 
	  {
	    layoutChildren (cur);
	  }
      }
    else if (tag.charAt(0) == 'h') 
      { 
	// Headers
	int halign = cur.halign;
	try {
	  if (cur.y > 0 || cur.x > cur.leftindent) 
	    {
	      cur.blankline ();
	    }
	  style = Font.BOLD;
	  switch (Integer.parseInt (tag.substring (1), 10))
	    {
	      case 7:
		size = size * 60 / 100;
		break;
	      case 6:
		size = size * 70 / 100;
		break;
	      case 5:
		size = size * 80 / 100;
		break;
	      case 4:
		size = size * 100 / 100;
		break;
	      case 3:
		size = size * 120 / 100;
		break;
	      case 2:
		size = size * 150 / 100;
		break;
	      case 1:
		size = size * 200 / 100;
		break;
	      default:
		size = size * 140 / 100;
		break;
	    }
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
	cur.halign = getAlign (getAttribute ("align"), cur.halign);
	cur.setFont (name, style, size);
	layoutChildren (cur);
	cur.setFont (fnt);
	cur.blankline ();
	cur.halign = halign;
      }
    else if (tag.equalsIgnoreCase ("span")) 
      { 
        // Change style
        // Currently we ignore any span information
        layoutChildren (cur);
      }
    else if (tag.equalsIgnoreCase ("u") || tag.equalsIgnoreCase ("ins")) 
      { 
	     
	// Underline
	boolean underline = cur.underline;
	cur.underline = true;
	layoutChildren (cur);
	cur.underline = underline;
      }
    else if (tag.equalsIgnoreCase ("strike") || tag.equalsIgnoreCase ("s") || tag.equalsIgnoreCase ("del")) 
      { 
	     
	// Strike
	boolean strike = cur.strike;
	cur.strike = true;
	layoutChildren (cur);
	cur.strike = strike;
      }
    else if (tag.equalsIgnoreCase ("pre")) 
      {
	// Preformatted
	boolean literal = cur.literal;
	cur.literal = true;
	cur.setFont ("Monospaced", style, size);
	layoutChildren (cur);
	cur.setFont(fnt);
	cur.literal = literal;
      }
    else if (tag.equalsIgnoreCase ("tt") || tag.equalsIgnoreCase ("code") || tag.equalsIgnoreCase ("samp") || tag.equalsIgnoreCase ("kbd"))
      { 
	// Teletype font
	cur.setFont ("Monospaced", style, size);
	layoutChildren (cur);
	cur.setFont(fnt);
      }
    else if (tag.equalsIgnoreCase ("body")) 
      { // General display things
	Color fgclr = cur.fgclr;
	Color bgclr = cur.bgclr;
	Color anchorclr = cur.anchorclr;
	Color activeclr = cur.activeclr;
	cur.anchorclr = getColor (getAttribute ("link"), cur.anchorclr);
	cur.activeclr = getColor (getAttribute ("alink"), cur.activeclr);

	// Set the background image (if there is one)
	String val = getAttribute ("background");
	if (val != null) 
	  {
	    Backdrop drop = getBackdrop ();
	    if (drop != null)
	      {
		drop.setBackground (getURLString (val));
	      }
	  }
	else
	  {
	    // Otherwise, set the background color
	    Color bclr = getColor (getAttribute ("bgcolor"), null);
	    if (bclr != null) 
	      {
		Backdrop drop = getBackdrop ();
		if (drop != null)
		  {
		    cur.bgclr = bclr;
		    drop.setBackground (bclr);
		  }
	      }
	  }
	cur.fgclr = getColor (getAttribute ("text"), cur.fgclr);
	layoutChildren (cur);
	cur.fgclr = fgclr;
	cur.bgclr = bgclr;
	cur.anchorclr = anchorclr;
	cur.activeclr = activeclr;
      }
  }

  static int getAlign (String val, int def) 
  {
    if (val != null) 
      {
	if ("left".equalsIgnoreCase (val)) 
	  {
	    return (LayoutCursor.LEFT);
	  }
	else if ("center".equalsIgnoreCase (val)) 
	  {
	    return (LayoutCursor.CENTER);
	  }
	else if ("middle".equalsIgnoreCase (val)) 
	  {
	    return (LayoutCursor.CENTER);
	  }
	else if ("right".equalsIgnoreCase (val)) 
	  {
	    return (LayoutCursor.RIGHT);
	  }
	else if ("top".equalsIgnoreCase (val)) 
	  {
	    return (LayoutCursor.TOP);
	  }
	else if ("bottom".equalsIgnoreCase (val)) 
	  {
	    return (LayoutCursor.BOTTOM);
	  }
      }
    return (def);
  }

  static Color getColor (String colorname, Color def) 
  {
    Color color = def;
    if (colorname != null && colorname.length () > 0) 
      {
	if (colorname.charAt (0) != '#') 
	  {
	    try { 
	      // try the colour names in java.awt.Color 
	      return ((Color)Color.class.getField (colorname).get (null));
	    } 
	    catch (NullPointerException npe) 
	      { 
		Logger.log ("org.xoe.display.dom.xhtml.Emphasis", npe);
	      }
	    catch (NoSuchFieldException nsfe) 
	      {
		//this gets thrown many times on bootup
		//Logger.log ("org.xoe.display.dom.xhtml.Emphasis", nsfe);
	      }
	    catch (SecurityException se) 
	      {
		Logger.log ("org.xoe.display.dom.xhtml.Emphasis", se);
	      }
	    catch (IllegalAccessException iae) 
	      {
		Logger.log ("org.xoe.display.dom.xhtml.Emphasis", iae);
	      }
	  }
	else 
	  {
	    colorname = colorname.substring (1);
	  }
	// Fall back on hex color values
	try 
	  {

	    int cval = Integer.parseInt (colorname, 16);
	    if (colorname.length () <= 6)
	      {
		cval |= 0xFF000000;
	      }
	    color = new Color ((cval >> 16) & 0xFF, (cval >> 8) & 0xFF, cval & 0xFF, (cval >> 24) & 0xFF);
	  }
	catch (NumberFormatException nfe) 
	  {
	    //this happens all the time when we're reading numbers
	    //Logger.log ("org.xoe.display.dom.xhtml.Emphasis", nfe);
	  }
      }
    return (color);
  }

  private Backdrop getBackdrop ()
  {
    for (Node n = this; n != null; n = n.getParentNode ())
      {
	if (n instanceof Backdrop)
	  {
	    return ((Backdrop)n);
	  }
      }
    return (null);
  }
}
