/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Form.java,v 1.11 2001/12/11 19:38:30 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.CheckboxGroup;
import java.net.URLEncoder;
import java.util.Hashtable;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.Callback;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathParser;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.Thread;
import org.xoe.display.dom.ViewElement;


public class Form
  extends ViewElement 
{
  private final static String ACTION = "action";
  private final static XPath FORM1 = XPathParser.singleton.parse("//xhtml:input[not(@type='submit' or @type='image')]");
  private final static XPath FORM2 = XPathParser.singleton.parse("//xhtml:select");
  private final static XPath FORM3 = XPathParser.singleton.parse("//xhtml:textarea");
  private final static XPath FORM4 = XPathParser.singleton.parse("//xoe:text"); // ?
  private final static XPath ALL = XPathParser.singleton.parse ("//xhtml:input[not(@type='submit' or @type='image')] | //xhtml:select | //xhtml:textarea");
  private Hashtable groups;
  
  public void layout (LayoutCursor cur) 
  {
    layoutChildren (cur);
  }
  
  void dispatchForm (ViewElement v) 
  {
    String action = getAttribute (ACTION);
    if (action == null)
      {
	action = "";
      }
    Thread.invoke(getCallback (ACTION), this, action + '?' + getSelectedAsString (v));
  }

  void resetForm ()
  {
    NodeList list = XPath.findNodes (this, ALL);
    int len = list.getLength ();
    for (int i = 0; i < len; i++) 
      {
	((FormElement) list.item (i)).resetValue ();
      }
  }

  synchronized CheckboxGroup getCheckboxGroup(String groupname)
  {
    if (groups == null)
      {
	groups = new Hashtable();
      }
    CheckboxGroup group = (CheckboxGroup)groups.get(groupname);
    if (group == null)
      {
	group = new CheckboxGroup();
	groups.put(groupname, group);
      }
    return (group);
  }
  
  private String getSelectedAsString (ViewElement v) 
  {
    StringBuffer buf = new StringBuffer ();

    add(buf, FORM1, v);
    add(buf, FORM2, v);
    add(buf, FORM3, v);
    add(buf, FORM4, v);

    // Add in the 'submit' button
    String name = v.getAttribute (NAME_ATTR);
    if (name != null && "image".equals (v .getAttribute ("type")) && v instanceof Input)
      {
	Input in = (Input) v;
	if (buf.length() > 0) 
	  {
	    buf.append ('&');
	  }
	buf.append (URLEncoder.encode (name) + ".x");
	buf.append ("=" + in.getX () + "&");
	buf.append (URLEncoder.encode (name) + ".y");
	buf.append ("=" + in.getY ());
      }
    else
      {	
	String value = ((FormElement)v).getFormValue ();
	if (name != null && value != null) 
	  {
	    if (buf.length() > 0) 
	      {
		buf.append ('&');
	      }
	    buf.append (URLEncoder.encode (name));
	    buf.append ('=');
	    buf.append (URLEncoder.encode (value));
	  }
      }
    return (buf.toString ());
  }

  private void add(StringBuffer buf, XPath xpath, Element v)
  {
    NodeList list = XPath.findNodes (this, xpath);
    int len = list.getLength ();
    for (int i = 0; i < len; i++) 
      {
	Element elem = (Element)list.item (i);
	if (elem != v)
	  {
	    String name = elem.getAttribute (NAME_ATTR);
	    String value = ((FormElement)elem).getFormValue ();
	    if (name != null && value != null) 
	      {
	        if (buf.length() > 0) 
	          {
		    buf.append ('&');
	          }
	        buf.append (URLEncoder.encode (name));
	        buf.append ('=');
	        buf.append (URLEncoder.encode (value));
	      }
	  }
      }
  }

}
