/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Frame.java,v 1.15 2001/12/08 01:12:28 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.net.URL;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import org.xoe.core.content.DocumentCache;
import org.xoe.core.dom.XDocument;
import org.xoe.core.xpath.XPath;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ScrollPane;
import org.xoe.display.dom.ViewElement;
import org.xoe.display.dom.Viewable;


public class Frame
  extends ScrollPane
  implements DocumentViewer 
{
  private XDocument redir;
  private URL url;

  public Frame () 
  {
  }

  public void layout (LayoutCursor ocur) 
  {
    if (redir == null) 
      {
	String src = getAttribute ("src");
	if (src != null) 
          {
	    redir = DocumentCache.fetchDocument (getURL (src));
	  }
      }

    if (redir != null) 
      {
	redir.clearModified ();
	redir.version = ((XDocument)getOwnerDocument ()).version;
      }

    super.layout (ocur);
  }


  public boolean hasChildNodes() 
  {
    return ((redir != null) || super.hasChildNodes ());
  }

  protected void layoutChildren (LayoutCursor cur)
  {
    Node child;
    if (redir != null)
      {
	child = redir.getDocumentElement ();
      }
    else
      {
	child = getFirstChild ();
      }
    for (; child != null; child = child.getNextSibling ())
      {
	if (child instanceof Viewable)
	  {
	    ((Viewable)child).layout (cur);
	  }
      }
  }

  public Document getDocument () 
  {
    return (redir);
  }

  public URL getURL ()
  {
    if (url != null)
      {
	return url;
      }
    else
      {
	return redir.getURL ();
      }
  }


  public void setDocument (Document doc)
  {
    if (doc != null && doc instanceof XDocument)
      {
	setDocument (doc, ((XDocument) doc).getURL ());
      }
    else
      {
	setDocument (doc, null);
      }
  }

  public void setDocument (Document doc, URL url) 
  {

    if (doc != getDocument ())
      {
	if (redir != null)
	  {
	    redir.setDocumentHolder (null);
	    ((ViewElement)redir.getDocumentElement ()).unlayout ();
	  }
	redir = (XDocument)doc;
	if (redir != null)
	  {
	    redir.setDocumentHolder (this);
	    if (!(redir.getDocumentElement () instanceof ViewElement))
	      {
		Logger.log (this, "Document not viewable: " + XMLWriter.getXML (redir));
		Logger.log (this, new Throwable ());
		redir.setDocumentHolder (null);
		redir = null;
	      }
	  }
	redrawNow (true);
      }

    if (url != null)
      {
	this.url = url;
	String src = url.toString ();
	int refIdx = src.lastIndexOf( '#');
	if ( refIdx > -1 ) 
	  {
	    String ref = src.substring( refIdx+1);
	    Anchor a = (Anchor)XPath.findFirstNode (doc, "//xhtml:a[@name='"+ref+"']");
	    Logger.log (this, " a = " + a + " a.refPosY = " + a.refPosY);
	    if ( a != null )
	      {	  
		setPosition( 0, a.refPosY);
		return;
	      }
	  }
      }
    Logger.log (this, "resetting position");
    resetPosition ();
  }


}
