/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: FrameSet.java,v 1.6 2001/11/07 20:53:11 tim Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.Container;

import kaffe.awt.AWT;

import org.w3c.dom.Node;

import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;


public class FrameSet
  extends ViewElement
{
  int[] rows = new int[1];
  int[] cols = new int[1];
  LayoutCursor cur = new LayoutCursor();
  String havestyle;

  public static class Render
    extends Container 
  {
    
    Render() 
    {
    } 
  }
  
  public Node cloneNode (boolean deep)
  {
    FrameSet nnode = (FrameSet)super.cloneNode (deep);
    nnode.rows = (int [])rows.clone ();
    nnode.cols = (int [])cols.clone ();
    nnode.cur = cur;
    nnode.havestyle = havestyle;
    return nnode;
  }

  public FrameSet () 
  {
  }

  private int[] calcVals (int va[], String val, int max) 
  {
    if (val == null) 
      {
	if (va.length > 1)
	  va = new int[1];
	va[0] = max;
      }
    else 
      {
	int vl = val.length ();
	char ca[] = new char[vl + 1];
	val.getChars ( 0, vl, ca, 0);
	ca[vl] = ',';
	
	int cv = numEntries (ca);
	if (va.length != cv) 
	  {
	    va = new int[cv];
	  }
	
	int sw = getSlackerSize (ca, max);
		
	int total = 0;
	int i1 = 0;
	int cc = 0;
	int rel = 0;
	for (int i = 0; i < ca.length; i++) 
	  {
			char c = ca[i];
			int t;
			if ( c == ',' ) {
				switch( ca[i-1] ) {
				case '%':
				  t = -max * intVal ( ca, i1, i-1) / 100;
				  rel -= t;
				  break;
				case '*':
				  t = sw * Math.max (intVal( ca, i1, i-1), 1);
				  break;
				default:
				  t = intVal (ca, i1, i);
				  break;
				}
				i1 = i+1;
				va[cc++] = t;
				total += (t > 0 ? t : -t);
			}
		}

		int diff = max - total;

		if (max > total) {
		  for (int i = 0; i < va.length; i++) {
		    if (va[i] < 0) 
		      {
			va[i] = -va[i];
		      }
		  }
		}
		else {
		  total -= max;
		  diff = 0;
		  for (int i = 0; i < va.length; i++) 
		    {
		      if (va[i] < 0) 
			{
			  int cdiff = -va[i] * total / rel;
			  va[i] = -va[i] - cdiff;
			  diff += cdiff;
			}
		    }
		  
		  diff -= total;
		}
		va[va.length-1] += diff;
      }
    return va;
  }
  
  private int getSlackerSize (char ca[], int max) 
  {
    int i1, ns, cw;
    
    cw = ns = i1 = 0;
    for (int i = 0; i < ca.length; i++) 
      {
	char c = ca[i];
	if (c == ',') 
	  {
	    switch(ca[i-1]) 
	      {
	      case '%':
		cw += max * intVal( ca, i1, i-1) / 100;
		break;
	      case '*':
		ns += Math.max(intVal( ca, i1, i-1), 1);
		break;
	      default:
		cw += intVal( ca, i1, i);
		break;
	      }
	    i1 = i+1;
	  }	
      }
    
    if (ns == 0 || max - cw < 0) 
      {
	return (0);
      }
    else {
      return ((max - cw) / ns);
    }
  }
  
  private int intVal (char[] ca, int sIdx, int eIdx) 
  {
    int val = 0;
    
    for (int i = sIdx; i < eIdx; i++) 
      {
	if (Character.isDigit (ca[i]))
	  {
	    val = (10 * val) + Character.digit (ca[i], 10);
	  }
	else
	  {
	    break;
	  }
      }
    return val;
  }
  
  public void layout (LayoutCursor ocur) 
  {
    cur.set (ocur, this);
	
    rows = calcVals (rows, getAttribute( "rows"), cur.height);
    cols = calcVals (cols, getAttribute("cols"), cur.width);

    ViewElement ve = (ViewElement)getFirstChild ();

    int y = 0;
    int x = 0;
    for (int r = 0; r < rows.length; r++) 
      {
	x = 0;
	for (int c = 0; c < cols.length; c++) 
	  {
	    cur.setArea(ocur.x + x, ocur.y + y, cols[c], rows[r]);
	    ve.layout (cur);
	    cur.finish ();
	    
	    for (Node n = ve.getNextSibling (); n != null; n = n.getNextSibling ()) 
	      {
		if ((n instanceof Frame) || (n instanceof FrameSet)) 
		  {
		    ve = (ViewElement)n;
		    break;
		  }
	      }
	    
	    x += cols[c];
	  }
	y += rows[r];
	}

    if (cur.isPhase3 ())
      {
	ocur.addVis (this);
	ocur.setLocation (vis);
	vis.setSize(x, y);
      }
    ocur.move(x, y);
  }

  protected void makeVis () 
  {
    vis = new Render ();
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    if ("xoe:style".equalsIgnoreCase(qualifiedName)) 
      {
	havestyle = value;
      }
    super.setAttributeNS (namespaceURI, qualifiedName, value);
  }

  private int numEntries(char[] ca) 
  {
    int ne = 0;
    
    for (int i = 0; i < ca.length; i++) 
      {
	if (ca[i] == ',')
	  {
	    ne++;
	  }
      }
	
    return ne; 
  }
}
