/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Image.java,v 1.22 2001/12/04 00:35:35 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Panel;
import java.awt.Toolkit;
import java.awt.event.MouseEvent;
import java.awt.image.ImageObserver;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;

import kaffe.awt.AWT;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.FloatViewElement;
import org.xoe.display.dom.ImageFactory;
import org.xoe.display.dom.ImageHolder;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.Thread;


public class Image
  extends FloatViewElement 
  implements ImageHolder 
{
  String url;
  String alt;
  java.awt.Image img;
  Anchor anchor;
  int width;
  int height;
  int border = 1;
  int align = LayoutCursor.NONE;
  int hspace = 0;
  int vspace = 0;
  int baseline = 0;
  String clientImageMap;
  boolean serverImageMap = false; 
  Color alink;
  int version;
  
  public Image () 
  {
  }

  public void imageDeliver (String url, java.awt.Image nimg)
  {
    if (vis != null && nimg != null && (vis.checkImage(nimg, null) & ImageObserver.FRAMEBITS) != 0)
      {
	// Animated - register observer so we get updates
	vis.checkImage(nimg, new ImageObserver ()
	  {
	    public boolean imageUpdate (java.awt.Image img, int infoflags, int x, int y, int width, int height)
	    {
	      redraw (false);
	      return (true);
	    }
	  }
        );
      }
    if (img != nimg)
      {
	img = nimg;
	if (img != null &&  (width <= hspace * 2 || height <= vspace * 2 
			     || img.getWidth(null) + vspace*2 != width  
			     || img.getHeight(null) + hspace*2 != height))
	  {
	    redraw (true);
	  }
	else
	  {
	    redraw (false);
	  }
      }
  }
 
  public void layout (LayoutCursor cur) 
  {
    width = 0;
    height = 0;

    int docversion = ((XDocument)getOwnerDocument ()).version;
    if (img != null && cur.isPhase1 () && version != docversion)
      {
	img.flush ();
      }
    version = docversion;

    if (img == null && url != null) 
      {
	String u = getURLString (url);

        //System.out.println("Image url is " + u); 
        
	if (u != null) 
	  {
	    ImageFactory.getImage (u, this);
	  }
      }
    if (img != null) 
      {
	width = img.getWidth (null);
	if (width < 0) 
	  {
	    width = 0;
	  }
	height = img.getHeight (null);
	if (height < 0) 
	  {
	    height = 0;
	  }
      }
    else if (alt != null)
      {

	java.awt.FontMetrics fm = cur.fontmetrics;
	width = fm.stringWidth (alt);
	height = fm.getHeight();
	baseline =  height - fm.getDescent ();
      }

    int pwidth = getPreferredWidth (cur.width);
    int pheight = getPreferredHeight (cur.height);
	
    if (pwidth >= 0 && pheight >= 0) 
      {
	width = pwidth;
	height = pheight;
      }
    else if (pwidth >= 0 && width > 0) 
      {
	height = height * pwidth / width;
	width = pwidth;
      }
    else if (pheight >= 0 && height > 0) 
      {
	width = width * pheight / height;
	height = pheight;
      }

    // Add in the spacing
    width += hspace * 2;
    height += vspace * 2;

    if (cur.isPhase3 ())
      {
	// Allow anchor linking
	if (cur.anchor != null || clientImageMap != null) 
	  { //System.out.println("Here in image");
	    flags |= HAS_MOUSE_LISTENER;
	    anchor = (Anchor)cur.anchor;
	    getVis ().setCursor (Anchor.anchorCursor);
	  }
	alink = cur.activeclr;
	cur.addVis (this);
	vis.setFont (cur.font);
	vis.setSize (width, height);
	vis.setForeground (cur.fgclr);
      }

    switch (align) 
      {
      case LayoutCursor.LEFT:
      case LayoutCursor.RIGHT:
	cur.flow (this, align, width, width, height);
	break;
      case LayoutCursor.TOP:
      case LayoutCursor.BOTTOM:
	cur.valign = align;
	// Fall through
      case LayoutCursor.NONE:
      default:
	cur.newline (width);
	if (cur.isPhase3 ()) 
	  {
	    cur.setLocation (vis);
	  }
	cur.move (width, height);
	break;
      }
  }

  protected void makeVis () 
  {
    vis = new Render ();
  }

  public void mousePressed (MouseEvent evt) 
  {
    Graphics g = vis.getGraphics();
    g.setColor(alink);
    vis.paint(g);
    g.dispose();
  }
  
  public void mouseClicked (MouseEvent evt) 
  {
    if (anchor != null) 
      {
        if (!serverImageMap)
	  {
            anchor.mouseClicked (evt);
          }
        else
	  {
            invoke (HREF_ATTR, anchor.getAttribute(HREF_ATTR) + "?" + evt.getX () + "," +  evt.getY ()) ;
          }
      }
    else if (clientImageMap != null)
      {
        try
          {
            String href = ImageMap.getAnchor (getOwnerDocument (), clientImageMap, evt.getX (), evt.getY ());
            if (href != null)
              {
                invoke (HREF_ATTR, href);
              }
          }
        catch (NoClassDefFoundError ncdfe)
          {
            // Image maps support not loaded - ignore
          }
      }    
    else
      {
        invoke (ONCLICK_ATTR);
      }

  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);
    if (localName.equalsIgnoreCase ("src") || localName.equalsIgnoreCase ("data")) 
      {
	url = value;
	if (vis != null) 
	  { //this will not get the base url changes 
	    ImageFactory.getImage (getURLString (url), this);
	  }
	
      }
    else if (localName.equalsIgnoreCase ("alt") && value.length () > 0) 
      {
	alt = "[" + value + "]";
      }
    else if (localName.equalsIgnoreCase ("align")) 
      {
	align = Emphasis.getAlign (value, align);
      }
    else if (localName.equalsIgnoreCase ("border")) 
      {
	try 
	  {
	    border = Integer.parseInt (value, 10);
	  }
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase ("hspace")) 
      {
	try 
	  {
	    hspace = Integer.parseInt (value, 10);
	  }
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase ("vspace")) 
      {
	try 
	  {
	    vspace = Integer.parseInt (value, 10);
	  }
	catch (NumberFormatException nfe) 
          {
	    Logger.log (this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase ("usemap")) 
      {
	clientImageMap = value;
      }
    else if (localName.equalsIgnoreCase ("ismap")) 
      {
	serverImageMap = true;
      }
    super.setAttributeNS( namespaceURI, localName, value);
  }

  public Node cloneNode (boolean deep) 
  {
    Image cloned = (Image)super.cloneNode (deep);
    cloned.url = url;
    cloned.alt = alt;
    cloned.img = img;
    cloned.anchor = null;
    cloned.width = width;
    cloned.height = height;
    cloned.border = border;
    cloned.align = align;
    cloned.hspace = hspace;
    cloned.vspace = vspace;
    cloned.clientImageMap = clientImageMap;
    cloned.version = version;
    
    return cloned;
  }

  public java.awt.Image getImage ()
  {
    return (img);
  }

  class Render
    extends Component 
  {
    
    public void paint (Graphics g) 
    {
      int w = Image.this.width;
      int h = Image.this.height;

      // If we don't need to draw the background then dont (to avoid flicker)
      if (img == null || hspace > 0 || vspace > 0)
	{
	  AWT.extension.paintBackground (this, g);
	}
	
      if (img != null)
	{
	  g.drawImage(img, hspace, vspace, w - (hspace * 2), h - (vspace * 2), null);
	}
      else if (alt != null) 
	{
	  g.drawString(alt, hspace, vspace + baseline);
	}
      if (anchor != null && border > 0)
	{
	  // HACK - is we have alpha set to 0 then we don't paint
	  if ((g.getColor().hashCode() & 0xFF000000) != 0) 
	    {
	      g.drawRect(hspace, vspace, w - (hspace * 2) - 1, h - (vspace * 2) - 1);
	    }
	}
    }
    
  }
}
