/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ImageMap.java,v 1.8 2001/10/26 18:39:35 philippe Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.util.StringTokenizer;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;

import org.xoe.core.dom.SingleNodeList;

class ImageMap
{
  static String getAnchor (Document doc, String imagemap, int mx, int my)
  {
    if (imagemap.charAt (0) != '#')
      {
        Logger.log (ImageMap.class, "image maps to other documents not yet supported");
        return (null);
      }

    String xpath = "//xhtml:map[@name='" + imagemap.substring(1) + "']/xhtml:area";
    NodeList areas = XPath.findNodes(doc, xpath);
    for (int i = areas.getLength() - 1; i >= 0; i--)
      {
        Element n = (Element)areas.item(i);

        
        String shapeString = n.getAttribute("shape");
	
        String coords = n.getAttribute("coords");
        
        if (coords != null)
          {
            StringTokenizer tok = new StringTokenizer(coords, ",");
            
           if(shapeString != null && (shapeString.equalsIgnoreCase("circle") || shapeString.equalsIgnoreCase("circ")))
	      {
                if (tok.countTokens() == 3)
	          {
                    try
                    {
                      int x = Integer.parseInt(tok.nextToken());
                      int y = Integer.parseInt(tok.nextToken());
                      int r = Integer.parseInt(tok.nextToken());
                     
                      if ( Math.sqrt((mx - x) * (mx - x) + (my - y) * (my - y)) <= (double) r )
		        { 
                          return (n.getAttribute("href"));
		        }
                    }
                    catch (NumberFormatException _)
                    {
                    }
                  }
                
              } //CIRCLE
	    else if(shapeString != null && (shapeString.equalsIgnoreCase("polygon") || shapeString.equalsIgnoreCase("poly")))
	      {
		  int numTokens = tok.countTokens();

		  if (numTokens % 2 == 0) //must have even number of coordinates
	          {
		    int mCoords[] = new int[numTokens];

                    try
                    {
		      for (int j = 0; j < numTokens; j++)
			 {
                           mCoords[j] = Integer.parseInt(tok.nextToken());
                         }
                    
                      if (isInsidePoly(mCoords, mx, my))
                        {
                          return (n.getAttribute("href"));
			}
                    }
                    catch (NumberFormatException _)
                    {
                    }
                  }
                
              } //POLYGON
	   else // RECTANGLE
	      {
                if (tok.countTokens() == 4)
	          {
                    try
                    {
                      int x = Integer.parseInt(tok.nextToken());
                      int y = Integer.parseInt(tok.nextToken());
                      int xw = Integer.parseInt(tok.nextToken());
                      int yh = Integer.parseInt(tok.nextToken());

                      if (mx >= x && mx <= xw && my >= y && my <= yh)
                        {
                          return (n.getAttribute("href"));
                        }
                    }
                    catch (NumberFormatException _)
                    {
                    }
                  }
                
              } //RECTANGLE  
	  } //coords != null
      

       }//for

    return null;
  }//method

  

  //adapted and translated from mozilla
  private static boolean isInsidePoly(int mCoords[], int x, int y)
  {
    int mNumCoords = mCoords.length;

    if (mNumCoords >= 6) 
      {
        int intersects = 0;
        int wherex = x;
        int wherey = y;
        int totalv = mNumCoords / 2;
        int totalc = totalv * 2;
        int xval = mCoords[totalc - 2];
        int yval = mCoords[totalc - 1];
        int end = totalc;
        int pointer = 1;
    
        if ((yval >= wherey) != (mCoords[pointer] >= wherey))
	  {
            if ((xval >= wherex) == (mCoords[0] >= wherex))
	      {
                intersects += (xval >= wherex) ? 1 : 0;
              }
            else
	      {
                intersects += ((xval - (yval - wherey) * (mCoords[0] - xval) / (mCoords[pointer] - yval)) >= wherex) ? 1 : 0;
              }
          }
  

     
        while (pointer < end)  
          {
            yval = mCoords[pointer];

            pointer += 2;

            if (yval >= wherey)  
              {
                while((pointer < end) && (mCoords[pointer] >= wherey))
		  {
                    pointer+=2;
                  }
                if (pointer >= end)
		  {
                    break;
                  }

                if ((mCoords[pointer-3] >= wherex) == (mCoords[pointer-1] >= wherex)) 
                  {
                    intersects += (mCoords[pointer-3] >= wherex) ? 1 : 0;
                  } 
                else 
                  {
                    intersects += ((mCoords[pointer-3] - (mCoords[pointer-2] - wherey) *
                                  (mCoords[pointer-1] - mCoords[pointer-3]) /
                                  (mCoords[pointer] - mCoords[pointer - 2])) >= wherex) ? 1:0;
                  }
              }  
            else  
              {
                while ((pointer < end) && (mCoords[pointer] < wherey))
		  {   
                    pointer+=2;
                  }
                if (pointer >= end)
		  {
                    break;
                  }

                if ((mCoords[pointer-3] >= wherex) == (mCoords[pointer-1] >= wherex)) 
                  {
                    intersects += (mCoords[pointer-3] >= wherex) ? 1:0;
                  } 
                else 
                  {
                    intersects += ((mCoords[pointer-3] - (mCoords[pointer-2] - wherey) *
                                  (mCoords[pointer-1] - mCoords[pointer-3]) /
                                  (mCoords[pointer] - mCoords[pointer - 2])) >= wherex) ? 1:0;
                  }
               }
        }
       
 
        if ((intersects & 1) != 0) 
          {
            return true;
          }
      }

      return false;

  } //method


}//class




