/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Input.java,v 1.24 2001/12/11 19:38:30 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;

import com.pocketlinux.awt.theme.StyleComponentDeco;
import com.pocketlinux.awt.theme.Theme;

import java.awt.Button;
import java.awt.Component;
import java.awt.Checkbox;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.net.URL;
import java.net.URLEncoder;
import java.net.URLDecoder;

import kaffe.awt.AWT;
import kaffe.awt.widgets.ImageButton;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.display.dom.ImageFactory;
import org.xoe.display.dom.ImageHolder;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;


public class Input
  extends ViewElement
  implements FormElement, ActionListener, KeyListener
{
  String type = "field";
  String value = "";
  String name = "";
  String src;
  Image img;
  boolean checked = false; // initial value only
  boolean editable = true;
  boolean enabled = true;
  int size = 20;
  int maxlength = 0;
  int border = 1;
  int min = 0;
  int max = 100;
  int pos = 0;
  int inc = 30;
  int ori = Scrollbar.VERTICAL;
  int clickX, clickY;
  String alt;

  String defaultValue = value;
  boolean defaultChecked = checked;
  boolean defaultEnabled = enabled;
  int defaultPos = pos;

  class ButtonRender
    extends ImageButton
  {
    ButtonRender() 
    {
      super (Input.this.img);
    }
  
    public boolean imageUpdate (Image img, int infoflags, int x, int y, int w, int h) 
    {
      if ((infoflags & (ALLBITS | FRAMEBITS)) != 0) 
	{
	  Input.this.redraw (true);
	}
      return (super.imageUpdate (img, infoflags, x, y, w, h));
    }

    public Dimension preferredSize () 
    {
      Insets ins = AWT.extension.getInsets (this);
      int width = ins.left + ins.right;
      int height = ins.top + ins.bottom;
      if (Input.this.img == null) 
        {
	  FontMetrics fm = getFontMetrics (getFont ());
	  width += fm.stringWidth (value);
	  height += fm.getHeight ();
        }
      else 
        {
	  width += Input.this.img.getWidth (this);
	  height += Input.this.img.getHeight (this);
        }
      return (new Dimension (width, height));
    }
  }
  
  public class BorderTextField
    extends TextField
  {
  }

  class ImageRender
    extends Component implements ImageHolder
  {

    public FontMetrics fm;

    public void paint (Graphics g) 
    {
      if (Input.this.img != null) 
	{
	  g.drawImage(Input.this.img, 0, 0, this);
	}
      else if (alt != null)
	{
	  fm = g.getFontMetrics ();
	  g.drawString (alt, 0, fm.getHeight () - fm.getDescent ());
	}
    }

    public void imageDeliver (String url, Image nimg)
    {
      Input.this.img = nimg;
      Input.this.redraw (true);
    }

    public Dimension preferredSize () 
    {
      Insets ins = AWT.extension.getInsets (this);
      int width = ins.left + ins.right;
      int height = ins.top + ins.bottom;
      if (Input.this.img != null) 
	{
	  width += Input.this.img.getWidth (this);
	  height += Input.this.img.getHeight (this);
	}
      else if (fm != null && alt != null)
	{
	  width += fm.stringWidth (alt);
	  height += fm.getHeight ();
	}
	return (new Dimension (width, height));
    }
  }

  class CheckboxRender
    extends Checkbox
  {
    CheckboxRender(String common)
    {
      if (common != null)
	{
	  Form form = getForm();
	  if (form != null)
	    {
	      setCheckboxGroup(form.getCheckboxGroup(common));
	    }
	}
      setState (checked);
    }
  }
  
  public Input () 
  {
  }

  public void layout(LayoutCursor cur)
  {
    if (getVis () instanceof TextField && cur.bgclr != null)
      {
	getVis (). setBackground (cur.bgclr);
      }
    super.layout (cur);
  }

  public Node cloneNode (boolean deep) 
  {
    Input nnode = (Input)super.cloneNode (deep);
    nnode.type = type;
    nnode.value = value;
    nnode.name = name;
    nnode.src = src;
    nnode.img = img;
    nnode.size = size;
    nnode.maxlength = maxlength;
    nnode.border = border;
    nnode.min = min;
    nnode.max = max;
    nnode.pos = pos;
    nnode.inc = inc;
    nnode.ori = ori;
    return nnode;
  }

  public String getButtonName() 
  { 
    if (vis instanceof Button)
      {
	return name;
      }
    else
      {
	return null;
      }
  }

  public String getText () 
  {
    if (vis instanceof TextField) 
      {
	return (((TextField)vis).getText ());
      }
    else {
      return ("");
    }
  }

  protected void makeVis () 
  {
    if (type.equalsIgnoreCase ("button")) 
      {
	ButtonRender ib = new ButtonRender ();
	if ( value.length() == 0 ) 
	  {
	    value = name;
	  }
	ib.setLabel (value);
	vis = ib;
        
      }
    else if (type.equalsIgnoreCase ("submit") || type.equalsIgnoreCase("reset")) 
      {
	ButtonRender ib = new ButtonRender ();
	ib.setLabel (value);
	vis = ib;
	flags |= HAS_MOUSE_LISTENER;
        
      }
    else if (type.equalsIgnoreCase ("hidden")) 
      {
	// Do nothing
      }
    else if (type.equalsIgnoreCase ("image")) 
      {
	ImageRender ib = new ImageRender ();
	ib.setCursor (Anchor.anchorCursor);
	vis = ib;
	String u = getURLString (src);
	if (u != null) 
	  {
	    ImageFactory.getImage (u, ib);
	  }
	flags |= HAS_MOUSE_LISTENER;
      }
    else if (type.equalsIgnoreCase ("scrollbar")) 
      {
	vis = new Scrollbar(ori, pos, inc, min, max);
      }
    else if (type.equalsIgnoreCase ("radio")) 
      {
	vis = new CheckboxRender (name);
      }
    else if (type.equalsIgnoreCase ("checkbox")) 
      {
	vis = new CheckboxRender (null);
      }
    else
      {
	TextField tf;
	if (border == 0) 
	  {
	    tf = new TextField ();
	  }
        else
	  {
	    tf = new BorderTextField ();
          }
        tf.setText (value);
        tf.setColumns (size);
        vis = tf;
        if (type.equalsIgnoreCase ("password")) 
	  {
	    tf.setEchoChar ('*');
	  }
	tf.addActionListener(this);
	tf.addKeyListener (this);

        tf.setEditable(editable);
      }

    if (vis != null)
      {
        vis.setEnabled(enabled);
      }
  }

  /**
   * When someone hits a submit, we have to walk up our parent
   * chain looking for the form we contain.  If we find one, we
   * dispatch the form instead.  Otherwise, we do the default.
   */
  public void mouseClicked (MouseEvent evt) 
  {
    if (type.equalsIgnoreCase("submit") ||
	type.equalsIgnoreCase("image")) 
      {
	clickX = evt.getX ();
	clickY = evt.getY ();
	Form f = getForm();
	if (f != null)
	  {
	    f.dispatchForm(this);
	    return;
	  }
      }
    else if (type.equalsIgnoreCase ("reset"))
      {
	Form f = getForm();
	if (f != null)
	  {
	    f.resetForm();
	    return;
	  }
      }
    super.mouseClicked (evt);
  }

  public int getX ()
  {
    return clickX;
  }

  public int getY ()
  {
    return clickY;
  }

  public void keyReleased (KeyEvent e)
  {
  }

  public void keyTyped (KeyEvent e)
  {
  }

  public void keyPressed (KeyEvent e) 
  {
    // only invoke the callback if the onkeypress attribute is set
    if (hasAttribute (ONKEYPRESS_ATTR))
      {
	String name = getAttribute ("name");
	if (name == null)
	  {
	    name = "value";
	  }

	String payload = getAttribute (ONKEYPRESS_ATTR)
                         + '?' 
                         + URLEncoder.encode (name) 
                         + '=' 
                         + URLEncoder.encode (getFormValue ());

	invoke (ONKEYPRESS_ATTR, payload);
      }
  }

  // If someone hits enter on a text entry widget, try to submit the
  // enclosing form.
  public void actionPerformed (ActionEvent evt) 
  {
    Form f = getForm ();
    if (type.equalsIgnoreCase ("text") && f != null)
      {
	f.dispatchForm (this);
	return;
      }
  }

  public Form getForm()
  {
    for (Node n = this; n != null; n = n.getParentNode ()) 
      {
        if (n instanceof Form) 
          {
	    return ((Form)n);
	  }
      }
    return (null);
  }
  
  public void setAttributeNS(String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);
    if (localName.equalsIgnoreCase ("type")) 
      {
	if (!type.equals (value)) 
	  {
	    type = value;
	    vis = null;
	  }
      }
    else if (localName.equalsIgnoreCase (VALUE_ATTR)) 
      {
	this.value = value;
	this.defaultValue = value;
	if (vis instanceof TextField) 
	  {
	    ((TextField)vis).setText (value);
	  }
      }
    else if (localName.equalsIgnoreCase ("size")) 
      {
	try {
	  size = Integer.parseInt (value);
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
	if (vis instanceof TextField) 
	  {
	    ((TextField)vis).setColumns (size);
	  }
      }
    else if (localName.equalsIgnoreCase ("maxlength")) 
      {
	try {
	  maxlength = Integer.parseInt (value);
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase (NAME_ATTR)) 
      {
	this.name = value;
	if (vis instanceof Button) 
	  {
	    ((Button)vis).setLabel (name);
	  }
      }
    else if (localName.equalsIgnoreCase ("src")) 
      {
	this.src = value;
      }
    else if (localName.equalsIgnoreCase ("img")) 
      {
	this.img = lookupIcon (value);
	if (vis instanceof ImageRender) 
	  {
	    vis.repaint ();
	  }
      }
    else if (localName.equalsIgnoreCase ("border")) 
      {
	try {
	  border = Integer.parseInt (value);
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
	}
        else if (localName.equalsIgnoreCase ("orientation")) 
	  {
	    if (value.equals("horizontal")) 
	      {
		ori = Scrollbar.HORIZONTAL;
		setPreferredWidth(-100);
	      }
	    else 
	      {
		ori = Scrollbar.VERTICAL;
		setPreferredHeight(-100);
	      }
	  }
    else if (localName.equalsIgnoreCase ("min")) 
      {
	try {
	  min = Integer.parseInt (value);
	}
	catch (NumberFormatException _) 
	  {
	  }
      }
    else if (localName.equalsIgnoreCase ("max")) 
      {
	try {
	  max = Integer.parseInt (value);
	}   
	catch (NumberFormatException _) 
	  {
	  }   
      }
        else if (localName.equalsIgnoreCase ("pos")) 
	  {
	    try 
	      {
		pos = Integer.parseInt (value);
		defaultPos = pos;
	      }
	    catch (NumberFormatException _) 
	      {
	      }
	  }
    else if (localName.equalsIgnoreCase ("inc")) 
      {
	try 
	  {
	    inc = Integer.parseInt (value);
	  }
	catch (NumberFormatException _) 
	  {
	  }
      }
    else if (localName.equalsIgnoreCase ("checked")) 
      {
	if ("on".equalsIgnoreCase (value) || "checked".equalsIgnoreCase (value))
	  {
	    checked = true;
	  }
	else if (value == null)
	  {
	    checked = false;
	    defaultChecked = false;
	  }
      }
    else if (localName.equalsIgnoreCase ("readonly")) 
      {
	if ("readonly".equals (value) || "true".equals (value))
	  {
            editable = false;
          }
	else if (value == null)
	  {
	    editable = true;
	  }
      }
    else if (localName.equalsIgnoreCase ("disabled")) 
      {
	if ("disabled".equalsIgnoreCase (value) || "true".equalsIgnoreCase (value))
	  {
            enabled = false;
	    defaultEnabled = false;
          }
	else if (value == null)
	  {
	    enabled = true;
	  }
	       
      }
    else if (localName.equalsIgnoreCase ("alt"))
      {
	alt = value;
      }

    super.setAttributeNS(namespaceURI, localName, value);
  }

  public void setButtonName (String bname) 
  { 
    if (vis instanceof Button) 
      { 
	name = bname;
	((Button) vis).setLabel (name);
	redraw (true);
      }
  }
  
  public void setText (String text) 
  {
    value = text;
    if (vis instanceof TextField) 
      {
	((TextField)vis).setText (text);
      }
  }

  public int getValue () 
  {
    if (vis instanceof Scrollbar) 
      {
	pos = ((Scrollbar)vis).getValue ();
      }
    return (pos);
  }

  public void setValue (int val) 
  {
    pos = val;
    if (vis instanceof Scrollbar) 
      {
	((Scrollbar)vis).setValue (pos);
      }
  }

  protected StyleComponentDeco getDeco () 
  {
    if (vis instanceof Scrollbar) 
      {
	return (((Theme)AWT.getTheme()).getDecoWithStyle(vis, style, (ori == Scrollbar.VERTICAL ? "vert" : "horz")));
      }
    else {
      return (super.getDeco ());
    }
  }

  /**
   *  For the first node in the list <code>arg</code>, get it's text content and
   *  initialized the input box with it.
   */ 
  public void watch (NodeList arg) 
  {
    super.watch (arg);
    if(arg != null) 
      {
	// if arg is an xpath result, grab the first item
	if (arg.getLength () > 0) 
	  {
            String text = DOMUtils.getTextContent (arg.item (0));
            if (text != null)
              {
		setText (text);
              }
	    else 
	      {
		Logger.log( this, "watch: null text for arg.item (0)  = " + XMLWriter.getXML (arg.item (0)));
	      }
	  }
	else 
	  {
	    Logger.log( this, "Nothing to watch for: " + getAttribute (NAME_ATTR));
	  }
      }
  }
  
  public String getFormValue () 
  {
    if (!enabled && !("hidden".equalsIgnoreCase (type)))
      {
	return (null); 
      } 

    if ("button".equalsIgnoreCase (type)) 
      {
	return (null);
      }
    else if ("submit".equalsIgnoreCase (type)) 
      {
	return (value);
      }
    else if ("hidden".equalsIgnoreCase (type)) 
      {
	return (value);
      }
    else if ("image".equalsIgnoreCase (type)) 
      {
	return (null); // We're suppose to send coordinates apparently
      }
    else if ("scrollbar".equalsIgnoreCase(type)) 
      {
	return ("" + getValue());
      }
    else if ("radio".equalsIgnoreCase(type) || "checkbox".equalsIgnoreCase(type)) 
      {
	if (vis != null && ((CheckboxRender)vis).getState())
	  {
	    return (value);
	  }
	else
	  {
	    return (null);
	  }
      }
    else 
      {
	    return (getText());
      }
  }

  // a bit hacky, since getAttributes will only have the original attributes and values
  public String getAttributeNS (String namespaceURI, String localName) 
  {
    if (namespaceURI == null && "value".equals (localName))
      {
	return getFormValue ();
      }
    else
      {
	return super.getAttributeNS (namespaceURI, localName);
      }
  }  
 
  
  public void resetValue ()
  {
    value = defaultValue;
    enabled = defaultEnabled;
    checked = defaultChecked;
    if (vis instanceof TextField)
      {
	TextField tf = (TextField) vis;
	tf.setText (defaultValue);
	tf.setEnabled (defaultEnabled);
      }
    else if (vis instanceof CheckboxRender)
      {
	CheckboxRender ck = (CheckboxRender) vis;
	ck.setState(defaultChecked);
	ck.setEnabled (defaultEnabled);
      }
    else if (vis instanceof Scrollbar)
      {
	((Scrollbar) vis).setValue (pos);
      }
  }
 
}
