/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ListItem.java,v 1.5 2001/11/06 00:15:43 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.Component;
import java.awt.Graphics;

import org.w3c.dom.Node;

import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;


public class ListItem
  extends ViewElement 
{
  private static final char[] indentTxt = { '8', '8', '8', '8' };

  private int indent;
  private int height;
  private int width;
  private char[] text;
  
  class Render extends Component 
  {
    
    public void paint (Graphics g) 
    {
      if (text != null) 
	{
	  g.drawChars (text, 0, text.length, indent * 4 / 5 - width, height);
	}
      else {
	g.drawOval (indent * 3 / 5, height / 2, 3, 3);
      }
    }

  }

  public void layout (LayoutCursor cur) 
  {
    Node pn = getParentNode ();
    String tagname = pn.getLocalName ();
    if ("ol".equalsIgnoreCase (tagname)) 
      {
	Itemize di = (Itemize)pn;
	int count = di.count++;
	switch (di.type) 
	  {
	  case 'A':
	    count = count % 26;
	    text = ((char)((int)'A' + count) + ".").toCharArray ();
	    break;
	  case 'a':
	    count = count % 26;
	    text = ((char)((int)'a' + count) + ".").toCharArray ();
	    break;
	  case 'I':
	  case 'i':
	    // Not supported yet
	  case '1':
	  default:
	    count = count % 100;
	    text = ((1+count) + ".").toCharArray ();
	    break;
	  }
	width = cur.fontmetrics.charsWidth (text, 0, text.length);
      }
    else 
      {	// <ul>, <menu> and <dir>
      }
    int ymove = cur.fontmetrics.getHeight ();
    height = ymove - cur.fontmetrics.getDescent ();
    indent = cur.fontmetrics.charsWidth (indentTxt, 0, indentTxt.length);
    
    int tindent = indent * cur.listlevel;
    cur.move (indent * (cur.listlevel - 1), 0);
    if (cur.isPhase3 ())
      {
	cur.addVis (this);
	vis.setFont (cur.font);
	vis.setForeground (cur.fgclr);
	cur.setLocation (vis);
	vis.setSize (indent, ymove);
      }
    cur.move (indent, ymove);
    
    cur.leftindent += tindent;
    layoutChildren (cur);
    cur.leftindent -= tindent;
    cur.newline ();
  }

  protected void makeVis () 
  {
    vis = new Render ();
  }

  public org.w3c.dom.Node cloneNode (boolean deep) 
  {
    ListItem nnode = (ListItem)super.cloneNode (deep);
    nnode.indent = indent;
    nnode.height = height;
    nnode.width = width;
    if (text != null )
      {
	nnode.text = (char[])text.clone ();
      }
    else 
      {
	nnode.text = null;
      }
    return nnode;
  }
}
