/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Select.java,v 1.43 2001/12/17 22:21:03 damian Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.Component;
import java.awt.Insets;
import java.awt.List;
import java.awt.Choice;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;

import java.io.IOException;
import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Vector;

import kaffe.awt.AWT;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.DocumentCache;
import org.xoe.core.dom.SimpleNodeList;
import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;


public class Select
  extends ViewElement 
  implements ItemListener, ActionListener, FormElement 
{
  protected final static XPath OPTION_PATH = XPathParser.singleton.parse ("xhtml:option");

  protected String dataValue;
  protected int size = 1;

  private String banner;
  private SimpleNodeList list;
  private String label = "Menu";
  private int selected = 0;
  private int defaultSelected = selected;

  private boolean multiple = false;
  private boolean forceSize = false;
  boolean enabled = true;


  private Vector builtOptions;

  // this xpath is used to build default options for the list
  protected XPath optionXPath;
  
  class ListRender 
    extends List 
  {
    ListRender () 
    {
      super (1);
      setMultipleMode (multiple);
    }

    public Dimension preferredSize () 
    {
      int rows = getRows ();
      FontMetrics fm = getFontMetrics (getFont ());
      
      int width = 0;
      for (int i = 0; i < rows; i++) 
	{
	  int iw = fm.stringWidth (getItem (i));
	  if (iw > width) 
	    {
	      width = iw;
	    }
	}
      if (multiple && !forceSize)
        {
          size = rows;
        }
      if (rows > size) 
	{
	  rows = size;
          
          // new width for the preffered size has to also include the
          // size of the menu bar and it does not happen. There is
          // a HACK to add 10 for the scrollbar but it should be given
          // an actual size.
          width += 10;
	}
      return (new Dimension (width + 10, rows * (fm.getHeight () + 1)));
    }
  }
  
  class ChoiceRender 
    extends Choice 
  {
    public Dimension preferredSize () 
    {
      Insets ins = AWT.extension.getInsets (this);
      int w = ins.left + ins.right;
      int h = ins.top + ins.bottom;
      
      FontMetrics mfm = getFontMetrics (getFont ());
      h += mfm.getHeight ();
      int maxw = 0;
      String ms = getSelectedItem ();
      if (ms != null) 
	{
	  maxw = mfm.stringWidth (ms);
	}
      int len = getItemCount ();
      for (int i = 0; i < len; i++) 
	{
	  int sw = mfm.stringWidth (getItem (i));
	  if (sw > maxw) 
	    {
	      maxw = sw;
	    }
	}
      w += maxw;
      
      // HACK
      w += 10;
      
      return (new Dimension (w, h));
    }
  }

  class MenuRender 
    extends Choice 
  {
    public String getSelectedItem () 
    {
      return (label);
    }

    public Dimension preferredSize () 
    {
      Insets insets = AWT.extension.getInsets (this);
      int w = insets.left + insets.right;
      int h = insets.top + insets.bottom;
      
      FontMetrics mfm = getFontMetrics (getFont ());
      w += mfm.stringWidth (label);
      h += mfm.getHeight ();
      
      return (new Dimension (w, h));
    }
  }


  public Select () 
  {
  }

  public void actionPerformed (ActionEvent e) 
  {
    invokeCallback ();
  }
  
  public Node cloneNode (boolean deep)
  {
    Select nnode = (Select)super.cloneNode (deep);
    nnode.banner = banner;
    nnode.list = null;
    nnode.label = label;
    nnode.size = size;
    return nnode;
  }

  protected void makeVis () 
  {
    if (size == 0 && !multiple) 
      {
	makeMenu ();
      }
    else if (size == 1 && !multiple) 
      {
	makeChoice ();
      }
    else 
      { 
	makeList ();
      }

    vis.setEnabled(enabled); 
  }
  
  protected void makeList () 
  {
    ListRender r = new ListRender ();
    r.addActionListener (this);
    vis = r;
  }

  protected void makeChoice () 
  {
    ChoiceRender r = new ChoiceRender ();
    r.addItemListener (this);
    vis = r;
  }
  
  protected void makeMenu () 
  {
    MenuRender r = new MenuRender ();
    r.addItemListener (this);
    vis = r;
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value)
  {
    String localName = getLocalName (qualifiedName);
    if (localName.equalsIgnoreCase ("size")) 
      {
	try 
	  {
	    size = Integer.parseInt (value);
            forceSize = true;
	  }
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase ("banner")) 
      {
	banner = value;
      }
    else if (localName.equalsIgnoreCase ("label")) 
      {
	label = value;
      }
    else if (localName.equalsIgnoreCase ("multiple")) 
      {
	if ("multiple".equalsIgnoreCase (value))
          { 
            multiple = true;
          }
      } 
   
    else if (localName.equalsIgnoreCase ("disabled"))
      {
        if ("disabled".equalsIgnoreCase (value) || "true".equalsIgnoreCase (value))
          {
            enabled = false;
          }
	else if (value == null)
	  {
	    enabled = true;
	  }
      }
    else if (localName.equalsIgnoreCase (XPATH_ATTR)) 
      {
	optionXPath = XPathParser.singleton.parse (value);
      }

    super.setAttributeNS (namespaceURI, qualifiedName, value);
  }

  public void itemStateChanged (ItemEvent e) 
  {
    if (e.getStateChange () == ItemEvent.SELECTED)
      {
	// Put selected item into a node list so we can retrieve
	// it from generic application code
	Object item = e.getItem ();
	if (item instanceof Integer) 
	  {
	    selected = ((Integer)item).intValue ();
	  }
	else if (item instanceof String) 
	  {
            // store the selected value
	    int oldSelected = selected;
            // get the new selected index
	    selected = ((Choice)vis).getSelectedIndex ();

            // check to see if this option is a trigger, which means we 
            // don't mark it as selected
            if (((Element)list.item (selected)).hasAttributeNS (DEFAULT_NAMESPACE, TRIGGER_ATTR))
              {
                selected = oldSelected;
              }
        
            // save the old data value
            dataValue = DOMUtils.getAssignedValue (list.item (selected));
	  }

        invokeCallback ();
      }
  }
    
  private void invokeCallback ()
  {
    // only invoke the callback if the onchange attribute is set
    if (hasAttribute (ONCHANGE_ATTR))
      {
	String name = getAttribute ("name");
	if (name == null)
	  {
	    name = "value";
	  }

	String payload = getAttribute (ONCHANGE_ATTR)
                         + '?' 
                         + URLEncoder.encode (name) 
                         + '=' 
                         + URLEncoder.encode (getFormValue ());

	invoke (ONCHANGE_ATTR, payload);
      }
  }
  
  /**
   * Override mouseClicked to avoid default.
   */
  public void mouseClicked (MouseEvent evt) 
  {
  }

  public void resetList () 
  {
    getCVis ().removeAll ();
    if (list == null) 
      {
	list = new SimpleNodeList ();
      }
    else 
      {
	list.removeAllElements ();
      }
  }

  /**
   *  we have to override this method because the the select tag has some
   *  more complex behaviour than your standard view element.  
   */
  public void onLoad (XPathContext xpathContext, Node contextNode) 
  {
    // only call the onload callback if neither a data nor option xpath was
    // specified since without either of them we can't initialize the select
    if (dataXPath == null && optionXPath == null) 
      {
	invoke(ONLOAD_ATTR);
      }
    else 
      {
        // first check for an optionXPath - if one exists, then we proceed
        // to get a nodelist of options for this select.
        if (optionXPath != null)
          {
	    Node optionContext = null;

	    // first thing, check for a context attribute. if it exists, it will 
	    // specify a context to use for building the options
	    String contextPath = getAttributeNS (DEFAULT_NAMESPACE, CONTEXT_ATTR);
	    if (contextPath != null)
	      {
                String pkgWorkspace = getOwnerDocument ().getDocumentElement ().getAttributeNS (DEFAULT_NAMESPACE, PKG_WORKSPACE_ATTR);
                String pkgArchive = getOwnerDocument ().getDocumentElement ().getAttributeNS (DEFAULT_NAMESPACE, PKG_ARCHIVE_ATTR);
                try
                  {
		    optionContext = DocumentCache.fetchFromPackageWorkspace (contextPath, pkgWorkspace, pkgArchive);
                  }
                catch (IOException ioe)
                  {
                    Logger.log (this, ioe);
                  }
	      }
	    // otherwise, use the context past in
	    else
	      {
		optionContext = contextNode;
	      }

	    if (optionContext == null)
	      {
		Logger.log (this, "onLoad: option context is null");
		Logger.log (this, "onLoad: this is " + XMLWriter.getXML (this));
	      }
            else
              {
		NodeList optionNodes = XPath.findNodes (xpathContext, optionContext, optionXPath);
                if (optionNodes != null)
                  {
                    buildNewOptions (optionNodes);                    
                  }
                else
                  {
		    Logger.log (this, "onLoad: no option nodes found");
                  }
              }
	  } 

        if (dataXPath != null && contextNode != null) 
          {
	    NodeList dataNodes = XPath.findNodes (xpathContext, contextNode, dataXPath);
	    if (dataNodes != null && dataNodes.getLength () > 0) 
	      { 
		dataValue = DOMUtils.getAssignedValue (dataNodes.item (0));
	      } 
	    else 
	      { 
		Logger.log (this, "onLoad: no data nodes found");
	      }
           }
      }
  }

  private void buildNewOptions (NodeList newOptions)
  {
    if (newOptions == null)
      {
        return;
      }

    if (builtOptions != null)
      {
        Enumeration oldOptions = builtOptions.elements ();
        while (oldOptions.hasMoreElements ())
          {
            Element oldOption = (Element)oldOptions.nextElement ();
            removeChild (oldOption);
          }

        builtOptions.removeAllElements ();
      }

    // the node to insert the generated options before
    Node refChild = null;

    String tmpPos = getAttributeNS (DEFAULT_NAMESPACE, INSERT_BEFORE_ATTR);
    if (tmpPos != null)
      {
        // get the insert-before attr 
        int insertBefore = Integer.valueOf (tmpPos).intValue ();

        // get all the pre-defined options
        NodeList predefinedOptions = this.getChildNodes ();

        // is the insert before position less than the number of 
        // pre-defined options?
        if (insertBefore < predefinedOptions.getLength ()) 
          {
            refChild = predefinedOptions.item (insertBefore);
          }
      }

    for (int i=0; i < newOptions.getLength (); i++)
      {
        Node newOptionNode = newOptions.item (i);
	if (newOptionNode.getNodeType () == Node.ELEMENT_NODE 
            && newOptionNode.getNodeName () == "option" 
	    && newOptionNode.getNamespaceURI () == XHTML_NAMESPACE)
	  {
	    addNewOption (newOptionNode.cloneNode (true), refChild);
	  }
	else 
	  {
	    Element newOption = getOwnerDocument ().createElementNS (XHTML_NAMESPACE, "option");
	    DOMUtils.setAssignedValue (newOption, DOMUtils.getBanner (newOptionNode, banner));
	    newOption.setAttribute ("value", XPath.findString (newOptionNode, "self::*"));
	    addNewOption (newOption, refChild);
	  }
      }
  }

  private void addNewOption (Node newOption, Node refChild)
  {
    if (builtOptions == null)
      {
        builtOptions = new Vector ();
      }
 
    builtOptions.addElement (newOption);
    insertBefore (newOption, refChild);
  }
  
  public void addOptionToList (Element option) 
  {
    // we want the text value of option tag to display
    String optionText = DOMUtils.getAssignedValue (option);

    Component c = getVis ();
    if (c instanceof List) 
      {
	((List)c).addItem (optionText);
      }
    else if (c instanceof Choice) 
      {
	((Choice)c).addItem (optionText);

        if (dataValue != null) 
          {
            // if there is a "value" attribute on the option, then
            // compare it to the data value.  if there isn't a "value"
            // attribute, compare the text value of the option to
            // the data value... if either matches, then mark this
            // option as selected.
            if (option.hasAttribute (VALUE_ATTR)
		&& dataValue.equals (option.getAttribute (VALUE_ATTR))
                || !option.hasAttribute (VALUE_ATTR)
		   && dataValue.equals (optionText))
              {
		option.setAttribute ("selected","true");
              }
            else
              {
                if (option.hasAttribute ("selected"))
                  {
		    option.removeAttribute ("selected");
                  }
              }
          }

	if (option.getAttribute ("selected") != null) 
	  { 
	    ((Choice)c).select (optionText);

            // we need to set the selected index to the selected item
	    selected = ((Choice)c).getSelectedIndex ();
	    defaultSelected = selected;
	  }
      }

    list.appendChild (option);
  }

  protected void setList (NodeList arg) 
  {
    // at this point, we can assume that this is a list of <option> elements
    resetList ();
    if (arg != null) 
      {
	int numItems = arg.getLength ();
	for (int i = 0; i < numItems; i++) 
	  {
	    addOptionToList ((Element) arg.item (i));
	  }
      }
    if (vis instanceof List)
      {
	((List)vis).select (selected);
      }
    else if (vis instanceof Choice)
      {
	((Choice)vis).select (selected);
      }
  }

  /**
   * When laying-out a select, we add in the option nodes if we're not watching
   * anything else.
   */
  public void layout (LayoutCursor cur) 
  {
    style = cur.style;
    if (cur.isPhase1())
      {
	NodeList nl = XPath.findNodes (this, OPTION_PATH);
	if (nl.getLength() > 0) 
	  {
	    setList (nl);
	  }
      }
    super.layout (cur);
  }

  public String getFormValue () 
  {
    if (!enabled)
      {
        return (null);
      }

    if (vis instanceof Choice) 
      {
	Choice c = (Choice) vis;
	selected = c.getSelectedIndex ();
	String val = ((Element) list.item (selected)).getAttribute (VALUE_ATTR);
	if (val == null) 
	  {
	    val = c.getSelectedItem ();
	  }
	return val;
      }
    else if (vis instanceof List)
      {
	List l = (List) vis;
	selected = l.getSelectedIndex ();
	String val = ((Element) list.item (selected)).getAttribute (VALUE_ATTR);
	if (val == null) 
	  {
	    val = l.getSelectedItem ();
	  }
	return val;
      }
    else
      {
	return (null);
      }
  }

  public String getText () 
  {
    return (null);
  }

  public XPath getXPath () 
  {
    return (null);
  }
  
  public void watch (NodeList data) 
  {
    buildNewOptions (data);
//    redraw (true);
  }

  public void resetValue ()
  {
    if (vis instanceof List)
      {
	((List)vis).select (defaultSelected);
      }
    else if (vis instanceof Choice)
      {
	((Choice)vis).select (defaultSelected);
      }
  }
  
}
