/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Table.java,v 1.14 2001/11/07 20:53:11 tim Exp $
 */ 
package org.xoe.display.dom.xhtml;

import com.pocketlinux.awt.theme.BackgroundImage;
import com.pocketlinux.awt.theme.Style;
import com.pocketlinux.awt.theme.Theme;

import java.awt.Container;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.net.URL;

import kaffe.awt.*;

import org.w3c.dom.*;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.FloatViewElement;
import org.xoe.display.dom.ImageFactory;
import org.xoe.display.dom.ImageHolder;
import org.xoe.display.dom.LayoutCursor;

public class Table
  extends FloatViewElement implements ImageHolder 
{
  private static final TableCell repeatCell = new TableCell ();
  
  private int halign = LayoutCursor.NONE;
  private int cellspacing = 2;
  private int cellpadding = 1;
  private int tableborder = 0;
  private int cellborder = 0;
  private Color bgcolor;

  private int twidth;
  private int[] widths;
  
  private int tminwidth;
  private int[] minwidths;
  
  private int tmaxwidth;
  private int[] maxwidths;

  private int[] pwidths;
  
  private int theight;
  private int[] heights;

  private int[] pheights;
  
  private int wslacker;
  private int hslacker;

  private int maxcolspan;
  private int mincolspan;
  private int maxrowspan;
  private int minrowspan;
  
  private int maxcol;
  private int maxrow;
  
  private TableCell[][] grid;
  
  private LayoutCursor cur = new LayoutCursor ();
  private Insets ins;
  private Image backgroundImg;
  private String backgroundName;
  private int version;  

  public class Render
    extends Container implements BackgroundImage 
  {
    Render() 
    {
    }

    public void paint (Graphics g) 
    {
      AWT.extension.paintBackground (this, g);
      for (int i = 0; i < tableborder; i++) 
	{
	  g.drawRect (i, i, twidth-i*2-1, theight-i*2-1);
	}
      super.paint(g);
    }

    public Image getBackgroundImage () 
    {
      return (backgroundImg);
    }
  }

  public Table () 
  {
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value)
  {
    String localName = getLocalName (qualifiedName);
    if ("align".equalsIgnoreCase(localName)) 
      {
	halign = Emphasis.getAlign (value, halign);
      }
    else if ("bgcolor".equalsIgnoreCase(localName)) 
      {
	bgcolor = Emphasis.getColor (value, null);
      }
    else if ("background".equalsIgnoreCase (localName)) 
      {
	backgroundName = value;
	backgroundImg = null;
      }
    else if (style == null && "cellspacing".equalsIgnoreCase (localName)) 
      {
	try {
	  cellspacing = Integer.parseInt (value);
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
      }
    else if (style == null && "cellpadding".equalsIgnoreCase (localName)) 
      {
	try 
	  {
	    cellpadding = Integer.parseInt (value);
	  }
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	  }
      }
    else if (style == null && "border".equalsIgnoreCase(localName)) 
      {
        try {
	  tableborder = Integer.parseInt (value);
          if (tableborder > 0) 
	    {
	      cellborder = 1;
	    }
	  else 
	    {
	      cellborder = 0;
	    }
	}
	catch (NumberFormatException nfe) 
	  {
	    if ("border".equalsIgnoreCase(value))
              {
                cellborder = 1;
                tableborder = 1;
	      } 
            else
	      {
	        Logger.log (this, nfe);
              }
	  }
      }
    else if ("xoe:style".equalsIgnoreCase (qualifiedName)) 
      {
        style = value;
        cellborder = 0;
        tableborder = 0;
        cellspacing = 0;
        cellpadding = 0;
      }
    super.setAttributeNS (namespaceURI, qualifiedName, value);
  }

  public void layout (LayoutCursor ocur) 
  {
    int ohalign = ocur.halign;
    if (halign == LayoutCursor.NONE && ocur.x > ocur.leftindent) 
      {
	ocur.newline ();
      }
    if (halign == LayoutCursor.CENTER)
      {
	if (ocur.x > ocur.leftindent)
	  {
	    ocur.newline ();
	  }
	ocur.halign = LayoutCursor.CENTER;
      }
    
    cur.set (ocur, this);

    if (style != null) 
      {
	Style mstyle = getDeco().style;
	cur.fgclr = mstyle.fg;
	cur.setFont (mstyle.fnt);
	cur.style = style;
      }

    ins = AWT.extension.getInsets (getVis());
    
    if (backgroundName != null) 
      {
	int docversion = ((XDocument)getOwnerDocument ()).version;
	if (backgroundImg != null && cur.isPhase1 () && version != docversion)
	  {
	    backgroundImg = null;
	  }
	if (backgroundImg == null) 
	  {
	    setBackground (getURLString (backgroundName));
	  }
	cur.havebackground = true;
      }

    int xmm = 0;
    int xm = 0;
    int ym = 0;
    int pad;
    
    switch (cur.phase) 
      {
      case LayoutCursor.PHASE1:
	// LayoutCursor.PHASE1 - we just work out the preferred widths
	
	// Only make a new grid if we need to
	if (changed) 
	  {
	    changed = false;
	    makeTableGrids ();
	  }

	pad = (ins.left+ins.right) + (tableborder * 2) + (cellspacing * (maxcol+1));
	int x = ins.left + tableborder + cellspacing;
	cur.width -= pad;
	for (int row = 0; row < maxrow; row++) 
	  {
	    for (int col = 0; col < maxcol; col++) 
	      {
		if (row == 0)
		  {
		    maxwidths[col] = 0;
		    minwidths[col] = 0;
		    widths[col] = 0;
		    pwidths[col] = 0;
		  }

		TableCell cell = grid[row][col];
		if (cell != null && cell != repeatCell) 
		  {
		    cell.setPadding (cellborder, cellpadding);
		    cur.setArea (x, 0, 0, 0);
		    cell.layout (cur);

		    // Handle preferred absolute width.
		    if (cell.preferredWidth > cell.cwidth) 
		      {
			cell.cwidth = cell.preferredWidth;
			cell.cmaxwidth = cell.cwidth;
		      }
		    // Work out the max and min cell widths
		    if (cell.colspan == 1) 
		      {
			if (cell.cmaxwidth > maxwidths[col]) 
			  {
			    maxwidths[col] = cell.cmaxwidth;
			  }
			if (cell.cwidth > minwidths[col]) 
			  {
			    minwidths[col] = cell.cwidth;
			  }
			if (pwidths[col] == 0) 
			  {
			    pwidths[col] = cell.preferredWidth;
			  }
		      }
		    cur.finish ();
		  }
	      }
	    cur.reset ();
	  }

	// If we have table cells which span multiple columns we need to
	// check that they have enough space.
	if (maxcolspan > 1) 
	  {
	    calcColSpans ();
	  }
	
	// Work out table width
	tminwidth = 0;
	tmaxwidth = 0;
	for (int i = minwidths.length-1; i >= 0; i--) 
	  {
	    tminwidth += minwidths[i];
	    tmaxwidth += maxwidths[i];
	  }

	xm = tminwidth + pad;
	xmm = tmaxwidth + pad;

	if (preferredWidth > xm) 
	  {
	    xm = preferredWidth;
	    if (xm > xmm) 
	      {
		xmm = xm;
	      }
	  }
	
	if (xmm > ocur.maxlinewidth) 
	  {
	    ocur.maxlinewidth = xmm;
	  }

	// Height doesn't matter yet
	theight = 0;
	ym = (ins.top+ins.bottom) + (tableborder * 2) + (cellspacing * (maxrow+1));
	break;

      case LayoutCursor.PHASE2:
	// Workout table heights based on selected widths
	
	// Handle preferred width.  If we don't have one then
	// we just use the current width.  If we do we use it
	// or the twidth (minwidth), whichever is larger
	if (preferredWidth != 0) 
	  {
	    int prefw = getPreferredWidth(ocur.width);
	    if (prefw < tminwidth) 
	      {
		prefw = tminwidth;
	      }
	    cur.width = prefw;
	  }

	// Calc cell widths
	calcWidths ();
	
	int y = ins.top + tableborder + cellspacing;
	for (int row = 0; row < maxrow; row++) 
	  {
	    int maxh = 0;
	    pheights[row] = 0;
	    if (grid[row][0] != null) 
	      {
		Node nrw = grid[row][0].getParentNode ();
		if (nrw instanceof TableRow) 
		  {
		    TableRow rw = (TableRow)nrw;
		    // Allow for bigger preferred heights (absolutes)
		    if (rw.preferredHeight > maxh) 
		      {
			maxh = rw.preferredHeight;
		      }
		    pheights[row] = rw.preferredHeight;
		  }
	      }
	    for (int col = 0; col < maxcol; col++) 
	      {
	        x = ins.left + tableborder + cellspacing;
		TableCell cell = grid[row][col];
		if (cell != null && cell != repeatCell) 
		  {
		    int w = widths[col];
		    for (int i = 1; i < cell.colspan; i++) 
		      {
			w += widths[col+i] + cellspacing;
		      }
		    cur.setArea (x, y, w, 0);
		    cell.layout (cur);
		    
		    if (cell.rowspan == 1) 
		      {
			// Handle preferred height (but only absolute for now)
			if (cell.preferredHeight > cell.cheight) 
			  {
			    cell.cheight = cell.preferredHeight;
			  }
			if (cell.cheight > maxh) 
			  {
			    maxh = cell.cheight;
			  }
		      }
		    cur.finish ();
		  }
		x += widths[col] + cellspacing;
	      }
	    heights[row] = maxh;
	    y = y + maxh;
	  }

	// If we have table cells which span multiple rows we need to
	// check that they have enough space.
	if (maxrowspan > 1) 
	  {
	    calcRowSpans ();
	  }

	theight = 0;
	for (int i = 0; i < maxrow; i++) 
	  {
	    theight += heights[i];
	  }

	xm = twidth + (ins.left+ins.right) + (tableborder * 2) + (cellspacing * (maxcol+1));
	xmm = xm;
	ym = theight + (ins.top+ins.bottom) + (tableborder * 2) + (cellspacing * (maxrow+1));
	
	// If absolute height is greater, use it instead
	if (preferredHeight > ym) 
	  {
	    ym = preferredHeight;
	  }

	break;
	
      case LayoutCursor.PHASE3:
	
	// Handle preferred height.  If we don't have one then
	// we just use the current height.  If we do we use it
	// or the theight, whichever is larger
	if (preferredHeight != 0) 
	  {
	    int prefh = getPreferredHeight(ocur.height);
	    if (prefh < theight) 
	      {
		prefh = theight;
	      }
	    cur.height = prefh;
	  }
	
	calcHeights();
	
	y = ins.top + tableborder + cellspacing;
	for (int row = 0; row < maxrow; row++) 
	  {
	    x = ins.left + tableborder + cellspacing;
	    for (int col = 0; col < maxcol; col++) 
	      {
		TableCell tc = grid[row][col];
		if (tc != null && tc != repeatCell) 
		  {
		    int h = heights[row];
		    for (int i = 1; i < tc.rowspan; i++) 
		      {
			h += heights[row+i] + cellspacing;
		      }
		    int w = widths[col];
		    for (int i = 1; i < tc.colspan; i++) 
		      {
			w += widths[col+i] + cellspacing;
		      }
		    cur.setArea(x, y, w, h);
		    tc.layout(cur);
		    cur.finish();
		  }
		x += widths[col] + cellspacing;
	      }
	    y += heights[row] + cellspacing;
	  }
	
	xm = twidth + (ins.left+ins.right) + (tableborder * 2) + (cellspacing * (maxcol+1));
	xmm = xm;
	ym = theight + (ins.top+ins.bottom) + (tableborder * 2) + (cellspacing * (maxrow+1));
	ocur.addVis (this);
	ocur.setLocation (vis);
	vis.setSize (xm, ym);
	if (bgcolor != null) 
	  {
	    setBackground(bgcolor);
	  }
	twidth = xm;
	theight = ym;
	break;
      }
    
    switch (halign) 
      {
      case LayoutCursor.CENTER:
      case LayoutCursor.NONE:
      default:
	ocur.move (xm, ym);
	ocur.newline ();
	break;

      case LayoutCursor.LEFT:
      case LayoutCursor.RIGHT:
	ocur.flow (this, halign, xm, xmm, ym);
	break;
      }
    ocur.halign = ohalign;
  }

  protected void makeVis () 
  {
    vis = new Render();
  }

  private void setBackground (String name) 
  {
    setBackground ((Color)null);
    if (name != null) 
      {
	ImageFactory.getImage (name, this);
      }
  }

  private void setBackground (Color clr) 
  {
    backgroundImg = null;
    getVis ().setBackground (clr);
  }

  public void imageDeliver (String url, Image nimg)
  {
    if (backgroundImg != nimg)
      {
	backgroundImg = nimg;
	if (vis != null)
	  {
	    vis.repaint ();
	  }
      }
  }
  
  protected void unmakeVis () 
  {
    backgroundImg = null;
    super.unmakeVis();
  }

  /**
   * Work out how big the table is based on what it contains
   * and the various colspans and rowspans of the cells.
   */
  private void makeTableGrids () 
  {
    maxrow = 1;
    maxcol = 1;
    maxcolspan = 1;
    mincolspan = 0x7FFFFFFF;
    maxrowspan = 1;
    minrowspan = 0x7FFFFFFF;
    hslacker = -1;
    wslacker = -1;
    
    // Count the number of rows;
    int rcnt = 0;
    for (TableRow row = getNextRow (this); row != null; row = getNextRow (row)) 
      {
	for (Node col = row.getFirstChild (); col != null; col = col.getNextSibling ()) 
	  {
	    if (col instanceof TableCell) 
	      {
		TableCell tc = (TableCell)col;
		if (rcnt + tc.rowspan > maxrow) 
		  {
		    maxrow = rcnt + tc.rowspan;
		  }
	      }
	  }
	rcnt++;
      }
    if (rcnt > maxrow) 
      {
	maxrow = rcnt;
      }

    // Make sure we have enough heights
    if (heights == null || maxrow != heights.length) 
      {
	heights = new int[maxrow];
	pheights = new int[maxrow];
      }
    else {
      for (int i = heights.length-1; i >= 0; i--) 
	{
	  heights[i] = 0;
	  pheights[i] = 0;
	}
    }

    rcnt = 0;
    for (TableRow row = getNextRow (this); row != null; row = getNextRow (row)) 
      {
	int ccnt = 0;
	for (Node col = row.getFirstChild (); col != null; col = col.getNextSibling ()) 
	  {
	    if (col instanceof TableCell) 
	      {
		TableCell tc = (TableCell)col;
		if (tc.colspan > maxcolspan) 
		  {
		    maxcolspan = tc.colspan;
		  }
		if (tc.colspan > 1 && tc.colspan < mincolspan) 
		  {
		    mincolspan = tc.colspan;
		  }
		if (tc.rowspan > maxrowspan) 
		  {
		    maxrowspan = tc.rowspan;
		  }
		if (tc.rowspan > 1 && tc.rowspan < minrowspan) 
		  {
		    minrowspan = tc.rowspan;
		  }
		for (int i = tc.rowspan - 1; i >= 0; i--) 
		  {
		    heights[rcnt+i] += tc.colspan;
		  }
	      }
	  }
	if (heights[rcnt] > maxcol) 
	  {
	    maxcol = heights[rcnt];
	  }
	heights[rcnt] = 0;
	rcnt++;
      }

    // Make sure we have enought widths
    if (maxwidths == null || maxcol != maxwidths.length) 
      {
	maxwidths = new int[maxcol];
	minwidths = new int[maxcol];
	widths = new int[maxcol];
	pwidths = new int[maxcol];
      }
    else {
      for (int i = maxwidths.length-1; i >= 0; i--) 
	{
	  maxwidths[i] = 0;
	  minwidths[i] = 0;
	  widths[i] = 0;
	  pwidths[i] = 0;
	}
    }

    // Adjust col and row spans
    if (maxcolspan > maxcol)
      {
	maxcolspan = maxcol;
	if (mincolspan > maxcol)
	  {
	    mincolspan = maxcol;
	  }
      }
    if (maxrowspan > maxrow)
      {
	maxrowspan = maxrow;
	if (minrowspan > maxrow)
	  {
	    minrowspan = maxrow;
	  }
      }

    // Create two dimension TableCell grid which we'll use
    // to layout the table.
    grid = new TableCell[maxrow][maxcol];
    
    // Populate it
    rcnt = 0;
    for (TableRow row = getNextRow(this); row != null; row = getNextRow(row)) 
      {
	// Note the first height slacker
	if ((row.flags & SLACK_HEIGHT) != 0 && hslacker == -1) 
	  {
	    hslacker = rcnt;
	  }
	int ccnt = 0;
	for (Node col = row.getFirstChild (); col != null; col = col.getNextSibling ()) 
	  {
	    if (col instanceof TableCell) 
	      {
		while (grid[rcnt][ccnt] != null) 
		  {
		    ccnt++;
		  }
		TableCell tc = (TableCell)col;
		for (int i = tc.rowspan - 1; i >= 0; i--) 
		  {
		    for (int j = tc.colspan - 1; j >= 0; j--) 
		      {
			grid[rcnt+i][ccnt+j] = repeatCell;
		      }
		  }
		grid[rcnt][ccnt] = tc;
		
		// Note the first width slacker
		if ((tc.flags & SLACK_WIDTH) != 0 && wslacker == -1) 
		  {
		    wslacker = ccnt;
		  }

		ccnt += tc.colspan;
	      }
	  }
	rcnt++;
      }
  }

  /**
   * Given a node, find the next TableRow node below this table.
   */
  private TableRow getNextRow (Node n) 
  {
    for (;;) 
      {
	if (n instanceof TableRow || n.getFirstChild () == null) 
	  {
	    if (n.getNextSibling () == null) 
	      {
		for (;;) 
		  {
		    n = n.getParentNode ();
		    if (n == this || n instanceof Document) 
		      {
			return (null);
		      }
		    else if (n.getNextSibling () != null) 
		      {
			break;
		      }
		  }
	      }
	    n = n.getNextSibling ();
	  }
	else
	  {
	    n = n.getFirstChild ();
	  }
	if (n instanceof TableRow) 
	  {
	    return ((TableRow)n);
	  }
      }
  }

  private void calcWidths () 
  {
    int pad = (ins.left+ins.right) + (tableborder * 2) + (cellspacing * (maxcol+1));
    cur.width -= pad;
    
    // Use minimums if they fit exactly or are too big
    if (tminwidth >= cur.width) 
      {
	twidth = tminwidth;
	for (int i = widths.length-1; i >= 0; i--) 
	  {
	    widths[i] = minwidths[i];
	  }
      }
    // Maximums fit - use them
    else if (tmaxwidth <= cur.width) 
      {
	twidth = tmaxwidth;
	for (int i = widths.length-1; i >= 0; i--) 
	  {
	    widths[i] = maxwidths[i];
	  }

	// Add in any slack
	if (wslacker != -1) 
	  {
	    widths[wslacker] += cur.width - twidth;
	    twidth = cur.width;
	  }
	// Otherwise we need to spread the extra space
	// amongst the cells.
	else {
	  // If there is no preferred width
	  // we work one out based on either the table's
	  // caculated width or one based on a column size and
	  // percentage of table it takes.
	  if (preferredWidth == 0) 
	    {
	      int tw = twidth;
	      for (int i = pwidths.length-1; i >= 0; i--) 
		{
		  int pw = pwidths[i];
		  if (pw < 0) 
		    {
		      pw = widths[i] * 100 / -pw;
		      if (pw > tw) 
			{
			  tw = pw;
			}
		    }
		}
	      cur.width = tw;
	    }
	  
	  // Convert the percentages into absolutes. 
	  for (int i = pwidths.length-1; i >= 0; i--) 
	    {
	      int pw = pwidths[i];
	      if (pw < 0) 
		{
		  pw = -pw * cur.width / 100;
		  pwidths[i] = pw;
		}
	    }
	  // We get the space we have left and we allocate it to the widths
	  // of the colomns.  If we run out of space we stop.  While we do this
	  // we keep track of the amount of expandable space (that without a
	  // preferred size).  We will need this later in case we still have
	  // space left
	  int npw = 0;
	  int adiff = cur.width - twidth;
	  for (int i = widths.length-1; i >= 0 && adiff > 0; i--) 
	    {
	      int pw = pwidths[i];
	      if (pw == 0) 
		{
		  npw += widths[i];
		}
	      else {
		int pdiff = pw - widths[i];
		if (pdiff > 0) 
		  {
		    if (pdiff > adiff) 
		      {
			pdiff = adiff;
		      }
		    adiff -= pdiff;
		    widths[i] += pdiff;
		    twidth += pdiff;
		  }
	      }
	    }
	  // Still got some space left - we distribute it amongst
	  // the cells.  We give the space to the cells without fixed
	  // sizes, but if that fails we give it to everyone
	  if (adiff > 0) 
	    {
	      if (npw > 0) 
		{
		  for (int i = widths.length-1; i >= 0; i--) 
		    {
		      if (pwidths[i] == 0) 
			{
			  widths[i] += (widths[i] * adiff) / npw;
			}
		    }
		}
	      else if (twidth > 0) 
		{
		  for (int i = widths.length-1; i >= 0; i--) 
		    {
		      widths[i] += (widths[i] * adiff) / twidth;
		    }
		}
	      else if (widths.length > 0) 
		{
		  adiff /= widths.length;
		  for (int i = widths.length-1; i >= 0; i--) 
		    {
		      widths[i] += adiff;
		    }
	  	}
	    }
	  twidth = cur.width;
	}
      }
    // Max too big, but mins fit - so pad the cells
    else 
      {
	twidth = tminwidth;
	for (int i = widths.length-1; i >= 0; i--) 
	  {
	    widths[i] = minwidths[i];
	  }

	// Add in any slack
	if (wslacker != -1) 
	  {
	    widths[wslacker] += cur.width - twidth;
	  }
		
	// Otherwise distribute the extra space
	else {
	  int adiff = cur.width - twidth;
	  int tdiff = tmaxwidth - twidth;
	  
	  // Convert the percentages into absolutes. 
	  for (int i = pwidths.length-1; i >= 0; i--) 
	    {
	      int pw = pwidths[i];
	      if (pw < 0) 
		{
		  pw = -pw * cur.width / 100;
		  pwidths[i] = pw;
		}
	    }

	  for (int i = widths.length-1; i >= 0 && adiff > 0; i--) 
	    {
	      int pw = pwidths[i];
	      if (pw != 0) 
		{
		  int pdiff = pw - widths[i];
		  if (pdiff > 0) 
		    {
		      if (pdiff > adiff) 
			{
			  pdiff = adiff;
			}
		      adiff -= pdiff;
		      tdiff -= pdiff;
		      widths[i] += pdiff;
		      twidth += pdiff;
		    }
		}
	    }
	  if (adiff > 0 && tdiff > 0) 
	    {
	      for (int i = widths.length-1; i >= 0; i--) 
		{
		  int d = maxwidths[i] - widths[i];
		  widths[i] += (adiff * d) / tdiff;
		}
	    }
	}
	twidth = cur.width;
      }
  }

  private void calcHeights () 
  {
    
    int pad = (ins.top+ins.bottom) + (tableborder * 2) + (cellspacing * (maxrow+1));
    cur.height -= pad;
    
    // If we have extra height we need to allocate it to the rows
    if (cur.height > theight) 
      {
	if (hslacker != -1) 
	  {
	    heights[hslacker] += cur.height - theight;
	  }
	else {
	  // If there is no preferred height
	  // we work one out based on either the table's
	  // caculated height or one based on a row size and
	  // percentage of table it takes.
	  if (preferredHeight == 0) 
	    {
	      cur.height = theight;
	      for (int i = pheights.length-1; i >= 0; i--) 
		{
		  int ph = pheights[i];
		  if (ph < 0) 
		    {
		      ph = heights[i] * 100 / -ph;
		      if (ph > cur.height) 
			{
			  cur.height = ph;
			}
		    }
		}
	    }
	  // Convert the percentages into absolutes. 
	  for (int i = pheights.length-1; i >= 0; i--) 
	    {
	      int ph = pheights[i];
	      if (ph < 0) 
		{
		  ph = -ph * cur.height / 100;
		  pheights[i] = ph;
		}
	    }
	  
	  // We get the space we have left and we allocate it to the heights
	  // of the rows.  If we run out of space we stop.  While we do this
	  // we keep track of the amount of expandable space (that without a
	  // preferred size).  We will need this later in case we still have
	  // space left
	  int nph = 0;
	  int adiff = cur.height - theight;
	  for (int i = heights.length-1; i >= 0 && adiff > 0; i--) 
	    {
	      int ph = pheights[i];
	      if (ph == 0) 
		{
		  nph += heights[i];
		}
	      else {
		int pdiff = ph - heights[i];
		if (pdiff > 0 && pdiff <= adiff) 
		  {
		    adiff -= pdiff;
		    heights[i] += pdiff;
		    theight += pdiff;
		  }
	      }
	    }
	  // Still got some space left - we distribute it amongst
	  // the cells.  We give the space to the cells without fixed
	  // sizes, but if that fails we give it to everyone
	  if (adiff > 0) 
	    {
	      if (nph > 0) {
		for (int i = heights.length-1; i >= 0; i--) 
		  {
		    if (pheights[i] == 0) {
		      heights[i] += (heights[i] * adiff) / nph;
		    }
		  }
	      }
	      else if (theight > 0) 
		{
		  for (int i = heights.length-1; i >= 0; i--) 
		    {
		      heights[i] += (heights[i] * adiff) / theight;
		    }
		}
	      else if (heights.length > 0)
		{
		  adiff /= heights.length;
		  for (int i = heights.length-1; i >= 0; i--) 
		    {
		      heights[i] += adiff;
		    }
		}
	    }
	}
	theight = cur.height;
      }
    else if (cur.height < theight) 
      {
	// Logger.log(this, "not enough height: " + (theight-cur.height));
      }
  }
  
  private void calcRowSpans () 
  {
    for (int span = minrowspan; span <= maxrowspan; span++) 
      {
	for (int row = 0; row < maxrow; row++) 
	  {
	    for (int col = 0; col < maxcol; col++) 
	      {
		TableCell cell = grid[row][col];
		if (cell == null || cell == repeatCell || cell.rowspan != span) 
		  {
		    continue;
		  }
		// This cell spans 'span' rows so we better make sure we have
		// enough space.
		int ch = heights[row];
		int limit = row + span;
		for (int i = row + 1; i < limit; i++) 
		  {
		    ch += heights[i] + cellspacing;
		  }
		// Not enough space - expand it in proportion
		if (ch < cell.cheight) 
		  {
		    if (hslacker >= row && hslacker < limit) 
		      {
			heights[hslacker] += cell.cheight - ch;
		      }
		    else if (ch <= 0) 
		      {
			int adiff = cell.cheight / span;
			for (int i = row; i < limit; i++) 
			  {
			    heights[i] += adiff;
			  }
		      }
		    else {
		      int adiff = cell.cheight - ch;
		      for (int i = row; i < limit; i++) 
			{
			  heights[i] += (heights[i] * adiff) / ch;
			}
		    }
		  }
	      }
	  }
      }
  }
  
  private void calcColSpans () 
  {
    for (int span = mincolspan; span <= maxcolspan; span++) 
      {
	for (int row = 0; row < maxrow; row++) 
	  {
	    for (int col = 0; col < maxcol; col++) 
	      {
		TableCell cell = grid[row][col];
		if (cell == null || cell == repeatCell || cell.colspan != span) 
		  {
		    continue;
		  }
		// This cell spans 'span' columns so we better make sure we have
		// enough space.
		int cw = minwidths[col];
		int cmw = maxwidths[col];
		int limit = col + span;
		for (int i = col + 1; i < limit; i++) 
		  {
		    cw += minwidths[i] + cellspacing;
		    cmw += maxwidths[i] + cellspacing;
		  }
		// Not enough space - expand it in proportion
		if (cw < cell.cwidth) 
		  {
		    if (wslacker >= col && wslacker < limit) 
		      {
			minwidths[wslacker] += cell.cwidth - cw;
		      }
		    else if (cw <= 0) 
		      {
			int adiff = cell.cwidth / span;
			int bdiff = cell.cwidth;
			for (int i = col; i < limit; i++) {
			  minwidths[i] += adiff;
			  bdiff -= adiff;
			}
			for (int i = col; i < limit && bdiff > 0; i++) 
			  {
			    minwidths[i]++;
			    bdiff--;
			  }
		      }
		    else {
		      int adiff = cell.cwidth - cw;
		      int bdiff = adiff;
		      for (int i = col; i < limit; i++) 
			{
			  int adjust = (minwidths[i] * adiff) / cw;
			  minwidths[i] += adjust;
			  bdiff -= adjust;
			}
		      for (int i = col; i < limit && bdiff > 0; i++) 
			{
			  minwidths[i]++;
			  bdiff--;
			}
		    }
		  }
		if (cmw < cell.cmaxwidth) 
		  {
		    if (wslacker >= col && wslacker < limit) 
		      {
			maxwidths[wslacker] += cell.cmaxwidth - cmw;
		      }
		    else if (cmw <= 0) 
		      {
			int adiff = cell.cmaxwidth / span;
			int bdiff = cell.cmaxwidth;
			for (int i = col; i < limit; i++) 
			  {
			    maxwidths[i] += adiff;
			    bdiff -= adiff;
			  }
			for (int i = col; i < limit && bdiff > 0; i++) 
			  {
			    maxwidths[i]++;
			    bdiff--;
			  }
		      }
		    else {
		      int adiff = cell.cmaxwidth - cmw;
		      int bdiff = adiff;
		      for (int i = col; i < limit; i++) 
			{
			  int adjust = (maxwidths[i] * adiff) / cmw;
			  maxwidths[i] += adjust;
			  bdiff -= adjust;
			}
		      for (int i = col; i < limit && bdiff > 0; i++) 
			{
			  maxwidths[i]++;
			  bdiff--;
			}
		    }
		  }
	      }
	  }
      }
  }
  
  public Node cloneNode (boolean deep) 
  {
    Table nnode = (Table)super.cloneNode (deep);
    
    nnode.halign = halign;
    nnode.cellspacing = cellspacing;
    nnode.cellpadding = cellpadding;
    nnode.tableborder = tableborder;
    nnode.cellborder = cellborder;
    nnode.bgcolor = bgcolor;
    
    nnode.twidth = 0;
    nnode.widths = null;
    nnode.tminwidth = 0;
    nnode.minwidths = null;
    nnode.tmaxwidth = 0;
    nnode.maxwidths = null;
    nnode.pwidths = null;
    nnode.theight = theight;
    nnode.heights = null;
    nnode.pheights = null;
    nnode.wslacker = -1;
    nnode.hslacker = -1;
    nnode.maxcolspan = 0;
    nnode.maxrowspan = 0;
    nnode.mincolspan = 0;
    nnode.minrowspan = 0;
    nnode.maxcol = 0;
    nnode.maxrow = 0;
    nnode.grid = null;
    
    nnode.cur = new LayoutCursor();
    nnode.ins = null;
    nnode.backgroundImg = backgroundImg;
    nnode.backgroundName = backgroundName;
    nnode.version = version;
    return nnode;
  }
}
