/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: TableCell.java,v 1.12 2001/11/07 20:53:11 tim Exp $
 */ 
package org.xoe.display.dom.xhtml;

import com.pocketlinux.awt.theme.*;
import com.pocketlinux.awt.theme.BackgroundImage;

import java.awt.Container;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.net.URL;

import kaffe.awt.AWT;

import org.w3c.dom.Node;

import org.xoe.core.dom.XDocument;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.ImageFactory;
import org.xoe.display.dom.ImageHolder;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;


public class TableCell
  extends ViewElement 
  implements ImageHolder 
{
  private int halign = LayoutCursor.NONE;	// Defaults to LEFT
  private int valign = LayoutCursor.NONE; // Defaults to CENTER
  
  int cwidth;
  int cheight;
  int cmaxwidth;
  int rowspan = 1;
  int colspan = 1;
  String backgroundName;
  Image backgroundImg;
  Color bgcolor;
  LayoutCursor cur = new LayoutCursor ();
  int border;
  int padding;
  int version;

  public class Render
    extends Container 
    implements BackgroundImage 
  {
    Render () 
    {
    }

    public void paint (Graphics g) 
    {
      AWT.extension.paintBackground (this, g);
      for (int i = 0; i < border; i++) 
	{
	  g.drawRect (i, i, cwidth-i*2-1, cheight-i*2-1);
	}
      super.paint (g);
    }

    public Image getBackgroundImage () 
    {
      return (backgroundImg);
    }
    
  }

  public TableCell () 
  {
  }

  public Node cloneNode (boolean deep)
  {
    TableCell nnode = (TableCell)super.cloneNode (deep);
    nnode.halign = halign;	// Defaults to LEFT
    nnode.valign = valign; // Defaults to CENTER
    nnode.cwidth = cwidth;
    nnode.cheight = cheight;
    nnode.cmaxwidth = cmaxwidth;
    nnode.rowspan = rowspan;
    nnode.colspan = colspan;
    nnode.backgroundName = backgroundName;
    nnode.backgroundImg = backgroundImg;
    nnode.bgcolor = bgcolor;
    nnode.cur = new LayoutCursor ();
    nnode.border = border;
    nnode.padding = padding;
    nnode.version = version;
    return nnode;
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);
    if (localName.equalsIgnoreCase ("halign") || localName.equalsIgnoreCase ("align"))
      {
	halign = Emphasis.getAlign(value, halign);
      }
    else if (localName.equalsIgnoreCase ("valign"))
      {
	valign = Emphasis.getAlign (value, valign);
      }
    else if (localName.equalsIgnoreCase( "colspan"))
      {
	try {
	  colspan = Integer.parseInt (value, 10);
	  if (colspan < 1) 
	    {
	      colspan = 1;
	    }
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	    colspan = 1;
	  }
      }
    else if (localName.equalsIgnoreCase ("rowspan"))
      {
	try {
	  rowspan = Integer.parseInt (value, 10);
	  if (rowspan < 1) 
	    {
	      rowspan = 1;
	    }
	}
	catch (NumberFormatException nfe) 
	  {
	    Logger.log (this, nfe);
	    rowspan = 1;
	  }
      }
    else if (localName.equalsIgnoreCase ("background"))
      {
	backgroundName = value;
	backgroundImg = null;
      }
    else if ("bgcolor".equalsIgnoreCase (localName)) 
      {
	bgcolor = Emphasis.getColor (value, null);
      }
    else if ("xoe:style".equalsIgnoreCase (qualifiedName)) 
      {
	style = value;
	border = 0;
	padding = 0;
      }
    super.setAttributeNS (namespaceURI, qualifiedName, value);
  }

/*
  public String toString () 
  {
    return (getClass ().getName () + "["+rowspan+","+colspan+"]");
  }
*/

  public void setPadding (int border, int padding) 
  {
    this.border = border;
    this.padding = padding;
  }

  public void layout (LayoutCursor ocur) 
  {
//System.out.println(this + " -> PHASE " + ocur.phase + " start");
    cur.set (ocur, this);

    if (style != null) 
      {
	Style mstyle = getDeco ().style;
	cur.fgclr = mstyle.fg;
	cur.setFont (mstyle.fnt);
	cur.style = style;
      }

    // Set alignment & color
    int halign = this.halign;
    int valign = this.valign;
    Color clr = bgcolor;

    Node pn = getParentNode ();
    if (pn instanceof TableRow) 
      {
	TableRow row = (TableRow)pn;
	if (clr == null) 
	  {
	    clr = row.bgcolor;
	  }
	if (halign == LayoutCursor.NONE) 
	  {
	    halign = row.align;
	  }
	if (valign == LayoutCursor.NONE) 
	  {
	    valign = row.valign;
	  }
      }

    cur.halign = halign;
    
    // Is background required?
    if (backgroundName != null) 
      {
	int docversion = ((XDocument)getOwnerDocument ()).version;
	if (backgroundImg != null && cur.isPhase1 () && version != docversion)
	  {
	    backgroundImg = null;
	  }
	version = docversion;

	// Not already setup? Do it now
	if (backgroundImg == null) 
	  {
	    setBackground (getURLString (backgroundName));
	  }
	cur.havebackground = true;
	clr = null;
      }

    // Handle insets
    Insets ins = AWT.extension.getInsets (getVis());
    int pad = (border + padding) * 2;
    cur.width -= (ins.left+ins.right) + pad;
    cur.height -= (ins.top+ins.bottom) + pad;
    cur.xoff = ins.left + border + padding;
    cur.yoff = ins.top + border + padding;

    // If we're fixing the layout, we should align the contents as
    // appropriate
    switch (cur.phase) 
      {
      case LayoutCursor.PHASE1:
	layoutChildren (cur);
	cur.finish ();
	cwidth = cur.maxwidth + (ins.left+ins.right) + pad;
	cmaxwidth = cur.maxlinewidth + (ins.left+ins.right) + pad;
	cheight = cur.y + (ins.top+ins.bottom) + pad;
	break;

      case LayoutCursor.PHASE2:
	layoutChildren (cur);
	cur.finish ();
	cwidth = ocur.width;
	cheight = cur.y + (ins.top+ins.bottom) + pad;
	break;

      case LayoutCursor.PHASE3:
	if (valign != LayoutCursor.TOP) 
	  {
	    cur.topindent = ocur.height - ocur.y - cheight;
	    cur.height -= cur.topindent;
	    if (valign == LayoutCursor.CENTER || valign == LayoutCursor.NONE) 
	      {
		cur.topindent /= 2;
	      }
	  }
	cur.useProxyVis ();
	layoutChildren (cur);
	cur.finish ();
	cwidth = ocur.width;
	cheight = ocur.height;
	ocur.addVis (this);
	ocur.setLocation (vis);
	vis.setSize (cwidth, cheight);
	if (clr != null) 
	  {
	    setBackground (clr);
	  }
	break;
      }
    
    ocur.move (cwidth, cheight);
//System.out.println(this + " -> PHASE " + ocur.phase + " end");
  }

  protected void makeVis () 
  {
    vis = new Render ();
  }

  public void setBackground (String name) 
  {
    setBackground ((Color)null);
    if (name != null) 
      {
	ImageFactory.getImage (name, this);
      }
  }

  public void setBackground (Color clr) 
  {
    backgroundImg = null;
    getVis ().setBackground (clr);
  }

  public void imageDeliver (String url, Image nimg)
  {
    if (nimg != backgroundImg)
      {
	backgroundImg = nimg;
	if (vis != null)
	  {
	    vis.repaint ();
	  }
      }
  }
}
