/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: Text.java,v 1.10 2001/12/12 03:41:46 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;

import java.awt.Color;
import java.awt.Component;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Vector;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.dom.DOMException;
import org.xoe.core.dom.XText;
import org.xoe.core.util.Logger;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.Viewable;
import org.xoe.display.dom.ViewElement;


public class Text
  extends XText
  implements Viewable, MouseListener
{
  int tlen;
  int fontheight;
  int fontdescent;
  Anchor anchor;
  boolean underline;
  boolean strike;
  int[] layoutinfo;
  Color alink;
  Vector texts = new Vector();
  boolean changed;

  Render c;
  int lx;
  int lw;
  int lcw;

  public Node cloneNode (boolean deep)
  {
    Text nnode = (Text) super.cloneNode (deep);
    nnode.setData (getData ());
    return (nnode);
  }

  public class Render extends Component {

    int start;
    int len;

    public void paint(Graphics g) {
      int h = fontheight;
       
      for (int i = 0; i < len; i++) {
	int o = (start+i) * 3;
	
        g.drawChars(text, layoutinfo[o+0], layoutinfo[o+1], 0, h - fontdescent);
	
        if (underline) {
                g.drawLine(0, h - 1, layoutinfo[o+2], h - 1);
      	}
        if (strike) {
		g.drawLine(0, h - (fontheight / 2), layoutinfo[o+2], h - (fontheight / 2));
        }
       
	h += fontheight;
      }
    }

  }

  public Text() {
  }

  private boolean isWhite(char ch) {
    switch (ch) {
    case ' ':
    case '\t':
    case '\n':
      return (true);
    default:
      return (false);
    }
  }

  private boolean isBreak(char ch) {
    // We can break on a whitespace or on any japanese/chinese character
    return (isWhite(ch) || (ch >= '\u3040' && ch < '\ufa2d'));
  }

  private void compressText ()
  {
      text = getData().toCharArray();
      tlen = text.length;

      // Compress the whitespace
      int wordstart = 0;
      int wordend = 0;
      if (tlen > 0 && isWhite(text[0])) {
	text[0] = ' ';
	wordstart++;
	wordend++;
      }
      outer: for (;;) {
      // Skip through whitespace
        for (;;) {
	  if (wordstart >= tlen) {
	    break outer;
	  }
	  if (!isWhite(text[wordstart])) {
	    break;
	  }
	  wordstart++;
        }
        // Find end of word
        for (;;) {
	  if (wordstart >= tlen) {
	    break;
	  }
	  if (isWhite(text[wordstart])) {
	    text[wordend] = ' ';
	    wordend++;
	    break;
	  }
	  text[wordend] = text[wordstart];
	  wordstart++;
	  wordend++;
        }
      }
      tlen = wordend;
  }

  /**
   * Layout the text as multiple lines.
   */
  public void layout(LayoutCursor cur) {
    underline = cur.underline;
    strike = cur.strike;
    alink = cur.activeclr;
    if (cur.anchor != null) {
      anchor = (Anchor)cur.anchor;
      underline = true;
    }
    FontMetrics fm = cur.fontmetrics;

    if (cur.isPhase1()) {
      texts.removeAllElements();
      fontheight = fm.getHeight();
      fontdescent = fm.getDescent();
      if (changed)
  	{
	  changed = false;
	  compressText();
	}
    }

    // Reset render
    c = null;
    int lines = 0;

    int wordend = 0;
    int linewidth = 0;
    int linelen = 0;
    int xpos = cur.x;
    int linestart = wordend;
    for (;;) {
      int wordstart = wordend;
      if (wordstart >= tlen) {
	break;
      }
      for (;;) {
	wordend++;
	if (wordend >= tlen) {
	  break;
	}
	if (isBreak(text[wordend]) && wordend+1 < tlen) {
	  break;
	}
      }

      // Find word length
      int wordwidth = fm.charsWidth(text, wordstart, wordend-wordstart);

      if (!cur.nobreaks) {
	if (cur.x + wordwidth > cur.width) {
	  if (cur.isPhase3()) {
	    addLay(lines, cur, linestart, linelen, xpos, cur.y, linewidth);
	  }
	  lines++;
	  linelen = 0;
	  linewidth = 0;
	  cur.newline(wordwidth);
	  xpos = cur.x;
	  linestart = wordstart;
	}
	// Make sure we don't put spaces at the
	// beginning of a line
	if (cur.x == cur.leftindent && text[wordstart] == ' ') {
	  wordstart++;
	  linestart++;
	  wordwidth -= fm.charWidth(' ');
	}
	cur.move(wordwidth, fontheight);
      }

      linewidth += wordwidth;
      linelen += wordend - wordstart;
    }
    if (cur.nobreaks) {
      cur.move(linewidth, fontheight);
    }
    if (cur.isPhase3 ()) {
      if (linewidth > 0) {
	addLay(lines, cur, linestart, linelen, xpos, cur.y, linewidth);
      }
      if (c != null) {
	cur.addVis(c);
        c = null;
      }
    }
    if (cur.isPhase2 ()) {
      if (linewidth > 0) {
	lines++;
      }
      layoutinfo = new int[lines * 3];
    }
  }

  private void addLay(int line, LayoutCursor cur, int pos, int len, int x, int y, int fwidth) {
    int i = line * 3;
    if (i + 3 > layoutinfo.length) {
      int[] old = layoutinfo;
      layoutinfo = new int[layoutinfo.length * 2];
      System.arraycopy(old, 0, layoutinfo, 0, old.length);
    }
    layoutinfo[i+0] = pos;
    layoutinfo[i+1] = len;
    layoutinfo[i+2] = fwidth;
    x += cur.xoff;
    y += cur.yoff;
    if (c == null || lx != x || lcw != cur.width) {
      if (c != null) {
	cur.addVis(c);
      }
      lx = x;
      lcw = cur.width;
      lw = fwidth;
      c = new Render();
      texts.addElement(c);
      c.start = line;
      c.len = 1;
      c.setFont(cur.font);
      c.setForeground(cur.fgclr);
      if (cur.anchor != null) {
	c.addMouseListener(this);
	c.setCursor(Anchor.anchorCursor);
      }
    }
    else {
      c.len++;
      if (lw < fwidth) {
	lw = fwidth;
      }
    }
    c.setBounds(x, y, lw, cur.fontmetrics.getHeight() * c.len);

    // If we're not left aligning then we need to add the elements in
    // one by one so alignment is correct
    if (cur.halign != LayoutCursor.LEFT) {
      cur.addVis(c);
      c = null;
    }
  }

  public void unlayout()
  {
  }

  public void mousePressed(MouseEvent evt)
  {
    highlight();
  }

  public void mouseReleased(MouseEvent evt) {
  }

  public void mouseClicked(MouseEvent evt) {
    anchor.mouseClicked(evt);
  }

  public void mouseEntered(MouseEvent evt) {
  }

  public void mouseExited(MouseEvent evt) {
  }

  public void setData(String data) 
    throws DOMException
  {
    changed = true;
    super.setData(data);
    redraw(); // relayout and redraw
  }


  public void replaceData(int offset, int count, String arg) 
    throws DOMException
  {
    changed = true;
    super.replaceData(offset, count, arg);
    redraw();
  }

  public void deleteData(int offset, int count) 
    throws DOMException
  {	
    changed = true;
    super.deleteData(offset, count);
    redraw();
  }

  public void insertData(int offset, String arg) 
    throws DOMException
  {  
    changed = true;
    super.insertData(offset, arg);
    redraw();
  }

  public void appendData(char[] arg, int start, int length) 
    throws DOMException
  {
    changed = true;
    super.appendData(arg, start, length);
    redraw();
  }

  public void appendData(String arg) 
    throws DOMException
  {
    changed = true;
    super.appendData(arg);
    redraw();
  }
  
  private void highlight()
  {
    for (int i = texts.size() - 1; i >= 0; i--)
      {
	Component c = (Component)texts.elementAt(i);
	Graphics g = c.getGraphics();
	if (g != null)
          {
            g.setColor(alink);
	    c.paint(g);
	    g.dispose();
          }
        else
          {
           Logger.status ("Could not get graphics to highlight text in Text.highlight()");
          }
      }
  }

  private void redraw() 
  {
    ViewElement ve = (ViewElement)getParentNode();
    if (ve != null)
      {
        ve.redraw(true);
      }
  }
}






