/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: TextArea.java,v 1.13 2001/12/11 19:38:30 suraj Exp $
 */ 
package org.xoe.display.dom.xhtml;


import java.awt.Graphics;
import java.awt.event.FocusEvent;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.ViewElement;

public class TextArea
  extends ViewElement
  implements FormElement
{
  int cols = 20;
  int rows = 3;
  int border = 1;
  int scrollbars = java.awt.TextArea.SCROLLBARS_VERTICAL_ONLY;
  boolean editable = true;
  boolean enabled = true;

  String defaultValue = "";
  
  public class BorderTextArea extends java.awt.TextArea 
  {
    
    BorderTextArea (String text, int rows, int cols, int scrolls) 
    {
      super (text, rows, cols, scrolls);
    }
    
  }
  
  public TextArea () 
  {
  }

  public Node cloneNode (boolean deep)
  {
    TextArea nnode = (TextArea)super.cloneNode (deep);
    nnode.cols = cols;
    nnode.rows = rows;
    nnode.border = border;
    nnode.scrollbars = scrollbars;
    return nnode;
  }

  protected void makeVis () 
  {
    vis = new java.awt.TextArea (null, rows, cols, scrollbars);
    /*
      if (border == 0) 
      {
      vis = new java.awt.TextArea(null, rows, cols, scrollbars);
      }
      else 
      {
      vis = new BorderTextArea(null, rows, cols, scrollbars);
      }
    */
    
    ((java.awt.TextArea)vis).setEditable(editable);   
    ((java.awt.TextArea)vis).setEnabled(enabled); 
    
    //initial text
    //assume has a unique child (if has one) and it is the text
    Node child = getFirstChild (); 
    if (child instanceof Text)
      {
	defaultValue = ((Text)child).getData();
        ((java.awt.TextArea)vis).setText(defaultValue);
      }
  }

  public String getText() 
  {
    if (vis == null) {
      return ("");
    }
    else {
      return (((java.awt.TextArea)vis).getText());
    }
  }

  public int getSelectionStart()
  {
    if (vis == null)
      {
	return 0;
      }
    else
      {
	return (((java.awt.TextArea)vis).getSelectionStart());
      }
  }

  public int getSelectionEnd()
  {
    if (vis == null)
      {
	return 0;
      }
    else
      {
	return (((java.awt.TextArea)vis).getSelectionEnd());
      }
  }

  public int getCaretPosition()
  {
    if (vis == null)
      {
	return 0;
      }
    else
      {
	return (((java.awt.TextArea)vis).getCaretPosition());
      }
  }

  public void setCaretPosition(int pos)
  {
    if (vis != null)
      {
	((java.awt.TextArea)vis).setCaretPosition(pos);
      }
  }
  
  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);
    if (localName.equalsIgnoreCase ("text")) 
      {
	setText (value);
      }
    else if (localName.equalsIgnoreCase("cols"))
      {
	try
	  {
	    cols = Integer.parseInt (value);
	  }
	catch (NumberFormatException nfe) 
	  {
	    Logger.log(this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase ("rows")) 
      {
	try 
	  {
	    rows = Integer.parseInt (value);
	  }
	catch (NumberFormatException _)
	  {
	    Logger.log(this, _);
	  }
      }
    else if (localName.equalsIgnoreCase ("scrollbars")) {
	if (value.equalsIgnoreCase ("both"))
	  {
	    scrollbars = java.awt.TextArea.SCROLLBARS_BOTH;
	  }
	else if (value.equalsIgnoreCase ("vertical"))
	  {
	    scrollbars=java.awt.TextArea.SCROLLBARS_VERTICAL_ONLY;
	  }
	else if (value.equalsIgnoreCase ("horizontal"))
	  {
	    scrollbars = java.awt.TextArea.SCROLLBARS_HORIZONTAL_ONLY;
	  }
	else if(value.equalsIgnoreCase ("none"))
	  {
	    scrollbars=java.awt.TextArea.SCROLLBARS_NONE;
	  }
      }
    else if (localName.equalsIgnoreCase("border")) 
      {
	try 
	  {
	    border = Integer.parseInt (value);
	  }
	catch (NumberFormatException nfe) 
	  {
	    Logger.log(this, nfe);
	  }
      }
    else if (localName.equalsIgnoreCase ("readonly")) 
      {
	if ("readonly".equalsIgnoreCase (value) || "true".equalsIgnoreCase (value))
	  {
            editable = false;
          }
	else
	  {
	    editable = true;
	  }
      } 
    else if (localName.equalsIgnoreCase ("disabled")) 
      {
	if ("disabled".equalsIgnoreCase (value) || "true".equalsIgnoreCase (value))
	  {
            enabled = false;
          }
	else 
	  {
	    enabled = true;
	  }
      }
    super.setAttributeNS (namespaceURI, qualifiedName, value);
  }

  public void setText (String text) 
  {
    if (text != null) // TODO find out who passes in a null and fix there if possible
      {
	((java.awt.TextArea)getVis()).setText( text);
      }
  }

  /**
   *  For the first node in the list <code>arg</code>, get it's text content and
   *  initialized the text area with it.
   */
  public void watch (NodeList arg) 
  {
    if (arg.getLength () > 0) 
      {
	String text = DOMUtils.getTextContent (arg.item (0));
	if (text != null)
	  {
	    setText (text);
	  }
	else 
	  {
	    Logger.log( this, "watch: null text for arg.item (0)  = " + XMLWriter.getXML (arg.item (0)));
	  }
      }
    else
      {
	Logger.log (this, "Nothing to watch for: " + getAttributeNS (DEFAULT_NAMESPACE, XPATH_ATTR));
      }
  }

  public String getFormValue () 
  {
     if (!enabled)
      {
	return (null); 
      } 
     else
      {
        return (getText ());
      }
  }

  public void resetValue ()
  {
    setText (defaultValue);
  }
}
