package org.xoe.editor.data; 

import java.awt.Dimension;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.ContentCache;
import org.xoe.core.content.ContentElement;
import org.xoe.core.content.ContentType;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.device.IScreen;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XLink;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceFactory;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.display.IconManager;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.IEditableDocumentBuilder;
import org.xoe.display.dom.xhtml.Dialog;
import org.xoe.servlet.XoeServlet;


public abstract class Main 
  extends XoeServlet
  implements ServiceFactory
{  
  // application states
  protected static final String DEFAULT = "default";
  protected static final String SELECTING_ITEM = "selecting_item";
  protected static final String DISPLAYING_ITEM = "displaying_item";
  protected static final String EDITING_ITEM = "editing_item";
  protected static final String EDITING_NOTE = "editing_note";

  protected static final String ITEM_TEMPLATE = "templates/template.";

  protected String currentState = DEFAULT;
  protected String previousState;

  private Node currentItem = null;
  private XDocument itemDoc = null;

  protected XDocument defaultMenu;
  protected Element menuFrame;

  protected XDocument itemSelectView;
  protected XDocument itemSelectTitle;
  protected XDocument itemSelectMenu;

  protected XDocument itemDisplayView;
  protected XDocument itemDisplayTitle;
  protected XDocument itemDisplayMenu;

  protected XDocument itemEditView;
  protected XDocument itemEditTitle;
  protected XDocument itemEditMenu;

  protected Dialog itemConfirmDeleteDialog;

  protected XDocument noteEditView;
  protected XDocument noteEditTitle;
  protected XDocument noteEditMenu;

  protected Dialog noteConfirmDeleteDialog;
  
  protected XPathContext xpathContext = null;
  
  private IEditableDocumentBuilder currentEDB = null;
  private XDocument currentEditableNote = null;


  // -------------------------------------------------------------------------
  // org.xoe.core.services.ServiceFactory

  public abstract IService newInstance();


  // -------------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    XDocument doc;
     
    res.setContentType (MIMETYPE_TEXT_HTML);
    String path =req.getPathInfo();

    if ("MENU_OPTIONS".equals (path))
      {
        path = req.getParameter ("value");
      }

    if ("MENU_HIDE".equals (path))
      {
        restoreTitle ();
      }
    else if ("ITEM_CANCEL".equals (path) || NO_PATH.equals (path))
      {
        //right now cancel just goes back to the original page without saving.  if we need
        //it to do something different later, we should fill this part in.
        handleItem_select (req, res);
      }
    else if ("ITEM_DISPLAY".equals (path))
      {
        handleItem_display (req, res);
      }
    else if ("ITEM_EDIT".equals (path))
      {
        handleItem_edit (req, res);
      }
    else if ("ITEM_SAVE".equals (path))
      {
        handleItem_save (req, res);
      }
    else if ("ITEM_DELETE".equals (path))
      {
        handleItem_confirmDelete (req, res);
      }
    else if ("ITEM_DELETE_CONFIRM".equals (path))
      {
        handleItem_remove (req, res);
      }
    else if ("ITEM_DELETE_CANCEL".equals (path))
      {
        handleItem_cancelDelete (req, res);
      }
    else if ("ITEM_NEW".equals (path))
      {
        handleItem_createNew (req, res);
      }
    else if ("MENU_SHOW".equals (path))
      {
        handleMenu ();
      }
    else if ("MENU_EDIT".equals (path))
      {
        handleClipboard (req.getParameter ("value"));
      }
    else if ("NOTE_EDIT".equals (path))
      {
        handleNote_edit (req, res);
      }
    else if ("NOTE_SAVE".equals (path))
      {
        handleNote_save (req, res);
      }
    else if ("NOTE_DELETE".equals (path))
      {
        handleNote_confirmDelete (req, res);
      }
    else if ("NOTE_DELETE_CONFIRM".equals (path))
      {
        handleNote_remove (req, res);
      }
    else if ("NOTE_DELETE_CANCEL".equals (path))
      {
        handleNote_cancelDelete (req, res);
      }
    else 
      {
        super.doGet (req, res);
      }
  }


  // -------------------------------------------------------------------------
  // org.xoe.editor.data.Main

  protected void changeState (String newState)
  {
    previousState = currentState;
    currentState = newState;
  }

  protected void restoreTitle ()
    throws IOException
  {
    if (SELECTING_ITEM.equals (currentState))
      {
        setItemSelectTitle ();
      }
    else if (DISPLAYING_ITEM.equals (currentState))
      {
        setItemDisplayTitle ();
      }
    else if (EDITING_ITEM.equals (currentState))
      {
        setItemEditTitle ();
      }
    else if (EDITING_NOTE.equals (currentState))
      {
        setNoteEditTitle ();
      }
  }

  protected abstract String getItemNamespaceURI ();

  protected abstract String getDoctypeName ();

  protected void handleItem_select (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    setCurrentItem (null);
    changeState (SELECTING_ITEM);
    res.getWriter ().print (getItemSelectView ());
    setItemSelectTitle ();
  }

  protected void handleItem_display (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    String selectedItemSrc = req.getParameter ("value");
    if (selectedItemSrc != null)
      {
	setCurrentItem (fetchItem (selectedItemSrc));
      }
    else
      {
        // if the current item isn't null, then we are already 
        // editing an item, so fall through
        if (currentItem == null)
          {
	    throw new ServletException ("no item selected in handleItem_display");
          }
      }

    changeState (DISPLAYING_ITEM);
    res.getWriter ().print (getItemDisplayView ());
    setItemDisplayTitle ();
  }

  protected void handleItem_edit (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    changeState (EDITING_ITEM);
    res.getWriter ().print (getItemEditView ());
    setItemEditTitle ();
  }

  protected void handleItem_save (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    saveEdits (req);
    //load the startup page again
    handleItem_select (req, res);
  }

  protected void handleItem_confirmDelete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getItemConfirmDeleteDialog ().setVisible (true);
  }

  protected void handleItem_cancelDelete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getItemConfirmDeleteDialog ().setVisible (false);
  }

  protected void handleItem_remove (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getItemConfirmDeleteDialog ().setVisible (false);
    removeCurrentItem ();
    handleItem_select (req, res);
  }

  protected void handleItem_createNew (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    createNewItem ();
    handleItem_edit (req, res);
  }

  protected void handleNote_edit (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (currentItem == null)
      {
	String selectedItemSrc = req.getParameter ("item");
	if (selectedItemSrc != null)
	  {
	    setCurrentItem (fetchItem (selectedItemSrc));
          }
        else
          {
	    throw new ServletException ("no item specified in handleNote_edit");
          }
      }

    changeState (EDITING_NOTE);

    // find the first note element
    Element note = (Element)XPath.findFirstNode (xpathContext, currentItem, "xoe:note[1]");
    if (note == null)
      {
	// create new note 
        note = itemDoc.createElementNS (DEFAULT_NAMESPACE, "note");
	currentItem.appendChild (note);
      }

    URL noteURL = null;
    String noteSrc = note.getAttributeNS (XLink.XLINK_NAMESPACE, "href");
    if (noteSrc != null)
      {
	try 
	  { 
	    noteURL = new URL (itemDoc.getURL (), noteSrc);
	  }
	catch (MalformedURLException mue)
	  {
	    throw new ServletException ("invalid href for note: " + noteSrc);
	  }
      }
     else
      {
	noteURL = Stash.createUniqueFilename (new URL (getDataDir ()), "note", "xhtml");
	note.setAttributeNS (XLink.XLINK_NAMESPACE, "href", noteURL.toString ());
      }

    ContentElement content = ContentCache.fetch (noteURL);
    if (content == null)
      {
	XDocument d = new XDocument ();
	Element html = d.createElementNS (XHTML_NAMESPACE, "html");
	d.appendChild (html);
	Element body = d.createElementNS (XHTML_NAMESPACE, "body");
        html.appendChild (body);
	d.setURL (noteURL);
	DocumentCache.putDocument (d);
	content = ContentCache.fetch (noteURL);
      }

    // right now we can only edit text/xml, so create a text/xml type object
    ContentType type = new ContentType ("text/xml");
    currentEDB = (IEditableDocumentBuilder)type.getService (IEditableDocumentBuilder.INTERFACE);
    if (currentEDB == null)
      {
	throw new ServletException ("no editable document builder service found for text/xml");
      }
	
    currentEditableNote = currentEDB.getEditableDocument (content);
    DocumentViewer editFrame = (DocumentViewer)getNoteEditView ().getElementById ("edit");
    if (editFrame == null)
      {
	throw new ServletException ("no tag with id = 'edit' found in note-edit-view");
      }

    editFrame.setDocument (currentEditableNote);
    res.getWriter ().print (getNoteEditView ());
    setNoteEditTitle ();
  }

  protected void handleNote_save (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (currentEDB != null)
      {
        currentEDB.save (currentEditableNote);
        saveCurrentItem ();
      }
    route (req, res);
  }

  protected void handleNote_confirmDelete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getNoteConfirmDeleteDialog ().setVisible (true);
  }

  protected void handleNote_cancelDelete (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getNoteConfirmDeleteDialog ().setVisible (false);
  }

  protected void handleNote_remove (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    getNoteConfirmDeleteDialog ().setVisible (false);
    Node note = XPath.findFirstNode (xpathContext, currentItem, "xoe:note[1]");
    if (note != null)
      {
        currentItem.removeChild (note);
        saveCurrentItem ();
      }
    route (req, res);
  }

  protected void route (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    if (DISPLAYING_ITEM.equals (previousState))
      {
	handleItem_display (req, res);
      }
    else if (EDITING_ITEM.equals (previousState))
      {
	handleItem_edit (req, res);
      }
    else if (SELECTING_ITEM.equals (previousState))
      {
	handleItem_select (req, res);
      }
  }

  protected void handleMenu ()
    throws ServletException, IOException
  {
    if (SELECTING_ITEM.equals (currentState))
      {
        setItemSelectMenu ();
      }
    else if (DISPLAYING_ITEM.equals (currentState))
      {
        setItemDisplayMenu ();
      }
    else if (EDITING_ITEM.equals (currentState))
      {
        setItemEditMenu ();
      }
    else if (EDITING_NOTE.equals (currentState))
      {
        setNoteEditMenu ();
      }
  }

  protected void setMenu (Document doc)
  {
    if (defaultMenu == null)
      {
        defaultMenu = new XDocument ();
        Element html = defaultMenu.createElementNS (XHTML_NAMESPACE, "html");
        defaultMenu.appendChild (html);

        Element body = defaultMenu.createElementNS (XHTML_NAMESPACE, "body");
        html.appendChild (body);

        Element table = defaultMenu.createElementNS (XHTML_NAMESPACE, "table");
        table.setAttributeNS (null, "width", "100%");
        body.appendChild (table);

        Element tr = defaultMenu.createElementNS (XHTML_NAMESPACE, "tr");
        table.appendChild (tr);

        Element td = defaultMenu.createElementNS (XHTML_NAMESPACE, "td");
        td.setAttributeNS (null, "valign", "top");
        tr.appendChild (td);

        menuFrame = defaultMenu.createElementNS (XHTML_NAMESPACE, "iframe");
        menuFrame.setAttributeNS (null, "width", "");
        menuFrame.setAttributeNS (null, "height", "100%");
        td.appendChild (menuFrame);

        td = defaultMenu.createElementNS (XHTML_NAMESPACE, "td");
        td.setAttributeNS (null, "width", "*");
        tr.appendChild (td);

        td = defaultMenu.createElementNS (XHTML_NAMESPACE, "td");
        td.setAttributeNS (null, "align", "right");
        tr.appendChild (td);

        Element font = defaultMenu.createElementNS (XHTML_NAMESPACE, "font");
        font.setAttributeNS (null, "size", "-1");
        td.appendChild (font);

        Element input = defaultMenu.createElementNS (XHTML_NAMESPACE, "input");
        input.setAttributeNS (null, "type", "button");
        input.setAttributeNS (null, "value", "x");
        input.setAttributeNS (null, "onclick", "MENU_HIDE");
        font.appendChild (input);
      }

    ((DocumentViewer)menuFrame).setDocument (doc);
    setTitleDocument (defaultMenu);
  }

  protected XDocument getPage (String viewNameParam, String defaultName)
  {
    String viewName = getServletConfig ().getInitParameter (viewNameParam);
    if (viewName.length () <= 0)
      {
	viewName = defaultName;
      }

    return (getRelativeDocument (viewName));
  }

  protected XDocument getItemSelectView ()
    throws IOException
  {
    if (itemSelectView == null)
      {
	itemSelectView = getPage ("item-select-view", "item-select-view.xhtml");
	if (itemSelectView == null)
	  { 
            throw new IOException ("view not found in getItemSelectView");
          }
      }

    DOMUtils.expandUsingXPathAttributes (itemSelectView, null, xpathContext);
    return (itemSelectView);
  }

  protected void setItemSelectTitle ()
    throws IOException
  {
    if (itemSelectTitle == null)
      {
	itemSelectTitle = getPage ("item-select-title", "item-select-title.xhtml");
      }

    if (itemSelectTitle != null)
      {
	DOMUtils.expandUsingXPathAttributes (itemSelectTitle, null, xpathContext);
      }
    setTitleDocument (itemSelectTitle);
  }

  protected void setItemSelectMenu ()
    throws IOException
  {
    if (itemSelectMenu == null)
      {
	itemSelectMenu = getPage ("item-select-menu", "item-select-menu.xhtml");
      }

    if (itemSelectMenu != null)
      {
	DOMUtils.expandUsingXPathAttributes (itemSelectMenu, null, xpathContext);
      }
    setMenu (itemSelectMenu);
  }

  protected XDocument getItemDisplayView ()
    throws IOException
  {
    if (itemDisplayView == null)
      {
	itemDisplayView = getPage ("item-display-view", "item-display-view.xhtml");
	if (itemDisplayView == null)
	  { 
	    throw new IOException ("view not found in getItemDisplayView");
	  }
      }
    DOMUtils.expandUsingXPathAttributes (itemDisplayView, currentItem, xpathContext);
    return (itemDisplayView);
  }

  protected void setItemDisplayTitle ()
    throws IOException
  {
    if (itemDisplayTitle == null)
      { 
	itemDisplayTitle = getPage ("item-display-title", "item-display-title.xhtml");
      }

    if (itemDisplayTitle != null)
      { 
	DOMUtils.expandUsingXPathAttributes (itemDisplayTitle, currentItem, xpathContext);
      }

    setTitleDocument (itemDisplayTitle);
  }

  protected void setItemDisplayMenu ()
    throws IOException
  {
    if (itemDisplayMenu == null)
      {
	itemDisplayMenu = getPage ("item-display-menu", "item-display-menu.xhtml");
      }

    if (itemDisplayMenu != null)
      {
	DOMUtils.expandUsingXPathAttributes (itemDisplayMenu, null, xpathContext);
      }
    setMenu (itemDisplayMenu);
  }

  protected XDocument getItemEditView ()
    throws IOException
  {
    if (itemEditView == null)
      {
	itemEditView = getPage ("item-edit-view", "item-edit-view.xhtml");
	if (itemEditView == null)
	  { 
            throw new IOException ("view not found in getItemEditView");
          }
      }
    DOMUtils.expandUsingXPathAttributes (itemEditView, currentItem, xpathContext);
    return (itemEditView);
  }

  protected void setItemEditTitle ()
    throws IOException
  {
    if (itemEditTitle == null)
      { 
	itemEditTitle = getPage ("item-edit-title", "item-edit-title.xhtml");
      }

    if (itemEditTitle != null)
      { 
	DOMUtils.expandUsingXPathAttributes (itemEditTitle, currentItem, xpathContext);
      }
    setTitleDocument (itemEditTitle);
  }

  protected void setItemEditMenu ()
    throws IOException
  {
    if (itemEditMenu == null)
      {
	itemEditMenu = getPage ("item-edit-menu", "item-edit-menu.xhtml");
      }

    if (itemEditMenu != null)
      {
	DOMUtils.expandUsingXPathAttributes (itemEditMenu, null, xpathContext);
      }
    setMenu (itemEditMenu);
  }

  protected XDocument getNoteEditView ()
    throws IOException
  {
    if (noteEditView == null)
      {
	noteEditView = getPage ("note-edit-view", "note-edit-view.xhtml");
	if (noteEditView == null)
	  { 
            throw new IOException ("view not found in getNoteEditView");
          }
      }
//    DOMUtils.expandUsingXPathAttributes (noteEditView, null, xpathContext);
    return (noteEditView);
  }

  protected void setNoteEditTitle ()
    throws IOException
  {
    if (noteEditTitle == null)
      { 
	noteEditTitle = getPage ("note-edit-title", "note-edit-title.xhtml");
      }

    if (noteEditTitle != null)
      { 
	DOMUtils.expandUsingXPathAttributes (noteEditTitle, currentItem, xpathContext);
      }
    setTitleDocument (noteEditTitle);
  }

  protected void setNoteEditMenu ()
    throws IOException
  {
    if (noteEditMenu == null)
      {
	noteEditMenu = getPage ("note-edit-menu", "note-edit-menu.xhtml");
      }

    if (noteEditMenu != null)
      {
	DOMUtils.expandUsingXPathAttributes (noteEditMenu, null, xpathContext);
      }
    setMenu (noteEditMenu);
  }

  protected Dialog buildDialog (XDocument dialogView)
  {
    Dialog dialog = new Dialog ("Tim Says: no title bar plz", false);
    dialog.setDocument (dialogView);
    dialog.pack ();
    Dimension screensize = ((IScreen) ServiceLocator.findImplementor (IScreen.INTERFACE)).getScreen ().getSize ();
    Dimension dialogsize = dialog.getSize();
    dialog.setLocation ((screensize.width - dialogsize.width) / 2, (screensize.height - dialogsize.height) / 2);
    return (dialog);
  }

  protected Dialog getItemConfirmDeleteDialog ()
    throws IOException
  {
    if (itemConfirmDeleteDialog == null)
      {
	XDocument dialogView = getPage ("item-confirm_delete-dialog", "item-confirm_delete-dialog.xhtml");
	if (dialogView == null)
	  { 
            throw new IOException ("view not found in getItemConfirmDeleteDialog");
          }

        // notify us if there is an action or onclick event
	dialogView.setCallback ("action", defaultCallback);
	dialogView.setCallback ("onclick", defaultCallback);

        itemConfirmDeleteDialog = buildDialog (dialogView);
      }

    DOMUtils.expandUsingXPathAttributes (itemConfirmDeleteDialog.getDocument (), currentItem, xpathContext);
    return (itemConfirmDeleteDialog);
  }

  protected Dialog getNoteConfirmDeleteDialog ()
    throws IOException
  {
    if (noteConfirmDeleteDialog == null)
      {
	XDocument dialogView = getPage ("note-confirm_delete-dialog", "note-confirm_delete-dialog.xhtml");
	if (dialogView == null)
	  { 
            throw new IOException ("view not found in getNoteConfirmDeleteDialog");
          }

        // notify us if there is an action or onclick event
	dialogView.setCallback ("action", defaultCallback);
	dialogView.setCallback ("onclick", defaultCallback);

        noteConfirmDeleteDialog = buildDialog (dialogView);
      }

    DOMUtils.expandUsingXPathAttributes (noteConfirmDeleteDialog.getDocument (), currentItem, xpathContext);
    return (noteConfirmDeleteDialog);
  }

  /**
   *  return true if enough data exists to consider this a true data element
   */
  protected abstract boolean foundData (Node n, String value);


  protected void saveEdits (HttpServletRequest req)
    throws IOException
  {
    // only write the data back if there was any
    if (saveEdits (req, currentItem))
      {
        saveCurrentItem ();
      }
    else
      {
        removeCurrentItem ();
      }
  }

  protected boolean saveEdits (HttpServletRequest req, Node contextNode)
    throws IOException
  {
    // flag to indicate some data was passed to us
    boolean foundData = false;

    // loop through the name/value pairs
    Enumeration names = req.getParameterNames ();
    while (names.hasMoreElements ())
      {
        String name = (String)names.nextElement ();
        String value = req.getParameter (name);
        if (value != null)
          {
	    Node n = XPath.findFirstNode (xpathContext, contextNode, name);
	    if (n != null)
	      {
		DOMUtils.setTextContent (n, value);
		foundData |= foundData (n, value);
	      }
	    else
	      {
		 Logger.log (this, "saveEdits: name doesn't match a data node -> " + name);
	      }
          }
	else
	  {
	     Logger.log (this, "saveEdits: null value for parameter --> " + name);
	  }
      }

    return (foundData);
  }

  protected void saveCurrentItem ()
    throws IOException
  {
    if (itemDoc.getURL () == null)
      {
        String fileName = "XOE" 
                          + ContentElement.encodeName (getFileNamePrefix ())
                          + Long.toString (System.currentTimeMillis ()) 
                          + '.'
                          + getDoctypeName ();

	itemDoc.setURL (new URL (getDataDir () + fileName));
      }

    DocumentCache.putDocument (itemDoc);
  }

  protected String getFileNamePrefix ()
  {
    return getFileNamePrefix (itemDoc);
  }

  /**
   *  return a unique string to use as the stash name for a new item
   */
  protected abstract String getFileNamePrefix (XDocument file);

  protected void removeCurrentItem ()
    throws IOException
  {
    // if the current item has a url and is in the
    // stash then remove it
    if (itemDoc.getURL () != null
        && Stash.contains (itemDoc.getURL ()))
      {
	Stash.remove (itemDoc.getURL ());
      }
  }

  public void setXPathContext (XPathContext xpathContext)
  {
    this.xpathContext = xpathContext;
  }

  protected URL getNoteIconURL ()
  {
    return (IconManager.getDefaultIcon ());
  }

  protected void createNewItem ()
    throws ServletException, IOException
  {
    itemDoc = new XDocument ();
    XDocument templateDoc = getFileFromPackageWorkspace (ITEM_TEMPLATE + getDoctypeName ());
    if (templateDoc == null)
      {
        throw new ServletException (ITEM_TEMPLATE + getDoctypeName () + " not found in createNewItem");
      }

    currentItem = itemDoc.importNode (templateDoc.getDocumentElement (), true);
    itemDoc.appendChild (currentItem);
  }

  protected Node fetchItem (String itemDocURI)
    throws ServletException, IOException
  {
    XDocument itemDoc = DocumentCache.fetchDocument (itemDocURI);
    if (itemDoc == null)
      {
	throw new ServletException ("item document not found in fetchItem: " + itemDocURI);
      }

    return (itemDoc.getDocumentElement ());
  } 

  protected Node getCurrentItem ()
  {
    return (currentItem);
  }

  protected void setCurrentItem (Node item)
  {
    currentItem = item;
    if (currentItem != null)
      {
        itemDoc = (XDocument) currentItem.getOwnerDocument ();
      }
    else
      {
        itemDoc = null;
      }
  }

  protected XDocument getItemDocument ()
  {
    return (itemDoc);
  }

}
