package org.xoe.display.dom;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;

import kaffe.util.IntegerHashtable;
import kaffe.awt.AWT;

import org.w3c.dom.NodeList;


public final class LayoutCursor 
{

  /**
   * Phases of layout.
   */
  final public static int PHASE1 = 1;
  final public static int PHASE2 = 2;
  final public static int PHASE3 = 3;

  /**
   * Various alignment settings.
   */
  final public static int NONE = 0;
  final public static int LEFT = 1;
  final public static int CENTER = 2;
  final public static int RIGHT = 3;
  final public static int TOP = 4;
  final public static int BOTTOM = 5;

  final public static int BIGY = Integer.MAX_VALUE;

  final private static String[] supportedFonts = Toolkit.getDefaultToolkit ().getFontList ();

  public static class DummyContainer extends Container {
  }

  final private static Container dummyVis = new DummyContainer();

  private static IntegerHashtable fonts = new IntegerHashtable();

  /**
   * Current layout phase
   */
  public int phase;

  /**
   * Current x,y position of the cursor in the rendering window.
   */
  public int x;
  public int y;

  /**
   * Current indent from margin
   */
  public int leftindent;
  public int topindent;

  /**
   * List indentation level.
   */
  public int listlevel;

  /**
   * Size of the rendering window.
   */
  public int width;
  public int height;

  /**
   * Offset for final layout.
   */
  public int xoff;
  public int yoff;

  /**
   * Maximum width/height of each element seen by this cursor.
   */
  public int maxwidth;
  public int maxheight;

  /**
   * Maximum X position assuming no soft-newlines seen.
   */
  public int mx;

  /**
   * Maximum nobreaking space.
   */
  public int nobrx;

  /**
   * Maximum line length of all elements see by this cursor.
   */
  public int maxlinewidth;

  /**
   * Current visual to add components to.
   */
  public Container vis;

  /**
   * Visual to flush components to.
   */
  public Container fvis;

  /**
   * Proxy visual for handling centering, etc.
   */
  private Container pvis;

  /**
   * Current anchor
   */
  public Object anchor;
  public Color anchorclr = Color.blue;
  public Color activeclr = Color.red;

  /**
   * Current layout attributes
   */
  public static Font deffont = new Font("Default", Font.PLAIN, 9);
  public Font font;
  public FontMetrics fontmetrics;
  public Color fgclr = Color.black;
  public Color bgclr = null;
  public boolean underline;
  public boolean strike;
  public int halign = LEFT;
  public int valign = BOTTOM;
  public boolean nobreaks;
  public boolean literal;
  public String style;

  /**
   * Have we seen a background image?
   */
  public boolean havebackground;

  /**
   * Current pending floating elements which have yet to be inserted
   */
  Vector pendingfloats;

  /**
   * Active floating elements
   */
  Stack leftfloats;
  Stack rightfloats;

  /**
   * Y position where we next have to remove left/right float.
   */
  int floatlefty;
  int floatrighty;

  /**
   * Extra width occupied by active floats.  Need this to get the maxwidth
   * right.
   */
  int floatwidth;

  /*
   * True if we have any active or pending floats.
   */
  boolean havefloats;

  // ----------------------------------------------------------------------- //

  public LayoutCursor() {
  }

  public void set(int w, int h, int l, int t) {
	  resetAll();
	  width = w;
	  height = h;
	  xoff = l;
	  yoff = t;
  }

  public void set(LayoutCursor ocur, ViewElement v) {
	  resetAll();
	  width = (ocur.width + ocur.floatwidth) - ocur.x;
	  height = ocur.height - ocur.y;
	  fvis = v.getCVis();
	  if (fvis != null) {
		  fvis.removeAll();
	  }
	  else {
		  fvis = dummyVis;
	  }
	  vis = fvis;
	  fgclr = ocur.fgclr;
	  bgclr = ocur.bgclr;
	  font = ocur.font;
	  fontmetrics = ocur.fontmetrics;
	  anchorclr = ocur.anchorclr;
	  activeclr = ocur.activeclr;
	  underline = ocur.underline;
	  havebackground = ocur.havebackground;
	  style = ocur.style;
	  phase = ocur.phase;
  }

  public void setArea(int x, int y, int w, int h) {
	  reset();
	  xoff = x;
	  yoff = y;
	  width = w;
	  height = h;
  }

  public void reset() {
	  x = 0;
	  mx = 0;
	  nobrx = 0;
	  y = 0;
	  maxheight = 0;
	  maxwidth = 0;
	  maxlinewidth = 0;
  }

  private void resetAll() {
	  x = 0;
	  y = 0;
	  leftindent = 0;
	  topindent = 0;
	  listlevel = 0;
	  width = 0;
	  height = 0;
	  xoff = 0;
	  yoff = 0;
	  maxwidth = 0;
	  maxheight = 0;
	  mx = 0;
	  nobrx = 0;
	  maxlinewidth = 0;
	  fvis = dummyVis;
	  vis = fvis;
	  anchor = null;
	  anchorclr = Color.blue;
	  activeclr = Color.red;
	  setFont(deffont);
	  fgclr = Color.black;
	  bgclr = null;
	  underline = false;
	  halign = LEFT;
	  valign = BOTTOM;
	  havebackground = false;
	  nobreaks = false;
	  literal = false;
	  if (pendingfloats != null) {
		  pendingfloats.removeAllElements();
	  }
	  if (leftfloats != null) {
		  leftfloats.removeAllElements();
	  }
	  if (rightfloats != null) {
		  rightfloats.removeAllElements();
	  }
	  floatlefty = BIGY;
	  floatrighty = BIGY;
	  floatwidth = 0;
	  havefloats = false;
	  style = "default";
	  phase = PHASE1;
  }

  public void setLocation(Component c) {
	  c.setLocation(xoff+x, yoff+y);
  }

  public void setBounds(Component c, Dimension size) {
	  setBounds(c, size.width, size.height);
  }

  public void setBounds(Component c, int width, int height) {
	  if (vis != dummyVis && c.getParent() != vis) {
		  vis.add(c);
	  }
	  c.setBounds(xoff+x, yoff+y, width, height);
  }

  public void newline(int nw) {
	  if (x + nw > width) {
		  // This is a soft-newline, so we keep hold of the
		  // mx value.
		  int omx = mx;
		  int onobrx = nobrx;
		  newline();
		  nobrx = onobrx;
		  mx = omx;
	  }
  }

  public void newline() {
	  int w = x;
	  int h = maxheight;
	  if (mx > maxlinewidth) {
		  maxlinewidth = mx;
	  }
	  if (nobreaks && nobrx > maxwidth) {
		  maxwidth = nobrx;
	  }

	  x = leftindent;
	  mx = leftindent;
	  nobrx = leftindent;
	  y += maxheight;
	  maxheight = 0;
	  flushline(w, h);
	  if (havefloats) {
		  removefloats();
		  insertfloats();
		  if (pendingfloats.isEmpty() && leftfloats.isEmpty() && rightfloats.isEmpty()) {
			  havefloats = false;
		  }
	  }
  }

  public void finish() {
	  newline();
	  flushfloats(CENTER);
  }

  public void flushfloats(int type) {
	  if (havefloats == false || type == NONE) {
		  return;
	  }
	  switch (type) {
	  case LEFT:
		  while (!leftfloats.isEmpty()) {
			  maxheight = floatlefty - y;
			  newline();
		  }
		  break;
	  case RIGHT:
		  while (!rightfloats.isEmpty()) {
			  maxheight = floatrighty - y;
			  newline();
		  }
		  break;
	  case CENTER:
	  default:
		  while (havefloats) {
			  int lh = floatlefty - y;
			  int rh = floatrighty - y;
			  maxheight = (lh < rh ? lh : rh);
			  newline();
		  }
		  break;
	  }
  }

  private void flushline(int w, int h) {
	  if (fvis == vis) {
		  return;
	  }
	  int len = vis.getComponentCount();
	  if (len > 0) {
		  int xoff;
		  switch (halign) {
		  case LEFT:
		  case NONE:
		  default:
			  xoff = 0;
			  break;
		  case CENTER:
			  xoff = (width - w) / 2;
			  break;
		  case RIGHT:
			  xoff = width - w;
			  break;
		  }
		  if (xoff < 0) {
			  xoff = 0;
		  }

		  // Align all components to bottom of line within the tablecell
		  for (int i = len - 1; i >= 0; i--) {
			  Component cmp = vis.getComponent(i);

			  int yoff;
			  switch (valign) {
			  case BOTTOM:
			  default:
				  yoff = h - AWT.extension.getHeight(cmp);
				  break;
			  case CENTER:
				  yoff = (h - AWT.extension.getHeight(cmp)) / 2;
				  break;
			  case TOP:
				  yoff = 0;
				  break;
			  }
			  cmp.setLocation(xoff + AWT.extension.getX(cmp), topindent + yoff + AWT.extension.getY(cmp));
			  fvis.add(cmp);
		  }
	  }
  }

  private void insertfloats() {
	  while (!pendingfloats.isEmpty()) {
		  FloatViewElement fe = (FloatViewElement)pendingfloats.elementAt(0);
		  // Will it fit?  No then return
		  if (fe.fminwidth > width && (!leftfloats.isEmpty() || !rightfloats.isEmpty())) {
			  return;
		  }
		  // Move element to active set
		  pendingfloats.removeElementAt(0);

		  if (fvis != vis) {
			  // Insert visual
			  Component myvis = fe.getVis();
			  fvis.add(myvis);
			  if (fe.falign == RIGHT) {
				  // Right align
				  myvis.setLocation(xoff + width - fe.fminwidth, yoff + y);
			  }
			  else {
				  // Left align
				  myvis.setLocation(xoff, yoff + y);
			  }
		  }
		  if (fe.falign == RIGHT) {
			  rightfloats.push(fe);
			  floatrighty = y + fe.fheight;
		  }
		  else {
			  xoff += fe.fminwidth;
			  leftfloats.push(fe);
			  floatlefty = y + fe.fheight;
		  }

		  // Adjust layout
		  fe.fendy = y + fe.fheight;
		  width -= fe.fminwidth;
		  floatwidth += fe.fminwidth;
		  if (floatwidth > maxwidth) {
			  maxwidth = floatwidth;
		  }
		  mx += fe.fmaxwidth;
		  nobrx += fe.fmaxwidth;
	  }
  }

  private void removefloats() {
	  // Floats still active? return now
	  if (y >= floatlefty) {
		  // Remove active floats which need to go inactive
		  floatlefty = BIGY;
		  while (!leftfloats.isEmpty()) {
			  FloatViewElement fe = (FloatViewElement)leftfloats.peek();
			  if (y < fe.fendy) {
				  floatlefty = fe.fendy;
				  break;
			  }
			  if (fe.falign == LEFT) {
				  xoff -= fe.fminwidth;
			  }
			  width += fe.fminwidth;
			  floatwidth -= fe.fminwidth;
			  leftfloats.pop();
		  }
	  }
	  if (y >= floatrighty) {
		  // Remove active floats which need to go inactive
		  floatrighty = BIGY;
		  while (!rightfloats.isEmpty()) {
			  FloatViewElement fe = (FloatViewElement)rightfloats.peek();
			  if (y < fe.fendy) {
				  floatrighty = fe.fendy;
				  break;
			  }
			  if (fe.falign == LEFT) {
				  xoff -= fe.fminwidth;
			  }
			  width += fe.fminwidth;
			  floatwidth -= fe.fminwidth;
			  rightfloats.pop();
		  }
	  }
  }

  public void move(int nw, int nh) {
	  mx += nw;
	  nobrx += nw;
	  // Add in leftindent if we've at the beginning of a line
	  if (x == leftindent) {
		  nw += leftindent;
		  x = nw;
	  }
	  else {
		  x += nw;
	  }
	  if (nw > maxwidth) {
		  maxwidth = nw;
	  }
	  if (nh > maxheight) {
		  maxheight = nh;
	  }
  }

  public void flow(FloatViewElement felem, int falign, int fw, int fmw, int fh) {
	  if (pendingfloats == null) {
		  pendingfloats = new Vector();
		  leftfloats = new Stack();
		  rightfloats = new Stack();
	  }
	  felem.setFloat(falign, fw, fmw, fh);
	  pendingfloats.addElement(felem);
	  if (x == leftindent) {
		  insertfloats();
	  }
	  havefloats = true;
  }

  public void blankline() {
	  newline();
	  maxheight = fontmetrics.getHeight() + fontmetrics.getDescent();
	  newline();
  }

  /**
   * Set the font to the name, style and size given.
   *  We go to a fair bit of effort to cache fonts here and reuse them
   *  rather than creating them each time since it's *much* more efficient.
   */
  public void setFont(String fname, int fstyle, int fsize) {
	  // Attempt to set the correct font by seeing what we support.
	  // Fall back on 'default' if we can't find a better match
	  String newfont = "Default";
	  for (int i = supportedFonts.length - 1; i >= 0; i--) {
		  if (supportedFonts[i].equalsIgnoreCase (fname)) {
			  newfont = fname;
			  break;
		  }
	  }

	  Font fnt = null;
	  if (fsize < 1) {
		  fsize = 1;
	  }
	  int key = fsize * 16 + fstyle;
	  Hashtable ftable = (Hashtable)fonts.get(key);
	  if (ftable != null) {
		  fnt = (Font)ftable.get(newfont);
	  }
	  if (fnt == null) {
		  if (ftable == null) {
			  ftable = new Hashtable(4);
			  fonts.put(key, ftable);
		  }
		  fnt = new Font(newfont, fstyle, fsize);
		  ftable.put(newfont, fnt);
	  }
	  setFont(fnt);
  }

  public void setFont(Font fnt) {
	  font = fnt;
	  fontmetrics = vis.getFontMetrics(font);
  }

  public void addVis(ViewElement v) {
	  addVis(v.getVis());
  }

  public void addVis(Component c) {
	  if (vis != dummyVis) {
		  vis.add(c);
	  }
  }

  public void useProxyVis() {
	  fvis = vis;
	  if (pvis == null) {
		  pvis = new DummyContainer();
	  }
	  vis = pvis;
  }

  public final boolean isPhase1() {
	  return (phase == PHASE1);
  }

  public final boolean isPhase2() {
	  return (phase == PHASE2);
  }

  public final boolean isPhase3() {
	  return (phase == PHASE3);
  }

  public String toString() {
	  return (getClass() + "[("+xoff+","+yoff+")("+width+","+height+")("+x+","+y+")("+maxwidth+","+maxheight+")]");
  }

}



















