/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: Render.java,v 1.17 2001/12/12 00:15:33 suraj Exp $
 */
package org.xoe.display.dom;

import java.awt.Container;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;

import kaffe.awt.AWT;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.device.IScreen;
import org.xoe.core.dom.XDocument;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;


public class Render
  extends Container 
{
  private LayoutCursor cur = new LayoutCursor();
  private XDocument doc;
  private Dimension prefsize;
  private int version = 1;

  /**
   * Return the document we're displaying.
   */
  public Document getDocument() {
	  return (doc);
  }

  /**
   * Search the document we're displaying and return the named frame.
   */
  public Node getFrame(String framename) {
	  if (doc == null) {
		  return (null);
	  }
	  else if ("_top".equals(framename)) {
		  return (doc.getDocumentElement ());
	  }
	  else {
		  return (XPath.findFirstNode(doc, "//xhtml:frame[@name='"+framename+"']"));
	  }
  }

  /**
   * Calculate the preferred size of the renders contents by giving it the
   * whole screen and seeing how much of it it wants.
   */
  public Dimension preferredSize() {
	  return (getPreferredSize(new Dimension(), true));
  }

  public synchronized Dimension getPreferredSize(Dimension limits, boolean expandheight)
  {
	  Dimension scrsz = null;

	  if (limits.width == 0 || limits.height == 0) {
		  scrsz = ((IScreen) ServiceLocator.findImplementor (IScreen.INTERFACE)).getScreen ().getSize ();
		  if (limits.width == 0) {
			  limits.width = scrsz.width;
		  }
		  if (limits.height == 0) {
			  limits.height = scrsz.height;
		  }
	  }
	  if (doc == null) {
		  return (new Dimension(0, 0));
	  }
	  else {
		  long time = System.currentTimeMillis ();

// 		  ImageFactory.stop();

		  doc.version = version;
		  ViewElement vis = (ViewElement)doc.getDocumentElement ();
		  cur.set(0, 0, 0, 0);
		  cur.phase = LayoutCursor.PHASE1;
		  vis.layout(cur);
		  cur.finish();
		  cur.set(limits.width, 0, 0, 0);
		  cur.phase = LayoutCursor.PHASE2;
		  vis.layout(cur);
		  cur.finish();
		  int height = (expandheight || cur.y > limits.height ? limits.height : cur.y);
		  cur.set(limits.width, height, 0, 0);
		  cur.phase = LayoutCursor.PHASE3;
		  vis.layout(cur);
		  cur.finish();

		  ImageFactory.flush();

		  prefsize = new Dimension(limits.width, height);

		  time = System.currentTimeMillis () - time;
		  Logger.log(this, "layout time " + time + "ms");

		  return (prefsize);
	  }
  }

  /**
   * Lay the contents out (if we have any).  If the version of the contents
   * and the version of the container don't match, we lay them out using PHASE1
   * and then propogate the current decoration (in case the theme changed).
   * Then, if the contents have changed, we relay them out.  If they haven't
   * changed we simply insert them into the container so we can paint them
   */
  public void layout() {
	  if (doc != null) {
		  doc.version = version;
		  doc.clearModified ();

		  removeAll();

		  Dimension size = getSize ();
		  if (!size.equals (prefsize)) {
			  getPreferredSize(size, true);
		  }

		  add(((ViewElement)doc.getDocumentElement ()).vis);

		  // System.out.println("========================================"); AWT.extension.dump("", ((ViewElement)doc.getDocumentElement()).vis);
	  }
  }

  /**
   * Paint the XHTML
   */
  public void paint(Graphics g) {
	  AWT.extension.paintBackground(this, g);
	  super.paint(g);
  }

  public void invalidate () {
	  prefsize = null;
	  super.invalidate ();
  }

  public void flush () {
	  version++;
	  invalidate ();
  }

  /**
   * Set the current document being displayed by the render
   * and invalidate the container so it get's layed out again.
   */
  public synchronized void setDocument (Document newdoc) {
	  if (newdoc == doc) { // Do nothing if document is the same
		  return;
	  }
	  if (doc != null) {
		  ViewElement vis = (ViewElement)doc.getDocumentElement ();
		  if (vis.vis != null) {
			  remove(vis.vis);
		  }
		  vis.unlayout();
	  }

	  doc = (XDocument)newdoc;
	  ViewElement vis = (ViewElement)doc.getDocumentElement ();
	  if (vis.vis != null) {
		  add(vis.vis);
	  }
	  invalidate();
  }

}
