package org.xoe.display.dom;

import java.awt.Component;
import java.util.Stack;

import org.w3c.dom.Node;

import org.xoe.core.dom.Callback;
import org.xoe.core.util.Logger;


public class Thread
  extends java.lang.Thread
{
  private static final int TIMEOUT = 20 * 1000;

  private static int cnt = 0;
  private static Stack freelist = new Stack();

  private Callback callback;
  private Node node;
  private Object payload;
  Component redraw;

  Thread()
  {
    super("XHTMLThread-" + cnt);
    cnt++;
  }

  public static void invoke(Callback call, Node node, Object payload)
  {
    if (call == null)
      {
	return;
      }
    Thread tid;
    if (freelist.empty ())
      {
	tid = new Thread ();
	tid.start();
      }
    else
      {
	tid = (Thread)freelist.pop ();
      }
    synchronized (tid)
      {
	tid.callback = call;
	tid.node = node;
	tid.payload = payload;
	tid.notify();
      }
    
  }

  public synchronized void run()
  {
    for (;;)
      {
	// No callback - wait to be given one
	if (callback == null)
	  {
	    try
	      {
		wait (TIMEOUT);
	      }
	    catch (InterruptedException _)
	      {
	      }
	    // Still no callback - timed out - terminate thread
	    if (callback == null)
	      {
		freelist.removeElement(this);
		Logger.log(this, "Running GC");
		System.gc ();
		break;
	      }
	  }

	// Invoke the callback
	try
	  {
	    callback.invoke(node, payload);
	    // Redraw
	    if (redraw != null && !redraw.isValid())
	      {
		Component r = redraw; // Because validating might change it!
		r.validate();
		r.repaint();
	      }
	  }
	catch (Throwable e)
	  {
	    Logger.log(this, e);
	  }

	// Clear invoke parameters and put on freelist for reuse
	callback = null;
	node = null;
	payload = null;
	redraw = null;
	freelist.push(this);
      }
  }

}
