/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *  
 * $tvt: ViewElement.java,v 1.31 2001/12/13 00:38:22 suraj Exp $
 */ 
package org.xoe.display.dom;

import org.xoe.core.util.XMLWriter;

import com.pocketlinux.awt.theme.StyleComponentDeco;
import com.pocketlinux.awt.theme.Theme;

import java.awt.Adjustable;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import kaffe.awt.AWT;

import org.w3c.dom.DOMException;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xoe.core.content.DocumentCache;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.Loadable;
import org.xoe.core.dom.XDocument;
import org.xoe.core.dom.XElement;
import org.xoe.core.dom.XNode;
import org.xoe.core.util.Logger;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathContext;
import org.xoe.core.xpath.XPathParser;


public class ViewElement
  extends XElement
  implements Viewable, MouseListener, AdjustmentListener, Loadable
{
  protected static final int HAS_WIDTH = 0x01;
  protected static final int HAS_HEIGHT = 0x02;
  protected static final int SLACK_WIDTH = 0x04;
  protected static final int SLACK_HEIGHT = 0x08;
  protected static final int HAS_MOUSE_LISTENER = 0x010;
  protected static final int HAS_ADJUSTMENT_LISTENER = 0x20;
 
  protected static final int LISTENERS = HAS_MOUSE_LISTENER|HAS_ADJUSTMENT_LISTENER;

  public int preferredWidth;
  public int preferredHeight;
  protected Component vis;
  protected boolean changed;

  protected String style;
  protected String fontmodifier;
  public int flags;
  protected XPath dataXPath;
  
  protected ViewElement () 
  {
    super ();
  }

  /**
   * Method called when mouse drags an element (e.g. scrollbar)
   */
  public void adjustmentValueChanged (AdjustmentEvent evt ) 
  {
    invoke (ONCHANGE_ATTR);
  }

  private ViewElement find (Component c) 
  {
    if (c == vis) 
      {
	return (this);
      }
    for (Node child = getFirstChild (); child != null; child = child.getNextSibling ()) 
      {
	if (child instanceof ViewElement) 
	  {
	    ViewElement f = ((ViewElement)child).find (c);
	    if (f != null) 
	      {
		return ((ViewElement)child);
	      }
	  }
      }
    return (null);
  }

  private ViewElement findViewElement (Component c) 
  {
    // Walk up parent chain to find the top ViewElement
    ViewElement start = this;
    for (Node cur = this; cur != null; cur = cur.getParentNode ()) 
      {
	if (cur instanceof ViewElement) 
	  {
	    start = (ViewElement)cur;
	  }
      }
    
    // Starting from this point, look down for the named component
    return (start.find (c));
  }

  public Container getCVis () 
  {
    return ((Container)getVis ());
  }

  public static Image getImage (String iname) 
  {
    return (((Theme)AWT.getTheme ()).getImage (iname));
  }
  
  /*
   * Set the style of the visual based on the current style paramater and
   * the visual type.
   */
  protected StyleComponentDeco getDeco () 
  {
    return (((Theme)AWT.getTheme ()).getDecoWithStyle (vis, style, null));
  }

  protected int getPreferredHeight (int parentHeight) 
  {
    if ((flags & HAS_HEIGHT) == 0) 
      {
	return (-1);
      }
    else if (preferredHeight >= 0) 
      {
	return (preferredHeight);
      }
    else 
      {
	return (parentHeight * -preferredHeight / 100);
      }
  }

  protected int getPreferredWidth (int parentWidth) 
  {
    if ((flags & HAS_WIDTH) == 0) 
      {
	return (-1);
      }
    else if (preferredWidth >= 0) 
      {
	return (preferredWidth);
      }
    else 
      {
	return (parentWidth * -preferredWidth / 100);
      }
  }

  protected Dimension getPreferredSize (LayoutCursor cur)
  {
    Dimension size = getVis ().getPreferredSize ();
    int v = getPreferredWidth (cur.width);
    if (v > size.width)
      {
	size.width = v;
      }
    v = getPreferredHeight (cur.height);
    if (v > size.height)
      {
	size.height = v;
      }
    return (size);
  }

  public Component getVis () 
  {
    if (vis == null) 
      {
	makeVis ();
	if (vis != null) 
	  {
	    if ((flags & HAS_MOUSE_LISTENER) != 0) 
	      {
		vis.addMouseListener (this);
	      }
	    if ((flags & HAS_ADJUSTMENT_LISTENER) != 0 
                && vis instanceof Adjustable) 
	      {
		((Adjustable)vis).addAdjustmentListener (this);
	      }
	    // HACK - some components aren't styled yet, so we return null
	    // for them so they use the default styling.
	    kaffe.awt.theme.ComponentDeco deco = getDeco ();
	    if (deco != null)
	      {
		AWT.extension.setDeco (vis, deco);
	      }
	  }
      }
    return (vis);
  }

  public XPath getXPath () 
  {
    return (dataXPath);
  }

  /**
   * The default layout assume we're just wrapping a standard java component
   * We get the preferred size and add it in, making allowances for newlines
   * as we go.
   */
  public void layout (LayoutCursor cur) 
  {
    style = cur.style;
    Component vis = getVis ();
    if (vis == null) 
      {
	layoutChildren (cur);
      }
    else 
      {
	vis.setFont (cur.font);
	vis.setForeground (cur.fgclr);

	Dimension size = getPreferredSize (cur);

	cur.newline (size.width);
	cur.setBounds (vis, size);
	cur.move (size.width, size.height);
      }
  }

  protected void layoutChildren (LayoutCursor cur) 
  {
    for (Node child = getFirstChild (); child != null; child = child.getNextSibling ()) 
      {
	// Try to tolerate bad children
	if (child instanceof Viewable) 
	  {
//System.out.println(this + " -> PHASE " + cur.phase + " start");
	    ((Viewable)child).layout (cur);
//System.out.println(this + " -> PHASE " + cur.phase + " end");
	  }
      }
  }

  protected void makeVis () 
  {
  }

  /**
   * All unlayout does is to traverse the DOM tree and remove all the
   * visuals.
   */
  public void unlayout () 
  {
    unmakeVis ();
    for (Node child = getFirstChild (); child != null; child = child.getNextSibling ()) 
      {
	// Try to tolerate bad children
	if (child instanceof Viewable) 
	  {
	    ((Viewable)child).unlayout ();
	  }
      }
  }

  protected void unmakeVis () 
  {
    if (vis instanceof Container) 
      {
	((Container)vis).removeAll ();
      }
    vis = null;
  }

  public static Image lookupIcon (String iconname) 
  {
    return (((Theme)AWT.getTheme ()).getIcon (iconname));
  }

  /**
   * Method called when element is 'clicked' on.
   */
  public void mouseClicked (MouseEvent evt) 
  {
    invoke (ONCLICK_ATTR);
  }

  public void mouseEntered (MouseEvent evt) 
  {
    invoke (ONMOUSEOVER_ATTR);
  }

  public void mouseExited (MouseEvent evt) 
  {
    invoke (ONMOUSEOUT_ATTR);
  }

  /**
   * Method called when mouse/pen is pressed on this element.
   */
  public void mousePressed (MouseEvent evt) 
  {
    invoke (ONMOUSEDOWN_ATTR);
  }
  
  /**
   * Method called when mouse/pen is released on this element.
   */
  public void mouseReleased (MouseEvent evt) 
  {
    ViewElement ve = null;
    Component c = AWT.extension.getComponentUnderMouse ();
    if (c != getVis ()) 
      {
	ve = findViewElement (c);
      }
    if (ve == null) 
      {
	ve = this;
      }
    ve.invoke (ONMOUSEUP_ATTR);
  }
  
  public void onLoad () 
  {
    onLoad (null, null);
  }

  /**
   * Method called just before element is displayed on the screen (loaded).
   *  If this is no overloaded then the default action is to get the data xpath
   *  and use it to bind the view element to the data it refers to.
   */
  public void onLoad (XPathContext xpathContext, Node contextNode) 
  {
    if (dataXPath == null) 
      {
	invoke(ONLOAD_ATTR);
      }
    else 
      {
	if (contextNode == null) 
	  {
	    String contextPath = getAttributeNS (DEFAULT_NAMESPACE, CONTEXT_ATTR);
	    if (contextPath != null) 
	      {
                String pkgWorkspace = getOwnerDocument ().getDocumentElement ().getAttributeNS (DEFAULT_NAMESPACE, PKG_WORKSPACE_ATTR);
                String pkgArchive = getOwnerDocument ().getDocumentElement ().getAttributeNS (DEFAULT_NAMESPACE, PKG_ARCHIVE_ATTR);
                try
                  {
		    contextNode = DocumentCache.fetchFromPackageWorkspace (contextPath, pkgWorkspace, pkgArchive);
                  }
                catch (IOException ioe)
                  {
                    Logger.log (this, ioe);
                  }
	      } 

	    if (contextNode == null)
	      {
		Logger.log (this, "onLoad: context is null");
		Logger.log (this, "onLoad: this is " + XMLWriter.getXML (this));
		return;
	      }
	  } 

	NodeList dataNodes = XPath.findNodes (xpathContext, contextNode, dataXPath);
	if (dataNodes != null) 
	  { 
	    watch (dataNodes);
	  } 
        else 
	  { 
	    Logger.log (this, "onLoad: empty nodelist for watch");
	  }
      }
  }

  public void redraw (boolean relayout) 
  {
    if (!changed && relayout)
      {
	XDocument doc = (XDocument)getOwnerDocument ();
	doc.setModified();
      }

    Component cvis = vis;
    if (cvis == null)
      {
	for (Node n = getParentNode (); n instanceof ViewElement && cvis == null; n = n.getParentNode ())
	  {
	    cvis = ((ViewElement)n).vis;
	  }
      }

    if (cvis != null)
      {
	if (relayout)
	  {
	    if (cvis.getParent () != null) 
	      {
		cvis.invalidate ();
	      }
	    Component c;
	    for (c = cvis; c.getParent () != null; c = c.getParent ())
		;
	    java.lang.Thread tid = Thread.currentThread();
	    if (tid instanceof Thread)
	      {
		((Thread)tid).redraw = c;
	      }
	    else
	      {
		c.validate();
		c.repaint();
	      }
	  }
	else
	  {
	    Graphics g = cvis.getGraphics ();
	    if (g != null)
	      {
		cvis.paint (g);
		g.dispose ();
	      }
	  }
      }
    changed |= relayout;
  }

  public void redrawNow (boolean relayout) 
  {
    if (!changed && relayout)
      {
	XDocument doc = (XDocument)getOwnerDocument ();
	doc.setModified();
      }

    Component cvis = vis;
    if (cvis == null)
      {
	for (Node n = getParentNode (); n instanceof ViewElement && cvis == null; n = n.getParentNode ())
	  {
	    cvis = ((ViewElement)n).vis;
	  }
      }

    if (cvis != null)
      {
	if (relayout)
	  {
	    if (cvis.getParent () != null) 
	      {
		cvis.invalidate ();
	      }
	    Component c;
	    for (c = cvis; c.getParent () != null; c = c.getParent ())
	      ;
	    
	    c.validate();
	    c.repaint();
	  }
	else
	  {
	    Graphics g = cvis.getGraphics ();
	    if (g != null)
	      {
		cvis.paint (g);
		g.dispose ();
	      }
	  }
      }
    changed |= relayout;
  }

  protected void viewTreeChanged () 
  {
    changed = true;
    XDocument doc = (XDocument)getOwnerDocument ();
    if (doc != null) 
      {
	doc.setModified ();
	if (vis != null && vis.getParent () != null) 
	  {
	    vis.invalidate ();
	  }
      }
  }

  public Node removeChild (Node oldChild) 
    throws DOMException 
  {
    redraw (true);

    Node nn = super.removeChild (oldChild);
    
    if ((nn != null) && (oldChild instanceof ViewElement)) 
      {
	ViewElement ve = (ViewElement)oldChild;
	Component rc = ve.getVis ();
	if (rc != null)  
	  {
	    Container pc = rc.getParent ();
	    if (pc != null) 
	      {
		Rectangle r = rc.getBounds ();
		pc.remove (rc);
		pc.repaint (r.x, r.y, r.width, r.height);
	      }
	  }
      }
    return (nn);
  }
  
  public Node replaceChild (Node newChild, Node oldChild) 
    throws DOMException 
  {
    Node n = super.replaceChild (newChild, oldChild);
    redraw (true);
    return (n);
  }

  public Node insertBefore (Node newChild, Node refChild) 
    throws DOMException 
  {
    Node n = super.insertBefore (newChild, refChild);
    redraw (true);
    return  (n);
  }
  
  public Node appendChild (Node newChild ) 
    throws DOMException 
  {
    Node n = super.appendChild (newChild);
    redraw (true);
    return (n);
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);

//     redraw (true); ?? unnecessary redraw. Commenting this does not seem to break anyng . If uncommenting, change Input.java (from xhtml) to not call super.setAttributeNS when setting a new "value" as a redraw is not required in that case.

    // Intern the name and value to make things easier both here
    // and when dispatching events to the app
    localName = localName.toLowerCase ().intern ();  // HTML attributes are case-insensitive
    value = value.intern ();

    if (localName == "width") 
      {
	if (value.equals ("")) 
	  {
	    flags &= ~(HAS_WIDTH|SLACK_WIDTH);
	    preferredWidth = 0;
	  }
	else if (value.equals ("*")) 
	  {
	    flags |= SLACK_WIDTH;
	  }
	else 
          {
	    try
	      {
		setPreferredWidth (sizeValue (value));
	      }
	    catch (NumberFormatException _)
	      {
	      }
	  }
      }
    else if (localName == "height") 
      {
	if (value.equals ("")) 
	  {
	    flags &= ~(HAS_HEIGHT|SLACK_HEIGHT);
	    preferredHeight = 0;
	  }
	else if (value.equals("*")) 
	  {
	    flags |= SLACK_HEIGHT;
	  }
	else 
          {
	    try
	      {
		setPreferredHeight (sizeValue (value));
	      }
	    catch (NumberFormatException _)
	      {
	      }
	  }
      }
    else if (localName == "font") 
      {
	fontmodifier = value;
      }
    else if (localName == ONCLICK_ATTR 
             || localName == ONMOUSEDOWN_ATTR 
             || localName == ONMOUSEOVER_ATTR 
             || localName == ONMOUSEOUT_ATTR 
             || localName == ONMOUSEUP_ATTR) 
      {
	flags |= HAS_MOUSE_LISTENER;
      }
    else if (localName == ONCHANGE_ATTR) 
      {
	flags |= HAS_ADJUSTMENT_LISTENER;
      }
    else if (localName == XPATH_ATTR) 
      {
        if (dataXPath == null)
          {
	    try 
	      { 
		dataXPath = XPathParser.singleton.parse (value);
	      } 
	    catch (Error e) 
	      {
		Logger.log (this, e.toString () + " ignore this unless you've explicitly put valid xpath in name attributes");
		dataXPath = null;
	      }
          }
      }
    else if (localName == NAME_ATTR) 
      {
	try 
	  { 
	    dataXPath = XPathParser.singleton.parse (value);
	  }
	catch (Error e) 
	  {
	    Logger.log (this, e.toString () + " ignore this unless you've explicitly put valid xpath in name attributes");
	    dataXPath = null;
	  }
      }
    super.setAttributeNS (namespaceURI, qualifiedName, value);
  }

  public String getAttributeNS (String namespaceURI, String localName) 
  {
    // HTML attributes are case-insensitive
    return super.getAttributeNS (namespaceURI, localName.toLowerCase ().intern ());
  }

  public void setData (String str) 
  {
    redraw (true);
  }

  protected void setPreferredHeight (int height) 
  {
    flags |= HAS_HEIGHT;
    preferredHeight = height;
  }

  protected void setPreferredWidth (int width) 
  {
    flags |= HAS_WIDTH;
    preferredWidth = width;
  }

  static int sizeValue (String dim) throws NumberFormatException
  {
    if (dim == null || dim.length () == 0)
      {
	return (0);
      }
    else if (dim.endsWith ("%"))
      {
	return -Integer.parseInt (dim.substring (0, dim.length () - 1));
      }
    else
      {
	return Integer.parseInt (dim);
      }
  }

  public void start () 
  {
  }

  public void stop () 
  {
  }

  public void watch (NodeList data) 
  {
  }

  protected void invoke (String attrname)
  {
    String val = getAttribute (attrname);
    if (val != null)
      {
	Thread.invoke (getCallback (attrname), this, val);
      }
  }

  public void invoke (String attrname, Object payload)
  {
    Thread.invoke (getCallback (attrname), this, payload);
  }

  public Node cloneNode (boolean deep) 
  {
    ViewElement nv = (ViewElement)super.cloneNode (deep); 
    nv.style = style;
    nv.preferredWidth = preferredWidth;
    nv.preferredHeight = preferredHeight;
    nv.flags = flags;
    nv.dataXPath = dataXPath;
    nv.changed = true;
    nv.fontmodifier = fontmodifier;
    return (nv);
  }

  /**
   * Construct an absolute URL using relativeURLString and the url the
   * owner document.
   */
  public URL getURL (String relativeURLString)
  {
    XDocument d = (XDocument) this.getOwnerDocument (); 
    try 
      {
	return (new URL (d.getURL (), relativeURLString));
      }
    catch (MalformedURLException mue) 
      {
	Logger.log (this, mue);
      }

    return (d.getURL ());
  }

  public String getURLString (String relativeURLString)
  {
    if (relativeURLString.indexOf (':') != -1)
      {
	return (relativeURLString);
      }
    else
      {
	return (getURL (relativeURLString).toString ());
      }
  }

  protected static String getLocalName (String qualifiedName)
  {
    if (qualifiedName != null)
      {
	return qualifiedName.substring (qualifiedName.indexOf (':') + 1);
      }
    else
      {
	return null;
      }
  }
  
}
