/*
 * Copyright (c) 1998-2001 Transvirtual Technologies, Inc.  All rights reserved.
 * See the file "license.terms" for details.
 *
 * $Id: DOM0Adaptor.java,v 1.2 2001/09/10 17:56:59 philippe Exp $
 */

package com.pocketlinux.xs;

import com.pocketlinux.awt.DocEmbed;
import com.pocketlinux.awt.DocForm;
import com.pocketlinux.awt.DocFrame;
import com.pocketlinux.awt.DocImage;
import com.pocketlinux.awt.DocText;
import com.pocketlinux.awt.XHTMLDocument;
import com.pocketlinux.dom.BaseElement;
import com.pocketlinux.dom.BaseNode;
import com.pocketlinux.dom.NodeCache;

import java.util.Stack;
import java.util.Vector;

import kaffe.xs.DocumentFactory;
import kaffe.xs.XsArray;
import kaffe.xs.XsObject;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class DOM0Adaptor
  extends XHTMLDocument
  implements DocumentFactory
{

  private XsObject root;

  private Stack pStack;
  private int curLevel;

  private Node curNode;
  private XHTMLXsObject curXs;

  private Vector onLoads;


  public DOM0Adaptor(){
    pStack = new Stack();
  }


  public void setXsRoot (XsObject root){
    this.root = root;
    pStack.removeAllElements ();
    pStack.push (this);
    curLevel = 0;
  }


  public XsObject getRoot (){
    return (root);
  }


  public void createElement (int level, String tag){
    insertNode(level, createElement (tag));
  }


  public void createTextElement (int level, String text){
    insertNode(level, createTextNode (text));
  }


  public void finish (){
    pStack.removeAllElements ();
    curLevel = 0;

    /* ================================================================= */

    /* Fixup the generated document to be more conformant to XHTML.
     * Perhaps this shouldn't happen here really but for the moment ....
     */
 
    //cases from slashdot.org     


    //base before html

    BaseNode rootnode = (BaseNode)getDocumentElement();
    BaseNode htmlnode = (BaseNode)rootnode.findFirstNode("html");
    Node basenode = rootnode.getFirstChild();

    if (basenode != htmlnode){
        rootnode.removeChild(basenode);

        Node headnode = htmlnode.findFirstNode("head"); 
	if(headnode == null){
	    headnode = createElement ("head");
	    Node bodynode = htmlnode.findFirstNode ("body");
            htmlnode.insertBefore(headnode, bodynode);
        }
        headnode.appendChild(basenode);
    }



    //li without ul as parent
    /* not working. wait for migration to new framework finished 
 
    potential problem: nodes in list get deleted after get list

    NodeList linodes = rootnode.findNodes("li");
  
    int len = linodes.getLength();
 
    System.out.println("got " + len + " li nodes");
    
    for(int i = 0; i < len; i++){

        System.out.println("HERE");

	BaseNode linode = (BaseNode) linodes.item(i);

	BaseNode parentnode = (BaseNode) linode.getParentNode();

	if(!parentnode.getNodeName().equals("ul")){        

        
	    BaseNode ulnode =  (BaseNode)createElementNS( XMLRoot.getDefaultNamespace(), "ul");
		        
	    //hopefully inserts at right spot
	    parentnode.insertBefore(ulnode, linode);
                               

	    NodeList linodelist = parentnode.getChildNodes();
                
                
	    int length = linodelist.getLength();

	    for(int j = 0; j < length; j++){
                
	        if(linodelist.item(j).getNodeName().equals("li")){
                         
	            ulnode.appendChild(linode);
                                
		    parentnode.removeChild(linode);
	        }
            }   
       }        
    }

    */ 





    /* ================================================================= */

    // Process the onLoad attributes (even though we're not necessarily
    // being put on the screen yet!)
    if ( onLoads != null ){
	int len = onLoads.size();
	for (int i = 0; i < len; i++){
	    ((XsObject)onLoads.elementAt(i)).invoke(null);
	}
	onLoads.removeAllElements();
    }
  }


  public int getCurrentLevel(){
    return curLevel;
  }


  private void insertNode(int level, Node elem){
    if (root == null){
	setXsRoot (new XsDocument (this));
    }

    // Adjust the level
    if (level < curLevel){
	for ( ; curLevel > level; curLevel-- ){
	    pStack.pop ();
	}
    }
    else if ( level > curLevel ){
	pStack.push (curNode);
    }
    curLevel = level;

    // Set current node to this one
    curNode = elem;
    curXs = null;

    // Append to appropriate parent
    ((Node)pStack.peek ()).appendChild (curNode);

    // Handle special DOM0 object arrays
    if (curNode instanceof DocEmbed){
	curXs = new XHTMLXsObject(elem);
 	root.putNamedArrayObject (curXs, "embeds", null);
    }
    else if (curNode instanceof DocImage){
	curXs = new XHTMLXsObject(elem);
 	root.putNamedArrayObject (curXs, "images", null);
    }
    else if (curNode instanceof DocForm){
	curXs = new XHTMLXsObject(elem);
 	root.putNamedArrayObject (curXs, "forms", null);
    }
    else if (curNode instanceof DocFrame){
	curXs = new XHTMLXsObject(elem);
    }
  }


  public void setAttribute (String key, String val){
    ((Element)curNode).setAttribute (key, val);

    if ("name".equalsIgnoreCase (key)){
	root.xsPut (val, curXs);
    }
    else if ("id".equalsIgnoreCase (key)){
	root.xsGet ("all").xsPut (val, curXs);
    }
  }


  public void setEventAttribute (String key, XsObject action){
    //this is currently meaningless in the sense that action does not override toString  
    ((Element)curNode).setAttribute (key, action.toString ());
    //was, in pocketlinux_tvb 
    //((BaseElement)curElem).setAttributeNS( null, key, "" );
    
    curXs.xsPut (key, action);
	
    if ("onLoad".equalsIgnoreCase (key)){
	if (onLoads == null){
	    onLoads = new Vector ();
	} 
	onLoads.addElement (action);
    }
  }



}









