package com.pocketlinux.xs;

import com.pocketlinux.awt.DocEmbed;
import com.pocketlinux.awt.DocForm;
import com.pocketlinux.awt.DocFrame;
import com.pocketlinux.awt.DocImage;
import com.pocketlinux.awt.ViewElement;
import com.pocketlinux.dom.BaseDocument;
import com.pocketlinux.dom.Callback;
import com.pocketlinux.dom.DocumentCache;

import java.awt.Component;
import java.awt.Image;
import java.awt.image.ImageObserver;
import java.awt.Toolkit;

import kaffe.xs.XsFunc;
import kaffe.xs.XsObject;
import kaffe.xs.XsNumber;
import kaffe.xs.XsString;

import org.w3c.dom.Node;

public class XHTMLXsObject
  extends XsObject
{
  private ViewElement delegate;

  public XHTMLXsObject(Node node)
  {
    delegate = (ViewElement)node;
  }

  public XsObject xsGet (String key)
  {
    if (delegate instanceof DocImage)
      {
	if ("width".equalsIgnoreCase (key) || "height".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (Integer.parseInt (delegate.getAttribute (key))));
	  }
	else if ("src".equalsIgnoreCase (key))
	  {
	    return (new XsString (delegate.getAttribute ("src")));
	  }
	else if ("complete".equalsIgnoreCase (key))
	  {
	    Image img = ((DocImage)delegate).getImage ();
	    if (img == null || (Toolkit.getDefaultToolkit ().checkImage (img, 0, 0, null) & ImageObserver.ALLBITS) == 0)
	      {
		return (xFalse);
	      }
	    else
	      {
		return (xTrue);
	      }
	  }
      }
    else if (delegate instanceof DocForm)
      {
      }
    else if (delegate instanceof DocEmbed)
      {
      }
    else if (delegate instanceof DocFrame)
      {
	if ("innerWidth".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (delegate.getVis ().getSize ().width));
	  }
	else if ("innerHeight".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (delegate.getVis ().getSize ().height));
	  }
	else if ("outerWidth".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (getTopVis (delegate.getVis ()).getSize ().height));
	  }
	else if ("outerHeight".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (getTopVis (delegate.getVis ()).getSize ().height));
	  }
	else if ("screenX".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (getTopVis (delegate.getVis ()).getLocation ().x));
	  }
	else if ("screenY".equalsIgnoreCase (key))
	  {
	    return (new XsNumber (getTopVis (delegate.getVis ()).getLocation ().y));
	  }
	else if ("width".equalsIgnoreCase (key))
	  {
	  }
	else if ("height".equalsIgnoreCase (key))
	  {
	  }
      }
    return (super.xsGet (key));
  }

  public XsObject xsPut (String key, XsObject val)
  {
    if (delegate instanceof DocImage)
      {
	if ("src".equalsIgnoreCase (key))
	  {
	    delegate.setAttribute("src", val.toString ());
	    return (val);
	  }
      }
    else if (delegate instanceof DocForm)
      {
      }
    else if (delegate instanceof DocEmbed)
      {
      }
    else if (delegate instanceof DocFrame)
      {
	if ("doc_open".equalsIgnoreCase (key))
	  {
	    DOM0Adaptor doc = (DOM0Adaptor)DocumentCache.fetchDocument (((DocForm)delegate).getURL (val.toString ()));
	    ((DocFrame)delegate).setDocument (doc);
	    top.xsGet ("update_browser").invoke (this);
	    return (doc.getRoot ());
	  }
      }

    if (registerEvents (key, val))
      {
	// We store the functions in the usual hash so we can
	// get them back easily
	super.xsPut (key, val);
	return (val);
      }

    return (super.xsPut (key, val));
  }

  private boolean registerEvents (String key, XsObject val)
  {
    if ("onclick".equalsIgnoreCase (key) ||
        "onmousedown".equalsIgnoreCase (key) ||
        "onmouseover".equalsIgnoreCase (key) ||
        "onmouseout".equalsIgnoreCase (key) ||
        "onmouseup".equalsIgnoreCase (key) ||
        "onchange".equalsIgnoreCase (key) ||
        "onload".equalsIgnoreCase (key))
      {
        final XsFunc func = (XsFunc)val;
        delegate.setCallback (key, new Callback()
	  {
	    public void invoke (Node node, Object payload)
	    {
	      func.invoke (XHTMLXsObject.this);
	    }
	  });
        // NB - need to set mouse listeners etc.
	return (true);
      }
    else
      {
	return (false);
      }
  }

  private Component getTopVis(Component c)
  {
    while (c.getParent() != null)
      {
	c = c.getParent();
      }
    return (c);
  }

}
