package org.xoe.display.dom.xhtml;

import java.io.IOException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.dom.XLink;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.Viewable;
import org.xoe.display.dom.ViewElement;
import org.xoe.services.IURLViewer;


public class Notes
  extends ViewElement 
{  
  private static final SimpleServiceQuery query = new SimpleServiceQuery (IURLViewer.INTERFACE);
  
  public void watch (NodeList notes) 
  {
    
    DOMUtils.removeChildNodes (this);
    int length = notes.getLength ();
    Document doc = getOwnerDocument ();
    if (length > 0)
      {
	Element table = doc.createElement ("table");
	table.setAttribute("width", "100%");
	String style = getAttributeNS (DEFAULT_NAMESPACE, "style");
	Element e;
	IURLViewer urlViewer = null;
	for (int i = 0 ; i < length; i++)
	  {
	    try 
	      {
		e = (Element) notes.item (i);
		String show = XLink.getShow (e);
		String href = XLink.getHref (e);
		Element tr = doc.createElement ("tr");
		Element td = doc.createElement ("td");
		if (style != null)
		  {
		    td.setAttributeNS (DEFAULT_NAMESPACE, "style", style);
		  }
		tr.appendChild (td);

		if (show == "embed")
		  {
		    Element iframe = doc.createElement ("iframe");
		    iframe.setAttribute ("width", "100%");
		    
                    // is there a url viewer service? 
		    if (urlViewer == null)
		      {
                        // get the url viewer service
			urlViewer = (IURLViewer) ServiceLocator.findService (query);
              
                        // if a viewer service is found, get the viewable document
                        // and display it
			if (urlViewer != null)
			  {
			    try 
			      {
				((DocumentViewer) iframe).setDocument (urlViewer.getPage (href));
				td.appendChild (iframe);
			      }
			    catch (IOException io)
			      {
				Logger.log (this, io);
			      }
                          }
                      }
		  }
		else if (show == "replace" || show == "new") // "new" should really pop up a new browser window when activated
		  {
		    Element a = getOwnerDocument ().createElement ("a");
		    a.setAttribute ("href", XLink.getHref(e));
		    a.appendChild (getOwnerDocument ().createTextNode (XLink.getHref(e)));
		    td.appendChild (a);
		  }
		else    // other
		  {
		  }
		table.appendChild (tr);
	      }
	    catch (ClassCastException cce)
	      {
		Logger.log (this, "Node " + notes.item (i) + " in watched nodelist is not an Element");
	      }
	  }
	appendChild (table);
	redraw (true);
      }
  }

}
  
