package org.xoe.display.dom.xhtml;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.util.Logger;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.xpath.XPath;
import org.xoe.display.dom.LayoutCursor;
import org.xoe.display.dom.Viewable;
import org.xoe.display.dom.ViewElement;


public class TextElement
  extends ViewElement 
  implements FormElement
{  
  private Text value;;

  /** 
   * Treat "value" like a child for layout purposes.
   */

  protected void layoutChildren (LayoutCursor cur) 
  {
    if (value != null) 
      { 
	((Viewable) value).layout (cur);
      }
    super.layoutChildren(cur);
  }

  public void setAttributeNS (String namespaceURI, String qualifiedName, String value) 
  {
    String localName = getLocalName (qualifiedName);
    super.setAttributeNS(namespaceURI, qualifiedName, value);
    if (localName.equals (VALUE_ATTR))
      {
	if (this.value == null)
	  {
	    this.value = getOwnerDocument ().createTextNode(value);
	  }
	else 
	  {
	    this.value.setNodeValue(value);
	  }
	redraw(true);
      }
  }


  public void watch (NodeList arg) 
  {
    super.watch (arg);

    Node data;

    if (arg instanceof Node)
      {
	data = (Node)arg;
      }
    else      
      {
	if (arg.getLength () > 0)
	  {
	    data = arg.item (0);
	  }
	else
	  {
	    Logger.log (this, "arg is a NodeList of length: " + arg.getLength ());
	    Logger.log (this, "local name: " + this.getLocalName ());
	    return;
	  }
      }
    
    // If we have a mapping, use it to get the data, otherwise leave
    // it with what the view gave it.
    if (data != null) 
      {
        String banner = getAttributeNS (DEFAULT_NAMESPACE, "banner");

        String value = null;
        if (banner != null)
          {
            value = DOMUtils.getBanner (data, banner);
          }
        else
          {
	    value = DOMUtils.getTextContent (data);
          }

        if (value != null)
          {
            DOMUtils.setTextContent (this, value);
          }
        else
          {
            Logger.log (this, "watch: value is null for data = " + data);
          }
      }

    redraw (true);
  }
  
  public String getFormValue() 
  {
    if (value != null)
      {
	return value.getNodeValue();
      }
    return "";
  }


  public String getText()
  {
    return null;
  }
  
  public XPath getXPath()
  {
    return null;
  }

  public String getBanner ()
  {
    return getNodeName ();
  }

  public void resetValue ()
  {
  }

}
