/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: XoeServlet.java,v 1.48 2001/12/10 22:51:21 suraj Exp $
 */
package org.xoe.servlet;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.net.URLDecoder;

import javax.servlet.GenericServlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xoe.core.Constants;
import org.xoe.core.content.DocumentCache;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.Callback;
import org.xoe.core.dom.DOMlet;
import org.xoe.core.dom.XDocument;
import org.xoe.core.packages.XoePackage;
import org.xoe.core.services.IService;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.services.SimpleServiceQuery;
import org.xoe.core.util.DOMUtils;
import org.xoe.core.util.Logger;
import org.xoe.core.util.XMLWriter;
import org.xoe.core.xpath.XPath;
import org.xoe.display.dom.DocumentViewer;
import org.xoe.display.dom.ViewElement;
import org.xoe.display.dom.xhtml.XHTMLDocument;
import org.xoe.services.IApplication;
import org.xoe.services.IClipboard;
import org.xoe.services.IHistory;


public abstract class XoeServlet
  extends HttpServlet
  implements IApplication
{
  // reserved path strings.
  private static final String XOE_PREFIX = "xoe.";
  private static final String XOE_BACK = XOE_PREFIX + "Back";

  private static final String XOE_FORWARD = XOE_PREFIX + "Forward";
  private static final String XOE_CUT = XOE_PREFIX + "Cut";
  private static final String XOE_COPY = XOE_PREFIX + "Copy";
  private static final String XOE_PASTE = XOE_PREFIX + "Paste";
  private static final String XOE_SELECT_ALL = XOE_PREFIX + "SelectAll";

  protected static final SimpleServiceQuery clipboardServiceQuery = new SimpleServiceQuery (IClipboard.INTERFACE);
  protected static final SimpleServiceQuery historyServiceQuery = new SimpleServiceQuery (IHistory.INTERFACE);

  // user config 
  protected static final String USER_CONFIG_FILENAME = "config.xml";

  /**
   *  xml document containing user settings
   */
  protected XDocument userConfig;

  /**
   * Propagate "history back" events to the enclosing document. Should
   * be set by the calling servlet.
   */
  protected boolean propagateHistory = false;

  // this is the real servlet.
  private GenericServlet servlet;

  // the document which provides the structure for the servlet screen
  private XDocument viewdoc;

  // the frame (in the viewdoc) where documents are set so they are visible
  private DocumentViewer viewer;

  // title holder
  private DocumentViewer title;

  // response information goes here for later handling after the
  // callback returns.
  private XoeServletResponse response = new XoeServletResponse ();

  // request used for various purposes.
  private XoeServletRequest request = new XoeServletRequest ();

  // are we processing a callback?
  private boolean running;

  // the history service object used to manage this servlet's history.
  private IHistory history;

  // the url of the installation directory of this package
  private String installedLocationURI;


  /**
   * Default callback
   */
  protected Callback defaultCallback = new Callback ()
  {
    public void invoke (Node node, Object payload)
    {
      if (running == true)
	{
	  return;
	}
      running = true;

      try
	{
	  Logger.log (this, " payload = [" + payload + "]");

	  String meth = ((Element)node).getAttribute ("method");
	  if (meth == null)
	    {
	      meth = "GET";
	    }
	  else 
	    {
	      meth = meth.toUpperCase ();
	    }
	  request.reset (meth, (String)payload);
	  request.setAttribute (XoeServletRequest.NODE, node);
	  long sservice = System.currentTimeMillis ();
	  XDocument d = service (request, response);
	  long eservice = System.currentTimeMillis ();
          if (d != null)
            {
	      setDisplayedDocument (d);
            }

          Logger.log (this, "service  = " + (eservice-sservice) + "ms");
	}
      finally
	{
	  running = false;
	}
    }
  };


  // ----------------------------------------------------------------------
  // constructors

  public XoeServlet () 
  {
     this (null);
  }

  /**
   * Wrap a Xoe Servlet round a standard Java servlet. I am a servlet
   * myself so wrap myself in myself. This allows Xoe apps to extends
   * XoeServlet.
   */
  public XoeServlet (GenericServlet servlet) 
  {
    if (servlet == null)
      {
	servlet = this;
      }
    this.servlet = servlet;
  }

  public XoeServlet (XoeServlet from)
  {
    redirect (from);
    this.servlet = this;
  }
  

  // ----------------------------------------------------------------------
  // org.xoe.services.IApplication

  public void init (Document configDoc) 
    throws Exception
  {
    XoeServletConfig config = new XoeServletConfig (configDoc, this);
    String name = config.getServletName ();
    if (name == null)
      {
	name = this.servlet.toString ();
      }

    config.context.putServlet (name, this.servlet);
    super.init (config);
    init (config);
  }

  public final String[] getFunctionalities ()
  {
    String [] funcs = copyInto (getServletFunctionalities (), 1);
    funcs[funcs.length - 1] = IApplication.FUNCTIONALITY;
    return (funcs);
  }

  public final String[] getInterfaces (String func)
  {
    if (IApplication.FUNCTIONALITY.equals (func))
      {
        return (new String [] { IApplication.INTERFACE });
      }

    return (copyInto (getServletInterfaces (func)));
  }
 
  public final String[] getFeatures (String func)
  {
    if (IApplication.FUNCTIONALITY.equals (func))
      {
	XoeServletConfig c = (XoeServletConfig) getServletConfig ();
        return (new String [] { IApplication.FEATURE_CONFIG, c==null ? null : c.getURI () });
      }

    return (copyInto (getServletFeatures (func)));
  }
 
  public String[] getInformation ()
  {
    return (null);
  }

  /**
   * Get the initial view on this component.
   */
  public XDocument getViewableDocument ()
  {
    synchronized (this)
      {
	if (getServletConfig () == null)
	  {
	    try 
	      {
		init ((Document) null);
	      }
	    catch (Exception e) 
	      {
		Logger.log (this, "Exception thrown when trying to init wrapped Servlet " + servlet + " before getViewableDocument ()");
		Logger.log (this, e);
	    }
	  }
    
	if (viewdoc == null)
	  {
            viewdoc = new XHTMLDocument ();
	    createCallbacks (viewdoc);
            viewdoc.setCallbackPropagate (false);

	    Element t = viewdoc.createElement ("table");
	    t.setAttribute ("cellspacing", "0");
	    t.setAttribute ("cellpadding", "0");
	    t.setAttribute ("width", "100%");
	    t.setAttribute ("height", "100%");
	    viewdoc.appendChild (t);

	    Element tr = viewdoc.createElement ("tr");
	    t.appendChild (tr);
	    Element td = viewdoc.createElement ("td");
	    td.setAttribute ("valign", "top");
	    td.setAttribute ("width", "100%");
	    tr.appendChild (td);

	    Element i = viewdoc.createElement ("iframe");
	    i.setAttribute ("width", "");
	    i.setAttribute ("height", "");
	    i.setAttribute ("id", "title");
	    td.appendChild (i);
	    title = (DocumentViewer) i;

	    tr = viewdoc.createElement ("tr");
	    tr.setAttribute ("height", "*");
	    t.appendChild (tr);
	    td = viewdoc.createElement ("td");
	    td.setAttribute ("valign", "top");
	    td.setAttribute ("width", "100%");
	    tr.appendChild (td);

	    i = viewdoc.createElement ("iframe");
	    i.setAttribute ("width", "100%");
	    i.setAttribute ("height", "100%");
	    td.appendChild (i);
	    viewer = (DocumentViewer)i;
	    if (viewer != null && history != null)
	      {
		history.trackHistory (viewer, propagateHistory);
	      }

	    request.reset ("GET", "");
	    XDocument d = service (request, response);
	    setDisplayedDocument (d);
	  }
      }
    return (viewdoc);
  }


  // ----------------------------------------------------------------------
  // javax.servlet.http.HttpServlet;

  public void init (ServletConfig config) 
    throws ServletException
  {
    if (servlet != this)
      {
	this.servlet.init (config);
      }
  }

  /**
   *  the default behaviour of XoeServlet is to render the file default.xhtml
   */
  protected void doGet (HttpServletRequest req, HttpServletResponse res)
    throws ServletException, IOException
  {
    XDocument doc;
     
    String path =req.getPathInfo ();

    if ("".equals (path))
	{
	  doc = getRelativeDocument (DEFAULT_VIEW);
	  if (doc == null)
	    { 
	      throw new ServletException (DEFAULT_VIEW + " not found");
	    }
	  DOMUtils.expandUsingXPathAttributes (doc, null, null);
	}
    else 
      {
	doc = getRelativeDocument (path);
	
	if (path.startsWith (XOE_PREFIX))
	  {
	    path = path.intern ();
	    if ((handleHistory (path) || handleClipboard (path)) && doc != null)
	      {
		throw new Error (path + " is a reserved XOE URL but exists in the resources of " + this);
	      }
	  }
	else if (doc != null)
	  {
	    res.setContentType (MIMETYPE_TEXT_HTML);
	    res.getWriter ().print (doc);
	  }
      }
  }

  /**
   * Dispatch the request to the servlet and process the reply.
   */
  private XDocument service (XoeServletRequest req, XoeServletResponse rep)
  {
    XDocument ndoc = null;
    try
      {
	req.setAttribute (XoeServletRequest.DOCUMENT, getDisplayedDocument ());
	for (;;)
	  {
	    rep.reset ();
	    long sservice = System.currentTimeMillis ();
	    servlet.service (req, rep);
	    long eservice = System.currentTimeMillis ();
	    Logger.log (this, "doGet   = " + (eservice-sservice) + "ms");
	    if (rep.redirect == null)
	      {
	        break;
	      }
	    req.reset ("GET", rep.redirect);
	  }
        Object data = rep.getOutputData ();
        if (data instanceof XDocument)
          {
	    ndoc = (XDocument)data;
          }
        else if (data instanceof ByteArrayOutputStream)
          {
	    ndoc = DOMlet.create (((ByteArrayOutputStream)data).toString (rep.getCharacterEncoding ()).toCharArray (), "unknown");
          }
      }
    catch (Throwable e)
      {
	StringWriter err = new StringWriter ();
	e.printStackTrace (new PrintWriter (err));
	servlet.log (err.toString ());
      }
    return (ndoc);
  }


  // ----------------------------------------------------------------------
  // org.xoe.servlet.XoeServlet

  public void redirect (XoeServlet from)
  {
    if (from != null) 
      { 
	this.viewdoc = from.viewdoc;
	this.viewer = from.viewer;
	this.title = from.title;
	this.response = from.response;
	this.request = from.request;
      }
  }

  protected String getDataDir ()
  {
    return (Stash.getPackageWorkspaceURI (USER_NAME, getPackageName ()));
  }

  protected XDocument getFileFromPackageWorkspace (String filename)
    throws IOException
  {
    return (DocumentCache.fetchFromPackageWorkspace (filename, getWorkspaceURI (), getArchiveURI ()));
  } 

  protected XDocument getUserConfig ()
    throws ServletException, IOException
  {
    if (userConfig == null)
      {
	userConfig = getFileFromPackageWorkspace (USER_CONFIG_FILENAME);
        if (userConfig == null)
          {
	    throw new ServletException ("no config.xml found");
          }
      }
    return (userConfig);
  }

  protected String getWorkspaceURI ()
  {
    return (Stash.getPackageWorkspaceURI (USER_NAME, getPackageName ()));
  }

  protected String getArchiveURI ()
  {
    return (Stash.getPackageArchiveURI (getPackageURI ()));
  }

  public String getPackageName ()
  {
    XoePackage pkg = XoePackage.getPackage (this.getClass ());
    if (pkg != null)
      {
	return (pkg.getName ());
      }
    return ("xoe-servlet");
  }

  public String getPackageURI ()
  {
    XoePackage pkg = XoePackage.getPackage (this.getClass ());
    if (pkg != null)
      {
	return (pkg.getPackageURI ());
      }
    return ("xoe-servlet");
  }

  /**
   * Put the document onto the screen.
   */
  protected void setDisplayedDocument (Document doc)
  {
    if (viewer != null && doc != null)
      {
	viewer.setDocument (doc);
	if (history != null)
	  {
	    history.forward (viewer, doc, null);
	  }
      }
    else
      {
	Logger.log (this, ( "WARNING: ignoring setDisplayedDocument call.\n"
			    + "viewer == " + viewer + " doc == " + doc));
      }
  }

  /**
   * Set the callbacks on the document so we get any interesting events.
   */
  protected void createCallbacks (XDocument doc)
  {
    doc.setCallback ("onload", defaultCallback);
    doc.setCallback ("onclick", defaultCallback);
    doc.setCallback ("onmouseover", defaultCallback);
    doc.setCallback ("onmouseout", defaultCallback);
    doc.setCallback ("onmousedown", defaultCallback);
    doc.setCallback ("onmouseup", defaultCallback);
    doc.setCallback ("onchange", defaultCallback);
    doc.setCallback ("href", defaultCallback);
    doc.setCallback ("action", defaultCallback);
  }


  protected void createCallbacks ()
  {
    createCallbacks (viewdoc);
  }

  /**
   * Populate the fields of the returned document with the name/value
   * pairs of the parameters passed with the request.
   */
  public static void populateFields (XDocument doc, Hashtable params)
  {
    Enumeration e = params.keys ();
    if (e != null)
      {
	while (e.hasMoreElements ())
	  {
	    String name = (String)e.nextElement ();
	    String value = (String) params.get (name);
	    NodeList list = XPath.findNodes (doc, "//*[@name='"+name+"']");
	    for (int i = list.getLength () - 1; i >= 0; i--)
	      {
		((Element)list.item (i)).setAttribute (VALUE_ATTR, value);
	      }
	  }
      }
  }

  public void populateFields (Hashtable params)
  {
    populateFields ((XDocument) getDisplayedDocument (), params);
  }

  /**
   * Get the document viewer
   **/
  protected DocumentViewer getDocumentViewer ()
  {
    return viewer;
  }

  /**
   * Get the document currently on the screen.
   */
  protected Document getDisplayedDocument ()
  {
    if (viewer != null)
      {
	return (viewer.getDocument ());
      }
    return (null);
  }

  /**
   * Set the title document
   */
  protected void setTitleDocument (Document doc)
  {
    if (title != null)
      {
	title.setDocument (doc);
      }
  }

  /**
   * Get the title document
   */
  protected Document getTitleDocument ()
  {
    if (title != null)
      {
	return (title.getDocument ());
      }
    return (null);
  }


  public String getRelativeSrc (String name) 
  {
    String installedLocationURI = ((XoeServletConfig) getServletConfig ()).getInstalledLocationURI ();
    if (installedLocationURI == null)
      {
        installedLocationURI = Stash.getWorkingDirectory (this).toString ();
      }
    if (!installedLocationURI.endsWith (File.separator))
      {
        installedLocationURI += File.separator;
      } 
    return (installedLocationURI + name);
  }
  
  public XDocument getRelativeDocument (String name) 
  {
    XDocument doc = DocumentCache.fetchDocument (getRelativeSrc (name));

    if (doc != null && doc.getDocumentElement () != null)
      {
	// add the package workspace attribute
	doc.getDocumentElement ().setAttributeNS (DEFAULT_NAMESPACE, 
						  PKG_WORKSPACE_ATTR,
						  getWorkspaceURI ());

	// add the package archive attribute
	doc.getDocumentElement ().setAttributeNS (DEFAULT_NAMESPACE, 
						  PKG_ARCHIVE_ATTR,
						  getArchiveURI ());
      }

    return (doc);
  }

  protected boolean handleClipboard (String cmd)
  {
    cmd = cmd.intern ();
    if (cmd == XOE_CUT)
      {
	((IClipboard)ServiceLocator.findService (clipboardServiceQuery)).cut ();
        return (true);
      }
    else if (cmd == XOE_COPY)
      {
	((IClipboard)ServiceLocator.findService (clipboardServiceQuery)).copy ();
        return (true);
      }
    else if (cmd == XOE_PASTE)
      {
	((IClipboard)ServiceLocator.findService (clipboardServiceQuery)).paste ();
        return (true);
      }
    else if (cmd == XOE_SELECT_ALL)
      {
	((IClipboard)ServiceLocator.findService (clipboardServiceQuery)).selectAll ();
        return (true);
      }
    
    return (false);
  }

  protected boolean handleHistory (String cmd)
  {
    if (XOE_BACK == cmd || XOE_FORWARD == cmd)
      {
	if (history == null)
	  {
	    throw new Error ("trying to go back using " + cmd + " with no IHistoryService object set.");
	  }
	if (XOE_BACK == cmd)
	  {
	    history.back (viewer);
	    return true;
	  }
	else 
	  {
	    history.forward (viewer);
	    return true;
	  }
      }
    return false;
  }


  public String getBackURLString () 
  {
    return XOE_BACK;
  }

  public void trackHistory ()
  {
    this.history = (IHistory) ServiceLocator.findService (historyServiceQuery);
  }

  public void setPropagateHistory (boolean propagateHistory)
  {
    this.propagateHistory = propagateHistory;
  }

  public boolean doesPropagateHistory()
  {
    return propagateHistory;
  }

  private String [] copyInto (String [] in)
  {
    return (copyInto (in, 0));
  }

  private String [] copyInto (String [] in, int extra)
  {
    int len = 0;

    if (in != null)
      {
	len = in.length;
      }

    // build the out array "extra" larger than the in array
    String [] out = new String [len + extra];

    if (len > 0)
      {
	System.arraycopy (in, 0, out, 0, len);
      }

    return (out);
  }

  protected String [] getServletFunctionalities ()
  {
    return (null);
  }

  protected String [] getServletInterfaces (String func)
  {
    return (null);
  }

  protected String [] getServletFeatures (String func)
  {
    return (null);
  }


  // ----------------------------------------------------------------------
  // statics 

  public static String getURLParameter (String url, String pname)
  {
    int idx = url.indexOf ('?');
    if (idx == -1)
      {
	return null;
      }

    String query = url.substring (idx+1);
    
    for (StringTokenizer tok = new StringTokenizer (query, "&"); tok.hasMoreTokens (); )
      {
	try
	  {
	    String name;
	    String value;
	    String token = tok.nextToken ();
	    idx = token.indexOf ('=');

	    if (idx == -1)
	      {
		name = token;
		value = "";
	      }
	    else
	      {
		name = token.substring (0, idx);
		value = token.substring (idx+1);
	      }

	    name = URLDecoder.decode (name);
	    if (name.equals (pname))
	      {
		return URLDecoder.decode (value);
	      }
	  }
	catch (Exception _)
	  {
	    Logger.log (XoeServlet.class, _);
	  }
      }
    return null;
  }
}
