/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: XoeServletConfig.java,v 1.9 2001/11/10 00:59:42 suraj Exp $
 */
package org.xoe.servlet;

import java.io.File;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

import org.xoe.core.Constants;
import org.xoe.core.content.Stash;
import org.xoe.core.dom.XDocument;
import org.xoe.core.xpath.XPath;
import org.xoe.core.xpath.XPathParser;


public class XoeServletConfig
  implements ServletConfig
{

  private static final String PARAM_NAME1  = "xoe:xoe-config/xoe:init-param[xoe:param-name='";
  private static final String PARAM_NAME2  = "']/xoe:param-value/text()";
  private static final XPath PARAM_NAME_XPATH = XPathParser.singleton.parse ("xoe:xoe-config/xoe:init-param/xoe:param-name/text()");
  private static final XPath NAME_XPATH = XPathParser.singleton.parse ("//xoe:servlet/xoe:servlet-name/*");

  /**
   * Servlet context is a singleton.
   */
  static XoeServletContext context = new XoeServletContext ();

  /**
   * Configuration document.
   */
  private Document config;

  private String installedLocationURI;
  private String xoeConfigURI;

  public XoeServletConfig (Document configDoc, XoeServlet s)
  {
    this.config = configDoc;
    if (config == null)
      {
        config = new XDocument ();
        config.appendChild (config.createElementNS (Constants.DEFAULT_NAMESPACE, "xoe-config"));
      }
    // ** HACK -- 
    // check to see if we were passed a XDocument.
    // if we were, then get it's URL and use that to
    // determine the working directory for this service.
    if (config instanceof XDocument)
      {
	// get the working dir from the location of the config doc
        URL xoeConfigURL = ((XDocument)config).getURL ();
        if (xoeConfigURL != null)
          {
            xoeConfigURI = xoeConfigURL.toString ();
            installedLocationURI = xoeConfigURI.substring (0, xoeConfigURI.lastIndexOf (File.separator));
          }
      }
    // if the working directory hasn't been set yet for this service, 
    // get it from the xoe pkg.
    if (installedLocationURI == null)
      {
        installedLocationURI = Stash.getWorkingDirectory (s).toString ();
      }
  }

  public ServletContext getServletContext ()
  {
    return (context);
  }

  public String getInitParameter (String name)
  {
    if (config != null)
      {
        return (XPath.findString (config, PARAM_NAME1 + name + PARAM_NAME2));
      }
    return (null);
  }

  public Enumeration getInitParameterNames ()
  {
    if (config == null)
      {
	return (null);
      }
    else
      {
        NodeList list = XPath.findNodes (config, PARAM_NAME_XPATH);
        Vector vec = new Vector ();
        for (int i = list.getLength () - 1; i >= 0; i--)
          {
	    vec.addElement (list.item (i).getNodeValue ());
          }
        return (vec.elements ());
      }
  }

  // -------------------------------------------------------------------

  String getServletName ()
  {
    if (config == null)
      {
	return (null);
      }
    else
      {
        return (XPath.findString (config, NAME_XPATH));
      }
  }

  String getInstalledLocationURI ()
  {
    return installedLocationURI;
  }

  String getURI ()
  {
    return xoeConfigURI;
  }
}
