/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: XoeServletRequest.java,v 1.4 2001/11/11 01:10:04 suraj Exp $
 */
package org.xoe.servlet;

import java.io.BufferedReader;
import java.io.IOException;
import java.net.URLDecoder;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import javax.servlet.ServletInputStream;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.w3c.dom.Node;

import org.xoe.core.util.Logger;


public class XoeServletRequest
  implements HttpServletRequest
{

  public static final String DOCUMENT = "org.w3c.dom.Document";
  public static final String NODE = "org.w3c.dom.Node";
  public static final String PARAMS = "org.xoe.servlet.XoeServletRequest.Params";
  
  private XoeSession session;
  private String method;
  private String path;
  private String query;
  private Hashtable params;
  private Object document;
  private Object node;

  XoeServletRequest ()
  {
  }

  XoeServletRequest (String method, String value)
  {
    reset (method, value);
  }

  void reset (String method, String value)
  {
    params = null;
    this.method = method;
    int idx = value.indexOf ('?');
    if (idx == -1)
      {
	path = value;
      }
    else
      {
	path = value.substring (0, idx);
	query = value.substring (idx+1);

	// If this query is in fact a set of parameters, parse them and stash
	// in a hashtable for later
	if (query.indexOf ('=') != -1)
	  {
	    params = new Hashtable ();
	    for (StringTokenizer tok = new StringTokenizer (query, "&"); tok.hasMoreTokens (); )
	      {
	        try
	          {
		    String n;
		    String v;
		    String val = tok.nextToken ();
		    idx = val.indexOf ('=');
		    if (idx == -1)
		      {
			n = val;
			v = "";
		      }
		    else
		      {
			n = val.substring (0, idx);
			v = val.substring (idx+1);
		      }
		    n = URLDecoder.decode (n);
		    v =URLDecoder.decode (v); 
		    Object o = params.get (n);
		    if (o == null)
		      {
			params.put (n, v);
		      }
		    else if (o instanceof String)
		      {
			Vector values = new Vector();
			values.addElement (o);
			values.addElement (v);
			params.put (n,values);
		      }
		    else if (o instanceof Vector)
		      {
			((Vector) o).addElement (v);
		      }
	          }
	        catch (Exception _)
	          {
		    Logger.log (this, _);
	          }
	      }
	  }
      }
  }

  // ----------------------------------------------------
  // HttpServletRequest

  public int getContentLength ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getContentType ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getProtocol ()
  {
    return ("http/1.1");
  }

  public String getScheme ()
  {
    return ("http");
  }

  public String getServerName ()
  {
    return ("localhost");
  }

  public int getServerPort ()
  {
    return (-1);
  }

  public String getRemoteAddr ()
  {
    return ("127.0.0.1");
  }

  public String getRemoteHost ()
  {
    return ("localhost");
  }

  public String getRealPath (String path)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public ServletInputStream getInputStream ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getParameter (String name)
  {
    if (params != null)
      {
	Object o = params.get (name);
	if (o instanceof String)
	  {
	    return (String) o;
	  }
	else if (o instanceof Vector)
	  {
	    return (String) ((Vector) o).firstElement ();
	  }
	else if (o instanceof String [])
	  {
	    return ((String []) o)[0];
	  }
      }
    return (null);
  }

  public String[] getParameterValues (String name)
  {

    if (params != null)
      {
	Object o = params.get (name);
	if (o instanceof String)
	  {
	    return (new String[]{ (String) o });
	  }
	else if (o instanceof Vector)
	  {
	    Vector v = (Vector) o;
	    String [] ret = new String [v.size ()];
	    v.copyInto (ret);
	    return ret;
	  }
	else if (o instanceof String [])
	  {
	    return ((String []) o);
	  }
      }
    return (null);
  }

  public Enumeration getParameterNames ()
  {
    if (params == null)
      {
	params = new Hashtable ();
      }
    return (params.keys ());
  }

  public void setAttribute (String name, Object value)
  {
    if (DOCUMENT.equals (name))
      {
	document = value;
      }
    else if (NODE.equals (name))
      {
	node = value;
      }
  }

  public Object getAttribute (String name)
  {
    if (DOCUMENT.equals (name))
      {
	return (document);
      }
    else if (NODE.equals (name))
      {
	return node;
      }
    else if (PARAMS.equals (name))
      {
	Logger.log (this, new Throwable ());
	Logger.log (this, "The structure of this hashtable has changed to accomodate multiple values for fields. So casting a result of get() to java.lang.String might not always work. Look at " + this.getClass ().getName () + " for details.");
	return params;
      }
    else
      {
	return (null);
      }
  }

  public BufferedReader getReader () throws IOException
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getCharacterEncoding ()
  {
    return ("UTF8");
  }

  public Cookie[] getCookies ()
  {
    return (null);
  }

  public String getMethod ()
  {
    return (method);
  }

  public String getRequestURI()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getServletPath ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getPathInfo ()
  {
    return (path);
  }

  public String getPathTranslated ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String getQueryString ()
  {
    return (query);
  }

  public String getRemoteUser ()
  {
    return (null);
  }

  public String getAuthType ()
  {
    return (null);
  }

  public String getHeader (String name)
  {
    return (null);
  }

  public int getIntHeader (String name)
  {
    return (-1);
  }

  public long getDateHeader (String name)
  {
    return (-1);
  }

  public Enumeration getHeaderNames ()
  {
    return new Enumeration () 
      {
	public boolean hasMoreElements () { return false; }
	public Object nextElement () { throw new NoSuchElementException (); }
      };
  }

  public HttpSession getSession (boolean create)
  {
    synchronized (this)
      {
	if (session == null && create)
	  {
	    session = new XoeSession ();
	  }
      }
    return (session);
  }

  public String getRequestedSessionId ()
  {
    return (getSession (true).getId ());
  }

  public boolean isRequestedSessionIdValid ()
  {
    return (session != null);
  }

  public boolean isRequestedSessionIdFromCookie ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public boolean isRequestedSessionIdFromUrl ()
  {
    throw new kaffe.util.NotImplemented ();
  }

}
