/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.
 * All rights reserved.  See the file "license.terms" for details.
 *
 * $tvt: XoeServletResponse.java,v 1.2 2001/10/10 18:13:50 samc Exp $
 */
package org.xoe.servlet;

import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;
import java.io.OutputStreamWriter;
import java.util.Date;
import java.util.Locale;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletOutputStream;

public class XoeServletResponse
  implements HttpServletResponse
{
  private static final String CONTENT_TYPE = "content-type";
  private static final String CONTENT_ENCODING = "charset=";
  private static final String DEFAULT_CONTENT_TYPE = "text/html";
  private static final String DEFAULT_CONTENT_ENCODING = "UTF8";

  private XoeServletOutputStream output;
  private XoePrintWriter writer;
  private String contentType = DEFAULT_CONTENT_TYPE;
  private String contentEncoding = DEFAULT_CONTENT_ENCODING;
  private int status;
  String redirect;

  public void setContentType (String ctype)
  {
    setHeader (CONTENT_TYPE, ctype);
  }

  public void setContentLength (int len)
  {
  }

  public ServletOutputStream getOutputStream () throws IOException
  {
    synchronized (this)
      {
	if (output == null)
	  {
	    output = new XoeServletOutputStream ();
	  }
      }
    return (output);
  }

  public PrintWriter getWriter () throws IOException
  {
    synchronized (this)
      {
	writer = new XoePrintWriter (new OutputStreamWriter (getOutputStream (), getCharacterEncoding ()), true);
      }
    return (writer);
  }

  public String getCharacterEncoding ()
  {
    return (contentEncoding);
  }

  public void addCookie (Cookie cookie)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public boolean containsHeader (String name)
  {
    if (CONTENT_TYPE.equals (name))
      {
	return (true);
      }
    else
      {
	return (false);
      }
  }

  public void setStatus (int sc, String sm)
  {
    setStatus (sc);
  }

  public void setStatus (int sc)
  {
    status = sc;
  }

  public void setHeader (String name, String value)
  {
    if (CONTENT_TYPE.equals (name))
      {
	int idx = value.indexOf (';');
	if (idx == -1)
	  {
	    contentType = value;
	    contentEncoding = DEFAULT_CONTENT_ENCODING;
	  }
	else
	  {
	    contentType = value.substring (0, idx).trim ();
	    String encoding = value.substring (idx+1);
	    if (CONTENT_ENCODING.equals (encoding))
	      {
		contentEncoding = encoding.substring (CONTENT_ENCODING.length ()).trim ();
	      }
	    else
	      {
		contentEncoding = DEFAULT_CONTENT_ENCODING;
	      }
	  }
      }
  }

  public void setIntHeader (String name, int value)
  {
    setHeader (name, Integer.toString (value));
  }

  public void setDateHeader (String name, long date)
  {
    setHeader (name, (new Date (date)).toString ());
  }

  public void sendError (int sc, String msg) throws IOException
  {
    status = sc;
    if (output != null)
      {
	output.reset ();
      }
    if (msg != null)
      {
	if (msg.startsWith ("<html>"))
	  {
	    setContentType ("text/html");
	  }
	else 
	  {
	    setContentType ("text/plain");
	  }
	getWriter ().print (msg);
      }
  }

  public void sendError (int sc) throws IOException
  {
    sendError (sc, null);
  }

  public void sendRedirect (String location) throws IOException
  {
    redirect = encodeRedirectUrl (location);
  }

  public String encodeUrl (String url)
  {
    return (url);
  }

  public String encodeRedirectUrl (String url)
  {
    return (url);
  }

  // -------------------------------------------------------------

  Object getOutputData ()
  {
    if (writer != null && writer.doc != null)
     {
	return (writer.doc);
     }
    else if (output == null || output.size () == 0)
      {
	return (null);
      }
    else
      {
	return (output.getBuffer ());
      }
  }

  public void reset ()
  {
    if (output != null)
      {
	output.reset ();
      }
    if (writer != null)
      {
	writer.doc = null;
      }
    contentType = DEFAULT_CONTENT_TYPE;
    contentEncoding = DEFAULT_CONTENT_ENCODING;
    status = 0;
    redirect = null;
  }


  public void addDateHeader (String name, long date)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public void addHeader (String name, String value)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public void addIntHeader (String name, int value)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String encodeRedirectURL(String url)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public String encodeURL(String url)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public void flushBuffer ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public int  getBufferSize ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public Locale  getLocale ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public void setLocale (Locale loc)
  {
    throw new kaffe.util.NotImplemented ();
  }

  public boolean  isCommitted ()
  {
    throw new kaffe.util.NotImplemented ();
  }

  public void  setBufferSize (int size)
  {
    throw new kaffe.util.NotImplemented ();
  }
}
