/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: XsArray.java,v 1.32 2001/08/14 17:03:06 philippe Exp $
 */
package kaffe.xs;

import java.util.*;

final
public class XsArray
  extends XsObject
    implements Cloneable
{
	public XsObject[] elements;
	public int count = 0;
	XsArray next;
	static XsArray cache;
	final static int ARG_SIZE = 9;  //whatever!
        XsNumber length;

        protected static XsObject array__proto__; 

static {
	//System.out.println( "<XsArray>");

	//init cache	
	XsArray xa;
	
	for ( int i =0; i<5; i++ ){
		xa = new XsArray(ARG_SIZE);
		xa.count=0;
		xa.next = cache;
		cache = xa;
	}
}
	

public XsArray (){
        elements = new XsObject[10];
	
        __proto__ = array__proto__;
}

XsArray ( int n ){
	elements = new XsObject[n];
	count = n;
	
        __proto__ = array__proto__;
}
	//ecma conformance to be reviewed	
public static XsObject _concat(XsObject xThis, XsArray args){
        //returns XsArray
        if(!(xThis instanceof XsArray)){
                //review: this does not seem to be ecma specs but it is js behavior
		System.err.println(xThis + ".concat is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        XsArray result = new XsArray(cur.count);
        
        System.arraycopy(cur.elements, 0, result.elements, 0, cur.count);

        result.count = cur.count;

        for(int i = 0; i < args.count; i++){
                if(args.elements[i] instanceof XsArray){
                        //flatten
			XsArray temp = (XsArray) args.elements[i];
                        for(int j = 0; j < temp.count; j++)
				result.push(temp.elements[j]);
                }                        
                else{
			result.push(args.elements[i]);
                }
        }        
        
        return result;
}

public static XsObject _join(XsObject xThis, XsArray args){
        //returns XsString
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".join is not a function");
                return _undefined;
        }
 
        String result = "";

        XsArray cur = ((XsArray)xThis);
        
        if(cur.count > 0){
                
                String delimiter;

		if(args.count == 0 || args.elements[0] == _undefined /* bug in js*/){
			delimiter = ",";
		}
		else{
			delimiter = args.elements[0].call("toString").toString();
		} 

		result = cur.elements[0].toString();
                   
		for ( int i = 1; i < cur.count; i++ ){

                        XsObject curEl = cur.elements[i];

                        String curS;

                        if(curEl == _undefined || curEl == _null){
				curS = "";
                        }
                        else{
				curS = curEl.call("toString").toString();
                        } 
                        
			result += delimiter + curS;
                }
	}
        else{
                result = ""; 
        }
        
        return new XsString(result); 
}

public static XsObject _pop(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".pop is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);
      
        if(cur.count > 0){
		return cur.elements[--cur.count];
        }
        else{
		return _undefined;
        }
}

public static XsObject _push(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".push is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        for(int i = 0; i < args.count; i++){
		cur.push(args.elements[i]);
        }

        return new XsNumber(cur.count);
}

public static XsObject _reverse(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".reverse is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        //don't recurse!!!!
        for(int i = 0, j = cur.count - 1; i < j; i++, j--){
                XsObject temp = cur.elements[i];
             	cur.elements[i] = cur.elements[j];
                cur.elements[j] = temp; 
        }

        return xThis;
}

public static XsObject _shift(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".shift is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);
	
        if(cur.count > 0){
		XsObject ret =  cur.elements[0];
                System.arraycopy(cur.elements, 1, cur.elements, 0, --cur.count);
                return ret;
        }
        else{
		return _undefined;
        } 
}

//review according to latest standard
public static XsObject _slice(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".slice is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        XsArray result = new XsArray();
	
        int beg_index, end_index = 0;

        if(args.count == 0){ 
		beg_index = 0;
                end_index = cur.count;
	}
        else{
 
                //get first index    
 		XsNumber xn = XsNumber.getNumberFromObject(args.elements[0]);

                if(xn == XsNumber.NaN)
			beg_index = 0;
		else{
                	if ( xn.type == XsNumber.TYPE_INT ){
				if(xn.ival < 0)
					beg_index = cur.count + xn.ival;
                                else
                                        beg_index = xn.ival;
                        }
			else{
				if((int)xn.ival < 0)
					beg_index = cur.count + (int)xn.ival;
                                else
                                        beg_index = (int)xn.ival;
                        } 
                }
             
                if(args.count == 1){   
			end_index = cur.count;
                } 

                //get second index  
		
		if(args.count >= 2){
			xn = XsNumber.getNumberFromObject(args.elements[1]);
			if(xn == XsNumber.NaN){
				if(args.elements[1] == _undefined){
					end_index = cur.count;
                                }
                                else{
					end_index = 0;
                                }
		        }
                        else{
				if ( xn.type == XsNumber.TYPE_INT ){
					if(xn.ival < 0)
						end_index = cur.count + xn.ival;
					else
						end_index = xn.ival;
                           	}
				else{
					if((int)xn.ival < 0)
						end_index = cur.count + (int)xn.ival;
					else
						end_index = (int)xn.ival;
				} 
			}
                }

		
                if(end_index < 0)
			end_index = 0;

                if(end_index > cur.count)
			end_index = cur.count;

                if(beg_index < 0)
			beg_index = 0;
                
                if(beg_index >= cur.count || end_index <= beg_index)
			beg_index = end_index = 0;
        }
       
        result.grow(end_index - beg_index);
        
        result.count = end_index - beg_index;         

        System.arraycopy(cur.elements, beg_index, result.elements, 0, result.count);

        return result;
}
//review held until 1.5 review
public static XsObject _sort(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".sort is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        //needed so that elements.length = count in sorting
        XsObject sortelements[] = new XsObject[cur.count];
        System.arraycopy(cur.elements, 0, sortelements, 0, cur.count);
        
        if(args.count == 0){
 		Arrays.sort(sortelements, new XsObjectSortingComparator());
        } 
        else{
      		if(!(args.elements[0] instanceof XsFunc)){
			System.err.println("TypeError: invalid Array.prototype.sort argument");
                        return _undefined;
                }
                
                Arrays.sort(sortelements, new XsObjectSortingComparator((XsFunc) args.elements[0]));
        }

        cur.elements = sortelements;

        return xThis;
}
//review held until 1.5 review
public static XsObject _splice(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".splice is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        XsArray result = new XsArray();
	
        int beg_index, length = 0;

        if(args.count == 0){ 
		result.count = 0;
                return result;
	}
        else{
                //get beg_index
                XsNumber xn = XsNumber.getNumberFromObject(args.elements[0]);
                if(xn == XsNumber.NaN)
			beg_index = 0;
                else{
                	if ( xn.type == XsNumber.TYPE_INT ){
				if(xn.ival < 0)
					beg_index = cur.count + xn.ival;
                                else
                                        beg_index = xn.ival;
                        }
			else{
				if((int)xn.ival < 0)
					beg_index = cur.count + (int)xn.ival;
                                else
                                        beg_index = (int)xn.ival;
                        } 
                }
   

                //get length
                if(args.count == 1){
			length = cur.count - beg_index;
                }
                else{
                        xn = XsNumber.getNumberFromObject(args.elements[1]);
			if(xn == XsNumber.NaN)
				length = 0;
			else{
				if ( xn.type == XsNumber.TYPE_INT ){
					if(xn.ival < 0)
						length = 0;
					else
						length = xn.ival;
                           	}
				else{
					if((int)xn.ival < 0)
						length = 0;
					else
						length = (int)xn.ival;
				} 
			}
                }

 
                if(beg_index + length  >= cur.count)
			length = cur.count - beg_index;
                
                if(beg_index < 0)
			beg_index = 0;

                if(beg_index >= cur.count){
			length = 0;
                        beg_index = cur.count;
                }  
 
                result.grow(length);
        
                result.count = length;  


		//System.out.println("args.count = " + args.count + " beg is " + beg_index + " length is " + length);

                //build result 
                System.arraycopy(cur.elements, beg_index, result.elements, 0, length);


                if(args.count == 1 || args.count == 2){
			System.arraycopy(cur.elements, beg_index + length, cur.elements, beg_index, cur.count - beg_index -length );
                        cur.count -= length;
                }
                else if(args.count >= 3){
			cur.grow(args.count - 2 - length + cur.count);
                        
                        System.arraycopy(cur.elements, beg_index + length, cur.elements, beg_index + args.count - 2,  cur.count - beg_index -length); 
                         
                        System.arraycopy(args.elements, 2, cur.elements, beg_index, args.count - 2);
                        
                        cur.count += args.count - 2  - length;
                } 

                return result;
        }
}

//not quite it, but good enough for now
public static XsObject _toLocaleString(XsObject xThis, XsArray args){
        return new XsString(((XsArray)xThis).toString());
}

public static XsObject _unshift(XsObject xThis, XsArray args){
        
        if(!(xThis instanceof XsArray)){
		System.err.println(xThis + ".unshift is not a function");
                return _undefined;
        }
       
        XsArray cur = ((XsArray)xThis);

        cur.grow(args.count);

        System.arraycopy(cur.elements, 0, cur.elements, args.count, cur.count);

        for(int i = 0; i < args.count; i++)
		cur.elements[i] = args.elements[i];

        cur.count += args.count;

        return new XsNumber(cur.count);
}

public static XsObject __Array ( XsObject xThis, XsArray args ){
       
	XsArray xa = (XsArray) xThis;
	
        if ( (args.count == 1) && (args.elements[0] instanceof XsNumber) ){
                 
		XsNumber xn = (XsNumber) args.elements[0];

		if ( xn.type == XsNumber.TYPE_INT ){
			xa.grow(xn.ival);
                        xa.count = xn.ival;

                }
                else{
                        xa.grow((int)xn.dval); // js: RangeError: invalid array length
                        xa.count = (int)xn.dval;
                }
      	}
	else if(args.count == 1){
                xa.count = 1;
                xa.elements[0] =  args.elements[0];         
        }
        else{
		/*
		for ( int i = 0; i<args.count; i++ ){
			xa.push(args.elements[i]);
		}
		*/
                int args_count = args.count;

                xa.grow(args_count);

                System.arraycopy(args.elements, 0, xa.elements, 0, args_count);

                xa.count = args_count; 
        }

        return xThis;
}

public static XsObject ___Array ( XsObject xThis, XsArray args ){
	return __Array(xThis, args);
}

public static XsObject _ArrayInitializer ( XsObject xThis, XsArray args ){
       
        XsArray xa = (XsArray) xThis;
        /*
        for ( int i = 0; i<args.count; i++ ){
	        xa.push(args.elements[i]);
	}
        */

        int args_count = args.count;

        xa.grow(args_count);

        System.arraycopy(args.elements, 0, xa.elements, 0, args_count);

        xa.count = args_count;

        xThis.xsPut("constructor",  XsObject._Array);

        XsPropertyAttributes propattrs = ((XsPropertyAttributes)xThis.properties_attributes.get("constructor"));
        propattrs.DontEnum = true;
        //propattrs.DontDelete = true;  ??
        

        return xThis;
}

public XsObject and ( String s ){
	int i = 0;
	double d = 0.0;
        int type = 0;
	
	try {
		if ( XsNumber.assumedFloat( s) ){
			d = Double.parseDouble(s);
			type = XsNumber.TYPE_FLOAT;
		}
		else {
			i = Integer.decode(s).intValue();
                        type = XsNumber.TYPE_INT;
		}
	}
	catch ( NumberFormatException x ) {
		return new XsNumber(0.0);
	}

        
        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			if ( type == XsNumber.TYPE_INT ){
				return new XsNumber(thisn.ival & i);
		        }
		        else {
			        return new XsNumber( thisn.ival & (int)d);
		        }
		}
		else {
			if ( type == XsNumber.TYPE_INT ){
			        return new XsNumber( (int)thisn.dval & i);
		        }
		        else {
			        return new XsNumber( (int)thisn.dval & (int)d);
		        }
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject and ( XsObject x ){
       	XsNumber thisn = XsNumber.getNumberFromObject(this);
        if(thisn != XsNumber.NaN){
		XsNumber xn = XsNumber.getNumberFromObject(x);
                
                if(xn != XsNumber.NaN){
			if ( thisn.type == XsNumber.TYPE_INT ){
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( thisn.ival & xn.ival);
				}
				else {
					return new XsNumber( thisn.ival & (int)xn.dval);
				}
			}
			else {
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( (int)thisn.dval & xn.ival);
				}
				else {
					return new XsNumber( (int)thisn.dval & (int)xn.dval);
				}
			}
                } 
                else
			return new XsNumber(0.0);
        }
        else
		return new XsNumber(0.0);

       
}

public XsObject and ( double d ){

        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival & (int)d);
		}
		else {
		        return new XsNumber( (int)thisn.dval & (int)d);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject and ( int i ){
	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival & i);
		}
		else {
		        return new XsNumber( (int)thisn.dval & i);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject andAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).and(s));
        return this;
}

public XsObject andAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).and(x));
        return this;
}

public XsObject andAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).and(d));
        return this;  
}
	
public XsObject andAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).and(i));
        return this;
}

protected Object clone(){
	XsArray out = new XsArray(this.count);
        System.arraycopy(this.elements, 0, out.elements, 0, this.count);
        return out;
}

public void dispose (){
	int n = elements.length;
	
	if ( (n > 0) && (n < 20) ){
		for ( int i=0; i<count; i++ ){
			elements[i] = _undefined;
		}
		
		count = 0;
	
		next = cache;
		cache = this;
	}
}

public XsObject divAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).div(s));
        return this;
}

public XsObject divAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).div(x));
        return this;
}

public XsObject divAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).div(d));
        return this;  
}
	
public XsObject divAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).div(i));
        return this;
}

public XsObject eq ( String s ){

        if(count == 0)
		return xFalse;
        else if(count == 1) 
		return xsGet(0).eq(s);
        else
                return xFalse;
}

public XsObject eq ( XsObject x ) {

        if(count == 0)
		return xFalse;
        else if(count == 1) 
		return xsGet(0).eq(x);
        else
                return xFalse;
}

public XsObject eq ( double d ) {
        if(count == 0)
		return xFalse;
        else if(count == 1) 
		return xsGet(0).eq(d);
        else
                return xFalse;
}

public XsObject eq ( int i ) {
        if(count == 0)
		return xFalse;
        else if(count == 1) 
		return xsGet(0).eq(i);
        else
                return xFalse;
}

void grow ( int idx ) {
	if ( elements.length <= idx ){
		int n = idx+1;
	
		XsObject[] newElements = new XsObject[n];

		if ( elements != null ){
			System.arraycopy( elements, 0, newElements, 0, count);
		}
		elements = newElements;
        }
}

public XsObject minusAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).minus(s));
        return this;
}

public XsObject minusAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).minus(x));
        return this;
}

public XsObject minusAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).minus(d));
        return this;  
}
	
public XsObject minusAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).minus(i));
        return this;
}

public XsObject modAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mod(s));
        return this;
}

public XsObject modAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mod(x));
        return this;
}

public XsObject modAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mod(d));
        return this;  
}
	
public XsObject modAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mod(i));
        return this;
}

public XsObject mulAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mul(s));
        return this;
}

public XsObject mulAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mul(x));
        return this;
}

public XsObject mulAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mul(d));
        return this;  
}
	
public XsObject mulAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).mul(i));
        return this;
}

public XsObject or ( String s ){
	int i = 0;
	double d = 0.0;
        int type = 0;
	
	try {
		if ( XsNumber.assumedFloat( s) ){
			d = Double.parseDouble(s);
			type = XsNumber.TYPE_FLOAT;
		}
		else {
			i = Integer.decode(s).intValue();
                        type = XsNumber.TYPE_INT;
		}
	}
	catch ( NumberFormatException x ) {
		return new XsNumber(0.0);
	}

        
        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			if ( type == XsNumber.TYPE_INT ){
				return new XsNumber(thisn.ival | i);
		        }
		        else {
			        return new XsNumber( thisn.ival | (int)d);
		        }
		}
		else {
			if ( type == XsNumber.TYPE_INT ){
			        return new XsNumber( (int)thisn.dval | i);
		        }
		        else {
			        return new XsNumber( (int)thisn.dval | (int)d);
		        }
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject or ( XsObject x ){
       	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		XsNumber xn = XsNumber.getNumberFromObject(x);
                
                if(xn != XsNumber.NaN){
			if ( thisn.type == XsNumber.TYPE_INT ){
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( thisn.ival | xn.ival);
				}
				else {
					return new XsNumber( thisn.ival | (int)xn.dval);
				}
			}
			else {
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( (int)thisn.dval | xn.ival);
				}
				else {
					return new XsNumber( (int)thisn.dval | (int)xn.dval);
				}
			}
                } 
                else
			return new XsNumber(0.0);
        }
        else
		return new XsNumber(0.0);

       
}

public XsObject or ( double d ){

        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival | (int)d);
		}
		else {
		        return new XsNumber( (int)thisn.dval | (int)d);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject or ( int i ){
	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival | i);
		}
		else {
		        return new XsNumber( (int)thisn.dval | i);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject orAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).or(s));
        return this;
}

public XsObject orAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).or(x));
        return this;
}

public XsObject orAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).or(d));
        return this;  
}
	
public XsObject orAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).or(i));
        return this;
}

public XsObject plus ( String s ){
        return new XsString(this.toString() + s);   
}

public XsObject plus ( XsObject x ){
          XsObject p;

          if(!(x instanceof XsDate)){
		  p = XsObject.getPrimitiveFromObject(x);
	  }
	  else{
		  p = XsObject.getPrimitiveFromDate(x);
	  }

          if(p instanceof XsString){
		  return plus(p.toString());
          }
          else if(p instanceof XsNumber){
		  if ( ((XsNumber)p).type == XsNumber.TYPE_INT ){
			  return plus(((XsNumber)p).ival);
		  }
		  else {
			  return plus(((XsNumber)p).dval);
		  }
          }
          else{   
                  //try toString
                  XsObject func;

                  if(!(x instanceof XsDate)){
			  func = x.xsGet("toString");
		  }
		  else{
			  func = x.xsGet("valueOf");
		  }     
                  
                  //check exceptions on invoke ??

                  if(!(func instanceof XsFunc)){
			  return xFalse;
                  }                                  	
	          else{   
			  return plus(func.invoke(x));
	         }
          } 
}

public XsObject plus ( double d ){
	return new XsString(this.toString() + d);  
}
	
public XsObject plus ( int i ){
        return new XsString(this.toString() + i);
}

public XsObject plusAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).plus(s));
        return this;
}

public XsObject plusAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).plus(x));
        return this;
}

public XsObject plusAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).plus(d));
        return this;  
}
	
public XsObject plusAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).plus(i));
        return this;
}

public XsArray push (String s) {
	return push(new XsString(s));
}

public XsArray push ( XsObject elem ) {
	if ( count >= elements.length ){
		grow( count + 10);
	}
	
	elements[count++] = elem;
        return this;
}

public XsArray push (int i) {
	return push(new XsNumber(i));
}

public XsArray push (double d) {
	return push(new XsNumber(d));
}

public static XsArray pushs (String s) {
	return pushs(new XsString(s));
}

public static XsArray pushs ( XsObject elem ) {
	XsArray a;

	if ( cache != null ){
                a = cache;
		cache = a.next;
                //reset array
                a.count = 1;
		a.next = null;
	}
	else {
		a = new XsArray(ARG_SIZE);
		a.count = 1;
	}
	
	a.elements[0] = elem;

	return a;
}

public static XsArray pushs (int i) {
	return pushs(new XsNumber(i));
}

public static XsArray pushs (double d) {
	return pushs(new XsNumber(d));
}

public XsObject shl ( String s ){
	int i = 0;
	double d = 0.0;
        int type = 0;
	
	try {
		if ( XsNumber.assumedFloat( s) ){
			d = Double.parseDouble(s);
			type = XsNumber.TYPE_FLOAT;
		}
		else {
			i = Integer.decode(s).intValue();
                        type = XsNumber.TYPE_INT;
		}
	}
	catch ( NumberFormatException x ) {
		return new XsNumber(0.0);
	}

        
        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			if ( type == XsNumber.TYPE_INT ){
				return new XsNumber(thisn.ival << i);
		        }
		        else {
			        return new XsNumber( thisn.ival << (int)d);
		        }
		}
		else {
			if ( type == XsNumber.TYPE_INT ){
			        return new XsNumber( (int)thisn.dval << i);
		        }
		        else {
			        return new XsNumber( (int)thisn.dval << (int)d);
		        }
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject shl ( XsObject x ){
       	XsNumber thisn = XsNumber.getNumberFromObject(this);
        if(thisn != XsNumber.NaN){
		XsNumber xn = XsNumber.getNumberFromObject(x);
                
                if(xn != XsNumber.NaN){
			if ( thisn.type == XsNumber.TYPE_INT ){
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( thisn.ival << xn.ival);
				}
				else {
					return new XsNumber( thisn.ival << (int)xn.dval);
				}
			}
			else {
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( (int)thisn.dval << xn.ival);
				}
				else {
					return new XsNumber( (int)thisn.dval << (int)xn.dval);
				}
			}
                } 
                else
			return new XsNumber(0.0);
        }
        else
		return new XsNumber(0.0);

       
}

public XsObject shl ( double d ){

        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival << (int)d);
		}
		else {
		        return new XsNumber( (int)thisn.dval << (int)d);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject shl ( int i ){
	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival << i);
		}
		else {
		        return new XsNumber( (int)thisn.dval << i);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject shlAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shl(s));
        return this;
}

public XsObject shlAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shl(x));
        return this;
}

public XsObject shlAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shl(d));
        return this;  
}
	
public XsObject shlAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shl(i));
        return this;
}

public XsObject shr ( String s ){
	int i = 0;
	double d = 0.0;
        int type = 0;
	
	try {
		if ( XsNumber.assumedFloat( s) ){
			d = Double.parseDouble(s);
			type = XsNumber.TYPE_FLOAT;
		}
		else {
			i = Integer.decode(s).intValue();
                        type = XsNumber.TYPE_INT;
		}
	}
	catch ( NumberFormatException x ) {
		return new XsNumber(0.0);
	}

        
        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			if ( type == XsNumber.TYPE_INT ){
				return new XsNumber(thisn.ival >> i);
		        }
		        else {
			        return new XsNumber( thisn.ival >> (int)d);
		        }
		}
		else {
			if ( type == XsNumber.TYPE_INT ){
			        return new XsNumber( (int)thisn.dval >> i);
		        }
		        else {
			        return new XsNumber( (int)thisn.dval >> (int)d);
		        }
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject shr ( XsObject x ){
       	XsNumber thisn = XsNumber.getNumberFromObject(this);
        if(thisn != XsNumber.NaN){
		XsNumber xn = XsNumber.getNumberFromObject(x);
                
                if(xn != XsNumber.NaN){
			if ( thisn.type == XsNumber.TYPE_INT ){
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( thisn.ival >> xn.ival);
				}
				else {
					return new XsNumber( thisn.ival >> (int)xn.dval);
				}
			}
			else {
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( (int)thisn.dval >> xn.ival);
				}
				else {
					return new XsNumber( (int)thisn.dval >> (int)xn.dval);
				}
			}
                } 
                else
			return new XsNumber(0.0);
        }
        else
		return new XsNumber(0.0);

       
}

public XsObject shr ( double d ){

        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival >> (int)d);
		}
		else {
		        return new XsNumber( (int)thisn.dval >> (int)d);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject shr ( int i ){
	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival >> i);
		}
		else {
		        return new XsNumber( (int)thisn.dval >> i);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject shrAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shr(s));
        return this;
}

public XsObject shrAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shr(x));
        return this;
}

public XsObject shrAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shr(d));
        return this;  
}
	
public XsObject shrAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).shr(i));
        return this;
}

public String toString() {
        StringBuffer b = new StringBuffer( elements.length*10);
	int i;
	XsObject e;
	
	for ( i=0; i < count; ){
		e = elements[i];
		if ( e != null ){
                        if(e != _undefined){
				b.append( e.toString());
                        }
                }
		if (++i < count ){
			b.append(',');
		}
	}
	
	return b.toString();
}

public XsObject toBoolean(){
	return xTrue;
}

public XsObject ushr ( String s ){
	int i = 0;
	double d = 0.0;
        int type = 0;
	
	try {
		if ( XsNumber.assumedFloat( s) ){
			d = Double.parseDouble(s);
			type = XsNumber.TYPE_FLOAT;
		}
		else {
			i = Integer.decode(s).intValue();
                        type = XsNumber.TYPE_INT;
		}
	}
	catch ( NumberFormatException x ) {
		return new XsNumber(0.0);
	}

        
        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			if ( type == XsNumber.TYPE_INT ){
				return new XsNumber(thisn.ival >>> i);
		        }
		        else {
			        return new XsNumber( thisn.ival >>> (int)d);
		        }
		}
		else {
			if ( type == XsNumber.TYPE_INT ){
			        return new XsNumber( (int)thisn.dval >>> i);
		        }
		        else {
			        return new XsNumber( (int)thisn.dval >>> (int)d);
		        }
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject ushr ( XsObject x ){
       	XsNumber thisn = XsNumber.getNumberFromObject(this);
        if(thisn != XsNumber.NaN){
		XsNumber xn = XsNumber.getNumberFromObject(x);
                
                if(xn != XsNumber.NaN){
			if ( thisn.type == XsNumber.TYPE_INT ){
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( thisn.ival >>> xn.ival);
				}
				else {
					return new XsNumber( thisn.ival >>> (int)xn.dval);
				}
			}
			else {
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( (int)thisn.dval >>> xn.ival);
				}
				else {
					return new XsNumber( (int)thisn.dval >>> (int)xn.dval);
				}
			}
                } 
                else
			return new XsNumber(0.0);
        }
        else
		return new XsNumber(0.0);
}

public XsObject ushr ( double d ){

        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival >>> (int)d);
		}
		else {
		        return new XsNumber( (int)thisn.dval >>> (int)d);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject ushr ( int i ){
	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival >>> i);
		}
		else {
		        return new XsNumber( (int)thisn.dval >>> i);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject ushrAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).ushr(s));
        return this;
}

public XsObject ushrAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).ushr(x));
        return this;
}

public XsObject ushrAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).ushr(d));
        return this;  
}
	
public XsObject ushrAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).ushr(i));
        return this;
}

public XsObject xor ( String s ){
	int i = 0;
	double d = 0.0;
        int type = 0;
	
	try {
		if ( XsNumber.assumedFloat( s) ){
			d = Double.parseDouble(s);
			type = XsNumber.TYPE_FLOAT;
		}
		else {
			i = Integer.decode(s).intValue();
                        type = XsNumber.TYPE_INT;
		}
	}
	catch ( NumberFormatException x ) {
		return new XsNumber(0.0);
	}

        
        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			if ( type == XsNumber.TYPE_INT ){
				return new XsNumber(thisn.ival ^ i);
		        }
		        else {
			        return new XsNumber( thisn.ival ^ (int)d);
		        }
		}
		else {
			if ( type == XsNumber.TYPE_INT ){
			        return new XsNumber( (int)thisn.dval ^ i);
		        }
		        else {
			        return new XsNumber( (int)thisn.dval ^ (int)d);
		        }
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject xor ( XsObject x ){
       	XsNumber thisn = XsNumber.getNumberFromObject(this);
        if(thisn != XsNumber.NaN){
		XsNumber xn = XsNumber.getNumberFromObject(x);
                
                if(xn != XsNumber.NaN){
			if ( thisn.type == XsNumber.TYPE_INT ){
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( thisn.ival ^ xn.ival);
				}
				else {
					return new XsNumber( thisn.ival ^ (int)xn.dval);
				}
			}
			else {
				if ( xn.type == XsNumber.TYPE_INT ){
					return new XsNumber( (int)thisn.dval ^ xn.ival);
				}
				else {
					return new XsNumber( (int)thisn.dval ^ (int)xn.dval);
				}
			}
                } 
                else
			return new XsNumber(0.0);
        }
        else
		return new XsNumber(0.0);

       
}

public XsObject xor ( double d ){

        XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival ^ (int)d);
		}
		else {
		        return new XsNumber( (int)thisn.dval ^ (int)d);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject xor ( int i ){
	XsNumber thisn = XsNumber.getNumberFromObject(this);

        if(thisn != XsNumber.NaN){
		if ( thisn.type == XsNumber.TYPE_INT ){
			return new XsNumber( thisn.ival ^ i);
		}
		else {
		        return new XsNumber( (int)thisn.dval ^ i);
		}
        }
        else
		return new XsNumber(0.0);
}

public XsObject xorAssign ( XsObject index, String s ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).xor(s));
        return this;
}

public XsObject xorAssign ( XsObject index, XsObject x ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).xor(x));
        return this;
}

public XsObject xorAssign ( XsObject index, double d ){
	String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).xor(d));
        return this;  
}
	
public XsObject xorAssign ( XsObject index, int i ){
        String index_ts = index.toString();
        xsPut( index_ts, xsGet(index_ts).xor(i));
        return this;
}

public XsObject xsGet(String skey){

	//        System.out.println("xsGet " + skey + "__proto__ = " + (__proto__ == array__proto__));
	char c = skey.charAt(0);
        
	if ( (c >= '0') && (c <= '9') ){
	        try {
	                int idx = Integer.parseInt( skey);
			return xsGet( idx);
		}
		catch ( Exception x ) {
                       return super.xsGet(skey);
                }
        }
        else if ( (c == 'l') && (skey.equals("length")) ){
		if ( length == null ){
			length = new XsNumber(count);
		}
		else {
			length.ival = count;
		}
		return length;
        }
	/*  else if( skey.equals( "prototype") ){
 
                System.out.println("IN HERE Array");
     		return array__proto__; 
		}*/
        else
		return super.xsGet(skey);
}

public XsObject xsGet(XsNumber xn){
        if ( xn.type == XsNumber.TYPE_INT ){
                return xsGet(xn.ival); 
	}
        else{
                return xsGet((int)xn.dval); 
        }  
}

public XsObject xsGet ( int idx ){
	if ( (idx < 0) || (idx >= count) ){
		return super.xsGet(new Integer(idx).toString());
	}
	
	XsObject v = elements[idx];
	return (v == null) ? _undefined : v;
}

	/*//can this be here ??
public XsObject xsGet ( double d ){
	if ( (d < 0.0) || (d >= count) ){
		return super.xsGet(new Integer((int)d).toString());
	}
	
	XsObject v = elements[(int)d];
	return (v == null) ? _undefined : v;
}
*/
public XsObject xsPut ( String key, XsObject elem ){

        char c = key.charAt(0);
	if ( (c >= '0') && (c <= '9') ){
		try {
			int idx = Integer.parseInt( key);
			return xsPut( idx, elem);
		}
		catch ( Exception x ) {
                        return super.xsPut( key, elem); 
                }
	}
        else if ( (c == 'l') && (key.equals("length")) ){

		XsNumber xn = XsNumber.getNumberFromObject(elem);

                if(xn == XsNumber.NaN){
			System.err.println("RangeError: invalid length");
			return _undefined;
                }
                else{  
                        int index;
                        
                        if ( xn.type == XsNumber.TYPE_INT ){
				index = xn.ival; 
			}
			else{
				index = (int)xn.dval; 
			} 


                        for(int i = index; i < count; i++){
				elements[i] = _undefined;
                        }


                        count = index;

                        return elem;
                }
        }//this is never called. similar for XsString etc. take out
        else if ( key.equals("prototype") ){
		/*
                  but Array.prototype = new Object;
                  does not do what I would expect in Nertscaqpe anyways 
		*/
                //System.out.println("RAMBO");
                if(canPut(key)){
			return array__proto__ = elem;
                }
                else{   //not sure ??			
                        return _undefined;
                }
	}
        else
		return super.xsPut( key, elem);
}

public XsObject xsPut( int idx, XsObject elem ){

        
        while ( (elements == null) || (elements.length <= idx) ){
		grow( idx + 10); 
	}
	
	elements[idx] = elem;
	
	if (idx >= count) {
		count = idx+1;
	} 

	return elem;
}

}

class XsObjectSortingComparator implements Comparator{

        XsFunc compareFunc = null;
        static Object[] argObj = new Object[2];

        static{
		argObj[0] = null;
        }

        public int compare(Object obj1, Object obj2){

                if(compareFunc == null)
			return ((XsObject)obj1).toString().compareTo(((XsObject)obj2).toString());	       
                else{
                        argObj[1] = XsArray.pushs((XsObject)obj1).push((XsObject)obj2);

		        Object ret = null;        

                        try{
				ret = compareFunc.mth.invoke(null, argObj);
                        }
                        catch(Exception e){ 
				return 0;
                        }
			
                        XsNumber xn = XsNumber.getNumberFromObject((XsObject)ret);

                        if(xn == XsNumber.NaN)
				return 0;

                        if ( xn.type == XsNumber.TYPE_INT )
				return xn.ival;
                        
			else
      				return (int)xn.dval;
                }
        }
        
        public XsObjectSortingComparator(XsFunc func){
		compareFunc = func;
        } 

        public XsObjectSortingComparator(){
        } 
	
} // class 







