/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: XsEngine.java,v 1.34 2001/09/24 22:15:14 philippe Exp $
 */
package kaffe.xs;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringBufferInputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Properties;
import java.util.StringTokenizer;
import kaffe.tools.compiler.*;
import java.io.FileOutputStream;

final public class XsEngine
	extends ClassLoader  
{
	
	final public static int MODE_SRC = 0;
	final public static int MODE_BC = 1;
	final public static int MODE_NATIVE = 2;
        //int mode = MODE_SRC;
        int mode = MODE_BC;
	static String srcCompiler;
	static String bcCompiler;
	static String bcClasspath;
	static String[] imports;
	static String baseClass;
	static boolean log = false;
	static String factoryName;
	static boolean cleanUp = true;
	static boolean useCache = false;
        //static boolean pretty_print_DOM_tree = true;
	static URL contextURL;

        private static long interpreted_code_sequence = 0;
        private static XsEngine interpreted_code_compiler; 
        private static String interp_compiler_args[] = new String[5];
        
        static int write_generated_sequence = 0;

        static int DocClassName_sequence = 0;

        //very temporary solution
        public static DocumentFactory curDocumentFactory;


static{

       
	System.loadLibrary("xs");


	Properties gp = new Properties();
	try {
		gp.load( new FileInputStream("xs.prp") );
	}
	catch ( IOException _) {
		System.err.println( "Cannot find XsEngine properties ( xs.prp), using defaults");
	}

	srcCompiler = gp.getProperty( "srcCompiler", "xs");
	bcCompiler = gp.getProperty( "bcCompiler", "javac"); //jikes bugs in CondExpression
	bcClasspath = gp.getProperty( "bcClasspath", ".:javalib/lib:../kaffe-all/build-gnu/libraries/javalib/lib:../pocketlinux/lib" );
	baseClass = gp.getProperty( "baseClass", "com/pocketlinux/dom/BaseDocument");
	//factoryName = gp.getProperty( "factoryName", "DocumentFactory");
        factoryName = XsObject.docFactoryCls = gp.getProperty("factoryName", "com.pocketlinux.dom.DOM0Adaptor");       
        String ip = gp.getProperty( "imports", "com.pocketlinux.dom.*");
	StringTokenizer it = new StringTokenizer( ip);
	imports = new String[ it.countTokens() ];
	for ( int i = 0; it.hasMoreTokens(); i++ ) {
		imports[i] = it.nextToken();
	}

        interp_compiler_args[0] =  "xs";
	interp_compiler_args[1] =  "-s";
	interp_compiler_args[2] =  "-f";
	interp_compiler_args[4] =  "kaffe/xs/XsObject";
}

//returns success or failure
native private static boolean compile(InputStream is, String[] cmda, URL url, XsObject[] domtree);


public XsObject createDocument ( URL url ) {
        //System.out.println("HERE");

	try {
     	     return (createDocument(  url.openConnection()));
	}
	catch ( IOException _ ) {
		return XsObject._null;
	}
}

public XsObject createDocument ( URL url, int mode ) {
	this.mode = mode; //wrong way to do, keep localized!!
	return createDocument(url);
}

public XsObject createWriteGeneratedDocument ( String src ) {
	
	//System.out.println("In createDocument");

        StringBufferInputStream sbis;

        sbis = new StringBufferInputStream(src);

	XsObject doc = null;

        boolean onlyDOMtree[] = new boolean[1]; // circumvent the lack of reference params 
	

        try{
            switch ( mode ){
	      case MODE_SRC:
			doc = createWriteGeneratedDocumentViaSrc( sbis, "write_generated" + write_generated_sequence++, onlyDOMtree);
			break;
	      case MODE_BC:
			doc = createWriteGeneratedDocumentViaBC( sbis, "write_generated" + write_generated_sequence++, onlyDOMtree);
			break;
	      case MODE_NATIVE:
			doc = createWriteGeneratedDocumentViaNative( sbis, "write_generated" + write_generated_sequence++, onlyDOMtree);
			break;
	    }

	    sbis.close();
            //to be done in an other way?? 
	    //DOM0Attrs.init( doc, null);  // find out what it should be

            if(!onlyDOMtree[0]){
		    //will build the DOM2 tree ( doc url already been set by location property )
		    doc.reset();
            }

	}
	catch ( IOException ex ){
		System.out.println( ex);
	}
	
        //System.out.println("HERE BEFORE returning doc");
        
        return doc;
}

XsObject createWriteGeneratedDocumentViaBC ( InputStream is, String clsName, boolean onlyDOMtree[]) {
	XsObject doc = null;
	String[] cmda;
        XsObject domtree[] = new XsObject[1]; // circumvent the lack of reference params

        try {
		boolean pp = !useCache || ! (new File( clsName + ".class")).exists();
		if (   pp ) {
			System.out.println( "PHASE 1: generate " + clsName);
			//--- phase 1. - generate *.class
			cmda = new String[5];
			cmda[0] = srcCompiler;
                        cmda[1] = "-w";
                        cmda[2] = clsName;
	       		cmda[3] = baseClass;
			cmda[4] = factoryName;
 
			if ( !compile(is, cmda, null, domtree)  ){
				System.err.println("source compilation failed: ");
                                onlyDOMtree[0] = true;
				return null;
			}

                }

                if(domtree[0] == null){ // if was not built by JNI code
	        
			//--- phase 2 - instantiate
			System.out.println( "PHASE 2: loading " + clsName);
			Class clazz = loadClass( clsName, true);
			//System.out.println( "Class loaded");
			doc = (XsObject) clazz.newInstance();
			//System.out.println( "New instance created");
                }
                 
	}
	catch ( Exception ex ){
               ex.printStackTrace();
	}

	if ( cleanUp) {
               (new File( clsName + ".class")).delete();
	}

        if(domtree[0] == null){
                onlyDOMtree[0] = false;
		//		System.out.println("doc.toXML " + ((com.pocketlinux.dom.BaseDocument)doc).toXML());
                return doc; 
        } 
        else{   
                onlyDOMtree[0] = true;
		//System.out.println("doc.toXML " + ((com.pocketlinux.dom.BaseDocument)domtree[0]).toXML());
		return domtree[0];
        }
}

XsObject createWriteGeneratedDocumentViaNative ( InputStream is, String clsName, boolean onlyDOMtree[] ) {
	System.err.println( "XMLCompiler error - native mode not supported yet");	
	return null;
}

XsObject createWriteGeneratedDocumentViaSrc ( InputStream is, String clsName, boolean onlyDOMtree[]) {
	System.err.println( "XsEngine error - source mode not supported yet");	
	return null;
}

	
public XsObject createDocument ( URLConnection uc ) {
	URL url = uc.getURL();
	InputStream is;
	XsObject doc = null;
	String dcname = getDocClassName( url);
        contextURL = url;

        boolean onlyDOMtree[] = new boolean[1]; // circumvent the lack of reference params
        
	try {
	        is = uc.getInputStream();
		if ( is == null ) {
			System.out.println( "cannot open: " + uc.getURL());
			return XsObject._null;
		}
		switch ( mode ){
		case MODE_SRC:
			doc = createDocumentViaSrc( is, dcname, onlyDOMtree);
			break;
		case MODE_BC:
			doc = createDocumentViaBC( is, dcname, url, onlyDOMtree);
			break;
		case MODE_NATIVE:
			doc = createDocumentViaNative( is, dcname, onlyDOMtree);
			break;
		}

		is.close();

                

                //XoeScript interface init
		DOM0Attrs.init( doc, new XsString(url));

                if(!onlyDOMtree[0]){
                        //will build the DOM2 tree ( doc url already been set by location property )
                        doc.reset();
                } 
               
     	}
	catch ( IOException ex ){
		System.out.println( ex);
	}
	//System.out.println("Returning doc in createDocument ");
        

        //System.out.println("doc.toXML " + ((com.pocketlinux.dom.BaseDocument)doc).toXML());

	return doc;
}

//current xoe interface
public Object createDocument ( URL url, InputStream is ) {
	Object doc = null;
	String dcname = getDocClassName( url);
        contextURL = url;

        boolean onlyDOMtree[] = new boolean[1]; // circumvent the lack of reference params
        
	try {
      		if ( is == null ) {
			System.out.println( "null inputstream");
			return XsObject._null;
		}

		switch ( mode ){

		case MODE_SRC:
			doc = createDocumentViaSrc( is, dcname, onlyDOMtree);
			break;
		case MODE_BC:
			doc = createDocumentViaBC( is, dcname, url, onlyDOMtree);
			break;
		case MODE_NATIVE:
			doc = createDocumentViaNative( is, dcname, onlyDOMtree);
			break;
		}

		is.close();

                //XoeScript interface init
		//DOM0Attrs.init( ((XsObject)doc), new XsString(url));

                if(!onlyDOMtree[0]){
                        //will build the DOM2 tree ( doc url already been set by location property )
                        ((XsObject)doc).reset();
                } 
               
     	}
	catch ( IOException ex ){
		System.out.println( ex);
	}
	System.out.println("Returning doc in createDocument ");
        

        //System.out.println("doc.toXML " + ((com.pocketlinux.dom.BaseDocument)doc).toXML());

	return doc;
}

XsObject createDocumentViaBC ( InputStream is, String clsName, URL url, boolean onlyDOMtree[]) {
	XsObject doc = null;
	String[] cmda;
        XsObject domtree[] = new XsObject[1]; // circumvent the lack of reference params

        try {
		boolean pp = !useCache || ! (new File( clsName + ".class")).exists();
		if (   pp ) {
			System.out.println( "PHASE 1: generate " + clsName);
			//--- phase 1. - generate *.class

			
			cmda = new String[2];
			cmda[0] = srcCompiler;
                        cmda[1] = clsName;
			
                        


                        /*
                        cmda = new String[2];
			cmda[0] = srcCompiler;
                        cmda[1] = "-p";
			*/
                        

			

			//System.out.println("before compile");
                        if ( !compile(is, cmda, url, domtree)  ){
				System.err.println("source compilation failed: ");
				return null;
			}

                }

                if(domtree[0] == null){ // if was not built by JNI code
	        
			//--- phase 2 - instantiate
			System.out.println( "PHASE 2: loading " + clsName);
			Class clazz = loadClass( clsName, true);
			//System.out.println( "Class loaded");
			doc = (XsObject) clazz.newInstance();
			//System.out.println( "New instance created");
                }
                 
	}
	catch ( Exception ex ){
               ex.printStackTrace();
	}

	if ( cleanUp) {
               (new File( clsName + ".class")).delete();
	}

        if(domtree[0] == null){
                onlyDOMtree[0] = false;
                return doc; 
        } 
        else{   
                //System.err.println(((com.pocketlinux.dom.BaseNode)domtree[0]).toXML());	
                onlyDOMtree[0] = true;
		return domtree[0];
        }
}

XsObject createDocumentViaNative ( InputStream is, String clsName, boolean onlyDOMtree[] ) {
	System.err.println( "XMLCompiler error - native mode not supported yet");	
	return null;
}

XsObject createDocumentViaSrc ( InputStream is, String clsName, boolean onlyDOMtree[]) {
        XsObject doc = null;
	String[] cmda;
	int i;
	String srcFile = getSourceFile( clsName);
     
        try {
		boolean pp = !useCache || ! (new File( srcFile)).exists();
		if (   pp ) {
			System.out.println( "PHASE 1: generate " + srcFile);
			//--- phase 1. - generate *.java sources
			cmda = new String[5 + imports.length];
			cmda[0] = srcCompiler;
                        cmda[1] = clsName;
			cmda[2] = srcFile;
			cmda[3] = baseClass;
			cmda[4] = factoryName;

			for ( i=0; i<imports.length; i++ ){
				cmda[5+i] = imports[i];
			}

                                                
                        /* old way

			byte[] buf = new byte[2048];
	                int n, ret;	
                        Process proc = Runtime.getRuntime().exec( cmda);
			OutputStream os = proc.getOutputStream();
			InputStream iss = null;
			
			if ( log ) {
				iss = proc.getInputStream();
			}
			


			while ( (n = is.read( buf)) >= 0 ){
				os.write( buf, 0, n);

				if ( log ) {
					if ( (n=iss.read(buf)) > 0 ){
						System.out.write(buf,0,n);
					}
					}
			}
			os.close();
			
			proc.waitFor();
			if ( (ret = proc.exitValue()) > 0 ){
				System.err.println("source compilation failed: " + ret);
				return null;
			}
                        */
			
			if ( !compile(is, cmda, null, null) ){ //THIS WON'T WORK with currenct code
				System.err.println("source compilation failed: ");
				return null;
		        }
                }
                
		pp = !useCache || ! ( new File( clsName + ".class")).exists();
		if (  pp ) {
			//--- phase 2. generate class file
			System.out.println( "PHASE 2: compiling " + srcFile);
			kaffe.tools.compiler.Compiler c = kaffe.tools.compiler.Compiler.getInstance( bcCompiler);
			c.setClasspath( bcClasspath);
			if ( ! c.compile(  srcFile) ) {
				System.err.println("bytecode compilation failed ");
				return null; // XsObject._null ??
			}
		}		

		//--- phase 3 - instantiate
		System.out.println( "PHASE 3: loading " + clsName);
		Class clazz = loadClass( clsName, true);
		doc = (XsObject) clazz.newInstance();
	}
	catch ( Exception ex ){
               ex.printStackTrace();
	}

	if ( cleanUp) {
		(new File( srcFile)).delete();
		(new File( clsName + ".class")).delete();
	}

        onlyDOMtree[0] = false;
        return doc;
}

public XsObject createFunction (InputStream is, String args[]) {

	switch ( mode ){
	case MODE_SRC:
                args[4] = "XsFunc";
		return createFunctionViaSrc( is, args);

	case MODE_BC:
                args[4] = "kaffe/xs/XsFunc";
		return createFunctionViaBC( is, args);

	case  MODE_NATIVE:
		return createFunctionViaNative( is, args);

	}

	return null; // XsObject._null ??
}

public XsObject createFunction (InputStream is, String args[], int mode) {
	this.mode = mode;
	return createFunction(is, args);
}

XsObject createFunctionViaBC ( InputStream is, String args[]) {
	XsObject doc = null;
       
        try {
		boolean pp = !useCache || ! (new File( args[3] + ".class")).exists();
		if (   pp ) {
			//--- phase 1. - generate *.class 

                        if ( !compile(is, args, null, null) ){
				System.err.println("source compilation failed");
				return null;
		        }
                        
                }
                
                //--- phase 2 - instantiate
		Class clazz = loadClass( args[3], true);
		doc = (XsObject) clazz.newInstance();
                
	}
	catch ( Exception ex ){
		ex.printStackTrace();
	}

        if ( cleanUp) {
		(new File( args[3] + ".class")).delete();
	}
	
        return doc; 
}

XsObject createFunctionViaNative ( InputStream is, String args[]) {
	System.err.println( "XMLCompiler error - native mode not supported yet");	
	return null;
}

public XsObject createFunctionViaSrc ( InputStream is, String args[] ) {
	XsObject doc = null;
	String[] cmda;
        String srcFile = getSourceFile( args[3]);
       
        try {
		boolean pp = !useCache || ! (new File( srcFile)).exists();
		if (   pp ) {
			//--- phase 1. - generate *.java sources
			cmda = new String[args.length + 1];

			cmda[0] = args[0];
			cmda[1] = args[1];
			cmda[2] = args[2];
			cmda[3] = args[3];
			cmda[4] = srcFile; 	    
			cmda[5] = args[4];
                        
			if ( !compile(is, cmda, null, null) ){  //THIS WON'T WORK with currenct code
				System.err.println("source compilation failed");
				return null;
			}
                        
                }

		pp = !useCache || ! ( new File( args[3] + ".class")).exists();
		if (  pp ) {
			//--- phase 2. generate class file
			kaffe.tools.compiler.Compiler c = kaffe.tools.compiler.Compiler.getInstance( bcCompiler);
			c.setClasspath( bcClasspath);
			if ( ! c.compile(  srcFile) ) {
				System.err.println("bytecode compilation failed ");
				return null; // XsObject._null ??
			}

		}		

		//--- phase 3 - instantiate
		Class clazz = loadClass( args[3], true);
		doc = (XsObject) clazz.newInstance();
	}
	catch ( Exception ex ){
		ex.printStackTrace();
	}
 
        //cleanup code removed by mistake 

	return doc;
}

protected void finalize() throws Throwable {
	System.out.println( "UNLOADING");
	super.finalize();
}

String getDocClassName( URL url ){
	int i0, i1;
	String  s = url.toExternalForm();

        if ( (i0 = s.lastIndexOf('/')) < 0 ){
		i0 = 0;
	}
	else {
		i0++;
	}

	if ( (i1 = s.lastIndexOf('.')) < 0 ){
		i1 = s.length()-1;
	}

        try{
		s = s.substring( i0, i1);
        }
        catch(StringIndexOutOfBoundsException e){
		return "document_" + DocClassName_sequence++;
        }
       
        return s;
}

public static InputStream getRelative(String file) throws MalformedURLException, IOException{
        return new URL(contextURL, file).openConnection().getInputStream();
}

private static final String getSourceFile ( String clsName ){
	return clsName + ".java";
}
	
private native static XsObject interp(XsObject context, XsObject xThis, String script);

//Achtung!! interpreted_code_sequence is a shared resource
//synchronize only compile run ??
public static /*synchronized */XsObject interpret(XsObject context, XsObject xThis, String script){

        XsObject ret;

        if((ret = interp(context, xThis, script)) == null){  
       
		System.out.println("interpreter failed on " + script);

                // if get error trying to interp, compile and run
 
               	
                 //compile
		
                interp_compiler_args[3] =  "InterpretedCode_" + interpreted_code_sequence++;
		
                XsObject code_class;

                if(interpreted_code_compiler == null)
		        interpreted_code_compiler = new XsEngine(); 

                code_class = interpreted_code_compiler.createFunction(
                                                                      new StringBufferInputStream("function " + interp_compiler_args[3] +
		                                                                                  " (){" + script + "}"
                                                                                                 )
			                                             ,
                                                                      interp_compiler_args
                                                                     );

                if(code_class == null)// means got a compile error, ex: new Function(5) or new Function("return 8 ** 9;");
			System.err.println("Got compile error in interpret"); // return xsObject._undefined ??
                    
                
                Method code = null;
                
                //run

                try{ 
			code = code_class.getClass().getMethod("_InterpretedCode_" + (interpreted_code_sequence - 1), 
                                                                XsFunc.argTypes
                                                               );  
                }
	        catch ( NoSuchMethodException x ) {
			x.printStackTrace();
                        return XsObject._undefined; // ???
		}
	
                Object[] argObj = new Object[2];  // make this a field, new only once
                
                argObj[0] = xThis;
                //make the following line better
                argObj[1] = new XsArray(0).push(context);
       
                try{    //this does not work in src-to-src
			ret = (XsObject) code.invoke( null, argObj );
                }
	        catch ( InvocationTargetException ite ) {
			//ite.printStackTrace();
                        System.out.println("Exception occured while executing compiled code " + ite.getTargetException());//get Invokation target Exception when get expcetion in called code 
                }
                catch(IllegalAccessException iae){}
        }

        return ret;
}
	


} // class


















