/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: XsMath.java,v 1.16 2001/08/14 17:03:07 philippe Exp $
 */
package kaffe.xs;


final public class XsMath
  extends XsObject
{
       
	final public static XsNumber E = new XsNumber(Math.E);
	final public static XsNumber PI = new XsNumber(Math.PI);
	final public static XsNumber LN10 = new XsNumber(Math.log(10.0));
	final public static XsNumber LN2 = new XsNumber(Math.log(2.0));
	final public static XsNumber LOG10E = new XsNumber(1.0 / Math.log(10.0));
	final public static XsNumber LOG2E = new XsNumber(1.0 / Math.log(2.0));
	final public static XsNumber SQRT1_2 = new XsNumber(1.0 / Math.sqrt(2.0));
	final public static XsNumber SQRT2 = new XsNumber(Math.sqrt(2.0));
	
        
        public static XsNumber _abs(XsObject xThis, XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
		XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.abs(x.ival));
                else  
                        return new XsNumber(Math.abs(x.dval));
        }


        public static XsNumber _acos(XsObject xThis,XsArray args){

                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);
 
                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.acos(x.ival));
                else 
                        return new XsNumber(Math.acos(x.dval));
        }
        
        
        public static XsNumber _asin(XsObject xThis,XsArray args){

                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.asin(x.ival));
                else 
                        return new XsNumber(Math.asin(x.dval));
        }

        
        public static XsNumber _atan(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.atan(x.ival));
                else 
                        return new XsNumber(Math.atan(x.dval));
        }


        public static XsNumber _atan2(XsObject xThis,XsArray args){
                 
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);
                XsNumber y =  XsNumber.getNumberFromObject(args.elements[1]);

                

                if(x.type == XsNumber.TYPE_INT && y.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.atan2(x.ival, y.ival));
                else if(x.type == XsNumber.TYPE_INT && y.type == XsNumber.TYPE_FLOAT)
			return new XsNumber(Math.atan2((double)x.ival, y.dval));
                else if(x.type == XsNumber.TYPE_FLOAT && y.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.atan2(x.dval, (double)y.ival));  
                else
                        return new XsNumber(Math.atan2(x.dval, y.dval)); 
        }
        
    
        public static XsNumber _ceil(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);
                
                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.ceil(x.ival));
                else 
                        return new XsNumber(Math.ceil(x.dval));
        }

        
        public static XsNumber _cos(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]); 
 
                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.cos(x.ival));
                else 
                        return new XsNumber(Math.cos(x.dval));
        }

        
        public static XsNumber _exp(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.exp(x.ival));
                else 
                        return new XsNumber(Math.exp(x.dval));
        }


        public static XsNumber _floor(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.floor(x.ival));
                else 
                        return new XsNumber(Math.floor(x.dval));
        }


        public static XsNumber _log(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.log(x.ival));
                else 
                        return new XsNumber(Math.log(x.dval));
        }


        public static XsNumber _max(XsObject xThis,XsArray args){
                //what is the type of the result if have a mix of ints and doubles?
                if(args.count == 0){
			return XsNumber.NaN;
	        }

                XsNumber xn;
                double max = Double.MIN_VALUE;

                for(int i = 0; i < args.count; i++){
			xn = XsNumber.getNumberFromObject(args.elements[i]);
			if(xn.type == XsNumber.TYPE_INT){
				max = Math.max(max, (double)xn.ival);
                        }
		        else{
				max = Math.max(max, xn.dval);
                        }
                }

                return new XsNumber(max); 
        }


        public static XsNumber _min(XsObject xThis,XsArray args){
                //what is the type of the result if have a mix of ints and doubles?
                if(args.count == 0){
			return XsNumber.NaN;
	        }

                XsNumber xn;
                double min = Double.MAX_VALUE;

                for(int i = 0; i < args.count; i++){
			xn = XsNumber.getNumberFromObject(args.elements[i]);
			if(xn.type == XsNumber.TYPE_INT){
				min = Math.min(min, (double)xn.ival);
                        }
		        else{
				min = Math.min(min, xn.dval);
                        }
                }

                return new XsNumber(min); 
        }


        public static XsNumber _pow(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);
                XsNumber y =  XsNumber.getNumberFromObject(args.elements[1]);


                if(x.type == XsNumber.TYPE_INT && y.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.pow(x.ival, y.ival));
                else if(x.type == XsNumber.TYPE_INT && y.type == XsNumber.TYPE_FLOAT)
			return new XsNumber(Math.pow((double)x.ival, y.dval));
                else if(x.type == XsNumber.TYPE_FLOAT && y.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.pow(x.dval, (double)y.ival));  
                else
                        return new XsNumber(Math.pow(x.dval, y.dval)); 
        }


        public static XsNumber _random(XsObject xThis,XsArray args){
                return new XsNumber(Math.random());
        }
 

        public static XsNumber _round(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]); 

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.round(x.ival));
                else 
                        return new XsNumber(Math.round(x.dval));
        }


        public static XsNumber _sin(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.sin(x.ival));
                else 
                        return new XsNumber(Math.sin(x.dval));
        }


        public static XsNumber _sqrt(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.sqrt(x.ival));
                else 
                        return new XsNumber(Math.sqrt(x.dval));
        }


        public static XsNumber _tan(XsObject xThis,XsArray args){
                
                if(args.count == 0)
			return XsNumber.NaN;
                XsNumber x =  XsNumber.getNumberFromObject(args.elements[0]);

                if(x.type == XsNumber.TYPE_INT)
			return new XsNumber(Math.tan(x.ival));
                else 
                        return new XsNumber(Math.tan(x.dval));
        }


        //this does not work yet, will rethink when know more
        public String toString() {
	        return "[object Math]";
        }

}


