/*
 * Copyright (c) 2000, 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: XsProps.java,v 1.9 2001/08/09 00:52:10 philippe Exp $
 */
package kaffe.xs;

import java.util.Enumeration;
import java.util.NoSuchElementException;

public class XsProps
{
	XsObject[] values;
	Object[] keys;
	int count;
	int threshold;
	static XsObject reclaimed = new XsObject();
	final static int LOAD_FACTOR = 80;

class Enumerator
  implements Enumeration
{
	int index = -1;
	Object next;

public boolean hasMoreElements () {
        XsObject o = null;

        if ( next == null ) {
                for ( index++; index < values.length; index++ ) {
                        if ( ((o = values[index]) != null) && (o != reclaimed) ){
                                break;
                        }
                }

                next = o;
                return (next != null);
        }
        else {
                return true;
        }
}

public Object nextElement () {
        Object ret;

        if ( next == null ) {
                if ( !hasMoreElements() )
                        throw new NoSuchElementException("PrimHashtabEnumerator");
        }

        ret = next;
        next = null;

        return ret;
}
}
class KeyEnumerator
  extends XsProps.Enumerator
{
public boolean hasMoreElements () {
        Object o = null;

        if ( next == null ) {
                for ( index++; index < keys.length; index++ ) {
                        if ( ((o = keys[index]) != null) && (o != reclaimed) )
                                break;
                }

                next = o;
                return (next != null);
        }
        else {
                return true;
        }
}
}

public XsProps() {
	this( 10);
}

public XsProps ( int initialCapacity ) {
	init(initialCapacity);
}

public synchronized void clear () {
	int i, n = values.length;

	if ( values.length > 32 ){
		values = new XsObject[n];
		keys = new Object[n];
	}
	else {
		for ( i=0; i<n; i++ ){
			values[i] = null;
			keys[i] = null;
		}
	}

	count = 0;
}

static int computeNearestPrime ( int n ) {
	return (n | 1);  // cheating !!
}

public void dump () {
	int i, n = count;
	Object o;

	if ( count == 0 ) System.out.println("{}");

	System.out.println( "{");
	for ( i=0; true; i++ ) {
		if ( ((o = keys[i]) != null) && (o != reclaimed) ){
			System.out.print( o);
			System.out.print( '=');
			System.out.print( values[i]);
			if ( --n > 0 )
				System.out.print( ", ");
			else
				break;
		}
	}
	System.out.println( "}");
}

public synchronized Enumeration elements () {
	return new XsProps.Enumerator();
}

public synchronized XsObject get ( Object key ) {
	Object  o;
	int     k = (key.hashCode() & Integer.MAX_VALUE);
	int     d, i = k % keys.length;

	if ( key.equals( keys[i] ) ){
		return values[i];
	}
	else if ( keys[i] == null ){   /* no "null" key value allowed */
		return null;
	}

	for ( d=(16 - (k & 0xf)), i+= d; true; i+= d ){
		i %= keys.length;

		if ( key.equals( keys[i]) ){
			return values[i];
		}
		else if ( keys[i] == null ){
			return null;
		}
	}
}

public synchronized void init(int initialCapacity){
        
        int n = (initialCapacity * 100 / LOAD_FACTOR);

	n = prime( n);
	threshold =  n * LOAD_FACTOR / 100;

	values = new XsObject[n];
	keys = new Object[n];
	count = 0;
}

public boolean isEmpty () {
	return (count == 0);
}

public synchronized Enumeration keys () {
	return new XsProps.KeyEnumerator();
}

static int prime ( int n ) {
	if ( n < 11 ) return 11;
	else if ( n < 17 ) return 17;
	else if ( n < 23 ) return 23;
	else if ( n < 31 ) return 31;
	else if ( n < 41 ) return 41;
	else if ( n < 67 ) return 67;
	else if ( n < 83 ) return 83;
	else if ( n < 151 ) return 151;
	else if ( n < 307 ) return 307;
	else if ( n < 503 ) return 503;
	else if ( n < 751 ) return 751;
	else if ( n < 1009 ) return 1009;
	else if ( n < 1511 ) return 1511;
	else if ( n < 2503 ) return 2503;
	else if ( n < 5003 ) return 5003;
	else if ( n < 7507 ) return 7507;
	else if ( n < 10007 ) return 10007;
	else if ( n < 15013 ) return 15013;
	else if ( n < 20011 ) return 20011;
	else if ( n < 30011 ) return 30011;
	else if ( n < 40009 ) return 40009;
	else if ( n < 50021 ) return 50021;
	else if ( n < 75011 ) return 75011;
	else if ( n < 100003 ) return 100003;
	else if ( n < 150001 ) return 150001;
	else if ( n < 200003 ) return 200003;

	else return computeNearestPrime( n);
}

public synchronized XsObject put ( Object key, XsObject value ) {
	//returns null if key was not stored previously, previous value if was 
        XsObject  o;
	int     k = (key.hashCode() & Integer.MAX_VALUE);
	int     d, i = k % keys.length;

	if ( count > threshold ) {
		rehash();
		i = k % keys.length;
	}

	if ( keys[i] == null ) {
		keys[i] = key; values[i] = value; count++;
		return null;
	}
	else if ( keys[i].equals( key) ){
		o = values[i];
		values[i] = value;
		return o;
	}

	for ( d=(16 - (k & 0xf)), i+= d; true; i+= d ){
		i %= keys.length;
		if ( keys[i] == null ) {
			keys[i] = key; values[i] = value; count++;
			return null;
		}
		else if ( keys[i].equals( key) ){
			o = values[i];
			values[i] = value;
			return o;
		}
	}
}

private boolean rehash () {
	int       i, j, k, d;
	Object    key, o;
	int       newLength = prime( ((values.length * 3 / 2) * 100 / LOAD_FACTOR));
	Object[]  newKeys   = new Object[newLength];
	XsObject[]  newValues = new XsObject[newLength];

	for ( j=0; j < values.length; j++ ) {
		if ( keys[j] != null ) {
			key = keys[j];
			k = (key.hashCode() & Integer.MAX_VALUE);
			i = k % newLength;

			if ( newKeys[i] == null ) {
				newKeys[i] = key; newValues[i] = values[j];
				continue;
			}

			for ( d=(16 - (k & 0xf)), i+= d; true; i+= d ){
				i %= newLength;
				if ( newKeys[i] == null ) {
					newKeys[i] = key; newValues[i] = values[j];
					break;
				}
			}
		}
	}

	threshold = (newLength * LOAD_FACTOR / 100);
	keys   = newKeys;
	values = newValues;

	return true;
}

public synchronized XsObject remove ( Object key ) {
	XsObject  o;
	int     k = (key.hashCode() & Integer.MAX_VALUE);
	int     d, i = k % keys.length;

	if ( key.equals( keys[i]) ) {
		o = values[i];
		values[i] = reclaimed;
		keys[i] = null;
		return o;
	}
	else if ( keys[i] == null ){             
		return null;
        }

	for ( d=(16 - (k & 0xf)), i+= d; true; i+= d ){
		i %= keys.length;
		if ( key.equals( keys[i]) ) {
			o = values[i];
			values[i] = reclaimed;
			keys[i] = null;
			return o;
		}
		else if ( keys[i] == null ){             
			return null;
                } 
	}
}
/*
public synchronized void removeAll ( ) {
	init(10);
}
*/
public int size () {
	return count;
}
}
