//ML_ONLY_START


/*
 * Generic AST node
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Philippe Laporte 
 * 
 * $tvt: AST.h,v 1.15 2001/09/28 20:28:55 philippe Exp $
 */
 
//this file used as a genral place to put stuff to be included, as well as the AST class 
 

#ifndef _AST_H_INCLUDED
#define _AST_H_INCLUDED

#include "ASTBase.h"
#include "AToken.h"

#include "ATokPtr.h"

typedef ANTLRCommonToken ANTLRToken;


//ML_ONLY_END



#include "symtab.h"

typedef struct symtab* symtabPtr;


#include "code.h"


//to differentiate between array literal and array reference
#define ARRAY_REFERENCE false
#define ARRAY_LITERAL true


#define GLOBAL true
#define LOCAL false


#define LHS false
#define RHS true


#define PRE false
#define POST true

#define RAW true
#define NOT_RAW false

#define NEEDUNDEF true
#define DEF   false


//for switch code generation
typedef struct SwitchLabel{
      int label;
      struct SwitchLabel* next;
} SwitchLabel;     


//whether Script is within xml or alone (-s option)
#define EMBEDDED true
#define ALONE false


//ML_ONLY_START

//#define NORMAL true
//#define UNWINDING false




class AST : public ASTDoublyLinkedBase{//  ASTBase {


private:
	ANTLRTokenPtr token; // pointer to the token found in input

		
public:	
//ML_ONLY_END

		
	//use some sort of union??
	
	// FUNCTION
	bool return_undef_required;
  
	
	symtabPtr localsym;
	
	//resources
	int labelcount;
        int localslimit;
	
	//code
	struct CODE *opcodes; 
	struct LABEL *labels;
	
	int nargs;  //number of arguments
	
        bool CallStackRequired;
	
	// ID (locals and parameters)
	bool side; // RHS or LHS
	int offset;  //code generation, which local it is
	bool ignore_decl; //whether it is a redundant name declaration and thus not associated to a resource 
	bool offsetAssigned;
	
	bool formal; //whether a local is a formal param or a local var
		
				
	//symbol analysis
	bool local_or_global;
	short level;
	
	
	bool isMember;
	
	bool nextIsAssignOp;
	
	//to differentiate between array literal and array reference
	bool array_kind;
	
	//whether an array literal is empty or not
	bool hasLiteral;
	
	
	// IF, to know which code pattern to generate
	bool else_present;  //syntactic predicates don't seem to work in sorcerer
	
	
	// LPAREN
	bool callHasArgs;
	
	//resources
	
	
	//labels
	int stoplabel;  //if-else, while
	int elselabel;  //if-else
	int startlabel; //while, do-while
	
       	int falselabel; //condexp
	
	int continuelabel; // for
	
	int truelabel;  //logOr
	
	int CallStackPoppinglabel; // function that have CallStackRequired true

        //link to symbol table, for IDs
	//symtabPtr symbol;
	
	
	// FOR
	CODE* whereBreakGoes;
	CODE* whereContinueGoes;
	
	//BREAK, CONTINUE
	AST* link;
	bool labelThere;
	
	// INC DEC
	bool pre_or_post;
	
        

        // ignore node while traversing. As if it were zapped.
        bool ignore;  
	
	//SWITCH
	int nlabels;
	// dummy place holder not needed but leave in otherwise crash -> investigate
	int* label_numbers;  
	struct SwitchLabel* Labels;
	
	//for AST diambiguiation between MINUS expr and MINUX expr expr
	bool take;
	
/*	//functions: do not generate a static field
	bool noField;
*/	
      
        //literal rawness
	bool raw;
      
	
	//array literal, for dealing with sparse arrays
	int item_number;
	//for LBRACK
	bool* item_flags;
	int n_items;			
	

//ML_ONLY_START
														
	// DOM nodes
	int nodelevel;												
																									
	
	//TXT_WORD
	bool containsEntities;




//methods
public:

        
	// called when #[tokentype,string] is seen in an action
	AST(ANTLRTokenType tok, char *s) { 
	 // printf("creating a new ANTLRTOKEN\n");
	  token = new ANTLRToken(tok,s);
	}



	// constructor called by parser for token references in grammar
	AST(ANTLRTokenPtr t) { 
           token = t;
	   
//ML_ONLY_END
	   
	   //strip ' or " from strings
           if(t->getType() == STRING_CONST){
	       char* t_text = t->getText();
	       int len = strlen(t_text);
	       char * temp =  (char*) calloc(len + 1, sizeof(char));
               strcpy(temp, t_text);
               temp[len - 1] = '\0';
	       t->setText(temp + 1);
	       free(temp);
	   }
	   
//ML_ONLY_START	   
	   
	}
	
        
	// copy constructor
/*	AST(AST &t){
	        token = t.token;
		//setDown(NULL);  
	//	setRight(NULL);
	        setDown(t.down());  
		setRight(t.right());

        }
*/


        /*  PCCTS_AST* shallowCopy() { 
	    return new AST(*this); 
	}*/

	
	
        int type(){
          return token->getType(); 
        }

        void setText(char* text){
          token->setText(text);
        }

	char* getText(){
          return token->getText();
        }
 
	// for preorder traversal
	void preorder_action(void* pData = NULL/* MR24*/) {
		char *s = token->getText();
		printf(" %s type is %d\n", s, token->getType());
       	}
	
	
	
	//make sure memory is allocated from kaffe's heap
	
	void* operator new(size_t size){
	    // printf("AST NEW called for size %d\n", size);
	    return (void*) malloc(size);
        }
	
        void operator delete(void* p, size_t size){
	    free(p);   
        }


};

#endif /* _AST_H_INCLUDED */


//ML_ONLY_END
