/*
 * coding utilities
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Philippe Laporte 
 *
 * $tvt: code.c,v 1.6 2001/07/28 02:50:00 philippe Exp $
 */

#include <stdio.h>
//#include <string.h>
#include <stdlib.h>

#include "code.h"



CODE *currentcode;
CODE *currenttail;
LABEL *currentlabels;

//for switch production
bool setCondition = false;
CODE* condition_end;

inline void appendCODE(CODE *c)
{ 
  if (currentcode==NULL) {
     currentcode = c;
     currenttail = c;
  } else {
     currenttail->next = c;
     currenttail = c;
     
     if(setCondition){
       condition_end = c;
       setCondition = false;
     }
  }
}

void freeCODE(CODE* c){

  CODE* prev, *cur; 
  
  for(cur = c; cur != NULL; ){
    
    prev = cur;
    
    cur = cur->next;
    
    switch(prev->kind){
            
	case newCK:          if(prev->toBeFreed){
	                         free(prev->val.newC);
                             }
                             break; 
	case checkcastCK:    if(prev->toBeFreed){
	                         free(prev->val.checkcastC);
                             }
                             break; 		     
	case instanceofCK:   if(prev->toBeFreed){
	                         free(prev->val.instanceofC);
                             }
                             break; 
		     
			     
	case ldc_stringCK:   if(prev->toBeFreed){
	                         free(prev->val.ldc_stringC);
			     }	 
			     break;
		     
    
        case getfieldCK:     if(prev->toBeFreed){
	                         free(prev->val.getfieldC);
			     }	 
			     break;
    
        case putfieldCK:     if(prev->toBeFreed){
	                         free(prev->val.putfieldC);
			     }	 
                             break;
			     
        case getstaticCK:    if(prev->toBeFreed){
	                         free(prev->val.getstaticC);
                             }
			     break;
			     
        case putstaticCK:    if(prev->toBeFreed){
	                         free(prev->val.putstaticC);
                             }
                             break; 
			     
        case invokevirtualCK:if(prev->toBeFreed){
	                         free(prev->val.invokevirtualC.text);
			     }	 
                             break;
			     
        case invokespecialCK:if(prev->toBeFreed){
	                         free(prev->val.invokespecialC.text);
                             }
			     break;
			     
	case invokestaticCK: if(prev->toBeFreed){
	                         free(prev->val.invokestaticC.text);
			     }	 
                             break;
			     
        case invokeinterfaceCK:if(prev->toBeFreed){
	                         free(prev->val.invokeinterfaceC.text);
                               }
			       break;
    }
    
    free(prev);
    
  }  
  
}





/*
void code_i2c()
{ appendCODE(makeCODEi2c(NULL));
}
*/
void code_new(char *arg, bool toBeFreed)
{ appendCODE(makeCODEnew(arg,NULL,toBeFreed));
}

void code_instanceof(char *arg, bool toBeFreed)
{ appendCODE(makeCODEinstanceof(arg,NULL,toBeFreed));
}

void code_checkcast(char *arg, bool toBeFreed)
{ appendCODE(makeCODEcheckcast(arg,NULL,toBeFreed));
}
/*
void code_imul()
{ appendCODE(makeCODEimul(NULL));
}
 
void code_ineg()
{ appendCODE(makeCODEineg(NULL));
}
 
void code_irem()
{ appendCODE(makeCODEirem(NULL));
}

void code_isub()
{ appendCODE(makeCODEisub(NULL));
}
 
void code_idiv()
{ appendCODE(makeCODEidiv(NULL));
}

void code_iadd()
{ appendCODE(makeCODEiadd(NULL));
}
*/

CODE *code_label(char *name, int label)
{ 
 // printf("in code_label, name is %s, currentlables is %p\n", name,currentlabels);
  currentlabels[label].name = name;
  currentlabels[label].sources = 1;
  appendCODE(makeCODElabel(label,NULL));
  currentlabels[label].position = currenttail;
  return currenttail;
}

void code_goto(int label)
{ appendCODE(makeCODEgoto(label,NULL));
}

void code_ifeq(int label)
{ appendCODE(makeCODEifeq(label,NULL));
}
/*
void code_ifne(int label)
{ appendCODE(makeCODEifne(label,NULL));
}
*/
void code_if_acmpeq(int label)
{ appendCODE(makeCODEif_acmpeq(label,NULL));
}

void code_if_acmpne(int label)
{ 
  //printf("label is %d\n", label);
  
  appendCODE(makeCODEif_acmpne(label,NULL));
}
/*
void code_ifnull(int label)
{ appendCODE(makeCODEifnull(label,NULL));
}

void code_ifnonnull(int label)
{ appendCODE(makeCODEifnonnull(label,NULL));
}

void code_if_icmpeq(int label)
{ appendCODE(makeCODEif_icmpeq(label,NULL));
}

void code_if_icmpgt(int label)
{ appendCODE(makeCODEif_icmpgt(label,NULL));
}

void code_if_icmplt(int label)
{ appendCODE(makeCODEif_icmplt(label,NULL));
}

void code_if_icmple(int label)
{ appendCODE(makeCODEif_icmple(label,NULL));
}

void code_if_icmpge(int label)
{ appendCODE(makeCODEif_icmpge(label,NULL));
}

void code_if_icmpne(int label)
{ appendCODE(makeCODEif_icmpne(label,NULL));
}

void code_ireturn()
{ appendCODE(makeCODEireturn(NULL));
}
*/
void code_areturn()
{ appendCODE(makeCODEareturn(NULL));
}

void code_return()
{ appendCODE(makeCODEreturn(NULL));
}
 
void code_aload(int arg)
{ appendCODE(makeCODEaload(arg,NULL));
}

void code_astore(int arg)
{ appendCODE(makeCODEastore(arg,NULL));
}
/*
void code_iload(int arg)
{ appendCODE(makeCODEiload(arg,NULL));
}

void code_istore(int arg)
{ appendCODE(makeCODEistore(arg,NULL));
}
*/
void code_dup()
{ appendCODE(makeCODEdup(NULL));
}

void code_pop()
{ appendCODE(makeCODEpop(NULL));
}
/*
void code_swap()
{ appendCODE(makeCODEswap(NULL));
}
*/
void code_ldc_int(int arg)
{ appendCODE(makeCODEldc_int(arg,NULL));
}

void code_ldc_float(double arg)
{ appendCODE(makeCODEldc_float(arg,NULL));
}

void code_ldc_string(char *arg, bool toBeFreed)
{ appendCODE(makeCODEldc_string(arg,NULL,toBeFreed));
}
/*
void code_aconst_null()
{ appendCODE(makeCODEaconst_null(NULL));
}

void code_getfield(char *arg, bool toBeFreed)
{ appendCODE(makeCODEgetfield(arg,NULL,toBeFreed));
}
*/
void code_putfield(char *arg, bool toBeFreed)
{ appendCODE(makeCODEputfield(arg,NULL,toBeFreed));
}

void code_getstatic(char *arg, bool toBeFreed)
{ appendCODE(makeCODEgetstatic(arg,NULL,toBeFreed));
}

void code_putstatic(char *arg, bool toBeFreed)
{ appendCODE(makeCODEputstatic(arg,NULL,toBeFreed));
}

void code_invokevirtual_toBeFreed(char *arg, int nargs, bool isVoid)
{ appendCODE(makeCODEinvokevirtual_toBeFreed(arg,nargs,isVoid,NULL));
}

void code_invokevirtual(char *arg, int nargs, bool isVoid)
{ appendCODE(makeCODEinvokevirtual(arg,nargs,isVoid,NULL));
}

void code_invokespecial(char *arg, int nargs, bool isVoid, bool toBeFreed)
{ appendCODE(makeCODEinvokespecial(arg,nargs,isVoid,NULL,toBeFreed));
}

void code_invokestatic(char *arg, int nargs, bool isVoid, bool toBeFreed)
{ appendCODE(makeCODEinvokestatic(arg,nargs,isVoid,NULL,toBeFreed));
}

void code_invokeinterface(char *arg, int nargs, bool isVoid, bool toBeFreed)
{ appendCODE(makeCODEinvokeinterface(arg,nargs,isVoid,NULL,toBeFreed));
}

void code_athrow()
{ appendCODE(makeCODEathrow(NULL));
}



/*

CODE *makeCODEi2c(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = i2cCK;
 //c->visited = 0;
  c->next = next;
  return c;
}
 */
CODE *makeCODEnew(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = newCK;
  c->visited = 0;
  c->val.newC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}
 
CODE *makeCODEinstanceof(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = instanceofCK;
  c->visited = 0;
  c->val.instanceofC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}

CODE *makeCODEcheckcast(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = checkcastCK;
  c->visited = 0;
  c->val.checkcastC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}
 /*
CODE *makeCODEimul(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = imulCK;
  //c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEineg(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = inegCK;
  //c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEirem(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = iremCK;
  //c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEisub(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = isubCK;
  //c->visited = 0;
  c->next = next;
  return c;
}

CODE *makeCODEidiv(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = idivCK;
  //c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEiadd(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = iaddCK;
  //c->visited = 0;
  c->next = next;
  return c;
}

CODE *makeCODEiinc(int offset, int amount, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = iincCK;
  //c->visited = 0;
  c->val.iincC.offset = offset;
  c->val.iincC.amount = amount;
  c->next = next;
  return c;
}
*/
CODE *makeCODElabel(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = labelCK;
  c->visited = 0;
  c->val.labelC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEgoto(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = gotoCK;
  c->visited = 0;
  c->val.gotoC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEifeq(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = ifeqCK;
  c->visited = 0;
  c->val.ifeqC = label;
  c->next = next;
  return c;
}
 /*
CODE *makeCODEifne(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = ifneCK;
  c->visited = 0;
  c->val.ifneC = label;
  c->next = next;
  return c;
}
 */
CODE *makeCODEif_acmpeq(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_acmpeqCK;
  c->visited = 0;
  c->val.if_acmpeqC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_acmpne(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_acmpneCK;
  c->visited = 0;
  c->val.if_acmpneC = label;
  c->next = next;
  return c;
}
/*
CODE *makeCODEifnull(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = ifnullCK;
  c->visited = 0;
  c->val.ifnullC = label;
  c->next = next;
  return c;
}

CODE *makeCODEifnonnull(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = ifnonnullCK;
  //c->visited = 0;
  c->val.ifnonnullC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_icmpeq(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_icmpeqCK;
  //c->visited = 0;
  c->val.if_icmpeqC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_icmpgt(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_icmpgtCK;
  //c->visited = 0;
  c->val.if_icmpgtC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_icmplt(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_icmpltCK;
  //c->visited = 0;
  c->val.if_icmpltC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_icmple(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_icmpleCK;
  //c->visited = 0;
  c->val.if_icmpleC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_icmpge(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_icmpgeCK;
  //c->visited = 0;
  c->val.if_icmpgeC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEif_icmpne(int label, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = if_icmpneCK;
  //c->visited = 0;
  c->val.if_icmpneC = label;
  c->next = next;
  return c;
}
 
CODE *makeCODEireturn(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = ireturnCK;
  //c->visited = 0;
  c->next = next;
  return c;
}
 */
CODE *makeCODEareturn(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = areturnCK;
  c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEreturn(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = returnCK;
  c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEaload(int arg, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = aloadCK;
  c->visited = 0;
  c->val.aloadC = arg;
  c->next = next;
  return c;
}

CODE *makeCODEastore(int arg, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = astoreCK;
  c->visited = 0;
  c->val.astoreC = arg;
  c->next = next;
  return c;
}
 /*
CODE *makeCODEiload(int arg, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = iloadCK;
  //c->visited = 0;
  c->val.iloadC = arg;
  c->next = next;
  return c;
}
 
CODE *makeCODEistore(int arg, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = istoreCK;
  //c->visited = 0;
  c->val.istoreC = arg;
  c->next = next;
  return c;
}
 */
CODE *makeCODEdup(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = dupCK;
  c->visited = 0;
  c->next = next;
  return c;
}
 
CODE *makeCODEpop(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = popCK;
  c->visited = 0;
  c->next = next;
  return c;
}
 /*
CODE *makeCODEswap(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = swapCK;
  //c->visited = 0;
  c->next = next;
  return c;
}
 */
CODE *makeCODEldc_int(int arg, CODE *next)
{ CODE *c;
  c = NEW(CODE); 
  c->kind = ldc_intCK; 
  c->visited = 0;
  c->val.ldc_intC = arg; 
  c->next = next;
  return c;
}

CODE *makeCODEldc_float(double arg, CODE *next)
{ CODE *c;
  c = NEW(CODE); 
  c->kind = ldc_floatCK; 
  c->visited = 0;
  c->val.ldc_floatC = arg; 
  c->next = next;
  return c;
}

CODE *makeCODEldc_string(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = ldc_stringCK;
  c->visited = 0;
  c->val.ldc_stringC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}
 /*
CODE *makeCODEaconst_null(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = aconst_nullCK;
  c->visited = 0;
  c->next = next;
  return c;
}
 */
CODE *makeCODEgetfield(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = getfieldCK;
  c->visited = 0;
  c->val.getfieldC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}
 
CODE *makeCODEputfield(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = putfieldCK;
  c->visited = 0;
  c->val.putfieldC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}

CODE *makeCODEgetstatic(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = getstaticCK;
  c->visited = 0;
  c->val.getstaticC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}
 
CODE *makeCODEputstatic(char *arg, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = putstaticCK;
  c->visited = 0;
  c->val.putstaticC = arg;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}

 
CODE *makeCODEinvokevirtual_toBeFreed(char *arg, int nargs, bool isVoid, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = invokevirtualCK;
  c->visited = 0;
  c->val.invokevirtualC.text = arg;
  c->val.invokevirtualC.nargs = nargs;
  c->val.invokevirtualC.isVoid = isVoid;
  c->toBeFreed = true;
  c->next = next;
  return c;
}

CODE *makeCODEinvokevirtual(char *arg, int nargs, bool isVoid, CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = invokevirtualCK;
  c->visited = 0;
  c->val.invokevirtualC.text = arg;
  c->val.invokevirtualC.nargs = nargs;
  c->val.invokevirtualC.isVoid = isVoid;
  c->toBeFreed = false;
  c->next = next;
  return c;
}

CODE *makeCODEinvokespecial(char *arg, int nargs, bool isVoid, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = invokespecialCK;
  c->visited = 0;
  c->val.invokespecialC.text = arg;
  c->val.invokespecialC.nargs = nargs;
  c->val.invokespecialC.isVoid = isVoid;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}

CODE *makeCODEinvokestatic(char *arg, int nargs, bool isVoid, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = invokestaticCK;
  c->visited = 0;
  c->val.invokestaticC.text = arg;
  c->val.invokestaticC.nargs = nargs;
  c->val.invokestaticC.isVoid = isVoid;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}

CODE *makeCODEinvokeinterface(char *arg, int nargs, bool isVoid, CODE *next, bool toBeFreed)
{ CODE *c;
  c = NEW(CODE);
  c->kind = invokeinterfaceCK;
  c->visited = 0;
  c->val.invokeinterfaceC.text = arg;
  c->val.invokeinterfaceC.nargs = nargs;
  c->val.invokeinterfaceC.isVoid = isVoid;
  c->toBeFreed = toBeFreed;
  c->next = next;
  return c;
}

CODE *makeCODEathrow(CODE *next)
{ CODE *c;
  c = NEW(CODE);
  c->kind = athrowCK;
  c->visited = 0;
  c->next = next;
  return c;
}



