/*
 * coding utilities
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Philippe Laporte 
 *
 * $tvt: code.h,v 1.6 2001/07/28 02:50:00 philippe Exp $
 */

#ifndef _CODE_H_INCLUDED
#define _CODE_H_INCLUDED

#include "xs_config.h"

#ifdef KAFFELIB  //if used as a kaffe jni lib

#include <jni.h>

#endif


#define NEW(type) (type*) malloc(sizeof(type))


typedef struct LABEL {
   char *name;
   int sources;
   struct CODE *position;
} LABEL;


enum CODE_KIND{
         //i2cCK,
	 instanceofCK,
	 checkcastCK,
	 newCK,
         //imulCK,inegCK,iremCK,isubCK,idivCK,iaddCK,iincCK,
         labelCK,gotoCK,ifeqCK,
	 //ifneCK,
	 if_acmpeqCK,if_acmpneCK,
        // ifnullCK,ifnonnullCK,
        // if_icmpeqCK,if_icmpgtCK,if_icmpltCK,
        // if_icmpleCK,if_icmpgeCK,if_icmpneCK,
         //ireturnCK,
	 areturnCK,returnCK,
         aloadCK,astoreCK,
	 //iloadCK,istoreCK,
	 dupCK,popCK,
	 //swapCK,
         ldc_intCK,ldc_stringCK,ldc_floatCK, 
	// aconst_nullCK,
         getfieldCK,
	 putfieldCK,
	 getstaticCK,putstaticCK,invokevirtualCK,
	 invokespecialCK, invokestaticCK, invokeinterfaceCK,
	 athrowCK
};

typedef struct CODE {

   enum CODE_KIND kind;

   //MAKE this a boolean
   int visited; 
 
   union {
     char *newC;
     char *instanceofC;
     char *checkcastC;
   //  struct {int offset; int amount;} iincC;
     int labelC;
     int gotoC;
     int ifeqC;
   //  int ifneC;
     int if_acmpeqC;
     int if_acmpneC;
    // int ifnullC;
    // int ifnonnullC;
    // int if_icmpeqC;
    // int if_icmpgtC;
   //  int if_icmpltC;
    // int if_icmpleC;
   //  int if_icmpgeC;
   //  int if_icmpneC;
     int aloadC;
     int astoreC;
   //  int iloadC;
   //  int istoreC;
     int ldc_intC;
     double ldc_floatC;
     char *ldc_stringC;
     char *getfieldC;
     char *putfieldC;
     char *getstaticC;
     char *putstaticC;
     struct {char *text; int nargs; bool isVoid;} invokevirtualC;
     struct {char *text; int nargs; bool isVoid;} invokespecialC;
     struct {char *text; int nargs; bool isVoid;} invokestaticC;
     struct {char *text; int nargs; bool isVoid;} invokeinterfaceC;
   } val;
   
   bool toBeFreed; //put inside val???
   
   struct CODE *next;
   
} CODE;



CODE *next(CODE *c);


void freeCODE(CODE* c);

//these are in optimize.c !!!
int replace(CODE **c, int k, CODE *r);
int replaceNoFree(CODE **c, int k, CODE *r);


inline void appendCODE(CODE *c);
//void code_i2c();
void code_new(char *arg, bool toBeFreed);
void code_instanceof(char *arg, bool toBeFreed);
void code_checkcast(char *arg, bool toBeFreed);
//void code_imul(); 
//void code_ineg(); 
//void code_irem();
//void code_isub(); 
//void code_idiv();
//void code_iadd();
CODE *code_label(char *name, int label);
void code_goto(int label);
void code_ifeq(int label);
//void code_ifne(int label);
void code_if_acmpeq(int label);
void code_if_acmpne(int label);
//void code_ifnull(int label);
//void code_ifnonnull(int label);
//void code_if_icmpeq(int label);
//void code_if_icmpgt(int label);
//void code_if_icmplt(int label);
//void code_if_icmple(int label);
//void code_if_icmpge(int label);
//void code_if_icmpne(int label);
//void code_ireturn();
void code_areturn();
void code_return(); 
void code_aload(int arg);
void code_astore(int arg);
//void code_iload(int arg);
//void code_istore(int arg);
void code_dup();
void code_pop();
//void code_swap();
void code_ldc_int(int arg);
void code_ldc_float(double arg);
void code_ldc_string(char *arg, bool toBeFreed);
//void code_aconst_null();
//void code_getfield(char *arg, bool toBeFreed);
void code_putfield(char *arg, bool toBeFreed);
void code_getstatic(char *arg, bool toBeFreed);
void code_putstatic(char *arg, bool toBeFreed);
void code_invokevirtual_toBeFreed(char *arg, int nargs, bool isVoid);  // since there are so many invokevirtual calls
void code_invokevirtual(char *arg, int nargs, bool isVoid);
void code_invokespecial(char *arg, int nargs, bool isVoid, bool toBeFreed);
void code_invokestatic(char *arg, int nargs, bool isVoid, bool toBeFreed);
void code_invokeinterface(char *arg, int nargs, bool isVoid, bool toBeFreed);
void code_athrow();

//CODE *makeCODEi2c(CODE *next);
CODE *makeCODEnew(char *arg, CODE *next, bool toBeFreed);
CODE *makeCODEinstanceof(char *arg, CODE *next, bool toBeFreed);
CODE *makeCODEcheckcast(char *arg, CODE *next, bool toBeFreed);
//CODE *makeCODEimul(CODE *next);
//CODE *makeCODEineg(CODE *next);
//CODE *makeCODEirem(CODE *next);
//CODE *makeCODEisub(CODE *next);
//CODE *makeCODEidiv(CODE *next);
//CODE *makeCODEiadd(CODE *next);
//CODE *makeCODEiinc(int offset, int amount, CODE *next);
CODE *makeCODElabel(int label, CODE *next);
CODE *makeCODEgoto(int label, CODE *next);
CODE *makeCODEifeq(int label, CODE *next);
//CODE *makeCODEifne(int label, CODE *next);
CODE *makeCODEif_acmpeq(int label, CODE *next);
CODE *makeCODEif_acmpne(int label, CODE *next);
//CODE *makeCODEifnull(int label, CODE *next);
//CODE *makeCODEifnonnull(int label, CODE *next);
//CODE *makeCODEif_icmpeq(int label, CODE *next);
//CODE *makeCODEif_icmpgt(int label, CODE *next);
//CODE *makeCODEif_icmplt(int label, CODE *next);
//CODE *makeCODEif_icmple(int label, CODE *next);
//CODE *makeCODEif_icmpge(int label, CODE *next);
//CODE *makeCODEif_icmpne(int label, CODE *next);
//CODE *makeCODEireturn(CODE *next);
CODE *makeCODEareturn(CODE *next);
CODE *makeCODEreturn(CODE *next);
CODE *makeCODEaload(int arg, CODE *next);
CODE *makeCODEastore(int arg, CODE *next);
//CODE *makeCODEiload(int arg, CODE *next);
//CODE *makeCODEistore(int arg, CODE *next);
CODE *makeCODEdup(CODE *next);
CODE *makeCODEpop(CODE *next);
//CODE *makeCODEswap(CODE *next);
CODE *makeCODEldc_int(int arg, CODE *next);
CODE *makeCODEldc_float(double arg, CODE *next);
CODE *makeCODEldc_string(char *arg, CODE *next, bool toBeFreed);
//CODE *makeCODEaconst_null(CODE *next);
CODE *makeCODEgetfield(char *arg, CODE *next, bool toBeFreed);
CODE *makeCODEputfield(char *arg, CODE *next, bool toBeFreed);
CODE *makeCODEgetstatic(char *arg, CODE *next, bool toBeFreed);
CODE *makeCODEputstatic(char *arg, CODE *next, bool toBeFreed);
CODE *makeCODEinvokevirtual(char *arg, int nargs, bool isVoid, CODE *next);
CODE *makeCODEinvokevirtual_toBeFreed(char *arg, int nargs, bool isVoid, CODE *next);
CODE *makeCODEinvokespecial(char *arg, int nargs, bool isVoid, CODE *next, bool toBeFreed);
CODE *makeCODEinvokestatic(char *arg, int nargs, bool isVoid, CODE *next, bool toBeFreed);
CODE *makeCODEinvokeinterface(char *arg, int nargs, bool isVoid, CODE *next, bool toBeFreed);
CODE *makeCODEathrow(CODE *next);

#endif /* _CODE_H_INCLUDED */



