/*
 * Label lookup. A simple stack
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Philippe Laporte 
 *
 * $tvt: labels.c,v 1.3 2001/07/13 22:26:31 philippe Exp $
 */

#include <stdlib.h>
#include <string.h>

#include "labels.h"
#include "xs_config.h"

#ifdef KAFFELIB  //if used as a kaffe jni lib
#include "jni.h"
#endif

#define INITIAL_MAX_LEVEL 4 // whatever

//rewrite the following as a struct ??
char** labels;
AST**  nodes;

int maxLevelLabels;

int levelLabels;


//review efficiency

void initLabels ()
{
	maxLevelLabels = INITIAL_MAX_LEVEL;
	labels = (char**) calloc(INITIAL_MAX_LEVEL, sizeof(char*));
	nodes = (AST**) calloc(INITIAL_MAX_LEVEL, sizeof(AST*));
	levelLabels = 0;
}

void growLabels ()
{
	int newMaxLevel = maxLevelLabels + 10;
	
	char** newLabels = (char**) calloc(newMaxLevel,sizeof(char*));
        memcpy( newLabels, labels, maxLevelLabels*sizeof(char*));
        free( labels);
	labels = newLabels;
	
	AST** newNodes = (AST**) calloc(newMaxLevel,sizeof(AST*));
	memcpy( newNodes, nodes, maxLevelLabels*sizeof(AST*));
        free( nodes);
	nodes = newNodes;

	maxLevelLabels = newMaxLevel;
}


// useless since will always delete through removeLabel + remove call
// or put levelLabels as the loop bound instead
void deleteLabels ()
{
	int i;
	for ( i=0; i < maxLevelLabels; i++ ){
		if ( labels[i] ){
			free( labels[i]);
		}
		//don't free AST nodes! done by AST->destroy
	}
	
	free( labels);
	free( nodes);
}


bool addLabel ( char* sym, AST* node)
{
        if(levelLabels == maxLevelLabels)
	  growLabels();
      
        if(isLabelDefined(sym))
	  return false;
	else{ 
	  nodes[levelLabels] = node; 
	  labels[levelLabels] = StrDup(sym);
	  levelLabels++;
          return true;
	}  
}

void addUnLabeled (AST* node)
{
        if(levelLabels == maxLevelLabels)
	  growLabels();
      
        nodes[levelLabels] = node; 
        labels[levelLabels] = StrDup("");
	levelLabels++;
}


bool isLabelDefined(char* sym)
{
       for(int i = levelLabels - 1; i >= 0; i--)
      	 if(labels[i])
           if(!strcmp(labels[i], sym))
	     return true;
       	 
       return false;
}

AST* getLabelNode(char* sym)
{
       for(int i = levelLabels - 1 ; i >= 0; i--)
      	 if(labels[i])
           if(!strcmp(labels[i], sym))
	     return nodes[i];
       	 
       return NULL;
}

AST* getFirstLabelNode()
{
       for(int i = levelLabels - 1; i >= 0; i--){
      	 if(labels[i]){
	     return nodes[i];
       	 }
       }	 
	 
       return NULL;
}

void removeLabel ()
{
        free(labels[--levelLabels]);
	//the nodes will be freed elsewhere ?? in sorcerer code ??
      
}

