//ML_ONLY_START


/*
 * JNI compiler/interpreter interface and main routine
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Philippe Laporte 
 * @author Peter Mehlitz
 *
 * $tvt: main.c,v 1.50 2001/10/26 22:48:09 philippe Exp $
 */
 
#include <stdio.h>
#include <unistd.h>

#include "xs_config.h"

#include "tokens.h"
#include "util.h"

//ML_ONLY_END

#include "emit.c"  // get include problems otherwise

//ML_ONLY_START

#include "queue.h"

#ifdef KAFFELIB
/*
 * Hack to support older jni.h files
 */
#ifndef JNIEXPORT
#define JNIEXPORT
#endif

#endif

#include "entities.h"







//global new and delete replacements to avoid need for libstdc++

void* __builtin_new(size_t size){
    return (void*) malloc(size);
}

void* __builtin_vec_new(size_t size){
    return __builtin_new( size );
}

void* __builtin_delete(void* o){
    free(o);
}

void __builtin_vec_delete( void* o ){
    __builtin_delete( o );
}

extern "C"
void __terminate(void)
{ //abort or something??
}






/**************************************************************************
 * document type functions
 */

#include "ht.h"

/*
 * return true if 'endTag' is supposed to be a valid end node of 'currentTag', 0 if not
 */
bool genericValidEndNode ( char* currentTag, char* endTag )
{
	return (strcasecmp( currentTag, endTag) == 0);
}

/*
 * return 0 if 'word' isn't a defined entity, "" if it expands to nil,
 * or otherwise its entity value
 */
char* genericExpand ( char* word )
{
	// 2DO - use a dynamically built table, here (PM)
	if ( !strcmp("test", word) ){
		return "MY TEST";
	}
	
	return 0;
}

DocType genericDocType = {
	0,     // we don't check parent-child relationships
	&genericValidEndNode,
	  
	expand//&genericExpand
};

DocType *docType = &genericDocType;


void setDocType ( char* type )
{
        // a configurable registry of some sort would be probably better (PM)
	// (but the main point is to get it out of xs.g)
	if ( type ){
		if ( !strcasecmp( type, "html") ){
			docType = &htmlDocType;
			
		}
	}
	else {
		docType = &genericDocType;
	}
}








/* generic utilities */

//need to rewrite since lib version never uses kaffe_malloc 
extern "C"
char* StrDup(char* s){

  char* d;

  if(s == NULL)
    return NULL;

  d = (char*) calloc(strlen(s) + 1, sizeof(char));
  
  if(d != NULL)
    strcpy(d, s);
  
  return d;
}  
  
  
//returns true if pattern can be found in s  
bool substring(char* s, char* pattern){

  int len = strlen(s);
  int i;
  int pat_len = strlen(pattern);
  
  for(i = 0; i < len; i++){
      if(!strncmp(s + i, pattern, pat_len)){
          return true;
      }
  }
    
  return false;
}      
  
  
//change to use perror ??
void fail(char* msg){
  fprintf(stderr, "%s\n", msg);
  //TODO different treatement when used as a Kaffe lib
  abort();
}







/******* Scanner stack for entities *******************************/

/*
DLGState** stack;
int sp;
int stacksize;


inline void initScannerStack(){

    stacksize = 5; 
    stack = (DLGState**) calloc(stacksize, sizeof(DLGState*));
    sp = 0;
}

inline void deleteScannerStack(){
    free(stack);
}


int getStackPointer(){
    return sp;
}


DLGState* pushScannerState(){
    
    if(sp == stacksize){
        int newstacksize = stacksize + 5;
	DLGState** newstack = (DLGState**) calloc(newstacksize, sizeof(DLGState*));
	memcpy(newstack, stack, stacksize*sizeof(DLGState*));
	free(stack);
	stack = newstack;
	stacksize = newstacksize;
    }
    
    DLGState* temp = new DLGState;
   
    stack[sp++] = temp;
    
    return temp;
}



DLGState* popScannerState(){

     return stack[--sp];
}

*/





/* helpers */

static void printVersion ()
{
	printf("%s v%s\n", PACKAGE, VERSION);
        printf("Copyright (c) 2001\nTransvirtual Technologies, Inc.  All rights reserved\n");
        printf("Compiled by  : %s\n",BUILD_USER);
        printf("Compiled on  : %s (%s)\n",BUILD_HOST,BUILD_ARCH);
        printf("Compile date : %s\n",BUILD_DATE);
#if defined(__GNUC__) & defined(__GNUC_MINOR__)
        printf("GCC Version  : %d.%d\n",__GNUC__,__GNUC_MINOR__);
#endif
        printf("Options      : %s\n",BUILD_OPTIONS);

}

static void printUsage ()
{
	printf("%s v%s\n", PACKAGE, VERSION);
	printf("Usage: xs [options] [<class_name> <base_class> <factory_class> <imports>]\n");
	printf("Options:\n");
	printf("   -s   - script-only mode\n");
	printf("   -v   - print version information\n");
	printf("   -h   - display this message\n");
}






#ifdef KAFFELIB

//ML_ONLY_END


/* Anonymous class management *************************************************/

#define INITIAL_MAX_TODELETE 25

char** toDelete;
int max_toDelete;
int n_todelete;


inline static void init_anonymous_class_management(void){
 
        toDelete = (char**) calloc(INITIAL_MAX_TODELETE, sizeof(char*));
	n_todelete = 0;
	max_toDelete = INITIAL_MAX_TODELETE;       
}


void remove_anonymous_classes(void){

        int i;
	char* temp; 
	
	for(i = 0; i < n_todelete; i++){
	    temp  = (char*) calloc(strlen(toDelete[i]) + 6 /*.class */ + 1, sizeof(char));
	    sprintf(temp, "%s.class", toDelete[i]); 
	    unlink(temp);
	    free(temp);
	    free(toDelete[i]);
	}
 
        free(toDelete);
}


inline static void grow_toDelete(){

        int newmax_toDelete = max_toDelete + INITIAL_MAX_TODELETE;
	char** newtoDelete = (char**) calloc(newmax_toDelete,sizeof(char*));
	memcpy(newtoDelete, toDelete, max_toDelete*sizeof(char*));
	free( toDelete);
	toDelete = newtoDelete;
	max_toDelete = newmax_toDelete;
}

extern "C"
void schedule_for_deletion(char* anonymous_class_name){

        if(n_todelete == max_toDelete){
	        grow_toDelete();
	}
	
	toDelete[n_todelete++] = anonymous_class_name;
}




//ML_ONLY_START


/*****************************************************************************
 * engines drivers and related utilities
 */
 



#include <setjmp.h>
#include <string.h>

jmp_buf jmpbuffer; //for exception handling

//#ifndef ML_ONLY
extern "C"
//#endif
void CheckException(){
    if(Genv->ExceptionOccurred())
        longjmp(jmpbuffer, 1);
        
}



JNIEnv* Genv;  //environment pointer used everywhere

jobject Gurl;  //url of parsed document

char*   GurlString;
char*   GurlFile;



//#define BUFFER_SIZE 4096
#define BUFFER_SIZE 16384



class DllExportPCCTS DLGJavaInputStream : public DLGInputStream {

 private:
        bool found_eof;
	
        jobject is;
	jclass  isclass;
	jmethodID read;
	JNIEnv* jni_env;
	
	jarray jbuf;    //Java data buffer
	
	jbyte buf[BUFFER_SIZE]; // data buffer
	
	jint buf_size;  //amount of data read
		
        short buf_idx; // index of next location to be read
	
//	jmethodID readOne;

	
 public:
        DLGJavaInputStream(JNIEnv* env, jobject istream) { 
	        is = istream; 
		
		isclass = env->GetObjectClass(istream); CheckException();
 
                read = env->GetMethodID(isclass, "read",  "([B)I"); CheckException();
		
		jbuf = env->NewByteArray((jsize)BUFFER_SIZE); CheckException();
		
		jni_env = env;
		
		found_eof = false;
		
		buf_size = 0;
		
		buf_idx = 0;  
		
	//	readOne = env->GetMethodID(istream, "read", "()I");
        }
	
        	
	int nextChar() {
	        		
	        if ( found_eof ) 
		        return EOF;
                else {
		       /* int c = jni_env->CallIntMethod(is, readOne); 
			printf("Char is %c\n", c);
			return c;
		*/
		        if(buf_idx < buf_size){
			 /* unsigned char c = buf[buf_idx++];
			  printf("Next Char is %c\n", c);
			  return (int) c;
 */
			  return (int) (unsigned char) buf[buf_idx++];   
			}
			else{
			
                          buf_size = jni_env->CallIntMethod(is, read, jbuf); CheckException();
	
			  if (buf_size == EOF){ 
			    found_eof = true;
		            return EOF;
			  }	
				  
		          jni_env->GetByteArrayRegion(jbuf, 0, buf_size, buf); CheckException();			
					   
			  buf_idx = 0;
			  
			 /* unsigned char c = buf[buf_idx++];
			  printf("Next Char is %c\n", c);
			  return (int) c;*/
			  return (int) (unsigned char) buf[buf_idx++]; 
		        }
		}
        }
	
	
	//make sure memory is allocated from kaffe's heap
	
	void* operator new(size_t size){
	    return (void*) malloc(size);
	}
	
	void operator delete(void* p, size_t size){
	    free(p);   
	}
	
	
	
};

#endif  





#include <ctype.h> //for isspace



  
class XsLexer : public DLGLexer {
 
 public:
  
 
  XsLexer (DLGFileInput *input) : DLGLexer(input/*, 5000 default is 2000*/) {}
  
  XsLexer (DLGStringInput *input) : DLGLexer(input/*, 5000*/) {}


#ifdef KAFFELIB  
  
  XsLexer (DLGJavaInputStream *input) : DLGLexer(input/*, 5000*/) {}
  
#endif  
 
  
  void errstd(const char* msg){
     warn("%s, %s, ignored", msg, lextext());
  }
  
  
  ANTLRChar * saved_token;  //not needed?  
  ANTLRTokenType saved_token_type;  
  int saved_token_line; //to do
	  
	          
  _ANTLRTokenPtr getToken() {
  
        
        
        if (token_to_fill == NULL) 
	  panic("NULL token_to_fill");
	  
        ANTLRTokenType tt;
        _ANTLRTokenPtr tk;
        
//printf("saved_token = %p\n", saved_token);

        if(!saved_token){
	    tt = nextTokenType();
	
            tk = token_to_fill->makeToken(tt, _lextext, _line);
        }
	else{
	    tt = saved_token_type;
	    tk = token_to_fill->makeToken(tt, saved_token, _line);
	    saved_token = NULL;
	}		  	
	
//ML_ONLY_END	
        if(tt == SC && NoLineTerminatorHere){
	  NoLineTerminatorHere = false;
	}  
	

	if(tt == NEWLINE && NoLineTerminatorHere){
	  tk = token_to_fill->makeToken(SC, ";", _line);
	  NoLineTerminatorHere = false;
	}  
	
	//Achtung nested functions
	if(inFunctionBody){
            if(tk->getType() != TEOF){
                fprintf(bodies_file, "%s ", tk->getText());
            }
	}
	
	
//ML_ONLY_START

        //needed ??  
	/*
        if(tt == DQUOTE && convertToTXT_WORD){
	   // printf("HERE text %s \n",tk->getText());
	    tk = token_to_fill->makeToken(TXT_WORD, tk->getText(), _line);
	    convertToTXT_WORD = false;
	} 
	*/
	  
//printf("BEFORE Got token type %d with text %s\n", tt, tk->getText());

        
        if(tt == XML_OPEN && docType == &htmlDocType){  //check if HTML, don't do for xml (others)
	
	   
	   // char* saved_lextext = StrDup(_lextext);
	    
	    saved_token_type = tt; 
	    
	   // printf("saved_lextext = %s, type = %d\n", saved_lextext, tt);
	    
	    tt = nextTokenType();
	    
	    
	    
	    if(tagIndex( _lextext) < 0){
	        //printf("A lextext is %s\n", _lextext);
		
		//printf("RAMBO\n");
		
	     //   saved_token = NULL; //not needed ?
	
	        char* temp = (char*) calloc(strlen(_lextext) + 1 + 1, sizeof(char));
                strcpy(temp, "<");
		strcat(temp, _lextext);
		
		tk = token_to_fill->makeToken(TXT_WORD, temp, _line);
		
		free(temp);  
		
		mode(TEXT_MODE);
		
		
	       // printf("Got token type %d with text %s\n", tt, tk->getText());
	    }
	    else{
	        //printf("B lextext is %s\n", _lextext);
              //  printf("Got token type %d with text %s\n", tt, tk->getText());
		
		
		saved_token_type = tt;  
	        saved_token = _lextext;
	
	    
	    }
	    
	    
	    
	  
	}





        
	
	//take tabs, newlines and carriage returns out of XML_TAGs
	//does not work since can have both fo\nnt and img\nsrc
	/*
	if(tt == XML_TAG){
	     
	    char* text = tk->getText();
	    
	    int len = strlen(text);
	  
	    char* result = (char*) calloc(len, sizeof(char));
	
	    int i, j;
	
	    for(i = 0, j = 0; i < len; i++){
	            //fastest way to do it since most will be ok??
		    if(!isspace(result[j] = text[i])){
		            j++;
		    }
	    }
	    
	    if(i != j){
	            if(j == 0){
		            //should probably check if this should be a TXT_NL
		            tk = token_to_fill->makeToken(TXT_BLANK, " ", _line);
		    }
		    else{
		            tk->setText(result);
		    }
	    }
	
	    free(result);  

	}
	*/						
	return tk;
  }
  
  //make sure memory is allocated from kaffe's heap
  
  void* operator new(size_t size){
      return (void*) malloc(size);
  }
	
  void operator delete(void* p, size_t size){
      free(p);   
  }

  

};



extern bool prettyPrint;
extern char* prettyPrintFileName;
extern FILE* prettyPrintFile;




inline void parseOptions(int argc, char* argv[]){

    int i;
       
    // parse options
    for (i=1; i<argc; i++){
 
//ML_ONLY_END   
        if ( !strcmp("-s",argv[i]) ){
            // Script-only mode (testing purposes)
            scriptOnly = true;
        } else 
//ML_ONLY_START	
	if ( !strcmp("-v",argv[i] ) ){
            printVersion();
            exit(0);
        } else if ( !strcmp("-h",argv[i] ) ){
            printUsage();
            exit(0);
	} 
//ML_ONLY_END	
	else if ( !strcmp("-f",argv[i] ) ){
	    functionOnly = true;  //implies scriptOnly
        } else if ( !strcmp("-w",argv[i] ) ){
	    writeGenerated = true;  
	}
//ML_ONLY_START		
	else if ( !strcmp("-p",argv[i] ) ){
	   // printf("GOT p\n");
	    prettyPrint = true;
	    i++;
	    if(i >= argc){
	        prettyPrintFileName = NULL; 
	    }
	    else if(argv[i][0] == '-'){
	        prettyPrintFileName = NULL;
            }
	    else{ 		    
	        prettyPrintFileName = argv[i];
	    }
	    
	    /*
	    #ifdef KAFFELIB
	    //not exactly true, then depends on if Script is there or not
	    fprintf(stderr, "pretty printing not supported in library mode, option ignored\n");
	    #endif
	    */
	    
	} 

	else break;
    }
//ML_ONLY_END
    // class name
    if ( i < argc ){
        clsName = argv[i];
        i++;
    }

    // base class name
    if ( i < argc ){
        baseName = argv[i];
        i++;
    }

    // DOM tree factory class
    if ( 
    
         !scriptOnly 
         && 
 
	 (i < argc) ){
        factoryName = argv[i];
       // i++;
    }
//ML_ONLY_START	    
}



//handles 
DLGInputStream* in;
XsLexer* scanner = NULL;
Parser* parser = NULL;


void compile (void)
{
  //  substring("loan.yahoo.com", "yahoo");


    int ret_signal;
    
    AST* scripT_ast_root = NULL;
    AST* ast_root_handle;        // needed since sorcerer seems not to rewind this pointer to the root after it traverses, so get a null leaf for the next sorcerer pass
    AST* ast_root_handle_for_destruction;
     
    Resourcer resourcer; //second pass, first AST pass
    
//ML_ONLY_END     
    Coder coder;         //third pass, second AST pass
//ML_ONLY_START     
        
    ANTLRTokenBuffer pipe(scanner);
    ANTLRToken tok;
    parser = new Parser(&pipe);
    scanner->setToken(&tok);
 
//ML_ONLY_END   
          
    if ( scriptOnly ){
      scanner->mode( scanner->SCRIPT_MODE);
    }

//ML_ONLY_START   
    
    parser->init();
    
    setbuf(stdout, NULL);
    setbuf(stderr, NULL);
    
//ML_ONLY_END  
        	
    //---- our own setup starts here
    initVars();
    initLabels();
   // initFieldHandles(); 
   
       
   
    //---- set up global XsObject props/funcs

    // use dummies instead of NULL	
    addVar( "top", T_PROP, NULL); 
    
    addVar( "window", T_PROP, NULL);
    addVar( "frames", T_PROP, NULL);
    addVar( "parent", T_PROP, NULL);
    addVar( "document", T_PROP, NULL);
    addVar( "navigator", T_PROP, NULL);
    
      
    addVar( "Array", T_FUNC, NULL);
    addVar( "Date", T_FUNC, NULL);
    
    addVar( "String", T_FUNC, NULL);
    addVar( "Number", T_FUNC, NULL);
    addVar( "Boolean", T_FUNC, NULL);
    addVar( "Object", T_FUNC, NULL);
    addVar( "Function", T_FUNC, NULL);
    
   //print, escape, etc...???
  
    
    
    #ifdef KAFFELIB
      if((bodies_file = fopen("xs.function_bodies", "a+")) == NULL){
          fail("Could not open file for function_bodies");
      }; 	        
    #else
      if((bodies_file = fopen("xs.function_bodies", "w+")) == NULL){
          fail("Could not open file for function_bodies");
      };    
    #endif	
    
//ML_ONLY_START
    
    initTagQueue();
   // initTagStack();

  //  printf("Compiler initialized \n");
    
    
//ML_ONLY_END
    
    //---- compile
    if ( scriptOnly ){
        parser->script((ASTBase **)&scripT_ast_root, &ret_signal);
    }
    else {
  
#ifndef ML_ONLY    
        parser->document((ASTBase **)&scripT_ast_root, &ret_signal);
#else
//ML_ONLY_START 
       parser->document((ASTBase **)&scripT_ast_root);
//ML_ONLY_END
#endif	               


    }
    
    
    
    fclose(bodies_file);
    
//ML_ONLY_START
	       	       
    deleteTagQueue();
   // deleteTagStack();		
  
    // TODO check ret_signal for exceptions
    /* not quite working
    switch ( ret_signal ) {
	 
	case NoViableAlt :
	case MismatchedToken :
	
	case NoSemViableAlt: printf("GOT PARSING ERROR");
		             break;  	
	case NoSignal: 	
		             break;
	default :
		fprintf(stderr, "Unknown error during parsing\n"); 
    }
*/

  //  printf("Parsing complete\n");
    
    

    if(scripT_ast_root == NULL){
        //this might also mean parse erros though, achtung
	#ifdef KAFFELIB
	  longjmp(jmpbuffer, 1);
	#else
	  fail("no input source");
        #endif
    }



    //pretty-print
  //  scripT_ast_root->preorder();  printf("\n");
   
  //  exit(0);
     
    //free parser resources
    delete parser;

   
   
    //save root aside, see above comment
    ast_root_handle_for_destruction = scripT_ast_root;
    
    
//ML_ONLY_END   
    
    #ifdef KAFFELIB
    
    if(hasScript /*|| !isHtml*/){
        
    #endif	



    //save root aside, see above comment
    ast_root_handle = scripT_ast_root;
    
    //  printf("here\n");
    //construct left and up pointers
    //at the moment, this is only used in generating array references
    //and adds an extra pass
    //override double-link since don't need the left pointers
    ast_root_handle->double_link(NULL, NULL);   
    
  //  printf("Double-linking complete\n");
    //resource



    reset.labelcount = 0;

    if(!scriptOnly){
    
        reset.offset = 2;  // xThis & _def, + this since not a static method (taken account for in emit)
        reset.localslimit = 2;
    }
    else{
        reset.offset = 3;  // xThis & _def & docfactory object, + this since not a static method (taken account for in emit)
        reset.localslimit = 3;
    }
     
    initFunctionHandles ();
    
    
    #ifdef KAFFELIB
    
    }
        
    #endif
    	
    
    
    if ( scriptOnly ){        
        resourcer.script((SORASTBase **)&scripT_ast_root); 
    }
    else{
    
//ML_ONLY_START
    
        // DOM needs no resources but still must go through the structure
        resourcer.document((SORASTBase **)&scripT_ast_root); 
	
//	printf("Resourcing complete\n");
//ML_ONLY_END
	
    } 	
    
     
  //  printf("Resourcing complete\n");
    //code
   
      
   
    #ifdef KAFFELIB
    
    if(hasScript /*|| !isHtml*/){
        
    #endif
    	
    
    
    reset.opcodes = NULL;
    reset.currenttail = NULL;
    reset.labels =  (LABEL*) malloc(reset.labelcount*sizeof(LABEL));

    if ( scriptOnly ){
       coder.script((SORASTBase **)&ast_root_handle, ALONE); 
    }
    else{
       coder.document((SORASTBase **)&ast_root_handle);
    }
    
   // printf("Coding complete\n");
    //emit
    emitDocument();
    
    
   // printf("Emitting complete\n");
    //free resources 
    deleteVars();
    
    deleteLabels();
    
    deleteFunctionHandles();
  //  deleteFieldHandles();
    
   
    
   
    #ifdef KAFFELIB
    
    }
        
    #endif	
    
//ML_ONLY_START        
    
    //some bug when have default in a switch thing
    
    ast_root_handle_for_destruction->destroy();
    
  //  printf("Compile complete\n");
   
}




//ML_ONLY_END

#ifndef KAFFELIB


//don't care too much about mem since it runs once

int main ( int argc, char* argv[] )                                     
{
 
 parseOptions(argc, argv);
 
 DLGFileInput* fileinput = new DLGFileInput(stdin);  //avoid some casting
 in = fileinput;
 scanner = new XsLexer(fileinput);
 
 InitEntities();
 
 compile();
 
 //assemble
 printf("Assembling\n");
 
 char* filename = (char*) calloc(strlen(clsName) + 1 + 2, sizeof(char));
 strcpy(filename, clsName);
 
 char* command = (char*) calloc(STRLEN("kaffe jasmin.Main ") + strlen(strcat(filename, ".j")) + 1, sizeof(char));
 //sprintf(command, "java jasmin.Main %s", filename);
 sprintf(command, "kaffe jasmin.Main %s", filename);
 
 system(command);
 
  
 //delete the jasmin file
 //unlink(filename);
 
 
}

#endif




// Java-JNI interface

#ifdef KAFFELIB

//ML_ONLY_START

#define CheckLocalException if(Genv->ExceptionOccurred()) {Genv = 0; if(strings) free(strings); delete isinput; delete scanner; return (jboolean) false;}



long number_of_compiles = 0;

extern jobject documenT;


//cached globals
jclass URL = NULL;
jmethodID getHost = NULL;
jmethodID getFile = NULL;




extern "C" {


JNIEXPORT jboolean JNICALL
#ifndef ML_ONLY
Java_kaffe_xs_XsEngine_compile
#else
Java_org_xoe_parser_html_MlEngine_compile
#endif
                              (JNIEnv* env, 
                               jclass clazz, 
                               jobject inputstream,
                               jarray cmda,
			       jobject url,
			       jarray domtree /* trick for output param*/)
{
 //save handle aside
 Genv = env;
 
// printf("HERE");
 
 //reset global variables
 if(number_of_compiles++ > 0){
 
//some of the following to take out for ml_only

//ML_ONLY_END
 
     scriptOnly = false;    
     functionOnly = false; 
     
//ML_ONLY_START    

      
     enterScript = false;
     sconstDelim = 0;
     inEventScript = false;
     inAttribute = false;
     wasTextMode = false;
 
    // isHtml = false; 
 
          
//ML_ONLY_END   
  
     //default values
     clsName     = "MyObject";
     baseName    = "kaffe/xs/XsObject";
     factoryName = "DocumentFactory";



     NoLineTerminatorHere = false;
     SrcInputStream = NULL; 

     hasScript = false;
     writeGenerated = false;
     //from code.c
     currentcode = NULL;
     
     //for switch production
     setCondition = false;
     

     //from resource.c
     label = 0;
     offset = 0;
     localslimit = 0;
     
//ML_ONLY_START      
     
     scanner = NULL;
     parser = NULL;
     
     docType = &genericDocType;

//ML_ONLY_END   

     parent_tag_set_in_this_run = false;
     
     
     
     CallStackRequired = false;
     inFunctionBody = false;

//ML_ONLY_START   
     prettyPrint = false;
     prettyPrintFileName = NULL;
     prettyPrintFile = NULL;

   

    // convertToTXT_WORD = false;
     
     didReplace = false;
 }
 
 
 else{
     InitEntities(); 
     
//ML_ONLY_END 
     init_anonymous_class_management();
     atexit(remove_anonymous_classes);
     //eventually might add call to atexit(FreeEntities)
//ML_ONLY_START 
 
 }
 


 
 
 jstring* strings = NULL;
 
 DLGJavaInputStream* isinput = NULL;
 
 
 

 //read in the args: build native argv
 
 jsize cmdc = env->GetArrayLength(cmda); CheckLocalException;

 char** args = (char**) calloc(cmdc, sizeof(char*));

 strings = (jstring*) calloc(cmdc, sizeof(jstring)); // array of objects saved aside for later release
 
 
 for(jsize i = 0; i < cmdc; i++){
   strings[i] = env->GetObjectArrayElement(cmda, i); CheckLocalException;
   args[i] =  (char*) env->GetStringUTFChars(strings[i], NULL); CheckLocalException;
 }
 
 
 parseOptions(cmdc, args);   
 
 free(args);  
  
 
 
 //for assembling, must declare before jump setup
 char* filename;
 
 jclass jasminMain; 
 jmethodID assemble;
 
 
 //save URL away
 Gurl = url;
 					 			 
 //save url string away
  					 
  
 if(URL == NULL){
     URL = env->FindClass("java.net.URL");  CheckLocalException;
 }    
 
 if(getHost == NULL){
     getHost = env->GetMethodID (URL, "getHost", "()Ljava/lang/String;");  CheckLocalException; 
 }
     	 	
 jstring urlstring = env->CallObjectMethod(url, getHost);  CheckLocalException;
 GurlString = (char*) env->GetStringUTFChars(urlstring, NULL); CheckLocalException;
 
 
 //save url file
 
// printf("URL is %s\n",GurlString); 		
	
 if(getFile == NULL){
     getFile = env->GetMethodID (URL, "getFile", "()Ljava/lang/String;");  CheckLocalException;
 } 
 	 	
 jstring urlfile = env->CallObjectMethod(url, getFile);  CheckLocalException;
 GurlFile = (char*) env->GetStringUTFChars(urlfile, NULL); CheckLocalException;	
 
// printf("URL File is %s\n",GurlFile); 

 
 
 //exception processing setup
 if(setjmp(jmpbuffer) != 0) 	
   goto GOT_EXCEPTION;

  	
  
 //setup the compiler's input stream
 isinput = new DLGJavaInputStream(env, inputstream);  //avoid casting
 in = isinput;
 scanner = new XsLexer(isinput);
 
 //compile
 compile();
 
//ML_ONLY_END 
  
 if(!hasScript/* && isHtml*/){
 
//ML_ONLY_START 

   //  printf("compile complete documenT = %p\n", documenT);
     

  
     //return the document via "output param"
     env->SetObjectArrayElement(domtree, 0, documenT); CheckLocalException;
     
     //release memory resources
     free(strings);
     delete isinput; 
     delete scanner;
     
     
     //release handle
     Genv = 0;
    // printf("RETURNING\n");
     return (jboolean) true;  //success
     
//ML_ONLY_END    
 }
 
    		    
 filename = (char*) calloc(strlen(clsName) + 2 + 1, sizeof(char));
 sprintf(filename, "%s.j", clsName);

 jasminMain = env->FindClass("jasmin.Main"); CheckLocalException;

 assemble =  env->GetStaticMethodID(jasminMain, "assemble", "(Ljava/lang/String;Ljava/lang/String;Z)V"); CheckLocalException;
 
 
 env->CallStaticVoidMethod(jasminMain, assemble, /*(jobject)*/ NULL, env->NewStringUTF(filename), (jboolean) false); CheckLocalException;

// printf("assembling complete\n");
  
 //delete the jasmin file
// unlink(filename);
 

 free(filename);
 

 //done
 
 //release handle
 Genv = 0;
 
 //release strings    causes NullpointerException. If can't fix, do not use strings array
 // for(jsize i = 0; i < cmdc; i++){
  // env->ReleaseStringUTFChars(strings[i], args[i]); CheckLocalException();
 // }
 
 //release memory resources
 free(strings);
 delete isinput; 
 delete scanner;
 
 
 
// printf("RETURNING\n");
 return (jboolean) true;  //success
 
 
//ML_ONLY_START 
 
 
 GOT_EXCEPTION:
 
               Genv = 0; 
 
               	      
               //release strings see above remark
               //for(jsize i = 0; i < cmdc; i++){
                // env->ReleaseStringUTFChars(strings[i], args[i]);CheckLocalException();
               //} 
	     //  printf("COMPILER FAILED");
	       
	       
               //release memory resources
               free(strings);
               delete isinput; 
               delete scanner; 
	       
	       
	       
	       
               return (jboolean) false;  //failure
 
}


}//extern "C" 





//ML_ONLY_END 




// Interpreter



DLGStringInput* stringinput = NULL;


static inline jobject 
interpretScript(char* script){

        int ret_signal;
	AST* scripT_ast_root = NULL; // dummy for inconsistency in framework
	
        //reset global variables
	sconstDelim = 0;
        NoLineTerminatorHere = false;
	stringinput = NULL;
	scanner = NULL;
	parser = NULL;
 
	stringinput = new DLGStringInput(script);  //avoid casting
        in = stringinput;
        scanner = new XsLexer(stringinput);
	ANTLRTokenBuffer pipe(scanner);
        ANTLRToken tok;
        parser = new Parser(&pipe);
        scanner->setToken(&tok);
        scanner->mode( scanner->SCRIPT_MODE);
        parser->init();
	
	return parser->call((ASTBase **)&scripT_ast_root, &ret_signal);
}


jobject Gcontext;  // The Ctx object
jmethodID Gmethod_args; // The call method in Ctx.call
jmethodID Gmethod_no_args; 
jobject GxThis; // the this object


#undef CheckLocalException
#define CheckLocalException if(Genv->ExceptionOccurred()) { Gcontext = 0; Genv = 0; Gmethod_args = 0; Gmethod_no_args = 0; GxThis = 0; free(script);  delete parser; delete scanner; delete stringinput; return NULL;}


extern "C" {


JNIEXPORT jobject JNICALL
Java_kaffe_xs_XsEngine_interp(JNIEnv* env, 
                              jclass clazz, 
                              jobject xsobject,
                              jobject xthis, 
                              jstring inscript)
{
 //save handles aside
 Gcontext = xsobject;
 Genv = env;        
 GxThis = xthis;

 jobject ret; // value returned by call (expression)
 

 char* script = (char*) env->GetStringUTFChars(inscript, NULL); CheckLocalException;  
    	
 jclass xsobjectclass = env->GetObjectClass(xsobject); CheckLocalException;

 
 jmethodID mid_args = env->GetMethodID(xsobjectclass, "call",  "(Ljava/lang/Object;Lkaffe/xs/XsArray;)Lkaffe/xs/XsObject;"); CheckLocalException;

 jmethodID mid_no_args = env->GetMethodID(xsobjectclass, "call",  "(Ljava/lang/Object;)Lkaffe/xs/XsObject;"); CheckLocalException;


 //save handles aside
 Gmethod_args = mid_args;
 Gmethod_no_args = mid_no_args;

 //exception processing setup
 if(setjmp(jmpbuffer) != 0)
 	goto GOT_EXCEPTION;
 
 
 ret = interpretScript(script);
 
 //not released if get exception, but only way since call resets exception status
 env->ReleaseStringUTFChars(inscript, script); //CheckLocalException(); TO PUT IN??


 free(script); 

 //lose the references
 Gcontext = 0;
 Genv = 0;        
 Gmethod_args = 0;
 Gmethod_no_args = 0;
 GxThis = 0;

 
 delete parser;
 delete scanner;
 delete stringinput;

 // printf("Interp succeeded\n");
 
 return ret;  //success
 

 GOT_EXCEPTION:
 
     //  printf("Interp failed\n");
 
     free(script); 
 
     //lose the references
     Gcontext = 0;
     Genv = 0;        
     Gmethod_args = 0;
     Gmethod_no_args = 0;
     GxThis = 0;
     
     //deleting NULL is no problem
     delete parser;
     delete scanner;
     delete stringinput;


     return NULL;  //failure
}


} // extern "C"





#endif  // KAFFELIB


//ML_ONLY_START


void warn ( char* fmt, ...)                                              
{
    va_list ap;
    
    va_start( ap, fmt);
                                         // was LT(1)->getLine()
#ifndef ML_ONLY					 
    fprintf( stderr, "%s:%d: ", clsName, scanner->line());
#else
    fprintf( stderr, "%d: ", scanner->line());  
#endif      
    vfprintf( stderr, fmt, ap);
    fputc( '\n', stderr);
    va_end( ap);    
}

//ML_ONLY_END



//System.out.println

//printf("System.out.println:\n");
   //jclass system = Genv->FindClass("java.lang.System"); CheckException(); 
   //jfieldID outid = Genv->GetStaticFieldID(system, "out","Ljava/io/PrintStream;"); CheckException();
   //jobject out =  Genv->GetStaticObjectField(system,outid);CheckException();
   //jclass printstream = Genv->FindClass("java.io.PrintStream");CheckException(); 
   //jmethodID println =  Genv->GetMethodID(printstream, "println","(Ljava/lang/Object;)V");CheckException();
   //Genv->CallObjectMethod(out, println, strings[i]); CheckException();












