//ML_ONLY_START


/*
 * Code analysis utilities
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Philippe Laporte 
 * @author Peter Mehlitz
 *
 * $tvt: util.h,v 1.8 2001/10/03 01:25:57 philippe Exp $
 */

#ifndef _UTIL_H_INCLUDED
#define _UTIL_H_INCLUDED

#include "xs_config.h"

#if defined KAFFELIB && ! defined BMATRIX_GEN

#include "jni.h"

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT


extern JNIEnv* Genv;
extern jobject Gurl;
extern char* GurlString;
extern char* GurlFile;

extern "C"
void CheckException();

#endif /* BMATRIX_GEN && ! BMATRIX_GEN */


/* generic utilities */

extern "C"
char* StrDup(char* s);

void warn ( char* fmt, ...);

extern "C"
void fail(char* message);



#define STRLEN(x) ( sizeof(x) / sizeof(char) - 1 )

//ML_ONLY_END

/* Anonymous class management *************************************************/

extern "C"
void schedule_for_deletion(char* anonymous_class_name);



// in (downwards) attributes
typedef enum {
	iNONE = 0,
	iBRK  = 1,            // break allowed
	iCNT  = 2,            // continue allowed
	iLBL = 4              //labeled   
} in_attr_t;


inline in_attr_t iadd ( in_attr_t attr, in_attr_t a )
{
	return (in_attr_t)(attr|a);
}

inline in_attr_t iloop ( in_attr_t attr )
{
	return (in_attr_t)(attr | iBRK | iCNT);
}

inline in_attr_t icase ( in_attr_t attr )
{
	return (in_attr_t)(attr | iBRK);
}
  
inline in_attr_t ilabeled ( in_attr_t attr )
{
	return (in_attr_t)(attr | iLBL);
}
 
inline in_attr_t iunlabeled ( in_attr_t attr )
{
	return (in_attr_t)(attr & ~iLBL);
}

  
inline bool itest ( in_attr_t attr, in_attr_t t )
{
	return ((attr & t) != 0);
}


	
// out (upwards) attributes
typedef enum {
        oNONE = 0,
	oBRK  = 1,  // break jump
	oCNT  = 2,  // continue jump
	oRET  = 4,  // return jump
	oDEAD = 8   // mixed jumps (ret,cnt,brk)
} out_attr_t;


inline out_attr_t oadd ( out_attr_t attr, out_attr_t a )
{
	return (out_attr_t)(attr|a);
}

inline bool otest ( out_attr_t attr, out_attr_t t )
{
	return ((attr & t) != 0);
}

inline bool isDeadEnd ( out_attr_t attr )
{
	return ((attr & (oBRK | oCNT | oRET | oDEAD)) != 0 );
}


// cond attributes
typedef enum {
	cNONE   = 0,
	cALWAYS = 1,
	cNEVER  = 2
} cond_attr_t;

inline cond_attr_t cadd ( cond_attr_t attr, cond_attr_t a )
{
	return (cond_attr_t)(attr|a);
}

inline bool ctest ( cond_attr_t attr, cond_attr_t t )
{
	return ((attr & t) != 0);
}


// X(HT)ML node types
typedef enum {
	nNONE   = 0,
	nPARENT = 1,
    nTEXT   = 2
} node_attr_t;

inline node_attr_t nadd ( node_attr_t attr, node_attr_t a )
{
	return (node_attr_t)(attr|a);
}

inline int ntest ( node_attr_t attr, node_attr_t t )
{
	return ((attr & t) != 0);
}



//problems when put this in resource.h
//for generating global methods (ex.: reset)
typedef struct TOPLEVEL {

   struct symtab* localsym;
   int labelcount;
   int offset;
   int localslimit;   
   struct CODE *opcodes;
   struct CODE *currenttail;
   struct LABEL *labels;
   
} TOPLEVEL;




//ML_ONLY_START


/***********************************************************************************
 * document type functions
 */

typedef struct {
	// structure check functions
	bool (*validChildNode) ( char* parentTag, char* child );
	bool (*validEndNode) ( char* currentTag, char* endTag );
	
	// macro expansion (e.g. entities)
	char* (*expand) ( char* entity );
} DocType;

extern DocType *docType;


void setDocType ( char* type );

#endif /* _UTIL_H_INCLUDED */


//ML_ONLY_END
