/*
 * variable management
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte 
 * 
 * $tvt: vars.c,v 1.3 2001/09/13 17:05:40 philippe Exp $
 */

#include "vars.h"
#include "symtab.h"
#include "xs_config.h"

#include <stdlib.h>
#include <string.h>

#ifdef KAFFELIB  //if used as a kaffe jni lib
#include <jni.h>
#endif

extern TOPLEVEL reset;


#define INITIAL_MAX_HANDLES 10  

//  keep an array of pointers to all symtab_t created and free them between compiles
symtab_t** handles;
int n_symtab_handles;
int maxhandles;

symtab_t* current;



void initVars ()
{
	maxhandles = INITIAL_MAX_HANDLES;
	
	handles = (symtab_t**) calloc(INITIAL_MAX_HANDLES, sizeof(symtab_t*));
	
	reset.localsym = current = st_create(10,1);
	
	n_symtab_handles = 0;
	
	handles[n_symtab_handles++] = current;
}

void growHandles ()
{
        int newMaxHandles = maxhandles + INITIAL_MAX_HANDLES;
	symtab_t** newHandles = (symtab_t**) calloc(newMaxHandles,sizeof(symtab_t*));
	memcpy(newHandles, handles, maxhandles*sizeof(symtab_t*));
	free( handles);
	handles = newHandles;
	maxhandles = newMaxHandles;
}

void deleteVars ()
{       
        int i;
	for ( i=0; i < n_symtab_handles; i++){
	        st_delete(handles[i]);
		free(handles[i]);
        }
	
        free(handles);
}

int varLevelt ( char* sym, int*type )
{
      	int i, t;
	symtab_t* temp;
	
	
	for(temp = current, i = getLevel(); temp != 0; temp = temp->next, i--){
            if (t = st_contains(temp, sym)){
               if ( type ) *type = t;
	       return i;
	    }
	}
	
	return -1;
}

void addVar ( char* sym, int type, AST* node )
{     //  fprintf(stderr, "current is %p, sym is %s, node is %p\n", current, sym, node);  
    // fprintf(stderr, "Adding %s\n", sym); 
        st_add( current, sym, type, node);	

}

void addVarTable (symtab_t* t, char* sym, int type, AST* node )
{
        st_add(t, sym, type, node);	

}

void incLevel ()
{ 
        symtab_t* t = st_create(10,1);
	
	t->next = current;
	
	current = t;

	if (  n_symtab_handles == maxhandles ){
		growHandles();
	}

        handles[ n_symtab_handles++] = t;
}

void decLevel ()
{      //check if current->next == NULL ?
        if(!current->next){
	  fail("error in decLevel");
	}
	
        current = current->next;
}

//cache the level value instead??
int getLevel()
{
  symtab_t* temp;
  int level = 0;
  for(temp =current; temp != NULL; temp = temp->next)
    level++;

  return level - 1;
}


int varLevel ( char* sym )
{
	return varLevelt( sym, 0);
}

int isVar ( char* sym )
{
	return ( varLevel( sym) >= 0 );
}

bool isVarAtLevel(char* sym)
{		
       if ( st_contains( current, sym) )
	        return true;
       else
	        return false;			
}

int getOffsetVar(char* sym)
{
        AST* node; 
    //   fprintf(stderr, "sym is %s\n",sym);  
        if ( node = st_contains_get_ast_node(current, sym) )
	        return node->offset;
        else
	        return -1;	
}

char** getAllLocalVars(symtab_t* t, int* count)
{
        char** ret = (char**) calloc(t->count, sizeof(char *));
	
	int j, i = 0;
			
	for ( j=0; j < t->len; j++ ) {
	
	        char* tested_string = t->sym[j];
		AST* tested_node = t->node[j];
	
		if ( tested_string != NULL && 
		     !tested_node->formal && 
		     strcmp(tested_string, "arguments")
		   )
	        {
	                ret[i++] = tested_string;
	        }
	}
 
        *count = i;
 
        return ret;
}


/*
void removeVar ( char* sym )
{
	if ( vars[level] ){
		st_remove( vars[level], sym);
	}
}*/
