/*
 * bmatrix.c - simple bit matrix implementation
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author Peter Mehlitz
 *
 * $tvt: bmatrix.c,v 1.4 2001/09/26 20:10:22 philippe Exp $
 */

#include <stdarg.h>
#include <stdlib.h>

#include "bmatrix.h"
#include "xs_config.h"

#if defined KAFFELIB && ! defined BMATRIX_GEN 
#include <jni.h>
#endif /* KAFFELIB && !BMATRIX_GEN */

bit_matrix_t* createBitMatrix ( int w, int h )                         /*{{{*/
{
	int n = (w + 7) / 8;

	bit_matrix_t* ba = (bit_matrix_t*) calloc( 1, n * h + sizeof(bit_matrix_t));
	ba->bits = (unsigned char*) ba + sizeof(bit_matrix_t);
	ba->l = n;
	ba->h = h;
	ba->w = w;

	return ba;
}                                                                      /*}}}*/

void freeBitMatrix ( bit_matrix_t* a )                                 /*{{{*/
{
	free( a);
}                                                                      /*}}}*/


/**********************************************************************************
 * test and aux functions (not used in compiler)
 */

#if defined BMATRIX_GEN 

#include <stdio.h>

void dumpBitMatrix ( bit_matrix_t* a )                                 /*{{{*/
{
	int i, j;
	for ( j=0; j<a->h; j++ ){
        printf( "%3d: ", j);
		for ( i=0; i<a->w; i++ ){
			putchar( isBitSet( a, i, j) ? 'o' : '.');
		}
		putchar('\n');
	}
}                                                                      /*}}}*/

void printBitMatrix ( bit_matrix_t* a, char* name )                    /*{{{*/
{
	int i, j, k=0;
//	int l1 = a->l - 1;
//	int h1 = a->h - 1;

	puts("/* automatically generated by bmatrix.c */");
	printf("bit_matrix_t %s = { %d, %d, %d, (unsigned char*)\n",
		   name, a->w, a->h, a->l);
	for ( j=0; j<a->h; j++ ){
		printf("    \"");
		for ( i=0; i<a->l; i++ ){
			printf( "\\%o", (a->bits)[k++]);
		}
		printf("\"\n");
	}
	printf("};\n");
}                                                                      /*}}}*/

void setRowBits ( bit_matrix_t* a, int row, int n, ... )               /*{{{*/
{
	va_list ap;
	int i, l, k;
	int k0 = (row * a->l); 
	

	va_start(ap, n);
	while ( n-- ){
	
		i = va_arg(ap,int);

		k = k0 + (i>>3);
		 
		l = i & 0x7;
		
		(a->bits)[k] |= (1 << l);
	}
	va_end(ap);
}                                                                      /*}}}*/

#endif /* BMATRIX_GEN */

#ifdef TEST

int main ( int argc, char* argv[] )                                    /*{{{*/
{
#define N 37
	
	int i, j;
	bit_matrix_t* b = createBitMatrix( N, N);
	
	for ( i = 0; i<N; i++ ) setBit( b, i, i);
	setRowBits(b,5,5, 1,3,5,7,9);

	dumpBitMatrix( b);
	printBitMatrix( b, "test");

	freeBitMatrix( b);

	//dumpBitMatrix( &ba);

	return 1;
}                                                                      /*}}}*/

#endif 
