/*
 * bmatrix.h - simple bit matrix implementation
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * @author Peter Mehlitz
 *
 * $tvt: bmatrix.h,v 1.2 2001/07/14 01:51:43 samc Exp $
 */

#ifndef _BMATRIX_H_INCLUDED
#define _BMATRIX_H_INCLUDED

typedef struct _bit_matrix {
	int   w, h;                /* matrix columns and rows */
        int   l;                   /* row length in bytes */
	unsigned char* bits;
} bit_matrix_t;


static inline void setBit ( bit_matrix_t* a, int i, int j )            /*{{{*/
{
	int k = (j * a->l) + (i >>3); 
	int l = i & 0x7;
	
	(a->bits)[k] |= (1 << l);
}                                                                      /*}}}*/

static inline int isBitSet ( bit_matrix_t* a, int i, int j )           /*{{{*/
{
	int k = (j * a->l) + (i >>3); 
	int l = i & 0x7;

	return ((a->bits)[k] & (1 << l));
}                                                                      /*}}}*/

bit_matrix_t* createBitMatrix ( int w, int h );
void freeBitMatrix ( bit_matrix_t* a );


/* these are just used for testing purposes or to generate data */
#if defined BMATRIX_GEN

void setRowBits ( bit_matrix_t* a, int row, int nBits, ...);
void dumpBitMatrix ( bit_matrix_t* a );
void printBitMatrix ( bit_matrix_t* a, char* name );

#endif /* BMATRIX_TEST .. */

#endif /* BMATRIX_H_INCLUDED */
