/*
 * helper functions for HTML tidy up
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte
 * 
 * $tvt: ht.c,v 1.23 2001/10/24 20:24:44 philippe Exp $
 */
 
#include <stdio.h>
#include <string.h>

#include "ht.h"
#include "bmatrix.h"

#include "xs_config.h"


#ifdef KAFFELIB  //if used as a kaffe jni lib (useless??)
#include <jni.h>
#else
#include <stdlib.h>
#endif


#include "entities.h"


/* this contains the matrix defining valid parent / child node combinations */
#include "ht_pc.h"



//the numbers here are wrong, make them right when no more tags added
char structTag[] = {                                                   
/* HTML_A           0  */ 1, 
/* HTML_ABBR        1  */ 0,
/* HTML_ACRONYM     2  */ 0,
/* HTML_ADDRESS     3  */ 0,
/* HTML_APPLET      4  */ 1, /* was 0  prompted by http://news.bbc.co.uk/ */
/* HTML_AREA        5  */ 0,
/* HTML_B           6  */ 0,
/* HTML_BASE        7  */ 0,
/* HTML_BASEFONT    8  */ 0,
/* HTML_BDO         9  */ 0,
/* HTML_BGSOUND     10 */ 0,
/* HTML_BIG         11 */ 0,
/* HTML_BLINK       12 */ 0,
/* HTML_BLOCKQUOTE  13 */ 0,
/* HTML_BODY        14 */ 1, 
/* HTML_BR          15 */ 0, 
/* HTML_BUTTON      16 */ 0,
/* HTML_CAPTION     17 */ 0,
/* HTML_CENTER      18 */ 1, /* was 0 prompted by slashdot.org */
/* HTML_CITE        19 */ 0,
/* HTML_CODE        20 */ 0,
/* HTML_COL         21 */ 0,
/* HTML_COLGROUP    22 */ 1, // was 0 prompted by www.cnn.com
/* HTML_COMMENT     23 */ 1, 
/* HTML_DD          24 */ 1,
/* HTML_DEL         25 */ 0,
/* HTML_DFN         26 */ 0,
/* HTML_DIR         27 */ 0,
/* HTML_DIV         28 */ 1,
/* HTML_DL          29 */ 0,
/* HTML_DT          30 */ 1,
/* HTML_EM          31 */ 0,
/* HTML_EMBED       32 */ 1,
/* HTML_FIELDSET    33 */ 1,
/* HTML_FONT        34 */ 1, /* was 0, prompted by slashdot.org line 500 or so*/
/* HTML_FORM        35 */ 1,
/* HTML_FRAME       36 */ 1,
/* HTML_FRAMESET    37 */ 1,
/* HTML_H1          38 */ 0,
/* HTML_H2          39 */ 0,
/* HTML_H3          40 */ 0,
/* HTML_H4          41 */ 0,
/* HTML_H5          42 */ 0,
/* HTML_H6          43 */ 0,
/* HTML_HEAD        44 */ 1,
/* HTML_HR          45 */ 0,
/* HTML_HTML        46 */ 1,
/* HTML_I           47 */ 0,
/* HTML_IFRAME      48 */ 1,
/* HTML_ILAYER      49 */ 1,
/* HTML_IMG         49 */ 0,
/* HTML_INPUT       50 */ 0, /* Was 1 */
/* HTML_INS         51 */ 0,
/* HTML_ISINDEX     52 */ 0,
/* HTML_KBD         53 */ 1,
/* HTML_LABEL       54 */ 1,
/* HTML_LAYER       55 */ 1,
/* HTML_LEGEND      56 */ 0,
/* HTML_LI          57 */ 1,  
/* HTML_LINK        58 */ 1,
/* HTML_MAP         59 */ 1,
/* HTML_MARQUEE     60 */ 0,
/* HTML_MENU        60 */ 1,
/* HTML_META        61 */ 0,
/* HTML_MULTICOL    62 */ 1,
/* HTML_NOBR        63 */ 0,
/* HTML_NOEMBED     64 */ 1,
/* HTML_NOFRAMES    65 */ 1,
/* HTML_NOSCRIPT    66 */ 1,
/* HTML_OBJECT      67 */ 1,
/* HTML_OL          68 */ 1,
/* HTML_OPTGROUP    69 */ 1,
/* HTML_OPTION      70 */ 1,
/* HTML_P           71 */ 0, 
/* HTML_PARAM       72 */ 0, 
/* HTML_PRE         73 */ 0,
/* HTML_Q           74 */ 0,
/* HTML_S           75 */ 0,
/* HTML_SAMP        76 */ 0,
/* HTML_SCRIPT      77 */ 1,
/* HTML_SELECT      78 */ 1, 
/* HTML_SMALL       79 */ 1, /* was 0 prompted by yahoo.com*/
/* HTML_SPACER      80 */ 0,
/* HTML_SPAN        81 */ 1,
/* HTML_STRIKE      82 */ 0,
/* HTML_STRONG      83 */ 0,
/* HTML_STYLE       84 */ 0,
/* HTML_SUB         85 */ 0,
/* HTML_SUP         86 */ 0,
/* HTML_TABLE       87 */ 1,
/* HTML_TBODY       88 */ 1,
/* HTML_TD          89 */ 1, 
/* HTML_TEXTAREA    90 */ 0,
/* HTML_TFOOT       91 */ 1,
/* HTML_TH          92 */ 1,
/* HTML_THEAD       93 */ 1,
/* HTML_TITLE       94 */ 0,
/* HTML_TR          95 */ 1,
/* HTML_TT          96 */ 1,
/* HTML_U           97 */ 0,
/* HTML_UL          98 */ 1,
/* HTML_VAR         99 */ 1,
/* HTML_WBR        100 */ 0
};                                                                     




/*****************************************************************************
 * internal helpers
 */

int match ( char* a, char* b, int matchVal )                           
{
	for ( ; *a && *b && ((*a | 32) == (*b | 32)); a++, b++ );
	return ( (*a == 0) && (*b == 0) ) ? matchVal : -1;
}                                                                      

inline int matchEnd ( char* s, int matchVal )                          
{
	return (*s) ? -1 : matchVal;
}                                                                      

inline int matchLast ( char* a, char b, int matchVal )                 
{
	return ( ((*a | 32) == b) && (*++a == 0) ) ? matchVal : -1;
}                                                                      




/*
 * return unique index for HTML tag 's' (used as index into parent-child-matrix)
 */
int tagIndex ( char* s )                                               /*{{{*/
{
	int ret;
	

	switch ( (*s | 32) ){
        case ('_'|32):          return match(++s,"text_", HTML_TEXT_);
	case 'a':
		switch ( (*++s | 32) ) {
		case 32:            return HTML_A;
		case 'b':           return match(++s,"br", HTML_ABBR);
		case 'c':           return match(++s,"ronym", HTML_ACRONYM);
		case 'd':           return match(++s,"dress", HTML_ADDRESS);
		case 'p':           return match(++s,"plet", HTML_APPLET);
                case 'r':           return match(++s,"ea", HTML_AREA);
		default:            return -1;
		}
	case 'b':
		switch ( (*++s | 32) ){
		case 32:            return HTML_B;
                case 'a':           ++s;
		                    if(*s != 's' && *s != 'S'){
				        return -1;
				    }
				    ++s;
				    if(*s != 'e' && *s != 'E'){
				        return -1;
				    }
                                    switch ( (*++s | 32) ){
			            case 'f':  return match(++s,"ont", HTML_BASEFONT);          
			            case ' ':  return HTML_BASE;
				    default:   return -1;
			            }
                case 'd':           return matchLast(++s,'o', HTML_BDO);
		case 'g':           return match(++s,"sound", HTML_BGSOUND);
		case 'i':           return matchLast(++s,'g', HTML_BIG);
		
		case 'l':           ret = match(++s,"ockquote", HTML_BLOCKQUOTE);
		                    if(ret < 0) ret = match(s,"ink", HTML_BLINK);
				    return ret;
				    
		case 'o':           return match(++s,"dy", HTML_BODY);
		case 'r':           return matchEnd(++s, HTML_BR);
		case 'u':           return match(++s,"tton", HTML_BUTTON);
		default:            return -1;
		}
	case 'c':
		switch ( (*++s | 32) ){
		case 'a':           return match(++s,"ption", HTML_CAPTION);
		case 'e':           return match(++s,"nter", HTML_CENTER);
		case 'i':           return match(++s,"te", HTML_CITE);
		case 'o':
			switch ( (*++s | 32) ){
			case 'd':       return matchLast(++s,'e', HTML_CODE);
			case 'l':       return (*++s) ? match(s,"group", HTML_COLGROUP) : HTML_COL;
			case 'm':       return match(++s,"ment", HTML_COMMENT);
			default:        return -1;
			}
		default:            return -1;
		}
	case 'd':
		switch ( (*++s | 32) ){
		case 'd':           return matchEnd(++s, HTML_DD);
		case 'e':           return matchLast(++s,'l', HTML_DEL);
		case 'f':           return matchLast(++s,'n', HTML_DFN);
		case 'i':
			switch ( (*++s | 32) ){
			case 'r':       return matchEnd(++s, HTML_DIR);
			case 'v':       return matchEnd(++s, HTML_DIV);
			default:        return -1;
			}
		case 'l':           return matchEnd(++s, HTML_DL);
		case 't':           return matchEnd(++s, HTML_DT);
		default:            return -1;
		}
	case 'e':
		                    ret = matchLast(++s,'m', HTML_EM);
		                    if ( ret < 0 ) ret = match(s,"mbed", HTML_EMBED);
		                    return ret;
	case 'f':
		switch ( (*++s | 32) ){
		case 'i':           return match(++s,"eldset", HTML_FIELDSET);
		case 'o':
			switch ( (*++s | 32) ){
			case 'n':       return matchLast(++s,'t', HTML_FONT);
			case 'r':       return matchLast(++s,'m', HTML_FORM);
			default:        return -1;
		}
		case 'r':
			                ret = match(++s,"ame", HTML_FRAME);
			                if ( ret < 0 ) ret = match(s,"ameset", HTML_FRAMESET);
                            return ret;
		default:            return -1;
		}
	case 'h':
		switch ( (*++s | 32) ) {
		case '1':           return matchEnd(++s, HTML_H1);
		case '2':           return matchEnd(++s, HTML_H2);
		case '3':           return matchEnd(++s, HTML_H3);
		case '4':           return matchEnd(++s, HTML_H4);
		case '5':           return matchEnd(++s, HTML_H5);
		case '6':           return matchEnd(++s, HTML_H6);
		case 'e':           return match(++s,"ad", HTML_HEAD);
		case 'r':           return matchEnd(++s, HTML_HR);
		case 't':           return match(++s,"ml", HTML_HTML);
		default:            return -1;
		}
	case 'i':
		switch ( (*++s | 32) ) {
		case 32:            return HTML_I;
		case 'f':           return match(++s,"rame", HTML_IFRAME);
		case 'l':           return match(++s,"ayer", HTML_ILAYER);
		case 'm':           return matchLast(++s,'g', HTML_IMG);
		case 'n':
			switch ( (*++s | 32) ){
			case 'p':       return match(++s,"ut", HTML_INPUT);
			case 's':       return matchEnd(++s, HTML_INS);
			default:        return -1;
			}
		case 's':           return match(++s,"index", HTML_ISINDEX);
		default:            return -1;
		}
	case 'k':               return match(++s,"bd", HTML_KBD);
	case 'l':
		switch ( (*++s | 32) ){
		case 'a':           ret = match(++s,"bel", HTML_LABEL);
			            if ( ret < 0 ) ret = match(s,"yer", HTML_LAYER);
			              return ret;
		case 'e':           return match(++s,"gend", HTML_LEGEND);
		case 'i':           return (*++s) ? match(s,"nk", HTML_LINK) : HTML_LI;
		default:            return -1;
		}
	case 'm':
		switch( (*++s | 32) ){
		case 'a':           ret = matchLast(++s,'p', HTML_MAP);
		                    if(ret < 0) ret = match(s,"rquee", HTML_MARQUEE);
				    return ret;
		case 'e':
			switch ( (*++s | 32) ){
			case 'n':       return matchLast(++s,'u', HTML_MENU);
			case 't':       return matchLast(++s,'a', HTML_META);
			default:        return -1;
			}
		case 'u':           return match(++s,"lticol", HTML_MULTICOL);	
		default:            return -1;
		}
	case 'n':
	         ret = match(++s,"oframes", HTML_NOFRAMES);
                 if ( ret < 0 ) 
		     ret = match(s,"oscript", HTML_NOSCRIPT);
		 if ( ret < 0 ) 
		     ret = match(s,"obr", HTML_NOBR);
		 if ( ret < 0 ) 
		     ret = match(s,"oembed", HTML_NOEMBED);    
                 return ret;
	case 'o':
		switch ( (*++s | 32) ){
		case 'b':           return match(++s,"ject", HTML_OBJECT);
		case 'l':           return matchEnd(++s, HTML_OL);
		case 'p':
			            ret = match(++s,"tgroup", HTML_OPTGROUP);
			            if ( ret < 0 ) ret = match(s,"tion", HTML_OPTION);
                                    return ret;
				    
		default:            return -1;
		}
	case 'p':
		switch ( (*++s | 32) ){
		case 32:            return HTML_P;
		case 'a':           return match(++s,"ram", HTML_PARAM);
		case 'r':           return matchLast(++s, 'e', HTML_PRE);
		default:            return -1;
		}
	case 'q':               return matchEnd(++s, HTML_Q);
	case 's':
		switch ( (*++s | 32) ){
		case 32:            return HTML_S;
		case 'a':           return match(++s,"mp", HTML_SAMP);
		case 'c':           return match(++s, "ript", HTML_SCRIPT);
		case 'e':           return match(++s, "lect", HTML_SELECT);
		case 'm':           return match(++s, "all", HTML_SMALL);
		
		case 'p':           ret = match(++s, "an", HTML_SPAN);
		                    if(ret < 0) ret = match(s, "acer", HTML_SPACER);
				    return ret;
		case 't':
			switch ( (*++s | 32) ){
			case 'r':
				switch ( (*++s | 32) ){
				case 'i':   return match(++s,"ke", HTML_STRIKE);
				case 'o':   return match(++s,"ng", HTML_STRONG);
				default:    return -1;
				}
			case 'y':       return match(++s,"le", HTML_STYLE);
			default:        return -1;
			}
		case 'u':
			switch ( (*++s | 32) ){
			case 'b':       return matchEnd(++s, HTML_SUB);
			case 'p':       return matchEnd(++s, HTML_SUP);
			default:        return -1;
			}
		default:            return -1;
		}
	case 't':
		switch ( (*++s | 32) ){
		case 'a':           return match(++s,"ble", HTML_TABLE);
		case 'b':           return match(++s, "ody", HTML_TBODY);
		case 'd':           return matchEnd(++s, HTML_TD);
		case 'e':           return match(++s, "xtarea", HTML_TEXTAREA);
		case 'f':           return match(++s, "oot", HTML_TFOOT);
		case 'h':           return (*++s) ? match(s,"ead", HTML_THEAD) : HTML_TH;
		case 'i':           return match(++s, "tle", HTML_TITLE);
		case 'r':           return matchEnd(++s, HTML_TR);
		case 't':           return matchEnd(++s, HTML_TT);
		default:            return -1;
		}
	case 'u':
		switch ( (*++s | 32) ) {
		case 32:            return HTML_U;
		case 'l':           return matchEnd(++s, HTML_UL);
		default:            return -1;
		}
	case 'v':               return match(++s,"ar", HTML_VAR);
	case 'w':               return match(++s,"br", HTML_WBR);
	default:                return -1;
        }
}                                                                      /*}}}*/


int isStructNode ( int idx )                                           /*{{{*/
{
       // printf("Index is %d\n", idx); 
        if(idx < 0){
	    return idx;
	}
	else{
            return structTag[idx];
	} 
}                                                                      /*}}}*/


/**********************************************************************************
 * exported functions
 */

/*
 * answer true if 'child' is a legal child node of 'parent'
 */
static bool validChildNode ( char* parent, char* child )                /*{{{*/
{
        int ret = 0;
	
/*	if(!strcasecmp("li", parent)){
	int i = 3;
	
	}
*/
	int row = tagIndex( parent);
	
	
	int col = tagIndex( child);
	
	
       // printf("row = %d, col, = %d\n", row, col);
	if ( (row >=0) && (col >= 0) ){
		ret = isBitSet( &ht_pc, col, row);
        }
	
	return (ret != 0) ? true : false;
}                                                                      /*}}}*/

/*
 * answer true if 'end' is a end-tag accepted in the context of node 'cur'
 */
static bool validEndNode ( char* cur, char* end )                       /*{{{*/
{

  /*
    if(!strcasecmp(cur, "b") && !strcasecmp(end, "a")){
                  int i = 5; 
    }
  
  */


    int row = tagIndex( cur);
    int col = tagIndex( end);
    
     
    
    if ( row == col ){
        return true;  /* a matching <tag>..</tag> at last */
    }
    else {
        /* If this is just an attribute propagation (<strong> and the like),
         * we gracefully accept it to cope with overlapping html tags. If
         * it is structural (<table> etc.), it has to match
         */
        return (isStructNode(col) == 0);
    }
}                                                                      


#include <string.h> //for atoi


char temp[2];


/*
char* expand ( char* word ){
         
     if(word[0] == '#'){ //ascii references
         temp[0] = atoi(word + 1);
	 temp[1]= '\0';
         return temp;	 
     }
     else if(!strcmp(word, "nbsp")){
         temp[0] = (char) 160;
	 temp[1]= '\0';
         return temp;	
     }//temporary
     else if(!strcmp(word, "lt")){
         return "<";           
     }
     else if(!strcmp(word, "copy")){
         temp[0] = (char) 169;
	 temp[1]= '\0';
         return temp;           
     }
     /*//*/recursive test
     else if(!strcmp(word, "somemore")){
         return "&gt;";
     }*//*
     else if(!strcmp(word, "gt")){
         return ">";
     }
     else if(!strcmp(word, "amp")){
         return "&";
     }
     else if(!strcmp(word, "apos")){
         return "'";
     }
     else if(!strcmp(word, "quot")){
         // TO TEST 
         #ifdef KAFFELIB
             temp[0] = (char) 34;
	     temp[1]= '\0';
             return temp;
	 #else
	     return "\\\"";
	 #endif
     }
     else 
         return "UNKNOWN";
}                                                                      
*/


char* expand ( char* word ){
     //printf("word is %s\n", word);    
	 
	 
     //quotes will be wrong for jasmin
     if((temp[0] = EntityCode(word)) != 0){
     
         temp[1]= '\0';
	/* if(temp[0] == 34){
	  int i = 6;
	 }
	 printf("temp is %s\n", temp);*/
	// printf("returning %s\n", temp);
         return temp;
     }
     else{
        // printf("returning word\n");	 
         return word;
     }	 
}                                 

DocType htmlDocType = {
	validChildNode,
	validEndNode,
	expand
};
