/*
 * buffer management
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte
 * 
 * $tvt: srcgen.c,v 1.4 2001/07/14 01:51:43 samc Exp $
 */

#include "srcgen.h"
#include "xs_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef KAFFELIB  //if used as a kaffe jni lib
#include <jni.h>
#endif

sbuf_stack* cur_stack;  //current buffer

sbuf_stack* tmp_stack;  //for arrays


sbuf_t* push_btmp(){
    
	if(tmp_stack->sp == tmp_stack->size)
	  grow_stack(tmp_stack);
    
        return tmp_stack->stack[tmp_stack->sp++] = create_sbuf(100);
}    
 
sbuf_t* pop_btmp(){

        return tmp_stack->stack[--tmp_stack->sp];
}    
 
sbuf_stack* create_sbuf_stack(int size){

        sbuf_stack* temp = (sbuf_stack *) malloc( sizeof(sbuf_stack));
        temp->size = size;
        temp->sp = 0;
        temp->stack = (sbuf_t **) malloc( size * sizeof(sbuf_t*));
        return temp;
}

void grow_stack (sbuf_stack* stack)
{
	int   newSize = stack->size * 3 / 2;
	sbuf_t** newstack = (sbuf_t**) malloc( newSize * sizeof(sbuf_t*));
	memcpy( newstack, stack->stack, stack->size);

	free(stack->stack);
	stack->stack = newstack;
	stack->size = newSize;
}

sbuf_t* bcur;

sbuf_t* create_sbuf ( int size )
{
	sbuf_t* b = (sbuf_t*) malloc( sizeof(sbuf_t));
	b->buf = (char*) malloc( size);
	b->size = size;
	b->pos = 0;
	b->indent = 0;
	
	return b;
}

void delete_sbuf ( sbuf_t* b )
{
	free( b->buf);
	free( b);
}

void grow_sbuf ( sbuf_t* b )
{
	int   newSize = b->size * 3 / 2;
	char* newBuf = (char*) malloc( newSize);
	memcpy( newBuf, b->buf, b->pos);

	free( b->buf);
	b->buf = newBuf;
	b->size = newSize;
}

void write_sbuf ( sbuf_t* b, char* fname )
{      
	int fd = STDOUT_FILENO;

	if ( fname ){
		fd = open( fname, O_CREAT|O_TRUNC|O_WRONLY,
				   S_IROTH|S_IRGRP|S_IRUSR|S_IWUSR);
                if( fd < 0){
		  char msg[100];
		  sprintf(msg, "unable to open file %s for writing", fname);
		  perror(msg);
		  exit(1);
	        }					   
	}
	
	write( fd, b->buf, b->pos);

	if( fd != STDOUT_FILENO) {
		close( fd);
	}
}


/*************************************************************************
 * output routines
 */

static char  indent_str[] = { ' ', ' '};
static int   ilen = sizeof(indent_str);


void voutb ( sbuf_t* b, char* fmt, va_list ap )
{
	int n, rem;

	while ( (n = vsnprintf( b->buf + b->pos,(rem = b->size - b->pos), fmt, ap)) >= rem)
	        grow_sbuf( b);
	       	
	b->pos += n;
//	vfprintf(stderr, fmt, ap);
}

void lnb ( sbuf_t* b )
{
	int n, i, j;

	n = b->indent * sizeof(indent_str) + 1;

	while ( (b->pos + n) > b->size ){
		grow_sbuf( b);
	}

	b->buf[b->pos++] = '\n';
	for ( i=0; i < b->indent; i++ ) {
		for ( j=0; j<ilen; j++ ){
			b->buf[b->pos++] = indent_str[j];
		}
	}
}


void lout ( char* fmt, ... )
{
	va_list ap;
	
	lnb( bcur);
        va_start( ap, fmt);
	voutb( bcur, fmt, ap);
	va_end( ap);	
}

void mergeb ( sbuf_t* dest, sbuf_t* src )
{
	while ( dest->size < (dest->pos + src->pos) ){
		grow_sbuf( dest);
	}

	memcpy( dest->buf + dest->pos, src->buf, src->pos);
	dest->pos += src->pos;
}

inline void writeb ( sbuf_t* b, int c )
{
	if ( b->pos == b->size ){
		grow_sbuf( b);
	}
	b->buf[b->pos++] = c;
}

void mergeb_str ( sbuf_t* dest, sbuf_t* src )
{
	int j, c;
	
	for ( j=0; j<src->pos; j++ ){
		c = src->buf[j];
		if ( c == '"' ){
			writeb( dest, '\\');
			writeb( dest, '"');
		}
		else {
			writeb( dest, c);
		}
	}
}


/************************************************************************
 * test stuff
 */

#ifdef TEST_SRCGEN

int main ( int argc, char* argv[] )
{
	sbuf_t* b1 = create_sbuf( 10);
	sbuf_t* b2 = create_sbuf( 10);

	set_sbuf( b1);

	out( "%s", "aaaa.aaaa");
	incInd();
	lout( "%s", "aaaaa");

	incInd(); ln();
	out( "%s", "aaaaa");
	decInd();

	setInd( set_sbuf( b2));
	
	out("%s", "bbbb.bbbb");
	incInd();
	ln();
	out( "%s", "bbbb.bbb");
	lout( "%s", "bbbb.bb");
	decInd(); ln();

	last_sbuf();
	
	ln();
	merge( b2);
	decInd();
	lout( "%s", "aaaaaaaa");
	ln();
	
	write_sbuf( b1, 0);
	delete_sbuf( b1);
	delete_sbuf( b2);

	return 0;
}

#endif
