/*
 * buffer management
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Philippe Laporte 
 * 
 * $tvt: srcgen.h,v 1.4 2001/07/14 01:51:43 samc Exp $
 */

#ifndef _SRCGEN_H_INCLUDED
#define _SRCGEN_H_INCLUDED

#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
	char* buf;
	int pos;
	int size;
	int indent;
} sbuf_t;

sbuf_t* push_btmp();
sbuf_t* pop_btmp();

extern sbuf_t* bcur;

sbuf_t* create_sbuf ( int size );
void grow_sbuf ( sbuf_t* b );
void delete_sbuf ( sbuf_t* b );
void write_sbuf ( sbuf_t* b, char* fname );

typedef struct {
        int sp;
	int size;
	sbuf_t** stack;	
} sbuf_stack;

sbuf_stack* create_sbuf_stack(int size);

void grow_stack (sbuf_stack* stack);

extern sbuf_stack* cur_stack;

inline void set_sbuf ( sbuf_t* b )
{
       	if(cur_stack->sp == cur_stack->size)
	  grow_stack (cur_stack);

	cur_stack->stack[cur_stack->sp++] = bcur;
	
	bcur = b;
}

inline sbuf_t* get_sbuf (void)
{
	return bcur;
}

inline void last_sbuf ()
{
        bcur = cur_stack->stack[--cur_stack->sp];
}

inline void reset_sbuf ( sbuf_t* b )
{
	b->pos = 0;
	b->indent = 0;
}

inline void setIndb( sbuf_t* b, sbuf_t* from )
{
	b->indent = from->indent;
}

inline void setInd ( sbuf_t* b )
{
	setIndb( bcur, b);
}

inline void setIndLevelb ( sbuf_t* b, int lvl )
{
	b->indent = lvl;
}

inline void setIndLevel ( int lvl )
{
	setIndLevelb( bcur, lvl);
}
	
inline void incIndb ( sbuf_t* b )
{
	b->indent++;
}

inline void incInd ()
{
	incIndb( bcur);
}

inline void decIndb ( sbuf_t* b )
{
	b->indent--;
}

inline void decInd ()
{
	decIndb( bcur);
}

inline int getPosb ( sbuf_t* b )
{
	return b->pos;
}
	
inline int getPos ()
{
	return getPosb( bcur);
}
	
inline void setPosb ( sbuf_t* b, int p )
{
	if ( (p >= 0) && (p < b->pos) ){
		b->pos = p;
	}
}

inline void setPosbOffset ( sbuf_t* b, int p )
{
       	if ( b->pos + p >= 0 ){
		b->pos += p;
	}
}

inline void setPos ( int p )
{
	setPosb( bcur, p);
}

inline void setPosOffset ( int p )
{
	setPosbOffset( bcur, p);
}

	
/********************************************************************************
 * output routines
 */

void voutb ( sbuf_t* b, char* fmt, va_list ap );
void lnb ( sbuf_t* b );
void lout ( char* fmt, ... );
void mergeb ( sbuf_t* dest, sbuf_t* src );
void mergeb_str ( sbuf_t* dest, sbuf_t* src );
	
inline void ln ()
{
	lnb( bcur);
}

inline void chrb ( sbuf_t* b, char c )
{
	if ( b->pos == b->size ){
		grow_sbuf( b);
	}
	b->buf[b->pos++] = c;
}

inline void chr( char c ){
	chrb( bcur, c);
}

inline void outb ( sbuf_t* b, char* fmt, ... )
{
	va_list ap;

	va_start( ap, fmt);
	voutb( b, fmt, ap);
	va_end( ap);
}

inline void out ( char* fmt, ... )
{
	va_list ap;

	va_start( ap, fmt);
	voutb( bcur, fmt, ap);
	va_end( ap);	
}

inline void merge ( sbuf_t* b )
{
	mergeb( bcur, b);
}
	
inline void merge_str ( sbuf_t* b )
{
	mergeb_str ( bcur, b);
}
	
#ifdef __cplusplus
}
#endif

#endif /* _SRCGEN_H_INCLUDED */
