/*
 * Code analysis utilities
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Philippe Laporte 
 * 
 * $tvt: util.h,v 1.4 2001/07/14 01:51:43 samc Exp $
 */

#ifndef _UTIL_H_INCLUDED
#define _UTIL_H_INCLUDED

extern "C"
char* StrDup(char* s);

void warn ( char* fmt, ...);

// in (downwards) attributes
typedef enum {
	iNONE = 0,
	iBRK  = 1,            // break allowed
	iCNT  = 2             // continue allowed
} in_attr_t;


inline in_attr_t iadd ( in_attr_t attr, in_attr_t a )
{
	return (in_attr_t)(attr|a);
}

inline in_attr_t iloop ( in_attr_t attr )
{
	return (in_attr_t)(attr | iBRK | iCNT);
}

inline in_attr_t icase ( in_attr_t attr )
{
	return (in_attr_t)(attr | iBRK);
}
  
inline bool itest ( in_attr_t attr, in_attr_t t )
{
	return ((attr & t) != 0);
}


	
// out (upwards) attributes
typedef enum {
        oNONE = 0,
	oBRK  = 1,  // break jump
	oCNT  = 2,  // continue jump
	oRET  = 4,  // return jump
	oDEAD = 8   // mixed jumps (ret,cnt,brk)
} out_attr_t;


inline out_attr_t oadd ( out_attr_t attr, out_attr_t a )
{
	return (out_attr_t)(attr|a);
}

inline bool otest ( out_attr_t attr, out_attr_t t )
{
	return ((attr & t) != 0);
}

inline bool isDeadEnd ( out_attr_t attr )
{
	return ((attr & (oBRK | oCNT | oRET | oDEAD)) != 0 );
}


// cond attributes
typedef enum {
	cNONE   = 0,
	cALWAYS = 1,
	cNEVER  = 2
} cond_attr_t;

inline cond_attr_t cadd ( cond_attr_t attr, cond_attr_t a )
{
	return (cond_attr_t)(attr|a);
}

inline bool ctest ( cond_attr_t attr, cond_attr_t t )
{
	return ((attr & t) != 0);
}



void closeParen ();



/*************************************************************************
 * document type functions
 */

typedef struct {
	// structure check functions
	bool (*validChildNode) ( char* parentTag, char* child );
	bool (*validEndNode) ( char* currentTag, char* endTag );
	
	// macro expansion (e.g. entities)
	char* (*expand) ( char* entity );
} DocType;

extern DocType *docType;


int setDocType ( char* type );

#endif /* _UTIL_H_INCLUDED */
