/*
 * Scanner, parser, AST building and interpreter
 *
 * Copyright (c) 2001
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 *
 * @author Peter Mehlitz
 * @author Philippe Laporte
 * 
 *  
 */
 
 

#header                                                                 /*{{{*/
<<

#include "util.h"
#include "srcgen.h"
#include "xs_config.h"

#include <stdarg.h>
#include <DLexerBase.h>
  
extern char  sconstDelim;
extern bool  enterScript;
extern bool  inEventScript;
extern bool  inAttribute;
extern bool  wasTextMode;
extern int   handles_number;

extern bool  scriptOnly;
extern bool  functionOnly;
extern char* clsName;
extern char* baseName;
extern char* factoryName;

extern sbuf_t* bcur;

extern sbuf_stack* tmp_stack;
extern sbuf_stack* cur_stack;

extern int levelLabels;

extern DLGState saved_state;
 
extern bool NoLineTerminatorHere;

extern bool memberLeftOp;



#ifdef KAFFELIB  //if used as a kaffe jni lib



#include <jni.h>

#undef JNIEXPORT
#define JNIEXPORT NATIVE_EXPORT

extern jobject SrcInputStream;

#include <setjmp.h>
extern jmp_buf jmpbuffer;


#else  //if used as a stand-alone process

#define jobject void*
#define jarray void*
#define jclass void*
#define jmethodID int
#define jfieldID int
#define jdouble double
#define jstring char*
#define jsize int

#endif

>>
/*}}}*/

<<
/* globals, includes etc. */                                            /*{{{*/


#include <stdlib.h>

typedef ANTLRCommonToken ANTLRToken;

#include "DLGLexer.h"

#include "vars.h"
#include "labels.h"
#include "srcgen.h"

bool  scriptOnly;    // this is only a script, don't parse XML when true
bool  functionOnly;  //only generating a function, ex: Function contructors

bool  enterScript;
char  sconstDelim;
bool  inEventScript;
bool  inAttribute;
bool  wasTextMode;

int   openParen;
bool  rawLiteral;    // whether to convert next STR_CONST, NUM_CONST into XsObjects

int   withLevel;             // level of with blocks
bool  statCall = false;      // direct static call to follow

char* clsName     = "MyObject";
char* baseName    = "XsObject";
char* factoryName = "DocumentFactory";

sbuf_t *bmain;   // the main buffer (where everything is merged, eventually)
sbuf_t *binit;   // this is where all toplevel expressions go
sbuf_t *btmp;    // temp expr buffer (postfix ops and the like)

sbuf_t *bfor;    // buffer for for loop code
sbuf_t *blocals; // buffer for function locals

sbuf_t *bdom;    // buffer to collect the DOM2 stuff 

// dummmy used for generation of condExp 
bool dummy_declared = false;
bool dummy_to_declare = false;


//for external script file processing
DLGState saved_state;

#ifdef KAFFELIB

jobject SrcInputStream = NULL;

#endif


//Virtual semicolon management
bool NoLineTerminatorHere = false;  


int Pos; // position of beginning of current stmt


bool memberLeftOp = false;


/*}}}*/

class XsParser : public Parser                                          /*{{{*/
{
public:
  XsParser (ANTLRTokenBuffer *input) : Parser(input) {}

};
/*}}}*/


/* main function and other auxiliaries (avoid the 4000 limit) */
#include "util.c"


>>




/******************************************************************************
 * Lexer part
 */

/* XML part */                                                          /*{{{*/

#lexclass START /********************************************************/
  
#token  "[\ \t]+"                  << skip(); >>
#token  "[\r\n]"                   << skip(); newline(); >>
#token  "\<!\-\-"                  << skip(); mode(XML_COMMENT_MODE); >>

//#token  REOF "@"  
                 
#token  XML_END_OPEN   "\</"
#token  XML_OPEN       "\<"
#token  XML_META_OPEN  "\<\?"
#token  XML_DTD_OPEN   "\<\!"
  
#token  SCRIPT_TAG     "script"    << enterScript = true; >>

#token  ONCLICK        "[oO][nN][cC][lL][iI][cC][kK]"     
#token  ONLOAD         "[oO][nN][lL][oO][aA][dD]"    
#token  ONMOUSEOVER    "[oO][nN][mM][oO][uU][sS][eE][oO][vV][eE][rR]"
#token  ONMOUSEOUT     "[oO][nN][mM][oO][uU][sS][eE][oO][uU][tT]"

#tokclass EVENT_TAG { ONCLICK ONLOAD ONMOUSEOVER ONMOUSEOUT }

#token  XML_TAG        "~[\=\<\>\"\'\ \t\n\r\?]+"
  
//#token  XML_ATTR       "~[\=\<\>\"\'\ \t\n\r]+"

#token  DQUOTE         "[\"\']"    << if ( ch == '\'' ) {
                                        replchar( '\"');
                                      }
                                      
                                      sconstDelim = ch;
  
                                      if ( inEventScript ){
                                        mode(SCRIPT_MODE);
                                      }
                                      else if ( inAttribute ) {
                                        mode(XML_STRING_MODE); more();
                                      }
                                   >>
  
#token  ASSIGN         "\="
#token  XML_ELEM_CLOSE "/\>"       << enterScript = false;
                                      mode( TEXT_MODE);
                                   >>
#token  XML_CLOSE      "\>"        << if ( enterScript ) {
                                        enterScript = false;
                                        mode( SCRIPT_MODE);
                                      }
                                      else {
                                        mode( TEXT_MODE);
                                      }
                                   >>

#token XML_META_CLOSE "\?\>"
  


#lexclass TEXT_MODE /************************************************/

#token TXT_BLANK  "[\ \t]+"
#token TXT_NL     "[\r\n]"         << newline(); >>
#token TXT_WORD   "~[\ \t\r\n\<\&\;]+"
#token TXT_AMP    "\&"
#token TXT_SEMI   "\;"
#token  "\<!\-\-"                 << skip(); mode(XML_COMMENT_MODE); wasTextMode=true; >>
  
  
/* thin ice, but the most intuitive solution - we have to get back
 * into START mode, but must not gobble the '<', so we simply rely
 * on DlgLexer to use the same token ID if we name the beast like in
 * the START automaton
 */
#token  XML_END_OPEN   "\</"       << mode(START); >>
#token  XML_OPEN       "\<"        << mode(START); >>
#token  XML_META_OPEN  "\<\?"      << mode(START); >>

#tokclass TXT_SKIP {
  TXT_BLANK TXT_NL
}

#lexclass XML_COMMENT_MODE /*****************************************/
                                   /* this is a regexpr weirdo - we need to recognize
                                    * '-->' and skip all other '-', but unfortunately
                                    * this is a prefix of the longer regexpr, and dlg for
                                    * some reason isn't looking for the longest match
                                    */
#token "\-+{\>}"                   << if ( (*_endexpr == '>') && (_begexpr[1] == '-')  ){
                                        if ( wasTextMode ){
                                          mode(TEXT_MODE); wasTextMode=false;
                                        }
                                        else {
                                          mode(START);
                                        }
                                      }
                                      skip();
                                   >>
#token "[\r\n]"                    << skip(); newline(); >>
#token "~[\-\r\n]"                 << skip(); >>


#lexclass   XML_STRING_MODE /****************************************/

#token  XML_STRING "[\"\']"        << if ( ch == sconstDelim ){
                                        if ( ch == '\'' ){
                                          replchar('\"'); 
                                        }
                                        mode(START);
                                      }
                                      else {
                                        more();
                                      }
                                   >>
                   
#token  "\\\""                     << replstr ("\\\""); more (); >>
#token  "\\~[]"                    << replchar (ch); more (); >> 
#token  "~[\"\'\\\n\r]+"           << more(); >>
  
/*}}}*/  

/* script part */                                                       /*{{{*/
  
#lexclass SCRIPT_MODE /**************************************************/

#token  "[\ \t]+"                  << skip(); >>

#token  NEWLINE "[\r\n]"           <<
                                      if(!NoLineTerminatorHere)
                                        skip();
                                     
                                      newline(); 
                                   >>
                                   
#token  TEOF "@"                   << 
                                      /* must not do this if in interp!!!
                                      closeParen();
                                      lout(";");
                                      */

                                      #ifdef KAFFELIB
                                        if(SrcInputStream != NULL){
                                          restoreState(&saved_state);
                                          skip();
                                        } 
                                      #endif            
                                   >>
                                   

#token  "//~[\n]*"                 << skip(); >>   // single line comment does not include NEWLINE

#token  "\<!\-\-~[\n]*"            << skip(); >>   // same for this one 
  
#token  "/\*"                      << skip(); mode(C_COMMENT_MODE); >>
  
/*}}}*/

/* C comment mode */                                                    /*{{{*/

#lexclass   C_COMMENT_MODE /*********************************************/

#token  "\*/"                      << skip (); mode (SCRIPT_MODE);  >>
#token  "[\r\n]"                   << skip (); newline(); >>
#token  "~[]"                      << skip (); >>

#lexclass SCRIPT_MODE /**************************************************/
  
/*}}}*/

/* non-op symbols */                                                    /*{{{*/
  
/* this brings us back into XML mode */
#token  SCRIPT_END "\</script\>"   << mode( START); >>

#token  JAVA_START "\<java\>"
#token  JAVA_END   "\</java\>"
  
#token  DOT       "\."   
#token  LBRACK    "\["
#token  RBRACK    "\]"
#token  LBRACE    "\{"
#token  RBRACE    "\}"
#token  LPAREN    "\("
#token  RPAREN    "\)"
#token  SC        "\;"  
#token  COMMA     "\,"
#token  COLON     "\:"

#tokclass SYMBOL {
  DOT LBRACK LBRACE LPAREN RBRACK RBRACE RPAREN SC COMMA
}

#tokclass BRACE {
  LBRACE RBRACE
}

/*}}}*/

/* operators */                                                         /*{{{*/
#token  STRICT_EQUAL "\=\=\="
#token  EQUAL        "\=\="
#token  ASSIGN       "\="
#token  STRICT_NOTEQUAL "!\=\="
#token  NOT_EQUAL    "!\="
#token  LOG_NOT      "!"
#token  USHR_ASSIGN  "\>\>\>\="
#token  USHR         "\>\>\>"
#token  SHR_ASSIGN   "\>\>\="
#token  SHR          "\>\>"
#token  GE           "\>\="
#token  GT           "\>"
#token  SHL_ASSIGN   "\<\<\="
#token  SHL          "\<\<"
#token  LE           "\<\="
#token  TLT           "\<"                 // beware of the LT(k) function
#token  MINUS_ASSIGN "\-\="
#token  DEC          "\-\-"
#token  MINUS        "\-"
#token  PLUS_ASSIGN  "\+\="
#token  INC          "\+\+"
#token  PLUS         "\+"
#token  MOD_ASSIGN   "%\="
#token  MOD          "%"
#token  MUL_ASSIGN   "\*\="
#token  MUL          "\*"
#token  DIV_ASSIGN   "\/\="
#token  DIV          "\/"
#token  OR_ASSIGN    "\|\="
#token  LOG_OR       "\|\|"
#token  OR           "\|"
#token  AND_ASSIGN   "&\="
#token  LOG_AND      "&&"
#token  AND          "&"
#token  XOR_ASSIGN   "^\="
#token  XOR          "^"
#token  COND         "\?"
#token  NOT          "\~"

#tokclass OPERATOR {
  STRICT_EQUAL EQUAL ASSIGN STRICT_NOTEQUAL NOT_EQUAL NOT GE GT LE TLT
  MINUS_ASSIGN DEC MINUS PLUS_ASSIGN INC PLUS MOD_ASSIGN MOD MUL_ASSIGN MUL
  DIV_ASSIGN DIV OR_ASSIGN LOG_OR OR AND_ASSIGN LOG_AND AND XOR_ASSIGN XOR
  USHR_ASSIGN USHR SHR_ASSIGN SHR SHL_ASSIGN SHL COND LOG_NOT
}

/*}}}*/

/* keywords */                                                          /*{{{*/
#token  IF        "if"
#token  ELSE      "else"
#token  FOR       "for"
#token  IN        "in"
#token  WHILE     "while"
#token  CONTINUE  "continue"
#token  RETURN    "return"
#token  DO        "do"
#token  BREAK     "break"
#token  SWITCH    "switch"
#token  CASE      "case"
#token  DEFAULT   "default"
#token  WITH      "with"
#token  VAR       "var"
#token  FUNCTION  "function"
#token  TRY       "try"
#token  THROW     "throw"
#token  CATCH     "catch"
#token  FINALLY   "finally"
#token  NEW       "new"
#token  DELETE    "delete"
#token  VOID      "void"
#token  TYPEOF    "typeof"
#token  INSTANCEOF "instanceof"
#token  TNULL     "null"
#token  TTRUE     "true"
#token  TFALSE    "false"
#token  TTHIS     "this"


// future keywords

#token "abstract"
#token "enum"
#token "int"
#token "short"
#token "boolean"
#token "export"
#token "interface"
#token "static"
#token "byte"
#token "extends"
#token "long"
#token "super"
#token "char"
#token "final"
#token "native"
#token "synchronized"
#token "class"
#token "float"
#token "package"
#token "throws"
#token "const"
#token "goto" 
#token "private"
#token "transient"
#token "debugger"
#token "implements"
#token "protected"
#token "volatile"
#token "double"
#token "import"
#token "public"


#tokclass KEYWORD {
  IF ELSE FOR IN WHILE CONTINUE RETURN 
  DO BREAK SWITCH CASE DEFAULT WITH VAR FUNCTION
  TRY THROW CATCH FINALLY NEW DELETE VOID TYPEOF INSTANCEOF
  TTHIS TNULL TTRUE TFALSE
}

/*}}}*/

/* static properties */                                                 /*{{{*/

/* Math */

/* methods */
#token MATH_ABS     "Math.abs"                           
#token MATH_ACOS    "Math.acos"
#token MATH_ASIN    "Math.asin"
#token MATH_ATAN    "Math.atan"
#token MATH_ATAN2   "Math.atan2"
#token MATH_CEIL    "Math.ceil"
#token MATH_COS     "Math.cos"
#token MATH_EXP     "Math.exp"
#token MATH_FLOOR   "Math.floor"
#token MATH_LOG     "Math.log"
#token MATH_MAX     "Math.max"
#token MATH_MIN     "Math.min"
#token MATH_POW     "Math.pow"
#token MATH_RANDOM  "Math.random"
#token MATH_ROUND   "Math.round"
#token MATH_SIN     "Math.sin"
#token MATH_SQRT    "Math.sqrt"
#token MATH_TAN     "Math.tan"

/*fields*/
#token MATH_E       "Math.E"
#token MATH_PI      "Math.PI"
#token MATH_LN10    "Math.LN10"
#token MATH_LN2     "Math.LN2"
#token MATH_LOG10E  "Math.LOG10E"
#token MATH_LOG2E   "Math.LOG2E"
#token MATH_SQRT1_2 "Math.SQRT1_2"
#token MATH_SQRT2   "Math.SQRT2"
       


#tokclass MATH_STATIC_METHODS{
  MATH_ABS    MATH_ACOS   MATH_ASIN   MATH_ATAN
  MATH_ATAN2  MATH_CEIL   MATH_COS    MATH_EXP  
  MATH_FLOOR  MATH_LOG    MATH_MAX    MATH_MIN  
  MATH_POW    MATH_RANDOM MATH_ROUND  MATH_SIN  
  MATH_SQRT   MATH_TAN 
}

#tokclass MATH_STATIC_FIELDS{
  MATH_E MATH_PI MATH_LN10 MATH_LN2 MATH_LOG10E MATH_LOG2E MATH_SQRT1_2 MATH_SQRT2
}




/* Number */

/* methods */

/*fields*/
#token NUMBER_MAX_VALUE         "Number.MAX_VALUE"
#token NUMBER_MIN_VALUE         "Number.MIN_VALUE"
#token NUMBER_NaN               "Number.NaN"
#token NUMBER_POSITIVE_INFINITY "Number.POSITIVE_INFINITY"
#token NUMBER_NEGATIVE_INFINITY "Number.NEGATIVE_INFINITY"



/* others */

/* fields */
#token NaN         "NaN"
#token INFINITY    "Infinity"

/* methods */
#token ISNAN       "isNaN"
#token ISFINITE    "isFinite"
#token PARSEFLOAT  "parseFloat"
#token PARSEINT    "parseInt"

#tokclass NUMBER_STATIC_METHODS{
  ISNAN ISFINITE PARSEFLOAT PARSEINT
}

#tokclass NUMBER_STATIC_FIELDS{
  NUMBER_MAX_VALUE NUMBER_MIN_VALUE NUMBER_NaN NUMBER_POSITIVE_INFINITY NUMBER_NEGATIVE_INFINITY 
  NaN INFINITY
}



/* String */

/* fields */

/* methods */
#token FROMCHARCODE   "String.fromCharCode"


#tokclass STRING_STATIC_METHODS{
  FROMCHARCODE
}


/* orther globals */

/* methods */
#token PRINT      "print"
#token ESCAPE     "escape"
#token UNESCAPE   "unescape"


//cannot put eval since can be invoked through an object

#tokclass OTHER_STATIC_METHODS{
  PRINT ESCAPE UNESCAPE
}


/* Converters */

/* methods */
#token OBJECT   "Object"
#token BOOLEAN  "Boolean"
#token STRING   "String"
#token NUMBER   "Number"

#tokclass CONVERTER_STATIC_METHODS{
  OBJECT BOOLEAN STRING  NUMBER
}





/* all */



#tokclass STATIC_PROP_METHOD{
  MATH_STATIC_METHODS NUMBER_STATIC_METHODS STRING_STATIC_METHODS CONVERTER_STATIC_METHODS OTHER_STATIC_METHODS
}

#tokclass STATIC_PROP_FIELD{
  MATH_STATIC_FIELDS NUMBER_STATIC_FIELDS 
}


/*}}}*/

/* identifier*/                                                         /*{{{*/

#token  ID           "[a-zA-Z_\$][a-zA-Z0-9_\$]*"

/*}}}*/

/* num literals */                                                      /*{{{*/

#token  EXP_CONST    "{[0-9]*}\.[0-9]+[eE]{[\+\-]}[0-9]+"
#token  FLT_CONST    "{[0-9]*}\.[0-9]+"
#token  INT_CONST    "[0-9]+"
#token  HEX_CONST    "0[xX][0-9a-fA-F]+"
#token  OCT_CONST    "0[0-7]*"

#tokclass NUM_CONST { INT_CONST HEX_CONST FLT_CONST EXP_CONST OCT_CONST}

/*}}}*/

/* string literal mode */                                               /*{{{*/

// again, the token name unification trick (START mode)
#token  DQUOTE "\""                << if ( inEventScript ) {
                                        mode(START);
                                      }
                                      else {
                                        sconstDelim = ch; more();
                                        mode(C_STRING_MODE);
                                      }
                                   >>
  
#token "\'"                        << sconstDelim = ch; more();
                                      mode(C_STRING_MODE);
                                   >>
  
#lexclass   C_STRING_MODE /**********************************************/

#token  STRING_CONST "[\"\'\n\r]"  << if (ch == sconstDelim){
                                        mode(SCRIPT_MODE);
                                      }
                                      else {
                                        if (ch == '\"'){
                                          replstr("\\\"");
                                        }  
                                        else if(ch == '\n'){
                      //should this be an error ???
                                          replchar(' '); //should really be nothing, find a way
                                        }
                                        more();                                      
                                      }
                                   >>
                   
#token  "\\\""                     << replstr ("\\\""); more (); >>
#token  "\\'"                      << replchar ('\'');  more (); >>
#token  "\\\\"                     << replstr ("\\\\"); more (); >>
#token  "\\b"                      << replstr ("\\b");  more (); >>
#token  "\\f"                      << replstr ("\\f");  more (); >>
#token  "\\n"                      << replstr ("\\n");  more (); >> 
#token  "\\r"                      << replstr ("\\r");  more (); >>
#token  "\\t"                      << replstr ("\\t");  more (); >>
#token  "\\v"                      << replstr ("");     more (); 
                                      warn("\\v not available in Java, ignored"); 
                                   >>  
#token  "\\~[]"                    << replchar (ch);    more (); >> 

#token  "~[\"\'\\\n\r]+"           << more(); >>
  
  
#lexclass SCRIPT_MODE /**************************************************/
  
/*}}}*/

/* token classes */                                                     /*{{{*/

#tokclass MULTIPLICATIVE_OP { MUL DIV MOD }

#tokclass ADDITIVE_OP { PLUS MINUS }

#tokclass SHIFT_OP { SHL SHR USHR }

#tokclass RELATIONAL_OP { TLT LE GT GE }

#tokclass EQUALITY_OP { EQUAL NOT_EQUAL STRICT_EQUAL STRICT_NOTEQUAL }

#tokclass UNARY_OP { INC DEC NOT LOG_NOT VOID TYPEOF}

#tokclass UNARY_OP_LEFTSIDE { INC DEC }
#tokclass UNARY_OP_UNARY {  NOT LOG_NOT VOID TYPEOF}

#tokclass ASSIGNMENT_OP { ASSIGN MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN PLUS_ASSIGN MINUS_ASSIGN SHL_ASSIGN SHR_ASSIGN USHR_ASSIGN AND_ASSIGN XOR_ASSIGN OR_ASSIGN }

#tokclass INCREMENT_OP { INC DEC }

#tokclass CONST { NUM_CONST STRING_CONST }

#tokclass VAR_END {SC COMMA}
  
/*}}}*/



/******************************************************************************
 * parser
 */
 
   
  
class Parser {

<< 

/* member functions */

void skipToBlockEnd ( int lvl )                                         /*{{{*/
{   // consume until, but not including block end
    int l = lvl;
    ANTLRTokenType tt;
    
    while ( true ){
        tt = LT(1)->getType();
            
        switch ( tt ){
        case LBRACE:
            l++; 
            break;
             
        case RBRACE:
            if ( l-- == lvl ) {
                return;
            }
            break;
            
        case TEOF:
        case SCRIPT_END:
            return;   
        }
            
        consume();
    } // while
}
/*}}}*/

void skipToCaseEnd ()                                                   /*{{{*/
{   // consume until, but not including case end
    int l = 0;
    ANTLRTokenType tt;
    
    while ( true ){
        tt = LT(1)->getType();
            
        switch ( tt ){
        case LBRACE:
            l++; 
            break;
             
        case RBRACE:
            if ( --l < 0)
              return;                    
            break;
                        
        case CASE:
        case DEFAULT:                      
        case TEOF:
        case SCRIPT_END:
            return;
        }
            
        consume();
    } // while
}
/*}}}*/

void skipToScriptEnd ()                                                 /*{{{*/
{
    if ( scriptOnly ){
        consumeUntilToken(TEOF);
    }
    else {
        consumeUntilToken(SCRIPT_END);
    }
}

/*}}}*/

int validNextNode ( char* tag )                                         /*{{{*/
{
    //printf("Tag is %s, LT(2)->type() is %d, LT(2)->getText() is %s\n", tag, LT(2)->getType(), LT(2)->getText());

    switch ( LT(1)->getType() ){
    case XML_OPEN:
        if ( docType->validChildNode && !(*docType->validChildNode)(tag,LT(2)->getText()) ){
            //printf("### bail %s <- %s\n", tag, LT(2)->getText());
            return 0;
        }
        break;
    case XML_END_OPEN:
        //printf("    test <%s>:</%s>\n", tag,LT(2)->getText());
        if ( docType->validEndNode && !(*docType->validEndNode)(tag,LT(2)->getText()) ){
            //printf("### bail %s <- end %s\n", tag, LT(2)->getText());
            return 0;
        }
        break;
    case TXT_WORD:
        if ( docType->validChildNode && !(*docType->validChildNode)(tag,"_text_") ){
            //printf("### bail %s <- _text_\n", tag);
            return 0;
        }
    }
    
    return 1;
}
/*}}}*/

/*
void warn ( char* fmt, ...)                                              
{
    va_list ap;
    
    va_start( ap, fmt);
    fprintf( stderr, "%s:%d: ", clsName, LT(1)->getLine());
    vfprintf( stderr, fmt, ap);
    fputc( '\n', stderr);
    va_end( ap);    
}

*/

>>

/* debugging stuff */                                                   /*{{{*/
/****
tokens :
  (
      a:KEYWORD            << printf( "keyword:  %s\n", $a->getText()); >>
    | b:ID                 << printf( "id:       %s\n", $b->getText()); >>
    | c:STRING_CONST       << printf( "s_const:  %s\n", $c->getText()); >>
    | d:NUM_CONST          << printf( "n_const:  %s\n", $d->getText()); >>
    | e:OPERATOR           << printf( "op:       %s\n", $e->getText()); >>
    | f:SYMBOL             << printf( "sym:      %s\n", $f->getText()); >>
    | g:XML_TAG            << printf( "XML_tag   %s\n", $g->getText()); >>
    | h:XML_ATTR           << printf( "XML_attr  %s\n", $h->getText()); >>
    | i:ASSIGN             << printf( "XML_asign %s\n", $i->getText()); >>
    | j:XML_ATTR_VAL       << printf( "XML_attr_val  %s\n", $j->getText()); >>
    | k:XML_ELEM_CLOSE     << printf( "XML_elem_close  %s\n", $k->getText()); >>
    | l:XML_CLOSE          << printf( "XML_close  %s\n", $l->getText()); >>
    | m:XML_OPEN           << printf( "XML_OPEN <\n"); >> 
    | n:XML_END_TAG        << printf( "XML_end_tag  %s\n", $n->getText()); >>
    | o:SCRIPT_TAG         << printf( "XML_script_start  %s\n", $o->getText()); >>
    | p:SCRIPT_END         << printf( "XML_script_end  %s\n", $p->getText()); >>
    | q:TXT_BLANK          << printf( "BLANK: '%s'\n", $q->getText()); >>
    | r:TXT_NL             << printf( "NL: '%s'\n", $r->getText()); >>
    | s:TXT_WORD           << printf( "XML_text  %s\n", $s->getText()); >>
  )*
;
****/
/*}}}*/



/******************************************************************************
 * Parser part A - XML
 */
 

//global exception handler: Means "for all" not "in case math none, at bottom of call stack"
//so don't put here, put in highedst node!
  
document                                                                /*{{{*/
:                                  << set_sbuf(bdom);
                                      out("void initDocument ()");
                                      lout("{"); incInd();
                                      lout( "final XsObject _def = Def;");
                                      lout("%s f = getDocumentFactory();", factoryName);
                                   >>
  ( node [0] )*
                                   << ln(); setIndLevel(1); lout("f.finish();");
                                      decInd(); lout("}");
                                   >>
;
/*}}}*/
         
node                                                                    /*{{{*/
  [int nodelevel]
:
  (
    XML_OPEN
    ( scriptNode
    | genericNode [nodelevel]
    )
  )
 |( XML_META_OPEN metaNode )
 |( XML_DTD_OPEN dtdNode )
;
/*}}}*/

genericNode                                                             /*{{{*/
  [int nodelevel]
:                                  << char* tag; char* aval; int pre=0; int gotVal; >>

  a:XML_TAG                        << tag = $a->getText();
                                      //printf("# entering: %s\n", tag);
                                      if ( !strcasecmp( tag,"html") ){
                                          setDocType( "html");
                                      }
                                      set_sbuf(bdom); 
                                      setIndLevel(nodelevel+1);
                                      ln();
                                      lout("f.createElement(%d,\"%s\");", nodelevel, tag);
                                   >>
  (                                // 2DO - we probably should turn EVENT_TAG into a
                                   // semantic predicate, and move it to docType (PM)
    ( b:EVENT_TAG                  << inEventScript = 1;
                                      lout("f.setEventAttribute(\"%s\",", $b->getText());
                                      out(" new XsObject (){"); incInd();
                                      lout("public XsObject invoke ( XsObject xThis ){"); incInd();
                                      
                                   >>
      ASSIGN
      DQUOTE
      script                       << inEventScript = 0;
                                      lout("return xThis;");
                                      decInd(); lout("}"); decInd(); lout("});");
                                   >>
      DQUOTE
    )
  | ( c:XML_TAG                    << gotVal = 0; >>
      { ASSIGN                     << inAttribute = 1; >>
        ( d:XML_STRING             << aval = $d->getText();
                                      lout("f.setAttribute(\"%s\",%s);", $c->getText(), aval);
                                      if ( (nodelevel == 1) && !strcasecmp( "name", $c->getText()) ){
                                        // named toplevel nodes become static properties
                                        lout("defProperty(\"%s\", f.currentNode());",aval);
                                      }
                                   >>
        | e:XML_TAG                << lout("f.setAttribute(\"%s\",\"%s\");",
                                           $c->getText(), $e->getText());
                                   >>
        )                          << inAttribute = 0; gotVal = 1; >>
      }
                                   << if ( !gotVal ) {
                                          lout("f.setAttribute(\"%s\",\"\");", $c->getText());
                                      }
                                   >>
    )
  )*
  (
    ( XML_ELEM_CLOSE (TXT_SKIP)* )     
    | ( XML_CLOSE (TXT_SKIP)*      << reset_sbuf( btmp);
                                      // 2DO - uggly HTML hardwired
                                      pre = (strcasecmp( tag, "pre") == 0);
                                      if ( !validNextNode(tag) ){
                                          // printf("RETURNING ON !validNextNode(tag)\n");
                                          decInd(); return;
                                      }
                                   >>
        ( pcData [pre]             << ln();  incInd();
                                      lout("f.createTextElement(%d,\"", nodelevel+1);
                                      decInd();
                                      merge_str(btmp); reset_sbuf(btmp);
                                      out("\");");
                                      if ( !validNextNode(tag) ){
                                          // printf("RETURNING ON !validNextNode(tag)\n");
                                          decInd(); return;
                                      }
                                   >>
         | node [nodelevel+1]      << if ( !validNextNode(tag) ){
                                          // printf("RETURNING ON !validNextNode(tag)\n"); 
                                          decInd(); return;
                                      }
                                      //printf( "## back in %s\n", tag);
                                   >>
         | ( XML_END_OPEN          // we accept mixed-in end-tags for (overlapping) HTML
             f:XML_TAG
             XML_CLOSE (TXT_SKIP)* << if ( !strcmp( tag, $f->getText()) ){
                                        // printf("RETURNING \n"); 
                                        return;
                                      }
                                   >>
           )
        )*
      )
  )
;
/*}}}*/

pcDataWord                                                              /*{{{*/
  [char* prefix]
:                                  << int mightBeEntity=0; char* val=0; int semiSeen=0; >>
  { TXT_AMP                        << mightBeEntity = 1; >>
  }
     
  ( a:TXT_WORD
     
    { TXT_SEMI                     << semiSeen = 1;
                                      if ( mightBeEntity && docType->expand ) {
                                        val = (*docType->expand)( $a->getText());
                                      }
                                   >>
    }
                                   << if ( prefix ) { outb(btmp,"%s",prefix); }
                                      if ( val ) {
                                        //printf("entity found: '%s' -> '%s'\n", $a->getText(),val);
                                        // 2DO - this has to be recursively parsed, but how do we
                                        // resync parser and scanner? (PM)
                                        outb(btmp,"%s",val);
                                      }
                                      else {
                                        if ( mightBeEntity ){ outb(btmp,"&"); }
                                        outb(btmp,"%s",$a->getText());
                                        if ( semiSeen ){ outb(btmp,";"); }
                                      }
                                   >>
  )
  | TXT_SEMI                       << if ( prefix ) { outb(btmp,"%s;",prefix); } >>
;
/*}}}*/
   
pcData                                                                  /*{{{*/
  [int conserveSpace]
:                                  << char* sep=0; >>
  pcDataWord [0]
  ( b:TXT_BLANK                    << sep = (conserveSpace) ? $b->getText() : (char*)" "; >>
  | c:TXT_NL                       << sep = (conserveSpace) ? (char*)"\n" : (char*)" "; >>
  | pcDataWord [sep]               << sep = 0; >>
  )*
;
/*}}}*/
   
scriptNode                                                              /*{{{*/
:                                  << 
                                      #ifdef KAFFELIB
                                      char* filename = NULL;
                                      bool external_source = false;
                                      #endif
                                   >>
   SCRIPT_TAG                      
   ( b:XML_TAG                     << 
                                      #ifdef KAFFELIB
                                      if(!strcmp(b->getText(), "src")){
                                        external_source = true; 
                                      }
                                      #endif
                                   >>
     ASSIGN                        << inAttribute = true; >>
    
     c:XML_STRING                  << 
                                      #ifdef KAFFELIB 
                                      if(external_source){
                                        filename =  (char*) calloc(strlen($c->getText()) - 1, sizeof(char));
                                        strcpy(filename, $c->getText() + 1);
                                        filename[strlen($c->getText()) - 2] = '\0';
                                                                                                                
                                        jclass xscompiler = Genv->FindClass("kaffe/xs/XsCompiler"); CheckException();
                                        jmethodID getRelative = Genv->GetStaticMethodID(xscompiler, "getRelative",
                                                                                        "(Ljava/lang/String;)Ljava/io/InputStream;"); CheckException();
                                        SrcInputStream = Genv->CallStaticObjectMethod(xscompiler, getRelative, Genv->NewStringUTF(filename));       
                                        CheckException();  // take out if put in commented out code                                  
                                           
                                                                                        
                                        if(SrcInputStream == NULL){
                                          warn("could not fetch file specified by src attribute, ignored");
                                        }
                                        
                                        /*
                                          jthrowable e;
                                          if(e = Genv->ExceptionOccurred()){
                                            printf("Exception Occurred\n");
                                            jclass malformedurlE =  Genv->FindClass("java/net/MalformedURLException"); CheckException();
                                          
                                            jclass IOE =  Genv->FindClass("java/io/IOException"); CheckException();

                                          
                                            if(Genv->IsInstanceOf(e, malformedurlE))
                                              warn("src attribute file not found, ignored"); // correct action ?
                                            else if(Genv->IsInstanceOf(e, IOE))
                                              warn("network error, src attribute ignored");
                                            
                                            SrcInputStream = NULL;  //already true, take out    
                                          }
                                         */
                                      }
                                      #endif
                                      
                                      inAttribute = false;
                                   >>
    
   )*
   XML_CLOSE                       << 
                                      #ifdef KAFFELIB
                                      if(SrcInputStream != NULL){
                                        scanner->saveState(&saved_state);
                                        DLGJavaInputStream isinput(Genv, SrcInputStream); 
                                        scanner->setInputStream(&isinput);
                                      }
                                      #endif
                                   >>
                                      
                                      
   script                          <<
                                      #ifdef KAFFELIB
                                      if(SrcInputStream != NULL){
                                        SrcInputStream = NULL;
                                      }
                                      if(filename != NULL){
                                        free(filename);
                                      }
                                      #endif    
                                   >>

   SCRIPT_END                      
;
/*}}}*/

metaNode                                                                /*{{{*/
:
   a:XML_TAG
   ( b:XML_TAG
     ASSIGN
     DQUOTE
     c:XML_TAG
     DQUOTE
   )*
   XML_META_CLOSE
;
/*}}}*/
   
dtdNode                                                                 /*{{{*/
:
   a:XML_TAG                      << inAttribute = 1; >>
   ( b:XML_TAG                    
   | c:XML_STRING                 
   )*
   XML_CLOSE (TXT_SKIP)*          << inAttribute = 0; >>
;
/*}}}*/


/******************************************************************************
 * Parser part B - XoeScript
 */


script                                                                  /*{{{*/
:                                  << out_attr_t r = oNONE; >>                             
    //leave function ambuiguity in -> more efficient
    ( function 
    | statement [iNONE] > [r]      << 
                                      if ( isDeadEnd( r) ){
                                        if(LT(1)->getType() != TEOF){
                                          warn("unreachable code, rest of script ignored");
                                          skipToScriptEnd();
                                        }       
                                      }
                                   >>
    )*
;
//exception reached top, bail out ?

exception catch MismatchedToken : << fprintf(stderr, "Parsing exception: MismatchedToken at line %d\n", scanner->line()); >>
          catch NoViableAlt :     << fprintf(stderr, "Parsing exception: NoViableAlt at line %d\n", scanner->line()); >>
          default :               << fprintf(stderr, "Parsing exception at line %d\n", scanner->line()); >>


/*}}}*/
     
function                                                                /*{{{*/
                                      
:                                  << int nargs = 0; char *fid; set_sbuf(bmain);
                                      out_attr_t r = oNONE; 
                                      int max_ids = 2;
                                      char** ids = (char**) calloc(max_ids, sizeof(char*)); 
                                   >>
    
    FUNCTION
    f:ID                           << fid = $f->getText(); addVar(fid,T_FUNC); ln();
                                      lout("public static XsObject _%s ( XsObject xThis, XsArray arguments )", fid);
                                      lout("{"); incInd(); incLevel();
                                   >>
    LPAREN
    { a:ID                         << lout("XsObject %s", $a->getText());
                                      
                                      ids[nargs++] = StrDup($a->getText());
                                      
                                      addVar( $a->getText(),T_PROP);
                                   >>
      ( COMMA
        b:ID                       << out(", %s", $b->getText()); 
                                      
                                      if(nargs == max_ids){
                                        // grow array 
                                        int newMaxIds = max_ids + 10;
                                        char** new_ids = (char**) calloc(newMaxIds, sizeof(char*));
                                        memcpy( new_ids, ids, max_ids * sizeof(char*));
                                        free(ids);
                                        ids = new_ids;
                                        max_ids = newMaxIds;
                                      }
                                      
                                      /*
                                       
                                      if(nargs == max_ids)
                                        ids = (char**) realloc((char**) ids, max_ids += 10);
                                      
                                      */
                                     
                                                                                
                                      ids[nargs++] = StrDup($b->getText());
                                      addVar( $b->getText(),T_PROP);
                                   >>
      )*                           << out(";"); >>
    }
    RPAREN                         << 
                                      if(!functionOnly){
                                        lout("XsObject _def = Def;");
                                      }
                                      else{
                                          lout("XsObject _def = arguments.elements[arguments.count - 1];");     
                                      }
                                      
                                      addVar("arguments",T_DARR);
                                      
                                      int j;    
                                      
                                      if(nargs != 0){
                                      
                                        int arg;  
                                        
                                        if(!functionOnly)
                                          lout("switch(arguments.count){");
                                        else
                                          lout("switch(arguments.count - 1){");  
                                       
                                       
                                        for(arg = 0; arg < nargs; arg++){
                                          lout("  case %d:\n", arg);
                                          
                                          for(j = 0; j < arg; j++)
                                           out("           %s = arguments.elements[%d];\n", ids[j], j);
                                          
                                          for(; j < nargs; j++)
                                           out("           %s = _undefined;\n", ids[j]); 
                                         
                                           out("           break;\n"); 
                                        }
                                        
                                           out("   default:\n");
                                        
                                        for(arg = 0; arg < nargs; arg++)
                                           out("           %s = arguments.elements[%d];\n", ids[arg], arg);
                                        
                                      
                                           out("           break;\n");
                                      
                                        out("  }");
         
                                      } 
                                                                             
                                      //release strings                                     
                                      for(j = 0; j < nargs;j++){
                                        free(ids[j]);
                                      }
                                   >>
    LBRACE
    ( 
     statement [iNONE] > [r]       << if ( isDeadEnd(r) ){
                                        if ( LT(1)->getType() != RBRACE && LT(1)->getType() != TEOF){
                                          warn("unreachable code, rest of block ignored");
                                          skipToBlockEnd(1);
                                        }
                                      }
                                   >> 
    )*                             
    RBRACE                         << 
                                      if( !otest(r,oRET) ){
                                        //function does not statically pessimistically return
                                        lout( "return _undefined;");
                                      }
                                      
                                      decInd();/*decInd();*/ lout("}"); 
                                      
                                      decLevel();
                                      
                                      if(!functionOnly){
                                        lout("static XsObject %s;", fid);
                                      }
                                      
                      if(dummy_to_declare && !dummy_declared){
                        lout("static XsObject __dummy__;");
                    dummy_to_declare = false;
                    dummy_declared = true;
                      } 
                      
                      
                                      set_sbuf(binit);
                                      
                                      if(!functionOnly){
                                        lout("%s = Def.defFunc(\"%s\", %d);", fid, fid, nargs);
                                      }
                      
                      
                                   >>
;
exception catch NoViableAlt:
          catch MismatchedToken : <<    
                                     //printf("RAMBO r = %d\n", r);
                                      
                                     if(LT(1)->getType() == RBRACE){
                                     
                                       closeParen(); out(";");
                                       
                                       /* what to do with this?
                          better to have an extra one than to be missing one
                                       if( !otest(r,oRET) ){
                                         //function does not statically pessimistically return
                                         lout( "return _undefined;");
                                       }*/
                                      
                                       lout( "return _undefined;"); 
                                      
                                       decInd(); decInd(); lout("}"); 
                                       
                                       decLevel();
                                      
                                       if(!functionOnly){
                                         lout("static XsObject %s;", fid);
                                       }
                                      
                       if(dummy_to_declare && !dummy_declared){
                         lout("static XsObject __dummy__;");
                     dummy_to_declare = false;
                     dummy_declared = true;
                       }    

                      
                                       set_sbuf(binit);
                                      
                                       if(!functionOnly){
                                         lout("%s = Def.defFunc(\"%s\", %d);", fid, fid, nargs);
                                       }
                                     
                                       consume();
                                        
                                                                              
                                     }  
                                  >>


/*}}}*/

block                                                                   /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]
  
:
  
    LBRACE                         << lout("{"); $oa = oNONE;>>
    ( statement [ia] > [$oa]       << 
                                      /*broken with Virtual ; if have { continue 
                                                        break } since NEWLINE is skipped */
                                      if ( isDeadEnd($oa) ){
                                        if ( LT(2)->getType() != RBRACE && LT(1)->getType() != TEOF){
                                          warn("unreachable code, rest of block ignored");
                                          skipToBlockEnd(1);
                                        }
                                      }
                      
                                   >> 
    )*
    RBRACE                         << lout("}"); >>
;  
exception catch NoViableAlt:
          catch MismatchedToken :  << 
                                     if(LT(1)->getType() == RBRACE){
                                       consume();
                                       closeParen(); out(";"); 
                                       lout("}");
                                     }  
                                   >>
                   

/*}}}*/

statement                                                               /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]

:                                  << Pos = getPos();  // save rollback point 
                                      //printf("In stmt, Pos = %d\n", Pos);
                                   >>

    SC                             << out(";"); $oa = oNONE; >>
|   ifStatement [ia] > [$oa]       
|   forStatement [ia] > [$oa]
|   whileStatement [ia] > [$oa]
|   doWhileStatement [ia] > [$oa]
|   breakStatement [ia] > [$oa]
|   continueStatement [ia] > [$oa]
|   switchStatement [ia] > [$oa]
|   withStatement [ia] > [$oa]
|   returnStatement [ia] > [$oa]
|   block [ia] > [$oa]
|   tryStatement [ia] > [$oa]
|   throwStatement [ia] > [$oa]
|   javaSection
|   (ID COLON)? labeledStatement [ia] > [$oa]
|   varOrExprStatement [ia] > [$oa]
;
exception catch NoViableAlt :
          catch MismatchedToken : <<               
                                    /* was working for Virtual ;
                    
                    if(!(LT(1)->getType() == TEOF))
                                      *$_retsignal = $_signal;
                                    else{
                                      
                      out(";");  
                                  
                    } */
                    
                    if(LT(1)->getType() == RBRACE)
                                      *$_retsignal = $_signal;
                                    else{
                                      if(LT(1)->getType() != TEOF){
                        setPos(Pos);
                        consumeUntilToken(SC);
                    consume();
                    warn("Syntax error, stmt ignored");
                                      }
                      else 
                        out(";");
                    } 
                    
                  >>

/*}}}*/

ifStatement                                                             /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]

:                                  << out_attr_t r1 = oNONE, r2 = oNONE; cond_attr_t ca = cNONE; // unused in static comp
                                      $oa=oNONE; 
                                   >>

    IF                             << lout("if "); >>
    condition > [ca]               << incInd(); >>
    statement [ia] > [r1]          << decInd(); >>
    { ELSE                         << lout("else"); incInd(); >>
      statement [ia] > [r2]            
    }                              << decInd();
                                      if( isDeadEnd(r1) && isDeadEnd(r2) ){
                                        $oa = oadd($oa,oDEAD);
                                        if( otest(r1,oRET) && otest(r2,oRET) ){
                                          $oa = oadd($oa,oRET); 
                                        }
                                      }
                                   >>


;
/*}}}*/

whileStatement                                                          /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]

:                                  << cond_attr_t ca; out_attr_t r;
                                      int pos=getPos(); $oa=oNONE;
                                   >> 
    
    WHILE                          << lout("while "); >>
    condition > [ca]               << incInd(); ln(); >>
    statement [iloop(ia)] > [r]    << decInd();
                                      
                                      if ( ctest(ca,cNEVER) ){
                                        setPos(pos); // roll back
                                        warn("while will never execute, ignored");
                                      }
                                      else if ( ctest(ca,cALWAYS) ){
                                        if ( otest(r,oRET) ){
                                          $oa = oadd($oa,oRET);
                                        }
                                      }
                                   >>
;
/*}}}*/

doWhileStatement                                                        /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]

:                                  << cond_attr_t ca = cNONE; 
                                      int pos=getPos(); 
                                      $oa=oNONE;
                                   >>
    
    DO                             << lout("do"); incInd(); ln(); >>
    statement [iloop(ia)] > [$oa]   
    WHILE                          << decInd(); lout("while "); >>
    condition > [ca]
    SC                             << out(";"); ln(); >>
;
/*}}}*/

switchStatement                                                         /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]

:                                  << bool hasDefault=false;
                                      out_attr_t r = oNONE; $oa=oNONE;
                                      int all_return = 1;
                                      int all_continue_or_return = 1;
                                   >>
   
    SWITCH                         << lout("switch "); >>
    condExpression
    LBRACE                         << setPosOffset(-1);
                                      out(".getNumber().getIntValue())");
                                      lout("{");
                                      incInd(); 
                                   >>
     
    ( CASE c:CONST COLON           << lout("case %s:", $c->getText()); incInd(); >>
      ( statement [icase(ia)] > [r]
                                   << if(isDeadEnd(r)){
                                        if ( LT(1)->getType() != RBRACE && LT(1)->getType() != CASE  &&
                                             LT(1)->getType() != DEFAULT){
                                          warn("unreachable code, rest of case ignored");
                                          skipToCaseEnd();
                                        }       
                                      }         
                                   >>
      )*                           << decInd();
                                      if ( otest(r,oRET)){ 
                                        all_return &= 1;
                                        all_continue_or_return &= 1; 
                                      }
                                      else if ( !otest(r,oCNT) && !otest(r, oBRK) ) { 
                                        all_return &= 1;  //status depends on the next one
                                        all_continue_or_return &= 1;
                                      } 
                                      else if(otest(r, oCNT)){ // but then break qualifies if has a label!
                                        all_return &= 0;
                                        all_continue_or_return &= 1;                                    
                                      }
                                      else{
                                        all_return &= 0;
                                        all_continue_or_return &= 0;
                                      }
      
                                   >>
    )*
    { DEFAULT COLON                << lout("default:"); incInd(); hasDefault=true; >>
      ( statement [icase($ia)] > [r]
                                   << if(isDeadEnd(r)){
                                        if ( LT(1)->getType() != RBRACE && LT(1)->getType() != CASE  &&
                                             LT(1)->getType() != DEFAULT){
                                          warn("unreachable code, rest of default ignored");
                                          skipToCaseEnd();
                                        }       
                                      } 
                                   >>
      )*                           << decInd();
      
                                      if ( otest(r,oRET)){ 
                                        all_return &= 1; 
                                        all_continue_or_return &= 1;
                                      }
                                      else if(otest(r, oCNT)){ // but then break qualifies if has a label!
                                        all_return &= 0;
                                        all_continue_or_return &= 1;                                    
                                      }
                                      else{
                                        all_return &= 0;
                                        all_continue_or_return &= 0;
                                      }
                                   >>
    }                              
    RBRACE                         << decInd(); lout("}"); decInd();
                                      if ( hasDefault ){
                                        if(all_return){
                                          $oa = oadd($oa,oRET);
                                        }
                                        if(all_continue_or_return){
                                          $oa = oadd($oa,oDEAD);
                                        }
                                      }
                                   >>
;
/*}}}*/

breakStatement                                                          /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << $oa = oNONE;
                                      if ( ! itest($ia,iBRK) ){
                                        warn("break not allowed here, ignored");
                                        consumeUntilToken(SC); consume();
                                        lout(";");
                                        return $oa;
                                      }
                                      
                                      NoLineTerminatorHere = true;
                                   >>
  
    b:BREAK                        << lout("break"); >>
    { a:ID                         << if(!isLabelDefined($a->getText())){
                                        warn("label %s not found, ignored", $a->getText());
                                      }
                                      else {
                                        out(" %s", $a->getText());
                                      }
                                   >>
    }
    SC                             << out(";"); >>
                                   << $oa = oadd($oa,oBRK); >>
;
/*}}}*/

continueStatement                                                       /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << $oa = oNONE;
                                      if (!itest($ia,iCNT)){
                                        warn("continue not allowed here, ignored");
                                        consumeUntilToken(SC); consume();
                                        lout(";");
                                        return $oa;
                                      }
                                      
                                      NoLineTerminatorHere = true;
                                   >>

    b:CONTINUE                     << lout("continue"); >>
    { a:ID                         << if(!isLabelDefined($a->getText())){
                                        warn("label %s not found, ignored", $a->getText());
                                      }
                                      else {
                                        out(" %s", $a->getText());
                                      }
                                   >>
    }
    SC                             << out(";"); >>
                                   << $oa = oadd($oa,oCNT); >>
;
/*}}}*/

//fix  in dyn comp, when doing full implement of , operator etc
withStatement                                                           /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << $oa = oNONE; >>

    WITH                           << withLevel++;
                                      lout("{"); incInd();
                                      lout("XsObject _def_%d = _def;", withLevel);
                                      lout("_def = ");
                                   >>
    LPAREN
    expression                     << closeParen(); out(";"); ln(); >>
    RPAREN
    statement [ia] > [$oa]         << lout("_def = _def_%d;", withLevel);
                                      decInd(); lout("}"); withLevel--;
                                   >>
;
/*}}}*/

//check out the closeParen(); thing maybe necessary in !hasExpr
returnStatement                                                         /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << bool hasExpr=false; $oa = oNONE; NoLineTerminatorHere = true;>>  

    RETURN                         << lout("return "); >>
    { expression                   << closeParen(); hasExpr=true; >>
    }                              
    SC                             << if (!hasExpr){
                                        if (getLevel() > 0){
                                          // all but toplevel returns must have a value
                                          out("_undefined");
                                        }
                                      }
                                      out(";");
                                      $oa = oadd($oa,oRET);
                                    >>
;
exception catch NoViableAlt :
          catch MismatchedToken : <<   //printf("In return\n");            
                                      
                       if (LT(1)->getType() == RBRACE){
                                         out("_undefined");
                                       }
                                       
                       *$_retsignal = $_signal;                    
                      >>


/*}}}*/


forStatement
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << bool enumKeys=false; bool declaring=false; char* id1, *id2;
                                      int pos = getPos(); int comma_pos;
                                      cond_attr_t ca=cNONE; out_attr_t r; $oa=oNONE;
                                      sbuf_t* before = get_sbuf();
                                      lout("");
                                      reset_sbuf(bfor);
                                      reset_sbuf(blocals);
                                      set_sbuf(bfor);
                                      setIndb(bfor, before);
                                      setIndb(blocals, before);
                                   >>

    FOR                            << lout("for (");  >>
    LPAREN
  ( // shouldn't be necessary, but pccts screws up, otherwise
    ( SC                           << out("; "); >>
      forCondAndLoop > [ca]
    )
  | ( { VAR                        << declaring = true; >>
      }
      a:ID                         << id1 = $a->getText();
                                     
                                      if(declaring){
                                        if ( getLevel() == 0 ){     // global var
                                          if(!isVarAtLevel(id1)){
                                            addVar(id1,T_PROP);
                                            set_sbuf(bmain);
                                            lout("static XsObject %s = _undefined;", id1);
                                            set_sbuf(bfor);
                                          }
                                        }
                                        else {                 // in function
                                          set_sbuf(blocals);
                                          if(!isVarAtLevel(id1)){
                                            addVar(id1,T_PROP); 
                                            
                                            lout("XsObject %s", id1);
                                            if( LT(2)->getType() != ASSIGN ){
                                              out(" = _undefined");
                                            }
                                          }
                                          else if(LT(2)->getType() == ASSIGN){
                                            out("%s", id1);
                                          }
                                        }
                                      }
                                      else{  // not declaring
                                      
                                        if ( getLevel() == 0 ){     // global var
                                          if(!isVarAtLevel(id1)){
                                            addVar(id1,T_PROP);
                                            set_sbuf(bmain);
                                            lout("static XsObject %s = _undefined;", id1);
                                            set_sbuf(bfor);
                                          }
                                        }
                                        else {                 // in function
                                          set_sbuf(blocals);
                                          
                                          int vtype;
                                          varLevelt(id1,&vtype);
                                          
                                          if(LT(2)->getType() == LPAREN){
                                            if ( (vtype == T_FUNC) && (withLevel == 0) ){
                                              statCall = true;
                                              lout("_%s(_def, ",id1); 
                                            }
                                            else if(!isVarAtLevel(id1)){ // not yet known 
                                              lout("_def.call(\"%s\"",id1);
                                            }
                                            else{
                                              lout("_def.call(%s",id1);
                                            }
                                          }
                                          else if(!isVarAtLevel(id1)){
                                              if (set_el(LT(2)->getType(), ASSIGNMENT_OP_set)){
                                                if(LT(2)->getType() == ASSIGN){
                                                  out("_def.xsPut(\"%s\", ",id1);
                                                }
                                                else{
                                                  out("_def.xsPut(\"%s\", _def.xsGet(\"%s\")",id1, id1);
                                                }
                                                openParen++;
                                              }         
                                          }
                                          else if(LT(2)->getType() == ASSIGN){
                                            out("%s = ", id1);
                                          }
                                          else if(LT(2)->getType() != INC && LT(2)->getType() != DEC ){
                                            out("%s = %s", id1, id1);
                                          }
                                        }
                                      } // not declaring
                                   >>
      ( // again, pccts trouble
        ( IN                       << set_sbuf(blocals);
                                      if(declaring){
                                        out(";");
                                      }
                                      lout("NameEnumeration _%s = (", id1); 
                                   >>
          expression               << // TODO: deal with expression list
                                      closeParen();
                                      out(").getPropertyNames();");
                                      set_sbuf(bfor);
                                      out(";_%s.hasMoreElements()", id1);
                                      enumKeys=true;
                                   >>
        )
      | ( 
          { b:ASSIGNMENT_OP        <<
                                      if(declaring){
                                        if($b->getType() != ASSIGN){
                                          warn("Parse error"); // do something else
                                        }  
                                      
                                        if ( getLevel() == 0 ) {  // global var
                                          set_sbuf(blocals);
                                          out("%s = Def.defProperty(\"%s\",", id1, id1);
                                          openParen++;
                                        }
                                        else {               // in function
                                          out(" = ");
                                        }
                                      }
                                      else{ // not declaring
                                        if ( getLevel() == 0 ) {  // global var
                                          set_sbuf(blocals);
                                          if($b->getType() == ASSIGN){
                                            lout("%s = Def.defProperty(\"%s\",", id1, id1);
                                          }
                                          else{
                                            lout("%s = Def.defProperty(\"%s\", %s.%s(", id1, id1, id1,
                                                                                    binOp($b->getType()));
                                            openParen++;
                                          }
                                          openParen++;
                                        }
                                        else if($b->getType() != ASSIGN){  // in function
                                          out(".%s(", binOp($b->getType()));
                                          openParen++; 
                                          rawLiteral = true;                              
                                        }

                                      
                                      } //not declaring
                                      
                                   >>
             assignmentExpression  << rawLiteral = false; >>    
           }
        | ( c:INCREMENT_OP         << if(declaring){
                                        warn("Parse error"); // do something else
                                      }  
        
                                      out("%s.%s()", id1, prefixOp($c->getType())); 
                                   >>
            )
        | <<declaring != true >>?
          callArguments[0]         << if(declaring)
                                         warn("Parse error");
                                   >>                    
         )
        ( COMMA                    << comma_pos = getPos(); >>
          d:ID                     << id2 = $d->getText();
          
                                      if(declaring){
                                       
                                        if ( getLevel() == 0 ){     // global var
                                          closeParen();
                                          out(";");
                                          if(!isVarAtLevel(id2)){
                                            addVar(id2,T_PROP);
                                            set_sbuf(bmain);
                                            lout("static XsObject %s = _undefined;", id2);
                                            set_sbuf(bfor);
                                          }
                                        }
                                        else {                 // in function
                                          out(";");
                                          set_sbuf(blocals);
                                          if(!isVarAtLevel(id2)){
                                            addVar(id2,T_PROP); 
                                            
                                            lout("XsObject %s", id2);
                                            if( LT(2)->getType() != ASSIGN ){
                                              out(" = _undefined");
                                            }
                                          }
                                          else if(LT(2)->getType() == ASSIGN){
                                            lout("%s", id2);
                                          }
                                          else{
                                           setPos(comma_pos);
                                          }
                                          
                                        }
                                      }
                                      else{ // not declaring
                                      
                                        if ( getLevel() == 0 ){     // global var
                                          closeParen();
                                          out(";");
                                          if(!isVarAtLevel(id2)){
                                            addVar(id2,T_PROP);
                                            set_sbuf(bmain);
                                            lout("static XsObject %s = _undefined;", id2);
                                            set_sbuf(bfor);
                                          }
                                        }
                                        else { 
                                          closeParen();
                                          out(";");                // in function
                                          set_sbuf(blocals);
                                          
                                          int vtype;
                                          varLevelt(id2,&vtype);
                                          
                                          if(LT(2)->getType() == LPAREN){
                                            if ( (vtype == T_FUNC) && (withLevel == 0) ){
                                              statCall = true;
                                              lout("_%s(_def, ",id2); 
                                            }
                                            else if(!isVarAtLevel(id2)){ // not yet known 
                                              lout("_def.call(\"%s\"",id2);
                                            }
                                            else{
                                              lout("_def.call(%s",id2);
                                            }

                                          }
                                          else if(!isVarAtLevel(id2)){
                                              if (set_el(LT(2)->getType(), ASSIGNMENT_OP_set)){
                                                if(LT(2)->getType() == ASSIGN){
                                                  lout("_def.xsPut(\"%s\", ",id2);
                                                }
                                                else{
                                                  lout("_def.xsPut(\"%s\", _def.xsGet(\"%s\")",id2, id2);
                                                }
                                                openParen++;
                                              }                                                   
                                          }
                                          else if(LT(2)->getType() == ASSIGN){
                                            lout("%s", id2);
                                          }
                                          else if(LT(2)->getType() != INC && LT(2)->getType() != DEC ){  
                                            lout("%s = %s", id2, id2);
                                          }
                                        }
                                      }
                                  >>
          ( 
           { e:ASSIGNMENT_OP      <<  if(declaring){
                                        if($e->getType() != ASSIGN){
                                          warn("Parse error"); // do something else
                                        }  
                                      
                                        if ( getLevel() == 0 ) {  // global var
                                          set_sbuf(blocals); 
                                          lout("%s = Def.defProperty(\"%s\",", id2, id2);
                                          openParen++;
                                        }
                                        else {               // in function
                                          out(" = ");
                                        } 
                                      }
                                      else{
                                      
                                         if ( getLevel() == 0 ) {  // global var
                                          set_sbuf(blocals);
                                          if($e->getType() == ASSIGN){
                                            lout("%s = Def.defProperty(\"%s\",", id2, id2);
                                          }
                                          else{
                                            lout("%s = Def.defProperty(\"%s\", %s.%s(", id2, id2, id2,
                                                                                    binOp($e->getType()));
                                            openParen++;
                                          }
                                          openParen++;
                                         }
                                         else if($e->getType() != ASSIGN){  // in function
                                          out(".%s(", binOp($e->getType()));
                                          openParen++; 
                                          rawLiteral = true;                              
                                         }
                                      }
                                   >>
             assignmentExpression  << rawLiteral = false; >>
           }
           
           | f:INCREMENT_OP        << if(declaring){
                                         warn("Parse error"); // do something else
                                      }  
        
                                      lout("%s.%s()", id2, prefixOp($f->getType())); 
                                   >>
                                   
           | <<declaring != true >>?
             callArguments[0]      << if(declaring)
                                         warn("Parse error");
                                   >>
          )

        )*
        SC                         << if(get_sbuf() == blocals){
                                        closeParen();
                                        out(";");
                                        set_sbuf(bfor);
                                      }
                                        
                                      closeParen(); out("; "); 
                                   >>
        forCondAndLoop > [ca]
      )
    )
  )
    RPAREN                         << set_sbuf(bfor);
                                     
                                      if ( enumKeys ){
                                        out(";)");
                                        incInd();
                                        out("{");
                                        lout("%s=_%s.nextElement();",id1,id1);
                                      }
                                      else{
                                        out(")");
                                        incInd();
                                        
                                      }
                                   >>
    statement [iloop(ia)] > [r]    << decInd();
    
                                      if ( enumKeys ){
                                        lout("}");
                                      }
                                                                            
                                      if ( ctest(ca,cNEVER) ){
                                        set_sbuf(before);
                                        setPos(pos); // roll back
                                      }
                                      else if ( ctest(ca,cALWAYS) ){
                                        if ( otest(r,oRET) ){ 
                                          $oa = oadd($oa,oRET); 
                                        }
                                        mergeb(blocals, bfor);
                                        mergeb(before,blocals);
                                        set_sbuf(before);
                                      }
                                      else{
                                        mergeb(blocals, bfor);
                                        mergeb(before,blocals);
                                        set_sbuf(before);
                                      }
                                      
                                   >>
;


/****
forStatement
  [in_attr_t ia]  > [out_attr_t oa]
:                           << bool declaring;
                              cond_attr_t ca=cNONE; out_attr_t r; $oa=oNONE; >>

    FOR
    LPAREN
  (
    ( SC
      forCondAndLoop > [ca]
    )
  | ( { VAR
      }
      a:ID
      ( // again, pccts trouble
        ( IN
          expression
        )
      |
        (
          { b:ASSIGNMENT_OP assignmentExpression }
        | c:INCREMENT_OP
        )
        ( COMMA
          d:ID
          ( 
           { e:ASSIGNMENT_OP
             assignmentExpression
           }           
           | f:INCREMENT_OP
          )

        )*
        SC
        forCondAndLoop > [ca]
      )
    )
  )
    RPAREN
    statement [iloop(ia)] > [r]
;
****/

/*
//redo: put the | later
forStatement                                                            
  [in_attr_t ia]  > [out_attr_t oa]
  
: (FOR LPAREN variablesOrExpression IN)?
  FOR                       << lout("for ("); //need var id from variablesOrExpression!!!
                               //out("NameEnumeration _%s = (", id1);
                            >>
  LPAREN
  variablesOrExpression
  IN                         
  expression
  RPAREN
  statement[ia]

  
  
 | FOR                     
  LPAREN
  {variablesOrExpression} 
  SC                        << lout("for ("); out(";"); >>
  forCondAndLoop
  RPAREN                    << out(")"); >>
  statement[ia]
  
;  
*/


forCondAndLoop                                                          /*{{{*/
  > [cond_attr_t ca]

:                                  << bool conditionThere = false;

                                      $ca=cNONE;
                                      
                                      if ( LT(2)->getType() == SC ){
                                        if ( LT(1)->getType() == TTRUE ){
                                          $ca = cadd($ca,cALWAYS);
                                        }
                                        else if ( LT(1)->getType() == TFALSE ){
                                          $ca = cadd($ca,cNEVER);
                                        }
                                      }
                                      else if(LT(1)->getType() == SC) 
                                        $ca = cadd($ca,cALWAYS);
                                   >>                               
    { expression                   << conditionThere = true; >>
    }             
    SC                             << if(conditionThere){
                                        out( ".toBoolean() == xTrue");
                                      }         
                                      out(";");  
                                   >>
    { expression }                 << closeParen(); >>
;
/*}}}*/

tryStatement                                                            /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << out_attr_t r; char* xid; $oa = oNONE; >>
    TRY                            << lout("try"); incInd(); >>
    LBRACE                         << lout("{"); >>
    ( statement [ia] > [$oa] 
    )*
    RBRACE                         << lout( "}"); decInd(); >>
    ( CATCH                        << lout( "catch"); >>
      LPAREN a:ID RPAREN           << xid = $a->getText();
                                      out("(XsException _%s)", xid); incInd();
                                   >>
      LBRACE                       << lout("{");
                                      lout("XsObject %s = _%s.getThrowObject();", xid, xid);
                                   >>
      ( statement [ia] > [r]
      )*
      RBRACE                       << lout("}"); decInd(); >>
    )*
    { FINALLY                      << lout( "finally"); incInd(); >>
      LBRACE                       << lout("{"); >>
      ( statement [ia] > [r]       << if ( isDeadEnd(r) ){
                                        // finally always gets executed, unreachable
                                        $oa = oadd($oa,r);
                                      }
                                   >>
      )*
      RBRACE                       << lout("}"); decInd(); >>
    }
;
/*}}}*/

throwStatement                                                          /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                  << $oa = oNONE; 
                                      NoLineTerminatorHere = true;  
                                   >>
    THROW                          << lout("throw new XsException("); >>
    expression
    SC                             << out(");"); >>
;
/*}}}*/

condition                                                               /*{{{*/
  > [cond_attr_t ca]

:                                  << $ca=cNONE; >>
    LPAREN                         << out("( (");
                                      // TODO: too trivial, push this down to expr
                                      if ( LT(3)->getType() == RPAREN ){
                                        if ( LT(2)->getType() == TTRUE ){
                                          $ca = cadd($ca,cALWAYS);
                                        }
                                        else if ( LT(2)->getType() == TFALSE ){
                                          $ca = cadd($ca,cNEVER);
                                        }
                                      }
                                      
                                   >>
    expression
    RPAREN                         << out(").toBoolean() == xTrue )"); >>
;
/*}}}*/
    
labeledStatement                                                        /*{{{*/
  [in_attr_t ia] > [out_attr_t oa]
:                                  <<  $oa = oNONE; 
                                       ANTLRTokenType tt; 
                                       // printf("In labeledStatement\n");
                                   >>
   a:ID 
   COLON                           << tt = LT(2)->getType();
                                      // for now we ignore all non-loop contruct labels
                                      if ( (tt==FOR) || (tt==WHILE) || (tt==DO) ){
                                        lout("%s:", $a->getText());
                                        if(!addLabel($a->getText()))
                                          warn("Duplicate label"); // what to do to patch? take the closest def of label?
                                      }
                                      else{
                                        warn("Label %s ignored", $a->getText());
                                      }
                                   >>
   statement[ia] > [$oa]           << //printf("At end of labeledStatement %s\n", $a->getText());
                                      removeLabel();
                                   >>                            
;
/*}}}*/
   
varOrExprStatement                                                      /*{{{*/
  [in_attr_t ia]  > [out_attr_t oa]

:                                   << $oa = oNONE; ln(); >>    
   
   variablesOrExpression  SC                                                       
                                    << closeParen(); out(";"); >>
  
;

exception catch NoViableAlt:
          catch MismatchedToken :   << // printf("In varOrExprStatement\n LT(-1) = %d, LT(0) = %d, LT(1) = %d\n", 
                                   //                         LT(-1)->getType(),LT(0)->getType(),LT(1)->getType() ); 
      
                               
                                       if (LT(1)->getType() == RBRACE){
                                         *$_retsignal = $_signal;
                                       }
                                       else{
                                         //consume();
                                         closeParen(); out(";"); 
                                       }
                                    >>


/*}}}*/

variablesOrExpression                                                   /*{{{*/
:      
   ( VAR
     variable
      ( COMMA                      << closeParen(); out(";"); ln();>>
        variable
      )*
    )
|   expression

;
exception catch NoViableAlt:
          catch MismatchedToken :   << // printf("RAMBO");
                                    // maybe have different handlers for expression and for VAR etc
      
                                    //resetPos at beg of stmt
                                    setPos(Pos); 
                    warn("Syntax error, stmt ignored");
                                   // printf("In variablesOrExpression LT(1)->getType() = %d", LT(1)->getType() ); 
      
                                /* try to manage such things as
                      var a, b, *u, g;
                       so that still get g
                      */ 
                                        consumeUntil(VAR_END_set);
                    consume();
                                    >>

/*}}}*/

variable                                                                /*{{{*/
:                                  << char* id;>>

    a:ID                           << id = $a->getText(); 
    
                                      if ( getLevel() == 0 ){     // global var
                                        if(!isVarAtLevel(id)){
                                          addVar(id,T_PROP);
                                          set_sbuf(bmain);
                                          lout("static XsObject %s = _undefined;", id);
                                          set_sbuf(binit);
                                        }
                                      }
                                      else {                 // in function
                                        if(!isVarAtLevel(id)){
                                          addVar(id,T_PROP);        
                                          out("XsObject %s", id);
                                          if ( LT(2)->getType() != ASSIGN ){
                                            out(" = _undefined");
                                          }
                                        }
                                        else if(LT(2)->getType() == ASSIGN){
                                          out("%s", id);
                                        }
                                        
                                      }

                                   >>
    { ASSIGN                       << 
                                      if ( getLevel() == 0 ) {  // global var
                                        lout("%s = Def.defProperty(\"%s\",", id, id);
                                        openParen++;
                                      }
                                      else {               // in function  
                                        out(" = ");
                                      }
                                   >>
      assignmentExpression 
    }
;
/*}}}*/

expression                                                              /*{{{*/
:  << rawLiteral = false; >>   
   assignmentExpression                   
   ( COMMA                         << closeParen(); out(","); >>
     assignmentExpression
   )*
  
;
/*}}}*/

assignmentExpression                                                    /*{{{*/
:                                    
   condExpression                     
   //(leftSideExpression[false] ASSIGNMENT_OP)?
  // leftSideExpression[false]
   
   { a:ASSIGNMENT_OP               << 
                                      if($a->getType() != ASSIGN && !memberLeftOp){
                                        out(".%s(", binOp($a->getType())); 
                                      }
                      else if(memberLeftOp){
                        memberLeftOp =false;
                      }
                   >>
     assignmentExpression          << if($a->getType() != ASSIGN && !memberLeftOp){
                                        out(")");
                                      }
                   >> 
   }
   
   
//| condExpression   
   
   
;
/*}}}*/

condExpression                                                          /*{{{*/
:  
   // ok this slows down the whole thing, but no choice for source-to-source because of weird javac and jikes behavior
   (logOrExpression COND)?         << //need to this because of lack of "function" buffer  
                                      if(getLevel() == 0){
                        if(!dummy_declared){
                                          set_sbuf(bmain);
                      lout("static XsObject __dummy__;");
                      last_sbuf();
                      dummy_declared = true;
                        }   
                                      }
                      else{
                        if(!dummy_declared){
                                          dummy_to_declare = true;
                        }   
                      }
                      
                      out("(__dummy__ = ");                         
                                   >>
   logOrExpression 
   { COND                          << out(".toBoolean() == xTrue ? "); >>
     assignmentExpression
     COLON                         << out(" : "); >>
     condExpression                << out(")"); >>// this should be assignmentExpression, but first fix ambiguity
   }
|
   logOrExpression 
;
/*}}}*/

logOrExpression                                                         /*{{{*/
:// need expression evaluator
   logAndExpression                
   ( a:LOG_OR                      << out(".%s(", binOp($a->getType())); >>
     logAndExpression              << out(")"); >>
   )*
;
/*}}}*/

logAndExpression                                                        /*{{{*/
:// need expression evaluator
   bitOrExpression                
   ( a:LOG_AND                     << out(".%s(", binOp($a->getType())); >>
     bitOrExpression               << out(")"); >>
   )*
;
/*}}}*/

bitOrExpression                                                         /*{{{*/
:
   bitXorExpression
   ( a:OR                          << out(".%s(", binOp($a->getType())); rawLiteral = true;>>
     bitXorExpression              << out(")"); rawLiteral = false;>>
   )*
;
/*}}}*/

bitXorExpression                                                        /*{{{*/
:
   bitAndExpression
   ( a:XOR                         << out(".%s(", binOp($a->getType())); rawLiteral = true;>>
     bitAndExpression              << out(")"); rawLiteral = false;>>
   )*
;
/*}}}*/
   
bitAndExpression                                                        /*{{{*/
:
   equalExpression
   ( a:AND                         << out(".%s(", binOp($a->getType())); rawLiteral = true;>>
     equalExpression               << out(")"); rawLiteral = false;>>
   )*
;
/*}}}*/

equalExpression                                                         /*{{{*/
: 
   relationalExpression
   ( a:EQUALITY_OP                 << out(".%s(", binOp($a->getType())); >>
     relationalExpression          << out(")"); >>
   )*
; 
/*}}}*/

relationalExpression                                                    /*{{{*/
:
   shiftExpression
   ( a:RELATIONAL_OP               << out(".%s(", binOp($a->getType())); >>
     shiftExpression               << out(")"); >>
   )*
;
/*}}}*/

shiftExpression                                                         /*{{{*/
:
   addExpression
   ( a:SHIFT_OP                    << out(".%s(", binOp($a->getType())); rawLiteral = true;>>
     addExpression                 << out(")"); rawLiteral = false;>>
   )*
;
/*}}}*/

addExpression                                                           /*{{{*/
:
   multExpression
   ( a:ADDITIVE_OP                 << out(".%s(", binOp($a->getType())); rawLiteral = true; >>
     multExpression                << out(")"); rawLiteral = false; >>
   )*
;
/*}}}*/

multExpression                                                          /*{{{*/
:
   unaryExpression
   ( a:MULTIPLICATIVE_OP           << out(".%s(", binOp($a->getType())); rawLiteral = true;>>
     unaryExpression               << out(")"); rawLiteral = false;>>
   )*
;
/*}}}*/

unaryExpression                                                         /*{{{*/
: 
    (MINUS                         << rawLiteral = false; >>
     memberExpression[false]       << out(".neg()"); >>
    ) 
    
|   ( memberExpression[false]     
      { a:INCREMENT_OP             << 
      
                                      int tt = LT(2)->getType();
                                      if (/* (tt == SC) ||maybe broke something in for*/ (tt == RPAREN) || (tt == COMMA) ){
                                        out(".%s()", prefixOp($a->getType()));
                                      }
                                      else {
                                        out(".%s()", postfixOp($a->getType()));
                                      }
                                   >>
      }
    )
    
  |  (TYPEOF LPAREN)?
     TYPEOF                        << rawLiteral = false;  >>
     LPAREN
     unaryExpression
     RPAREN                        << out(".xsGetType()"); >> 
         
  | ( b:UNARY_OP                   << rawLiteral = false;  >>  
      unaryExpression              << out(".%s()", prefixOp($b->getType())); >>
    )
  | ( DELETE                       << out("delete"); >> // need expression evaluator
      unaryExpression )            
  
;
/*}}}*/
   
/*
leftSideExpression [bool isMember]
:  //added errors to music objects
  << sbuf_t *  btemp; // handle for freeing 
     char* id;
     int vtype;
     int lvl;
  >>

   b:ID                              << id = $b->getText();  lvl = varlevelt(id,&vtype);
                                      
                                       if (!isMember ) {
                                          if ( lvl < 0 ){ // not yet def
                                            if ( level==0 ) {
                                              // define it as top variable
                                              set_sbuf(bmain);
                                              lout("static XsObject %s = _undefined;", id);
                                              if(inEventScript)
                                                set_sbuf(bdom);
                                              else 
                                                set_sbuf(binit);
                                              if(LT(2)->getType() == ASSIGN){
                                                out("%s = _def.defProperty(\"%s\", ", id, id);
                                              }
                                              else{
                                                out("%s = _def.defProperty(\"%s\", %s", id, id, id);
                                              }
                                              openParen++; 
                                              addVar(id,T_PROP);
                                            }
                                            else {
                                              // register as _def property
                                              if(LT(2)->getType() == ASSIGN){
                                                out("_def.xsPut(\"%s\", ",id);
                                              }
                                              else{
                                                out("_def.xsPut(\"%s\", _def.xsGet(\"%s\")",id, id);
                                              }
                                              openParen++;
                                            }
                                          }
                                          else { // a var
                                            if ( lvl == 0 ){
                                              // we need to store the prop, too
                                              if(LT(2)->getType() == ASSIGN){
                                                out("%s = _def.xsPut(\"%s\", ", id, id);
                                              }
                                              else{
                                                out("%s = _def.xsPut(\"%s\", %s", id, id, id);
                                              }
                                              openParen++;
                                            }
                                            else {
                                              if(LT(2)->getType() == ASSIGN){ 
                                                out("%s = ",id);
                                              }
                                              else{
                                                out("%s = %s", id, id);
                                              }
                                            }
                                          }
                                        }
                                        else { // a prop
                                          if(LT(2)->getType() == ASSIGN){
                                            out("xsPut(\"%s\",", id);
                                          }
                                          else{
                                            // ????
                                          }    
                                          openParen++;
                                        }
                                         

   
     
       
                               >>  
   ( LBRACK                        << set_sbuf(push_btmp()); >>
     assignmentExpression          //should be expression, when dealt with commas                               
     RBRACK                        << 
                                      last_sbuf();
                      
                      if ( LT(2)->getType() == ASSIGN ){
                      //if (set_el(LT(2)->getType(), ASSIGNMENT_OP_set)){

                                        out(".xsPut("); merge(btemp = pop_btmp()); out(",");
                                        openParen++;
                                      }
                                      else{
                      
                        out(".xsGet("); merge(btemp = pop_btmp()); out(")");
                      }
                   
                      free(btemp);
                       >>

  )*          
  (DOT                             << out("."); >>
  
  leftSideExpression[true]
  )*


;
*/


memberExpression                                                        /*{{{*/
  [bool isMember]
:
  << sbuf_t *  btemp; /* handle for freeing */ >>

   primaryExpression[isMember]                                            
   
   ( LBRACK                        << set_sbuf(push_btmp()); >>
     assignmentExpression          //should be expression, when dealt with commas                               
     RBRACK                        << 
                                      last_sbuf();
                      
                      if ( LT(2)->getType() == ASSIGN ){
                        out(".xsPut("); merge(btemp = pop_btmp()); out(",");
                                        openParen++;
                                      }
                      else if (set_el(LT(2)->getType(), ASSIGNMENT_OP_set)){
                                        out(".%s(", memberBinOp(LT(2)->getType())); merge(btemp = pop_btmp()); out(", ");
                                        memberLeftOp = true;
                      }
                      else if ( LT(2)->getType() == LPAREN ){ 
                        out(".call("); merge(btemp = pop_btmp()); 
                                  }
                      else{
                        out(".xsGet("); merge(btemp = pop_btmp()); out(")");
                      }
                   
                      free(btemp);
                       >>

   )*                                 
   { (DOT ID)?                     //ambiguity: a.(8) versus a.b
     DOT                           << out("."); >>
     memberExpression[true]           
   }
   { callArguments[false]
   }
;
/*}}}*/

callArguments                                                           /*{{{*/
  [bool isCtor]
:

   LPAREN                          << if ( isCtor ){
                                        out(".callCtor(_def");
                                      }
                                      
                                      if ( LT(2)->getType() != RPAREN ){
                                        if(!statCall){
                                          out(",");
                                        }
                                        out(" XsArray.pushs(");
                                      }
                                      else if ( statCall ){
                                        out(" XsFunc.noArgs");
                                      }
                                      
                                      
                                      statCall = false;
                                   >>
 
   { assignmentExpression          << out(")"); >>
     ( COMMA                       << out(".push("); >>
       assignmentExpression        << out(")"); >>
     )* 
   }
   RPAREN                          << out( ")"); >>
;
/*}}}*/

primaryExpression                                                       /*{{{*/
  [bool isMember]
:                                  << char* id; int lvl; int vtype; 
                                      //only IDs allowed on the left hand-side 
                                      //but (a) = 4; is OK!!!
                                      /*if(LT(1)->getType() != ID 
                                             && ( isAssignOp(LT(2)->getType()) ||
                                                   LT(2)->getType() == LPAREN  )){
                                          warn("parse error, left-value required");
                                        }*/
                    
                //  fprintf(stderr, "In primaryExpression\n");
                                   >>
                                   
   ( LPAREN                        << out("("); >>
     expression
     RPAREN                        << out(")"); >>
   )                               
    
| objectLiteral

| arrayLiteral

|  ( {  PLUS
     }
     a:NUM_CONST                   << if ( rawLiteral ) {
                                        rawLiteral = false;
                                        out("%s", $a->getText());
                                      }
                                      else {
                                        out("new XsNumber(%s)", $a->getText());
                                      }
                                   >>
   )
   
|  b:ID                            << 

                                      id = $b->getText(); lvl = varLevelt(id,&vtype);
                                      if (set_el(LT(2)->getType(), ASSIGNMENT_OP_set)){
                                        // property/var assignment
                                        if (!isMember ) {
                                          if ( lvl < 0 ){ // not yet def
                                            if ( getLevel()==0 ) {
                                              // define it as top variable
                                              set_sbuf(bmain);
                                              lout("static XsObject %s = _undefined;", id);
                                              if(inEventScript)
                                                set_sbuf(bdom);
                                              else 
                                                set_sbuf(binit);
                                              if(LT(2)->getType() == ASSIGN){
                                                out("%s = _def.defProperty(\"%s\", ", id, id);
                                              }
                                              else{
                                                out("%s = _def.defProperty(\"%s\", %s", id, id, id);
                                              }
                                              openParen++; 
                                              addVar(id,T_PROP);
                                            }
                                            else {
                                              // register as _def property
                                              if(LT(2)->getType() == ASSIGN){
                                                out("_def.xsPut(\"%s\", ",id);
                                              }
                                              else{
                                                out("_def.xsPut(\"%s\", _def.xsGet(\"%s\")",id, id);
                                              }
                                              openParen++;
                                            }
                                          }
                                          else { // a var
                                            if ( lvl == 0 ){
                                              // we need to store the prop, too
                                              if(LT(2)->getType() == ASSIGN){
                                                out("%s = _def.xsPut(\"%s\", ", id, id);
                                              }
                                              else{
                                                out("%s = _def.xsPut(\"%s\", %s", id, id, id);
                                              }
                                              openParen++;
                                            }
                                            else {
                                              if(LT(2)->getType() == ASSIGN){ 
                                                out("%s = ",id);
                                              }
                                              else{
                                                out("%s = %s", id, id);
                                              }
                                            }
                                          }
                                        }
                                        else { // a prop
                                          if(LT(2)->getType() == ASSIGN){
                                            out("xsPut(\"%s\",", id);
                        openParen++;
                                          }
                                          else{
                        out("%s(\"%s\", ", memberBinOp(LT(2)->getType()), id);
                        memberLeftOp = true;
                                          }  
                    }
                                      }
                                      else if ( LT(2)->getType() == LPAREN ){ 
                                                        // a call
                                        if ( !isMember ){
                     //all calls now go through call so can setup call stack, function properties, etc
                    
                                        /*if ( vtype == T_FUNC && withlevel == 0){
                                            statCall = true;
                                            out("_%s(_def,",id); 
                                          }
                                          else if(!isVarAtlevel(id)){ // not yet known, need second pass for exactitude
                                            out("_def.call(\"%s\"",id);
                                          }
                                          else{
                                            out("_def.call(%s",id);
                                          }
                      */
                      
                      if(!isVarAtLevel(id)){ // not yet known, need second pass for exactitude
                                            out("_def.call(\"%s\"",id);
                                          }
                                          else{
                                            out("_def.call(%s",id);
                                          }
                      
                    }
                                        else {
                                          out("call(\"%s\"",id);
                                        }
                                      }
                                      else {
                        
                                        // just a ref, no assignment
                                        if (!isMember){  
                                                                
                                          if(!strcmp("undefined", id) && !isVar(id))
                                            id = "_undefined"; 
                                          
                                          if(isVar(id) && withLevel == 0){
                      
                        // what is this ???
                        //also Object, Boolean, String will never match here because of STATIC_PROP_METHOD
                        
                        
                                            if(!strcmp(id, "Object"))
                                              id = "_Object";
                                            else if(!strcmp(id, "Array"))
                                              id = "_Array";
                                            else if(!strcmp(id, "Date"))
                                              id = "_Date";
                                            else if(!strcmp(id, "Boolean"))
                                              id = "_Boolean";
                                            else if(!strcmp(id, "String"))
                                              id = "_String";

                                            out("%s", id);
                                          }  
                                          else{
                        
                                            out("_def.xsGet(\"%s\")", id);
                                          }
                                        }
                                        else {
                                          out("xsGet(\"%s\")", id);
                                        }
                                      }
                      
                      
                      
                                   >>
    {callArguments[false]
    }                  
                   

|   c:STRING_CONST                 << //fprintf(stderr, "rawLiteral = %d\n", rawLiteral);

                                      char * temp = NULL;

                                      if($c->getText()[0] == '\''){
                                        temp =  (char*) calloc(strlen($c->getText()) + 1, sizeof(char));
                                        strcpy(temp, $c->getText());
                                        temp[strlen($c->getText()) -1] = '\"';
                                        temp[0] = '\"';
                                        $c->setText(temp);
                                      }
                                
                                      if ( rawLiteral ){
                                        rawLiteral = false;
                                        out("%s", $c->getText());
                                      }
                                      else {
                                        out("new XsString(%s)", $c->getText());
                                      }
                      
                                      if (temp != NULL){
                                        free(temp); 
                                      }
                                   >>
|   d:TFALSE                       << out("xFalse"); >>
|   e:TTRUE                        << out("xTrue"); >>
|   f:TNULL                        << out("_null"); >>
|   g:TTHIS                        << out("xThis"); >>

|   h:STATIC_PROP_METHOD           << char number_buf[sizeof("Number.isFinite") + 1] = "Number.";
                                      char object_buf[sizeof("Object.__Object") + 1] = "";
                      
                                      if(!strcmp($h->getText(), "isNaN"))
                                        id = strcat(number_buf,$h->getText()); 
                                      else if(!strcmp($h->getText(), "isFinite"))
                                        id = strcat(number_buf,$h->getText()); 
                                      else if(!strcmp($h->getText(), "parseFloat"))
                                        id = strcat(number_buf,$h->getText()); 
                                      else if(!strcmp($h->getText(), "parseInt"))
                                        id = strcat(number_buf,$h->getText());                 
                                      else if(!strcmp($h->getText(), "Object"))
                                        id = strcat(object_buf,"Object._Object"); 
                      else if(!strcmp($h->getText(), "String"))
                                        id = strcat(object_buf,"Object.String");
                      else if(!strcmp($h->getText(), "Number"))
                                        id = strcat(object_buf,"Object.Number");
                      else if(!strcmp($h->getText(), "Boolean"))
                                        id = strcat(object_buf,"Object._Boolean");
                      else if(!strcmp($h->getText(), "Boolean"))
                                        id = strcat(object_buf,"Object._Boolean");
                                  else if(!strcmp($h->getText(), "print"))
                                        id = strcat(object_buf,"Object.print");
                                      else if(!strcmp($h->getText(), "escape"))
                                        id = strcat(object_buf,"Object.escape");
                                      else if(!strcmp($h->getText(), "unescape"))
                                        id = strcat(object_buf,"Object.unescape");
                      else  
                                        id = $h->getText();
                    
                        
                                      if(LT(2)->getType() == LBRACK){
                                        warn("%s is not an object", id);
                                        //what to do next??? -> exceptions
                                      }
                                      else {                                        
                        //must put in the _ 
                                        char* before = strtok(id, ".");
                                        char* after = strtok(NULL, ".");
                                        
                                        char* underscore = (char *) calloc(strlen(after) + 2, sizeof(char));
                                        underscore[0] = '_';
                                        
                                        after = strcat(underscore, after);
                                        
                                        if(LT(2)->getType() == LPAREN){  //must watch out ex: f(Math.sin)
                                          out("Xs%s.%s(_def,", before, after); 
                                        }
                                        else{
                                          out("Xs%s.%s", before, after); 
                                        }
                                   
                                        free(underscore);
                                        statCall = true;
                          }         
                                   >> 
|   i:STATIC_PROP_FIELD            << id = $i->getText();
                                      if(!strcmp(id, "NaN")){
                                        id = "Number.NaN";
                                      }
                                      else if(!strcmp(id, "Infinity")){
                                        id = "Number.POSITIVE_INFINITY";
                                      }
 
                                      out("Xs%s", id);
                                      
                                      if ( LT(2)->getType() == LPAREN ){
                                        warn("%s is not a function", id);
                                        //what to do next??? -> exceptions
                                      }
                                      else if(LT(2)->getType() == LBRACK){
                                        warn("%s is not an object", id);
                                        //what to do next??? -> exceptions
                                      }
                                   >>
                   
                   
| NEW
  constructor

;
/*}}}*/

constructor                                                             /*{{{*/
:                                  << bool hasArgs = false; char* id; >>
    (  TTHIS                       << out("xThis"); >>
     | a:ID                        << id = $a->getText();
                                      if ( isVar(id) ) {
                                        if(!strcmp(id, "Object"))
                                          id = "_Object";
                                        else if(!strcmp(id, "Array"))
                                          id = "_Array";
                                        else if(!strcmp(id, "Date"))
                                          id = "_Date";
                                        else if(!strcmp(id, "Boolean"))
                                          id = "_Boolean";
                                        else if(!strcmp(id, "String"))
                                          id = "_String";
                                        
                                        out("%s", id);
                                      }
                                      else {
                                        out( "_def.xsGet(\"%s\")", id);
                          }
                                   >>
     |  b:STATIC_PROP_METHOD       << id = $b->getText();
                                      if ( isVar(id) ) {
                                        if(!strcmp(id, "Object"))
                                          id = "_Object";
                                        else if(!strcmp(id, "Array"))
                                          id = "_Array";
                                        else if(!strcmp(id, "Date"))
                                          id = "_Date";
                                        else if(!strcmp(id, "Boolean"))
                                          id = "_Boolean";
                                        else if(!strcmp(id, "String"))
                                          id = "_String";
                                        
                                        out("%s", id);
                                      }
                                      else {
                                        out( "_def.xsGet(\"%s\")", id);
                          }
                                   >>
    )
    ( DOT c:ID                     << out(".xsGet(\"%s\")", $c->getText()); >>
    
    
    
    )*
    { callArguments [true]         << hasArgs = true; >>
    }                              
                                   << if ( !hasArgs ) out( ".callCtor(_def)"); >>
;
/*}}}*/

objectLiteral                                                           /*{{{*/
: 
                                   << bool hasIntializer = false; >>
    LBRACE                         << out(" ObjectInitializer.callCtor(_def, "); >>
    { a:ID COLON                   << out("XsArray.pushs(new XsString(\"%s\")).push(", $a->getText()); 
                                      bool hasIntializer = true; 
                                   >>
      assignmentExpression        
      ( COMMA                      << out(")"); >>
        b:ID COLON                 << out(".push(new XsString(\"%s\")).push(", $b->getText()); >>
        assignmentExpression 
      )*                           << out(")"); >>
    }                                 
    RBRACE                         << if(!hasIntializer){
                                        out("XsFunc.noArgs");
                                      }
                                      out(")"); 
                                   >>
;
/*}}}*/

arrayLiteral:                                                           /*{{{*/

                                   << bool hasLiteral = false; >> 
   ( LBRACK                        << out(" ArrayInitializer.callCtor(_def, ");
                                      if ( LT(2)->getType() != RBRACK ){
                                        out("XsArray.pushs(");
                                      }
                                   >>
     { assignmentExpression        << hasLiteral = true; >>   
       ( COMMA                     << out(").push("); >>
         assignmentExpression
       )*                          << out(")"); >>
     }
     RBRACK                        << if(!hasLiteral){
                                        out("XsFunc.noArgs");
                                      }
                                      out(")"); 
                                   >>
   )
;

/*}}}*/

   
//Interpreter ********************************************************
                                                                        

//put in function composition ?? ex: f(g())

call > [jobject ret]                                                    /*{{{*/
:

(functionCall > [$ret]

|
 constructorCall > [$ret]
)

{SC}
;

exception default :               << #ifdef KAFFELIB
                                     longjmp(jmpbuffer, 1); //return with failure
                                     #endif
                                  >> 
/*}}}*/

tweakedExpression > [jobject object]                                    /*{{{*/
:  <<
     #ifdef KAFFELIB 

     jclass xsobjectclass = Genv->GetObjectClass(Gcontext); CheckException();
     
     jmethodID mid = Genv->GetMethodID(xsobjectclass, "xsGet",  
                                       "(Ljava/lang/Object;)Lkaffe/xs/XsObject;"); CheckException();
                                       
     #endif                                    
   >>

// member expression
((ID|TTHIS) DOT)?

   ( id:ID                <<#ifdef KAFFELIB
                            $object =  Genv->CallObjectMethod(Gcontext, mid, Genv->NewStringUTF($id->getText())); CheckException();
                            #endif
                          >> 
    | 
     TTHIS                <<#ifdef KAFFELIB 
                            $object = GxThis;
                            #endif
                          >> 
   )
   (
     DOT
     dottedid:ID          <<#ifdef KAFFELIB
                            $object = Genv->CallObjectMethod($object, mid, Genv->NewStringUTF($dottedid->getText())); CheckException();
                            #endif
                          >> 
   )*

|  constructorCall > [$object]

|  (ID (DOT ID)* LPAREN)?

   functionCall > [$object]

|  primaryExpressionE > [$object] 

;
/*}}}*/

constructorCall > [jobject object]                                      /*{{{*/
:  
  << 
     jobject temp;
     #ifdef KAFFELIB
     jclass clazz = Genv->FindClass("kaffe/xs/XsArray"); CheckException();
     
     jmethodID push = Genv->GetMethodID(clazz, "push", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsArray;"); CheckException();

     jmethodID pushs =  Genv->GetStaticMethodID(clazz, "pushs", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsArray;"); CheckException();

     jobject xsarray =  NULL;
     #endif
     
  >>

  NEW
  id:ID                           
  LPAREN                     
  {
   addExpressionE > [temp]       <<#ifdef KAFFELIB
                                   xsarray = Genv->CallStaticObjectMethod(clazz, pushs,  temp); CheckException();
                                   #endif
                                 >> 
   ( COMMA                 
     addExpressionE > [temp]     <<#ifdef KAFFELIB  
                                   xsarray = Genv->CallObjectMethod(xsarray, push, temp); CheckException();
                                   #endif
                                 >> 
     )*
   }
   RPAREN 

   << 
   
      #ifdef KAFFELIB
      jclass xthis = Genv->GetObjectClass(GxThis); CheckException();
    
      jmethodID xsGet = Genv->GetMethodID(xthis, "xsGet", "(Ljava/lang/String;)Lkaffe/xs/XsObject;"); CheckException();
      
      jobject handle = Genv->CallObjectMethod(GxThis, xsGet, Genv->NewStringUTF($id->getText())); CheckException();

      
      jclass xsfunc =  Genv->FindClass("kaffe/xs/XsFunc"); CheckException();

      jmethodID callCtor;
      

      if(xsarray != NULL){ // some arguments

        callCtor = Genv->GetMethodID(xsfunc, "callCtor", "(Lkaffe/xs/XsObject;Lkaffe/xs/XsArray;)Lkaffe/xs/XsObject;"); CheckException();
 
        $object = Genv->CallObjectMethod(handle, callCtor, Gcontext, xsarray); 
       
       
        /* putting this one in causes a weird bug, investigate 
    
           on input
       var a = eval('new Function("return 6;")');
      
       same when have arguments, below
      
           CheckException(); */
       
      }
      else{
      
        callCtor = Genv->GetMethodID(xsfunc, "callCtor", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsObject;"); CheckException();

        $object = Genv->CallObjectMethod(handle, callCtor, Gcontext); //CheckException();
       
      }
      #endif
   >>
;
/*}}}*/

addExpressionE > [jobject object]                                       /*{{{*/
:
   <<
     jobject returned;
    
     jclass returnedclass;
     jmethodID mid;
   >>


   multExpressionE  > [returned]    <<   $object = returned; >>  
   (
    op:ADDITIVE_OP
    multExpressionE > [returned]    <<   #ifdef KAFFELIB
                                         returnedclass = Genv->GetObjectClass($object); CheckException();

                                         ANTLRTokenType op_type = op->getType();
                      
                                         if(op_type == PLUS){
                                                mid = Genv->GetMethodID(returnedclass, "plus", 
                                                                        "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsObject;"); CheckException();
                                         }
                                         else if(op_type == MINUS){
                                                mid = Genv->GetMethodID(returnedclass, "minus", 
                                                                        "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsObject;"); CheckException();
                                         }

                                         $object = Genv->CallObjectMethod($object, mid, returned); CheckException();
                                      
                                         #endif
                                    >>   
   )*

;
/*}}}*/

multExpressionE > [jobject object]                                      /*{{{*/
:
   <<
     jobject returned;
    
     jclass returnedclass;
     jmethodID mid;

   >>


   tweakedExpression  > [returned]    << $object = returned; >>  
   (
    op:MULTIPLICATIVE_OP
    tweakedExpression > [returned]    << 
                                         #ifdef KAFFELIB
                                         returnedclass = Genv->GetObjectClass($object); CheckException();

                                         ANTLRTokenType op_type = op->getType();
                      
                                         if(op_type == MUL){
                                                mid = Genv->GetMethodID(returnedclass, "mul", 
                                                                        "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsObject;"); CheckException();
                                         }
                                         else if(op_type == DIV){
                                                mid = Genv->GetMethodID(returnedclass, "div", 
                                                                        "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsObject;"); CheckException();
                                         }
                                         else if(op_type == MOD){
                                                mid = Genv->GetMethodID(returnedclass, "mod", 
                                                                        "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsObject;"); CheckException();
                                         }
                                        
                                         $object = Genv->CallObjectMethod($object, mid, returned); CheckException();
                                         #endif 
                                      >>   
   )*

;
/*}}}*/

functionCall > [jobject ret]                                            /*{{{*/
:   <<
       jobject temp;
       #ifdef KAFFELIB
       
       jclass clazz = Genv->FindClass("kaffe/xs/XsArray"); CheckException();

       jmethodID push = Genv->GetMethodID(clazz, "push", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsArray;"); CheckException();

       jmethodID pushs =  Genv->GetStaticMethodID(clazz, "pushs", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsArray;"); CheckException();
       #endif
       jobject xsarray =  NULL;
       
       int nids = 0;
       int max_ids = 2;
       char** ids = (char**) calloc(max_ids, sizeof(char*)); 
       
       bool hasTop = false;
       
    >>
 
    a:ID                           << if(!strcmp("top", $a->getText()))
                                        hasTop = true;
                                      ids[nids++] = StrDup($a->getText()); 
                                   >>
    (DOT 
     b:ID                          << if(nids == max_ids){
                                        // grow array 
                                        int newMaxIds = max_ids + 10;
                                        char** new_ids = (char**) calloc(newMaxIds, sizeof(char*));
                                        memcpy( new_ids, ids, max_ids * sizeof(char*));
                                        free(ids);
                                        ids = new_ids;
                                        max_ids = newMaxIds;
                                      }
                                        
                                      ids[nids++] = StrDup($b->getText());
                                   >>
    )*
    LPAREN                         
    {
     addExpressionE > [temp]       <<#ifdef KAFFELIB
                                     xsarray = Genv->CallStaticObjectMethod(clazz, pushs,  temp); CheckException();
                                     #endif
                                    >> 
     ( COMMA                 
       addExpressionE > [temp]     <<#ifdef KAFFELIB                                         
                                     xsarray = Genv->CallObjectMethod(xsarray, push, temp); CheckException();
                                     #endif
                                   >> 
     )*
    }
    RPAREN 
                            
    <<
      #ifdef KAFFELIB
      int i; 
   
      jclass xsobjectclass = Genv->FindClass("kaffe/xs/XsObject"); CheckException();
      
      jmethodID xsget = Genv->GetMethodID(xsobjectclass, "xsGet", "(Ljava/lang/String;)Lkaffe/xs/XsObject;"); CheckException();

      jobject caller;
      
      if(!hasTop){
        caller = Gcontext;
      
        for(i = 0; i < nids - 1; i++){
          caller = Genv->CallObjectMethod(caller, xsget, Genv->NewStringUTF(ids[i])); CheckException();
        }
      }
      else{
        jfieldID topID = Genv->GetStaticFieldID(xsobjectclass, "top", "Lkaffe/xs/XsObject;"); CheckException();

        caller = Genv->GetStaticObjectField(xsobjectclass, topID); CheckException();
        
        for(i = 1; i < nids - 1; i++){
          caller = Genv->CallObjectMethod(caller, xsget, Genv->NewStringUTF(ids[i])); CheckException();
        }
      }                 
                        
      if(xsarray != NULL) // some arguments
        $ret = Genv->CallObjectMethod(caller, Gmethod_args, Genv->NewStringUTF(ids[nids - 1]), xsarray); 
      else
        $ret = Genv->CallObjectMethod(caller, Gmethod_no_args, Genv->NewStringUTF(ids[nids - 1])); 
      
      CheckException(); 

      //release strings                                    
      for(i = 0; i < nids; i++)
        free(ids[i]);
    
      #endif    
    >>
;    
/*}}}*/

primaryExpressionE > [jobject object]                                   /*{{{*/
:                                 
    <<
      jclass clazz;
      jmethodID  CtorId;
      jfieldID   fieldID;
    >>
  

    ( LPAREN                    
      addExpressionE > [$object]
      RPAREN 
    )                  

|   arrayLiteralE  > [$object]

|   num:NUM_CONST
                 
                  <<  #ifdef KAFFELIB
                      clazz = Genv->FindClass("kaffe/xs/XsNumber"); CheckException();

                      if($num->getType() == INT_CONST){ 
                        CtorId = Genv->GetMethodID(clazz, "<init>", "(I)V"); CheckException();
                        $object = Genv->NewObject(clazz, CtorId, (jint) atoi($num->getText())); CheckException();
                      }
                      else{
                        //this does not work
                        jdouble number =  (jdouble) atof($num->getText()); 
                      
                        CtorId = Genv->GetMethodID( clazz, "<init>", "(D)V"); CheckException();

                        printf("Value is %f\n",  number);

                        $object = Genv->NewObject( clazz, CtorId, number); CheckException();
 
                        printf("Got object\n");
                      }
                      #endif
                  >> 

//indexed array
|  ( ID LBRACK)? 
 arrayName:ID
           LBRACK
           index:INT_CONST
           RBRACK            << jobject array; 
                                #ifdef KAFFELIB   
                                jclass xsobjectclass = Genv->GetObjectClass(Gcontext); CheckException();
                                jmethodID mid = Genv->GetMethodID(xsobjectclass, "xsGet",  
                                                         "(Ljava/lang/Object;)Lkaffe/xs/XsObject;"); CheckException();
                                //might miss an exception here
                                array =  Genv->CallObjectMethod(Gcontext, mid, Genv->NewStringUTF($arrayName->getText())); CheckException();
                                 
                                jclass xsarrayC = Genv->FindClass("kaffe/xs/XsArray"); CheckException();
                                 
                               
                                mid =  Genv->GetMethodID(xsarrayC, "xsGet", "(I)Lkaffe/xs/XsObject;"); CheckException(); 

                                $object =  Genv->CallObjectMethod(array, mid, (jint) atoi($index->getText())); CheckException();
                                #endif
                             >>


|       id:ID     <<  #ifdef KAFFELIB 
                      jclass xsobjectclass = Genv->GetObjectClass(Gcontext); CheckException();
                      jmethodID mid = Genv->GetMethodID(xsobjectclass, "xsGet",  
                                                         "(Ljava/lang/Object;)Lkaffe/xs/XsObject;"); CheckException();
                           
                      
                      $object =  Genv->CallObjectMethod(Gcontext, mid, Genv->NewStringUTF($id->getText())); CheckException();
                      #endif
                  >>                   

|       string:STRING_CONST <<
                              //strip '' or ""
                              char * temp =  (char*) calloc(strlen($string->getText()) + 1, sizeof(char));
                              strcpy(temp, $string->getText());
                              temp[strlen($string->getText()) - 1] = '\0';
                              $string->setText(temp + 1);
                              
                              #ifdef KAFFELIB
                              clazz = Genv->FindClass( "kaffe/xs/XsString"); CheckException(); 

                              CtorId = Genv->GetMethodID( clazz, "<init>", "(Ljava/lang/String;)V"); CheckException();

                              $object = Genv->NewObject( clazz, CtorId, Genv->NewStringUTF($string->getText())); CheckException(); 
                              #endif
                            >>
|       TFALSE              << #ifdef KAFFELIB
                               clazz = Genv->FindClass( "kaffe/xs/XsObject"); CheckException();
                               fieldID =  Genv->GetStaticFieldID(clazz, "xFalse", "Lkaffe/xs/XsObject;"); CheckException(); 
                               $object =  Genv->GetStaticObjectField(clazz, fieldID); CheckException();
                               #endif
                            >>      
|       TTRUE               << #ifdef KAFFELIB
                               clazz = Genv->FindClass( "kaffe/xs/XsObject"); CheckException();
                               fieldID =  Genv->GetStaticFieldID(clazz, "xTrue", "Lkaffe/xs/XsObject;"); CheckException(); 
                               $object =  Genv->GetStaticObjectField(clazz, fieldID); CheckException();
                               #endif
                            >>            
//|     TNULL                     
|       TTHIS               << #ifdef KAFFELIB
                               $object = GxThis; 
                               #endif
                            >>
;
/*}}}*/

arrayLiteralE > [jobject object]                                        /*{{{*/
:
    <<#ifdef KAFFELIB
      jobject xsarray =  NULL;
       
      jclass xsarrayC = Genv->FindClass("kaffe/xs/XsArray"); CheckException();

      jfieldID ArrayID =  Genv->GetStaticFieldID(xsarrayC, "Array", "Lkaffe/xs/XsFunc;"); CheckException();
                                           
      jobject Array =  Genv->GetStaticObjectField(xsarrayC, ArrayID); CheckException();
       
      jclass xsfuncC = Genv->FindClass("kaffe/xs/XsFunc"); CheckException();
     
      jmethodID callCtor_args = Genv->GetMethodID(xsfuncC, "callCtor", "(Lkaffe/xs/XsArray;)Lkaffe/xs/XsObject;"); CheckException();
      jmethodID callCtor_no_args = Genv->GetMethodID(xsfuncC, "callCtor", "()Lkaffe/xs/XsObject;"); CheckException();
  
      jmethodID push = Genv->GetMethodID(xsarrayC, "push", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsArray;"); CheckException();
      jmethodID pushs =  Genv->GetStaticMethodID(xsarrayC, "pushs", "(Lkaffe/xs/XsObject;)Lkaffe/xs/XsArray;"); CheckException();
      #endif
      jobject temp; 
    >> 

  
     
    LBRACK                   
    { addExpressionE  > [temp]     <<#ifdef KAFFELIB 
                                     xsarray = Genv->CallStaticObjectMethod(xsarrayC, pushs,  temp); CheckException();
                                     #endif
                                    >>
      ( COMMA                  
        addExpressionE > [temp]    <<#ifdef KAFFELIB
                                     xsarray = Genv->CallObjectMethod(xsarray, push, temp); CheckException();
                                     #endif
                                   >> 
      )*                      
    }
    RBRACK                    
    
    <<#ifdef KAFFELIB 
      if(xsarray != NULL){
         $object = Genv->CallObjectMethod(Array, callCtor_args, xsarray); CheckException();    
      }
      else{
         $object = Genv->CallObjectMethod(Array, callCtor_no_args); CheckException();
      }
      #endif
    >>
;
/*}}}*/


/******************************************************************************
 * Parser part C - Java inserts (just simple literal copy for now)
 */

javaSection                                                             /*{{{*/
:                                  << char* sym; int dontBreak=0; int sep=0; >>
  JAVA_START                       << ln(); lout("//--------- begin Java"); ln(); >>
  (
      a:KEYWORD                    << out("%s", $a->getText()); sep=1;>>
    | b:ID                         << if ( sep ){  out(" "); }
                                      out("%s", $b->getText()); sep=1;
                                   >>
    | c:STRING_CONST               << //allow XoeScript strings within Java
                                      char * temp = NULL;
                                      if($c->getText()[0] == '\''){
                                        temp =  (char*) calloc(strlen($c->getText()) + 1, sizeof(char));
                                        strcpy(temp, $c->getText());
                                        temp[strlen($c->getText()) -1] = '\"';
                                        temp[0] = '\"';
                                        $c->setText(temp);
                                      }

                                      out("%s", $c->getText());
                   
                                      if(temp != NULL){ free(temp); }
                                   >>
    | d:NUM_CONST                  << out("%s", $d->getText()); >>
    | e:OPERATOR                   << out("%s", $e->getText()); sep = 0; >>
    | f:SYMBOL                     << sym = $f->getText(); sep = 0;
                                      if ( *sym == '{' ) {
                                        incInd();
                                        lout("{"); ln();
                                      }
                                      else if ( *sym == '}' ){
                                        decInd();
                                        out("}"); ln();
                                      }
                                      else if ( *sym == ';' ){
                                        out(";");
                                        if ( !dontBreak ) ln();
                                      }
                                      else {
                                        if ( *sym == '(' ){
                                          dontBreak=1;
                                        }
                                        else if ( *sym == ')' ){
                                          dontBreak=0;
                                        }
                                        out("%s", sym);
                                      }
                                   >>
  )*
  JAVA_END                         << out("//--------- end Java"); ln(); >>
;
/*}}}*/


}


