/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: BuildDepends.java,v 1.6 2001/10/25 21:48:21 samc Exp $
 */
package org.xoe.core.packages;

import org.xoe.core.Engine;
import org.xoe.core.content.Stash;
import org.xoe.core.services.ServiceLocator;
import org.xoe.core.util.Logger;

import java.io.FileInputStream;
import java.util.Vector;

/**
 * Builds a list of XOE packages from a package-in.xml file.
 * All this is used for is building the CLASSPATH for package
 * building.
 */
class BuildDepends
{
  public static void main (String[] args) 
    throws Exception
  {
    if (args.length != 1)
      {
        System.err.println ("usage: " + BuildDepends.class.getName () + " <package-in.xml>");
	System.exit (1);
      }

    long timer = System.currentTimeMillis ();
    Engine.bootstrap ();
    FileInputStream pkgFile = new FileInputStream (args[0]);
    PackageDescription desc = new PackageDescription (pkgFile);

    Vector requiredPackages = new Vector ();
    IPackageQuery[] deps = desc.getDependencies ();
    int len = deps.length;

    for (int i = 0; i < len; i++)
      {
        Logger.log (BuildDepends.class, "Looking for: " + deps[i].toXPath () + "\n");
        SearchResults res = PackageManager.search (deps[i]);
        if (res.getInstalledPackages ().length != 0)
          {
            Logger.log (BuildDepends.class, "Already installed: " + deps[i].toXPath ());
            continue;
          }
        InstallOption[] options = res.getInstallablePackages ();
        if (options.length > 0)
          {
            //requiredPackages.addElement ();
            PackageDescription[] newInstalls = options[0].getNewInstalls ();
            for (int j = 0; j < newInstalls.length; j++)
              {
                Logger.log (BuildDepends.class, "Adding: " + newInstalls[j]);
                String uri = newInstalls[j].getPackageURI ();
                boolean found = false;
                for (int k = 0; k < requiredPackages.size (); k++)
                  {
                    PackageDescription depDesc = (PackageDescription) requiredPackages.elementAt (k);
                    if (depDesc.getPackageURI ().equals (uri))
                      {
                        found = true;
                      }
                  }
                if (!found)
                  {
                    requiredPackages.addElement (newInstalls[j]);
                  }
              }
          }
        else
          {
            throw new Exception ("No installable packages for: " + deps[i].toXPath ());
          }
      }

    timer = System.currentTimeMillis () - timer;
    Logger.log (BuildDepends.class, "BuildDepends took " + timer + " ms.");

    String path = Stash.getPath () + "/packages/";
    boolean first = true;

    for (int i = 0; i < requiredPackages.size (); i++)
      {
        PackageDescription depDesc = (PackageDescription) requiredPackages.elementAt (i);
	System.err.println ("depends on: " + depDesc.getName ());
        if (first == true) { first = false; }
        else               { System.out.print (":"); }
        System.out.print (path + depDesc.getXoeFilename ());
      }

    System.exit (0);
  }
}
