/*
 * Copyright (c) 2001 Transvirtual Technologies, Inc.  
 * All rights reserved. See the file "COPYING" for details.
 *
 * $tvt: GetPackageInfoFromXML.java,v 1.3 2001/10/31 00:06:46 samc Exp $
 */
package org.xoe.core.packages;

import org.xoe.core.dom.DOMlet;
import org.xoe.core.xpath.XPath;

import org.w3c.dom.Document;
import org.w3c.dom.Text;

import java.io.IOException;
import java.io.FileInputStream;

class GetPackageInfoFromXML
{
  public static void printUsage()
  {
    System.err.println ("usage: GetPackageInfoFromXML [-name | -version | -filename] <package-xml>");
    System.exit (1);
  }

  public static void main (String[] args) 
    throws IOException
  {
    if (args.length != 2)
      {
        printUsage();
      }
    FileInputStream pkgFile = new FileInputStream (args[1]);
    Document doc = DOMlet.create (pkgFile, "-- package desc --");

    String name = doc.getDocumentElement().getAttribute("name");
    String version = doc.getDocumentElement().getAttribute("version");

    if (args[0].equals ("-name"))
      {
        System.out.println (name);
      }
    else if (args[0].equals ("-maintainer"))
      {
        String maintainer = "";
        Text textNode = (Text) XPath.findFirstNode (doc, "package/info/maintainer/name/text()");
        if (textNode != null) 
          {
            maintainer = textNode.getNodeValue ();
          }
        else
          {
            System.err.println ("GetPackageInfoFromXML: package has no maintainer/name");
          }
        textNode = (Text) XPath.findFirstNode (doc, "package/info/maintainer/email/text()");
        if (textNode != null) 
          {
            maintainer = maintainer + " <" + textNode.getNodeValue () + ">";
          }
        else
          {
            System.err.println ("GetPackageInfoFromXML: package has no maintainer/email");
          }
        System.out.println (maintainer);
      }
    else if (args[0].equals ("-version"))
      {
        System.out.println (version);
      }
    else if (args[0].equals ("-filename"))
      {
        String arch = null;
        if (name.endsWith ("-native"))
          {
            arch = System.getProperty ("xoe.hostarch");
          }
        System.out.println (PackageDescription.getXoeFilename (name, arch, version));
      }
    else
      {
        printUsage();
      }
    System.exit (0);
  }
}
