package java.awt;

import kaffe.awt.Defaults;

/**
 * class Label - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Label
  extends Component
{
	final private static long serialVersionUID = 3094126758329070636L;
	final public static int LEFT = 0;
	final public static int CENTER = 1;
	final public static int RIGHT = 2;
	int align;
	String label;
	FontMetrics fm;
	static int named;

public Label() {
	this( null );
}

public Label( String label) {
	this( label, LEFT);
}

public Label( String label, int align) {
	setText(label != null ? label : "");
	setAlignment(align);
}

public int getAlignment() {
	return align;
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "label" + (named++);
}

public String getText() {
	return label;
}

public boolean isFocusTraversable() {
	return false;
}

public void paint( Graphics g) {
	int sw = fm.stringWidth( label);
	int y0 = height - (height-fm.getHeight())/2 - fm.getDescent();
	int x0;

	switch( align) {
	case CENTER:
		x0 = (width - sw)/2;
		break;
	case RIGHT:
		x0 = width - insets.right - sw;
		break;
	case LEFT:
	default:
		x0 = insets.left+1;
		break;
	}

	if ( Defaults.LabelTxtCarved ) {
		g.setColor( Color.white);
		g.drawString( label, x0+1, y0+1 );
	}
	
	g.setColor( fgClr);
	g.drawString( label, x0, y0);
}

protected String paramString() {
	return ( super.paramString() + ",Label: " + label);
}

/**
 * @deprecated
 */
public Dimension preferredSize() {
	Dimension d = new Dimension( insets.left + insets.right, insets.top + insets.bottom);
	String ms = (label != null ) ? label : "xxx";
	FontMetrics mfm = (fm != null) ? fm : Defaults.WndFontMetrics;
	
	d.width += mfm.stringWidth( ms) * 4 / 3; 
	d.height += mfm.getHeight() * 3 / 2;

	return d;
}

void propagateFont( Font fnt) {
	super.propagateFont( fnt);
//	if ( fnt != null )
		fm = getFontMetrics( fnt);
	
	if ( (flags & IS_SHOWING) > 0 )
		repaint();
}

public void setAlignment( int align) {
	if ( align < LEFT || align > RIGHT ) {
		throw new IllegalArgumentException();
	}
	this.align = align;
	if ( isShowing() ) {
		repaint();
	}
}

public void setText( String label) {
	if ( label == null ) {
		label = "";
	}
	if ( (this.label == null ) || ! (this.label.equals( label)) ){
		this.label = label;
		//maybe high frequent
		if ( (flags & IS_SHOWING) == IS_SHOWING ) {
			Graphics g = getGraphics();
			if ( g != null ) {
				paintBackground( g, this, 0, 0, width, height);
				paint( g);
				g.dispose();
			}
		}
	}
}
}
