package java.awt;

import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.TextEvent;
import java.awt.event.TextListener;
import kaffe.awt.AWT;
import kaffe.awt.Defaults;

/**
 * class TextComponent - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
abstract public class TextComponent
  extends Container
{
	final private static long serialVersionUID = -2214773872412987419L;
	protected transient TextListener textListener;
	boolean isEditable = true;

public synchronized void addTextListener( TextListener l) {
	eventMask |= AWTEvent.TEXT_EVENT_MASK;
}

void copyToClipboard () {
	Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();
	String sel = getSelectedText();
	if ( (sel != null) && (sel.length() > 0) ) {
		StringSelection ss = new StringSelection( sel);
		cb.setContents( ss, ss);
	}
}

public synchronized int getCaretPosition(){
	return -1;
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

public synchronized String getSelectedText(){
	return null;
}

public synchronized int getSelectionEnd(){
	return -1;
}

public synchronized int getSelectionStart(){
	return -1;
}

public synchronized String getText(){
	return null;
}

public boolean isEditable() {
	return isEditable;
}

public boolean isFocusTraversable() {
	return ( ((flags & IS_SHOWING) == IS_SHOWING) && 
	         ((eventMask & AWTEvent.DISABLED_MASK) == 0) &&
	         isEditable );
}

boolean isPrintableTyped( KeyEvent e) {
	int mods = e.getModifiers();
	int chr  = e.getKeyChar();

	if ( (mods != 0) && (mods != e.SHIFT_MASK) )
		return false;

	switch( chr ) {
		case 8:		//BACKSPACE
		case 9:		//TAB
		case 10:	//ENTER
		case 27:	//ESC
		case 127:	//DEL
			return false;
	}
	
	return true;		
}

protected String paramString() {
	return super.paramString();
}

void pasteFromClipboard () {
	Clipboard cb = Toolkit.getDefaultToolkit().getSystemClipboard();
	Transferable tf = cb.getContents( this);
	
	if ( tf != null ){
		try {
			String s = (String) tf.getTransferData( DataFlavor.stringFlavor);
			replaceSelectionWith( s);
		}
		catch ( Exception x ) {}
	}
}

void process ( TextEvent e ) {
	if ( (textListener != null) || (eventMask & AWTEvent.TEXT_EVENT_MASK) != 0 )
		processEvent( e);
}

protected void processTextEvent( TextEvent e) {
	if ( textListener != null ) {
		textListener.textValueChanged( e);
	}
}

public synchronized void removeTextListener( TextListener l) {
	textListener = AWTEventMulticaster.remove( textListener, l);
}

void replaceSelectionWith ( String s ) {
}

public synchronized void select( int start, int end){
}

public synchronized void selectAll(){
}

public synchronized void setCaretPosition( int pos){
}

void setEditMenu( Component owner) {

	PopupMenu p = new PopupMenu();
	p.add( new MenuItem("Cut")).setShortcut( new MenuShortcut( KeyEvent.VK_U, false) );
	p.add( new MenuItem("Copy")).setShortcut( new MenuShortcut( KeyEvent.VK_O, false) );
	p.add( new MenuItem("Paste")).setShortcut( new MenuShortcut( KeyEvent.VK_A, false) );
	p.addSeparator();
	p.add( new MenuItem("Select All")).setShortcut( new MenuShortcut( KeyEvent.VK_S, false) );
	
	owner.add( p);

	p.addActionListener( new ActionListener() {
		public void actionPerformed( ActionEvent e) {
			String cmd = e.getActionCommand();
	
			if ( "Copy".equals( cmd) ) {
				copyToClipboard();
				//reset selection
				setCaretPosition( getCaretPosition() );
			}
			else if ( "Paste".equals( cmd) )
				pasteFromClipboard();
			else if ( "Select All".equals( cmd) )
				selectAll();
			else if ( "Cut".equals( cmd) ) {
				copyToClipboard();
				replaceSelectionWith( null);
			}
		}
	});
}

public synchronized void setEditable( boolean edit) {
	// We need to maintain a separate state for this (i.e. we can't just map it to
	// setEnabled) because - another spec glitch - non-editable TextComponents can have
	// pointer-selections, and they also report as isEnabled=true in the JDK.
	
	if ( isEditable != edit ) {
		isEditable = edit;

		// Hrmm, this is incomplete, in case we have explicitly set colors, we would
		// loose them when the state is restored	
		setForeground( isEditable ? Defaults.TextAreaTxtClr : Defaults.TextAreaRoTxtClr);
		setBackground( isEditable ? Defaults.TextAreaBgClr : Defaults.TextAreaRoBgClr);
	}
}

public synchronized void setSelectionEnd( int end){
}

public synchronized void setSelectionStart( int start){
}

public synchronized void setText( String text){
}

void updateAWT () {
	TextComponent tc = AWT.lastTextComponentTgt;
	
	if ( (tc != null) && (tc != this) ) {
		//will remove selection
		tc.setCaretPosition( tc.getCaretPosition() );
	}
	
	AWT.lastTextComponentTgt = this;
}
}
