package java.awt;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import kaffe.awt.Defaults;
import kaffe.awt.ImageDict;
import kaffe.awt.ImageSpec;
import kaffe.awt.WidgetTheme;

/**
 * class Button - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Button
  extends Component
  implements ShortcutConsumer
{
	final private static long serialVersionUID = -8774683716313001058L;
	String label;
	ActionListener aListener;
	String aCmd;
	FontMetrics fm;
	ImageSpec imgs;
	MenuShortcut shortcut;
	int state;
	static int named;

public Button () {
	this( "");
}

public Button ( String label) {
	cursor = Cursor.getPredefinedCursor( Cursor.HAND_CURSOR);
	
	flags |= IS_MOUSE_AWARE;
	state |= WidgetTheme.NEEDS_BORDER_UPDATE;
	
  WidgetTheme.btnSetDefaults(this);
	setLabel( label);
}

public void addActionListener ( ActionListener a) {
	aListener = AWTEventMulticaster.add( aListener, a);
}

public void addNotify () {
	super.addNotify();

	if ( shortcut != null )
		ShortcutHandler.addShortcut( shortcut, getToplevel(), this);
		
	MenuShortcut s1 = new MenuShortcut( this, KeyEvent.VK_ENTER, 0);
	MenuShortcut s2 = new MenuShortcut( this, KeyEvent.VK_SPACE, 0);
	ShortcutHandler.addShortcut( s1, this, this);
	ShortcutHandler.addShortcut( s2, this, this);
}

void animate() {
	Graphics g = getGraphics();
	
	state |= WidgetTheme.IS_ACTIVE;
	paint( g);
	Toolkit.tlkSync();
	try { Thread.sleep( 100); }
	catch ( Exception _x) {}
	state &= ~WidgetTheme.IS_ACTIVE;
	paint( g);
	Toolkit.tlkSync();
	
	g.dispose();
}

void checkMouseAware () {
	// we are always mouse aware if not disabled
	if ( (eventMask & AWTEvent.DISABLED_MASK) == 0 ){
		flags |= IS_MOUSE_AWARE;
	}
	else {
		flags &= ~IS_MOUSE_AWARE;
	}
}

void drawImage ( Graphics g) {
	Image img = imgs.getImage();
	int iw = img.getWidth( this);
	int ih = img.getHeight( this);
	int di = ((state & WidgetTheme.IS_ACTIVE) > 0) ? 1 : 0;
	int x = (width - iw) / 2 + di;
	int y = (height - ih) / 2 + di;
	
	g.drawImage( img, x, y, this);
}

public String getActionCommand () {
	return ( (aCmd != null) ? aCmd : label);
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "button" + (named++);
}

public String getLabel() {
	return label;
}

public void handleShortcut( MenuShortcut ms) {
	int mods = (ms != null) ? ms.mods : 0;

	if ( (state & WidgetTheme.IS_ACTIVE) == 0 )
		animate();
		
	if ( (aListener != null) ||
	     ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) ||
	     ((flags & IS_OLD_EVENT) != 0) ) {
		Toolkit.eventQueue.postEvent( ActionEvt.getEvent( this, ActionEvent.ACTION_PERFORMED,
		                                                  getActionCommand(), mods));
	}
}

public void paint ( Graphics g) {
	if ( (imgs == null) || ! imgs.isPlain() ) {
		WidgetTheme.btnPaint( this, g, deco, state);
		if ( imgs != null )
			drawImage( g);
	}
	else {
		g.setColor( parent.getBackground() );
		g.fillRect( 0, 0, width, height);
		if ( (state & WidgetTheme.IS_POINTED) > 0 ) {
			g.setColor( Defaults.BtnPointClr);
			g.draw3DRect( 0, 0, width-1, height-1, ((state & WidgetTheme.IS_ACTIVE) == 0));
		}
		drawImage( g);
	}
}

void paintBackground( Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.btnPaintBackground( this, g, x+deco.x, y+deco.y, width-deco.width, height-deco.width, flags);
}

protected String paramString() {
	return (super.paramString() + ",Label: " + label);
}

public Dimension preferredSize () {
	Dimension d = new Dimension();
	
	if ( imgs != null ) {
		d.width = imgs.getImage().getWidth( this) + 2*BORDER_WIDTH;
		d.height = imgs.getImage().getHeight( this) + 2*BORDER_WIDTH;
	}

	//let Theme decide about text and direction preferrences		
	return WidgetTheme.btnPreferredSize( this, d);
}

void process ( ActionEvent e ) {
	if ( (aListener != null) || ((eventMask & AWTEvent.ACTION_EVENT_MASK) != 0) ){
		processEvent( e);
	}

	if ( (flags & IS_OLD_EVENT) != 0 ) postEvent( Event.getEvent( e));
}

void process ( FocusEvent e ) {
	switch ( e.getID() ){
		case e.FOCUS_GAINED:
			state |= WidgetTheme.HAS_FOCUS;
			break;
		case e.FOCUS_LOST:
			state &= ~WidgetTheme.HAS_FOCUS;
			break;			
	}
	repaint();
	super.process( e);
}

protected void processActionEvent( ActionEvent e) {
	if ( aListener != null ) {
		aListener.actionPerformed( e);
	}
}

void processMouse ( MouseEvent e ) {
	switch ( e.id ) {
		case e.MOUSE_ENTERED:
			state |= WidgetTheme.IS_POINTED;
			repaint();
			break;
		case e.MOUSE_EXITED:
			state &= ~WidgetTheme.IS_POINTED;
			repaint();
			break;
		case e.MOUSE_PRESSED:
			state |= WidgetTheme.IS_ACTIVE;
			if ( AWTEvent.keyTgt != Button.this )
				requestFocus();
			else
				repaint();
			break;
		case e.MOUSE_RELEASED:
			if ( contains( e.getX(), e.getY()))
				handleShortcut( null);
			state &= ~WidgetTheme.IS_ACTIVE;
			repaint();
			break;
	}

	super.processMouse( e);
}

void propagateFont( Font fnt) {
	super.propagateFont( fnt);
	fm = getFontMetrics( font);
	repaint();
}

public void removeActionListener ( ActionListener a) {
	aListener = AWTEventMulticaster.remove( aListener, a);
}

public void removeNotify () {
	if ( shortcut != null )
		ShortcutHandler.removeFromOwner( getToplevel(), shortcut);
	ShortcutHandler.removeShortcuts( this);
	super.removeNotify();
}

public void setActionCommand ( String aCmd) {
	this.aCmd = aCmd;
}

public void setEnabled( boolean isEnabled) {
	super.setEnabled( isEnabled);
	repaint();
}

public void setLabel ( String label) {

	if ( label != null ){
		int ti = label.indexOf( '~');

		if ( ti > -1){
			this.label = label.substring( 0, ti) + label.substring( ti+1);
			shortcut = new MenuShortcut( this, Character.toUpperCase( label.charAt(ti+1)), KeyEvent.ALT_MASK);
		}
		else if ( label.startsWith( " ") && label.endsWith( " ") ) {
			imgs = ImageDict.getDefaultDict().getSpec( label.substring( 1, label.length() - 1), null, this );
			this.label = (imgs != null) ? null : label;
		}
		else {
			this.label = label;
		}
	}
	else {
		this.label = null;
	}

	if ( isShowing() )
		repaint();
}
}
