package java.awt;

import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import kaffe.awt.Defaults;
import kaffe.awt.WidgetTheme;

/**
 * class Checkbox - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Checkbox
  extends Component
  implements ItemSelectable, MouseListener, KeyListener
{
	final private static long serialVersionUID = 7270714317450821763L;
	CheckboxGroup group;
	int state;
	String label;
	ItemListener iListener;
	static int named;

public Checkbox () {
	this( "", false, null);
}

public Checkbox ( String label) {
	this( label, false, null);
}

public Checkbox ( String label, CheckboxGroup group, boolean state) {
	this( label, state, group);
}

public Checkbox ( String label, boolean state) {
	this( label, state, null);
}

public Checkbox ( String label, boolean state, CheckboxGroup group) {
	this.label = (label == null) ? "" : label;
	setCheckboxGroup( group);
	setState( state);
	
	setForeground( Color.black);
	setFont( Defaults.TextFont);
	
	deco = WidgetTheme.cbGetDeco( this);
	
	addMouseListener( this);
	addKeyListener( this);
}

public synchronized void addItemListener ( ItemListener il) {
	iListener = AWTEventMulticaster.add( iListener, il);
}

void checkMouseAware () {
	// we are always mouse aware if not disabled
	if ( (eventMask & AWTEvent.DISABLED_MASK) == 0 ){
		flags |= IS_MOUSE_AWARE;
	}
	else {
		flags &= ~IS_MOUSE_AWARE;
	}
}

public CheckboxGroup getCheckboxGroup () {
	return group;
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "checkbox" + (named++);
}

public String getLabel () {
	return label;
}

public Object[] getSelectedObjects () {
	Object[] oa;
	if ( (state &  WidgetTheme.IS_ACTIVE) > 0 ) {
		oa = new Object[1];
		oa[0] = this;
	}
	else
		oa = new Object[0];
		
	return oa;
}

public boolean getState () {
	return ((state &  WidgetTheme.IS_ACTIVE) > 0);
}

public void keyPressed ( KeyEvent e) {
}

public void keyReleased ( KeyEvent e) {
}

public void keyTyped ( KeyEvent e) {
	char c = e.getKeyChar();

	switch ( c) {
		case ' ':
		case 0xA:	//ENTER
			if ( ((state &  WidgetTheme.IS_ACTIVE) > 0) && (group != null) )
				return;
			setState ( (state &  WidgetTheme.IS_ACTIVE) == 0 );
			break;
	}

}

public void mouseClicked ( MouseEvent e) {
}

public void mouseEntered ( MouseEvent e) {
	state |=  WidgetTheme.IS_POINTED;
	repaint();
}

public void mouseExited ( MouseEvent e) {
	state &= ~ WidgetTheme.IS_POINTED;
	repaint();
}

public void mousePressed ( MouseEvent e) {
	requestFocus();
	if ( ((state &  WidgetTheme.IS_ACTIVE) > 0) && (group != null) )
		return;
	setState( (state &  WidgetTheme.IS_ACTIVE) == 0);
}

public void mouseReleased ( MouseEvent e) {
}

void notifyItem () {
	if ( (iListener != null) ||
	     ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) ||
	     ((flags & IS_OLD_EVENT) != 0) ) {
		int id = ((state &  WidgetTheme.IS_ACTIVE) > 0) ? ItemEvent.SELECTED : ItemEvent.DESELECTED;
		Toolkit.eventQueue.postEvent( ItemEvt.getEvent( this, ItemEvent.ITEM_STATE_CHANGED, label, id));
	}
}

public void paint ( Graphics g) {
	WidgetTheme.cbPaint( this, g, deco, state);
}

void paintBackground ( Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.cbPaintBackground(this, g, x, y, width, height, flags);
}

protected String paramString() {
	return super.paramString();
}

/** 
 * @deprecated
 */
public Dimension preferredSize () {
	Dimension d = new Dimension();
	//let Theme decide about text and direction preferrences		
	return WidgetTheme.cbPreferredSize( this, d);
}

void process ( FocusEvent e ) {
	switch ( e.getID() ){
		case e.FOCUS_GAINED:
			state |= WidgetTheme.HAS_FOCUS;
			break;
		case e.FOCUS_LOST:
			state &= ~WidgetTheme.HAS_FOCUS;
			break;			
	}
	repaint();
	super.process( e);
}

void process ( ItemEvent e ) {
	if ( (iListener != null) || ((eventMask & AWTEvent.ITEM_EVENT_MASK) != 0) ){
		processEvent( e);
	}
	
	if ( (flags & IS_OLD_EVENT) > 0 ) {
		postEvent( Event.getEvent( e));
	}
}

protected void processItemEvent( ItemEvent e) {
	if (iListener != null) {
		iListener.itemStateChanged( e);
	}
}

public synchronized void removeItemListener ( ItemListener il) {
	iListener = AWTEventMulticaster.remove( iListener, il);
}

public void setCheckboxGroup ( CheckboxGroup group) {
	if ( this.group == group )
		return;
		
	if ( this.group != null)
		this.group.boxes.removeElement( this);
		
	this.group = group;
	
	if ( this.group != null)
		this.group.boxes.addElement( this);
}

public synchronized void setLabel ( String label) {
	this.label = (label == null) ? "" : label;
	if ( isShowing() )
		repaint();
}

public void setState ( boolean state) {
	if ( state && (group != null) ) {
		group.setSelectedCheckbox( this);
	}
	else {
		setStateImpl( state);
	}
}

void setStateImpl ( boolean state) {

	if ( ((this.state &  WidgetTheme.IS_ACTIVE) > 0) == state ){
		return;
	}

	if ( state ){	
		this.state |=  WidgetTheme.IS_ACTIVE;
	}
	else {
		this.state &= ~WidgetTheme.IS_ACTIVE;
	}

	if ( (flags & IS_SHOWING) == IS_SHOWING ){
		repaint();
	}

	if ( (group == null) || state ){
		// we don't inform in case this is a DESELECT and we have a group (JDK behavior)
		notifyItem();
	}
}
}
