package java.awt;

import kaffe.awt.Defaults;
import kaffe.awt.RenderDirection;
import kaffe.awt.WidgetTheme;

/**
 * class Label - 
 *
 * Copyright (c) 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */
public class Label
  extends Component
{
	final private static long serialVersionUID = 3094126758329070636L;
	final public static int LEFT = 0;
	final public static int CENTER = 1;
	final public static int RIGHT = 2;
	int align;
	String label;
	FontMetrics fm;
	int state;
	static int named;

public Label() {
	this( null );
}

public Label( String label) {
	this( label, LEFT);
}

public Label( String label, int align) {
	WidgetTheme.lblSetDefaults(this);
	setText(label != null ? label : "");
	setAlignment(align);
	
	deco = WidgetTheme.lblGetDeco( this);
}

public int getAlignment() {
	return align;
}

ClassProperties getClassProperties () {
	return ClassAnalyzer.analyzeAll( getClass(), true);
}

String getGenericName () {
	// Apparently, the JDK increments the running number each time
	// *after* there was a genericName request. And yes, the stem is of course
	// not the qualified class name
	
	return "label" + (named++);
}

public String getText() {
	return label;
}

public boolean isFocusTraversable() {
	return false;
}

public void paint( Graphics g) {
	WidgetTheme.lblPaint( this, g, deco, state);
}

void paintBackground (Graphics g, int x, int y, int width, int height, int flags) {
	WidgetTheme.lblPaintBackground( this, g, x+deco.x, y+deco.y, width-deco.width, height-deco.width, flags);
}

protected String paramString() {
	return ( super.paramString() + ",Label: " + label);
}

/**
 * @deprecated
 */
public Dimension preferredSize() {
	Dimension d = new Dimension();
	//let Theme decide about text and direction preferrences		
	return WidgetTheme.lblPreferredSize( this, d);
}

void propagateFont( Font fnt) {
	super.propagateFont( fnt);
	fm = getFontMetrics( font);
	repaint();
}

public void setAlignment( int align) {
	if ( align < LEFT || align > RIGHT ) {
		throw new IllegalArgumentException();
	}
	this.align = align;
	if ( isShowing() ) {
		repaint();
	}
}

public void setText( String label) {
	if ( label == null ) {
		label = "";
	}
	if ( (this.label == null ) || ! (this.label.equals( label)) ){
		this.label = label;
		if ( label.startsWith(" ") && label.endsWith( " ") ) {
			state |= WidgetTheme.NEEDS_BORDER_UPDATE;
		}
		else {
			state &= ~WidgetTheme.NEEDS_BORDER_UPDATE;
		}
		//high frequent sync paint
		if ( (flags & IS_SHOWING) == IS_SHOWING ) {
			Graphics g = getGraphics();
			if ( g != null ) {
				update( g);
				g.dispose();
			}
		}
		//repaint();
	}
}
}
